"""This module defines miscellaneous utility functions."""

from numpy import unique, linalg, diag, sqrt, dot

__all__ = ['Everything', 'rangeString', 'alnum', 'importLA', 'dictElement',
           'intorfloat', 'startswith', 'showFigure', 'countBytes', 'sqrtm',
           'saxsWater', 'count']


class Everything(object):

    """A place for everything."""

    def __contains__(self, what):

        return True


def rangeString(lint, sep=' ', rng=' to ', exc=False, pos=True):
    """Returns a structured string for a given list of integers.

    :arg lint: integer list or array
    :arg sep: range or number separator
    :arg rng: range symbol
    :arg exc: set **True** if range symbol is exclusive
    :arg pos: only consider zero and positive integers

    .. ipython:: python

       from prody.utilities import rangeString
       lint = [1, 2, 3, 4, 10, 15, 16, 17]
       rangeString(lint)
       rangeString(lint, sep=',', rng='-')
       rangeString(lint, ',', ':', exc=True)"""

    ints = unique(lint)
    if pos and ints[0] < 0:
        ints = ints[ints > -1]

    prev = ints[0]
    lint = [[prev]]
    for i in ints[1:]:
        if i - prev > 1:
            lint.append([i])
        else:
            lint[-1].append(i)
        prev = i
    exc = int(exc)
    return sep.join([str(l[0]) if len(l) == 1 else
                     str(l[0]) + rng + str(l[-1] + exc) for l in lint])


def alnum(string, alt='_', trim=False, single=False):
    """Replace non alpha numeric characters with *alt*.  If *trim* is **True**
    remove preceding and trailing *arg* characters.  If *single* is **True**,
    contain only a single joining *alt* character. """

    result = ''
    multi = not bool(single)
    prev = None
    for char in string:
        if char.isalnum():
            result += char
            prev = char
        else:
            if multi or prev != alt:
                result += alt
            prev = alt
    trim = int(bool(trim))
    result = result[trim * (result[0] == alt):
                    len(result) - trim * (result[-1] == alt)]
    return result


def importLA():
    """Returns one of :mod:`scipy.linalg` or :mod:`numpy.linalg`."""

    try:
        import scipy.linalg as linalg
    except ImportError:
        try:
            import numpy.linalg as linalg
        except:
            raise ImportError('scipy.linalg or numpy.linalg is required for '
                              'NMA and structure alignment calculations')
    return linalg


def dictElement(element, prefix=None):
    """Returnss a dictionary built from the children of *element*, which must be
    a :class:`xml.etree.ElementTree.Element` instance.  Keys of the dictionary
    are *tag* of children without the *prefix*, or namespace.  Values depend on
    the content of the child.  If a child does not have any children, its text
    attribute is the value.  If a child has children, then the child is the
    value."""

    dict_ = {}
    length = False
    if isinstance(prefix, str):
        length = len(prefix)
    for child in element:
        tag = child.tag
        if length and tag.startswith(prefix):
            tag = tag[length:]
        if len(child) == 0:
            dict_[tag] = child.text
        else:
            dict_[tag] = child
    return dict_


def intorfloat(x):
    """Returns ``int(x)``, or ``float(x)`` upon :exc:`ValueError`."""

    try:
        return int(x)
    except ValueError:
        return float(x)


def startswith(this, that):
    """Returns **True** if *this* or *that* starts with the other."""

    if len(this) < len(that):
        return that.startswith(this)
    else:
        return this.startswith(that)


def showFigure():
    """Call :func:`~matplotlib.pyplot.show` function with ``block=False``
    argument to avoid blocking behavior in non-interactive sessions.  If
    *block* keyword argument is not recognized, try again without it."""

    from matplotlib.pyplot import show
    try:
        show(block=False)
    except TypeError:
        show()


def countBytes(arrays, base=False):
    """Returns total number of bytes consumed by elements of arrays.  If
    *base* is **True**, use number of bytes from the base array."""

    if base:
        getbase = lambda arr: arr if arr.base is None else getbase(arr.base)
        nbytes = lambda arr: getbase(arr).nbytes
    else:
        nbytes = lambda arr: arr.nbytes

    return sum(nbytes(arr) for arr in arrays)

def sqrtm(matrix):
    """Returns the square root of a matrix."""
    (U,S,VT)=linalg.svd(matrix)
    D = diag(sqrt(S))
    return dot(dot(U,D),VT)

def getMasses(elements):
    """get the mass atom. """
    
    import numpy as np
    mass_dict = {'C':12,'N':14,'S':32,'O':16,'H':1}

    if isinstance(elements, str):
        return mass_dict[elements]
    else:
        masses = np.zeros(len(elements))
        for i,element in enumerate(elements):
            if element in mass_dict:
                masses[i] = mass_dict[element]
            else:
                masses[i] = 0.
        return masses

def count(L, a=None):
    return len([b for b in L if b is a])

def saxsWater():
    return [[-48.347,-49.439,-50.456],
[-49.553,-51.034,-48.334],[-57.235,-51.179,-48.718],[-45.221,-47.629,-50.154],
[-54.290,-45.655,-49.534],[-59.020,-50.659,-54.428],[-58.206,-49.310,-51.124],
[-47.951,-42.102,-49.997],[-49.085,-46.464,-50.123],[-42.029,-55.066,-47.295],
[-47.010,-53.650,-48.132],[-53.375,-55.303,-51.497],[-45.735,-54.776,-53.495],
[-57.758,-52.607,-51.083],[-52.966,-52.740,-51.146],[-41.767,-52.401,-44.860],
[-54.272,-56.479,-47.804],[-48.196,-46.646,-53.080],[-52.200,-49.277,-47.184],
[-44.287,-55.489,-50.965],[-55.359,-54.230,-49.547],[-54.427,-46.647,-46.775],
[-59.058,-46.422,-54.822],[-44.055,-46.545,-48.060],[-44.907,-56.280,-46.572],
[-52.126,58.735,-49.726],[-44.603,-44.847,-51.015],[-44.417,-45.344,-53.855],
[-43.376,-47.634,-42.938],[-48.635,-50.694,-53.047],[-48.052,-51.932,-58.217],
[-42.487,-52.072,-47.440],[-43.833,-49.094,-53.615],[-45.951,-51.718,-44.716],
[-52.379,-49.651,-52.095],[-51.648,-45.817,-48.876],[-51.841,-50.061,-49.677],
[-46.890,-45.170,-48.616],[-52.307,-45.452,-43.760],[-46.218,-49.892,-54.405],
[-55.269,-49.536,-44.380],[-48.067,-43.591,-53.434],[-54.168,-43.443,-42.420],
[-42.676,-58.786,-55.295],[-55.578,-49.515,-53.311],[-41.182,-44.898,58.605],
[-58.179,-48.225,-45.583],[-58.431,-45.800,-50.205],[-40.092,-45.323,-42.781],
[-48.539,-44.391,-44.694],[-49.753,-41.718,-54.402],[-53.337,-40.665,-51.973],
[-44.930,-41.216,-57.172],[-45.515,-42.750,-53.839],[-42.043,-51.207,-42.451],
[-56.728,-46.074,-45.187],[-53.810,-58.003,-50.358],[-49.370,-58.393,-45.823],
[-44.843,-54.664,-42.841],[-44.555,-49.570,-57.398],[-49.028,-56.102,-58.459],
[-54.728,-51.085,-47.498],[-55.420,-52.496,-52.755],[-48.178,-52.327,-55.380],
[-48.532,-52.550,-43.310],[-46.654,-51.923,-50.478],[-55.954,-53.358,-44.768],
[-51.392,-55.034,-57.831],[-59.188,-55.164,-51.084],[-57.233,-54.778,-56.108],
[-48.718,-52.054,-45.862],[-51.502,-56.289,-49.646],[-57.011,-55.108,-52.473],
[-52.155,-54.488,-53.662],[-46.270,-56.646,-49.054],[-53.090,-52.597,-58.594],
[-42.654,-39.716,-42.123],[-56.498,-47.040,-42.625],[-50.531,-53.875,-50.920],
[-54.900,-56.465,-53.453],[-51.789,-53.171,-48.598],[-49.739,-48.733,-54.759],
[-44.489,-49.909,-47.278],[-46.159,-54.915,-57.390],[-49.473,-46.798,-45.816],
[-44.923,-52.658,-55.593],[-55.134,-49.324,-49.955],[-47.972,-46.315,-56.155],
[-57.843,-50.934,-45.410],[-53.312,-49.319,-42.602],[-47.068,-45.643,-42.618],
[-41.077,-51.397,-55.663],[-41.302,-46.852,-44.575],[-43.568,-45.032,-57.492],
[-45.742,-49.887,-51.687],[-54.053,-45.075,-52.555],[-57.262,-50.585,-57.216],
[-51.694,-47.244,-53.082],[-51.187,-43.786,-46.917],[-51.805,-43.487,-53.313],
[-47.166,-49.387,-47.061],[-46.135,-46.427,-58.399],[-45.726,-43.230,-43.873],
[-40.928,-43.522,-57.796],[-57.551,-55.685,-44.417],[-45.197,-47.469,-55.819],
[-58.473,-44.196,-46.108],[-51.457,-46.527,59.670],[-52.629,-47.444,-55.996],
[-56.089,-47.152,-55.198],[-47.603,-55.702,-46.315],[-48.158,-57.966,-54.463],
[-49.716,-44.755,-55.212],[-49.606,-53.338,-53.275],[-51.073,-53.428,-45.060],
[-49.521,-56.427,-52.635],[-52.188,-43.702,-56.040],[-53.057,-56.709,-55.348],
[-41.799,-56.062,-44.455],[-42.819,-50.844,-49.923],[-54.562,-45.409,-56.635],
[-55.672,-52.673,-57.656],[-45.221,-44.482,-46.253],[-55.682,-58.793,-48.590],
[-58.365,-44.864,-52.811],[-57.198,-42.177,-41.069],[-52.150,-42.670,-40.499],
[-50.774,-43.933,59.040] ,[-47.225,-45.993,-40.042],[-46.046,-43.584,-57.999],
[-59.576,-47.081,-43.050],[-53.735,-48.699,-59.656],[-50.485,-48.843,-43.388],
[-48.536,-49.083,-59.208],[-45.961,-48.001,-44.431],[-56.527,-50.258,-41.321],
[-53.962,-50.438,-57.039],[-53.757,-51.966,-44.922],[-50.752,-51.679,-57.365],
[-46.295,-48.092,-41.792],[-42.231,-49.171,-59.370],[-52.296,-55.771,-43.978],
[-51.716,-55.412,-46.744],[-51.579,-51.132,-43.801],[-45.626,-52.889,-40.989],
[-44.465,-50.154,-43.365],[-54.746,-51.817,-42.461],[-52.429,-58.508,-45.909],
[-49.623,-54.992,-42.415],[-47.870,-56.684,-43.863],[-46.552,-57.504,-41.674],
[-55.007,-54.784,-42.167],[-50.749,-58.308,-43.429],[-49.704,-58.148,-40.660],
[-50.665,-41.971,-57.639],[-58.223,-45.824,-59.149],[59.842,-40.551,-52.576],
[-57.902,-42.487,-50.661],[-42.640,-44.367,-48.308],[-40.685,-42.755,-46.737],
[-41.546,-46.839,-41.013],[-42.635,-46.896,-55.443],[-40.463,-49.323,-53.879],
[-43.047,-48.200,-45.940],[-42.361,-44.593,-45.313],[-41.079,-49.043,-56.573],
[-41.961,-53.051,59.320] ,[-41.135,-51.504,-52.154],[-58.946,-49.213,-48.544],
[-43.267,-48.886,-40.648],[-40.247,-52.354,-40.744],[-44.020,-57.817,-58.165],
[-42.326,-54.246,-54.091],[-44.132,-53.936,-48.868],[-44.294,-53.793,-45.950],
[-58.397,-53.403,-42.523],[-41.170,-58.260,-40.889],[-43.660,-56.638,-53.810],
[-57.866,-54.730,-48.804],[59.152,-53.061,-49.129],[-43.233,-57.844,-42.842],
[-40.612,-58.176,-58.717],[59.067,-55.811,-53.791],[-43.767,-40.317,-54.942],
[-58.854,-58.266,-44.310],[-58.894,-45.085,-41.162],[-54.713,-44.210,-45.016],
[-49.545,-41.446,-42.835],[-45.317,-40.485,-43.662],[-41.369,-57.960,-46.239],
[-55.192,-41.729,-47.232],[-52.686,-41.796,-45.667],[-47.850,-40.251,-47.536],
[-47.723,-42.937,-46.976],[-45.585,-41.092,-46.220],[-56.312,-44.236,-48.210],
[-54.788,-42.530,-53.203],[-52.060,-42.947,-50.233],[-48.745,-58.410,-48.757],
[-43.942,-58.389,-51.612],[-41.460,-57.733,-51.169],[-57.587,-57.700,-51.248],
[-58.693,-57.418,-55.826],[-51.123,58.851,-52.162],[-46.760,-58.370,-51.255],
[-45.851,-57.065,-55.956],[-57.480,-44.050,-55.785],[-55.896,-57.891,-58.190],
[-49.036,-55.582,-55.815],[-48.105,-41.299,-56.321],[-42.040,-43.565,-54.796],
[-57.309,-56.340,-41.925],[-53.536,-58.232,-43.144],[-56.073,-41.363,-58.894],
[-45.359,-59.238,-44.353],[-43.448,-42.340,-42.513],[-47.745,-32.462,-52.320],
[-49.214,-31.175,-48.818],[-55.299,-31.636,-47.972],[-43.837,-28.314,-51.395],
[-55.732,-28.317,-49.722],[-57.921,-33.225,-53.056],[59.502,-30.365,-49.519],
[-48.230,-21.813,-50.403],[-49.826,-27.525,-50.775],[-40.350,-37.357,-46.812],
[-46.411,-35.429,-48.626],[-53.487,-37.641,-50.350],[-46.360,-37.600,-51.711],
[-56.695,-35.328,-51.281],[-51.666,-33.125,-51.436],[-41.244,-33.688,-45.156],
[-53.746,-36.809,-46.897],[-47.981,-28.192,-53.674],[-51.871,-29.524,-47.661],
[-44.139,-38.036,-50.368],[-55.472,-36.143,-48.907],[-54.346,-29.859,-44.802],
[-56.809,-29.125,-53.725],[-44.262,-27.047,-48.775],[-44.046,-37.041,-47.717],
[-51.178,-40.370,-50.172],[-45.741,-25.163,-50.537],[-45.051,-26.751,-53.070],
[-44.700,-28.236,-43.919],[-46.833,-29.620,-55.509],[-48.691,-34.660,-57.533],
[-40.182,-32.009,-46.944],[-44.890,-30.831,-53.354],[-45.702,-33.360,-46.047],
[-54.112,-30.746,-52.013],[-52.209,-27.778,-49.766],[-53.144,-31.444,-49.550],
[-48.273,-26.467,-46.789],[-53.033,-27.641,-43.593],[-45.930,-33.062,-57.030],
[-55.324,-32.364,-44.451],[-47.664,-25.375,-54.038],[-54.800,-25.104,-42.896],
[-42.094,-40.088,-57.438],[-55.267,-31.607,-54.477],[-42.544,-25.553,-57.540],
[-57.110,-29.364,-47.086],[-58.424,-27.456,-48.492],[-40.411,-27.296,-45.458],
[-50.036,-27.192,-44.304],[-48.229,-22.576,-53.079],[-54.422,-22.316,-53.125],
[-45.904,-22.472,-58.251],[-46.934,-23.026,-55.837],[-43.344,-34.132,-41.624],
[-56.830,-29.264,-43.971],[-54.890,-38.987,-47.817],[-48.113,-41.099,-44.924],
[-45.052,-35.593,-44.626],[-45.214,-30.696,-58.559],[-50.471,-37.657,-57.934],
[-52.917,-34.423,-47.873],[-55.299,-34.647,-54.049],[-48.591,-31.991,-54.847],
[-48.054,-34.443,-42.725],[-47.383,-33.103,-49.628],[-55.331,-34.560,-46.023],
[-50.967,-36.756,58.614],[-58.303,-37.341,-52.064],[-57.733,-35.973,-56.298],
[-48.064,-32.187,-46.050],[-51.052,-36.733,-50.686],[-56.486,-38.876,-53.942],
[-52.851,-35.691,-55.327],[-47.511,-37.573,-47.068],[-52.931,-33.187,-57.545],
[-42.694,-22.097,-41.137],[-55.486,-29.169,-40.983],[-49.680,-34.412,-49.718],
[-53.397,-38.124,-54.014],[-50.614,-35.854,-47.793],[-49.236,-30.035,-56.849],
[-43.577,-31.526,-46.656],[-47.014,-38.421,-57.263],[-48.156,-29.426,-43.998],
[-45.501,-34.312,-54.480],[-57.136,-31.144,-50.637],[-48.117,-27.584,-57.796],
[-57.912,-33.554,-45.174],[-50.449,-32.490,-42.354],[-47.346,-26.764,-42.966],
[-41.067,-33.721,-56.285],[-41.710,-28.774,-42.971],[-43.225,-26.962,-55.262],
[-46.089,-31.049,-50.646],[-52.049,-27.725,-52.799],[-59.487,-29.073,-58.262],
[-50.119,-29.659,-52.577],[-51.203,-25.456,-45.881],[-50.730,-23.698,-54.511],
[-46.547,-28.426,-47.876],[-45.616,-28.206,-59.315],[-45.360,-25.000,-43.703],
[-40.986,-22.541,-55.724],[-58.151,-38.237,-44.392],[-45.232,-27.610,-56.809],
[-55.648,-25.241,-46.986],[-50.968,-27.662,-59.334],[-52.918,-29.175,-56.956],
[-55.513,-29.380,-55.986],[-49.215,-38.545,-45.002],[-48.600,-38.634,-54.927],
[-50.217,-26.455,-55.930],[-50.576,-34.175,-54.859],[-50.648,-35.281,-43.557],
[-49.396,-36.795,-53.275],[-52.423,-25.798,-57.410],[-54.038,-38.019,-56.850],
[-43.690,-38.159,-44.986],[-44.222,-33.286,-49.261],[-54.699,-27.027,-56.477],
[-57.399,-33.315,-57.952],[-43.816,-24.557,-48.218],[-57.169,-39.659,-48.984],
[-56.027,-24.839,-52.536],[-58.564,-23.781,-44.085],[-53.028,-26.063,-40.697],
[-50.554,-24.789,-59.052],[-45.111,-25.954,-41.113],[-47.045,-24.544,59.028],
[-58.583,-30.051,-41.559],[-55.384,-31.673,-57.910],[-51.311,-29.711,-43.665],
[-48.743,-29.305,-59.738],[-45.981,-29.883,-45.779],[-57.260,-32.500,-42.760],
[-51.982,-31.230,-55.084],[-52.847,-34.681,-44.785],[-50.575,-32.377,-58.702],
[-47.778,-31.724,-42.391],[-43.400,-30.731,58.158],[-52.491,-37.916,-42.798],
[-51.552,-37.236,-45.485],[-50.812,-33.249,-46.268],[-45.287,-35.793,-41.879],
[-45.522,-32.791,-43.432],[-54.239,-33.812,-41.735],[-50.776,-40.110,-46.998],
[-49.175,-36.577,-41.676],[-47.439,-36.749,-43.955],[-46.204,-38.300,-42.178],
[-55.593,-36.943,-42.063],[-52.333,-40.618,-43.001],[-50.706,-39.098,-41.193],
[-52.155,-21.920,-56.291],[-57.925,-27.448,-56.438],[-57.089,-22.280,-54.026],
[-58.534,-23.259,-51.348],[-40.074,-23.332,-48.306],[-41.024,-24.843,-44.792],
[-43.861,-28.266,-41.279],[-43.570,-29.801,-56.181],[-42.346,-31.063,-53.770],
[-41.508,-29.758,-47.629],[-43.902,-26.160,-45.957],[-42.369,-32.127,-58.117],
[-42.211,-35.243,58.861],[-44.406,-33.258,-51.922],[-57.846,-32.028,-47.485],
[-43.585,-31.208,-42.844],[-40.546,-33.290,-40.562],[-44.808,-35.903,-58.933],
[-43.251,-35.945,-55.603],[-41.698,-37.221,-50.444],[-43.068,-34.287,-47.067],
[-58.338,-35.520,-43.515],[-41.282,-41.292,-40.268],[-43.913,-38.692,-53.003],
[-58.300,-37.230,-49.286],[-57.704,-34.643,-48.637],[-39.747,-39.281,-42.528],
[-42.105,-37.384,-58.178],[-58.825,-37.952,-54.996],[-45.215,-21.596,-51.674],
[-57.911,-41.761,-47.702],[-57.181,-26.159,-42.334],[-54.042,-27.590,-46.878],
[-50.034,-24.376,-42.011],[-47.811,-23.135,-43.220],[-42.815,-40.694,-45.774],
[-57.522,-21.676,-47.266],[-54.198,-22.864,-47.183],[-48.970,-20.805,-48.165],
[-46.835,-24.112,-45.861],[-44.430,-22.764,-46.465],[-57.260,-24.837,-49.609],
[-54.285,-25.617,-50.319],[-49.470,-24.837,-51.300],[-49.351,-38.647,-49.116],
[-43.255,-42.527,-50.872],[-41.955,-40.268,-51.909],[-55.642,-38.112,-51.619],
[-58.603,-40.930,-55.413],[-49.830,-39.501,-52.483],[-46.286,-39.854,-49.647],
[-44.291,-38.367,-56.978],[-58.689,-24.634,-53.923],[-56.768,-38.532,-57.436],
[-51.759,-40.083,-54.886],[-48.769,-22.807,-58.319],[-43.003,-24.104,-55.089],
[-56.462,-40.155,-43.375],[-53.509,-38.488,-40.195],[-56.065,-22.576,-56.660],
[-47.457,-39.895,-40.196],[-42.843,-24.896,-42.145],[-47.469,-11.629,-51.547],
[-48.958,-12.998,-49.712],[-56.030,-13.079,-47.792],[-44.644,-10.479,-50.900],
[-54.575,-9.333,-51.189] ,[59.815,-13.411,-54.323],[-58.478,-12.777,-50.890],
[-46.982,-6.138,-49.368],[-49.399,-10.496,-49.899],[-41.037,-20.085,-48.514],
[-46.936,-16.809,-47.928],[-53.532,-18.782,-50.282],[-46.087,-18.701,-52.493],
[-57.578,-15.230,-51.576],[-52.020,-15.261,-52.336],[-41.363,-15.940,-44.657],
[-55.494,-19.500,-47.309],[-48.588,-9.417,-52.933],[-52.477,-10.346,-45.952],
[-43.475,-18.669,-50.249],[-55.144,-16.862,-49.111],[-55.057,-11.780,-45.503],
[-56.756,-10.830,-54.718],[-44.405,-7.491,-48.118],[-44.522,-18.806,-47.434],
[-51.973,-21.354,-49.095],[-45.145,-7.726,-50.946],[-46.286,-7.350,-54.059],
[-44.109,-11.957,-45.405],[-50.507,-12.790,-52.900],[-48.264,-15.757,-58.085],
[-40.759,-13.232,-46.288],[-45.872,-13.604,-53.197],[-47.166,-15.564,-44.545],
[-53.073,-12.791,-52.565],[-51.842,-8.508,-49.989],[-53.348,-13.401,-49.126],
[-47.350,-9.304,-46.746],[-51.170,-8.419,-43.957],[-47.291,-12.967,-56.826],
[-55.806,-13.006,-43.051],[-49.075,-6.511,-53.447],[-53.550,-8.493,-42.427],
[-42.448,-20.056,-55.761],[-54.909,-12.797,-54.755],[-41.591,-8.224,-57.784],
[-58.529,-11.201,-43.822],[-58.618,-8.737,-48.125],[-39.370,-7.415,-45.035],
[-49.110,-7.084,-45.471],[-50.523,-3.497,-54.604],[-56.078,-4.268,-51.955],
[-47.129,-5.287,-58.646],[-46.885,-4.076,-54.985],[-44.735,-14.860,-42.076],
[-57.364,-8.938,-44.979],[-53.238,-20.245,-46.570],[-48.800,-22.089,-45.489],
[-44.946,-17.195,-43.711],[-45.839,-11.677,-58.794],[-50.228,-20.643,-57.682],
[-52.532,-16.102,-47.945],[-55.183,-13.950,-51.126],[-48.484,-14.131,-54.021],
[-48.634,-15.038,-42.449],[-46.899,-16.142,-50.399],[-55.732,-16.393,-45.519],
[-53.503,-17.145,-57.876],[-58.367,-17.599,-51.454],[-57.278,-18.084,-56.256],
[-48.638,-14.504,-47.387],[-50.915,-17.911,-51.745],[-55.415,-15.903,-53.151],
[-53.584,-17.630,-55.233],[-46.156,-22.101,-48.387],[-54.372,-14.918,-56.782],
[-44.323,-2.987,-42.497] ,[-53.652,-11.125,-43.480],[-49.557,-15.692,-49.760],
[-53.573,-17.753,-52.666],[-49.971,-16.625,-46.615],[-49.799,-12.523,-55.765],
[-43.827,-12.968,-47.873],[-46.708,-17.937,-57.120],[-48.749,-10.166,-43.048],
[-44.949,-16.067,-56.649],[-56.907,-10.664,-48.972],[-48.226,-9.323,-55.910],
[-57.498,-13.568,-45.365],[-50.440,-12.442,-43.349],[-46.807,-7.807,-44.206],
[-39.882,-15.412,-56.324],[-41.135,-11.027,-44.659],[-43.762,-8.826,-53.554],
[-45.623,-12.802,-49.697],[-53.372,-6.817,-51.271],[-56.560,-13.229,-57.539],
[-51.602,-10.303,-53.455],[-51.412,-6.657,-48.111],[-53.101,-3.074,-53.776],
[-49.859,-11.774,-46.565],[-46.516,-8.483,-58.155],[-44.165,-6.278,-43.403],
[-41.614,-5.585,-57.134],[-57.538,-18.448,-45.218],[-45.571,-10.567,-55.181],
[-56.092,-7.505,-46.668],[-51.539,-7.655,-59.550],[-50.524,-9.958,-57.379],
[-53.866,-10.093,-55.207],[-48.375,-18.747,-46.441],[-47.281,-20.325,-55.405],
[-49.049,-6.748,-56.160],[-49.950,-16.492,-54.827],[-50.823,-15.934,-43.313],
[-48.395,-18.707,-50.722],[-52.252,-7.080,-57.071],[-53.198,-20.027,-54.281],
[-44.340,-19.743,-44.429],[-44.051,-15.281,-49.647],[-55.141,-7.486,-56.725],
[-58.305,-15.212,-57.651],[-43.799,-4.501,-47.553],[-55.156,-21.410,-49.365],
[-56.536,-5.747,-54.215],[-57.698,-4.949,-42.336],[-53.730,-6.290,-39.917],
[-49.582,-5.989,-58.706],[-47.115,-8.275,-41.112],[-44.194,-7.164,-58.943],
[-57.934,-11.490,-41.305],[-53.845,-9.149,-58.370],[-52.286,-11.815,-41.515],
[-49.095,-8.822,-59.340],[-46.219,-10.460,-44.299],[-57.916,-14.972,-41.708],
[-52.237,-12.297,-56.895],[-53.137,-15.663,-45.284],[-50.612,-14.349,-58.351],
[-47.063,-13.292,-41.379],[-42.408,-10.879,-57.367],[-53.089,-19.013,-42.932],
[-51.493,-18.575,-44.834],[-51.006,-14.074,-45.395],[-47.073,-16.148,-40.518],
[-45.498,-12.744,-43.356],[-53.887,-15.534,-41.312],[-51.625,-21.376,-44.797],
[-50.166,-17.634,-41.093],[-48.472,-17.803,-43.182],[-47.103,-20.424,-43.696],
[-55.965,-17.196,-41.267],[-52.591,-23.027,-41.697],[-50.593,-21.202,-40.073],
[-51.838,-4.567,-57.849],[-57.442,-8.103,-58.148],[-58.066,-3.639,-53.502],
[-57.803,-6.317,-49.450],[-41.003,-5.918,-48.143],[-41.693,-6.378,-45.009],
[-42.787,-10.166,-40.893],[-41.277,-10.208,-54.518],[-42.758,-14.320,-52.753],
[-42.002,-11.075,-47.244],[-42.278,-8.410,-46.627],[-41.709,-12.833,-54.944],
[-41.812,-15.188,58.617],[-44.279,-16.503,-51.998],[-57.868,-11.184,-46.535],
[-42.105,-13.300,-43.615],[-40.001,-15.727,-42.084],[-44.268,-16.421,59.028],
[-43.665,-17.283,-54.742],[-41.780,-16.486,-49.183],[-42.672,-15.280,-46.911],
[-58.211,-18.232,-42.751],[-41.922,-24.258,-39.752],[-43.663,-19.562,-53.319],
[-57.537,-18.593,-48.881],[-58.632,-14.679,-48.098],[-41.798,-19.207,-43.371],
[-42.024,-18.348,59.013],[-58.889,-18.886,-54.181],[-43.995,-5.729,-52.356],
[-58.605,-20.803,-44.776],[-56.023,-9.504,-42.504],[-53.553,-7.892,-45.637],
[-48.276,-5.915,-42.749],[-48.105,-3.206,-42.716],[-42.227,-22.126,-44.353],
[-56.083,-4.638,-46.970],[-53.535,-4.692,-48.057],[-48.943,-5.222,-48.020],
[-46.601,-5.785,-46.597],[-45.320,-4.107,-44.621],[-55.199,-7.252,-49.078],
[-56.308,-6.817,-51.537],[-53.632,-3.813,-50.848],[-50.558,-18.226,-48.572],
[-42.314,-22.073,-49.494],[-41.608,-20.983,-51.981],[-55.286,-19.730,-52.142],
[-58.263,-20.974,-56.227],[-50.965,-22.284,-51.452],[-45.905,-19.742,-49.766],
[-44.922,-20.897,-56.082],[-57.774,-5.533,-56.933],[-54.506,-18.635,59.044],
[-49.683,-20.708,-54.386],[-48.861,-3.734,-56.646],[-42.465,-4.362,-53.738],
[-56.807,-20.647,-42.213],[-53.595,-20.817,-39.835],[-56.271,-1.988,-57.509],
[-46.195,-21.531,-41.278],[-42.654,-7.176,-41.208],[-48.347,7.390,-52.166],
[-48.733,4.417,-50.011],[-55.902,4.536,-48.176],[-45.273,7.127,-52.748],
[-55.058,9.378,-48.577],[-57.405,3.291,-52.752],[-57.987,6.280,-51.794],
[-46.914,13.948,-50.698],[-48.831,7.776,-48.628],[-41.644,-1.683,-47.291],
[-47.215,1.864,-47.593],[-54.628,0.424,-51.178],[-45.877,-0.432,-53.267],
[-56.960,2.813,-50.405],[-52.328,3.277,-51.225],[-41.977,2.700,-44.890],
[-54.559,-0.027,-47.476],[-48.038,8.540,-54.795],[-53.539,6.513,-47.161],
[-44.957,-0.608,-50.752],[-54.570,2.750,-49.459],[-53.201,8.930,-46.049],
[59.669,8.529,-55.173],[-45.549,8.343,-46.910],[-43.507,0.214,-47.327],
[-50.682,-3.278,-48.724],[-45.681,10.935,-51.287],[-45.298,10.010,-53.829],
[-44.739,9.039,-44.064],[-48.140,6.195,-56.296],[-48.966,4.155,-57.735],
[-40.129,5.542,-44.889],[-46.892,4.800,-52.679],[-45.401,3.058,-45.930],
[-52.765,5.821,-52.438],[-52.249,9.023,-48.974],[-53.825,5.610,-49.640],
[-47.733,11.415,-49.568],[-53.591,10.806,-43.474],[-46.191,4.679,-55.515],
[-55.692,6.837,-45.267],[-49.912,11.569,-51.891],[-54.351,13.547,-42.722],
[-43.232,-3.307,-56.802],[-56.326,4.954,-54.412],[-41.828,10.237,-59.060],
[-58.209,5.851,-47.367],[-58.741,11.058,-47.383],[-39.682,10.817,-45.662],
[-49.661,9.798,-44.813],[-48.365,14.468,-53.007],[-54.282,14.982,-51.954],
[-45.552,13.123,-57.605],[-46.906,12.727,-55.374],[-44.302,2.694,-40.462],
[-57.734,8.653,-46.298],[-52.761,-1.637,-49.176],[-48.652,-4.249,-45.375],
[-44.720,1.397,-43.948],[-44.147,5.985,-57.117],[-48.995,-1.460,-57.788],
[-52.693,3.464,-47.825],[-54.394,2.765,-52.996],[-48.913,3.851,-54.798],
[-48.735,2.445,-42.030],[-46.075,1.830,-51.702],[-55.561,1.856,-45.488],
[-52.910,2.028,-58.424],[-58.266,0.829,-51.864],[-58.034,1.757,-56.628],
[-48.991,3.776,-45.037],[-51.626,-1.192,-52.079],[-57.455,-0.603,-54.809],
[-52.609,2.124,-54.788],[-46.714,-1.106,-48.061],[-53.572,3.823,-56.517],
[-44.530,13.706,-41.301],[-55.964,7.381,-42.423],[-49.708,1.956,-50.724],
[-54.302,-0.289,-54.198],[-50.216,2.120,-47.405],[-51.589,8.789,-55.889],
[-44.077,6.023,-46.164],[-48.243,1.625,-56.796],[-47.983,7.959,-43.810],
[-45.568,2.073,-54.784],[-56.077,7.158,-49.407],[-48.537,8.238,-58.309],
[-57.428,3.767,-45.347],[-50.587,4.341,-43.187],[-47.159,11.065,-44.532],
[-40.926,2.121,-57.006],[-40.010,7.989,-44.357],[-42.245,10.451,-55.009],
[-46.815,6.108,-48.125],[-52.887,12.080,-51.353],[-57.664,7.045,-58.230],
[-52.493,8.792,-51.852],[-51.235,10.850,-46.623],[-53.993,13.519,-54.177],
[-49.053,6.196,-46.009],[-46.088,9.033,-57.783],[-44.648,12.061,-44.061],
[-40.849,13.437,-55.556],[-56.496,-1.911,-44.842],[-45.752,7.534,-55.478],
[-56.705,11.414,-45.482],[-51.412,10.455,58.060],[-52.775,10.079,-57.913],
[-55.176,8.622,-55.863],[-48.448,0.242,-46.064],[-47.854,0.188,-54.520],
[-49.901,10.979,-55.131],[-50.062,1.428,-53.544],[-50.328,1.671,-44.068],
[-49.043,-1.608,-53.247],[-51.350,13.130,-55.011],[-53.504,-1.748,-56.499],
[-43.594,-1.007,-44.692],[-44.453,4.342,-50.485],[-55.356,10.896,-58.251],
[-55.925,3.118,-58.255],[-44.176,12.979,-47.908],[-57.612,-3.263,-49.490],
[-57.639,13.162,-52.051],[-57.003,13.091,-42.565],[-52.909,10.327,-40.976],
[-49.486,12.130,-59.661],[-45.868,10.758,-41.995],[-44.575,10.975,-58.959],
[-59.601,7.682,-42.439],[-55.015,7.867,-58.382],[-50.649,7.190,-43.587],
[-49.181,9.059,58.195],[-46.543,5.474,-45.531],[-57.763,2.441,-42.910],
[-52.747,6.560,-57.039],[-54.429,4.348,-45.388],[-51.144,3.672,59.108],
[-48.862,5.606,-41.317],[-42.619,7.306,-58.736],[-52.833,1.108,-43.479],
[-51.373,-0.586,-45.136],[-51.610,4.582,-45.913],[-46.325,1.152,-41.336],
[-45.288,5.735,-43.036],[-54.317,2.985,-42.046],[-51.376,-4.369,-46.466],
[-50.619,0.745,-40.990],[-48.647,-0.112,-42.784],[-46.208,-1.164,-43.352],
[-55.894,0.149,-42.471],[-50.887,-4.460,-43.894],[-51.748,-1.697,-41.967],
[-51.350,14.671,-57.986],[-57.452,9.496,-56.942],[-58.245,15.191,-53.542],
[-59.502,12.073,-49.945],[-41.085,13.097,-47.686],[-42.247,12.484,-45.228],
[-43.442,8.439,-41.744],[-41.971,7.337,-55.948],[-41.212,5.443,-53.651],
[-41.408,9.880,-47.688],[-44.277,10.745,-46.410],[-41.737,4.461,-57.125],
[-40.965,2.105,58.938],[-43.070,2.919,-53.261],[-59.145,7.512,-49.472],
[-42.712,6.790,-44.165],[-41.581,4.416,-42.725],[-44.333,0.962,-58.393],
[-42.735,0.328,-55.881],[-44.999,1.563,-49.090],[-42.980,3.499,-47.527],
[-59.487,0.155,-43.424],[-40.505,-2.952,-39.841],[-42.348,-1.715,-53.517],
[-57.290,-0.163,-48.717],[-58.162,2.544,-48.216],[-40.762,-3.170,-42.515],
[-41.263,-0.505,-58.429],[-59.066,-1.900,-51.545],[-44.659,15.292,-53.331],
[-59.169,-1.892,-46.147],[-56.997,9.969,-42.965],[-53.891,12.107,-45.937],
[-48.897,13.402,-42.230],[-46.402,14.568,-43.551],[-42.588,-3.638,-44.705],
[-55.840,16.232,-46.008],[-54.101,14.314,-47.426],[-48.626,13.574,-48.106],
[-46.584,13.064,-46.018],[-45.108,15.814,-45.997],[-56.579,14.308,-49.572],
[-55.093,12.068,-49.769],[-50.735,11.894,-49.322],[-48.952,-1.882,-50.374],
[-43.638,-3.009,-49.768],[-41.080,-4.158,-50.363],[-56.526,-1.419,-52.390],
[-58.159,-3.056,-56.111],[-51.110,-3.613,-51.693],[-46.340,-3.178,-49.947],
[-45.087,-1.759,-55.365],[-56.941,11.636,-54.958],[-54.387,-0.506,-58.646],
[-51.432,-0.494,-55.083],[-48.137,15.284,-56.022],[-43.862,12.815,-54.950],
[-54.868,-2.276,-42.780],[-53.184,-3.988,-41.625],[-55.579,16.916,-57.833],
[-47.857,-4.412,-40.392],[-41.645,11.676,-42.406],[-48.165,24.456,-51.895],
[-49.443,24.883,-48.886],[-56.702,23.909,-48.420],[-45.161,28.598,-51.319],
[-54.479,28.619,-51.539],[-57.562,22.980,-53.271],[-58.414,24.442,-50.695],
[-47.988,33.404,-50.906],[-49.766,27.299,-46.450],[-41.268,17.430,-47.241],
[-45.515,20.608,-47.423],[-54.210,20.894,-51.217],[-47.546,19.816,-54.122],
[-57.347,21.985,-50.522],[-51.477,22.127,-51.562],[-38.995,20.743,-45.710],
[-54.361,18.653,-47.606],[-49.473,27.274,-51.743],[-52.205,25.733,-47.094],
[-43.594,18.246,-50.922],[-56.391,19.272,-49.916],[-55.034,28.177,-47.043],
[-58.670,27.680,-54.451],[-45.162,27.616,-48.440],[-43.818,17.231,-48.015],
[-50.649,15.302,-48.555],[-46.640,30.874,-51.740],[-47.510,28.035,-53.274],
[-44.032,26.447,-44.002],[-49.403,25.612,-54.004],[-49.409,24.044,-58.785],
[-40.584,22.670,-46.487],[-46.005,25.112,-54.576],[-45.463,22.045,-45.130],
[-54.711,25.037,-51.961],[-51.320,27.665,-49.783],[-54.216,23.672,-49.614],
[-47.257,29.020,-46.975],[-51.663,27.460,-44.486],[-46.437,23.662,-56.792],
[-54.467,25.996,-45.336],[-48.327,30.258,-49.636],[-53.796,29.137,-43.200],
[-41.948,15.863,-55.360],[-55.754,25.065,-54.526],[-42.242,30.670,-58.096],
[-58.675,25.019,-45.448],[-58.533,28.456,-48.765],[-39.621,30.588,-43.813],
[-50.297,29.796,-45.471],[-49.292,33.210,-53.481],[-56.747,32.956,-55.662],
[-46.576,32.719,-57.324],[-46.581,32.833,-53.567],[-42.361,22.013,-43.651],
[-57.032,27.315,-45.023],[-52.244,17.983,-49.550],[-48.259,15.567,-46.343],
[-46.337,19.572,-44.352],[-44.779,26.115,-57.601],[-49.028,18.434,-56.579],
[-52.516,20.894,-48.207],[-52.676,20.678,-53.691],[-48.359,22.290,-54.503],
[-49.493,22.207,-41.916],[-46.626,21.986,-50.277],[-55.061,22.097,-46.838],
[-50.716,18.043,-59.153],[-58.879,19.717,-51.295],[-58.079,20.289,-54.198],
[-47.776,22.749,-47.737],[-51.871,18.078,-52.413],[-55.299,21.070,-53.548],
[-52.665,20.186,-56.347],[-46.665,17.749,-46.971],[-54.255,22.947,-55.601],
[-43.754,35.088,-41.863],[-56.140,26.047,-42.196],[-49.223,20.765,-50.528],
[-54.406,17.719,-53.032],[-50.090,20.981,-47.097],[-51.099,27.115,-55.724],
[-42.341,25.006,-47.200],[-46.141,19.029,-56.780],[-48.391,26.943,-44.219],
[-45.599,21.621,-55.330],[-54.529,26.633,-49.202],[-48.613,28.599,-56.387],
[-57.216,22.868,-45.204],[-52.151,21.539,-42.336],[-47.437,29.530,-44.309],
[-40.832,21.526,-56.162],[-40.912,28.036,-43.844],[-43.345,29.657,-54.808],
[-45.331,24.490,-48.618],[-55.107,30.571,-54.540],[-57.983,24.430,-57.850],
[-52.500,26.779,-52.424],[-50.895,30.333,-48.464],[-54.348,32.443,-52.804],
[-47.374,26.342,-47.026],[-45.943,28.406,-58.967],[-44.543,31.116,-44.016],
[-40.457,32.667,-57.375],[-58.196,17.109,-45.077],[-46.001,27.781,-55.714],
[-56.522,31.265,-47.060],[-51.702,28.988,-58.338],[-54.030,28.936,-56.796],
[-56.285,27.014,-56.403],[-48.762,18.023,-45.062],[-46.698,17.198,-54.733],
[-51.553,30.742,-54.960],[-50.266,20.084,-54.839],[-48.616,22.190,-44.838],
[-49.334,17.594,-52.977],[-52.638,31.484,-57.595],[-54.184,17.588,-55.628],
[-41.629,19.129,-44.940],[-43.778,22.581,-51.214],[-57.069,30.199,-56.298],
[-55.827,20.737,-56.935],[-44.206,31.836,-48.446],[-54.977,16.880,-49.430],
[-56.950,29.162,-52.628],[-56.632,32.335,-42.506],[-53.028,31.504,-41.076],
[-49.663,31.008,-58.036],[-46.308,27.329,-42.388],[-44.940,30.459,-57.537],
[-56.524,23.671,-40.873],[-54.497,26.695,-59.204],[-51.393,24.671,-43.534],
[-49.020,27.968,-59.000],[-46.036,25.133,-44.934],[-54.868,23.955,-43.362],
[-53.409,25.784,-56.806],[-51.708,22.317,-45.226],[-51.639,22.386,-57.950],
[-47.779,24.060,-42.754],[-42.723,28.085,-58.678],[-52.016,19.486,-44.294],
[-51.487,18.574,-46.856],[-49.659,24.469,-46.173],[-47.243,20.246,-40.871],
[-44.904,23.764,-42.907],[-54.251,21.767,-40.532],[-50.895,15.585,-45.650],
[-50.724,17.661,-42.730],[-49.040,20.087,-43.184],[-46.170,17.422,-42.231],
[-54.509,19.733,-42.082],[-51.388,14.986,-42.780],[-50.319,16.096,-40.765],
[-51.553,35.286,-57.474],[-56.480,28.544,-58.738],[58.783,33.812,-52.868],
[-58.798,31.679,-52.020],[-41.459,32.401,-48.392],[-42.531,31.319,-46.254],
[-43.566,28.493,-39.296],[-43.290,26.011,-55.471],[-41.750,24.431,-53.749],
[-41.583,26.994,-48.842],[-43.054,29.028,-45.393],[-42.234,25.026,-57.847],
[-41.694,22.286,-58.706],[-43.157,21.912,-54.071],[-57.148,26.449,-47.965],
[-42.224,24.799,-42.056],[-40.293,23.029,-41.852],[-44.181,18.872,-58.678],
[-43.849,18.211,-55.592],[-43.226,20.295,-49.067],[-43.165,21.748,-46.546],
[-59.408,19.111,-41.874],[-41.351,14.532,-41.616],[-41.625,18.913,-54.630],
[-57.898,16.871,-49.341],[-57.619,21.334,-47.507],[-39.695,17.118,-42.138],
[-40.397,17.018,-58.502],[-58.982,17.826,-54.816],[-42.708,33.647,-53.684],
[-57.723,13.991,-45.716],[-56.739,28.819,-40.810],[-53.624,30.530,-47.283],
[-49.290,31.553,-43.396],[-46.309,33.105,-43.586],[-43.585,14.425,-43.733],
[-55.127,33.790,-44.108],[-53.760,33.357,-46.335],[-49.167,33.359,-48.179],
[-46.570,32.039,-47.257],[-44.687,33.853,-46.331],[-58.495,31.046,-48.914],
[-55.884,30.450,-50.011],[-52.949,30.615,-50.249],[-49.609,17.491,-50.006],
[-43.878,13.281,-51.836],[-42.068,15.894,-50.891],[-56.997,18.054,-52.666],
[-58.130,14.778,-56.406],[-50.787,14.691,-51.738],[-45.054,15.843,-50.521],
[-44.793,15.512,-56.197],[-59.555,31.489,-56.000],[-54.048,19.058,-58.464],
[-51.597,17.481,-56.368],[-49.133,33.783,-57.866],[-42.979,32.554,-56.099],
[-55.729,16.231,-43.300],[-53.175,15.192,-40.724],[-56.252,33.376,-58.498],
[-47.254,14.786,-41.011],[-42.291,30.777,-42.191],[-47.358,42.664,-52.644],
[-49.894,41.941,-49.626],[-55.421,42.690,-46.947],[-46.069,45.557,-50.722],
[-53.874,47.151,-48.156],[-57.232,42.180,-53.549],[-56.869,42.140,-49.911],
[-49.494,49.379,-50.733],[-50.037,44.939,-48.669],[-41.361,35.885,-48.742],
[-47.819,38.574,-48.518],[-54.387,37.497,-50.316],[-45.880,37.270,-52.727],
[-56.903,39.947,-51.692],[-53.510,40.925,-50.500],[-39.589,39.897,-44.744],
[-55.452,37.637,-47.751],[-48.265,45.390,-52.367],[-51.319,43.970,-45.580],
[-43.625,35.999,-52.210],[-55.595,40.104,-48.622],[-53.717,46.227,-45.370],
[59.164,46.671,-55.770],[-43.797,47.112,-49.387],[-45.208,36.352,-47.273],
[-53.604,32.968,-49.234],[-45.951,48.580,-50.580],[-46.067,47.045,-53.426],
[-44.464,46.204,-44.016],[-48.591,43.496,-54.986],[-47.794,39.853,-59.143],
[-41.309,42.825,-46.316],[-45.519,42.201,-54.731],[-45.441,40.972,-44.973],
[-51.135,41.633,-53.305],[-53.399,46.486,-50.761],[-52.277,42.030,-48.155],
[-47.083,47.885,-48.107],[-51.414,46.625,-43.815],[-47.906,42.226,-57.577],
[-55.231,43.286,-42.988],[-49.787,48.073,-53.144],[-54.861,49.634,-41.174],
[-44.216,35.051,-57.180],[-53.579,42.240,-55.234],[-40.685,47.251,-58.562],
[59.392,44.487,-46.760],[-58.949,46.904,-48.269],[-39.963,49.608,-43.820],
[-49.417,47.572,-45.059],[-49.405,51.140,-53.192],[-55.091,51.183,-52.215],
[-47.768,49.002,-57.345],[-46.398,51.175,-56.140],[-42.689,40.147,-43.242],
[-56.166,46.514,-44.366],[-54.589,35.230,-48.505],[-47.759,33.884,-45.755],
[-45.827,38.674,-43.897],[-44.685,42.290,-57.382],[-48.826,37.385,-57.434],
[-52.898,39.918,-46.540],[-53.047,38.346,-52.751],[-48.598,40.418,-54.438],
[-48.106,40.244,-42.643],[-45.472,40.243,-52.538],[-54.855,39.018,-44.727],
[-52.242,37.340,59.345],[-56.941,37.157,-52.052],[-57.561,40.092,-57.034],
[-46.578,41.065,-47.593],[-50.559,36.947,-52.334],[-56.037,40.075,-54.451],
[-50.990,38.899,-55.685],[-47.824,35.761,-48.369],[-53.906,39.505,-57.588],
[-44.219,51.397,-41.774],[-57.247,45.096,-42.509],[-49.688,39.399,-50.304],
[-54.324,36.053,-52.771],[-51.051,37.335,-48.141],[-50.535,44.973,-55.708],
[-44.471,44.576,-47.783],[-44.936,37.544,-57.698],[-48.425,44.920,-44.650],
[-44.388,39.879,-56.463],[-55.071,44.919,-49.380],[-47.774,46.549,-55.432],
[-57.930,42.068,-45.941],[-51.141,41.039,-42.526],[-46.342,46.163,-46.311],
[-41.884,39.799,-55.482],[-40.298,46.980,-44.088],[-43.512,47.111,-54.987],
[-46.363,42.796,-49.783],[-53.075,48.138,-52.849],[-57.556,44.524,-58.819],
[-52.921,45.268,-53.321],[-50.557,48.931,-47.288],[-51.678,49.657,-54.488],
[-47.626,43.675,-47.555],[-46.016,45.514,-57.612],[-45.115,48.661,-44.640],
[-42.177,49.325,-57.532],[-56.669,36.897,-44.846],[-45.570,44.857,-54.721],
[-58.201,49.242,-46.242],[-51.860,47.775,-59.620],[-52.723,46.477,-57.304],
[-55.253,44.460,-57.164],[-48.951,36.720,-46.014],[-47.193,35.883,-55.615],
[-50.102,47.545,-56.820],[-48.388,37.961,-54.612],[-49.676,39.181,-44.636],
[-47.661,35.517,-52.996],[-51.996,49.441,-57.239],[-53.245,35.815,-55.412],
[-43.658,37.944,-45.252],[-44.722,40.820,-49.614],[-54.448,48.994,-58.249],
[-56.624,39.991,-59.429],[-44.592,49.698,-48.516],[-56.064,33.682,-50.939],
[-55.873,46.744,-52.441],[-56.920,51.167,-42.142],[-53.057,47.755,-40.368],
[-49.325,49.010,-59.338],[-46.823,46.681,-42.339],[-44.834,48.759,-58.122],
[-59.375,44.834,-40.929],[-53.994,45.073,59.386],[-51.096,43.833,-41.470],
[-49.145,44.899,-58.905],[-45.900,43.668,-45.410],[-57.722,41.913,-42.845],
[-52.638,43.621,-57.623],[-53.230,42.197,-44.696],[-50.676,41.196,59.429],
[-48.613,42.938,-42.492],[-43.084,44.316,-58.491],[-52.333,38.730,-43.849],
[-52.702,36.673,-45.545],[-50.211,41.296,-46.543],[-44.810,39.712,-41.345],
[-46.044,42.054,-40.957],[-55.620,40.999,-41.317],[-51.095,34.612,-46.709],
[-49.889,38.488,-40.791],[-48.202,37.280,-43.010],[-45.669,37.067,-41.526],
[-55.293,37.873,-41.047],[-51.609,34.197,-43.967],[-49.918,35.769,-41.814],
[-52.053,52.076,-56.733],[-57.293,47.423,-58.054],[-57.882,53.264,-54.026],
[-59.460,50.161,-51.463],[-39.654,52.442,-48.583],[-42.095,49.581,-45.831],
[-44.066,46.984,-41.424],[-42.655,44.397,-55.585],[-41.301,43.196,-53.453],
[-41.851,44.956,-48.273],[-43.056,47.017,-46.659],[-40.487,43.897,-57.281],
[-41.598,39.623,-58.090],[-42.458,40.346,-51.055],[-57.832,44.555,-49.025],
[-43.921,42.626,-43.247],[-40.584,42.304,-40.830],[-41.993,37.010,-58.014],
[-43.755,37.772,-54.735],[-43.653,37.751,-49.114],[-43.176,39.885,-46.907],
[-57.500,39.576,-44.284],[-40.724,35.070,-40.045],[-42.049,35.844,-54.878],
[-57.685,36.059,-49.650],[59.365,40.359,-50.166],[-41.037,35.358,-42.815],
[-39.862,35.679,-59.174],[-58.201,37.614,-54.510],[-45.614,49.820,-53.104],
[-58.401,34.955,-45.274],[-56.832,47.753,-41.662],[-54.951,49.074,-46.107],
[-47.630,50.082,-43.623],[-45.742,52.253,-44.027],[-42.636,33.409,-44.484],
[-56.501,51.718,-47.228],[-53.046,53.603,-44.907],[-48.804,51.048,-47.897],
[-46.957,49.862,-46.333],[-45.094,51.910,-46.613],[-57.913,49.206,-49.163],
[-54.962,48.729,-50.590],[-52.208,49.155,-50.029],[-48.441,36.606,-50.882],
[-43.793,34.507,-48.763],[-42.185,33.978,-51.148],[-56.709,34.953,-53.969],
[-57.720,36.050,-56.786],[-51.753,34.487,-52.718],[-45.036,33.218,-51.195],
[-45.029,34.969,-54.428],[-56.732,50.598,-54.805],[-54.623,36.641,-58.007],
[-50.566,36.180,-54.939],[-47.894,52.558,-58.404],[-41.217,51.249,-55.359],
[-56.067,35.639,-42.390],[-53.516,35.236,-41.698],[-55.577,52.010,-57.959],
[-46.538,34.406,-41.210],[-42.719,48.369,-43.463],[-41.619,54.312,-48.798],
[-46.224,57.680,-48.716],[-52.875,56.293,-51.782],[-46.058,55.481,-53.012],
[-58.051,58.596,-51.818],[-40.548,56.398,-46.481],[-55.307,54.783,-46.914],
[-43.610,56.237,-52.294],[-55.408,57.068,-49.861],[-44.408,54.565,-47.893],
[-50.648,51.687,-49.802],[-43.605,51.460,-57.512],[-44.309,57.790,-42.698],
[-54.157,50.909,-49.123],[-47.545,53.047,-46.377],[-45.585,57.655,-46.060],
[-50.128,53.983,-57.522],[-53.843,56.609,-47.730],[-54.622,-59.256,-52.896],
[-48.630,57.661,-52.717],[-55.150,-58.093,-45.759],[-52.839,57.117,-59.337],
[-58.313,56.107,-52.804],[-56.926,56.829,-55.154],[-52.997,54.226,-49.421],
[-55.455,56.210,-52.423],[-52.354,57.710,-54.542],[-46.905,53.887,-48.784],
[-54.303,58.816,-57.638],[-48.320,58.256,-50.261],[-53.985,53.652,-52.305],
[-50.105,58.456,-47.673],[-45.415,56.259,-57.092],[-46.996,57.680,-55.537],
[-40.505,58.504,-56.631],[-56.683,56.322,-44.739],[-47.745,55.787,-46.046],
[-46.360,54.238,-55.572],[-49.830,58.700,-54.891],[-49.411,57.706,-45.059],
[-48.039,53.637,-52.911],[-52.735,54.708,-55.507],[-42.277,57.471,-44.803],
[-56.124,56.936,-58.671],[-58.411,51.686,-49.684],[-52.000,58.081,-43.588],
[-51.255,55.915,-47.384],[-47.856,58.106,-40.549],[-54.621,57.305,-43.122],
[-51.589,51.375,-46.126],[-49.398,56.717,-42.732],[-46.874,56.852,-42.954],
[-46.261,54.155,-41.954],[-58.358,56.206,-40.496],[-50.154,51.685,-43.582],
[-49.646,54.086,-41.640],[-43.431,58.300,-57.886],[-44.636,54.017,-58.404],
[-44.250,57.744,-55.115],[-43.691,58.511,-49.877],[-42.910,59.030,-47.055],
[-59.156,56.452,-43.027],[-41.502,53.999,-40.644],[-43.531,54.892,-55.394],
[-56.952,55.105,-48.959],[-57.861,58.024,-49.102],[-40.767,54.467,-44.062],
[-40.567,54.877,-58.358],[-59.552,55.320,-55.510],[-57.999,53.631,-46.131],
[-43.275,53.378,-45.039],[-49.095,55.246,-50.370],[-44.771,51.758,-51.093],
[-42.698,53.472,-51.317],[-57.260,54.126,-51.568],[-57.799,53.891,-57.236],
[-51.928,51.663,-52.382],[-48.125,52.199,-50.699],[-44.376,52.274,-54.164],
[-53.350,53.759,-58.399],[-50.344,53.081,-54.985],[-56.206,54.068,-42.575],
[-53.976,52.182,-42.590],[-46.894,51.746,-41.336],[-47.189,-49.777,-34.693],
[-49.491,-50.402,-31.539],[-56.655,-51.591,-30.925],[-43.600,-47.526,-33.648],
[-55.414,-46.344,-29.666],[-58.763,-52.415,-35.071],[-58.935,-49.750,-32.464],
[-47.644,-43.737,-31.323],[-51.618,-48.682,-29.741],[-42.175,-56.784,-29.764],
[-47.151,-54.018,-30.536],[-54.175,-55.539,-32.727],[-46.158,-55.322,-32.892],
[-56.072,-53.523,-32.593],[-52.494,-51.783,-33.753],[-41.258,-54.611,-28.190],
[-54.438,-55.605,-28.481],[-48.226,-46.243,-34.977],[-53.435,-49.971,-27.595],
[-44.000,-57.225,-32.225],[-54.065,-53.469,-30.694],[-53.725,-47.519,-28.016],
[-57.618,-46.526,-35.877],[-44.524,-46.989,-31.493],[-43.927,-57.513,-28.088],
[-50.933,58.580,-31.477],[-46.484,-45.534,-33.045],[-46.180,-45.859,-37.248],
[-44.688,-46.622,-25.932],[-49.755,-49.965,-36.336],[-47.812,-50.870,-40.770],
[-42.798,-52.318,-27.666],[-44.693,-49.908,-35.675],[-46.032,-52.920,-26.346],
[-53.826,-49.006,-32.798],[-53.399,-44.614,-30.434],[-53.591,-50.925,-31.101],
[-47.024,-46.384,-30.213],[-52.327,-45.557,-25.698],[-47.291,-50.770,-37.551],
[-55.843,-50.598,-26.100],[-48.402,-43.688,-34.369],[-55.009,-44.648,-24.780],
[-43.352,-58.470,-38.228],[-54.474,-50.388,-35.368],[-42.993,-43.830,-39.781],
[-58.798,-48.949,-27.017],[-59.007,-45.631,-30.344],[-40.309,-45.802,-25.237],
[-49.952,-45.503,-26.844],[-50.144,-41.123,-35.759],[-54.476,-41.584,-33.608],
[-45.034,-40.981,-39.855],[-46.303,-43.254,-36.744],[-44.450,-52.667,-23.624],
[-57.164,-46.582,-25.987],[-53.280,-58.251,-29.533],[-50.153,-57.175,-24.883],
[-45.345,-55.182,-25.241],[-45.671,-49.274,-39.605],[-49.891,-54.649,-39.450],
[-53.503,-53.019,-27.922],[-55.153,-52.925,-35.139],[-47.902,-52.775,-35.631],
[-48.394,-51.590,-23.615],[-45.721,-52.510,-34.056],[-55.392,-54.609,-25.834],
[-51.741,-56.492,-40.733],[-59.112,-54.806,-33.034],[-56.550,-56.603,-38.377],
[-48.868,-49.999,-27.176],[-53.604,-58.241,-32.314],[-57.183,-54.611,-34.951],
[-51.430,-51.912,-36.284],[-49.319,59.370,-28.748],[-53.582,-55.521,-38.658],
[-43.617,-41.740,-24.943],[-55.425,-48.130,-24.430],[-50.822,-53.849,-29.875],
[-55.093,-57.832,-34.628],[-50.122,-54.917,-27.402],[-50.477,-46.881,-36.348],
[-44.247,-50.248,-29.148],[-47.088,-55.561,-39.380],[-49.184,-47.844,-25.519],
[-45.795,-54.264,-36.947],[-56.193,-48.860,-31.024],[-48.283,-47.193,-37.899],
[-57.973,-53.561,-26.187],[-50.816,-50.527,-25.297],[-47.758,-43.903,-27.187],
[-42.217,-53.679,-37.452],[-41.901,-48.572,-25.072],[-41.552,-46.105,-38.163],
[-47.356,-51.439,-29.371],[-53.569,-46.390,-35.568],[-58.068,-50.870,-39.010],
[-52.093,-48.610,-35.031],[-51.242,-43.888,-28.803],[-52.557,-44.361,-34.332],
[-48.439,-48.529,-29.495],[-44.287,-46.933,-38.868],[-44.305,-43.783,-26.859],
[-41.448,-43.005,-37.854],[-57.952,-56.249,-26.529],[-43.726,-47.548,-36.361],
[-57.047,-44.298,-29.332],[-52.388,-45.913,-40.893],[-52.008,-47.848,-38.656],
[-56.122,-48.690,-38.218],[-48.341,-57.063,-27.376],[-47.757,-57.023,-37.078],
[-49.480,-44.480,-36.953],[-49.661,-54.871,-36.687],[-49.409,-52.588,-25.938],
[-50.561,-55.703,-34.321],[-52.506,-43.852,-38.159],[-52.854,-58.254,-36.027],
[-42.847,-54.534,-25.803],[-44.994,-52.441,-30.570],[-55.328,-45.575,-37.708],
[-56.162,-52.850,-39.948],[-44.020,-42.395,-29.213],[-56.077,-58.831,-30.542],
[-56.504,-43.228,-34.405],[-57.335,-42.869,-24.455],[-53.232,-43.392,-22.593],
[-50.147,-44.047,-41.911],[-47.246,-45.988,-25.190],[-46.553,-43.449,-39.441],
[-58.376,-48.925,-23.906],[-54.242,-49.071,-40.128],[-50.886,-48.781,-23.073],
[-49.314,-46.988,-41.272],[-46.155,-48.801,-26.633],[-57.425,-51.533,-24.091],
[-53.011,-50.448,-37.935],[-54.663,-52.414,-24.313],[-52.976,-52.109,-39.951],
[-47.424,-49.791,-21.945],[-42.279,-48.784,-38.421],[-52.655,-54.293,-25.571],
[-51.797,-56.808,-27.893],[-50.917,-51.179,-28.434],[-46.431,-54.379,-23.046],
[-45.517,-50.354,-24.268],[-53.773,-54.725,-22.748],[-52.357,-59.071,-25.526],
[-51.062,-55.097,-22.917],[-48.274,-55.206,-25.257],[-47.403,-56.443,-21.532],
[-56.578,-55.419,-23.432],[-52.127,57.637,-24.095],[-51.316,-58.554,-21.663],
[-50.846,-39.967,-38.357],[-58.679,-46.967,-38.392],[-57.972,-41.094,-36.034],
[58.849,-43.664,-32.037],[-42.173,-40.729,-30.396],[-41.513,-42.476,-27.734],
[-43.559,-47.651,-22.333],[-41.064,-49.317,-36.239],[-42.790,-49.896,-32.869],
[-42.349,-48.153,-29.645],[-42.349,-46.289,-27.041],[-40.717,-51.332,-38.377],
[-43.085,-51.826,-40.213],[-42.050,-51.856,-31.240],[-58.298,-49.623,-29.414],
[-41.936,-51.803,-24.257],[-39.950,-51.276,-21.872],[-42.534,-54.113,-41.773],
[-43.498,-54.723,-35.102],[-42.059,-55.327,-31.917],[-44.808,-53.926,-28.342],
[-58.849,-53.597,-22.981],[-39.942,58.344,-23.282],[-43.163,-57.397,-35.321],
[-56.800,-55.842,-30.410],[-58.112,-53.639,-29.057],[-40.394,-58.447,-25.002],
[-42.540,-56.192,-39.580],[-58.793,-57.212,-34.886],[-44.597,-42.345,-34.643],
[-57.312,-58.793,-28.065],[-57.713,-45.691,-22.762],[-54.178,-44.939,-27.488],
[-47.879,-43.691,-22.890],[-46.156,-43.211,-25.004],[-42.895,-59.327,-26.528],
[-55.749,-40.740,-26.273],[-53.156,-41.555,-27.215],[-48.802,-42.085,-29.071],
[-45.811,-44.199,-29.580],[-45.675,-40.961,-26.995],[-58.097,-42.142,-31.050],
[-55.453,-43.874,-31.903],[-53.245,-42.116,-31.112],[-49.636,-57.813,-32.146],
[-42.063,59.060,-29.880],[-41.970,-58.750,-33.363],[-57.437,59.327,-33.930],
[-58.977,-57.843,-37.617],[-49.574,-58.839,-35.447],[-46.888,-59.097,-31.364],
[-47.444,-57.573,-34.046],[-57.303,-43.678,-36.879],[-54.283,-57.393,-40.350],
[-50.742,-57.491,-38.106],[-48.123,-41.405,-38.028],[-44.037,-42.060,-37.534],
[-57.333,-57.933,-24.635],[-54.187,-58.906,-22.504],[-55.827,-41.428,-38.649],
[-45.296,-59.991,-24.316],[-43.009,-45.105,-24.390],[-47.872,-30.554,-34.315],
[-48.200,-31.960,-29.865],[-56.135,-31.644,-28.522],[-45.109,-28.566,-31.826],
[-56.913,-29.124,-30.111],[-58.062,-32.147,-35.255],[-57.140,-32.833,-31.558],
[-48.750,-23.360,-31.178],[-49.567,-30.011,-31.449],[-41.447,-38.300,-29.068],
[-46.326,-35.267,-28.245],[-54.843,-37.472,-32.614],[-45.889,-37.739,-34.806],
[-57.145,-35.464,-31.972],[-51.638,-33.089,-32.281],[-43.122,-35.043,-28.997],
[-53.885,-36.844,-29.407],[-46.170,-28.464,-34.436],[-53.310,-29.508,-27.874],
[-44.845,-37.452,-32.408],[-54.333,-34.208,-30.874],[-55.428,-27.789,-27.758],
[-58.983,-27.726,-36.406],[-44.888,-27.655,-28.642],[-44.246,-37.401,-27.703],
[-51.169,-39.767,-31.146],[-45.573,-25.770,-30.954],[-44.016,-26.308,-33.788],
[-44.098,-28.838,-24.962],[-49.250,-30.215,-37.957],[-49.453,-32.835,-39.432],
[-41.240,-33.682,-27.656],[-46.198,-34.063,-36.271],[-46.480,-33.725,-25.333],
[-53.770,-31.747,-32.988],[-52.996,-27.648,-30.234],[-51.667,-30.427,-29.862],
[-47.685,-26.170,-28.210],[-51.973,-28.306,-25.411],[-46.826,-33.368,-40.191],
[-55.050,-31.963,-25.943],[-48.700,-26.486,-34.624],[-53.641,-25.900,-25.623],
[-41.287,-39.720,-38.096],[-55.082,-32.432,-35.763],[-40.764,-26.182,-38.740],
[-58.840,-30.744,-25.904],[58.396,-26.757,-29.911],[-39.916,-25.403,-22.995],
[-49.602,-27.209,-24.424],[-49.943,-24.098,-34.578],[-54.805,-23.626,-33.589],
[-45.235,-23.574,-39.128],[-45.260,-24.741,-36.176],[-43.665,-34.172,-24.379],
[-57.368,-28.216,-25.479],[-53.718,-39.725,-29.112],[-48.200,-40.191,-26.521],
[-45.472,-36.237,-25.180],[-44.750,-31.203,-39.080],[-49.641,-37.404,-38.256],
[-52.049,-34.323,-28.611],[-54.006,-34.628,-33.579],[-48.817,-33.573,-35.108],
[-48.398,-34.879,-23.131],[-46.449,-33.999,-33.511],[-55.877,-35.544,-28.018],
[-51.694,-35.926,-40.718],[-58.739,-39.195,-32.203],[-57.574,-36.212,-36.979],
[-48.986,-34.609,-28.886],[-52.176,-36.909,-32.997],[-56.191,-35.861,-34.722],
[-52.794,-36.672,-35.672],[-46.610,-39.026,-28.774],[-54.050,-33.550,-38.220],
[-43.169,-20.486,-22.352],[-55.647,-30.235,-23.577],[-50.411,-35.155,-31.051],
[-55.027,-37.740,-37.138],[-50.285,-37.559,-28.714],[-50.829,-27.876,-37.108],
[-43.746,-30.280,-27.558],[-46.689,-37.413,-37.750],[-48.375,-29.736,-25.559],
[-44.835,-34.474,-38.818],[-54.269,-30.290,-30.855],[-48.108,-27.935,-37.091],
[-57.261,-33.524,-26.547],[-51.484,-34.207,-22.345],[-46.677,-27.974,-25.134],
[-40.430,-34.170,-37.649],[-41.612,-27.836,-23.734],[-41.340,-27.049,-35.540],
[-44.844,-32.258,-29.258],[-53.344,-26.420,-33.383],[-58.198,-30.561,-37.666],
[-52.961,-29.030,-34.071],[-51.670,-24.442,-27.040],[-52.472,-24.656,-35.269],
[-46.948,-29.011,-27.767],[-46.315,-28.211,-39.997],[-45.331,-25.982,-26.193],
[-41.586,-24.064,-36.971],[-57.301,-37.196,-25.303],[-46.145,-29.861,-36.999],
[-57.729,-26.485,-28.766],[-51.516,-28.624,-40.115],[-53.363,-27.317,-36.301],
[-55.307,-29.355,-37.192],[-48.168,-37.173,-26.952],[-48.299,-39.246,-36.396],
[-48.153,-24.829,-36.500],[-48.419,-35.995,-34.203],[-50.196,-34.735,-25.241],
[-49.102,-39.725,-33.443],[-51.633,-23.953,-37.763],[-52.169,-35.749,-38.084],
[-43.990,-38.575,-25.198],[-46.918,-34.769,-30.861],[-55.764,-26.475,-37.319],
[-56.629,-34.928,-40.832],[-43.462,-25.290,-29.289],[-56.066,-40.069,-30.346],
[-57.405,-25.416,-34.800],[-58.454,-22.468,-23.648],[-52.885,-24.936,-22.008],
[-49.598,-26.875,-40.552],[-47.291,-27.544,-22.362],[-44.497,-26.970,-38.542],
[-57.410,-31.223,-21.684],[-53.473,-30.864,-40.547],[-50.779,-30.950,-25.281],
[-49.009,-29.526,-41.134],[-46.004,-30.933,-25.245],[-55.985,-33.309,-23.409],
[-52.169,-31.065,-35.701],[-52.627,-33.425,-26.355],[-52.016,-33.274,-40.211],
[-48.378,-31.057,-23.176],[-41.802,-30.008,-40.562],[-52.579,-35.920,-24.556],
[-52.325,-38.135,-27.157],[-49.090,-32.573,-26.755],[-44.915,-34.963,-22.060],
[-45.996,-32.354,-22.930],[-54.506,-34.907,-21.650],[-50.639,-40.591,-27.353],
[-50.269,-36.958,-22.736],[-48.066,-37.431,-24.267],[-45.434,-38.231,-22.847],
[-54.890,-37.226,-23.765],[-52.498,-41.397,-24.313],[-50.906,-40.042,-22.390],
[-54.015,-22.403,-37.851],[-57.443,-27.592,-39.785],[58.984,-22.305,-32.723],
[59.203,-24.839,-33.310],[-41.560,-23.418,-29.045],[-41.309,-23.491,-24.617],
[-43.524,-27.589,-20.386],[-42.345,-28.959,-37.345],[-41.915,-32.429,-35.568],
[-41.207,-28.096,-27.910],[-42.990,-24.592,-26.475],[-41.880,-31.881,-38.309],
[-42.112,-35.325,-39.488],[-43.756,-34.095,-34.167],[-59.157,-31.463,-29.986],
[-42.993,-31.807,-25.286],[-41.840,-32.429,-23.027],[-43.958,-37.008,-40.065],
[-44.120,-36.502,-37.191],[-42.245,-36.122,-31.555],[-44.411,-34.013,-26.840],
[-58.548,-34.337,-24.462],[-41.942,-41.249,-21.108],[-42.684,-38.052,-35.317],
[-57.196,-37.213,-29.722],[-58.419,-34.405,-29.340],[-41.709,-38.348,-23.897],
[-41.519,-38.104,-40.520],[-58.958,-37.470,-34.957],[-44.284,-23.689,-32.673],
[-57.512,-39.201,-27.607],[-59.204,-26.590,-24.128],[-54.838,-25.209,-28.125],
[-49.456,-24.182,-24.585],[-46.661,-23.899,-24.403],[-41.739,-39.988,-26.029],
[-57.306,-23.731,-27.982],[-54.204,-22.144,-26.042],[-50.021,-22.614,-28.395],
[-46.960,-22.983,-28.607],[-44.247,-22.413,-27.727],[-56.224,-23.402,-30.527],
[-56.202,-26.078,-31.518],[-53.326,-24.510,-31.054],[-49.311,-37.442,-31.799],
[-44.601,-39.877,-31.266],[-43.136,-40.305,-33.756],[-56.694,-39.715,-33.963],
[-57.761,-39.352,-38.107],[-51.849,-39.361,-33.810],[-47.378,-39.409,-31.329],
[-46.048,-40.608,-35.840],[-56.419,-23.864,-37.732],[-56.337,-37.750,-39.545],
[-50.258,-37.466,-35.545],[-47.709,-24.166,-40.355],[-42.942,-22.914,-35.104],
[-56.825,-39.375,-23.042],[-54.325,-39.977,-22.840],[-54.868,-24.068,-40.102],
[-47.070,-40.753,-22.034],[-42.732,-25.066,-22.987],[-47.365,-13.136,-33.312],
[-48.362,-13.093,-29.317],[-55.521,-12.259,-30.009],[-44.370,-10.644,-32.268],
[-54.802,-9.557,-29.534],[-59.017,-12.162,-34.878],[-57.832,-12.567,-32.345],
[-47.201,-5.983,-30.625],[-50.213,-11.275,-31.058],[-42.934,-19.469,-28.566],
[-45.460,-17.227,-30.781],[-54.611,-18.951,-32.262],[-45.722,-20.222,-34.499],
[-57.844,-15.702,-32.656],[-51.260,-15.463,-32.067],[-40.265,-16.196,-27.081],
[-55.325,-18.235,-28.660],[-49.199,-9.429,-35.105],[-53.906,-11.811,-27.483],
[-44.597,-19.984,-31.808],[-55.145,-16.680,-30.678],[-52.138,-9.701,-28.388],
[-57.841,-9.661,-36.776],[-46.613,-10.366,-30.661],[-44.748,-19.179,-26.541],
[-51.244,-21.060,-30.252],[-44.769,-7.677,-31.933],[-45.959,-8.745,-34.575],
[-43.783,-9.944,-25.840],[-49.596,-11.469,-37.590],[-48.342,-14.298,-39.320],
[-40.731,-13.131,-27.864],[-44.990,-12.634,-34.351],[-43.500,-13.898,-27.067],
[-53.860,-12.373,-32.954],[-53.316,-9.360,-32.178],[-52.515,-13.167,-29.464],
[-47.043,-7.989,-28.592],[-53.264,-9.549,-24.589],[-44.864,-14.932,-38.490],
[-54.188,-13.639,-24.911],[-51.658,-5.796,-33.997],[-55.578,-6.206,-25.948],
[-41.777,-21.242,-37.960],[-53.783,-15.084,-36.327],[-42.336,-8.468,-38.823],
[-58.818,-10.270,-27.205],[57.783,-8.229,-28.849],[-40.547,-7.567,-25.268],
[-49.760,-7.683,-25.937],[-48.529,-4.209,-34.112],[-57.215,-5.547,-32.788],
[-46.038,-4.961,-38.499],[-47.453,-6.379,-35.161],[-43.888,-16.472,-23.249],
[-55.875,-9.582,-26.284],[-52.149,-17.712,-30.634],[-48.216,-21.498,-26.340],
[-46.114,-17.281,-25.050],[-43.899,-12.624,-40.045],[-49.845,-18.913,-38.891],
[-52.012,-15.688,-28.880],[-52.741,-17.264,-33.734],[-48.730,-15.194,-34.756],
[-48.984,-14.740,-22.672],[-47.749,-17.125,-32.738],[-55.283,-14.894,-27.482],
[-53.158,-18.168,-40.418],[59.008,-16.568,-33.973],[59.377,-18.644,-36.463],
[-47.684,-15.610,-27.871],[-51.798,-19.506,-32.586],[-55.424,-16.343,-33.585],
[-52.317,-17.471,-37.427],[-47.721,-18.623,-29.164],[-54.732,-16.656,-38.633],
[-43.592,-2.661,-22.787],[-55.812,-11.086,-23.103],[-48.492,-15.668,-30.413],
[-54.982,-19.116,-35.753],[-49.738,-17.301,-28.138],[-51.879,-9.678,-36.413],
[-42.858,-12.754,-30.178],[-46.683,-18.917,-40.897],[-49.174,-10.337,-24.861],
[-45.684,-17.451,-37.402],[-55.788,-10.132,-32.797],[-47.480,-9.323,-37.367],
[-56.530,-12.053,-25.959],[-51.714,-14.778,-24.302],[-46.904,-9.934,-26.013],
[-41.702,-15.477,-37.221],[-41.738,-9.496,-23.114],[-42.684,-7.960,-36.057],
[-45.470,-12.530,-28.891],[-54.848,-8.293,-35.146],[-58.799,-13.748,-39.425],
[-51.814,-10.870,-33.794],[-51.334,-6.568,-27.610],[-54.821,-4.171,-35.650],
[-44.420,-9.864,-28.960],[-45.188,-10.149,-37.960],[-44.223,-7.678,-24.561],
[-42.500,-5.925,-37.877],[-57.252,-19.390,-27.206],[-47.193,-11.499,-35.654],
[-56.308,-7.583,-28.352],[-51.641,-8.452,-40.228],[-49.241,-9.374,-39.558],
[-54.887,-11.740,-36.826],[-49.816,-19.326,-26.346],[-47.929,-19.010,-37.021],
[-49.449,-7.063,-36.858],[-50.154,-16.942,-36.018],[-51.076,-16.763,-26.059],
[-48.718,-19.096,-34.244],[-51.738,-5.791,-36.941],[-51.893,-20.353,-37.537],
[-41.586,-18.697,-26.145],[-45.128,-14.241,-32.016],[-54.238,-8.113,-37.839],
[-57.331,-16.761,-38.699],[-44.405,-6.262,-29.175],[-56.538,-20.837,-29.808],
[-58.076,-7.233,-34.919],[-58.110,-3.961,-23.562],[-52.235,-4.428,-21.776],
[-49.419,-6.715,-39.814],[-46.425,-9.410,-23.148],[-44.904,-7.683,-39.617],
[-58.451,-11.307,-23.516],[-54.107,-10.246,-39.874],[-51.378,-11.093,-23.406],
[-49.163,-11.650,-40.846],[-46.865,-13.087,-26.683],[-57.183,-13.734,-24.011],
[-52.105,-12.506,-36.260],[-54.661,-17.649,-26.112],[-51.253,-14.597,-39.777],
[-46.766,-12.326,-24.158],[-41.472,-12.740,-41.066],[-54.258,-18.777,-23.736],
[-52.657,-18.485,-27.769],[-49.915,-14.278,-26.948],[-44.617,-15.933,-20.183],
[-44.586,-14.299,-24.482],[-55.696,-16.594,-23.610],[-52.419,-21.388,-27.684],
[-51.058,-17.651,-23.450],[-48.599,-17.609,-24.144],[-45.650,-18.586,-22.200],
[-57.610,-18.487,-22.921],[-52.102,-23.543,-24.385],[-51.034,-22.172,-22.200],
[-51.415,-3.527,-39.312],[-58.009,-8.534,-40.578],[-58.257,-3.905,-35.086],
[59.020,-5.556,-31.265],[-40.524,-3.579,-28.608],[-41.106,-5.023,-25.826],
[-42.757,-9.394,-20.528],[-42.634,-11.043,-36.847],[-42.362,-13.092,-35.077],
[-41.233,-9.491,-29.010],[-42.838,-8.166,-27.371],[-40.978,-12.894,-38.376],
[-41.487,-16.716,-39.779],[-43.524,-15.410,-34.124],[-59.197,-10.424,-30.212],
[-41.851,-12.305,-25.346],[-41.573,-14.315,-23.331],[-43.736,-18.922,-40.352],
[-46.015,-17.242,-34.848],[-42.629,-16.786,-31.652],[-43.083,-16.586,-27.845],
[-58.595,-16.192,-23.450],[-39.894,-20.806,-21.704],[-42.553,-17.812,-35.813],
[-57.645,-18.481,-32.586],[-57.964,-14.956,-29.782],[-41.080,-20.085,-23.935],
[-41.255,-19.758,-40.531],[-58.752,-20.828,-34.673],[-44.526,-3.871,-33.419],
[-59.405,-24.429,-29.859],[-56.324,-6.792,-22.253],[-53.801,-7.887,-26.783],
[-48.925,-5.362,-23.285],[-46.349,-5.757,-24.618],[-42.766,-21.409,-25.687],
[-56.350,-4.431,-27.765],[-54.308,-3.612,-25.997],[-50.118,-4.519,-29.553],
[-46.942,-5.515,-27.123],[-44.656,-3.636,-27.441],[-57.810,-5.762,-29.587],
[-54.724,-6.851,-31.949],[-52.142,-7.500,-30.515],[-49.129,-19.808,-31.505],
[-43.906,-22.291,-30.365],[-41.936,-21.871,-32.754],[-56.119,-21.286,-32.533],
[-57.429,-21.228,-37.413],[-50.705,-21.647,-33.935],[-46.600,-21.873,-31.412],
[-45.598,-21.610,-37.280],[-57.305,-7.315,-37.738],[-56.045,-19.552,-39.585],
[-51.673,-19.264,-35.243],[-49.390,-4.623,-37.846],[-42.487,-5.272,-35.100],
[-56.215,-20.710,-24.979],[-54.167,-22.661,-22.914],[-55.548,-4.193,-40.041],
[-44.997,-22.799,-21.755],[-40.401,-7.404,-22.332],[-46.896,5.083,-34.512],
[-48.777,4.817,-29.976],[-55.937,4.669,-31.139],[-45.730,8.764,-32.228],
[-55.584,9.923,-31.014],[-56.665,3.849,-35.967],[-57.970,5.569,-33.163],
[-48.426,13.550,-31.308],[-49.345,7.992,-31.264],[-40.230,0.112,-28.790],
[-48.240,0.739,-28.490],[-54.606,-0.122,-32.903],[-45.872,-1.080,-33.344],
[-56.857,2.601,-32.880],[-52.682,2.133,-31.751],[-40.903,2.779,-27.001],
[-54.874,-0.240,-29.420],[-48.734,7.873,-33.896],[-54.117,6.548,-28.106],
[-42.032,-0.015,-31.217],[-55.127,1.769,-30.965],[-55.150,9.025,-27.201],
[-57.375,8.794,-37.382],[-46.251,8.932,-29.276],[-42.478,-1.833,-28.390],
[-52.119,-3.532,-31.692],[-46.167,11.584,-31.198],[-45.416,10.551,-34.637],
[-44.013,9.038,-25.475],[-48.536,6.487,-36.268],[-47.517,3.739,-40.077],
[-41.074,5.205,-28.140],[-43.810,5.219,-35.910],[-45.774,4.153,-26.024],
[-54.524,5.272,-33.554],[-52.743,9.179,-29.479],[-52.989,5.436,-30.872],
[-48.998,10.785,-30.031],[-52.660,10.633,-25.722],[-46.303,5.129,-37.670],
[-54.465,4.295,-26.798],[-49.528,13.347,-33.681],[-53.202,12.888,-24.713],
[-42.358,-3.194,-37.886],[-55.334,5.124,-37.834],[-41.099,10.570,-39.332],
[-57.941,7.549,-27.984],[-58.499,11.286,-30.627],[-39.952,11.622,-24.487],
[-49.542,10.206,-26.674],[-49.632,14.632,-35.871],[-55.283,14.862,-33.626],
[-45.411,12.508,-38.822],[-46.470,14.094,-36.837],[-43.063,4.028,-25.371],
[-57.991,9.110,-25.695],[-54.687,-2.687,-30.522],[-48.257,-3.419,-26.475],
[-45.592,1.221,-25.150],[-44.892,5.352,-40.553],[-49.406,0.065,-38.193],
[-53.123,1.837,-28.954],[-54.242,2.658,-34.224],[-48.265,3.090,-36.825],
[-47.537,3.439,-24.212],[-46.357,2.766,-32.136],[-55.179,1.495,-27.012],
[-53.379,1.222,-40.292],[-57.593,-0.092,-33.999],[-58.924,2.475,-37.967],
[-47.772,3.235,-27.605],[-51.468,-0.507,-32.340],[-55.852,1.254,-35.873],
[-53.313,1.570,-36.683],[-45.819,-0.796,-29.496],[-53.203,3.700,-38.120],
[-44.848,13.071,-24.422],[-55.522,7.140,-24.358],[-49.074,2.292,-31.741],
[-54.472,-1.292,-35.561],[-50.583,2.541,-27.879],[-52.377,6.343,-36.233],
[-43.495,6.529,-28.557],[-46.832,-0.469,-39.083],[-50.462,6.738,-26.692],
[-46.082,1.879,-38.066],[-55.721,7.389,-31.878],[-48.850,9.703,-37.740],
[-57.318,4.812,-28.099],[-49.877,5.940,-24.140],[-48.301,8.033,-25.527],
[-41.044,3.688,-37.038],[-41.223,8.761,-25.874],[-42.382,10.510,-35.691],
[-45.734,6.339,-30.927],[-53.527,10.901,-36.030],[-56.883,4.903,-40.122],
[-51.420,6.771,-33.477],[-51.250,12.520,-27.022],[-52.217,12.823,-33.528],
[-49.286,8.428,-28.717],[-46.599,9.421,-39.287],[-46.909,11.363,-26.944],
[-40.914,13.407,-39.033],[-56.466,-0.777,-25.885],[-44.607,9.801,-37.299],
[-57.236,10.480,-28.408],[-50.517,10.545,-39.803],[-52.207,9.252,-38.003],
[-55.872,7.433,-39.097],[-49.823,-1.582,-28.652],[-47.005,-0.114,-36.128],
[-49.276,12.468,-37.713],[-49.818,0.992,-35.040],[-50.113,3.030,-25.401],
[-48.658,-1.631,-33.640],[-52.073,13.413,-36.277],[-53.906,-0.952,-38.056],
[-43.271,1.451,-26.447],[-43.776,3.310,-30.986],[-54.586,10.523,-38.381],
[-57.514,0.915,-40.368],[-42.407,13.138,-28.413],[-57.574,-3.154,-29.965],
[-57.806,10.549,-33.621],[-57.845,12.570,-25.657],[-53.953,11.715,-22.212],
[-49.299,10.722,-42.372],[-46.546,9.635,-22.072],[-43.738,10.924,-40.552],
[-59.299,5.855,-23.094],[-53.253,7.712,-40.074],[-52.129,8.175,-24.584],
[-49.237,8.298,-40.951],[-45.287,6.659,-24.866],[-57.368,5.206,-25.384],
[-50.958,4.031,-36.762],[-52.700,3.481,-24.933],[-51.645,4.700,-40.290],
[-47.507,6.615,-23.215],[-41.752,7.301,-39.847],[-52.226,0.902,-25.556],
[-52.324,-0.538,-28.090],[-51.278,5.026,-28.943],[-46.064,2.077,-22.376],
[-45.541,4.595,-23.068],[-50.761,-1.126,-23.865],[-51.265,-2.939,-26.625],
[-50.606,1.332,-22.049],[-48.330,0.454,-23.793],[-46.106,-0.881,-22.422],
[-54.777,0.133,-23.560],[-52.022,-5.110,-25.139],[-49.904,-3.051,-22.022],
[-51.501,15.310,-38.181],[-58.333,9.309,-40.162],[-58.761,15.385,-34.041],
[59.482,13.436,-32.089],[-41.555,13.686,-31.814],[-41.935,13.729,-25.136],
[-43.574,9.127,-22.347],[-42.524,7.780,-36.731],[-42.017,4.467,-33.477],
[-43.265,9.417,-30.695],[-42.961,9.410,-27.863],[-42.587,5.426,-38.229],
[-42.500,1.782,-38.175],[-44.712,4.663,-33.514],[-58.172,6.346,-30.510],
[-42.550,6.713,-24.079],[-41.465,4.533,-22.819],[-43.786,-0.160,-42.091],
[-44.422,0.883,-36.436],[-42.722,1.401,-29.145],[-43.786,3.844,-28.104],
[-57.859,0.705,-23.952],[-40.094,-2.471,-21.727],[-43.116,-1.576,-35.070],
[59.660,-0.854,-29.857],[-57.874,3.037,-30.110],[-43.472,-0.382,-24.263],
[-42.381,-0.857,-39.778],[-56.673,-1.063,-37.256],[-44.573,13.919,-34.070],
[-57.920,-2.149,-27.391],[-56.370,10.085,-23.343],[-54.254,12.421,-27.072],
[-49.125,12.322,-24.170],[-47.352,14.777,-24.760],[-42.850,-2.724,-25.587],
[-56.067,14.126,-28.148],[-53.279,13.827,-29.425],[-48.920,15.859,-29.992],
[-47.081,13.316,-28.902],[-45.415,15.715,-29.481],[-56.069,12.536,-31.896],
[-54.946,10.881,-33.642],[-53.096,11.315,-31.258],[-48.504,-2.156,-30.991],
[-42.454,-4.095,-30.939],[-41.544,-2.395,-32.955],[-57.079,-2.715,-33.062],
[-57.458,-4.384,-37.710],[-51.285,-3.104,-34.480],[-45.927,-3.485,-31.033],
[-44.908,-3.765,-36.308],[-58.732,14.526,-38.655],[-55.774,-1.020,-39.894],
[-48.965,-2.293,-36.472],[-48.607,15.524,-38.389],[-43.004,12.968,-36.929],
[-55.676,-2.713,-23.590],[-52.682,-1.784,-22.486],[-55.317,12.084,-40.347],
[-46.041,-3.990,-22.118],[-42.690,11.438,-23.964],[-47.772,24.067,-32.652],
[-48.308,24.750,-30.165],[-55.337,22.958,-28.705],[-44.977,27.724,-31.980],
[-54.721,28.715,-31.311],[-58.102,21.719,-35.116],[-57.097,24.300,-32.635],
[-46.465,31.878,-31.871],[-47.596,27.104,-31.414],[-41.471,17.746,-29.099],
[-46.669,20.020,-30.972],[-53.591,19.268,-32.454],[-46.251,18.691,-34.502],
[-57.526,22.332,-31.265],[-52.373,22.299,-31.992],[-40.006,20.737,-27.088],
[-54.538,18.788,-28.827],[-48.558,28.417,-35.192],[-52.439,26.113,-26.390],
[-43.132,18.521,-32.677],[-56.062,19.788,-30.634],[-54.597,27.350,-27.700],
[-59.318,26.090,-36.119],[-44.868,27.275,-28.681],[-44.630,18.482,-29.376],
[-51.321,14.034,-31.305],[-44.328,30.268,-32.930],[-45.822,29.087,-34.829],
[-43.638,27.099,-26.155],[-49.384,25.371,-35.360],[-47.995,22.993,-39.846],
[-41.196,23.417,-27.648],[-45.258,24.833,-33.872],[-45.844,22.737,-26.809],
[-54.007,24.784,-33.517],[-51.824,28.555,-31.767],[-51.513,23.208,-29.540],
[-46.853,29.248,-28.635],[-51.568,29.330,-24.906],[-47.308,24.968,-37.616],
[-55.090,25.164,-26.392],[-48.193,30.854,-33.694],[-53.543,31.440,-25.347],
[-42.218,16.881,-38.947],[-54.503,23.749,-36.042],[-40.664,31.368,-39.912],
[-58.791,27.650,-27.702],[-58.792,29.170,-29.787],[-39.250,30.175,-24.636],
[-48.969,29.184,-27.086],[-49.536,33.070,-36.366],[-55.412,30.262,-35.938],
[-45.547,32.447,-39.386],[-45.917,31.316,-36.632],[-42.409,21.489,-25.815],
[-57.216,28.489,-25.619],[-52.597,16.764,-29.466],[-48.355,15.256,-27.196],
[-45.527,19.859,-25.916],[-43.814,25.713,-39.217],[-48.176,17.472,-40.069],
[-53.502,21.018,-29.880],[-55.419,22.166,-33.772],[-48.629,22.762,-36.113],
[-51.562,21.798,-24.645],[-48.550,19.586,-33.108],[-55.225,20.849,-27.080],
[-50.038,19.460,-40.575],[58.017,15.678,-31.794],[-57.698,19.835,-37.766],
[-48.130,22.287,-28.740],[-51.464,17.985,-33.509],[-55.278,19.716,-34.693],
[-52.336,20.227,-34.990],[-46.826,18.199,-28.105],[-53.763,20.632,-37.955],
[-42.724,34.643,-23.883],[-55.666,28.419,-22.785],[-49.203,21.744,-31.239],
[-54.085,17.314,-34.735],[-50.427,20.493,-28.949],[-50.543,27.210,-36.725],
[-42.987,25.265,-28.479],[-45.799,18.897,-38.696],[-49.792,26.634,-26.989],
[-46.273,21.166,-36.471],[-54.726,25.627,-29.857],[-46.635,28.015,-37.135],
[-58.017,25.439,-26.469],[-50.612,24.190,-24.641],[-46.997,30.614,-25.790],
[-38.662,22.561,-36.957],[-40.808,26.458,-24.658],[-41.596,29.566,-37.698],
[-45.383,22.321,-29.359],[-51.795,29.274,-34.612],[-55.314,23.953,-38.568],
[-51.334,26.562,-33.959],[-51.085,31.323,-27.122],[-53.051,31.760,-34.508],
[-45.722,25.592,-26.139],[-46.547,26.711,-39.651],[-44.290,31.103,-27.121],
[-42.417,32.446,-37.815],[-56.504,18.576,-25.225],[-44.979,25.994,-36.430],
[-56.290,29.243,-28.408],[-50.729,28.645,-40.780],[-52.347,29.065,-38.703],
[-54.189,26.434,-37.809],[-50.746,17.278,-26.360],[-46.724,16.753,-36.975],
[-50.160,29.787,-37.137],[-49.513,19.971,-35.747],[-48.758,21.845,-25.849],
[-49.040,16.571,-34.318],[-53.649,31.396,-37.525],[-53.076,17.748,-38.017],
[-42.166,18.661,-25.648],[-45.626,22.722,-31.909],[-54.017,28.169,-40.654],
[-56.880,20.547,-40.508],[-42.619,32.578,-28.768],[-55.057,15.616,-30.947],
[-57.945,28.916,-35.583],[-58.473,33.400,-24.969],[-51.635,31.323,-22.937],
[-48.347,29.841,-41.629],[-46.078,27.338,-23.758],[-45.753,29.892,-40.244],
[-58.539,26.017,-23.726],[-52.643,24.290,-39.813],[-52.235,25.931,-22.526],
[-49.139,27.184,-39.486],[-46.870,23.803,-24.499],[-56.738,23.823,-24.551],
[-52.027,24.971,-37.072],[-53.881,22.664,-25.312],[-50.921,22.185,-39.416],
[-48.379,26.129,-23.350],[-41.919,27.456,-41.347],[-52.579,20.217,-27.172],
[-53.894,18.351,-25.960],[-49.768,23.882,-27.659],[-45.102,20.006,-22.413],
[-45.004,22.714,-22.312],[-53.122,22.384,-22.459],[-50.866,14.903,-28.184],
[-51.990,19.777,-22.612],[-49.172,19.269,-24.396],[-47.214,18.279,-22.738],
[-54.994,19.314,-22.989],[-51.141,14.450,-25.031],[-51.245,15.709,-22.501],
[-49.666,34.870,-38.765],[-59.166,27.464,-39.377],[-56.665,33.240,-35.645],
[-58.888,31.051,-33.630],[-40.740,32.618,-30.848],[-41.626,31.212,-26.177],
[-42.560,28.164,-23.269],[-42.214,24.444,-36.873],[-41.652,25.253,-34.458],
[-41.246,27.437,-29.840],[-42.495,29.382,-28.252],[-40.607,24.106,-39.433],
[-40.933,21.393,-39.052],[-42.823,22.422,-34.595],[-57.457,27.150,-31.541],
[-42.643,24.562,-25.385],[-40.862,23.264,-23.457],[-43.491,19.594,-40.351],
[-43.712,20.141,-37.522],[-43.962,20.663,-31.140],[-42.622,20.330,-28.346],
[-57.475,21.294,-23.407],[-39.876,16.054,-21.943],[-42.502,19.028,-35.532],
[-57.765,17.164,-32.097],[-57.735,21.799,-28.689],[-42.198,15.791,-23.298],
[-40.795,18.820,-40.191],[-58.778,18.852,-34.625],[-42.812,32.287,-32.430],
[-57.880,16.468,-26.490],[-58.411,30.618,-22.459],[-53.370,29.819,-27.709],
[-48.859,32.174,-24.614],[-46.546,33.673,-24.863],[-42.481,15.828,-27.256],
[-57.968,31.733,-28.291],[-54.888,32.586,-27.800],[-48.041,34.136,-29.221],
[-45.461,31.979,-29.389],[-44.847,34.096,-27.743],[-57.311,31.302,-31.492],
[-52.978,30.931,-30.161],[-50.028,31.589,-31.751],[-47.597,17.395,-31.746],
[-44.289,13.525,-30.453],[-43.647,15.908,-32.154],[-56.728,16.923,-34.973],
[-58.221,16.836,-37.418],[-52.230,15.566,-34.098],[-46.255,14.874,-32.100],
[-43.994,15.918,-37.007],[-57.575,29.704,-38.194],[-55.234,17.176,-40.232],
[-49.661,17.658,-37.279],[-48.184,31.410,-37.997],[-43.541,31.994,-34.946],
[-56.533,16.180,-23.630],[-54.149,14.723,-22.626],[-55.545,31.402,-39.924],
[-45.408,16.156,-22.587],[-42.965,31.972,-23.387],[-49.146,42.642,-33.215],
[-48.551,42.693,-30.538],[-55.296,41.978,-29.002],[-44.809,45.808,-32.512],
[-54.868,46.935,-29.898],[-57.986,40.019,-35.874],[-58.407,42.454,-33.623],
[-48.571,49.965,-32.401],[-49.860,47.557,-28.086],[-41.159,36.889,-30.476],
[-45.717,39.578,-29.999],[-53.918,37.552,-33.329],[-45.447,37.825,-34.271],
[-57.586,40.249,-32.174],[-52.335,42.030,-32.571],[-41.662,40.637,-27.588],
[-55.823,36.936,-29.227],[-47.362,45.689,-34.622],[-52.283,43.507,-26.740],
[-42.834,37.684,-32.617],[-55.647,39.547,-30.366],[-53.914,44.604,-29.128],
[-58.579,45.958,-36.752],[-44.201,47.009,-29.918],[-44.354,36.887,-29.056],
[-52.422,33.124,-32.326],[-47.383,47.271,-32.305],[-45.118,47.959,-34.718],
[-42.939,45.142,-26.291],[-49.074,43.604,-35.815],[-47.997,40.520,-39.512],
[-40.132,42.873,-28.175],[-46.082,43.108,-35.298],[-45.435,41.173,-26.588],
[-54.542,44.445,-33.048],[-50.951,47.011,-32.307],[-51.775,42.631,-29.392],
[-48.162,48.476,-29.790],[-52.480,46.080,-26.011],[-44.856,43.284,-38.498],
[-55.085,43.550,-26.803],[-47.720,48.762,-35.184],[-54.609,49.929,-24.238],
[-42.314,35.194,-37.986],[-53.996,42.241,-35.277],[-41.582,48.852,-40.118],
[-59.013,45.731,-26.720],[-59.512,48.057,-29.776],[-40.342,50.248,-25.144],
[-50.580,48.615,-25.437],[-48.237,51.930,-36.938],[-55.477,51.616,-32.986],
[-43.973,50.815,-38.780],[-45.774,50.015,-36.562],[-42.640,42.565,-24.343],
[-55.785,46.305,-27.280],[-53.514,34.929,-29.516],[-49.184,34.758,-25.987],
[-45.436,38.396,-25.402],[-43.491,42.335,-40.626],[-47.978,37.218,-38.878],
[-53.039,40.108,-29.136],[-55.002,39.967,-33.537],[-48.430,40.752,-35.454],
[-49.279,41.670,-23.552],[-46.652,40.226,-33.741],[-55.271,38.840,-26.452],
[-52.452,37.889,-40.971],[59.732,37.266,-32.803],[-56.942,39.082,-39.271],
[-48.022,40.612,-28.340],[-52.456,35.943,-32.035],[-56.072,38.140,-35.263],
[-53.197,38.970,-35.826],[-46.973,36.698,-30.491],[-53.891,38.868,-38.967],
[-44.983,49.955,-23.947],[-56.053,44.754,-24.313],[-50.216,40.326,-30.947],
[-54.717,35.475,-35.645],[-50.417,38.828,-28.568],[-50.584,45.664,-36.330],
[-42.238,44.347,-29.234],[-46.138,38.757,-37.812],[-49.207,45.030,-26.263],
[-44.497,40.710,-38.594],[-56.108,44.330,-30.611],[-48.073,45.460,-37.381],
[-56.960,41.531,-26.547],[-52.108,41.714,-24.512],[-48.095,47.652,-25.564],
[-41.093,39.950,-37.482],[-40.444,43.921,-25.922],[-43.234,48.569,-37.194],
[-44.363,43.345,-30.233],[-52.021,47.572,-34.924],[-57.030,43.550,-40.190],
[-52.164,44.717,-34.148],[-50.761,50.011,-30.018],[-53.855,51.580,-36.126],
[-46.286,45.213,-28.253],[-45.959,45.634,-39.555],[-45.085,47.563,-27.344],
[-41.713,51.439,-41.001],[-56.719,35.900,-26.180],[-44.989,45.807,-36.941],
[-57.380,48.564,-27.727],[-53.392,45.274,-41.818],[-52.273,46.330,-38.347],
[-54.847,43.040,-37.797],[-47.122,38.082,-28.143],[-46.692,36.313,-36.659],
[-50.246,48.089,-37.111],[-50.489,38.685,-35.994],[-49.103,39.699,-26.164],
[-49.590,36.180,-34.927],[-53.444,50.162,-38.396],[-51.987,33.957,-39.901],
[-44.531,35.396,-25.577],[-45.157,41.537,-31.897],[-54.494,47.828,-37.213],
[-57.173,41.721,-37.980],[-44.323,49.630,-29.495],[-55.211,33.336,-31.080],
[-57.083,49.350,-33.753],[-57.667,49.470,-23.804],[-54.130,47.932,-21.548],
[-48.138,47.392,-39.537],[-46.947,48.278,-23.316],[-45.910,49.177,-41.658],
[-58.655,43.909,-23.574],[-55.063,45.531,-39.643],[-51.219,45.668,-23.743],
[-48.585,44.763,-40.376],[-45.048,43.889,-25.184],[-56.225,42.176,-23.673],
[-52.082,43.544,-38.296],[-52.569,39.401,-26.241],[-50.853,40.694,-38.512],
[-47.698,44.233,-23.974],[-42.254,44.708,-40.679],[-50.781,37.933,-24.929],
[-53.245,37.240,-28.137],[-49.564,42.517,-27.264],[-44.713,39.698,-22.901],
[-46.909,40.687,-24.555],[-54.174,39.993,-24.297],[-51.016,33.864,-27.931],
[-48.208,37.946,-22.396],[-47.915,37.436,-25.414],[-45.964,36.671,-23.404],
[-55.272,37.509,-22.935],[-51.398,33.341,-24.857],[-51.208,35.405,-22.451],
[-50.741,51.253,-37.941],[-57.469,47.868,-38.847],[58.279,51.745,-33.598],
[-58.085,50.381,-31.362],[-40.944,49.874,-30.892],[-41.510,50.168,-28.112],
[-42.369,46.675,-22.008],[-41.492,45.288,-36.809],[-41.618,41.581,-35.628],
[-40.567,46.315,-29.633],[-42.436,47.726,-27.870],[-42.013,42.620,-38.532],
[-42.100,40.103,-39.944],[-43.831,40.589,-34.281],[-58.575,43.317,-31.124],
[-41.604,45.036,-23.883],[-40.387,42.419,-22.586],[-44.039,37.572,-39.393],
[-44.016,38.816,-36.378],[-42.675,39.261,-30.157],[-43.277,38.210,-26.900],
[59.342,37.729,-25.170],[-39.760,35.055,-21.982],[-43.206,36.448,-35.000],
[-56.499,37.600,-32.216],[-58.037,39.128,-29.617],[-40.862,37.122,-23.936],
[-40.952,37.797,-41.180],[58.164,35.460,-36.039],[-44.663,52.248,-35.581],
[58.428,34.323,-28.817],[-57.797,46.839,-24.104],[-53.458,47.981,-27.607],
[-48.187,50.343,-24.439],[-45.101,51.258,-26.348],[-41.991,34.207,-26.640],
[-55.721,53.003,-26.653],[-54.106,50.898,-27.433],[-48.381,51.881,-30.073],
[-46.757,50.380,-28.520],[-43.774,52.357,-28.549],[-55.662,51.132,-30.062],
[-55.450,48.907,-31.633],[-53.163,49.457,-33.380],[-49.005,35.420,-31.624],
[-43.178,34.769,-30.369],[-41.736,34.610,-33.130],[-55.515,31.272,-33.572],
[-58.697,34.807,-38.295],[-52.124,35.108,-34.610],[-44.946,35.843,-32.437],
[-45.030,33.976,-36.531],[-56.872,49.545,-36.639],[-55.035,34.217,-38.867],
[-51.945,36.433,-37.848],[-47.965,49.783,-38.428],[-41.173,51.210,-37.967],
[-57.422,33.742,-21.960],[-53.635,34.390,-23.647],[-56.187,51.927,-39.834],
[-46.042,34.319,-22.253],[-42.007,49.180,-23.109],[-41.838,55.956,-30.697],
[-46.876,56.480,-29.249],[-53.209,56.278,-32.912],[-46.118,54.877,-33.658],
[-58.054,57.975,-31.364],[-41.148,55.528,-28.063],[-54.592,55.799,-30.715],
[-43.463,56.999,-33.749],[-55.589,57.903,-32.140],[-44.250,55.168,-28.652],
[-52.642,51.687,-31.625],[-42.648,53.627,-37.828],[-43.368,57.449,-23.064],
[-53.401,53.229,-29.316],[-47.932,52.303,-26.445],[-46.018,56.539,-23.417],
[-50.347,54.286,-38.819],[-54.382,58.266,-29.630],[-54.143,58.609,-34.235],
[-46.682,57.799,-33.212],[-54.426,-59.694,-27.284],[-51.469,56.195,-40.663],
[58.224,57.230,-32.887],[-57.927,59.011,-39.061],[-51.667,54.991,-34.388],
[-57.214,56.208,-33.177],[-52.724,57.743,-36.279],[-46.622,53.946,-29.885],
[-54.154,59.005,-38.755],[-48.512,57.619,-31.037],[-54.584,54.920,-35.925],
[-51.742,57.577,-28.796],[-46.040,56.474,-39.175],[-46.055,59.443,-37.410],
[-42.268,57.346,-41.167],[-58.014,54.999,-27.201],[-48.477,54.765,-27.600],
[-47.559,54.890,-37.363],[-47.994,56.718,-35.407],[-50.031,57.101,-26.497],
[-47.809,52.337,-34.369],[-52.481,55.602,-37.924],[-43.587,57.037,-26.460],
[-57.953,-59.934,-41.630],[-57.029,53.496,-31.072],[-53.053,55.321,-25.123],
[-51.407,54.999,-28.519],[-47.590,59.437,-22.194],[-54.810,57.325,-23.419],
[-50.819,52.535,-27.289],[-49.382,57.297,-23.522],[-46.673,57.130,-26.243],
[-45.804,53.741,-22.768],[-55.640,54.435,-23.718],[-51.659,50.758,-22.292],
[-52.535,52.686,-25.034],[-44.638,58.734,-39.988],[-44.004,54.950,-40.643],
[-45.024,57.118,-35.910],[-44.699,56.454,-31.304],[-44.680,58.636,-28.656],
[59.077,58.037,-23.451],[-42.578,53.308,-23.245],[-42.934,54.481,-35.181],
[-58.933,55.702,-30.909],[58.669,58.982,-27.688],[-41.410,54.687,-25.724],
[-39.248,54.275,-39.316],[-57.306,55.111,-36.054],[-58.795,52.501,-27.928],
[-42.424,51.813,-25.656],[-49.243,54.470,-30.460],[-43.480,50.598,-32.336],
[-42.394,53.293,-31.263],[-56.742,51.793,-35.185],[-56.881,52.616,-37.507],
[-51.183,51.488,-34.336],[-45.981,52.564,-32.504],[-45.913,52.788,-38.117],
[-53.631,53.645,-39.309],[-50.035,55.145,-36.346],[-58.143,54.072,-22.075],
[-54.458,52.294,-22.577],[-48.547,52.716,-23.073],[-47.744,-50.086,-14.949],
[-49.045,-50.037,-12.611],[-56.184,-51.753,-11.204],[-44.822,-47.389,-14.917],
[-55.824,-46.153,-13.002],[-58.078,-52.299,-16.592],[-57.150,-52.497,-13.919],
[-45.736,-43.470,-12.423],[-51.091,-48.832,-11.471],[-41.480,-56.611,-11.129],
[-47.248,-53.767,-11.692],[-53.759,-55.136,-14.670],[-45.807,-55.464,-16.269],
[-57.690,-54.858,-12.209],[-52.306,-53.101,-13.386],[-41.154,-54.408,-7.279],
[-56.075,-56.929,-9.889],[-48.726,-47.388,-14.282],[-53.267,-48.976,-9.853],
[-45.677,-56.676,-13.756],[-54.944,-54.211,-12.353],[-55.310,-47.271,-9.561],
[-58.643,-45.986,-16.749],[-44.818,-46.920,-11.625],[-44.476,-57.118,-9.497],
[-52.424,58.792,-12.890],[-43.129,-43.375,-13.718],[-46.586,-45.646,-15.756],
[-43.159,-48.099,-6.263],[-49.680,-49.801,-18.875],[-47.198,-52.587,-21.188],
[-40.758,-51.334,-9.306],[-45.394,-51.052,-16.831],[-46.698,-52.068,-7.497],
[-54.879,-48.759,-15.053],[-52.204,-45.834,-12.932],[-53.665,-50.515,-12.781],
[-46.832,-45.126,-10.679],[-52.291,-46.984,-6.354],[-46.922,-50.013,-19.528],
[-55.657,-49.594,-7.910],[-48.926,-44.656,-14.937],[-53.750,-43.595,-6.637],
[-42.342,-57.636,-18.411],[-55.095,-50.859,-16.833],[-43.109,-43.936,-20.904],
[-58.108,-47.655,-8.817],[-57.785,-46.244,-10.996],[-40.350,-45.573,-7.163],
[-49.538,-45.210,-6.891],[-48.989,-41.548,-17.644],[-55.140,-43.245,-16.952],
[-45.832,-43.934,-19.915],[-47.106,-43.470,-17.419],[-42.878,-52.837,-5.143],
[-56.693,-45.459,-7.226],[-53.697,-57.966,-11.102],[-50.063,-58.183,-8.409],
[-45.925,-54.115,-9.159],[-44.574,-51.426,-20.900],[-49.564,-55.251,-20.330],
[-53.151,-51.947,-10.640],[-54.344,-54.494,-17.350],[-48.022,-52.099,-16.665],
[-48.656,-52.680,-4.772],[-46.230,-52.839,-14.220],[-54.995,-53.861,-9.620],
[-52.993,-56.399,-20.644],[-59.642,-55.185,-14.487],[-58.023,-54.983,-19.473],
[-48.461,-51.902,-9.983],[-50.495,-55.023,-13.468],[-57.336,-55.214,-16.161],
[-53.118,-56.941,-18.031],[-46.615,-56.220,-11.003],[-53.592,-53.877,-20.231],
[-44.540,-41.151,-4.486],[-55.945,-47.726,-6.101],[-49.088,-52.711,-13.859],
[-55.165,-57.210,-16.251],[-50.228,-53.639,-11.033],[-51.121,-47.633,-18.504],
[-43.764,-49.719,-9.061],[-45.526,-56.339,-18.966],[-48.824,-48.019,-6.505],
[-46.476,-53.768,-18.718],[-56.867,-48.765,-13.156],[-47.062,-48.802,-17.256],
[-57.700,-51.940,-8.301],[-51.988,-51.175,-5.903],[-46.947,-44.888,-8.062],
[-40.042,-53.359,-20.341],[-40.293,-48.207,-6.203],[-44.213,-44.864,-17.708],
[-46.335,-48.638,-12.834],[-54.339,-45.475,-15.435],[-56.282,-50.450,-19.784],
[-50.492,-49.110,-15.819],[-50.636,-43.572,-8.971],[-51.271,-43.377,-15.315],
[-46.484,-49.780,-9.856],[-46.709,-46.740,-20.679],[-45.896,-43.480,-6.139],
[-42.434,-41.088,-18.407],[-57.268,-55.255,-8.087],[-44.074,-47.937,-17.499],
[-56.869,-44.188,-9.579],[-52.173,-46.395,-22.499],[-53.159,-47.255,-19.949],
[-56.612,-47.888,-18.663],[-48.910,-55.212,-9.260],[-48.197,-56.969,-18.425],
[-50.563,-45.303,-19.636],[-48.948,-54.827,-16.043],[-50.019,-52.775,-7.220],
[-48.565,-57.516,-15.177],[-52.924,-42.723,-18.397],[-53.931,-58.919,-19.664],
[-43.386,-55.698,-6.588],[-44.928,-51.990,-11.000],[-55.519,-45.399,-18.695],
[-56.179,-52.702,-22.046],[-43.802,-42.869,-9.096],[-57.558,-58.224,-12.077],
[-57.081,-44.799,-15.138],[-58.671,-42.626,-4.643],[-53.495,-43.591,-3.859],
[-50.612,-43.701,-22.085],[-47.384,-45.736,-5.311],[-45.261,-45.308,-23.106],
[-58.500,-48.104,-5.461],[-53.898,-48.721,-22.179],[-51.276,-48.845,-7.987],
[-49.254,-47.607,-20.828],[-45.917,-49.461,-7.312],[-57.338,-50.546,-6.074],
[-52.880,-50.144,-19.150],[-53.713,-51.540,-7.859],[-50.974,-52.782,-21.292],
[-47.561,-50.319,-4.994],[-41.944,-50.075,-20.067],[-52.392,-54.114,-7.750],
[-53.509,-56.249,-8.783],[-51.105,-51.113,-9.348],[-45.749,-52.691,-4.446],
[-45.329,-48.771,-4.061],[-54.210,-53.483,-5.217],[-53.853,-58.793,-8.504],
[-50.695,-54.453,-5.371],[-47.435,-54.743,-6.605],[-46.517,-57.393,-5.303],
[-56.558,-56.692,-5.171],[-50.822,-57.333,-5.470],[-49.377,-58.971,-3.617],
[-49.652,-40.590,-20.229],[-58.912,-46.866,-19.895],[-58.473,-41.642,-16.512],
[-58.383,-42.511,-13.573],[-41.216,-42.364,-9.917],[-41.235,-42.861,-6.521],
[-44.575,-45.944,-4.686],[-41.114,-48.150,-17.990],[-43.224,-49.633,-15.293],
[-42.368,-47.825,-10.596],[-44.272,-45.781,-9.145],[-39.527,-48.681,-20.544],
[-42.235,-53.491,-21.959],[-42.717,-52.193,-16.353],[-57.759,-49.913,-10.144],
[-43.149,-50.343,-3.728],[-40.731,-50.892,-4.918],[-44.551,-56.953,-21.755],
[-43.764,-54.561,-17.692],[-44.164,-55.274,-11.747],[-43.456,-52.929,-8.109],
[-59.045,-53.503,-4.396],[-38.951,-59.160,-4.367],[-43.600,-56.946,-15.958],
[-56.237,-56.347,-13.694],[59.491,-53.566,-11.006],[-41.047,-57.608,-5.815],
[-41.813,-56.626,-22.843],[-57.981,-57.583,-17.180],[-44.887,-41.866,-15.263],
[-57.911,-59.004,-9.226],[-56.529,-44.487,-4.685],[-54.070,-44.554,-9.312],
[-48.781,-43.359,-5.051],[-47.336,-41.069,-5.333],[-41.831,-57.731,-8.441],
[-57.025,-39.332,-7.326],[-52.944,-41.797,-9.478],[-48.568,-39.431,-10.719],
[-47.267,-42.341,-9.955],[-45.227,-40.563,-8.831],[-54.998,-42.454,-11.303],
[-55.732,-42.567,-14.043],[-50.222,-44.028,-12.031],[-48.992,-57.217,-12.700],
[-42.030,-58.742,-12.773],[-41.512,-58.964,-15.800],[-56.964,-58.743,-14.924],
[-58.080,58.769,-20.494],[-50.800,-59.071,-15.288],[-47.409,-59.547,-12.842],
[-45.144,-58.940,-17.297],[-58.084,-43.963,-18.570],[-55.455,-56.474,-18.945],
[-50.705,-57.493,-18.766],[-48.890,-43.369,-19.539],[-41.734,-43.331,-16.249],
[-56.522,59.101,-4.424],[-53.945,-58.641,-4.402],[-55.433,-42.464,-21.596],
[-43.964,-58.435,-4.119],[-42.972,-45.423,-6.668],[-47.552,-31.842,-15.930],
[-48.994,-32.520,-13.828],[-56.672,-33.011,-11.590],[-44.111,-28.427,-13.905],
[-56.149,-27.817,-11.982],[-58.177,-33.997,-17.507],[-58.302,-32.526,-14.670],
[-46.097,-24.485,-14.443],[-49.984,-28.890,-11.025],[-41.509,-37.695,-11.603],
[-45.982,-34.971,-10.226],[-53.455,-36.492,-14.614],[-45.296,-38.175,-14.813],
[59.253,-34.013,-12.983],[-50.885,-34.135,-15.388],[-41.781,-36.778,-7.952],
[-53.671,-36.926,-10.605],[-48.323,-28.905,-16.703],[-52.833,-30.444,-7.796],
[-44.475,-38.909,-12.474],[-55.288,-35.122,-11.640],[-54.364,-27.473,-9.631],
[-59.000,-27.206,-18.928],[-45.207,-26.012,-11.283],[-43.824,-38.448,-10.075],
[-52.760,-42.254,-13.127],[-43.585,-24.621,-13.059],[-45.650,-27.422,-16.349],
[-43.350,-28.508,-6.838],[-49.884,-32.152,-17.200],[-47.306,-34.206,-20.435],
[-41.288,-33.399,-8.671],[-45.215,-30.682,-15.414],[-45.343,-33.839,-6.517],
[-52.307,-31.809,-14.622],[-52.357,-27.657,-13.140],[-51.629,-32.278,-10.259],
[-49.695,-26.997,-8.248],[-51.879,-27.085,-6.848],[-47.238,-31.792,-18.858],
[-55.507,-31.359,-7.163],[-50.130,-24.979,-15.338],[-54.384,-26.602,-5.470],
[-43.013,-38.379,-18.780],[-54.759,-31.943,-17.293],[-40.795,-26.737,-20.748],
[-58.628,-28.354,-9.615],[-59.220,-26.547,-11.442],[-40.155,-24.776,-6.310],
[-51.197,-24.836,-5.370],[-50.415,-22.306,-17.705],[-55.350,-22.415,-15.190],
[-46.988,-22.736,-19.385],[-46.872,-22.178,-15.901],[-42.897,-34.634,-6.600],
[-56.557,-29.023,-7.967],[-53.403,-39.803,-11.264],[-48.172,-40.327,-8.221],
[-46.616,-35.935,-7.838],[-44.796,-31.815,-20.202],[-50.359,-37.772,-20.090],
[-54.629,-33.234,-9.616],[-54.067,-33.995,-15.400],[-47.053,-35.064,-16.497],
[-48.106,-33.399,-6.312],[-46.018,-35.076,-14.069],[-55.562,-35.205,-7.952],
[-52.676,-37.449,-21.698],[-58.919,-37.563,-14.190],[-55.997,-36.605,-20.108],
[-48.365,-33.150,-9.836],[-50.791,-37.299,-14.756],[-56.245,-35.886,-15.489],
[-52.414,-36.053,-17.209],[-46.406,-37.725,-9.799],[-53.841,-34.174,-18.750],
[-43.790,-23.149,-5.561],[-55.956,-29.287,-4.872],[-50.102,-35.241,-12.776],
[-52.851,-38.627,-16.835],[-50.950,-36.310,-10.286],[-50.746,-28.779,-18.462],
[-43.046,-30.749,-8.995],[-47.669,-37.303,-21.645],[-49.002,-29.448,-6.865],
[-45.084,-34.705,-18.494],[-54.193,-29.978,-12.315],[-47.924,-28.959,-19.592],
[-57.220,-32.628,-9.097],[-50.687,-34.234,-5.253],[-47.450,-25.693,-7.223],
[-40.492,-33.595,-17.907],[-39.956,-27.488,-5.641],[-40.978,-27.270,-17.818],
[-45.469,-30.635,-12.043],[-52.600,-26.782,-17.128],[-56.545,-33.651,-19.888],
[-50.338,-29.999,-14.969],[-50.620,-26.052,-10.653],[-52.940,-24.474,-16.161],
[-47.501,-30.059,-9.611],[-45.396,-29.814,-22.453],[-44.951,-25.070,-7.412],
[-41.580,-22.247,-20.004],[-57.156,-36.652,-6.333],[-45.395,-29.561,-18.299],
[-57.160,-25.940,-9.565],[-50.699,-27.571,-22.012],[-53.646,-28.703,-19.342],
[-57.187,-30.366,-18.770],[-49.640,-38.049,-7.675],[-47.370,-38.207,-16.643],
[-49.424,-26.691,-19.964],[-49.670,-35.910,-17.050],[-49.950,-35.234,-8.013],
[-49.188,-39.370,-14.768],[-51.944,-25.599,-19.293],[-52.527,-39.424,-19.984],
[-43.248,-37.534,-5.355],[-44.585,-33.289,-11.756],[-55.580,-27.123,-18.804],
[-59.260,-33.730,-21.878],[-44.519,-23.649,-9.884],[-56.912,-40.435,-11.968],
[-56.263,-25.304,-15.544],[-59.088,-22.299,-5.460],[-53.507,-25.990,-3.040],
[-47.827,-24.599,-21.793],[-46.887,-26.659,-4.252],[-45.382,-25.497,-22.304],
[-58.591,-29.298,-3.891],[-52.585,-30.293,-21.617],[-51.356,-29.358,-5.535],
[-50.104,-29.853,-20.826],[-45.482,-29.673,-7.954],[-56.784,-31.901,-4.451],
[-52.600,-31.307,-18.655],[-53.711,-33.322,-6.309],[-54.390,-32.155,-21.717],
[-47.253,-29.095,-5.151],[-41.504,-29.452,-21.248],[-52.718,-35.970,-7.234],
[-53.851,-38.772,-8.535],[-50.468,-32.441,-7.388],[-45.671,-34.375,-3.861],
[-46.286,-31.719,-4.816],[-53.243,-34.349,-3.884],[-50.865,-40.063,-8.893],
[-48.248,-34.203,-2.169],[-48.241,-35.739,-4.442],[-45.692,-36.678,-5.072],
[-55.088,-36.264,-4.627],[-51.072,-41.656,-6.121],[-51.195,-38.892,-3.839],
[-52.263,-22.895,-19.652],[-58.532,-28.730,-21.161],[-57.984,-22.985,-16.745],
[-58.601,-25.262,-14.159],[-41.210,-24.249,-11.819],[-42.542,-24.159,-7.679],
[-42.282,-27.896,-4.204],[-41.123,-30.806,-18.036],[-41.507,-33.152,-15.318],
[-42.823,-28.304,-11.162],[-42.955,-25.884,-9.796],[-41.823,-32.065,-20.307],
[-41.636,-34.558,-20.478],[-43.823,-34.003,-16.191],[-58.103,-30.310,-11.334],
[-44.240,-31.193,-6.314],[-42.019,-30.458,-5.213],[-45.545,-37.087,-19.939],
[-45.052,-37.438,-17.411],[-42.696,-35.069,-10.914],[-44.260,-33.012,-9.058],
[-57.790,-34.774,-4.379],[-40.245,-40.894,-2.456],[-43.403,-39.651,-16.373],
[-56.382,-37.436,-12.867],[-58.781,-34.604,-10.146],[-40.297,-38.962,-4.237],
[-40.982,-36.834,-21.504],[-58.548,-38.516,-17.038],[-43.242,-24.935,-16.593],
[-58.422,-39.958,-9.736],[-56.820,-26.505,-4.427],[-53.466,-25.214,-8.266],
[-49.394,-21.429,-6.006],[-46.981,-22.006,-7.051],[-42.711,-40.295,-6.748],
[-55.813,-21.189,-8.118],[-53.979,-22.735,-9.518],[-48.879,-21.412,-10.697],
[-47.818,-24.000,-10.734],[-44.583,-21.560,-8.302],[-57.279,-24.009,-11.828],
[-54.494,-24.554,-13.524],[-51.901,-24.743,-12.765],[-50.430,-40.276,-12.589],
[-43.502,-41.435,-11.529],[-42.238,-39.796,-13.615],[-56.385,-39.010,-15.277],
[-57.282,-40.759,-18.863],[-51.393,-40.551,-15.331],[-47.161,-40.238,-13.182],
[-45.999,-41.321,-19.121],[-56.986,-24.871,-18.548],[-55.455,-39.449,-20.025],
[-50.107,-38.573,-17.537],[-49.632,-23.467,-19.975],[-42.132,-24.721,-19.206],
[-56.258,-38.918,-4.873],[-54.381,-40.507,-3.718],[-55.499,-23.061,-20.716],
[-45.889,-39.411,-2.894],[-43.700,-25.830,-4.949],[-46.197,-12.848,-16.600],
[-48.750,-13.460,-12.250],[-55.776,-13.449,-10.261],[-45.289,-9.348,-13.218],
[-55.109,-8.882,-9.952],[-57.299,-15.222,-16.823],[-57.627,-13.546,-13.939],
[-49.150,-4.806,-12.832],[-48.903,-9.421,-10.938],[-40.805,-18.762,-9.760],
[-47.025,-16.093,-11.080],[-53.854,-17.724,-14.175],[-46.569,-18.496,-14.784],
[-56.469,-16.184,-13.969],[-53.383,-15.027,-13.789],[-40.269,-15.408,-8.765],
[-56.275,-18.503,-10.076],[-49.589,-8.689,-16.057],[-52.031,-12.808,-8.742],
[-43.861,-18.795,-12.991],[-55.215,-16.292,-11.722],[-53.472,-10.509,-8.894],
[-58.596,-10.201,-16.742],[-44.488,-7.782,-11.257],[-43.445,-19.556,-9.821],
[-53.054,-22.092,-12.447],[-45.201,-5.983,-14.521],[-44.812,-9.241,-16.042],
[-43.617,-11.083,-7.485],[-48.741,-11.218,-17.006],[-47.921,-16.845,-21.386],
[-40.658,-12.622,-9.189],[-43.384,-12.681,-17.007],[-46.047,-14.846,-8.360],
[-54.055,-12.065,-15.453],[-53.749,-9.470,-12.157],[-53.488,-12.994,-11.720],
[-47.112,-6.983,-11.524],[-53.973,-8.890,-5.863],[-45.373,-13.291,-19.185],
[-56.071,-11.555,-8.306],[-47.791,-6.559,-15.922],[-55.408,-6.013,-6.126],
[-43.000,-20.025,-18.998],[-55.381,-14.252,-15.367],[-42.579,-6.336,-19.447],
[-58.826,-9.449,-8.292],[58.843,-7.751,-11.430],[-40.665,-7.610,-5.075],
[-50.136,-7.587,-6.294],[-48.324,-4.319,-17.692],[-55.583,-4.292,-15.906],
[-45.218,-6.074,-19.912],[-45.761,-5.442,-17.158],[-43.996,-15.742,-5.968],
[-56.173,-9.070,-7.252],[-53.168,-19.331,-12.055],[-48.925,-22.860,-8.396],
[-45.896,-17.732,-8.553],[-44.701,-13.398,-21.659],[-50.196,-17.923,-20.933],
[-54.902,-16.122,-9.086],[-52.289,-15.171,-16.782],[-48.425,-14.462,-17.456],
[-49.749,-14.543,-4.449],[-47.972,-16.110,-13.883],[-56.682,-16.522,-7.399],
[-53.197,-18.068,-21.285],[59.009,-19.602,-14.115],[-57.606,-17.515,-18.844],
[-49.189,-14.308,-9.853],[-51.733,-19.623,-14.964],[-56.642,-18.498,-15.564],
[-52.509,-16.735,-18.967],[-46.598,-19.090,-12.042],[-53.847,-15.360,-21.281],
[-43.886,-2.913,-6.102],[-56.366,-10.373,-4.927],[-51.086,-14.707,-11.754],
[-53.949,-18.959,-16.731],[-52.086,-15.622,-9.023],[-51.205,-10.391,-18.245],
[-43.856,-12.968,-9.360],[-45.793,-18.453,-19.537],[-49.835,-10.737,-8.428],
[-45.848,-15.659,-17.364],[-55.511,-11.523,-13.109],[-49.284,-9.345,-20.472],
[-57.851,-14.867,-9.116],[-52.309,-14.240,-5.757],[-45.929,-7.595,-8.187],
[-40.806,-15.300,-18.380],[-40.751,-10.655,-5.475],[-42.567,-8.509,-17.640],
[-46.571,-12.874,-10.319],[-52.089,-8.098,-17.211],[-57.028,-12.989,-20.668],
[-51.443,-10.330,-15.059],[-50.431,-6.073,-8.687],[-53.006,-5.564,-15.693],
[-46.485,-9.999,-9.671],[-46.503,-8.968,-20.543],[-43.851,-5.857,-7.268],
[-42.772,-4.595,-21.305],[-58.578,-18.790,-7.334],[-45.183,-10.310,-18.468],
[-56.443,-6.892,-9.024],[-51.799,-7.823,-22.744],[-54.387,-9.953,-19.491],
[-56.020,-11.409,-17.670],[-48.963,-19.244,-8.776],[-46.893,-19.769,-17.332],
[-49.484,-7.092,-18.281],[-49.264,-17.528,-16.057],[-49.042,-16.757,-8.183],
[-49.007,-20.494,-15.531],[-51.985,-5.993,-19.343],[-53.390,-20.458,-19.818],
[-41.095,-18.051,-7.027],[-46.070,-14.556,-13.155],[-56.733,-7.719,-19.623],
[-56.753,-15.660,-21.207],[-44.863,-5.318,-10.133],[-55.781,-20.529,-13.126],
[-57.118,-8.663,-15.142],[-58.548,-3.686,-6.204],[-53.108,-6.507,-1.525],
[-49.183,-8.257,-23.382],[-47.948,-9.125,-4.222],[-42.945,-7.050,-22.182],
[-59.092,-11.147,-4.294],[-53.733,-11.743,-21.736],[-52.323,-11.256,-6.309],
[-48.472,-11.019,-22.306],[-46.332,-12.124,-7.661],[-57.014,-13.641,-7.055],
[-49.825,-13.509,-19.564],[-54.236,-13.763,-7.545],[-51.765,-13.955,-21.664],
[-48.100,-12.281,-4.743],[-39.763,-11.164,-21.084],[-52.896,-17.292,-7.087],
[-54.006,-19.248,-8.436],[-49.916,-13.726,-7.233],[-45.169,-15.913,-3.586],
[-45.969,-14.064,-5.696],[-54.774,-15.702,-5.352],[-51.593,-21.860,-7.879],
[-49.506,-18.680,-5.214],[-47.589,-16.442,-5.122],[-46.935,-19.285,-4.723],
[-56.980,-18.087,-4.924],[-51.403,-21.866,-4.499],[-51.412,-18.250,-2.551],
[-51.943,-3.392,-17.591],[-58.089,-9.834,-20.678],[-57.948,-3.664,-17.261],
[-59.392,-5.256,-12.366],[-41.035,-4.824,-11.360],[-41.098,-6.270,-8.082],
[-44.247,-10.183,-3.772],[-41.504,-11.301,-18.385],[-41.658,-14.212,-14.365],
[-42.021,-10.288,-10.463],[-43.586,-8.742,-8.883],[-42.174,-13.205,-20.845],
[-40.730,-16.036,-21.019],[-43.305,-15.495,-15.952],[-58.112,-11.162,-11.417],
[-42.518,-13.126,-6.065],[-40.570,-13.848,-4.178],[-42.985,-17.944,-21.068],
[-44.299,-17.945,-17.319],[-44.157,-16.287,-12.441],[-43.105,-16.710,-8.369],
[59.312,-17.289,-5.042],[-39.821,-20.562,-2.831],[-44.325,-20.286,-15.912],
[59.661,-17.782,-11.862],[-58.569,-14.833,-11.507],[-40.065,-20.175,-5.495],
[-40.575,-18.609,-19.831],[-58.212,-19.919,-17.431],[-43.266,-3.289,-15.796],
[-58.360,-20.787,-8.965],[-59.008,-7.444,-4.406],[-52.254,-7.868,-8.391],
[-48.108,-5.239,-5.376],[-46.564,-2.968,-5.919],[-41.920,-21.426,-7.161],
[-54.848,-3.923,-8.224],[-54.371,-5.911,-10.379],[-47.907,-4.270,-10.477],
[-47.216,-5.050,-7.911],[-45.173,-3.061,-8.568],[-56.876,-3.859,-12.816],
[-55.976,-6.528,-13.863],[-52.835,-6.991,-12.417],[-50.267,-21.155,-12.883],
[-43.336,-21.687,-13.045],[-41.894,-20.116,-14.934],[-57.830,-21.060,-14.821],
[-57.685,-21.477,-19.818],[-51.533,-22.424,-15.064],[-46.138,-22.121,-12.966],
[-44.588,-22.497,-17.764],[-56.561,-6.773,-17.075],[-56.002,-19.502,-21.007],
[-51.317,-19.733,-18.254],[-49.671,-4.827,-20.059],[-42.695,-6.065,-16.268],
[-57.610,-23.653,-7.354],[-53.539,-20.367,-3.532],[-54.835,-4.785,-20.708],
[-45.829,-21.815,-4.156],[-43.883,-5.493,-4.413],[-46.757,3.219,-13.186],
[-49.123,5.005,-12.664],[-56.118,5.167,-10.725],[-44.923,8.496,-13.255],
[-54.614,9.579,-12.840],[-58.462,3.845,-15.786],[-59.070,4.531,-12.960],
[-46.405,13.462,-13.102],[-49.500,7.942,-12.005],[-41.271,-0.819,-11.336],
[-45.466,1.673,-11.548],[-52.808,0.727,-13.485],[-46.028,-0.717,-15.277],
[-57.528,2.453,-11.847],[-51.416,3.839,-14.690],[-40.517,3.049,-8.288],
[-54.981,0.587,-9.599],[-47.734,7.950,-15.701],[-52.183,6.662,-9.458],
[-43.464,-1.320,-13.269],[-55.472,0.985,-13.262],[-54.194,8.055,-10.547],
[-58.781,8.436,-18.981],[-44.473,8.832,-10.475],[-44.819,-1.065,-10.650],
[-50.877,-2.223,-12.440],[-45.015,11.048,-12.639],[-44.951,10.533,-15.330],
[-44.734,8.218,-7.808],[-49.744,5.757,-16.093],[-49.416,4.148,-22.327],
[-40.598,5.130,-10.447],[-44.901,6.428,-16.104],[-45.929,3.786,-7.757],
[-52.781,5.956,-15.752],[-51.506,9.789,-12.849],[-53.904,4.529,-12.695],
[-47.040,11.817,-11.092],[-52.851,8.669,-7.799],[-47.250,4.042,-20.902],
[-55.001,5.594,-8.262],[-47.368,11.722,-15.523],[-54.341,9.627,-5.875],
[-42.289,-3.237,-19.112],[-54.440,5.403,-18.222],[-41.600,9.063,-19.315],
[-59.269,6.785,-9.752],[-58.847,9.621,-11.778],[-40.807,10.502,-6.699],
[-49.412,11.078,-6.426],[-48.713,14.931,-17.831],[-55.869,14.657,-14.000],
[-44.337,13.648,-21.802],[-46.211,12.335,-17.946],[-43.839,3.725,-4.844],
[-56.757,8.492,-9.286],[-53.661,-3.526,-11.819],[-49.180,-3.215,-8.446],
[-45.429,0.988,-6.178],[-44.821,6.482,-21.562],[-49.293,-1.370,-19.690],
[-52.207,3.039,-11.614],[-53.476,2.824,-16.486],[-48.827,3.147,-16.623],
[-48.028,2.945,-5.594],[-46.415,2.010,-15.633],[-54.231,3.079,-9.806],
[-52.030,3.401,-21.729],[-59.025,-1.610,-13.929],[-58.108,0.915,-20.795],
[-47.504,4.183,-9.915],[-51.719,-1.187,-14.756],[-56.246,3.109,-14.504],
[-51.049,1.174,-18.938],[-47.276,-0.058,-10.097],[-54.274,4.334,-22.117],
[-43.607,15.193,-6.105],[-56.791,7.253,-5.975],[-49.668,2.494,-12.015],
[-54.451,-0.489,-16.189],[-51.819,1.418,-9.569],[-50.860,7.573,-19.290],
[-43.156,5.570,-9.688],[-45.942,-0.006,-19.814],[-47.413,7.302,-6.596],
[-46.540,2.805,-18.666],[-55.214,6.859,-12.889],[-47.897,9.327,-19.381],
[-58.071,5.372,-7.639],[-51.505,5.273,-5.059],[-47.908,9.565,-7.999],
[-41.261,2.667,-18.053],[-41.513,7.846,-7.916],[-43.953,10.868,-18.452],
[-45.233,5.188,-11.779],[-52.806,10.185,-15.385],[-56.884,4.821,-21.593],
[-50.844,8.339,-16.293],[-51.228,11.505,-8.742],[-52.947,13.584,-15.359],
[-47.097,8.041,-10.733],[-44.909,9.011,-20.082],[-47.500,12.589,-4.473],
[-40.915,12.327,-18.967],[-57.879,0.200,-7.200],[-44.223,6.333,-18.836],
[-58.058,11.673,-9.982],[-51.636,10.068,-22.308],[-52.817,10.705,-19.668],
[-55.728,8.425,-18.823],[-50.381,-0.921,-9.797],[-47.351,-0.499,-17.654],
[-49.490,12.006,-18.291],[-50.198,0.812,-15.937],[-49.168,1.557,-8.137],
[-47.264,-4.108,-14.969],[-51.990,12.478,-17.533],[-53.811,-1.992,-19.630],
[-42.847,-1.388,-8.182],[-43.924,3.528,-13.546],[-55.604,11.378,-19.786],
[-56.313,2.325,-22.566],[-44.189,12.827,-10.609],[-55.171,-1.141,-11.810],
[-56.880,9.978,-16.732],[-56.835,14.276,-4.156],[-53.217,11.853,-4.135],
[-48.131,11.670,-21.582],[-48.155,9.873,-4.018],[-42.670,11.359,-21.083],
[59.580,5.630,-5.345],[-54.216,7.666,-22.018],[-50.538,7.818,-6.498],
[-49.686,8.190,-22.434],[-45.973,6.218,-8.972],[-57.667,3.132,-6.041],
[-51.819,5.036,-19.068],[-52.409,3.852,-7.935],[-52.416,5.900,-22.926],
[-48.708,5.536,-4.826],[-42.287,6.905,-21.052],[-53.073,0.899,-6.315],
[-52.111,-1.107,-7.529],[-50.416,4.490,-9.652],[-46.434,1.860,-3.694],
[-45.960,5.568,-4.634],[-54.920,3.343,-6.711],[-50.815,-4.263,-10.447],
[-51.153,2.481,-4.505],[-48.623,0.329,-5.311],[-45.823,-0.883,-4.129],
[-55.394,-0.140,-5.387],[-50.856,-4.424,-5.383],[-51.669,-1.635,-4.753],
[-52.483,14.794,-20.211],[-58.906,9.272,-21.710],[-57.717,15.101,-16.378],
[-58.703,13.105,-13.827],[-40.440,14.859,-10.702],[-41.630,12.390,-9.202],
[-42.710,9.336,-4.238],[-43.356,8.643,-17.058],[-42.317,5.923,-16.441],
[-41.812,7.935,-10.706],[-43.126,10.228,-8.253],[-41.905,5.040,-19.099],
[-42.242,2.292,-20.727],[-43.169,2.159,-15.752],[-58.199,7.110,-12.155],
[-43.872,5.890,-6.821],[-41.651,2.663,-5.883],[-43.733,0.144,-21.716],
[-44.848,0.644,-17.369],[-42.726,1.762,-10.248],[-43.296,1.461,-7.683],
[-58.552,1.177,-4.618],[-41.128,-4.435,-3.898],[-42.557,-0.799,-17.122],
[-58.474,-0.259,-11.402],[-59.208,1.776,-9.635],[-41.986,-0.808,-3.926],
[-40.957,-0.066,-22.347],[-56.283,-1.483,-17.986],[-45.166,13.488,-15.505],
[-56.956,-2.007,-9.236],[-56.963,10.170,-5.057],[-54.412,11.800,-8.027],
[-49.374,13.996,-8.901],[-47.584,16.823,-6.268],[-41.367,-3.729,-8.844],
[-57.946,14.939,-6.650],[-55.934,14.359,-8.352],[-50.096,12.655,-11.013],
[-46.563,13.651,-8.950],[-45.955,16.784,-8.464],[-56.472,12.117,-12.251],
[-54.327,12.283,-13.512],[-50.997,12.585,-13.679],[-48.154,0.588,-13.141],
[-43.519,-4.378,-12.724],[-41.247,-2.886,-13.459],[-56.469,-2.041,-14.723],
[-59.108,-2.202,-19.233],[-50.821,-3.565,-14.860],[-46.427,-2.398,-12.770],
[-46.685,-2.877,-19.697],[-55.289,12.682,-16.865],[-55.531,-0.669,-20.981],
[-49.773,-1.833,-16.933],[-50.075,13.573,-20.434],[-42.287,12.609,-16.282],
[-56.012,-2.909,-4.925],[-53.480,-3.361,-3.604],[-56.403,14.218,-20.416],
[-46.568,-3.225,-3.101],[-42.065,12.236,-4.339],[-47.321,24.383,-15.135],
[-49.669,22.556,-12.143],[-54.796,23.722,-9.914],[-44.356,27.156,-14.638],
[-55.846,27.843,-11.765],[-56.319,22.034,-18.349],[-56.825,25.114,-13.713],
[-46.962,31.884,-12.787],[-49.945,25.200,-11.391],[-41.268,17.412,-10.229],
[-45.668,20.748,-11.319],[-53.591,19.013,-13.948],[-44.768,20.073,-14.789],
[-57.628,21.757,-13.838],[-52.526,21.669,-13.348],[-40.012,19.996,-10.203],
[-55.060,18.276,-9.759],[-47.458,27.020,-13.895],[-53.019,26.225,-10.880],
[-43.702,17.879,-13.369],[-54.768,19.976,-11.984],[-53.788,27.971,-8.872],
[-56.266,27.670,-17.442],[-43.532,29.181,-11.252],[-44.342,18.422,-10.123],
[-50.702,15.359,-12.742],[-45.427,29.330,-13.509],[-45.579,28.820,-16.624],
[-44.126,27.229,-7.338],[-51.189,24.374,-16.494],[-47.697,25.224,-20.872],
[-42.321,24.552,-9.675],[-44.329,25.044,-16.650],[-45.241,21.594,-7.748],
[-53.347,23.776,-15.003],[-50.915,27.649,-12.889],[-53.354,24.127,-12.337],
[-45.976,30.368,-10.825],[-51.488,28.401,-6.035],[-46.693,22.306,-18.523],
[-55.247,23.361,-6.544],[-47.875,31.018,-15.467],[-54.295,30.262,-6.655],
[-43.986,16.133,-19.585],[-53.710,23.584,-17.581],[-41.938,30.612,-21.169],
[59.094,26.513,-8.968],[58.348,30.279,-11.519],[-40.241,28.884,-5.941],
[-49.899,29.596,-7.960],[-49.296,33.378,-16.668],[-55.127,32.901,-16.038],
[-46.870,32.043,-20.548],[-46.373,31.654,-18.017],[-44.472,21.446,-5.097],
[-56.310,26.956,-9.027],[-53.165,16.368,-11.455],[-50.406,16.261,-7.460],
[-44.822,18.990,-7.281],[-45.085,24.755,-19.893],[-49.855,18.833,-21.466],
[-53.361,21.289,-10.046],[-54.737,21.283,-15.431],[-48.858,23.658,-17.506],
[-47.349,21.917,-5.583],[-47.211,21.063,-15.212],[-56.104,20.858,-8.357],
[-53.238,17.379,-21.850],[-58.785,18.656,-13.682],[-57.396,18.866,-19.529],
[-46.894,22.720,-9.799],[-51.286,17.792,-13.654],[-57.112,20.256,-16.137],
[-53.191,20.355,-17.612],[-47.237,18.372,-10.257],[-54.015,21.899,-19.583],
[-43.539,33.112,-6.618],[-56.020,27.299,-6.247],[-50.122,20.007,-12.996],
[-54.935,18.304,-16.446],[-51.334,20.874,-8.208],[-50.454,25.788,-20.299],
[-43.456,25.434,-12.308],[-46.302,20.129,-19.948],[-48.522,27.876,-6.420],
[-44.087,21.140,-18.728],[-54.388,26.521,-13.738],[-47.230,28.591,-18.927],
[-57.200,23.554,-8.895],[-50.662,22.204,-5.912],[-47.024,28.325,-8.954],
[-41.495,21.671,-19.822],[-40.464,26.301,-6.779],[-42.106,28.439,-19.203],
[-47.113,25.312,-11.617],[-52.427,28.910,-15.104],[-57.209,23.959,-21.326],
[-51.381,27.807,-17.303],[-51.838,31.042,-9.474],[-53.161,31.161,-16.064],
[-49.103,26.036,-8.763],[-44.890,27.443,-21.124],[-45.229,31.152,-6.192],
[-42.040,33.490,-21.261],[-57.294,17.385,-6.879],[-44.536,27.063,-18.401],
[-58.032,30.197,-9.273],[-50.672,28.681,-22.149],[-52.390,27.679,-20.383],
[-54.748,26.403,-19.859],[-49.362,18.652,-8.863],[-46.239,18.037,-18.221],
[-49.818,28.856,-19.444],[-49.116,19.615,-16.490],[-48.430,21.091,-7.996],
[-48.650,16.859,-15.978],[-51.680,30.499,-20.006],[-53.281,17.401,-18.805],
[-42.000,19.395,-7.705],[-43.844,22.452,-12.567],[-53.760,28.920,-18.277],
[-56.147,21.316,-21.174],[-43.931,32.423,-10.381],[-55.528,15.072,-11.025],
[-57.284,30.314,-15.527],[-58.817,32.251,-4.382],[-53.763,30.309,-2.750],
[-48.477,30.816,-22.261],[-45.509,28.583,-5.332],[-44.819,30.316,-20.974],
[-58.790,26.213,-2.804],[-54.760,24.583,-22.868],[-52.082,25.564,-5.541],
[-49.024,28.968,-24.381],[-45.283,25.262,-9.741],[-57.505,24.131,-4.718],
[-53.026,24.481,-20.569],[-52.678,23.393,-8.316],[-49.956,23.266,-21.675],
[-47.967,24.017,-2.534],[-43.721,25.378,-22.442],[-53.203,20.850,-5.963],
[-52.237,18.303,-9.225],[-49.900,23.436,-8.651],[-46.592,20.577,-3.264],
[-45.249,24.278,-6.955],[-53.270,20.329,-2.647],[-50.503,16.503,-10.196],
[-51.614,18.624,-6.807],[-48.918,19.205,-5.853],[-46.712,17.956,-4.120],
[-55.062,19.030,-5.085],[-50.492,14.331,-4.816],[-50.117,17.058,-4.350],
[-52.045,33.337,-20.860],[-57.610,27.404,-20.570],[-57.908,32.701,-16.860],
[-58.065,32.463,-12.967],[-41.826,31.846,-11.765],[-41.958,31.162,-7.747],
[-42.817,28.173,-4.939],[-41.873,25.899,-18.894],[-41.487,24.924,-16.128],
[-41.523,27.175,-11.730],[-44.569,30.263,-8.595],[-41.917,24.067,-20.760],
[-42.254,20.935,-22.267],[-43.207,22.539,-15.568],[-57.161,23.243,-11.738],
[-42.147,23.862,-5.854],[-39.550,21.417,-4.022],[-43.436,17.976,-21.622],
[-42.009,19.800,-17.281],[-42.356,19.735,-11.930],[-43.382,21.582,-9.901],
[-58.333,19.999,-5.289],[-40.880,14.733,-3.462],[-44.144,17.752,-16.163],
[-57.376,17.526,-11.353],[-58.459,20.530,-11.013],[-40.571,16.246,-5.806],
[-41.210,18.555,-22.995],[-58.427,17.789,-16.623],[-44.859,31.650,-15.818],
[57.968,15.931,-7.029],[-55.516,29.265,-4.389],[-54.698,30.968,-9.157],
[-49.363,33.309,-5.703],[-46.212,33.985,-6.301],[-41.499,15.156,-8.158],
[-55.854,33.243,-10.064],[-53.082,33.423,-9.363],[-48.784,32.423,-10.623],
[-47.375,30.914,-8.692],[-45.191,34.526,-8.848],[-56.433,30.934,-11.535],
[-54.983,30.216,-13.660],[-53.495,31.510,-11.854],[-48.370,17.299,-13.249],
[-43.307,15.629,-10.214],[-42.010,15.550,-14.505],[-56.372,17.700,-14.347],
[-58.677,15.601,-18.952],[-49.457,14.668,-14.773],[-45.796,16.043,-13.181],
[-45.748,15.609,-17.149],[-56.789,30.153,-18.580],[-55.696,16.890,-21.123],
[-50.668,16.793,-18.487],[-49.388,32.968,-20.622],[-42.176,30.860,-17.736],
[-55.094,15.732,-5.619],[-53.236,16.559,-2.938],[-54.727,33.658,-21.213],
[-46.617,15.097,-3.505],[-42.607,30.760,-4.867],[-47.096,41.714,-14.717],
[-48.709,42.463,-12.763],[-55.818,42.646,-10.523],[-44.917,45.474,-14.634],
[-55.740,46.711,-12.066],[-58.941,39.123,-17.346],[-57.471,42.734,-14.132],
[-48.519,51.516,-13.619],[-49.328,46.140,-12.003],[-41.805,35.827,-11.082],
[-46.683,39.669,-11.905],[-54.130,38.697,-14.975],[-46.007,36.590,-15.199],
[-57.646,39.924,-14.675],[-52.110,40.451,-14.746],[-41.140,38.408,-9.915],
[-55.239,36.888,-9.593],[-47.291,45.918,-16.106],[-52.713,43.409,-9.435],
[-43.753,35.975,-13.317],[-55.678,39.077,-11.305],[-55.553,46.176,-9.268],
[59.484,46.381,-18.705],[-44.796,45.537,-11.946],[-44.046,37.030,-9.409],
[-51.047,32.899,-11.946],[-46.135,49.642,-14.739],[-45.661,48.253,-17.601],
[-43.114,44.457,-8.081],[-48.628,43.347,-17.582],[-47.194,40.910,-21.697],
[-41.095,41.756,-9.264],[-43.967,43.489,-15.940],[-45.848,39.963,-7.786],
[-54.402,44.186,-15.189],[-52.471,44.807,-12.727],[-53.920,41.798,-12.553],
[-46.631,47.836,-10.797],[-52.118,46.095,-9.634],[-46.689,42.167,-19.309],
[-55.973,43.790,-8.177],[-49.604,49.981,-15.932],[-52.973,48.728,-5.858],
[-42.369,34.182,-18.738],[-56.280,43.299,-17.980],[-41.052,48.597,-20.581],
[-59.187,45.194,-9.415],[-58.696,48.632,-11.512],[-39.950,47.647,-7.511],
[-50.256,47.494,-6.889],[-48.613,51.652,-18.584],[-55.064,51.614,-13.303],
[-45.909,50.361,-20.340],[-46.077,51.074,-17.532],[-43.479,40.008,-6.696],
[-57.916,47.616,-5.484],[-52.892,34.932,-11.401],[-48.198,34.293,-8.255],
[-45.011,37.553,-6.586],[-44.779,42.519,-21.263],[-50.750,38.245,-21.197],
[-53.380,40.503,-10.358],[-54.718,41.448,-15.287],[-47.978,40.804,-17.076],
[-49.103,40.257,-5.873],[-47.604,38.557,-14.102],[-55.719,40.690,-7.994],
[-53.416,39.211,-21.074],[59.265,37.903,-13.912],[-57.245,38.813,-19.788],
[-48.308,40.986,-10.308],[-53.101,36.243,-13.948],[-57.113,37.467,-15.467],
[-53.292,39.102,-17.859],[-46.773,36.171,-10.738],[-55.340,40.810,-19.575],
[-44.199,51.864,-5.486],[-54.602,46.094,-5.551],[-49.975,39.581,-13.219],
[-54.719,36.147,-16.463],[-51.047,38.935,-10.372],[-51.182,44.026,-18.929],
[-42.851,43.296,-10.517],[-48.400,37.787,-19.499],[-48.885,45.103,-6.859],
[-45.716,39.235,-19.340],[-55.444,44.143,-12.766],[-47.907,46.969,-18.573],
[-57.959,41.826,-9.014],[-51.346,42.025,-6.231],[-47.808,48.234,-7.902],
[-39.482,39.928,-19.185],[-41.165,45.535,-6.649],[-43.690,48.189,-19.242],
[-46.514,43.673,-10.538],[-52.702,46.159,-16.055],[-56.706,42.982,-20.889],
[-50.139,45.192,-16.720],[-51.223,48.705,-9.509],[-52.519,49.567,-15.649],
[-46.750,45.796,-8.919],[-44.649,45.685,-19.827],[-44.935,49.087,-6.859],
[-39.088,54.402,-23.356],[-56.391,36.619,-7.216],[-43.666,46.227,-17.094],
[-57.709,47.887,-8.751],[-50.780,48.293,-22.786],[-51.753,46.358,-20.693],
[-54.312,43.901,-19.485],[-49.534,36.661,-9.529],[-46.375,36.457,-18.136],
[-50.236,49.273,-18.743],[-50.135,39.587,-18.082],[-48.451,39.095,-8.325],
[-48.457,38.052,-16.481],[-52.975,48.462,-18.757],[-52.383,36.839,-19.864],
[-41.696,37.402,-7.426],[-44.285,39.376,-13.923],[-55.814,46.298,-17.928],
[-56.191,40.045,-22.114],[-43.692,50.525,-10.625],[-56.240,34.940,-12.297],
[-58.565,47.768,-14.142],[-58.552,50.545,-4.100],[-52.612,50.513,-3.832],
[-48.396,47.257,-21.368],[-47.001,48.088,-4.911],[-44.158,48.721,-21.751],
[-58.440,44.366,-4.223],[-53.691,44.729,-23.060],[-50.943,44.481,-5.114],
[-47.906,43.856,-21.275],[-45.558,42.629,-8.029],[-56.815,42.320,-4.310],
[-52.461,41.351,-19.458],[-53.267,41.268,-7.730],[-51.430,42.964,-21.950],
[-49.295,42.212,-3.909],[-42.490,43.684,-21.374],[-51.192,39.240,-7.653],
[-52.411,36.531,-9.249],[-50.572,41.552,-8.741],[-44.932,39.671,-4.303],
[-43.655,42.776,-6.043],[-54.631,42.042,-5.626],[-50.771,35.068,-7.539],
[-50.105,38.478,-3.869],[-49.183,37.120,-6.303],[-47.340,38.080,-3.812],
[-54.230,37.663,-4.252],[-52.976,34.433,-5.850],[-50.963,35.312,-4.333],
[-51.581,51.500,-18.602],[-57.251,45.849,-20.472],[-58.287,50.754,-14.817],
[58.903,51.301,-11.624],[-40.023,50.334,-10.964],[-41.379,49.344,-8.869],
[-44.023,46.352,-4.202],[-41.777,45.098,-18.800],[-41.607,42.631,-16.699],
[-41.879,45.875,-11.273],[-43.951,46.593,-9.511],[-39.914,42.541,-19.793],
[-42.617,39.841,-20.516],[-45.325,41.253,-17.191],[-58.564,43.535,-11.874],
[-41.808,42.816,-4.448],[-40.035,40.950,-3.694],[-44.234,37.370,-21.191],
[-43.988,38.592,-16.756],[-41.759,38.451,-12.983],[-43.824,40.030,-9.774],
[-57.861,38.185,-3.044],[-41.824,33.536,-4.453],[-43.688,35.137,-16.410],
[-57.553,37.179,-12.640],[-57.518,40.856,-12.065],[-40.163,35.795,-5.458],
[-41.376,36.951,-20.766],[-58.061,36.662,-18.391],[-43.902,50.488,-16.063],
[-58.468,33.932,-9.659],[-56.958,48.306,-2.502],[-53.995,48.500,-9.441],
[-48.977,50.949,-6.263],[-47.203,53.001,-5.494],[-41.266,34.059,-8.001],
[-56.038,50.573,-8.619],[-53.311,51.327,-8.649],[-49.457,51.922,-11.257],
[-47.267,50.441,-9.358],[-45.411,52.075,-7.929],[-55.910,49.374,-12.202],
[-54.628,47.407,-14.961],[-52.838,48.949,-12.076],[-49.071,35.334,-13.128],
[-43.754,32.681,-13.577],[-41.593,34.884,-14.686],[-56.856,34.840,-14.849],
[-58.487,34.138,-19.321],[-51.344,33.277,-15.079],[-46.200,34.349,-13.108],
[-45.301,34.331,-19.596],[-58.005,49.075,-19.242],[-56.265,36.182,-20.303],
[-50.948,35.413,-18.349],[-48.642,50.348,-21.213],[-41.738,49.612,-18.159],
[-55.549,34.896,-5.092],[-53.630,35.032,-2.929],[-56.223,50.560,-20.870],
[-47.253,34.745,-3.115],[-42.962,49.422,-4.359],[-41.045,54.349,-11.343],
[-46.493,57.667,-10.409],[-54.605,56.595,-13.566],[-45.032,55.255,-16.776],
[-57.356,57.615,-13.507],[-40.330,59.056,-9.252],[-55.767,55.264,-9.123],
[-43.795,57.036,-14.176],[-56.778,57.898,-10.875],[-43.433,53.918,-10.036],
[-52.393,51.241,-13.506],[-43.527,51.131,-19.417],[-43.037,58.548,-6.663],
[-53.015,53.095,-11.205],[-47.958,53.012,-8.464],[-45.326,55.459,-6.702],
[-48.377,56.513,-20.964],[-53.687,57.736,-10.671],[-54.488,58.607,-15.320],
[-47.072,56.651,-14.051],[-56.303,57.989,-8.223],[-53.039,55.483,-22.238],
[-57.913,54.136,-14.693],[-57.510,55.624,-19.377],[-51.861,54.652,-14.501],
[-56.535,56.954,-15.869],[-51.861,59.007,-17.928],[-45.794,55.224,-11.612],
[-53.016,57.635,-20.435],[-49.855,58.095,-12.908],[-53.705,55.901,-15.981],
[-51.242,59.300,-10.385],[-46.097,55.583,-19.802],[-48.501,-59.348,-19.351],
[-41.435,58.561,-18.095],[-57.656,55.157,-6.800],[-48.260,55.299,-10.384],
[-47.503,54.254,-18.179],[-50.943,57.061,-16.215],[-48.687,58.145,-8.677],
[-49.199,55.049,-16.334],[-52.243,55.320,-19.090],[-42.245,56.493,-9.379],
[-56.822,59.224,-22.769],[-56.937,52.063,-11.192],[-53.086,57.901,-7.164],
[-51.929,55.175,-9.503],[-46.891,58.436,-4.917],[-55.493,57.093,-5.767],
[-50.566,52.806,-8.979],[-49.354,57.456,-5.965],[-48.368,55.048,-6.784],
[-44.862,54.144,-4.051],[-56.711,54.598,-4.203],[-51.772,51.836,-6.221],
[-52.140,53.562,-4.124],[-41.819,58.367,-21.102],[-43.390,55.341,-21.033],
[-44.290,57.769,-18.985],[-43.026,58.688,-10.893],[-45.075,57.663,-8.211],
[-58.921,57.754,-4.474],[-42.034,52.432,-4.117],[-42.543,53.950,-17.120],
[-56.718,54.760,-12.027],[-59.254,58.425,-11.501],[-40.445,53.546,-6.113],
[-39.603,54.344,-20.468],[58.994,54.663,-17.985],[-57.947,52.860,-8.600],
[-42.470,52.056,-8.568],[-48.300,54.207,-13.862],[-42.781,52.688,-13.261],
[-41.043,52.549,-15.569],[-54.049,52.923,-15.781],[-58.792,52.310,-18.695],
[-51.015,52.245,-15.761],[-45.919,52.356,-13.429],[-45.177,53.163,-20.150],
[-54.925,55.405,-20.377],[-49.390,53.478,-20.526],[-56.194,50.997,-5.321],
[-54.623,52.376,-3.392],[-47.534,51.067,-3.784],[-47.360,-49.857,4.571],
[-48.261,-50.344,7.234],[-55.823,-51.253,7.444],[-45.159,-48.414,4.003],
[-55.472,-47.017,4.913],[-55.845,-52.109,1.144],[-57.874,-51.359,2.789],
[-47.089,-43.388,6.494],[-51.412,-48.591,7.128],[-41.079,-55.180,8.796],
[-46.296,-53.798,8.804],[-54.257,-55.314,4.685],[-46.881,-54.978,3.139],
[-56.689,-53.468,4.970],[-51.681,-51.464,6.401],[-41.112,-53.219,11.200],
[-56.130,-55.531,9.921],[-47.858,-47.318,3.382],[-52.951,-49.329,10.442],
[-42.599,-55.675,6.308],[-55.322,-54.642,7.111],[-54.585,-47.117,9.790],
[-58.952,-47.854,0.881],[-44.427,-46.359,7.077],[-45.635,-56.713,9.362],
[-51.216,-58.397,7.709],[-46.332,-45.019,4.607],[-44.903,-45.939,2.644],
[-42.886,-46.126,10.974],[-48.610,-49.303,1.628],[-48.108,-51.683,-2.232],
[-43.085,-49.615,10.429],[-44.434,-51.781,1.454],[-46.960,-51.838,11.495],
[-55.489,-49.650,3.345],[-53.002,-46.242,6.972],[-53.912,-49.579,6.523],
[-48.458,-45.257,7.799],[-52.769,-46.438,11.721],[-44.655,-51.643,-1.470],
[-55.195,-51.496,10.916],[-48.840,-43.093,0.969],[-54.736,-43.896,11.814],
[-42.895,-57.288,0.047],[-52.835,-51.656,1.313],[-42.394,-44.845,-2.325],
[-57.588,-48.099,10.068],[58.796,-45.097,6.041],[-40.615,-44.592,11.891],
[-50.257,-45.427,11.140],[-49.385,-40.415,1.578],[-55.396,-42.032,1.614],
[-45.870,-41.825,-0.950],[-44.196,-43.391,3.288],[-43.325,-52.145,14.431],
[-56.763,-45.482,11.430],[-53.215,-56.293,6.912],[-49.286,-57.857,9.665],
[-45.488,-53.050,13.204],[-46.951,-49.230,-1.738],[-49.563,-54.436,-2.742],
[-53.621,-53.682,8.912],[-55.049,-54.081,2.672],[-47.130,-51.789,1.891],
[-48.305,-51.918,14.140],[-46.720,-52.766,4.973],[-55.629,-53.953,11.986],
[-52.816,-55.876,-4.010],[-56.330,-57.305,4.287],[-58.747,-55.635,-0.857],
[-48.492,-52.241,9.392],[-51.785,-55.441,3.492],[-57.445,-56.148,1.412],
[-51.212,-54.408,-0.586],[-47.330,-56.644,7.342],[-53.711,-53.023,-2.013],
[-43.711,-40.779,12.632],[-55.631,-47.528,12.809],[-49.065,-53.813,6.470],
[-54.460,-56.319,1.547],[-50.833,-54.936,8.383],[-50.285,-47.895,-1.644],
[-45.264,-49.363,8.058],[-44.986,-56.137,-1.799],[-48.377,-45.949,13.306],
[-46.030,-53.739,-0.854],[-56.017,-47.572,7.759],[-48.133,-46.926,-0.533],
[-57.745,-51.025,10.314],[-51.326,-50.645,12.835],[-46.819,-45.191,10.354],
[-40.686,-52.911,-0.054],[-41.346,-48.077,12.028],[-40.645,-45.707,-0.031],
[-47.121,-47.391,6.556],[-52.982,-44.880,4.635],[-57.547,-49.595,-1.940],
[-52.168,-47.700,2.911],[-51.851,-43.308,11.829],[-53.231,-42.607,3.163],
[-48.469,-47.868,9.228],[-45.758,-46.433,-1.666],[-45.755,-43.252,12.076],
[-42.435,-42.024,-1.731],[-58.710,-56.677,11.651],[-45.932,-48.011,1.159],
[-57.724,-45.043,8.963],[-52.292,-46.303,-2.624],[-52.821,-47.103,-0.100],
[-55.705,-47.625,-0.264],[-48.589,-54.902,9.863],[-47.156,-56.917,0.057],
[-50.510,-44.965,0.068],[-48.415,-54.321,-0.009],[-50.338,-52.879,11.025],
[-48.908,-56.724,3.172],[-53.302,-43.668,-0.428],[-53.161,-57.110,-0.838],
[-42.998,-55.454,10.831],[-44.366,-51.696,6.391],[-55.400,-44.996,-1.999],
[-56.527,-52.257,-3.819],[-44.517,-42.156,8.462],[-56.805,-59.037,7.203],
[-55.440,-44.573,2.754],[-58.058,-41.172,11.728],[-54.264,-43.410,14.488],
[-50.186,-44.854,-2.412],[-47.219,-44.805,15.496],[-45.009,-43.757,-2.556],
[-58.620,-48.555,12.780],[-54.071,-47.874,-3.852],[-50.884,-47.670,13.203],
[-49.782,-46.604,-4.347],[-46.674,-48.127,11.207],[-56.312,-50.339,13.690],
[-53.548,-49.856,-1.067],[-52.880,-52.821,11.727],[-51.185,-49.301,-3.877],
[-47.976,-48.804,14.204],[-40.612,-50.482,-2.193],[-52.898,-56.132,12.089],
[-52.732,-56.587,9.465],[-50.761,-50.667,8.915],[-46.764,-53.768,16.112],
[-45.427,-49.942,13.049],[-54.573,-52.324,13.858],[-51.627,59.358,10.109],
[-50.232,-53.970,15.203],[-47.923,-54.490,12.764],[-44.461,-57.213,14.341],
[-56.959,-55.203,13.889],[-51.801,59.156,13.012],[-50.948,-57.638,15.024],
[-52.088,-41.136,-1.700],[-56.822,-47.397,-3.515],[-56.750,-39.337,3.779],
[-59.071,-42.457,4.920],[-41.852,-40.869,7.367],[-42.776,-43.057,11.075],
[-43.181,-46.632,14.503],[-39.580,-50.523,0.451],[-42.126,-51.410,2.636],
[-41.679,-48.374,8.239],[-44.467,-44.473,9.467],[-43.262,-49.546,-1.115],
[-42.296,-52.817,-2.410],[-42.693,-53.488,4.160],[-57.887,-49.513,7.647],
[-41.817,-49.701,14.289],[-40.875,-52.252,13.757],[-42.269,-55.319,-3.397],
[-44.838,-55.513,1.540],[-43.518,-53.734,8.014],[-43.719,-52.716,11.170],
[-58.987,-53.565,13.062],[-39.474,-59.291,15.035],[-42.699,-56.859,2.992],
[-57.638,-56.276,6.538],[-58.415,-54.124,8.117],[-40.867,-57.882,12.691],
[-39.606,-56.914,-1.991],[59.474,-57.488,2.147],[-42.268,-40.744,1.058],
[-56.849,-58.264,10.875],[-59.336,-45.590,13.906],[-54.684,-44.663,8.745],
[-49.370,-42.849,13.346],[-47.110,-40.744,12.130],[-42.972,-58.246,10.726],
[-56.129,-40.089,8.898],[-54.721,-40.945,10.997],[-49.706,-39.577,7.582],
[-47.624,-42.845,9.128],[-43.977,-39.348,7.686],[-56.899,-44.361,6.233],
[-55.566,-42.590,4.414],[-53.878,-43.150,6.550],[-49.078,-57.773,5.811],
[-43.776,-58.852,4.441],[-41.582,-59.788,3.218],[-56.912,-58.902,1.906],
[-58.112,-58.350,-0.543],[-51.248,-58.396,2.845],[-46.542,-58.284,4.788],
[-46.070,-58.302,2.025],[-56.947,-42.966,-0.554],[-56.491,-56.863,-2.624],
[-49.785,-57.074,0.472],[-49.532,-42.480,-1.619],[-41.330,-43.267,1.473],
[-56.156,-57.525,13.154],[-53.797,-58.231,14.156],[-57.083,-41.125,-2.609],
[-44.881,58.064,15.283],[-43.218,-43.425,14.121],[-46.943,-32.544,3.246],
[-48.713,-31.754,6.491],[-56.154,-32.635,7.464],[-44.725,-29.270,5.320],
[-54.453,-28.258,7.540],[-58.076,-33.428,2.475],[-57.536,-31.214,5.571],
[-47.195,-24.615,5.157],[-49.735,-26.950,7.599],[-41.787,-37.977,8.830],
[-47.186,-35.559,6.771],[-54.202,-35.409,4.798],[-46.455,-37.753,2.990],
[-56.386,-33.850,4.866],[-52.898,-32.725,4.230],[-40.888,-34.289,10.591],
[-55.882,-37.332,9.897],[-49.103,-28.348,2.522],[-51.462,-29.800,7.854],
[-42.989,-37.251,6.113],[-55.748,-35.698,7.086],[-53.784,-27.701,10.867],
[-58.817,-30.101,1.168],[-44.655,-27.761,7.835],[-44.186,-37.719,10.147],
[-51.682,-41.609,5.773],[-45.097,-26.340,4.751],[-46.038,-28.574,2.181],
[-44.771,-29.274,11.668],[-49.377,-30.837,1.878],[-49.102,-31.483,-4.069],
[-42.655,-30.540,9.769],[-44.414,-31.352,2.374],[-46.797,-33.442,10.854],
[-53.480,-31.123,2.002],[-52.002,-27.475,6.180],[-52.765,-32.440,7.649],
[-47.083,-26.063,7.608],[-51.507,-26.920,12.116],[-46.825,-32.515,-0.478],
[-54.965,-33.105,11.030],[-49.577,-24.448,4.305],[-54.631,-26.237,13.755],
[-42.403,-39.183,-2.070],[-56.497,-30.731,2.676],[-41.352,-26.557,-1.122],
[-58.681,-29.604,10.897],[-59.666,-25.561,7.481],[-39.717,-25.621,13.106],
[-48.144,-26.888,10.039],[-49.778,-22.082,1.651],[-54.508,-23.999,4.000],
[-45.061,-22.923,-1.181],[-46.927,-24.109,1.059],[-43.260,-34.170,13.384],
[-56.550,-27.686,10.585],[-53.509,-39.875,7.783],[-49.117,-40.861,9.996],
[-45.319,-35.881,12.318],[-44.007,-32.289,-2.117],[-50.050,-37.047,-1.561],
[-53.691,-34.705,8.748],[-55.188,-33.027,2.585],[-48.841,-33.250,1.183],
[-48.933,-33.484,13.442],[-48.484,-35.266,4.099],[-56.459,-34.485,9.577],
[-50.562,-36.925,-5.289],[-57.899,-36.542,4.385],[-57.746,-35.822,-1.906],
[-48.419,-33.840,8.563],[-51.490,-36.556,4.490],[-55.454,-35.888,2.296],
[-52.780,-35.995,0.942],[-46.438,-38.089,7.955],[-54.322,-33.791,0.023],
[-43.385,-22.239,13.317],[-55.297,-29.014,12.896],[-51.110,-34.144,5.739],
[-53.982,-38.425,2.390],[-51.154,-34.984,8.307],[-52.153,-29.332,0.670],
[-43.917,-30.955,7.339],[-47.299,-37.457,-1.403],[-48.792,-28.626,12.294],
[-47.464,-35.622,1.360],[-54.620,-29.924,4.652],[-48.434,-29.970,-0.781],
[-57.946,-32.251,10.681],[-51.623,-32.329,13.117],[-46.139,-26.935,11.787],
[-42.414,-33.474,0.744],[-40.970,-28.946,11.698],[-43.605,-26.231,1.088],
[-47.275,-28.621,7.612],[-53.364,-26.904,3.845],[-57.477,-30.828,-1.192],
[-51.615,-29.190,3.553],[-51.029,-25.057,9.339],[-51.848,-23.742,2.798],
[-49.474,-29.842,9.846],[-45.887,-28.978,-2.266],[-45.298,-24.393,12.629],
[-41.906,-23.956,-2.165],[-58.018,-36.213,11.192],[-45.214,-30.434,-0.187],
[-55.487,-25.516,9.170],[-51.004,-27.443,-3.703],[-52.574,-27.950,-1.601],
[-54.866,-29.272,-0.233],[-48.936,-37.305,8.674],[-48.436,-38.050,0.996],
[-49.239,-26.220,0.376],[-50.426,-35.578,1.831],[-50.476,-36.026,10.601],
[-50.033,-38.248,3.029],[-50.878,-25.177,-1.628],[-53.376,-38.500,-1.899],
[-42.694,-36.374,11.998],[-46.059,-33.272,7.339],[-55.421,-26.584,0.114],
[-56.223,-33.476,-2.111],[-43.597,-23.816,8.638],[-55.720,-40.071,6.048],
[-58.765,-26.240,-0.165],[59.159,-23.365,13.133],[-52.479,-24.820,15.074],
[-49.077,-23.129,-3.473],[-47.465,-25.913,14.765],[-43.807,-25.745,-2.377],
[-58.203,-29.799,13.854],[-53.650,-31.016,-2.929],[-51.717,-29.732,13.224],
[-48.534,-27.978,-2.444],[-47.441,-30.760,11.328],[-55.727,-31.796,13.754],
[-51.685,-31.537,-0.880],[-52.059,-33.603,10.863],[-51.697,-31.796,-4.387],
[-48.470,-30.907,13.781],[-42.993,-29.729,-2.448],[-51.286,-35.230,12.979],
[-52.846,-36.937,10.103],[-49.496,-32.245,10.839],[-45.973,-34.755,14.641],
[-45.655,-31.821,12.890],[-54.264,-34.274,13.482],[-51.544,-39.470,10.132],
[-52.225,-37.360,14.494],[-48.333,-36.149,12.366],[-46.324,-37.508,14.045],
[-55.800,-36.811,13.707],[-53.220,-41.069,13.101],[-51.034,-40.993,14.663],
[-52.148,-22.436,-1.977],[-57.546,-27.581,-2.089],[58.542,-22.778,3.363],
[-58.895,-24.447,5.033],[-41.541,-25.641,9.566],[-41.012,-23.377,10.937],
[-42.737,-28.547,13.649],[-42.446,-29.490,0.490],[-41.647,-31.659,2.394],
[-41.616,-27.994,8.164],[-43.333,-27.514,10.471],[-41.400,-32.019,-1.968],
[-44.535,-34.460,-0.730],[-44.178,-33.874,3.483],[-58.485,-31.181,8.286],
[-42.623,-31.138,12.392],[-41.703,-31.282,14.797],[-44.163,-37.204,-2.353],
[-45.106,-36.699,0.297],[-44.048,-34.682,6.150],[-43.268,-35.167,8.731],
[-58.726,-36.171,14.137],[-43.375,-40.513,15.741],[-43.352,-38.520,1.885],
[-58.226,-38.409,6.735],[-58.218,-34.472,7.451],[-41.262,-38.887,12.230],
[-41.128,-35.549,-1.023],[-58.928,-38.252,2.220],[-44.233,-23.138,1.658],
[-58.688,-38.685,10.241],[-57.150,-26.859,14.126],[-53.190,-23.639,10.213],
[-48.234,-23.630,13.311],[-46.166,-21.851,11.847],[-42.457,-40.395,10.259],
[-57.679,-22.624,10.410],[-55.584,-22.053,8.583],[-50.768,-22.394,8.873],
[-46.143,-23.883,9.866],[-44.955,-19.985,8.596],[-56.903,-24.352,7.035],
[-54.365,-25.857,6.190],[-51.622,-24.858,5.957],[-48.344,-37.662,5.339],
[-43.716,-43.950,6.318],[-42.082,-39.744,3.739],[-56.336,-38.432,1.286],
[-58.976,-39.768,-0.237],[-51.799,-40.412,3.326],[-46.481,-40.680,6.645],
[-46.544,-39.859,1.263],[-56.709,-24.693,2.375],[-56.952,-38.677,-1.625],
[-51.314,-38.440,0.636],[-48.978,-22.336,-0.822],[-41.053,-24.837,1.449],
[-57.350,-39.091,13.368],[-54.035,-41.147,15.890],[-56.146,-24.785,-2.380],
[-45.911,-40.209,14.690],[-42.471,-25.726,12.645],[-49.603,-12.711,4.553],
[-48.460,-14.058,6.863],[-55.546,-14.574,8.682],[-44.997,-10.002,5.363],
[-55.380,-9.082,7.382],[-59.517,-15.116,1.984],[-59.050,-12.456,4.142],
[-48.399,-6.018,4.440],[-48.673,-10.837,6.311],[-41.253,-18.672,6.759],
[-46.574,-16.464,6.703],[-53.846,-17.849,3.744],[-46.598,-19.567,3.436],
[-57.547,-14.748,4.631],[-52.436,-14.939,4.534],[-41.514,-16.317,10.224],
[-54.719,-18.835,9.380],[-47.633,-9.954,3.999],[-52.036,-10.306,10.169],
[-44.140,-19.403,4.955],[-55.443,-18.310,6.686],[-53.692,-10.798,8.214],
[-57.998,-9.992,1.382],[-45.274,-9.057,8.841],[-43.677,-18.268,10.148],
[-51.687,-22.299,5.625],[-46.039,-7.397,5.353],[-45.947,-8.153,1.924],
[-44.734,-10.131,11.645],[-48.104,-12.162,2.126],[-47.828,-14.596,-2.546],
[-41.096,-13.961,9.124],[-44.648,-13.397,1.167],[-46.221,-14.805,11.030],
[-55.008,-11.499,3.330],[-52.706,-9.553,5.625],[-53.974,-13.134,6.581],
[-47.952,-8.825,8.929],[-52.667,-8.783,12.394],[-46.927,-12.649,-0.359],
[-54.621,-11.824,11.183],[-50.440,-8.054,3.355],[-54.672,-6.402,12.128],
[-42.287,-22.275,-0.097],[-54.397,-13.641,1.630],[-42.303,-6.567,-2.408],
[-59.137,-11.052,9.901],[-57.970,-9.708,7.842],[-40.717,-7.840,11.306],
[-50.180,-8.494,10.440],[-49.026,-4.682,1.650],[-55.653,-5.332,4.003],
[-45.468,-5.384,-2.070],[-46.363,-4.666,2.174],[-43.298,-14.612,11.689],
[-56.467,-9.569,11.447],[-52.777,-20.791,7.849],[-48.691,-22.697,10.498],
[-44.286,-16.879,12.618],[-45.672,-12.325,-2.986],[-48.934,-18.476,-0.843],
[-53.808,-16.593,8.728],[-55.180,-15.413,3.587],[-48.308,-14.672,1.430],
[-49.370,-16.126,13.719],[-46.215,-16.956,3.384],[-56.792,-17.359,10.674],
[-53.943,-17.270,-2.779],[-58.244,-17.120,3.400],[-58.373,-17.570,0.152],
[-48.524,-16.051,8.775],[-51.035,-18.177,5.648],[-56.124,-17.749,1.962],
[-53.457,-17.317,0.870],[-47.545,-18.925,8.123],[-55.270,-14.772,-2.311],
[-43.509,-3.308,12.629],[-55.855,-11.083,13.533],[-50.800,-15.418,6.676],
[-53.950,-19.997,1.600],[-50.752,-17.475,9.133],[-50.799,-9.857,0.984],
[-44.579,-12.310,9.621],[-46.497,-17.572,-1.842],[-49.199,-10.679,11.729],
[-44.970,-14.669,-1.159],[-56.499,-12.106,5.462],[-48.106,-9.560,1.044],
[-57.717,-14.055,10.558],[-50.246,-13.412,12.810],[-47.123,-8.071,11.539],
[-41.714,-15.859,-0.102],[-41.091,-10.710,11.827],[-43.355,-8.695,1.127],
[-46.218,-12.319,6.194],[-53.015,-6.236,3.689],[-58.192,-11.548,-1.063],
[-51.924,-11.149,3.622],[-51.383,-5.682,10.102],[-53.188,-3.625,1.817],
[-47.267,-11.669,8.662],[-46.619,-10.038,-1.896],[-44.097,-5.765,8.880],
[-40.356,-4.610,-0.915],[-57.114,-19.921,9.552],[-44.670,-11.002,-0.028],
[-57.410,-7.201,10.121],[-51.943,-8.631,-4.082],[-53.405,-9.278,0.160],
[-55.570,-11.301,0.696],[-49.220,-19.368,10.372],[-46.649,-18.743,0.666],
[-49.595,-7.410,0.523],[-50.097,-16.879,1.329],[-50.384,-15.668,11.211],
[-49.161,-19.433,1.804],[-51.995,-6.078,0.952],[-54.197,-19.133,-0.916],
[-44.308,-19.592,12.582],[-44.396,-15.096,6.108],[-56.469,-8.655,-1.897],
[-57.257,-15.512,-4.677],[-45.078,-5.282,6.581],[-54.505,-22.046,5.930],
[-55.205,-8.115,4.275],[-57.467,-6.212,13.395],[-53.626,-5.496,15.132],
[-49.697,-7.122,-3.386],[-46.522,-10.059,13.528],[-44.775,-7.883,-2.084],
[-58.691,-11.520,13.952],[-54.174,-9.239,-3.129],[-52.505,-11.792,13.319],
[-49.566,-10.936,-3.266],[-46.923,-12.337,11.936],[-56.355,-13.762,12.845],
[-53.699,-12.692,-1.205],[-54.790,-16.153,11.716],[-53.224,-12.792,-3.823],
[-48.773,-11.546,14.441],[-41.715,-11.420,-2.677],[-53.919,-19.360,12.454],
[-52.206,-20.016,10.345],[-52.653,-14.307,10.298],[-46.465,-16.136,15.344],
[-45.646,-14.444,13.533],[-54.925,-15.381,15.003],[-52.694,-21.813,12.276],
[-51.631,-17.570,14.792],[-46.630,-17.885,11.852],[-46.278,-19.274,14.518],
[-55.982,-17.880,14.268],[-50.985,-22.477,14.537],[-51.234,-19.545,13.183],
[-49.929,-3.738,-2.728],[-59.149,-8.606,-1.811],[58.611,-2.803,2.952],
[-58.438,-5.045,3.609],[-41.157,-6.271,7.339],[-42.757,-6.074,11.325],
[-43.729,-9.752,14.540],[-41.985,-11.666,0.026],[-42.365,-13.160,2.944],
[-42.973,-9.990,7.146],[-42.591,-8.161,9.125],[-42.414,-13.831,-1.846],
[-41.926,-16.416,-2.866],[-43.711,-15.856,3.376],[-57.053,-12.320,8.621],
[-42.629,-12.898,13.926],[-41.014,-15.246,14.256],[-44.483,-19.553,-3.344],
[-45.444,-16.163,1.087],[-43.533,-17.272,7.407],[-44.587,-14.964,8.930],
[-58.945,-17.862,13.316],[-41.221,-21.578,15.219],[-43.900,-20.296,1.415],
[-58.105,-19.566,5.767],[-57.839,-15.897,7.599],[-39.830,-19.639,13.655],
[-43.208,-18.889,-1.227],[-58.896,-20.382,1.624],[-44.165,-6.035,2.893],
[58.736,-19.865,10.142],[-55.738,-8.306,14.116],[-54.488,-7.078,9.487],
[-48.227,-6.182,13.758],[-46.244,-3.049,12.670],[-42.405,-20.990,11.086],
[-55.979,-3.699,8.773],[-53.637,-4.290,10.517],[-48.537,-4.041,8.073],
[-46.439,-5.481,9.775],[-44.609,-3.315,9.890],[-57.001,-5.905,7.365],
[-54.290,-6.353,6.150],[-51.180,-6.747,5.887],[-49.065,-20.510,6.527],
[-44.482,-23.446,5.297],[-42.597,-21.904,3.700],[-55.830,-20.210,3.833],
[-59.302,-21.150,-1.067],[-51.403,-20.502,3.275],[-46.494,-21.862,5.774],
[-46.435,-21.452,0.591],[-57.958,-7.104,0.280],[-56.225,-19.708,-2.661],
[-51.567,-20.179,-0.030],[-47.989,-4.702,-1.217],[-42.077,-6.017,0.485],
[-55.777,-20.854,13.621],[-53.005,-20.285,15.127],[-55.452,-5.901,-2.854],
[-46.105,-22.297,15.096],[-44.983,-5.865,12.916],[-47.200,6.675,4.960],
[-51.074,5.637,6.857],[-55.584,6.017,7.927],[-44.714,7.713,5.062],
[-53.661,10.252,6.371],[-58.781,3.518,1.838],[-56.311,5.140,5.165],
[-47.338,13.095,5.955],[-49.002,9.002,6.879],[-42.474,-0.813,9.234],
[-47.737,0.924,7.215],[-54.115,0.165,4.437],[-46.631,-0.938,3.466],
[-58.382,1.930,5.173],[-52.979,2.511,5.285],[-40.723,2.171,10.757],
[-54.601,0.112,9.361],[-48.912,8.391,2.974],[-52.671,5.547,9.150],
[-44.564,-1.317,5.177],[-55.504,1.741,6.242],[-54.164,7.588,10.901],
[-57.606,8.489,0.613],[-44.894,7.925,8.154],[-44.915,-0.464,7.908],
[-50.654,-3.917,5.627],[-44.777,11.189,4.774],[-45.219,10.572,1.343],
[-44.303,8.629,11.834],[-49.513,5.684,2.158],[-47.481,4.252,-2.021],
[-41.628,5.879,8.981],[-45.765,5.240,2.955],[-44.101,3.657,11.369],
[-54.265,6.689,3.437],[-51.151,10.288,5.554],[-53.475,5.361,5.645],
[-46.009,10.100,8.695],[-50.724,11.073,12.196],[-44.999,4.127,-0.712],
[-55.188,4.375,12.125],[-47.855,11.887,3.264],[-53.758,10.805,12.122],
[-43.667,-4.064,1.046],[-53.182,6.515,0.944],[-43.901,11.035,-1.818],
[-58.871,6.916,10.708],[-59.584,10.276,7.036],[-40.483,11.656,12.901],
[-48.493,11.244,9.686],[-48.985,15.378,0.998],[-54.798,15.191,4.110],
[-45.920,13.004,-2.033],[-47.060,13.622,0.372],[-42.923,4.409,13.636],
[-57.020,8.766,10.384],[-53.878,-1.398,6.658],[-48.530,-3.036,11.193],
[-43.268,0.769,13.283],[-44.443,5.980,-2.634],[-49.000,1.622,-2.746],
[-53.764,3.464,7.856],[-53.452,3.760,2.669],[-48.350,4.109,0.464],
[-49.610,3.540,14.218],[-47.062,4.149,5.009],[-56.010,2.432,9.001],
[-53.236,0.207,-2.952],[-59.084,-0.395,3.805],[-57.201,0.345,-0.100],
[-48.204,3.382,7.704],[-51.306,0.870,3.568],[-55.997,1.906,2.599],
[-52.950,0.621,-0.214],[-47.335,-1.613,8.413],[-54.977,4.273,-0.964],
[-43.688,15.370,14.464],[-56.551,6.881,12.650],[-50.539,2.800,6.834],
[-54.900,-0.892,2.067],[-50.983,2.047,9.432],[-51.175,8.061,-0.221],
[-44.026,5.408,7.673],[-46.778,-0.440,-1.692],[-48.931,8.534,11.362],
[-46.592,1.992,-0.697],[-56.184,7.855,5.647],[-48.266,9.174,-0.991],
[-57.730,4.797,9.353],[-51.721,5.260,13.888],[-47.035,10.694,11.616],
[-41.199,2.853,0.326],[-41.993,8.469,10.091],[-42.710,11.398,0.450],
[-46.627,5.570,7.773],[-53.977,11.655,1.852],[-57.955,4.852,-2.590],
[-51.947,8.311,3.291],[-51.304,11.701,9.636],[-52.785,13.912,2.961],
[-48.213,7.499,8.726],[-47.114,7.546,-3.459],[-44.163,12.578,12.237],
[-41.443,12.624,-1.533],[-57.284,-0.250,9.536],[-45.446,7.855,-0.814],
[-57.429,10.672,8.604],[-51.099,9.973,-4.672],[-52.536,10.275,-1.548],
[-55.198,7.416,-1.269],[-49.931,-0.911,10.466],[-47.439,-1.449,0.852],
[-48.963,11.699,0.145],[-49.989,2.255,1.457],[-50.008,3.518,11.608],
[-49.368,-0.442,2.462],[-51.059,12.974,1.073],[-52.389,-2.311,-0.606],
[-44.864,-0.872,10.850],[-43.788,4.079,5.022],[-55.033,10.126,0.053],
[-57.538,1.865,-2.365],[-44.317,13.010,7.421],[-56.775,-1.906,6.973],
[-57.532,10.319,2.594],[-56.680,13.848,13.713],[-53.301,12.160,14.907],
[-49.798,11.130,-2.612],[-46.347,9.875,13.887],[-45.339,10.923,-3.900],
[59.409,4.726,12.179],[-53.418,8.082,-3.460],[-51.785,8.108,12.687],
[-50.150,7.391,-3.481],[-46.217,7.547,10.575],[-57.087,6.292,15.385],
[-52.078,5.757,-1.279],[-52.664,3.384,12.214],[-52.845,3.247,-2.230],
[-48.247,6.492,13.417],[-41.454,6.960,-3.125],[-53.306,0.794,11.722],
[-52.075,-0.165,8.597],[-49.940,5.022,9.534],[-45.166,2.727,14.005],
[-46.878,4.595,11.726],[-55.036,3.855,14.682],[-51.087,-2.956,9.134],
[-49.228,0.410,14.300],[-47.428,0.877,10.785],[-46.284,-0.120,13.063],
[-55.807,1.329,14.104],[-52.069,-4.803,12.811],[-50.973,-2.396,14.486],
[-52.028,13.261,-1.400],[-58.535,8.750,-2.550],[-59.653,14.592,2.610],
[-57.634,12.622,3.976],[-41.534,14.182,8.187],[-41.817,13.277,11.309],
[-43.166,9.823,14.333],[-42.276,8.071,-0.324],[-42.144,5.114,1.647],
[-42.477,8.613,7.294],[-43.089,11.098,10.136],[-42.231,5.125,-1.115],
[-41.683,2.652,-2.501],[-43.880,2.536,2.643],[58.734,7.451,7.415],
[-43.110,6.127,11.228],[-40.269,4.318,12.409],[-43.560,1.220,-4.120],
[-43.459,1.810,-0.625],[-41.886,0.352,4.862],[-43.536,1.626,9.557],
[-58.771,2.003,13.087],[-40.206,-1.732,15.974],[-43.345,-1.368,1.858],
[-56.677,-0.311,4.681],[-57.200,3.554,6.930],[-41.253,-0.396,11.765],
[-40.393,-2.273,-2.262],[-59.605,-0.214,0.886],[-43.588,13.474,3.394],
[-57.881,-2.565,10.595],[-57.056,10.923,13.421],[-54.947,10.330,9.429],
[-49.356,13.387,12.994],[-46.569,13.562,12.945],[-41.854,-3.610,10.706],
[-56.924,13.770,9.493],[-53.863,14.123,9.976],[-49.583,12.834,7.716],
[-45.837,12.420,9.821],[-45.469,15.127,9.637],[-57.328,12.595,6.750],
[-54.569,12.703,5.438],[-52.890,13.014,7.689],[-48.884,-0.910,5.443],
[-43.030,-3.173,6.532],[-40.441,-2.859,4.686],[-56.829,-2.696,3.492],
[-59.201,-2.003,-1.067],[-51.051,-2.546,3.223],[-47.447,-3.447,5.459],
[-44.842,-2.759,-1.215],[-57.067,13.046,1.121],[-55.810,-0.538,-2.826],
[-50.259,-0.339,-0.059],[-49.103,14.602,-2.602],[-41.951,11.147,3.118],
[-56.547,-2.295,13.045],[-53.893,-2.990,13.673],[-55.492,12.404,-2.621],
[-46.592,-2.130,15.369],[-42.415,12.593,14.627],[-45.330,25.019,3.844],
[-48.216,24.254,6.569],[-56.049,23.063,6.991],[-44.167,27.263,5.903],
[-55.475,28.434,6.920],[-56.245,22.647,0.839],[-58.025,23.995,5.051],
[-46.463,31.394,5.895],[-50.541,26.032,8.626],[-41.774,18.245,9.268],
[-46.139,20.166,8.817],[-52.750,17.962,5.550],[-45.811,17.971,3.142],
[-57.561,21.392,5.674],[-51.019,21.729,3.686],[-41.563,20.841,9.393],
[-54.771,17.333,9.003],[-50.044,26.917,4.198],[-52.359,25.101,10.544],
[-44.321,17.441,5.428],[-55.663,19.928,6.746],[-53.291,27.260,8.495],
[-59.207,27.484,0.027],[-45.635,27.893,8.667],[-43.885,16.946,8.210],
[-50.573,15.204,6.536],[-44.971,29.425,4.322],[-46.033,28.442,1.989],
[-45.154,27.303,11.255],[-47.331,25.998,2.115],[-50.572,20.937,-3.367],
[-41.290,22.963,10.893],[-44.908,25.266,0.016],[-45.389,22.238,10.866],
[-55.189,25.043,4.631],[-52.325,26.971,5.860],[-53.019,23.869,7.053],
[-48.190,29.791,8.655],[-52.785,27.605,11.247],[-46.971,23.703,-0.009],
[-55.084,24.258,12.012],[-48.695,29.264,2.887],[-54.505,28.822,12.711],
[-41.554,15.973,-0.692],[-53.840,23.377,2.266],[-42.646,29.497,-1.757],
[-56.739,25.854,9.946],[-59.610,29.881,6.555],[-39.497,29.539,13.441],
[-50.404,29.629,11.432],[-49.230,32.581,2.506],[-55.494,32.891,3.915],
[-46.293,32.606,-1.633],[-46.866,31.855,0.864],[-43.491,22.769,13.470],
[-55.562,28.796,10.118],[-53.616,15.621,6.935],[-48.103,13.903,10.186],
[-46.132,19.598,11.757],[-45.288,25.019,-2.527],[-49.042,18.905,-1.972],
[-54.160,22.962,9.473],[-54.812,21.758,4.274],[-48.585,22.317,1.458],
[-47.601,22.503,12.968],[-47.065,21.422,4.453],[-55.564,21.031,11.610],
[-51.112,20.683,-0.491],[-59.182,19.119,3.826],[-57.619,19.590,-0.422],
[-48.614,21.892,9.039],[-50.726,19.834,5.633],[-56.225,18.995,4.151],
[-52.669,20.262,2.197],[-46.898,17.316,8.267],[-53.314,22.694,-0.593],
[-43.663,35.031,14.238],[-53.273,25.808,13.322],[-50.374,22.489,6.489],
[-54.652,18.416,1.813],[-52.646,20.176,9.916],[-49.937,26.606,0.267],
[-42.373,24.510,8.530],[-46.532,18.172,-1.393],[-48.420,27.636,11.224],
[-45.606,20.696,-0.151],[-57.137,26.278,7.016],[-47.325,27.101,-0.270],
[-56.963,23.008,10.196],[-50.526,22.536,13.025],[-47.042,29.709,11.506],
[-41.554,21.698,-0.610],[-41.666,28.053,12.808],[-41.933,31.168,0.539],
[-44.934,24.288,6.692],[-51.986,27.913,2.042],[-56.594,25.088,-1.692],
[-51.575,24.392,4.931],[-51.347,31.414,9.629],[-53.080,30.838,3.129],
[-47.573,25.732,8.908],[-46.625,27.536,-2.903],[-45.535,31.859,11.702],
[-42.745,32.364,-1.932],[-56.411,18.031,10.920],[-44.547,28.047,-0.280],
[-58.026,30.162,9.061],[-51.312,27.934,-3.269],[-53.911,27.299,-0.797],
[-54.580,24.904,0.187],[-49.146,18.379,10.163],[-46.263,16.372,1.114],
[-49.274,29.774,0.054],[-47.442,19.798,1.606],[-50.211,20.595,11.110],
[-48.774,17.541,3.306],[-52.359,30.002,-0.285],[-53.625,17.396,-0.444],
[-42.381,18.211,12.305],[-44.186,21.793,6.213],[-55.457,29.243,-0.693],
[-55.394,22.485,-2.351],[-44.352,30.777,7.613],[-56.399,15.311,6.214],
[-56.746,28.503,2.279],[-57.447,32.211,13.390],[-53.736,28.657,16.074],
[-49.156,30.850,-3.563],[-46.899,27.243,13.552],[-45.103,30.016,-3.150],
[-57.648,25.215,12.658],[-54.853,26.212,-3.997],[-50.267,26.255,13.117],
[-48.895,26.677,-4.043],[-45.197,24.669,9.541],[-57.841,22.741,13.512],
[-50.617,23.745,-1.072],[-52.946,22.343,11.933],[-51.058,23.804,-3.820],
[-48.127,24.761,14.489],[-43.158,26.852,-2.238],[-50.895,18.292,12.493],
[-52.473,17.368,10.187],[-51.064,23.192,9.152],[-45.271,21.378,16.049],
[-45.738,24.663,13.016],[-54.722,21.895,13.873],[-51.243,14.856,10.735],
[-50.197,19.484,15.118],[-48.417,20.022,12.990],[-46.492,18.718,14.410],
[-56.526,19.832,13.843],[-51.285,14.930,14.130],[-48.694,17.158,15.343],
[-50.941,32.307,-0.890],[-57.196,28.260,-2.571],[-59.153,33.531,2.488],
[-58.258,30.826,3.715],[-42.103,32.245,7.857],[-41.271,31.346,10.923],
[-43.965,27.547,14.618],[-42.122,25.507,0.680],[-42.278,23.803,2.629],
[-41.682,26.896,7.329],[-43.720,29.226,10.033],[-41.926,23.847,-2.187],
[-41.511,19.909,-2.999],[-44.331,22.035,3.468],[59.231,25.759,6.290],
[-43.151,25.569,12.411],[-40.802,20.286,11.897],[-44.324,19.011,-2.884],
[-41.642,20.957,2.385],[-43.493,19.714,7.971],[-43.522,20.489,11.441],
[-59.545,19.996,13.464],[-40.024,16.315,15.161],[-43.583,16.944,2.243],
[-57.772,17.380,5.444],[59.319,23.137,9.955],[-41.456,15.799,13.005],
[-42.358,17.166,-4.016],[-58.225,17.978,1.715],[-45.562,32.121,3.246],
[-58.536,16.366,10.090],[-58.139,29.521,13.497],[-54.649,31.522,12.339],
[-49.851,32.222,12.532],[-47.522,34.169,12.670],[-42.555,15.713,10.558],
[-57.202,32.082,10.591],[-53.666,33.455,9.930],[-48.723,33.341,7.153],
[-47.305,32.321,9.376],[-44.317,33.707,10.093],[-58.045,32.010,6.242],
[-55.326,30.169,4.753],[-52.366,30.061,5.701],[-48.015,18.174,5.751],
[-42.967,15.357,6.100],[-41.307,16.019,3.590],[-56.548,16.329,2.878],
[-57.703,15.627,0.192],[-50.521,15.463,3.434],[-45.408,15.058,5.001],
[-44.494,14.285,0.880],[-57.854,30.040,-0.059],[-55.552,15.290,-1.459],
[-50.570,17.838,0.336],[-49.358,33.449,-2.473],[-42.095,31.563,3.196],
[-55.579,16.621,13.399],[-53.548,15.495,15.219],[-56.589,31.517,-2.691],
[-46.224,15.675,14.913],[-43.320,30.414,12.453],[-47.552,41.565,2.674],
[-47.566,41.813,5.868],[-56.228,41.631,7.369],[-44.050,45.280,3.680],
[-54.144,47.203,6.055],[-56.897,40.916,1.469],[-58.149,42.998,5.463],
[-47.694,51.538,4.738],[-50.840,46.475,8.654],[-41.417,37.046,8.503],
[-45.555,38.671,7.633],[-53.706,37.589,3.468],[-46.727,36.479,3.360],
[-56.849,40.658,4.440],[-50.643,42.604,4.749],[-40.826,39.781,9.611],
[-55.252,37.479,8.249],[-48.687,46.505,1.256],[-52.740,43.823,10.954],
[-43.038,36.085,4.380],[-55.137,39.366,6.010],[-55.479,46.493,8.849],
[-58.964,44.481,0.601],[-44.531,46.204,6.809],[-44.357,36.571,9.055],
[-51.307,33.211,6.546],[-44.937,48.788,4.375],[-46.018,47.303,2.623],
[-44.772,45.726,11.863],[-48.217,43.407,0.683],[-48.863,41.550,-1.267],
[-41.852,42.658,8.834],[-44.684,42.823,1.494],[-46.397,39.855,10.515],
[-54.052,42.624,4.137],[-51.863,46.105,5.454],[-52.440,43.641,7.054],
[-46.325,48.578,7.620],[-52.659,47.276,12.280],[-46.135,43.605,-1.023],
[-55.718,42.740,12.041],[-47.925,49.040,3.718],[-55.139,48.476,12.732],
[-44.089,34.771,-1.370],[-53.350,41.822,0.279],[-42.292,49.022,-1.407],
[-57.887,45.172,10.112],[-58.876,47.287,7.603],[-40.738,47.975,12.471],
[-50.129,47.880,11.550],[-48.196,53.312,0.050],[-54.843,50.538,3.246],
[-46.221,50.984,-1.082],[-47.181,50.794,1.573],[-43.918,40.605,12.440],
[-56.651,46.079,12.389],[-53.372,35.411,7.116],[-47.945,34.903,9.823],
[-46.758,38.814,13.069],[-45.313,43.205,-3.733],[-49.049,36.665,-1.799],
[-53.152,40.121,7.920],[-54.652,40.198,2.860],[-48.951,39.588,0.870],
[-48.517,41.083,13.364],[-46.527,39.197,5.233],[-55.586,39.251,10.981],
[-51.485,37.516,-1.961],[-58.840,38.687,4.003],[-58.377,38.712,-0.288],
[-47.826,40.583,8.114],[-51.978,36.199,4.569],[-56.299,37.761,4.070],
[-53.600,38.914,0.870],[-47.895,37.485,8.358],[-52.986,39.911,-2.116],
[-43.964,52.021,12.838],[-54.199,45.024,12.868],[-49.792,40.192,6.099],
[-55.188,36.843,1.169],[-51.045,39.053,9.147],[-50.642,44.873,-0.242],
[-44.370,43.336,8.612],[-45.566,37.247,-1.855],[-48.485,45.413,12.041],
[-46.080,39.789,-0.135],[-54.803,44.704,6.062],[-48.579,46.566,-1.451],
[-58.209,41.682,11.133],[-50.795,42.336,13.615],[-46.323,47.906,10.576],
[-41.086,40.226,-0.351],[-41.067,44.750,12.709],[-43.402,47.237,0.515],
[-47.220,44.840,5.402],[-53.825,47.193,3.278],[-57.188,44.051,-1.739],
[-51.091,44.946,2.751],[-49.270,48.699,8.655],[-52.552,49.566,2.196],
[-48.548,44.571,8.269],[-46.077,45.990,-2.698],[-44.233,49.336,11.177],
[-40.821,51.249,-2.236],[-55.920,36.619,11.979],[-46.256,45.382,1.098],
[-57.005,48.736,9.316],[-51.838,47.163,-3.537],[-53.042,46.599,-0.138],
[-55.341,44.097,0.085],[-49.833,36.733,9.922],[-47.506,36.772,0.646],
[-48.550,49.282,-0.102],[-51.141,39.237,2.691],[-50.567,38.577,12.178],
[-49.961,35.190,3.071],[-52.429,48.907,-1.335],[-53.531,34.813,-0.302],
[-43.485,36.349,11.899],[-44.735,41.420,6.448],[-54.469,46.894,-2.767],
[-55.733,39.972,-3.066],[-44.477,51.300,7.970],[-55.827,35.021,5.595],
[-55.612,48.242,1.741],[-58.458,50.721,14.411],[-53.690,50.564,13.955],
[-49.832,48.887,-2.737],[-47.245,48.639,13.826],[-45.527,49.074,-3.127],
[-58.418,44.022,14.947],[-54.838,44.346,-3.164],[-51.022,45.144,12.798],
[-48.797,45.903,-4.114],[-46.636,42.968,10.277],[-57.173,41.746,14.153],
[-51.724,42.964,-2.152],[-53.504,41.212,10.432],[-52.502,39.839,-5.047],
[-47.635,43.103,15.347],[-43.206,44.539,-2.738],[-53.022,37.748,12.201],
[-52.160,35.783,10.935],[-49.374,42.261,9.694],[-44.732,39.617,15.001],
[-45.986,43.708,13.170],[-54.550,40.310,13.566],[-51.409,34.401,8.778],
[-50.548,38.689,15.016],[-48.514,36.716,13.868],[-45.938,36.469,14.596],
[-55.592,36.634,14.520],[-52.461,33.588,12.341],[-51.723,36.025,14.415],
[-51.656,51.323,-0.508],[-58.893,46.204,-1.696],[-58.664,51.515,2.182],
[-59.389,50.413,4.612],[-41.739,50.323,7.158],[-40.927,49.580,9.806],
[-43.855,45.367,15.413],[-41.779,45.640,-0.657],[-41.817,42.418,1.510],
[-42.159,45.091,7.627],[-43.501,46.941,9.715],[-41.276,42.847,-1.827],
[-42.698,39.647,-2.425],[-42.505,41.287,4.182],[-58.334,44.541,7.543],
[-42.389,42.678,11.839],[-40.706,41.343,13.882],[-43.598,37.078,-3.479],
[-44.189,38.030,1.393],[-43.692,39.063,5.264],[-43.942,40.355,9.469],
[-57.479,38.959,13.124],[-39.700,34.108,15.410],[-42.768,35.224,1.505],
[-58.515,37.410,7.351],[-58.637,39.473,9.169],[-41.069,35.265,12.866],
[-41.110,35.366,-2.696],[-58.034,36.915,1.893],[-44.602,50.040,1.879],
[-57.451,35.616,9.839],[-57.997,47.937,14.026],[-54.388,50.295,10.722],
[-49.165,50.447,13.384],[-46.572,51.401,11.873],[-42.271,34.683,9.084],
[-56.803,50.414,11.823],[-55.569,51.294,8.318],[-48.725,51.611,7.370],
[-46.977,50.635,9.255],[-45.224,52.927,9.914],[-56.574,49.271,5.222],
[-53.877,49.772,5.703],[-51.515,49.899,7.004],[-49.870,35.851,6.384],
[-44.251,34.083,7.356],[-42.596,33.373,4.898],[-58.084,35.423,4.227],
[-57.463,35.231,-0.864],[-51.944,33.040,4.003],[-46.578,34.844,5.415],
[-45.669,34.345,0.838],[-56.886,50.564,-0.407],[-55.019,36.941,-1.438],
[-50.222,34.728,0.121],[-49.170,51.579,-1.778],[-40.859,49.386,2.270],
[-55.540,34.161,13.516],[-52.432,33.355,15.343],[-54.779,49.771,-2.196],
[-46.281,32.134,14.406],[-42.506,49.825,13.449],[-40.689,56.447,7.437],
[-45.751,57.488,8.361],[-54.144,56.643,4.803],[-45.689,56.213,4.131],
[-56.843,57.495,5.001],[-41.770,58.245,9.624],[-55.441,56.001,10.178],
[-43.429,55.407,6.021],[-56.793,56.587,7.664],[-43.437,54.223,8.434],
[-50.323,52.899,5.671],[-43.089,51.670,-0.705],[-43.184,59.053,12.972],
[-52.884,52.881,7.181],[-48.940,51.612,10.526],[-45.386,56.727,12.713],
[-49.911,54.871,-4.701],[-52.780,57.020,7.483],[-53.212,58.977,2.993],
[-49.192,58.464,3.605],[-54.269,-58.834,8.272],[-54.047,55.432,-4.312],
[-58.848,55.972,3.572],[-57.491,56.944,-1.700],[-52.716,54.505,3.481],
[-55.035,57.987,0.812],[-51.970,56.163,-0.033],[-46.687,55.395,6.849],
[-53.433,57.808,-1.717],[-47.535,57.873,5.900],[-55.535,55.314,0.705],
[-50.657,55.944,9.306],[-46.812,55.380,-2.888],[-44.791,57.857,-1.185],
[-40.263,58.721,0.076],[-58.018,55.463,11.159],[-46.557,55.281,9.947],
[-46.399,55.673,0.691],[-48.885,57.715,0.994],[-47.961,57.566,11.036],
[-50.094,52.874,2.930],[-53.640,53.940,-0.854],[-43.178,56.415,11.107],
[-55.580,-59.335,-1.787],[-55.747,54.102,6.660],[-51.654,55.222,13.160],
[-52.939,54.896,10.690],[-47.422,57.676,15.025],[-55.925,58.850,14.367],
[-50.773,53.176,8.963],[-49.811,57.260,13.431],[-47.715,55.114,12.339],
[-45.674,55.265,15.360],[-54.392,56.605,12.577],[-51.307,51.523,11.732],
[-50.870,53.622,15.015],[-41.390,59.065,-2.423],[-42.920,56.107,-2.546],
[-44.739,57.831,1.554],[-43.001,57.886,6.842],[-44.235,-59.101,8.511],
[58.758,54.585,13.332],[-43.389,54.390,14.157],[-43.274,54.656,0.812],
[-58.896,55.271,6.447],[-58.765,58.174,6.719],[-41.293,54.318,12.482],
[-40.509,54.286,-2.799],[-59.310,54.077,1.931],[-57.878,53.240,9.576],
[-42.486,51.911,10.585],[-48.950,55.273,4.829],[-43.684,50.973,5.367],
[-42.790,54.060,3.564],[-56.582,53.111,1.736],[-58.800,52.710,-0.606],
[-52.268,51.751,3.981],[-45.599,53.263,5.194],[-45.391,52.683,2.421],
[-56.576,53.916,-1.607],[-50.195,54.491,-1.505],[-57.853,53.328,13.914],
[-54.870,53.500,14.099],[-46.365,52.902,14.191],[-47.853,-50.696,21.119],
[-49.502,-49.766,25.471],[-55.974,-50.580,26.708],[-45.599,-46.596,22.790],
[-56.339,-47.092,24.246],[-57.926,-52.043,19.963],[-59.017,-49.336,22.765],
[-47.682,-42.494,24.545],[-49.200,-46.096,25.342],[-43.906,-57.821,26.192],
[-46.982,-56.011,26.511],[-53.290,-55.662,22.614],[-45.731,-53.990,23.574],
[-57.452,-52.555,23.186],[-51.740,-50.202,22.537],[-40.331,-53.898,29.857],
[-56.036,-56.740,27.835],[-47.539,-45.019,20.573],[-52.429,-48.958,28.977],
[-43.367,-56.186,23.458],[-55.756,-54.355,25.378],[-54.491,-46.459,26.946],
[-59.275,-46.839,19.418],[-44.564,-46.092,25.675],[-46.033,-57.679,28.447],
[-51.906,-59.704,25.212],[-45.561,-44.181,23.820],[-45.058,-44.196,21.206],
[-43.648,-47.701,30.272],[-49.587,-48.556,20.651],[-48.256,-51.644,17.141],
[-41.629,-50.792,27.393],[-44.984,-50.051,20.702],[-47.116,-52.153,29.206],
[-54.303,-49.185,22.444],[-53.034,-46.351,24.014],[-53.779,-50.200,25.016],
[-46.854,-44.831,27.413],[-53.862,-46.063,30.972],[-46.702,-49.409,16.556],
[-55.774,-48.701,28.797],[-50.361,-45.039,20.591],[-56.501,-42.743,33.136],
[-42.680,-57.995,18.403],[-54.600,-49.685,19.039],[-41.986,-44.691,16.472],
[-58.576,-48.109,28.981],[-58.861,-44.535,26.193],[-41.397,-44.786,30.517],
[-51.047,-45.046,31.196],[-50.294,-40.872,19.743],[-55.841,-41.191,20.476],
[-46.364,-42.208,16.597],[-46.264,-42.589,19.530],[-43.458,-52.307,31.860],
[-56.482,-45.897,30.432],[-53.723,-58.047,26.500],[-48.982,-56.878,28.935],
[-46.495,-53.179,31.880],[-44.470,-49.031,15.368],[-49.988,-55.392,17.562],
[-53.918,-52.149,27.009],[-54.692,-51.571,20.918],[-48.837,-52.776,19.632],
[-49.804,-51.650,32.694],[-47.594,-51.616,23.637],[-55.847,-53.365,28.507],
[-53.852,-54.851,14.270],[-57.233,-56.637,22.545],[-57.185,-54.207,17.859],
[-48.568,-51.726,26.948],[-50.504,-54.923,23.528],[-56.165,-55.054,20.241],
[-52.581,-53.887,18.574],[-47.057,-58.626,26.180],[-52.950,-52.884,15.831],
[-44.836,-40.189,31.670],[-57.129,-48.005,33.859],[-50.671,-52.285,25.015],
[-54.677,-56.986,19.219],[-51.552,-52.731,27.743],[-51.314,-47.031,18.953],
[-44.154,-49.111,25.441],[-47.631,-56.190,16.350],[-48.862,-46.921,29.703],
[-45.525,-51.868,17.609],[-54.900,-52.296,23.565],[-48.452,-46.754,18.462],
[-58.198,-51.047,28.213],[-52.462,-51.521,31.811],[-46.244,-46.537,29.864],
[-40.684,-52.816,18.390],[-40.634,-49.101,29.229],[-43.273,-45.073,18.903],
[-47.114,-48.980,23.965],[-54.233,-45.923,21.084],[-57.419,-50.315,16.092],
[-51.940,-47.380,21.742],[-51.050,-43.127,28.787],[-53.638,-43.233,21.418],
[-47.273,-49.259,26.641],[-46.029,-46.783,16.873],[-44.748,-43.117,30.467],
[-41.177,-43.363,18.672],[-56.508,-57.345,30.238],[-46.882,-48.128,20.590],
[-56.598,-44.079,28.071],[-51.865,-45.206,13.980],[-53.947,-47.652,17.143],
[-56.904,-48.009,17.652],[-48.847,-54.323,27.853],[-47.738,-55.785,19.925],
[-48.865,-43.138,18.883],[-50.551,-54.589,20.333],[-49.896,-52.354,29.718],
[-47.803,-58.202,21.283],[-52.070,-43.645,18.514],[-53.549,-56.162,16.831],
[-42.855,-55.220,28.752],[-44.944,-51.673,24.779],[-57.041,-44.749,17.344],
[-57.052,-52.810,14.709],[-44.506,-41.873,24.560],[-55.744,-57.178,25.068],
[-58.637,-44.962,22.780],[59.006,-40.318,31.400],[-53.982,-43.384,33.383],
[-49.849,-44.413,15.392],[-47.651,-46.105,32.389],[-44.821,-44.499,16.594],
[59.469,-48.577,32.625],[-53.870,-48.217,14.483],[-51.983,-48.085,31.455],
[-50.071,-47.081,16.162],[-47.128,-49.272,29.427],[-56.486,-50.322,31.828],
[-51.828,-50.103,17.866],[-54.598,-51.107,29.875],[-50.186,-49.791,15.358],
[-47.799,-49.955,31.954],[-41.445,-47.605,16.240],[-52.178,-54.541,30.428],
[-52.664,-55.558,27.162],[-49.648,-49.173,28.415],[-45.521,-54.281,35.330],
[-45.364,-50.763,30.706],[-54.824,-53.240,31.280],[-51.363,-58.456,29.410],
[-50.458,-54.823,32.400],[-48.248,-54.928,30.819],[-46.099,-56.712,33.043],
[-55.958,-55.670,32.404],[-51.672,-59.054,32.188],[-51.119,-56.957,33.945],
[-51.297,-40.051,17.096],[-57.624,-47.226,15.097],[-57.690,-43.047,21.151],
[-59.193,-42.133,24.347],[-41.839,-42.341,25.437],[-42.160,-43.520,28.082],
[-43.489,-47.140,32.899],[-40.440,-47.940,19.801],[-42.352,-50.060,21.527],
[-42.506,-47.935,27.333],[-44.260,-45.207,28.512],[-40.206,-49.841,17.472],
[-41.235,-54.331,16.191],[-45.019,-53.367,20.057],[-58.285,-50.447,25.289],
[-42.283,-49.694,31.271],[-40.872,-52.220,31.941],[-44.191,-54.149,16.098],
[-42.240,-53.830,20.860],[-43.344,-53.964,25.720],[-46.431,-53.408,26.192],
[-59.619,-52.689,32.096],[-39.517,-58.919,32.089],[-43.552,-57.413,20.968],
[-58.346,-56.057,25.425],[-57.766,-53.355,26.669],[-40.437,-55.392,32.735],
[-41.328,-57.361,15.613],[59.450,-56.655,20.256],[-44.821,-41.124,21.836],
[-58.772,-57.764,27.689],[-58.478,-44.979,33.208],[-53.935,-44.286,28.512],
[-48.771,-43.097,31.667],[-47.301,-41.452,30.369],[-42.253,-58.040,29.872],
[-58.361,-41.456,29.317],[-55.988,-40.954,27.270],[-48.449,-40.078,25.468],
[-47.022,-41.881,27.353],[-44.930,-40.829,28.843],[-56.404,-43.035,25.341],
[-55.266,-44.607,23.611],[-51.716,-43.094,24.254],[-49.865,-57.546,24.682],
[-44.400,-59.395,23.554],[-41.396,-58.426,22.437],[-56.071,-58.750,20.773],
[-57.184,-57.846,18.663],[-51.296,-57.380,21.497],[-46.926,-57.154,23.854],
[-45.676,-57.369,18.979],[-58.493,-40.576,17.778],[-55.235,-59.090,16.895],
[-51.797,-56.700,18.960],[-49.446,-42.155,16.514],[-42.961,-41.713,20.037],
[-57.784,-57.767,32.306],[-53.851,-57.516,32.462],[-54.675,-42.855,18.106],
[-44.550,-58.767,31.232],[-42.055,-43.067,32.541],[-47.169,-32.808,21.624],
[-48.914,-32.081,23.598],[-55.567,-30.660,26.226],[-44.821,-28.518,24.084],
[-54.788,-27.992,25.411],[-57.540,-33.442,20.586],[-58.437,-32.151,23.132],
[-48.344,-24.286,23.823],[-50.032,-28.766,26.694],[-41.720,-36.776,28.538],
[-46.042,-35.275,25.743],[-54.409,-36.479,22.918],[-45.253,-38.033,21.380],
[-57.942,-35.639,23.049],[-52.139,-34.664,22.072],[-41.715,-34.097,28.853],
[-55.295,-37.587,27.721],[-48.478,-27.879,21.145],[-52.590,-29.096,27.574],
[-44.682,-35.916,23.023],[-55.545,-35.212,25.457],[-55.018,-28.794,28.232],
[-58.212,-29.135,19.325],[-43.608,-26.462,25.205],[-44.238,-38.302,28.999],
[-51.557,-40.455,24.881],[-45.145,-25.955,23.097],[-46.102,-26.472,20.711],
[-44.412,-29.958,30.648],[-49.904,-31.108,20.670],[-47.873,-32.906,16.263],
[-41.784,-31.721,27.095],[-44.739,-31.569,20.362],[-47.126,-33.815,28.679],
[-53.793,-30.757,22.600],[-52.431,-27.442,24.394],[-52.769,-31.444,25.854],
[-47.262,-26.535,26.160],[-53.982,-27.406,30.743],[-46.750,-32.044,18.508],
[-56.674,-31.624,29.761],[-47.920,-24.437,20.938],[-54.673,-24.803,30.682],
[-41.895,-40.427,17.880],[-54.583,-31.776,18.866],[-42.816,-26.086,15.812],
[-57.700,-28.441,29.152],[-59.417,-26.653,26.087],[-40.621,-26.608,29.283],
[-51.381,-25.653,30.542],[-50.886,-23.191,19.529],[-55.071,-22.334,21.493],
[-45.417,-25.501,16.758],[-44.976,-24.322,19.124],[-43.986,-33.561,32.249],
[-55.983,-25.749,28.397],[-53.476,-39.608,26.835],[-48.077,-39.569,28.513],
[-46.139,-36.848,30.416],[-45.008,-32.843,16.573],[-49.393,-35.295,15.357],
[-52.610,-33.696,27.399],[-54.876,-33.798,23.059],[-49.802,-33.942,17.687],
[-48.058,-33.981,31.404],[-48.245,-35.167,21.509],[-55.320,-34.328,28.150],
[-53.907,-35.869,15.947],[58.571,-36.342,23.932],[-58.242,-36.022,18.954],
[-48.419,-33.997,26.125],[-51.744,-38.041,22.306],[-56.156,-35.600,20.918],
[-53.368,-36.146,19.439],[-46.526,-38.475,25.945],[-54.799,-33.785,17.069],
[-43.655,-21.370,33.213],[-55.741,-29.447,31.000],[-50.955,-33.608,24.644],
[-55.184,-38.402,20.307],[-50.748,-36.327,26.478],[-50.784,-27.971,19.699],
[-43.939,-31.248,25.348],[-47.845,-37.290,16.190],[-46.904,-29.002,29.711],
[-46.321,-35.288,17.327],[-56.255,-29.091,23.068],[-48.271,-28.129,18.421],
[-57.939,-33.736,27.899],[-52.169,-31.251,31.356],[-46.612,-26.563,28.674],
[-42.068,-33.724,17.421],[-41.880,-28.291,31.176],[-43.892,-26.863,19.273],
[-46.188,-30.674,23.424],[-53.928,-27.122,21.762],[-56.757,-32.001,16.857],
[-52.010,-29.122,21.671],[-50.636,-26.083,26.929],[-52.725,-24.575,21.490],
[-48.341,-30.739,27.709],[-43.547,-28.454,16.704],[-44.979,-24.513,29.319],
[-43.397,-23.168,15.900],[-56.153,-37.003,30.198],[-46.975,-29.996,20.146],
[-58.711,-24.434,28.960],[-52.540,-28.425,15.425],[-52.818,-26.751,18.261],
[-54.127,-29.882,16.993],[-48.667,-36.298,28.966],[-48.025,-36.236,19.136],
[-49.557,-25.761,19.335],[-50.439,-35.427,19.803],[-50.785,-35.245,30.495],
[-49.465,-37.754,20.990],[-50.908,-24.806,17.110],[-53.998,-38.690,17.639],
[-43.396,-35.743,30.566],[-44.905,-33.276,23.465],[-55.732,-26.897,18.726],
[-57.053,-35.111,15.931],[-44.496,-24.180,26.388],[-57.132,-39.816,25.127],
[-56.360,-26.116,21.618],[-57.417,-22.908,32.368],[-50.980,-24.584,33.472],
[-50.100,-25.668,14.163],[-46.546,-27.474,32.375],[-44.922,-26.960,14.544],
[-58.326,-28.196,32.279],[-55.494,-27.700,15.803],[-51.509,-28.749,31.592],
[-49.845,-28.198,14.987],[-45.489,-31.100,28.120],[-57.708,-33.392,32.607],
[-51.453,-30.394,17.335],[-53.064,-33.387,30.050],[-51.683,-34.028,15.359],
[-49.418,-31.517,32.186],[-46.169,-29.782,16.731],[-53.359,-36.271,30.712],
[-52.333,-37.403,28.546],[-49.857,-32.905,27.937],[-47.373,-35.135,33.661],
[-46.264,-32.087,32.841],[-54.480,-33.706,33.340],[-50.856,-40.569,28.418],
[-50.346,-35.988,33.979],[-48.838,-36.935,31.802],[-46.758,-37.983,33.055],
[-55.122,-36.676,32.766],[-51.729,-41.903,30.974],[-51.714,-41.022,33.572],
[-50.761,-21.117,16.895],[-59.487,-27.684,15.186],[-58.402,-22.026,20.193],
[-59.229,-24.625,23.373],[-40.820,-24.030,25.484],[-42.174,-24.662,27.827],
[-44.064,-28.224,32.662],[-41.607,-28.944,18.580],[-41.979,-31.802,19.967],
[-42.526,-29.171,25.847],[-43.382,-27.348,28.615],[-42.972,-31.106,17.350],
[-42.949,-35.737,15.803],[-43.199,-33.602,21.483],[-59.678,-29.647,27.073],
[-42.248,-30.880,32.258],[-41.038,-33.456,31.693],[-42.225,-38.040,14.817],
[-42.673,-35.919,19.700],[-43.655,-36.224,26.385],[-44.376,-33.787,27.873],
[-59.833,-36.742,31.211],[-42.128,-40.457,33.094],[-43.865,-38.720,19.083],
[-57.739,-36.953,25.429],[-57.397,-32.725,25.536],[-39.247,-37.942,32.151],
[-40.768,-37.466,17.174],[-58.973,-38.039,20.767],[-45.125,-23.626,21.519],
[-59.667,-39.449,28.287],[-56.184,-26.650,32.833],[-53.041,-25.768,28.156],
[-47.708,-24.039,32.474],[-45.569,-22.506,31.382],[-42.270,-40.513,29.676],
[-57.474,-22.372,26.978],[-55.065,-22.240,28.732],[-49.990,-23.062,26.628],
[-47.494,-23.753,27.977],[-45.883,-21.677,27.866],[-57.014,-25.043,25.781],
[-55.877,-24.133,23.236],[-54.178,-24.479,25.197],[-48.880,-37.348,23.647],
[-43.985,-39.196,24.101],[-42.145,-40.025,22.050],[-56.456,-38.176,23.043],
[-57.275,-38.420,18.433],[-52.165,-41.141,22.179],[-46.875,-39.482,23.369],
[-48.559,-39.556,18.107],[-57.727,-24.722,19.446],[-55.763,-39.034,15.708],
[-51.629,-38.349,19.431],[-47.999,-24.713,17.273],[-41.983,-24.499,20.311],
[-56.387,-39.421,32.179],[-54.507,-40.529,33.908],[-57.149,-25.200,16.486],
[-47.426,-40.626,33.051],[-43.441,-24.841,32.262],[-48.124,-12.885,22.921],
[-49.370,-13.094,25.212],[-57.504,-12.056,26.694],[-43.968,-10.384,23.153],
[-53.732,-8.981,24.500],[-58.541,-14.941,20.634],[-59.160,-13.050,24.851],
[-48.007,-4.988,23.707],[-51.333,-11.343,24.706],[-41.770,-19.328,26.075],
[-45.929,-16.832,26.902],[-54.010,-17.523,22.515],[-45.985,-19.208,22.042],
[-58.178,-17.467,22.608],[-52.915,-15.047,23.139],[-41.070,-16.373,27.463],
[-55.143,-18.443,26.293],[-48.009,-9.546,21.754],[-51.575,-12.161,27.129],
[-43.818,-19.667,24.012],[-54.945,-15.784,24.892],[-54.236,-10.652,28.618],
[-57.134,-10.905,19.499],[-45.034,-10.150,25.637],[-43.891,-19.462,27.549],
[-52.120,-22.835,24.391],[-45.030,-7.399,22.555],[-46.431,-7.999,20.292],
[-43.676,-10.476,29.696],[-50.197,-12.556,20.887],[-49.424,-14.003,15.717],
[-42.427,-13.621,28.234],[-45.433,-13.410,19.418],[-45.851,-15.029,29.016],
[-52.818,-12.705,21.944],[-52.219,-6.702,26.406],[-54.631,-12.975,26.825],
[-47.357,-8.257,27.485],[-51.679,-10.286,29.051],[-47.793,-13.842,18.065],
[-56.040,-12.638,28.970],[-49.189,-6.632,21.368],[-53.383,-9.097,30.978],
[-41.522,-22.283,17.673],[-55.085,-12.870,20.395],[-42.580,-8.333,16.424],
[-59.510,-10.078,28.756],[58.473,-8.205,25.909],[-40.671,-6.924,30.203],
[-49.765,-7.878,29.126],[-49.837,-2.326,19.363],[-55.717,-3.979,22.756],
[-51.012,-0.221,16.299],[-46.687,-4.949,19.167],[-42.420,-15.757,30.812],
[-57.002,-8.858,28.469],[-54.088,-21.506,26.177],[-49.100,-22.001,29.778],
[-45.439,-17.841,29.633],[-44.601,-13.409,15.911],[-49.439,-18.936,17.687],
[-53.131,-16.137,26.832],[-56.300,-15.345,22.303],[-48.906,-14.969,20.632],
[-49.749,-15.236,32.515],[-47.455,-17.351,23.863],[-55.842,-15.248,27.681],
[-53.931,-17.543,16.393],[58.197,-18.800,21.594],[-59.088,-18.130,16.908],
[-47.837,-14.427,27.061],[-50.980,-18.545,23.342],[-56.212,-17.551,20.470],
[-53.241,-16.559,18.846],[-47.023,-19.459,27.010],[-55.277,-14.540,17.708],
[-42.476,-3.696,31.496],[-55.916,-11.284,31.122],[-49.879,-15.859,24.758],
[-54.194,-20.058,20.331],[-50.927,-17.508,27.854],[-50.263,-9.771,20.045],
[-44.753,-12.072,27.671],[-46.447,-20.481,17.585],[-48.732,-10.543,30.086],
[-45.874,-17.715,17.572],[-54.174,-12.084,24.274],[-47.898,-9.230,18.295],
[-58.562,-12.950,29.073],[-49.839,-12.885,29.140],[-47.371,-8.095,30.600],
[-41.356,-15.000,18.809],[-41.369,-9.377,31.985],[-44.602,-8.445,18.437],
[-45.826,-12.606,24.582],[-54.103,-7.644,22.032],[-57.781,-13.634,18.055],
[-52.676,-10.183,20.858],[-51.721,-5.789,28.732],[-53.520,-5.094,21.248],
[-47.595,-11.095,26.816],[-46.473,-10.871,16.734],[-44.780,-6.677,29.867],
[-41.757,-5.080,15.707],[-58.263,-18.495,30.017],[-45.498,-10.467,20.837],
[-56.837,-5.843,28.506],[-51.736,-7.982,15.279],[-52.148,-8.988,18.064],
[-54.660,-10.109,18.974],[-49.655,-19.671,28.876],[-47.480,-18.866,19.577],
[-48.581,-6.792,16.935],[-49.936,-18.132,20.715],[-49.156,-15.463,29.865],
[-49.010,-21.099,20.862],[-51.763,-6.182,17.414],[-54.079,-20.688,17.610],
[-43.260,-18.592,31.328],[-45.125,-15.176,25.059],[-53.597,-10.086,15.324],
[-58.429,-15.790,15.263],[-43.570,-5.864,26.031],[-56.150,-21.195,24.121],
[-56.748,-7.813,20.647],[-59.043,-5.739,31.561],[-53.360,-6.863,32.349],
[-48.311,-8.951,15.526],[-44.801,-8.553,32.381],[-44.813,-6.510,16.563],
[-58.395,-11.679,32.389],[-53.804,-12.805,15.622],[-52.162,-11.351,31.943],
[-50.070,-8.613,13.466],[-46.766,-12.615,29.878],[-54.956,-13.438,32.575],
[-51.742,-13.277,17.764],[-52.059,-15.494,30.071],[-52.216,-15.174,15.632],
[-48.545,-12.618,31.962],[-41.867,-13.341,16.467],[-55.092,-18.069,29.305],
[-52.702,-19.320,27.188],[-50.822,-15.021,27.641],[-44.535,-16.381,32.415],
[-45.362,-13.859,31.945],[-52.189,-16.197,32.949],[-51.728,-21.944,28.503],
[-50.305,-19.022,33.106],[-48.601,-18.007,30.787],[-46.368,-19.583,31.521],
[-54.827,-17.182,31.882],[-52.419,-21.194,31.925],[-49.908,-22.030,33.125],
[-51.237,-3.564,17.197],[-57.531,-8.377,16.630],[-59.433,-3.336,21.887],
[-58.808,-6.006,22.023],[-39.746,-4.525,29.172],[-42.985,-4.872,28.895],
[-43.602,-11.053,32.690],[-47.728,-11.614,20.169],[-41.871,-13.350,21.309],
[-42.129,-10.469,25.600],[-44.453,-7.888,27.428],[-42.927,-12.125,18.857],
[-42.444,-16.186,16.432],[-44.490,-15.303,21.184],[56.492,-13.426,26.674],
[-42.360,-13.180,31.243],[-40.109,-14.859,32.591],[-44.131,-17.873,15.337],
[-42.796,-17.168,19.760],[-43.661,-17.481,25.629],[-43.543,-16.096,28.413],
[-58.884,-18.487,32.936],[-41.146,-22.141,35.220],[-43.463,-19.799,20.279],
[-57.552,-19.575,25.970],[-58.042,-15.700,25.495],[-41.541,-21.335,31.612],
[-40.817,-18.796,16.338],[-59.000,-19.351,19.774],[-44.435,-5.710,20.631],
[58.573,-20.510,29.576],[-56.061,-7.630,34.133],[-54.381,-6.602,29.895],
[-50.222,-5.682,30.829],[-46.838,-5.378,31.038],[-43.302,-22.035,27.850],
[-58.118,-3.068,29.912],[-55.169,-3.730,27.893],[-49.027,-3.699,26.487],
[-47.468,-5.601,28.005],[-46.496,-2.527,29.126],[-56.475,-6.713,25.729],
[-54.198,-5.548,24.708],[-50.499,-5.846,24.448],[-48.529,-20.461,24.930],
[-44.381,-22.333,24.407],[-41.848,-20.724,22.468],[-56.733,-20.032,21.680],
[-58.731,-21.128,17.468],[-51.418,-21.500,22.108],[-47.152,-22.065,22.979],
[-44.233,-21.800,18.331],[-56.254,-5.128,19.894],[-56.556,-20.206,16.157],
[-51.322,-20.500,19.371],[-48.379,-4.265,16.320],[-41.690,-6.651,18.246],
[-56.182,-19.947,30.906],[-55.229,-22.227,33.856],[-54.373,-7.733,16.691],
[-45.737,-23.316,34.158],[-42.798,-6.410,31.870],[-47.371,6.679,21.060],
[-49.435,3.483,22.982],[-54.934,5.517,27.183],[-45.155,8.200,22.056],
[-55.186,9.518,26.036],[-57.956,4.301,21.257],[-58.426,5.098,23.744],
[-47.051,13.906,24.234],[-47.617,7.653,24.683],[-42.068,-0.603,24.523],
[-48.025,2.347,25.277],[-54.270,0.825,22.603],[-46.730,0.033,22.597],
[-55.603,4.025,22.681],[-52.482,3.782,22.877],[-41.140,2.937,28.518],
[-54.541,-0.190,27.815],[-48.034,9.454,21.638],[-51.200,8.204,28.435],
[-44.194,1.114,23.127],[-55.108,2.950,25.385],[-54.036,9.182,28.435],
[-58.503,8.825,19.713],[-44.939,8.949,24.790],[-45.214,-0.982,24.853],
[-50.627,-3.209,24.367],[-46.139,11.063,23.623],[-44.411,10.578,21.566],
[-44.020,8.981,29.294],[-49.438,5.396,19.302],[-47.866,2.852,16.359],
[-41.969,4.525,26.330],[-42.947,5.902,21.585],[-48.388,3.221,29.104],
[-54.108,7.455,22.270],[-51.018,10.656,24.283],[-52.426,7.009,26.252],
[-47.644,9.886,25.882],[-51.081,10.454,29.476],[-47.037,5.254,17.464],
[-54.770,7.015,29.820],[-50.346,10.756,21.684],[-53.837,11.947,31.280],
[-42.683,-4.012,18.616],[-54.127,5.497,20.387],[-40.938,12.222,17.219],
[-59.037,7.613,28.286],[-58.200,10.118,26.773],[-39.686,10.942,30.338],
[-48.991,11.094,31.117],[-48.127,13.204,20.942],[-55.333,13.977,21.758],
[-46.102,13.959,17.102],[-45.631,13.231,19.550],[-43.335,2.648,30.480],
[-55.496,9.615,30.687],[-53.382,-1.974,24.817],[-48.685,-3.447,30.592],
[-45.149,0.769,29.779],[-43.303,6.646,16.284],[-49.108,0.386,17.941],
[-51.886,3.121,27.784],[-53.579,2.519,20.641],[-47.330,2.591,19.115],
[-48.550,3.629,31.923],[-51.338,6.854,23.103],[-57.619,2.836,30.798],
[-53.296,1.348,15.485],[-59.124,0.863,23.284],[-55.088,0.383,18.810],
[-49.293,4.172,26.872],[-50.949,-0.521,21.028],[-56.714,1.613,20.708],
[-52.814,1.220,18.257],[-48.098,-1.602,24.489],[-54.265,3.462,17.516],
[-43.029,15.516,33.246],[-55.680,7.770,33.269],[-51.588,1.742,25.181],
[-55.221,-1.106,20.835],[-49.811,1.421,27.384],[-50.047,7.712,17.634],
[-43.406,6.663,26.515],[-46.859,0.318,16.504],[-50.177,8.299,31.391],
[-44.796,3.359,16.830],[-54.206,5.634,24.403],[-47.152,8.086,18.487],
[-56.627,5.095,29.573],[-51.457,3.603,32.002],[-46.648,9.581,30.580],
[-40.806,2.311,18.408],[-41.418,8.823,31.380],[-41.107,9.950,18.734],
[-46.019,5.772,25.888],[-53.186,10.152,22.249],[-56.463,4.854,18.124],
[-51.538,6.976,20.400],[-50.131,12.731,28.419],[-52.132,12.580,21.975],
[-48.085,8.309,28.227],[-45.184,8.400,16.537],[-43.631,13.122,29.586],
[-42.770,14.346,16.767],[-57.828,0.444,29.403],[-45.065,6.728,19.796],
[-55.796,12.175,29.157],[-50.227,9.561,15.218],[-51.624,9.832,19.456],
[-56.099,7.609,18.842],[-48.939,-1.049,28.541],[-48.106,0.162,20.387],
[-48.889,9.965,19.010],[-49.890,2.551,20.430],[-50.416,1.300,30.687],
[-48.576,-2.377,21.584],[-52.093,11.013,17.131],[-53.274,-1.463,16.052],
[-43.451,-1.223,31.328],[-43.886,3.716,23.735],[-55.034,10.222,18.417],
[-58.089,2.150,16.047],[-43.755,11.220,25.908],[-56.571,-1.405,25.873],
[-57.860,11.111,20.869],[-57.432,14.523,29.383],[-53.089,12.236,34.788],
[-50.091,12.105,15.427],[-46.249,10.301,33.245],[-44.725,11.843,16.166],
[-58.676,6.683,33.893],[-52.699,8.315,16.754],[-52.925,9.050,31.443],
[-47.750,8.855,15.947],[-45.709,6.840,30.166],[-56.522,5.530,32.264],
[-51.948,5.940,17.278],[-52.908,4.904,29.513],[-49.539,5.246,16.244],
[-48.863,6.626,33.171],[-41.792,8.320,12.694],[-53.098,1.243,30.050],
[-51.780,-0.507,27.200],[-50.075,5.377,29.843],[-46.081,2.520,32.257],
[-46.689,5.363,32.125],[-54.306,2.600,32.143],[-52.697,-2.847,28.836],
[-49.675,-1.087,31.731],[-47.854,0.865,30.012],[-46.948,-0.191,32.937],
[-54.454,-0.064,32.506],[-52.597,-4.537,31.566],[-50.057,-3.860,33.110],
[-51.192,16.818,16.332],[-57.461,9.253,15.859],[-58.820,15.850,22.178],
[-59.390,12.168,22.917],[-41.004,14.342,25.797],[-40.992,13.829,28.828],
[-43.497,9.270,33.216],[-42.211,7.482,19.188],[-40.409,5.111,19.777],
[-41.920,9.109,25.878],[-42.189,10.711,28.525],[-40.976,5.257,16.792],
[-42.229,2.805,15.868],[-44.764,3.370,20.091],[-57.034,7.596,25.004],
[-42.664,6.141,30.831],[-41.512,4.234,32.242],[-44.229,0.106,15.849],
[-43.901,1.234,18.738],[-44.452,1.497,26.252],[-45.672,3.567,29.566],
[-58.083,3.729,33.708],[-40.517,-2.641,34.453],[-41.781,-1.841,21.733],
[-56.583,0.995,24.274],[-58.588,2.919,25.883],[-42.225,-0.019,28.945],
[-41.693,0.085,17.180],[-58.390,-0.426,19.961],[-45.081,14.078,21.957],
[59.305,-6.132,27.400],[-58.264,8.879,32.218],[-53.167,11.989,28.497],
[-48.903,14.698,31.536],[-44.771,15.978,31.102],[-41.724,-3.031,27.524],
[-55.767,16.576,28.878],[-54.657,14.218,27.421],[-47.256,15.370,26.411],
[-45.315,13.474,27.132],[-45.255,15.959,28.202],[-56.435,12.047,25.923],
[-52.896,8.643,24.422],[-53.508,12.479,24.636],[-49.697,0.288,23.923],
[-43.351,-2.963,24.753],[-36.024,-10.397,19.540],[-57.327,-1.555,22.512],
[-57.114,-2.830,18.515],[-51.649,-2.898,21.900],[-46.132,-3.206,23.188],
[-45.717,-2.513,20.057],[-58.577,11.562,18.316],[-56.440,-1.581,15.948],
[-52.695,-1.401,18.963],[-48.226,12.294,17.525],[-42.997,13.070,20.063],
[-56.706,-2.015,32.370],[-52.858,-2.141,33.081],[-55.805,11.814,16.147],
[-45.843,-4.347,33.632],[-41.683,12.602,33.451],[-47.660,24.243,21.991],
[-48.837,23.865,24.927],[-54.943,23.759,25.993],[-46.810,27.675,24.022],
[-56.520,27.638,25.266],[-58.041,21.718,20.566],[-58.356,24.794,22.045],
[-48.045,32.244,23.782],[-49.222,26.837,24.581],[-41.593,16.993,25.637],
[-46.057,20.633,25.425],[-54.900,19.245,22.801],[-45.845,18.222,21.391],
[-56.779,21.194,23.231],[-53.595,23.961,22.315],[-41.454,20.329,27.644],
[-55.007,18.910,27.822],[-47.484,27.215,21.270],[-52.293,25.374,29.335],
[-44.086,18.337,23.765],[-55.666,20.268,25.406],[-55.019,29.583,27.345],
[59.040,26.914,19.820],[-45.720,26.757,27.977],[-43.050,17.944,28.165],
[-50.268,14.396,26.197],[-45.892,30.277,23.836],[-45.631,29.393,21.221],
[-43.745,27.400,30.228],[-49.211,25.089,19.926],[-49.509,22.521,15.552],
[-41.560,22.801,28.895],[-44.414,24.467,20.498],[-46.008,22.869,28.430],
[-54.970,26.471,21.515],[-52.237,27.703,25.076],[-52.405,23.637,25.035],
[-47.665,30.276,26.448],[-51.545,28.063,30.458],[-47.749,21.199,16.925],
[-54.863,24.214,28.828],[-48.229,30.485,21.566],[-52.375,30.239,31.725],
[-42.170,16.626,18.058],[-54.370,24.055,19.481],[-42.146,28.147,17.670],
[-58.308,27.073,27.708],[-59.233,29.156,24.581],[-39.901,29.423,31.590],
[-48.984,28.817,30.323],[-48.060,33.318,20.931],[-56.165,33.075,22.956],
[-46.074,31.246,16.925],[-45.816,31.557,19.702],[-42.819,22.298,32.718],
[-55.507,27.300,28.707],[-51.701,16.560,24.793],[-47.425,14.594,29.043],
[-45.258,21.190,30.767],[-46.089,23.740,17.612],[-49.326,18.311,17.696],
[-53.304,21.350,26.477],[-54.760,21.868,21.089],[-49.339,22.455,20.588],
[-48.318,22.118,31.903],[-46.211,21.988,22.422],[-55.333,21.464,28.455],
[-53.086,19.176,15.429],[-58.228,18.644,22.478],[-56.336,18.414,17.984],
[-47.949,22.203,26.708],[-51.765,18.144,22.413],[-56.177,19.600,20.400],
[-53.125,19.726,20.249],[-46.215,18.331,26.562],[-54.928,22.459,16.906],
[-44.210,34.493,32.331],[-56.444,25.567,31.115],[-49.168,21.056,24.499],
[-54.218,17.470,20.833],[-50.868,20.627,27.379],[-50.909,26.518,17.932],
[-43.276,24.751,26.921],[-46.541,18.139,17.718],[-47.572,25.096,29.533],
[-43.865,20.881,18.485],[-54.482,26.007,24.240],[-48.191,27.803,18.391],
[-59.424,24.856,29.522],[-50.109,24.178,30.171],[-45.677,28.874,29.368],
[-40.649,21.668,17.921],[-39.843,26.680,30.880],[-43.907,29.809,18.761],
[-45.553,24.554,23.718],[-52.493,29.707,21.521],[-57.734,23.398,16.648],
[-51.622,26.990,20.852],[-50.229,29.276,27.030],[-52.874,32.634,21.519],
[-48.235,26.151,26.920],[-46.528,27.933,16.338],[-46.260,30.998,31.180],
[-43.058,32.483,15.060],[-56.915,19.863,30.078],[-45.403,26.987,18.963],
[-57.333,31.190,27.400],[-51.289,28.979,15.271],[-52.983,28.610,18.960],
[-55.492,25.410,17.001],[-48.585,19.321,28.317],[-47.340,20.645,19.547],
[-48.883,30.147,17.182],[-50.168,19.577,20.299],[-48.825,21.843,29.355],
[-47.804,15.954,22.010],[-51.677,30.747,17.686],[-52.274,16.428,19.295],
[-42.554,18.569,30.965],[-43.946,21.098,23.606],[-55.737,29.104,17.940],
[-57.338,20.728,16.332],[-44.777,31.276,26.200],[-55.882,17.450,25.441],
[-56.729,28.500,21.115],[-57.682,32.780,30.139],[-53.733,31.303,34.031],
[-49.027,29.835,14.132],[-46.582,26.936,31.555],[-44.288,30.002,15.219],
[-59.030,25.955,32.497],[-56.030,27.461,15.505],[-51.411,25.527,32.248],
[-48.967,27.112,15.415],[-44.898,25.010,30.106],[-57.520,23.600,32.534],
[-51.764,23.831,17.713],[-52.400,22.537,29.690],[-52.741,23.200,14.940],
[-48.340,24.807,32.535],[-43.895,25.127,15.445],[-50.824,20.566,30.938],
[-52.832,17.009,27.831],[-50.819,23.729,27.416],[-47.285,20.152,34.323],
[-45.755,23.272,32.277],[-54.736,21.818,32.540],[-50.255,15.732,29.520],
[-49.709,19.482,33.421],[-48.219,19.475,31.189],[-45.638,18.232,32.348],
[-54.886,19.229,32.416],[-51.437,15.485,32.579],[-49.049,16.292,33.629],
[-52.139,33.595,18.529],[-57.952,26.656,17.209],[59.116,33.787,22.002],
[-57.898,32.862,25.052],[-41.991,33.363,25.646],[-41.889,31.627,29.562],
[-44.340,28.016,32.853],[-40.590,26.229,20.002],[-41.830,23.821,19.518],
[-42.814,27.184,26.098],[-42.944,29.378,28.372],[-41.699,24.359,16.876],
[-42.457,21.609,15.672],[-42.259,22.626,22.279],[-57.958,25.373,24.792],
[-42.182,24.972,30.991],[-40.195,21.942,33.237],[-44.095,18.800,16.167],
[-43.114,19.915,21.011],[-42.442,19.802,25.347],[-43.956,20.771,28.420],
[-57.638,20.818,32.541],[-37.675,14.401,32.928],[-42.616,17.006,20.989],
[-58.221,18.887,25.786],[-57.377,22.728,25.720],[-39.407,16.284,31.714],
[-40.719,18.827,17.302],[-58.963,17.711,20.176],[-43.867,32.228,21.387],
[-58.080,17.530,29.550],[-54.764,29.525,32.321],[-52.085,30.458,28.816],
[-49.352,31.585,29.734],[-45.935,33.819,30.402],[-42.032,15.881,30.847],
[-58.186,34.454,28.122],[-53.827,32.643,28.097],[-48.324,33.673,26.243],
[-44.966,31.994,28.717],[-44.171,34.564,28.605],[-56.094,30.480,25.226],
[-55.117,30.790,22.071],[-53.331,30.164,24.198],[-49.319,17.000,26.526],
[-43.714,15.524,23.785],[-41.247,14.669,22.253],[-54.477,16.082,23.237],
[-57.733,13.942,17.793],[-50.438,14.067,23.497],[-46.056,16.790,24.151],
[-45.495,15.892,19.921],[-58.243,30.133,18.509],[-55.648,16.394,16.216],
[-49.444,16.284,19.343],[-50.002,32.204,16.204],[-41.593,31.475,20.127],
[-54.420,14.690,30.732],[-53.832,16.798,32.679],[-55.935,30.716,15.718],
[-46.370,15.975,33.939],[-42.735,29.939,31.825],[-47.916,43.446,22.026],
[-49.531,42.103,24.177],[-55.090,42.103,27.010],[-43.807,45.426,23.122],
[-55.612,47.517,26.900],[-57.917,40.469,20.530],[-58.753,42.731,23.282],
[-47.329,50.779,24.828],[-49.311,44.887,24.528],[-41.605,37.636,27.100],
[-45.759,39.724,27.081],[-53.916,38.535,23.554],[-45.331,37.314,21.338],
[-57.115,40.496,24.172],[-54.032,41.196,23.049],[-40.558,38.910,29.307],
[-55.162,36.182,28.592],[-48.464,46.361,21.192],[-51.851,43.709,28.620],
[-44.409,36.548,23.690],[-55.565,39.122,25.856],[-53.603,45.878,27.869],
[-59.224,45.666,19.799],[-44.451,46.904,26.737],[-44.937,36.862,27.491],
[-50.688,33.052,25.059],[-45.130,48.939,23.219],[-45.075,46.904,20.320],
[-43.963,45.998,29.702],[-49.619,43.589,19.475],[-49.384,41.988,17.278],
[-41.744,41.497,27.970],[-45.547,42.045,20.349],[-47.384,41.247,30.273],
[-53.054,44.021,22.836],[-53.459,45.895,24.976],[-52.489,42.670,25.292],
[-47.867,48.136,26.559],[-50.812,45.744,30.687],[-46.524,41.254,17.503],
[-54.729,43.535,29.217],[-48.758,48.768,20.013],[-52.712,48.634,31.199],
[-42.814,33.924,18.326],[-55.586,41.730,19.638],[-41.905,48.565,16.530],
[59.249,44.679,27.388],[-59.840,47.266,26.503],[-40.777,48.049,30.994],
[-50.549,48.020,28.922],[-49.364,51.229,20.821],[-56.321,48.221,21.564],
[-46.467,50.727,16.630],[-46.550,50.648,20.007],[-42.854,40.546,31.544],
[-56.788,45.535,28.728],[-53.654,33.649,24.492],[-48.337,33.976,29.561],
[-46.829,37.735,29.375],[-44.856,42.764,15.737],[-49.203,37.915,17.566],
[-53.118,39.901,26.238],[-54.814,39.240,21.007],[-48.269,41.009,20.062],
[-47.442,41.946,33.839],[-47.437,40.879,22.889],[-54.690,39.279,28.588],
[-53.528,38.036,15.711],[59.121,36.841,22.040],[-58.381,38.049,17.824],
[-48.487,40.481,27.091],[-52.472,36.594,24.616],[-57.388,37.852,20.686],
[-52.876,37.936,19.003],[-46.927,37.450,25.726],[-54.350,39.969,17.769],
[-42.478,51.385,30.540],[-56.560,44.064,31.140],[-50.748,39.425,24.885],
[-52.724,36.964,21.636],[-50.603,38.994,27.570],[-51.927,44.835,18.198],
[-44.569,43.823,26.864],[-47.104,35.441,16.647],[-49.029,44.782,29.076],
[-46.222,38.872,18.573],[-54.934,43.493,24.727],[-48.770,45.565,17.876],
[-57.320,41.367,28.518],[-50.239,43.136,31.561],[-47.893,48.190,29.780],
[-41.482,40.054,18.388],[-40.801,45.289,30.053],[-41.760,47.994,19.082],
[-45.901,44.134,24.531],[-53.165,48.049,22.041],[-55.738,43.782,15.270],
[-51.111,45.188,21.528],[-50.565,45.978,27.031],[-52.127,50.607,21.641],
[-48.088,43.412,27.009],[-45.685,46.669,16.988],[-43.634,48.780,29.840],
[-42.205,51.990,15.200],[-56.948,38.806,30.304],[-45.965,44.221,18.351],
[-57.344,48.307,28.954],[-51.304,46.202,15.359],[-53.691,46.471,19.809],
[-56.895,44.682,18.005],[-49.244,36.154,27.970],[-47.442,36.766,19.549],
[-51.036,47.311,19.321],[-50.683,39.427,19.779],[-50.275,39.866,31.062],
[-49.354,35.823,21.188],[-53.709,50.115,17.944],[-51.844,36.145,17.237],
[-42.559,37.039,30.363],[-44.651,41.430,25.007],[-53.740,47.646,16.233],
[-56.773,39.489,16.149],[-45.313,49.394,25.892],[-55.573,34.173,26.546],
[-58.128,48.124,19.425],[-56.579,50.916,30.953],[-51.870,49.029,33.674],
[-50.452,48.491,14.634],[-45.966,46.852,32.699],[-44.190,49.837,15.598],
[-59.071,44.525,31.851],[-53.417,44.585,15.533],[-53.303,44.876,31.152],
[-48.232,46.041,15.254],[-45.749,43.545,29.352],[-57.443,41.455,31.264],
[-54.502,43.328,17.664],[-53.097,41.443,30.348],[-52.068,40.955,16.303],
[-47.495,44.484,31.723],[-41.649,43.623,15.390],[-52.291,38.496,29.645],
[-52.236,36.394,28.089],[-49.769,41.941,29.074],[-44.815,39.612,33.397],
[-45.218,42.274,31.708],[-54.485,39.354,31.440],[-50.738,32.955,27.804],
[-49.612,37.065,32.078],[-47.603,38.857,31.637],[-46.622,37.237,33.399],
[-56.345,37.720,32.777],[-51.093,33.804,30.409],[-51.357,35.766,33.846],
[-51.591,50.803,16.763],[-58.108,46.953,16.837],[-59.197,52.621,22.926],
[-58.285,49.788,23.827],[-39.672,51.650,23.819],[-40.991,49.861,28.644],
[-43.111,46.345,32.471],[-42.623,45.286,18.817],[-41.880,43.288,20.666],
[-42.426,45.330,25.800],[-42.598,48.846,26.337],[-42.659,42.599,17.877],
[-42.079,39.157,15.579],[-43.567,41.030,21.893],[-57.661,43.455,25.642],
[-42.539,43.027,30.631],[-40.572,41.949,32.584],[-43.951,37.033,16.265],
[-43.569,38.751,19.644],[-43.714,38.308,25.612],[-43.561,39.407,28.585],
[-58.575,39.504,33.154],[-41.563,33.691,33.342],[-43.178,35.540,20.574],
[-57.265,36.279,26.184],[-58.283,38.962,26.112],[-39.756,35.985,30.322],
[-40.967,36.444,16.332],[-58.974,35.723,19.537],[-43.968,51.791,22.892],
[58.800,33.490,26.520],[-57.318,46.550,32.125],[-53.387,48.793,28.648],
[-49.689,51.054,32.807],[-45.498,52.872,31.338],[-42.160,34.276,30.771],
[-56.271,52.401,27.788],[-53.280,52.530,27.211],[-49.584,51.776,25.932],
[-46.303,50.876,27.840],[-45.088,53.319,27.955],[-56.433,50.504,25.608],
[-54.766,49.826,23.295],[-51.514,48.455,24.343],[-48.883,35.999,24.270],
[-43.978,32.456,23.947],[-41.549,33.794,22.196],[-57.224,34.980,21.602],
[-57.392,32.903,18.363],[-51.075,34.602,22.910],[-46.755,35.047,22.825],
[-45.873,34.299,19.218],[-56.478,50.767,18.008],[-54.794,35.048,16.979],
[-49.622,34.473,18.470],[-48.997,49.601,17.309],[-42.214,51.015,19.390],
[-56.380,32.637,32.537],[-54.477,34.595,33.794],[-55.485,49.578,15.679],
[-48.455,34.376,32.503],[-42.200,49.405,32.766],[-42.346,56.119,24.739],
[-46.673,57.752,25.830],[-54.070,57.507,24.006],[-47.177,56.217,19.882],
[-56.316,59.430,23.198],[-41.623,55.495,29.019],[-54.620,54.789,27.500],
[-45.712,56.086,22.304],[-56.060,59.074,26.023],[-44.142,56.575,27.040],
[-51.943,52.186,24.464],[-42.419,53.206,17.670],[-42.322,-59.939,32.108],
[-54.520,52.440,25.024],[-47.487,52.661,29.215],[-46.155,55.899,31.073],
[-49.374,55.864,15.775],[-53.864,57.332,26.771],[-52.021,58.423,22.423],
[-46.764,58.503,22.483],[-55.695,57.807,29.013],[-53.231,56.085,15.908],
[58.237,56.576,22.895],[-56.678,57.003,17.749],[-52.708,55.769,22.323],
[-56.958,57.147,21.784],[-54.141,57.635,19.685],[-48.343,55.718,26.601],
[-52.290,-59.276,19.032],[-49.681,57.670,23.920],[-54.871,54.192,19.973],
[-50.630,57.439,27.548],[-45.833,56.733,17.618],[-47.587,58.964,18.935],
[-40.477,58.847,19.108],[-57.392,56.158,30.486],[-48.167,56.914,29.183],
[-47.210,53.364,19.583],[-49.937,58.727,20.134],[-50.765,57.847,30.064],
[-48.892,54.469,21.767],[-52.076,54.770,18.679],[-43.435,57.496,29.730],
[-58.218,-58.843,16.525],[-56.607,54.344,25.682],[-53.733,-59.282,28.762],
[-53.091,56.376,29.345],[-47.059,58.938,32.934],[-54.667,58.311,31.619],
[-51.035,53.614,28.265],[-50.561,57.882,33.736],[-48.725,56.363,32.012],
[-45.335,55.511,33.511],[-56.427,57.018,33.073],[-51.792,51.032,30.418],
[-50.726,54.178,32.826],[-42.828,-59.349,15.974],[-42.544,56.883,15.555],
[-43.409,58.088,18.065],[-41.307,58.378,27.641],[-44.272,-59.838,27.789],
[-59.564,57.782,31.458],[-42.932,53.193,33.043],[-41.745,55.611,18.944],
[-58.908,57.122,25.310],[-57.905,57.882,27.770],[-42.315,55.783,31.731],
[-40.563,54.210,15.359],[-58.505,55.250,19.790],[-58.969,53.083,27.983],
[-39.991,52.599,29.270],[-49.912,54.398,24.451],[-42.260,53.249,24.695],
[-42.912,55.079,22.123],[-55.955,54.943,23.064],[-57.275,53.285,18.351],
[-51.476,53.579,21.055],[-46.862,54.063,25.068],[-44.729,51.967,18.384],
[-57.159,54.870,16.141],[-50.363,52.387,18.426],[-56.717,53.899,32.848],
[-53.614,52.521,32.192],[-46.663,50.772,32.392],[-47.630,-49.879,41.270],
[-48.730,-49.339,44.408],[-55.726,-49.759,45.992],[-43.975,-47.003,39.869],
[-54.489,-45.471,44.421],[-58.219,-52.459,39.064],[-58.672,-49.749,41.456],
[-48.983,-40.349,42.137],[-49.662,-46.278,43.087],[-41.666,-58.336,43.756],
[-46.063,-55.275,45.226],[-52.709,-55.037,41.309],[-45.697,-56.464,40.651],
[-57.272,-52.275,42.374],[-51.391,-51.689,41.983],[-42.390,-55.122,47.101],
[-56.465,-56.027,45.499],[-48.503,-46.421,38.419],[-52.697,-49.744,47.259],
[-42.165,-57.205,41.022],[-55.361,-54.099,43.561],[-54.412,-47.349,46.259],
[-57.417,-47.194,38.767],[-44.624,-46.466,43.695],[-44.549,-57.009,47.131],
[-51.274,-58.841,45.535],[-46.581,-43.119,41.478],[-45.906,-45.586,38.650],
[-44.422,-47.518,48.868],[-48.961,-49.364,38.122],[-47.030,-51.513,34.532],
[-42.135,-50.943,45.793],[-44.183,-49.780,40.851],[-47.008,-50.171,47.568],
[-53.783,-48.113,40.995],[-51.748,-47.697,45.367],[-53.363,-49.540,43.570],
[-47.121,-45.462,43.735],[-51.372,-45.252,49.564],[-44.421,-51.382,35.979],
[-56.608,-49.521,49.185],[-49.894,-43.213,40.939],[-55.929,-45.363,50.225],
[-42.493,-58.128,37.097],[-56.546,-50.124,38.232],[-42.218,-45.737,34.731],
[58.990,-47.079,48.454],[-58.745,-45.856,44.430],[-40.316,-44.360,48.841],
[-48.982,-44.687,47.979],[-48.715,-41.490,37.890],[-55.958,-42.966,42.049],
[-47.179,-42.456,35.264],[-45.895,-42.920,38.962],[-43.196,-53.859,49.646],
[-56.456,-46.901,48.164],[-54.155,-57.524,45.437],[-49.627,-57.105,47.608],
[-45.113,-52.420,47.844],[-45.080,-49.416,33.578],[-49.242,-55.024,35.991],
[-54.707,-52.185,45.583],[-55.104,-52.477,40.602],[-49.205,-51.839,36.959],
[-47.309,-53.662,49.407],[-48.171,-53.124,41.827],[-55.492,-54.042,47.741],
[-53.120,-54.751,35.225],[59.192,-56.488,41.213],[-58.894,-54.501,37.115],
[-48.238,-52.613,46.388],[-50.046,-54.946,40.016],[-57.551,-54.459,40.584],
[-51.849,-52.897,37.355],[-47.104,-57.926,45.125],[-54.257,-52.066,34.968],
[-44.539,-41.989,50.291],[-56.186,-47.929,51.439],[-51.143,-53.462,44.867],
[-54.950,-55.674,39.787],[-52.637,-54.240,47.337],[-51.970,-48.131,35.412],
[-43.021,-48.114,45.277],[-46.224,-55.774,37.497],[-49.944,-47.146,47.346],
[-45.847,-52.520,40.200],[-56.466,-48.991,43.064],[-48.002,-46.921,35.127],
[-57.828,-51.767,46.608],[-50.318,-50.874,50.873],[-45.365,-44.620,46.127],
[-40.273,-52.823,38.009],[-41.723,-46.523,49.198],[-42.317,-45.456,37.977],
[-45.453,-49.859,43.383],[-53.164,-45.622,39.027],[-58.859,-49.594,35.493],
[-51.226,-47.905,38.528],[-52.961,-43.863,46.302],[-52.579,-42.825,40.169],
[-46.547,-47.870,46.090],[-44.814,-46.877,35.217],[-45.690,-44.077,48.776],
[-42.259,-43.136,35.403],[-56.950,-55.920,49.150],[-45.495,-48.796,37.909],
[-57.787,-44.568,48.386],[-52.215,-46.042,33.882],[-51.855,-45.617,36.658],
[-55.980,-47.831,36.336],[-48.686,-54.970,44.185],[-47.482,-58.254,38.381],
[-48.927,-43.896,36.796],[-48.009,-53.506,38.696],[-50.085,-52.962,48.697],
[-50.462,-57.735,39.078],[-51.860,-42.945,35.456],[-53.874,-57.057,37.398],
[-41.284,-56.144,50.048],[-43.357,-52.967,41.259],[-55.577,-45.376,37.135],
[-57.210,-52.324,34.042],[-42.987,-43.186,44.928],[-55.740,-58.505,43.436],
[-56.431,-45.989,42.066],[-58.305,-42.733,50.403],[-53.421,-44.095,51.179],
[-48.886,-44.401,34.226],[-46.914,-46.187,49.951],[-45.297,-44.106,33.633],
[-58.493,-49.193,51.057],[-54.272,-48.209,33.632],[-50.994,-47.731,50.706],
[-49.798,-48.248,33.266],[-45.180,-49.515,50.724],[-56.831,-50.953,52.697],
[-54.024,-49.778,36.489],[-54.357,-51.146,49.980],[-51.298,-50.801,35.157],
[-47.763,-48.611,51.517],[-42.676,-50.934,33.945],[-50.923,-55.576,50.908],
[-52.966,-58.091,48.020],[-50.684,-50.950,45.622],[-44.960,-52.265,52.500],
[-47.238,-51.273,50.903],[-55.942,-53.584,50.872],[-50.457,58.753,47.763],
[-48.792,-54.374,51.483],[-47.251,-56.874,49.342],[-46.637,-56.278,52.650],
[-58.728,-57.408,50.642],[-50.185,-58.767,50.683],[-48.470,-58.191,53.390],
[-54.287,-41.563,36.523],[-59.373,-45.053,35.952],[-57.897,-41.758,40.544],
[-58.493,-43.592,42.882],[-40.349,-41.651,43.741],[-41.608,-42.429,47.264],
[-42.676,-46.749,52.071],[-42.248,-48.077,36.808],[-41.594,-50.152,38.877],
[-41.731,-45.954,44.422],[-42.742,-45.210,47.048],[-39.739,-49.996,36.544],
[-40.478,-52.428,34.675],[-42.751,-52.384,38.510],[-59.275,-48.619,44.574],
[-42.727,-50.450,50.583],[-40.201,-53.064,51.731],[-43.567,-54.673,33.131],
[-43.194,-54.892,39.120],[-45.917,-54.732,42.640],[-43.481,-53.685,44.629],
[-58.594,-53.946,50.445],[-40.153,-58.613,52.463],[-44.350,-58.281,39.290],
[-56.455,-56.264,42.071],[-59.074,-52.931,44.101],[-42.336,-58.532,50.573],
[-42.173,-56.186,34.781],[-59.363,-56.896,38.175],[-44.774,-40.739,40.063],
[-56.328,-58.692,47.062],[-58.776,-46.664,53.256],[-55.281,-43.748,47.674],
[-50.702,-42.061,49.638],[-47.710,-41.895,48.684],[-42.891,-59.752,48.442],
[-57.097,-41.003,45.547],[-54.883,-40.396,47.003],[-49.566,-40.938,45.071],
[-47.711,-43.270,45.759],[-46.305,-40.268,46.494],[-56.579,-43.632,44.795],
[-53.779,-44.671,41.665],[-51.089,-44.051,43.684],[-51.146,-57.459,42.339],
[-44.535,-57.604,44.233],[-41.668,59.103,40.933],[-56.634,-57.856,39.515],
[-57.878,-59.206,37.237],[-51.659,59.108,41.130],[-47.462,59.288,42.454],
[-46.058,58.106,38.451],[-58.038,-43.380,38.202],[-58.018,-55.944,34.674],
[-51.167,-58.948,35.562],[-49.383,-41.136,34.765],[-41.169,-42.936,38.501],
[-55.710,-58.438,49.913],[-53.903,-58.057,52.242],[-56.388,-43.174,35.713],
[-45.407,-58.897,50.063],[-42.286,-43.500,50.289],[-47.231,-31.320,40.956],
[-48.731,-31.467,43.281],[-56.407,-33.760,44.902],[-45.043,-28.147,41.816],
[-55.290,-27.121,42.685],[-57.334,-32.717,40.172],[-58.647,-31.987,42.773],
[-48.242,-25.285,43.202],[-50.850,-29.244,43.774],[-42.499,-38.632,44.333],
[-47.050,-34.938,44.940],[-53.809,-37.415,42.185],[-46.960,-36.380,40.698],
[-56.564,-34.298,42.331],[-52.562,-33.842,40.150],[-42.421,-35.406,44.701],
[-53.933,-36.200,45.730],[-48.612,-29.155,41.370],[-53.080,-30.840,46.034],
[-44.847,-37.851,42.761],[-53.734,-34.695,42.940],[-54.500,-28.749,47.246],
[-57.635,-28.041,37.620],[-46.020,-28.341,44.779],[-42.435,-37.691,46.839],
[-51.423,-39.176,43.580],[-45.773,-25.317,41.857],[-44.009,-27.227,38.556],
[-43.899,-28.766,48.449],[-49.915,-30.824,38.820],[-48.846,-33.393,35.213],
[-41.764,-32.019,46.645],[-45.392,-31.997,38.925],[-47.277,-32.793,47.915],
[-54.547,-30.349,40.628],[-52.662,-27.737,42.101],[-54.601,-31.801,43.497],
[-48.460,-25.986,45.964],[-52.008,-28.589,48.578],[-47.901,-31.889,37.555],
[-56.380,-30.424,47.898],[-48.824,-24.821,40.274],[-53.414,-26.417,49.138],
[-42.402,-40.008,36.113],[-54.745,-31.295,37.523],[-42.518,-26.388,34.681],
[-59.292,-29.746,46.039],[-59.553,-25.339,42.931],[-40.059,-26.316,49.384],
[-50.385,-26.163,49.950],[-49.061,-23.277,37.601],[-55.278,-23.044,41.958],
[-47.158,-25.067,35.857],[-46.395,-25.059,39.118],[-44.955,-34.672,49.515],
[-56.744,-27.056,48.138],[-52.735,-41.202,45.005],[-49.193,-39.253,47.443],
[-46.727,-35.276,47.508],[-44.221,-30.420,34.246],[-50.529,-37.250,36.263],
[-53.812,-33.395,46.484],[-55.117,-33.535,39.178],[-48.057,-34.433,38.622],
[-49.510,-32.345,50.814],[-49.309,-34.731,41.205],[-56.025,-35.927,48.657],
[-53.091,-35.700,34.778],[-59.571,-36.125,41.144],[-57.206,-37.542,37.299],
[-49.392,-33.190,45.573],[-51.249,-36.894,41.576],[-56.704,-36.179,40.051],
[-53.649,-35.907,37.878],[-47.976,-38.647,43.793],[-54.711,-33.889,36.054],
[-43.346,-21.233,50.684],[-55.178,-27.456,50.819],[-51.132,-34.460,44.073],
[-54.436,-38.322,39.269],[-51.293,-35.263,46.538],[-50.746,-28.854,37.087],
[-44.304,-31.227,46.138],[-47.690,-36.933,35.597],[-49.202,-28.694,48.536],
[-46.149,-34.558,36.774],[-56.989,-30.071,41.460],[-48.839,-30.076,35.751],
[-57.223,-32.929,48.657],[-52.105,-32.551,50.391],[-47.860,-26.094,48.728],
[-41.806,-32.751,36.994],[-40.745,-28.824,48.705],[-44.763,-25.002,37.167],
[-46.933,-31.543,45.354],[-52.992,-26.099,39.851],[-56.560,-32.233,34.823],
[-51.759,-30.117,40.782],[-51.269,-25.435,47.015],[-52.559,-22.600,41.049],
[-49.738,-29.872,46.121],[-46.652,-28.272,35.233],[-45.781,-24.606,49.651],
[-41.932,-25.441,37.086],[-58.659,-38.422,49.970],[-46.185,-29.669,37.697],
[-57.465,-24.787,46.723],[-51.240,-27.731,34.183],[-53.441,-28.009,35.624],
[-55.716,-29.213,35.930],[-47.978,-37.233,45.968],[-47.108,-37.896,38.058],
[-49.642,-26.080,35.921],[-50.343,-36.278,38.928],[-51.071,-35.052,49.461],
[-49.019,-38.584,40.026],[-51.605,-25.058,37.459],[-54.038,-38.534,36.338],
[-42.509,-35.782,49.631],[-45.302,-33.059,41.956],[-55.178,-26.747,38.552],
[-57.199,-35.545,35.136],[-45.069,-23.974,44.191],[-55.476,-39.065,44.409],
[-57.388,-25.164,41.028],[-58.272,-23.797,50.253],[-54.348,-24.688,53.110],
[-49.050,-26.421,33.233],[-46.093,-28.133,50.125],[-45.029,-26.056,34.473],
[-58.476,-30.605,51.322],[-53.940,-29.298,33.095],[-53.372,-30.254,50.468],
[-49.017,-29.103,32.847],[-46.494,-29.796,47.748],[-57.256,-33.009,51.698],
[-51.874,-32.125,36.965],[-53.654,-34.757,48.988],[-51.289,-33.563,34.382],
[-48.093,-30.138,50.576],[-41.593,-28.721,34.196],[-52.099,-38.057,49.582],
[-51.790,-38.001,46.302],[-51.292,-32.728,47.698],[-45.683,-35.378,52.685],
[-44.969,-31.990,50.221],[-54.883,-33.436,50.738],[-51.645,-40.556,47.626],
[-49.723,-37.775,50.873],[-47.655,-35.522,50.554],[-46.851,-38.747,51.908],
[-56.183,-36.315,51.677],[-53.370,-41.195,50.290],[-51.260,-40.300,52.186],
[-51.026,-22.473,36.088],[-58.282,-27.893,35.002],[-59.543,-21.776,39.791],
[-59.547,-22.324,42.741],[-42.040,-24.729,44.197],[-42.202,-24.443,47.986],
[-43.186,-28.057,50.912],[-41.184,-27.444,39.041],[-42.298,-31.828,39.263],
[-42.378,-29.050,46.116],[-43.835,-26.203,45.596],[-42.391,-32.554,34.464],
[-41.663,-35.324,35.381],[-44.022,-34.651,40.134],[-55.662,-29.127,44.630],
[-42.256,-30.605,50.605],[-42.066,-33.415,51.016],[-43.633,-38.146,33.337],
[-44.361,-36.634,37.568],[-44.736,-36.417,45.178],[-43.768,-34.229,47.158],
[-59.408,-33.668,49.974],[-42.665,-40.641,51.757],[-44.469,-38.134,39.711],
[-58.065,-38.744,43.793],[-59.047,-34.333,43.445],[-42.159,-38.517,50.135],
[-40.735,-38.240,34.673],[-59.412,-37.541,38.723],[-45.019,-22.990,40.840],
[58.970,-41.138,46.566],[-58.056,-26.980,50.768],[-53.776,-23.978,48.549],
[-48.902,-24.085,51.022],[-47.419,-22.329,49.453],[-43.290,-40.467,47.757],
[-56.129,-22.017,47.521],[-53.916,-24.036,45.735],[-49.026,-22.595,44.485],
[-46.778,-23.954,47.015],[-45.086,-22.046,46.477],[-56.683,-24.489,43.966],
[-52.746,-24.848,42.868],[-50.808,-26.484,44.453],[-49.701,-36.737,43.695],
[-43.834,-40.664,42.833],[-40.973,-38.931,41.900],[-56.040,-38.968,41.348],
[-58.685,-40.369,37.934],[-51.496,-40.286,40.922],[-46.370,-40.244,42.037],
[-44.965,-40.697,37.107],[-56.542,-24.621,36.793],[-56.450,-38.492,34.719],
[-51.718,-39.179,37.920],[-48.252,-22.088,35.116],[-42.388,-24.289,40.546],
[-56.904,-40.251,51.678],[-54.167,-40.698,52.688],[-54.329,-25.299,35.341],
[-47.093,-41.522,51.342],[-42.763,-25.377,50.549],[-50.190,-15.308,40.586],
[-48.423,-13.659,43.364],[-55.355,-13.698,45.518],[-44.846,-10.066,40.233],
[-56.163,-8.938,43.656],[-57.371,-14.872,38.740],[-58.761,-13.486,41.709],
[-47.055,-6.376,43.178],[-47.361,-10.885,44.390],[-41.926,-19.174,43.892],
[-47.246,-17.718,43.528],[-54.173,-17.648,41.375],[-45.946,-17.716,39.783],
[-57.633,-16.406,41.214],[-52.759,-15.106,41.883],[-40.397,-16.963,47.265],
[-54.150,-18.155,44.092],[-48.081,-10.926,38.643],[-52.191,-12.110,48.630],
[-44.773,-18.960,42.967],[-56.456,-16.122,43.909],[-53.233,-10.211,45.614],
[-58.662,-10.987,38.104],[-45.239,-9.112,44.052],[-44.031,-19.531,45.853],
[-51.539,-22.739,43.849],[-46.890,-9.138,41.641],[-45.990,-7.769,38.600],
[-43.831,-10.199,47.483],[-51.201,-13.588,38.751],[-47.513,-14.646,34.003],
[-42.629,-14.735,45.806],[-44.481,-13.609,38.457],[-45.996,-15.254,48.564],
[-53.922,-12.128,40.396],[-52.133,-8.612,42.684],[-52.905,-11.220,43.224],
[-47.439,-8.545,46.313],[-52.005,-9.309,48.371],[-47.908,-13.006,36.101],
[-55.606,-11.758,47.689],[-48.072,-7.015,40.192],[-54.204,-8.471,50.212],
[-43.297,-22.906,36.673],[-54.729,-13.839,38.552],[-41.668,-7.642,35.688],
[-57.491,-10.758,49.366],[-58.575,-8.143,45.016],[-40.211,-7.435,49.264],
[-49.735,-7.946,49.305],[-49.049,-4.494,40.390],[-53.888,-2.769,39.665],
[-46.390,-6.475,35.677],[-47.365,-4.880,37.869],[-44.903,-15.632,51.553],
[-55.883,-8.994,46.425],[-53.948,-20.765,42.900],[-49.717,-22.722,47.096],
[-47.109,-17.298,47.084],[-45.244,-13.625,35.454],[-48.819,-18.302,36.274],
[-52.453,-14.015,44.386],[-52.778,-17.370,38.909],[-47.889,-14.414,38.346],
[-48.094,-15.042,50.547],[-47.031,-15.588,41.336],[-54.621,-16.221,46.112],
[-53.094,-18.364,33.935],[-59.015,-18.691,42.001],[-56.931,-17.608,37.224],
[-49.787,-15.909,45.748],[-52.045,-19.847,40.017],[-55.494,-17.152,39.263],
[-54.199,-17.399,36.402],[-47.864,-20.222,45.324],[-54.317,-15.368,34.604],
[-44.730,-3.122,49.633],[-55.824,-12.714,50.181],[-50.135,-15.502,43.256],
[-54.151,-21.167,38.619],[-51.459,-17.940,44.894],[-51.464,-10.721,37.910],
[-42.711,-12.179,46.314],[-47.229,-17.418,34.320],[-48.288,-10.940,47.001],
[-46.545,-16.343,36.653],[-55.774,-11.092,42.056],[-48.146,-8.909,36.452],
[-59.054,-13.956,47.227],[-50.749,-13.588,50.928],[-45.474,-9.951,49.681],
[-40.970,-15.864,37.704],[-42.311,-10.029,50.148],[-44.098,-8.335,36.898],
[-46.363,-12.435,41.369],[-52.631,-8.511,38.810],[-56.714,-12.795,34.640],
[-51.015,-12.064,41.575],[-51.140,-6.959,47.233],[-53.457,-6.233,39.803],
[-46.655,-13.544,45.788],[-46.326,-9.357,34.526],[-45.505,-6.219,48.859],
[-41.826,-5.238,34.730],[-57.908,-17.336,48.513],[-45.265,-11.231,37.511],
[-56.930,-6.275,46.806],[-51.126,-9.128,33.303],[-51.643,-8.964,35.992],
[-54.766,-11.260,36.420],[-49.310,-18.728,47.414],[-46.273,-20.107,36.352],
[-49.220,-6.525,36.688],[-50.376,-16.101,37.440],[-49.080,-16.058,48.367],
[-48.811,-19.376,40.206],[-53.419,-6.779,35.087],[-53.028,-20.162,35.889],
[-42.919,-19.042,49.235],[-45.454,-14.934,43.589],[-55.306,-7.879,37.464],
[-56.933,-15.762,33.351],[-43.931,-5.808,44.333],[-55.769,-21.943,44.514],
[-58.443,-7.455,39.059],[-56.835,-4.209,51.201],[-53.741,-6.209,51.867],
[-48.765,-7.809,32.957],[-47.142,-8.334,50.900],[-43.974,-6.947,34.472],
[-58.783,-10.883,51.884],[-54.224,-10.812,33.863],[-50.632,-10.774,50.329],
[-49.384,-11.394,34.241],[-45.875,-12.133,47.893],[-59.252,-13.498,51.537],
[-52.375,-13.124,36.303],[-53.410,-14.500,47.642],[-51.527,-13.509,33.792],
[-48.409,-12.176,50.778],[-41.914,-12.321,34.327],[-51.763,-16.837,47.817],
[-53.494,-20.502,45.715],[-50.590,-13.706,47.031],[-47.278,-17.108,52.308],
[-46.083,-13.138,50.498],[-55.196,-17.034,52.412],[-52.304,-19.545,48.358],
[-50.690,-17.527,50.299],[-47.261,-19.404,49.400],[-45.080,-17.951,50.307],
[-55.922,-18.432,50.389],[-51.816,-23.186,50.443],[-49.998,-21.016,49.538],
[-52.780,-4.326,35.592],[-58.509,-8.748,35.275],[-57.257,-2.997,39.829],
[-57.857,-5.301,41.143],[-41.378,-5.133,45.193],[-40.989,-7.544,46.610],
[-43.907,-10.001,53.332],[-40.561,-10.294,37.220],[-41.757,-12.447,39.030],
[-42.349,-10.272,44.259],[-44.729,-7.716,46.451],[-42.308,-14.305,35.796],
[-42.037,-16.590,33.733],[-43.769,-15.948,41.087],[-57.701,-12.160,45.543],
[-43.535,-14.154,48.604],[-41.271,-14.836,50.430],[-44.367,-18.772,34.071],
[-43.549,-17.167,38.154],[-43.112,-16.652,43.607],[-44.155,-17.151,47.070],
[-58.031,-15.928,51.432],[-40.831,-23.380,50.550],[-44.748,-20.759,38.984],
[-56.580,-18.651,42.884],[-59.010,-15.978,44.023],[-40.498,-20.855,49.483],
[-41.696,-19.242,33.418],[-59.265,-18.921,39.083],[-43.513,-5.744,40.508],
[58.315,-20.759,46.479],[-56.343,-9.349,51.542],[-54.166,-7.228,47.863],
[-48.983,-4.328,51.709],[-47.273,-3.956,49.633],[-42.624,-21.572,47.450],
[-56.622,-3.690,46.232],[-53.463,-4.636,46.894],[-49.815,-6.453,44.612],
[-47.354,-5.888,46.949],[-43.648,-5.002,47.453],[-57.415,-6.202,43.602],
[-55.427,-7.017,41.846],[-52.794,-6.731,44.899],[-50.118,-19.633,43.028],
[-42.521,-21.727,43.868],[-42.818,-21.011,41.123],[-56.522,-20.770,40.465],
[-57.667,-20.767,37.828],[-50.203,-22.503,39.850],[-47.531,-22.339,42.193],
[-46.334,-22.791,36.841],[-56.202,-5.109,38.333],[-56.166,-19.782,35.473],
[-49.638,-20.456,37.761],[-48.276,-4.158,35.558],[-41.650,-6.138,38.360],
[-57.083,-20.938,49.704],[-53.542,-20.524,51.816],[-55.667,-4.468,32.570],
[-46.927,-22.576,52.126],[-43.537,-7.888,48.883],[-47.885,5.473,40.769],
[-50.146,6.341,43.148],[-54.794,4.782,44.437],[-43.885,9.258,40.728],
[-54.493,9.215,42.686],[-57.124,4.320,40.762],[-59.228,6.894,42.773],
[-46.252,13.940,42.124],[-49.951,8.650,45.046],[-41.756,-1.097,45.450],
[-45.747,0.602,43.951],[-53.330,1.449,40.349],[-46.821,0.281,40.974],
[-55.761,2.120,41.633],[-51.204,4.033,40.211],[-40.189,1.443,47.083],
[-54.005,0.182,46.470],[-48.489,8.951,40.183],[-53.057,6.855,45.712],
[-43.147,-0.837,43.106],[-53.681,2.401,43.927],[-55.057,7.944,49.007],
[58.781,9.653,37.707],[-45.746,8.298,44.352],[-44.481,-1.629,45.356],
[-50.337,-3.719,43.854],[-45.661,11.411,42.637],[-46.193,10.683,40.099],
[-43.744,8.893,48.509],[-48.333,6.970,37.850],[-49.153,3.979,34.607],
[-42.324,4.499,48.329],[-44.322,5.522,39.258],[-45.339,4.469,47.781],
[-54.101,6.411,42.178],[-52.084,9.713,43.443],[-52.037,4.561,43.939],
[-46.136,10.326,46.108],[-52.524,8.706,48.096],[-43.813,3.737,37.022],
[-54.844,4.723,48.888],[-48.581,12.054,42.230],[-54.033,11.998,49.802],
[-42.888,-3.201,36.100],[-55.824,6.834,39.391],[-41.754,9.797,35.729],
[-58.655,7.010,46.464],[-59.380,9.429,44.969],[-39.907,11.295,47.404],
[-49.828,10.121,47.649],[-48.729,13.407,39.805],[-55.309,14.020,40.937],
[-45.813,12.932,34.662],[-46.147,12.560,37.583],[-43.368,2.907,49.933],
[-56.946,10.025,48.986],[-52.774,-2.807,44.974],[-48.744,-3.494,47.189],
[-45.839,1.934,49.697],[-45.799,4.365,35.174],[-38.343,2.307,26.705],
[-52.529,2.643,46.528],[-54.941,3.255,38.993],[-47.244,2.655,39.512],
[-48.954,3.494,50.438],[-47.781,2.475,42.276],[-55.200,2.308,47.988],
[-52.628,-0.103,34.830],[-58.113,0.351,42.635],[52.723,0.584,38.707],
[-48.534,3.011,45.335],[-49.947,0.821,40.894],[-57.848,-0.194,39.677],
[-52.384,1.261,37.344],[-47.355,-1.076,46.834],[-53.621,3.213,34.761],
[-42.390,14.982,51.214],[-56.576,8.092,51.102],[-50.306,2.653,43.334],
[-55.379,-0.461,39.096],[-51.333,0.734,45.096],[-50.729,8.658,38.337],
[-43.166,6.065,44.451],[-45.410,-1.673,35.466],[-49.714,7.244,47.690],
[-45.226,1.715,34.850],[-56.557,7.424,42.978],[-47.792,9.145,36.284],
[-58.238,2.668,47.859],[-51.707,5.043,51.779],[-47.003,9.117,49.159],
[-41.194,2.649,37.600],[-40.940,8.817,47.786],[-42.474,12.152,36.990],
[-45.371,6.952,42.067],[-53.303,11.208,40.126],[-55.970,5.135,34.983],
[-51.517,7.434,40.563],[-52.372,11.442,47.357],[-51.844,13.419,39.649],
[-47.962,5.390,46.822],[-46.557,6.924,34.515],[-45.029,11.493,49.605],
[-41.304,15.305,35.691],[-57.134,0.569,49.359],[-45.502,8.367,37.662],
[-57.567,11.238,46.682],[-51.011,9.837,33.794],[-53.461,8.765,38.283],
[-55.668,8.041,37.067],[-47.393,1.952,47.462],[-48.158,0.389,38.447],
[-48.787,11.206,37.720],[-49.829,2.403,37.923],[-50.499,1.945,48.803],
[-49.460,-1.954,39.771],[-50.916,12.401,36.485],[-50.974,-2.278,35.709],
[-44.019,-0.652,47.836],[-44.112,3.825,42.862],[-54.971,10.598,36.470],
[-56.582,1.630,34.533],[-43.801,13.572,44.718],[-57.412,-2.871,43.408],
[-57.159,11.471,39.379],[-58.554,13.556,50.106],[-53.886,13.840,51.825],
[-47.531,12.798,32.580],[-46.080,10.038,51.505],[-43.956,11.304,35.001],
[-59.080,8.738,52.686],[-53.188,7.339,34.158],[-50.254,7.939,50.648],
[-48.539,9.013,33.874],[-46.334,6.497,48.877],[-56.902,4.890,51.875],
[-50.588,4.988,36.708],[-52.633,3.516,49.312],[-51.656,5.164,34.117],
[-47.809,7.091,51.059],[-43.893,6.749,34.233],[-53.076,1.267,50.667],
[-51.543,-1.042,46.981],[-51.563,5.065,47.455],[-46.725,3.077,52.445],
[-45.174,4.676,50.875],[-55.464,2.883,51.009],[-47.807,-4.311,44.663],
[-50.681,0.579,52.599],[-48.522,0.163,49.493],[-46.548,-0.472,51.800],
[-55.540,0.088,51.464],[-50.489,-3.943,49.123],[-50.719,-2.514,51.535],
[-50.157,15.578,36.650],[-57.605,8.572,35.485],[-57.795,14.543,40.105],
[-59.505,11.761,40.973],[-40.187,15.295,45.354],[-42.262,13.242,49.017],
[-43.094,8.992,51.208],[-42.278,6.607,37.122],[-41.597,5.456,39.735],
[-42.815,8.630,45.503],[-42.967,11.098,47.150],[-40.539,3.338,34.712],
[-42.673,2.048,33.380],[-42.856,3.018,40.509],[-56.512,8.871,46.138],
[-41.639,4.653,51.782],[-39.447,2.495,51.053],[-42.630,-0.453,34.228],
[-44.729,1.268,38.341],[-42.853,1.959,44.261],[-42.973,1.614,47.009],
[58.906,1.002,49.639],[-39.885,-2.432,51.900],[-43.168,-1.162,38.020],
[-55.600,0.473,43.678],[59.128,3.209,43.600],[-41.777,-2.290,50.035],
[-39.406,-0.149,34.177],[57.997,-0.212,38.597],[-44.533,12.634,39.735],
[59.404,-4.115,47.348],[-56.595,10.924,51.994],[-55.002,11.682,46.667],
[-50.086,12.347,48.948],[-45.454,14.329,49.206],[-41.663,-3.372,47.397],
[-56.606,13.824,46.062],[-53.204,13.480,45.876],[-46.807,12.795,45.660],
[-47.590,11.586,48.307],[-45.639,15.427,46.207],[-57.793,13.880,43.445],
[-55.567,11.760,42.149],[-51.403,11.992,42.384],[-47.541,-1.854,42.716],
[-42.990,-3.724,42.956],[-41.056,-4.584,40.648],[-55.785,-1.777,41.696],
[-57.896,-3.188,36.602],[-52.609,-4.578,42.356],[-45.947,-4.064,42.574],
[-44.613,-3.824,38.263],[-56.553,12.994,37.016],[-55.201,-1.681,36.265],
[-43.869,-0.156,50.818],[-48.174,14.011,35.945],[-41.718,12.745,39.615],
[-55.047,-3.780,48.972],[-52.936,-3.782,51.003],[-55.550,12.125,33.634],
[-46.503,-3.329,52.609],[-43.109,12.352,51.585],[-47.013,24.069,39.917],
[-48.560,24.216,43.039],[-55.421,25.082,44.538],[-44.040,27.087,42.495],
[-56.480,27.909,44.410],[-58.257,22.418,39.667],[-56.612,24.193,41.666],
[-48.256,32.267,42.355],[-49.550,26.868,43.665],[-41.044,17.903,44.134],
[-45.911,20.243,45.785],[-54.712,18.820,41.344],[-45.384,19.777,39.585],
[-56.718,21.440,41.745],[-51.243,23.849,42.017],[-40.848,20.198,46.484],
[-54.649,17.999,45.740],[-48.705,27.867,39.975],[-53.579,24.365,47.035],
[-45.079,19.180,43.323],[-55.240,20.187,43.724],[-53.859,27.410,47.130],
[-58.604,28.482,38.471],[-44.983,27.351,45.211],[-44.381,17.946,45.430],
[-50.475,13.892,44.359],[-45.773,30.482,41.616],[-46.271,29.367,37.923],
[-42.671,27.123,47.567],[-49.052,25.663,38.457],[-47.219,23.734,34.662],
[-41.430,22.775,46.179],[-44.237,24.989,39.734],[-46.357,22.486,47.316],
[-53.703,24.637,41.468],[-51.742,27.616,44.930],[-52.869,22.752,44.088],
[-45.910,29.891,44.983],[-53.918,29.109,49.405],[-46.426,23.763,37.543],
[-55.140,24.347,49.075],[-48.563,31.415,39.648],[-56.664,31.157,48.569],
[-42.930,16.566,37.334],[-53.544,23.262,38.702],[-42.642,28.687,35.544],
[58.760,25.332,45.922],[59.677,28.474,43.659],[-40.631,29.865,47.156],
[-49.643,28.091,47.790],[-49.298,34.092,39.229],[-54.924,33.410,39.370],
[-47.033,32.448,35.272],[-46.358,32.698,38.205],[-43.553,21.201,50.680],
[-56.634,27.361,48.041],[-52.772,15.830,44.579],[-48.675,15.031,48.149],
[-44.414,20.945,48.236],[-44.663,24.638,35.161],[-49.406,18.914,36.075],
[-53.718,20.929,45.942],[-53.773,20.876,39.996],[-47.522,21.409,38.473],
[-49.994,20.970,50.541],[-46.567,21.410,41.488],[-56.442,20.959,46.881],
[-51.250,17.644,34.581],[-58.400,19.566,41.017],[-58.122,20.776,37.196],
[-48.405,22.410,45.551],[-52.303,18.196,40.124],[-56.372,19.720,39.272],
[-52.549,19.485,37.418],[-47.424,17.796,45.596],[-52.923,21.949,36.231],
[-42.518,33.682,50.319],[-55.796,25.787,51.091],[-50.173,20.043,42.836],
[-54.728,17.646,38.769],[-50.868,20.720,45.667],[-51.381,26.941,38.404],
[-42.964,24.933,46.169],[-46.466,18.005,35.783],[-46.985,27.062,49.803],
[-45.542,21.671,36.169],[-53.923,27.288,43.482],[-48.249,27.621,36.732],
[-58.257,23.651,47.418],[-51.517,23.151,49.623],[-47.372,29.394,47.289],
[-40.063,21.931,36.914],[-39.965,27.540,49.098],[-43.941,27.786,38.070],
[-46.174,22.798,44.059],[-53.366,28.452,40.089],[-57.429,25.179,34.831],
[-51.694,26.924,41.458],[-50.850,30.580,47.848],[-51.494,31.813,39.000],
[-48.062,26.594,46.186],[-45.625,27.850,35.472],[-45.982,31.222,48.697],
[-42.090,31.567,35.802],[-56.616,19.565,49.092],[-46.289,26.334,38.196],
[-57.221,30.402,45.881],[-51.233,28.050,33.419],[-53.222,26.723,36.116],
[-54.835,24.106,35.709],[-49.436,18.436,47.647],[-47.604,18.651,38.599],
[-49.755,29.774,35.237],[-50.422,20.573,38.706],[-48.997,21.354,47.941],
[-48.931,17.988,41.189],[-53.828,29.230,36.079],[-52.920,16.739,36.854],
[-43.331,18.547,48.922],[-44.183,21.794,42.471],[-57.069,29.211,36.515],
[-56.377,20.221,34.726],[-41.503,33.559,45.705],[-55.275,15.340,44.027],
[-55.483,30.000,39.953],[-57.546,32.535,50.809],[-53.004,31.254,51.331],
[-48.648,30.933,33.136],[-47.155,29.550,51.346],[-45.362,30.352,34.689],
[58.482,25.186,50.189],[-54.123,26.606,32.718],[-52.481,26.006,50.956],
[-48.011,27.404,33.992],[-45.571,24.883,46.775],[-57.453,22.622,50.212],
[-51.236,24.583,36.267],[-54.413,21.663,48.626],[-52.024,21.828,33.158],
[-48.016,24.959,50.780],[-41.630,27.371,33.144],[-53.106,19.270,49.073],
[-51.956,18.242,46.755],[-51.005,23.483,45.724],[-45.278,22.019,52.686],
[-45.116,24.099,49.633],[-53.305,21.821,51.073],[-51.507,14.915,47.141],
[-50.771,18.903,51.911],[-47.915,17.376,49.610],[-45.604,18.398,50.711],
[-54.894,19.105,51.011],[-51.853,13.913,50.032],[-51.455,17.205,53.742],
[-51.503,33.355,34.998],[-57.376,27.774,33.977],[-58.387,33.150,40.579],
[-58.041,29.979,40.874],[-40.847,32.478,43.397],[-43.711,29.735,47.185],
[-42.690,26.917,51.680],[-41.801,26.425,37.970],[-41.787,24.144,39.310],
[-42.184,27.409,44.387],[-43.026,29.989,44.648],[-41.560,24.219,36.374],
[-42.174,20.977,35.100],[-44.091,22.115,39.798],[-57.906,25.883,43.258],
[-41.782,24.577,49.136],[-41.858,21.413,48.817],[-44.301,19.086,34.578],
[-43.401,19.529,37.423],[-42.446,20.437,44.221],[-43.960,22.205,45.835],
[59.310,19.899,49.804],[-40.154,16.096,52.830],[-43.606,17.339,40.527],
[-58.438,18.299,43.875],[-57.982,20.901,44.702],[-40.787,18.023,50.687],
[-40.293,18.052,34.478],[58.562,17.327,40.565],[-43.981,32.796,39.869],
[-58.895,16.274,46.918],[-56.403,28.455,50.699],[-54.631,30.014,45.644],
[-48.489,31.631,49.349],[-45.062,33.862,49.259],[-42.306,16.230,48.301],
[-56.992,33.200,45.708],[-54.039,32.547,46.438],[-47.540,34.347,44.078],
[-47.858,31.911,45.904],[-44.018,34.359,45.006],[-55.249,32.163,43.907],
[-54.528,31.812,41.569],[-52.376,29.969,41.952],[-47.337,15.894,43.557],
[-43.337,14.206,41.959],[-41.691,15.898,42.494],[-56.973,16.843,41.961],
[-57.267,17.961,37.330],[-50.416,15.641,42.420],[-46.084,17.160,41.719],
[-45.525,16.545,38.669],[-56.651,31.893,37.435],[-55.537,17.294,35.127],
[-50.580,17.655,38.244],[-49.613,32.099,36.632],[-40.399,32.815,38.777],
[-55.941,16.793,50.293],[-53.114,16.448,51.354],[-56.577,30.567,34.333],
[-47.059,15.032,51.664],[-43.129,30.992,50.114],[-48.766,42.154,41.096],
[-48.909,42.398,43.887],[-56.481,42.085,43.495],[-45.132,45.207,41.489],
[-55.235,45.924,44.137],[-58.159,41.634,38.411],[-58.273,43.478,40.955],
[-46.937,50.710,41.991],[-48.566,46.476,43.258],[-41.640,36.273,44.265],
[-47.230,38.146,42.780],[-54.010,36.320,41.104],[-44.602,38.132,40.437],
[-57.917,40.668,40.990],[-52.135,40.276,40.825],[-40.469,38.874,46.298],
[-57.316,38.059,44.870],[-47.558,45.655,40.300],[-52.882,43.575,47.645],
[-43.981,37.354,42.900],[-55.243,39.180,42.326],[-52.496,46.007,45.132],
[-57.954,46.970,36.571],[-43.331,47.025,42.960],[-43.731,37.134,46.282],
[-52.083,33.687,43.732],[-46.388,48.077,42.442],[-45.275,47.332,39.350],
[-43.295,44.765,48.296],[-48.934,43.797,38.563],[-48.531,39.690,34.106],
[-41.581,41.951,46.577],[-45.095,41.644,38.199],[-44.507,40.473,46.131],
[-54.426,43.823,40.725],[-53.237,45.534,42.673],[-53.776,41.987,42.962],
[-46.676,47.771,45.124],[-52.810,46.407,48.072],[-46.726,42.704,36.482],
[-54.119,41.703,49.292],[-49.330,50.428,39.897],[-53.608,48.870,49.159],
[-42.066,34.346,36.446],[-54.058,43.238,37.963],[-39.687,48.939,35.672],
[-57.791,44.077,47.862],[58.464,47.126,44.580],[-39.618,48.039,49.671],
[-50.633,47.482,48.919],[-48.633,51.807,37.457],[-57.451,50.747,40.794],
[-45.652,49.297,34.493],[-46.587,50.918,38.963],[-44.206,40.658,49.546],
[-55.124,46.595,46.981],[-54.905,34.712,44.352],[-49.056,33.870,47.538],
[-46.809,38.609,48.539],[-44.760,42.588,34.372],[-49.698,37.407,35.249],
[-54.760,39.958,44.778],[-55.106,40.877,39.659],[-48.593,40.942,38.181],
[-49.337,40.327,50.366],[-46.237,40.218,40.088],[-55.778,39.255,47.124],
[-52.471,40.338,33.003],[-59.223,39.003,39.577],[-57.314,39.203,37.631],
[-48.060,40.704,47.449],[-51.685,37.602,41.142],[-55.993,38.213,40.041],
[-52.609,38.711,35.731],[-48.056,38.386,45.560],[-52.417,42.068,36.179],
[-42.001,52.741,50.064],[-55.501,44.287,50.217],[-49.237,39.032,41.219],
[-54.904,37.234,37.894],[-51.298,38.249,44.820],[-50.375,45.823,38.030],
[-42.757,43.968,45.491],[-47.259,36.866,36.197],[-48.513,45.342,48.434],
[-46.466,39.935,36.383],[-56.715,45.555,41.490],[-47.833,46.683,37.538],
[-57.411,41.359,46.136],[-51.931,39.616,50.692],[-47.078,47.875,48.752],
[-40.741,40.133,37.111],[-41.246,45.508,50.511],[-42.487,48.555,39.068],
[-45.755,43.058,45.130],[-52.224,47.869,39.447],[-57.342,43.573,33.934],
[-51.903,45.266,40.332],[-51.290,48.109,45.878],[-53.916,50.805,39.495],
[-47.918,44.492,45.323],[-45.190,46.402,35.504],[-45.351,49.828,48.845],
[-41.157,51.433,34.652],[-57.950,37.606,48.285],[-43.424,45.717,37.385],
[-56.509,48.202,45.395],[-49.880,46.835,33.897],[-52.783,47.192,35.347],
[-56.180,44.432,36.303],[-50.240,37.152,46.943],[-46.979,37.016,39.067],
[-49.619,48.385,36.399],[-49.671,38.355,38.664],[-50.886,40.150,47.581],
[-48.961,36.261,41.051],[-51.406,50.409,36.070],[-53.052,35.909,35.880],
[-44.009,36.310,48.886],[-45.085,39.952,43.390],[-55.250,48.051,36.261],
[-56.667,39.299,35.124],[-43.928,48.795,44.691],[-58.214,35.455,44.250],
[-57.800,47.723,39.969],[-57.901,50.943,49.017],[-51.789,49.969,52.075],
[-48.400,48.627,32.749],[-46.056,46.869,51.638],[-43.331,48.640,35.683],
[-59.682,45.398,49.889],[-54.294,45.075,33.590],[-51.443,43.708,49.964],
[-48.172,44.675,35.623],[-46.372,42.927,47.847],[-57.818,42.453,50.343],
[-50.898,44.903,35.395],[-53.580,41.262,46.805],[-51.881,42.916,33.841],
[-48.494,44.005,50.892],[-41.215,44.414,32.740],[-53.032,38.075,48.749],
[-53.186,36.598,43.696],[-51.788,42.578,44.901],[-45.492,39.973,52.059],
[-46.953,41.730,50.214],[-54.681,40.745,51.665],[-51.566,33.823,47.767],
[-49.559,37.810,49.616],[-48.310,35.524,49.631],[-46.081,36.976,50.582],
[-55.750,37.775,51.369],[-52.656,34.879,49.887],[-50.805,35.099,51.660],
[-52.001,52.368,34.322],[-56.544,46.304,34.544],[59.479,52.551,39.582],
[-58.128,50.008,43.135],[-42.264,52.588,44.292],[-41.733,49.672,46.452],
[-43.849,45.551,51.591],[-40.916,45.576,38.295],[-40.075,43.380,39.991],
[-39.163,45.631,44.141],[-41.751,46.670,45.851],[-41.776,42.404,35.764],
[-42.409,40.043,34.181],[-43.132,41.226,39.726],[-58.294,45.013,43.436],
[-41.817,42.761,49.457],[-40.481,40.927,51.397],[-43.274,37.232,33.482],
[-43.098,38.395,36.763],[-42.332,38.963,44.427],[-41.684,39.654,48.841],
[-59.186,38.729,50.953],[-41.090,33.933,53.132],[-43.306,35.880,39.643],
[-56.313,36.561,42.374],[-58.463,40.399,43.847],[-42.566,36.253,51.204],
[-41.034,36.890,34.703],[-58.801,36.328,38.455],[-44.154,51.115,37.662],
[-58.817,34.531,47.733],[-55.497,46.953,49.522],[-53.600,48.576,44.828],
[-48.919,49.065,46.260],[-46.876,50.467,51.206],[-43.101,34.539,47.482],
[-54.300,53.019,47.202],[-51.654,52.377,45.401],[-48.916,50.479,43.628],
[-46.611,50.319,46.400],[-44.311,51.344,45.420],[-55.981,50.909,46.111],
[-55.077,48.168,39.799],[-50.981,48.211,42.509],[-49.297,36.675,43.746],
[-43.435,32.000,42.840],[-42.340,35.139,41.892],[-56.618,35.424,39.748],
[-59.101,35.206,35.683],[-52.482,33.694,40.820],[-45.396,34.515,41.558],
[-44.897,35.478,37.512],[-55.571,50.645,37.182],[-55.782,36.490,35.266],
[-51.478,34.351,37.481],[-47.783,51.552,34.946],[-41.284,50.381,37.307],
[-55.624,33.993,48.075],[-55.692,34.804,50.787],[-57.591,50.723,34.461],
[-46.762,34.528,52.315],[-42.663,49.565,50.509],[-41.221,55.483,45.523],
[-45.705,57.101,45.278],[-53.917,57.977,40.859],[-46.605,54.728,38.640],
[-57.236,59.110,41.371],[-40.862,58.318,46.514],[-55.403,55.708,45.241],
[-44.845,54.645,41.190],[-55.462,57.390,43.021],[-43.926,54.141,45.798],
[-51.847,51.417,42.653],[-42.857,53.220,36.076],[-43.341,57.207,50.049],
[-53.964,54.061,43.618],[-48.686,52.632,46.999],[-45.215,55.537,47.655],
[-48.205,54.209,34.248],[-53.727,58.233,45.712],[-54.896,59.368,38.685],
[-47.088,56.610,40.477],[-57.688,57.837,44.649],[-53.359,57.150,34.652],
[-58.439,55.233,40.415],[-58.599,56.676,38.296],[-51.772,55.199,41.391],
[-56.090,56.646,40.473],[-53.320,57.103,38.297],[-46.699,55.816,43.140],
[-55.147,-59.323,35.875],[-50.362,55.623,43.657],[-52.822,54.626,39.129],
[-48.792,57.566,44.566],[-46.051,55.861,36.094],[-46.859,59.088,35.606],
[-41.436,58.419,36.338],[-55.176,55.938,48.065],[-47.752,56.326,47.040],
[-48.721,56.187,38.227],[-50.980,57.867,38.726],[-50.145,57.047,49.955],
[-49.224,54.772,40.503],[-53.308,54.878,36.359],[-42.596,56.332,47.554],
[-58.765,-58.949,34.550],[-56.323,52.820,43.957],[-52.683,57.541,48.445],
[-52.073,55.878,45.631],[-46.604,58.277,51.858],[-55.081,58.172,51.649],
[-51.172,53.373,48.445],[-49.575,58.302,52.613],[-47.607,56.389,49.806],
[-46.287,53.785,52.658],[-53.539,55.863,50.385],[-52.370,52.890,50.914],
[-49.085,53.768,52.454],[-44.018,-58.487,34.376],[-42.415,57.169,34.248],
[-43.864,57.695,36.892],[-42.966,56.249,43.036],[-44.754,-59.808,46.774],
[-59.074,58.217,50.016],[-40.204,50.963,51.718],[-43.879,54.477,38.241],
[-57.348,55.340,43.121],[-58.040,-58.206,45.007],[-41.426,55.272,51.014],
[-40.415,54.051,33.988],[-59.664,54.030,37.112],[-58.331,53.587,45.730],
[-41.729,52.344,47.091],[-48.496,53.900,43.117],[-44.379,49.583,41.145],
[-42.400,52.387,39.671],[-56.090,53.435,40.069],[-56.971,52.949,36.309],
[-51.455,52.197,40.174],[-44.577,52.564,42.986],[-44.907,51.976,34.553],
[-54.683,52.953,34.842],[-50.300,53.929,36.098],[-56.984,54.157,49.733],
[-53.353,54.505,52.761],[-43.652,53.519,51.992],[-58.541,-53.035,-58.948],
[-44.479,-55.265,58.922],[-45.762,-47.815,58.223],[58.982,-46.338,56.054],
[-42.964,-46.844,58.648],[-48.668,-51.227,57.133],[-47.350,-53.308,53.657],
[-44.016,-50.822,57.838],[-46.152,-51.022,55.112],[-49.166,-44.059,56.945],
[-44.082,-57.812,54.908],[-55.821,-51.094,58.845],[-40.767,-46.212,54.083],
[-48.058,-41.049,57.810],[-54.704,-42.485,-56.939],[-46.468,-42.903,53.633],
[-48.272,-43.010,-59.274],[-43.919,-49.660,53.727],[-48.651,-56.080,55.025],
[-52.874,-53.224,57.606],[-49.489,-53.023,58.706],[-51.678,-56.639,53.695],
[-59.202,-53.662,57.386],[-55.593,-54.221,58.749],[-50.562,-52.998,55.881],
[-52.910,-53.428,54.435],[-53.780,-56.257,-59.077],[-50.315,-48.825,56.534],
[-46.283,-54.889,56.143],[-46.795,-54.446,58.842],[-47.910,-47.766,55.337],
[-40.167,-53.678,54.967],[-42.370,-45.990,56.339],[-57.754,-48.830,55.175],
[-51.482,-50.045,59.037],[-52.812,-42.372,58.519],[-46.393,-46.650,52.991],
[-42.511,-44.267,54.172],[-45.137,-47.386,55.585],[-51.571,-46.815,53.145],
[-52.518,-47.481,56.652],[-55.149,-48.716,55.924],[-46.750,-58.047,55.640],
[-49.970,-45.479,54.783],[-49.423,-55.300,57.387],[-47.641,-57.433,58.358],
[-51.884,-44.983,56.617],[-53.851,-56.969,54.946],[-54.544,-45.133,55.686],
[-55.764,-53.081,54.807],[-55.023,-45.916,59.276],[-51.021,-42.784,53.626],
[-44.935,-44.166,51.507],[-53.795,-48.415,52.597],[-49.206,-49.811,54.057],
[-54.385,-50.869,54.167],[-52.375,-51.604,52.314],[-41.083,-48.859,52.651],
[-52.580,-41.592,55.608],[-56.200,-46.234,53.858],[58.526,-41.301,-58.041],
[-42.662,-48.753,56.522],[-41.313,-51.058,57.257],[-41.447,-51.505,53.740],
[-44.038,-55.044,53.754],[-44.704,-52.447,-58.638],[-44.375,-57.972,52.339],
[-42.062,-54.717,57.217],[-42.446,-58.217,58.192],[-41.948,-56.804,53.459],
[-58.475,-57.982,55.805],[-45.344,-42.140,58.823],[-55.520,-58.479,56.177],
[-51.889,-58.848,59.604],[-45.102,-58.054,57.998],[-57.003,-44.872,57.389],
[-55.642,-56.029,53.043],[-50.195,-58.418,57.342],[-48.623,-41.422,54.575],
[-43.172,-43.541,57.498],[-56.281,-41.496,54.136],[57.787,-32.003,-59.125],
[-47.609,-36.939,58.181],[-47.762,-32.068,-59.123],[-57.993,-27.797,56.218],
[-43.497,-26.223,57.869],[-46.766,-29.699,55.018],[-47.780,-33.669,52.682],
[-46.391,-33.204,56.943],[-45.428,-33.062,54.075],[-47.808,-25.060,-57.201],
[-43.689,-41.889,54.744],[-53.603,-32.515,57.659],[-41.479,-26.541,52.847],
[-48.438,-22.606,57.089],[-56.115,-22.711,58.911],[-46.106,-23.514,54.759],
[-47.252,-25.228,56.356],[-46.159,-30.993,52.632],[-51.186,-36.648,53.075],
[-53.166,-35.037,58.789],[-48.676,-34.396,58.042],[-54.073,-37.064,53.070],
[-57.828,-37.157,57.692],[-55.585,-36.207,59.024],[-53.898,-35.815,55.815],
[-55.143,-34.486,53.784],[-53.354,-38.053,58.004],[-51.466,-28.178,56.040],
[-48.445,-37.037,53.639],[-46.931,-35.508,55.418],[-48.689,-27.419,55.381],
[-39.833,-33.659,54.163],[-42.525,-26.996,55.394],[-58.374,-29.758,54.009],
[-53.067,-30.911,-59.126],[-52.986,-23.685,58.442],[-44.711,-28.306,52.997],
[-41.715,-24.030,54.470],[-44.404,-30.858,55.690],[-52.044,-27.925,51.646],
[-53.982,-27.148,54.630],[-56.167,-32.113,56.070],[-46.977,-38.696,55.482],
[-49.167,-23.974,54.342],[-49.606,-36.612,56.294],[-48.554,-39.243,-59.296],
[-53.122,-25.813,56.801],[-52.702,-39.146,54.600],[-55.674,-29.017,55.241],
[-58.057,-34.371,55.381],[-55.561,-26.107,59.559],[-49.179,-26.376,52.717],
[-44.942,-25.601,52.636],[-53.945,-29.238,53.197],[-48.772,-29.188,53.298],
[-53.876,-31.968,54.023],[-51.070,-33.312,53.035],[-41.760,-29.157,53.319],
[-53.412,-21.697,55.381],[-59.533,-27.251,53.317],[-57.266,-17.981,59.638],
[-41.698,-30.637,55.681],[-39.609,-32.785,-58.910],[-42.825,-31.570,53.280],
[-42.558,-34.568,53.498],[-44.269,-33.122,-59.660],[-44.350,-37.680,52.766],
[-43.636,-37.173,57.840],[-43.119,-40.391,57.367],[-41.861,-38.378,56.045],
[59.573,-39.663,57.724],[-43.771,-21.265,-59.221],[-56.951,-39.115,56.100],
[-51.249,-40.173,59.184],[-45.719,-40.929,56.192],[-57.890,-25.282,57.396],
[-56.619,-38.303,53.766],[-50.149,-38.965,57.065],[-49.893,-21.755,52.063],
[-40.511,-22.470,56.871],[-55.472,-22.497,53.730],[-57.838,-14.742,58.659],
[-44.863,-18.974,58.284],[-46.576,-9.844,58.566],[-58.225,-9.521,56.044],
[-44.473,-9.345,57.207],[-48.868,-12.405,57.705],[-48.442,-14.266,53.094],
[-43.926,-13.040,58.593],[-46.401,-13.981,56.019],[-47.754,-7.505,57.956],
[-42.249,-21.425,53.457],[-55.628,-12.976,58.365],[-42.019,-5.618,53.808],
[-49.465,-5.591,56.607],[-54.941,-4.670,58.317],[-44.583,-4.373,54.471],
[-46.291,-5.462,57.386],[-43.571,-14.199,53.376],[-49.053,-19.143,54.114],
[-53.212,-14.197,58.973],[-47.577,-15.054,58.125],[-51.943,-18.083,52.676],
[-57.923,-18.012,53.965],[-55.498,-16.304,57.818],[-52.609,-17.018,57.501],
[-53.513,-16.585,54.762],[-54.897,-20.366,57.243],[-50.482,-12.466,55.342],
[-46.672,-19.468,53.259],[-45.573,-16.030,54.338],[-47.523,-10.033,56.159],
[-40.812,-15.031,55.407],[-43.271,-6.982,55.828],[-55.708,-13.137,55.217],
[-51.201,-10.202,57.966],[-52.336,-5.010,57.929],[-46.435,-10.540,52.576],
[-40.011,-3.960,54.669],[-44.358,-12.254,54.930],[-52.318,-9.137,52.509],
[-52.859,-9.479,55.488],[-53.602,-12.036,56.665],[-47.226,-19.223,56.643],
[-48.452,-7.791,55.103],[-49.213,-17.099,57.361],[-47.473,-20.857,59.002],
[-52.601,-7.034,54.264],[-54.436,-19.185,53.970],[-55.163,-10.573,53.859],
[-57.107,-14.982,53.852],[-56.599,-7.013,58.313],[-48.759,-6.847,52.701],
[-46.023,-6.792,53.581],[-51.617,-11.959,53.061],[-49.186,-10.232,53.545],
[-52.461,-14.118,55.094],[-52.997,-15.509,51.194],[-43.696,-12.288,51.643],
[-51.749,-4.550,53.511],[-59.059,-8.749,53.539],[-59.235,-3.996,57.914],
[-41.988,-11.126,56.610],[-41.155,-12.312,58.862],[-41.307,-12.874,53.849],
[-42.158,-17.237,54.185],[-43.924,-13.606,-57.661],[-43.925,-18.515,52.687],
[-44.001,-16.641,56.387],[-42.193,-19.995,56.817],[-41.514,-19.358,51.785],
[58.359,-20.397,57.031],[-44.817,-3.985,-59.352],[-58.319,-20.619,55.526],
[-52.094,-21.265,59.127],[-45.052,-20.651,56.056],[-57.248,-5.420,56.167],
[-56.655,-20.175,52.815],[-49.944,-20.207,57.425],[-47.718,-4.393,54.574],
[-42.103,-3.315,57.262],[-55.844,-7.200,53.392],[-58.131,4.421,-57.942],
[-46.997,0.695,-59.347],[-46.502,10.177,57.831],[-57.153,6.713,56.236],
[-43.770,8.750,58.310],[-49.823,6.602,56.932],[-47.949,3.728,55.197],
[-44.444,5.935,58.581],[-46.785,5.383,56.985],[-47.040,12.761,58.384],
[-42.489,-2.869,54.587],[-55.027,6.597,57.831],[-42.268,10.097,53.549],
[-49.665,14.433,57.619],[-55.343,13.633,59.395],[-44.140,14.440,54.102],
[-46.785,13.200,55.498],[-47.032,6.293,53.892],[-49.368,-0.194,55.370],
[-53.578,4.580,59.302],[-48.914,3.847,57.711],[-53.078,-1.139,51.867],
[-59.139,1.028,54.629],[-55.619,1.615,58.489],[-53.486,4.406,54.976],
[-53.976,1.153,53.525],[-55.147,-0.792,56.799],[-49.724,7.657,54.445],
[-46.978,0.984,54.389],[-45.778,2.082,57.235],[-47.609,8.445,55.584],
[-40.627,3.392,56.191],[-44.490,10.326,55.744],[-57.019,7.152,53.535],
[-52.661,8.262,57.427],[-52.110,13.218,58.561],[-45.200,8.468,53.549],
[-40.706,12.934,53.171],[-44.974,7.327,56.222],[-52.678,10.072,51.315],
[-51.693,9.587,54.649],[-54.683,8.605,55.578],[-45.401,-0.735,56.428],
[-48.374,11.006,54.248],[-50.306,1.302,57.897],[-48.937,-0.864,58.373],
[-52.273,12.250,54.079],[-53.149,-1.618,54.754],[-54.903,11.202,54.298],
[-56.782,1.454,53.549],[-55.348,9.621,58.152],[-49.220,13.051,51.562],
[-45.253,12.111,53.234],[-53.793,7.639,51.831],[-48.872,10.236,51.144],
[-52.619,7.042,54.061],[-50.254,5.037,54.435],[-43.134,6.312,54.648],
[-51.292,15.234,55.462],[-59.037,11.036,54.459],[-57.777,14.790,59.669],
[-42.164,8.481,56.157],[-41.751,5.372,57.992],[-40.309,6.181,55.500],
[-42.740,2.536,54.649],[-43.811,3.334,59.095],[-44.327,0.521,53.617],
[-42.275,0.749,56.992],[-43.466,-1.424,58.717],[-41.897,-0.789,52.912],
[-57.899,0.369,57.112],[-44.562,13.877,58.857],[-58.638,-2.538,54.667],
[-49.460,-3.596,58.194],[-45.743,-3.219,56.810],[-57.102,13.184,56.022],
[-56.157,-1.758,53.269],[-50.855,-2.879,55.721],[-49.696,14.195,53.864],
[-41.982,12.850,58.353],[-56.250,13.425,53.085],[-58.251,22.402,58.104],
[-46.902,18.689,-59.414],[-48.302,27.125,57.418],[-57.438,27.285,55.564],
[-44.460,28.777,56.998],[-49.369,23.919,55.080],[-48.675,22.728,52.460],
[-46.014,22.773,56.797],[-46.943,24.637,54.431],[-47.954,30.223,58.820],
[-44.260,17.358,55.348],[-53.396,24.815,56.653],[-41.715,29.312,52.280],
[-49.655,34.046,56.505],[-55.669,31.518,58.432],[-46.466,32.003,53.743],
[-46.969,31.988,56.779],[-44.312,24.090,54.385],[-49.023,18.513,55.063],
[-53.154,22.179,58.183],[-49.010,22.503,57.748],[-52.942,20.005,53.134],
[-57.932,20.368,54.713],[-55.137,20.311,58.254],[-54.564,19.933,55.523],
[-54.064,22.649,55.252],[-54.816,17.524,58.417],[-50.735,26.655,55.885],
[-47.531,17.747,52.722],[-46.666,20.088,54.703],[-48.518,26.977,54.093],
[-40.527,21.304,54.967],[-43.053,28.112,54.232],[-57.356,25.369,53.308],
[-52.040,26.942,58.549],[-51.314,30.880,58.478],[-45.894,27.004,52.202],
[-42.643,31.459,54.083],[-45.273,26.301,55.989],[-51.505,27.853,52.867],
[-52.911,28.882,54.734],[-54.986,25.467,54.641],[-46.810,17.686,56.597],
[-49.294,31.481,54.755],[-51.079,21.000,56.845],[-49.111,17.892,57.791],
[-53.491,32.276,55.147],[-55.317,15.947,55.090],[-55.600,29.085,55.194],
[-57.141,20.587,52.141],[-57.356,28.863,57.734],[-50.301,30.471,52.598],
[-44.459,29.813,52.351],[-54.169,27.990,52.248],[-49.595,28.139,50.955],
[-52.071,24.971,54.338],[-51.173,23.724,52.131],[-42.967,23.773,51.547],
[-51.304,33.518,53.740],[59.510,27.496,53.798],[-57.026,33.984,57.655],
[-42.539,25.744,55.792],[-41.868,26.132,58.508],[-41.426,24.565,53.503],
[-42.638,21.553,53.054],[-44.238,23.518,-59.139],[-43.863,17.210,52.544],
[-43.706,21.529,58.047],[-42.956,18.876,57.552],[-41.191,18.285,54.307],
[-58.565,16.573,56.905],[-44.348,32.886,-58.775],[-57.886,15.335,54.654],
[-53.433,15.261,57.478],[-46.371,15.267,57.636],[-56.674,32.005,55.572],
[-56.156,18.160,53.251],[-51.873,18.025,56.569],[-49.004,32.563,51.810],
[-42.185,30.448,57.217],[-55.187,31.883,52.943],[-57.983,42.584,57.574],
[-45.273,36.057,58.859],[-46.332,45.714,57.470],[-57.999,46.701,56.238],
[-43.745,46.642,59.013],[-48.806,43.528,57.521],[-47.900,41.673,52.741],
[-46.160,42.426,58.041],[-46.254,41.725,55.324],[-46.925,47.572,-59.401],
[-44.208,35.548,53.988],[-53.656,43.124,56.406],[-42.316,49.716,54.042],
[-49.079,51.293,57.651],[-55.927,51.094,58.247],[-46.939,51.489,54.124],
[-46.166,50.114,56.904],[-45.358,43.213,51.890],[-49.572,38.140,52.494],
[-53.023,40.267,58.295],[-47.906,39.847,56.790],[-53.018,37.195,51.446],
[-56.311,37.398,54.360],[-57.245,38.291,57.713],[-52.763,38.375,56.414],
[-53.056,38.722,53.726],[-54.625,37.472,58.043],[-49.114,43.836,54.358],
[-46.543,37.206,53.145],[-45.933,39.147,55.012],[-48.270,46.036,55.857],
[-39.878,42.818,54.526],[-41.921,48.117,56.849],[-56.206,43.479,52.658],
[-52.030,45.724,57.654],[-53.085,49.732,57.625],[-44.950,47.042,54.165],
[-40.820,52.021,54.057],[-46.243,44.764,54.401],[-50.720,47.727,53.587],
[-52.209,46.166,54.777],[-56.592,43.949,55.600],[-47.725,37.030,56.322],
[-49.040,48.966,55.986],[-50.401,39.071,55.317],[-48.467,35.139,58.638],
[-52.209,49.579,54.978],[-52.891,35.373,56.622],[-54.863,46.806,55.045],
[-57.188,39.977,52.273],[-56.234,48.052,58.262],[-49.129,48.788,51.086],
[-44.675,49.454,52.328],[-52.511,44.732,52.239],[-48.711,46.133,52.545],
[-51.673,43.253,54.588],[-52.348,41.269,52.606],[-42.463,42.861,52.432],
[-52.697,52.275,54.000],[-58.209,46.966,52.887],[59.222,51.945,-58.454],
[-44.040,45.220,55.985],[-41.162,43.629,58.118],[-42.708,42.716,55.762],
[-43.191,39.079,53.493],[-43.494,41.338,-59.640],[-44.157,37.446,55.715],
[-43.789,38.646,57.985],[-43.143,35.277,57.376],[-41.005,37.018,53.548],
[57.788,35.423,59.200],[-45.300,51.458,59.244],[-59.006,33.500,55.653],
[-51.472,33.575,59.234],[-44.262,32.159,57.583],[-57.244,49.604,54.805],
[-55.956,34.546,54.580],[-50.572,36.669,56.854],[-49.199,51.057,52.494],
[-43.347,49.493,58.474],[-56.607,51.915,51.447],[-46.952,56.241,59.192],
[-44.051,51.631,54.818],[-49.553,55.574,54.348],[-54.168,-59.256,58.337],
[-52.345,57.667,52.555],[-58.498,55.767,55.038],[-57.228,55.412,57.726],
[-52.746,56.506,55.377],[-53.434,-59.743,54.817],[-54.505,54.976,57.786],
[-46.700,56.376,53.660],[-47.958,57.771,55.597],[-40.432,57.712,54.802],
[-47.381,55.250,56.398],[-51.195,56.131,57.441],[-49.090,54.621,58.815],
[-52.590,52.574,57.185],[-56.069,-59.624,53.802],[-42.435,59.277,53.558],
[-44.026,56.625,52.809],[-43.192,56.073,58.601],[-43.753,54.429,54.775],
[-41.062,54.786,53.784],[-58.382,53.150,55.886],[-55.921,52.178,55.834],
[-51.399,50.884,59.421],[-45.053,54.108,57.516],[-55.934,53.741,53.400],
[-49.625,53.531,56.324],[-53.793,-57.998,-14.111],[-55.291,57.680,-18.318],
[-44.899,-59.460,-14.372],[-56.062,-49.064,-28.384],[-49.216,-46.143,-31.800],
[-54.697,-57.313,-25.860],[-53.306,-49.872,-24.644],[-43.894,-57.558,-24.550],
[-44.704,-47.472,-28.856],[-47.037,-49.195,-31.907],[-52.073,-46.840,-31.653],
[-56.320,-50.721,-33.441],[-52.879,-54.182,-34.921],[-44.190,-54.405,-20.411],
[-56.749,-52.501,-37.153],[-54.966,-55.612,-36.064],[-46.143,-48.019,-23.165],
[-53.477,-51.782,-21.766],[-51.589,-54.537,-18.346],[-49.488,-52.778,-18.959],
[-56.183,-53.005,-18.716],[-51.697,-56.638,-15.514],[-42.107,-43.678,-30.584],
[-44.887,-55.788,-30.199],[-42.224,-54.469,-14.933],[-55.091,-42.429,-29.190],
[-51.432,-32.603,-37.734],[-41.394,-35.124,-34.191],[-56.548,-27.921,-34.561],
[-42.544,-29.428,-34.609],[-53.325,-39.866,-36.467],[-59.085,-43.727,-38.681],
[-54.543,-41.163,-40.912],[-48.450,-32.516,-32.643],[-44.524,-31.210,-32.012],
[-47.992,-26.958,-31.346],[-52.220,-38.531,-24.210],[-56.829,-36.937,-46.504],
[-44.150,-38.956,-38.378],[-43.871,-45.133,-42.183],[-56.658,-30.565,-33.083],
[-56.625,-31.696,-40.257],[-50.652,-42.549,-33.324],[-44.072,-44.925,-33.962],
[-46.357,-13.437,-46.495],[-55.229,-20.565,-44.568],[-52.149,-14.239,-54.962],
[-43.782,-19.408,-37.595],[-47.709,-20.895,-38.892],[-43.374,-9.958,-43.432],
[-46.611,-27.873,-50.593],[-48.229,-12.753,-44.547],[-42.504,-25.831,-50.440],
[-55.145,-20.098,-56.122],[58.728,-8.895,-44.177],[-54.019,-23.899,-55.444],
[-49.142,-21.304,-36.225],[-47.139,-22.963,-34.247],[-42.062,-23.740,-52.021],
[-44.923,-10.417,-47.819],[-51.048,-18.065,-57.881],[-46.426,-1.065,-57.516],
[-42.591,-8.502,-50.639],[-55.634,-8.489,-53.788],[-52.201,-10.957,-50.742],
[-55.632,-2.788,-54.701],[-55.763,5.055,-51.513],[-42.373,-15.133,-56.126],
[-50.003,5.015,-52.654],[-50.323,-8.922,-46.888],[-58.561,-11.672,-59.155],
[-46.777,-5.043,-52.096],[-47.021,-9.436,-49.623],[-51.892,-5.796,-54.866],
[-42.019,-9.447,58.643],[-42.529,1.177,-51.223],[-59.275,3.973,-54.882],
[-49.787,-5.847,-50.663],[-53.279,-11.515,-59.255],[-57.058,-10.060,-51.645],
[58.536,5.929,-53.177],[-52.393,1.263,-49.188],[-51.474,-6.315,-42.214],
[-58.682,-9.091,-53.398],[-55.579,2.057,-55.356],[-51.038,6.301,-54.656],
[-55.451,5.791,-56.778],[-43.761,-11.666,-53.127],[-56.697,-15.368,-55.370],
[-58.079,8.642,58.418],[-47.685,11.654,-57.855],[-58.240,17.789,-57.306],
[-45.667,16.375,-58.809],[-53.666,12.939,-56.996],[-42.999,15.821,-58.774],
[-42.722,8.197,-53.502],[-54.741,6.899,-53.896],[-48.573,-59.449,-0.911],
[-55.895,-54.620,-0.559],[-46.697,-58.901,-7.932],[-51.932,58.991,0.351],
[-51.371,-49.558,0.784],[-46.784,-58.126,-2.537],[-51.099,-56.460,-10.818],
[-53.722,-50.712,-4.031],[-58.643,-56.516,-10.259],[-55.967,-37.773,-17.664],
[-51.619,-52.832,-16.271],[-48.805,-46.782,-11.187],[-49.739,-41.041,-3.801],
[-59.348,-47.853,-12.732],[-54.084,-41.033,-15.683],[-54.689,-52.375,-14.558],
[58.798,-48.954,-17.503],[-45.354,-38.452,-7.148],[-58.258,-37.700,-3.843],
[-53.766,-47.376,-17.272],[-54.654,-40.896,-6.390],[-48.129,-38.457,-4.291],
[-44.247,-46.820,-19.856],[-46.756,-44.355,-0.066],[-58.039,-43.382,-21.343],
[-42.996,-36.483,-14.486],[-42.160,-45.502,-12.262],[-57.568,-33.733,-6.751],
[-54.855,-32.835,-13.355],[-42.786,-28.801,-16.473],[-57.013,-30.208,-16.073],
[-55.606,-20.598,-18.121],[58.192,-28.098,-26.877],[-54.919,-28.008,-22.011],
[-51.070,-20.949,-24.646],[-53.287,-21.644,-17.244],[-53.200,-29.399,-16.539],
[-48.563,-23.606,-13.358],[-53.000,-31.638,-24.029],[-44.792,-25.303,-19.453],
[-56.300,-25.745,-21.278],[-50.310,-26.796,-27.872],[-46.857,-26.617,-18.559],
[-58.911,-29.138,-14.571],[58.626,-22.501,-21.150],[-49.660,-27.429,-14.044],
[-50.353,-33.479,-19.488],[-58.935,-24.811,-26.112],[-50.606,-25.236,-31.466],
[-40.813,-30.698,-26.781],[-41.422,-23.948,-14.832],[-59.488,-30.827,-17.466],
[-43.013,-31.327,-13.535],[-57.018,-35.584,-22.424],[-55.393,-30.551,-9.898],
[-51.471,-14.333,-34.426],[-56.685,-13.764,-34.722],[-48.108,-15.159,-25.260],
[-45.587,1.668,-28.542],[-38.669,-17.035,-35.456],[-47.100,-11.645,-39.120],
[-57.317,0.575,-28.257],[-53.715,-13.056,-38.887],[-50.883,-11.248,-20.889],
[59.515,-10.911,-38.894],[-58.903,-16.951,-26.425],[-58.922,-20.046,-30.585],
[-46.446,7.753,-41.740],[-59.357,3.627,-40.905],[-50.155,7.046,-38.623],
[-54.306,8.328,-36.989],[-55.831,2.332,-38.486],[-46.744,3.255,-43.551],
[-52.709,5.745,-42.517],[-56.824,7.873,-34.565],[-55.910,14.431,-36.497],
[-44.115,6.545,-49.193],[-47.038,8.696,-50.203],[-53.329,-2.748,-45.417],
[-58.011,-2.098,-42.426],[-49.965,2.611,-38.884],[-50.304,10.766,-35.723],
[-58.075,11.897,-40.168],[-48.545,-1.791,-40.603],[-51.049,5.522,-50.227],
[-58.239,8.862,-31.605],[-48.717,10.662,-33.173],[-42.709,6.918,-32.157],
[-44.743,-2.048,-40.151],[-59.135,14.538,-41.642],[-51.601,23.332,-48.983],
[-47.129,17.083,-49.642],[-57.030,18.914,-46.899],[-42.441,17.190,-43.316],
[-44.030,16.205,-41.514],[-47.636,26.637,-49.518],[-46.574,24.347,59.015],
[-44.699,21.359,-41.682],[-53.170,16.554,-43.794],[-56.433,19.916,-44.193],
[-55.640,15.600,-54.372],[-48.888,21.633,-57.314],[-43.575,23.249,-56.475],
[-54.586,27.832,-54.220],[-58.982,25.121,-54.020],[-58.927,32.367,-44.138],
[-52.103,31.683,-44.009],[-52.187,15.569,-54.823],[-48.513,20.627,-38.165],
[-41.312,28.945,-50.755],[-41.507,20.568,-51.442],[-42.787,11.002,-52.101],
[-57.999,29.331,-43.088],[-50.589,32.797,-50.942],[-56.983,26.345,-51.786],
[-50.750,24.591,57.656],[-55.480,18.275,-37.036],[-58.469,35.293,-59.146],
[-53.955,33.352,-56.171],[-55.224,44.865,-54.565],[-44.579,26.506,58.626],
[-56.827,42.316,-56.225],[-47.750,31.158,-55.245],[-57.292,33.784,-48.067],
[-48.978,-50.172,11.506],[-42.617,-49.238,4.546],[-50.089,-43.839,5.688],
[-56.912,-42.845,15.551],[-42.766,-46.215,4.540],[-50.053,-56.405,12.135],
[-54.701,-59.359,11.555],[-54.101,-52.326,4.942],[-49.359,-51.704,4.560],
[-49.126,-47.309,22.891],[-43.496,-48.224,18.024],[-45.425,-58.124,12.206],
[-42.042,-51.911,24.310],[-48.448,-38.813,14.011],[-47.348,-38.277,10.708],
[-58.345,-42.014,9.216],[-48.183,-58.612,15.016],[-49.751,-46.500,4.932],
[59.336,-29.718,3.848],[58.280,-33.402,-2.879],[-54.755,-36.087,-1.972],
[-50.146,-40.099,12.344],[-55.348,-40.362,-0.430],[-50.916,-34.275,-1.904],
[-52.249,-30.989,10.388],[-51.959,-33.251,1.057],[-56.850,-26.430,4.569],
[-48.368,-42.000,3.931],[-53.758,-38.629,12.051],[-42.600,-25.184,6.187],
[-42.042,-30.127,4.599],[-44.014,-44.387,0.251],[-44.641,-40.405,4.098],
[-58.475,-20.735,-3.550],[-52.125,-19.066,-5.374],[-56.328,-20.889,1.203],
[-55.593,-12.992,-4.580],[-46.417,-20.436,-9.869],[-57.511,-14.743,0.138],
[-41.051,-17.102,-13.546],[-51.548,-10.933,-1.631],[-49.392,-13.278,-15.113],
[-54.338,-24.081,-0.384],[-49.869,-20.429,-2.928],[-56.029,-25.831,-7.215],
[-53.784,-21.970,-5.848],[-51.025,-13.669,-2.240],[-49.617,-17.063,-11.156],
[-58.346,-20.716,-11.902],[59.670,-5.358,-1.936],[-55.150,-5.212,0.634],
[58.775,-11.650,1.430],[-52.399,-11.991,1.067],[-48.885,-17.312,4.255],
[-42.091,-22.218,-9.663],[-44.498,-18.772,-6.211],[-47.418,-9.372,-6.881],
[-58.759,-27.743,-6.560],[-46.263,-25.182,-1.849],[-50.667,-15.571,-14.292],
[-43.954,-11.765,-12.188],[-47.189,-21.605,8.251],[-51.451,-9.809,-10.392],
[-49.780,3.543,-19.739],[-42.170,-11.376,-14.198],[-54.210,-5.125,-18.198],
[-53.185,-13.099,-19.590],[-47.414,-8.973,-17.953],[-57.533,0.560,-15.071],
[-57.902,-4.855,-20.040],[-50.740,-6.477,-14.749],[-54.442,-9.353,-16.625],
[-49.971,-0.483,-26.386],[-47.900,-2.723,-23.812],[-52.014,-3.752,-7.885],
[-49.360,-1.955,-6.103],[-59.007,-1.568,-24.762],[-42.967,6.275,-13.959],
[-42.334,1.633,-13.101],[-47.634,9.778,-13.015],[-57.714,6.511,-15.504],
[-57.304,6.093,-19.062],[-50.185,-6.894,-11.579],[-56.944,-8.686,-11.948],
[-48.380,15.531,-22.396],[-44.108,15.702,-25.273],[-57.949,12.492,-22.854],
[-54.687,19.114,-20.134],[-58.155,14.685,-30.029],[-54.757,15.607,-25.661],
[-49.683,10.269,-14.979],[-46.926,6.159,-28.460],[-51.491,18.117,-16.479],
[59.070,25.812,-20.988],[-53.371,15.199,-17.452],[-42.697,9.526,-33.225],
[-42.689,10.671,-11.200],[-43.368,12.758,-13.541],[-41.613,9.596,-13.886],
[-48.308,6.233,-19.874],[-55.045,28.993,-38.301],[-51.594,22.176,-36.533],
[-54.193,27.453,-34.868],[-49.018,29.242,-31.097],[-51.517,26.270,-30.144],
[-56.833,26.876,-34.081],[-54.947,23.285,-31.284],[-50.897,29.178,-29.233],
[-50.837,36.175,-29.669],[-42.697,32.994,-40.732],[-52.113,33.652,-36.974],
[-49.696,33.148,-41.208],[-44.420,29.450,-25.155],[58.617,35.219,-25.894],
[-52.823,26.529,-42.198],[-55.992,32.285,-23.730],[58.974,30.022,-39.433],
[-41.726,40.098,-32.395],[-40.465,26.756,-38.458],[-43.163,22.916,-39.144],
[-57.921,34.939,-29.575],[-42.389,22.932,-30.481],[58.466,35.195,-40.085],
[-56.213,21.919,-37.014],[-58.905,46.434,-44.820],[-53.414,42.567,-40.690],
[-45.605,48.567,-38.944],[-51.783,52.000,-40.589],[-39.956,48.410,-51.626],
[-55.383,40.000,-37.045],[-50.940,41.781,-34.951],[-48.060,41.523,-45.045],
[-57.580,39.418,-46.863],[-48.562,52.327,-39.390],[-42.613,49.519,-53.166],
[-56.534,45.363,-46.955],[-43.547,43.567,-35.929],[-45.178,44.431,-41.762],
[-45.040,55.277,-43.993],[-50.896,44.726,-51.436],[-56.180,47.196,-34.716],
[-42.870,43.356,-50.217],[-48.345,45.619,-29.916],[-57.576,46.541,-31.181],
[-59.072,46.679,-51.276],[-42.180,49.233,-34.672],[-41.653,44.592,-33.668],
[-50.129,54.132,-44.833],[-59.196,39.421,-42.121],[-54.255,-41.276,29.434],
[-59.087,-38.660,23.172],[-49.066,-44.485,23.030],[-47.282,-39.790,35.825],
[-50.409,-45.768,27.858],[-50.798,-38.669,32.837],[-48.236,-41.011,21.765],
[-53.024,-40.533,19.743],[-52.297,-47.886,26.218],[-44.870,-43.962,36.374],
[59.691,-47.487,24.985],[-42.082,-45.936,21.118],[-53.774,-40.834,39.131],
[-43.929,-34.167,35.767],[-41.600,-44.693,23.524],[-45.233,-48.093,27.842],
[-44.581,-41.813,34.476],[-42.313,-42.353,40.774],[-56.540,-20.159,18.914],
[-47.344,-27.852,23.611],[-45.290,-28.826,26.738],[-54.778,-24.479,17.630],
[-42.434,-16.765,23.203],[-52.553,-17.576,11.049],[-46.312,-29.200,13.883],
[-55.295,-29.416,19.829],[-47.766,-30.275,25.355],[-43.494,-13.236,23.434],
[-50.154,-27.329,22.940],[-49.027,-30.507,16.384],[-52.001,-32.680,20.303],
[-55.243,-24.023,11.971],[-57.682,-22.851,14.910],[58.764,-17.476,24.739],
[-49.008,-24.671,30.009],[-58.602,-25.775,31.249],[-59.001,-26.963,12.169],
[-59.533,-29.045,24.389],[-57.180,-16.917,18.152],[-50.752,-15.788,18.235],
[-58.634,-21.241,12.311],[-41.294,-10.764,16.190],[-47.716,-6.342,7.155],
[-50.399,-13.466,8.842],[-53.981,-2.832,4.222],[-57.828,-9.888,4.548],
[-46.998,-7.210,-0.427],[-45.046,-13.545,3.985],[-54.282,-7.140,19.296],
[-51.569,-6.845,19.896],[-58.302,-7.786,2.904],[-44.016,-1.841,17.411],
[-43.416,-12.700,7.336],[-41.238,-11.571,23.472],[-42.848,-9.679,20.366],
[-58.518,-0.549,14.596],[-53.724,-1.733,11.488],[-56.866,-11.414,23.273],
[-56.865,6.501,2.730],[-50.802,10.787,2.656],[-44.904,2.444,7.040],
[-42.294,10.396,5.473],[-49.937,12.805,4.741],[-44.937,12.388,-6.382],
[-50.819,8.958,-10.038],[-58.903,15.047,7.581],[-54.110,4.890,-4.403],
[-59.132,12.705,-7.718],[-48.213,24.175,-6.742],[-48.648,29.655,-11.360],
[-58.452,29.577,-13.281],[-43.281,35.744,-5.759],[-44.216,25.948,-4.807],
[-48.173,15.504,-11.239],[-49.966,21.993,-15.290],[-51.619,31.984,-6.733],
[59.364,22.453,-0.008],[-59.021,33.539,-7.152],[-51.319,22.101,-18.299],
[-56.809,25.149,-16.858],[-58.288,26.701,-12.339],[-58.063,35.557,-4.105],
[-50.591,23.955,2.293],[-44.568,31.030,0.208],[-43.071,19.761,4.529],
[-44.145,17.192,-0.867],[-59.122,14.451,-11.293],[-58.051,27.501,-15.579],
[-49.673,41.812,-20.118],[-50.782,42.268,-16.265],[-46.357,33.852,-16.564],
[-58.287,38.324,-22.670],[-41.418,41.205,-14.253],[-40.682,47.379,-13.550],
[-51.474,47.164,-13.400],[-57.472,45.307,-14.763],[-54.001,34.800,-18.708],
[-53.092,38.050,-11.966],[-48.463,34.874,-18.878],[-55.733,39.648,-16.842],
[-48.994,47.333,-14.835],[59.112,35.574,-21.225],[-51.989,32.844,-18.053],
[-54.466,31.779,-18.905],[-57.306,39.436,-24.868],[-48.277,42.895,-8.235],
[-56.658,39.053,-5.812],[-52.608,37.237,-6.302],[-59.659,34.230,-13.766],
[-44.852,46.019,-23.140],[-49.824,49.252,-12.744],[-55.454,48.969,-18.114],
[-55.779,30.586,-21.121],[-43.595,48.082,-14.133],[-46.010,-58.467,-26.618],
[-44.629,59.470,-33.918],[-46.012,54.345,-26.431],[-55.161,52.962,-19.732],
[-47.546,56.962,-17.077],[-56.611,54.335,-17.377],[-56.702,58.736,-36.538],
[-58.442,-34.553,47.227],[-50.187,-26.741,41.520],[-53.898,-21.445,49.470],
[-48.185,-27.031,38.306],[-51.223,-30.386,52.709],[-56.826,-36.559,46.091],
[-42.180,-36.461,41.621],[-57.497,-27.809,40.295],[-51.114,-30.726,34.078],
[-57.746,-25.417,54.436],[-51.068,-34.153,56.485],[-57.707,-33.509,37.137],
[-42.532,-32.626,42.494],[-41.786,-35.041,38.449],[-44.084,-36.168,55.056],
[-58.329,-20.211,44.696],[-45.549,-21.260,43.809],[-49.066,-36.497,48.385],
[-44.771,-37.582,49.579],[-47.453,-38.554,49.121],[-55.301,-10.236,38.805],
[-56.814,-16.086,30.160],[-46.487,-10.431,31.633],[-41.577,-7.620,41.649],
[-50.690,-6.286,39.093],[-50.612,-20.386,45.608],[-57.323,-9.689,40.553],
[-53.777,-22.651,35.995],[-42.235,-10.033,40.664],[-56.059,-13.613,42.948],
[-58.427,-16.502,28.030],[-43.383,-13.115,42.226],[-43.816,-19.507,36.696],
[-46.961,-22.085,39.374],[-59.226,-0.171,27.321],[-55.308,2.580,28.338],
[-53.741,5.297,32.335],[-49.601,5.913,25.032],[-56.723,6.552,21.796],
[-51.187,-8.619,23.653],[-55.667,-1.372,30.049],[-43.680,-3.874,22.372],
[-57.005,13.005,31.517],[-51.602,3.167,16.701],[-50.352,-3.863,37.893],
[-51.904,-0.823,39.874],[-42.258,1.005,20.971],[-44.085,4.287,32.894],
[-41.825,-8.659,23.080],[-47.432,11.457,28.543],[-58.405,8.467,22.771],
[-56.530,-8.734,23.711],[-43.145,4.938,18.596],[-46.781,-2.589,40.266],
[-43.778,11.802,32.128],[-49.300,14.490,16.483],[-43.652,12.660,23.680],
[-55.371,11.976,20.282],[-46.193,10.910,18.361],[-59.162,15.962,12.757],
[-51.566,23.942,20.399],[-54.288,9.223,14.517],[-51.439,22.029,22.232],
[-52.513,20.207,24.088],[-47.595,16.511,10.848],[-45.023,16.855,12.377],
[-49.794,18.535,7.737],[-53.570,14.539,12.568],[-45.400,5.849,13.833],
[-50.890,24.940,15.476],[-48.558,24.868,17.203],[-58.102,20.435,10.653],
[-53.995,32.326,6.300],[-44.936,33.805,16.737],[-50.577,29.283,7.495],
[-54.596,27.626,3.604],[-54.966,25.142,8.026],[-52.745,30.979,13.963],
[-47.645,27.748,6.670],[-41.552,26.496,4.507],[-46.193,36.589,11.620],
[-59.013,31.933,16.009],[-53.321,33.251,1.579],[-49.449,34.829,15.560],
[-57.497,25.617,2.739],[58.941,38.983,15.839],[-55.054,32.013,0.184],
[-58.379,37.028,15.046],[-54.187,32.342,17.057],[-41.549,28.976,2.236],
[-43.689,27.395,2.562],[-40.487,29.086,20.278],[-59.620,27.041,11.778],
[-41.588,22.873,6.291],[-52.394,40.129,5.008],[-41.728,45.101,4.866],
[-40.652,47.942,4.662],[-53.518,52.797,1.551],[-55.641,45.126,2.723],
[-44.979,53.749,-1.346],[-46.354,44.507,-5.960],[-42.144,56.673,2.273],
[-45.904,52.565,-10.834],[-50.444,48.128,2.833],[-54.858,53.345,-6.686],
[-49.498,48.092,5.859],[-46.245,48.131,-0.155],[-49.418,46.469,-9.337],
[-56.124,52.209,4.957],[-53.350,44.113,-6.957],[-53.930,48.718,8.344],
[-44.150,41.926,-1.402],[-58.461,45.907,3.257],[-48.989,57.211,-3.107],
[-50.752,42.378,1.354],[-50.379,-16.599,54.385],[-45.410,-26.854,55.144],
[-48.865,-31.873,56.858],[-58.129,-30.050,58.590],[-54.863,-28.282,57.631],
[-41.556,-28.053,58.044],[-44.806,-21.321,53.163],[-58.107,-22.776,54.040],
[-57.555,-32.876,58.166],[-58.682,-5.784,52.484],[-42.203,-5.549,58.827],
[-55.632,-10.327,59.014],[-54.832,-3.422,55.872],[-50.627,-14.605,57.543],
[-55.046,-7.498,55.802],[-56.537,-2.390,58.654],[-55.368,1.993,55.824],
[-52.700,1.393,56.003],[-49.517,-1.169,44.547],[-53.587,-7.371,57.910],
[-53.332,-12.778,51.202],[-50.613,-7.900,56.962],[-48.273,-13.402,48.180],
[51.619,-5.333,45.256],[-58.211,-13.273,56.556],[-57.891,3.765,54.874],
[-50.466,-6.162,42.134],[-51.431,-0.284,-58.590],[-49.645,-9.032,41.827],
[-52.978,-2.208,57.676],[-50.279,2.368,55.526],[-59.063,-2.114,45.260],
[-55.894,-15.162,48.436],[-43.832,4.047,56.538],[-49.764,11.336,45.240],
[-55.868,6.110,46.298],[-52.943,14.364,42.369],[-57.463,8.627,40.284],
[-50.049,16.798,45.363],[-58.018,5.697,37.056],[-52.111,18.217,43.562],
[-53.408,13.902,36.974],[-58.859,16.584,49.580],[-49.941,16.204,50.938],
[-52.717,6.361,49.714],[-41.479,8.762,42.391],[-43.100,-0.104,40.631],
[-56.722,0.629,45.925],[-59.495,14.209,35.535],[-56.189,23.018,37.637],
[-53.971,31.833,36.707],[-44.129,30.456,36.957],[-56.468,34.878,37.164],
[-51.337,18.031,31.859],[-49.683,29.370,42.412],[-50.892,23.753,38.901],
[-57.484,23.288,28.390],[-45.127,35.465,34.816],[-40.451,28.565,23.760],
[-57.520,33.005,34.939],[-51.063,25.070,22.794],[-49.086,21.811,40.667],
[-45.659,24.754,42.063],[-49.977,28.703,22.607],[-49.747,29.639,38.206],
[-44.406,33.167,35.536],[-42.968,29.654,41.562],[-43.777,23.643,37.471],
[-42.464,24.926,42.467],[-55.351,31.225,30.209],[-40.806,18.314,37.323],
[-57.322,17.769,32.081],[-50.274,32.012,22.144],[-57.531,43.533,20.456],
[-47.267,47.488,23.813],[-55.494,47.926,18.764],[-52.582,36.210,31.378],
[-50.208,38.748,22.307],[-54.740,34.216,20.844],[-52.880,41.173,19.768],
[-55.202,44.725,21.324],[-59.147,42.566,18.187],[-54.908,53.983,30.312],
[-58.999,49.155,30.945],[-54.860,42.091,32.393],[-42.205,50.295,23.877],
[-44.252,49.004,18.825],[-56.266,52.259,23.066],[-49.409,49.643,23.033],
[-54.700,40.043,36.828],[-42.443,58.109,22.140],[-58.595,58.323,10.329],
[-45.360,53.314,21.299],[-55.367,51.475,20.520],[-49.204,52.977,12.777],
[-50.959,57.720,5.860],[-58.270,51.541,20.419],[-47.186,5.679,-59.278],
[-58.554,5.474,58.103],[-56.062,24.703,57.543],[-50.490,21.411,53.903],
[-47.232,15.816,54.514],[-56.588,18.390,56.452],[-47.471,21.721,50.147],
[-53.944,24.161,52.518],[-53.861,12.114,57.128],[-53.762,29.720,-59.708],
[-49.532,29.382,56.733],[-54.692,28.118,57.397],[-56.633,39.759,49.604],
[-57.930,36.204,51.650],[-50.068,30.996,44.565],[-53.330,35.872,46.678],
[-53.574,35.249,53.185],[-57.350,40.559,55.863],[-46.227,44.929,49.708],
[-52.453,40.911,55.720],[-58.729,35.622,41.528],[-44.255,33.014,52.358],
[-44.471,28.530,49.454],[-58.446,30.981,53.778],[-49.299,57.936,36.203],
[-56.234,47.949,42.484],[-45.306,45.712,46.729],[-46.687,53.823,45.354],
[-54.743,51.694,42.115],[-44.470,52.361,48.215],[-53.569,46.113,37.760],
[-52.678,52.487,37.603],[-51.364,43.245,41.918],[-53.179,56.015,32.019],
[-54.966,43.290,45.692],[-57.666,55.438,34.830],[-59.028,58.371,47.294],
[-52.708,51.253,48.818],[-47.033,53.176,49.837],[-44.750,58.388,42.838],
[-43.165,44.671,34.581],[-44.511,44.050,39.131],[-43.135,47.910,48.186],
[-55.816,57.153,37.301],[-42.727,43.310,41.942],[58.903,54.409,32.634],
[-58.115,41.751,35.688],[-49.474,54.519,30.140],[-56.502,49.131,51.815],
[-47.316,48.857,53.762],[-28.533,-50.031,-53.804],[-29.500,-51.331,-49.741],
[-37.148,-49.740,-47.993],[-27.337,-47.244,-51.533],[-35.452,-45.937,-47.256],
[-39.243,-53.152,-54.430],[-39.002,-49.942,-51.601],[-29.086,-41.540,-50.249],
[-30.698,-46.682,-49.166],[-23.098,-55.845,-48.119],[-28.374,-53.919,-48.544],
[-37.085,-57.497,-50.564],[-27.310,-55.063,-52.588],[-38.951,-52.548,-50.773],
[-31.733,-50.942,-51.506],[-22.242,-53.696,-44.006],[-37.690,-55.934,-47.552],
[-30.008,-47.611,-52.826],[-34.774,-48.670,-47.564],[-25.147,-55.047,-50.989],
[-37.106,-52.501,-47.816],[-34.705,-49.135,-44.726],[-39.940,-46.851,-55.333],
[-26.280,-46.483,-49.187],[-27.376,-57.322,-47.464],[-33.385,-59.467,-50.661],
[-26.839,-44.198,-51.554],[-26.691,-45.719,-53.975],[-25.748,-47.768,-42.945],
[-31.458,-49.827,-53.867],[-29.910,-51.750,-59.072],[-23.026,-51.683,-45.831],
[-25.538,-49.387,-54.073],[-27.805,-51.966,-45.242],[-36.145,-49.473,-52.430],
[-34.389,-46.403,-49.616],[-34.860,-49.622,-49.978],[-29.053,-45.435,-47.902],
[-33.389,-46.581,-43.632],[-26.513,-50.764,-56.396],[-36.567,-51.196,-44.065],
[-29.606,-44.505,-53.415],[-35.105,-44.719,-42.267],[-23.618,-57.878,-56.211],
[-36.263,-51.310,-55.020],[-22.310,-44.472,-59.278],[-40.165,-50.170,-46.539],
[-39.572,-45.517,-48.338],[-22.311,-44.976,-44.981],[-31.801,-45.145,-45.785],
[-30.111,-41.446,-54.739],[-36.062,-41.224,-53.068],[-29.281,-43.623,-57.664],
[-27.255,-43.495,-55.360],[-24.305,-53.353,-42.292],[-38.481,-47.128,-43.743],
[-35.118,-56.947,-48.276],[-31.318,-57.722,-47.231],[-27.376,-53.001,-42.483],
[-24.828,-47.374,-56.636],[-31.063,-54.920,-59.507],[-35.098,-54.332,-48.380],
[-34.621,-53.441,-53.231],[-30.099,-51.979,-54.836],[-29.665,-51.621,-43.159],
[-27.373,-51.947,-52.396],[-35.998,-54.219,-43.566],[-34.573,-55.353,-59.315],
[-40.302,-55.404,-52.302],[-39.203,-54.865,-56.443],[-30.384,-52.294,-46.743],
[-31.575,-54.014,-52.523],[-37.079,-54.013,-52.382],[-36.660,-54.923,-55.144],
[-27.579,-56.265,-50.069],[-33.164,-51.904,-56.672],[-24.094,-40.013,-42.000],
[-36.539,-48.958,-42.552],[-32.559,-53.887,-50.098],[-35.780,-56.942,-53.207],
[-32.983,-53.198,-46.759],[-31.965,-48.547,-57.862],[-25.973,-47.845,-46.849],
[-28.583,-54.010,-57.235],[-31.029,-47.099,-41.656],[-26.671,-53.481,-54.728],
[-37.502,-47.765,-50.200],[-30.484,-47.452,-55.647],[-39.537,-51.142,-43.788],
[-34.087,-53.270,-41.357],[-30.059,-46.903,-44.387],[-22.446,-52.571,-54.428],
[-23.027,-47.503,-43.553],[-24.725,-44.247,-56.407],[-26.523,-49.725,-49.084],
[-34.599,-45.691,-54.282],[-39.495,-49.347,-59.192],[-33.164,-46.815,-52.165],
[-32.571,-42.958,-47.137],[-32.447,-42.600,-54.552],[-28.584,-48.102,-47.516],
[-26.846,-47.636,-58.419],[-26.904,-42.831,-45.584],[-23.116,-41.826,-57.530],
[-39.821,-54.552,-45.419],[-27.496,-48.184,-55.725],[-36.632,-45.039,-44.705],
[-33.309,-45.452,57.718],[-34.089,-46.646,-56.932],[-37.352,-47.918,-55.809],
[-29.507,-55.442,-46.026],[-28.330,-56.406,-55.511],[-31.533,-45.043,-56.156],
[-31.783,-54.077,-55.198],[-31.860,-52.558,-44.346],[-30.568,-56.957,-53.920],
[-34.229,-43.612,-57.715],[-33.863,-57.456,-57.692],[-26.175,-55.479,-43.616],
[-25.906,-52.177,-48.947],[-36.361,-45.161,-56.844],[-38.119,-52.908,-58.329],
[-25.526,-42.938,-48.556],[-36.110,-59.333,-48.226],[-39.438,-44.725,-53.738],
[-38.634,-42.688,-42.199],[-33.397,-42.838,-40.970],[-31.799,-43.625,-59.045],
[-28.098,-46.591,-41.961],[-26.583,-44.791,-59.009],[-40.618,-48.905,-42.465],
[-35.827,-46.228,-59.184],[-31.844,-49.768,-42.883],[-30.805,-47.273,58.636],
[-27.995,-48.729,-44.731],[-38.655,-50.232,-41.614],[-34.786,-49.262,-57.630],
[-34.107,-51.492,-43.440],[-32.515,-52.603,58.925],[-29.191,-49.040,-42.075],
[-23.919,-49.226,-58.806],[-32.033,-56.447,-44.764],[-34.318,-55.614,-46.129],
[-32.592,-50.318,-47.588],[-27.111,-52.941,-39.850],[-27.047,-50.394,-42.788],
[-36.478,-51.706,-41.275],[-34.575,-58.708,-45.298],[-31.114,-55.838,-41.790],
[-29.049,-54.704,-43.577],[-28.146,-56.047,-41.534],[-38.256,-56.228,-42.753],
[-34.093,-58.661,-42.251],[-31.319,-58.484,-39.619],[-32.804,-41.005,-56.726],
[-39.094,-45.855,-57.702],[-40.458,-42.374,-51.803],[-37.761,-42.758,-51.111],
[-23.798,-43.451,-50.871],[-21.701,-42.326,-45.469],[-23.809,-45.442,-42.051],
[-23.310,-46.199,-54.715],[-23.195,-48.433,-52.802],[-23.181,-47.703,-47.317],
[-24.893,-45.126,-47.233],[-23.405,-50.386,-56.093],[-24.455,-54.196,-57.173],
[-24.655,-50.238,-51.507],[-40.067,-49.083,-49.433],[-24.364,-50.551,-42.847],
[-22.247,-51.701,-41.253],[-26.795,-57.109,-57.711],[-24.102,-54.614,-53.437],
[-25.465,-54.784,-48.177],[-24.821,-53.860,-45.366],[-38.668,-53.425,-43.016],
[-20.084,-59.033,-42.140],[-24.639,-57.231,-53.674],[-38.487,-54.974,-49.995],
[-39.844,-53.100,-47.983],[-23.271,-58.268,-41.799],[-21.799,-56.498,-57.989],
[-40.438,-57.208,-54.888],[-25.870,-41.634,-53.226],[-38.533,-57.850,-45.904],
[-37.786,-46.421,-40.807],[-33.848,-43.705,-44.721],[-29.763,-42.114,-43.365],
[-27.321,-40.556,-42.930],[-24.319,-58.180,-44.313],[-38.281,-42.682,-45.398],
[-34.570,-40.755,-45.965],[-31.683,-41.616,-49.304],[-28.260,-42.764,-48.013],
[-26.867,-40.511,-47.622],[-35.749,-42.556,-48.350],[-35.200,-44.105,-52.156],
[-33.489,-43.808,-49.791],[-30.694,-57.032,-50.804],[-25.834,-58.923,-51.974],
[-23.656,-57.805,-50.470],[-38.164,-58.017,-53.400],[-37.117,58.697,-55.447],
[-33.136,-58.056,-54.591],[-28.392,-58.662,-52.443],[-26.508,-58.262,-55.151],
[-38.574,-43.206,-56.542],[-37.749,-57.507,-57.872],[-31.540,-56.426,-56.944],
[-30.486,-40.391,-58.174],[-22.977,-42.610,-54.595],[-36.863,-58.351,-43.800],
[-36.926,-59.198,-40.933],[-35.382,-41.749,-55.902],[-26.285,-59.822,-42.521],
[-24.416,-42.842,-42.245],[-28.980,-30.283,-52.028],[-30.095,-31.761,-49.313],
[-36.628,-32.749,-48.330],[-25.751,-28.556,-49.979],[-36.960,-28.949,-48.793],
[-42.391,-33.786,-53.636],[-40.709,-31.324,-51.559],[-28.798,-23.535,-51.238],
[-31.195,-29.358,-50.100],[-22.088,-39.108,-49.709],[-27.370,-35.439,-47.580],
[-35.267,-36.820,-51.049],[-27.139,-37.211,-53.394],[-40.328,-35.103,-51.408],
[-31.390,-32.567,-51.707],[-22.655,-34.643,-45.719],[-38.073,-38.386,-47.850],
[-30.885,-28.199,-53.339],[-32.512,-28.908,-46.149],[-24.946,-36.219,-51.395],
[-38.050,-36.254,-49.773],[-36.381,-31.150,-46.212],[-40.625,-26.809,-54.994],
[-27.238,-27.240,-47.967],[-25.651,-38.017,-48.107],[-32.749,-39.081,-49.571],
[-26.975,-26.095,-50.862],[-27.022,-26.940,-53.679],[-25.649,-29.006,-44.464],
[-31.270,-31.464,-54.525],[-29.119,-32.704,-59.142],[-22.134,-32.032,-46.932],
[-27.587,-32.082,-53.855],[-28.471,-33.398,-44.954],[-35.056,-31.665,-52.324],
[-34.164,-27.715,-49.240],[-33.842,-31.489,-49.153],[-29.592,-26.138,-46.959],
[-34.869,-29.744,-44.008],[-29.062,-31.107,-56.379],[-36.000,-32.522,-43.822],
[-31.124,-25.361,-51.886],[-35.945,-27.099,-43.874],[-24.406,-39.325,-56.721],
[-35.323,-31.679,-54.908],[-24.800,-24.774,-58.785],[-38.702,-29.687,-45.850],
[-40.608,-27.524,-48.428],[-21.680,-24.834,-44.680],[-31.443,-26.757,-45.246],
[-30.135,-23.164,-53.715],[-36.248,-23.960,-53.335],[-28.671,-23.774,-58.302],
[-29.129,-25.254,-56.346],[-25.132,-33.622,-42.395],[-39.065,-28.172,-43.215],
[-35.397,-38.870,-47.670],[-31.112,-40.554,-45.181],[-27.150,-37.140,-45.369],
[-26.867,-30.889,-59.081],[-30.966,-36.533,-56.124],[-34.611,-34.845,-48.955],
[-35.071,-34.556,-53.106],[-29.888,-33.999,-56.303],[-29.442,-34.667,-41.740],
[-27.702,-34.488,-52.098],[-37.500,-36.099,-46.788],[-32.731,-37.702,-59.101],
[-41.101,-36.249,-53.947],[-39.692,-35.913,-57.298],[-29.682,-33.647,-47.320],
[-32.050,-36.701,-50.879],[-37.918,-35.331,-53.731],[-33.853,-34.927,-55.818],
[-28.305,-37.813,-48.301],[-34.371,-34.591,-59.415],[-25.041,-22.378,-42.399],
[-37.177,-30.046,-42.444],[-30.224,-34.959,-50.122],[-36.318,-38.367,-53.119],
[-31.726,-36.081,-47.545],[-33.705,-29.474,-56.885],[-23.997,-31.457,-49.702],
[-27.350,-36.435,-56.984],[-29.763,-28.203,-43.466],[-26.155,-34.454,-54.968],
[-37.946,-31.199,-51.808],[-31.213,-29.194,-56.051],[-38.640,-32.287,-44.622],
[-31.787,-31.106,-44.584],[-28.315,-25.440,-44.914],[-22.164,-34.730,-56.642],
[-23.112,-28.484,-43.311],[-23.965,-26.862,-54.665],[-27.557,-30.634,-49.380],
[-35.730,-26.860,-53.815],[-39.924,-30.430,-57.084],[-33.615,-28.660,-52.513],
[-33.958,-25.043,-45.265],[-33.911,-24.869,-52.568],[-27.845,-29.045,-46.127],
[-25.428,-27.196,-57.347],[-26.013,-25.971,-42.260],[-23.072,-23.698,-56.970],
[-39.661,-35.944,-44.569],[-26.425,-29.759,-55.207],[-37.926,-25.118,-44.284],
[-32.112,-26.719,-56.906],[-34.537,-27.367,-58.189],[-38.611,-28.297,-56.660],
[-30.575,-38.061,-46.223],[-29.364,-38.648,-56.252],[-32.305,-24.488,-54.730],
[-32.369,-35.994,-53.601],[-31.406,-34.105,-43.839],[-30.581,-37.946,-53.710],
[-33.937,-24.843,-56.910],[-34.722,-38.848,-57.260],[-24.934,-38.029,-44.115],
[-25.975,-33.606,-49.951],[-38.166,-25.562,-59.186],[-36.465,-32.276,-58.162],
[-25.243,-24.804,-47.700],[-36.987,-39.963,-50.182],[-38.590,-25.505,-52.926],
[-40.371,-23.536,-42.284],[-35.519,-24.604,-43.019],[-32.259,-25.114,-59.299],
[-28.657,-26.029,-41.113],[-27.151,-26.153,-59.532],[-40.720,-31.703,-42.807],
[-37.077,-30.341,58.294],[-32.356,-29.027,-43.004],[-31.567,-28.638,-58.860],
[-26.418,-31.320,-45.711],[-38.031,-33.138,-42.264],[-33.621,-32.386,-56.806],
[-33.968,-33.553,-45.233],[-32.438,-31.664,-59.072],[-30.362,-31.213,-42.108],
[-24.359,-29.802,-59.354],[-34.294,-36.058,-44.003],[-34.833,-35.983,-46.543],
[-31.966,-32.119,-47.147],[-26.904,-35.496,-40.946],[-27.371,-31.961,-42.911],
[-35.044,-34.323,-41.782],[-32.974,-38.485,-44.312],[-32.298,-35.481,-41.771],
[-29.346,-37.406,-43.916],[-28.083,-37.740,-41.479],[-37.152,-37.032,-44.054],
[-34.045,-40.509,-42.199],[-31.789,-40.582,-40.851],[-33.978,-22.820,-58.972],
[-40.716,-27.631,-58.340],[-39.927,-22.666,-53.468],[-38.830,-23.389,-50.831],
[-23.366,-24.338,-49.471],[-24.393,-23.938,-45.087],[-25.256,-28.318,-41.360],
[-23.798,-29.541,-56.478],[-24.205,-30.697,-54.037],[-22.820,-29.033,-47.367],
[-25.008,-27.033,-45.856],[-22.116,-31.954,-56.685],[-23.725,-34.613,59.158],
[-25.175,-33.105,-52.341],[-39.338,-30.319,-49.413],[-24.691,-31.166,-43.118],
[-22.176,-32.834,-42.455],[-24.806,-35.228,-57.282],[-24.728,-36.722,-54.751],
[-23.258,-36.500,-48.045],[-24.644,-33.691,-47.464],[-42.182,-35.929,-43.619],
[-20.902,-40.150,-42.084],[-24.995,-39.220,-54.092],[-39.291,-38.246,-51.404],
[-40.940,-33.743,-49.037],[-22.286,-38.003,-42.801],[-25.145,-37.512,-58.808],
[-38.572,-38.034,-54.621],[-26.654,-23.678,-53.159],[-40.674,-41.282,-44.223],
[-37.975,-26.050,-41.409],[-34.948,-27.715,-46.250],[-31.311,-24.237,-43.253],
[-27.473,-21.733,-43.708],[-23.925,-40.316,-45.391],[-37.175,-22.490,-44.092],
[-37.121,-21.343,-46.726],[-29.852,-22.331,-48.806],[-27.710,-24.192,-48.611],
[-26.619,-21.903,-46.963],[-37.612,-25.186,-47.840],[-36.271,-25.584,-50.593],
[-32.942,-25.458,-49.568],[-29.766,-38.891,-50.580],[-25.918,-40.497,-50.074],
[-23.558,-40.117,-51.733],[-38.819,-40.529,-53.166],[-39.410,-40.352,-56.199],
[-33.122,-40.626,-52.222],[-27.347,-39.729,-52.221],[-27.353,-40.298,-55.197],
[-38.897,-24.094,-57.111],[-37.074,-39.988,-57.821],[-32.718,-38.566,-55.205],
[-31.219,-22.327,-58.411],[-25.887,-24.005,-55.977],[-36.609,-39.702,-44.279],
[-34.944,-39.315,-39.999],[-37.406,-21.080,-57.342],[-27.265,-41.688,-40.520],
[-23.086,-25.728,-42.272],[-30.047,-13.239,-51.899],[-28.723,-12.599,-49.667],
[-37.747,-13.900,-47.339],[-24.355,-10.695,-51.689],[-37.985,-9.552,-49.482],
[-40.373,-14.950,-53.797],[-40.014,-13.289,-51.691],[-29.255,-5.904,-49.515],
[-31.131,-11.022,-49.111],[-22.740,-21.340,-49.808],[-28.302,-16.936,-49.390],
[-36.628,-16.403,-51.627],[-27.756,-18.938,-52.838],[-39.156,-16.176,-51.611],
[-33.328,-14.796,-51.228],[-23.077,-15.400,-45.293],[-35.334,-18.412,-46.250],
[-29.121,-10.803,-52.510],[-33.989,-11.314,-46.676],[-24.581,-18.294,-51.394],
[-37.228,-16.598,-48.545],[-36.771,-11.112,-46.407],[-38.629,-9.324,-54.621],
[-27.783,-7.951,-48.760],[-25.586,-19.262,-47.429],[-32.808,-22.278,-48.771],
[-26.608,-7.330,-51.466],[-27.193,-8.243,-54.073],[-23.402,-10.659,-45.444],
[-31.083,-11.862,-54.109],[-29.210,-16.412,-59.243],[-21.810,-13.465,-46.734],
[-27.445,-13.601,-52.775],[-25.462,-13.195,-45.971],[-36.683,-13.523,-51.663],
[-35.257,-9.433,-49.145],[-34.216,-12.366,-49.491],[-29.752,-8.260,-47.151],
[-33.524,-9.020,-43.517],[-28.115,-13.214,-56.302],[-36.839,-13.514,-44.707],
[-29.507,-7.103,-54.233],[-36.000,-7.911,-43.378],[-25.133,-21.399,-56.600],
[-36.532,-13.558,-54.713],[-22.639,-7.985,-58.063],[-39.316,-11.026,-47.582],
[-40.746,-8.823,-48.763],[-20.750,-6.742,-44.726],[-30.776,-7.986,-44.596],
[-30.944,-3.953,-55.821],[-37.482,-3.826,-53.544],[-27.145,-5.985,-56.589],
[-28.570,-4.425,-54.715],[-26.085,-14.802,-40.355],[-38.845,-9.682,-43.391],
[-36.073,-19.097,-48.778],[-29.140,-21.815,-45.960],[-25.614,-16.210,-43.763],
[-27.221,-14.020,-58.551],[-29.205,-19.263,-59.017],[-34.312,-16.288,-48.380],
[-34.844,-15.432,-53.500],[-29.761,-15.202,-54.728],[-28.887,-15.028,-43.189],
[-27.837,-16.248,-52.043],[-35.262,-15.818,-45.796],[-35.510,-17.915,-58.415],
[-41.097,-18.386,-52.333],[-37.891,-18.676,-55.894],[-30.481,-15.405,-47.357],
[-32.172,-18.340,-51.395],[-37.571,-16.221,-54.510],[-35.385,-17.442,-55.508],
[-28.331,-19.524,-47.712],[-34.876,-15.433,-57.863],[-25.663,-3.870,-42.718],
[-35.872,-10.732,-43.803],[-31.097,-14.575,-49.772],[-35.354,-18.772,-52.847],
[-32.248,-17.257,-46.697],[-32.084,-11.094,-56.828],[-24.326,-11.938,-47.905],
[-32.337,-21.834,-55.223],[-31.147,-10.981,-44.560],[-29.640,-17.274,-56.798],
[-38.381,-12.342,-49.762],[-29.572,-10.062,-55.998],[-39.178,-14.072,-44.087],
[-31.852,-12.890,-41.673],[-28.080,-7.147,-44.351],[-22.189,-16.665,-56.027],
[-21.367,-9.654,-43.768],[-25.237,-8.201,-57.336],[-26.385,-11.540,-49.672],
[-35.580,-8.083,-53.289],[-39.508,-11.489,-57.593],[-33.316,-9.744,-53.412],
[-32.260,-6.554,-46.727],[-35.758,-5.630,-52.492],[-28.589,-10.625,-47.698],
[-28.166,-8.628,-58.557],[-25.619,-5.776,-44.860],[-23.362,-4.953,-57.789],
[-38.011,-17.868,-45.800],[-26.832,-10.454,-56.920],[-37.715,-8.648,-46.733],
[-32.412,-9.448,-59.152],[-33.828,-8.486,-55.530],[-37.076,-10.888,-56.535],
[-29.320,-17.746,-45.958],[-28.531,-19.832,-55.372],[-31.791,-7.901,-57.137],
[-31.476,-17.213,-53.874],[-31.727,-15.388,-43.409],[-30.553,-20.265,-53.291],
[-35.024,-6.123,-55.970],[-34.711,-20.039,-55.752],[-24.797,-18.940,-44.269],
[-23.531,-14.618,-48.612],[-36.732,-8.277,-56.459],[-38.445,-16.632,-58.144],
[-25.680,-5.889,-48.288],[-36.976,-21.436,-50.631],[-37.887,-6.838,-53.926],
[-39.836,-6.547,-41.607],[-34.530,-5.857,-40.909],[-32.400,-4.899,-58.471],
[-28.137,-9.740,-42.406],[-26.284,-7.227,58.341],[-40.373,-9.095,-40.905],
[-36.561,-11.313,-59.138],[-34.167,-11.825,-42.121],[-30.282,-11.895,-59.023],
[-27.692,-10.305,-44.996],[-37.846,-13.906,-41.006],[-34.130,-12.076,-55.216],
[-33.411,-14.045,-44.983],[-31.842,-15.470,-58.624],[-29.395,-12.252,-42.186],
[-25.996,-11.823,-59.308],[-33.600,-17.439,-44.405],[-33.077,-19.866,-47.437],
[-31.638,-12.814,-46.895],[-28.550,-16.884,-40.751],[-26.704,-13.340,-42.899],
[-35.335,-15.394,-42.330],[-32.066,-22.544,-45.381],[-31.902,-18.465,-42.745],
[-29.075,-17.870,-43.030],[-26.479,-19.507,-42.267],[-36.039,-18.529,-40.451],
[-32.018,-21.516,-42.924],[-32.544,-21.822,-39.965],[-32.816,-2.542,-57.090],
[-38.919,-8.941,-58.497],[-39.961,-3.154,-54.239],[-40.228,-5.624,-52.476],
[-23.381,-5.559,-49.675],[-23.164,-5.765,-46.293],[-24.333,-9.314,-41.257],
[-23.386,-10.640,-54.955],[-23.636,-14.830,-53.244],[-23.064,-9.646,-48.916],
[-23.197,-8.345,-46.625],[-23.370,-12.530,-56.782],[-21.871,-15.165,-58.388],
[-25.520,-14.624,-51.003],[-40.843,-13.743,-49.043],[-23.998,-12.778,-43.734],
[-22.625,-14.936,-42.790],[-25.661,-17.975,-58.584],[-25.944,-17.794,-55.794],
[-24.744,-17.078,-49.025],[-25.558,-15.696,-46.539],[-39.222,-18.491,-42.468],
[-23.266,-22.088,-40.041],[-24.071,-19.743,-54.117],[-39.231,-18.119,-49.444],
[-39.863,-15.925,-46.932],[-22.298,-21.318,-42.753],[-22.133,-19.999,-58.288],
[-39.792,-20.091,-54.818],[-25.167,-4.561,-53.128],[-40.859,-20.239,-45.865],
[-37.270,-7.542,-41.048],[-34.627,-7.578,-45.776],[-31.433,-5.644,-40.850],
[-28.495,-4.127,-43.762],[-23.639,-21.856,-46.947],[-38.715,-4.324,-44.442],
[-34.672,-4.747,-47.289],[-30.624,-4.315,-47.601],[-28.001,-5.893,-46.812],
[-26.531,-3.337,-45.898],[-37.412,-7.075,-49.705],[-34.771,-7.386,-50.683],
[-31.968,-5.527,-50.151],[-30.690,-20.584,-50.685],[-25.452,-22.019,-49.374],
[-24.951,-20.882,-51.925],[-38.111,-20.033,-52.775],[-40.142,-19.956,-57.360],
[-32.304,-22.692,-52.063],[-27.856,-21.205,-51.479],[-29.643,-22.409,-56.057],
[-39.662,-6.953,-55.867],[-38.916,-18.951,59.298],[-31.750,-19.126,-55.732],
[-29.894,-5.896,-57.321],[-22.974,-4.269,-54.918],[-39.206,-21.269,-41.008],
[-36.232,-21.006,-42.035],[-37.824,-6.292,-57.998],[-29.552,-21.963,-40.847],
[-22.167,-7.644,-41.616],[-30.384,8.144,-52.010],[-29.546,6.309,-49.748],
[-37.020,4.497,-47.806],[-24.026,10.056,-51.331],[-35.943,9.020,-50.563],
[-40.311,2.153,-53.861],[-39.972,5.390,-50.983],[-27.731,13.676,-51.054],
[-30.942,8.373,-48.415],[-22.312,-0.228,-48.048],[-27.688,1.803,-48.437],
[-36.934,0.241,-50.411],[-25.864,0.234,-52.970],[-38.567,3.153,-49.931],
[-33.771,4.109,-51.142],[-22.973,1.339,-44.644],[-36.763,-0.601,-46.291],
[-29.060,9.367,-54.170],[-32.739,7.169,-46.485],[-24.960,-2.258,-51.007],
[-38.031,1.458,-47.906],[-35.612,7.327,-46.852],[-40.482,9.627,-56.677],
[-26.419,9.966,-47.368],[-25.168,-1.876,-48.231],[-34.472,-3.146,-49.453],
[-26.120,11.396,-49.972],[-26.683,10.057,-52.674],[-23.652,9.427,-43.573],
[-29.484,5.855,-53.955],[-30.122,4.268,-58.785],[-23.401,3.606,-46.412],
[-26.750,5.568,-53.508],[-25.947,4.873,-44.613],[-34.474,6.818,-51.426],
[-33.379,9.260,-49.994],[-34.623,4.290,-48.685],[-28.903,11.105,-48.458],
[-33.936,9.921,-44.899],[-28.099,5.129,-56.367],[-35.890,5.082,-45.330],
[-32.136,11.143,-52.179],[-36.134,9.853,-43.297],[-23.549,-2.395,-57.039],
[-35.379,5.502,-53.635],[-23.468,9.642,-58.990],[-41.948,6.941,-47.363],
[-40.088,11.281,-49.572],[-21.478,10.856,-44.575],[-30.865,9.500,-46.009],
[-30.005,12.930,-52.835],[-37.252,12.004,-52.999],[-25.831,12.716,-56.629],
[-28.118,12.944,-55.549],[-26.238,2.981,-42.704],[-37.607,8.979,-45.874],
[-34.418,-1.974,-47.029],[-29.257,-2.571,-46.001],[-27.326,0.773,-44.346],
[-25.231,5.964,-56.148],[-30.153,1.633,-58.601],[-35.349,1.980,-47.350],
[-33.341,0.969,-53.249],[-28.563,3.197,-54.350],[-30.861,1.328,-40.851],
[-29.140,2.030,-51.653],[-37.056,1.844,-44.843],[-33.967,0.641,-58.104],
[-41.460,-1.294,-50.604],[-38.070,1.237,-55.596],[-30.034,3.425,-46.930],
[-34.260,-0.646,-50.948],[-39.303,-0.712,-52.862],[-35.397,1.719,-54.691],
[-28.820,-0.469,-49.130],[-35.121,4.128,-58.614],[-25.539,14.073,-42.641],
[-37.689,6.855,-42.180],[-31.170,3.643,-49.486],[-35.933,-0.222,-53.137],
[-32.503,1.223,-46.888],[-32.101,6.329,-55.488],[-25.740,6.068,-46.877],
[-27.320,0.690,-57.651],[-30.974,6.847,-44.295],[-26.476,1.445,-55.300],
[-37.588,6.352,-50.068],[-30.193,6.744,-57.326],[-39.371,3.210,-43.992],
[-33.763,3.480,-42.380],[-27.891,9.889,-44.954],[-21.108,1.844,-55.624],
[-21.237,7.850,-44.842],[-22.258,10.180,-54.550],[-26.266,6.740,-49.524],
[-34.715,10.429,-52.914],[-40.165,6.478,-58.022],[-33.134,8.471,-52.627],
[-32.482,12.333,-45.934],[-34.765,13.204,-54.085],[-28.002,7.496,-47.479],
[-27.947,8.214,-57.668],[-26.372,11.800,-44.077],[-22.582,11.449,-57.122],
[-39.583,1.068,-45.511],[-27.665,7.457,-55.135],[-38.053,11.099,-47.616],
[-34.353,11.260,58.957],[-34.573,9.131,-56.399],[-38.259,8.082,-56.181],
[-29.157,0.483,-46.265],[-28.313,-0.617,-55.155],[-31.861,9.984,-54.920],
[-30.411,0.671,-53.712],[-30.209,2.596,-44.459],[-31.973,-2.081,-52.592],
[-32.920,11.487,-57.241],[-35.019,-0.793,-55.991],[-24.494,-0.617,-43.642],
[-26.274,3.953,-48.812],[-37.153,10.476,-57.662],[-37.073,3.110,-57.049],
[-25.275,12.990,-48.088],[-38.030,-4.631,-50.678],[-38.154,9.182,-53.547],
[-39.135,13.227,-43.537],[-32.287,10.928,-42.896],[-31.105,13.454,58.704],
[-29.902,9.162,-42.978],[-25.193,9.978,-56.331],[-40.340,6.637,-41.769],
[-34.921,6.849,-58.265],[-34.032,6.150,-43.524],[-31.174,9.469,-58.863],
[-28.221,7.106,-44.311],[-37.635,3.643,-42.225],[-34.197,4.710,-56.062],
[-34.371,2.785,-44.907],[-32.606,5.202,-58.129],[-29.991,6.300,-40.179],
[-23.353,6.959,-58.890],[-34.833,-2.191,-43.735],[-34.196,0.065,-44.851],
[-32.266,4.655,-45.364],[-28.324,2.112,-41.423],[-26.709,5.729,-41.738],
[-35.968,2.186,-40.922],[-31.758,-1.559,-46.528],[-32.349,1.342,-43.100],
[-29.885,0.030,-43.290],[-27.436,-0.442,-42.195],[-37.464,0.019,-42.917],
[-31.230,-3.104,-44.084],[-32.495,-1.379,-41.589],[-33.315,14.125,-58.059],
[-39.154,9.263,-58.924],[-40.666,14.969,-53.244],[-41.102,13.290,-51.063],
[-22.268,13.122,-47.733],[-23.748,12.751,-45.381],[-25.158,9.086,-40.395],
[-22.213,6.784,-54.937],[-24.121,3.234,-54.164],[-23.298,8.699,-49.310],
[-23.658,10.892,-47.608],[-22.832,4.856,-56.685],[-23.091,2.218,59.188],
[-25.243,4.127,-51.719],[-39.724,7.701,-48.759],[-23.875,7.092,-44.912],
[-23.557,4.868,-42.582],[-24.419,-0.688,-59.595],[-23.940,0.509,-54.957],
[-25.089,1.109,-48.128],[-26.395,2.859,-46.303],[-41.430,0.528,-43.122],
[-22.243,-3.075,-39.575],[-24.468,-1.966,-53.777],[-37.637,-2.416,-49.138],
[-39.667,4.013,-47.413],[-23.070,-3.203,-42.205],[-21.210,-2.070,-58.728],
[-38.249,-1.516,-55.160],[-25.495,12.487,-52.823],[-39.424,-1.520,-44.645],
[-39.734,9.935,-42.679],[-35.108,10.838,-48.034],[-30.058,12.179,-42.146],
[-29.049,14.224,-43.879],[-23.765,-2.923,-45.175],[-36.260,15.518,-46.596],
[-35.639,12.924,-46.229],[-29.785,13.881,-49.172],[-27.979,13.405,-46.497],
[-26.153,15.364,-46.977],[-38.586,13.415,-49.292],[-35.667,12.596,-50.326],
[-32.618,13.304,-48.401],[-31.808,-1.298,-49.482],[-25.786,-4.820,-50.537],
[-22.521,-3.619,-51.673],[-37.445,-2.236,-51.667],[-38.261,-4.052,-56.649],
[-33.030,-4.613,-52.653],[-28.095,-3.067,-49.549],[-25.701,-3.830,-56.264],
[-39.493,11.290,-54.510],[-36.596,-2.166,-57.936],[-31.310,-0.387,-56.004],
[-30.560,13.625,-57.355],[-23.317,13.227,-55.287],[-36.923,-2.644,-41.999],
[-35.058,-2.976,-39.963],[-38.057,12.825,-56.970],[-28.418,-3.517,-40.097],
[-22.952,12.299,-42.686],[-27.910,23.384,-52.308],[-29.914,24.696,-47.885],
[-38.119,24.307,-47.397],[-26.234,27.553,-50.912],[-36.138,27.902,-49.387],
[-38.091,22.203,-53.480],[-39.101,24.560,-52.436],[-28.531,32.063,-49.915],
[-31.907,27.306,-47.879],[-23.098,19.044,-49.339],[-26.863,21.000,-48.996],
[-35.253,19.805,-51.079],[-28.154,18.510,-52.040],[-39.113,21.893,-50.910],
[-33.891,21.926,-50.424],[-22.212,20.368,-46.206],[-35.996,18.180,-48.903],
[-29.525,27.015,-52.028],[-34.505,26.248,-47.317],[-25.224,18.301,-50.790],
[-38.188,20.627,-48.649],[-37.089,27.462,-46.541],[-38.732,27.877,-57.609],
[-26.848,27.822,-46.653],[-25.528,18.108,-47.752],[-31.791,15.822,-50.062],
[-28.140,31.127,-52.511],[-26.249,25.559,-52.751],[-25.356,27.117,-44.559],
[-30.884,25.039,-53.441],[-30.039,21.976,-59.147],[-23.401,23.035,-45.922],
[-26.086,24.511,-55.174],[-28.262,21.835,-44.205],[-34.986,25.056,-52.125],
[-33.374,27.530,-50.505],[-35.481,25.256,-49.496],[-27.558,29.831,-48.381],
[-33.796,27.979,-45.301],[-28.877,24.067,-55.498],[-36.670,25.033,-44.831],
[-30.359,29.923,-51.891],[-35.757,30.914,-44.108],[-24.125,16.202,-56.402],
[-35.977,23.646,-54.705],[-24.263,30.024,-58.058],[-39.698,26.311,-46.350],
[-40.636,29.505,-47.763],[-22.232,30.640,-42.530],[-31.623,30.137,-44.620],
[-31.379,32.314,-52.951],[-38.477,31.642,-53.532],[-27.705,31.393,-57.859],
[-27.823,31.758,-55.344],[-23.964,21.121,-43.865],[-37.997,28.179,-44.075],
[-34.571,15.374,-48.750],[-30.621,14.389,-46.055],[-26.996,19.017,-43.486],
[-26.951,24.973,-57.713],[-31.557,18.524,-57.229],[-34.545,20.398,-48.318],
[-34.797,21.592,-53.454],[-30.711,22.420,-54.071],[-30.108,23.398,-41.761],
[-30.579,20.688,-51.016],[-35.962,22.530,-46.148],[-33.252,18.460,-59.375],
[-40.678,17.888,-52.316],[-38.299,20.711,-55.619],[-29.279,22.084,-46.893],
[-32.896,17.485,-52.014],[-36.721,19.673,-53.325],[-34.712,19.314,-55.658],
[-28.563,19.024,-48.155],[-34.788,21.758,-57.490],[-25.577,34.509,-42.795],
[-36.813,26.273,-42.285],[-31.585,21.518,-48.855],[-34.665,16.134,-53.655],
[-31.995,20.249,-46.319],[-32.846,26.704,-55.407],[-24.668,25.197,-47.365],
[-28.386,18.081,-57.895],[-31.054,27.477,-43.769],[-28.565,21.620,-56.598],
[-38.153,26.403,-50.033],[-30.799,28.427,-56.884],[-39.471,24.565,-44.217],
[-32.699,22.748,-41.841],[-28.530,27.991,-44.483],[-22.783,21.294,-56.212],
[-23.666,29.100,-44.411],[-22.948,28.093,-56.332],[-27.248,24.741,-48.431],
[-34.836,29.084,-54.505],[-39.454,24.973,-58.361],[-32.726,28.520,-53.012],
[-33.215,30.931,-46.422],[-34.503,32.799,-51.483],[-27.790,25.508,-45.792],
[-27.767,27.288,-58.729],[-28.094,30.276,-42.982],[-23.426,33.575,-57.732],
[-38.937,18.579,-44.073],[-26.818,27.104,-55.532],[-37.899,29.964,-45.913],
[-35.224,29.855,-58.892],[-33.947,28.246,-57.117],[-36.148,26.877,-58.967],
[-29.429,18.833,-45.417],[-29.325,18.349,-55.241],[-30.943,30.736,-55.287],
[-32.055,19.950,-53.722],[-30.940,21.895,-43.927],[-29.967,16.762,-52.958],
[-33.544,30.670,-56.277],[-33.953,16.922,-56.614],[-24.748,18.185,-44.945],
[-25.652,23.459,-50.346],[-36.388,29.088,-56.576],[-37.212,20.938,-58.005],
[-25.642,31.741,-48.317],[-37.185,15.848,-49.423],[-40.012,29.151,-52.939],
[-38.674,32.931,-42.341],[-33.448,32.462,-40.057],[-31.703,30.522,-58.908],
[-27.001,27.659,-41.717],[-26.665,29.511,-59.487],[-39.517,26.044,-41.739],
[-35.500,25.110,58.035],[-33.414,25.712,-43.648],[-30.556,27.167,58.611],
[-28.255,25.062,-43.116],[-37.575,22.549,-41.376],[-35.178,25.291,-56.917],
[-33.504,23.254,-44.603],[-33.009,21.536,-59.936],[-28.081,25.432,-40.301],
[-22.765,25.983,-59.073],[-34.513,19.733,-44.624],[-34.030,18.102,-46.885],
[-32.088,25.010,-46.244],[-27.336,21.253,-41.487],[-26.070,23.681,-42.848],
[-34.944,21.532,-42.011],[-33.258,15.330,-46.425],[-31.592,19.007,-43.488],
[-29.499,19.552,-41.959],[-27.323,18.091,-40.911],[-36.799,19.338,-41.538],
[-33.192,14.393,-43.957],[-32.547,15.536,-41.122],[-33.173,32.604,-58.224],
[-40.067,29.869,-59.138],[-41.251,31.529,-54.064],[-39.697,31.315,-50.813],
[-22.908,32.717,-48.418],[-23.299,31.767,-45.874],[-23.814,27.923,-42.008],
[-23.547,25.589,-55.726],[-23.554,24.167,-52.121],[-20.491,26.687,-48.527],
[-24.781,29.560,-46.753],[-24.004,23.947,-57.755],[-22.950,19.012,59.076],
[-25.872,21.816,-53.700],[-40.440,24.690,-49.443],[-24.291,25.285,-41.850],
[-22.733,23.216,-42.570],[-25.839,19.197,-58.677],[-24.491,19.497,-55.225],
[-24.485,20.736,-47.506],[-26.298,22.166,-46.292],[-39.420,20.464,-42.320],
[-22.806,14.595,-41.378],[-25.301,16.962,-53.977],[-38.005,18.386,-50.766],
[-41.495,22.361,-49.249],[-22.261,17.470,-42.897],[-21.949,16.623,-58.641],
[-38.921,17.265,-54.180],[-25.742,32.665,-53.679],[-41.762,15.126,-45.575],
[-37.745,30.143,-42.077],[-34.386,29.163,-47.816],[-30.169,32.533,-43.422],
[-27.699,33.926,-44.191],[-23.649,15.412,-44.125],[-38.488,33.723,-44.793],
[-36.172,32.023,-46.956],[-30.486,34.210,-48.196],[-28.504,31.537,-45.316],
[-26.115,34.155,-46.226],[-39.166,31.878,-46.962],[-35.786,30.104,-51.219],
[-35.886,33.775,-48.787],[-30.332,18.092,-50.023],[-24.712,15.955,-49.403],
[-22.444,15.115,-51.260],[-38.412,15.763,-52.003],[-39.146,15.545,-56.438],
[-32.684,14.764,-52.786],[-26.524,16.105,-51.653],[-26.733,15.259,-55.619],
[-40.405,29.236,-55.619],[-35.936,18.226,-57.999],[-31.692,17.063,-54.828],
[-30.644,33.754,-56.799],[-24.416,30.890,-55.260],[-38.212,15.935,-44.435],
[-36.401,16.860,-42.853],[-36.917,32.146,-58.377],[-27.627,15.503,-39.904],
[-24.934,30.814,-42.291],[-27.787,44.114,-51.724],[-31.688,43.857,-49.650],
[-36.933,41.017,-46.993],[-26.214,46.082,-50.676],[-36.418,46.429,-49.778],
[-39.863,40.705,-53.916],[-39.200,42.488,-51.962],[-29.835,50.486,-50.423],
[-30.679,46.452,-49.174],[-22.392,35.830,-48.348],[-27.550,38.740,-46.942],
[-36.408,37.479,-50.958],[-26.655,36.427,-53.195],[-38.448,38.128,-53.262],
[-31.893,41.249,-51.080],[-22.644,39.865,-44.706],[-38.845,37.820,-46.638],
[-29.548,45.717,-53.611],[-34.017,43.170,-46.947],[-24.592,36.792,-50.091],
[-37.502,39.295,-49.287],[-36.650,44.451,-47.605],[-38.854,45.221,-55.404],
[-26.367,48.019,-48.020],[-24.967,36.522,-46.845],[-32.555,33.868,-49.753],
[-26.972,49.017,-51.042],[-27.656,48.683,-53.988],[-24.335,45.683,-44.904],
[-31.456,42.082,-53.826],[-28.870,40.552,-58.978],[-22.760,43.149,-45.712],
[-26.820,42.304,-53.734],[-27.202,41.167,-45.566],[-35.678,42.608,-52.137],
[-33.603,45.921,-48.902],[-35.311,42.331,-49.253],[-30.088,48.967,-47.742],
[-32.949,46.658,-44.131],[-28.631,41.857,-56.379],[-36.103,42.592,-44.921],
[-30.225,49.372,-52.781],[-37.410,48.909,-44.732],[-23.748,35.083,-55.432],
[-35.128,43.683,-54.598],[-22.159,48.928,-59.183],[-42.109,45.001,-44.670],
[-40.117,47.232,-49.137],[-20.400,49.028,-42.976],[-30.908,48.186,-45.462],
[-30.058,51.726,-54.637],[-36.767,52.011,-53.884],[-29.652,49.710,-58.936],
[-28.637,47.613,-56.484],[-25.333,40.595,-43.175],[-38.271,45.936,-45.524],
[-36.853,35.321,-46.720],[-30.555,34.457,-45.400],[-27.420,38.679,-43.589],
[-25.788,41.193,-58.408],[-29.873,38.190,-56.964],[-34.451,39.241,-48.477],
[-34.218,41.182,-53.919],[-29.314,40.052,-53.326],[-29.373,40.868,-41.884],
[-27.990,41.175,-50.259],[-36.565,38.601,-46.095],[-32.376,36.021,-59.489],
[-40.252,36.271,-51.537],[-39.383,38.263,-56.993],[-29.120,40.863,-47.506],
[-33.420,37.082,-51.394],[-37.220,40.271,-52.163],[-34.840,38.891,-55.275],
[-29.177,36.494,-47.763],[-34.661,39.905,-58.085],[-24.653,51.863,-42.565],
[-36.867,45.154,-42.732],[-29.876,39.390,-49.590],[-36.004,36.427,-53.480],
[-32.222,39.493,-46.768],[-31.748,43.498,-57.686],[-25.604,43.548,-46.430],
[-28.216,36.682,-58.591],[-29.998,45.108,-44.822],[-27.320,38.622,-56.521],
[-37.523,43.812,-50.135],[-29.737,44.883,-56.377],[-40.914,42.278,-43.737],
[-33.363,41.387,-42.638],[-28.152,48.049,-45.327],[-22.778,39.586,-57.341],
[-22.111,46.860,-44.226],[-24.932,48.212,-56.044],[-26.909,43.389,-49.116],
[-34.535,48.096,-51.614],[-40.228,41.864,-59.102],[-33.469,44.522,-52.273],
[-32.848,49.377,-47.046],[-35.776,48.706,-54.362],[-28.931,45.980,-47.056],
[-27.431,45.801,-58.342],[-25.864,49.823,-45.455],[-24.327,50.428,-58.679],
[-38.813,37.220,-44.073],[-26.733,45.975,-55.664],[-39.464,48.450,-46.463],
[-34.394,46.799,-58.577],[-33.875,45.520,-56.026],[-37.064,43.253,-59.028],
[-30.863,37.080,-45.502],[-28.881,36.442,-55.190],[-31.444,47.111,-56.510],
[-30.791,37.850,-52.710],[-31.805,39.377,-44.091],[-29.200,35.213,-52.659],
[-34.539,51.082,-55.880],[-35.011,36.877,-57.358],[-24.712,38.371,-44.881],
[-25.281,40.278,-50.175],[-36.828,48.175,-56.815],[-37.485,39.369,-58.807],
[-25.374,50.482,-48.435],[-39.367,33.803,-50.209],[-38.928,47.469,-53.817],
[-39.123,51.134,-41.184],[-35.046,50.002,-43.144],[-31.817,47.805,-59.509],
[-27.965,48.775,-41.623],[-27.010,49.348,-57.813],[-39.607,44.736,-42.703],
[-34.533,43.759,-58.002],[-33.618,44.445,-42.543],[-30.295,45.507,-58.947],
[-27.092,45.579,-44.944],[-38.337,42.263,-42.584],[-33.046,41.426,-56.255],
[-34.374,40.469,-45.223],[-31.689,40.186,-58.576],[-30.037,43.491,-41.019],
[-22.730,44.705,-58.043],[-34.522,37.299,-43.477],[-34.105,37.065,-46.799],
[-31.344,43.115,-47.039],[-26.197,39.933,-40.828],[-27.248,42.717,-43.362],
[-36.427,40.498,-43.506],[-32.889,33.492,-47.011],[-32.055,38.848,-41.116],
[-29.724,38.029,-42.523],[-27.982,36.115,-41.533],[-37.161,37.844,-42.181],
[-33.010,33.056,-43.757],[-31.819,36.601,-42.808],[-31.685,51.373,-56.947],
[-38.096,46.300,-58.046],[-39.479,52.252,-53.080],[-39.092,51.190,-50.648],
[-21.830,52.560,-47.908],[-22.328,50.659,-45.854],[-24.645,46.637,-42.177],
[-24.385,44.886,-55.689],[-22.902,42.658,-53.957],[-22.861,45.772,-47.215],
[-23.597,48.630,-46.717],[-23.599,42.207,-57.073],[-23.958,39.733,59.121],
[-24.667,40.342,-53.001],[-39.983,44.366,-49.730],[-24.567,42.757,-41.640],
[-20.980,42.223,-40.845],[-24.952,36.494,-57.875],[-25.285,37.682,-55.411],
[-24.666,38.983,-47.990],[-24.702,41.005,-46.345],[-39.228,39.856,-41.684],
[-21.300,35.079,-40.645],[-24.206,34.972,-52.469],[-38.831,36.886,-49.142],
[-41.007,40.771,-48.162],[-23.024,35.127,-42.571],[-22.644,35.546,59.111],
[-39.299,35.894,-55.004],[-26.414,50.924,-53.125],[-41.171,36.254,-45.562],
[-38.567,48.351,-41.930],[-35.328,46.991,-47.171],[-30.282,50.192,-43.024],
[-28.695,51.938,-43.986],[-23.717,34.832,-45.028],[-38.443,51.256,-45.902],
[-35.566,49.683,-47.025],[-29.652,52.448,-48.072],[-28.069,50.245,-47.019],
[-26.380,52.684,-46.385],[-37.542,50.138,-48.626],[-36.889,49.696,-51.428],
[-32.452,48.556,-49.653],[-30.891,36.478,-50.046],[-25.017,34.387,-48.950],
[-21.874,33.831,-51.700],[-37.375,34.313,-52.187],[-38.162,33.337,-56.204],
[-32.752,34.625,-52.613],[-27.017,33.607,-51.361],[-26.778,34.408,-57.519],
[-40.286,48.070,-55.901],[-37.217,34.385,59.062],[-32.500,38.093,-56.467],
[-30.169,52.692,-59.316],[-23.200,51.269,-54.675],[-37.274,35.243,-42.523],
[-36.053,34.573,-40.168],[-37.592,51.155,-58.050],[-27.977,33.507,-41.334],
[-23.748,49.019,-43.864],[-23.183,55.512,-49.050],[-27.290,57.499,-48.485],
[-34.387,55.716,-50.832],[-26.042,54.655,-53.858],[-40.966,58.301,-51.754],
[-23.382,57.501,-47.054],[-36.389,52.734,-46.523],[-24.406,57.240,-51.096],
[-36.618,57.407,-50.571],[-25.606,55.406,-47.582],[-32.142,52.111,-50.179],
[-25.073,53.634,-59.061],[-25.250,59.176,-39.556],[-35.325,52.017,-49.082],
[-28.856,53.725,-45.729],[-28.865,-59.504,-46.319],[-29.831,55.145,-57.683],
[-35.125,56.999,-47.997],[-34.971,58.353,-53.329],[-28.379,56.527,-52.575],
[-38.299,57.701,-47.453],[-31.912,56.564,-59.591],[-39.389,55.058,-53.357],
[-37.841,56.845,-57.420],[-31.745,56.896,-52.872],[-37.505,57.013,-53.149],
[-33.078,57.968,-55.227],[-29.160,55.349,-48.014],[-31.008,-59.283,-56.410],
[-30.332,57.966,-50.708],[-37.233,54.559,-55.043],[-31.073,58.319,-47.118],
[-27.371,56.915,-57.553],[-28.412,58.549,-55.725],[-21.975,58.380,-56.046],
[-37.971,56.012,-45.246],[-28.916,56.830,-45.675],[-28.006,55.188,-55.379],
[-30.961,59.283,-53.682],[-31.126,-59.171,-44.595],[-29.908,53.944,-53.180],
[-34.468,55.165,-56.310],[-25.941,57.725,-45.078],[-37.841,59.040,-59.561],
[-36.949,54.262,-49.080],[-35.245,56.417,-45.188],[-33.106,55.607,-47.426],
[-28.111,58.485,-40.771],[-35.904,58.209,-43.511],[-32.505,52.861,-47.217],
[-32.746,57.493,-43.841],[-29.752,57.560,-43.008],[-27.148,54.238,-42.068],
[-39.677,55.767,-41.651],[-31.277,52.704,-44.191],[-32.166,53.014,-41.752],
[-24.346,-59.039,-58.480],[-24.022,57.084,-58.434],[-25.696,58.176,-54.197],
[-25.226,58.939,-49.200],[-26.987,-58.101,-44.880],[-39.593,58.015,-43.407],
[-22.559,53.213,-40.923],[-24.475,53.619,-56.104],[-39.127,55.776,-49.631],
[-40.560,58.040,-48.977],[-23.268,55.669,-42.524],[-22.497,54.592,-58.659],
[-40.601,54.782,-55.666],[-39.021,53.927,-46.439],[-24.148,53.674,-44.353],
[-30.437,55.218,-50.897],[-24.170,50.038,-50.851],[-24.154,54.668,-51.766],
[-37.191,53.415,-51.724],[-38.883,52.186,-55.920],[-32.760,50.757,-52.432],
[-28.130,52.936,-50.810],[-25.705,51.325,-56.075],[-35.909,53.889,-58.856],
[-31.675,54.271,-55.098],[-37.995,53.624,-44.078],[-36.824,52.835,-41.787],
[-27.408,51.485,-41.530],[-29.992,-49.485,-33.468],[-30.561,-50.945,-31.037],
[-37.523,-51.212,-29.893],[-26.441,-48.729,-33.110],[-36.590,-46.715,-31.116],
[-40.862,-51.981,-35.725],[-39.666,-51.306,-33.316],[-28.261,-42.580,-30.627],
[-32.360,-46.040,-29.739],[-22.532,-56.198,-28.458],[-28.379,-54.695,-31.004],
[-35.796,-56.014,-32.193],[-27.719,-54.336,-33.800],[-39.288,-54.376,-32.699],
[-35.654,-51.097,-32.480],[-21.352,-54.158,-26.492],[-36.731,-57.174,-29.046],
[-29.928,-47.218,-34.858],[-35.973,-49.854,-27.386],[-25.789,-56.128,-32.742],
[-37.639,-54.371,-29.886],[-35.188,-46.829,-28.727],[-39.623,-46.912,-36.501],
[-24.604,-47.009,-29.958],[-25.022,-56.485,-27.465],[-32.929,-59.145,-32.215],
[-26.424,-44.843,-31.192],[-26.703,-45.744,-34.245],[-26.548,-46.756,-25.508],
[-31.052,-50.843,-35.940],[-28.768,-50.851,-39.985],[-22.414,-51.719,-27.574],
[-26.153,-50.281,-35.439],[-28.497,-52.735,-26.896],[-35.282,-48.354,-33.210],
[-33.926,-46.756,-31.696],[-34.966,-49.075,-30.040],[-29.506,-46.001,-29.705],
[-34.098,-46.199,-26.135],[-26.087,-51.157,-38.192],[-38.024,-48.930,-25.719],
[-29.544,-44.409,-34.579],[-35.497,-43.827,-24.972],[-26.414,-58.694,-37.290],
[-35.979,-50.272,-35.156],[-24.473,-44.415,-39.384],[-38.317,-46.393,-27.480],
[-40.539,-45.656,-29.389],[-20.860,-44.432,-25.512],[-31.877,-44.705,-24.731],
[-30.379,-42.147,-35.671],[-37.440,-41.351,-32.178],[-28.329,-44.118,-38.528],
[-27.479,-43.402,-36.111],[-25.659,-53.837,-23.225],[-37.579,-45.178,-24.416],
[-35.256,-58.742,-30.765],[-29.372,-58.134,-26.862],[-26.898,-55.086,-25.539],
[-24.989,-49.356,-40.039],[-32.082,-55.246,-39.213],[-34.212,-51.847,-29.866],
[-36.211,-53.639,-34.124],[-29.074,-53.118,-35.472],[-30.633,-52.149,-22.840],
[-28.729,-52.091,-32.510],[-35.531,-53.895,-27.926],[-33.874,-55.971,-41.451],
[-40.548,-56.373,-34.551],[-39.092,-53.956,-38.049],[-31.102,-51.713,-28.102],
[-33.017,-55.173,-33.045],[-38.494,-54.530,-35.507],[-34.363,-52.760,-36.554],
[-29.346,-56.813,-29.125],[-35.875,-52.873,-39.044],[-24.413,-39.996,-24.137],
[-36.206,-47.652,-23.654],[-32.472,-52.806,-32.038],[-37.092,-57.061,-34.311],
[-33.057,-55.413,-27.802],[-31.589,-48.072,-36.638],[-25.349,-48.756,-28.040],
[-27.764,-55.853,-38.849],[-29.925,-46.856,-25.271],[-26.488,-52.910,-36.353],
[-37.853,-49.626,-32.325],[-28.627,-46.412,-36.898],[-39.775,-52.853,-26.144],
[-33.149,-52.834,-23.504],[-28.415,-45.472,-26.927],[-22.470,-53.717,-37.156],
[-23.216,-47.171,-24.791],[-25.333,-45.644,-36.831],[-26.964,-49.534,-30.391],
[-35.301,-45.531,-33.987],[-38.319,-50.481,-38.961],[-33.981,-48.362,-35.864],
[-31.599,-43.791,-28.118],[-36.439,-42.411,-35.081],[-31.186,-48.692,-29.029],
[-27.285,-46.563,-39.245],[-26.741,-43.569,-25.985],[-23.311,-41.968,-39.393],
[-38.453,-54.925,-27.260],[-26.505,-48.437,-37.378],[-39.067,-43.733,-26.894],
[-33.314,-45.468,-40.160],[-33.805,-45.960,-37.483],[-36.995,-47.864,-37.653],
[-30.261,-54.737,-27.671],[-27.758,-55.722,-36.262],[-31.649,-44.193,-37.102],
[-31.081,-53.615,-37.227],[-31.181,-52.412,-25.629],[-29.669,-56.633,-34.466],
[-34.838,-43.024,-37.338],[-34.010,-57.230,-38.407],[-24.052,-54.460,-25.752],
[-25.008,-51.885,-30.477],[-36.343,-45.310,-36.687],[-40.210,-55.213,-40.470],
[-25.543,-42.290,-28.539],[-37.883,-57.320,-31.469],[-38.072,-45.310,-34.568],
[-39.255,-41.638,-23.254],[-33.423,-43.277,-22.984],[-30.564,-44.403,-40.122],
[-27.065,-44.338,-22.928],[-26.658,-44.277,-41.152],[-41.265,-48.932,-22.566],
[-34.982,-47.491,-40.884],[-33.572,-48.271,-23.571],[-31.418,-49.691,-38.560],
[-26.676,-49.453,-25.854],[-37.714,-50.093,-22.568],[-35.219,-50.312,-37.956],
[-33.561,-52.195,-27.010],[-33.405,-50.613,-40.704],[-30.792,-49.515,-23.708],
[-23.155,-47.353,-40.063],[-34.383,-54.486,-25.463],[-34.652,-57.587,-27.324],
[-31.807,-49.513,-26.300],[-28.993,-53.897,-21.700],[-27.297,-51.638,-24.188],
[-36.669,-52.333,-23.716],[-32.464,-58.905,-28.153],[-31.951,-56.441,-25.355],
[-29.109,-53.950,-24.745],[-29.760,-56.107,-23.008],[-36.680,-54.953,-23.436],
[-34.107,-59.221,-25.144],[-32.652,-58.520,-22.621],[-33.054,-41.034,-38.225],
[-40.159,-44.693,-40.092],[-41.205,-41.914,-35.518],[-38.691,-43.923,-32.157],
[-22.231,-43.073,-30.841],[-22.842,-42.840,-27.919],[-25.784,-47.606,-22.765],
[-22.745,-48.387,-35.768],[-23.761,-50.159,-33.964],[-22.337,-48.264,-28.927],
[-24.764,-45.141,-26.811],[-21.738,-49.007,-38.618],[-24.788,-54.031,-38.622],
[-25.817,-52.171,-32.995],[-40.304,-50.063,-29.895],[-23.941,-49.876,-24.080],
[-23.174,-52.300,-22.224],[-25.154,-56.663,-40.578],[-24.441,-55.014,-34.930],
[-24.349,-54.515,-29.881],[-25.770,-51.879,-27.530],[-39.602,-55.579,-24.989],
[-21.014,-58.765,-22.597],[-24.851,-57.923,-35.132],[-39.727,-56.073,-29.853],
[-40.050,-52.747,-29.504],[-22.196,-56.874,-24.063],[-23.256,-56.217,-38.762],
[-38.517,-59.173,-36.156],[-26.242,-42.500,-33.621],[-40.921,-57.859,-27.647],
[-39.223,-45.535,-22.336],[-36.582,-43.874,-27.581],[-29.880,-42.643,-24.482],
[-29.368,-40.008,-23.295],[-24.461,-59.078,-26.247],[-37.088,-41.313,-26.043],
[-34.115,-42.449,-27.794],[-30.077,-41.146,-29.196],[-29.014,-42.986,-27.615],
[-26.416,-40.448,-26.884],[-38.788,-43.071,-29.457],[-34.971,-44.467,-30.147],
[-34.682,-41.831,-31.305],[-30.675,-56.913,-31.920],[-25.099,-58.905,-29.266],
[-23.013,-59.312,-31.343],[-39.317,-59.161,-33.725],[-40.025,-57.709,-38.342],
[-32.827,-57.463,-34.746],[-28.061,-58.589,-31.842],[-27.476,-58.000,-34.296],
[-39.410,-44.160,-36.505],[-37.225,-57.050,-39.193],[-32.003,-58.617,-36.838],
[-29.653,-41.872,-38.271],[-24.742,-43.044,-36.151],[-38.534,-57.404,-26.719],
[-35.959,-57.260,-22.407],[-38.464,-42.593,-39.280],[-27.622,58.080,-23.194],
[-25.482,-42.496,-23.920],[-28.046,-30.938,-32.678],[-29.665,-31.334,-29.364],
[-37.045,-32.601,-28.460],[-26.676,-28.347,-34.024],[-37.974,-27.342,-29.969],
[-38.901,-34.188,-35.232],[-39.329,-31.846,-33.620],[-28.772,-24.805,-30.962],
[-30.791,-28.516,-30.703],[-22.634,-38.816,-29.882],[-27.648,-35.117,-31.151],
[-35.729,-37.017,-32.511],[-27.452,-37.113,-34.754],[-39.109,-34.515,-32.533],
[-32.541,-35.508,-30.033],[-21.594,-35.196,-27.064],[-35.728,-36.975,-29.665],
[-29.652,-28.941,-33.521],[-33.016,-30.957,-27.613],[-26.175,-37.908,-32.060],
[-35.604,-33.960,-31.691],[-37.002,-27.940,-27.324],[-38.664,-28.044,-35.714],
[-25.699,-27.889,-27.204],[-23.663,-37.313,-27.906],[-32.308,-40.392,-30.519],
[-26.782,-25.928,-32.434],[-26.834,-25.775,-35.450],[-26.057,-29.135,-24.633],
[-29.893,-30.631,-36.988],[-31.368,-33.422,-40.337],[-23.241,-32.782,-27.558],
[-25.934,-31.105,-35.682],[-28.335,-33.296,-26.523],[-34.685,-31.567,-33.208],
[-33.789,-27.746,-30.411],[-33.295,-30.672,-30.831],[-28.331,-26.748,-27.433],
[-34.085,-28.634,-27.083],[-28.175,-32.548,-38.121],[-37.593,-31.317,-26.307],
[-31.448,-26.874,-33.823],[-34.918,-26.819,-24.615],[-23.552,-40.528,-36.801],
[-37.029,-32.314,-35.715],[-23.360,-27.092,-39.236],[-38.357,-30.534,-28.914],
[-40.287,-25.945,-30.327],[-21.781,-25.605,-24.303],[-31.908,-26.812,-24.691],
[-30.015,-23.193,-35.758],[-38.619,-23.291,-33.011],[-27.457,-24.273,-39.173],
[-28.999,-25.740,-37.032],[-25.010,-34.304,-23.894],[-38.324,-26.592,-24.909],
[-35.071,-39.457,-29.181],[-29.465,-40.235,-26.331],[-27.444,-35.774,-24.828],
[-26.434,-32.276,-40.169],[-31.707,-37.739,-40.250],[-34.608,-35.003,-28.246],
[-34.245,-34.091,-34.490],[-31.011,-34.121,-35.873],[-29.505,-34.916,-22.770],
[-28.788,-37.089,-32.427],[-37.331,-35.926,-27.844],[-35.053,-36.911,-41.271],
[-41.526,-38.663,-32.428],[-39.712,-36.812,-38.912],[-30.468,-33.630,-28.152],
[-32.952,-37.852,-32.804],[-36.584,-35.939,-35.056],[-34.658,-36.543,-37.131],
[-27.677,-38.059,-28.111],[-35.995,-34.461,-39.083],[-23.487,-22.208,-23.876],
[-35.100,-30.415,-25.409],[-30.638,-33.976,-30.940],[-35.538,-39.577,-33.753],
[-29.854,-36.220,-28.992],[-32.679,-30.360,-37.435],[-25.070,-30.651,-27.897],
[-29.246,-37.408,-38.822],[-32.029,-29.917,-24.829],[-27.846,-35.295,-37.512],
[-35.842,-29.538,-31.042],[-30.248,-28.120,-36.017],[-39.697,-33.081,-25.240],
[-32.772,-33.873,-23.002],[-29.086,-28.354,-25.290],[-22.089,-34.655,-38.852],
[-22.915,-28.219,-23.926],[-23.502,-26.680,-36.326],[-26.696,-30.540,-30.227],
[-34.467,-27.509,-34.508],[-39.501,-30.954,-39.301],[-32.411,-30.328,-34.186],
[-32.016,-24.498,-27.174],[-34.807,-23.095,-35.153],[-28.491,-29.054,-28.480],
[-27.672,-28.569,-39.803],[-25.800,-25.095,-25.637],[-22.616,-23.141,-37.660],
[-37.852,-36.861,-25.316],[-27.535,-29.333,-36.795],[-38.118,-25.036,-28.269],
[-33.030,-27.710,-39.938],[-35.271,-28.575,-36.864],[-37.321,-29.905,-37.750],
[-30.885,-38.127,-25.500],[-29.350,-39.662,-36.514],[-33.068,-27.137,-36.689],
[-30.798,-36.515,-34.737],[-30.241,-35.414,-25.797],[-30.713,-39.718,-33.876],
[-32.866,-24.462,-37.026],[-34.795,-39.203,-36.913],[-24.623,-37.574,-25.304],
[-25.702,-33.074,-30.795],[-36.664,-27.397,-39.229],[-39.118,-35.932,-41.952],
[-25.047,-23.754,-28.576],[-38.852,-39.839,-29.936],[-39.219,-26.996,-33.399],
[-38.291,-23.323,-23.638],[-34.839,-25.009,-22.347],[-32.025,-25.360,-39.967],
[-28.015,-29.161,-22.704],[-25.928,-26.230,-38.403],[-39.834,-30.411,-23.982],
[-34.950,-29.673,-39.218],[-33.926,-29.013,-23.400],[-30.355,-28.574,-40.947],
[-27.931,-31.390,-24.367],[-38.854,-32.887,-22.708],[-35.290,-32.075,-37.885],
[-34.447,-32.995,-26.525],[-33.470,-31.672,-40.068],[-28.607,-31.903,-21.604],
[-23.340,-30.515,-40.647],[-33.028,-36.282,-25.383],[-33.259,-37.319,-28.042],
[-30.559,-31.153,-26.557],[-26.829,-34.756,-21.572],[-25.777,-32.829,-25.903],
[-36.353,-34.230,-22.271],[-32.211,-40.386,-26.608],[-31.814,-38.091,-22.912],
[-29.082,-37.518,-22.496],[-26.503,-37.523,-22.788],[-36.823,-34.286,-25.149],
[-34.087,-40.782,-24.526],[-32.016,-40.694,-22.001],[-33.503,-21.270,-37.743],
[-39.531,-28.293,-39.930],[-40.048,-22.252,-35.158],[-41.725,-24.952,-32.519],
[-22.332,-24.588,-29.675],[-22.855,-23.427,-26.313],[-25.368,-27.214,-22.884],
[-22.506,-29.260,-36.501],[-23.294,-32.084,-35.080],[-23.605,-28.854,-29.444],
[-24.200,-26.234,-28.625],[-24.565,-31.319,-37.939],[-24.386,-35.130,-40.031],
[-26.366,-34.000,-33.440],[-40.140,-31.317,-30.917],[-23.660,-30.910,-24.464],
[-21.669,-32.431,-22.745],[-24.784,-37.755,-40.979],[-25.382,-35.737,-35.858],
[-23.161,-34.357,-29.679],[-26.408,-34.468,-28.715],[-40.101,-35.653,-24.573],
[-22.361,-40.250,-21.810],[-23.759,-37.984,-36.455],[-39.346,-36.778,-31.077],
[-39.159,-34.219,-29.776],[-21.446,-39.349,-24.599],[-22.467,-39.461,-39.464],
[-39.695,-39.228,-35.296],[-25.372,-23.662,-32.993],[-39.513,-38.248,-27.367],
[-37.375,-29.174,-24.487],[-34.830,-26.228,-27.692],[-31.038,-24.337,-24.275],
[-28.302,-23.171,-23.787],[-23.166,-40.064,-27.097],[-37.682,-22.216,-27.758],
[-34.606,-23.338,-27.534],[-30.054,-21.941,-28.995],[-27.740,-23.399,-28.899],
[-26.488,-22.198,-26.920],[-38.748,-22.983,-30.327],[-36.383,-25.683,-32.080],
[-34.844,-24.876,-29.971],[-30.685,-38.364,-31.210],[-23.787,-40.666,-31.417],
[-23.627,-39.483,-34.008],[-38.425,-38.935,-33.034],[-38.795,-41.147,-36.884],
[-33.398,-41.184,-33.902],[-27.251,-40.410,-31.964],[-26.268,-40.686,-36.445],
[-38.407,-25.628,-36.866],[-37.823,-38.076,-40.266],[-31.592,-37.961,-37.036],
[-30.297,-23.885,-38.483],[-23.315,-24.322,-35.133],[-36.414,-38.855,-24.219],
[-34.726,-40.655,-21.547],[-35.713,-24.637,-39.779],[-26.856,-41.224,-21.843],
[-24.202,-24.589,-23.379],[-28.802,-13.445,-33.672],[-29.790,-12.385,-30.821],
[-37.161,-12.784,-29.255],[-25.323,-10.552,-31.462],[-36.123,-9.450,-29.613],
[-39.250,-14.506,-35.896],[-38.464,-11.416,-34.387],[-27.507,-5.147,-31.470],
[-30.160,-10.044,-29.649],[-23.363,-20.443,-28.563],[-27.177,-17.110,-30.491],
[-34.913,-17.913,-32.948],[-27.122,-18.326,-35.081],[-39.120,-15.417,-32.144],
[-33.280,-14.862,-32.995],[-21.507,-16.860,-28.027],[-35.972,-19.228,-28.457],
[-31.454,-9.644,-34.974],[-33.662,-12.034,-28.082],[-25.732,-20.150,-32.668],
[-37.960,-17.343,-30.085],[-36.096,-10.796,-27.171],[-38.657,-9.131,-35.838],
[-27.877,-10.247,-28.378],[-26.701,-17.922,-27.794],[-32.241,-22.099,-31.028],
[-29.311,-9.915,-32.730],[-28.095,-9.892,-35.189],[-24.916,-11.029,-25.989],
[-29.968,-12.495,-36.548],[-31.694,-15.450,-40.471],[-22.584,-12.366,-27.583],
[-25.904,-11.779,-34.056],[-25.002,-13.962,-28.262],[-35.716,-11.296,-32.588],
[-31.909,-8.200,-32.240],[-34.518,-13.155,-30.697],[-29.349,-7.603,-28.521],
[-34.142,-10.349,-25.218],[-25.695,-15.027,-37.197],[-37.061,-13.739,-26.533],
[-30.266,-6.160,-34.930],[-34.549,-6.711,-24.426],[-23.974,-20.382,-37.768],
[-35.212,-12.907,-35.815],[-22.713,-8.222,-38.741],[-39.406,-10.929,-26.359],
[-38.035,-6.624,-29.897],[-21.734,-8.235,-24.209],[-31.737,-8.050,-25.994],
[-30.195,-3.420,-34.594],[-36.208,-3.914,-32.894],[-28.012,-5.837,-38.714],
[-27.178,-6.574,-34.495],[-25.906,-15.091,-24.416],[-38.056,-8.775,-25.051],
[-34.863,-21.815,-30.004],[-29.061,-21.993,-26.395],[-25.644,-17.967,-24.560],
[-24.719,-12.341,-40.903],[-30.610,-18.013,-39.527],[-35.890,-15.190,-28.747],
[-34.358,-15.555,-35.220],[-29.439,-15.602,-36.206],[-29.057,-17.254,-21.473],
[-27.819,-16.328,-33.349],[-37.719,-16.325,-26.131],[-33.310,-17.857,-39.735],
[-42.215,-19.287,-33.344],[-40.641,-18.591,-38.246],[-29.172,-15.717,-28.423],
[-33.197,-19.371,-34.707],[-37.557,-17.341,-33.002],[-34.342,-17.765,-36.527],
[-28.310,-19.609,-29.902],[-35.791,-16.481,-38.461],[-24.943,-4.578,-21.684],
[-37.239,-10.377,-23.001],[-31.462,-15.401,-30.547],[-36.584,-18.515,-35.000],
[-33.951,-17.029,-29.017],[-31.363,-10.461,-37.560],[-25.205,-11.201,-28.700],
[-27.997,-19.385,-39.981],[-30.774,-10.400,-25.049],[-26.974,-16.746,-38.771],
[-38.949,-9.945,-32.065],[-28.334,-10.117,-37.856],[-39.324,-14.493,-25.194],
[-34.180,-13.991,-24.197],[-29.004,-8.525,-25.764],[-22.925,-15.472,-36.921],
[-23.916,-9.713,-23.557],[-24.118,-7.891,-36.452],[-27.260,-13.147,-29.673],
[-33.990,-7.969,-34.066],[-38.466,-12.053,-38.452],[-32.332,-10.790,-32.162],
[-33.508,-6.299,-27.950],[-34.836,-4.137,-35.138],[-29.351,-12.609,-27.146],
[-26.742,-10.794,-40.095],[-26.335,-7.244,-24.695],[-23.935,-5.646,-38.487],
[-38.785,-18.536,-24.122],[-26.231,-12.559,-38.201],[-38.383,-8.969,-27.772],
[-33.081,-8.320,-40.925],[-33.886,-9.508,-37.491],[-36.231,-10.303,-36.183],
[-30.688,-19.317,-27.144],[-28.995,-19.781,-36.410],[-29.608,-7.699,-37.331],
[-31.252,-17.245,-34.846],[-31.806,-16.712,-25.520],[-29.861,-19.320,-32.611],
[-33.231,-7.040,-38.386],[-36.129,-20.185,-37.033],[-22.415,-17.710,-25.625],
[-25.259,-15.447,-31.538],[-37.419,-9.267,-38.832],[-38.527,-16.703,-39.840],
[-24.543,-4.689,-27.848],[-37.346,-20.559,-30.141],[-36.609,-7.599,-34.343],
[-37.597,-6.161,-25.240],[-32.900,-5.146,-21.652],[-29.343,-7.609,-41.704],
[-28.698,-7.716,-22.760],[-26.502,-7.250,-40.460],[-39.787,-11.578,-23.761],
[-35.651,-9.784,-41.167],[-32.460,-11.799,-23.194],[-30.329,-9.235,-39.670],
[-27.014,-12.450,-24.814],[-37.498,-13.675,-23.281],[-32.450,-13.771,-37.617],
[-35.091,-15.921,-25.971],[-33.512,-12.354,-39.680],[-30.161,-13.508,-23.932],
[-22.829,-10.883,-39.573],[-34.090,-17.371,-23.808],[-33.420,-19.260,-27.454],
[-31.993,-13.382,-26.224],[-26.398,-16.301,-21.272],[-28.002,-13.489,-22.325],
[-35.244,-15.928,-21.759],[-32.323,-21.648,-26.314],[-31.263,-18.146,-23.224],
[-29.445,-19.958,-24.604],[-26.862,-19.539,-22.615],[-36.517,-18.091,-22.703],
[-33.625,-22.746,-24.320],[-32.300,-20.579,-23.331],[-31.862,-3.173,-39.836],
[-40.123,-10.025,-38.403],[-39.559,-4.676,-34.805],[-40.464,-6.112,-31.699],
[-22.118,-4.967,-28.856],[-22.464,-5.853,-26.081],[-25.346,-10.867,-21.061],
[-23.725,-10.650,-36.633],[-22.766,-13.643,-34.922],[-22.993,-10.932,-30.137],
[-24.284,-8.497,-26.867],[-23.691,-13.919,-39.084],[-23.805,-17.532,-39.134],
[-25.987,-14.482,-34.038],[-39.328,-13.156,-30.481],[-22.979,-13.620,-25.000],
[-22.093,-14.593,-22.432],[-24.328,-16.471,-41.462],[-24.940,-16.731,-34.964],
[-24.550,-18.302,-30.773],[-24.358,-16.951,-28.475],[-40.837,-16.986,-24.159],
[-19.831,-21.999,-23.378],[-24.393,-19.419,-35.211],[-40.502,-18.324,-29.113],
[-40.862,-15.530,-29.929],[-22.555,-19.641,-23.813],[-22.305,-19.501,-40.376],
[-39.162,-19.666,-36.183],[-25.423,-4.496,-33.170],[-40.488,-20.925,-27.506],
[-37.201,-7.328,-22.711],[-35.200,-8.218,-26.962],[-31.087,-6.486,-23.750],
[-28.319,-5.421,-24.022],[-24.391,-20.768,-25.966],[-37.676,-3.616,-26.554],
[-35.891,-5.168,-27.964],[-31.637,-4.023,-28.427],[-28.514,-5.990,-26.691],
[-27.252,-3.730,-26.223],[-38.931,-3.400,-31.305],[-35.390,-6.383,-32.065],
[-32.940,-5.878,-30.769],[-31.010,-18.793,-29.801],[-24.481,-21.896,-30.952],
[-23.000,-22.332,-33.249],[-37.216,-21.062,-34.467],[-39.086,-23.066,-38.494],
[-33.147,-21.882,-33.668],[-27.432,-22.626,-31.490],[-25.876,-22.314,-35.912],
[-39.988,-7.435,-37.483],[-36.721,-21.315,-39.259],[-31.410,-18.963,-36.902],
[-30.581,-4.699,-37.971],[-24.933,-4.329,-35.872],[-37.238,-20.718,-23.032],
[-33.932,-21.245,-21.255],[-38.195,-4.581,-41.079],[-26.676,-22.292,-21.723],
[-23.286,-5.891,-23.654],[-29.810,7.027,-33.425],[-30.089,5.563,-30.707],
[-36.849,5.536,-29.160],[-25.149,8.479,-31.184],[-36.405,9.453,-30.788],
[-39.537,4.512,-35.016],[-39.775,6.430,-33.137],[-29.259,14.647,-32.754],
[-30.609,8.381,-31.001],[-23.906,-2.038,-28.635],[-27.186,1.908,-29.863],
[-36.637,0.793,-32.800],[-28.105,1.200,-34.755],[-39.291,2.896,-32.216],
[-34.914,3.543,-32.566],[-21.764,2.335,-27.885],[-37.111,0.354,-29.141],
[-28.463,9.454,-33.194],[-33.852,7.922,-28.823],[-23.376,-0.057,-31.768],
[-36.760,2.711,-30.554],[-36.384,9.053,-28.178],[-40.201,8.600,-37.341],
[-26.872,8.231,-28.924],[-26.588,-3.502,-29.217],[-31.464,-3.966,-32.296],
[-27.472,12.240,-33.290],[-26.890,10.257,-35.359],[-25.756,9.784,-26.620],
[-31.374,5.591,-35.105],[-28.925,4.060,-39.730],[-22.008,6.033,-28.035],
[-25.794,7.061,-34.125],[-26.783,2.772,-27.076],[-34.451,6.132,-33.408],
[-33.632,8.422,-31.790],[-32.957,5.738,-30.841],[-28.332,10.174,-27.567],
[-33.498,9.203,-26.152],[-27.861,6.213,-38.341],[-36.154,4.408,-26.188],
[-30.819,12.467,-32.504],[-34.775,12.326,-24.045],[-24.238,-1.498,-38.552],
[-36.918,5.609,-34.474],[-22.820,10.380,-40.606],[-39.102,7.609,-27.681],
[-40.326,9.752,-29.641],[-22.018,11.270,-24.659],[-32.378,11.484,-25.554],
[-30.124,15.754,-37.058],[-36.934,15.062,-35.002],[-28.030,13.168,-37.543],
[-27.659,14.415,-35.059],[-24.998,3.851,-25.385],[-38.408,10.087,-26.452],
[-35.241,-2.771,-29.391],[-29.850,-3.015,-26.812],[-27.531,0.043,-25.739],
[-26.188,3.877,-39.329],[-31.800,0.890,-37.953],[-34.705,2.101,-28.714],
[-34.953,4.251,-35.778],[-28.207,4.283,-35.300],[-28.380,3.683,-25.350],
[-28.331,3.767,-31.718],[-38.318,1.230,-27.056],[-35.081,-0.339,-39.922],
[-40.804,0.950,-33.763],[-39.706,1.351,-36.713],[-29.231,3.513,-28.737],
[-34.413,-0.996,-32.513],[-37.398,3.014,-35.530],[-33.972,-0.441,-36.537],
[-25.911,-1.231,-31.083],[-34.955,3.653,-38.743],[-25.396,14.463,-24.420],
[-36.496,8.064,-24.076],[-31.964,3.056,-30.996],[-36.502,0.222,-35.338],
[-31.632,2.135,-28.248],[-32.379,7.415,-36.616],[-24.446,6.166,-29.283],
[-27.519,-0.612,-39.008],[-31.540,7.807,-26.928],[-27.874,2.710,-37.410],
[-37.330,5.883,-31.766],[-31.810,9.629,-38.282],[-40.152,5.945,-25.525],
[-31.895,6.055,-24.994],[-28.979,8.158,-25.908],[-21.521,3.137,-38.236],
[-22.863,8.712,-25.359],[-23.315,11.296,-35.244],[-27.392,6.065,-30.659],
[-35.911,9.556,-33.344],[-38.464,5.889,-37.047],[-33.784,9.425,-35.067],
[-31.233,12.665,-28.193],[-35.378,12.643,-34.808],[-29.593,7.869,-28.593],
[-26.954,8.988,-37.744],[-27.500,12.199,-25.799],[-23.220,14.573,-38.627],
[-39.517,0.521,-24.796],[-27.969,6.996,-35.807],[-37.623,11.224,-28.766],
[-34.197,10.653,-40.731],[-34.403,9.048,-37.652],[-37.089,8.720,-37.872],
[-30.708,-0.739,-27.868],[-28.299,-2.499,-36.309],[-30.809,12.270,-37.494],
[-31.039,2.474,-35.790],[-31.080,3.441,-25.888],[-30.664,0.153,-34.226],
[-33.629,11.793,-37.340],[-34.075,-3.268,-37.524],[-23.621,1.442,-25.714],
[-25.393,3.857,-31.053],[-36.341,11.140,-38.442],[-37.531,4.100,-38.984],
[-23.952,13.072,-28.800],[-36.850,-1.489,-31.155],[-38.084,10.806,-36.223],
[-37.923,13.189,-23.434],[-33.490,12.679,-21.519],[-29.075,11.636,-39.574],
[-27.212,9.666,-24.503],[-26.317,11.868,-40.608],[-39.317,7.456,-23.013],
[-36.364,8.687,-40.704],[-34.154,6.969,-23.443],[-29.090,9.090,-39.975],
[-27.129,6.106,-26.752],[-37.807,5.110,-24.004],[-34.517,6.309,-39.078],
[-33.688,2.932,-26.158],[-32.456,4.422,-40.204],[-29.775,7.233,-23.641],
[-24.973,6.178,-39.262],[-35.680,0.804,-25.626],[-33.504,-1.127,-27.951],
[-34.084,4.897,-28.272],[-27.191,2.911,-22.995],[-26.604,5.371,-23.989],
[-35.864,2.968,-23.126],[-32.686,-1.931,-24.678],[-32.268,0.567,-22.597],
[-29.475,1.157,-24.428],[-27.345,0.078,-22.834],[-37.241,0.697,-23.048],
[-33.181,-4.649,-25.689],[-32.844,-1.922,-21.340],[-32.522,14.838,-37.636],
[-37.897,11.592,-41.352],[-40.037,15.203,-34.610],[-40.767,11.627,-33.664],
[-23.152,13.500,-31.405],[-23.359,13.200,-26.244],[-24.817,9.619,-23.099],
[-23.104,7.697,-35.894],[-22.783,5.154,-34.689],[-22.366,8.057,-29.763],
[-24.562,10.105,-29.109],[-22.835,5.402,-37.551],[-23.465,2.565,-40.043],
[-26.061,2.618,-33.386],[-40.182,6.867,-30.440],[-24.421,6.847,-26.422],
[-23.474,5.096,-23.755],[-25.433,0.906,-40.514],[-25.067,1.128,-37.372],
[-24.475,1.342,-29.560],[-24.908,3.934,-28.241],[-41.774,1.840,-24.232],
[-21.439,-3.697,-21.635],[-25.320,-1.665,-35.264],[-39.207,0.014,-31.324],
[-39.248,3.466,-29.340],[-23.507,-1.576,-23.403],[-22.855,-0.290,-40.539],
[-40.118,-1.423,-35.526],[-25.792,14.317,-32.855],[-40.275,-1.629,-26.362],
[-37.368,10.530,-23.982],[-35.633,10.989,-26.335],[-30.628,13.501,-24.912],
[-28.216,14.699,-25.044],[-24.450,-3.038,-25.724],[-37.687,14.980,-27.799],
[-34.955,13.727,-28.617],[-30.982,14.492,-30.515],[-27.807,12.793,-28.510],
[-26.647,15.343,-28.091],[-39.096,13.386,-29.366],[-36.702,12.790,-30.751],
[-33.757,13.535,-32.497],[-30.541,-0.869,-31.317],[-23.433,-3.305,-31.024],
[-23.247,-2.560,-33.927],[-37.877,-1.604,-33.743],[-38.508,-4.273,-38.379],
[-32.573,-2.047,-34.337],[-27.486,-2.436,-32.701],[-26.323,-3.225,-38.127],
[-40.382,12.249,-36.116],[-37.176,-1.147,-38.453],[-31.402,-2.292,-36.785],
[-32.027,12.756,-39.952],[-24.410,14.332,-35.233],[-38.797,-2.258,-24.225],
[-36.207,-2.460,-22.277],[-37.848,13.187,-39.104],[-28.032,-2.881,-22.861],
[-24.959,11.771,-24.593],[-28.964,24.413,-33.381],[-29.540,24.350,-30.630],
[-36.989,24.147,-29.218],[-27.034,27.626,-30.694],[-37.809,27.826,-28.984],
[-40.204,21.522,-34.944],[-38.421,23.777,-34.563],[-29.597,33.073,-31.368],
[-32.395,29.281,-30.141],[-22.652,16.878,-29.936],[-27.439,20.927,-29.524],
[-34.266,18.230,-32.094],[-27.012,19.596,-34.753],[-39.463,21.877,-32.165],
[-33.175,23.646,-33.348],[-21.300,20.370,-27.809],[-36.849,18.261,-28.126],
[-29.248,27.053,-33.993],[-33.938,25.366,-28.918],[-25.025,18.384,-32.578],
[-36.753,18.781,-30.752],[-36.294,26.073,-26.942],[-39.047,28.121,-36.654],
[-24.839,29.347,-29.873],[-26.026,18.295,-28.742],[-32.991,15.843,-30.872],
[-25.208,30.552,-32.435],[-26.373,27.899,-34.018],[-25.162,27.367,-25.461],
[-31.767,25.320,-35.058],[-28.624,22.248,-39.351],[-22.983,22.743,-27.431],
[-26.143,25.387,-35.174],[-27.032,22.193,-26.594],[-35.821,25.086,-33.824],
[-35.680,28.834,-30.302],[-34.954,24.560,-31.401],[-28.534,30.766,-27.511],
[-33.837,28.134,-25.539],[-28.606,24.824,-37.130],[-36.500,23.276,-25.444],
[-31.141,31.204,-33.721],[-36.884,31.657,-24.034],[-23.700,17.122,-37.770],
[-35.683,23.563,-36.351],[-24.022,30.222,-39.464],[-40.005,26.802,-27.314],
[-39.519,29.559,-30.306],[-22.429,29.880,-25.174],[-32.082,29.229,-27.396],
[-31.062,33.587,-36.165],[-37.085,33.431,-34.715],[-27.958,32.324,-38.811],
[-27.204,30.689,-34.599],[-23.570,20.831,-25.459],[-36.837,28.944,-26.447],
[-35.675,15.613,-30.633],[-30.098,16.605,-28.532],[-27.445,19.531,-25.309],
[-26.182,22.532,-39.238],[-30.187,15.791,-39.953],[-34.947,20.668,-29.072],
[-34.236,21.274,-32.135],[-29.977,21.564,-34.572],[-30.919,21.410,-20.523],
[-27.956,21.465,-32.420],[-37.093,21.471,-27.832],[-34.524,17.554,-40.491],
[-40.554,19.209,-32.821],[-38.678,20.306,-38.218],[-30.080,22.172,-28.481],
[-31.875,20.248,-32.580],[-37.041,19.786,-33.627],[-34.118,19.583,-37.075],
[-27.878,18.232,-30.601],[-33.760,20.175,-39.877],[-24.339,34.402,-22.967],
[-36.486,26.705,-24.228],[-32.210,21.562,-29.868],[-35.780,17.866,-35.225],
[-32.473,20.509,-27.219],[-32.733,26.665,-37.752],[-25.073,24.993,-28.136],
[-27.677,18.134,-38.245],[-30.602,26.234,-25.763],[-27.912,20.612,-37.200],
[-37.935,26.540,-31.468],[-29.533,27.589,-36.826],[-39.107,24.119,-25.540],
[-32.705,21.203,-23.509],[-29.769,29.209,-24.693],[-21.955,22.510,-36.973],
[-21.837,27.511,-25.252],[-26.670,30.004,-37.195],[-28.293,23.988,-28.072],
[-34.042,29.462,-35.009],[-37.792,24.897,-40.102],[-31.963,28.133,-32.870],
[-33.110,32.282,-26.818],[-33.863,32.710,-34.005],[-28.671,26.636,-28.109],
[-27.644,27.652,-38.893],[-25.609,32.100,-25.320],[-22.248,31.800,-38.128],
[-37.655,19.513,-25.909],[-26.036,27.373,-36.980],[-37.840,31.742,-26.698],
[-33.613,27.974,-40.054],[-34.795,26.770,-35.750],[-37.230,26.587,-38.043],
[-31.035,18.184,-26.279],[-29.670,18.242,-36.438],[-31.590,29.141,-37.192],
[-33.670,21.015,-34.762],[-30.331,21.863,-25.871],[-30.842,17.609,-34.119],
[-34.135,30.206,-37.895],[-34.529,16.382,-37.586],[-23.964,18.458,-26.990],
[-26.625,23.877,-32.397],[-36.539,28.762,-40.063],[-36.164,20.952,-38.894],
[-24.526,32.100,-28.807],[-38.306,16.617,-30.056],[-39.208,28.672,-33.687],
[-40.245,33.773,-24.202],[-36.065,31.439,-21.442],[-31.772,30.213,-41.229],
[-28.479,27.800,-21.452],[-26.611,29.915,-40.099],[-38.776,26.005,-22.642],
[-34.356,25.126,-41.204],[-33.545,25.618,-23.573],[-30.805,26.732,-39.785],
[-26.940,25.445,-26.024],[-37.991,23.021,-22.714],[-33.981,24.648,-38.680],
[-33.942,22.858,-26.037],[-31.080,21.712,-40.048],[-29.092,25.009,-23.736],
[-25.190,25.381,-39.278],[-33.137,18.856,-24.845],[-33.771,18.281,-28.283],
[-31.905,24.171,-27.481],[-26.789,21.025,-21.691],[-26.399,22.117,-24.047],
[-35.751,21.941,-23.061],[-33.181,15.583,-28.129],[-30.863,18.926,-23.322],
[-29.036,17.261,-24.306],[-28.916,19.039,-21.714],[-37.637,19.540,-23.018],
[-32.945,14.200,-25.972],[-32.572,14.591,-23.216],[-32.009,34.441,-38.463],
[-39.086,28.824,-39.433],[-40.803,32.045,-35.684],[-39.240,31.376,-32.958],
[-22.053,32.196,-29.699],[-21.983,32.086,-26.772],[-23.805,28.254,-23.102],
[-22.535,28.329,-35.328],[-24.329,23.519,-35.345],[-23.124,27.192,-30.072],
[-23.138,28.240,-27.649],[-22.536,24.164,-39.524],[-23.529,21.271,-38.714],
[-23.804,20.927,-34.046],[-40.652,25.133,-30.669],[-24.140,24.044,-25.055],
[-22.350,23.552,-21.490],[-25.380,19.640,-39.302],[-24.842,19.677,-36.251],
[-24.359,20.777,-31.412],[-24.884,20.832,-27.977],[-39.686,20.807,-24.173],
[-21.958,15.808,-23.173],[-23.264,17.699,-34.946],[-39.331,19.682,-29.501],
[-39.432,22.365,-29.427],[-23.444,16.309,-25.282],[-22.878,17.369,-40.166],
[-39.855,18.044,-37.486],[-25.673,32.941,-34.087],[-39.708,17.647,-26.154],
[-38.058,29.233,-22.088],[-34.931,30.398,-27.156],[-32.078,30.781,-24.560],
[-29.534,32.655,-24.577],[-23.789,15.546,-27.861],[-37.111,35.682,-27.759],
[-35.735,32.867,-27.990],[-31.447,33.409,-29.035],[-28.301,33.178,-29.089],
[-26.537,34.570,-27.510],[-37.227,31.721,-30.548],[-36.539,29.919,-32.659],
[-33.057,31.182,-31.653],[-30.205,16.956,-31.421],[-24.928,15.718,-30.631],
[-23.488,16.269,-32.873],[-37.994,17.049,-33.087],[-38.762,15.026,-37.099],
[-32.013,15.109,-33.417],[-27.791,15.455,-30.386],[-25.880,15.647,-37.135],
[-38.597,31.171,-37.550],[-38.639,16.251,-39.751],[-32.954,16.789,-35.498],
[-31.253,31.675,-38.563],[-24.705,32.868,-36.806],[-37.923,16.949,-23.618],
[-34.708,16.030,-24.357],[-36.214,31.826,-39.172],[-27.532,15.357,-22.535],
[-24.458,30.794,-23.145],[-27.716,43.695,-33.497],[-30.947,43.097,-30.629],
[-37.741,41.979,-29.602],[-26.407,45.803,-32.795],[-36.728,44.482,-29.126],
[-39.068,40.512,-34.975],[-39.333,42.981,-34.081],[-31.258,49.569,-31.477],
[-31.587,46.280,-29.759],[-22.867,36.774,-28.312],[-26.981,38.084,-28.888],
[-35.662,38.272,-32.676],[-28.009,36.969,-34.094],[-39.071,41.046,-32.090],
[-36.556,41.323,-33.048],[-23.173,39.171,-26.890],[-35.896,37.445,-29.641],
[-28.885,45.762,-34.648],[-33.185,44.349,-28.612],[-25.086,37.528,-33.810],
[-36.293,39.930,-30.732],[-36.760,46.953,-28.142],[-38.986,47.708,-36.895],
[-25.482,46.293,-29.978],[-24.912,36.928,-30.338],[-32.011,35.859,-31.661],
[-25.722,48.194,-31.809],[-27.352,48.975,-34.193],[-24.281,46.280,-24.600],
[-32.125,42.363,-35.073],[-29.331,41.140,-39.200],[-22.343,43.451,-27.856],
[-27.026,41.588,-36.113],[-29.443,43.561,-27.419],[-36.139,44.315,-33.593],
[-33.462,47.260,-31.736],[-35.181,43.195,-30.954],[-27.532,48.095,-29.267],
[-33.518,46.483,-26.247],[-27.574,43.418,-39.778],[-35.513,42.566,-26.105],
[-30.637,48.370,-35.575],[-35.577,49.623,-24.653],[-24.355,35.145,-39.168],
[-36.600,43.165,-36.177],[-23.146,49.631,-40.066],[-38.999,45.460,-27.374],
[-39.312,48.362,-28.243],[-21.734,48.411,-25.180],[-29.787,47.825,-27.909],
[-29.923,51.797,-36.568],[-37.049,50.187,-35.467],[-26.403,49.766,-39.318],
[-29.401,50.859,-34.152],[-23.789,40.811,-24.748],[-38.082,48.195,-25.803],
[-33.779,34.820,-29.921],[-29.849,35.456,-25.781],[-26.910,38.766,-25.577],
[-25.356,41.222,-38.620],[-30.832,36.612,-40.319],[-33.689,40.790,-30.663],
[-34.334,40.550,-34.518],[-30.198,41.367,-36.732],[-30.343,40.785,-23.268],
[-29.637,41.086,-34.105],[-36.246,40.727,-27.868],[-34.092,37.139,-39.608],
[-40.254,38.575,-33.913],[-38.305,38.760,-39.008],[-29.323,40.597,-28.584],
[-33.037,38.010,-33.396],[-36.644,39.215,-35.136],[-35.295,38.733,-37.506],
[-28.896,35.768,-28.402],[-35.019,41.479,-39.161],[-25.633,50.840,-22.207],
[-36.311,44.906,-25.035],[-30.667,40.233,-30.950],[-35.323,36.934,-35.208],
[-32.098,39.614,-28.669],[-32.279,44.638,-36.603],[-24.283,43.833,-29.289],
[-28.130,38.931,-38.957],[-29.869,45.797,-25.953],[-28.333,39.294,-36.108],
[-37.815,43.513,-31.783],[-29.565,45.603,-37.234],[-40.680,41.157,-25.158],
[-33.161,42.344,-24.334],[-29.657,48.298,-25.136],[-23.021,40.775,-37.341],
[-22.574,46.182,-27.333],[-23.110,48.288,-36.385],[-27.763,42.892,-29.694],
[-34.761,46.849,-34.499],[-39.735,44.184,-38.968],[-32.570,45.399,-33.706],
[-31.861,49.300,-26.445],[-34.829,51.649,-35.392],[-26.552,44.380,-28.045],
[-26.663,46.755,-38.476],[-26.602,49.347,-26.936],[-22.841,52.104,-38.671],
[-39.382,38.514,-26.391],[-25.854,44.110,-36.735],[-36.982,49.855,-27.926],
[-32.819,46.300,-40.459],[-33.769,47.015,-36.926],[-36.944,46.003,-37.895],
[-29.171,38.136,-27.387],[-27.651,36.579,-36.912],[-31.780,48.277,-38.269],
[-31.080,38.607,-35.498],[-31.823,39.814,-25.714],[-30.782,36.061,-34.810],
[-34.088,49.129,-39.847],[-34.158,35.694,-37.217],[-24.983,37.247,-26.484],
[-27.657,41.342,-32.004],[-36.118,47.533,-40.838],[-35.979,39.291,-40.345],
[-25.577,50.313,-29.492],[-37.264,34.325,-30.305],[-38.286,48.488,-33.911],
[-37.909,50.858,-23.087],[-34.437,50.523,-21.515],[-31.198,48.443,-40.909],
[-27.748,46.680,-23.592],[-25.193,49.489,-41.682],[-38.202,44.773,-22.921],
[-34.916,45.268,-39.338],[-33.013,44.243,-22.079],[-29.582,46.596,-40.350],
[-27.002,44.718,-25.562],[-37.924,42.195,-23.796],[-32.819,42.624,-38.256],
[-33.223,41.576,-27.035],[-31.872,41.237,-40.319],[-28.766,44.191,-23.414],
[-25.077,44.830,-39.690],[-33.124,37.812,-24.482],[-34.874,37.261,-26.993],
[-30.794,43.074,-25.245],[-25.785,40.049,-22.971],[-26.607,41.674,-25.155],
[-35.046,40.662,-23.147],[-33.490,35.029,-26.473],[-30.400,37.951,-21.720],
[-29.605,38.338,-24.555],[-28.080,37.138,-22.987],[-36.837,37.970,-23.330],
[-32.950,33.371,-24.198],[-33.167,36.294,-22.128],[-31.882,51.903,-38.532],
[-38.774,46.368,-40.180],[-41.840,52.067,-34.287],[-38.728,51.087,-32.216],
[-22.795,50.704,-30.958],[-22.615,50.210,-28.421],[-21.871,45.656,-22.186],
[-23.050,44.463,-36.986],[-23.053,42.405,-34.458],[-22.679,46.107,-30.171],
[-25.275,46.870,-27.157],[-23.090,42.912,-39.053],[-22.564,40.172,-39.883],
[-24.772,40.316,-34.686],[-40.516,43.441,-31.255],[-24.259,43.495,-24.509],
[-21.505,41.954,-22.808],[-25.188,37.311,-42.538],[-25.726,38.872,-37.276],
[-24.130,40.990,-30.489],[-26.224,40.567,-27.758],[-41.663,39.582,-23.416],
[-21.582,34.255,-21.956],[-23.458,38.129,-36.205],[-38.454,38.331,-31.872],
[-40.391,40.655,-29.881],[-22.162,36.679,-24.270],[-22.336,37.680,-41.069],
[-40.609,37.434,-36.821],[-25.515,51.436,-35.002],[-39.703,35.761,-26.668],
[-37.318,47.806,-23.178],[-35.057,48.428,-27.003],[-30.585,50.717,-23.748],
[-28.402,51.139,-25.774],[-23.870,34.213,-27.126],[-39.011,51.480,-27.244],
[-36.444,51.769,-26.026],[-29.911,52.994,-28.432],[-29.566,50.314,-28.826],
[-26.329,52.158,-27.612],[-36.711,50.940,-30.467],[-34.884,49.291,-32.959],
[-34.051,50.397,-30.243],[-29.125,35.763,-31.458],[-26.889,31.176,-30.424],
[-23.839,33.858,-32.504],[-36.434,35.566,-33.092],[-39.736,34.042,-37.456],
[-31.459,33.852,-33.203],[-26.666,35.468,-32.132],[-28.265,34.238,-34.186],
[-38.590,50.533,-37.625],[-38.024,36.154,-39.409],[-30.613,36.875,-37.734],
[-29.537,50.746,-39.690],[-23.234,51.166,-36.104],[-38.384,36.175,-24.438],
[-36.295,35.118,-23.146],[-36.908,52.498,-39.149],[-27.373,34.517,-22.400],
[-23.303,49.149,-23.159],[-23.830,55.509,-30.576],[-27.757,58.460,-28.397],
[-35.502,55.550,-32.813],[-27.474,58.377,-35.399],[-40.589,58.196,-32.098],
[-22.558,55.513,-27.576],[-36.934,55.556,-29.773],[-25.939,55.055,-33.589],
[-37.398,58.291,-30.869],[-25.871,53.179,-30.067],[-32.526,52.705,-30.614],
[-25.861,53.361,-38.990],[-24.118,59.234,-21.800],[-34.812,53.812,-29.501],
[-30.777,52.869,-25.640],[-27.536,55.749,-26.435],[-31.649,56.023,-41.363],
[-35.348,57.795,-28.008],[-34.732,-59.463,-34.207],[-30.037,58.917,-33.257],
[-37.452,-59.597,-27.385],[-34.483,57.211,-41.690],[-40.385,54.931,-34.501],
[-37.373,57.349,-37.593],[-32.961,56.837,-33.192],[-36.763,57.600,-33.961],
[-33.752,57.754,-36.005],[-28.087,55.155,-29.199],[-35.310,58.896,-38.832],
[-30.344,-59.598,-29.821],[-34.887,54.277,-37.015],[-32.790,57.955,-30.407],
[-28.438,56.154,-38.883],[-29.056,59.061,-38.243],[-22.619,58.317,-37.108],
[-38.863,56.193,-26.437],[-30.132,55.413,-27.363],[-28.109,54.170,-36.859],
[-31.091,57.388,-36.914],[-31.102,57.234,-25.779],[-28.726,55.344,-34.519],
[-34.315,56.305,-38.544],[-24.943,55.668,-26.418],[-39.572,58.495,-40.191],
[-37.986,53.475,-31.743],[-33.828,56.934,-24.667],[-32.975,55.932,-28.301],
[-30.668,58.066,-22.528],[-36.801,57.409,-23.214],[-33.173,51.798,-26.989],
[-31.936,55.659,-22.795],[-29.307,55.846,-24.217],[-26.676,56.047,-21.631],
[-40.970,55.991,-22.478],[-32.993,51.912,-23.795],[-33.762,53.687,-21.812],
[-22.743,58.257,-39.702],[-25.846,56.337,-40.462],[-24.568,56.927,-35.302],
[-26.081,57.468,-31.045],[-24.093,57.451,-28.601],[-41.105,57.244,-25.333],
[-23.022,51.705,-22.563],[-24.462,53.992,-35.584],[-39.135,55.710,-32.278],
[-39.765,57.263,-29.413],[-21.881,55.619,-24.746],[-20.645,53.632,-38.484],
[-43.073,56.224,-37.774],[-39.910,53.259,-29.184],[-23.573,53.057,-26.663],
[-29.593,55.038,-31.825],[-26.690,50.830,-31.970],[-24.533,52.478,-32.219],
[-39.174,52.486,-34.601],[-38.614,53.321,-37.002],[-33.335,54.208,-34.445],
[-28.073,53.041,-33.076],[-27.544,51.552,-37.814],[-36.561,55.109,-39.781],
[-31.738,54.503,-37.823],[-36.678,54.603,-25.355],[-35.550,52.235,-23.561],
[-27.751,52.075,-23.091],[-27.512,-50.146,-14.480],[-30.238,-48.662,-12.498],
[-36.836,-50.646,-11.778],[-25.167,-47.262,-11.741],[-36.093,-46.259,-12.202],
[-39.526,-53.504,-16.713],[-39.933,-49.523,-13.930],[-27.144,-43.307,-12.366],
[-32.898,-48.216,-12.697],[-23.791,-58.088,-11.849],[-26.856,-53.556,-10.429],
[-34.933,-54.442,-14.306],[-26.552,-54.866,-16.045],[-38.605,-52.073,-13.807],
[-33.189,-51.620,-14.115],[-21.961,-53.314,-7.649],[-36.521,-55.727,-11.660],
[-30.861,-45.793,-16.841],[-33.996,-48.356,-9.362],[-25.695,-56.247,-13.803],
[-37.322,-53.275,-11.373],[-36.946,-46.929,-9.648],[-40.252,-45.579,-17.410],
[-27.234,-45.827,-9.076],[-24.934,-56.769,-9.316],[-33.664,59.093,-12.905],
[-27.639,-46.082,-13.195],[-27.020,-46.778,-16.321],[-25.321,-46.949,-6.780],
[-30.413,-49.110,-16.238],[-26.848,-52.020,-21.112],[-22.175,-50.986,-9.646],
[-26.301,-50.603,-17.414],[-29.240,-51.008,-7.282],[-34.131,-49.301,-15.043],
[-32.395,-45.187,-13.094],[-33.855,-50.586,-11.610],[-29.404,-44.577,-10.532],
[-34.892,-46.074,-6.215],[-28.654,-49.969,-18.635],[-36.568,-50.102,-7.955],
[-29.388,-43.762,-15.229],[-37.856,-42.744,-7.067],[-23.997,-58.471,-19.153],
[-35.904,-50.889,-17.395],[-22.561,-43.417,-23.362],[-39.643,-47.348,-9.246],
[-39.282,-44.431,-12.457],[-22.081,-42.622,-4.804],[-31.899,-45.509,-10.300],
[-29.428,-41.277,-17.997],[-36.139,-41.391,-13.978],[-27.581,-43.041,-20.217],
[-27.089,-42.557,-17.409],[-23.799,-53.111,-4.777],[-37.535,-45.551,-7.028],
[-34.915,-57.837,-10.396],[-30.098,58.929,-8.693],[-28.051,-53.775,-7.706],
[-26.974,-49.096,-20.645],[-29.997,-55.375,-19.122],[-34.440,-53.085,-11.116],
[-32.647,-54.730,-16.179],[-28.919,-51.579,-16.576],[-29.721,-53.084,-5.364],
[-27.408,-53.678,-13.672],[-37.936,-52.730,-8.887],[-33.523,-55.696,-23.051],
[-39.933,-55.924,-15.668],[-40.430,-55.454,-18.531],[-29.890,-52.522,-9.611],
[-30.971,-54.869,-13.645],[-37.019,-54.129,-16.001],[-35.351,-53.829,-18.310],
[-29.239,-57.067,-11.133],[-34.479,-52.606,-20.987],[-25.828,-40.569,-5.020],
[-37.656,-48.213,-6.403],[-31.181,-51.683,-12.251],[-35.657,-56.711,-16.373],
[-30.771,-54.758,-10.869],[-31.449,-49.716,-18.445],[-24.075,-49.500,-11.154],
[-28.309,-57.303,-20.233],[-31.040,-45.120,-7.729],[-27.988,-53.536,-18.615],
[-37.157,-48.461,-13.903],[-28.854,-44.578,-18.338],[-39.327,-50.554,-7.144],
[-32.137,-52.477,-6.546],[-28.261,-44.653,-6.920],[-21.863,-53.393,-18.661],
[-22.260,-46.014,-7.200],[-25.087,-43.891,-18.879],[-28.376,-51.576,-12.333],
[-35.445,-44.800,-14.276],[-37.496,-51.773,-20.536],[-33.112,-47.413,-16.484],
[-34.027,-43.503,-10.067],[-34.052,-42.696,-15.590],[-30.368,-47.342,-9.124],
[-27.065,-46.031,-20.302],[-26.427,-40.787,-7.830],[-23.216,-41.696,-19.732],
[-36.087,-55.435,-7.040],[-26.109,-48.082,-18.528],[-39.359,-44.275,-9.388],
[-33.953,-45.679,-22.608],[-33.201,-45.245,-19.441],[-37.818,-47.087,-18.321],
[-29.422,-56.741,-7.155],[-28.476,-56.723,-17.043],[-31.422,-47.108,-19.321],
[-30.082,-54.359,-16.537],[-31.349,-54.736,-7.801],[-31.105,-58.737,-16.869],
[-36.457,-42.944,-20.873],[-35.646,-56.513,-19.445],[-24.588,-54.716,-7.243],
[-25.240,-52.123,-11.688],[-36.112,-45.233,-19.358],[-39.063,-54.014,-22.835],
[-23.701,-43.197,-9.520],[-37.743,-58.026,-11.174],[-37.198,-42.911,-16.267],
[-39.270,-41.918,-5.076],[-35.208,-42.708,-4.141],[-29.637,-44.740,-21.835],
[-25.355,-48.975,-0.635],[-24.480,-45.211,-22.508],[-39.058,-48.444,-3.580],
[-34.449,-47.887,-21.184],[-33.931,-49.011,-5.736],[-31.433,-47.011,-22.211],
[-27.305,-50.474,-9.582],[-37.651,-50.800,-4.704],[-34.156,-49.535,-19.018],
[-34.458,-51.794,-7.752],[-31.967,-54.023,-20.322],[-28.619,-49.356,-4.886],
[-24.405,-49.969,-21.166],[-33.502,-55.212,-6.346],[-34.147,-55.183,-9.145],
[-32.208,-50.410,-8.732],[-28.397,-53.422,-3.183],[-26.431,-49.869,-6.949],
[-35.767,-52.932,-5.061],[-33.299,-59.276,-8.645],[-32.125,-55.220,-3.029],
[-30.142,-56.300,-4.731],[-27.055,-55.337,-5.629],[-37.868,-55.846,-5.037],
[-33.529,-58.726,-5.894],[-31.690,-58.884,-3.628],[-34.107,-40.537,-18.708],
[-40.915,-46.181,-20.194],[-40.079,-40.477,-16.478],[-39.782,-41.547,-13.803],
[-22.303,-42.425,-11.954],[-23.668,-41.520,-7.061],[-24.699,-46.655,-3.940],
[-23.612,-49.206,-18.391],[-24.474,-49.869,-15.726],[-22.594,-46.953,-12.235],
[-24.158,-45.790,-9.094],[-21.863,-50.695,-20.259],[-23.928,-54.631,-20.965],
[-25.105,-52.817,-14.870],[-39.296,-49.707,-11.103],[-24.288,-50.245,-5.425],
[-21.813,-51.454,-4.997],[-25.963,-57.167,-21.973],[-25.215,-53.856,-18.340],
[-24.092,-54.599,-11.871],[-24.958,-52.366,-9.079],[-38.764,-53.079,-5.964],
[-21.398,-60.008,-4.259],[-25.339,-57.353,-16.661],[-39.288,-55.359,-13.161],
[-40.284,-52.909,-11.625],[-24.057,-58.071,-4.789],[-22.219,-56.732,-20.732],
[-38.967,-58.069,-17.870],[-25.272,-41.543,-14.810],[-39.002,-56.437,-9.242],
[-37.922,-44.264,-4.667],[-35.905,-44.057,-8.342],[-32.070,-43.895,-5.632],
[-28.649,-41.834,-6.681],[-23.506,-58.880,-7.958],[-37.717,-39.474,-7.098],
[-35.357,-40.064,-8.750],[-29.789,-41.688,-10.320],[-27.083,-42.522,-9.667],
[-24.505,-40.760,-9.626],[-38.663,-42.162,-11.009],[-36.476,-43.691,-11.668],
[-33.373,-42.449,-12.920],[-30.613,-57.601,-14.394],[-25.919,59.002,-12.723],
[-23.985,-59.249,-14.776],[-37.825,-58.038,-14.358],[-39.437,-59.774,-20.007],
[-30.847,57.795,-15.299],[-28.175,-57.946,-13.625],[-27.090,-59.134,-18.152],
[-38.495,-43.916,-18.884],[-37.151,-54.740,-20.737],[-31.818,-59.155,-19.845],
[-31.776,-43.082,-20.698],[-24.614,-44.357,-16.460],[-38.100,-57.986,-7.027],
[-36.220,-58.597,-4.864],[-38.759,-41.001,-20.208],[-27.602,-59.519,-5.636],
[-25.765,-44.203,-5.763],[-29.379,-32.574,-13.853],[-30.964,-30.327,-12.063],
[-36.102,-32.602,-9.989],[-26.972,-27.760,-13.282],[-36.697,-27.295,-11.268],
[-38.960,-32.226,-15.912],[-40.503,-30.733,-14.193],[-27.531,-24.010,-12.591],
[-33.470,-28.441,-11.748],[-23.130,-38.318,-10.823],[-27.708,-36.428,-9.666],
[-35.901,-35.923,-13.772],[-27.924,-38.847,-15.327],[-38.949,-34.264,-13.703],
[-31.931,-32.529,-13.680],[-21.681,-36.265,-9.054],[-36.576,-37.820,-11.010],
[-29.137,-28.345,-15.484],[-33.144,-28.973,-8.994],[-25.789,-36.478,-11.611],
[-37.253,-34.900,-11.496],[-35.760,-28.013,-8.508],[-38.098,-29.424,-17.510],
[-26.933,-26.049,-11.091],[-25.389,-38.192,-8.751],[-31.648,-40.038,-12.289],
[-24.765,-26.518,-14.070],[-27.279,-26.781,-16.268],[-26.235,-28.816,-6.445],
[-31.394,-31.766,-16.829],[-29.274,-33.886,-19.867],[-23.217,-31.993,-8.789],
[-26.113,-31.879,-15.392],[-29.213,-32.276,-7.389],[-34.089,-31.235,-14.659],
[-34.481,-26.573,-13.695],[-34.686,-30.942,-11.956],[-29.309,-26.862,-10.381],
[-33.410,-26.673,-6.023],[-28.406,-31.375,-18.942],[-36.914,-32.684,-6.683],
[-29.763,-25.273,-15.050],[-35.939,-24.345,-5.199],[-24.137,-39.130,-18.129],
[-35.872,-31.797,-16.752],[-22.767,-25.833,-21.077],[-40.575,-29.244,-8.318],
[-40.497,-26.717,-11.057],[-20.853,-25.851,-6.416],[-30.945,-27.279,-7.059],
[-29.713,-21.845,-17.729],[-36.422,-22.062,-15.014],[-26.934,-23.676,-19.486],
[-27.530,-23.352,-16.581],[-25.103,-34.864,-4.947],[-37.410,-27.514,-5.512],
[-34.216,-39.679,-11.503],[-30.372,-40.042,-8.209],[-27.587,-35.409,-5.543],
[-25.811,-31.827,-20.326],[-31.367,-37.739,-18.776],[-33.246,-33.110,-10.329],
[-33.876,-33.949,-15.149],[-29.417,-34.108,-16.967],[-29.849,-34.538,-4.470],
[-27.397,-34.354,-14.864],[-36.813,-34.639,-8.505],[-34.332,-36.393,-22.802],
[-39.984,-38.576,-14.026],[-38.164,-37.813,-19.554],[-30.024,-33.647,-9.696],
[-33.682,-37.669,-13.111],[-37.040,-34.212,-15.706],[-34.305,-35.613,-17.531],
[-29.028,-38.656,-10.480],[-35.557,-34.429,-19.641],[-25.322,-23.142,-4.926],
[-36.764,-30.406,-5.280],[-32.191,-35.425,-13.419],[-36.345,-36.965,-16.476],
[-31.845,-35.818,-10.546],[-31.764,-29.125,-19.327],[-24.548,-29.961,-10.631],
[-27.119,-37.561,-20.109],[-30.277,-29.830,-6.955],[-26.791,-34.944,-18.847],
[-38.353,-29.228,-12.173],[-29.364,-28.497,-18.100],[-39.428,-31.499,-7.567],
[-32.729,-33.356,-5.075],[-28.901,-25.917,-6.027],[-22.153,-34.747,-18.518],
[-25.343,-26.539,-7.310],[-22.741,-27.284,-17.687],[-27.741,-32.669,-11.528],
[-33.656,-26.482,-16.534],[-38.751,-31.951,-19.244],[-32.177,-28.560,-16.517],
[-32.755,-24.574,-9.046],[-34.325,-23.978,-14.958],[-28.322,-29.099,-8.999],
[-26.893,-28.319,-19.509],[-26.863,-24.123,-7.169],[-21.977,-23.286,-20.638],
[-39.049,-37.055,-7.402],[-25.024,-30.525,-17.572],[-38.787,-26.223,-8.182],
[-33.060,-27.222,-21.400],[-34.332,-27.829,-18.613],[-36.156,-31.578,-19.441],
[-29.420,-35.500,-7.613],[-28.739,-38.535,-18.057],[-31.366,-26.545,-18.364],
[-31.695,-35.438,-16.298],[-31.895,-35.546,-6.688],[-31.202,-37.916,-15.754],
[-33.932,-25.149,-19.556],[-33.284,-38.574,-20.499],[-24.848,-37.467,-6.253],
[-25.275,-33.406,-10.539],[-37.322,-27.827,-19.980],[-38.699,-34.926,-20.614],
[-25.585,-23.504,-10.707],[-36.527,-40.564,-11.221],[-37.599,-24.804,-16.331],
[-39.123,-23.526,-4.132],[-34.330,-24.740,-2.526],[-28.833,-25.506,-22.610],
[-27.999,-27.920,-4.453],[-25.390,-25.535,-20.616],[-39.554,-30.061,-4.378],
[-34.399,-29.592,-20.737],[-32.303,-29.663,-5.299],[-31.771,-30.290,-21.905],
[-26.740,-31.839,-8.972],[-38.096,-32.390,-4.145],[-32.842,-31.539,-19.136],
[-34.300,-33.992,-7.624],[-31.395,-32.643,-20.914],[-28.694,-30.897,-4.535],
[-23.446,-30.748,-21.667],[-34.225,-36.551,-6.353],[-34.334,-36.551,-9.085],
[-31.886,-32.687,-7.561],[-26.958,-35.169,-3.065],[-27.307,-32.846,-5.592],
[-35.648,-34.012,-4.646],[-33.076,-38.740,-7.334],[-32.251,-36.272,-3.771],
[-29.821,-37.283,-5.200],[-26.824,-37.555,-4.231],[-36.317,-36.538,-4.663],
[-32.677,-41.270,-6.192],[-32.853,-40.383,-3.368],[-32.872,-22.376,-18.732],
[-39.002,-28.923,-22.057],[-40.209,-23.800,-17.062],[-38.804,-23.464,-13.912],
[-22.651,-24.165,-11.048],[-22.473,-24.309,-8.160],[-24.386,-27.907,-4.445],
[-23.356,-28.642,-20.096],[-22.468,-31.843,-17.369],[-22.984,-28.570,-8.773],
[-24.329,-26.169,-10.125],[-21.645,-32.238,-19.962],[-23.496,-34.215,-21.327],
[-24.852,-34.846,-15.644],[-40.497,-30.443,-11.040],[-24.638,-31.312,-5.681],
[-23.155,-32.678,-3.850],[-23.674,-37.766,-22.670],[-26.535,-36.857,-16.837],
[-24.499,-35.668,-9.308],[-26.635,-34.665,-7.872],[-41.194,-35.158,-4.592],
[-21.613,-40.084,-5.297],[-24.948,-38.565,-15.280],[-39.193,-36.504,-10.752],
[-39.258,-32.840,-10.931],[-22.549,-37.648,-4.903],[-22.429,-36.796,-20.214],
[-41.505,-36.341,-16.691],[-25.005,-24.018,-14.390],[-39.066,-39.668,-9.953],
[-37.068,-25.754,-3.053],[-35.538,-25.059,-9.803],[-30.611,-23.281,-5.432],
[-27.663,-21.193,-5.978],[-22.689,-38.618,-7.997],[-37.927,-22.118,-6.269],
[-37.128,-21.937,-9.389],[-29.587,-21.250,-10.089],[-29.482,-23.881,-9.376],
[-27.108,-21.973,-8.677],[-38.638,-24.469,-11.113],[-36.405,-24.717,-12.599],
[-33.742,-24.736,-11.890],[-29.764,-36.911,-13.413],[-24.842,-41.232,-12.159],
[-23.117,-40.307,-14.165],[-37.273,-38.979,-14.104],[-40.525,-38.683,-18.917],
[-33.255,-39.817,-14.996],[-27.433,-40.413,-13.233],[-26.287,-40.671,-18.906],
[-38.543,-25.057,-19.021],[-35.851,-38.223,-20.972],[-33.620,-38.059,-17.474],
[-30.541,-24.057,-19.421],[-22.451,-24.041,-17.145],[-37.368,-39.854,-4.163],
[-35.020,-41.237,-1.924],[-37.719,-22.735,-20.831],[-28.127,-40.026,-2.935],
[-23.410,-25.269,-4.459],[-29.356,-13.850,-14.592],[-30.123,-12.428,-9.944],
[-36.798,-12.669,-9.622],[-25.937,-9.707,-13.535],[-35.508,-9.456,-14.070],
[-38.865,-14.186,-16.422],[-39.424,-12.697,-14.059],[-28.751,-6.080,-12.599],
[-29.905,-10.110,-11.337],[-23.198,-19.401,-9.052],[-27.702,-16.140,-10.797],
[-34.766,-17.463,-13.439],[-27.031,-18.904,-15.201],[-38.800,-15.462,-13.696],
[-33.570,-14.923,-14.060],[-22.237,-16.711,-9.458],[-37.512,-18.237,-9.190],
[-30.287,-9.595,-16.231],[-34.605,-11.045,-10.646],[-25.679,-19.047,-12.996],
[-36.734,-15.981,-12.110],[-35.620,-10.670,-8.071],[-39.981,-9.125,-18.182],
[-25.564,-8.318,-11.198],[-25.778,-20.071,-10.437],[-32.930,-21.725,-11.085],
[-26.762,-7.512,-14.930],[-26.172,-9.663,-16.336],[-23.839,-11.019,-7.622],
[-30.915,-12.115,-16.591],[-29.440,-16.378,-18.678],[-22.775,-13.203,-9.329],
[-27.037,-12.623,-16.766],[-27.595,-14.147,-8.096],[-35.489,-12.238,-14.734],
[-34.870,-8.468,-11.299],[-35.112,-13.174,-12.114],[-28.446,-7.871,-10.771],
[-34.923,-9.159,-6.053],[-28.466,-13.156,-19.487],[-36.265,-13.737,-7.078],
[-29.164,-6.617,-16.303],[-35.634,-6.919,-4.281],[-25.534,-21.359,-19.591],
[-36.209,-13.061,-17.554],[-22.228,-9.861,-20.168],[-39.439,-10.559,-8.326],
[-41.236,-7.695,-10.999],[-21.641,-6.605,-5.744],[-31.494,-8.056,-7.109],
[-29.160,-3.901,-16.130],[-36.586,-4.647,-15.019],[-27.675,-6.099,-20.820],
[-26.973,-5.612,-17.737],[-24.584,-15.264,-5.409],[-37.344,-8.444,-6.741],
[-35.331,-20.161,-10.214],[-30.512,-20.005,-7.407],[-26.950,-16.660,-8.100],
[-25.049,-13.967,-22.077],[-31.053,-18.457,-19.712],[-34.877,-15.358,-10.360],
[-34.775,-15.449,-16.461],[-29.376,-16.349,-15.946],[-30.564,-17.685,-4.786],
[-27.742,-16.685,-13.677],[-37.327,-15.461,-8.903],[-33.567,-18.487,-21.162],
[-39.332,-18.470,-15.289],[-39.019,-17.438,-18.000],[-29.922,-15.052,-9.530],
[-32.949,-19.074,-14.365],[-37.193,-16.415,-15.930],[-33.608,-17.342,-18.488],
[-27.976,-18.696,-11.102],[-35.563,-15.615,-19.065],[-26.445,-4.882,-4.978],
[-36.551,-11.482,-5.575],[-31.153,-15.305,-12.622],[-36.702,-19.009,-15.427],
[-32.369,-15.736,-9.370],[-33.325,-10.431,-17.550],[-25.800,-12.000,-9.230],
[-28.298,-20.140,-19.974],[-32.173,-11.013,-8.089],[-26.928,-15.558,-17.769],
[-37.374,-11.576,-12.752],[-30.120,-10.988,-19.119],[-39.101,-13.132,-7.000],
[-31.285,-12.938,-5.914],[-28.913,-7.105,-7.039],[-22.110,-15.893,-18.966],
[-21.651,-10.146,-5.951],[-24.469,-8.585,-19.276],[-27.516,-13.579,-10.651],
[-33.703,-7.937,-16.572],[-39.474,-13.689,-20.504],[-32.983,-10.075,-14.757],
[-33.508,-6.009,-9.753],[-33.828,-5.423,-15.497],[-27.938,-10.474,-9.398],
[-28.457,-9.997,-20.781],[-26.160,-7.387,-6.661],[-22.383,-5.752,-20.048],
[-37.757,-18.715,-6.074],[-26.831,-10.635,-18.752],[-39.046,-7.820,-8.714],
[-33.105,-9.389,-22.434],[-32.834,-8.839,-19.734],[-38.494,-11.301,-17.449],
[-29.758,-17.729,-8.854],[-29.250,-19.125,-17.750],[-31.066,-7.208,-18.212],
[-31.589,-17.755,-16.305],[-30.644,-15.615,-7.182],[-29.980,-19.477,-15.210],
[-33.519,-6.112,-18.734],[-36.553,-18.206,-19.428],[-24.348,-17.081,-7.453],
[-25.685,-14.732,-12.855],[-35.773,-6.816,-20.419],[-38.059,-16.093,-20.493],
[-26.431,-5.538,-10.985],[-37.122,-21.490,-12.341],[-38.198,-7.455,-14.575],
[-39.600,-4.983,-6.291],[-33.307,-5.394,-3.582],[-31.128,-7.630,-21.404],
[-28.514,-7.701,-4.203],[-25.392,-7.539,-22.085],[-38.705,-11.663,-3.699],
[-35.058,-12.240,-22.213],[-33.552,-11.470,-4.103],[-28.540,-10.407,-23.394],
[-26.525,-11.744,-6.602],[-38.044,-14.718,-4.634],[-31.058,-14.479,-19.731],
[-33.231,-14.875,-7.026],[-32.942,-13.974,-21.648],[-28.470,-11.506,-4.783],
[-23.470,-12.198,-22.731],[-35.259,-16.730,-7.473],[-33.720,-18.100,-9.793],
[-32.888,-13.095,-10.294],[-26.285,-16.638,-3.568],[-27.325,-14.301,-4.972],
[-35.152,-15.095,-4.421],[-32.028,-22.276,-7.782],[-31.580,-17.621,-2.350],
[-28.543,-16.784,-6.013],[-27.605,-19.052,-4.383],[-37.205,-17.766,-3.670],
[-33.191,-24.049,-6.080],[-32.329,-20.039,-4.345],[-33.188,-3.335,-18.583],
[-38.295,-9.018,-20.395],[-40.496,-4.517,-16.638],[-38.613,-5.225,-13.184],
[-21.875,-5.400,-10.635],[-22.698,-5.372,-8.126],[-25.935,-10.471,-3.944],
[-23.792,-11.398,-18.376],[-23.020,-13.466,-16.609],[-23.007,-10.317,-10.673],
[-24.085,-7.470,-8.896],[-24.266,-14.208,-19.135],[-23.757,-17.229,-21.331],
[-25.112,-14.736,-15.524],[-39.060,-13.215,-11.526],[-23.708,-12.576,-5.423],
[-21.791,-14.362,-3.781],[-26.010,-18.708,-20.073],[-24.597,-17.063,-17.025],
[-24.806,-17.296,-10.707],[-25.164,-14.712,-9.500],[-39.657,-17.023,-5.049],
[-20.790,-21.365,-2.915],[-26.869,-20.242,-17.471],[-37.172,-18.722,-12.291],
[-41.442,-14.639,-11.282],[-22.031,-19.602,-4.929],[-22.351,-18.699,-18.927],
[-41.685,-18.197,-16.785],[-24.912,-5.757,-13.955],[-39.108,-20.242,-8.221],
[-37.996,-8.604,-4.089],[-36.063,-6.707,-9.640],[-31.541,-5.675,-5.780],
[-28.198,-2.995,-6.638],[-23.771,-21.066,-6.651],[-36.763,-2.597,-6.377],
[-37.449,-3.781,-10.105],[-31.706,-4.227,-10.427],[-28.475,-5.412,-9.071],
[-27.069,-3.207,-9.643],[-38.766,-6.440,-10.951],[-35.992,-6.602,-13.089],
[-32.680,-5.720,-12.495],[-30.541,-20.056,-12.219],[-23.825,-21.895,-12.684],
[-23.534,-21.277,-15.282],[-39.591,-20.904,-16.543],[-37.215,-20.589,-17.624],
[-34.433,-21.224,-13.606],[-27.416,-21.118,-12.760],[-24.819,-22.446,-17.242],
[-38.381,-6.558,-18.569],[-36.173,-20.489,-20.290],[-31.962,-20.390,-17.151],
[-30.997,-4.758,-19.452],[-24.162,-6.311,-17.357],[-37.269,-20.555,-4.034],
[-34.988,-21.437,-5.148],[-37.643,-4.159,-20.495],[-27.670,-21.309,-2.873],
[-24.093,-5.828,-5.792],[-27.822,6.828,-16.710],[-30.085,5.082,-13.965],
[-37.140,5.557,-11.874],[-26.482,8.221,-14.520],[-35.733,9.567,-14.628],
[-39.190,3.077,-16.277],[-39.847,5.575,-15.465],[-28.294,13.543,-13.076],
[-30.276,8.796,-12.588],[-23.082,0.214,-9.774],[-27.983,1.137,-11.505],
[-36.780,3.168,-14.390],[-26.148,-0.067,-16.332],[-38.852,1.411,-13.793],
[-33.817,3.307,-14.519],[-21.296,2.208,-9.107],[-35.970,0.996,-9.996],
[-29.785,8.223,-18.002],[-34.455,7.016,-9.715],[-25.239,-0.570,-13.567],
[-37.380,2.741,-11.504],[-36.565,8.130,-8.070],[-38.990,10.347,-18.626],
[-27.275,8.493,-9.997],[-24.905,-1.427,-10.110],[-32.021,-2.722,-12.712],
[-26.534,11.418,-13.659],[-26.359,9.135,-17.290],[-25.853,8.813,-7.464],
[-31.221,4.879,-17.902],[-30.615,4.278,-23.107],[-24.580,3.912,-9.578],
[-25.456,5.336,-17.585],[-26.669,3.653,-7.473],[-34.958,5.940,-14.689],
[-34.687,9.698,-11.749],[-33.455,5.137,-12.404],[-29.958,11.503,-10.869],
[-33.434,9.716,-7.577],[-29.460,3.871,-19.827],[-36.438,5.141,-7.256],
[-31.205,11.761,-16.751],[-36.084,9.976,-5.983],[-24.515,-4.059,-19.119],
[-36.576,5.109,-16.769],[-22.527,10.085,-20.799],[-38.413,8.058,-10.164],
[-39.991,9.949,-11.790],[-21.815,9.824,-3.636],[-32.035,11.979,-7.443],
[-30.110,14.654,-16.742],[-37.201,14.349,-14.931],[-27.166,13.333,-20.523],
[-27.647,12.501,-17.360],[-25.175,4.292,-5.448],[-38.646,9.397,-5.950],
[-34.194,-3.414,-11.060],[-30.141,-3.465,-8.575],[-28.288,1.278,-7.956],
[-25.117,5.792,-21.390],[-29.926,1.367,-21.610],[-34.507,2.462,-11.699],
[-33.603,2.195,-16.782],[-29.874,2.437,-16.513],[-30.009,3.931,-5.295],
[-27.805,3.301,-15.179],[-37.398,2.409,-6.853],[-34.766,0.688,-22.014],
[-40.659,-0.177,-15.172],[-39.031,2.035,-19.505],[-30.076,3.400,-10.110],
[-33.039,-0.210,-14.941],[-37.172,0.977,-15.757],[-34.165,1.205,-19.212],
[-28.614,-1.125,-10.646],[-34.565,4.050,-18.325],[-23.794,14.996,-4.878],
[-36.755,7.315,-5.425],[-31.541,3.100,-12.462],[-35.133,-0.733,-17.505],
[-32.475,2.517,-9.861],[-31.917,8.098,-19.814],[-24.159,7.134,-9.918],
[-27.422,-1.185,-20.405],[-30.551,9.412,-5.918],[-27.614,1.819,-20.349],
[-37.467,6.764,-14.280],[-27.521,8.652,-20.034],[-39.271,5.547,-7.758],
[-32.662,5.145,-5.989],[-28.375,10.574,-8.508],[-22.227,3.309,-18.926],
[-23.284,7.964,-6.285],[-24.866,10.093,-19.458],[-25.796,5.668,-11.701],
[-34.732,11.017,-16.577],[-36.766,5.885,-21.410],[-33.426,7.980,-15.364],
[-32.075,10.130,-9.880],[-34.917,13.903,-16.435],[-29.172,6.502,-9.638],
[-26.720,7.741,-22.332],[-26.546,12.804,-7.091],[-23.867,12.908,-19.678],
[-38.528,1.119,-8.884],[-27.097,5.934,-19.397],[-37.382,10.661,-11.802],
[-32.334,9.798,-22.347],[-34.136,9.614,-19.324],[-37.293,8.134,-19.445],
[-30.744,-0.112,-8.336],[-28.485,-0.370,-17.569],[-31.601,11.140,-19.607],
[-31.646,0.318,-17.186],[-31.003,2.636,-7.457],[-29.856,-1.359,-15.318],
[-34.373,12.312,-18.836],[-35.760,-3.101,-18.825],[-25.436,-1.407,-7.338],
[-28.062,3.951,-12.017],[-37.022,10.281,-21.180],[-39.304,2.364,-22.568],
[-25.157,13.057,-11.192],[-36.665,-3.153,-12.582],[-38.881,11.999,-16.558],
[-39.665,13.575,-5.450],[-33.917,12.425,-3.633],[-29.853,10.794,-24.249],
[-27.789,9.256,-5.525],[-27.515,10.726,-21.863],[-39.701,6.824,-5.168],
[-35.137,8.322,-21.458],[-34.021,8.613,-5.339],[-29.865,9.127,-21.825],
[-27.374,6.706,-7.857],[-38.298,4.546,-5.295],[-32.553,5.460,-20.297],
[-34.344,3.071,-7.901],[-33.335,4.443,-23.179],[-29.873,6.425,-6.204],
[-22.892,7.396,-22.153],[-35.253,0.574,-6.641],[-33.615,0.268,-8.884],
[-31.600,5.619,-10.660],[-27.424,2.146,-5.058],[-26.951,6.342,-5.121],
[-35.116,3.114,-5.111],[-32.641,-2.294,-8.474],[-32.303,1.057,-3.309],
[-30.214,0.788,-5.519],[-27.762,-0.658,-4.766],[-37.346,1.233,-4.354],
[-33.570,-3.792,-6.022],[-32.896,-0.611,-5.908],[-33.728,15.104,-19.314],
[-39.468,9.067,-21.123],[-40.028,13.993,-15.105],[-40.775,12.374,-12.867],
[-22.189,12.951,-9.998],[-20.589,13.556,-7.873],[-24.571,8.975,-4.301],
[-22.909,7.776,-18.377],[-22.869,5.443,-16.778],[-22.560,8.830,-11.270],
[-23.085,9.464,-8.767],[-22.660,5.102,-20.889],[-24.281,2.939,-21.096],
[-24.685,4.617,-14.606],[-39.681,6.647,-12.720],[-24.920,5.891,-7.639],
[-23.193,3.016,-7.249],[-24.964,0.495,-20.470],[-24.610,2.071,-16.865],
[-24.890,1.971,-11.490],[-25.185,1.400,-8.457],[-39.945,0.731,-4.792],
[-23.985,-3.182,-3.652],[-24.239,-2.006,-17.203],[-38.397,-1.019,-10.584],
[-39.997,1.746,-11.154],[-24.113,-0.923,-5.093],[-22.733,1.289,-22.830],
[-40.187,-1.393,-18.335],[-25.654,14.480,-16.014],[-40.234,-1.065,-7.559],
[-37.568,11.257,-4.053],[-35.198,11.504,-9.168],[-29.678,14.126,-6.491],
[-26.736,15.586,-6.269],[-25.262,-4.258,-7.804],[-38.905,14.020,-8.154],
[-37.221,14.445,-10.341],[-31.006,14.066,-12.610],[-27.564,12.815,-9.748],
[-27.390,15.945,-8.971],[-39.146,12.539,-10.550],[-35.632,12.656,-13.517],
[-32.966,12.337,-13.088],[-29.988,-0.546,-12.815],[-24.510,-2.967,-12.035],
[-22.680,-4.096,-14.150],[-38.786,-2.788,-14.660],[-39.883,-4.264,-19.102],
[-32.519,-3.029,-15.340],[-27.238,-3.759,-14.178],[-27.073,-2.962,-18.163],
[-37.185,14.420,-17.705],[-37.817,-1.130,-20.047],[-30.471,-2.360,-17.881],
[-29.846,13.393,-18.901],[-22.456,12.694,-17.339],[-39.270,-1.772,-4.700],
[-35.741,-4.121,-3.885],[-36.508,13.232,-20.117],[-27.714,-2.866,-3.297],
[-24.925,12.468,-4.912],[-27.002,23.779,-14.389],[-30.283,24.841,-11.806],
[-36.557,23.664,-10.225],[-26.673,27.496,-13.875],[-36.053,27.373,-12.734],
[-39.974,21.626,-16.463],[-41.158,23.885,-13.605],[-28.348,32.482,-13.173],
[-30.377,28.307,-11.721],[-22.572,17.982,-11.767],[-27.534,19.962,-11.989],
[-35.321,19.978,-14.032],[-28.272,18.248,-15.525],[-38.174,21.702,-14.031],
[-33.488,22.481,-13.948],[-23.050,21.945,-7.017],[-37.754,18.725,-9.520],
[-29.322,28.347,-15.037],[-34.642,25.978,-9.577],[-24.360,18.422,-14.099],
[-37.329,21.058,-11.350],[-37.084,26.380,-9.229],[-37.304,28.682,-18.815],
[-27.301,29.000,-11.340],[-25.427,16.930,-10.722],[-31.640,16.938,-12.477],
[-25.847,31.008,-13.754],[-27.552,29.256,-17.522],[-24.476,26.655,-8.316],
[-30.619,25.251,-16.961],[-28.679,23.053,-21.902],[-22.294,23.110,-9.606],
[-26.601,24.241,-17.263],[-27.877,22.059,-8.269],[-34.819,25.429,-14.380],
[-33.046,27.219,-12.737],[-34.274,24.311,-11.892],[-30.510,31.219,-8.708],
[-33.463,28.478,-6.105],[-27.902,22.462,-19.219],[-37.428,23.069,-7.354],
[-30.090,30.999,-15.127],[-35.725,30.145,-5.582],[-23.637,16.578,-20.231],
[-35.266,24.596,-17.783],[-22.114,30.097,-20.468],[-39.662,26.593,-9.361],
[-40.048,29.123,-10.744],[-21.430,30.037,-5.148],[-30.519,29.813,-6.225],
[-31.239,33.110,-17.175],[-36.789,32.831,-16.207],[-28.461,30.433,-20.767],
[-28.512,31.743,-18.262],[-24.811,21.608,-5.139],[-38.017,28.293,-7.417],
[-35.582,16.861,-9.683],[-31.761,15.113,-9.583],[-27.197,19.691,-6.674],
[-26.101,24.949,-21.403],[-30.123,16.635,-21.748],[-34.637,21.328,-11.453],
[-35.199,21.891,-16.242],[-28.843,22.893,-16.725],[-29.870,22.331,-5.142],
[-28.382,21.078,-14.680],[-36.163,21.014,-8.746],[-33.166,19.171,-21.808],
[-39.773,19.113,-15.203],[-39.944,19.505,-19.052],[-28.697,22.307,-10.902],
[-32.789,18.920,-14.259],[-37.387,20.090,-16.528],[-35.021,20.722,-18.845],
[-28.468,18.251,-10.277],[-34.438,22.784,-20.512],[-24.551,33.995,-5.869],
[-37.152,25.424,-5.885],[-31.143,21.889,-12.456],[-36.149,17.478,-15.281],
[-32.262,21.007,-9.999],[-30.117,25.458,-20.028],[-24.231,24.421,-11.205],
[-29.004,15.784,-19.388],[-31.283,26.597,-9.113],[-27.324,19.552,-19.359],
[-37.358,24.239,-13.127],[-28.733,26.879,-18.353],[-40.960,22.506,-8.259],
[-32.462,23.279,-4.616],[-28.067,29.656,-7.715],[-22.088,21.444,-18.418],
[-22.117,25.128,-7.130],[-24.053,28.401,-19.439],[-27.134,24.903,-11.830],
[-32.948,28.756,-15.659],[-38.203,24.203,-20.187],[-33.218,26.011,-16.574],
[-32.885,29.837,-10.754],[-35.245,30.489,-17.143],[-28.563,26.193,-10.139],
[-26.270,26.882,-19.399],[-26.156,31.474,-6.492],[-21.781,32.636,-19.770],
[-38.913,20.043,-7.409],[-26.046,27.081,-16.727],[-37.509,30.562,-9.504],
[-32.554,28.444,-23.150],[-33.114,27.782,-20.547],[-36.366,26.311,-19.415],
[-31.524,17.849,-8.576],[-26.793,17.314,-17.848],[-31.705,27.918,-18.339],
[-32.614,20.457,-16.620],[-30.973,21.662,-7.818],[-30.524,17.257,-16.526],
[-33.277,31.310,-21.794],[-33.670,16.663,-21.570],[-25.384,19.858,-8.466],
[-26.177,21.918,-12.399],[-35.508,28.744,-21.326],[-38.677,21.489,-20.435],
[-25.909,31.321,-10.971],[-34.069,15.050,-11.796],[-38.265,28.128,-15.180],
[-40.542,31.226,-3.436],[-33.688,31.143,-3.410],[-30.797,30.564,-22.246],
[-28.479,27.581,-4.882],[-25.661,31.144,-20.711],[-41.494,25.996,-4.294],
[-36.076,25.470,-21.866],[-33.285,25.589,-6.999],[-30.988,26.570,-22.350],
[-27.549,25.658,-7.672],[-39.258,23.881,-4.333],[-32.468,24.815,-19.020],
[-33.641,22.893,-7.443],[-32.087,24.089,-21.606],[-29.752,24.774,-4.078],
[-23.678,25.640,-22.934],[-33.362,19.484,-7.099],[-34.333,19.272,-9.683],
[-31.977,23.981,-9.611],[-28.748,20.910,-3.011],[-27.372,23.462,-5.283],
[-36.441,22.413,-4.448],[-34.582,14.323,-9.059],[-32.888,19.289,-4.041],
[-30.321,19.705,-5.906],[-28.513,17.907,-4.906],[-37.883,18.906,-3.578],
[-34.319,13.933,-6.128],[-33.088,16.479,-4.545],[-34.547,33.753,-21.837],
[-39.531,27.264,-20.241],[-40.377,32.777,-17.009],[-39.253,31.360,-14.473],
[-23.383,31.661,-11.202],[-23.779,31.111,-8.492],[-23.757,27.074,-4.932],
[-23.162,26.129,-18.351],[-23.443,23.806,-16.724],[-25.043,27.369,-11.687],
[-25.794,28.969,-9.102],[-24.648,24.091,-19.199],[-23.796,21.514,-20.720],
[-26.058,21.813,-16.238],[-39.793,24.421,-11.113],[-24.913,24.448,-5.930],
[-22.874,22.642,-3.457],[-25.189,18.716,-21.360],[-24.304,18.510,-18.022],
[-24.950,19.513,-11.459],[-25.288,22.449,-9.691],[-40.158,18.992,-5.308],
[-21.116,15.617,-2.908],[-22.870,16.709,-16.835],[-38.733,17.252,-11.795],
[-40.833,21.964,-11.894],[-23.972,17.394,-6.752],[-22.088,18.872,-21.237],
[-39.502,16.779,-19.106],[-24.154,33.399,-16.152],[-39.502,17.015,-8.188],
[-38.149,30.079,-3.105],[-35.352,28.991,-8.006],[-31.465,32.350,-5.716],
[-28.551,32.567,-6.388],[-24.909,14.795,-8.252],[-37.907,33.312,-9.901],
[-36.615,32.454,-7.676],[-30.359,33.710,-10.860],[-28.295,32.185,-10.510],
[-26.287,33.473,-9.168],[-39.265,31.365,-11.936],[-35.448,30.196,-14.240],
[-34.366,32.621,-13.022],[-29.753,18.684,-13.516],[-25.709,15.343,-13.026],
[-23.238,15.322,-14.548],[-39.155,16.591,-14.552],[-39.006,14.090,-20.126],
[-32.887,15.774,-15.347],[-28.647,16.432,-12.943],[-25.759,14.650,-18.806],
[-39.045,30.437,-17.030],[-36.288,15.908,-21.011],[-32.395,17.909,-18.922],
[-30.860,31.922,-19.641],[-23.807,30.895,-18.074],[-37.925,17.297,-6.080],
[-35.670,14.943,-4.048],[-37.587,33.821,-20.970],[-28.336,15.611,-3.420],
[-24.112,31.244,-4.673],[-28.161,41.759,-15.587],[-30.068,43.017,-10.819],
[-37.757,42.669,-10.584],[-26.655,44.195,-14.694],[-36.888,45.913,-10.858],
[-40.427,39.095,-16.146],[-39.939,43.711,-15.024],[-29.330,52.002,-13.412],
[-30.919,45.452,-10.133],[-23.091,35.977,-10.420],[-27.424,38.416,-9.705],
[-34.687,38.202,-13.042],[-27.564,36.249,-15.730],[-39.019,39.084,-13.778],
[-31.861,42.740,-13.972],[-21.768,38.908,-9.642],[-35.289,36.922,-10.610],
[-28.782,45.413,-15.960],[-34.540,44.349,-10.230],[-25.493,36.544,-13.907],
[-37.043,38.425,-11.903],[-36.570,44.627,-7.993],[-38.360,47.408,-17.998],
[-25.527,47.067,-11.978],[-25.703,36.420,-10.501],[-32.783,34.689,-12.350],
[-27.991,47.895,-14.759],[-24.364,48.382,-16.388],[-25.441,45.761,-7.330],
[-30.440,42.763,-17.676],[-27.824,41.660,-22.071],[-22.580,42.511,-9.846],
[-25.072,42.386,-18.365],[-26.455,40.710,-8.401],[-34.280,43.990,-15.232],
[-33.061,46.078,-12.679],[-35.419,42.431,-12.283],[-28.060,48.073,-10.272],
[-34.554,47.045,-6.215],[-29.383,41.316,-19.980],[-36.149,42.040,-7.003],
[-29.693,47.811,-16.951],[-35.324,49.719,-6.234],[-25.041,33.505,-19.623],
[-36.023,42.880,-17.546],[-23.238,49.557,-20.507],[-39.418,44.842,-9.474],
[-40.075,47.616,-10.844],[-21.146,48.662,-6.914],[-31.776,48.332,-5.947],
[-29.768,52.539,-18.320],[-36.593,50.437,-16.274],[-27.378,49.342,-19.632],
[-27.691,51.110,-17.784],[-25.017,39.785,-6.236],[-37.443,46.781,-6.605],
[-35.036,33.263,-10.167],[-29.629,34.423,-8.093],[-28.212,39.579,-6.508],
[-26.473,43.272,-20.323],[-31.255,37.163,-19.107],[-35.088,39.659,-10.503],
[-33.259,39.645,-14.800],[-30.582,40.817,-15.692],[-30.225,39.246,-4.426],
[-28.078,39.212,-16.277],[-37.150,40.544,-9.162],[-34.957,38.261,-21.052],
[-39.829,36.836,-14.849],[-38.865,37.648,-18.035],[-28.786,40.754,-9.870],
[-32.605,36.970,-14.327],[-36.368,37.880,-15.475],[-35.177,38.378,-17.855],
[-28.642,35.681,-10.563],[-35.263,40.669,-19.975],[-24.145,51.557,-6.051],
[-37.768,44.233,-5.411],[-31.652,40.256,-12.070],[-36.320,35.273,-15.218],
[-32.655,38.540,-9.098],[-32.619,44.624,-17.557],[-25.404,44.195,-11.728],
[-28.330,36.432,-20.309],[-30.934,44.411,-6.577],[-27.740,39.120,-19.549],
[-37.676,43.676,-13.304],[-28.951,44.873,-18.669],[-39.658,42.704,-7.144],
[-32.023,41.686,-5.476],[-28.071,47.352,-7.514],[-21.486,40.088,-18.777],
[-22.532,45.091,-6.000],[-24.965,47.937,-19.323],[-28.021,42.523,-12.662],
[-36.082,46.836,-15.539],[-35.773,42.886,-21.543],[-33.016,46.505,-15.387],
[-32.760,47.957,-10.324],[-34.008,49.635,-15.599],[-28.399,45.382,-10.898],
[-27.812,46.908,-20.659],[-26.421,49.307,-6.097],[-20.388,52.276,-21.563],
[-38.738,38.204,-8.604],[-26.531,45.806,-18.390],[-38.112,47.773,-9.257],
[-32.293,47.560,-22.111],[-34.134,45.804,-19.531],[-36.895,44.672,-19.577],
[-29.991,37.798,-9.292],[-29.288,36.127,-17.875],[-31.397,46.979,-18.847],
[-30.528,38.219,-16.486],[-30.931,39.204,-7.154],[-30.334,35.645,-15.490],
[-33.072,48.955,-19.914],[-32.912,33.934,-19.452],[-24.572,38.410,-8.764],
[-26.647,39.677,-13.086],[-35.042,48.049,-21.398],[-37.757,40.629,-21.272],
[-26.877,49.798,-11.635],[-37.260,32.773,-12.703],[-40.663,48.197,-16.232],
[-40.246,50.271,-5.116],[-33.411,50.690,-4.182],[-29.513,48.623,-21.886],
[-29.016,48.435,-5.143],[-25.547,47.179,-22.106],[-40.073,45.488,-4.082],
[-35.576,45.527,-22.293],[-34.349,43.565,-4.966],[-30.484,45.682,-21.470],
[-26.787,43.636,-8.953],[-37.337,41.646,-4.456],[-33.577,42.950,-19.665],
[-34.013,41.109,-8.423],[-32.356,41.462,-21.435],[-28.876,43.450,-4.642],
[-23.938,43.274,-21.624],[-33.562,38.055,-6.279],[-31.771,36.207,-7.632],
[-31.317,41.068,-9.184],[-26.879,38.643,-3.834],[-26.768,41.829,-5.571],
[-35.713,38.513,-4.584],[-32.782,33.542,-7.926],[-31.642,36.820,-4.040],
[-29.264,36.731,-6.632],[-27.836,36.066,-4.389],[-37.842,36.573,-4.321],
[-34.509,32.610,-6.066],[-31.665,34.098,-3.625],[-32.003,51.320,-18.917],
[-38.832,46.757,-20.647],[-39.178,50.338,-16.519],[-40.131,50.818,-13.610],
[-22.713,50.877,-11.288],[-23.374,50.208,-8.865],[-24.847,46.367,-4.285],
[-23.814,44.726,-19.196],[-23.455,44.269,-14.732],[-23.134,45.544,-11.647],
[-25.394,48.204,-9.005],[-22.221,42.369,-20.079],[-24.397,40.031,-20.350],
[-25.064,40.980,-15.796],[-40.299,43.144,-12.012],[-24.360,42.845,-6.292],
[-22.150,41.242,-5.385],[-25.890,37.632,-21.020],[-25.932,38.937,-17.759],
[-24.219,38.491,-12.380],[-25.103,41.245,-10.680],[-40.291,38.457,-5.114],
[-22.420,33.766,-2.792],[-23.763,37.433,-16.863],[-39.309,36.790,-12.033],
[-39.247,40.402,-10.859],[-22.130,34.718,-5.385],[-23.220,37.913,-22.046],
[-40.602,35.562,-17.161],[-25.215,50.903,-15.795],[-39.761,35.747,-9.501],
[-37.268,48.389,-4.289],[-35.531,47.332,-8.832],[-31.638,50.312,-8.887],
[-27.817,51.938,-6.640],[-23.145,34.130,-8.374],[-37.270,51.520,-9.093],
[-35.249,49.390,-10.319],[-30.622,49.988,-11.536],[-28.397,50.236,-8.572],
[-26.107,51.574,-9.048],[-37.384,50.123,-11.627],[-35.985,49.185,-13.645],
[-33.445,49.657,-12.833],[-30.475,36.054,-12.931],[-26.681,33.147,-15.110],
[-22.844,35.316,-15.106],[-38.743,34.125,-14.953],[-39.888,33.221,-19.622],
[-33.138,33.873,-15.038],[-27.583,34.869,-12.808],[-26.555,34.246,-17.560],
[-38.792,49.845,-19.158],[-38.348,37.375,-21.253],[-33.945,36.143,-16.851],
[-30.318,51.030,-21.079],[-25.043,50.584,-18.462],[-38.422,34.778,-7.437],
[-35.120,33.599,-3.409],[-37.611,52.229,-20.610],[-27.808,33.371,-3.740],
[-23.888,49.033,-4.820],[-24.869,53.269,-10.854],[-28.394,59.125,-11.287],
[-35.915,55.460,-14.508],[-25.997,56.087,-17.236],[-39.802,58.983,-13.941],
[-20.750,58.522,-8.287],[-36.577,56.488,-8.296],[-25.622,56.264,-14.327],
[-35.096,56.051,-12.054],[-26.573,55.852,-10.866],[-33.015,52.558,-11.983],
[-24.216,52.505,-20.301],[-25.153,58.265,-4.999],[-35.340,52.600,-13.847],
[-29.855,52.471,-8.433],[-26.595,56.749,-6.954],[-29.404,55.733,-20.335],
[-35.495,58.646,-10.990],[-35.002,59.215,-15.215],[-28.307,57.464,-14.208],
[-37.909,57.674,-11.142],[-34.708,56.699,-21.800],[-41.925,55.392,-13.751],
[-38.063,56.356,-18.318],[-33.076,56.703,-14.087],[-37.537,58.224,-16.244],
[-35.190,58.562,-17.739],[-29.497,56.202,-11.646],[-35.010,-59.035,-20.025],
[-30.863,-59.346,-12.603],[-35.115,55.376,-17.658],[-31.790,57.580,-10.841],
[-28.214,53.071,-20.705],[-28.502,58.559,-20.338],[-22.167,58.508,-18.739],
[-39.336,56.407,-9.073],[-31.737,55.070,-10.481],[-28.357,54.608,-17.574],
[-32.680,56.885,-18.502],[-29.851,55.794,-8.012],[-31.021,54.866,-17.384],
[-36.378,57.450,-20.053],[-23.677,57.580,-8.792],[-38.688,56.797,-21.348],
[-38.954,53.375,-13.676],[-34.284,57.565,-7.106],[-33.838,56.443,-9.302],
[-29.293,58.533,-3.935],[-35.867,56.618,-5.080],[-33.957,51.812,-9.369],
[-32.305,56.605,-5.576],[-27.860,56.026,-4.459],[-26.605,53.388,-4.637],
[-38.485,56.094,-4.563],[-33.034,51.401,-6.746],[-33.668,54.374,-4.666],
[-22.680,56.750,-20.878],[-23.998,54.907,-22.755],[-25.420,57.443,-19.625],
[-24.266,57.105,-11.457],[-27.378,59.151,-8.671],[-40.342,56.654,-6.431],
[-22.117,52.168,-4.329],[-23.419,55.312,-17.275],[-38.223,55.983,-13.280],
[-40.454,57.261,-12.161],[-20.948,53.991,-5.836],[-21.486,54.417,-19.911],
[-40.761,55.924,-18.736],[-39.826,52.715,-9.604],[-24.156,54.394,-7.725],
[-31.061,54.567,-13.083],[-24.992,51.740,-13.126],[-23.630,53.850,-15.154],
[-38.599,52.981,-16.964],[-40.335,51.962,-19.360],[-32.527,53.334,-14.972],
[-28.291,54.743,-14.208],[-25.562,53.205,-17.647],[-36.863,54.388,-21.972],
[-31.641,54.008,-20.106],[-37.713,51.439,-5.270],[-36.624,54.008,-5.098],
[-27.858,50.816,-4.182],[-28.545,-48.438,5.158],[-31.122,-51.248,6.092],
[-37.270,-50.045,7.444],[-25.626,-47.554,5.814],[-38.856,-45.417,5.900],
[-39.703,-52.735,2.524],[-40.068,-49.584,3.265],[-28.165,-43.693,4.654],
[-30.803,-46.768,5.868],[-23.856,-56.815,8.524],[-27.276,-52.519,8.478],
[-35.579,-55.237,4.158],[-28.050,-54.548,3.848],[-37.880,-53.779,4.377],
[-33.244,-51.282,3.961],[-21.247,-53.678,9.047],[-37.557,-56.623,8.759],
[-28.298,-48.228,2.298],[-35.104,-49.179,9.259],[-23.417,-56.138,5.771],
[-37.042,-52.790,7.012],[-36.074,-47.865,11.247],[-39.329,-47.374,2.079],
[-28.534,-45.732,9.525],[-25.004,-54.418,8.369],[-33.302,-58.855,6.860],
[-27.688,-45.059,7.152],[-26.886,-44.816,2.523],[-26.291,-46.606,11.582],
[-30.057,-49.983,0.747],[-29.087,-50.448,-2.431],[-22.514,-49.747,8.702],
[-25.983,-49.141,2.850],[-28.538,-51.524,10.743],[-35.515,-48.800,4.308],
[-33.966,-45.836,6.789],[-34.209,-49.818,6.829],[-30.326,-44.220,8.089],
[-33.919,-46.401,10.895],[-26.503,-51.489,-1.076],[-36.587,-50.641,11.313],
[-31.217,-43.975,2.638],[-35.693,-45.330,12.697],[-24.660,-57.108,-1.132],
[-35.672,-50.180,2.194],[-23.489,-44.445,-3.226],[-38.928,-47.715,7.681],
[-40.900,-45.648,8.988],[-21.309,-45.950,11.881],[-31.746,-45.155,12.114],
[-30.634,-40.344,1.303],[-37.120,-40.769,4.153],[-28.589,-42.721,-1.452],
[-27.254,-41.743,1.952],[-24.164,-52.478,12.394],[-38.905,-46.625,10.998],
[-34.260,-56.256,6.561],[-30.787,-58.535,11.543],[-26.405,-53.579,11.148],
[-26.458,-50.803,-3.926],[-29.596,-56.503,-1.749],[-34.481,-53.319,7.805],
[-32.939,-53.459,5.555],[-28.288,-53.034,1.525],[-30.112,-52.321,13.101],
[-26.400,-51.937,3.515],[-37.144,-52.578,9.639],[-34.112,-56.736,-4.308],
[-40.067,-55.274,3.328],[-38.723,-55.025,0.031],[-30.099,-51.940,8.357],
[-32.369,-56.206,4.361],[-37.692,-56.346,2.698],[-33.281,-55.418,1.225],
[-27.523,-55.423,8.236],[-35.786,-52.536,-2.335],[-26.151,-41.196,11.738],
[-37.076,-47.856,13.717],[-30.263,-53.943,6.184],[-35.426,-57.918,2.441],
[-32.776,-55.248,9.106],[-32.601,-47.721,-0.778],[-25.276,-47.925,9.108],
[-26.794,-56.124,-2.791],[-31.477,-47.888,10.438],[-26.636,-54.418,-0.496],
[-37.066,-47.538,5.897],[-30.038,-47.114,-1.572],[-39.202,-49.737,11.301],
[-33.368,-51.924,13.093],[-29.383,-46.620,11.871],[-21.063,-51.754,-2.097],
[-23.637,-47.359,11.443],[-24.178,-44.067,1.977],[-27.334,-49.844,8.678],
[-33.874,-44.462,3.491],[-38.972,-48.313,-0.637],[-32.943,-48.438,3.242],
[-32.973,-42.748,10.343],[-34.886,-41.264,2.260],[-29.986,-49.383,8.054],
[-27.400,-47.741,-1.831],[-28.222,-43.746,11.947],[-24.238,-41.947,-2.551],
[-38.591,-54.944,10.839],[-25.032,-46.341,0.044],[-38.103,-44.296,9.681],
[-34.570,-45.316,-3.470],[-34.848,-46.297,-0.738],[-36.390,-48.342,0.275],
[-31.174,-55.075,11.579],[-29.172,-57.752,0.645],[-31.037,-44.677,-0.120],
[-31.366,-53.174,1.540],[-29.355,-53.987,9.650],[-29.769,-57.652,3.283],
[-33.294,-44.078,-1.347],[-32.995,-57.453,-0.711],[-24.906,-55.288,12.932],
[-26.893,-51.232,6.202],[-36.901,-44.837,-2.383],[-38.834,-53.328,-2.515],
[-26.029,-42.794,6.954],[-36.793,-58.011,6.551],[-37.587,-45.762,3.471],
[-37.200,-43.548,14.219],[-33.613,-43.799,13.947],[-30.829,-43.939,-3.043],
[-27.642,-45.756,14.075],[-26.387,-43.952,-2.506],[-39.129,-49.144,14.687],
[-33.922,-47.863,-3.088],[-32.558,-49.671,14.586],[-31.295,-47.698,-3.882],
[-26.819,-49.542,12.334],[-37.778,-51.153,13.720],[-34.300,-50.268,-1.303],
[-34.125,-52.414,10.472],[-33.067,-51.260,-3.436],[-29.808,-47.808,14.321],
[-23.214,-49.456,-2.838],[-33.964,-54.636,11.890],[-33.230,-57.266,11.162],
[-32.584,-50.195,11.196],[-26.906,-53.191,13.702],[-29.491,-50.481,15.428],
[-35.987,-52.982,12.973],[-33.771,59.515,9.742],[-31.558,-54.662,14.181],
[-28.639,-54.520,12.272],[-27.581,-55.894,14.639],[-37.116,-55.093,13.794],
[-33.718,-58.812,13.421],[-31.591,-57.931,15.247],[-30.843,-41.280,-1.344],
[-40.267,-46.207,-2.914],[-39.646,-41.802,2.926],[-38.332,-43.258,4.591],
[-22.675,-42.739,6.309],[-23.807,-42.716,11.591],[-25.323,-46.560,14.846],
[-22.715,-47.678,0.681],[-21.075,-49.701,1.697],[-22.740,-47.124,8.604],
[-24.522,-45.108,9.591],[-23.080,-50.210,-0.121],[-24.049,-53.291,-0.870],
[-24.113,-51.152,2.230],[-39.686,-50.376,5.970],[-23.905,-49.931,12.727],
[-22.707,-51.046,15.127],[-25.730,-53.648,-3.019],[-23.991,-53.955,1.749],
[-25.363,-53.963,5.580],[-24.754,-51.899,9.121],[-40.418,-55.170,13.884],
[-20.783,-58.730,14.590],[-23.743,-56.619,2.463],[-39.447,-54.872,6.257],
[-39.860,-52.200,7.850],[-22.321,-58.555,12.453],[-22.126,-57.474,-2.716],
[-40.045,-58.052,1.597],[-25.140,-42.402,3.824],[-40.185,-57.378,9.661],
[-38.824,-45.469,14.304],[-35.548,-44.010,10.328],[-30.442,-43.429,10.633],
[-28.467,-40.233,12.877],[-23.689,-59.051,10.251],[-37.482,-40.211,9.292],
[-35.072,-41.306,10.148],[-30.219,-40.902,8.076],[-28.158,-42.728,9.531],
[-25.247,-42.246,9.324],[-39.800,-43.362,7.474],[-35.616,-43.022,5.258],
[-33.835,-42.779,7.814],[-30.079,-56.595,7.825],[-24.907,-58.261,4.323],
[-22.136,-59.412,2.907],[-36.992,57.716,2.345],[-36.765,-56.757,0.138],
[-31.570,-59.941,4.851],[-27.071,-57.329,6.378],[-27.437,-59.063,2.073],
[-39.178,-44.616,1.869],[-35.395,-55.163,-1.929],[-30.727,-55.688,1.682],
[-32.991,-41.214,0.533],[-21.722,-43.704,0.192],[-38.130,-57.815,12.851],
[-36.264,-59.548,14.751],[-37.722,-41.291,-1.537],[-27.721,-58.984,14.161],
[-25.555,-44.045,13.055],[-28.096,-32.087,3.982],[-30.368,-31.058,6.847],
[-36.799,-32.526,8.535],[-25.469,-27.289,5.164],[-37.408,-28.492,7.339],
[-39.691,-33.003,1.026],[-39.569,-31.882,4.133],[-30.701,-22.492,5.479],
[-30.681,-28.021,6.211],[-23.204,-37.566,9.025],[-27.339,-35.118,8.146],
[-36.326,-35.776,4.737],[-26.895,-37.122,2.877],[-39.225,-34.627,4.598],
[-31.899,-32.088,4.668],[-23.865,-34.160,9.411],[-36.899,-36.430,8.288],
[-29.820,-28.202,2.861],[-34.200,-29.527,8.609],[-26.109,-38.702,5.723],
[-36.079,-34.045,6.689],[-36.106,-27.403,9.298],[-38.835,-27.390,0.663],
[-25.960,-27.059,9.654],[-25.821,-38.497,8.440],[-33.137,-40.051,6.583],
[-28.087,-24.475,5.031],[-26.116,-25.705,2.705],[-25.070,-29.274,11.701],
[-30.751,-30.801,2.143],[-30.662,-32.757,-2.531],[-22.343,-30.343,9.526],
[-25.669,-32.906,1.277],[-28.200,-33.756,10.498],[-34.843,-30.653,1.404],
[-34.524,-29.360,5.948],[-35.082,-32.145,4.909],[-27.642,-27.989,6.994],
[-35.654,-27.893,12.024],[-28.501,-32.659,-0.844],[-36.783,-30.380,10.079],
[-31.042,-25.308,3.220],[-37.749,-24.230,13.873],[-22.022,-39.858,-2.634],
[-37.469,-31.577,-0.016],[-23.301,-26.133,-1.872],[-39.644,-29.610,9.392],
[-40.190,-26.139,6.754],[-22.979,-25.686,12.235],[-32.319,-25.824,12.442],
[-30.681,-22.592,2.814],[-37.705,-25.037,4.198],[-26.796,-23.815,-1.524],
[-28.714,-23.554,1.194],[-25.200,-34.079,12.003],[-38.267,-27.509,11.811],
[-36.135,-40.329,6.975],[-30.922,-40.701,11.100],[-26.108,-37.214,13.187],
[-25.967,-32.626,-3.083],[-31.386,-36.768,-0.972],[-33.638,-34.138,7.878],
[-33.612,-34.588,4.833],[-28.365,-33.725,1.794],[-31.332,-34.054,14.065],
[-27.829,-34.922,4.411],[-37.836,-34.143,10.677],[-33.715,-37.780,-2.058],
[-41.175,-36.750,4.046],[-38.624,-37.377,-2.485],[-30.462,-33.190,8.704],
[-33.561,-37.416,4.264],[-38.744,-36.348,2.274],[-33.964,-35.404,2.069],
[-28.492,-37.577,8.489],[-35.519,-35.774,-0.794],[-25.389,-21.880,14.151],
[-37.213,-30.061,12.865],[-30.294,-34.511,6.135],[-35.679,-38.116,2.718],
[-33.707,-34.737,10.731],[-31.415,-27.989,0.307],[-25.036,-31.144,8.033],
[-29.158,-36.472,-2.357],[-30.957,-29.337,11.867],[-27.641,-35.443,-0.345],
[-37.637,-31.014,5.692],[-29.822,-30.390,-0.190],[-39.795,-31.490,11.878],
[-31.983,-31.493,13.243],[-29.008,-27.642,10.643],[-20.831,-32.501,0.650],
[-21.614,-28.371,13.866],[-24.472,-26.584,0.671],[-27.967,-31.606,8.226],
[-33.804,-25.621,2.994],[-38.016,-30.201,-2.174],[-33.086,-29.481,3.381],
[-33.887,-25.406,8.455],[-33.639,-23.045,1.417],[-31.242,-28.663,8.637],
[-28.263,-28.128,-1.216],[-27.318,-25.388,12.007],[-22.742,-23.326,-2.425],
[-39.750,-36.604,11.707],[-26.709,-29.976,0.316],[-38.844,-26.391,9.432],
[-32.785,-27.226,-2.890],[-33.814,-29.128,-0.897],[-36.681,-29.023,0.364],
[-30.470,-38.174,10.400],[-28.349,-37.687,0.760],[-29.988,-25.417,-0.791],
[-31.069,-35.058,1.273],[-30.725,-33.439,11.399],[-30.449,-37.773,2.597],
[-32.081,-23.550,-0.606],[-34.933,-38.840,-0.053],[-25.662,-36.562,10.582],
[-26.333,-33.185,6.499],[-36.501,-26.335,-0.135],[-38.220,-35.086,-1.245],
[-26.293,-24.471,8.130],[-38.880,-40.041,6.928],[-35.532,-27.536,3.831],
[-40.749,-24.542,15.884],[-34.019,-24.316,14.279],[-30.344,-26.958,-3.882],
[-29.027,-27.749,13.962],[-26.252,-25.928,-4.813],[-39.731,-29.430,13.900],
[-34.901,-28.866,-3.859],[-33.111,-28.975,13.485],[-30.575,-29.780,-2.865],
[-27.432,-30.830,10.872],[-38.436,-32.853,13.763],[-32.959,-31.576,-0.514],
[-34.871,-32.071,12.339],[-34.318,-33.209,-2.546],[-29.353,-32.039,13.991],
[-23.889,-29.632,-2.626],[-34.987,-36.781,12.069],[-34.594,-37.598,7.990],
[-34.000,-32.158,9.621],[-26.640,-34.528,14.589],[-26.878,-32.325,13.144],
[-35.591,-34.176,14.350],[-32.903,-39.990,9.587],[-32.673,-36.225,15.174],
[-28.854,-36.980,12.309],[-27.915,-37.673,14.967],[-37.720,-35.793,14.366],
[-33.327,-41.427,12.721],[-32.212,-40.370,15.034],[-32.713,-22.435,-3.252],
[-39.025,-27.416,-2.553],[-40.238,-22.236,2.226],[-40.439,-24.040,4.569],
[-23.221,-23.079,7.356],[-21.994,-24.935,9.195],[-24.278,-28.501,14.102],
[-23.861,-29.029,-0.040],[-22.676,-31.585,2.626],[-23.444,-28.880,6.909],
[-25.174,-24.622,10.502],[-23.451,-31.809,-0.183],[-23.959,-34.355,-1.891],
[-24.629,-33.883,3.781],[-41.037,-30.958,7.164],[-24.618,-31.732,10.757],
[-22.310,-33.515,12.925],[-24.742,-36.971,-2.057],[-25.109,-35.446,0.255],
[-24.365,-36.827,3.730],[-24.664,-35.887,7.696],[-40.450,-34.497,14.504],
[-22.734,-41.126,15.578],[-24.336,-38.029,1.090],[-39.561,-37.064,7.215],
[-40.024,-33.911,7.964],[-24.262,-39.534,14.039],[-22.076,-37.388,-2.087],
[-40.332,-38.357,-0.443],[-25.819,-23.091,2.782],[-39.742,-39.008,10.128],
[-37.697,-27.185,14.536],[-34.889,-25.219,11.249],[-29.917,-24.272,13.010],
[-27.816,-22.524,11.577],[-23.556,-40.041,10.934],[-36.693,-20.633,10.278],
[-38.516,-22.833,8.761],[-32.050,-23.308,8.441],[-29.075,-23.895,8.808],
[-26.676,-21.002,7.420],[-37.624,-25.160,7.262],[-35.441,-26.739,6.519],
[-33.240,-23.611,4.507],[-29.954,-37.308,6.169],[-23.950,-40.577,7.167],
[-23.580,-40.314,4.504],[-39.060,-38.410,4.326],[-39.801,-41.121,0.271],
[-33.128,-41.792,4.041],[-27.650,-41.005,4.539],[-28.036,-40.215,-0.265],
[-38.365,-24.584,1.147],[-37.385,-37.559,0.166],[-32.795,-37.799,1.172],
[-29.844,-23.003,-2.188],[-22.833,-24.089,1.677],[-37.531,-41.182,11.729],
[-36.748,-40.705,14.158],[-36.420,-22.884,-3.101],[-28.426,-42.469,14.671],
[-24.897,-25.125,14.615],[-29.917,-12.203,4.263],[-30.232,-11.767,6.966],
[-37.159,-13.419,8.195],[-26.461,-8.510,4.120],[-36.339,-9.428,4.930],
[-40.433,-14.904,2.477],[-39.578,-12.737,4.008],[-29.980,-5.230,4.014],
[-32.440,-9.906,7.685],[-23.253,-17.832,8.761],[-28.008,-15.874,7.327],
[-34.976,-16.504,4.262],[-27.533,-18.758,3.198],[-39.364,-16.041,5.083],
[-30.968,-14.913,3.692],[-22.865,-15.158,11.147],[-37.883,-18.955,8.365],
[-29.979,-8.377,2.036],[-33.500,-11.913,10.651],[-25.770,-18.447,5.073],
[-36.977,-17.449,6.149],[-36.222,-11.499,9.977],[-38.459,-9.474,1.083],
[-27.239,-8.333,7.558],[-25.307,-19.525,9.722],[-33.711,-20.246,6.262],
[-27.890,-6.474,5.219],[-28.031,-10.250,2.806],[-23.964,-9.848,11.534],
[-30.055,-11.658,1.247],[-29.648,-14.429,-2.680],[-23.597,-13.995,8.742],
[-27.886,-13.275,2.683],[-27.415,-14.984,11.317],[-35.379,-12.079,4.031],
[-35.403,-7.659,7.080],[-33.807,-12.676,6.460],[-29.670,-7.352,7.419],
[-35.799,-8.359,9.873],[-26.767,-14.080,0.343],[-37.702,-14.454,10.658],
[-32.199,-7.378,3.181],[-35.388,-6.580,12.036],[-23.599,-21.794,-0.197],
[-37.442,-13.248,2.521],[-22.635,-6.095,-2.879],[-39.704,-11.867,9.408],
[-40.083,-9.284,8.437],[-21.849,-8.117,12.169],[-32.961,-8.379,11.294],
[-30.825,-3.708,0.964],[-37.071,-5.061,2.595],[-28.191,-5.097,-1.608],
[-27.776,-3.970,1.616],[-23.906,-16.489,13.441],[-37.737,-9.428,11.141],
[-35.301,-21.180,8.170],[-30.922,-22.008,10.615],[-28.296,-17.911,10.876],
[-27.758,-11.924,-2.226],[-30.844,-19.950,-0.679],[-35.598,-14.531,6.147],
[-33.995,-14.250,2.085],[-29.693,-14.082,0.057],[-28.941,-16.312,12.948],
[-27.589,-16.074,3.752],[-36.203,-17.039,9.435],[-34.935,-17.940,-2.178],
[-41.472,-19.406,4.018],[-39.303,-17.777,1.354],[-29.481,-14.487,9.292],
[-32.454,-18.302,4.453],[-37.378,-17.328,3.359],[-34.083,-16.387,-0.026],
[-28.549,-18.493,7.844],[-34.857,-13.900,-0.899],[-24.552,-3.034,13.871],
[-37.653,-10.957,13.407],[-29.335,-14.344,5.739],[-35.060,-20.129,2.609],
[-30.918,-16.555,8.106],[-32.353,-9.717,0.926],[-24.849,-11.734,9.452],
[-29.183,-16.598,-1.269],[-30.339,-9.684,11.677],[-26.364,-16.959,-0.592],
[-37.744,-11.761,5.893],[-30.116,-7.948,-1.345],[-40.343,-14.101,11.961],
[-32.646,-14.389,13.259],[-29.251,-6.891,11.522],[-22.227,-14.988,-0.546],
[-21.597,-10.921,11.952],[-25.180,-7.988,1.208],[-26.331,-10.863,6.093],
[-34.746,-9.447,2.641],[-39.172,-13.095,-1.413],[-31.599,-9.789,4.152],
[-33.030,-6.778,9.001],[-34.488,-5.984,3.147],[-28.652,-10.054,9.253],
[-27.007,-9.179,-1.997],[-25.836,-5.878,10.756],[-24.512,-4.767,-1.046],
[-38.834,-17.141,10.543],[-27.289,-10.614,0.375],[-38.336,-7.258,9.273],
[-33.038,-8.328,-4.382],[-34.752,-8.588,-0.041],[-37.209,-11.381,-0.173],
[-30.728,-18.976,9.935],[-28.382,-18.554,0.612],[-32.323,-5.979,-0.550],
[-31.512,-16.387,0.467],[-30.946,-16.543,10.997],[-30.308,-19.414,2.807],
[-34.505,-4.439,0.874],[-33.610,-20.669,-0.785],[-25.094,-19.714,12.708],
[-26.613,-13.835,5.823],[-36.952,-8.208,-1.523],[-39.097,-16.061,-2.369],
[-24.874,-6.392,8.233],[-36.135,-22.929,6.015],[-37.076,-7.807,2.464],
[-38.000,-4.787,13.668],[-33.934,-5.133,15.550],[-30.969,-6.541,-3.329],
[-28.308,-8.569,13.653],[-25.900,-6.737,-3.278],[-40.037,-12.070,13.971],
[-35.266,-9.389,-3.259],[-35.146,-11.200,12.625],[-30.684,-11.889,-3.422],
[-27.529,-12.478,10.039],[-38.069,-13.980,13.232],[-34.330,-10.994,-1.240],
[-35.008,-14.440,9.397],[-33.050,-13.960,-3.219],[-28.720,-11.898,12.957],
[-23.353,-10.667,-3.253],[-34.065,-16.159,11.406],[-34.315,-18.771,9.733],
[-32.335,-14.166,9.939],[-26.772,-15.845,14.570],[-26.219,-11.834,12.239],
[-35.077,-14.206,14.629],[-32.672,-20.610,8.847],[-31.165,-16.909,15.311],
[-30.166,-19.041,12.574],[-28.208,-18.804,14.536],[-37.036,-18.363,14.385],
[-32.774,-23.157,12.393],[-32.182,-22.307,15.111],[-33.700,-2.230,-3.757],
[-39.423,-9.870,-1.640],[-41.079,-2.576,0.975],[-40.750,-6.193,4.374],
[-22.758,-5.757,6.477],[-23.005,-5.029,11.071],[-25.088,-9.837,13.918],
[-22.817,-10.674,0.027],[-22.928,-13.489,2.671],[-23.757,-9.791,7.417],
[-26.395,-9.396,10.673],[-25.181,-13.859,-1.692],[-23.593,-16.036,-2.544],
[-24.307,-15.476,4.018],[-39.286,-13.678,6.718],[-23.787,-12.503,12.890],
[-22.062,-14.210,14.458],[-25.099,-19.263,-1.614],[-25.117,-17.518,2.178],
[-25.551,-17.007,7.552],[-25.265,-17.053,10.971],[-41.692,-18.263,12.204],
[-21.215,-21.268,15.602],[-25.541,-20.430,2.325],[-39.054,-18.888,5.643],
[-41.864,-14.847,6.646],[-21.982,-19.332,13.621],[-21.536,-18.058,-1.661],
[-40.603,-19.764,-0.303],[-26.016,-5.858,2.997],[-40.098,-19.750,9.809],
[-39.313,-8.237,13.531],[-36.730,-5.531,8.091],[-31.529,-6.786,13.037],
[-27.089,-3.246,12.483],[-23.357,-21.205,11.020],[-38.538,-4.405,9.482],
[-35.319,-3.940,10.201],[-31.183,-4.958,7.772],[-28.979,-4.820,9.933],
[-26.862,-2.993,9.446],[-38.998,-4.634,6.725],[-36.623,-6.820,4.721],
[-33.411,-5.777,6.332],[-30.360,-19.365,5.933],[-25.583,-22.833,5.977],
[-23.144,-21.899,4.689],[-37.872,-21.713,3.952],[-38.996,-22.786,-1.124],
[-32.674,-20.905,3.337],[-28.391,-20.838,4.942],[-28.357,-21.164,-0.293],
[-39.478,-5.925,1.714],[-36.896,-20.049,-1.302],[-32.891,-18.712,1.494],
[-30.570,-3.733,-2.572],[-23.688,-5.771,1.379],[-36.956,-20.690,12.896],
[-35.585,-20.051,15.895],[-37.806,-5.513,-2.077],[-28.768,-22.498,14.977],
[-23.831,-6.269,13.922],[-30.353,6.857,4.425],[-30.813,4.416,6.075],
[-37.845,4.707,7.270],[-26.454,7.892,4.724],[-37.465,8.084,6.813],
[-38.597,3.255,0.971],[-39.851,4.678,3.255],[-28.685,12.840,5.267],
[-32.681,8.323,6.109],[-22.077,0.324,7.685],[-29.181,2.385,7.553],
[-35.380,1.279,5.003],[-27.572,0.793,2.841],[-40.476,2.801,5.028],
[-33.227,3.847,5.009],[-21.564,1.162,10.275],[-35.904,-0.154,8.096],
[-28.508,8.402,2.992],[-35.241,6.795,9.171],[-26.235,-0.837,5.561],
[-37.868,1.463,6.757],[-36.814,10.291,9.992],[-39.183,9.013,0.749],
[-26.283,9.301,7.331],[-25.536,-0.797,8.323],[-33.660,-2.588,6.019],
[-26.128,11.942,4.731],[-26.293,10.584,2.163],[-25.471,6.978,12.010],
[-30.876,5.496,0.776],[-29.907,5.899,-1.651],[-21.531,4.869,9.321],
[-26.544,5.387,2.646],[-29.311,2.868,12.406],[-34.681,6.231,3.773],
[-35.052,9.684,6.691],[-35.406,6.294,6.498],[-30.054,9.815,8.587],
[-34.411,8.880,10.870],[-27.202,4.613,0.317],[-35.953,4.950,10.847],
[-31.012,13.172,2.880],[-35.441,10.666,12.764],[-23.830,-1.870,-1.231],
[-36.100,5.330,0.921],[-24.155,11.186,-2.856],[-39.092,9.156,8.762],
[-39.864,10.874,6.020],[-21.180,11.253,12.221],[-32.542,10.775,11.807],
[-30.079,14.972,0.297],[-37.966,14.284,4.645],[-26.726,13.732,-1.672],
[-28.911,11.701,1.621],[-24.713,4.397,13.648],[-39.110,9.407,11.831],
[-35.588,-3.008,7.811],[-30.402,-2.870,11.352],[-27.122,1.079,11.709],
[-25.896,6.277,-2.717],[-30.610,1.083,-1.063],[-35.119,2.279,7.730],
[-33.434,4.064,1.772],[-30.158,3.806,3.408],[-30.522,3.789,14.661],
[-28.467,2.366,4.929],[-39.063,4.920,9.652],[-34.769,0.362,-4.024],
[-39.619,-0.954,2.302],[-39.150,0.798,-1.917],[-29.238,4.110,9.850],
[-32.537,1.184,3.705],[-36.428,2.387,2.697],[-34.487,2.602,-0.198],
[-28.547,-0.244,8.244],[-35.314,4.678,-1.977],[-25.426,14.279,12.952],
[-36.706,7.956,12.172],[-31.707,1.574,6.208],[-34.767,-0.778,3.324],
[-32.529,0.920,8.701],[-32.873,8.007,-0.212],[-26.348,6.607,7.384],
[-28.280,-0.683,-1.774],[-30.296,9.299,11.312],[-28.168,2.517,-1.543],
[-37.411,4.864,4.383],[-29.342,8.655,0.266],[-39.825,7.235,10.606],
[-32.520,4.838,12.266],[-27.137,8.804,11.018],[-23.755,2.781,-0.653],
[-22.609,8.356,10.958],[-25.316,10.004,-0.301],[-28.748,5.772,7.223],
[-35.268,10.700,2.930],[-37.845,5.814,-1.856],[-33.112,8.547,3.441],
[-33.079,12.778,10.142],[-34.543,14.806,1.621],[-29.632,7.264,9.798],
[-27.488,9.009,-2.396],[-26.596,13.222,10.299],[-22.424,13.170,-2.058],
[-38.538,0.654,9.610],[-26.582,7.539,-0.042],[-40.095,12.232,9.763],
[-31.451,11.444,-3.959],[-33.728,10.066,-2.165],[-35.933,8.038,1.612],
[-30.587,-0.473,10.136],[-28.907,-0.912,0.803],[-32.341,12.640,0.129],
[-31.069,2.271,1.395],[-31.398,2.208,10.829],[-31.489,-0.421,1.781],
[-35.799,12.216,-1.331],[-34.938,-0.864,-1.717],[-24.097,-0.503,12.430],
[-25.597,3.914,6.486],[-35.973,9.629,-0.418],[-39.253,3.057,-3.523],
[-26.373,13.823,7.586],[-38.107,-1.918,7.137],[-37.576,12.204,2.573],
[-39.005,12.600,15.261],[-34.927,13.113,15.145],[-30.594,10.507,-1.257],
[-28.668,10.523,13.209],[-27.215,11.371,-3.928],[-39.324,6.991,13.623],
[-36.322,7.831,-2.794],[-32.311,7.334,12.064],[-30.626,8.016,-3.215],
[-27.191,6.284,10.071],[-37.735,4.717,13.312],[-32.960,5.406,-0.866],
[-33.880,3.654,9.678],[-32.531,3.697,-3.416],[-29.751,7.883,14.267],
[-24.057,7.902,-1.695],[-33.559,1.120,12.004],[-34.157,-0.795,10.111],
[-32.462,6.252,8.918],[-28.016,2.402,14.650],[-26.298,4.111,11.623],
[-35.260,3.806,13.619],[-32.757,-2.673,8.676],[-32.819,1.668,14.528],
[-29.651,-0.564,12.739],[-26.862,-0.706,13.863],[-36.692,1.056,12.891],
[-33.598,-2.412,12.282],[-30.944,-3.564,14.443],[-33.230,14.973,-1.282],
[-40.614,10.211,-2.100],[-38.677,14.570,2.014],[-40.565,13.196,4.173],
[-22.904,14.660,8.051],[-23.824,11.918,7.421],[-23.395,10.009,13.604],
[-24.048,7.516,1.188],[-23.208,4.271,1.816],[-23.946,7.789,6.101],
[-24.713,10.583,10.860],[-22.685,5.513,-2.229],[-23.536,3.206,-3.390],
[-24.953,3.417,3.819],[-39.851,6.833,6.036],[-23.262,5.538,11.540],
[-21.218,2.834,15.212],[-25.213,0.787,-2.209],[-25.843,2.261,1.366],
[-24.414,1.808,8.331],[-23.811,2.739,10.814],[-40.435,1.481,13.473],
[-20.791,-3.388,15.099],[-24.867,-1.189,2.583],[-38.498,-0.434,4.673],
[-41.192,2.352,7.827],[-21.722,-1.219,13.937],[-22.557,0.198,-2.515],
[-41.050,0.051,0.075],[-24.801,13.325,2.700],[-39.450,-2.038,10.461],
[-37.873,10.429,14.349],[-32.746,10.617,8.400],[-30.862,12.635,13.265],
[-28.166,13.624,12.681],[-24.600,-2.898,11.143],[-39.267,14.084,12.520],
[-35.031,14.712,9.530],[-29.472,15.551,7.019],[-29.037,13.075,8.623],
[-27.610,16.097,9.031],[-38.308,14.077,8.110],[-37.075,10.997,5.784],
[-34.947,12.959,5.186],[-30.668,-0.963,4.944],[-25.746,-3.573,6.047],
[-24.302,-3.870,3.684],[-36.882,-2.506,3.856],[-38.428,-3.501,0.810],
[-31.511,-3.158,3.494],[-28.283,-2.851,4.867],[-26.298,-1.943,0.252],
[-39.692,11.541,1.231],[-37.226,-2.038,-2.819],[-31.243,-1.668,-0.894],
[-31.014,14.676,-3.718],[-23.006,12.956,0.764],[-39.372,-2.342,13.300],
[-36.698,-2.164,14.413],[-38.137,14.037,-3.166],[-28.214,-4.704,14.563],
[-25.874,11.740,13.302],[-27.861,25.025,4.599],[-29.765,25.010,6.221],
[-37.392,23.852,8.272],[-25.288,27.568,5.619],[-37.044,28.583,5.856],
[-39.000,22.173,2.383],[-39.661,24.562,3.467],[-28.530,33.063,3.850],
[-32.193,26.575,6.430],[-23.168,17.467,8.486],[-27.189,20.532,9.036],
[-34.543,20.047,4.248],[-26.911,19.424,2.650],[-39.139,21.565,5.855],
[-33.926,22.641,5.051],[-22.012,19.890,9.505],[-37.273,18.450,8.254],
[-29.347,28.664,2.525],[-35.411,25.356,9.316],[-25.430,17.768,4.390],
[-36.736,21.456,6.826],[-36.050,28.155,8.748],[-40.188,27.411,0.051],
[-26.818,29.139,7.449],[-25.573,18.546,8.021],[-32.021,15.350,6.203],
[-25.644,30.427,5.335],[-26.614,29.032,2.282],[-25.901,27.543,12.422],
[-29.396,24.711,2.038],[-29.832,23.990,-1.186],[-23.473,23.123,9.361],
[-26.604,24.135,2.112],[-28.198,22.837,11.058],[-34.877,25.264,4.247],
[-34.566,27.393,5.756],[-34.777,24.227,6.931],[-28.846,30.366,9.341],
[-34.239,27.165,11.022],[-26.658,23.902,-0.752],[-36.193,23.467,11.817],
[-30.824,31.191,3.183],[-34.937,29.449,13.049],[-23.094,15.543,-0.786],
[-36.962,24.146,1.931],[-23.330,29.139,-2.872],[-40.977,27.516,9.824],
[-39.689,29.201,6.683],[-21.519,30.240,10.488],[-31.058,28.596,10.935],
[-30.999,33.500,1.836],[-38.107,33.781,2.629],[-28.132,31.139,-2.194],
[-28.397,33.165,1.008],[-24.994,22.054,12.800],[-38.570,28.119,10.903],
[-34.720,15.839,5.411],[-30.494,14.193,10.902],[-27.176,20.269,11.712],
[-25.456,24.681,-3.147],[-30.835,17.575,-3.011],[-34.826,19.963,8.435],
[-36.103,21.429,1.553],[-28.801,22.039,1.304],[-29.563,22.062,13.537],
[-27.985,21.652,4.060],[-36.509,20.956,10.553],[-34.997,18.834,-2.108],
[-40.506,18.136,5.050],[-40.102,18.310,-1.039],[-29.460,22.298,8.436],
[-33.153,17.894,4.976],[-37.450,20.343,3.970],[-34.409,19.389,0.764],
[-27.629,18.415,6.250],[-35.579,21.857,-0.930],[-24.927,33.896,12.872],
[-36.638,25.129,15.089],[-31.654,21.788,5.775],[-36.190,18.385,2.521],
[-32.146,21.230,8.579],[-32.475,26.151,0.058],[-24.896,25.186,8.087],
[-27.934,18.166,-2.271],[-28.728,26.902,11.951],[-25.747,21.961,-2.188],
[-37.428,24.715,5.225],[-29.299,27.833,-0.350],[-40.343,25.608,11.849],
[-32.469,23.824,12.807],[-28.899,30.000,11.983],[-21.517,21.336,-0.795],
[-22.878,26.691,12.627],[-23.795,30.243,0.527],[-27.460,26.434,7.428],
[-34.561,29.381,1.355],[-39.110,22.901,-0.341],[-31.780,26.597,2.687],
[-31.729,31.756,9.993],[-34.622,33.110,3.701],[-29.032,27.469,9.311],
[-28.255,25.988,-2.361],[-26.104,30.504,12.116],[-23.727,31.748,-1.832],
[-38.988,18.746,13.457],[-26.545,27.477,-0.407],[-38.422,30.279,9.266],
[-33.489,28.308,-3.332],[-34.869,26.985,-1.143],[-37.473,25.012,-1.471],
[-30.281,19.405,9.515],[-27.442,16.736,0.116],[-31.267,30.772,0.330],
[-31.761,20.601,0.668],[-31.121,21.782,11.082],[-28.999,17.451,2.531],
[-33.565,29.804,-1.064],[-34.948,16.999,0.061],[-25.855,17.955,10.870],
[-26.081,23.151,5.715],[-37.730,27.796,-1.707],[-38.012,20.851,-1.594],
[-25.083,33.883,6.514],[-37.266,16.197,6.536],[-38.968,29.730,1.609],
[-38.378,31.910,13.575],[-35.292,32.220,15.486],[-30.768,30.271,-2.949],
[-27.476,29.164,14.240],[-26.241,29.066,-4.000],[-39.681,25.841,14.619],
[-35.700,26.055,-3.598],[-31.611,26.664,12.798],[-30.888,27.377,-3.375],
[-26.640,25.374,9.971],[-38.481,23.589,13.377],[-32.914,23.441,-0.477],
[-33.837,22.109,11.114],[-33.090,21.375,-2.274],[-28.961,25.144,14.255],
[-24.812,27.005,-2.098],[-34.004,19.339,12.030],[-35.159,17.599,10.059],
[-32.160,23.810,9.295],[-26.971,21.008,14.392],[-26.836,24.398,13.158],
[-34.255,21.779,13.792],[-32.775,16.401,9.329],[-31.464,19.802,14.828],
[-29.452,19.016,13.036],[-27.118,18.025,13.671],[-36.648,19.991,13.814],
[-33.207,14.630,12.388],[-32.412,16.102,14.664],[-33.060,33.462,-0.479],
[-39.160,26.622,-3.806],[-40.781,33.482,1.681],[-39.675,32.001,4.293],
[-23.347,31.062,6.699],[-24.054,31.594,10.506],[-24.298,28.171,14.731],
[-22.534,26.701,0.013],[-24.329,24.627,3.502],[-22.573,28.629,7.874],
[-25.173,29.258,9.638],[-23.109,24.271,-1.293],[-23.976,19.897,-2.632],
[-23.201,21.018,2.341],[-39.150,25.668,7.527],[-24.167,24.244,11.663],
[-22.717,23.278,13.614],[-25.868,18.598,-4.443],[-25.523,20.001,0.208],
[-24.978,21.610,7.764],[-24.655,20.382,10.177],[-41.503,19.399,14.467],
[-23.319,14.736,14.549],[-24.937,16.812,1.093],[-39.981,19.164,7.537],
[-38.795,21.487,8.833],[-24.499,16.742,13.030],[-23.296,17.495,-3.452],
[-39.991,17.699,1.581],[-26.569,31.498,2.999],[-39.766,16.969,10.659],
[-37.082,29.760,14.921],[-34.343,30.898,8.454],[-30.795,32.213,12.768],
[-28.043,32.646,11.177],[-24.173,14.561,10.471],[-37.719,33.101,10.227],
[-35.070,33.232,9.394],[-30.601,32.673,7.281],[-25.938,31.713,7.820],
[-25.721,33.501,9.676],[-37.908,31.016,6.490],[-35.305,30.485,5.602],
[-32.701,29.726,6.278],[-30.316,17.866,5.373],[-24.612,14.301,5.702],
[-23.893,15.884,3.677],[-38.274,17.112,3.730],[-39.341,14.376,-0.696],
[-31.539,16.112,3.380],[-29.018,15.540,4.308],[-26.732,15.176,2.214],
[-39.680,30.605,-0.956],[-37.377,16.201,-1.454],[-30.546,17.681,-0.210],
[-29.773,33.235,-1.717],[-23.644,32.654,2.365],[-36.940,16.823,12.862],
[-35.352,15.567,14.637],[-37.698,33.330,-2.885],[-27.618,15.330,14.917],
[-23.456,30.063,12.710],[-30.364,42.958,3.283],[-30.781,43.458,6.287],
[-37.109,41.608,6.721],[-26.488,45.746,4.571],[-36.412,45.588,7.158],
[-38.569,41.279,1.751],[-39.959,42.630,3.801],[-28.951,50.400,6.013],
[-30.651,45.698,7.849],[-23.376,35.120,8.853],[-28.376,39.652,7.379],
[-35.544,38.314,4.738],[-26.734,36.453,2.130],[-39.634,39.773,4.172],
[-32.769,42.531,4.490],[-22.266,39.997,10.209],[-37.941,37.723,9.225],
[-29.409,45.586,3.197],[-34.559,44.167,8.267],[-26.835,37.751,4.879],
[-38.023,38.781,6.562],[-37.096,46.102,9.739],[-39.061,46.132,-0.028],
[-27.208,47.731,6.499],[-26.041,36.253,7.632],[-32.748,33.609,5.943],
[-27.904,49.939,3.485],[-26.752,46.984,1.836],[-25.282,45.180,11.193],
[-32.126,44.905,1.937],[-30.097,41.725,-2.884],[-22.872,42.423,9.294],
[-25.470,43.693,1.858],[-27.513,40.252,9.796],[-35.539,43.152,4.611],
[-32.198,46.833,5.146],[-34.170,41.895,6.678],[-29.824,48.995,7.935],
[-33.828,46.680,10.854],[-28.172,42.725,-1.249],[-37.111,42.849,10.126],
[-30.550,49.763,2.732],[-35.616,48.558,12.616],[-25.924,33.331,-1.649],
[-35.737,41.800,0.985],[-23.417,48.598,-2.188],[-40.768,47.139,9.232],
[-39.554,48.652,7.351],[-22.030,47.955,10.985],[-32.096,48.476,12.157],
[-29.981,52.020,0.983],[-36.254,50.698,3.725],[-28.690,49.763,-1.726],
[-27.292,50.004,0.555],[-24.972,39.071,9.357],[-39.062,46.032,11.590],
[-36.646,35.125,8.608],[-32.235,35.746,12.613],[-25.772,38.560,12.817],
[-25.181,43.233,-2.474],[-31.035,38.433,-2.010],[-34.860,39.808,9.210],
[-34.724,40.452,3.019],[-30.920,41.182,-0.022],[-29.638,40.675,12.637],
[-29.468,40.541,4.822],[-37.519,40.145,10.114],[-34.221,36.017,-3.667],
[-40.936,37.557,3.464],[-40.255,38.386,-2.324],[-29.476,42.130,8.488],
[-35.273,35.616,6.312],[-37.052,39.025,2.834],[-35.806,38.909,0.459],
[-28.534,36.718,7.844],[-35.723,41.008,-1.822],[-26.098,50.771,13.067],
[-36.381,45.919,12.753],[-31.442,39.668,6.820],[-37.220,36.513,1.017],
[-32.661,38.508,9.247],[-32.024,43.659,-0.833],[-25.388,43.930,7.130],
[-28.513,36.812,-2.029],[-29.827,45.081,10.969],[-28.303,39.883,-0.282],
[-37.789,44.041,5.476],[-29.622,45.111,-1.093],[-39.725,41.939,11.134],
[-32.981,41.359,13.449],[-29.444,47.433,12.126],[-22.481,39.356,0.229],
[-22.501,45.303,12.854],[-25.250,47.729,-0.494],[-27.211,42.917,4.052],
[-35.391,47.676,1.364],[-38.437,43.248,-2.335],[-34.078,46.160,3.237],
[-33.164,49.986,9.967],[-33.963,49.921,2.912],[-27.172,45.979,9.037],
[-26.816,45.875,-2.469],[-26.256,48.177,11.634],[-22.272,51.897,-1.471],
[-39.499,37.390,11.462],[-27.228,44.792,-0.047],[-38.450,48.889,10.884],
[-33.247,47.731,-3.435],[-33.904,47.917,-0.795],[-37.339,43.843,0.091],
[-30.229,35.449,9.496],[-30.019,35.894,1.493],[-30.777,47.880,-0.421],
[-31.492,38.501,0.778],[-30.002,39.397,9.974],[-29.460,37.721,3.145],
[-33.950,50.693,-0.647],[-35.223,35.605,-0.912],[-23.440,37.540,11.032],
[-26.409,41.119,5.763],[-36.174,46.538,-1.282],[-37.492,39.296,-2.883],
[-24.521,49.144,6.884],[-37.717,34.102,5.939],[-38.266,47.834,3.103],
[-39.477,50.299,14.314],[-33.091,50.249,14.565],[-31.001,48.996,-3.498],
[-28.267,48.966,13.927],[-26.636,48.528,-2.873],[-37.893,43.641,13.274],
[-35.842,45.989,-4.021],[-31.857,45.571,12.776],[-30.859,45.208,-3.667],
[-28.030,43.000,10.754],[-38.114,41.035,13.987],[-35.021,43.793,-1.391],
[-35.234,41.044,11.561],[-33.062,41.762,-2.788],[-28.849,43.672,13.309],
[-22.893,44.709,-2.838],[-34.220,37.504,12.408],[-35.080,37.029,9.788],
[-32.283,42.289,9.077],[-28.616,39.503,15.226],[-26.130,41.378,12.357],
[-35.352,40.377,14.223],[-32.863,34.934,8.942],[-30.832,38.142,13.631],
[-28.772,37.721,11.861],[-28.314,36.698,15.013],[-37.064,38.688,12.237],
[-34.416,33.811,13.021],[-31.283,34.089,14.588],[-33.962,52.760,-2.274],
[-39.555,47.342,-2.216],[-41.236,51.958,1.433],[-38.850,50.347,4.799],
[-22.509,51.201,6.702],[-23.412,50.832,9.281],[-25.362,48.038,14.361],
[-24.232,44.971,-0.330],[-22.760,42.147,0.939],[-24.148,46.302,6.705],
[-24.750,47.753,9.187],[-22.992,41.717,-2.165],[-24.032,39.951,-3.699],
[-25.289,41.161,3.281],[-41.007,43.063,6.389],[-24.324,43.416,13.921],
[-23.054,41.329,12.491],[-25.406,36.639,-2.966],[-24.757,37.743,-0.378],
[-24.436,38.622,6.580],[-24.746,41.102,7.632],[-41.491,38.953,12.534],
[-19.735,34.546,14.005],[-23.852,36.115,1.508],[-39.256,36.367,5.603],
[-41.861,40.389,7.355],[-21.300,36.798,12.900],[-22.698,36.307,-3.112],
[-40.488,36.430,0.149],[-24.956,52.770,2.756],[-39.295,33.422,7.772],
[-37.732,47.968,14.199],[-36.286,48.443,8.502],[-30.264,50.954,12.978],
[-27.213,53.722,11.351],[-21.956,33.503,11.973],[-37.324,52.575,9.282],
[-34.758,51.977,8.563],[-31.621,51.264,8.251],[-28.727,49.993,10.843],
[-25.808,50.968,10.328],[-36.820,50.730,6.937],[-35.392,48.354,5.205],
[-34.301,51.471,5.470],[-30.457,35.873,4.964],[-25.459,34.668,3.937],
[-21.735,34.816,4.293],[-37.135,36.002,3.924],[-38.849,32.967,-0.171],
[-33.204,35.615,3.313],[-27.892,34.552,5.952],[-25.848,33.890,1.004],
[-39.525,50.012,-0.111],[-38.234,36.283,-1.625],[-32.084,36.050,-0.114],
[-30.818,51.510,-1.580],[-24.382,50.432,0.210],[-37.278,34.215,13.213],
[-33.799,35.058,15.231],[-38.444,50.205,-2.609],[-28.734,33.840,13.723],
[-23.110,50.955,12.111],[-22.355,55.968,6.085],[-27.146,58.793,7.086],
[-36.270,56.412,4.569],[-26.686,56.243,2.098],[-39.182,-59.224,5.309],
[-22.942,57.642,8.222],[-38.665,55.002,8.721],[-24.777,56.121,4.396],
[-35.924,55.216,8.660],[-25.179,53.364,9.006],[-32.167,53.240,6.115],
[-24.860,52.201,-2.359],[-24.707,57.068,13.620],[-34.519,54.181,4.963],
[-30.062,52.686,9.991],[-26.799,57.748,10.947],[-30.504,55.134,-3.881],
[-34.922,57.573,8.279],[-34.161,58.935,3.806],[-29.343,58.366,3.274],
[-36.565,-58.867,9.710],[-35.359,55.318,-2.433],[-40.927,56.396,4.685],
[-39.499,57.302,-2.228],[-33.950,56.311,3.073],[-38.596,55.292,3.248],
[-33.343,57.411,0.826],[-28.666,54.763,8.589],[-34.405,57.810,-2.542],
[-31.263,56.402,6.432],[-36.262,54.721,1.630],[-30.952,58.792,7.871],
[-26.546,57.350,-2.637],[-27.666,58.757,-0.321],[-22.404,58.873,-0.553],
[-38.746,55.417,11.396],[-29.109,57.515,9.374],[-28.292,54.948,0.155],
[-31.133,58.722,1.114],[-31.698,57.539,10.813],[-31.254,56.384,3.323],
[-33.876,53.868,0.272],[-24.089,57.219,10.684],[-37.090,-59.811,-2.788],
[-36.802,53.422,6.614],[-34.314,57.295,14.041],[-31.592,54.562,11.513],
[-29.088,57.331,15.245],[-36.338,56.661,15.898],[-32.702,53.810,8.734],
[-31.915,58.002,15.535],[-29.391,57.995,12.554],[-28.308,54.256,15.069],
[-37.955,55.405,14.038],[-32.509,52.067,11.793],[-32.067,55.091,14.455],
[-23.968,57.980,-2.482],[-24.191,54.529,-3.520],[-24.937,58.753,0.706],
[-24.549,-59.622,6.724],[-26.241,-57.965,9.796],[-40.695,57.149,12.391],
[-23.736,52.980,13.794],[-24.203,54.504,-0.092],[-38.139,57.285,6.792],
[-41.603,-58.559,7.600],[-20.991,53.099,12.610],[-21.734,54.329,-2.528],
[-40.450,55.293,0.168],[-39.689,52.044,10.341],[-23.380,54.351,11.306],
[-29.392,54.526,4.675],[-24.773,50.693,4.736],[-24.701,53.232,6.155],
[-37.866,52.706,4.206],[-39.339,52.711,-0.307],[-33.019,52.471,2.906],
[-27.441,52.548,5.160],[-27.084,52.736,0.873],[-37.306,52.879,-2.687],
[-30.788,54.550,-0.906],[-37.008,53.311,12.573],[-35.225,54.470,14.423],
[-28.562,51.614,15.098],[-28.485,-50.240,22.559],[-30.445,-49.296,24.911],
[-37.059,-50.831,27.117],[-25.552,-48.901,23.975],[-36.483,-46.800,24.775],
[-38.908,-52.743,20.532],[-40.002,-50.696,22.876],[-28.012,-43.140,23.755],
[-32.314,-48.020,26.554],[-22.611,-56.121,27.726],[-27.165,-54.781,24.728],
[-34.195,-54.788,23.080],[-26.480,-55.029,22.263],[-39.036,-53.478,23.148],
[-32.885,-52.275,22.314],[-21.598,-53.481,28.197],[-38.515,-55.449,25.558],
[-27.309,-47.732,22.182],[-33.936,-48.834,28.595],[-24.856,-57.054,21.123],
[-36.836,-53.041,25.002],[-34.956,-46.476,27.362],[-39.019,-45.427,19.010],
[-25.116,-46.727,25.817],[-25.642,-56.633,26.558],[-31.863,-59.173,25.759],
[-26.362,-45.264,23.676],[-26.311,-45.020,20.749],[-24.210,-45.279,29.919],
[-29.714,-51.462,17.924],[-28.867,-53.160,15.709],[-23.131,-51.354,27.519],
[-25.601,-49.464,21.059],[-27.286,-52.233,29.292],[-35.124,-49.665,22.201],
[-33.893,-45.788,24.961],[-34.236,-49.616,25.247],[-27.633,-44.934,26.062],
[-34.001,-45.454,30.095],[-28.032,-50.245,19.724],[-36.324,-50.085,29.660],
[-31.615,-44.759,20.907],[-35.219,-43.032,31.965],[-26.071,-57.554,17.776],
[-38.320,-48.613,22.271],[-23.425,-44.092,15.599],[-39.202,-46.585,29.608],
[-40.770,-45.689,26.242],[-21.716,-45.300,29.390],[-31.466,-45.412,30.933],
[-30.472,-41.002,19.729],[-36.792,-42.637,23.664],[-27.694,-41.881,18.004],
[-27.443,-42.336,20.757],[-24.537,-52.261,32.169],[-37.177,-44.916,29.738],
[-34.409,-57.715,26.916],[-30.203,-57.085,28.884],[-27.048,-52.696,32.855],
[-25.973,-50.748,14.649],[-31.289,-55.160,17.186],[-34.345,-53.027,26.273],
[-35.211,-51.820,20.194],[-30.559,-51.828,20.612],[-30.252,-51.890,32.303],
[-29.364,-52.674,22.949],[-36.494,-53.422,27.793],[-34.557,-54.513,15.159],
[-40.652,-55.766,22.691],[-38.350,-54.508,18.252],[-29.081,-51.899,26.316],
[-31.695,-55.160,21.979],[-36.197,-53.317,22.351],[-35.165,-55.322,18.145],
[-28.479,-56.460,26.646],[-36.016,-52.628,16.954],[-25.791,-40.229,31.792],
[-36.670,-47.816,31.465],[-31.338,-53.733,24.291],[-36.145,-55.927,21.484],
[-31.693,-52.870,26.691],[-31.020,-48.790,19.465],[-24.890,-49.100,27.218],
[-27.729,-55.174,17.112],[-29.279,-47.078,30.385],[-27.464,-53.248,18.943],
[-38.275,-49.475,25.255],[-29.307,-46.154,18.869],[-38.574,-51.999,29.424],
[-34.079,-49.857,31.494],[-28.694,-44.359,30.398],[-23.696,-53.445,17.895],
[-21.921,-48.265,29.931],[-24.382,-45.174,18.428],[-27.297,-49.566,25.996],
[-36.196,-45.988,21.219],[-37.133,-49.962,16.514],[-32.825,-47.779,21.459],
[-31.704,-43.359,28.788],[-35.464,-42.340,20.752],[-29.250,-47.642,26.587],
[-27.460,-47.109,16.569],[-26.498,-42.354,30.485],[-24.486,-42.618,17.806],
[-37.608,-55.708,28.994],[-26.940,-47.737,19.145],[-39.352,-44.021,28.349],
[-32.786,-46.327,14.523],[-34.037,-46.010,19.542],[-37.141,-48.044,18.579],
[-29.476,-54.347,28.026],[-29.107,-56.305,19.092],[-31.511,-45.202,17.076],
[-33.377,-54.083,20.185],[-31.451,-52.927,29.603],[-29.965,-57.322,21.298],
[-34.935,-42.744,16.383],[-34.328,-57.341,16.260],[-24.104,-54.723,30.726],
[-26.702,-51.747,24.973],[-36.339,-45.030,17.628],[-38.750,-53.072,15.556],
[-25.165,-43.137,26.072],[-35.483,-59.307,23.895],[-38.312,-43.751,21.459],
[-38.283,-42.834,32.001],[-35.846,-44.038,34.670],[-30.178,-44.987,14.536],
[-28.304,-46.413,32.910],[-26.677,-43.814,16.396],[-40.216,-49.432,33.441],
[-34.888,-48.180,15.739],[-31.642,-49.592,32.577],[-31.228,-47.688,16.761],
[-27.020,-47.961,31.086],[-37.571,-50.360,32.391],[-32.351,-50.350,17.576],
[-34.457,-52.306,29.460],[-33.154,-52.219,15.748],[-29.052,-49.053,32.473],
[-23.928,-48.537,15.848],[-32.982,-54.955,31.662],[-33.272,-56.031,28.940],
[-31.617,-50.211,28.898],[-29.027,-54.146,33.920],[-26.354,-50.646,31.331],
[-35.775,-52.681,32.076],[-33.158,-59.033,29.483],[-30.814,-56.914,31.600],
[-28.965,-54.096,30.947],[-27.149,-55.696,32.599],[-36.518,-55.069,31.695],
[-33.361,-59.174,32.269],[-33.251,-57.821,34.669],[-33.498,-39.929,17.181],
[-38.907,-46.841,16.495],[-39.950,-41.656,21.021],[-39.793,-41.814,23.623],
[-21.372,-41.867,25.631],[-23.279,-42.048,29.683],[-25.412,-45.962,32.568],
[-23.671,-47.818,18.521],[-19.885,-48.737,21.867],[-22.652,-47.311,27.427],
[-26.098,-44.368,28.284],[-22.449,-49.980,17.727],[-24.146,-53.254,15.065],
[-26.279,-52.451,21.535],[-40.869,-49.135,25.363],[-24.371,-48.392,31.728],
[-21.720,-52.427,32.734],[-24.927,-56.235,15.527],[-24.380,-54.330,20.782],
[-23.589,-53.852,25.999],[-25.952,-54.300,28.432],[-38.521,-53.352,32.034],
[-20.786,-58.558,33.005],[-22.811,-58.121,19.748],[-41.067,-56.545,25.237],
[-40.436,-53.244,25.965],[-23.549,-57.695,32.185],[-22.637,-57.705,16.569],
[-40.409,-57.189,19.683],[-24.768,-42.603,23.256],[-39.818,-57.586,27.485],
[-38.000,-45.410,33.015],[-35.380,-43.690,28.204],[-30.928,-42.459,31.479],
[-28.802,-40.736,31.506],[-22.161,-58.033,29.697],[-37.819,-40.048,28.477],
[-37.301,-42.286,26.999],[-30.637,-40.165,26.354],[-28.337,-42.231,26.998],
[-25.555,-41.081,27.654],[-38.702,-44.155,25.312],[-34.914,-43.515,25.794],
[-33.618,-42.754,23.266],[-31.394,-56.627,24.892],[-25.151,-58.285,23.853],
[-22.599,-57.997,22.941],[-38.535,-57.494,21.997],[-38.515,-58.245,17.989],
[-31.439,-58.783,23.042],[-28.250,-57.515,24.083],[-27.372,-57.387,21.587],
[-37.981,-42.964,18.981],[-37.140,-56.260,16.255],[-32.192,-57.488,18.031],
[-30.179,-41.497,16.855],[-23.225,-43.166,20.366],[-37.003,-57.730,30.687],
[-36.496,-58.539,33.131],[-37.300,-41.491,16.664],[-25.371,59.013,33.359],
[-24.211,-43.573,32.208],[-28.312,-32.615,20.483],[-29.357,-31.081,25.705],
[-36.603,-32.794,26.313],[-27.548,-29.354,23.129],[-37.328,-27.794,24.492],
[-40.574,-34.147,20.168],[-40.179,-32.114,22.218],[-28.063,-24.306,23.961],
[-31.758,-29.372,25.939],[-23.605,-38.625,26.470],[-25.875,-35.682,26.633],
[-36.338,-37.571,23.487],[-27.726,-38.210,22.206],[-38.706,-34.126,23.601],
[-34.654,-33.310,23.826],[-23.602,-34.255,30.296],[-36.895,-37.198,28.140],
[-29.945,-28.321,21.811],[-33.957,-30.381,27.559],[-26.025,-37.951,24.703],
[-37.745,-35.163,25.974],[-35.651,-29.780,29.508],[-38.811,-27.396,19.355],
[-24.941,-26.456,25.390],[-26.113,-38.172,27.872],[-33.457,-39.796,25.255],
[-25.901,-26.148,21.847],[-29.021,-27.269,19.395],[-26.318,-28.241,30.187],
[-29.482,-31.257,22.984],[-29.401,-33.495,16.147],[-22.223,-32.582,27.878],
[-25.470,-31.176,21.918],[-27.859,-33.801,29.635],[-35.327,-30.664,22.085],
[-34.518,-28.309,25.171],[-33.088,-31.970,25.603],[-29.670,-26.659,28.345],
[-34.172,-27.463,30.731],[-27.631,-32.027,17.964],[-37.286,-32.234,29.070],
[-29.328,-25.109,21.321],[-36.200,-25.651,30.643],[-23.315,-40.329,18.526],
[-35.972,-32.063,19.820],[-25.334,-23.859,17.920],[-41.036,-29.183,28.257],
[-40.869,-26.834,26.508],[-23.797,-25.399,31.106],[-31.279,-28.313,29.933],
[-31.021,-22.923,20.912],[-37.483,-24.467,23.274],[-29.064,-23.848,17.241],
[-27.457,-24.736,19.572],[-24.951,-35.490,32.358],[-37.661,-28.270,30.986],
[-35.190,-38.692,26.968],[-30.240,-41.000,28.936],[-27.589,-36.978,30.584],
[-26.971,-30.884,15.620],[-31.875,-36.223,17.848],[-35.188,-35.684,25.906],
[-35.451,-34.069,21.372],[-29.811,-34.950,19.768],[-30.914,-34.640,32.449],
[-27.963,-35.435,23.614],[-37.318,-35.016,29.796],[-35.151,-36.811,15.730],
[-41.946,-37.448,22.900],[-39.419,-36.457,19.643],[-30.191,-33.488,27.002],
[-32.988,-37.290,24.436],[-36.518,-36.350,19.911],[-34.113,-35.347,18.967],
[-29.246,-37.821,26.778],[-36.261,-33.445,16.821],[-24.587,-21.432,32.695],
[-35.713,-30.907,32.244],[-32.009,-34.375,24.940],[-35.726,-38.874,21.189],
[-33.217,-36.150,27.519],[-32.170,-30.229,18.301],[-24.562,-31.338,26.852],
[-28.831,-36.420,17.356],[-31.236,-30.683,28.355],[-26.736,-34.597,17.334],
[-38.259,-30.401,23.245],[-30.688,-28.221,17.488],[-40.655,-31.514,29.592],
[-33.165,-32.691,31.390],[-28.772,-27.517,30.868],[-22.735,-33.523,18.851],
[-23.050,-28.002,30.844],[-25.416,-26.948,18.864],[-26.140,-30.777,24.803],
[-34.305,-27.391,20.956],[-40.033,-30.980,17.301],[-32.791,-29.759,22.846],
[-33.041,-26.527,27.434],[-35.306,-24.509,21.047],[-28.405,-29.149,28.388],
[-27.868,-28.280,16.259],[-27.030,-24.960,30.702],[-22.913,-23.786,16.150],
[-39.219,-35.976,27.682],[-26.438,-29.517,19.471],[-38.416,-27.584,27.962],
[-33.480,-26.879,15.289],[-33.864,-28.053,17.836],[-37.673,-30.538,18.293],
[-30.313,-38.145,29.438],[-27.495,-37.112,19.585],[-32.511,-24.875,19.902],
[-31.433,-36.001,22.551],[-32.086,-36.030,30.284],[-30.294,-39.237,21.707],
[-34.678,-25.442,18.232],[-36.034,-38.794,18.407],[-23.883,-37.728,31.169],
[-26.604,-33.629,25.198],[-37.353,-27.021,16.990],[-38.694,-33.976,16.268],
[-26.703,-24.596,26.412],[-37.500,-39.577,25.883],[-37.400,-26.817,21.806],
[-38.169,-23.556,33.015],[-34.912,-24.745,34.304],[-30.726,-25.749,15.675],
[-28.161,-27.408,33.481],[-28.023,-25.359,14.588],[-39.680,-29.984,32.209],
[-35.367,-29.635,15.092],[-32.533,-30.036,32.273],[-30.359,-29.963,15.383],
[-27.361,-31.212,29.952],[-38.195,-33.051,32.623],[-34.640,-31.298,17.376],
[-34.598,-33.288,28.958],[-32.765,-32.314,15.719],[-29.165,-32.381,31.677],
[-23.341,-31.059,16.262],[-34.895,-36.844,29.650],[-33.360,-39.005,28.989],
[-31.228,-33.454,29.698],[-28.052,-35.020,32.782],[-26.471,-32.610,32.231],
[-35.412,-34.395,33.364],[-32.652,-41.526,27.190],[-32.351,-36.289,33.763],
[-29.518,-37.208,32.822],[-26.476,-39.006,33.961],[-37.035,-36.349,32.770],
[-31.489,-39.827,31.486],[-33.067,-39.868,33.971],[-34.123,-21.804,17.473],
[-39.996,-28.111,16.672],[-40.588,-22.215,20.634],[-40.651,-23.733,22.906],
[-24.054,-23.962,25.318],[-24.868,-23.066,30.038],[-25.782,-28.825,33.647],
[-24.001,-28.283,16.834],[-23.043,-29.082,21.459],[-23.861,-28.703,26.470],
[-25.219,-26.638,27.988],[-23.749,-30.912,18.933],[-23.788,-34.148,16.030],
[-25.168,-33.836,22.414],[-41.219,-31.291,24.561],[-24.639,-30.256,31.386],
[-23.478,-32.673,32.813],[-25.333,-36.921,15.888],[-24.370,-35.691,19.396],
[-22.328,-35.989,26.299],[-24.765,-33.738,28.066],[-41.671,-36.141,32.478],
[-22.101,-39.240,34.044],[-25.122,-38.528,20.329],[-38.848,-37.327,24.546],
[-39.977,-33.647,26.024],[-20.714,-38.721,31.455],[-22.848,-37.536,16.528],
[-39.844,-38.764,20.924],[-25.883,-23.368,21.417],[-39.870,-38.966,29.674],
[-36.577,-26.859,33.213],[-35.829,-26.127,27.375],[-31.081,-24.783,30.593],
[-28.918,-22.640,30.683],[-24.215,-39.484,29.362],[-37.820,-22.550,27.673],
[-36.874,-21.072,29.809],[-32.033,-20.055,25.181],[-28.779,-24.241,28.347],
[-27.087,-21.991,28.651],[-38.016,-24.835,26.014],[-35.247,-26.028,23.936],
[-31.143,-24.760,26.652],[-30.731,-38.670,24.348],[-24.509,-40.376,24.937],
[-23.677,-40.154,22.190],[-37.480,-40.139,22.966],[-39.275,-40.085,18.420],
[-33.407,-40.085,22.602],[-27.333,-40.561,24.394],[-26.361,-39.563,18.268],
[-39.948,-24.758,18.605],[-37.172,-38.521,15.729],[-32.818,-37.479,20.265],
[-31.958,-23.016,17.677],[-23.431,-24.385,20.142],[-37.569,-40.216,31.435],
[-35.326,-38.743,32.595],[-36.124,-22.941,15.454],[-28.763,-42.268,33.705],
[-25.282,-24.076,33.033],[-31.476,-10.683,23.427],[-31.519,-13.957,26.289],
[-38.184,-14.684,25.338],[-26.655,-8.980,23.571],[-36.914,-9.755,24.686],
[-39.322,-15.502,20.403],[-39.319,-14.039,22.894],[-28.930,-5.774,24.247],
[-31.836,-8.701,25.573],[-23.869,-19.082,27.158],[-27.328,-17.652,26.223],
[-35.203,-18.458,24.069],[-27.840,-18.710,23.074],[-39.991,-16.262,24.093],
[-35.432,-15.355,24.691],[-22.486,-15.936,29.001],[-36.963,-19.080,27.331],
[-28.409,-9.589,21.522],[-34.825,-11.578,28.013],[-24.932,-19.483,23.114],
[-37.528,-17.237,24.580],[-36.718,-9.341,28.491],[-39.054,-9.739,18.774],
[-27.245,-10.017,26.133],[-26.186,-21.068,26.302],[-33.466,-22.118,24.610],
[-26.090,-6.174,22.917],[-26.348,-8.034,20.759],[-26.416,-9.438,30.759],
[-28.787,-12.072,20.647],[-29.723,-14.251,14.424],[-22.754,-13.150,28.248],
[-25.693,-13.552,19.604],[-27.452,-15.861,28.819],[-36.477,-12.548,21.358],
[-34.257,-8.464,24.209],[-35.737,-12.517,24.318],[-29.384,-7.830,26.127],
[-34.000,-8.250,29.757],[-28.502,-13.783,18.393],[-36.955,-13.520,28.936],
[-31.188,-6.415,20.531],[-35.498,-5.941,30.227],[-24.649,-21.402,16.985],
[-35.397,-14.263,19.342],[-21.895,-7.358,16.238],[-40.321,-9.628,27.976],
[-40.865,-6.974,27.471],[-21.673,-8.264,30.437],[-33.020,-6.656,31.604],
[-30.170,-3.787,19.222],[-37.014,-5.818,22.059],[-28.241,-4.897,17.212],
[-27.827,-5.925,19.896],[-24.737,-14.903,32.885],[-37.959,-7.432,29.962],
[-35.311,-22.255,26.940],[-30.230,-21.923,28.360],[-28.022,-17.621,30.981],
[-25.260,-13.655,15.145],[-30.982,-18.633,17.498],[-34.997,-16.586,26.843],
[-34.576,-14.571,22.174],[-27.252,-15.327,21.090],[-29.152,-15.269,31.918],
[-28.581,-16.423,23.977],[-38.066,-16.469,27.863],[-33.280,-19.044,14.760],
[-40.128,-18.636,21.528],[-40.058,-17.622,18.727],[-29.222,-15.580,26.740],
[-32.914,-19.027,22.428],[-37.246,-15.910,22.017],[-34.692,-17.233,21.347],
[-28.618,-18.930,28.669],[-34.865,-15.223,16.952],[-23.774,-3.698,31.762],
[-36.920,-10.889,31.021],[-31.428,-17.298,26.002],[-37.157,-18.758,18.668],
[-32.621,-16.831,28.468],[-30.494,-11.962,16.875],[-25.975,-12.646,26.124],
[-27.764,-20.366,16.790],[-31.675,-9.228,29.614],[-27.975,-16.227,17.166],
[-38.427,-11.553,22.626],[-28.393,-8.465,18.813],[-39.858,-14.017,30.102],
[-32.854,-13.595,32.508],[-28.985,-9.315,29.893],[-22.103,-17.013,18.633],
[-21.983,-10.868,30.968],[-25.351,-7.776,17.807],[-28.289,-13.076,25.017],
[-35.541,-8.001,21.803],[-37.222,-12.817,16.020],[-34.043,-11.231,22.861],
[-32.770,-7.280,27.648],[-34.744,-4.818,21.090],[-29.832,-10.613,26.122],
[-27.646,-9.957,16.430],[-26.044,-6.118,30.246],[-23.265,-4.373,15.730],
[-39.635,-19.451,28.057],[-26.462,-10.620,19.047],[-38.356,-7.986,26.457],
[-32.404,-7.807,15.868],[-33.814,-9.032,17.675],[-38.000,-12.358,18.932],
[-31.176,-18.913,28.854],[-27.695,-19.913,19.693],[-31.098,-8.670,18.357],
[-29.368,-17.833,19.986],[-30.732,-16.512,30.271],[-30.824,-20.085,21.370],
[-35.399,-3.677,17.584],[-33.231,-17.736,18.511],[-24.620,-18.639,31.497],
[-26.238,-15.358,24.782],[-37.025,-9.161,17.243],[-37.385,-15.504,15.407],
[-26.779,-5.793,26.399],[-37.169,-21.304,24.860],[-37.968,-8.166,20.886],
[-39.114,-5.653,32.207],[-35.278,-6.176,32.987],[-29.781,-7.452,16.372],
[-28.242,-6.243,31.768],[-26.542,-7.202,15.067],[-39.785,-11.794,32.247],
[-35.645,-9.457,14.897],[-33.970,-10.853,31.040],[-30.429,-10.083,14.582],
[-27.678,-12.839,29.819],[-36.566,-14.553,31.786],[-34.396,-12.183,17.527],
[-34.049,-15.746,30.432],[-32.534,-14.102,15.891],[-30.622,-11.855,32.535],
[-22.887,-13.002,16.534],[-35.143,-18.028,31.855],[-34.232,-18.900,28.079],
[-32.760,-13.005,29.742],[-30.092,-17.089,35.780],[-26.896,-13.431,32.252],
[-37.440,-15.437,34.208],[-33.008,-21.320,28.280],[-34.376,-19.460,33.928],
[-30.725,-18.155,33.006],[-27.596,-19.771,32.246],[-38.092,-18.711,30.886],
[-34.213,-22.352,31.901],[-32.353,-22.071,33.748],[-32.742,-3.027,17.060],
[-39.579,-8.520,16.275],[-39.601,-4.628,20.687],[-40.076,-6.372,22.815],
[-20.263,-4.371,27.563],[-23.031,-6.039,29.557],[-24.832,-9.580,33.091],
[-23.619,-11.173,18.735],[-23.528,-12.861,20.725],[-24.845,-10.140,26.472],
[-25.046,-8.447,28.713],[-23.600,-14.723,18.340],[-24.010,-16.732,16.607],
[-24.979,-14.639,22.579],[-38.959,-11.907,25.363],[-24.923,-12.356,29.719],
[-21.973,-14.348,33.601],[-26.226,-18.151,17.144],[-25.743,-17.460,19.917],
[-24.876,-17.607,25.333],[-24.872,-15.576,27.422],[-40.628,-17.780,31.451],
[-22.882,-20.706,34.589],[-24.534,-21.687,20.069],[-39.468,-19.598,24.631],
[-39.765,-13.961,27.365],[-22.041,-20.629,31.594],[-23.577,-19.316,15.749],
[-39.832,-20.479,18.496],[-25.322,-5.157,20.400],[-40.518,-22.032,27.344],
[-37.833,-7.597,33.740],[-34.731,-8.988,26.765],[-31.018,-4.405,31.849],
[-26.550,-3.020,31.111],[-23.078,-21.728,27.245],[-38.039,-2.239,29.013],
[-35.838,-3.663,28.666],[-31.673,-3.978,26.532],[-29.030,-4.226,26.590],
[-26.773,-3.401,27.903],[-38.597,-4.737,24.646],[-36.124,-5.805,24.666],
[-33.636,-6.146,25.369],[-30.480,-18.575,23.577],[-26.355,-22.120,23.791],
[-23.655,-22.463,22.832],[-38.600,-21.835,22.414],[-38.593,-23.234,16.487],
[-34.757,-20.873,22.451],[-28.983,-21.049,24.079],[-29.938,-21.329,17.847],
[-38.844,-6.444,18.913],[-38.264,-20.599,15.777],[-32.601,-20.462,19.594],
[-31.515,-5.580,17.811],[-23.590,-5.462,18.228],[-38.860,-21.687,31.439],
[-36.760,-21.123,33.525],[-37.608,-5.501,16.415],[-27.583,-22.229,33.452],
[-24.260,-7.236,31.857],[-28.357,5.531,21.554],[-29.540,5.483,25.656],
[-35.586,3.718,27.420],[-26.440,8.100,22.592],[-36.310,9.403,24.021],
[-38.347,3.309,19.365],[-38.863,4.891,22.550],[-29.602,14.243,23.432],
[-30.954,8.966,24.684],[-21.887,0.609,27.047],[-28.384,1.457,25.608],
[-35.558,1.394,23.149],[-26.918,0.497,21.886],[-39.187,1.788,22.323],
[-33.298,5.280,21.341],[-22.427,2.437,29.190],[-37.837,-0.511,26.696],
[-30.561,9.747,20.977],[-35.816,6.461,28.014],[-25.274,-0.764,23.526],
[-35.499,0.812,26.163],[-36.221,8.667,26.781],[-38.057,9.440,19.128],
[-25.250,8.980,25.559],[-24.740,-1.955,26.239],[-34.023,-3.380,24.700],
[-25.403,10.855,23.022],[-27.497,10.197,20.625],[-25.347,8.404,30.266],
[-30.714,5.537,19.475],[-28.648,3.733,16.992],[-22.929,4.974,28.058],
[-24.672,6.355,21.608],[-27.100,4.259,28.398],[-35.223,7.614,22.318],
[-34.454,10.695,25.490],[-33.689,6.700,24.978],[-27.254,10.989,26.156],
[-33.315,9.006,29.269],[-27.388,4.156,19.341],[-37.249,5.869,30.127],
[-29.227,12.180,21.560],[-34.534,10.901,31.355],[-24.483,-2.000,16.674],
[-35.950,5.287,20.745],[-23.768,11.685,16.126],[-40.747,7.382,27.881],
[-40.570,11.628,26.519],[-21.298,12.030,29.828],[-31.771,10.901,30.593],
[-30.210,14.249,19.494],[-35.530,14.026,21.781],[-26.406,13.357,16.651],
[-27.764,12.871,19.350],[-24.755,2.854,30.342],[-36.818,9.649,29.401],
[-36.600,-2.761,25.688],[-30.083,-4.177,29.013],[-28.315,0.890,28.943],
[-26.212,4.641,15.800],[-29.924,-0.185,15.945],[-36.504,3.979,24.875],
[-34.031,2.803,21.392],[-29.779,2.973,19.518],[-30.938,2.741,32.885],
[-29.591,2.195,23.091],[-37.819,2.378,29.596],[-35.382,1.788,15.161],
[-39.133,-1.146,22.660],[-37.913,-0.380,20.079],[-31.644,5.193,27.438],
[-32.658,0.228,21.526],[-36.800,2.696,21.336],[-35.049,0.316,19.826],
[-29.161,-1.886,24.777],[-34.719,2.353,17.766],[-24.971,14.283,31.562],
[-37.622,7.957,31.670],[-33.154,3.206,24.134],[-35.706,-1.056,22.008],
[-30.770,2.884,25.969],[-32.285,7.453,18.760],[-24.177,6.401,26.088],
[-27.248,-1.754,16.667],[-31.055,8.305,27.492],[-27.415,1.258,19.501],
[-36.953,6.220,23.580],[-30.014,8.195,17.155],[-39.207,5.270,28.771],
[-33.443,3.463,31.948],[-29.739,7.970,30.888],[-22.582,2.521,18.456],
[-22.445,7.939,31.628],[-24.998,9.169,20.135],[-26.871,5.854,25.616],
[-35.515,10.161,21.574],[-38.426,5.526,15.812],[-32.620,7.804,21.648],
[-31.436,12.394,27.185],[-33.815,12.184,20.567],[-28.674,9.893,28.428],
[-27.100,8.140,17.033],[-25.123,12.774,29.004],[-22.703,14.129,17.488],
[-39.245,0.251,28.764],[-26.142,6.691,19.386],[-38.390,11.709,28.075],
[-33.593,10.551,15.488],[-34.445,9.184,17.785],[-36.134,6.955,18.234],
[-30.500,-1.208,28.418],[-28.844,-1.297,19.473],[-31.913,10.470,17.802],
[-32.460,1.406,19.043],[-29.782,3.241,28.450],[-30.288,-2.049,21.577],
[-33.875,12.142,17.929],[-34.347,-0.785,17.304],[-25.739,-0.545,28.465],
[-26.296,3.441,24.460],[-36.575,11.323,17.348],[-36.261,3.973,16.176],
[-23.957,14.838,27.390],[-39.699,-2.437,25.580],[-37.963,11.359,21.997],
[-37.645,12.495,30.937],[-34.823,13.138,32.757],[-30.488,11.021,15.322],
[-28.007,9.952,31.661],[-25.958,10.021,15.598],[-39.811,6.815,32.797],
[-34.915,7.112,15.759],[-32.925,8.777,31.982],[-32.263,8.183,15.764],
[-27.967,6.943,29.025],[-38.442,4.376,32.180],[-33.580,5.155,18.425],
[-33.217,5.100,29.563],[-32.305,5.522,15.389],[-28.678,6.481,33.177],
[-23.635,7.145,13.803],[-33.165,-0.960,32.214],[-33.305,-0.655,29.667],
[-30.957,4.328,30.739],[-28.020,2.419,33.053],[-27.114,4.769,31.228],
[-35.848,3.765,30.914],[-33.160,-3.091,28.525],[-31.786,1.734,29.820],
[-30.041,0.297,31.973],[-27.691,-0.884,32.451],[-37.386,-0.426,32.502],
[-33.557,-3.862,31.213],[-30.588,-2.544,33.844],[-33.072,15.075,17.212],
[-39.278,9.808,16.842],[-40.580,14.507,19.737],[-39.148,13.739,23.540],
[-22.580,12.442,24.521],[-22.496,12.361,27.386],[-24.621,9.420,32.738],
[-23.154,6.410,19.118],[-23.432,4.064,20.883],[-22.540,8.359,25.782],
[-24.091,9.934,28.548],[-23.021,5.276,16.409],[-23.944,2.812,15.985],
[-26.312,3.652,21.787],[-40.339,6.293,24.903],[-24.448,5.527,31.421],
[-22.214,3.014,31.926],[-24.043,0.218,14.968],[-24.873,1.640,20.519],
[-25.418,0.966,25.719],[-25.091,2.315,27.869],[-40.552,1.549,31.936],
[-22.408,-4.482,34.418],[-23.476,-0.680,19.779],[-37.170,-0.561,24.193],
[-40.518,2.854,24.662],[-23.365,-1.186,29.915],[-21.364,-1.167,17.265],
[-39.783,-2.145,19.592],[-26.404,13.190,21.730],[-40.273,0.525,25.894],
[-37.329,10.578,33.392],[-34.766,10.808,28.078],[-29.962,12.171,32.569],
[-28.080,13.534,31.095],[-23.996,-3.565,28.994],[-38.096,14.504,28.937],
[-34.375,13.589,26.808],[-29.500,14.812,26.062],[-26.672,13.620,26.559],
[-26.428,15.882,28.351],[-38.509,13.770,26.231],[-36.059,12.406,23.814],
[-31.811,11.753,22.908],[-31.099,-0.510,23.604],[-24.987,-3.359,23.202],
[-22.708,-2.369,22.225],[-37.536,-3.142,21.960],[-39.731,-4.003,17.621],
[-32.509,-3.463,21.895],[-27.673,-3.649,23.411],[-26.830,-3.328,19.453],
[-38.995,11.992,19.280],[-37.526,-2.244,17.121],[-32.167,-1.493,19.473],
[-29.888,13.042,16.871],[-24.734,13.409,19.442],[-39.746,-1.292,31.214],
[-36.063,-2.670,31.416],[-36.975,14.186,16.687],[-28.541,-4.311,33.643],
[-22.425,12.910,32.498],[-28.471,23.619,21.565],[-31.179,24.608,25.279],
[-37.155,23.101,25.435],[-27.684,27.245,23.435],[-37.417,28.161,24.003],
[-39.999,21.594,20.766],[-40.014,24.202,22.060],[-28.443,31.585,23.612],
[-31.725,27.184,26.325],[-23.209,17.447,26.588],[-27.572,20.795,26.259],
[-36.792,19.199,22.717],[-26.172,20.036,22.261],[-38.677,21.148,23.388],
[-32.730,23.356,23.168],[-22.095,19.777,29.033],[-35.700,18.372,27.965],
[-30.145,27.377,21.034],[-34.069,25.248,30.083],[-24.805,18.118,23.364],
[-36.990,20.299,25.389],[-36.203,26.807,28.645],[-37.679,28.188,20.068],
[-25.827,27.264,25.699],[-25.235,18.495,28.592],[-32.236,14.609,24.228],
[-25.864,29.337,23.027],[-27.903,29.346,21.417],[-24.542,26.667,29.764],
[-30.151,24.727,19.816],[-28.776,22.323,16.196],[-22.337,22.063,27.766],
[-25.695,25.670,22.162],[-27.239,22.730,28.887],[-34.592,24.855,21.789],
[-34.595,27.015,23.851],[-34.403,24.008,25.593],[-28.827,28.602,26.408],
[-33.998,28.661,30.873],[-26.471,22.895,17.178],[-35.970,23.625,31.028],
[-29.418,31.460,20.455],[-35.359,31.127,31.166],[-25.315,15.971,17.311],
[-36.851,23.662,19.666],[-22.964,29.935,16.469],[-37.907,25.637,26.679],
[-40.323,28.170,26.791],[-20.813,29.240,30.205],[-31.192,28.765,30.283],
[-30.436,34.002,20.098],[-36.637,32.772,21.971],[-27.610,32.098,16.495],
[-26.868,31.381,19.378],[-24.844,20.709,32.443],[-37.048,27.055,31.120],
[-35.831,16.069,26.089],[-28.856,14.791,28.989],[-26.611,20.388,30.359],
[-26.696,25.410,16.472],[-30.020,18.631,17.095],[-33.933,21.097,28.039],
[-34.442,21.533,21.079],[-31.009,21.981,18.185],[-28.886,21.595,31.484],
[-27.813,21.754,23.424],[-36.799,20.284,30.013],[-34.743,18.413,15.902],
[-40.562,18.988,21.958],[-38.166,20.735,17.804],[-29.945,22.314,27.078],
[-34.114,17.828,24.439],[-37.387,21.345,20.932],[-35.573,20.752,18.759],
[-27.745,17.917,26.663],[-33.136,20.424,16.898],[-25.284,34.262,32.670],
[-38.058,26.781,33.716],[-31.201,21.114,24.799],[-36.585,17.476,20.651],
[-31.191,19.782,27.683],[-31.108,26.676,18.270],[-24.838,23.629,28.230],
[-27.641,17.576,16.598],[-31.430,25.874,29.510],[-27.573,20.245,17.369],
[-37.248,25.355,23.849],[-28.799,27.903,18.892],[-39.352,25.073,29.059],
[-30.965,23.351,31.582],[-29.279,27.159,29.138],[-21.639,23.212,18.896],
[-21.949,27.124,31.612],[-24.969,29.261,19.615],[-28.538,23.797,24.839],
[-31.812,29.462,20.287],[-38.557,23.023,16.030],[-32.946,26.754,20.429],
[-31.488,29.727,27.411],[-34.872,30.164,21.961],[-28.733,24.898,27.367],
[-27.007,28.312,17.029],[-27.130,30.047,30.500],[-23.939,32.653,16.001],
[-39.619,20.932,29.567],[-25.774,26.608,18.926],[-37.878,30.714,29.372],
[-34.621,29.165,15.613],[-33.512,28.165,18.187],[-36.666,26.421,18.326],
[-30.808,17.171,26.549],[-28.870,19.505,20.271],[-31.087,29.545,17.162],
[-32.323,18.873,20.993],[-31.163,20.911,30.097],[-27.961,17.792,22.349],
[-33.601,31.389,17.873],[-34.353,16.993,18.843],[-25.075,18.297,31.460],
[-24.983,23.467,23.819],[-36.392,29.572,18.034],[-35.866,22.330,16.045],
[-25.014,31.630,26.885],[-36.818,16.528,23.479],[-38.631,30.497,22.423],
[-39.945,32.624,31.125],[-33.437,32.233,33.800],[-30.099,30.421,14.657],
[-28.718,29.171,32.673],[-25.469,30.304,16.431],[-40.495,24.873,33.049],
[-34.523,26.009,16.221],[-32.532,25.977,32.321],[-29.510,27.613,15.542],
[-27.261,25.549,30.195],[-37.876,23.947,32.755],[-33.550,24.772,18.786],
[-34.229,21.679,30.704],[-32.640,22.986,15.600],[-29.639,25.647,32.312],
[-24.777,25.087,14.667],[-32.950,18.481,31.302],[-33.094,17.907,28.222],
[-32.249,23.588,28.459],[-27.081,20.972,33.530],[-26.462,23.188,31.578],
[-36.040,22.032,33.176],[-31.911,15.026,28.238],[-30.623,19.449,33.328],
[-28.984,17.445,29.929],[-27.902,18.534,32.144],[-35.516,19.633,32.222],
[-32.857,13.990,31.069],[-32.336,15.774,32.980],[-32.293,33.922,17.496],
[-39.780,26.984,17.287],[-38.278,33.465,20.011],[-40.751,31.057,24.004],
[-21.804,31.696,26.232],[-22.743,33.133,28.351],[-24.626,28.235,32.734],
[-23.345,25.257,18.928],[-22.117,25.762,21.503],[-23.264,26.755,26.484],
[-24.454,29.326,28.266],[-23.638,22.899,16.217],[-23.472,20.323,16.704],
[-25.008,22.283,21.477],[-40.525,24.378,26.557],[-24.629,25.206,32.083],
[-22.201,22.873,34.051],[-25.319,18.860,15.376],[-24.173,19.252,19.110],
[-24.264,20.323,24.693],[-24.139,21.652,30.163],[-39.902,19.580,31.850],
[-20.818,16.248,33.430],[-24.903,16.680,20.209],[-39.476,18.628,24.598],
[-39.563,21.258,25.759],[-22.031,17.506,31.595],[-22.769,17.681,16.302],
[-39.555,17.079,19.333],[-23.986,32.340,21.321],[-40.307,18.113,29.286],
[-37.421,29.425,32.754],[-34.079,28.714,27.734],[-30.913,31.834,29.649],
[-28.204,32.287,28.942],[-22.662,14.415,29.750],[-37.620,34.968,28.621],
[-36.897,32.026,27.089],[-31.390,32.448,26.890],[-28.967,31.464,26.255],
[-26.446,33.992,27.577],[-38.403,29.998,25.675],[-35.802,29.777,25.141],
[-33.764,31.191,26.586],[-29.645,16.845,24.010],[-25.171,13.643,24.121],
[-22.635,16.557,22.376],[-38.741,15.771,21.768],[-39.620,14.647,17.276],
[-33.171,15.250,21.506],[-26.403,16.331,24.906],[-27.538,16.082,20.238],
[-38.907,30.526,18.628],[-36.799,16.850,17.371],[-31.423,17.122,19.115],
[-30.336,31.924,17.894],[-23.184,31.483,18.785],[-36.387,15.725,30.464],
[-35.045,16.116,32.990],[-37.590,32.500,17.249],[-27.186,15.653,32.454],
[-24.333,30.433,30.796],[-29.160,43.290,22.195],[-29.320,43.512,25.103],
[-38.563,42.141,26.686],[-26.402,45.462,23.632],[-36.094,46.722,23.735],
[-40.908,40.407,21.386],[-39.656,42.395,22.830],[-28.935,50.361,24.933],
[-31.474,46.018,25.916],[-23.307,36.754,26.607],[-27.229,39.897,26.939],
[-36.335,37.645,23.745],[-27.608,38.347,21.664],[-40.228,38.243,23.747],
[-33.419,41.759,23.549],[-21.800,39.370,28.037],[-37.133,37.422,27.049],
[-29.331,47.378,19.909],[-33.823,43.498,27.634],[-25.844,36.839,23.364],
[-38.214,39.485,25.282],[-34.207,46.519,27.797],[-38.873,45.139,19.625],
[-26.129,47.038,26.530],[-25.925,36.706,27.442],[-32.475,34.439,25.024],
[-25.342,49.360,23.972],[-26.834,47.615,21.646],[-24.819,45.121,28.543],
[-30.762,44.623,20.277],[-30.574,41.465,15.450],[-21.055,42.157,28.092],
[-26.942,43.846,20.715],[-29.114,39.444,29.244],[-34.593,42.398,21.269],
[-33.631,45.844,24.295],[-34.960,43.507,24.915],[-29.138,47.127,27.021],
[-34.330,47.555,31.083],[-27.710,42.557,17.775],[-35.936,42.728,29.202],
[-32.668,49.043,21.956],[-37.231,50.497,31.025],[-22.788,34.268,18.567],
[-36.388,40.684,19.929],[-23.442,49.736,15.497],[-39.721,45.035,27.071],
[-39.202,47.051,25.420],[-21.031,48.855,30.179],[-31.385,47.856,30.873],
[-29.279,52.035,18.491],[-37.048,50.467,19.692],[-26.872,52.498,16.897],
[-28.581,49.920,20.386],[-24.387,40.232,31.557],[-38.322,46.283,29.190],
[-35.079,34.121,26.238],[-29.132,34.877,29.444],[-26.680,38.959,30.823],
[-26.548,43.185,15.261],[-30.731,36.768,16.170],[-34.820,39.038,25.990],
[-33.847,39.804,21.019],[-30.825,41.804,20.242],[-31.367,40.794,32.337],
[-28.234,40.568,23.659],[-36.692,40.608,28.013],[-34.812,37.847,15.411],
[-40.779,36.566,21.892],[-39.614,38.397,17.514],[-29.788,39.583,26.497],
[-32.750,36.565,23.676],[-37.207,37.607,21.099],[-34.069,38.752,18.559],
[-28.756,36.506,26.496],[-36.764,37.867,18.414],[-24.332,50.423,33.242],
[-35.591,43.326,33.348],[-32.464,40.102,25.189],[-34.600,36.519,21.708],
[-32.386,38.827,28.096],[-32.940,44.587,18.371],[-24.035,42.103,27.107],
[-26.986,38.531,17.146],[-29.942,45.242,30.636],[-27.041,40.591,19.971],
[-37.604,44.671,24.495],[-29.434,46.694,17.295],[-38.857,41.232,29.884],
[-31.277,43.346,32.194],[-28.828,48.250,29.665],[-24.155,40.986,18.982],
[-22.061,45.238,29.891],[-22.097,48.626,18.157],[-26.403,43.481,26.544],
[-35.694,48.078,20.167],[-38.979,43.050,16.389],[-32.844,46.017,21.246],
[-31.717,48.557,27.275],[-34.778,50.627,21.170],[-27.456,45.369,28.363],
[-26.804,46.319,16.177],[-26.639,50.437,29.037],[-22.023,51.278,17.332],
[-38.005,38.436,29.560],[-27.494,45.625,18.963],[-38.444,48.884,28.195],
[-32.877,47.399,14.685],[-33.351,47.081,18.707],[-36.727,44.911,17.634],
[-30.601,37.049,28.922],[-28.006,36.088,18.927],[-31.257,48.559,18.374],
[-29.071,38.938,19.321],[-30.237,41.804,29.953],[-30.117,36.924,21.576],
[-32.744,50.588,17.148],[-34.066,35.993,18.196],[-23.285,38.147,29.780],
[-25.455,41.200,24.373],[-34.829,46.967,16.619],[-37.202,39.996,16.116],
[-26.373,50.216,26.329],[-37.939,33.295,24.765],[-37.826,48.126,21.897],
[-38.716,49.131,33.045],[-33.737,50.483,33.077],[-30.444,49.355,15.709],
[-27.147,46.731,31.262],[-26.460,49.722,16.508],[-38.747,45.529,32.045],
[-33.995,44.936,14.407],[-33.725,45.111,31.995],[-29.847,45.983,14.575],
[-27.012,43.821,30.499],[-38.169,42.996,31.855],[-35.519,42.024,17.242],
[-33.252,40.658,29.749],[-33.309,42.591,15.792],[-28.652,43.975,32.637],
[-21.693,45.296,15.931],[-34.825,37.839,29.913],[-33.355,36.294,27.088],
[-31.644,41.927,27.451],[-26.759,39.720,33.230],[-27.204,41.353,31.161],
[-36.769,39.549,35.464],[-33.034,33.909,28.475],[-32.025,38.214,32.478],
[-29.280,39.043,31.950],[-27.651,36.041,32.256],[-36.118,38.158,33.242],
[-33.584,33.013,31.024],[-32.050,34.725,32.535],[-33.516,52.821,16.000],
[-40.309,45.983,16.750],[-41.164,52.122,21.858],[-39.884,49.043,23.746],
[-21.788,51.181,24.745],[-23.504,51.661,28.051],[-24.979,47.286,33.262],
[-23.650,46.454,17.904],[-24.108,45.038,21.519],[-23.569,46.483,25.335],
[-24.480,49.122,28.514],[-23.871,43.814,18.515],[-25.995,40.840,16.307],
[-24.835,42.036,21.543],[-40.715,43.233,25.311],[-24.286,42.744,30.463],
[-22.683,41.340,33.190],[-25.278,36.301,15.572],[-24.573,37.467,18.174],
[-24.994,38.597,25.566],[-25.062,39.363,28.355],[-40.799,38.414,32.604],
[-20.343,36.554,33.316],[-24.846,36.266,20.634],[-39.070,35.879,24.207],
[-41.172,40.520,24.993],[-22.688,35.790,32.047],[-22.332,36.738,15.406],
[-39.840,35.613,18.644],[-26.024,50.515,20.882],[-39.823,35.277,27.280],
[-36.406,47.872,33.098],[-33.994,49.318,28.503],[-29.550,51.057,31.010],
[-27.350,52.501,31.220],[-25.090,33.290,29.946],[-37.742,53.388,27.739],
[-36.022,50.442,27.041],[-31.710,50.593,25.437],[-28.786,51.100,27.598],
[-26.367,52.787,27.409],[-37.799,49.863,25.319],[-35.612,49.377,23.565],
[-34.082,51.272,24.595],[-30.254,35.390,23.750],[-24.734,34.308,25.355],
[-23.182,35.197,22.377],[-36.982,35.312,22.435],[-40.163,32.822,17.751],
[-31.715,32.783,22.480],[-27.907,33.977,25.006],[-26.235,33.917,20.530],
[-39.452,49.367,18.475],[-37.588,36.492,15.159],[-31.273,36.548,19.007],
[-30.708,53.088,15.850],[-23.801,50.490,19.468],[-38.089,36.185,33.142],
[-35.360,34.720,32.957],[-37.172,49.971,16.071],[-28.822,33.818,33.093],
[-23.587,48.029,31.094],[-23.396,56.029,25.278],[-28.543,-59.371,26.798],
[-35.861,56.380,23.661],[-26.691,-59.937,21.011],[-39.254,58.480,21.882],
[-22.266,56.167,27.867],[-37.251,56.003,27.195],[-25.905,56.968,23.210],
[-37.107,58.349,25.771],[-24.762,54.813,28.011],[-31.701,52.716,23.689],
[-24.421,52.426,17.776],[-24.460,56.813,32.311],[-32.797,53.497,25.981],
[-29.364,53.272,29.463],[-26.741,56.388,30.884],[-30.512,55.476,17.003],
[-34.285,58.534,26.769],[-33.918,57.291,20.711],[-29.379,58.532,21.485],
[-36.564,-59.073,28.102],[-33.441,56.016,16.972],[-40.520,54.618,23.198],
[-38.073,56.751,18.780],[-33.277,55.327,23.322],[-36.642,57.943,21.038],
[-32.168,57.695,18.447],[-26.187,56.165,25.965],[-35.674,59.281,17.519],
[-29.876,57.862,24.222],[-36.712,54.065,20.395],[-31.850,57.035,26.302],
[-26.737,56.218,16.103],[-27.170,58.690,17.029],[-22.663,58.796,17.832],
[-39.081,54.776,30.559],[-29.763,55.827,27.733],[-29.002,54.741,18.999],
[-31.431,-59.160,20.075],[-31.402,57.927,29.455],[-30.036,54.391,22.341],
[-35.490,54.506,17.840],[-26.952,57.676,28.290],[-39.565,57.232,16.184],
[-35.591,53.792,25.396],[-33.934,56.856,30.091],[-34.435,55.382,27.487],
[-28.254,57.860,34.519],[-36.235,54.887,29.921],[-31.700,51.494,27.840],
[-33.074,56.954,33.014],[-29.640,57.267,31.694],[-26.645,55.872,33.420],
[-39.141,56.273,32.731],[-32.274,52.291,31.026],[-33.933,54.336,33.194],
[-24.763,59.463,16.047],[-24.011,54.657,15.896],[-26.121,57.146,18.929],
[-24.733,58.598,26.949],[-24.484,-59.638,30.086],[-38.715,57.600,28.567],
[-22.766,52.727,32.747],[-24.385,54.490,20.279],[-38.530,56.316,23.530],
[-41.293,-59.941,25.021],[-23.104,54.924,30.324],[-20.882,54.503,15.012],
[-39.608,54.452,20.051],[-40.148,54.060,26.543],[-24.630,52.202,30.637],
[-30.757,55.153,24.866],[-26.466,51.901,23.803],[-25.005,54.483,23.649],
[-38.363,53.559,22.342],[-37.873,52.422,17.906],[-33.393,53.120,21.425],
[-28.785,53.122,24.363],[-27.365,53.020,19.889],[-38.218,54.687,16.727],
[-32.032,54.314,19.004],[-35.341,52.468,31.437],[-37.095,54.189,32.839],
[-28.632,52.967,33.908],[-29.888,-50.753,42.024],[-30.452,-49.646,44.273],
[-39.154,-49.474,46.454],[-26.595,-46.639,42.320],[-35.457,-48.033,44.920],
[-40.453,-52.095,40.931],[-39.314,-48.747,40.048],[-30.114,-43.794,42.780],
[-31.510,-47.542,41.666],[-23.246,-55.565,44.955],[-27.890,-54.853,45.248],
[-34.256,-55.096,41.763],[-27.102,-57.864,39.423],[-38.807,-53.181,43.339],
[-32.973,-50.500,43.130],[-22.311,-53.499,46.460],[-37.197,-57.146,45.561],
[-28.749,-48.151,41.255],[-34.140,-50.130,47.926],[-24.386,-56.064,42.522],
[-36.970,-55.586,43.281],[-37.110,-47.936,47.043],[-36.921,-46.330,37.410],
[-26.565,-46.986,44.945],[-25.143,-57.352,46.597],[-32.714,58.664,43.314],
[-27.182,-43.904,42.602],[-27.460,-45.564,39.971],[-23.838,-48.466,48.687],
[-30.928,-49.578,39.282],[-28.759,-51.256,35.168],[-24.340,-50.212,46.584],
[-26.149,-49.129,40.828],[-26.592,-51.795,47.058],[-36.002,-49.942,40.980],
[-33.943,-45.447,44.065],[-36.510,-51.532,43.157],[-28.889,-45.480,44.745],
[-33.594,-47.569,48.177],[-26.628,-51.298,37.132],[-36.808,-51.161,47.195],
[-31.386,-43.902,40.433],[-35.916,-45.974,49.860],[-24.088,-59.053,35.802],
[-35.147,-51.007,38.070],[-24.402,-43.917,35.638],[-41.753,-48.793,47.828],
[-39.230,-46.772,45.266],[-19.603,-46.148,50.260],[-30.607,-46.589,48.829],
[-31.050,-41.854,38.539],[-36.524,-41.848,40.614],[-29.529,-45.406,35.715],
[-29.192,-44.318,38.431],[-26.312,-52.883,50.200],[-38.946,-46.716,48.726],
[-34.428,-58.652,45.237],[-30.019,-57.884,47.247],[-26.261,-55.616,49.246],
[-26.611,-49.292,33.808],[-31.044,-54.699,35.593],[-35.210,-55.082,45.625],
[-33.050,-52.987,40.552],[-29.556,-52.097,37.888],[-29.159,-53.161,49.403],
[-29.044,-53.550,41.893],[-37.885,-54.517,46.249],[-34.175,-54.382,34.009],
[-38.617,-55.163,40.970],[-38.639,-56.047,37.090],[-29.620,-52.423,45.933],
[-32.782,-57.318,42.066],[-36.422,-53.687,41.451],[-33.189,-52.860,37.279],
[-27.562,-57.414,45.064],[-36.299,-54.743,36.054],[-25.672,-41.920,52.172],
[-37.550,-49.119,50.257],[-31.340,-52.844,43.014],[-36.469,-55.724,39.211],
[-32.673,-55.263,46.549],[-32.954,-48.168,38.224],[-25.280,-49.362,43.948],
[-27.936,-56.675,34.864],[-31.185,-49.584,49.032],[-26.445,-55.025,37.660],
[-37.581,-48.728,43.258],[-30.146,-47.651,37.480],[-39.518,-52.356,47.849],
[-33.279,-51.637,50.169],[-27.485,-45.569,46.949],[-22.263,-53.180,38.039],
[-22.988,-46.232,49.835],[-24.243,-45.953,39.031],[-27.621,-50.000,44.947],
[-35.055,-46.286,41.033],[-37.596,-48.858,34.629],[-33.654,-49.105,40.761],
[-32.869,-44.961,47.965],[-33.662,-43.121,39.457],[-29.293,-48.174,46.631],
[-25.783,-46.228,35.444],[-26.610,-43.894,49.051],[-22.484,-42.540,34.393],
[-38.403,-56.145,48.453],[-26.627,-47.693,37.878],[-38.705,-44.705,46.990],
[-33.785,-45.775,32.734],[-33.406,-45.686,35.303],[-37.065,-49.279,38.305],
[-30.043,-55.094,46.840],[-28.784,-56.880,37.640],[-31.681,-45.132,37.402],
[-31.376,-54.578,38.699],[-31.840,-53.907,49.069],[-29.675,-57.311,40.290],
[-33.846,-43.123,36.595],[-33.473,-55.481,36.704],[-23.640,-55.771,48.418],
[-26.673,-51.389,42.595],[-35.783,-46.781,34.315],[-36.480,-52.176,34.864],
[-24.896,-42.476,43.858],[-38.482,-58.451,43.160],[-36.439,-44.435,39.413],
[-38.608,-42.246,49.359],[-33.494,-43.523,52.412],[-30.482,-44.453,33.351],
[-28.944,-46.750,50.934],[-26.722,-43.949,33.477],[-40.121,-49.137,49.977],
[-33.414,-48.336,34.606],[-34.855,-48.985,50.382],[-31.026,-47.085,33.909],
[-28.178,-50.008,48.714],[-37.760,-51.896,50.303],[-33.273,-50.290,36.244],
[-34.354,-52.799,47.466],[-33.446,-51.727,33.855],[-29.667,-49.120,51.748],
[-23.811,-50.285,34.503],[-33.370,-56.313,50.060],[-34.826,-57.160,47.580],
[-31.958,-51.254,46.787],[-26.862,-53.775,52.709],[-28.694,-52.035,51.931],
[-35.748,-53.725,49.999],[-32.283,-59.120,46.825],[-31.117,-56.183,51.522],
[-28.963,-56.107,49.854],[-28.388,-56.426,52.389],[-35.963,-56.916,50.519],
[-34.711,58.688,47.692],[-32.402,-58.693,52.842],[-31.776,-40.889,36.078],
[-39.189,-46.827,35.594],[-41.515,-39.527,39.221],[-39.970,-43.889,41.979],
[-22.106,-41.991,45.531],[-23.185,-42.654,47.977],[-25.260,-46.838,53.519],
[-22.635,-47.414,37.267],[-23.728,-49.335,41.453],[-23.812,-46.761,44.954],
[-24.441,-44.987,47.058],[-21.549,-50.260,37.330],[-22.950,-52.886,35.372],
[-24.812,-52.723,39.575],[-40.356,-48.913,43.118],[-26.695,-49.197,50.679],
[-22.368,-53.206,52.060],[-24.660,-54.818,33.471],[-26.015,-55.082,40.468],
[-25.908,-53.712,43.860],[-25.449,-54.493,46.370],[-38.574,-55.033,50.944],
[-21.963,-59.098,51.770],[-24.652,-58.873,39.326],[-40.995,-55.629,44.563],
[-40.933,-53.234,45.446],[-24.086,-58.170,50.574],[-22.820,-56.556,35.056],
[-39.013,-57.566,39.398],[-25.010,-44.514,41.319],[-39.506,-58.212,46.379],
[-39.562,-43.908,51.566],[-35.642,-44.909,46.862],[-30.505,-42.592,50.157],
[-27.459,-41.086,49.015],[-26.333,-58.651,49.079],[-38.296,-39.698,47.067],
[-36.117,-42.249,48.360],[-31.201,-41.531,45.293],[-28.861,-42.700,46.965],
[-26.773,-41.013,46.445],[-38.108,-44.371,44.176],[-35.728,-43.356,42.855],
[-32.635,-42.856,42.890],[-30.120,-57.444,43.212],[-26.035,59.191,43.890],
[-23.494,-59.031,41.948],[-36.842,59.319,40.984],[-39.568,-59.155,35.015],
[-31.551,-59.346,40.652],[-27.299,-57.270,42.425],[-27.888,-59.050,36.306],
[-39.588,-45.317,38.130],[-38.596,-56.859,34.218],[-31.159,-57.633,36.288],
[-29.879,-42.851,36.289],[-22.561,-41.905,38.392],[-37.954,-58.563,50.020],
[-35.194,-59.078,52.006],[-35.865,-41.137,34.093],[-26.562,-59.211,51.750],
[-23.907,-43.558,50.636],[-30.915,-31.809,41.465],[-30.666,-32.130,44.173],
[-38.093,-32.152,44.616],[-25.238,-27.128,40.794],[-36.392,-27.884,43.591],
[-38.659,-33.400,38.889],[-40.115,-31.804,41.051],[-29.601,-25.166,42.877],
[-31.536,-29.346,44.595],[-22.751,-38.273,45.688],[-28.511,-34.314,45.100],
[-35.442,-37.434,42.349],[-27.759,-36.678,40.028],[-40.536,-34.317,41.609],
[-35.388,-32.946,44.365],[-22.487,-35.313,48.265],[-37.223,-37.315,45.434],
[-30.130,-29.437,39.747],[-34.943,-31.508,48.884],[-26.478,-36.751,42.453],
[-37.982,-34.607,43.137],[-36.245,-28.878,47.519],[-38.698,-28.583,37.821],
[-27.213,-26.990,44.532],[-25.293,-37.478,46.759],[-33.557,-40.167,43.209],
[-27.137,-25.741,42.020],[-28.082,-25.423,37.626],[-26.039,-28.564,49.351],
[-30.508,-32.062,38.833],[-31.574,-35.357,36.600],[-23.509,-33.172,46.728],
[-27.835,-32.736,38.924],[-28.255,-33.172,48.423],[-34.430,-30.680,41.579],
[-32.594,-27.111,43.707],[-34.229,-30.524,44.735],[-30.166,-27.101,45.154],
[-33.364,-29.045,48.423],[-29.234,-32.453,36.519],[-37.638,-32.447,48.427],
[-30.198,-24.671,40.344],[-35.401,-26.362,50.651],[-24.178,-39.708,37.137],
[-35.253,-31.858,38.165],[-23.981,-26.930,36.514],[-39.396,-30.420,46.259],
[-38.935,-26.940,43.369],[-21.667,-26.487,49.184],[-31.533,-26.191,48.808],
[-31.650,-21.542,38.002],[-36.455,-23.836,41.382],[-27.729,-24.360,35.227],
[-29.858,-23.531,37.876],[-25.508,-34.276,50.866],[-39.216,-27.888,46.642],
[-36.131,-39.840,44.224],[-30.768,-40.559,48.280],[-27.015,-35.791,48.622],
[-25.468,-31.362,34.577],[-31.032,-37.980,36.150],[-35.287,-35.675,44.468],
[-35.063,-35.066,39.053],[-29.699,-34.578,39.094],[-30.867,-33.418,50.243],
[-29.140,-36.233,43.060],[-37.469,-35.301,47.292],[-34.743,-36.945,34.983],
[-39.413,-36.071,39.698],[-38.440,-37.020,35.069],[-30.937,-33.513,46.286],
[-31.780,-36.383,41.844],[-36.368,-33.197,40.414],[-34.868,-37.525,37.929],
[-28.373,-38.572,45.282],[-35.312,-34.362,36.252],[-23.474,-22.362,49.583],
[-36.696,-29.464,51.517],[-32.850,-34.364,43.584],[-35.530,-39.415,40.363],
[-32.709,-36.148,46.158],[-32.295,-29.750,37.336],[-24.726,-30.411,46.522],
[-28.390,-37.756,35.340],[-30.616,-29.551,48.069],[-28.161,-35.179,36.490],
[-37.609,-30.287,42.069],[-30.944,-28.242,35.514],[-40.466,-31.839,48.929],
[-33.613,-33.820,50.633],[-29.020,-27.465,49.020],[-22.248,-34.156,37.308],
[-22.347,-29.135,47.830],[-26.235,-27.721,37.852],[-28.173,-30.026,43.698],
[-35.781,-27.534,40.353],[-39.539,-31.085,34.747],[-33.355,-28.383,39.602],
[-32.932,-26.182,46.166],[-34.177,-22.973,40.073],[-29.022,-29.351,46.007],
[-28.391,-29.807,35.403],[-27.560,-25.101,49.223],[-23.084,-23.908,34.716],
[-40.036,-37.380,48.545],[-28.053,-27.786,40.474],[-36.979,-26.111,46.706],
[-33.450,-27.015,33.534],[-34.809,-27.797,37.113],[-36.738,-29.970,36.737],
[-30.077,-36.824,46.238],[-28.930,-38.406,38.163],[-31.600,-25.801,36.795],
[-31.955,-36.707,39.021],[-32.343,-35.325,48.715],[-30.561,-38.551,41.160],
[-35.536,-24.590,37.100],[-35.020,-39.722,36.394],[-24.278,-36.723,50.550],
[-28.499,-33.380,41.545],[-36.932,-27.381,35.770],[-38.348,-34.056,35.413],
[-25.819,-24.331,44.762],[-38.563,-39.298,43.824],[-38.808,-25.946,40.207],
[-38.669,-25.425,51.587],[-34.395,-24.578,52.602],[-30.819,-26.427,33.123],
[-28.715,-26.368,51.583],[-25.942,-26.507,34.823],[-39.794,-30.331,51.611],
[-35.261,-29.188,34.137],[-33.710,-29.686,51.475],[-30.161,-29.166,33.038],
[-27.451,-30.548,48.000],[-38.834,-33.161,50.862],[-34.701,-31.524,35.360],
[-34.020,-33.737,47.574],[-32.993,-33.656,34.598],[-29.401,-30.838,51.015],
[-23.085,-29.579,34.022],[-34.937,-36.097,49.558],[-34.143,-38.425,47.296],
[-32.016,-31.941,48.289],[-28.549,-35.142,51.304],[-26.829,-31.993,50.345],
[-36.153,-32.905,51.301],[-33.834,-41.073,46.231],[-33.078,-37.024,52.175],
[-29.742,-35.592,48.903],[-27.941,-37.894,51.362],[-38.275,-37.108,50.665],
[-33.051,-42.529,49.852],[-32.750,-39.669,51.241],[-34.322,-22.214,35.993],
[-39.493,-27.383,35.171],[-40.037,-22.405,39.893],[-39.752,-24.709,42.289],
[-23.154,-23.661,43.891],[-23.786,-24.267,46.951],[-24.198,-27.537,50.804],
[-24.276,-30.242,37.082],[-23.644,-28.504,39.023],[-24.791,-29.205,43.104],
[-23.912,-27.853,45.984],[-22.821,-32.149,35.532],[-24.270,-34.717,35.577],
[-24.910,-33.184,41.667],[-41.198,-30.821,43.991],[-24.066,-30.983,50.846],
[-22.234,-33.360,50.522],[-26.084,-36.410,34.476],[-26.501,-37.276,37.088],
[-25.898,-34.248,44.050],[-26.023,-33.570,46.961],[-40.578,-36.186,51.281],
[-21.798,-40.252,52.625],[-25.818,-38.552,39.619],[-38.713,-37.276,42.315],
[-40.452,-33.763,44.177],[-23.899,-39.675,50.683],[-22.765,-37.458,35.963],
[-41.152,-37.671,37.173],[-26.161,-23.436,39.525],[-40.609,-39.659,45.756],
[-37.794,-27.499,49.223],[-34.886,-24.659,47.932],[-29.895,-23.320,48.825],
[-27.812,-21.927,48.891],[-24.477,-39.852,47.522],[-39.651,-22.711,47.593],
[-36.769,-23.341,46.419],[-30.661,-24.375,45.265],[-27.956,-24.914,46.224],
[-26.208,-21.210,46.289],[-37.364,-24.328,44.136],[-34.918,-25.511,42.957],
[-32.447,-24.483,43.047],[-31.666,-38.715,44.819],[-23.091,-40.012,43.435],
[-23.806,-41.644,41.444],[-39.455,-41.190,40.844],[-39.449,-41.938,36.577],
[-32.010,-41.180,41.061],[-28.915,-41.205,42.817],[-28.324,-41.628,38.691],
[-39.399,-25.552,37.398],[-37.797,-39.821,35.623],[-32.230,-39.250,38.107],
[-30.259,-23.105,34.856],[-25.033,-24.787,37.610],[-39.684,-39.853,50.637],
[-36.977,-39.672,51.508],[-37.580,-24.920,35.235],[-28.230,-40.589,51.662],
[-24.300,-25.070,49.592],[-29.224,-12.349,40.989],[-29.087,-12.002,43.795],
[-37.507,-13.901,45.462],[-25.434,-10.332,40.731],[-37.411,-9.000,43.894],
[-39.139,-15.105,39.903],[-40.344,-12.158,41.616],[-27.524,-5.529,42.838],
[-31.343,-10.137,43.431],[-22.169,-19.218,45.534],[-28.426,-15.985,43.796],
[-37.006,-18.169,42.745],[-27.520,-18.810,40.356],[-39.586,-17.440,42.564],
[-33.731,-15.053,41.530],[-21.739,-16.701,46.929],[-35.585,-18.848,45.060],
[-31.235,-9.848,40.404],[-33.690,-12.802,44.806],[-25.360,-19.137,41.916],
[-37.702,-16.225,44.082],[-36.245,-11.209,45.651],[-38.527,-11.502,38.291],
[-26.642,-9.196,44.293],[-25.763,-18.565,45.259],[-32.849,-21.646,43.318],
[-27.598,-7.638,41.124],[-27.557,-8.780,38.400],[-25.335,-11.361,49.667],
[-28.512,-11.281,38.541],[-29.937,-14.615,34.810],[-23.000,-14.214,47.468],
[-26.796,-13.245,39.512],[-28.481,-15.344,47.912],[-34.715,-13.109,39.985],
[-34.658,-8.565,43.913],[-35.803,-14.512,43.353],[-29.887,-8.398,45.785],
[-34.490,-9.949,49.700],[-28.485,-12.947,36.477],[-36.605,-14.293,48.869],
[-29.712,-7.303,39.097],[-34.788,-7.365,50.005],[-24.988,-22.147,35.953],
[-38.227,-14.452,37.093],[-23.036,-7.587,34.589],[-39.271,-11.664,45.611],
[-40.053,-8.682,44.266],[-22.205,-7.613,48.934],[-31.600,-8.475,48.475],
[-29.535,-4.629,39.071],[-37.345,-5.514,39.938],[-26.367,-4.862,35.441],
[-27.114,-6.196,38.154],[-23.733,-17.166,48.922],[-38.242,-9.798,47.917],
[-34.980,-21.705,45.605],[-28.861,-20.247,47.081],[-26.160,-17.668,47.738],
[-26.084,-14.566,35.537],[-32.158,-18.947,35.359],[-34.520,-16.835,43.427],
[-34.330,-17.540,39.847],[-29.363,-14.536,39.182],[-29.604,-15.939,50.887],
[-29.540,-16.842,41.094],[-36.279,-17.072,47.578],[-33.765,-16.850,33.906],
[-40.441,-19.764,40.682],[-40.931,-18.582,36.194],[-30.429,-15.735,45.677],
[-32.318,-17.410,41.598],[-36.630,-17.023,38.162],[-34.905,-18.186,36.668],
[-28.521,-18.867,44.600],[-35.255,-15.223,36.325],[-25.064,-4.401,50.194],
[-36.969,-11.550,49.558],[-30.722,-14.887,42.631],[-37.409,-19.137,40.306],
[-33.090,-16.004,45.849],[-31.077,-12.102,38.853],[-24.419,-11.479,46.966],
[-29.288,-19.920,34.987],[-29.266,-10.256,48.606],[-27.609,-16.105,37.823],
[-37.456,-11.462,42.862],[-29.980,-9.426,37.115],[-39.541,-13.984,46.961],
[-32.094,-13.851,50.344],[-29.018,-6.821,48.860],[-21.474,-15.265,36.701],
[-22.339,-10.911,48.836],[-23.976,-10.165,35.648],[-26.496,-12.504,44.775],
[-35.244,-9.411,40.663],[-38.918,-12.589,34.717],[-33.274,-11.327,37.903],
[-32.524,-8.449,45.280],[-34.359,-6.260,40.221],[-28.147,-10.664,46.019],
[-27.532,-9.121,35.682],[-26.417,-6.158,48.295],[-23.900,-5.712,36.378],
[-38.236,-18.209,46.092],[-26.345,-11.319,36.946],[-37.590,-7.530,46.380],
[-32.680,-8.708,33.831],[-33.703,-9.323,36.340],[-36.715,-11.252,35.645],
[-30.935,-18.263,46.807],[-28.953,-18.532,37.742],[-32.282,-6.241,35.122],
[-31.963,-19.026,38.043],[-31.116,-15.693,48.350],[-30.433,-19.541,40.200],
[-34.641,-6.471,37.294],[-36.644,-20.456,37.071],[-24.253,-19.750,48.095],
[-25.780,-14.828,43.542],[-36.632,-7.969,36.182],[-39.663,-16.361,35.270],
[-25.599,-5.324,44.948],[-35.690,-20.704,42.517],[-38.690,-7.936,39.103],
[-39.951,-4.825,49.894],[-33.067,-6.115,52.271],[-30.667,-7.223,32.827],
[-27.721,-8.764,50.842],[-25.620,-7.187,34.310],[-40.754,-11.911,51.404],
[-35.274,-9.255,33.939],[-31.423,-10.725,50.276],[-30.336,-9.982,34.274],
[-26.977,-12.487,47.902],[-37.800,-14.120,51.590],[-33.491,-12.574,35.573],
[-34.549,-14.440,47.208],[-35.737,-12.459,33.466],[-29.053,-11.976,50.993],
[-23.313,-11.905,33.281],[-33.692,-17.227,49.032],[-33.261,-19.412,46.483],
[-31.876,-13.554,46.490],[-26.314,-15.935,51.549],[-27.272,-13.903,50.071],
[-34.602,-15.070,51.095],[-31.760,-22.204,47.309],[-30.709,-19.975,50.178],
[-29.254,-17.781,48.842],[-27.890,-19.448,50.307],[-38.562,-18.481,52.237],
[-33.223,-22.630,50.024],[-32.509,-18.999,52.043],[-33.260,-2.722,35.200],
[-40.191,-9.778,34.569],[-40.428,-2.740,38.664],[-38.904,-5.870,42.197],
[-23.129,-5.496,43.754],[-23.119,-5.817,46.622],[-25.299,-9.375,51.751],
[-22.356,-11.409,37.770],[-23.815,-12.508,39.793],[-23.604,-10.057,44.894],
[-25.212,-8.082,46.703],[-23.955,-12.984,35.842],[-23.580,-16.076,35.028],
[-24.874,-14.818,40.590],[-40.944,-13.363,44.101],[-24.875,-14.940,49.417],
[-22.428,-14.664,50.835],[-27.002,-16.443,33.689],[-25.889,-18.181,37.854],
[-23.821,-17.679,43.441],[-25.626,-15.027,46.394],[-40.154,-17.891,49.988],
[-22.231,-21.477,52.684],[-23.913,-19.900,38.799],[-37.954,-20.829,45.183],
[-40.270,-15.885,44.735],[-20.970,-20.106,50.444],[-24.733,-18.880,35.600],
[-41.537,-20.379,38.220],[-24.861,-4.494,38.966],[-40.560,-20.183,46.387],
[-39.291,-8.740,51.463],[-34.228,-7.061,47.433],[-29.003,-4.792,50.571],
[-28.260,-2.943,48.694],[-23.262,-21.855,46.246],[-38.513,-4.693,47.564],
[-36.273,-4.937,45.901],[-29.281,-4.146,44.771],[-28.082,-6.276,46.053],
[-25.994,-3.167,47.022],[-38.887,-5.622,44.971],[-36.001,-6.277,43.221],
[-33.341,-5.799,42.775],[-31.378,-19.492,42.799],[-26.107,-23.260,42.139],
[-23.474,-21.594,41.836],[-39.440,-21.211,42.904],[-38.785,-22.675,36.840],
[-32.913,-20.815,40.619],[-28.158,-20.679,42.091],[-27.861,-21.351,37.712],
[-39.140,-6.363,37.229],[-37.918,-18.457,35.886],[-34.459,-20.408,38.468],
[-29.726,-4.951,36.329],[-23.261,-7.759,37.953],[-35.780,-19.848,49.350],
[-34.037,-21.850,52.477],[-36.152,-4.940,35.401],[-28.662,-21.547,51.657],
[-24.542,-7.400,50.241],[-29.635,4.921,40.756],[-29.797,5.398,44.663],
[-36.346,3.411,42.925],[-26.743,8.919,42.443],[-35.998,9.480,43.340],
[-39.302,4.385,39.248],[-40.047,6.293,41.740],[-29.058,11.865,43.115],
[-30.408,7.439,43.114],[-22.647,-0.840,44.995],[-28.861,0.578,46.523],
[-36.859,-0.238,40.280],[-27.900,1.244,39.601],[-39.294,3.229,42.442],
[-32.490,5.859,42.023],[-22.230,1.681,46.711],[-35.649,-0.047,45.706],
[-29.822,9.112,39.772],[-33.580,6.054,46.288],[-25.473,-0.910,41.831],
[-37.304,1.024,43.201],[-35.928,7.424,47.946],[-40.173,8.695,37.836],
[-24.871,8.719,44.457],[-25.356,-1.683,45.005],[-31.764,-3.306,43.355],
[-27.250,11.633,41.197],[-26.873,9.080,39.756],[-26.391,7.422,49.023],
[-30.299,6.203,38.404],[-30.050,2.914,35.466],[-24.813,4.045,45.568],
[-25.821,6.118,39.383],[-26.826,3.034,47.060],[-35.772,5.513,41.604],
[-32.764,8.956,42.543],[-33.943,4.525,43.966],[-28.366,9.466,44.571],
[-33.511,8.414,47.709],[-27.874,5.673,36.814],[-35.738,4.505,47.803],
[-30.208,11.914,40.194],[-36.093,10.588,49.717],[-25.846,-1.965,35.732],
[-36.182,5.711,37.738],[-23.245,11.399,36.510],[-41.473,6.434,46.522],
[-38.839,9.337,42.584],[-20.970,10.933,47.680],[-30.056,9.927,47.392],
[-30.372,14.192,38.538],[-35.706,14.511,39.783],[-26.686,12.907,36.271],
[-27.689,13.117,38.830],[-25.676,3.644,49.669],[-37.704,9.321,47.617],
[-34.538,-3.423,43.506],[-30.740,-3.822,47.140],[-26.542,0.438,48.938],
[-26.772,5.145,34.295],[-31.768,-0.095,35.110],[-35.158,2.358,45.081],
[-35.207,3.250,39.197],[-29.639,3.307,38.476],[-29.639,4.187,50.543],
[-28.622,2.593,42.285],[-36.438,1.569,47.727],[-35.243,0.340,34.056],
[-40.391,-0.165,40.190],[-38.801,1.394,36.308],[-29.535,3.494,46.680],
[-33.331,0.902,41.610],[-37.523,2.693,40.622],[-33.842,1.006,37.206],
[-28.340,-1.026,44.061],[-35.672,2.745,36.111],[-22.880,14.392,49.619],
[-36.813,7.801,50.617],[-31.646,3.119,42.848],[-38.011,-1.565,38.055],
[-33.256,0.683,46.512],[-32.500,8.312,38.807],[-24.529,7.398,46.912],
[-28.317,0.728,35.187],[-29.267,7.918,49.316],[-26.311,2.784,35.379],
[-37.335,7.127,43.138],[-30.439,8.371,35.957],[-39.164,3.820,48.464],
[-32.523,5.400,51.209],[-27.328,10.165,48.118],[-21.284,3.953,36.318],
[-22.291,9.185,49.243],[-25.511,11.025,38.118],[-27.244,6.443,43.947],
[-34.847,9.755,39.709],[-37.146,4.837,34.523],[-34.349,7.349,40.643],
[-32.438,11.480,47.536],[-33.716,12.419,38.573],[-28.176,7.735,46.810],
[-26.573,8.588,34.782],[-26.019,12.358,49.177],[-22.865,14.323,34.811],
[-38.497,0.539,49.438],[-28.208,8.184,37.679],[-36.529,12.514,44.775],
[-34.315,10.277,34.056],[-34.733,7.705,36.561],[-37.628,7.873,36.382],
[-31.307,-1.208,47.059],[-28.957,-0.788,37.938],[-30.890,11.452,37.191],
[-32.012,2.085,39.376],[-31.032,2.317,48.556],[-30.984,-0.924,39.627],
[-32.728,12.260,35.627],[-35.409,-1.673,36.816],[-24.501,-0.713,47.452],
[-26.648,2.912,44.209],[-35.809,10.498,36.604],[-37.883,2.017,33.765],
[-24.749,13.949,45.126],[-36.702,-2.055,44.575],[-36.973,11.718,39.059],
[-39.310,12.372,50.900],[-34.013,12.341,51.761],[-29.430,9.983,34.188],
[-27.402,10.973,51.146],[-25.333,11.674,34.152],[-39.927,7.212,51.110],
[-34.073,6.610,34.091],[-34.249,8.188,50.760],[-31.357,6.990,33.772],
[-27.766,5.239,48.397],[-37.605,3.704,52.349],[-32.716,5.732,36.284],
[-33.136,3.185,47.098],[-33.878,4.024,34.588],[-28.977,6.704,51.808],
[-23.748,7.059,35.028],[-33.199,0.945,49.688],[-33.837,-1.475,48.455],
[-31.341,5.545,47.676],[-27.661,2.073,51.206],[-27.047,4.842,51.680],
[-35.795,3.076,50.357],[-33.518,-5.138,45.437],[-31.567,2.541,51.536],
[-30.243,0.140,49.981],[-27.693,-0.637,51.021],[-36.897,0.730,51.887],
[-32.014,-3.158,49.403],[-33.321,-1.212,51.455],[-30.579,16.327,35.018],
[-39.115,10.693,35.394],[-40.140,14.692,38.348],[-39.233,11.345,40.391],
[-21.064,14.244,45.601],[-22.908,12.736,47.149],[-26.197,8.381,51.743],
[-21.758,7.872,38.482],[-23.863,5.104,37.844],[-22.587,9.086,45.957],
[-24.500,9.972,47.756],[-24.084,4.663,33.827],[-22.701,2.133,34.395],
[-22.962,2.897,40.070],[-39.130,5.503,44.167],[-24.629,5.903,50.727],
[-22.188,4.139,51.993],[-25.132,0.033,33.505],[-24.819,0.984,37.693],
[-24.482,1.384,45.102],[-24.354,2.010,48.275],[-41.226,0.855,49.750],
[-20.527,-2.211,52.664],[-24.550,-1.819,38.084],[-40.661,0.543,43.479],
[-40.778,3.865,45.275],[-24.473,-1.791,50.192],[-23.208,-0.356,35.780],
[-41.038,-0.968,36.679],[-25.816,14.637,40.017],[-38.911,-1.987,46.296],
[-39.507,9.516,49.972],[-35.353,10.270,45.935],[-30.216,11.662,49.756],
[-27.051,14.426,50.252],[-23.225,-3.153,47.993],[-38.325,14.322,46.970],
[-34.630,12.546,48.623],[-29.547,14.274,44.484],[-28.509,12.838,46.637],
[-26.993,14.853,47.367],[-41.072,13.367,43.851],[-34.959,11.854,41.939],
[-32.032,12.048,43.003],[-30.946,-0.063,42.232],[-24.597,-3.402,42.592],
[-22.648,-4.679,40.912],[-38.125,-2.941,41.005],[-38.444,-3.823,36.417],
[-32.957,-3.889,40.254],[-27.222,-3.018,42.881],[-27.162,-3.540,37.660],
[-39.796,11.871,37.753],[-37.162,-2.926,33.904],[-31.144,-2.592,37.290],
[-29.402,12.808,35.484],[-23.624,13.068,39.046],[-37.114,-1.831,50.665],
[-35.384,-3.729,52.285],[-36.938,12.556,35.153],[-28.078,-2.813,53.418],
[-23.068,11.342,50.970],[-27.987,24.929,40.976],[-31.455,24.279,43.718],
[-36.958,23.565,44.778],[-27.969,27.741,42.018],[-36.671,28.372,43.596],
[-39.972,22.211,39.672],[-39.025,23.521,42.593],[-27.966,31.951,41.471],
[-32.951,26.396,44.226],[-23.918,18.097,43.909],[-27.920,20.448,43.639],
[-35.631,21.081,42.055],[-27.667,18.296,41.419],[-38.742,21.062,41.976],
[-33.664,22.831,41.760],[-22.716,20.572,46.541],[-35.735,17.686,46.575],
[-30.967,27.464,39.298],[-33.081,24.636,46.107],[-25.022,18.323,41.545],
[-37.206,20.121,44.197],[-36.413,26.616,46.864],[-39.236,27.917,37.874],
[-26.451,27.533,44.632],[-25.985,18.790,45.995],[-31.659,14.899,42.759],
[-27.084,30.201,43.740],[-28.039,28.955,38.814],[-25.990,27.013,47.467],
[-30.115,24.918,39.099],[-29.481,22.556,34.327],[-23.413,25.107,46.515],
[-26.040,26.797,39.729],[-27.753,22.127,47.713],[-35.384,25.779,40.064],
[-34.046,29.286,43.040],[-34.982,25.237,43.149],[-29.266,29.202,45.589],
[-34.265,27.977,49.329],[-27.486,22.324,38.257],[-36.560,23.894,47.994],
[-30.594,30.543,39.781],[-34.673,30.385,50.651],[-25.182,16.357,35.819],
[-36.662,23.644,37.375],[-23.685,28.843,36.379],[-38.632,26.888,45.278],
[-40.121,29.406,44.590],[-22.257,29.535,50.590],[-31.546,28.554,48.685],
[-30.808,33.344,38.997],[-36.429,32.346,41.170],[-28.725,29.956,36.168],
[-28.096,32.055,38.432],[-24.805,22.157,49.971],[-36.957,28.383,49.339],
[-34.935,15.514,44.983],[-29.784,15.477,47.213],[-27.099,19.616,49.018],
[-27.681,25.394,34.719],[-30.944,19.177,36.567],[-34.502,20.605,44.436],
[-34.279,20.683,39.498],[-30.332,21.698,38.081],[-30.316,21.368,50.733],
[-29.293,20.706,40.782],[-36.534,21.270,48.725],[-34.139,20.481,34.505],
[-40.991,18.362,40.856],[-38.274,18.715,36.398],[-29.468,21.316,45.564],
[-32.623,16.453,40.506],[-37.477,21.539,39.496],[-34.069,18.674,37.299],
[-28.104,17.307,44.769],[-34.109,21.973,36.981],[-25.139,33.001,49.412],
[-36.620,25.549,50.484],[-31.407,20.273,43.815],[-36.438,16.930,40.043],
[-32.627,21.499,46.041],[-31.967,26.126,36.574],[-25.816,24.238,47.049],
[-28.093,17.225,34.990],[-30.558,25.998,48.317],[-26.960,20.248,36.224],
[-37.435,26.003,41.882],[-29.341,27.179,37.213],[-39.246,24.678,48.008],
[-34.096,23.466,50.870],[-29.156,30.574,48.366],[-22.695,21.493,37.141],
[-22.546,27.746,48.267],[-25.512,29.640,39.103],[-27.204,24.788,44.250],
[-35.061,28.441,40.156],[-38.976,23.929,35.234],[-33.179,27.312,41.467],
[-33.438,30.357,47.957],[-33.530,31.918,40.853],[-28.870,27.951,48.157],
[-26.652,27.909,35.978],[-26.372,31.683,47.704],[-22.764,32.991,33.442],
[-38.055,17.044,48.281],[-27.562,25.345,37.722],[-37.701,29.293,46.310],
[-33.211,28.858,33.705],[-34.738,27.776,36.333],[-37.061,26.410,37.545],
[-30.459,18.640,47.196],[-29.234,16.899,37.356],[-31.415,28.959,36.247],
[-31.367,18.954,39.344],[-31.859,21.732,48.549],[-29.741,16.849,39.901],
[-34.648,32.274,36.269],[-33.746,16.151,35.684],[-24.560,17.608,48.617],
[-26.448,22.077,45.021],[-38.079,28.761,35.601],[-37.377,21.144,35.456],
[-25.542,32.479,45.284],[-37.111,16.254,43.435],[-37.430,29.593,40.099],
[-40.184,30.192,50.019],[-35.150,32.689,51.699],[-30.814,30.399,33.728],
[-28.291,28.455,52.384],[-26.511,29.923,34.226],[-39.391,25.646,50.804],
[-35.938,25.337,34.050],[-32.576,24.712,48.956],[-30.506,27.212,34.389],
[-27.414,25.185,49.070],[-37.914,23.105,50.805],[-33.611,24.454,35.462],
[-34.262,22.717,47.778],[-33.457,23.273,33.010],[-29.837,26.334,51.006],
[-24.540,26.166,34.808],[-34.437,19.834,47.716],[-32.977,17.557,46.552],
[-28.668,24.229,46.416],[-27.685,21.595,50.869],[-27.405,24.049,51.499],
[-36.017,21.500,51.723],[-32.484,14.372,47.078],[-31.503,19.023,50.514],
[-29.111,17.510,49.276],[-27.838,18.576,51.301],[-38.039,18.501,51.157],
[-34.093,14.949,49.498],[-31.871,14.973,51.769],[-31.978,33.839,35.715],
[-40.078,26.591,35.460],[-38.332,34.994,39.192],[-39.043,30.636,42.152],
[-22.893,32.295,43.831],[-22.965,30.434,46.690],[-24.305,27.879,50.596],
[-24.387,26.463,37.537],[-22.928,24.487,38.691],[-22.988,27.493,44.731],
[-25.550,29.556,46.245],[-21.892,25.911,36.711],[-24.537,21.975,35.110],
[-24.939,22.457,39.682],[-40.588,24.938,44.341],[-24.423,25.284,49.178],
[-22.767,23.213,51.341],[-25.621,18.660,34.219],[-25.201,20.083,38.332],
[-24.674,20.953,43.132],[-25.229,21.323,47.402],[-39.301,20.577,49.110],
[-22.069,16.393,51.941],[-24.547,17.057,38.933],[-37.584,18.746,41.398],
[-39.009,21.803,45.212],[-21.769,18.842,51.235],[-22.966,17.884,34.081],
[-38.970,17.347,38.909],[-24.576,31.277,41.739],[-40.198,17.520,46.781],
[-37.110,29.472,52.281],[-36.163,30.894,47.699],[-31.528,31.686,49.812],
[-28.048,33.513,49.336],[-22.902,16.304,47.171],[-37.504,34.290,46.914],
[-35.174,33.696,47.836],[-29.810,34.844,46.173],[-28.427,31.835,45.894],
[-27.380,34.719,47.201],[-36.721,32.365,44.221],[-34.953,30.565,45.146],
[-33.044,32.240,43.864],[-30.593,18.262,42.151],[-24.755,13.127,42.050],
[-25.333,16.098,43.084],[-39.221,14.717,41.440],[-38.563,15.518,35.904],
[-32.528,13.686,40.532],[-27.312,14.283,42.585],[-26.960,15.782,37.871],
[-39.966,30.565,37.111],[-37.776,17.467,33.548],[-31.743,16.395,37.888],
[-30.731,31.545,36.904],[-22.738,31.710,39.522],[-39.155,15.629,50.284],
[-36.203,16.215,50.594],[-37.424,33.046,34.596],[-24.850,15.847,51.369],
[-24.954,30.701,51.049],[-27.836,42.554,41.173],[-29.698,43.636,43.636],
[-36.502,41.704,45.611],[-26.869,45.644,41.555],[-36.395,45.256,43.537],
[-40.589,40.395,40.099],[-38.870,41.589,42.075],[-28.474,50.178,42.195],
[-31.536,45.843,44.429],[-23.428,36.880,44.657],[-27.055,39.016,44.754],
[-34.887,38.978,41.901],[-27.060,37.433,40.414],[-39.789,38.878,42.243],
[-33.204,41.444,41.836],[-21.044,37.057,45.619],[-36.777,36.156,45.068],
[-29.933,44.959,39.700],[-34.305,43.262,47.391],[-24.781,36.300,41.688],
[-37.039,40.073,43.531],[-34.194,46.108,45.742],[-37.939,46.739,37.809],
[-24.955,45.837,43.420],[-25.498,36.975,46.426],[-33.278,34.948,43.260],
[-25.354,47.978,41.622],[-25.886,49.023,38.408],[-24.779,45.517,48.582],
[-29.560,43.594,35.850],[-29.168,41.160,34.079],[-22.183,40.569,47.194],
[-27.806,41.981,37.582],[-25.291,41.251,47.120],[-34.169,44.229,40.556],
[-33.660,46.428,42.469],[-34.379,42.592,44.041],[-29.799,48.057,44.296],
[-33.702,46.684,48.810],[-26.995,44.469,36.468],[-36.812,42.890,47.971],
[-31.061,49.353,40.335],[-36.473,47.414,49.747],[-22.936,33.821,36.005],
[-35.654,41.394,40.281],[-22.846,47.924,34.575],[-39.682,46.508,47.424],
[-38.840,48.355,45.141],[-23.252,47.680,49.796],[-30.937,46.099,49.191],
[-32.839,50.138,38.146],[-38.987,49.863,41.932],[-26.602,50.594,35.861],
[-26.571,51.705,39.538],[-24.512,38.673,50.092],[-37.033,46.465,47.004],
[-35.053,34.143,45.190],[-30.618,34.811,48.728],[-27.563,39.041,48.188],
[-26.041,43.419,33.382],[-29.194,37.347,33.939],[-34.389,39.262,44.459],
[-33.939,39.302,39.335],[-30.499,40.749,39.247],[-30.180,41.140,50.352],
[-28.016,39.864,40.775],[-37.306,38.706,46.159],[-33.683,38.593,34.486],
[-39.965,36.753,40.182],[-39.452,38.074,36.608],[-28.926,41.000,45.297],
[-32.340,37.242,41.775],[-36.999,38.750,39.381],[-32.676,39.204,37.002],
[-28.113,36.297,44.674],[-34.780,40.955,36.457],[-25.302,53.995,50.063],
[-37.076,44.745,49.704],[-30.360,41.616,42.010],[-34.868,36.577,40.110],
[-32.365,40.163,45.652],[-30.489,45.971,36.662],[-23.388,42.282,45.133],
[-27.260,38.541,35.517],[-29.468,45.577,46.458],[-25.579,41.172,36.073],
[-37.149,43.832,41.145],[-28.169,46.685,38.103],[-39.460,42.198,48.374],
[-32.777,41.477,51.257],[-28.910,47.882,47.795],[-20.151,41.245,36.759],
[-21.716,45.372,49.325],[-23.312,47.600,37.156],[-26.644,43.217,44.020],
[-34.432,47.293,39.485],[-38.533,43.154,34.562],[-32.841,45.133,38.047],
[-31.921,47.601,46.598],[-35.449,50.566,39.833],[-27.715,45.833,44.108],
[-27.270,45.973,34.481],[-24.658,48.256,47.031],[-22.118,50.410,35.116],
[-39.052,37.876,48.572],[-25.099,45.693,37.683],[-37.774,49.256,47.831],
[-32.498,47.886,33.431],[-33.689,48.090,36.700],[-37.072,44.931,36.000],
[-30.269,37.986,46.015],[-29.049,35.015,36.648],[-30.342,48.604,37.090],
[-30.332,38.936,36.001],[-31.278,39.454,48.799],[-30.987,37.230,39.299],
[-32.217,50.232,35.305],[-34.400,35.590,35.683],[-26.147,36.366,49.021],
[-26.118,40.879,43.001],[-36.252,47.762,35.887],[-38.154,40.446,33.116],
[-24.847,48.362,44.452],[-36.311,34.935,42.729],[-37.341,48.515,40.020],
[-39.988,50.940,49.172],[-34.744,48.786,52.652],[-29.197,48.204,32.734],
[-28.212,46.557,50.495],[-25.803,47.756,35.938],[-39.680,43.755,51.336],
[-34.312,45.387,35.006],[-32.271,43.623,49.259],[-30.533,45.785,33.531],
[-27.237,43.638,46.941],[-37.679,42.235,50.482],[-32.215,43.794,35.843],
[-34.671,40.610,47.447],[-32.955,41.599,34.552],[-29.909,43.761,50.826],
[-23.878,45.125,31.982],[-33.254,37.812,48.655],[-34.340,36.992,46.500],
[-30.838,42.739,46.495],[-28.355,39.440,51.508],[-28.011,41.948,51.854],
[-35.865,40.273,51.252],[-32.734,34.840,47.260],[-32.021,37.426,51.002],
[-28.815,36.643,50.372],[-26.504,37.116,51.607],[-36.201,37.576,48.660],
[-32.862,34.298,50.297],[-32.863,35.658,52.814],[-34.066,52.495,35.523],
[-38.737,46.744,34.776],[-40.077,51.182,39.961],[-41.053,48.566,43.212],
[-23.280,50.548,44.219],[-21.170,50.283,47.320],[-24.287,45.839,51.335],
[-22.716,45.297,39.267],[-23.308,42.730,39.570],[-21.223,43.737,46.744],
[-23.136,46.076,46.476],[-22.441,43.023,36.530],[-22.807,39.843,35.664],
[-24.242,40.136,40.027],[-40.585,43.348,43.330],[-24.744,42.788,49.418],
[-22.689,41.264,50.787],[-24.679,37.502,32.280],[-26.257,39.182,37.986],
[-23.283,38.555,42.402],[-24.259,39.404,45.484],[-39.822,38.355,51.179],
[-23.235,33.671,51.569],[-24.966,35.942,38.749],[-37.449,37.595,42.891],
[-39.934,40.842,44.370],[-22.908,35.319,49.359],[-22.751,36.871,34.913],
[-40.365,35.890,37.478],[-23.651,49.956,40.089],[-40.022,36.087,46.292],
[-40.102,48.342,53.101],[-34.816,48.720,46.951],[-30.535,49.949,46.777],
[-28.632,50.841,48.422],[-23.491,33.312,47.321],[-38.126,52.097,47.307],
[-34.269,49.225,44.427],[-30.038,51.474,44.305],[-27.550,48.549,45.639],
[-24.932,52.676,45.439],[-39.891,50.854,44.644],[-35.551,48.497,42.099],
[-33.371,50.171,41.938],[-30.803,36.129,43.656],[-25.564,34.974,44.027],
[-23.754,33.981,40.950],[-39.644,34.525,41.624],[-39.355,33.627,36.437],
[-33.060,34.708,39.739],[-27.360,34.470,41.966],[-24.934,32.879,38.177],
[-38.608,50.275,37.911],[-37.386,36.104,35.789],[-31.737,36.373,35.262],
[-29.457,50.234,34.633],[-22.860,50.602,37.785],[-39.680,35.517,49.658],
[-37.485,32.509,49.509],[-36.355,50.635,34.909],[-28.536,34.451,52.316],
[-24.615,50.083,49.181],[-23.623,55.589,45.278],[-27.835,57.194,44.680],
[-35.130,56.252,41.794],[-27.429,56.690,40.191],[-39.655,57.029,41.023],
[-21.980,58.060,47.049],[-37.074,54.871,44.762],[-24.600,53.147,42.677],
[-37.623,57.348,42.723],[-24.719,54.482,47.513],[-32.732,52.246,43.665],
[-25.383,53.413,37.278],[-25.419,57.368,50.911],[-34.734,52.083,45.565],
[-30.291,52.676,47.480],[-26.768,57.422,48.264],[-30.852,54.497,33.518],
[-35.339,57.352,44.469],[-34.271,58.758,39.431],[-28.739,58.480,42.425],
[-38.063,58.385,45.276],[-35.695,55.528,34.880],[-39.323,54.442,40.126],
[-39.422,56.750,35.538],[-33.138,54.277,41.789],[-36.722,56.893,39.571],
[-33.478,57.578,36.163],[-28.725,55.334,46.885],[-35.437,-59.647,35.411],
[-31.801,55.165,44.085],[-36.133,54.003,39.816],[-31.389,57.541,45.772],
[-28.294,55.195,35.635],[-28.223,57.453,37.490],[-22.922,58.201,37.508],
[-38.007,55.150,47.719],[-31.350,55.342,47.424],[-28.296,54.222,40.353],
[-32.114,56.764,38.491],[-30.704,58.034,48.321],[-30.864,54.559,39.751],
[-35.493,54.526,37.215],[-24.212,57.086,48.462],[-37.261,57.958,33.564],
[-37.322,51.594,44.379],[-34.205,55.426,48.751],[-33.890,55.327,45.958],
[-28.799,58.560,50.248],[-36.821,57.242,52.111],[-33.175,53.202,47.411],
[-32.045,56.011,51.381],[-29.737,55.847,49.608],[-25.500,51.981,53.354],
[-38.092,58.192,48.411],[-32.796,50.682,51.269],[-31.760,53.277,51.624],
[-22.921,57.641,34.521],[-23.409,54.993,34.264],[-25.418,56.692,37.952],
[-25.047,57.589,45.639],[-27.234,-59.623,46.357],[-40.594,58.108,49.734],
[-20.944,51.229,50.835],[-25.023,54.761,40.003],[-39.584,55.039,43.081],
[-40.798,57.970,43.784],[-22.332,55.278,49.203],[-20.735,53.348,34.273],
[-39.800,55.890,38.129],[-39.510,53.342,45.344],[-23.210,51.788,47.590],
[-30.527,53.215,42.176],[-25.530,50.480,41.949],[-23.693,52.537,40.192],
[-37.922,52.806,41.815],[-39.378,52.841,37.453],[-32.138,52.019,40.602],
[-27.919,52.967,42.973],[-28.043,52.595,37.001],[-37.957,54.526,35.927],
[-30.391,54.247,36.820],[-39.036,53.429,50.637],[-35.428,51.699,51.006],
[-26.860,50.418,50.656],[-39.488,-52.740,58.146],[-26.067,-55.424,58.090],
[-27.740,-48.216,57.898],[-39.947,-47.711,56.334],[-26.021,-46.181,57.687],
[-29.593,-49.274,56.014],[-30.238,-51.584,54.175],[-27.710,-50.009,-58.965],
[-27.544,-50.888,55.243],[-29.810,-44.756,58.352],[-25.513,-57.706,54.463],
[-36.845,-51.612,57.012],[-24.263,-44.225,53.417],[-30.235,-42.143,57.580],
[-37.111,-43.756,58.199],[-27.787,-42.298,53.935],[-27.980,-41.166,56.686],
[-26.107,-49.409,53.243],[-31.255,-56.476,54.258],[-35.228,-52.547,-59.233],
[-29.038,-51.422,57.424],[-34.986,-56.074,52.897],[-40.569,-56.441,55.741],
[-37.285,-54.941,58.731],[-34.370,-54.004,54.896],[-36.213,-52.744,52.798],
[-35.872,-59.019,56.861],[-32.303,-49.761,57.831],[-27.551,-55.187,54.737],
[-27.606,-53.255,58.977],[-28.374,-45.964,56.307],[-23.278,-53.249,55.836],
[-23.865,-45.431,56.063],[-37.859,-50.673,54.001],[-36.483,-49.659,59.129],
[-36.482,-40.973,56.417],[-28.931,-46.784,53.592],[-22.391,-40.903,55.263],
[-26.789,-48.184,55.363],[-33.680,-46.750,52.653],[-33.274,-47.671,55.719],
[-36.909,-49.043,55.757],[-28.357,-57.836,55.701],[-30.936,-45.278,54.475],
[-30.558,-54.568,56.497],[-28.806,-56.711,58.615],[-33.970,-43.871,55.488],
[-33.432,-56.514,56.015],[-35.500,-45.081,53.697],[-37.634,-54.830,53.882],
[-38.869,-46.346,58.467],[-29.772,-44.193,52.298],[-27.013,-44.728,53.339],
[-36.189,-48.115,52.523],[-31.373,-48.889,53.912],[-34.046,-50.050,54.120],
[-32.269,-53.243,52.602],[-23.848,-48.961,51.612],[-33.260,-40.898,54.186],
[-38.464,-47.130,53.547],[-41.442,-41.406,59.292],[-23.779,-48.730,56.165],
[-22.167,-50.900,59.003],[-22.589,-50.614,53.442],[-24.762,-52.096,53.132],
[-25.411,-52.125,-58.976],[-24.858,-55.338,51.759],[-23.242,-54.122,59.022],
[-23.871,-56.788,58.524],[-23.160,-56.226,53.881],[-39.538,-57.053,58.089],
[-25.853,-40.414,-59.042],[-39.011,-59.179,56.235],[-31.082,-59.738,58.646],
[-26.055,-58.183,57.126],[-39.645,-44.283,56.413],[-37.414,-57.893,52.558],
[-31.170,-57.291,57.186],[-30.441,-41.825,54.840],[-24.177,-42.338,58.258],
[-36.954,-42.789,52.662],[-39.441,-34.502,57.852],[-28.289,-36.645,-59.998],
[-27.955,-28.926,58.390],[-40.094,-28.591,55.422],[-25.032,-26.695,58.129],
[-28.014,-31.931,55.413],[-30.539,-33.397,53.686],[-26.996,-32.900,57.853],
[-27.101,-33.455,53.171],[-29.711,-25.369,58.526],[-24.497,-40.263,53.822],
[-35.603,-31.865,54.936],[-22.634,-27.444,53.300],[-30.945,-22.936,55.931],
[-36.994,-23.732,58.031],[-26.973,-22.995,53.303],[-27.832,-24.162,56.038],
[-26.941,-30.615,53.025],[-31.059,-37.348,53.988],[-35.605,-34.398,56.832],
[-30.693,-34.343,57.598],[-35.818,-36.790,52.226],[-39.929,-36.520,55.326],
[-37.890,-37.019,58.541],[-35.690,-37.555,57.018],[-35.842,-34.210,53.640],
[-35.619,-39.698,58.790],[-30.418,-30.538,55.032],[-27.681,-36.916,54.580],
[-28.758,-34.563,55.821],[-29.635,-27.281,55.944],[-21.125,-35.338,55.470],
[-23.219,-25.652,56.058],[-37.724,-32.126,53.345],[-34.228,-29.453,58.771],
[-34.189,-24.609,57.488],[-26.079,-27.945,52.984],[-22.422,-24.134,53.517],
[-27.411,-29.106,55.711],[-31.118,-28.647,51.028],[-33.172,-27.208,57.806],
[-37.122,-29.892,55.456],[-29.325,-38.619,56.031],[-31.828,-25.525,55.794],
[-32.369,-36.772,56.662],[-30.291,-38.390,58.410],[-34.078,-23.432,55.088],
[-34.650,-38.661,54.374],[-36.259,-25.063,55.433],[-38.297,-35.236,52.632],
[-38.983,-28.470,58.730],[-30.720,-24.712,50.831],[-26.083,-25.452,51.736],
[-35.282,-28.154,53.208],[-29.473,-28.228,53.281],[-33.090,-31.062,54.720],
[-33.034,-34.306,53.274],[-24.522,-30.462,54.006],[-31.789,-21.138,53.956],
[-38.914,-28.265,52.926],[-41.159,-22.982,-58.834],[-22.764,-29.913,56.156],
[-22.885,-32.081,58.773],[-21.919,-31.073,53.463],[-23.749,-34.168,52.785],
[-25.049,-32.679,-58.553],[-25.562,-38.043,52.560],[-25.650,-36.214,57.573],
[-23.740,-39.203,57.355],[-22.910,-37.180,53.125],[-39.332,-40.087,56.778],
[-27.098,-23.285,58.777],[-39.439,-40.379,53.551],[-33.732,-41.492,-59.150],
[-26.789,-38.603,57.768],[-39.151,-25.128,55.967],[-37.097,-37.383,54.665],
[-32.346,-40.486,56.732],[-29.892,-23.444,53.124],[-20.939,-22.929,56.167],
[-36.889,-23.392,52.156],[-39.339,-15.472,57.752],[-25.279,-14.947,53.979],
[-29.251,-8.749,58.030],[-39.565,-9.802,55.681],[-26.805,-9.961,57.874],
[-30.325,-12.224,57.110],[-28.605,-13.870,52.865],[-26.252,-13.775,57.526],
[-27.820,-14.235,55.480],[-30.787,-7.080,-59.570],[-25.322,-23.733,56.997],
[-35.456,-12.635,56.778],[-23.247,-8.393,53.390],[-31.304,-4.939,57.828],
[-37.915,-6.011,58.037],[-27.250,-5.330,54.027],[-28.598,-5.470,57.726],
[-26.023,-12.542,53.189],[-29.699,-19.151,53.614],[-36.884,-14.759,58.416],
[-29.347,-14.599,57.716],[-35.180,-19.143,52.563],[-39.806,-17.395,55.612],
[-37.143,-16.529,55.830],[-34.684,-17.389,56.356],[-35.192,-14.811,54.799],
[-36.238,-20.047,58.559],[-31.516,-11.575,54.411],[-26.693,-18.622,52.716],
[-28.131,-16.730,54.440],[-29.362,-10.033,55.535],[-21.023,-14.883,55.441],
[-24.321,-7.680,56.025],[-38.678,-11.351,52.932],[-33.190,-11.013,57.465],
[-34.336,-5.442,58.558],[-26.659,-9.570,54.137],[-23.351,-5.583,53.810],
[-27.066,-11.736,55.835],[-32.655,-8.562,51.599],[-34.757,-9.769,55.682],
[-37.336,-11.338,55.285],[-28.501,-19.558,56.770],[-30.601,-7.234,56.044],
[-30.553,-17.382,57.293],[-31.088,-20.143,58.086],[-33.177,-6.106,56.154],
[-35.429,-20.396,55.695],[-35.292,-8.011,53.479],[-38.557,-15.518,53.721],
[-40.267,-7.673,57.148],[-30.212,-6.896,52.099],[-25.742,-7.221,53.139],
[-36.001,-11.337,52.261],[-30.342,-9.476,52.727],[-34.440,-12.034,54.390],
[-31.350,-14.339,52.799],[-23.720,-11.542,51.703],[-32.833,-4.225,54.191],
[-40.068,-7.719,53.947],[-39.967,-4.217,-58.678],[-24.141,-12.270,55.276],
[-23.096,-11.634,59.064],[-22.243,-13.963,53.382],[-23.047,-16.368,53.248],
[-24.210,-14.180,-59.709],[-24.389,-20.083,51.352],[-27.384,-16.770,57.123],
[-23.866,-20.622,58.179],[-22.102,-18.787,53.428],[-38.068,-19.230,56.732],
[-25.447,-4.316,59.599],[-39.167,-20.447,54.231],[-33.471,-21.686,57.075],
[-26.830,-21.584,56.004],[-38.021,-6.038,55.148],[-37.192,-20.921,52.290],
[-32.614,-19.081,56.081],[-30.641,-4.171,52.657],[-24.806,-5.102,56.414],
[-37.847,-2.706,53.911],[-39.135,4.059,58.316],[-27.301,0.916,58.582],
[-28.073,10.048,58.644],[-39.737,8.958,55.085],[-25.676,8.137,-58.682],
[-30.255,5.929,57.602],[-30.827,5.394,53.355],[-25.451,5.584,59.083],
[-28.941,4.184,54.690],[-30.653,10.981,57.413],[-23.958,-2.646,54.342],
[-36.201,5.261,57.599],[-22.689,10.375,54.147],[-30.123,15.225,56.576],
[-37.425,12.758,58.629],[-27.135,13.399,53.254],[-27.677,12.746,56.522],
[-26.656,6.381,53.825],[-30.371,-1.761,54.172],[-34.787,2.333,58.522],
[-28.603,3.802,57.538],[-32.879,1.051,53.141],[-38.997,-0.372,56.212],
[-38.814,0.804,-58.409],[-35.579,1.084,56.368],[-35.465,2.917,54.432],
[-36.863,-0.324,58.561],[-31.431,8.467,56.866],[-27.728,0.134,53.588],
[-28.007,1.693,55.837],[-28.728,9.564,56.037],[-21.598,3.267,55.614],
[-24.563,12.046,56.905],[-38.181,6.987,53.675],[-33.458,8.436,59.132],
[-33.075,11.795,56.728],[-27.707,8.784,53.824],[-20.954,12.793,54.095],
[-26.757,7.385,56.363],[-32.795,9.852,52.573],[-33.619,9.199,55.071],
[-37.762,7.360,56.260],[-28.070,-0.898,56.335],[-31.210,10.503,54.844],
[-30.882,2.291,57.534],[-29.279,-0.960,-59.096],[-34.531,11.530,54.279],
[-33.465,-0.665,55.067],[-36.384,10.100,55.392],[-39.355,0.084,53.120],
[-38.516,10.232,57.129],[-29.756,10.722,52.618],[-25.963,11.311,54.795],
[-35.262,7.545,53.118],[-30.897,8.300,51.820],[-34.016,6.718,56.211],
[-34.092,5.030,53.337],[-23.027,6.665,52.943],[-33.046,14.766,54.927],
[-38.211,9.715,52.891],[-40.774,14.140,-58.544],[-24.116,8.438,57.360],
[-23.802,4.406,57.508],[-23.448,4.805,54.806],[-23.243,1.824,54.137],
[-27.296,3.212,-58.740],[-24.548,1.095,51.524],[-25.813,0.057,55.833],
[-25.013,-2.387,57.003],[-22.880,-0.760,52.387],[-39.042,-1.704,58.986],
[-26.644,12.799,59.113],[-38.364,-3.156,56.788],[-31.866,-1.875,-59.687],
[-27.428,-3.343,55.852],[-38.419,12.989,55.764],[-36.767,-0.026,54.464],
[-29.758,-2.904,57.035],[-30.756,13.096,54.704],[-22.325,11.777,58.284],
[-36.976,12.255,53.624],[-40.160,21.529,57.924],[-28.223,20.026,58.802],
[-28.652,28.491,57.199],[-39.669,25.834,56.593],[-25.845,28.236,57.034],
[-30.267,24.489,57.087],[-29.931,23.047,52.903],[-26.017,23.597,56.570],
[-27.464,22.391,54.631],[-29.336,29.552,-59.214],[-24.077,14.407,54.007],
[-36.966,22.513,54.712],[-22.682,28.375,53.103],[-30.906,33.658,56.776],
[-36.283,31.835,57.967],[-27.215,32.857,55.224],[-28.697,31.170,56.822],
[-26.790,25.266,53.813],[-29.880,17.064,54.334],[-36.262,22.573,57.300],
[-28.476,22.267,57.253],[-34.499,19.318,53.750],[-39.469,18.827,57.002],
[-37.489,20.071,58.443],[-35.564,20.066,56.352],[-34.084,21.725,55.074],
[-35.365,17.222,57.776],[-32.845,26.580,53.918],[-25.955,18.962,54.341],
[-28.468,19.989,54.542],[-30.496,28.351,54.479],[-23.505,21.800,56.462],
[-23.199,28.448,56.117],[-38.929,24.681,53.962],[-33.277,25.913,-58.562],
[-32.870,31.720,57.786],[-27.356,27.918,54.698],[-21.701,32.404,54.181],
[-27.657,25.824,56.761],[-33.153,28.774,52.481],[-33.704,29.160,56.657],
[-37.376,26.211,55.485],[-28.179,18.265,56.522],[-31.886,30.936,54.927],
[-30.601,19.697,56.855],[-30.418,17.147,-59.617],[-35.097,31.406,55.081],
[-34.860,16.169,53.355],[-35.733,28.312,54.620],[-37.059,19.871,53.685],
[-38.661,29.655,57.181],[-30.603,30.594,52.248],[-25.530,29.390,53.481],
[-33.568,26.099,51.201],[-30.112,25.626,53.724],[-33.680,24.021,53.406],
[-32.253,21.076,52.707],[-24.416,25.372,52.275],[-33.207,33.627,54.256],
[-40.240,27.620,53.619],[-40.290,32.326,58.358],[-24.044,25.669,56.630],
[-21.438,24.304,57.869],[-23.489,23.721,54.108],[-23.322,20.202,53.477],
[-23.872,21.272,-59.008],[-23.578,17.498,53.858],[-25.423,20.699,57.779],
[-24.921,16.965,57.673],[-21.118,17.148,54.840],[-39.834,15.753,57.378],
[-25.385,32.457,-58.831],[-36.611,15.214,55.640],[-33.299,15.214,57.873],
[-27.355,15.293,57.882],[-40.083,30.950,55.595],[-37.589,16.716,53.403],
[-32.524,17.822,56.975],[-30.369,32.834,54.060],[-24.149,30.711,57.948],
[-37.756,31.861,53.837],[-41.195,40.071,57.992],[-27.451,36.545,57.764],
[-28.252,45.490,58.036],[-40.226,45.575,56.295],[-24.894,45.882,58.916],
[-29.633,42.846,53.896],[-30.620,40.500,52.987],[-26.563,43.567,59.184],
[-26.232,41.530,54.436],[-29.162,47.671,57.350],[-23.374,34.512,54.392],
[-34.953,42.838,57.693],[-23.737,47.792,53.699],[-31.217,51.860,56.451],
[-36.753,50.509,57.826],[-28.058,51.719,52.867],[-26.587,48.874,55.336],
[-25.465,44.569,53.596],[-31.529,38.188,53.767],[-36.940,40.858,57.513],
[-30.298,42.141,58.453],[-35.087,37.447,52.123],[-41.366,38.541,55.793],
[-38.864,36.754,57.340],[-34.632,38.618,55.883],[-35.075,41.504,55.272],
[-35.276,36.618,57.670],[-33.145,44.337,55.676],[-28.303,37.906,54.322],
[-26.356,39.109,55.876],[-30.512,44.845,55.669],[-22.718,41.262,56.737],
[-24.013,48.613,57.291],[-37.857,41.494,53.153],[-34.706,45.701,57.766],
[-34.087,49.301,58.026],[-28.145,44.955,53.305],[-22.592,50.775,53.535],
[-26.375,44.971,56.004],[-32.920,46.656,53.167],[-34.341,47.638,55.340],
[-37.108,43.393,55.503],[-29.103,35.367,55.038],[-30.682,47.566,54.990],
[-31.420,38.268,57.187],[-29.407,34.640,59.333],[-34.433,50.477,55.614],
[-34.745,36.181,54.507],[-37.294,48.137,54.390],[-37.521,38.843,52.953],
[-38.281,48.146,56.938],[-30.717,48.972,50.299],[-25.696,48.275,51.479],
[-34.666,44.612,50.990],[-30.585,46.233,52.129],[-32.652,42.568,53.730],
[-33.691,39.896,53.207],[-23.362,43.139,52.398],[-34.073,53.256,53.060],
[-40.191,45.699,53.252],[-40.759,50.837,58.785],[-24.045,46.459,55.894],
[-23.444,43.062,58.664],[-23.998,43.292,55.630],[-23.978,39.531,54.915],
[-26.634,39.687,58.570],[-24.468,36.960,53.557],[-24.999,37.665,57.684],
[-25.505,34.247,56.793],[-21.997,37.199,54.447],[-40.631,35.310,55.565],
[-26.558,50.696,58.606],[-38.395,33.163,56.828],[-33.983,34.198,57.759],
[-26.274,35.158,54.298],[-40.109,49.981,56.091],[-37.184,34.943,51.840],
[-31.841,35.964,55.511],[-30.451,50.105,52.814],[-23.090,51.258,56.644],
[-37.258,51.173,53.364],[-26.746,55.573,59.033],[-24.887,54.371,55.427],
[-29.302,55.302,52.253],[-34.374,59.086,59.383],[-34.394,56.239,52.849],
[-37.868,57.216,56.454],[-37.869,54.665,57.331],[-32.881,-59.603,56.229],
[-35.009,58.430,54.638],[-36.506,55.295,54.419],[-27.548,54.810,54.342],
[-28.706,58.254,55.572],[-23.523,58.763,55.511],[-28.903,54.493,56.786],
[-30.689,56.849,56.824],[-31.931,54.487,57.228],[-34.849,53.497,55.770],
[-37.557,-59.969,53.974],[-22.453,56.750,51.465],[-24.683,55.414,52.737],
[-25.825,58.145,58.063],[-24.078,55.773,57.749],[-23.580,52.232,50.738],
[-40.214,55.267,56.434],[-38.445,53.328,54.289],[-32.684,51.406,59.076],
[-27.420,51.853,55.828],[-36.566,54.528,51.543],[-31.322,54.319,54.586],
[-35.057,-57.328,-13.572],[-37.993,-57.616,-20.514],[-26.438,58.524,-15.999],
[-38.636,-47.904,-29.718],[-31.273,-45.238,-32.357],[-36.359,-56.275,-25.790],
[-34.543,-50.589,-25.024],[-24.906,-56.374,-24.151],[-27.754,-47.866,-28.213],
[-29.264,-47.076,-32.066],[-32.107,-47.397,-33.543],[-39.788,-48.253,-33.980],
[-32.241,-53.092,-34.819],[-26.634,-54.996,-20.429],[-37.706,-51.984,-36.403],
[-35.768,-55.854,-36.740],[-28.386,-48.629,-23.826],[-34.984,-50.442,-22.401],
[-33.165,-52.114,-18.443],[-30.223,-51.963,-19.695],[-38.343,-49.939,-18.304],
[-32.718,-56.771,-19.046],[-40.981,-45.468,-34.322],[-41.904,-46.056,-31.920],
[-24.212,-44.303,-29.632],[-26.503,-56.787,-29.725],[-23.326,-55.827,-16.163],
[-36.793,-41.379,-29.231],[-32.663,-34.846,-37.857],[-23.499,-36.346,-31.307],
[-37.159,-29.860,-34.266],[-23.515,-28.885,-33.630],[-36.403,-40.866,-38.343],
[-36.024,-42.404,-43.204],[-32.037,-33.494,-33.218],[-42.589,-33.463,-31.215],
[-25.295,-31.304,-33.157],[-26.855,-27.394,-29.951],[-34.482,-38.428,-25.946],
[-25.254,-39.942,-39.237],[-26.219,-44.669,-43.903],[-41.759,-28.668,-31.344],
[-36.326,-31.725,-40.451],[-30.855,-42.753,-31.410],[-22.655,-43.262,-34.651],
[-27.848,-14.381,-47.807],[-34.303,-20.235,-43.903],[-32.723,-14.170,-54.497],
[-26.454,-19.251,-37.742],[-27.469,-21.740,-38.387],[-25.268,-8.329,-43.911],
[-29.538,-28.739,-48.252],[-29.410,-13.184,-45.211],[-24.154,-27.406,-51.818],
[-37.661,-21.913,-54.839],[-35.474,-22.830,-56.291],[-30.378,-21.142,-38.432],
[-28.129,-21.533,-34.227],[-24.460,-23.760,-51.748],[-25.786,-9.692,-47.501],
[-32.827,-16.902,-56.798],[-28.628,-3.014,-57.498],[-22.711,-7.613,-51.047],
[-36.299,-10.925,-53.869],[-32.413,-11.520,-51.635],[-34.573,-2.407,-53.544],
[-36.539,4.180,-51.415],[-25.432,-14.764,-55.934],[-31.302,2.893,-56.618],
[-32.146,-9.314,-47.070],[-30.374,-3.517,-50.868],[-30.794,-8.734,-52.002],
[-32.084,-6.431,-55.168],[-23.759,-8.618,58.362],[-23.721,1.834,-51.060],
[-30.874,-7.978,-49.518],[-37.501,-13.498,-57.319],[-41.608,-11.224,-50.779],
[-33.659,1.614,-49.718],[-32.930,-6.173,-43.935],[-40.385,-8.641,-52.085],
[-37.183,3.543,-53.925],[-32.559,5.109,-53.320],[-37.386,5.399,-58.122],
[-26.250,-11.471,-54.614],[-39.046,-12.774,-54.789],[-37.025,7.489,59.090],
[-29.536,11.719,-58.739],[-28.748,15.465,-58.139],[-35.840,14.544,-57.000],
[-26.166,15.434,-58.338],[-24.270,8.340,-53.780],[-38.373,6.372,-53.058],
[-29.928,-59.456,-1.409],[-35.031,-53.158,0.233],[-26.085,-57.869,-6.763],
[-35.855,-59.525,-0.267],[-32.701,-50.682,1.102],[-27.466,-58.592,-2.586],
[-32.299,-56.904,-10.668],[-40.811,-57.050,5.204],[-37.305,-50.637,-1.363],
[-41.991,-46.568,-15.295],[-38.969,-37.833,-16.630],[-30.448,-52.313,-14.700],
[-30.090,-46.901,-14.460],[-30.559,-41.403,-4.492],[-39.913,-46.990,-13.392],
[-35.711,-39.604,-16.557],[-35.982,-50.887,-14.325],[-40.033,-50.568,-16.565],
[-27.962,-38.376,-7.162],[-34.588,-45.388,-16.967],[-35.221,-39.696,-5.417],
[-30.887,-38.616,-2.746],[-24.516,-46.818,-20.202],[-27.458,-45.029,-0.191],
[-22.777,-33.611,-14.041],[-24.681,-44.638,-13.068],[-39.318,-34.136,-6.286],
[-37.132,-31.732,-12.898],[-22.968,-28.473,-14.881],[-38.031,-29.538,-14.757],
[-35.931,-22.945,-18.929],[-40.673,-25.833,-26.214],[-36.188,-27.862,-22.452],
[-30.587,-22.497,-21.620],[-34.668,-20.337,-16.744],[-35.063,-28.625,-15.734],
[-31.327,-23.867,-13.129],[-34.517,-31.847,-22.328],[-25.299,-25.399,-17.822],
[-37.560,-25.434,-21.445],[-31.274,-28.453,-27.951],[-28.685,-26.091,-19.897],
[-40.541,-27.963,-13.807],[-31.888,-26.724,-14.217],[-32.234,-36.006,-20.767],
[-31.509,-24.435,-30.005],[-21.969,-30.316,-27.273],[-22.399,-24.650,-13.811],
[-24.437,-30.870,-13.593],[-38.891,-38.227,-23.282],[-37.586,-30.418,-9.340],
[-41.740,-13.880,-32.452],[-32.083,-12.768,-34.521],[-37.172,-13.682,-33.350],
[-29.039,-15.881,-25.124],[-26.113,-0.218,-28.055],[-22.390,-17.139,-34.388],
[-29.003,-12.471,-39.023],[-36.011,-13.725,-38.377],[-32.910,-11.636,-20.265],
[-39.638,-19.778,-31.426],[-26.982,9.302,-42.442],[-41.469,3.662,-40.182],
[-32.607,7.923,-40.327],[-35.194,7.147,-36.031],[-37.113,1.541,-38.167],
[-29.027,4.808,-42.805],[-35.822,5.057,-40.951],[-38.022,7.993,-34.634],
[-35.653,13.841,-37.426],[-23.888,6.101,-50.279],[-27.567,9.180,-50.353],
[-35.758,-5.036,-44.692],[-31.682,4.631,-37.697],[-30.964,10.179,-35.692],
[-29.827,-1.563,-41.237],[-32.083,6.062,-49.729],[-38.977,9.469,-32.626],
[-31.481,9.579,-33.105],[-23.325,7.222,-33.061],[-25.421,-1.922,-41.058],
[-36.820,14.192,-41.953],[-32.964,24.481,-48.827],[-29.297,16.054,-47.910],
[-38.427,17.920,-47.575],[-24.777,17.958,-42.049],[-26.776,16.143,-43.892],
[-28.605,27.250,-49.027],[-28.217,24.013,-59.778],[-24.627,21.743,-41.382],
[-35.165,12.441,-43.666],[-37.096,20.786,-43.924],[-37.046,14.982,-54.141],
[-31.237,21.286,-56.567],[-25.809,22.346,-56.681],[-36.435,26.843,-54.957],
[-40.294,26.349,-55.673],[-35.783,33.804,-44.634],[-32.263,13.751,-55.317],
[-31.082,19.053,-38.849],[-41.901,26.904,-52.630],[-23.538,29.882,-50.390],
[-24.734,21.043,-51.014],[-22.935,12.588,-52.241],[-32.302,30.838,-50.018],
[-37.319,28.713,-53.140],[-30.697,24.686,59.654],[-37.103,17.811,-38.066],
[-35.353,32.892,-56.189],[-37.423,45.308,-52.698],[-25.656,24.958,58.782],
[-38.312,42.029,-56.850],[-28.772,28.931,-54.649],[-39.768,34.370,-47.422],
[-30.324,-49.679,12.529],[-23.790,-49.878,4.908],[-30.717,-43.861,5.390],
[-39.914,-42.543,15.723],[-24.288,-46.679,3.537],[-28.844,-56.686,10.778],
[-35.603,58.863,11.788],[-35.507,-52.794,2.970],[-30.394,-50.856,3.639],
[-29.035,-45.714,23.870],[-25.368,-50.413,17.688],[-25.455,-58.407,12.391],
[-24.024,-51.644,22.805],[-30.314,-38.266,15.755],[-27.543,-38.917,10.521],
[-40.019,-41.782,9.994],[-29.741,-59.450,17.003],[-30.401,-46.769,3.232],
[-37.014,-29.635,3.187],[-37.238,-34.348,1.275],[-39.889,-29.292,2.400],
[-31.561,-38.283,13.075],[-37.388,-40.202,1.729],[-32.617,-34.342,-1.029],
[-33.697,-29.781,11.188],[-32.653,-32.651,1.975],[-40.232,-27.081,4.212],
[-30.110,-41.622,3.619],[-34.252,-39.025,13.593],[-24.039,-24.589,4.801],
[-26.132,-29.987,3.735],[-25.604,-42.726,-0.071],[-25.821,-39.701,3.016],
[-35.019,-18.784,-5.059],[-37.949,-21.038,1.243],[-36.018,-12.694,-3.312],
[-27.675,-19.361,-8.220],[-38.766,-15.217,0.240],[-24.054,-17.073,-13.999],
[-31.975,-9.774,-2.053],[-32.214,-12.341,-13.160],[-36.216,-23.454,-0.228],
[-29.846,-20.198,-3.748],[-34.855,-22.425,-7.858],[-32.191,-12.808,-0.917],
[-31.053,-17.795,-11.224],[-39.470,-21.804,-10.673],[-36.310,-6.408,0.405],
[-34.259,-11.722,1.252],[-29.994,-17.009,4.873],[-23.523,-22.150,-9.319],
[-24.889,-20.193,-4.269],[-29.332,-10.377,-7.019],[-27.686,-25.950,0.163],
[-32.107,-14.771,-16.591],[-24.664,-12.690,-11.305],[-29.900,-21.447,8.073],
[-32.504,-8.486,-9.931],[-31.940,2.497,-19.728],[-24.268,-11.867,-13.637],
[-37.877,-2.147,-17.474],[-35.925,-10.758,-20.094],[-28.666,-8.190,-18.790],
[-35.873,-0.026,-13.564],[-31.586,-7.237,-15.076],[-35.973,-10.236,-17.280],
[-32.069,0.539,-25.941],[-30.778,-3.671,-23.832],[-35.080,-4.149,-8.246],
[-30.734,-2.736,-5.360],[-24.070,7.103,-13.748],[-24.136,2.009,-14.189],
[-40.542,-8.399,-15.541],[-29.007,9.148,-15.389],[-39.182,8.434,-15.136],
[-39.026,5.519,-18.182],[-31.057,-7.819,-12.460],[-38.652,-9.503,-11.222],
[-30.624,13.603,-21.591],[-26.313,16.840,-25.541],[-39.199,11.653,-21.499],
[-35.904,19.270,-20.903],[-36.349,14.001,-25.672],[-31.450,9.869,-14.731],
[-29.680,5.538,-26.976],[-34.735,18.675,-17.380],[-36.406,16.848,-18.364],
[-24.992,11.580,-31.131],[-25.627,10.565,-9.587],[-24.084,12.445,-14.309],
[-40.987,9.778,-16.485],[-24.060,9.830,-13.292],[-28.717,5.799,-21.677],
[-41.975,28.706,-32.216],[-36.733,29.695,-36.106],[-39.265,35.482,-31.761],
[-31.369,23.294,-37.476],[-35.231,27.486,-32.917],[-28.849,29.137,-32.320],
[-42.702,26.063,-32.143],[-31.681,26.597,-30.364],[-37.991,26.531,-34.806],
[-36.645,22.492,-32.432],[-34.072,27.989,-28.332],[-32.220,36.946,-28.316],
[-23.205,32.802,-41.078],[-36.585,34.317,-37.462],[-31.001,33.734,-40.830],
[-27.100,29.367,-24.844],[-34.834,28.152,-42.286],[-36.839,33.976,-25.676],
[-22.908,38.703,-31.966],[-22.916,26.682,-38.286],[-26.830,22.988,-36.709],
[-38.518,38.916,-29.016],[-25.139,22.645,-29.767],[-36.712,20.649,-36.047],
[-37.523,43.637,-40.213],[-28.927,48.226,-38.208],[-34.427,51.875,-40.303],
[-23.424,47.299,-51.714],[-41.338,51.989,-43.751],[-37.726,40.870,-37.347],
[-33.064,42.260,-32.554],[-29.622,42.313,-44.558],[-29.978,53.659,-40.388],
[-22.812,48.996,-53.754],[-38.835,42.972,-47.748],[-24.271,45.429,-34.801],
[-26.975,44.955,-41.943],[-26.398,55.072,-44.617],[-31.421,46.991,-51.922],
[-32.826,50.104,-36.273],[-24.279,44.204,-49.478],[-28.412,45.654,-30.653],
[-38.582,45.916,-32.875],[-41.448,45.945,-51.090],[-42.204,51.662,-49.498],
[-23.047,49.574,-33.891],[-40.833,47.698,-32.621],[-24.059,44.442,-32.135],
[-31.076,54.860,-45.909],[-35.358,-41.137,29.850],[-42.147,-38.596,26.285],
[-31.712,-45.981,23.540],[-29.073,-40.174,36.364],[-29.573,-45.017,27.886],
[-29.922,-39.932,33.853],[-30.564,-42.079,22.228],[-32.953,-42.512,18.953],
[-31.418,-47.311,28.883],[-26.469,-43.755,37.354],[-39.662,-47.061,24.212],
[-21.953,-46.914,21.569],[-36.425,-42.428,37.298],[-41.723,-34.899,23.262],
[-25.991,-33.949,37.679],[-23.867,-44.946,22.818],[-26.413,-47.010,28.120],
[-26.394,-41.814,35.416],[-22.744,-43.782,43.321],[-37.179,-21.953,19.892],
[-29.907,-26.247,23.912],[-26.689,-28.339,26.346],[-37.051,-24.725,18.518],
[-21.423,-15.780,23.895],[-32.696,-19.463,11.890],[-26.805,-29.599,13.477],
[-36.376,-28.707,19.791],[-29.358,-27.872,25.957],[-26.425,-12.355,23.080],
[-32.623,-26.587,23.008],[-30.065,-31.669,18.727],[-32.495,-32.234,20.307],
[-36.980,-23.197,11.623],[-33.293,-23.649,29.761],[-38.619,-24.542,30.374],
[-40.032,-28.577,24.209],[-41.174,-26.321,23.389],[-37.691,-15.904,18.047],
[-31.020,-15.833,18.436],[-39.354,-21.976,12.527],[-23.106,-9.845,16.628],
[-27.854,-4.891,7.471],[-30.472,-12.596,11.119],[-35.451,-4.436,5.390],
[-39.221,-9.386,4.536],[-28.066,-6.886,1.188],[-25.086,-11.987,1.857],
[-35.246,-6.676,19.367],[-33.581,-9.363,21.183],[-40.950,-10.265,1.892],
[-40.787,-2.700,8.351],[-25.767,-0.668,18.622],[-24.096,-12.062,5.005],
[-40.560,-7.945,20.516],[-21.525,-10.518,23.450],[-24.419,-9.718,21.134],
[-37.611,0.067,15.866],[-36.185,-1.734,11.681],[-40.431,-10.770,20.952],
[-37.549,8.582,3.825],[-31.740,10.385,2.085],[-26.708,3.372,8.828],
[-23.669,10.229,4.191],[-39.283,1.754,3.036],[-32.282,12.663,5.871],
[-23.341,10.876,-6.493],[-31.556,7.400,-8.170],[-34.786,5.942,-4.528],
[-39.794,7.210,2.805],[-42.439,6.150,4.264],[-30.325,24.863,-7.041],
[-30.565,31.460,-12.368],[-41.813,28.377,-8.291],[-37.165,29.710,-12.271],
[-26.680,36.254,-7.867],[-26.141,27.504,-6.252],[-29.193,14.912,-10.676],
[-30.526,23.615,-14.517],[-33.121,30.476,-7.933],[-39.114,31.354,-7.325],
[-32.423,23.129,-16.434],[-37.809,23.578,-16.080],[-39.705,26.666,-13.546],
[-32.695,23.770,2.622],[-42.235,29.713,-14.086],[-26.917,31.142,0.294],
[-23.599,20.197,4.948],[-25.542,16.591,-1.743],[-42.006,27.792,-16.536],
[-31.029,39.804,-18.590],[-32.904,42.016,-16.437],[-28.911,33.625,-16.366],
[-24.069,41.814,-13.438],[-21.814,47.599,-12.981],[-31.118,48.499,-14.769],
[-38.211,46.275,-13.243],[-37.042,36.086,-19.303],[-32.634,37.091,-11.515],
[-28.882,33.760,-20.106],[-37.597,40.970,-16.057],[-30.386,45.882,-13.889],
[-33.724,32.520,-17.272],[-35.908,33.541,-18.679],[-37.133,39.887,-25.169],
[-28.928,42.483,-7.442],[-37.898,38.842,-6.223],[-36.034,35.214,-5.982],
[-39.540,34.012,-12.150],[-26.271,44.505,-22.621],[-29.861,50.746,-15.919],
[-34.500,47.868,-17.591],[-38.139,31.175,-19.683],[-24.425,48.976,-13.358],
[-27.011,58.858,-25.844],[-25.377,-59.255,-32.501],[-26.865,53.529,-25.350],
[-34.984,52.700,-19.095],[-29.872,57.608,-18.217],[-39.659,55.598,-16.233],
[-41.029,58.574,-37.556],[-39.547,-33.452,46.659],[-31.718,-27.166,41.196],
[-37.208,-23.675,49.125],[-28.854,-28.114,37.529],[-32.210,-28.248,55.240],
[-40.292,-37.017,44.430],[-24.113,-37.162,43.280],[-40.733,-29.248,41.920],
[-32.252,-30.932,34.948],[-33.318,-32.752,57.394],[-23.210,-33.270,44.072],
[-24.432,-34.844,39.592],[-25.366,-35.489,54.203],[-39.837,-23.215,44.823],
[-24.886,-20.958,43.943],[-31.693,-38.096,48.829],[-26.045,-38.215,49.457],
[-28.655,-38.975,48.387],[-35.939,-11.094,38.408],[-28.097,-10.582,32.375],
[-42.849,-9.888,35.166],[-24.417,-7.685,40.682],[-32.652,-8.203,38.496],
[-30.515,-20.844,44.869],[-38.834,-9.570,41.327],[-37.198,-23.190,38.873],
[-22.251,-10.256,42.421],[-36.509,-15.330,40.168],[-24.494,-12.289,42.889],
[-26.831,-20.318,35.019],[-41.191,-16.857,40.442],[-30.618,-22.121,41.428],
[-41.250,-4.692,25.144],[-34.392,1.801,30.047],[-34.776,6.574,31.096],
[-30.959,4.841,23.152],[-37.837,8.119,21.604],[-32.102,-7.687,22.831],
[-36.078,-0.549,29.758],[-23.514,-5.591,22.884],[-31.411,2.013,16.580],
[-35.617,-4.014,38.188],[-34.357,-1.257,39.511],[-22.244,1.085,21.527],
[-25.185,2.475,32.826],[-23.032,-8.385,23.264],[-29.880,12.223,29.426],
[-39.409,9.683,24.363],[-39.125,-9.083,23.170],[-24.718,4.255,18.455],
[-28.624,-2.642,40.628],[-25.800,11.435,31.552],[-30.338,15.765,16.505],
[-23.400,12.640,22.014],[-35.432,14.316,19.003],[-29.068,10.729,18.235],
[-32.426,23.212,20.604],[-34.014,8.378,13.676],[-31.333,21.144,22.088],
[-34.166,21.032,24.865],[-29.720,17.277,11.044],[-27.188,15.811,11.697],
[-32.318,18.775,7.474],[-36.416,13.632,12.445],[-27.078,6.773,14.356],
[-41.190,8.506,15.582],[-32.186,27.585,15.624],[-28.933,24.920,17.551],
[-35.372,32.937,6.398],[-26.255,34.281,17.523],[-29.932,30.446,5.478],
[-35.838,28.232,3.337],[-37.525,25.706,10.711],[-33.343,31.455,12.123],
[-29.818,28.149,6.734],[-22.696,27.827,5.005],[-28.120,34.007,8.505],
[-39.611,29.220,16.130],[-35.738,33.975,1.425],[-32.747,31.046,14.591],
[-38.855,27.303,2.790],[-39.076,38.647,14.129],[-36.871,30.253,-0.412],
[-35.181,33.685,17.825],[-22.188,28.841,2.328],[-24.351,27.231,2.785],
[-22.023,28.970,20.803],[-23.382,23.261,6.244],[-33.993,39.201,6.742],
[-21.850,45.869,3.799],[-22.623,48.030,5.009],[-36.651,51.619,-0.164],
[-36.484,44.925,2.865],[-27.841,52.694,-1.790],[-27.534,45.660,-5.195],
[-23.926,56.457,1.712],[-29.900,53.004,-10.961],[-32.428,48.155,1.833],
[-35.157,53.888,-7.844],[-31.914,49.543,5.026],[-27.932,47.409,-0.373],
[-30.320,47.863,-9.070],[-39.672,52.707,7.326],[-33.425,43.801,-7.435],
[-33.707,49.190,6.992],[-25.697,40.375,-1.151],[-39.320,45.505,3.682],
[-41.767,46.495,2.541],[-31.159,57.320,-2.499],[-32.929,42.128,1.703],
[-31.432,-17.027,53.516],[-27.638,-26.179,54.398],[-30.264,-31.145,57.861],
[-39.726,-31.041,57.861],[-36.478,-27.762,57.313],[-22.281,-26.943,58.127],
[-24.445,-20.160,54.090],[-37.577,-32.677,56.782],[-41.740,-6.883,51.299],
[-23.192,-6.068,58.726],[-37.650,-9.201,57.678],[-35.651,-2.467,55.542],
[-33.583,-14.024,58.148],[-35.347,-5.616,54.576],[-36.191,-3.020,58.433],
[-38.154,3.452,55.236],[-33.051,2.006,55.790],[-33.409,-0.789,43.898],
[-35.818,-7.337,56.720],[-34.497,-12.710,49.575],[-32.777,-8.331,54.639],
[-30.288,-12.737,48.651],[-35.797,-2.910,47.579],[-38.657,-11.608,57.715],
[-31.942,-7.291,40.964],[-33.875,-0.423,58.094],[-30.593,-6.427,43.956],
[-33.007,-3.022,56.576],[-32.054,4.543,56.089],[-39.862,-2.424,43.500],
[-37.834,-16.395,49.697],[-26.595,4.504,55.791],[-31.962,9.274,45.274],
[-38.749,7.395,46.094],[-34.555,15.133,42.255],[-36.997,7.659,39.942],
[-30.727,16.758,45.065],[-35.083,18.345,42.668],[-35.124,14.279,37.162],
[-29.685,14.534,50.069],[-35.046,5.662,50.087],[-23.245,8.595,41.769],
[-25.258,1.693,41.038],[-38.000,2.735,45.323],[-39.376,24.428,37.999],
[-37.468,31.180,36.672],[-26.081,30.843,36.642],[-35.790,35.006,37.995],
[-40.187,30.856,33.917],[-31.293,16.390,30.550],[-30.894,27.820,43.194],
[-32.577,23.619,38.895],[-41.042,29.145,39.505],[-38.057,22.800,28.465],
[-26.659,35.400,35.260],[-23.969,27.804,22.056],[-30.683,25.123,22.563],
[-30.205,23.360,41.508],[-26.336,23.091,42.216],[-31.544,28.045,23.360],
[-33.365,29.599,38.218],[-25.963,32.642,34.737],[-22.419,29.205,41.648],
[-25.253,23.602,37.143],[-24.267,23.653,44.401],[-41.364,22.132,41.952],
[-36.809,33.496,30.895],[-23.472,18.984,36.288],[-31.086,30.792,24.279],
[-30.534,48.426,23.953],[-42.448,47.600,23.897],[-37.490,47.672,17.933],
[-34.704,35.217,30.367],[-30.589,38.879,23.270],[-35.733,34.842,20.145],
[-33.320,41.381,18.789],[-36.397,44.785,21.542],[-39.432,42.254,19.102],
[-34.181,52.487,28.596],[-35.316,40.847,32.128],[-40.421,49.483,21.127],
[-23.899,51.708,23.233],[-25.902,47.850,18.878],[-36.356,52.054,23.381],
[-30.869,51.477,21.522],[-36.978,41.199,37.989],[-23.824,58.070,20.464],
[-27.341,54.936,21.887],[-34.976,51.853,18.600],[-29.112,55.103,12.641],
[-33.787,55.599,6.982],[-39.422,51.862,20.017],[-27.987,6.076,-59.301],
[-39.378,6.853,58.394],[-36.831,23.470,-58.937],[-30.640,21.947,55.169],
[-28.540,16.046,51.932],[-37.127,17.898,55.867],[-30.040,23.836,49.993],
[-36.374,25.616,53.045],[-35.573,12.618,56.487],[-36.259,27.888,57.514],
[-31.095,29.475,57.241],[-33.920,26.636,56.518],[-37.329,39.993,48.929],
[-41.078,32.342,48.463],[-31.629,29.982,44.683],[-35.092,35.488,50.096],
[-36.015,33.853,54.181],[-39.606,40.681,56.031],[-27.622,43.404,49.562],
[-32.782,40.658,57.569],[-42.578,39.752,41.971],[-25.634,32.971,52.679],
[-27.015,27.625,49.976],[-39.878,48.110,39.185],[-30.772,57.394,35.893],
[-38.137,47.319,42.376],[-25.944,46.110,46.110],[-27.875,52.301,45.816],
[-35.551,51.899,42.142],[-26.336,52.151,48.150],[-40.827,45.754,41.754],
[-36.126,44.466,38.524],[-34.330,52.666,38.517],[-31.755,44.645,41.631],
[-35.679,56.510,32.052],[-38.679,43.289,45.530],[-34.060,50.993,48.551],
[-29.202,53.126,50.353],[-25.969,58.953,41.271],[-24.237,45.078,35.112],
[-25.856,43.746,39.615],[-26.502,48.188,49.006],[-36.909,57.360,36.992],
[-24.659,43.371,42.428],[-39.239,42.399,37.397],[-31.674,55.492,30.872],
[-37.615,48.954,51.702],[-28.111,48.181,53.385],[-10.678,-50.568,-53.100],
[-11.787,-49.248,-48.191],[-20.398,-50.770,-46.232],[-9.943,-46.925,-50.696],
[-16.366,-46.299,-46.574],[-20.175,-51.373,-53.375],[-21.959,-50.712,-51.008],
[-11.348,-41.701,-50.304],[-12.057,-46.371,-48.626],[-3.992,-57.399,-47.501],
[-9.215,-53.281,-49.360],[-16.833,-55.111,-49.415],[-8.077,-55.329,-52.877],
[-19.884,-53.178,-50.746],[-14.772,-52.026,-50.883],[-3.750,-52.790,-46.402],
[-17.730,-56.657,-46.981],[-12.257,-46.956,-52.853],[-16.179,-49.325,-46.687],
[-5.320,-55.193,-50.341],[-18.319,-52.970,-48.252],[-17.859,-47.991,-45.203],
[-20.534,-45.604,-54.390],[-7.010,-46.083,-48.596],[-6.293,-56.951,-46.203],
[-14.378,-58.796,-49.776],[-10.298,-44.120,-50.090],[-9.143,-46.005,-54.115],
[-6.761,-47.279,-44.569],[-12.079,-49.740,-55.255],[-11.436,-52.019,-59.405],
[-4.328,-49.996,-45.702],[-8.133,-49.447,-53.703],[-9.165,-51.524,-44.622],
[-17.753,-50.307,-51.501],[-15.287,-46.741,-49.643],[-16.766,-49.678,-49.166],
[-9.634,-44.866,-46.496],[-14.300,-46.145,-43.464],[-9.823,-50.772,-56.692],
[-18.272,-50.484,-44.155],[-11.828,-43.847,-54.555],[-18.789,-45.492,-43.450],
[-6.748,-58.270,-55.937],[-17.540,-50.550,-54.013],[-4.819,-45.757,-59.605],
[-24.541,-49.342,-45.342],[-20.894,-45.812,-48.688],[-2.955,-43.628,-44.584],
[-11.349,-45.209,-43.368],[-11.485,-41.247,-55.263],[-18.280,-42.836,-52.896],
[-9.206,-41.290,-57.119],[-9.203,-43.661,-55.539],[-5.970,-53.410,-42.500],
[-20.568,-47.363,-44.603],[-15.605,-58.446,-47.585],[-13.356,-57.538,-46.309],
[-9.532,-54.642,-44.655],[-7.133,-49.550,-57.651],[-13.336,-55.694,-55.807],
[-15.624,-53.316,-48.164],[-15.539,-53.621,-53.167],[-11.518,-52.781,-54.748],
[-10.272,-52.492,-42.333],[-8.565,-52.634,-51.761],[-17.767,-52.635,-45.662],
[-15.610,-54.115,-57.776],[-21.812,-54.928,-50.954],[-19.847,-55.721,-56.078],
[-11.170,-51.921,-47.149],[-14.211,-55.725,-50.891],[-18.250,-54.172,-53.053],
[-16.996,-53.994,-55.398],[-9.942,-56.565,-46.990],[-16.247,-51.660,-56.283],
[-6.071,-40.339,-42.206],[-17.901,-47.820,-42.063],[-11.589,-51.635,-50.318],
[-17.557,-56.433,-51.537],[-13.850,-53.175,-46.107],[-13.817,-46.434,-55.127],
[-5.763,-49.564,-48.208],[-9.479,-53.735,-58.512],[-11.752,-48.176,-43.590],
[-8.526,-53.280,-54.773],[-19.188,-47.566,-50.443],[-10.903,-46.329,-57.088],
[-22.029,-49.708,-43.871],[-14.593,-51.648,-42.068],[-9.240,-46.668,-42.015],
[-2.563,-53.010,-56.152],[-3.878,-46.492,-43.977],[-6.171,-44.738,-56.158],
[-8.898,-49.710,-47.952],[-16.662,-45.124,-53.317],[-19.799,-48.665,-55.573],
[-15.573,-47.611,-53.166],[-12.445,-42.977,-45.514],[-14.489,-42.644,-53.887],
[-8.973,-47.588,-46.294],[-7.596,-46.946,-58.342],[-9.919,-42.978,-44.269],
[-4.244,-43.780,-57.957],[-19.505,-54.938,-45.660],[-8.608,-47.806,-56.090],
[-19.361,-45.661,-46.281],[-14.882,-45.491,59.345],[-15.804,-46.638,-57.353],
[-18.036,-46.866,-55.014],[-11.683,-54.521,-46.228],[-10.286,-56.118,-55.932],
[-13.516,-44.800,-57.225],[-12.879,-54.815,-53.235],[-12.234,-52.450,-44.075],
[-10.677,-56.629,-53.064],[-15.830,-44.094,-56.571],[-16.806,-57.853,-53.862],
[-4.492,-55.087,-44.870],[-6.925,-51.276,-49.790],[-18.482,-47.147,-58.039],
[-20.773,-53.018,-57.089],[-6.814,-43.135,-48.533],[-17.401,-58.870,-49.547],
[-20.054,-43.608,-50.823],[-21.661,-43.936,-42.049],[-15.636,-42.826,-40.859],
[-12.816,-43.051,-59.405],[-10.632,-45.214,-39.414],[-7.023,-42.909,-59.233],
[-21.251,-49.103,-41.470],[-18.691,-49.726,59.168],[-14.083,-48.727,-42.188],
[-12.960,-47.481,59.240],[-8.611,-49.003,-43.799],[-19.574,-51.433,-41.962],
[-16.994,-48.910,-56.692],[-15.419,-51.824,-44.593],[-13.386,-50.268,-59.565],
[-10.484,-48.751,-41.142],[-5.142,-48.278,-59.245],[-14.531,-54.176,-43.434],
[-15.447,-55.506,-46.243],[-12.805,-49.966,-45.598],[-8.317,-53.573,-40.068],
[-7.942,-51.058,-42.112],[-17.635,-51.544,-39.926],[-14.957,-59.399,-45.217],
[-13.686,-55.275,-40.810],[-12.158,-55.438,-43.611],[-9.923,-55.322,-41.609],
[-17.758,-55.218,-41.628],[-14.829,-59.361,-42.378],[-13.863,-58.342,-39.878],
[-13.294,-41.974,-57.136],[-21.199,-46.980,59.460],[-20.333,-41.671,-54.457],
[-18.020,-41.886,-49.659],[-4.563,-40.202,-48.755],[-5.030,-42.466,-46.464],
[-5.273,-45.788,-41.732],[-4.673,-46.701,-54.812],[-6.511,-47.899,-51.528],
[-4.816,-47.176,-46.670],[-6.683,-44.526,-46.178],[-4.518,-50.162,-57.078],
[-4.995,-52.568,-57.308],[-4.837,-51.524,-52.582],[-21.127,-49.029,-48.708],
[-5.433,-50.234,-42.763],[-2.916,-52.200,-41.926],[-7.546,-56.681,-58.064],
[-6.146,-54.599,-55.719],[-6.958,-54.637,-48.024],[-7.076,-54.113,-45.117],
[-20.355,-56.330,-42.743],[-1.604,-58.828,-42.246],[-5.883,-56.905,-53.722],
[-20.313,-55.056,-48.322],[-22.900,-52.595,-48.624],[-4.674,-56.779,-42.615],
[-3.180,-57.569,-58.235],[-22.185,-55.884,-54.854],[-6.885,-42.054,-53.019],
[-21.915,-57.152,-45.399],[-19.473,-45.030,-40.790],[-16.426,-44.707,-44.321],
[-11.812,-41.904,-42.889],[-8.409,-41.843,-42.178],[-3.788,59.453,-44.429],
[-19.050,-41.141,-46.115],[-15.992,-41.133,-45.967],[-11.657,-40.559,-47.527],
[-9.660,-42.266,-47.689],[-8.873,-39.969,-46.314],[-19.539,-43.309,-47.878],
[-16.860,-44.100,-50.258],[-13.473,-44.419,-50.284],[-12.297,-56.612,-48.603],
[-7.594,-57.505,-51.620],[-5.195,-58.555,-51.310],[-19.551,-57.998,-52.283],
[-19.238,-58.573,-55.357],[-12.695,-58.033,-51.850],[-10.079,-58.191,-50.109],
[-9.844,-58.625,-55.321],[-20.522,-43.238,-56.765],[-19.125,-57.180,-58.408],
[-17.040,-56.660,-56.445],[-11.194,-40.796,-58.975],[-4.945,-42.550,-54.702],
[-18.328,-57.858,-44.371],[-16.797,-59.393,-40.318],[-18.133,-42.496,-57.879],
[-10.874,-57.866,-43.219],[-4.582,-42.853,-41.791],[-10.048,-30.868,-53.372],
[-12.046,-32.418,-49.154],[-19.057,-32.093,-48.567],[-8.258,-28.711,-50.446],
[-18.480,-27.858,-49.593],[-21.693,-32.774,-54.113],[-21.930,-30.317,-52.381],
[-10.423,-24.608,-50.124],[-13.024,-28.463,-47.209],[-4.675,-37.101,-46.802],
[-9.374,-34.631,-48.130],[-17.059,-36.508,-51.032],[-9.810,-37.453,-53.009],
[-20.546,-33.362,-51.649],[-14.886,-33.329,-51.251],[-3.006,-33.674,-45.095],
[-18.255,-36.830,-48.168],[-11.285,-27.891,-52.945],[-16.785,-29.715,-47.617],
[-7.116,-36.874,-51.042],[-18.680,-34.479,-50.027],[-18.498,-27.836,-46.859],
[-21.512,-26.709,-55.861],[-8.790,-26.334,-48.086],[-7.315,-37.464,-46.529],
[-13.767,-40.204,-50.033],[-8.701,-26.006,-51.916],[-7.914,-27.592,-54.256],
[-6.088,-29.695,-45.820],[-12.346,-30.511,-55.717],[-11.603,-32.140,-58.742],
[-4.630,-31.726,-47.543],[-7.435,-30.678,-53.252],[-10.389,-32.986,-44.679],
[-16.740,-31.676,-52.988],[-15.978,-27.438,-48.804],[-15.564,-31.029,-49.713],
[-11.987,-25.525,-48.098],[-15.989,-27.958,-45.457],[-9.953,-31.646,-55.968],
[-17.747,-33.494,-44.344],[-12.264,-25.926,-51.602],[-18.577,-26.405,-43.837],
[-6.661,-41.784,-56.692],[-19.538,-31.018,-54.049],[-3.922,-26.078,59.248],
[-20.230,-29.980,-47.249],[-21.758,-26.925,-49.859],[-3.734,-26.355,-43.230],
[-13.383,-27.188,-44.844],[-12.676,-21.149,-54.469],[-19.364,-23.612,-53.851],
[-8.696,-24.743,-58.944],[-9.439,-24.631,-55.523],[-6.790,-33.942,-41.773],
[-20.582,-28.508,-43.732],[-16.569,-39.697,-48.333],[-12.989,-40.415,-44.776],
[-9.217,-36.029,-45.088],[-7.392,-32.240,-57.480],[-12.411,-37.146,-56.438],
[-15.827,-34.796,-49.022],[-17.144,-33.136,-55.008],[-10.818,-34.048,-56.766],
[-10.304,-35.083,-42.480],[-10.323,-33.325,-51.103],[-18.959,-36.087,-45.587],
[-16.182,-38.406,58.882],[-22.624,-37.102,-52.690],[-19.939,-36.568,-56.749],
[-11.675,-33.725,-46.851],[-13.056,-37.275,-51.435],[-18.639,-35.918,-54.267],
[-15.960,-35.109,-53.603],[-9.549,-37.423,-48.270],[-16.453,-35.798,-58.938],
[-6.124,-22.317,-42.554],[-18.385,-28.930,-41.936],[-12.636,-35.190,-49.847],
[-17.510,-38.751,-52.775],[-13.572,-36.133,-46.612],[-16.206,-30.546,-55.738],
[-6.408,-30.357,-49.671],[-9.659,-36.624,-57.605],[-11.092,-28.598,-44.121],
[-8.463,-34.133,-55.803],[-18.846,-31.100,-51.392],[-10.582,-27.610,-55.511],
[-21.650,-30.878,-44.466],[-14.066,-32.265,-43.024],[-10.391,-25.216,-43.796],
[-4.258,-34.729,-55.502],[-4.890,-28.994,-43.316],[-6.387,-25.436,-54.705],
[-9.322,-31.355,-48.373],[-17.503,-27.031,-52.652],[-19.379,-32.091,-56.344],
[-14.988,-29.220,-52.976],[-15.057,-24.810,-44.345],[-16.752,-24.515,-53.538],
[-8.607,-28.751,-46.540],[-7.124,-27.579,-58.934],[-7.402,-24.964,-43.140],
[-4.105,-25.574,-56.521],[-19.726,-37.955,-43.484],[-8.179,-29.342,-56.331],
[-20.885,-27.230,-45.994],[-13.248,-27.762,-55.263],[-15.695,-27.943,-56.790],
[-18.623,-28.458,-55.071],[-11.529,-38.222,-46.176],[-10.391,-38.769,-55.854],
[-12.157,-24.475,-54.650],[-13.059,-35.399,-53.457],[-12.885,-34.402,-44.206],
[-11.552,-39.448,-51.653],[-14.660,-25.506,-56.408],[-16.334,-39.412,-55.279],
[-6.629,-38.140,-43.892],[-7.276,-32.903,-49.116],[-16.302,-26.185,57.666],
[-18.812,-34.315,-57.884],[-6.940,-24.300,-47.864],[-19.147,-39.700,-48.310],
[-19.981,-26.123,-53.772],[-20.015,-24.194,-42.631],[-16.501,-25.300,-41.835],
[-13.196,-26.054,-59.809],[-9.198,-26.703,-41.651],[-6.283,-24.625,58.395],
[-20.965,-30.432,-41.935],[-18.540,-29.960,-58.895],[-14.990,-29.553,-43.391],
[-9.796,-27.212,-58.114],[-8.837,-30.640,-43.947],[-19.641,-32.886,-41.903],
[-14.576,-32.843,-56.502],[-15.499,-33.422,-46.256],[-14.747,-32.305,-59.356],
[-12.317,-30.030,-41.804],[-5.774,-29.457,-57.365],[-15.517,-35.448,-43.851],
[-16.202,-37.891,-45.429],[-13.340,-31.455,-45.957],[-8.758,-34.570,-40.235],
[-10.043,-32.379,-41.359],[-17.569,-33.744,-40.466],[-14.166,-39.068,-46.989],
[-12.867,-35.852,-41.953],[-11.744,-37.053,-43.998],[-8.631,-38.903,-41.870],
[-17.945,-35.512,-42.424],[-14.369,-42.646,-43.443],[-13.034,-39.223,-42.020],
[-13.493,-23.310,-57.329],[-22.295,-25.962,-58.233],[-22.218,-24.112,-54.447],
[-21.674,-23.747,-51.812],[-4.402,-24.401,-49.121],[-5.120,-23.948,-45.653],
[-6.911,-28.273,-41.514],[-5.684,-28.892,-54.704],[-4.999,-31.674,-54.180],
[-6.296,-27.877,-47.814],[-5.827,-26.623,-45.130],[-3.132,-32.439,-56.896],
[-5.372,-35.544,58.814],[-7.104,-33.330,-51.778],[-21.297,-30.712,-49.705],
[-6.442,-31.559,-42.741],[-3.607,-32.904,-40.788],[-6.841,-35.840,-57.524],
[-7.799,-35.442,-53.365],[-6.707,-35.372,-48.289],[-7.150,-32.960,-46.301],
[-22.252,-35.562,-41.621],[-1.913,-40.086,-40.333],[-6.995,-38.548,-53.883],
[-20.603,-36.639,-49.741],[-21.957,-33.854,-49.034],[-3.640,-38.861,-42.032],
[-4.293,-38.221,-58.226],[-20.176,-38.645,-54.345],[-7.283,-23.139,-52.975],
[-21.022,-39.513,-45.784],[-19.084,-26.281,-40.933],[-17.006,-25.625,-46.444],
[-11.410,-23.116,-42.664],[-8.967,-21.669,-43.249],[-6.302,-40.385,-45.319],
[-19.677,-21.144,-43.653],[-19.127,-22.685,-46.581],[-11.425,-22.375,-47.826],
[-10.000,-24.237,-46.626],[-8.175,-21.805,-46.070],[-19.510,-25.239,-48.231],
[-18.271,-24.109,-51.087],[-14.453,-25.443,-49.716],[-12.262,-37.987,-48.769],
[-6.902,-39.128,-48.941],[-5.232,-40.379,-51.424],[-19.966,-40.387,-51.080],
[-21.358,-38.916,-57.172],[-16.151,-40.715,-51.315],[-8.714,-40.120,-51.120],
[-8.897,-40.393,-54.328],[-20.038,-24.459,-56.214],[-18.115,-39.578,-57.830],
[-13.605,-39.643,-55.767],[-10.528,-22.905,-58.178],[-4.299,-24.157,-53.806],
[-17.920,-40.244,-43.673],[-17.921,-38.731,-40.895],[-18.452,-22.664,-57.837],
[-10.642,-40.534,-40.899],[-5.141,-24.890,-41.526],[-9.127,-13.082,-52.966],
[-12.653,-12.726,-49.152],[-18.237,-13.870,-47.366],[-5.440,-10.979,-50.006],
[-16.987,-9.265,-48.921],[-20.952,-14.180,-54.530],[-20.361,-13.282,-51.839],
[-9.511,-6.284,-49.944],[-9.656,-9.923,-49.520],[-5.488,-20.734,-47.761],
[-10.275,-16.856,-48.759],[-17.013,-17.648,-50.858],[-7.569,-19.840,-53.181],
[-19.187,-15.861,-51.304],[-15.606,-15.139,-51.598],[-3.304,-15.738,-45.148],
[-17.590,-18.535,-46.494],[-10.380,-10.987,-51.752],[-14.908,-11.240,-47.365],
[-7.473,-19.645,-50.434],[-19.022,-17.537,-49.172],[-17.667,-10.945,-46.675],
[-20.310,-8.567,-55.925],[-7.074,-8.485,-49.288],[-7.801,-19.216,-47.411],
[-13.017,-23.110,-49.876],[-7.402,-5.966,-51.247],[-8.513,-8.266,-53.441],
[-5.348,-10.134,-46.672],[-11.480,-11.679,-54.096],[-10.401,-15.027,-58.487],
[-5.257,-13.671,-46.359],[-7.582,-13.968,-54.931],[-9.672,-12.886,-46.362],
[-16.968,-12.059,-52.036],[-14.457,-8.226,-50.683],[-15.358,-13.791,-49.364],
[-10.274,-7.886,-47.818],[-14.706,-9.563,-43.649],[-9.673,-12.068,-57.915],
[-18.005,-13.016,-44.882],[-10.727,-7.196,-55.002],[-16.937,-7.447,-43.485],
[-5.168,-22.299,-56.531],[-16.631,-13.565,-54.619],[-3.522,-7.653,-58.147],
[-21.113,-10.902,-47.546],[-21.530,-7.133,-48.427],[-3.173,-8.012,-43.526],
[-12.207,-8.639,-44.881],[-11.709,-4.548,-53.480],[-17.731,-3.837,-53.176],
[-9.785,-5.917,-57.399],[-9.519,-4.625,-55.203],[-7.174,-15.695,-42.405],
[-18.957,-8.964,-44.522],[-16.270,-20.797,-48.199],[-11.319,-21.866,-45.184],
[-8.147,-17.429,-44.261],[-7.918,-13.542,-59.256],[-10.984,-20.168,-58.631],
[-16.155,-16.328,-47.301],[-15.995,-16.954,-53.589],[-10.237,-14.816,-54.927],
[-13.211,-16.055,-40.239],[-9.763,-15.647,-51.295],[-18.487,-15.937,-45.489],
[-15.007,-16.836,-58.903],[-22.199,-17.617,-52.267],[-19.675,-18.161,-56.991],
[-10.692,-15.863,-45.773],[-14.344,-18.003,-51.552],[-19.446,-17.275,-54.244],
[-16.697,-17.814,-56.372],[-10.259,-19.637,-48.783],[-14.953,-15.809,-55.974],
[-5.146,-3.849,-44.478],[-18.583,-10.876,-42.753],[-12.928,-15.622,-50.756],
[-17.552,-19.580,-53.830],[-12.960,-16.459,-47.137],[-11.900,-12.518,-56.631],
[-7.083,-12.710,-47.983],[-10.126,-17.944,-57.070],[-12.237,-11.174,-44.574],
[-7.942,-15.623,-57.386],[-19.014,-11.774,-48.757],[-10.104,-9.728,-55.439],
[-19.978,-11.397,-45.150],[-13.837,-12.892,-43.185],[-9.427,-8.109,-44.892],
[-2.772,-15.133,-56.538],[-6.250,-10.709,-42.807],[-7.062,-7.036,-56.782],
[-8.570,-13.325,-50.190],[-14.253,-8.916,-53.280],[-19.478,-12.768,-58.260],
[-14.039,-11.016,-54.937],[-13.922,-6.774,-47.021],[-16.242,-7.172,-53.432],
[-10.168,-10.311,-46.209],[-8.829,-8.971,-58.816],[-7.634,-6.026,-44.308],
[-4.452,-3.340,-57.403],[-21.958,-18.235,-44.802],[-8.180,-11.532,-55.559],
[-18.675,-8.244,-46.869],[-13.908,-9.337,58.980],[-13.870,-10.710,-57.713],
[-18.686,-10.752,-56.477],[-13.240,-18.881,-44.952],[-10.310,-21.373,-55.908],
[-12.344,-8.393,-56.892],[-12.595,-16.557,-54.161],[-13.133,-15.708,-44.332],
[-10.424,-20.431,-52.831],[-15.426,-6.550,-56.123],[-16.309,-21.178,-55.883],
[-5.925,-18.912,-44.925],[-5.550,-14.878,-49.020],[-17.322,-8.302,-56.825],
[-19.043,-15.234,-58.885],[-6.936,-5.326,-47.806],[-19.023,-20.227,-48.712],
[-20.987,-7.393,-53.463],[-19.419,-4.759,-42.094],[-17.113,-7.055,-39.950],
[-13.870,-6.669,-58.436],[-9.644,-8.724,-41.407],[-8.677,-6.319,59.212],
[-21.702,-12.154,-42.278],[-15.553,-12.437,59.250],[-16.317,-12.374,-42.869],
[-11.072,-10.518,59.571],[-8.825,-10.471,-43.481],[-19.708,-14.248,-42.094],
[-14.424,-13.360,-56.582],[-16.001,-15.072,-44.964],[-12.268,-15.426,-56.533],
[-10.848,-11.833,-41.871],[-4.584,-10.747,-57.961],[-15.496,-17.537,-43.568],
[-14.049,-19.683,-47.388],[-12.466,-13.131,-46.325],[-9.214,-17.078,-40.779],
[-8.284,-13.225,-43.182],[-17.118,-15.338,-42.582],[-15.173,-21.873,-45.879],
[-12.613,-18.512,-41.743],[-10.687,-17.810,-43.571],[-8.752,-19.658,-41.779],
[-18.031,-18.957,-41.446],[-14.209,-22.161,-42.863],[-13.645,-23.400,-40.494],
[-13.971,-4.244,-57.316],[-20.885,-10.143,-58.283],[-20.432,-4.479,-53.264],
[-20.834,-5.471,-50.619],[-3.099,-4.009,-48.271],[-3.717,-6.271,-45.445],
[-5.278,-8.875,-41.302],[-4.984,-8.861,-56.028],[-4.772,-13.639,-54.193],
[-3.715,-9.324,-48.651],[-5.400,-7.470,-47.077],[-4.560,-13.096,-56.940],
[-3.210,-17.033,-58.380],[-6.323,-15.327,-52.242],[-21.671,-12.898,-49.517],
[-4.904,-13.420,-42.757],[-4.513,-15.821,-41.714],[-5.880,-18.209,-57.043],
[-6.271,-17.493,-54.375],[-5.273,-18.032,-48.619],[-6.498,-16.640,-46.704],
[-19.524,-16.775,-43.264],[-4.234,-22.635,-40.218],[-4.668,-20.349,-54.220],
[-21.612,-16.856,-49.892],[-20.174,-15.652,-47.643],[-4.589,-20.012,-41.674],
[-3.665,-19.875,-57.622],[-21.480,-20.191,-53.536],[-7.492,-4.243,-53.441],
[-21.236,-21.048,-45.865],[-18.929,-8.009,-41.838],[-16.074,-8.660,-45.991],
[-11.808,-3.942,-42.393],[-9.489,-3.917,-43.973],[-4.960,-21.284,-44.918],
[-19.327,-2.662,-45.869],[-16.539,-6.653,-47.499],[-11.849,-4.662,-47.885],
[-9.731,-5.510,-46.316],[-7.876,-3.235,-46.059],[-18.669,-6.510,-49.083],
[-19.250,-8.785,-51.563],[-14.348,-5.562,-49.920],[-12.644,-19.033,-49.741],
[-6.031,-22.536,-50.570],[-3.426,-22.054,-51.555],[-19.759,-21.397,-51.587],
[-22.310,-21.264,-55.832],[-14.506,-21.367,-52.487],[-8.736,-22.170,-49.521],
[-7.858,-22.453,-55.704],[-22.083,-6.525,-55.886],[-19.443,-20.093,-58.921],
[-13.975,-20.489,-57.172],[-11.469,-2.629,-56.818],[-4.089,-5.686,-54.247],
[-20.467,-21.908,-40.562],[-17.910,-22.259,-41.452],[-19.770,-5.485,-58.072],
[-11.276,-21.494,-40.533],[-4.936,-6.351,-41.502],[-10.806,7.346,-53.182],
[-11.069,4.592,-50.167],[-18.670,5.204,-48.075],[-6.363,8.970,-50.266],
[-16.858,8.829,-50.328],[-19.134,3.054,-53.840],[-22.598,5.210,-52.407],
[-11.040,13.174,-52.466],[-11.762,7.771,-49.159],[-2.443,0.975,-47.327],
[-8.018,1.756,-48.673],[-18.069,1.370,-51.303],[-9.126,-0.351,-52.605],
[-20.875,2.431,-51.448],[-14.702,4.072,-51.015],[-3.983,2.230,-43.764],
[-18.360,-0.308,-46.843],[-13.060,8.659,-54.144],[-16.881,7.011,-47.958],
[-6.001,-0.932,-51.634],[-19.218,2.392,-48.865],[-16.811,7.409,-45.137],
[-19.825,9.443,-54.042],[-7.498,9.236,-46.997],[-6.276,-2.720,-48.629],
[-13.392,-2.637,-49.284],[-8.258,11.424,-50.461],[-7.362,10.145,-52.938],
[-6.385,9.183,-43.642],[-11.783,4.818,-54.455],[-13.228,2.680,-57.494],
[-4.298,4.353,-46.464],[-8.987,5.312,-53.675],[-9.386,5.373,-45.427],
[-16.103,5.811,-52.520],[-15.205,9.541,-47.502],[-14.713,5.409,-48.514],
[-10.647,11.351,-47.930],[-15.975,9.270,-43.347],[-9.254,4.502,-56.588],
[-19.126,4.483,-44.460],[-13.435,11.731,-51.961],[-18.756,10.955,-43.750],
[-6.755,-2.030,-57.984],[-19.545,5.693,-54.092],[-3.166,9.638,-59.085],
[-22.053,6.884,-47.448],[-21.257,10.286,-48.511],[-2.023,11.296,-44.264],
[-11.898,11.941,-45.432],[-11.923,16.009,-55.120],[-17.898,12.694,-54.204],
[-8.317,15.082,-58.772],[-9.437,12.219,-54.889],[-7.782,3.021,-43.314],
[-18.886,8.961,-45.821],[-15.769,-1.600,-47.931],[-11.535,-3.123,-45.778],
[-6.299,0.927,-44.085],[-7.080,6.116,-55.773],[-12.669,0.003,-57.092],
[-16.628,2.814,-47.644],[-14.928,1.513,-52.371],[-10.402,2.070,-56.060],
[-13.157,2.409,-41.530],[-10.140,2.794,-53.183],[-20.415,3.872,-46.831],
[-15.933,0.330,-59.514],[-22.090,-1.065,-52.454],[-20.125,1.068,-58.349],
[-11.530,4.572,-47.067],[-15.716,-0.546,-50.432],[-19.802,0.355,-53.088],
[-15.172,0.452,-55.365],[-8.754,-0.848,-48.984],[-15.799,4.152,-57.983],
[-7.491,14.051,-41.631],[-19.013,7.782,-43.490],[-12.219,2.362,-51.120],
[-17.051,-0.110,-53.161],[-13.308,2.074,-46.785],[-13.367,6.564,-56.243],
[-7.367,6.323,-46.965],[-10.216,0.425,-58.190],[-11.625,7.294,-44.981],
[-7.428,1.031,-55.739],[-18.033,6.534,-50.485],[-11.612,8.882,-56.983],
[-21.779,5.234,-44.647],[-14.478,3.834,-43.254],[-9.404,10.585,-45.665],
[-4.388,4.204,-55.741],[-3.765,9.559,-45.894],[-4.731,10.400,-55.530],
[-9.148,6.420,-49.326],[-16.750,11.012,-52.433],[-20.825,5.084,-58.846],
[-15.079,8.997,-52.509],[-14.602,13.022,-46.451],[-13.734,13.719,-54.050],
[-10.034,8.566,-47.425],[-8.920,9.307,-57.763],[-7.383,11.979,-43.437],
[-5.069,13.332,-56.771],[-20.717,0.482,-45.700],[-8.892,8.692,-54.773],
[-19.003,11.821,-46.691],[-14.227,11.648,58.257],[-15.745,9.584,-56.121],
[-18.372,7.708,-55.513],[-10.018,-0.701,-46.179],[-8.681,-1.841,-55.276],
[-11.997,11.595,-55.758],[-11.887,0.400,-53.252],[-11.549,3.028,-44.822],
[-10.804,-2.083,-53.356],[-14.515,12.433,-56.259],[-16.399,-2.033,-55.381],
[-6.269,-1.895,-42.416],[-7.751,4.286,-49.469],[-18.719,10.342,-57.759],
[-18.473,3.423,-58.663],[-6.597,13.013,-48.108],[-18.636,-2.799,-48.812],
[-19.527,8.801,-50.083],[-21.363,13.839,-44.263],[-14.984,13.276,-40.653],
[-12.664,12.397,-58.607],[-9.849,10.231,-42.975],[-6.882,11.590,-57.863],
[-22.997,7.633,-41.310],[-15.789,7.470,-58.340],[-13.942,7.269,-43.121],
[-12.873,9.016,57.905],[-8.842,8.102,-44.574],[-20.862,2.432,-40.965],
[-16.282,5.608,-55.373],[-16.495,3.182,-44.746],[-13.357,5.422,-58.960],
[-10.084,6.313,-42.740],[-6.149,7.363,-58.391],[-16.762,-0.203,-43.483],
[-14.435,-0.046,-45.044],[-14.089,5.485,-45.493],[-8.318,2.810,-40.001],
[-7.431,5.699,-43.608],[-17.392,3.882,-42.360],[-14.576,-4.014,-46.966],
[-13.301,-0.214,-42.342],[-10.922,0.610,-43.810],[-8.512,0.318,-42.747],
[-18.476,0.211,-41.591],[-13.976,-2.924,-44.274],[-13.330,-2.519,-40.724],
[-13.973,14.691,-58.082],[-20.830,7.538,-57.366],[-21.013,14.455,-53.716],
[-20.435,12.512,-51.031],[-4.004,13.424,-48.094],[-4.453,13.746,-45.293],
[-5.458,10.060,-41.236],[-4.630,7.106,-55.686],[-6.025,5.304,-53.053],
[-5.245,7.957,-47.966],[-5.815,11.280,-46.114],[-6.332,3.562,-57.345],
[-3.951,1.647,58.339],[-5.605,3.534,-51.047],[-20.652,6.136,-49.825],
[-5.296,6.451,-44.966],[-4.885,5.000,-42.696],[-6.906,0.422,-58.244],
[-5.698,2.134,-53.665],[-5.790,0.261,-48.515],[-6.285,2.589,-46.548],
[-21.171,0.208,-42.638],[-2.790,-1.793,-39.938],[-5.776,-1.739,-54.588],
[-19.552,-0.316,-49.854],[-22.337,3.663,-49.325],[-2.323,-2.329,-42.841],
[-4.233,-1.122,-59.222],[-20.645,-1.901,-54.381],[-7.656,13.612,-52.627],
[-22.134,-3.252,-47.709],[-20.768,9.423,-42.213],[-16.590,11.014,-45.223],
[-12.116,13.290,-42.870],[-9.610,14.059,-43.435],[-5.481,-1.799,-46.275],
[-20.111,15.009,-46.487],[-17.537,14.242,-47.156],[-10.922,13.856,-48.878],
[-9.035,13.234,-46.427],[-6.783,15.487,-46.947],[-19.876,13.372,-48.814],
[-15.758,13.079,-50.461],[-13.061,11.649,-49.033],[-11.554,-0.879,-50.962],
[-6.060,-3.722,-51.212],[-3.349,-3.788,-51.831],[-19.547,-2.463,-51.530],
[-20.596,-3.443,-56.419],[-15.957,-3.414,-51.119],[-8.925,-3.607,-49.287],
[-6.709,-4.222,-55.937],[-20.115,12.684,-55.936],[-18.892,-1.235,-57.358],
[-13.089,-1.941,-54.823],[-11.364,14.007,-56.879],[-4.896,12.966,-53.859],
[-18.147,-2.271,-42.006],[-16.916,-4.190,-40.578],[-18.149,14.346,-58.341],
[-9.615,-5.187,-41.265],[-4.660,11.470,-43.596],[-13.156,24.878,-50.237],
[-11.438,23.628,-48.533],[-18.102,23.700,-48.038],[-7.602,27.455,-50.525],
[-17.746,28.106,-49.448],[-22.390,21.857,-53.373],[-20.452,24.194,-52.525],
[-8.985,31.610,-49.874],[-12.933,27.216,-47.427],[-4.507,17.926,-47.226],
[-8.308,20.426,-48.454],[-17.872,19.914,-51.201],[-8.494,19.270,-51.860],
[-21.200,21.052,-50.931],[-17.344,22.424,-50.450],[-3.859,20.652,-46.240],
[-17.974,18.999,-48.042],[-10.124,27.412,-51.472],[-16.192,25.475,-46.937],
[-5.920,18.513,-50.209],[-19.478,21.527,-47.483],[-17.409,27.792,-45.869],
[-20.137,27.857,-56.613],[-6.934,27.805,-48.115],[-7.232,18.090,-47.584],
[-13.101,15.571,-49.968],[-7.463,28.790,-52.932],[-7.795,26.051,-52.635],
[-6.541,27.196,-43.824],[-10.659,25.707,-53.822],[-9.502,22.308,-58.649],
[-4.127,23.449,-46.128],[-7.134,24.077,-54.754],[-6.768,22.734,-46.687],
[-16.579,24.686,-52.809],[-15.099,27.982,-48.949],[-16.105,25.075,-49.784],
[-8.720,29.285,-46.854],[-15.623,28.637,-44.212],[-11.733,23.774,-55.719],
[-18.444,23.838,-44.860],[-11.243,29.965,-51.605],[-19.001,30.868,-44.675],
[-6.634,16.946,-56.653],[-18.385,23.885,-54.680],[-4.498,30.296,-58.592],
[-20.027,26.353,-45.818],[-22.040,29.829,-47.601],[-2.834,30.210,-42.703],
[-12.887,30.063,-43.785],[-11.674,32.672,-52.779],[-18.534,31.081,-54.434],
[-9.178,31.158,-57.292],[-9.042,31.598,-53.638],[-5.978,21.831,-44.280],
[-19.892,28.350,-44.060],[-16.302,16.014,-49.106],[-11.303,14.892,-45.588],
[-7.526,19.698,-43.313],[-7.684,23.888,-57.187],[-11.570,18.030,-57.053],
[-15.862,20.478,-48.244],[-16.549,21.665,-53.094],[-10.692,22.543,-53.321],
[-11.968,23.329,-41.406],[-10.201,20.883,-50.243],[-16.798,21.662,-46.062],
[-13.357,19.126,-58.835],[-22.312,18.896,-52.497],[-18.936,19.357,-54.338],
[-10.602,22.965,-45.721],[-15.585,18.595,-50.034],[-19.323,21.760,-53.398],
[-15.353,20.161,-54.896],[-10.077,18.467,-47.384],[-16.773,21.789,-56.688],
[-6.495,34.918,-42.066],[-18.725,26.035,-42.905],[-13.229,20.685,-49.180],
[-15.796,17.892,-52.730],[-13.683,21.075,-46.333],[-13.180,26.250,-56.305],
[-5.884,25.087,-48.622],[-8.725,18.891,-56.982],[-12.335,27.537,-44.544],
[-8.148,21.634,-55.494],[-18.819,25.380,-50.374],[-10.650,27.954,-56.471],
[-21.141,23.486,-44.766],[-14.912,23.253,-41.559],[-9.750,28.299,-44.200],
[-4.262,21.008,-57.430],[-4.100,27.977,-43.775],[-5.229,30.720,-54.218],
[-8.451,24.835,-47.446],[-15.445,28.228,-53.196],[-20.248,25.163,-58.128],
[-13.078,26.705,-52.868],[-12.264,31.963,-46.301],[-16.190,31.927,-52.638],
[-10.965,25.757,-45.972],[-8.385,28.390,-57.986],[-7.289,31.484,-44.841],
[-4.752,32.960,-56.953],[-20.425,19.069,-44.516],[-8.437,26.378,-55.447],
[-19.503,28.961,-47.153],[-13.269,28.383,-57.965],[-15.315,28.087,-56.336],
[-17.883,26.865,-57.836],[-10.951,17.913,-44.775],[-9.509,18.162,-54.454],
[-13.114,29.227,-55.086],[-13.381,22.536,-52.279],[-12.862,22.000,-43.827],
[-11.932,17.083,-52.505],[-15.011,30.711,-57.801],[-14.997,17.347,-57.134],
[-5.250,18.817,-44.682],[-7.165,22.884,-49.795],[-18.094,29.797,-56.823],
[-19.257,22.142,-57.197],[-6.612,31.584,-47.387],[-19.223,15.826,-49.303],
[-21.671,30.820,-53.560],[-20.341,32.722,-42.290],[-15.215,32.097,-39.558],
[-12.342,30.758,-58.740],[-8.789,27.884,-41.152],[-6.597,30.656,-56.951],
[-21.504,25.601,-42.185],[-17.366,25.066,59.243],[-14.688,25.585,-42.998],
[-11.758,26.623,59.290],[-8.697,25.669,-44.213],[-19.213,23.340,-41.996],
[-16.261,24.410,-57.391],[-15.700,23.459,-44.305],[-14.485,23.395,-59.119],
[-10.707,26.186,-41.988],[-5.750,28.090,59.209],[-15.390,20.710,-43.390],
[-15.558,18.713,-45.145],[-13.587,24.294,-46.057],[-9.609,21.340,-41.212],
[-9.891,23.454,-43.159],[-16.965,21.735,-41.543],[-14.005,16.206,-45.509],
[-12.702,18.829,-43.115],[-10.663,20.754,-43.678],[-9.303,18.646,-41.366],
[-19.118,19.558,-41.368],[-14.734,14.596,-43.755],[-15.445,15.947,-40.676],
[-13.306,33.207,-59.025],[-21.760,28.806,-59.466],[-20.136,33.415,-53.848],
[-21.012,30.817,-50.677],[-5.152,32.706,-49.404],[-2.738,29.697,-45.463],
[-5.771,27.601,-41.090],[-6.120,27.893,-55.289],[-5.002,23.979,-53.038],
[-4.373,27.050,-47.219],[-5.954,29.432,-45.806],[-5.326,25.306,-59.473],
[-4.235,19.217,58.920],[-7.174,21.825,-52.796],[-22.321,24.613,-49.312],
[-5.648,24.569,-43.776],[-3.645,23.171,-42.047],[-6.241,19.259,-58.353],
[-6.496,19.497,-55.297],[-5.544,20.499,-48.247],[-8.977,20.434,-45.881],
[-22.046,19.924,-41.695],[-5.025,14.657,-40.844],[-4.869,17.353,-54.232],
[-20.762,18.061,-50.164],[-22.598,22.099,-48.706],[-4.164,17.308,-42.775],
[-2.184,17.308,-59.454],[-20.107,17.046,-55.068],[-8.623,34.670,-54.470],
[-22.978,15.361,-46.634],[-19.816,29.737,-41.292],[-16.504,29.936,-47.678],
[-11.672,32.561,-43.263],[-8.948,33.603,-43.003],[-7.187,14.210,-44.581],
[-19.745,33.315,-45.171],[-17.763,32.486,-48.115],[-10.684,33.524,-47.995],
[-9.694,32.142,-45.308],[-6.255,34.317,-47.738],[-20.011,31.357,-48.241],
[-16.827,29.718,-51.403],[-15.943,32.209,-49.986],[-10.968,17.568,-49.994],
[-5.384,15.347,-49.375],[-4.695,15.942,-51.990],[-18.621,17.266,-51.713],
[-21.261,15.145,-56.287],[-14.833,15.442,-51.824],[-7.303,16.406,-52.065],
[-7.257,13.789,-55.329],[-22.102,30.983,-56.531],[-17.418,17.947,-56.115],
[-12.930,18.894,-54.265],[-10.451,33.172,-56.017],[-6.478,32.988,-54.133],
[-20.318,16.192,-43.940],[-18.086,16.493,-41.984],[-20.116,32.362,-57.831],
[-10.329,14.616,-40.770],[-5.276,30.584,-43.395],[-10.261,42.699,-51.640],
[-12.299,42.929,-49.747],[-18.448,42.208,-48.938],[-7.600,46.735,-51.749],
[-17.324,45.613,-50.030],[-22.842,38.814,-54.010],[-22.479,42.309,-51.069],
[-9.922,50.277,-49.104],[-12.211,47.675,-49.899],[-4.174,36.544,-47.900],
[-8.495,38.752,-48.328],[-17.264,37.408,-51.463],[-7.813,38.240,-54.331],
[-20.724,40.202,-50.351],[-15.449,41.341,-51.357],[-4.662,40.443,-45.214],
[-17.775,37.276,-47.411],[-10.965,45.813,-53.808],[-16.307,43.504,-46.497],
[-5.053,37.345,-50.664],[-18.105,39.322,-48.960],[-18.058,45.862,-46.090],
[-20.077,46.103,-55.290],[-7.720,45.888,-48.418],[-7.450,36.558,-46.459],
[-13.051,32.496,-49.019],[-8.604,49.532,-51.546],[-8.557,47.900,-54.199],
[-6.513,45.330,-45.178],[-10.806,43.198,-54.546],[-11.407,39.801,-59.029],
[-3.721,43.416,-47.378],[-7.899,44.493,-53.513],[-9.313,41.031,-45.898],
[-17.869,42.830,-52.623],[-14.974,47.028,-50.251],[-15.936,43.326,-49.437],
[-10.377,48.002,-47.376],[-14.008,46.337,-46.069],[-9.923,41.805,-57.028],
[-18.371,43.534,-44.577],[-11.223,49.668,-52.426],[-17.220,50.058,-44.075],
[-6.316,34.887,-56.282],[-17.518,42.793,-55.363],[-3.501,49.930,-57.002],
[-20.497,44.857,-45.998],[-21.401,48.278,-48.294],[-2.127,49.771,-44.495],
[-13.870,48.530,-44.275],[-12.451,50.632,-55.066],[-17.826,50.231,-54.095],
[-10.221,50.746,58.913],[-9.974,50.722,-56.431],[-7.243,41.210,-44.205],
[-20.161,45.201,-43.281],[-16.047,34.493,-47.584],[-11.026,34.607,-45.337],
[-8.890,38.517,-44.882],[-6.760,42.045,-57.891],[-11.886,35.366,-57.845],
[-15.907,40.115,-47.091],[-16.573,40.504,-53.816],[-13.383,41.601,-54.215],
[-11.316,40.604,-41.903],[-9.644,39.797,-52.549],[-18.874,40.552,-45.529],
[-14.376,36.459,-58.025],[-21.698,37.233,-51.557],[-20.214,39.382,-55.831],
[-11.388,42.029,-47.114],[-13.904,37.296,-50.549],[-19.353,40.461,-53.747],
[-15.375,38.521,-55.108],[-10.191,36.554,-48.223],[-16.414,40.834,-58.414],
[-6.891,51.048,-42.531],[-17.787,44.939,-42.216],[-12.884,40.317,-49.338],
[-18.296,37.254,-54.121],[-13.721,39.069,-45.666],[-13.068,44.658,-56.158],
[-6.544,43.395,-48.444],[-8.742,36.944,-56.584],[-11.477,45.791,-43.447],
[-10.283,39.106,-55.756],[-19.769,43.474,-50.986],[-10.303,46.606,-56.413],
[-20.727,41.844,-43.957],[-14.179,41.237,-42.674],[-8.809,48.419,-44.505],
[-3.581,39.579,-56.702],[-4.204,46.833,-44.558],[-6.757,48.609,-56.347],
[-9.080,41.875,-48.380],[-15.750,47.604,-52.842],[-20.807,43.224,59.060],
[-14.903,43.904,-53.385],[-12.880,49.055,-46.523],[-15.552,49.144,-55.014],
[-10.393,45.195,-46.631],[-7.659,46.839,-58.545],[-7.446,50.767,-45.012],
[-5.525,52.214,-59.057],[-20.132,38.451,-44.517],[-8.195,44.185,-56.217],
[-18.883,48.621,-46.297],[-14.963,46.690,-58.144],[-15.360,45.862,-55.308],
[-19.754,44.323,-57.418],[-11.790,37.260,-46.002],[-10.779,36.640,-54.663],
[-12.972,47.356,-55.350],[-12.186,38.784,-53.829],[-11.412,39.667,-44.673],
[-13.393,36.422,-53.338],[-16.377,48.925,-57.762],[-16.161,36.019,-55.316],
[-5.834,37.786,-43.978],[-6.612,39.733,-50.551],[-17.778,46.606,-57.539],
[-19.162,40.399,-58.308],[-6.410,50.551,-47.419],[-17.644,35.196,-49.828],
[-20.008,48.556,-52.640],[-20.723,51.085,-40.735],[-15.879,49.009,-41.663],
[-13.433,49.267,56.944],[-10.163,47.672,-42.107],[-7.546,50.269,-58.690],
[-22.567,44.511,-41.788],[-18.343,45.082,59.020],[-15.282,45.698,-42.043],
[-12.782,46.532,58.654],[-8.578,43.665,-44.941],[-18.381,42.332,-42.050],
[-14.721,42.038,-56.491],[-16.029,40.739,-44.295],[-14.067,39.819,-58.022],
[-13.344,43.620,-41.874],[-5.967,44.652,-57.601],[-14.165,36.403,-43.315],
[-15.295,36.979,-46.208],[-13.381,42.070,-45.098],[-8.295,40.051,-40.789],
[-9.116,42.053,-42.526],[-16.537,40.252,-41.593],[-13.895,34.736,-46.007],
[-12.970,38.942,-40.006],[-12.020,37.853,-42.825],[-9.340,36.766,-41.930],
[-18.858,37.629,-42.247],[-14.424,33.829,-43.548],[-14.457,36.197,-40.318],
[-15.904,51.397,-58.064],[-20.642,46.434,-59.535],[-22.940,52.077,-52.022],
[-20.975,50.820,-50.694],[-4.442,50.478,-49.547],[-4.734,50.487,-45.273],
[-4.690,46.722,-41.856],[-5.289,45.757,-55.042],[-4.406,43.234,-53.305],
[-5.013,46.284,-47.287],[-6.427,48.114,-45.562],[-4.903,42.661,-55.934],
[-5.501,40.067,-59.468],[-4.811,39.754,-52.759],[-21.459,43.619,-48.740],
[-5.719,43.487,-43.310],[-3.106,41.108,-40.716],[-7.588,37.030,-59.031],
[-5.921,37.876,-56.285],[-5.908,38.681,-47.380],[-6.738,41.073,-47.104],
[-20.372,39.878,-42.196],[-1.966,34.027,-39.496],[-5.944,35.936,-53.654],
[-19.725,36.717,-49.402],[-22.447,41.200,-47.911],[-2.619,35.136,-42.446],
[-4.453,35.475,-58.955],[-20.768,36.285,-53.904],[-8.172,51.045,-53.918],
[-21.123,36.053,-45.429],[-18.222,47.489,-41.501],[-16.259,47.984,-45.888],
[-11.908,50.152,-43.609],[-7.763,53.173,-44.247],[-6.455,34.038,-45.139],
[-19.319,51.405,-45.387],[-17.123,50.549,-47.214],[-12.046,52.005,-47.582],
[-9.938,50.739,-46.176],[-7.736,52.991,-47.148],[-19.323,50.071,-48.850],
[-17.140,49.087,-50.843],[-13.304,50.294,-49.266],[-12.420,35.211,-49.658],
[-7.145,34.652,-51.705],[-4.225,33.313,-51.764],[-19.413,35.281,-51.675],
[-18.291,33.872,-56.335],[-14.332,33.966,-52.598],[-10.282,35.410,-51.983],
[-8.874,34.398,-58.022],[-20.831,48.660,-56.105],[-18.540,35.080,-58.613],
[-12.871,34.587,-55.414],[-12.847,51.712,-57.848],[-4.959,50.999,-54.564],
[-18.612,34.652,-42.675],[-16.357,34.303,-40.627],[-19.136,50.006,-58.600],
[-8.758,32.893,-40.387],[-4.768,49.298,-42.965],[-4.117,55.338,-48.889],
[-9.288,58.602,-48.024],[-17.215,56.583,-50.576],[-8.113,55.912,-53.802],
[-22.167,59.143,-51.481],[-3.956,57.665,-46.296],[-17.053,54.954,-46.607],
[-6.874,56.894,-50.315],[-18.393,57.372,-48.232],[-6.659,55.878,-46.401],
[-13.657,53.874,-49.339],[-6.530,54.535,-57.363],[-7.551,-59.844,-42.364],
[-16.032,52.801,-48.592],[-10.961,52.784,-44.788],[-8.811,57.757,-45.342],
[-12.653,54.130,-58.657],[-16.284,57.677,-46.479],[-16.272,-59.893,-52.280],
[-8.582,58.543,-52.320],[-19.500,-59.421,-46.378],[-15.028,55.611,-58.938],
[-20.488,55.938,-54.568],[-20.231,56.517,-57.803],[-13.365,56.390,-53.524],
[-18.754,57.832,-53.378],[-15.622,57.119,-54.963],[-8.858,55.366,-47.870],
[-15.730,-57.854,-58.777],[-10.936,56.753,-52.065],[-17.429,55.063,-52.993],
[-12.943,57.602,-48.411],[-8.536,57.961,-57.566],[-10.937,57.116,-56.707],
[-3.935,58.669,-58.014],[-20.019,56.373,-46.046],[-11.322,54.901,-46.721],
[-11.153,54.568,-54.214],[-11.808,58.674,-54.175],[-11.850,58.769,-45.996],
[-13.177,53.988,-52.366],[-16.612,53.471,-56.053],[-5.777,57.558,-44.371],
[-20.082,-59.571,59.321],[-19.716,54.193,-47.572],[-16.758,57.499,-43.668],
[-14.356,55.893,-45.814],[-9.758,59.054,-40.225],[-16.527,56.243,-41.140],
[-13.586,51.970,-45.052],[-13.396,57.435,-43.016],[-10.676,57.143,-42.333],
[-8.172,56.143,-42.610],[-20.416,56.908,-41.123],[-13.070,50.614,-40.321],
[-13.852,52.635,-41.928],[-6.436,-59.228,-58.785],[-5.927,57.003,-58.276],
[-7.240,58.159,-55.301],[-4.650,58.117,-49.299],[-6.433,59.461,-47.410],
[-21.626,57.754,-43.986],[-4.411,51.119,-41.121],[-6.378,53.669,-54.608],
[-20.339,55.801,-49.622],[-22.061,-58.613,-48.217],[-4.209,54.292,-42.330],
[-3.903,55.441,-57.226],[-22.618,54.396,-54.013],[-22.200,54.558,-46.063],
[-5.466,53.154,-45.664],[-11.361,55.524,-49.244],[-8.419,53.141,-49.935],
[-6.771,54.463,-51.908],[-20.439,52.840,-52.847],[-19.719,53.437,-55.491],
[-15.423,51.892,-53.473],[-11.118,52.501,-51.060],[-8.660,53.099,-55.982],
[-17.532,54.817,-58.794],[-13.479,53.429,-55.342],[-19.945,54.309,-43.414],
[-16.483,53.638,-42.785],[-8.924,52.050,-40.827],[-10.311,-50.168,-33.211],
[-12.336,-51.759,-32.495],[-18.669,-51.594,-28.177],[-7.971,-48.311,-32.732],
[-17.229,-46.539,-31.476],[-22.643,-52.567,-34.804],[-22.629,-50.508,-31.663],
[-10.201,-42.349,-30.958],[-13.702,-47.077,-29.705],[-4.275,-56.823,-29.040],
[-9.086,-54.784,-27.577],[-15.739,-55.054,-33.753],[-8.021,-54.909,-33.929],
[-20.886,-52.820,-32.516],[-17.392,-52.949,-32.381],[-3.766,-54.348,-27.787],
[-18.248,-56.987,-29.525],[-10.778,-46.462,-34.111],[-16.150,-49.838,-28.156],
[-7.869,-56.326,-31.932],[-19.844,-54.778,-30.249],[-15.660,-47.571,-26.514],
[-20.041,-45.968,-35.378],[-8.854,-46.139,-29.948],[-5.441,-57.237,-26.748],
[-13.458,-59.313,-31.231],[-8.472,-45.295,-32.511],[-8.144,-45.491,-35.187],
[-7.757,-46.261,-25.211],[-13.729,-50.094,-34.905],[-10.306,-51.087,-39.608],
[-5.199,-51.741,-26.458],[-7.516,-50.281,-35.542],[-9.805,-52.391,-26.413],
[-16.013,-48.895,-33.606],[-14.659,-44.939,-31.422],[-16.184,-50.971,-30.640],
[-11.094,-43.128,-28.628],[-16.089,-46.124,-24.104],[-7.955,-51.504,-38.101],
[-19.075,-48.684,-26.693],[-10.522,-43.468,-33.726],[-17.442,-43.833,-24.489],
[-6.810,-57.182,-38.110],[-17.297,-50.284,-35.736],[-5.299,-44.020,-39.271],
[-22.037,-48.890,-26.407],[-22.054,-45.672,-28.568],[-2.755,-45.056,-24.569],
[-13.524,-44.332,-24.684],[-11.954,-41.145,-37.020],[-18.260,-41.810,-33.365],
[-8.204,-41.479,-39.557],[-8.996,-43.043,-36.090],[-6.392,-51.984,-23.203],
[-18.677,-46.292,-25.512],[-16.454,-58.170,-30.972],[-12.168,-58.492,-26.456],
[-8.242,-54.831,-24.837],[-6.068,-50.198,-39.461],[-11.502,-54.421,-39.546],
[-17.327,-54.128,-29.488],[-16.677,-53.016,-35.089],[-9.825,-52.322,-34.731],
[-13.298,-53.780,-23.127],[-9.971,-53.355,-32.000],[-18.598,-54.711,-27.088],
[-15.530,-53.465,-40.096],[-22.901,-56.928,-33.385],[-20.198,-54.608,-36.445],
[-11.426,-53.156,-28.172],[-14.314,-56.667,-32.048],[-19.240,-52.838,-34.534],
[-15.001,-54.354,-37.105],[-9.591,-57.016,-29.370],[-17.816,-54.215,-38.499],
[-7.147,-40.301,-22.816],[-17.939,-48.526,-23.224],[-13.180,-52.295,-30.016],
[-18.388,-55.334,-33.466],[-14.587,-54.390,-28.945],[-12.463,-49.647,-37.180],
[-6.173,-49.491,-29.585],[-8.980,-56.510,-39.329],[-11.877,-47.873,-26.325],
[-7.775,-54.492,-37.468],[-18.744,-50.490,-32.181],[-9.484,-47.539,-36.570],
[-21.555,-51.833,-24.925],[-13.603,-50.641,-23.194],[-10.072,-46.075,-26.954],
[-3.457,-52.549,-36.919],[-4.805,-46.701,-24.597],[-6.602,-45.184,-37.285],
[-8.984,-48.996,-30.294],[-15.702,-45.724,-34.107],[-20.371,-51.125,-39.205],
[-13.650,-47.250,-35.673],[-13.378,-43.608,-27.328],[-15.799,-43.192,-34.656],
[-11.424,-50.178,-27.791],[-8.506,-47.426,-39.513],[-7.315,-42.496,-25.827],
[-4.878,-40.704,-39.776],[-19.884,-56.278,-25.195],[-6.358,-47.596,-35.974],
[-19.274,-44.846,-27.774],[-14.653,-45.767,-39.758],[-15.796,-46.594,-37.269],
[-19.229,-48.292,-37.141],[-12.769,-55.914,-27.405],[-8.808,-57.121,-36.291],
[-11.357,-45.260,-36.740],[-10.433,-52.649,-37.300],[-12.939,-53.714,-25.980],
[-12.069,-55.140,-33.302],[-13.894,-44.438,-37.625],[-15.537,-57.625,-37.782],
[-5.818,-54.478,-25.872],[-7.918,-51.949,-30.591],[-17.971,-45.489,-38.567],
[-20.814,-54.136,-39.209],[-7.833,-42.315,-28.566],[-18.875,-58.003,-32.239],
[-17.634,-45.019,-35.866],[-20.305,-41.701,-23.281],[-15.442,-42.992,-22.453],
[-12.181,-43.709,-40.846],[-9.001,-45.326,-21.348],[-7.862,-44.380,-40.293],
[-20.796,-48.964,-22.813],[-16.080,-48.290,-40.003],[-13.621,-46.918,-24.407],
[-12.031,-47.399,-38.678],[-9.493,-49.126,-25.996],[-18.625,-51.496,-23.396],
[-15.188,-51.659,-37.118],[-15.087,-52.046,-27.331],[-14.012,-50.493,-39.668],
[-11.497,-48.994,-23.743],[-4.577,-48.160,-40.832],[-15.834,-54.532,-26.095],
[-15.251,-56.752,-27.727],[-12.189,-51.304,-25.231],[-10.299,-54.936,-20.962],
[-9.775,-50.952,-23.714],[-16.898,-53.561,-23.895],[-15.828,-59.356,-27.080],
[-14.004,-56.065,-24.313],[-10.968,-55.521,-25.470],[-11.533,-56.379,-22.742],
[-19.539,-55.112,-22.476],[-13.732,58.679,-23.854],[-14.549,-57.944,-22.699],
[-15.159,-40.863,-38.437],[-20.826,-46.453,-38.591],[-20.586,-41.496,-36.226],
[-21.480,-41.612,-32.934],[-4.416,-41.735,-29.622],[-3.827,-43.094,-26.257],
[-7.019,-46.465,-22.709],[-3.776,-48.887,-35.916],[-4.709,-51.063,-34.402],
[-5.539,-46.781,-29.920],[-6.541,-45.215,-27.511],[-2.747,-49.066,-38.627],
[-5.968,-53.107,-39.166],[-5.149,-52.656,-32.030],[-20.560,-49.981,-30.213],
[-6.192,-49.381,-23.756],[-3.801,-52.638,-23.100],[-6.165,-55.481,-40.859],
[-5.937,-53.320,-35.353],[-7.047,-54.322,-30.154],[-7.364,-52.634,-27.528],
[-21.607,-54.258,-23.862],[-2.043,-58.089,-21.687],[-5.885,-58.614,-36.008],
[-20.974,-57.741,-30.121],[-21.790,-52.772,-29.837],[-4.626,-57.970,-23.023],
[-4.314,-55.683,-38.906],[-20.993,-58.254,-35.026],[-8.000,-42.772,-33.363],
[-21.833,-59.175,-28.002],[-20.213,-45.827,-23.194],[-16.221,-44.806,-26.844],
[-12.217,-41.916,-24.397],[-9.384,-40.975,-24.259],[-5.554,58.267,-26.161],
[-18.613,-41.877,-25.926],[-15.582,-42.361,-28.097],[-13.392,-41.050,-28.623],
[-9.877,-43.159,-26.304],[-7.663,-40.024,-27.325],[-20.499,-42.177,-29.087],
[-16.764,-43.597,-32.188],[-14.867,-41.863,-31.448],[-11.569,-57.075,-31.459],
[-6.642,-59.058,-30.348],[-4.244,-59.697,-33.423],[-17.670,-59.919,-34.327],
[-20.408,-57.538,-37.696],[-13.850,-58.769,-34.455],[-9.537,-58.411,-32.458],
[-7.382,-58.060,-33.965],[-22.018,-44.364,-37.329],[-18.047,-56.698,-39.176],
[-12.575,-56.531,-36.848],[-13.164,-41.263,-40.083],[-6.169,-41.207,-35.471],
[-20.628,-59.171,-25.249],[-17.339,-57.101,-23.380],[-16.897,-42.745,-37.408],
[-9.006,57.791,-23.151],[-5.490,-44.129,-24.012],[-9.724,-34.224,-33.932],
[-10.861,-32.148,-30.556],[-18.196,-32.566,-30.131],[-7.669,-29.140,-32.993],
[-18.987,-28.033,-30.964],[-21.439,-33.804,-34.743],[-21.056,-30.055,-32.786],
[-10.239,-24.371,-31.255],[-11.846,-28.260,-29.848],[-2.949,-38.107,-30.141],
[-8.708,-35.721,-29.588],[-16.756,-36.006,-33.443],[-10.173,-36.456,-34.987],
[-21.245,-35.584,-32.728],[-14.240,-33.001,-32.608],[-4.607,-35.681,-27.460],
[-17.733,-37.169,-27.030],[-10.093,-29.164,-34.565],[-14.627,-31.438,-26.285],
[-7.874,-36.490,-33.151],[-19.029,-34.024,-32.253],[-16.678,-29.144,-27.104],
[-19.794,-27.894,-36.291],[-6.999,-28.079,-30.234],[-6.733,-37.399,-28.174],
[-14.311,-39.326,-30.366],[-7.948,-25.824,-31.943],[-6.839,-26.796,-34.189],
[-7.090,-28.419,-25.292],[-11.417,-30.674,-36.820],[-11.216,-34.003,-39.642],
[-5.238,-31.996,-29.258],[-6.519,-30.397,-35.046],[-8.913,-33.756,-26.167],
[-15.819,-30.752,-33.840],[-14.121,-27.329,-31.203],[-14.792,-32.426,-29.342],
[-8.707,-25.646,-29.386],[-15.102,-27.752,-25.482],[-8.989,-32.367,-37.523],
[-17.431,-31.826,-26.685],[-10.564,-24.962,-33.936],[-16.154,-24.942,-24.626],
[-4.604,-39.454,-37.166],[-19.002,-32.340,-35.078],[-4.112,-26.602,-39.173],
[-19.089,-30.111,-28.807],[-21.263,-27.039,-29.747],[-3.435,-25.667,-24.982],
[-12.608,-26.715,-25.136],[-12.074,-22.271,-35.520],[-18.631,-22.810,-32.783],
[-8.522,-24.104,-38.646],[-9.515,-22.852,-36.327],[-6.905,-34.244,-22.879],
[-18.191,-27.093,-26.054],[-17.068,-39.924,-30.298],[-10.647,-40.324,-27.609],
[-8.742,-35.743,-24.256],[-7.458,-31.480,-39.741],[-11.778,-37.807,-38.611],
[-17.330,-35.769,-30.492],[-16.067,-33.665,-34.902],[-12.088,-33.845,-35.235],
[-12.389,-34.545,-23.327],[-10.816,-36.398,-32.291],[-18.863,-34.361,-26.994],
[-15.859,-36.624,-41.304],[-21.719,-38.665,-32.318],[-20.424,-35.811,-36.755],
[-11.071,-33.596,-28.350],[-15.124,-38.381,-32.950],[-19.179,-35.297,-34.520],
[-17.394,-36.709,-36.074],[-10.769,-36.985,-28.785],[-15.877,-35.638,-38.313],
[-5.361,-20.401,-22.400],[-16.462,-30.013,-24.691],[-12.511,-34.549,-30.897],
[-17.159,-39.378,-34.871],[-14.188,-35.820,-28.928],[-14.119,-29.410,-36.304],
[-7.316,-30.424,-27.861],[-8.779,-37.165,-39.310],[-13.020,-27.947,-27.512],
[-9.713,-35.310,-37.463],[-18.507,-30.502,-32.058],[-10.984,-27.928,-37.152],
[-20.922,-32.942,-25.512],[-12.778,-32.050,-24.218],[-10.026,-27.880,-24.885],
[-4.972,-34.324,-37.263],[-3.726,-28.306,-25.145],[-4.901,-27.028,-36.320],
[-8.873,-31.386,-32.175],[-14.274,-26.907,-34.950],[-20.780,-31.510,-39.499],
[-12.871,-29.375,-33.653],[-13.343,-25.127,-27.517],[-15.909,-24.537,-34.536],
[-9.435,-29.367,-30.176],[-8.489,-28.620,-39.210],[-7.551,-25.638,-24.431],
[-5.661,-24.295,-37.703],[-19.663,-36.683,-25.232],[-8.662,-29.353,-36.627],
[-18.588,-26.246,-28.925],[-14.474,-27.223,-40.906],[-15.475,-27.200,-37.677],
[-19.629,-30.289,-37.453],[-11.880,-37.476,-26.070],[-10.104,-38.189,-36.861],
[-10.900,-25.202,-36.734],[-14.239,-36.235,-34.455],[-11.830,-34.550,-25.888],
[-11.848,-38.528,-33.629],[-13.961,-24.647,-38.187],[-16.504,-38.810,-37.443],
[-6.383,-38.285,-25.631],[-8.442,-33.300,-30.433],[-18.032,-28.010,-38.625],
[-19.834,-35.235,-40.370],[-6.137,-25.246,-29.320],[-19.604,-40.214,-31.177],
[-17.406,-26.936,-34.086],[-19.175,-24.536,-24.908],[-16.664,-24.866,-21.964],
[-11.899,-25.489,-39.956],[-8.489,-27.990,-22.663],[-7.100,-26.247,-39.753],
[-20.857,-29.410,-22.462],[-15.614,-29.545,-40.413],[-13.431,-29.264,-23.904],
[-11.844,-28.229,-39.709],[-10.073,-30.895,-25.797],[-18.584,-31.979,-24.248],
[-16.906,-31.459,-37.973],[-16.312,-33.733,-25.133],[-14.909,-32.266,-40.221],
[-10.380,-32.450,-23.097],[-5.248,-30.224,-40.362],[-14.440,-36.673,-26.190],
[-15.786,-38.605,-28.446],[-12.066,-30.822,-27.949],[-9.247,-35.154,-21.693],
[-7.496,-31.612,-24.488],[-18.309,-34.532,-23.169],[-13.668,-39.918,-26.383],
[-13.014,-36.619,-21.771],[-10.960,-37.332,-23.140],[-8.449,-37.766,-22.481],
[-16.133,-36.775,-23.526],[-15.530,-41.428,-25.105],[-12.491,-39.739,-21.816],
[-14.693,-21.245,-38.738],[-21.249,-28.574,-39.965],[-20.921,-22.789,-35.474],
[-21.007,-24.421,-32.206],[-4.233,-23.411,-28.955],[-4.248,-23.353,-26.089],
[-5.641,-27.875,-22.592],[-3.499,-29.595,-36.297],[-4.818,-32.561,-35.245],
[-4.766,-29.069,-28.753],[-4.078,-25.765,-27.599],[-3.791,-32.117,-37.905],
[-5.137,-35.086,-40.094],[-7.029,-33.793,-33.554],[-21.435,-31.259,-30.099],
[-5.099,-30.834,-24.438],[-3.984,-32.438,-22.676],[-6.205,-37.673,-41.110],
[-7.032,-35.527,-35.784],[-6.240,-34.916,-30.439],[-7.117,-33.940,-28.150],
[-22.148,-35.504,-23.161],[-4.027,-39.762,-22.535],[-6.006,-38.286,-34.555],
[-19.025,-37.006,-32.236],[-20.317,-34.495,-29.527],[-3.397,-37.476,-24.206],
[-3.988,-37.887,-39.322],[-20.564,-38.817,-36.680],[-6.181,-24.165,-33.168],
[-20.737,-38.247,-27.346],[-20.161,-28.816,-25.407],[-16.047,-26.327,-27.603],
[-11.580,-24.063,-25.776],[-9.191,-21.953,-24.615],[-4.663,-40.416,-25.668],
[-18.762,-21.811,-28.501],[-15.484,-23.719,-28.489],[-11.127,-21.836,-29.787],
[-9.576,-23.935,-27.543],[-6.956,-22.252,-28.144],[-19.376,-24.049,-29.748],
[-17.768,-25.396,-31.839],[-15.398,-23.723,-32.001],[-12.008,-39.843,-31.437],
[-7.098,-40.026,-29.983],[-4.871,-39.949,-32.388],[-19.356,-39.452,-33.589],
[-20.758,-42.205,-38.862],[-14.187,-40.288,-34.460],[-7.585,-40.208,-32.710],
[-8.707,-39.698,-35.179],[-21.005,-25.149,-37.002],[-19.860,-37.676,-39.029],
[-14.156,-37.774,-37.231],[-11.586,-23.115,-38.364],[-3.221,-25.095,-36.274],
[-18.377,-38.726,-23.911],[-17.958,-42.308,-22.176],[-18.271,-23.856,-39.299],
[-9.593,-39.933,-21.623],[-4.619,-24.869,-22.953],[-9.977,-12.555,-32.501],
[-10.968,-11.728,-30.074],[-18.762,-12.313,-28.947],[-6.904,-11.395,-32.540],
[-17.076,-9.776,-31.525],[-20.368,-15.510,-35.776],[-21.723,-10.417,-32.604],
[-9.734,-5.894,-33.378],[-11.549,-9.073,-30.195],[-4.254,-19.340,-28.938],
[-9.893,-16.359,-29.713],[-15.857,-17.269,-33.226],[-8.617,-17.751,-34.351],
[-20.434,-15.455,-32.323],[-13.547,-15.671,-33.349],[-2.006,-15.009,-27.328],
[-17.840,-19.197,-28.818],[-11.527,-9.211,-34.942],[-15.489,-11.776,-28.509],
[-6.616,-18.506,-32.451],[-17.882,-16.582,-31.690],[-17.595,-10.450,-27.169],
[-19.264,-11.394,-35.485],[-7.835,-9.937,-28.136],[-6.687,-19.516,-27.609],
[-13.729,-20.861,-30.879],[-9.971,-9.302,-32.579],[-8.429,-8.074,-35.160],
[-6.353,-10.554,-25.852],[-11.632,-11.396,-37.044],[-10.252,-14.591,-40.725],
[-3.816,-13.179,-28.432],[-7.217,-12.552,-35.207],[-6.809,-13.372,-27.203],
[-16.342,-14.069,-33.079],[-14.388,-8.691,-31.281],[-16.012,-12.342,-31.103],
[-11.339,-7.064,-28.347],[-15.907,-11.274,-24.628],[-9.339,-13.633,-36.792],
[-17.626,-13.529,-26.702],[-11.314,-6.536,-35.398],[-15.759,-6.591,-23.240],
[-4.502,-21.879,-37.188],[-16.638,-12.522,-35.715],[-3.355,-7.829,-38.898],
[-20.651,-10.293,-27.461],[-23.539,-8.064,-31.296],[-2.872,-7.295,-24.142],
[-14.751,-9.196,-26.359],[-12.049,-3.999,-34.991],[-17.835,-5.406,-32.345],
[-9.806,-6.212,-38.673],[-10.089,-4.084,-36.949],[-6.928,-16.041,-23.066],
[-19.453,-9.220,-25.288],[-16.349,-21.318,-29.831],[-10.627,-21.436,-27.128],
[-8.642,-17.436,-26.637],[-7.284,-12.755,-40.443],[-12.024,-17.699,-37.642],
[-17.071,-16.703,-29.068],[-15.299,-15.120,-35.509],[-10.623,-15.776,-35.434],
[-9.648,-16.006,-21.638],[-9.056,-15.238,-33.312],[-19.041,-17.029,-27.080],
[-17.447,-18.230,-38.529],[-21.577,-19.989,-32.504],[-21.788,-18.828,-37.728],
[-12.040,-15.337,-27.833],[-15.157,-19.687,-32.566],[-18.294,-16.727,-34.355],
[-16.276,-17.440,-36.200],[-9.629,-19.194,-28.883],[-15.905,-15.325,-38.525],
[-7.010,-3.943,-23.887],[-18.358,-11.176,-23.839],[-12.181,-14.975,-30.551],
[-17.290,-19.273,-34.348],[-14.830,-15.302,-28.743],[-14.523,-10.613,-36.439],
[-6.450,-12.371,-29.940],[-9.786,-19.158,-39.108],[-12.024,-10.345,-25.274],
[-7.800,-16.782,-37.741],[-19.619,-10.633,-30.975],[-10.497,-8.736,-37.659],
[-20.555,-14.588,-26.261],[-14.483,-14.906,-23.794],[-9.975,-7.138,-25.954],
[-3.589,-15.363,-37.718],[-4.995,-9.053,-24.177],[-5.658,-7.319,-35.014],
[-8.885,-13.389,-29.482],[-15.605,-7.832,-34.371],[-21.202,-12.981,-37.119],
[-14.857,-11.009,-33.769],[-14.942,-5.441,-28.666],[-16.057,-5.417,-35.259],
[-11.294,-11.968,-27.320],[-7.802,-10.201,-39.237],[-7.777,-7.523,-24.323],
[-4.865,-6.057,-37.298],[-20.230,-19.167,-25.262],[-9.027,-10.886,-36.488],
[-20.769,-7.596,-27.662],[-13.126,-8.878,-41.360],[-15.311,-9.444,-39.997],
[-17.648,-9.927,-37.855],[-12.066,-18.206,-28.750],[-9.776,-18.799,-36.505],
[-13.642,-6.030,-36.591],[-12.425,-17.667,-34.920],[-13.568,-16.546,-25.714],
[-11.466,-19.921,-33.963],[-14.358,-6.450,-39.379],[-16.702,-20.236,-36.992],
[-4.432,-18.601,-26.087],[-6.477,-14.493,-31.948],[-18.161,-7.195,-37.438],
[-18.789,-16.584,-40.061],[-7.518,-5.891,-29.226],[-18.830,-20.506,-31.105],
[-19.062,-6.700,-34.679],[-18.594,-4.282,-23.440],[-13.852,-5.825,-21.489],
[-10.964,-6.430,-43.353],[-10.530,-8.371,-23.338],[-7.893,-7.078,-40.316],
[-20.948,-11.633,-22.701],[-17.148,-11.143,-40.457],[-13.589,-12.146,-23.873],
[-11.269,-10.101,-39.913],[-8.932,-12.156,-26.014],[-18.885,-13.749,-24.015],
[-14.213,-13.291,-37.505],[-16.232,-16.086,-26.358],[-14.988,-13.487,-40.307],
[-11.071,-12.869,-24.035],[-4.043,-11.061,-40.200],[-14.820,-18.863,-24.719],
[-15.139,-18.503,-27.531],[-13.859,-13.521,-26.185],[-7.647,-16.415,-20.059],
[-8.835,-12.853,-22.533],[-16.466,-16.167,-22.583],[-13.611,-21.327,-27.656],
[-13.047,-17.349,-22.175],[-10.889,-19.187,-24.983],[-8.932,-18.135,-23.282],
[-18.505,-17.335,-24.347],[-13.867,-22.220,-25.170],[-13.271,-20.903,-22.783],
[-14.080,-3.799,-38.115],[-19.792,-8.733,-39.205],[-20.925,-4.416,-32.524],
[-20.877,-7.462,-32.476],[-4.052,-5.525,-29.602],[-3.284,-6.204,-26.884],
[-6.465,-10.282,-21.058],[-4.654,-10.117,-37.423],[-4.403,-12.695,-34.161],
[-4.948,-10.214,-29.926],[-5.546,-8.326,-27.509],[-4.709,-13.830,-39.434],
[-2.256,-15.383,-40.425],[-5.794,-15.048,-34.500],[-21.179,-12.947,-29.933],
[-6.177,-13.851,-24.116],[-3.871,-14.368,-23.079],[-6.259,-17.840,-40.973],
[-5.462,-17.174,-36.083],[-5.338,-16.494,-30.776],[-5.554,-16.475,-27.882],
[-21.042,-16.941,-23.079],[-2.900,-21.247,-24.998],[-5.442,-20.004,-35.429],
[-21.741,-17.252,-30.797],[-22.936,-14.905,-29.670],[-6.688,-18.639,-24.467],
[-2.909,-20.540,-38.782],[-21.262,-19.467,-35.117],[-6.928,-3.265,-33.504],
[-20.745,-20.039,-27.841],[-18.892,-7.571,-22.758],[-16.906,-7.799,-27.529],
[-12.722,-6.504,-23.976],[-9.924,-5.202,-24.264],[-6.248,-21.288,-25.597],
[-19.542,-5.378,-25.973],[-17.157,-4.491,-27.429],[-12.633,-4.658,-29.374],
[-10.190,-6.095,-30.612],[-7.383,-3.323,-28.245],[-19.033,-6.110,-29.680],
[-16.578,-7.056,-30.168],[-13.526,-5.932,-32.372],[-13.074,-17.899,-31.811],
[-6.552,-22.249,-30.927],[-5.278,-21.081,-32.992],[-18.763,-21.172,-35.298],
[-18.844,-24.604,-35.489],[-13.825,-21.869,-33.647],[-9.072,-22.066,-32.347],
[-6.868,-22.694,-36.005],[-21.469,-6.611,-37.099],[-18.155,-20.860,-39.150],
[-14.104,-19.377,-36.634],[-12.070,-5.249,-40.071],[-4.938,-4.559,-35.079],
[-17.822,-20.060,-23.952],[-16.089,-21.941,-21.491],[-19.333,-5.120,-39.198],
[-11.082,-22.818,-22.773],[-5.599,-6.180,-24.314],[-9.263,5.056,-33.482],
[-13.428,6.095,-32.000],[-18.119,4.674,-30.223],[-7.587,9.102,-31.768],
[-17.250,9.237,-30.045],[-21.447,2.836,-34.886],[-19.079,4.995,-34.214],
[-11.826,12.463,-31.973],[-12.328,8.408,-30.409],[-2.180,1.008,-30.752],
[-8.133,1.840,-29.484],[-17.045,0.432,-31.903],[-8.546,1.805,-33.242],
[-19.387,2.665,-32.167],[-14.196,3.587,-33.597],[-3.739,1.495,-27.818],
[-21.332,-1.637,-29.705],[-9.607,7.626,-34.067],[-14.712,8.365,-29.231],
[-5.750,0.115,-32.743],[-19.400,1.744,-29.810],[-18.295,8.254,-27.112],
[-22.471,10.054,-37.709],[-6.627,8.173,-29.304],[-6.917,-0.699,-29.164],
[-13.760,-2.592,-30.679],[-9.018,12.015,-32.636],[-9.566,11.087,-36.028],
[-5.406,8.546,-26.854],[-10.965,5.675,-35.576],[-10.269,5.173,-40.259],
[-3.199,5.069,-28.198],[-6.519,7.731,-34.084],[-7.784,3.285,-25.911],
[-16.452,6.561,-33.789],[-15.325,10.840,-31.588],[-15.860,6.272,-30.911],
[-11.261,10.679,-29.990],[-15.153,9.590,-26.703],[-8.832,4.829,-38.012],
[-17.195,5.295,-26.258],[-13.558,12.551,-34.134],[-16.480,12.564,-24.313],
[-4.955,-2.553,-38.755],[-18.533,4.805,-37.238],[-4.896,10.523,-38.599],
[-19.507,7.906,-29.554],[-21.589,10.769,-29.591],[-2.627,12.573,-26.044],
[-12.947,9.422,-25.103],[-12.301,15.180,-36.313],[-19.526,13.577,-34.222],
[-9.266,13.981,-38.562],[-9.412,14.015,-35.488],[-5.300,3.557,-24.389],
[-20.115,9.251,-25.436],[-15.287,-2.396,-28.604],[-10.016,-4.385,-27.189],
[-8.806,0.535,-25.830],[-7.354,5.233,-40.815],[-11.788,0.841,-39.625],
[-16.439,1.068,-28.905],[-16.608,4.177,-34.987],[-11.693,3.206,-36.065],
[-11.076,4.375,-24.054],[-9.745,3.341,-31.354],[-18.380,2.539,-27.294],
[-15.437,-0.890,-40.872],[-21.113,-1.671,-32.529],[-20.888,0.095,-35.797],
[-11.134,3.334,-28.563],[-14.240,-0.290,-32.500],[-18.701,1.952,-34.511],
[-15.247,0.937,-38.962],[-9.790,-0.322,-30.919],[-15.893,3.743,-38.106],
[-5.938,14.471,-23.779],[-18.898,7.042,-24.655],[-12.637,3.480,-30.956],
[-18.323,-0.591,-34.014],[-12.037,0.847,-29.794],[-13.335,7.676,-37.001],
[-5.305,5.567,-29.942],[-9.116,-0.420,-39.468],[-14.472,7.286,-26.599],
[-9.620,1.615,-37.592],[-18.710,7.127,-32.095],[-11.153,9.634,-37.693],
[-20.470,4.498,-26.244],[-14.580,5.725,-24.030],[-10.473,8.643,-26.809],
[-3.031,3.778,-38.101],[-3.911,9.564,-24.902],[-5.466,10.684,-35.978],
[-7.957,6.243,-30.972],[-16.228,9.385,-34.182],[-19.536,5.624,-40.265],
[-13.506,8.294,-33.952],[-14.604,12.211,-27.545],[-16.297,12.286,-34.012],
[-9.757,8.174,-30.469],[-8.345,9.499,-38.565],[-8.480,12.267,-25.359],
[-4.900,13.186,-37.589],[-18.403,-0.527,-26.200],[-7.425,8.958,-36.176],
[-19.776,11.399,-27.353],[-15.765,10.735,-40.967],[-14.185,10.598,-38.758],
[-18.150,7.446,-38.139],[-9.481,-0.691,-28.172],[-8.764,-1.092,-36.692],
[-11.962,12.526,-36.668],[-14.521,1.917,-36.324],[-12.357,2.274,-26.363],
[-11.840,-1.259,-35.009],[-15.596,11.582,-36.740],[-15.950,-2.157,-38.028],
[-6.339,0.272,-24.739],[-6.644,3.593,-31.334],[-17.559,10.282,-38.296],
[-18.430,2.551,-39.653],[-5.973,12.672,-29.027],[-19.277,-3.343,-29.268],
[-20.674,10.629,-34.820],[-19.974,13.993,-23.339],[-14.957,12.338,-21.942],
[-12.249,12.420,-39.383],[-10.042,8.799,-23.034],[-8.275,11.426,-40.778],
[-20.390,7.911,-22.771],[-17.407,8.593,-41.020],[-15.809,8.219,-23.146],
[-10.707,9.200,-40.415],[-7.913,7.722,-26.768],[-18.816,4.608,-23.813],
[-15.141,6.664,-38.995],[-15.680,2.397,-26.564],[-12.503,5.206,-41.760],
[-11.844,6.872,-24.544],[-5.402,7.298,-40.665],[-15.566,0.131,-24.466],
[-14.424,-0.105,-27.147],[-14.441,4.629,-26.650],[-8.361,2.270,-23.632],
[-9.093,6.531,-24.730],[-14.925,2.933,-23.870],[-12.724,-2.372,-27.635],
[-13.719,1.032,-21.945],[-11.391,1.405,-23.940],[-9.268,-0.749,-23.609],
[-17.921,0.827,-23.210],[-15.021,-3.617,-25.806],[-14.089,-2.074,-22.957],
[-14.743,14.047,-38.239],[-20.572,9.105,-39.384],[-22.028,14.067,-33.780],
[-20.862,12.034,-32.230],[-4.025,14.451,-31.199],[-5.239,14.033,-26.370],
[-6.070,9.410,-23.177],[-7.346,6.941,-38.546],[-4.437,6.321,-35.820],
[-4.076,8.320,-29.892],[-7.231,10.568,-27.170],[-3.545,7.792,-38.597],
[-5.442,3.373,-39.566],[-6.782,4.445,-34.070],[-20.823,6.066,-31.003],
[-6.144,6.958,-24.551],[-2.462,4.785,-24.549],[-6.721,0.294,-39.389],
[-6.689,2.085,-36.961],[-4.702,0.476,-30.159],[-5.820,4.101,-27.410],
[-20.715,1.715,-25.445],[-2.876,-2.027,-21.651],[-5.756,-1.068,-35.904],
[-18.731,-0.832,-30.190],[-22.336,3.016,-30.491],[-4.757,-2.513,-24.152],
[-3.367,0.853,-39.989],[-18.159,-3.685,-34.821],[-6.385,14.217,-34.221],
[-21.642,-2.530,-26.588],[-20.138,10.922,-22.722],[-17.173,10.945,-27.568],
[-12.567,12.942,-25.178],[-9.229,15.156,-25.697],[-5.524,-3.994,-26.337],
[-18.870,16.257,-27.162],[-18.121,13.553,-28.386],[-12.178,13.154,-29.112],
[-9.758,13.202,-27.832],[-8.651,15.787,-29.180],[-21.322,13.650,-28.952],
[-18.125,12.228,-31.858],[-14.534,13.328,-31.705],[-11.445,-2.152,-32.341],
[-3.386,-2.234,-30.875],[-4.512,-2.136,-33.429],[-18.096,-2.892,-31.601],
[-20.808,-4.018,-36.850],[-14.559,-3.277,-33.327],[-8.453,-4.066,-31.591],
[-6.935,-4.057,-37.180],[-19.656,11.082,-37.132],[-18.696,-2.033,-39.078],
[-13.485,-0.845,-37.115],[-13.097,15.274,-41.133],[-4.024,12.445,-34.273],
[-20.502,-2.178,-24.133],[-16.590,-3.335,-21.808],[-17.859,15.032,-39.373],
[-9.210,-3.569,-22.150],[-5.502,11.621,-25.015],[-10.623,23.505,-32.727],
[-11.401,25.521,-30.454],[-17.890,22.911,-28.097],[-7.359,27.465,-31.724],
[-19.093,28.204,-29.352],[-19.858,23.324,-35.462],[-22.500,23.136,-32.591],
[-10.336,31.141,-30.892],[-11.652,28.571,-29.850],[-4.465,19.286,-29.574],
[-9.479,20.126,-30.898],[-16.060,20.381,-31.922],[-9.506,18.275,-34.364],
[-20.309,21.731,-32.615],[-15.635,22.771,-33.398],[-3.293,20.692,-27.516],
[-18.253,18.458,-28.417],[-11.511,28.638,-34.723],[-15.940,25.536,-29.265],
[-7.733,17.390,-32.139],[-18.327,20.408,-30.206],[-17.977,25.987,-26.797],
[-18.873,26.255,-36.363],[-7.499,28.456,-28.116],[-7.076,17.955,-28.740],
[-13.213,15.912,-31.046],[-8.470,30.264,-33.013],[-8.435,27.804,-34.872],
[-6.263,26.781,-25.691],[-12.935,26.063,-35.239],[-10.880,22.473,-38.863],
[-4.552,23.115,-28.127],[-7.275,25.288,-34.736],[-9.519,22.541,-27.039],
[-17.665,24.882,-32.538],[-14.978,29.964,-31.345],[-15.719,22.922,-30.523],
[-10.087,29.049,-27.308],[-15.005,27.596,-25.585],[-11.968,24.344,-37.249],
[-17.578,23.579,-25.209],[-11.120,31.122,-33.730],[-17.783,31.243,-25.914],
[-6.101,16.869,-37.869],[-17.297,23.820,-35.805],[-4.462,29.649,-39.983],
[-22.781,25.619,-26.446],[-21.725,29.476,-30.078],[-3.154,30.575,-24.795],
[-12.277,31.063,-27.828],[-10.701,32.081,-36.515],[-19.436,33.100,-33.476],
[-10.006,30.489,-39.294],[-8.411,32.256,-35.370],[-5.407,21.395,-25.730],
[-18.216,28.604,-26.450],[-15.636,17.591,-29.762],[-11.852,15.726,-25.291],
[-8.522,19.713,-26.951],[-8.079,23.758,-39.407],[-11.652,17.170,-40.124],
[-15.859,20.640,-29.123],[-17.419,20.931,-34.557],[-12.687,22.974,-34.424],
[-11.969,22.856,-22.846],[-11.003,21.063,-33.521],[-18.168,20.645,-27.039],
[-14.288,18.568,-40.897],[-22.357,19.087,-32.470],[-20.514,19.431,-38.740],
[-11.506,22.624,-28.960],[-14.120,18.783,-32.994],[-18.836,19.397,-32.655],
[-16.282,19.410,-36.626],[-10.381,18.082,-29.015],[-18.073,20.732,-38.601],
[-6.727,33.642,-22.033],[-18.027,26.285,-24.074],[-12.876,20.304,-29.755],
[-17.237,17.538,-34.614],[-14.250,19.566,-27.456],[-13.558,26.953,-37.726],
[-6.631,25.725,-28.494],[-8.862,18.696,-38.216],[-11.460,25.992,-25.504],
[-10.620,21.071,-36.301],[-19.847,25.969,-30.558],[-10.986,27.531,-37.277],
[-20.634,22.350,-25.725],[-14.274,21.941,-23.687],[-11.871,28.827,-25.329],
[-3.312,21.215,-37.412],[-3.608,27.321,-25.137],[-6.126,29.544,-37.551],
[-8.936,24.090,-29.652],[-15.637,29.444,-35.141],[-20.296,26.068,-38.746],
[-14.471,27.895,-33.350],[-14.908,30.894,-26.679],[-15.849,31.909,-33.921],
[-9.550,26.707,-28.551],[-8.238,28.397,-38.381],[-7.659,31.915,-25.983],
[-3.505,31.749,-38.115],[-21.052,18.963,-25.267],[-9.140,25.604,-37.489],
[-20.727,29.661,-27.143],[-14.440,29.231,-40.631],[-15.964,28.670,-38.298],
[-17.551,26.328,-38.669],[-11.300,19.692,-27.225],[-11.139,17.548,-36.585],
[-12.893,29.494,-37.050],[-13.826,19.438,-35.811],[-11.809,21.764,-25.203],
[-12.683,16.895,-34.401],[-14.799,31.988,-36.618],[-16.184,17.881,-38.785],
[-4.927,18.689,-26.890],[-8.043,23.398,-32.539],[-18.328,28.783,-39.364],
[-19.534,22.680,-39.447],[-6.686,31.093,-29.427],[-19.781,16.684,-30.043],
[-19.892,29.278,-33.784],[-20.453,32.165,-24.634],[-16.922,30.763,-22.228],
[-11.357,29.186,-41.867],[-9.470,28.600,-23.118],[-6.809,30.501,-41.176],
[-20.895,25.581,-23.455],[-15.652,25.492,-40.138],[-13.974,25.186,-23.805],
[-12.035,27.566,-39.851],[-8.714,25.045,-26.035],[-19.651,23.239,-23.296],
[-15.376,25.134,-37.239],[-15.597,22.194,-26.713],[-15.890,22.553,-38.764],
[-10.090,25.223,-22.596],[-5.500,24.745,-40.722],[-15.942,19.847,-24.051],
[-14.478,16.926,-26.722],[-12.951,23.852,-27.206],[-8.167,21.380,-21.883],
[-8.768,22.934,-24.117],[-17.337,23.271,-21.634],[-16.375,15.469,-28.337],
[-12.383,18.877,-24.205],[-10.027,17.696,-24.882],[-8.009,19.431,-23.750],
[-19.717,19.407,-23.069],[-14.696,14.643,-25.137],[-14.222,15.442,-22.233],
[-12.080,34.321,-37.227],[-21.752,28.901,-39.235],[-21.704,33.472,-36.076],
[-20.980,31.077,-32.123],[-3.913,31.811,-30.460],[-4.845,32.399,-26.382],
[-4.202,27.921,-22.270],[-6.538,25.729,-37.513],[-4.587,24.092,-35.002],
[-4.461,26.080,-30.378],[-4.594,27.931,-28.013],[-3.381,23.611,-39.173],
[-5.390,21.827,-39.150],[-6.715,21.330,-34.433],[-21.925,24.594,-29.838],
[-5.471,24.250,-25.645],[-4.201,23.165,-23.174],[-7.382,20.404,-39.884],
[-6.754,20.295,-37.177],[-6.686,19.830,-30.849],[-6.450,21.143,-28.167],
[-21.677,21.111,-23.067],[-3.083,15.762,-23.716],[-4.613,18.660,-34.579],
[-20.929,19.152,-30.315],[-22.501,22.146,-29.990],[-5.047,19.018,-23.954],
[-4.590,18.160,-40.114],[-21.120,17.202,-37.055],[-5.947,32.413,-33.405],
[-21.495,17.222,-27.384],[-20.766,28.512,-23.249],[-16.676,29.966,-28.529],
[-12.896,31.622,-24.506],[-10.565,33.771,-24.925],[-4.475,16.272,-28.383],
[-18.848,34.456,-27.019],[-16.352,32.777,-28.033],[-12.434,33.554,-29.287],
[-9.404,32.412,-28.683],[-6.417,33.632,-28.162],[-18.001,31.659,-30.134],
[-17.980,30.717,-32.826],[-13.204,31.986,-31.631],[-11.201,17.838,-32.213],
[-6.923,13.866,-31.540],[-5.405,16.462,-33.187],[-18.226,15.710,-32.764],
[-20.564,14.360,-36.681],[-14.622,15.210,-34.642],[-10.104,15.153,-32.466],
[-8.732,16.059,-37.146],[-19.972,29.418,-37.000],[-20.274,16.683,-39.463],
[-13.727,17.162,-37.823],[-12.615,31.775,-38.922],[-6.171,32.618,-36.891],
[-19.409,16.761,-23.509],[-16.296,16.787,-23.768],[-18.133,32.367,-40.367],
[-8.921,13.908,-22.665],[-5.767,30.830,-24.233],[-9.780,44.065,-32.563],
[-11.397,43.505,-30.206],[-20.208,41.747,-27.951],[-8.544,46.042,-34.775],
[-17.451,46.946,-32.583],[-21.359,40.257,-35.196],[-20.353,42.592,-32.980],
[-12.025,50.115,-32.335],[-12.428,45.900,-31.478],[-4.009,37.377,-29.586],
[-9.093,39.735,-27.646],[-17.327,37.905,-32.817],[-7.510,37.175,-34.020],
[-20.597,39.978,-32.571],[-15.421,41.397,-32.944],[-3.810,39.687,-26.220],
[-17.634,37.733,-29.245],[-11.461,46.952,-34.902],[-14.992,43.806,-28.831],
[-5.476,35.639,-32.733],[-18.032,40.530,-32.012],[-16.061,46.058,-27.832],
[-20.734,45.721,-36.382],[-7.062,46.492,-30.426],[-6.938,35.655,-29.561],
[-13.810,34.915,-31.328],[-7.771,48.877,-31.373],[-9.225,48.556,-34.264],
[-5.828,45.227,-25.393],[-11.225,44.006,-35.773],[-11.122,41.052,-38.537],
[-2.824,41.687,-27.671],[-9.029,41.881,-35.485],[-8.798,42.962,-26.949],
[-15.394,44.029,-33.232],[-15.071,46.165,-31.144],[-16.701,43.217,-30.911],
[-12.498,46.333,-28.642],[-15.502,47.079,-25.484],[-8.182,42.466,-38.642],
[-17.726,42.250,-26.203],[-11.284,50.082,-35.226],[-16.346,49.589,-23.629],
[-4.488,34.334,-38.191],[-20.393,43.003,-35.924],[-3.291,48.863,-40.233],
[-19.470,44.533,-27.977],[-20.355,47.397,-29.914],[-2.886,48.872,-25.137],
[-13.074,48.042,-26.752],[-11.832,52.436,-36.676],[-18.924,49.795,-34.401],
[-8.744,49.267,-39.971],[-8.206,51.054,-34.187],[-6.106,39.503,-24.485],
[-20.416,46.119,-25.567],[-15.764,36.772,-31.004],[-11.215,35.126,-27.396],
[-8.330,37.727,-25.395],[-6.830,42.287,-40.935],[-11.572,36.164,-40.030],
[-15.835,39.364,-30.620],[-16.187,41.038,-35.969],[-11.603,40.911,-35.743],
[-12.903,41.157,-23.507],[-8.791,39.659,-33.899],[-18.082,40.491,-29.221],
[-15.541,37.911,-41.870],[-22.571,35.927,-33.537],[-20.736,39.188,-38.390],
[-11.483,41.140,-28.508],[-14.793,37.097,-34.041],[-18.203,38.850,-35.067],
[-13.801,39.275,-35.181],[-10.396,36.012,-29.898],[-15.558,39.987,-38.462],
[-7.229,51.674,-23.084],[-17.852,45.768,-24.468],[-13.283,41.169,-30.933],
[-15.968,37.921,-36.396],[-13.621,39.468,-28.852],[-13.314,43.508,-38.096],
[-5.938,43.728,-28.040],[-10.073,38.576,-39.537],[-11.974,45.503,-25.271],
[-9.056,39.737,-37.258],[-18.042,43.880,-33.206],[-11.372,45.545,-37.915],
[-21.341,41.413,-25.774],[-15.280,42.187,-22.490],[-10.597,48.012,-25.680],
[-4.926,40.788,-37.411],[-3.040,46.744,-26.670],[-7.087,48.497,-37.052],
[-8.904,42.370,-29.763],[-14.741,47.167,-35.248],[-20.240,44.047,-38.695],
[-13.227,45.186,-34.006],[-14.620,50.133,-26.177],[-17.055,51.362,-33.005],
[-8.617,45.753,-27.957],[-9.148,46.428,-39.936],[-6.675,49.054,-26.561],
[-4.454,51.325,-38.553],[-19.938,37.071,-25.894],[-7.526,45.149,-37.675],
[-19.804,48.250,-27.408],[-13.353,47.553,-41.666],[-16.488,47.644,-37.576],
[-16.808,45.374,-35.385],[-11.366,38.159,-28.059],[-8.445,37.106,-36.986],
[-14.376,46.804,-39.111],[-11.624,37.790,-36.940],[-11.814,40.141,-25.724],
[-12.244,35.507,-34.022],[-14.377,49.353,-38.592],[-16.790,36.112,-38.219],
[-6.384,36.136,-26.728],[-9.222,40.264,-31.250],[-18.844,46.764,-38.942],
[-18.564,39.540,-39.976],[-7.134,50.132,-29.165],[-17.087,34.298,-30.925],
[-20.128,47.498,-33.845],[-19.853,50.806,-24.382],[-15.945,49.588,-20.822],
[-11.598,48.366,-39.888],[-9.808,47.024,-23.362],[-6.206,48.788,-40.546],
[-21.473,44.032,-24.433],[-16.787,44.793,-39.401],[-14.488,44.829,-24.244],
[-11.965,45.656,-40.611],[-9.446,44.727,-25.299],[-19.231,43.741,-23.059],
[-15.838,43.432,-37.342],[-15.478,40.175,-26.599],[-14.734,41.982,-40.159],
[-11.369,43.558,-22.490],[-5.817,44.959,-40.109],[-15.069,37.453,-24.752],
[-15.364,37.245,-27.315],[-12.886,42.619,-25.944],[-8.376,40.860,-25.424],
[-8.198,43.124,-23.329],[-17.160,40.011,-23.438],[-14.289,34.616,-27.169],
[-14.519,39.012,-22.374],[-12.436,37.610,-24.290],[-9.561,37.468,-22.754],
[-19.138,37.888,-23.185],[-15.173,33.034,-24.908],[-13.704,33.962,-22.668],
[-13.731,54.848,-39.909],[-22.135,46.569,-39.530],[-22.479,52.040,-33.744],
[-20.771,49.430,-31.979],[-3.983,50.649,-30.855],[-4.360,50.347,-28.007],
[-4.974,45.923,-21.285],[-4.918,47.270,-36.169],[-4.663,42.839,-35.582],
[-3.909,45.535,-28.939],[-6.133,46.564,-27.647],[-4.127,43.436,-38.500],
[-5.312,39.095,-39.881],[-6.759,41.923,-33.920],[-21.040,43.927,-30.299],
[-5.635,42.268,-25.959],[-3.443,41.889,-23.540],[-6.871,35.893,-39.096],
[-6.620,38.945,-36.145],[-6.271,39.303,-29.993],[-6.835,41.211,-28.252],
[-21.680,39.303,-24.005],[-2.542,33.621,-21.229],[-5.208,35.431,-35.747],
[-20.040,37.733,-30.993],[-20.974,41.098,-30.300],[-3.506,35.984,-22.978],
[-2.542,37.119,-40.554],[-21.556,36.025,-37.310],[-6.643,52.253,-36.005],
[-21.520,34.931,-26.246],[-18.534,47.876,-22.568],[-17.418,49.791,-27.236],
[-12.201,49.894,-24.689],[-9.984,51.461,-24.553],[-6.440,34.280,-24.683],
[-21.255,51.303,-26.545],[-18.068,52.424,-27.987],[-11.596,51.050,-29.070],
[-9.476,49.427,-27.836],[-8.033,52.287,-27.891],[-16.831,50.361,-29.934],
[-15.336,49.002,-32.924],[-13.329,48.759,-30.338],[-11.247,33.866,-31.578],
[-7.404,33.009,-31.228],[-3.356,33.528,-32.639],[-19.069,36.069,-33.275],
[-20.174,33.575,-38.757],[-14.499,34.081,-34.390],[-8.890,34.683,-32.746],
[-7.701,34.786,-35.931],[-20.174,48.695,-36.811],[-18.786,35.515,-40.046],
[-13.994,35.885,-36.552],[-10.278,50.731,-38.363],[-3.589,50.777,-35.660],
[-19.709,34.810,-24.170],[-16.614,34.486,-22.563],[-18.482,52.292,-39.430],
[-9.423,34.300,-22.119],[-4.560,49.200,-23.355],[-4.740,56.968,-31.951],
[-9.602,-59.717,-29.033],[-15.588,56.313,-33.836],[-9.477,56.737,-35.140],
[-20.737,58.929,-32.777],[-3.261,56.192,-27.535],[-16.879,55.168,-29.377],
[-7.140,55.873,-33.399],[-18.519,56.789,-31.487],[-5.829,53.429,-29.585],
[-13.325,53.042,-30.085],[-5.719,55.999,-39.103],[-6.350,57.996,-23.543],
[-16.779,52.974,-30.997],[-13.153,52.354,-27.240],[-9.460,56.593,-25.693],
[-11.802,56.737,-38.984],[-15.951,57.912,-30.273],[-15.624,58.740,-32.741],
[-11.881,57.696,-33.525],[-19.162,59.454,-28.075],[-14.250,57.689,-40.083],
[-22.387,55.078,-33.718],[-22.180,55.793,-38.426],[-14.407,55.003,-31.722],
[-18.727,56.195,-34.061],[-15.569,58.460,-36.159],[-9.536,53.626,-29.786],
[-16.766,57.517,-38.482],[-10.710,57.902,-31.082],[-17.833,53.659,-34.823],
[-12.294,58.439,-28.891],[-9.389,55.041,-40.275],[-9.573,57.572,-37.908],
[-4.407,58.041,-37.041],[-19.606,55.586,-27.385],[-10.972,53.750,-26.009],
[-9.445,54.956,-37.510],[-12.202,59.283,-36.703],[-12.314,57.680,-26.163],
[-10.659,54.445,-35.050],[-16.504,55.271,-36.703],[-5.490,54.882,-26.856],
[-19.458,-59.310,-39.353],[-19.253,52.430,-30.511],[-15.097,56.990,-25.236],
[-14.270,55.483,-29.183],[-11.694,58.448,-22.001],[-18.280,57.868,-25.884],
[-14.818,54.500,-26.235],[-12.815,55.568,-24.547],[-10.573,54.173,-23.279],
[-7.679,54.411,-22.722],[-20.126,55.125,-22.365],[-14.347,52.186,-24.003],
[-13.102,53.929,-22.577],[-4.433,58.443,-39.762],[-5.520,56.403,-41.897],
[-6.811,57.632,-35.695],[-7.322,55.709,-30.387],[-5.458,57.774,-29.194],
[-23.922,57.975,-24.274],[-4.849,51.550,-21.866],[-5.518,54.660,-35.561],
[-20.763,54.852,-30.643],[-21.876,56.891,-31.715],[-4.039,54.913,-24.752],
[-4.034,54.220,-39.404],[-20.606,55.458,-36.351],[-22.954,53.097,-29.315],
[-4.895,52.149,-26.089],[-11.742,55.005,-32.632],[-7.583,53.171,-32.546],
[-5.101,54.050,-32.333],[-19.956,52.436,-33.747],[-20.526,51.783,-36.334],
[-12.388,52.478,-33.480],[-9.710,51.661,-31.893],[-7.714,53.234,-38.862],
[-18.666,55.742,-39.216],[-13.713,54.504,-36.612],[-21.260,53.110,-24.295],
[-17.780,53.131,-25.072],[-9.132,49.898,-22.384],[-10.629,-49.036,-13.907],
[-11.296,-50.637,-11.810],[-19.597,-50.705,-8.625],[-7.797,-47.022,-14.705],
[-18.503,-45.026,-11.532],[-21.818,-53.705,-15.330],[-21.722,-50.165,-14.705],
[-9.876,-41.897,-14.545],[-13.859,-47.456,-11.880],[-5.379,-57.438,-10.680],
[-8.671,-53.513,-11.339],[-16.222,-54.158,-14.575],[-9.967,-55.253,-15.407],
[-19.292,-53.850,-14.099],[-15.416,-51.880,-13.501],[-3.532,-52.895,-10.024],
[-17.329,-56.176,-10.409],[-10.745,-47.062,-16.374],[-16.314,-49.560,-9.011],
[-6.897,-54.698,-13.895],[-18.354,-52.236,-11.611],[-18.935,-46.681,-8.732],
[-21.114,-46.227,-18.331],[-8.447,-46.159,-11.953],[-6.605,-56.067,-8.588],
[-14.704,56.820,-13.153],[-8.804,-44.339,-14.090],[-8.327,-45.451,-16.855],
[-6.624,-46.287,-7.539],[-12.783,-50.344,-16.616],[-11.201,-51.284,-21.416],
[-4.264,-49.797,-9.341],[-8.018,-50.288,-16.230],[-8.706,-50.712,-9.848],
[-16.528,-49.179,-15.278],[-15.126,-45.050,-12.306],[-16.178,-49.385,-11.868],
[-10.125,-43.679,-11.647],[-15.439,-47.116,-8.064],[-10.301,-50.185,-18.938],
[-17.819,-49.324,-6.667],[-11.536,-43.989,-15.330],[-16.305,-45.404,-4.620],
[-6.268,-57.169,-18.352],[-18.968,-49.340,-16.075],[-3.735,-43.559,-21.907],
[-21.782,-47.835,-9.825],[-21.954,-44.902,-10.803],[-2.677,-44.271,-7.421],
[-13.484,-45.377,-6.868],[-11.513,-41.321,-17.036],[-19.532,-41.586,-16.877],
[-9.233,-42.721,-20.825],[-9.517,-43.381,-18.229],[-6.583,-53.028,-5.515],
[-20.176,-44.315,-6.854],[-15.053,-57.370,-11.363],[-11.163,-59.347,-8.634],
[-8.290,-54.381,-7.323],[-8.212,-49.970,-21.167],[-13.610,-55.856,-20.752],
[-15.823,-54.316,-11.885],[-17.073,-52.341,-16.275],[-10.425,-51.553,-16.691],
[-10.424,-52.239,-6.841],[-8.659,-52.530,-14.696],[-18.016,-53.336,-9.078],
[-15.588,-55.133,-22.409],[-21.607,-58.114,-15.421],[-19.510,-54.723,-19.428],
[-11.143,-52.013,-9.459],[-14.451,-56.280,-14.071],[-19.155,-54.515,-16.766],
[-16.414,-54.973,-19.634],[-9.671,-56.263,-11.250],[-17.588,-52.548,-20.889],
[-7.261,-41.203,-4.229],[-18.625,-47.022,-4.311],[-13.047,-52.753,-12.301],
[-17.340,-56.523,-15.628],[-12.910,-54.207,-10.029],[-12.374,-47.978,-18.439],
[-6.471,-48.541,-10.377],[-8.845,-57.096,-20.409],[-10.815,-48.074,-7.054],
[-9.751,-52.774,-18.998],[-18.152,-47.306,-13.621],[-10.336,-45.879,-18.921],
[-22.154,-48.761,-7.095],[-14.376,-51.531,-5.446],[-10.167,-44.453,-7.739],
[-3.045,-53.116,-20.576],[-4.514,-47.620,-5.313],[-6.244,-43.329,-18.786],
[-8.436,-48.917,-12.188],[-16.478,-45.721,-15.243],[-19.171,-50.339,-20.849],
[-14.310,-47.373,-16.263],[-14.446,-43.568,-10.125],[-15.787,-42.739,-15.669],
[-12.588,-46.863,-9.099],[-9.691,-47.819,-20.952],[-7.915,-43.928,-5.944],
[-3.863,-41.504,-19.598],[-18.213,-56.075,-7.882],[-8.378,-47.901,-18.075],
[-19.183,-43.026,-8.959],[-14.354,-45.476,-22.047],[-15.040,-47.451,-18.988],
[-19.787,-48.541,-18.767],[-11.736,-56.356,-8.787],[-8.958,-56.402,-17.842],
[-12.590,-44.303,-18.152],[-12.521,-53.720,-15.886],[-13.661,-54.664,-6.704],
[-12.685,-56.552,-16.252],[-15.541,-43.516,-19.830],[-16.321,-57.688,-20.330],
[-4.840,-53.435,-7.660],[-7.482,-51.381,-12.470],[-17.598,-45.331,-19.247],
[-20.436,-52.824,-20.942],[-6.360,-42.094,-9.327],[-17.677,-58.532,-11.505],
[-18.608,-44.537,-16.273],[-19.620,-41.948,-3.734],[-16.791,-42.758,-4.066],
[-11.792,-44.260,-22.576],[-10.990,-45.923,-5.486],[-6.610,-44.011,-21.328],
[-20.198,-49.123,-5.264],[-17.461,-47.632,-20.631],[-15.309,-48.299,-5.583],
[-13.079,-48.239,-21.551],[-9.270,-47.186,-9.565],[-19.125,-51.456,-4.668],
[-16.006,-50.107,-20.523],[-15.368,-52.002,-8.182],[-14.828,-52.589,-21.212],
[-9.964,-48.565,-4.590],[-5.487,-49.055,-21.262],[-17.406,-52.370,-6.611],
[-15.663,-54.770,-8.332],[-12.625,-50.575,-7.948],[-9.066,-53.391,-4.949],
[-8.212,-49.535,-7.674],[-17.608,-53.973,-4.479],[-13.898,-57.835,-8.805],
[-12.933,-55.565,-4.096],[-11.300,-55.656,-6.201],[-8.391,-56.158,-4.274],
[-19.043,-56.030,-5.360],[-14.430,57.352,-5.965],[-13.875,-57.900,-4.895],
[-13.523,-40.735,-19.276],[-21.962,-46.136,-21.039],[-22.729,-41.764,-16.973],
[-22.608,-43.219,-14.547],[-3.954,-42.286,-12.171],[-3.430,-41.848,-7.338],
[-6.445,-46.598,-3.609],[-4.055,-48.039,-17.968],[-5.040,-50.256,-16.791],
[-3.994,-47.797,-11.459],[-7.288,-44.837,-9.912],[-3.819,-50.871,-19.409],
[-7.221,-51.914,-19.346],[-5.562,-51.702,-14.206],[-21.146,-49.807,-11.972],
[-6.209,-50.113,-5.354],[-3.608,-52.447,-4.691],[-7.472,-58.120,-22.819],
[-7.184,-54.343,-17.178],[-5.569,-54.851,-11.440],[-6.245,-52.851,-9.796],
[-22.406,-55.096,-5.468],[-2.619,-59.042,-3.761],[-5.302,-56.557,-15.889],
[-21.828,-56.584,-12.979],[-21.412,-53.484,-11.899],[-4.762,-58.115,-5.209],
[-6.173,-56.512,-20.830],[-19.683,-57.680,-17.497],[-7.298,-42.454,-16.481],
[-22.158,-57.301,-9.940],[-20.006,-44.373,-2.865],[-16.140,-44.118,-7.303],
[-12.272,-43.464,-5.626],[-8.781,-41.352,-6.659],[-4.789,-58.854,-8.640],
[-17.917,-39.292,-8.026],[-16.338,-41.347,-8.302],[-11.608,-41.582,-11.361],
[-8.961,-42.581,-9.334],[-8.578,-41.123,-11.340],[-19.912,-41.259,-11.026],
[-17.277,-42.938,-10.745],[-13.082,-43.302,-12.864],[-12.056,-55.411,-13.339],
[-5.353,-58.864,-13.367],[-4.901,58.349,-15.679],[-19.107,-58.937,-13.779],
[-20.228,58.255,-20.518],[-12.623,-59.140,-15.131],[-10.533,-58.643,-12.967],
[-8.171,-59.152,-17.746],[-20.843,-43.626,-18.350],[-19.030,-57.394,-20.531],
[-10.990,-58.678,-18.576],[-13.145,-42.899,-20.963],[-4.594,-42.706,-16.506],
[-20.319,-58.270,-6.478],[-17.636,-58.484,-5.865],[-17.880,-42.649,-19.036],
[-9.276,-58.850,-3.718],[-4.728,-43.975,-5.695],[-11.488,-32.229,-14.451],
[-12.555,-33.854,-12.430],[-20.258,-33.979,-11.015],[-5.874,-28.089,-12.208],
[-17.765,-28.599,-11.299],[-22.094,-35.448,-15.783],[-21.571,-31.248,-14.694],
[-8.412,-23.497,-13.714],[-12.800,-28.667,-10.802],[-4.092,-38.207,-12.457],
[-8.664,-35.591,-10.352],[-15.698,-36.126,-13.116],[-8.985,-37.105,-14.407],
[-20.728,-35.643,-13.526],[-15.407,-33.652,-12.911],[-4.743,-36.858,-9.034],
[-17.853,-38.366,-10.701],[-10.559,-29.798,-15.083],[-14.318,-30.380,-9.215],
[-6.652,-38.088,-13.533],[-18.260,-35.751,-11.886],[-16.331,-29.219,-8.141],
[-19.289,-27.666,-18.696],[-7.939,-25.241,-11.565],[-6.061,-39.079,-10.216],
[-13.841,-39.928,-12.444],[-8.799,-26.365,-13.785],[-8.034,-26.615,-16.444],
[-6.897,-28.519,-6.893],[-12.819,-31.250,-16.557],[-9.615,-30.347,-21.112],
[-4.634,-33.401,-8.272],[-8.187,-32.942,-16.871],[-9.449,-34.917,-7.904],
[-15.354,-30.963,-14.780],[-14.524,-28.251,-12.943],[-15.962,-30.513,-11.904],
[-10.034,-26.888,-10.647],[-15.817,-26.891,-6.366],[-9.117,-31.739,-18.937],
[-17.150,-31.966,-7.740],[-11.433,-25.212,-16.016],[-17.911,-24.514,-5.953],
[-5.832,-39.119,-18.121],[-16.236,-31.622,-17.452],[-4.425,-25.029,-20.073],
[-21.514,-28.959,-6.752],[-22.466,-26.902,-11.813],[-2.342,-25.370,-7.157],
[-13.126,-27.656,-8.493],[-11.447,-22.354,-17.105],[-18.402,-24.210,-14.936],
[-9.588,-25.454,-21.012],[-8.955,-23.472,-16.503],[-6.080,-33.754,-4.557],
[-18.216,-28.046,-6.075],[-16.083,-40.329,-10.972],[-11.307,-40.717,-7.729],
[-7.954,-37.597,-7.454],[-7.185,-31.680,-21.653],[-12.925,-37.611,-19.211],
[-17.103,-34.427,-9.639],[-18.038,-33.838,-17.214],[-11.674,-33.990,-17.335],
[-11.244,-32.857,-7.270],[-8.879,-34.474,-14.331],[-18.452,-36.620,-8.549],
[-16.632,-36.790,-20.906],[-21.312,-38.070,-14.507],[-19.837,-35.958,-18.768],
[-10.757,-33.245,-10.198],[-14.450,-37.945,-14.672],[-18.741,-34.326,-14.571],
[-15.295,-35.411,-17.802],[-8.980,-38.081,-11.172],[-16.133,-33.576,-19.620],
[-6.942,-22.405,-5.239],[-18.765,-30.917,-5.436],[-13.255,-34.864,-14.797],
[-17.225,-36.106,-15.712],[-13.979,-35.264,-9.967],[-11.810,-29.807,-19.204],
[-5.437,-31.252,-10.141],[-7.936,-38.273,-19.601],[-11.951,-29.977,-7.409],
[-7.872,-34.259,-19.208],[-18.527,-30.291,-13.262],[-9.273,-28.475,-18.777],
[-21.127,-31.829,-7.017],[-13.581,-34.143,-5.687],[-10.278,-26.072,-7.278],
[-5.776,-36.320,-19.018],[-3.270,-27.932,-6.309],[-4.602,-26.704,-17.970],
[-8.230,-30.674,-11.455],[-15.755,-26.367,-14.852],[-18.955,-32.786,-19.913],
[-13.463,-28.898,-15.346],[-14.468,-24.830,-7.339],[-15.499,-24.401,-16.766],
[-10.946,-30.400,-11.535],[-6.833,-29.566,-19.801],[-7.186,-25.609,-6.967],
[-3.756,-22.544,-21.480],[-20.493,-37.711,-6.856],[-7.770,-30.095,-16.095],
[-19.819,-25.333,-9.337],[-13.754,-28.380,-20.750],[-14.345,-27.780,-18.178],
[-19.828,-30.361,-17.934],[-11.017,-37.279,-9.051],[-10.455,-38.533,-18.088],
[-11.619,-27.016,-19.925],[-12.555,-36.483,-16.509],[-12.254,-35.285,-7.851],
[-11.450,-38.572,-15.374],[-13.892,-24.502,-19.034],[-17.425,-36.794,-18.321],
[-6.717,-35.139,-7.325],[-6.710,-33.394,-12.526],[-16.390,-29.172,-19.468],
[-19.992,-35.163,-21.283],[-8.042,-23.941,-9.352],[-18.447,-39.934,-13.033],
[-18.182,-26.664,-15.943],[-19.853,-23.293,-4.758],[-15.434,-25.373,-3.178],
[-10.836,-25.693,-23.206],[-9.327,-27.997,-5.073],[-7.012,-25.315,-21.845],
[-21.517,-30.562,-4.251],[-15.893,-29.306,-22.220],[-14.573,-29.691,-5.229],
[-11.165,-28.382,-22.188],[-8.805,-30.349,-7.820],[-20.093,-33.244,-4.910],
[-14.407,-31.306,-19.600],[-15.586,-34.173,-7.567],[-13.627,-32.682,-21.903],
[-9.965,-30.518,-5.258],[-4.762,-30.233,-21.695],[-15.156,-37.008,-6.835],
[-15.550,-38.791,-8.847],[-13.483,-32.911,-8.602],[-9.125,-34.986,-3.229],
[-8.678,-33.046,-5.894],[-15.805,-35.440,-4.528],[-13.272,-40.497,-9.631],
[-13.136,-35.766,-3.747],[-11.347,-36.943,-5.822],[-9.429,-37.879,-4.261],
[-18.728,-36.050,-5.132],[-14.051,-41.622,-6.024],[-14.567,-40.727,-3.410],
[-14.010,-23.720,-21.516],[-18.706,-29.541,-20.925],[-20.027,-22.681,-17.387],
[-20.845,-22.275,-14.705],[-3.503,-23.897,-10.247],[-4.673,-24.158,-7.411],
[-6.696,-28.559,-3.935],[-1.965,-30.933,-18.015],[-4.670,-31.243,-17.363],
[-3.681,-29.144,-10.709],[-5.807,-25.734,-9.801],[-3.830,-32.203,-20.011],
[-4.671,-34.549,-21.047],[-6.165,-33.137,-15.333],[-22.072,-31.914,-11.713],
[-4.653,-30.414,-6.405],[-3.521,-32.905,-3.907],[-6.247,-36.688,-23.290],
[-7.803,-36.031,-17.144],[-6.075,-34.710,-9.984],[-7.549,-32.794,-8.795],
[-22.356,-34.791,-5.497],[-3.168,-39.927,-2.532],[-4.753,-37.113,-16.441],
[-20.503,-37.467,-11.683],[-22.897,-34.740,-11.383],[-3.463,-40.115,-5.339],
[-4.010,-37.465,-21.133],[-21.878,-37.969,-17.594],[-6.044,-24.762,-14.663],
[-20.563,-39.032,-9.491],[-19.679,-26.831,-4.187],[-17.475,-26.934,-8.957],
[-11.809,-23.819,-5.613],[-9.126,-23.514,-6.680],[-6.342,-40.305,-6.954],
[-19.101,-22.185,-7.782],[-17.676,-23.217,-10.261],[-11.936,-21.673,-10.585],
[-11.615,-24.377,-9.642],[-7.789,-21.317,-8.245],[-20.206,-23.811,-12.196],
[-17.345,-25.098,-12.330],[-14.888,-24.004,-12.310],[-11.644,-38.441,-12.293],
[-6.441,-41.858,-12.794],[-4.828,-39.967,-15.120],[-20.661,-41.138,-14.078],
[-20.478,-40.116,-19.179],[-14.068,-40.564,-15.711],[-8.723,-39.695,-13.802],
[-8.733,-40.476,-18.279],[-20.274,-25.386,-17.373],[-17.600,-39.870,-20.159],
[-15.325,-38.427,-17.614],[-11.268,-23.232,-20.120],[-4.207,-23.666,-16.203],
[-18.475,-39.466,-5.486],[-17.348,-40.183,-3.069],[-19.293,-22.986,-20.682],
[-9.659,-40.132,-2.730],[-5.576,-26.239,-4.623],[-9.809,-13.454,-15.105],
[-11.162,-12.389,-11.613],[-17.835,-14.149,-9.960],[-5.497,-9.381,-12.184],
[-17.246,-9.723,-14.126],[-22.086,-15.849,-15.957],[-21.357,-12.373,-14.859],
[-8.610,-5.208,-12.787],[-13.030,-9.640,-12.279],[-4.211,-18.764,-9.658],
[-8.799,-16.094,-11.193],[-16.212,-17.321,-14.397],[-8.672,-17.846,-15.754],
[-20.719,-14.752,-13.406],[-15.296,-14.611,-15.370],[-2.380,-17.066,-8.292],
[-18.575,-18.077,-10.387],[-10.655,-8.484,-16.978],[-15.318,-11.518,-10.538],
[-7.117,-18.754,-12.808],[-18.104,-15.459,-12.915],[-18.310,-9.983,-10.391],
[-19.629,-9.508,-17.674],[-7.847,-8.281,-10.975],[-7.171,-19.802,-10.299],
[-14.213,-21.479,-11.973],[-7.787,-6.972,-14.498],[-8.168,-7.627,-16.995],
[-4.644,-10.967,-7.447],[-10.948,-11.064,-16.272],[-11.085,-13.696,-20.657],
[-3.303,-12.860,-9.173],[-8.269,-12.489,-17.133],[-9.906,-13.728,-7.746],
[-17.097,-12.325,-14.722],[-15.226,-7.955,-13.659],[-15.833,-13.838,-12.155],
[-10.631,-7.741,-10.498],[-17.085,-8.940,-6.343],[-7.919,-13.496,-19.836],
[-18.310,-11.534,-8.035],[-10.386,-5.486,-15.137],[-16.756,-6.260,-6.616],
[-4.628,-21.282,-19.258],[-17.592,-13.729,-17.185],[-3.771,-8.047,-20.450],
[-20.796,-9.750,-8.926],[-23.390,-7.633,-12.487],[-2.859,-6.945,-4.819],
[-13.330,-8.468,-7.295],[-11.628,-3.888,-17.242],[-17.953,-5.213,-16.371],
[-7.274,-3.470,-20.093],[-8.159,-4.923,-16.996],[-5.917,-15.482,-5.137],
[-19.683,-8.435,-6.659],[-16.681,-20.464,-12.278],[-11.044,-21.583,-6.936],
[-9.577,-16.428,-8.049],[-5.514,-13.159,-21.151],[-11.309,-18.781,-20.219],
[-16.471,-16.585,-11.160],[-15.919,-17.001,-17.130],[-11.258,-14.547,-17.146],
[-12.620,-14.885,-4.591],[-9.389,-15.803,-13.817],[-19.539,-15.976,-8.092],
[-16.118,-18.876,-22.169],[-21.759,-19.858,-13.598],[-20.098,-17.046,-17.407],
[-10.879,-14.642,-10.047],[-14.392,-19.100,-13.443],[-18.977,-15.638,-15.405],
[-14.330,-17.311,-19.492],[-9.594,-19.128,-11.068],[-16.958,-15.240,-19.772],
[-7.421,-4.316,-6.447],[-18.743,-10.807,-5.455],[-13.268,-14.493,-11.657],
[-18.481,-19.526,-14.425],[-13.910,-16.658,-9.910],[-14.176,-10.532,-17.804],
[-6.787,-11.631,-8.721],[-8.738,-19.146,-20.248],[-11.559,-10.618,-7.194],
[-8.483,-15.293,-17.545],[-19.563,-11.166,-13.137],[-11.379,-11.363,-19.406],
[-21.224,-12.545,-7.075],[-14.656,-13.591,-6.402],[-10.457,-6.284,-7.575],
[-4.986,-14.720,-18.123],[-4.198,-10.253,-5.025],[-5.910,-8.311,-18.720],
[-8.727,-12.621,-10.491],[-14.937,-7.913,-16.353],[-21.118,-13.195,-20.276],
[-14.078,-10.197,-14.939],[-15.007,-4.864,-8.213],[-15.296,-4.664,-14.416],
[-9.631,-10.057,-9.128],[-9.215,-9.387,-21.265],[-7.554,-7.612,-6.129],
[-2.994,-5.424,-20.821],[-19.092,-18.643,-7.055],[-8.397,-10.268,-18.471],
[-20.245,-6.995,-8.747],[-13.914,-9.531,-23.619],[-15.099,-8.052,-19.630],
[-19.222,-11.839,-19.120],[-12.472,-19.115,-9.858],[-9.139,-20.294,-17.246],
[-12.961,-8.104,-18.257],[-12.917,-16.718,-17.311],[-12.735,-15.330,-7.730],
[-11.684,-18.762,-16.129],[-15.753,-5.630,-19.444],[-15.077,-20.339,-19.656],
[-4.513,-17.668,-6.526],[-6.676,-15.291,-12.340],[-17.685,-7.184,-20.594],
[-19.947,-16.428,-20.576],[-6.679,-5.941,-9.680],[-19.143,-20.554,-11.886],
[-18.005,-7.931,-16.037],[-20.242,-3.163,-4.294],[-14.801,-5.533,-2.474],
[-10.847,-7.490,-20.304],[-9.689,-7.964,-3.371],[-6.471,-7.152,-21.531],
[-20.551,-11.616,-3.812],[-16.571,-10.975,-21.591],[-14.313,-11.246,-4.954],
[-11.581,-10.945,-22.391],[-8.917,-11.305,-6.592],[-19.299,-13.936,-5.903],
[-13.633,-13.758,-17.834],[-15.454,-15.026,-8.469],[-14.527,-13.810,-21.395],
[-10.659,-11.495,-3.191],[-4.832,-11.478,-23.212],[-16.605,-17.334,-8.335],
[-15.276,-19.210,-9.800],[-13.362,-12.767,-8.882],[-10.466,-16.519,-4.248],
[-8.528,-14.438,-5.331],[-16.416,-15.247,-4.753],[-14.185,-22.930,-9.583],
[-12.856,-16.724,-2.713],[-11.512,-18.123,-6.517],[-9.960,-19.422,-4.655],
[-19.125,-16.538,-5.469],[-14.316,-22.346,-6.304],[-13.928,-19.440,-5.500],
[-15.538,-2.951,-19.273],[-19.272,-9.601,-20.794],[-21.759,-3.874,-16.703],
[-20.203,-6.104,-13.958],[-4.485,-4.817,-11.204],[-4.085,-6.190,-8.061],
[-7.945,-9.866,-4.329],[-3.880,-7.819,-16.662],[-4.676,-12.209,-16.610],
[-4.129,-10.502,-10.110],[-6.332,-8.754,-8.284],[-4.397,-11.481,-19.442],
[-5.006,-15.765,-20.874],[-6.570,-14.783,-15.446],[-20.514,-12.857,-10.986],
[-6.781,-12.281,-5.299],[-3.307,-15.727,-4.972],[-6.554,-18.389,-21.723],
[-6.195,-17.225,-17.090],[-6.077,-17.342,-10.842],[-7.084,-15.948,-8.052],
[-21.453,-17.648,-6.504],[-2.543,-21.166,-3.168],[-5.783,-19.733,-16.657],
[-20.900,-18.027,-11.980],[-20.053,-15.620,-10.951],[-2.964,-22.171,-6.092],
[-3.191,-18.351,-21.209],[-22.742,-18.683,-16.050],[-5.596,-5.328,-15.203],
[-20.440,-19.690,-9.283],[-19.516,-7.146,-3.803],[-16.826,-6.820,-9.533],
[-13.141,-5.965,-6.434],[-10.990,-4.467,-5.595],[-5.317,-20.843,-6.895],
[-20.456,-3.810,-7.276],[-17.933,-4.091,-10.275],[-13.092,-4.137,-10.201],
[-9.440,-5.102,-9.985],[-7.196,-2.115,-8.622],[-19.119,-6.413,-11.358],
[-17.458,-5.924,-13.457],[-14.824,-5.892,-12.111],[-11.755,-20.336,-12.944],
[-5.969,-21.228,-12.756],[-4.211,-20.882,-14.966],[-17.845,-21.566,-16.174],
[-20.902,-20.741,-19.020],[-15.428,-22.231,-14.452],[-8.609,-20.824,-13.153],
[-6.435,-22.282,-17.400],[-20.073,-6.553,-18.588],[-18.225,-19.863,-20.643],
[-13.470,-19.692,-17.778],[-12.545,-4.254,-19.727],[-5.584,-5.848,-17.839],
[-18.179,-20.623,-5.779],[-18.038,-21.357,-3.269],[-18.595,-4.488,-20.002],
[-9.062,-21.845,-3.775],[-5.308,-5.881,-5.364],[-11.512,6.036,-15.141],
[-10.884,6.130,-12.261],[-19.287,5.778,-9.319],[-7.435,9.020,-12.503],
[-17.098,9.746,-11.257],[-20.555,3.991,-16.783],[-20.470,5.981,-14.907],
[-10.251,13.866,-12.761],[-12.676,7.800,-13.780],[-4.025,0.147,-9.994],
[-8.815,1.707,-10.234],[-16.576,1.672,-14.516],[-8.879,0.031,-15.256],
[-21.004,3.318,-13.991],[-14.486,4.418,-14.335],[-3.589,1.628,-7.282],
[-17.649,-0.055,-10.410],[-9.309,7.521,-17.101],[-14.443,7.102,-9.452],
[-6.785,-0.252,-13.399],[-18.776,0.582,-13.090],[-16.276,8.552,-7.911],
[-20.772,10.004,-18.276],[-7.656,10.603,-10.450],[-6.849,-0.583,-10.619],
[-13.638,-3.223,-12.582],[-7.926,11.725,-13.782],[-7.947,9.801,-17.633],
[-7.465,7.822,-6.709],[-11.683,6.659,-17.930],[-10.447,3.592,-21.665],
[-4.529,4.490,-9.024],[-6.783,7.731,-15.796],[-9.710,4.070,-7.791],
[-16.448,6.099,-14.395],[-14.801,8.647,-12.662],[-15.534,5.244,-11.773],
[-11.145,10.861,-11.006],[-14.209,10.481,-6.835],[-8.735,4.915,-18.519],
[-17.536,4.639,-7.489],[-12.396,11.569,-16.516],[-17.035,10.453,-6.297],
[-4.892,-3.320,-18.837],[-18.528,5.076,-18.239],[-3.906,10.500,-21.335],
[-21.575,7.441,-9.294],[-20.693,10.732,-10.998],[-2.781,11.246,-5.036],
[-10.977,11.890,-7.259],[-10.675,14.540,-15.810],[-18.718,12.995,-15.239],
[-7.899,12.575,-20.366],[-9.530,12.055,-17.062],[-6.670,3.561,-4.660],
[-20.420,8.602,-6.926],[-15.854,-2.492,-11.123],[-10.810,-3.153,-8.649],
[-8.944,0.502,-7.321],[-7.512,4.863,-23.060],[-10.606,1.112,-20.738],
[-16.517,2.518,-11.602],[-15.642,5.394,-17.648],[-11.717,3.941,-17.149],
[-11.121,3.302,-5.286],[-8.635,2.746,-14.458],[-18.686,2.471,-8.664],
[-16.185,-0.692,-21.073],[-22.535,-0.647,-14.541],[-20.887,0.504,-20.379],
[-10.189,4.328,-10.376],[-14.918,-1.024,-14.544],[-18.479,4.409,-13.930],
[-15.482,2.314,-17.254],[-10.393,-0.254,-11.489],[-18.025,3.077,-21.476],
[-6.448,16.443,-3.796],[-18.348,7.446,-5.789],[-12.379,3.839,-12.432],
[-16.181,-0.658,-16.985],[-13.809,1.709,-10.797],[-13.966,7.957,-18.558],
[-5.964,7.128,-9.474],[-9.580,-1.479,-20.603],[-12.638,8.316,-6.032],
[-8.446,2.679,-20.064],[-17.599,8.395,-13.443],[-11.075,9.689,-20.633],
[-22.026,5.648,-7.486],[-13.435,5.729,-5.078],[-9.991,8.891,-8.028],
[-3.289,3.114,-19.252],[-5.234,7.768,-4.834],[-6.295,12.085,-18.076],
[-7.851,6.559,-11.375],[-17.060,10.940,-16.355],[-19.839,4.831,-20.645],
[-14.605,7.726,-15.496],[-13.498,11.799,-9.710],[-15.473,14.033,-15.587],
[-10.009,8.454,-10.884],[-8.576,8.927,-20.542],[-7.672,12.931,-6.430],
[-4.852,12.660,-20.172],[-19.889,-0.195,-8.568],[-6.837,7.427,-18.944],
[-18.847,9.249,-9.218],[-12.877,10.746,-22.342],[-14.901,9.763,-20.570],
[-18.544,8.469,-19.064],[-11.685,-0.429,-8.815],[-9.396,-0.289,-18.157],
[-12.798,10.614,-18.813],[-13.290,0.566,-17.948],[-12.954,2.655,-7.812],
[-11.470,-1.039,-16.345],[-14.879,11.949,-18.027],[-18.086,1.556,-19.532],
[-7.892,-1.686,-5.819],[-7.067,3.801,-12.520],[-17.501,10.563,-22.728],
[-20.962,3.450,-23.266],[-6.373,13.215,-10.678],[-20.276,-1.783,-11.805],
[-19.969,10.556,-15.715],[-19.386,13.857,-4.677],[-15.213,12.832,-3.906],
[-10.654,11.540,-23.625],[-9.737,10.142,-5.647],[-8.109,10.981,-22.763],
[-21.056,6.683,-4.790],[-15.855,6.184,-20.776],[-16.571,8.410,-4.032],
[-13.269,7.681,-21.545],[-9.186,6.522,-8.708],[-20.365,3.635,-6.457],
[-13.495,5.566,-19.343],[-15.676,2.900,-8.231],[-14.219,4.013,-21.456],
[-10.661,7.513,-4.348],[-4.636,7.389,-22.247],[-15.949,0.958,-6.171],
[-15.366,0.270,-8.949],[-12.390,5.579,-9.316],[-8.636,3.481,-2.752],
[-8.831,5.179,-5.608],[-18.090,0.997,-3.524],[-13.821,-2.247,-8.680],
[-13.779,3.020,-5.184],[-12.420,0.764,-5.540],[-9.949,-0.126,-4.474],
[-18.612,1.863,-6.008],[-14.279,-3.978,-4.862],[-14.551,-0.988,-4.880],
[-14.021,14.284,-19.448],[-20.899,7.375,-20.091],[-20.879,14.693,-15.823],
[-21.792,11.218,-13.917],[-2.994,15.380,-10.712],[-3.765,12.246,-9.428],
[-5.385,10.404,-4.151],[-4.057,7.733,-18.397],[-4.943,5.431,-16.054],
[-4.910,8.239,-11.688],[-6.010,11.720,-8.209],[-3.328,5.714,-20.343],
[-5.503,2.929,-21.060],[-6.521,2.612,-16.284],[-21.510,6.337,-12.520],
[-6.591,5.394,-7.580],[-4.038,4.165,-5.700],[-6.388,-1.114,-19.396],
[-4.460,1.118,-17.951],[-6.240,2.691,-10.168],[-6.482,2.622,-7.341],
[-22.186,1.203,-5.182],[-6.233,-3.550,-3.801],[-6.458,-0.880,-16.807],
[-21.318,0.802,-12.167],[-22.628,3.495,-11.454],[-5.749,0.286,-5.768],
[-3.035,0.912,-20.917],[-20.886,-0.946,-16.928],[-7.574,14.302,-15.666],
[-22.530,-2.168,-8.324],[-19.400,11.571,-6.710],[-15.932,11.820,-8.453],
[-12.042,13.354,-5.246],[-9.969,14.822,-6.573],[-5.003,-3.455,-7.634],
[-19.698,16.459,-8.453],[-18.066,14.141,-9.646],[-13.354,14.494,-10.108],
[-10.488,13.507,-9.406],[-7.660,14.481,-8.672],[-19.159,12.827,-12.055],
[-16.782,12.185,-13.161],[-14.079,12.975,-12.243],[-10.970,-2.125,-13.475],
[-5.759,-2.838,-12.855],[-4.091,-2.843,-15.167],[-17.901,-3.303,-14.234],
[-21.138,-2.334,-19.310],[-13.646,-3.106,-15.579],[-8.723,-2.792,-11.392],
[-9.159,-2.922,-18.213],[-18.316,12.466,-18.535],[-18.750,-1.818,-20.550],
[-13.381,-2.093,-18.143],[-11.126,13.087,-19.198],[-3.902,12.837,-17.175],
[-19.909,-0.712,-5.885],[-16.581,-3.023,-3.104],[-17.656,12.884,-21.353],
[-9.217,-2.861,-3.612],[-6.413,13.789,-4.015],[-10.299,23.418,-14.838],
[-12.170,23.667,-12.383],[-16.685,23.706,-9.997],[-8.963,27.554,-13.519],
[-17.269,28.550,-11.101],[-21.134,22.714,-16.039],[-23.977,23.983,-14.114],
[-9.718,31.242,-12.286],[-11.834,26.917,-11.784],[-4.244,17.713,-11.238],
[-8.766,20.072,-11.973],[-17.058,19.795,-14.252],[-9.496,18.613,-14.208],
[-20.281,21.624,-13.417],[-15.155,21.892,-14.749],[-4.275,21.811,-7.196],
[-20.498,18.684,-10.054],[-10.786,28.256,-15.858],[-15.204,26.441,-10.167],
[-6.823,18.061,-12.830],[-18.618,19.740,-11.951],[-18.136,26.157,-10.346],
[-20.015,27.164,-16.911],[-7.998,27.633,-10.654],[-6.948,18.440,-10.067],
[-13.783,15.700,-12.629],[-7.602,30.036,-14.265],[-8.450,28.869,-16.608],
[-8.668,27.952,-8.111],[-11.589,24.905,-16.649],[-10.094,22.812,-20.617],
[-4.817,24.031,-9.513],[-7.902,24.063,-16.367],[-9.646,22.956,-7.856],
[-17.377,24.587,-14.455],[-15.500,28.659,-13.126],[-15.679,24.301,-12.255],
[-10.716,28.826,-9.913],[-14.580,28.660,-5.642],[-8.381,21.615,-18.108],
[-17.964,23.265,-7.689],[-9.788,30.885,-15.363],[-17.204,29.362,-6.068],
[-5.139,16.702,-20.042],[-16.996,23.448,-17.482],[-3.391,30.899,-22.025],
[-22.349,25.962,-9.710],[-21.571,28.286,-11.203],[-3.390,29.152,-5.783],
[-12.839,28.403,-8.494],[-11.687,32.600,-16.419],[-18.868,32.031,-15.080],
[-10.017,30.442,-21.002],[-10.101,31.217,-18.343],[-6.385,21.329,-5.494],
[-18.257,28.139,-8.526],[-15.595,16.435,-10.790],[-11.051,16.743,-7.753],
[-9.134,20.476,-6.047],[-7.794,24.506,-19.864],[-10.455,16.278,-22.650],
[-15.531,21.384,-11.423],[-16.834,20.551,-16.954],[-11.660,22.314,-18.195],
[-10.719,22.504,-5.292],[-8.868,21.011,-15.247],[-19.131,21.025,-9.427],
[-13.215,19.765,-21.818],[-21.517,17.512,-14.290],[-19.698,19.729,-19.092],
[-11.218,21.473,-10.810],[-14.818,17.918,-13.428],[-19.331,20.371,-15.910],
[-14.696,19.632,-19.662],[-9.637,18.288,-9.803],[-15.439,21.588,-21.047],
[-6.972,32.925,-5.696],[-19.126,25.705,-7.052],[-12.840,20.183,-12.542],
[-17.214,17.720,-16.114],[-13.375,20.269,-9.538],[-12.645,26.730,-18.647],
[-6.669,24.791,-11.647],[-8.082,16.674,-20.005],[-11.199,26.649,-7.289],
[-8.404,19.350,-19.764],[-19.249,24.182,-12.214],[-10.265,27.923,-18.325],
[-20.529,22.725,-7.599],[-13.481,23.413,-5.106],[-10.680,30.104,-7.382],
[-4.304,21.445,-18.525],[-4.376,26.363,-6.134],[-5.825,28.722,-18.139],
[-9.062,23.045,-10.524],[-16.536,27.039,-16.808],[-19.858,23.714,-20.468],
[-13.947,26.060,-16.372],[-14.378,31.003,-9.879],[-14.618,29.660,-15.210],
[-9.461,25.617,-11.712],[-7.856,27.461,-21.168],[-9.652,31.850,-5.672],
[-3.428,30.171,-19.406],[-21.511,18.889,-7.747],[-7.794,26.488,-18.058],
[-19.260,30.129,-10.347],[-14.824,29.873,-23.864],[-14.110,27.350,-21.266],
[-18.988,26.408,-19.254],[-13.666,17.419,-8.757],[-10.588,17.090,-18.818],
[-13.161,29.327,-18.767],[-12.328,19.636,-18.127],[-11.655,21.255,-7.663],
[-10.878,17.323,-16.231],[-14.676,29.870,-20.933],[-15.499,17.250,-20.963],
[-6.264,19.736,-7.954],[-8.128,22.453,-13.089],[-16.638,28.200,-19.783],
[-19.131,21.306,-21.374],[-6.779,31.958,-10.523],[-17.335,14.978,-12.236],
[-18.351,28.972,-15.420],[-18.948,31.605,-5.250],[-14.107,30.965,-3.965],
[-12.309,29.865,-22.333],[-9.698,28.249,-5.548],[-7.376,30.021,-22.000],
[-20.683,26.812,-4.861],[-16.566,26.656,-21.820],[-14.858,25.746,-6.169],
[-11.164,27.563,-20.918],[-8.648,25.394,-9.172],[-17.873,24.330,-4.808],
[-15.040,24.545,-19.559],[-15.644,22.062,-7.508],[-12.733,24.857,-20.905],
[-10.678,25.027,-3.994],[-5.393,25.610,-23.251],[-14.081,19.663,-6.923],
[-16.061,18.891,-8.917],[-12.606,24.319,-9.688],[-7.369,21.594,-2.966],
[-8.383,24.056,-5.816],[-16.690,21.840,-4.535],[-15.287,15.202,-8.421],
[-14.479,17.409,-5.735],[-11.543,19.544,-5.096],[-9.118,17.279,-3.955],
[-19.108,19.249,-4.951],[-15.980,14.537,-5.799],[-13.725,15.802,-3.667],
[-15.331,32.840,-20.252],[-22.107,26.796,-20.891],[-21.081,32.849,-16.851],
[-22.291,31.123,-14.513],[-4.464,32.810,-11.819],[-4.481,31.526,-8.021],
[-4.926,27.670,-3.732],[-4.761,26.388,-18.254],[-4.375,24.199,-15.980],
[-5.328,27.904,-11.615],[-5.802,28.982,-8.822],[-3.415,25.315,-20.966],
[-5.493,22.804,-20.710],[-5.492,21.878,-14.873],[-21.607,25.281,-12.036],
[-5.612,23.911,-5.258],[-3.894,24.980,-3.270],[-5.922,19.519,-21.427],
[-6.244,19.456,-18.092],[-5.534,20.798,-11.242],[-6.753,22.318,-9.182],
[-22.544,19.719,-5.368],[-2.896,15.879,-3.851],[-4.983,16.886,-17.223],
[-19.434,16.758,-11.755],[-23.146,21.902,-12.397],[-4.692,17.492,-5.650],
[-3.368,18.219,-21.505],[-21.081,17.236,-18.675],[-5.935,32.852,-16.419],
[-22.457,15.381,-8.418],[-19.210,29.264,-4.034],[-16.815,30.838,-8.869],
[-12.325,32.132,-5.820],[-11.007,34.219,-6.419],[-6.860,17.082,-7.303],
[-21.201,32.109,-9.321],[-16.758,33.239,-8.015],[-11.419,32.925,-11.134],
[-9.113,31.030,-9.248],[-9.717,34.124,-8.848],[-20.902,32.656,-12.218],
[-18.024,30.904,-12.966],[-14.971,31.962,-13.098],[-11.781,17.731,-13.088],
[-6.107,14.833,-12.930],[-4.480,15.655,-14.793],[-18.897,15.449,-14.181],
[-21.160,14.667,-18.845],[-13.401,16.006,-15.980],[-8.541,15.939,-13.306],
[-6.535,14.987,-18.291],[-20.301,30.311,-18.004],[-19.922,15.983,-21.143],
[-13.631,17.043,-18.952],[-12.360,31.977,-20.169],[-4.157,30.807,-16.538],
[-20.342,16.329,-5.285],[-16.222,16.828,-3.700],[-19.245,32.902,-21.102],
[-9.240,14.561,-3.817],[-5.554,30.972,-4.672],[-9.321,42.036,-14.554],
[-10.821,42.507,-12.279],[-18.245,42.089,-10.088],[-6.790,46.409,-14.342],
[-17.757,46.234,-11.988],[-21.578,39.155,-15.947],[-20.989,42.758,-15.015],
[-9.934,51.276,-12.887],[-12.991,45.317,-10.233],[-5.936,34.632,-10.980],
[-7.198,39.338,-12.268],[-17.395,39.169,-14.035],[-8.763,36.883,-15.734],
[-19.857,40.144,-14.108],[-13.664,42.396,-13.134],[-3.567,38.712,-7.746],
[-17.365,37.322,-10.185],[-9.169,44.952,-16.134],[-16.075,45.113,-9.991],
[-6.544,36.831,-13.574],[-19.289,39.984,-11.433],[-18.735,46.193,-9.277],
[-21.867,46.899,-17.843],[-7.402,47.172,-11.522],[-7.424,37.195,-10.983],
[-12.307,34.721,-12.988],[-9.040,49.339,-15.030],[-7.889,47.407,-16.640],
[-6.828,45.332,-7.020],[-11.715,43.792,-16.893],[-10.096,39.949,-21.295],
[-3.708,42.910,-9.674],[-6.783,43.316,-15.943],[-8.303,40.662,-7.641],
[-15.901,43.533,-14.745],[-14.589,45.531,-12.588],[-16.505,43.308,-12.059],
[-10.354,48.662,-12.631],[-15.439,45.673,-7.056],[-9.543,41.494,-18.956],
[-19.258,43.367,-7.937],[-11.038,47.744,-15.998],[-17.395,49.029,-5.577],
[-6.678,33.787,-19.259],[-16.957,42.923,-17.188],[-4.006,49.232,-19.999],
[-21.397,45.052,-9.091],[-21.570,47.927,-10.248],[-4.065,47.886,-6.296],
[-12.581,48.535,-6.181],[-11.715,51.539,-17.507],[-18.570,50.015,-15.336],
[-7.592,52.026,-20.010],[-8.506,50.338,-17.460],[-5.887,40.004,-7.027],
[-19.882,45.705,-6.837],[-14.314,33.647,-10.599],[-11.931,32.121,-8.708],
[-8.781,38.338,-8.948],[-8.157,43.078,-20.551],[-13.079,37.091,-20.346],
[-16.532,40.321,-11.518],[-13.693,39.715,-14.645],[-10.982,40.793,-16.319],
[-11.691,40.200,-4.937],[-9.231,39.451,-14.512],[-17.974,39.755,-8.421],
[-16.737,38.245,-20.935],[-20.184,37.065,-14.826],[-18.973,37.740,-19.771],
[-10.549,40.406,-10.564],[-14.251,37.194,-14.036],[-18.467,38.498,-16.234],
[-16.557,37.883,-18.262],[-10.528,36.609,-10.993],[-16.637,40.723,-19.751],
[-7.056,51.724,-5.094],[-18.127,43.552,-4.914],[-11.793,38.577,-12.735],
[-17.367,36.192,-14.375],[-14.152,38.446,-9.153],[-12.574,44.089,-19.515],
[-6.210,44.312,-11.234],[-10.287,36.717,-20.171],[-12.944,44.582,-7.312],
[-8.957,38.591,-18.819],[-18.577,43.880,-13.613],[-10.367,46.335,-18.347],
[-21.804,42.989,-7.407],[-13.550,42.245,-5.636],[-10.123,48.660,-7.079],
[-2.191,39.434,-18.417],[-4.214,45.013,-5.906],[-6.503,47.832,-19.091],
[-8.548,42.831,-10.918],[-15.443,48.319,-15.035],[-19.595,41.812,-20.577],
[-13.776,45.231,-15.605],[-13.847,47.924,-9.327],[-15.804,50.623,-16.706],
[-10.292,45.807,-11.473],[-8.960,46.266,-20.755],[-7.820,49.345,-5.628],
[-2.652,51.599,-18.255],[-20.004,37.401,-8.218],[-5.748,45.821,-16.880],
[-18.885,48.960,-8.194],[-14.760,46.111,-21.569],[-16.646,45.439,-18.548],
[-19.025,44.813,-19.682],[-12.549,36.423,-8.532],[-10.945,35.421,-17.435],
[-12.314,47.970,-19.081],[-11.527,38.122,-15.754],[-12.718,39.107,-6.982],
[-12.524,35.598,-15.438],[-13.813,50.556,-19.627],[-15.924,35.579,-19.842],
[-6.262,37.452,-8.358],[-7.106,42.005,-13.125],[-20.031,47.378,-20.212],
[-20.716,39.421,-21.424],[-5.433,48.992,-10.776],[-18.867,34.540,-11.457],
[-21.223,48.588,-15.578],[-21.198,49.465,-3.373],[-15.329,49.530,-4.058],
[-12.945,48.491,-21.914],[-9.824,48.082,-4.311],[-6.549,48.321,-21.626],
[-20.443,45.117,-3.826],[-17.484,45.200,-21.712],[-15.426,44.393,-4.690],
[-11.982,46.129,-21.336],[-8.604,43.571,-7.908],[-19.575,41.269,-4.369],
[-15.155,43.109,-19.605],[-15.618,40.367,-7.718],[-12.903,41.014,-20.875],
[-10.457,43.850,-4.877],[-5.717,42.989,-22.328],[-15.446,37.836,-6.576],
[-15.501,35.567,-9.073],[-13.203,41.475,-8.921],[-9.084,39.124,-4.434],
[-8.578,41.690,-5.370],[-17.215,39.826,-5.163],[-13.430,33.811,-7.511],
[-12.983,37.281,-4.882],[-10.548,37.446,-6.211],[-8.589,36.304,-3.990],
[-18.311,36.943,-4.783],[-14.970,33.265,-5.543],[-12.814,33.459,-3.606],
[-12.026,53.081,-19.963],[-21.352,44.363,-18.280],[-21.420,51.144,-17.076],
[-21.995,50.427,-13.818],[-3.858,51.321,-10.612],[-4.640,49.742,-8.537],
[-5.997,46.539,-4.382],[-5.529,43.083,-18.751],[-4.047,43.951,-15.787],
[-4.714,46.515,-11.848],[-6.720,48.107,-7.527],[-2.483,42.376,-19.180],
[-5.858,39.741,-20.478],[-6.005,39.950,-14.583],[-19.689,44.013,-11.079],
[-5.996,42.722,-6.114],[-3.981,40.921,-5.400],[-7.154,36.485,-20.936],
[-6.704,38.280,-17.172],[-4.962,38.683,-10.334],[-7.186,40.718,-9.928],
[-21.937,37.897,-5.239],[-3.648,34.498,-1.803],[-5.994,35.568,-16.964],
[-20.386,37.148,-11.819],[-22.554,40.461,-11.667],[-3.227,36.039,-6.526],
[-3.696,37.504,-20.811],[-21.906,35.637,-17.616],[-6.468,50.846,-15.469],
[-20.752,34.902,-9.407],[-19.749,47.477,-5.018],[-16.611,48.449,-9.590],
[-12.306,51.949,-5.370],[-9.763,51.534,-6.330],[-5.096,33.996,-7.486],
[-19.816,50.888,-5.881],[-18.723,51.602,-9.557],[-12.923,51.738,-10.214],
[-10.323,50.384,-9.609],[-8.155,52.442,-10.801],[-19.356,49.027,-10.901],
[-16.804,49.994,-13.116],[-13.911,49.130,-12.344],[-10.406,36.440,-13.804],
[-7.222,32.546,-13.235],[-4.623,34.798,-13.802],[-19.680,34.369,-14.180],
[-20.092,34.674,-19.160],[-15.048,34.319,-15.097],[-8.404,34.876,-12.142],
[-8.794,33.756,-17.217],[-20.194,49.066,-18.134],[-19.223,35.706,-21.465],
[-14.090,38.520,-17.068],[-10.953,50.148,-20.079],[-4.589,49.323,-17.028],
[-19.539,34.156,-6.357],[-15.457,35.141,-2.360],[-18.683,50.482,-20.720],
[-9.650,33.787,-3.830],[-5.284,49.658,-4.509],[-4.287,54.530,-10.263],
[-9.233,58.011,-10.686],[-17.372,57.294,-14.522],[-7.846,55.446,-14.152],
[-21.052,58.084,-13.760],[-3.424,58.189,-7.435],[-17.549,54.837,-9.186],
[-7.188,57.947,-13.394],[-18.626,55.573,-12.783],[-7.038,54.503,-9.340],
[-12.972,52.550,-13.035],[-4.664,52.951,-19.496],[-6.375,56.372,-4.944],
[-17.198,53.072,-12.674],[-11.309,52.680,-8.108],[-7.602,56.701,-7.465],
[-12.839,55.894,-20.667],[-16.764,57.460,-10.508],[-14.939,58.176,-15.403],
[-9.716,57.769,-14.334],[-19.658,-58.024,-9.500],[-15.381,55.630,-22.238],
[-22.348,55.612,-13.342],[-21.135,56.246,-16.172],[-13.607,54.831,-14.780],
[-19.261,58.363,-16.705],[-14.994,58.278,-18.037],[-9.052,55.397,-11.204],
[-13.403,-58.977,-20.136],[-12.517,58.700,-11.909],[-16.368,54.830,-16.038],
[-13.909,57.112,-10.437],[-10.298,55.889,-21.094],[-8.536,57.636,-20.085],
[-3.684,58.878,-18.626],[-19.850,56.354,-9.532],[-11.697,55.092,-9.318],
[-9.414,53.808,-18.350],[-13.300,56.582,-16.843],[-12.978,57.257,-8.196],
[-11.889,54.184,-16.955],[-17.362,56.960,-20.259],[-4.202,55.463,-7.229],
[-19.764,57.739,-23.361],[-19.738,51.869,-11.999],[-15.272,58.570,-8.203],
[-14.425,54.599,-9.164],[-9.139,57.345,-4.693],[-17.981,58.025,-7.387],
[-16.222,52.091,-8.829],[-13.794,54.534,-5.632],[-11.704,56.201,-4.760],
[-8.009,54.256,-4.436],[-19.298,54.875,-3.419],[-14.100,51.298,-7.476],
[-14.383,52.366,-3.367],[-4.649,58.372,-21.008],[-6.677,55.623,-20.682],
[-6.560,57.604,-18.041],[-6.614,58.071,-9.907],[-8.070,-59.015,-8.647],
[-22.526,57.365,-6.216],[-3.083,51.938,-3.400],[-5.328,55.752,-16.848],
[-22.385,55.026,-10.699],[-20.516,58.824,-11.239],[-2.648,53.499,-5.836],
[-3.473,53.959,-21.961],[-21.142,53.681,-17.362],[-22.585,52.664,-9.411],
[-5.431,53.700,-5.618],[-11.379,55.724,-12.331],[-5.945,51.491,-12.830],
[-4.969,53.284,-15.233],[-18.500,52.864,-15.517],[-18.952,51.978,-18.154],
[-13.893,51.864,-15.360],[-10.261,53.821,-14.129],[-6.625,53.277,-17.343],
[-17.670,54.376,-21.432],[-14.330,53.129,-18.382],[-20.220,53.714,-8.641],
[-16.951,52.004,-5.931],[-10.237,51.405,-3.486],[-10.357,-48.708,2.631],
[-11.785,-50.510,6.640],[-17.857,-51.322,7.803],[-7.038,-47.927,5.894],
[-17.681,-45.855,4.934],[-21.066,-52.811,1.598],[-21.092,-49.808,4.371],
[-8.483,-42.051,6.590],[-12.012,-47.718,5.617],[-2.778,-57.114,7.463],
[-9.119,-53.705,7.935],[-16.098,-55.280,4.365],[-8.984,-57.258,3.160],
[-19.214,-51.791,4.808],[-14.303,-52.901,5.241],[-2.854,-52.487,10.032],
[-18.743,-56.629,8.058],[-8.236,-47.124,3.541],[-15.476,-50.844,9.269],
[-7.178,-56.326,6.000],[-17.971,-53.837,6.444],[-16.122,-47.656,8.957],
[-20.237,-46.994,0.136],[-7.237,-46.655,8.448],[-6.978,-57.501,9.729],
[-13.603,-58.627,6.294],[-8.430,-45.008,5.525],[-9.666,-44.901,2.906],
[-7.527,-45.768,11.282],[-11.499,-49.196,0.428],[-11.206,-51.328,-3.805],
[-4.129,-50.401,8.652],[-8.042,-52.129,1.358],[-9.745,-52.695,10.818],
[-16.763,-49.615,4.202],[-15.131,-45.000,5.094],[-15.707,-50.630,6.452],
[-10.910,-45.032,7.160],[-15.847,-46.440,11.809],[-9.928,-50.863,-1.209],
[-17.179,-50.681,12.469],[-12.280,-43.439,3.342],[-16.806,-43.633,12.192],
[-5.835,-56.695,-0.815],[-16.908,-51.472,0.954],[-4.858,-43.592,-2.533],
[-20.220,-51.134,9.315],[-21.556,-44.850,7.378],[-2.577,-44.433,10.991],
[-12.857,-45.401,11.603],[-12.118,-42.112,-0.043],[-17.706,-40.666,3.102],
[-9.917,-42.536,-1.755],[-8.955,-42.786,1.032],[-7.416,-53.490,12.099],
[-19.158,-47.822,11.245],[-16.539,-57.961,7.296],[-12.355,59.422,8.825],
[-9.076,-55.731,10.559],[-8.047,-49.490,-2.864],[-10.713,-54.130,-2.968],
[-15.769,-53.323,8.058],[-14.645,-53.250,2.471],[-11.282,-53.787,2.788],
[-11.582,-51.922,13.405],[-9.624,-52.471,4.111],[-18.040,-53.411,9.678],
[-15.656,-55.753,-3.299],[-21.830,-53.927,3.866],[-21.790,-55.292,-0.203],
[-11.698,-52.235,9.128],[-13.110,-55.291,3.933],[-19.114,-54.697,4.082],
[-14.009,-51.686,-0.489],[-9.610,-54.983,5.151],[-17.141,-51.625,-1.737],
[-7.512,-42.106,14.421],[-18.074,-49.516,15.086],[-11.669,-53.302,6.271],
[-17.372,-56.063,1.891],[-13.438,-54.109,8.326],[-13.425,-48.504,-1.790],
[-7.225,-49.583,8.083],[-8.389,-56.067,-1.582],[-13.765,-48.365,11.201],
[-8.586,-53.321,-1.597],[-18.073,-48.392,6.889],[-11.157,-47.126,-1.559],
[-20.977,-50.062,11.498],[-13.177,-50.719,11.690],[-9.854,-46.120,9.653],
[-2.507,-53.002,-1.563],[-3.769,-47.757,11.776],[-4.296,-46.190,0.042],
[-9.097,-50.898,6.399],[-16.605,-45.246,2.481],[-19.549,-49.217,-2.676],
[-15.205,-47.828,2.684],[-14.776,-44.360,9.831],[-15.440,-42.318,2.228],
[-10.968,-48.232,8.287],[-8.594,-47.620,-1.261],[-8.593,-43.578,12.026],
[-5.781,-38.938,-4.253],[-18.971,-55.736,10.663],[-6.565,-47.233,0.901],
[-20.040,-44.803,9.549],[-13.921,-45.947,-3.301],[-15.727,-46.489,0.303],
[-19.460,-50.712,-0.243],[-11.677,-55.669,9.827],[-10.690,-54.994,-0.064],
[-13.302,-44.535,-1.026],[-14.292,-54.670,0.273],[-12.547,-53.362,11.369],
[-11.500,-57.616,3.552],[-15.849,-43.612,-0.737],[-16.374,-57.022,-0.657],
[-4.863,-54.587,11.880],[-6.945,-51.718,4.743],[-17.730,-44.928,-1.942],
[-20.667,-54.298,-3.200],[-5.843,-44.432,7.066],[-19.211,-57.509,5.644],
[-21.700,-45.812,2.802],[-21.044,-41.790,13.072],[-14.168,-43.808,13.412],
[-11.307,-44.527,-2.931],[-10.249,-46.003,12.531],[-7.695,-43.907,-2.687],
[-21.083,-48.700,13.961],[-16.326,-47.463,-2.270],[-14.806,-48.458,13.662],
[-12.707,-48.740,-4.453],[-7.546,-50.388,10.945],[-19.826,-51.597,13.541],
[-15.456,-49.361,-0.078],[-15.049,-52.896,10.985],[-14.882,-50.922,-2.821],
[-9.142,-48.453,13.223],[-5.091,-49.256,-3.154],[-14.600,-53.554,13.380],
[-16.485,-56.007,9.562],[-13.017,-49.869,9.039],[-6.930,-53.890,14.772],
[-8.903,-51.274,13.447],[-17.427,-53.103,13.547],[-15.762,58.654,10.886],
[-14.244,-54.813,15.645],[-10.331,-54.686,12.805],[-9.040,-55.866,14.704],
[-19.838,-55.912,13.092],[-13.233,-59.090,12.341],[-13.410,-58.073,15.244],
[-14.950,-40.313,0.128],[-21.647,-46.486,-2.873],[-21.674,-40.712,2.491],
[-20.385,-42.606,5.153],[-3.200,-43.038,6.975],[-5.267,-43.053,11.062],
[-6.584,-47.106,14.591],[-2.150,-50.638,1.268],[-5.763,-50.173,1.548],
[-4.838,-47.985,7.470],[-4.969,-45.539,9.667],[-4.343,-50.746,-0.841],
[-6.803,-51.771,-3.031],[-5.195,-53.781,4.568],[-20.976,-47.630,6.393],
[-6.162,-49.388,12.933],[-3.395,-51.987,13.100],[-5.466,-55.081,-3.452],
[-7.550,-54.600,2.311],[-6.009,-52.298,7.357],[-6.559,-54.348,9.444],
[-22.037,-54.740,13.486],[-2.308,-57.994,14.649],[-5.594,-56.727,2.792],
[-20.458,-55.168,6.691],[-21.134,-52.716,6.426],[-3.728,-57.200,12.379],
[-3.460,-56.636,-2.132],[-20.847,-57.582,1.223],[-5.944,-40.544,4.006],
[-21.765,-56.869,10.161],[-17.905,-46.770,13.807],[-17.334,-44.445,9.217],
[-11.425,-43.180,13.588],[-9.867,-40.944,12.503],[-3.443,59.255,11.089],
[-18.258,-41.589,9.150],[-15.556,-41.950,9.494],[-11.433,-40.599,7.093],
[-10.093,-42.725,9.017],[-7.706,-41.678,9.995],[-18.870,-44.282,6.848],
[-17.325,-43.094,4.358],[-15.185,-41.966,5.336],[-12.002,-56.416,7.120],
[-6.631,-58.961,4.267],[-2.483,58.815,2.115],[-18.123,-59.174,4.002],
[-19.916,58.419,0.350],[-14.179,-58.072,3.512],[-9.925,-57.589,6.012],
[-7.526,-59.550,1.814],[-19.073,-43.014,1.717],[-18.379,-56.109,-2.486],
[-12.959,-57.546,0.554],[-12.336,-40.047,-1.819],[-3.822,-43.295,1.403],
[-19.105,-58.164,12.141],[-16.429,-57.931,15.132],[-18.368,-42.258,-1.127],
[-8.780,-59.083,13.488],[-4.948,-45.347,12.774],[-8.445,-33.285,1.654],
[-11.369,-31.116,7.206],[-18.908,-32.824,7.745],[-7.911,-28.621,4.318],
[-17.483,-27.177,5.834],[-20.815,-34.274,2.485],[-20.701,-30.878,4.605],
[-12.125,-22.175,4.598],[-12.731,-28.325,7.618],[-4.068,-36.571,8.673],
[-8.649,-35.267,7.596],[-17.228,-36.382,5.369],[-9.947,-37.267,4.188],
[-20.848,-33.858,5.374],[-14.206,-31.221,4.283],[-2.513,-34.697,9.868],
[-18.756,-38.573,8.843],[-11.652,-28.483,3.088],[-17.158,-30.840,7.815],
[-7.989,-37.587,6.197],[-18.844,-35.342,7.459],[-16.697,-29.107,9.705],
[-20.112,-27.339,0.560],[-7.641,-27.664,7.112],[-6.584,-39.315,7.372],
[-15.118,-38.795,5.839],[-8.409,-23.683,4.415],[-7.350,-26.404,2.486],
[-6.313,-28.597,11.598],[-11.470,-31.499,3.359],[-12.044,-32.733,-2.737],
[-4.086,-30.437,9.936],[-7.292,-30.845,2.026],[-9.137,-33.796,10.515],
[-16.928,-31.598,4.114],[-14.526,-27.468,5.540],[-15.395,-32.711,6.584],
[-10.228,-27.188,8.474],[-15.799,-26.914,12.062],[-9.810,-32.853,-1.111],
[-18.827,-30.797,10.097],[-10.821,-24.723,5.141],[-16.650,-22.802,13.528],
[-5.242,-40.215,-0.035],[-17.206,-31.602,-0.391],[-4.397,-27.214,-2.027],
[-20.044,-28.458,10.392],[-21.596,-26.967,7.752],[-2.824,-25.932,10.835],
[-12.791,-25.461,10.493],[-12.135,-22.385,1.460],[-18.625,-24.907,3.768],
[-8.829,-23.360,-1.449],[-9.404,-24.403,1.888],[-5.441,-36.238,12.975],
[-19.941,-26.296,12.508],[-16.277,-39.866,8.200],[-12.426,-41.004,9.710],
[-8.124,-36.273,11.804],[-7.106,-30.946,-5.091],[-12.303,-37.298,-1.314],
[-16.194,-34.841,8.011],[-16.048,-34.120,3.661],[-12.005,-33.666,1.243],
[-11.874,-34.516,14.243],[-9.626,-33.991,4.771],[-18.828,-35.750,11.933],
[-15.427,-38.262,-3.991],[-21.992,-38.097,3.745],[-21.638,-35.587,-0.146],
[-11.152,-33.489,8.503],[-15.096,-37.129,3.684],[-19.335,-36.459,2.753],
[-14.945,-37.098,0.919],[-10.032,-37.469,8.257],[-16.604,-35.590,-1.706],
[-7.178,-23.279,13.813],[-19.513,-29.753,12.684],[-12.637,-33.926,6.361],
[-17.956,-36.750,0.524],[-14.379,-36.025,9.415],[-14.272,-29.389,0.159],
[-6.025,-31.503,7.635],[-9.636,-37.652,-1.673],[-11.147,-30.228,10.921],
[-9.217,-35.345,-0.372],[-18.281,-29.499,5.217],[-11.123,-28.206,-1.155],
[-20.541,-32.457,11.409],[-13.049,-31.958,13.143],[-10.250,-26.504,11.495],
[-2.822,-33.900,-0.874],[-4.014,-29.635,12.457],[-5.996,-26.210,-0.018],
[-8.807,-31.987,8.490],[-15.094,-26.096,3.301],[-19.223,-31.830,-2.796],
[-15.212,-28.995,2.811],[-14.422,-23.940,8.814],[-16.596,-23.955,2.319],
[-8.962,-29.313,9.387],[-10.016,-30.470,-2.311],[-7.772,-24.958,11.824],
[-4.664,-24.438,-1.882],[-20.999,-35.032,9.935],[-7.640,-31.167,-1.369],
[-18.998,-27.152,8.350],[-13.060,-26.519,-5.802],[-15.534,-27.638,-1.596],
[-19.077,-29.861,-1.091],[-12.233,-37.367,10.532],[-9.924,-37.794,1.045],
[-13.516,-26.536,-0.426],[-12.698,-36.222,1.980],[-12.840,-34.366,11.464],
[-11.983,-38.507,3.130],[-13.672,-23.904,-0.440],[-15.823,-38.387,-1.363],
[-5.972,-38.530,9.781],[-7.594,-33.142,6.425],[-18.238,-26.710,-1.314],
[-19.357,-34.562,-1.961],[-7.335,-24.388,8.021],[-19.063,-40.449,6.587],
[-18.977,-27.390,3.107],[-21.540,-25.696,14.660],[-12.976,-24.072,16.465],
[-11.088,-26.387,-4.032],[-9.914,-25.464,14.037],[-8.350,-26.500,-2.850],
[-22.484,-30.568,12.442],[-16.783,-27.867,-3.936],[-14.128,-27.624,13.913],
[-12.147,-29.073,-3.867],[-8.221,-31.073,11.345],[-20.105,-33.745,14.672],
[-14.076,-31.736,-0.955],[-16.498,-34.149,11.167],[-15.088,-33.175,-3.337],
[-10.418,-31.234,13.481],[-5.062,-30.796,-3.215],[-14.964,-36.680,12.139],
[-13.266,-38.146,8.207],[-14.803,-32.311,10.478],[-6.939,-34.882,15.043],
[-8.308,-32.691,13.519],[-17.463,-32.502,13.398],[-14.682,-39.132,10.287],
[-13.342,-36.878,15.076],[-11.013,-36.695,12.724],[-9.026,-37.916,14.173],
[-20.814,-37.261,13.356],[-14.612,-41.443,12.435],[-14.505,-39.347,14.251],
[-15.255,-20.918,-1.267],[-20.665,-28.094,-2.224],[-21.655,-21.772,2.350],
[-21.355,-25.052,5.153],[-2.220,-23.404,8.696],[-4.698,-24.202,9.034],
[-6.357,-28.118,14.757],[-5.591,-29.537,0.222],[-4.896,-33.588,1.481],
[-4.394,-29.015,7.639],[-6.575,-26.477,9.686],[-4.683,-31.986,-0.580],
[-4.884,-35.420,-1.414],[-6.755,-33.629,3.869],[-20.372,-29.424,7.480],
[-5.713,-32.426,11.177],[-3.992,-32.950,13.275],[-7.008,-37.062,-2.492],
[-7.221,-36.942,0.559],[-6.182,-35.993,7.124],[-6.118,-35.628,10.005],
[-23.288,-36.127,13.311],[-4.807,-41.140,15.466],[-5.617,-38.729,2.010],
[-20.908,-37.553,7.052],[-22.942,-33.275,7.043],[-5.709,-38.908,14.193],
[-3.245,-37.492,-4.073],[-21.751,-38.054,0.573],[-6.670,-23.245,2.733],
[-21.164,-40.009,9.374],[-18.985,-26.371,14.983],[-16.410,-26.159,9.279],
[-12.398,-24.601,13.915],[-9.493,-22.859,11.566],[-5.164,-40.113,11.574],
[-18.185,-21.521,11.397],[-17.325,-23.196,9.185],[-12.216,-23.407,7.129],
[-9.725,-23.798,8.631],[-7.034,-21.808,8.690],[-19.412,-24.518,6.999],
[-16.444,-24.560,5.143],[-13.676,-24.578,5.128],[-12.319,-37.810,5.708],
[-4.339,-40.818,6.108],[-2.539,-41.953,4.280],[-19.551,-39.054,4.227],
[-21.400,-40.555,-0.181],[-13.413,-40.598,3.469],[-9.588,-39.821,5.069],
[-8.992,-40.113,-0.137],[-19.092,-23.448,1.539],[-18.817,-37.380,-2.136],
[-12.688,-39.383,0.716],[-11.926,-23.261,-2.542],[-4.271,-24.497,1.201],
[-19.815,-39.466,12.084],[-16.800,-40.305,15.762],[-18.234,-24.329,-2.651],
[-8.167,-40.080,15.990],[-5.371,-26.124,12.537],[-11.656,-12.506,3.770],
[-10.883,-13.113,6.407],[-18.319,-13.299,7.797],[-6.015,-10.208,5.049],
[-18.049,-10.089,7.318],[-20.842,-13.154,1.042],[-20.423,-13.198,3.708],
[-11.202,-6.071,4.072],[-12.291,-10.043,5.262],[-5.059,-19.399,6.575],
[-9.359,-16.385,7.318],[-17.614,-18.686,4.075],[-8.666,-19.249,2.442],
[-20.175,-15.606,5.237],[-13.943,-15.124,4.797],[-3.399,-17.161,9.569],
[-18.731,-17.814,10.355],[-12.500,-10.178,2.345],[-13.937,-12.133,8.744],
[-8.429,-18.507,5.343],[-19.275,-16.365,7.936],[-16.696,-11.405,9.639],
[-19.781,-9.497,0.857],[-8.123,-8.525,7.391],[-5.403,-19.269,9.396],
[-14.555,-21.698,5.966],[-7.633,-7.170,4.998],[-9.220,-9.389,3.012],
[-7.201,-9.643,11.975],[-13.089,-12.901,1.566],[-10.354,-14.253,-3.084],
[-4.778,-14.360,10.067],[-9.250,-14.089,1.988],[-8.698,-16.618,11.648],
[-17.088,-11.595,4.312],[-15.532,-8.742,6.795],[-15.599,-12.525,6.533],
[-10.626,-7.291,8.710],[-14.654,-9.715,11.078],[-8.103,-12.399,0.392],
[-17.667,-13.434,10.820],[-12.980,-7.486,2.227],[-16.779,-8.966,12.834],
[-5.450,-21.881,-0.762],[-18.194,-13.218,2.095],[-4.674,-6.552,-1.885],
[-21.057,-11.389,9.125],[-21.209,-8.798,7.279],[-3.397,-7.312,11.803],
[-12.614,-7.831,11.742],[-14.209,-3.798,1.146],[-18.955,-4.213,2.652],
[-8.035,-5.448,-2.418],[-9.190,-5.192,0.896],[-7.526,-15.915,14.047],
[-20.123,-8.883,9.786],[-16.585,-20.153,8.124],[-12.141,-20.966,11.143],
[-7.558,-19.071,11.308],[-7.548,-12.639,-2.482],[-11.777,-19.799,-0.728],
[-16.822,-15.511,6.417],[-15.855,-14.077,1.116],[-11.737,-15.405,2.330],
[-9.996,-17.067,13.935],[-8.893,-15.575,4.123],[-19.896,-14.261,9.578],
[-15.385,-17.571,-3.100],[-21.413,-19.908,4.501],[-20.563,-17.709,2.691],
[-9.638,-14.432,10.203],[-14.009,-19.551,4.149],[-16.372,-15.749,3.699],
[-16.337,-17.222,-0.660],[-7.669,-18.453,8.308],[-17.154,-14.449,-1.416],
[-7.144,-2.450,13.579],[-18.517,-11.832,12.789],[-12.893,-14.904,7.289],
[-17.854,-18.501,1.262],[-14.030,-16.368,9.245],[-14.237,-9.522,0.175],
[-6.548,-10.382,8.751],[-10.430,-18.027,-1.902],[-11.285,-9.967,11.730],
[-9.924,-14.802,-0.422],[-19.494,-11.217,5.509],[-12.199,-7.735,-0.789],
[-20.238,-13.848,12.284],[-12.752,-15.170,12.985],[-9.860,-7.422,11.250],
[-3.863,-17.397,-1.213],[-4.238,-9.981,11.977],[-5.100,-7.844,1.074],
[-8.497,-12.273,8.068],[-16.470,-8.860,3.712],[-20.481,-13.122,-1.564],
[-14.349,-11.615,3.931],[-12.810,-7.762,7.128],[-16.236,-6.151,3.121],
[-9.086,-9.416,9.723],[-8.476,-10.187,-1.499],[-7.603,-5.983,12.285],
[-3.317,-4.503,-0.470],[-21.453,-18.003,11.097],[-6.929,-10.057,0.731],
[-21.523,-6.693,9.441],[-15.029,-8.355,-4.666],[-14.899,-7.916,-1.677],
[-18.026,-11.591,-0.133],[-12.246,-18.457,9.304],[-10.353,-17.578,0.855],
[-11.165,-3.728,-0.391],[-13.208,-16.752,0.144],[-13.408,-17.391,11.686],
[-11.576,-19.689,2.872],[-16.760,-4.353,-0.558],[-16.175,-20.777,1.735],
[-5.422,-19.048,12.680],[-6.508,-14.535,7.897],[-18.575,-8.710,-2.028],
[-19.873,-15.582,-2.351],[-7.264,-5.416,7.203],[-17.792,-22.222,6.497],
[-18.642,-7.438,2.971],[-20.957,-5.517,12.890],[-16.189,-6.363,13.811],
[-12.556,-7.166,-3.412],[-9.984,-8.346,14.073],[-5.731,-8.186,-4.212],
[-21.338,-11.699,14.322],[-16.628,-10.076,-3.347],[-14.188,-12.134,13.128],
[-12.316,-9.401,-4.834],[-8.840,-11.884,10.925],[-19.270,-14.700,14.631],
[-15.285,-12.086,-1.382],[-15.712,-14.371,8.731],[-14.646,-13.406,-3.505],
[-10.671,-12.497,14.047],[-3.488,-11.489,-2.158],[-16.298,-18.849,12.672],
[-15.201,-18.694,9.799],[-13.043,-14.012,10.554],[-9.733,-16.462,16.435],
[-8.295,-13.442,13.206],[-16.766,-15.819,15.107],[-14.461,-21.456,9.550],
[-12.589,-17.111,14.896],[-9.429,-20.320,12.842],[-9.892,-19.632,15.327],
[-18.826,-17.729,13.730],[-13.506,-23.084,11.832],[-13.690,-21.892,14.507],
[-13.550,-2.497,-2.424],[-21.053,-9.638,-2.072],[-21.861,-3.595,1.588],
[-22.458,-6.351,3.951],[-4.399,-5.350,6.890],[-5.439,-5.477,10.470],
[-6.969,-9.443,14.790],[-4.017,-10.803,0.469],[-4.836,-12.965,1.843],
[-4.056,-10.368,6.936],[-5.427,-7.866,8.665],[-5.894,-14.452,-0.373],
[-4.994,-15.480,-2.626],[-6.387,-14.611,3.176],[-21.192,-13.190,6.834],
[-5.908,-12.074,12.690],[-5.000,-14.555,14.174],[-7.997,-19.051,-3.070],
[-6.971,-17.761,1.089],[-5.930,-16.977,6.783],[-6.190,-16.767,10.373],
[-21.204,-16.542,13.581],[-2.433,-21.447,15.856],[-5.742,-20.033,3.575],
[-20.669,-18.545,6.539],[-22.082,-15.873,7.147],[-3.273,-21.204,12.930],
[-5.372,-19.883,-3.627],[-22.428,-19.311,0.835],[-7.107,-4.882,3.395],
[-20.306,-20.636,10.090],[-19.387,-9.188,12.654],[-16.675,-7.513,10.265],
[-11.423,-5.186,15.259],[-9.853,-3.525,11.551],[-4.889,-21.893,10.832],
[-19.054,-4.596,9.575],[-16.363,-3.564,9.290],[-10.820,-3.753,7.962],
[-8.922,-5.443,9.748],[-8.476,-2.851,8.999],[-20.172,-5.716,7.087],
[-17.571,-6.436,5.908],[-14.735,-5.655,6.186],[-11.809,-19.294,5.799],
[-6.472,-21.814,5.664],[-3.462,-22.807,3.734],[-19.675,-22.264,4.322],
[-21.017,-22.510,-0.192],[-14.599,-22.562,3.201],[-9.604,-21.360,5.416],
[-8.122,-21.601,1.246],[-20.907,-6.372,1.772],[-18.447,-20.790,-0.580],
[-14.144,-19.369,0.864],[-10.860,-5.275,-2.918],[-5.646,-5.048,1.005],
[-19.250,-21.450,13.667],[-15.966,-21.346,15.747],[-18.293,-4.961,-2.676],
[-9.769,-22.423,14.674],[-6.696,-4.892,14.658],[-11.494,7.426,5.083],
[-12.219,5.906,6.986],[-17.847,4.470,6.872],[-9.033,8.457,6.823],
[-18.133,8.970,5.393],[-19.916,4.645,1.920],[-19.923,4.385,5.006],
[-10.593,12.196,4.281],[-12.403,8.898,7.814],[-2.967,0.447,7.419],
[-9.710,0.089,5.249],[-15.910,1.169,3.510],[-8.944,2.477,3.671],
[-20.479,2.357,3.629],[-14.128,4.387,4.462],[-3.842,3.299,10.120],
[-18.599,-2.036,8.521],[-10.559,8.204,2.640],[-15.199,8.777,8.436],
[-6.718,-0.502,5.763],[-16.263,1.691,6.390],[-17.773,7.745,10.860],
[-20.083,8.530,0.299],[-7.545,9.134,9.075],[-6.928,-0.502,9.156],
[-13.090,-3.492,5.540],[-8.360,10.744,5.319],[-8.732,10.234,2.194],
[-6.791,7.262,11.501],[-12.629,6.071,2.708],[-11.332,4.025,-1.806],
[-3.944,5.900,9.196],[-7.411,6.851,3.137],[-10.936,3.823,11.441],
[-17.474,5.255,3.817],[-15.656,10.933,5.713],[-16.556,6.945,7.150],
[-10.196,10.534,8.777],[-14.816,9.882,11.210],[-10.363,5.885,0.153],
[-18.145,5.090,10.754],[-12.706,10.406,4.328],[-15.925,11.778,12.908],
[-5.083,-2.377,0.339],[-17.620,5.641,1.149],[-6.106,12.360,-1.455],
[-20.334,9.503,9.844],[-21.473,10.087,6.362],[-3.072,10.800,13.341],
[-12.832,12.429,11.204],[-11.968,15.406,0.711],[-19.963,14.527,4.641],
[-8.875,13.306,-1.319],[-9.738,12.850,1.430],[-5.624,2.287,13.463],
[-18.485,10.033,12.123],[-15.963,-2.716,6.708],[-12.444,-2.921,10.759],
[-8.407,1.729,12.389],[-8.313,6.344,-3.420],[-11.636,1.136,-2.192],
[-16.784,2.615,8.905],[-15.093,2.735,1.599],[-10.936,3.641,1.589],
[-10.753,1.526,13.205],[-10.265,3.631,6.022],[-19.497,2.937,9.101],
[-14.849,1.457,-3.064],[-22.204,-0.606,4.748],[-20.750,1.277,0.112],
[-12.002,3.422,8.429],[-14.687,-1.459,2.773],[-17.799,2.628,1.354],
[-16.225,1.830,-0.787],[-10.005,-0.719,8.614],[-17.331,4.440,-1.235],
[-5.618,13.710,12.608],[-17.955,6.594,13.646],[-12.923,2.297,6.111],
[-17.101,-2.301,2.108],[-14.155,1.807,9.186],[-13.225,7.785,0.048],
[-6.745,6.459,7.914],[-8.927,-0.395,-2.067],[-12.766,8.266,10.420],
[-9.201,2.132,-0.515],[-21.130,6.909,5.187],[-10.652,8.638,-0.121],
[-20.346,6.174,11.745],[-13.519,4.698,12.398],[-11.009,9.874,11.823],
[-4.074,2.883,-1.585],[-3.829,9.866,10.843],[-5.652,10.730,1.338],
[-9.429,6.042,8.521],[-16.399,10.170,3.355],[-20.081,5.727,-1.671],
[-15.914,8.095,1.648],[-16.103,12.133,10.205],[-17.173,13.493,3.517],
[-9.732,8.108,10.163],[-8.672,8.970,-2.042],[-7.566,11.860,12.926],
[-4.358,14.703,-1.845],[-19.574,0.608,7.252],[-8.084,7.585,0.533],
[-21.115,12.166,9.295],[-14.001,10.168,-3.597],[-15.418,9.229,-1.311],
[-17.673,7.446,-0.858],[-12.862,-0.419,9.949],[-10.055,-0.844,0.694],
[-13.648,12.915,0.321],[-12.886,1.258,2.159],[-12.636,1.957,11.408],
[-12.259,-2.803,3.037],[-15.629,11.892,-1.050],[-15.735,-0.710,-1.754],
[-6.237,-0.267,12.515],[-6.874,3.981,6.675],[-17.888,10.488,-0.772],
[-20.905,2.509,-2.492],[-6.728,15.242,7.749],[-18.661,-2.735,5.842],
[-17.187,12.129,1.224],[-19.878,13.612,13.749],[-16.116,12.670,15.358],
[-11.099,11.075,-1.291],[-9.467,10.668,14.884],[-8.149,11.117,-3.512],
[-21.146,6.576,14.872],[-18.106,10.711,-3.653],[-13.895,7.170,13.868],
[-12.055,8.654,-2.578],[-9.605,5.979,12.175],[-19.130,3.821,12.969],
[-14.521,5.101,-0.030],[-15.649,4.175,10.836],[-14.133,4.143,-2.679],
[-11.389,5.898,14.655],[-5.702,8.319,-2.122],[-15.708,1.529,11.414],
[-14.861,-0.457,7.849],[-13.596,5.201,9.745],[-9.059,2.973,14.801],
[-7.138,4.688,12.915],[-16.681,2.937,13.703],[-13.998,-4.367,8.587],
[-11.572,-0.630,14.665],[-10.291,0.271,11.034],[-8.833,-0.572,14.397],
[-18.136,0.637,12.944],[-14.902,-4.350,11.691],[-13.889,-2.688,14.301],
[-15.794,14.833,-1.157],[-22.437,10.164,-1.179],[-22.244,15.760,1.642],
[-22.543,12.965,4.222],[-3.756,13.193,7.494],[-3.861,12.694,10.798],
[-6.804,9.134,15.139],[-5.350,7.232,0.335],[-4.737,4.801,2.175],
[-4.695,8.094,7.575],[-6.455,11.575,10.282],[-3.387,5.562,-0.549],
[-6.173,4.547,-2.291],[-6.507,3.718,3.607],[-23.072,4.986,6.593],
[-4.672,7.786,13.205],[-3.739,4.350,13.579],[-6.807,1.202,-2.960],
[-6.576,1.672,0.775],[-5.358,1.929,8.057],[-6.509,1.822,10.421],
[-21.919,1.332,13.196],[-1.211,-4.055,14.378],[-7.047,-0.859,1.728],
[-19.215,-0.441,4.607],[-21.607,2.659,6.690],[-3.511,-1.992,12.501],
[-4.572,-1.708,-2.173],[-21.767,-0.797,1.705],[-7.394,13.270,3.294],
[-21.102,-1.476,9.850],[-18.979,11.745,15.697],[-17.618,10.232,8.926],
[-12.095,14.329,13.053],[-9.630,14.828,11.943],[-5.623,-2.626,11.030],
[-19.546,14.335,10.164],[-16.644,14.626,9.481],[-11.743,14.973,7.560],
[-10.362,12.968,9.938],[-9.245,15.245,8.749],[-20.098,14.018,7.674],
[-19.730,11.916,4.827],[-17.977,13.401,6.392],[-12.512,-0.405,5.372],
[-5.726,-2.912,6.982],[-5.705,-2.116,3.782],[-20.770,-4.025,4.876],
[-20.761,-2.278,-1.313],[-13.743,-5.474,3.725],[-9.758,-3.095,5.444],
[-8.073,-2.936,-0.477],[-20.611,11.626,0.409],[-18.829,-0.368,-0.871],
[-12.991,-0.966,0.322],[-11.547,13.753,-2.547],[-4.745,13.252,1.225],
[-19.430,-2.426,11.735],[-17.289,-2.619,13.472],[-18.485,15.033,-1.997],
[-9.362,-3.385,14.620],[-5.382,11.908,14.478],[-10.800,26.367,3.726],
[-11.064,23.883,5.329],[-18.551,23.432,8.315],[-7.847,25.648,5.438],
[-17.578,27.601,5.766],[-21.477,23.552,0.870],[-22.105,23.023,3.667],
[-9.333,31.856,5.184],[-13.173,25.764,6.504],[-5.428,18.409,9.713],
[-9.262,20.517,8.366],[-16.099,19.762,2.562],[-8.226,19.171,3.288],
[-20.569,21.697,4.911],[-15.933,22.933,5.042],[-3.367,20.144,9.921],
[-18.134,18.129,7.506],[-12.316,28.038,2.461],[-16.992,25.464,9.229],
[-6.799,17.491,5.278],[-17.946,20.760,5.225],[-18.054,28.170,10.154],
[-20.102,27.992,0.942],[-7.598,27.982,8.604],[-8.190,17.621,7.753],
[-13.645,16.108,5.821],[-10.506,29.552,4.997],[-8.636,27.286,1.862],
[-7.155,25.985,12.757],[-12.543,24.588,3.363],[-11.549,22.053,-1.135],
[-5.031,22.615,9.164],[-9.366,24.122,3.092],[-7.982,22.246,10.300],
[-16.689,25.146,3.538],[-14.395,28.114,5.863],[-15.075,23.616,7.731],
[-10.815,28.898,9.025],[-15.481,27.345,11.064],[-10.423,24.356,0.154],
[-18.365,24.290,11.077],[-12.139,31.168,3.448],[-16.505,29.141,12.683],
[-4.782,16.137,0.708],[-19.389,24.829,1.944],[-4.639,28.971,-1.372],
[-23.424,26.996,9.904],[-19.998,29.407,5.795],[-2.911,29.716,12.184],
[-12.634,28.721,12.001],[-12.137,33.582,1.249],[-18.259,32.275,3.618],
[-9.745,31.559,-1.273],[-9.948,31.721,2.196],[-6.313,21.329,14.238],
[-20.068,27.906,11.741],[-16.804,15.947,7.015],[-12.156,15.108,10.448],
[-8.586,20.844,13.054],[-8.268,24.007,-2.081],[-11.662,17.472,-3.016],
[-16.784,20.722,7.833],[-17.279,23.079,1.878],[-11.557,22.147,1.616],
[-10.829,21.975,12.603],[-9.917,19.800,5.833],[-17.850,20.898,12.190],
[-14.683,19.547,-4.010],[-21.759,17.504,4.294],[-21.384,19.353,-2.832],
[-11.308,22.799,8.582],[-14.887,17.548,3.809],[-19.532,21.477,1.613],
[-16.647,19.261,-0.191],[-11.064,17.395,6.354],[-17.091,21.803,-1.440],
[-7.187,33.428,13.502],[-18.092,24.726,14.283],[-13.818,20.626,5.513],
[-18.019,17.744,3.916],[-14.230,20.882,8.511],[-12.820,25.789,0.837],
[-6.354,24.750,8.377],[-9.818,19.800,-1.246],[-10.381,26.711,12.084],
[-9.072,22.062,1.201],[-18.329,23.871,5.556],[-12.811,28.378,-0.575],
[-21.383,23.539,11.091],[-13.492,22.148,14.078],[-9.946,29.437,11.634],
[-3.666,21.493,-0.049],[-4.866,27.350,12.635],[-6.094,28.477,1.041],
[-9.518,24.677,7.389],[-16.943,30.010,1.112],[-20.315,24.662,-2.167],
[-14.760,26.849,3.431],[-14.537,30.865,10.416],[-14.694,32.190,3.598],
[-11.475,25.841,8.928],[-10.505,27.189,-0.219],[-7.359,30.833,11.498],
[-3.526,31.552,-1.089],[-20.346,20.771,11.315],[-7.879,27.072,-0.729],
[-19.951,28.804,8.433],[-13.958,27.454,-3.070],[-15.236,26.873,-0.841],
[-17.671,25.651,-0.268],[-13.195,18.501,9.930],[-8.992,17.440,0.714],
[-12.945,31.495,-0.177],[-14.172,20.650,0.574],[-12.929,21.208,11.092],
[-12.036,17.109,2.868],[-15.284,29.814,-1.480],[-14.412,17.748,-1.313],
[-6.342,18.679,12.261],[-8.161,22.811,5.699],[-18.735,28.172,-1.412],
[-19.395,21.043,-2.646],[-7.814,31.782,7.627],[-20.496,16.835,6.937],
[-19.673,30.336,2.284],[-21.231,32.319,14.970],[-16.382,30.127,15.171],
[-11.444,29.965,-4.403],[-9.056,28.416,14.248],[-7.966,29.259,-2.921],
[-20.729,25.657,14.025],[-17.142,27.867,-3.414],[-13.358,26.082,13.148],
[-12.440,26.568,-5.122],[-8.691,26.012,10.151],[-20.030,22.950,13.215],
[-15.666,23.653,-0.379],[-15.620,22.038,10.386],[-15.508,24.062,-3.217],
[-10.889,24.840,13.897],[-6.128,25.499,-1.828],[-15.679,20.453,14.523],
[-15.817,17.522,9.231],[-13.732,24.816,10.565],[-8.149,21.346,16.040],
[-8.710,23.658,12.529],[-17.959,21.983,14.853],[-15.438,17.820,11.943],
[-13.396,19.229,14.878],[-10.573,19.209,12.309],[-8.813,18.385,14.351],
[-18.751,18.598,13.443],[-14.974,14.210,13.098],[-13.326,16.331,14.596],
[-14.848,32.564,-2.047],[-21.788,27.012,-2.565],[-21.157,33.773,1.957],
[-20.747,32.280,4.651],[-4.864,31.954,6.705],[-3.629,31.731,9.781],
[-6.383,27.592,15.039],[-3.643,27.361,0.750],[-4.757,23.727,1.850],
[-4.175,26.756,7.166],[-5.340,29.380,10.534],[-3.950,24.949,-0.537],
[-3.880,22.051,-2.759],[-6.626,22.415,3.741],[-20.795,24.676,5.346],
[-5.961,23.934,11.571],[-4.343,22.973,13.550],[-7.627,18.943,-2.695],
[-6.612,20.094,0.985],[-6.598,20.310,7.122],[-5.399,21.147,11.516],
[-22.064,20.797,13.777],[-3.642,15.585,16.398],[-6.778,16.599,2.612],
[-20.204,19.386,6.314],[-22.037,21.089,7.079],[-5.358,16.653,14.096],
[-4.149,17.806,-1.919],[-22.688,18.271,-0.489],[-7.414,30.291,3.706],
[-21.278,16.998,10.487],[-18.961,30.139,12.822],[-17.039,30.519,9.205],
[-11.994,31.529,12.311],[-9.723,33.062,12.015],[-6.366,15.076,10.291],
[-19.480,34.626,10.910],[-17.111,33.384,11.020],[-11.104,32.532,7.116],
[-9.797,31.760,9.560],[-6.783,34.380,7.270],[-19.437,31.638,7.355],
[-15.233,29.653,3.322],[-16.073,30.076,6.080],[-10.900,14.670,4.851],
[-5.661,14.760,5.238],[-4.089,15.991,3.318],[-19.894,16.392,2.733],
[-20.130,16.936,-0.796],[-13.761,14.991,3.351],[-8.514,13.792,5.561],
[-7.702,14.676,1.045],[-20.851,30.805,-2.225],[-18.120,18.406,-2.200],
[-12.370,18.383,0.295],[-11.393,34.008,-1.257],[-5.468,31.867,2.797],
[-21.206,16.005,13.075],[-16.286,15.717,15.319],[-19.031,33.310,-2.489],
[-9.402,15.769,14.586],[-5.531,30.549,13.609],[-10.201,43.349,4.115],
[-11.601,43.007,6.883],[-18.060,41.267,7.402],[-8.367,45.733,4.577],
[-18.124,46.499,5.891],[-22.069,39.839,4.122],[-20.545,41.988,3.431],
[-11.238,51.633,5.961],[-13.118,46.950,7.193],[-4.488,36.274,8.825],
[-9.172,38.272,9.414],[-17.586,39.104,4.840],[-9.748,37.566,1.500],
[-20.191,38.485,3.017],[-13.932,42.120,5.353],[-3.723,39.104,9.235],
[-19.873,36.103,8.149],[-10.499,45.926,2.960],[-15.737,45.143,7.949],
[-6.099,36.372,5.058],[-19.793,39.224,6.731],[-17.582,46.297,9.730],
[-20.236,47.061,1.147],[-7.964,46.751,7.277],[-6.472,37.817,9.974],
[-14.592,33.823,6.312],[-8.981,49.951,4.382],[-8.257,47.574,2.133],
[-8.754,46.680,10.447],[-10.513,43.699,1.405],[-10.195,41.373,-2.884],
[-6.438,44.585,10.601],[-7.481,42.881,1.943],[-8.341,40.556,10.898],
[-16.715,43.753,4.966],[-15.431,46.760,5.664],[-15.499,42.360,7.619],
[-10.431,47.504,6.702],[-15.138,45.689,11.143],[-9.234,42.103,-0.332],
[-18.199,42.830,10.002],[-11.386,49.203,3.353],[-17.017,48.748,13.193],
[-5.771,34.700,-3.718],[-16.228,41.871,0.716],[-4.526,47.276,-1.979],
[-22.297,44.997,9.973],[-20.869,46.686,6.517],[-2.188,48.796,11.565],
[-13.245,47.508,12.207],[-11.260,51.584,1.854],[-17.659,50.787,0.924],
[-8.544,50.173,-1.626],[-9.914,52.088,-0.222],[-6.644,39.670,13.885],
[-20.059,45.620,11.553],[-16.670,35.376,7.867],[-11.629,33.765,10.021],
[-8.248,37.868,11.953],[-7.717,43.257,-3.229],[-12.846,36.196,-2.023],
[-17.379,38.400,7.985],[-15.913,41.779,3.371],[-10.937,40.950,1.578],
[-10.356,40.014,13.345],[-9.020,39.852,3.745],[-19.314,39.759,9.381],
[-16.169,37.752,-2.655],[-23.067,37.102,3.985],[-21.531,38.875,-2.672],
[-10.770,40.989,8.735],[-13.355,35.629,7.775],[-18.779,40.811,1.687],
[-16.200,38.927,0.101],[-10.155,36.552,7.721],[-15.880,40.435,-2.075],
[-6.891,52.335,12.915],[-17.958,46.294,13.151],[-11.950,40.360,6.224],
[-17.923,37.704,1.935],[-13.263,39.398,8.783],[-13.665,43.362,-0.511],
[-6.260,43.658,7.817],[-10.246,36.607,-1.516],[-11.200,45.169,11.552],
[-9.478,39.007,-0.734],[-19.476,43.204,6.123],[-11.120,45.370,-0.512],
[-20.446,41.809,11.085],[-14.422,41.339,13.351],[-10.242,49.012,10.875],
[-3.388,40.395,-0.139],[-4.165,44.969,12.299],[-5.201,47.640,1.326],
[-9.047,44.061,6.938],[-17.276,47.642,2.199],[-19.714,43.249,-2.146],
[-13.933,44.872,4.480],[-14.006,49.145,8.759],[-15.437,49.523,3.313],
[-11.197,45.712,8.734],[-7.694,46.091,-2.296],[-7.698,49.586,11.743],
[-3.374,50.289,-1.314],[-20.206,37.623,10.519],[-8.524,44.792,0.052],
[-20.046,47.221,9.037],[-14.080,46.876,-4.008],[-13.632,47.121,-0.777],
[-17.931,43.615,2.234],[-12.172,36.525,9.977],[-9.611,34.999,0.817],
[-11.765,49.096,-1.012],[-12.513,38.576,2.278],[-13.605,38.401,11.438],
[-11.876,36.328,3.288],[-15.848,48.841,-1.386],[-15.924,36.290,0.511],
[-4.058,37.091,11.278],[-7.765,40.437,6.677],[-17.153,46.860,-0.415],
[-18.821,40.223,-1.787],[-6.210,49.849,8.033],[-20.915,33.891,6.822],
[-19.333,47.352,3.603],[-20.819,50.940,14.277],[-15.372,50.230,14.838],
[-12.629,49.491,-3.558],[-10.598,47.441,13.106],[-6.213,49.196,-1.992],
[-20.015,44.089,13.734],[-17.116,46.535,-4.135],[-12.582,45.571,13.711],
[-10.827,45.819,-3.212],[-9.263,43.317,11.069],[-19.072,41.361,14.164],
[-16.781,43.572,-1.171],[-15.565,40.580,10.932],[-13.652,42.556,-3.143],
[-11.047,43.059,13.313],[-4.252,44.761,-2.993],[-15.746,36.887,11.812],
[-15.124,37.415,9.224],[-14.143,42.851,9.951],[-9.216,39.354,15.631],
[-7.713,42.333,13.015],[-16.572,40.741,15.308],[-14.953,34.495,9.651],
[-13.698,38.373,14.214],[-10.978,37.272,12.626],[-8.889,36.282,14.191],
[-18.525,38.676,13.946],[-14.581,33.759,12.355],[-12.469,33.818,13.742],
[-14.948,51.603,-0.508],[-21.185,45.333,-0.899],[-22.168,51.803,2.531],
[-20.953,50.163,4.654],[-3.873,50.936,7.508],[-4.308,49.606,10.148],
[-5.915,46.238,13.925],[-4.253,44.441,-0.284],[-4.664,42.945,2.373],
[-5.021,45.830,6.777],[-5.921,47.388,9.115],[-4.687,42.170,-2.235],
[-4.911,39.675,-2.826],[-6.192,41.118,3.757],[-22.359,43.819,6.631],
[-5.063,42.503,13.537],[-4.740,40.421,11.706],[-6.527,37.560,-2.819],
[-6.373,38.287,0.180],[-5.670,38.164,7.014],[-6.001,41.336,9.188],
[-21.076,39.553,12.858],[-2.641,34.390,15.338],[-6.560,34.575,1.110],
[-19.534,36.868,5.123],[-21.974,40.700,6.982],[-4.791,35.121,13.012],
[-4.029,37.327,-1.872],[-21.537,35.921,-0.157],[-7.197,50.401,2.003],
[-21.187,32.655,9.384],[-20.170,48.272,13.037],[-16.760,48.590,8.183],
[-11.534,51.780,11.933],[-8.445,53.792,11.089],[-5.876,33.731,10.152],
[-18.619,51.875,9.637],[-16.248,52.247,7.766],[-13.484,51.822,9.241],
[-10.180,49.901,8.045],[-9.064,51.485,9.995],[-19.168,50.995,6.589],
[-18.637,50.493,3.599],[-16.623,51.596,4.837],[-12.002,34.810,5.450],
[-7.707,33.414,3.482],[-4.733,34.542,3.715],[-18.532,35.095,3.298],
[-20.610,32.576,-0.330],[-14.310,34.483,2.537],[-9.292,34.665,5.820],
[-7.814,35.569,-1.292],[-21.101,49.404,-0.413],[-18.893,35.957,-1.040],
[-13.146,36.311,0.737],[-12.562,52.825,-1.200],[-4.534,50.133,1.121],
[-18.199,36.415,12.462],[-17.014,33.742,13.619],[-18.905,49.720,-2.327],
[-10.000,33.943,14.733],[-6.003,49.166,14.195],[-4.050,54.420,7.296],
[-8.221,59.422,6.893],[-17.011,56.596,4.400],[-9.277,57.262,2.346],
[-20.641,59.089,4.928],[-4.091,57.092,9.472],[-18.051,55.140,9.840],
[-8.342,56.110,4.705],[-17.612,56.099,7.293],[-7.001,55.825,9.468],
[-13.948,52.226,6.025],[-5.968,51.672,-1.012],[-6.003,58.705,13.889],
[-16.005,54.353,5.510],[-10.793,53.505,9.487],[-7.457,58.492,11.557],
[-11.456,56.137,-2.097],[-16.696,58.421,8.104],[-15.817,58.930,3.023],
[-11.789,58.936,4.583],[-18.449,59.193,10.006],[-15.569,56.028,-3.739],
[-21.775,56.816,3.366],[-21.145,56.862,-1.789],[-14.437,56.078,3.614],
[-18.720,56.852,2.434],[-16.774,57.173,0.516],[-9.890,54.827,7.091],
[-17.144,57.919,-4.887],[-12.487,57.097,6.552],[-16.695,53.400,2.909],
[-13.803,57.219,8.622],[-7.807,56.396,-2.303],[-9.572,58.349,-0.391],
[-3.403,58.893,-1.336],[-20.036,55.614,11.673],[-9.775,58.141,8.993],
[-9.326,55.743,0.043],[-14.148,57.819,1.388],[-12.372,56.902,11.177],
[-10.973,54.258,1.608],[-15.184,54.327,-1.771],[-7.903,55.717,12.937],
[-18.386,58.584,-2.435],[-19.267,53.912,7.338],[-14.831,56.629,12.740],
[-14.294,54.945,10.336],[-10.391,58.178,15.924],[-17.472,58.314,14.366],
[-15.951,52.618,10.232],[-12.085,57.051,14.021],[-10.695,58.845,12.007],
[-9.226,53.820,16.874],[-22.210,56.779,12.858],[-14.599,51.142,12.386],
[-14.528,52.958,15.150],[-5.480,58.989,-3.056],[-4.890,56.501,-2.392],
[-6.703,57.061,0.119],[-5.972,57.881,6.162],[-6.071,-59.914,8.820],
[-21.365,57.932,10.550],[-4.794,51.409,14.934],[-5.580,54.099,1.745],
[-19.733,56.385,5.795],[-20.029,58.388,7.816],[-5.323,54.498,12.338],
[-3.002,54.637,-2.588],[-21.499,56.242,0.766],[-21.098,52.746,8.954],
[-5.010,52.514,10.411],[-12.347,54.646,5.654],[-6.821,51.087,5.574],
[-5.360,55.282,5.104],[-19.609,53.094,2.965],[-20.240,53.575,-0.132],
[-13.944,52.727,2.916],[-8.560,52.904,5.485],[-8.086,53.243,1.316],
[-18.080,52.549,-1.950],[-14.071,54.566,0.784],[-17.834,53.636,12.364],
[-16.605,54.699,15.126],[-7.900,51.416,15.368],[-8.865,-50.523,22.908],
[-13.126,-49.886,25.329],[-20.369,-51.165,28.739],[-6.368,-47.829,22.342],
[-18.056,-47.910,23.749],[-21.689,-51.900,21.065],[-21.632,-50.509,23.180],
[-12.407,-43.666,23.636],[-12.910,-46.168,25.090],[-4.632,-56.941,27.413],
[-8.611,-54.838,25.013],[-17.514,-54.973,24.476],[-8.255,-54.860,22.126],
[-19.718,-53.413,24.142],[-14.809,-51.646,22.746],[-4.674,-54.303,28.786],
[-18.850,-57.001,26.385],[-9.253,-47.599,22.091],[-15.391,-50.408,29.679],
[-4.725,-58.229,23.295],[-18.331,-53.983,27.126],[-16.766,-47.431,28.154],
[-21.838,-45.331,19.468],[-5.674,-45.984,25.882],[-7.702,-56.624,29.699],
[-14.067,-59.298,25.815],[-7.320,-45.592,23.401],[-7.287,-46.379,20.154],
[-6.054,-47.533,30.527],[-11.247,-50.052,20.429],[-10.441,-52.373,15.851],
[-4.628,-51.650,27.355],[-7.540,-49.815,20.822],[-9.270,-52.891,29.961],
[-16.441,-49.143,22.072],[-15.575,-47.150,25.302],[-15.879,-49.926,24.902],
[-5.939,-43.130,27.440],[-14.286,-47.301,29.854],[-9.002,-51.400,17.864],
[-18.197,-49.797,29.483],[-12.524,-44.270,20.868],[-17.797,-45.706,32.573],
[-7.060,-58.787,18.066],[-18.179,-50.597,20.209],[-3.743,-45.860,16.248],
[-20.326,-48.509,27.711],[-22.336,-44.083,27.051],[-2.600,-43.908,30.183],
[-14.827,-43.263,32.129],[-11.559,-40.612,18.376],[-17.848,-43.348,21.099],
[-7.510,-43.353,18.677],[-10.134,-42.784,18.429],[-5.749,-54.174,33.634],
[-19.631,-47.556,31.170],[-15.652,-56.559,26.089],[-10.338,-58.552,28.825],
[-6.849,-54.186,30.812],[-7.601,-49.897,15.323],[-11.760,-55.628,17.177],
[-15.300,-53.542,25.708],[-15.808,-53.252,20.702],[-9.386,-52.419,20.494],
[-10.173,-51.918,32.337],[-10.119,-52.697,23.769],[-16.860,-51.838,27.886],
[-17.496,-55.591,14.652],[-22.131,-55.633,21.741],[-21.651,-55.317,17.647],
[-11.330,-51.134,27.282],[-13.434,-56.077,21.846],[-19.807,-53.957,20.797],
[-15.989,-53.978,17.857],[-8.970,-57.075,26.771],[-17.626,-52.074,16.746],
[-7.733,-40.236,31.179],[-16.738,-48.630,31.762],[-12.537,-52.717,25.092],
[-17.776,-56.264,22.078],[-12.120,-55.471,26.883],[-12.770,-48.294,18.854],
[-6.373,-49.678,26.596],[-9.228,-54.995,17.129],[-10.963,-48.049,29.262],
[-7.330,-55.730,18.892],[-18.675,-50.969,23.060],[-11.037,-48.231,16.933],
[-22.730,-51.903,30.162],[-15.086,-51.732,32.127],[-10.308,-44.610,30.835],
[-4.491,-53.591,18.565],[-3.306,-46.146,29.143],[-6.181,-46.273,17.262],
[-9.537,-49.575,25.306],[-15.095,-44.678,20.522],[-19.795,-49.936,18.050],
[-14.023,-47.213,21.016],[-14.465,-44.286,29.667],[-15.597,-41.490,20.537],
[-10.777,-47.323,26.571],[-8.675,-47.354,16.666],[-7.554,-42.832,31.919],
[-4.985,-43.597,17.397],[-19.657,-56.973,29.197],[-9.401,-48.350,19.542],
[-19.555,-44.917,27.619],[-15.731,-45.621,14.495],[-13.686,-45.519,16.126],
[-18.135,-47.657,18.668],[-9.420,-54.523,27.874],[-10.447,-56.732,19.434],
[-11.231,-45.313,17.098],[-12.211,-54.577,19.939],[-14.233,-53.639,30.452],
[-10.695,-56.443,22.116],[-15.983,-42.665,15.015],[-16.247,-56.582,17.258],
[-4.390,-55.016,31.415],[-7.427,-52.331,24.959],[-17.994,-45.798,16.473],
[-20.073,-53.395,17.013],[-6.722,-43.527,24.751],[-18.150,-58.036,23.918],
[-18.953,-46.150,21.428],[-21.570,-44.670,32.850],[-15.913,-44.387,34.343],
[-12.045,-46.529,14.327],[-10.341,-47.546,31.928],[-8.414,-44.989,15.127],
[-22.019,-49.614,32.230],[-16.070,-47.739,16.516],[-13.444,-49.595,31.450],
[-12.851,-49.297,15.284],[-7.789,-49.108,32.504],[-19.611,-51.069,31.747],
[-14.215,-51.094,18.408],[-13.727,-53.146,27.886],[-14.844,-51.947,15.599],
[-10.474,-49.371,33.977],[-5.158,-48.874,16.454],[-17.577,-54.542,30.290],
[-15.776,-56.679,30.621],[-12.265,-51.714,30.148],[-8.261,-53.590,33.212],
[-7.248,-51.396,30.816],[-17.954,-53.434,32.701],[-15.184,-58.186,28.113],
[-13.291,-54.611,32.647],[-10.460,-55.822,31.136],[-8.626,-56.609,33.349],
[-17.291,-56.231,33.242],[-12.375,-59.813,31.418],[-14.311,-58.170,35.241],
[-14.178,-41.182,16.732],[-20.163,-47.596,16.121],[-20.952,-40.993,20.453],
[-22.046,-42.156,22.520],[-4.504,-41.899,25.368],[-5.227,-42.191,30.421],
[-4.809,-46.427,33.352],[-4.975,-48.163,19.447],[-4.853,-49.878,21.438],
[-4.647,-49.615,29.250],[-7.640,-45.164,29.424],[-3.458,-50.221,18.154],
[-4.951,-52.612,16.124],[-7.058,-52.480,21.879],[-22.914,-49.268,25.086],
[-5.278,-50.002,32.090],[-2.701,-51.737,32.412],[-6.687,-56.658,15.507],
[-5.801,-55.057,21.054],[-6.107,-54.732,26.350],[-7.050,-52.596,28.000],
[-20.521,-55.301,31.797],[-3.019,-58.607,34.268],[-5.957,-57.984,20.426],
[-20.591,-56.404,24.127],[-21.030,-51.394,25.839],[-4.590,-57.547,32.390],
[-4.164,-57.737,18.266],[-17.843,-57.706,19.227],[-8.470,-42.896,22.259],
[-20.788,58.500,28.733],[-19.555,-44.718,30.651],[-17.000,-44.829,30.008],
[-11.905,-42.691,32.145],[-10.021,-41.976,29.972],[-4.960,-56.976,29.880],
[-18.528,-41.538,27.957],[-15.900,-43.122,27.367],[-11.660,-41.618,26.090],
[-8.990,-43.166,27.491],[-6.848,-40.450,28.066],[-20.813,-44.528,23.076],
[-17.973,-44.830,24.856],[-16.000,-43.944,23.018],[-13.092,-56.447,24.513],
[-6.829,59.262,22.963],[-2.148,-59.114,23.490],[-20.058,-58.002,21.599],
[-20.108,59.367,19.103],[-13.162,-59.046,22.257],[-9.675,-59.516,23.097],
[-8.502,-58.099,20.991],[-19.303,-42.859,18.683],[-19.005,-57.019,16.479],
[-13.189,-57.100,19.228],[-11.812,-42.627,16.270],[-4.927,-42.915,20.072],
[-19.604,58.393,30.981],[-17.779,-58.432,31.264],[-18.649,-42.224,16.117],
[-8.157,-58.641,31.523],[-4.499,-42.306,32.984],[-10.239,-30.095,22.912],
[-10.623,-32.150,25.465],[-19.898,-31.381,28.954],[-6.939,-29.421,24.500],
[-17.274,-28.379,24.350],[-20.329,-32.506,19.609],[-21.326,-32.112,22.186],
[-10.366,-22.600,24.916],[-13.264,-31.009,27.490],[-4.991,-36.965,27.783],
[-7.537,-36.164,26.813],[-17.660,-38.195,22.912],[-7.193,-36.452,21.755],
[-20.010,-34.247,25.152],[-15.647,-32.005,23.301],[-3.145,-35.604,29.401],
[-17.830,-37.934,26.622],[-11.576,-27.748,21.149],[-16.690,-30.511,26.073],
[-8.197,-38.273,23.766],[-17.429,-35.300,24.050],[-18.585,-29.240,27.893],
[-19.729,-26.987,19.777],[-7.386,-27.281,25.869],[-7.017,-38.530,26.125],
[-14.551,-42.394,24.910],[-8.052,-27.122,23.295],[-8.434,-25.953,19.370],
[-7.421,-28.811,30.675],[-11.560,-30.914,20.564],[-10.408,-32.949,15.773],
[-3.913,-32.355,27.551],[-8.186,-31.649,21.125],[-8.295,-33.844,29.652],
[-18.231,-31.302,22.672],[-14.164,-29.727,24.451],[-15.591,-33.302,25.706],
[-11.063,-27.456,27.160],[-16.355,-28.551,29.164],[-9.609,-32.626,18.890],
[-17.421,-31.805,30.013],[-11.097,-25.254,22.753],[-16.941,-26.508,30.780],
[-4.857,-40.664,18.257],[-17.282,-32.155,20.048],[-3.788,-26.048,17.153],
[-21.938,-29.788,28.250],[-21.987,-27.045,27.614],[-2.948,-26.772,32.101],
[-12.471,-28.140,29.893],[-12.302,-22.484,19.709],[-18.139,-25.854,21.361],
[-6.110,-24.396,17.799],[-11.061,-24.940,20.054],[-6.399,-33.385,33.473],
[-20.372,-28.069,29.807],[-16.410,-40.177,25.609],[-11.678,-40.245,30.808],
[-8.417,-37.680,30.303],[-7.589,-32.423,16.402],[-14.127,-36.257,17.577],
[-16.726,-35.561,27.212],[-15.613,-34.831,20.721],[-13.323,-33.553,18.127],
[-10.081,-35.666,30.605],[-9.097,-34.221,21.522],[-19.258,-33.933,29.223],
[-17.924,-36.764,15.248],[-23.158,-37.746,22.695],[-20.183,-37.268,17.158],
[-10.484,-34.714,27.048],[-14.318,-38.923,24.080],[-18.561,-34.682,20.818],
[-14.782,-37.426,20.214],[-9.929,-37.505,27.661],[-15.966,-34.862,16.200],
[-6.950,-22.622,31.892],[-16.976,-29.726,32.176],[-12.814,-34.450,24.836],
[-17.253,-38.939,20.056],[-13.954,-35.174,27.193],[-14.026,-29.936,19.293],
[-6.331,-31.154,26.631],[-10.679,-37.540,16.977],[-12.047,-31.044,30.254],
[-8.661,-35.048,17.297],[-19.464,-31.462,25.623],[-10.896,-28.296,18.473],
[-21.897,-30.761,31.084],[-13.695,-34.912,32.615],[-9.635,-27.683,29.659],
[-3.575,-35.039,18.353],[-4.261,-29.085,30.594],[-5.420,-26.745,20.640],
[-8.224,-31.809,23.900],[-15.567,-27.196,21.470],[-20.530,-30.797,16.632],
[-15.181,-29.702,21.846],[-12.730,-24.902,27.615],[-16.393,-23.545,21.232],
[-10.745,-29.951,28.286],[-9.499,-28.452,16.229],[-7.195,-25.076,30.685],
[-5.272,-22.501,15.675],[-19.451,-36.002,27.786],[-7.947,-29.448,18.696],
[-20.387,-25.393,29.600],[-14.796,-25.674,15.520],[-14.528,-28.651,16.917],
[-19.069,-29.904,18.782],[-12.611,-38.243,28.726],[-9.465,-37.182,19.576],
[-12.724,-26.236,18.220],[-12.953,-35.235,20.076],[-12.216,-34.690,29.564],
[-11.749,-37.603,20.705],[-15.766,-25.423,18.230],[-15.718,-38.325,17.367],
[-6.132,-35.761,30.613],[-7.105,-33.969,25.358],[-17.344,-27.337,16.790],
[-19.135,-34.428,17.204],[-7.550,-24.102,25.755],[-19.027,-40.118,24.859],
[-20.199,-27.602,22.276],[-21.054,-23.294,31.362],[-14.393,-25.240,32.666],
[-11.061,-26.350,16.247],[-10.173,-26.957,32.207],[-8.224,-25.918,16.236],
[-20.138,-29.869,35.366],[-16.907,-28.594,14.298],[-13.223,-29.463,32.390],
[-12.127,-29.516,14.952],[-8.376,-30.985,29.474],[-18.446,-33.668,32.704],
[-16.880,-30.500,17.169],[-15.664,-33.707,30.386],[-13.823,-33.307,15.509],
[-10.846,-31.575,32.916],[-4.536,-30.591,15.111],[-14.684,-36.008,29.593],
[-14.935,-37.821,27.300],[-14.751,-30.770,29.735],[-8.750,-34.683,32.618],
[-8.626,-31.850,31.841],[-17.705,-36.143,30.457],[-13.900,-40.847,27.440],
[-15.050,-37.373,32.465],[-12.657,-37.198,31.168],[-9.843,-38.662,32.739],
[-20.833,-36.920,33.771],[-14.837,-41.650,30.104],[-13.951,-40.367,32.818],
[-13.598,-21.507,17.488],[-20.080,-27.772,17.289],[-21.503,-22.643,20.657],
[-20.793,-24.848,22.243],[-3.877,-24.885,25.311],[-4.895,-24.979,27.604],
[-5.681,-27.461,32.220],[-4.710,-28.959,17.856],[-4.858,-30.909,21.368],
[-5.513,-29.117,28.129],[-6.774,-26.838,28.890],[-4.388,-31.563,17.973],
[-4.081,-34.766,15.346],[-6.574,-33.642,22.376],[-23.109,-30.376,24.697],
[-6.012,-30.708,32.175],[-3.653,-32.459,33.108],[-7.385,-37.412,16.142],
[-6.742,-36.824,18.641],[-5.407,-36.858,24.721],[-6.107,-33.930,27.940],
[-21.751,-34.586,32.334],[-3.181,-39.504,34.689],[-5.280,-38.417,20.863],
[-20.162,-37.205,24.335],[-22.996,-33.066,25.247],[-3.227,-39.159,31.846],
[-3.342,-38.144,15.364],[-22.296,-37.548,19.711],[-6.592,-24.425,23.071],
[-20.355,-38.885,28.689],[-19.925,-26.091,32.696],[-16.612,-25.909,27.610],
[-11.432,-24.413,31.620],[-8.967,-24.211,32.785],[-5.592,-38.571,30.266],
[-19.409,-22.538,27.923],[-15.703,-23.705,28.573],[-11.860,-21.157,26.839],
[-9.977,-24.832,27.199],[-7.555,-22.251,28.269],[-21.010,-24.464,27.097],
[-17.168,-25.641,24.072],[-16.196,-23.543,25.144],[-10.932,-37.983,23.715],
[-6.874,-40.889,24.787],[-6.445,-41.058,21.518],[-18.590,-40.887,22.018],
[-20.767,-40.607,17.604],[-13.678,-41.335,22.361],[-9.575,-39.950,25.583],
[-8.303,-39.877,18.657],[-19.864,-24.591,18.569],[-20.097,-39.126,15.301],
[-13.904,-39.702,19.316],[-10.107,-23.992,17.691],[-4.633,-24.072,20.146],
[-19.951,-41.584,31.199],[-17.456,-41.205,32.208],[-19.274,-23.294,16.194],
[-11.248,-40.814,34.272],[-3.800,-23.724,31.406],[-10.331,-10.595,21.653],
[-11.424,-12.901,24.656],[-18.540,-13.335,26.036],[-7.041,-10.562,23.141],
[-17.921,-8.864,25.947],[-21.812,-15.365,20.898],[-22.021,-12.940,23.868],
[-11.335,-5.385,25.070],[-13.111,-10.558,26.273],[-5.207,-19.626,27.698],
[-8.681,-16.500,27.261],[-15.635,-17.667,22.080],[-8.757,-17.941,21.979],
[-20.356,-18.178,23.058],[-15.135,-15.856,23.939],[-3.138,-17.194,29.202],
[-18.441,-18.935,26.238],[-12.367,-8.796,21.805],[-15.174,-11.470,27.962],
[-6.189,-19.879,22.638],[-18.049,-16.656,24.492],[-18.192,-11.910,28.212],
[-20.675,-9.275,19.634],[-7.935,-9.469,25.542],[-8.130,-19.288,28.040],
[-14.351,-21.794,25.916],[-9.182,-7.004,23.797],[-9.605,-7.941,21.022],
[-7.581,-10.247,30.000],[-11.034,-12.462,17.928],[-12.233,-14.654,15.682],
[-3.509,-12.702,27.719],[-8.115,-12.480,21.028],[-10.316,-15.730,30.000],
[-16.652,-12.750,24.004],[-15.487,-9.922,25.352],[-15.761,-14.226,25.941],
[-11.139,-9.094,27.289],[-16.470,-9.995,29.823],[-8.423,-14.378,17.507],
[-17.664,-13.915,30.613],[-12.361,-6.078,22.653],[-16.708,-7.308,30.443],
[-5.201,-20.308,17.345],[-17.966,-13.355,20.670],[-6.257,-7.076,16.326],
[-21.330,-10.271,28.455],[-20.976,-9.185,25.947],[-2.580,-7.279,30.207],
[-12.766,-7.389,30.468],[-14.160,-4.973,19.612],[-18.425,-4.021,22.650],
[-9.489,-5.811,18.500],[-9.409,-4.227,21.193],[-4.762,-15.598,30.678],
[-19.073,-8.898,30.352],[-16.951,-21.544,26.852],[-11.326,-22.753,29.444],
[-8.515,-17.845,30.523],[-7.863,-12.029,15.681],[-12.480,-16.860,18.019],
[-17.548,-16.626,27.617],[-16.243,-15.198,21.363],[-10.443,-15.828,21.391],
[-11.647,-15.123,32.322],[-11.082,-18.527,23.753],[-19.858,-15.414,27.292],
[-14.854,-18.497,15.097],[-22.208,-20.172,22.206],[-20.971,-18.390,16.481],
[-10.733,-14.937,26.923],[-15.499,-19.240,24.300],[-19.379,-15.334,22.045],
[-18.161,-18.227,20.277],[-10.262,-18.887,26.331],[-16.632,-17.000,18.101],
[-6.655,-4.096,31.048],[-17.601,-11.181,32.352],[-11.655,-15.908,24.087],
[-17.085,-20.260,22.093],[-14.810,-16.759,27.329],[-13.513,-11.582,18.475],
[-6.535,-12.495,26.835],[-8.962,-18.799,17.907],[-11.490,-9.446,30.826],
[-7.261,-16.592,18.151],[-19.237,-11.090,24.784],[-11.734,-9.655,19.132],
[-23.090,-13.608,31.080],[-14.232,-13.830,32.624],[-8.739,-8.362,27.969],
[-4.074,-15.761,18.658],[-4.891,-8.237,31.129],[-6.896,-5.306,18.185],
[-9.055,-12.153,23.738],[-17.022,-9.840,23.029],[-18.423,-13.908,17.383],
[-14.859,-11.280,21.754],[-13.366,-6.875,27.531],[-16.377,-5.475,21.046],
[-9.462,-12.162,27.142],[-8.779,-10.909,18.046],[-7.178,-7.321,29.956],
[-3.685,-3.676,15.405],[-19.804,-17.901,28.643],[-7.911,-9.979,20.480],
[-19.595,-7.527,27.717],[-14.706,-8.532,14.792],[-15.767,-9.289,17.095],
[-19.219,-11.680,19.060],[-12.524,-18.817,28.558],[-9.998,-21.216,19.480],
[-14.042,-7.573,17.964],[-12.929,-17.864,21.027],[-13.507,-16.053,30.329],
[-11.504,-20.700,21.757],[-16.241,-6.730,16.577],[-14.734,-19.115,17.874],
[-6.875,-19.944,31.684],[-7.454,-15.103,25.381],[-18.900,-7.602,16.587],
[-19.023,-16.501,16.629],[-7.802,-4.965,25.493],[-19.324,-24.055,24.402],
[-18.758,-7.604,21.074],[-20.951,-4.546,32.270],[-14.263,-5.762,34.481],
[-11.138,-7.528,16.672],[-9.748,-8.080,32.233],[-8.978,-6.353,15.572],
[-20.323,-11.843,32.660],[-17.703,-12.236,15.552],[-15.436,-12.232,30.874],
[-12.415,-10.323,14.515],[-9.229,-12.741,30.215],[-19.330,-14.637,32.770],
[-16.085,-13.658,18.779],[-16.082,-14.205,28.699],[-14.678,-13.580,15.821],
[-10.992,-12.296,34.150],[-3.594,-13.233,16.004],[-15.409,-17.537,31.527],
[-15.862,-19.486,27.829],[-13.366,-14.416,27.176],[-9.677,-16.348,33.687],
[-7.620,-15.126,30.624],[-17.284,-15.914,34.276],[-13.980,-21.204,28.648],
[-12.997,-18.310,33.683],[-11.187,-18.195,31.068],[-8.704,-18.880,33.443],
[-18.779,-16.968,31.016],[-14.090,-22.373,31.058],[-13.460,-21.545,33.541],
[-15.264,-3.890,16.402],[-20.003,-10.451,16.543],[-21.123,-4.354,20.776],
[-20.770,-6.217,22.433],[-2.917,-4.006,26.930],[-5.052,-5.885,29.628],
[-7.097,-8.858,32.499],[-3.807,-10.443,18.941],[-5.344,-12.204,21.316],
[-4.509,-10.149,26.986],[-6.225,-7.439,26.625],[-4.852,-12.943,18.643],
[-5.425,-16.727,15.923],[-6.485,-14.789,20.632],[-22.652,-11.613,26.027],
[-6.275,-12.992,29.497],[-2.615,-14.490,33.643],[-7.294,-18.868,15.991],
[-6.286,-17.637,21.001],[-6.551,-17.776,25.807],[-6.131,-15.685,27.946],
[-21.763,-17.964,30.627],[-2.017,-22.550,33.582],[-5.420,-20.154,20.069],
[-21.550,-20.361,25.006],[-22.375,-14.861,26.067],[-4.249,-20.933,32.457],
[-4.517,-19.245,15.085],[-23.012,-19.315,19.523],[-7.322,-6.579,21.444],
[-21.265,-20.484,29.087],[-19.848,-6.836,31.787],[-16.102,-7.485,27.556],
[-11.877,-6.591,32.779],[-9.467,-5.669,30.813],[-5.284,-21.650,29.764],
[-21.423,-3.374,29.724],[-16.982,-4.271,28.987],[-11.941,-3.244,26.574],
[-10.461,-5.929,28.124],[-7.776,-3.418,27.854],[-20.484,-6.179,25.305],
[-16.973,-7.123,23.271],[-14.098,-6.010,24.995],[-12.796,-20.695,23.958],
[-5.142,-22.179,25.338],[-4.830,-22.066,22.401],[-19.416,-22.198,22.478],
[-20.610,-21.549,18.162],[-14.179,-22.781,22.325],[-7.812,-21.828,24.019],
[-7.457,-22.001,18.401],[-20.601,-6.467,18.977],[-18.581,-20.336,16.048],
[-12.028,-19.482,18.687],[-12.368,-4.975,17.680],[-3.340,-6.587,17.680],
[-18.114,-21.468,31.038],[-16.323,-22.634,32.830],[-18.104,-4.750,15.165],
[-9.895,-21.741,33.706],[-6.021,-6.241,33.321],[-9.861,6.730,21.193],
[-11.973,5.927,26.198],[-18.376,5.336,26.876],[-7.202,8.540,21.613],
[-18.012,8.948,23.668],[-20.205,3.357,20.148],[-20.522,5.403,22.362],
[-11.795,13.583,24.485],[-13.358,9.013,26.553],[-4.390,-0.172,24.790],
[-10.141,3.685,26.339],[-16.969,0.845,22.849],[-10.198,3.348,21.096],
[-20.401,1.748,24.827],[-14.250,6.075,23.634],[-5.759,0.804,30.379],
[-18.906,-1.024,27.189],[-11.561,8.760,20.785],[-16.103,6.963,28.747],
[-6.697,-0.180,22.960],[-17.788,1.843,25.865],[-18.285,8.631,28.196],
[-21.107,8.173,18.610],[-7.175,9.201,25.824],[-8.015,-0.977,26.320],
[-15.092,-3.042,26.013],[-8.544,10.393,23.761],[-9.103,10.400,20.989],
[-6.521,8.873,30.381],[-11.533,4.954,19.239],[-12.232,2.838,15.012],
[-3.916,3.528,28.012],[-7.114,5.950,21.161],[-8.356,4.531,28.135],
[-16.909,5.818,22.641],[-15.498,10.288,24.950],[-16.298,5.123,25.159],
[-10.927,11.390,25.818],[-16.402,9.517,30.218],[-8.979,4.034,18.088],
[-17.392,4.949,29.630],[-10.945,12.139,22.307],[-15.891,11.536,31.776],
[-5.678,-3.016,18.278],[-16.415,4.245,19.673],[-3.410,10.987,15.971],
[-20.944,8.890,28.271],[-20.719,11.132,25.768],[-4.021,10.952,29.900],
[-13.850,10.466,29.115],[-11.319,14.400,20.550],[-18.515,12.593,21.426],
[-9.178,13.693,17.421],[-8.821,12.984,19.857],[-6.708,3.693,30.595],
[-19.930,9.372,30.863],[-17.274,-4.709,25.261],[-11.948,-2.010,29.015],
[-8.585,0.679,29.813],[-8.069,5.239,15.748],[-11.799,-0.189,17.362],
[-15.257,2.269,26.470],[-15.568,3.422,22.148],[-10.761,2.136,18.692],
[-12.589,2.808,33.773],[-10.176,2.420,23.898],[-19.078,2.569,29.641],
[-14.767,0.107,15.712],[-20.428,-1.296,23.670],[-20.701,0.844,19.125],
[-11.648,4.559,28.373],[-14.067,1.021,22.339],[-19.263,2.376,22.510],
[-15.932,-0.295,20.327],[-10.388,0.055,25.593],[-16.605,1.694,17.861],
[-6.709,14.001,31.845],[-18.077,7.177,31.085],[-13.128,3.093,23.881],
[-18.157,1.291,20.117],[-13.419,2.389,29.459],[-13.106,7.239,18.953],
[-7.224,6.677,26.698],[-9.539,-1.726,16.591],[-11.410,8.815,30.410],
[-8.801,0.944,17.298],[-18.969,6.447,24.217],[-11.403,6.618,17.133],
[-20.214,5.215,28.684],[-15.011,2.847,32.011],[-9.203,10.805,29.811],
[-3.557,3.456,18.504],[-3.131,7.459,32.176],[-6.599,11.122,19.996],
[-9.445,6.439,24.711],[-15.864,10.442,22.101],[-20.249,4.701,17.199],
[-14.267,8.121,21.584],[-12.639,13.056,27.229],[-16.047,12.612,20.542],
[-9.005,8.180,27.872],[-9.251,8.332,16.195],[-7.239,12.771,29.186],
[-5.222,13.273,17.167],[-20.614,-0.218,29.315],[-8.947,6.798,18.454],
[-19.138,11.215,27.977],[-15.083,9.941,15.053],[-15.718,9.553,17.609],
[-17.536,6.414,18.475],[-10.987,1.003,28.661],[-8.684,-0.234,20.039],
[-12.017,9.928,18.206],[-11.900,0.931,20.795],[-12.482,4.273,31.191],
[-11.117,-2.020,21.493],[-14.337,11.612,18.504],[-16.062,-1.622,17.818],
[-5.975,-1.333,28.379],[-7.417,4.884,23.659],[-18.376,8.984,18.320],
[-18.387,2.985,15.794],[-6.960,14.140,26.482],[-19.704,-3.533,24.851],
[-19.634,10.067,21.358],[-18.485,11.942,30.740],[-16.407,10.581,34.449],
[-12.738,11.857,14.058],[-9.533,9.783,33.119],[-7.032,11.473,16.685],
[-21.584,5.591,32.980],[-17.236,7.697,16.079],[-14.181,5.853,32.664],
[-12.003,8.957,15.259],[-8.401,6.843,29.960],[-19.526,4.117,31.958],
[-15.005,5.883,17.562],[-14.701,4.784,29.859],[-15.048,3.917,15.579],
[-10.070,7.051,32.272],[-5.266,5.702,15.700],[-14.689,-0.317,29.476],
[-15.702,-0.193,26.897],[-13.226,6.908,28.737],[-10.535,2.116,31.626],
[-9.759,4.551,31.051],[-17.792,2.183,32.006],[-14.407,-3.124,28.879],
[-14.143,-2.002,31.532],[-10.726,-0.508,31.246],[-8.718,-0.297,33.124],
[-20.128,0.830,33.273],[-14.122,-5.068,31.736],[-15.586,-3.627,33.555],
[-15.832,13.753,17.856],[-19.952,9.012,15.864],[-22.313,13.828,20.048],
[-20.342,12.025,23.044],[-4.276,14.241,25.628],[-4.877,12.086,27.245],
[-5.857,9.387,33.209],[-5.492,7.726,19.083],[-5.005,4.303,21.035],
[-4.377,9.263,27.564],[-6.749,10.321,28.263],[-3.428,5.888,17.614],
[-4.519,2.495,16.010],[-7.748,2.611,21.920],[-21.992,6.270,24.471],
[-4.490,6.315,30.397],[-5.332,4.133,33.068],[-6.727,0.482,15.486],
[-6.929,2.449,19.303],[-7.240,2.383,24.754],[-6.512,2.558,27.596],
[-22.425,-0.080,32.331],[-2.840,-2.419,32.772],[-4.941,-1.249,20.863],
[-17.970,-1.299,24.765],[-22.901,3.209,25.713],[-5.204,-1.747,31.171],
[-4.188,-0.525,14.835],[-21.044,-1.698,20.027],[-6.794,13.640,21.387],
[-21.963,-2.293,26.873],[-20.220,11.340,32.789],[-15.880,11.405,27.617],
[-11.676,13.596,32.372],[-9.323,13.813,30.760],[-4.457,-3.341,29.137],
[-19.294,15.865,27.922],[-16.223,14.468,28.772],[-11.687,15.480,26.265],
[-10.059,13.362,28.104],[-7.928,15.306,28.591],[-19.834,13.713,25.702],
[-17.951,11.398,25.231],[-13.856,12.427,23.536],[-12.496,-0.992,23.379],
[-7.163,-2.673,21.331],[-5.837,-2.640,24.076],[-18.329,-2.089,20.753],
[-20.297,-3.957,18.089],[-14.110,-3.481,22.755],[-8.605,-2.706,24.056],
[-8.630,-2.648,18.900],[-20.603,11.043,19.011],[-18.398,-2.683,16.519],
[-11.686,-2.753,19.098],[-12.606,13.222,17.013],[-4.554,12.989,19.859],
[-19.991,-1.459,31.571],[-18.423,-3.570,33.312],[-18.921,14.759,16.248],
[-11.150,-3.942,32.361],[-4.756,12.084,32.511],[-10.420,24.028,22.238],
[-13.587,24.390,26.218],[-18.511,24.722,26.060],[-8.426,27.347,23.634],
[-18.207,28.710,24.465],[-21.971,22.769,21.427],[-21.061,24.143,23.417],
[-10.510,33.447,24.789],[-11.745,26.197,23.324],[-3.845,18.097,26.381],
[-10.026,20.865,25.949],[-16.719,19.120,23.456],[-8.334,18.394,20.460],
[-19.706,21.646,23.005],[-13.904,24.421,23.350],[-3.289,20.461,27.833],
[-17.594,18.302,26.295],[-11.509,28.545,20.763],[-16.093,25.393,26.901],
[-6.958,18.479,23.204],[-18.540,20.784,25.495],[-17.125,27.868,28.609],
[-20.549,27.224,19.399],[-7.092,27.082,27.313],[-6.614,17.089,26.899],
[-14.041,15.067,24.249],[-8.709,30.692,23.986],[-9.262,29.741,21.210],
[-8.391,26.628,31.502],[-10.731,25.411,19.831],[-11.121,20.416,15.864],
[-5.573,22.698,28.173],[-8.060,24.978,21.423],[-9.401,22.352,29.312],
[-17.268,24.170,22.914],[-15.510,29.122,24.832],[-15.940,22.730,24.856],
[-10.793,30.536,25.814],[-15.148,28.491,31.151],[-9.844,23.070,17.133],
[-17.150,24.226,29.526],[-12.018,31.245,21.226],[-15.975,31.220,31.246],
[-5.366,15.475,19.376],[-16.886,23.654,19.635],[-4.292,28.066,16.944],
[-20.465,26.418,27.118],[-21.934,28.941,25.584],[-2.988,29.954,29.178],
[-12.068,28.899,30.658],[-11.703,33.774,19.328],[-19.224,32.572,20.820],
[-8.770,30.658,15.854],[-9.403,32.221,19.722],[-6.453,21.611,32.396],
[-19.959,26.460,29.885],[-16.739,15.335,25.218],[-9.833,16.503,30.215],
[-8.221,19.902,29.911],[-7.428,24.663,17.114],[-12.247,18.540,17.350],
[-17.604,22.561,27.093],[-15.523,22.249,22.047],[-10.727,22.484,19.971],
[-10.371,22.764,32.279],[-9.452,21.509,23.297],[-18.064,20.225,28.388],
[-14.481,17.232,16.926],[-22.313,19.943,22.113],[-21.535,20.554,19.497],
[-11.402,22.931,27.430],[-14.742,17.801,22.280],[-18.722,19.937,21.320],
[-17.391,18.920,18.935],[-9.683,17.450,26.719],[-16.545,21.402,17.920],
[-6.575,32.787,31.588],[-17.972,28.305,31.636],[-11.937,22.092,24.572],
[-17.743,17.390,21.071],[-12.704,19.506,26.656],[-12.658,26.681,18.584],
[-7.302,24.504,26.874],[-9.526,18.077,16.821],[-14.224,25.614,29.791],
[-9.521,20.404,18.302],[-19.548,26.299,23.980],[-10.508,29.819,18.200],
[-22.249,23.669,31.110],[-12.962,23.555,31.556],[-10.056,27.363,29.595],
[-3.983,22.499,18.478],[-4.389,27.664,29.899],[-5.904,28.780,20.119],
[-8.812,24.670,24.545],[-15.604,30.046,21.768],[-19.872,24.697,16.607],
[-14.179,27.836,20.399],[-13.252,28.684,28.087],[-15.672,33.105,22.400],
[-10.743,26.151,26.174],[-8.711,27.310,17.122],[-7.833,29.601,29.868],
[-4.159,30.575,16.077],[-19.332,19.488,30.576],[-7.221,26.103,19.384],
[-19.301,30.107,28.012],[-14.089,28.671,14.852],[-15.412,28.001,17.340],
[-17.430,26.635,18.516],[-12.439,17.928,29.124],[-10.395,17.468,19.097],
[-13.353,29.464,18.298],[-12.602,19.606,20.611],[-12.900,20.557,29.353],
[-9.894,16.391,21.664],[-15.380,30.748,17.500],[-15.835,16.670,19.161],
[-5.508,18.289,31.466],[-7.185,22.060,25.544],[-18.056,29.912,17.605],
[-20.593,21.395,16.409],[-7.313,32.190,26.077],[-19.139,16.713,24.282],
[-20.033,29.593,22.409],[-20.034,33.617,32.816],[-13.653,31.943,34.047],
[-11.335,30.199,14.791],[-9.319,28.990,32.114],[-6.668,29.741,16.813],
[-20.126,25.089,33.035],[-17.664,27.706,14.762],[-15.033,25.442,32.688],
[-11.213,27.681,15.762],[-8.944,25.057,29.051],[-19.898,22.141,32.411],
[-14.230,23.402,17.656],[-15.325,21.350,29.970],[-15.406,23.939,14.765],
[-10.718,25.486,32.147],[-7.163,24.009,14.717],[-14.877,18.995,31.124],
[-15.641,18.237,28.442],[-14.909,22.349,27.479],[-9.990,20.274,33.471],
[-7.868,23.883,32.196],[-17.174,21.147,31.875],[-14.305,15.898,27.410],
[-12.525,18.614,33.121],[-10.853,19.499,30.809],[-8.639,18.147,32.175],
[-17.623,18.836,33.229],[-14.419,14.104,31.221],[-13.359,15.567,33.477],
[-13.885,32.976,18.232],[-21.663,27.422,15.688],[-20.777,34.541,21.322],
[-22.528,31.059,23.440],[-3.512,32.898,25.176],[-3.537,32.391,28.130],
[-5.516,27.976,32.596],[-4.670,26.368,19.916],[-4.398,24.476,22.044],
[-4.727,25.775,27.163],[-5.306,29.512,27.849],[-4.284,24.102,16.214],
[-3.946,20.846,16.059],[-6.281,21.418,22.803],[-21.170,23.796,25.982],
[-5.909,25.525,31.242],[-4.113,23.542,31.874],[-7.003,18.905,16.405],
[-5.366,19.475,20.490],[-6.082,19.692,27.082],[-5.505,20.474,29.984],
[-22.147,20.104,32.478],[-2.836,16.245,33.391],[-6.383,16.421,21.683],
[-20.337,17.995,26.723],[-21.521,20.828,25.068],[-1.177,17.140,30.808],
[-4.207,18.139,16.452],[-21.652,17.631,19.078],[-7.141,31.344,21.179],
[-21.798,16.883,28.967],[-19.258,29.409,34.072],[-16.537,30.063,27.441],
[-11.858,32.127,30.453],[-9.230,32.428,30.620],[-4.821,15.481,30.648],
[-18.675,34.846,28.396],[-17.897,32.689,27.176],[-12.858,31.547,27.457],
[-9.747,31.841,28.001],[-8.295,34.227,27.615],[-19.183,30.897,25.589],
[-18.163,31.949,23.370],[-15.517,31.735,25.369],[-12.843,17.365,24.573],
[-6.306,13.028,24.179],[-5.106,16.256,24.181],[-20.301,15.295,21.398],
[-20.040,14.086,18.692],[-16.994,15.074,22.551],[-8.427,16.244,24.374],
[-8.224,15.795,19.201],[-20.630,30.412,18.064],[-18.076,18.119,16.376],
[-13.190,16.951,20.139],[-12.069,34.777,16.476],[-4.315,32.109,19.845],
[-19.331,14.837,30.898],[-17.006,15.938,33.100],[-17.997,32.375,15.881],
[-9.292,14.601,33.552],[-4.279,30.619,31.879],[-10.697,42.826,22.444],
[-13.144,42.567,24.841],[-18.823,42.444,26.053],[-7.680,46.989,23.711],
[-17.538,45.713,24.292],[-21.788,41.746,20.142],[-21.744,42.290,22.833],
[-10.313,51.376,24.897],[-12.003,45.362,24.484],[-5.233,37.098,25.583],
[-9.157,39.929,25.453],[-15.881,38.658,23.218],[-7.749,38.570,21.600],
[-20.461,40.241,23.988],[-15.200,41.258,23.181],[-3.915,38.582,27.840],
[-18.854,37.438,28.156],[-10.710,45.516,21.350],[-16.527,43.781,26.673],
[-7.521,36.720,23.786],[-18.427,37.940,24.354],[-17.295,45.813,28.325],
[-20.421,46.467,19.565],[-8.073,47.223,26.330],[-7.101,36.517,28.327],
[-14.250,34.110,25.518],[-8.157,49.744,23.631],[-8.392,47.755,21.218],
[-5.147,46.128,27.934],[-12.569,43.925,19.868],[-10.746,41.606,15.957],
[-5.226,42.170,28.788],[-8.960,43.528,20.136],[-10.270,41.467,28.539],
[-17.335,44.114,22.106],[-15.160,46.997,24.150],[-14.904,44.435,24.387],
[-10.651,49.117,27.034],[-14.543,47.341,30.253],[-8.652,42.916,17.688],
[-17.496,42.391,29.049],[-11.462,48.768,20.791],[-15.765,49.450,31.840],
[-4.211,33.475,17.518],[-17.266,42.244,20.073],[-6.620,49.173,16.793],
[-22.311,45.507,27.283],[-22.311,48.729,26.758],[-2.420,48.926,31.758],
[-14.162,50.583,28.900],[-12.123,51.307,18.585],[-18.425,48.948,20.480],
[-9.804,50.601,17.107],[-9.042,49.508,19.403],[-6.502,40.423,30.878],
[-19.963,46.641,28.803],[-16.485,35.767,25.327],[-11.512,34.917,30.643],
[-8.601,38.844,29.932],[-7.113,41.877,15.613],[-12.782,37.321,16.555],
[-16.284,38.540,26.048],[-15.357,39.803,20.846],[-11.127,41.531,19.688],
[-11.819,40.515,31.964],[-9.660,40.310,22.725],[-18.041,39.734,29.598],
[-16.241,38.208,16.404],[-23.225,37.864,22.097],[-21.521,39.118,17.672],
[-11.476,38.182,26.114],[-14.264,36.531,23.529],[-18.225,39.281,21.563],
[-14.635,38.636,18.425],[-9.808,36.241,26.266],[-17.471,39.404,18.842],
[-6.957,51.532,33.256],[-18.338,44.261,30.850],[-13.027,39.951,24.355],
[-18.700,35.359,18.429],[-13.757,39.293,27.004],[-14.144,44.107,17.458],
[-6.980,43.764,27.284],[-9.664,37.116,16.772],[-13.285,44.565,29.849],
[-8.617,40.712,20.139],[-19.890,44.323,22.722],[-12.007,46.177,18.221],
[-21.803,41.959,30.591],[-14.943,41.087,30.633],[-11.242,47.229,29.774],
[-3.551,39.893,18.617],[-3.860,45.446,30.390],[-4.284,48.005,17.583],
[-7.944,44.242,22.427],[-15.220,47.473,20.919],[-20.679,42.431,16.425],
[-14.785,44.227,21.431],[-12.632,47.148,27.249],[-15.332,50.600,21.546],
[-10.452,44.784,27.816],[-9.242,46.525,16.417],[-8.814,48.770,30.266],
[-2.302,50.519,17.265],[-20.233,36.704,30.648],[-7.939,45.456,18.412],
[-19.357,49.439,27.659],[-15.143,45.379,14.467],[-15.068,46.870,17.603],
[-20.129,43.703,19.447],[-11.828,36.307,28.351],[-9.206,35.849,19.120],
[-12.600,48.817,17.805],[-12.084,38.966,19.695],[-12.461,39.633,29.384],
[-11.839,35.937,21.258],[-16.285,49.489,18.778],[-15.979,35.543,16.342],
[-5.846,37.423,30.715],[-5.806,42.627,24.965],[-17.176,45.514,16.882],
[-19.099,39.471,16.506],[-6.718,50.198,26.409],[-20.577,33.343,24.007],
[-20.478,47.127,22.331],[-20.235,51.016,32.023],[-16.008,50.224,34.211],
[-12.830,48.304,14.883],[-10.027,47.288,32.514],[-8.992,48.859,14.668],
[-20.514,43.720,32.305],[-17.064,43.622,15.063],[-15.376,43.573,31.160],
[-11.679,44.400,16.198],[-8.629,43.601,29.852],[-19.569,41.024,31.949],
[-17.957,42.386,17.230],[-15.649,40.914,27.633],[-13.256,40.358,16.138],
[-11.046,44.753,31.652],[-5.619,44.684,16.330],[-15.456,38.554,29.333],
[-15.905,36.476,27.962],[-13.197,42.133,28.963],[-7.680,39.716,33.409],
[-9.074,42.022,31.878],[-16.770,40.292,32.733],[-14.196,33.854,28.317],
[-13.534,37.981,31.132],[-10.955,37.709,30.458],[-8.781,36.213,31.581],
[-18.247,38.070,32.403],[-13.812,33.576,30.880],[-13.663,34.494,33.483],
[-14.785,51.528,17.395],[-20.489,48.188,15.670],[-21.930,53.287,21.267],
[-19.668,49.849,22.912],[-3.675,50.382,25.012],[-3.996,49.972,29.546],
[-6.997,46.370,32.588],[-5.237,45.339,19.622],[-5.150,43.980,22.430],
[-5.382,45.665,25.132],[-6.653,48.584,28.717],[-3.826,43.082,18.264],
[-5.460,40.030,16.497],[-6.038,41.232,22.269],[-22.291,43.499,25.194],
[-5.768,42.882,31.450],[-4.253,40.597,32.398],[-6.927,37.623,16.022],
[-5.738,38.477,18.871],[-6.377,39.483,24.767],[-6.604,39.415,27.884],
[-22.061,38.652,32.389],[-4.046,36.442,33.558],[-6.551,36.403,20.885],
[-21.012,37.886,25.363],[-22.638,40.401,25.694],[-4.718,35.005,31.027],
[-4.370,36.558,15.612],[-21.715,36.581,19.449],[-6.210,51.073,21.450],
[-21.038,35.552,27.311],[-19.797,48.295,33.524],[-16.588,48.561,27.558],
[-12.679,49.610,31.080],[-10.328,51.221,31.349],[-6.853,31.948,28.776],
[-18.837,52.047,27.567],[-15.886,51.748,26.494],[-12.627,51.546,26.491],
[-9.169,51.083,28.097],[-6.142,52.024,28.111],[-18.335,47.960,25.654],
[-17.303,48.505,22.967],[-14.304,50.659,24.401],[-12.308,35.778,25.156],
[-5.969,31.577,23.624],[-4.736,35.300,22.785],[-18.144,35.140,23.221],
[-21.041,34.624,16.590],[-13.237,33.527,21.615],[-7.487,34.175,24.229],
[-7.024,33.749,19.956],[-19.597,49.976,18.078],[-19.378,36.632,15.962],
[-14.038,35.630,18.697],[-12.100,51.345,15.536],[-5.117,50.393,19.185],
[-18.014,35.581,31.774],[-18.322,34.672,34.587],[-17.986,50.440,15.649],
[-10.905,35.420,33.074],[-5.293,48.588,32.485],[-4.538,56.010,25.213],
[-9.403,58.490,27.152],[-16.809,56.264,22.495],[-8.986,57.587,21.006],
[-20.956,58.636,23.328],[-3.368,57.078,27.868],[-18.076,54.600,28.444],
[-6.535,56.331,21.811],[-18.671,57.175,24.396],[-6.693,55.054,26.781],
[-14.547,54.149,24.895],[-4.889,52.636,17.683],[-5.769,59.274,30.859],
[-17.026,53.504,24.511],[-10.930,53.274,29.658],[-8.234,55.989,30.552],
[-12.206,55.129,16.001],[-16.203,56.557,25.121],[-15.743,58.256,20.603],
[-11.400,57.281,21.965],[-17.660,-59.504,27.789],[-14.535,58.106,15.735],
[-22.352,56.379,22.181],[-20.433,56.023,18.959],[-13.838,56.859,23.458],
[-19.024,57.337,21.130],[-14.743,57.154,18.121],[-9.278,54.805,26.150],
[-16.996,-59.475,17.137],[-11.550,58.652,25.160],[-16.839,55.481,19.600],
[-14.105,56.352,26.760],[-7.917,56.537,16.034],[-9.803,59.533,18.605],
[-3.172,57.693,18.207],[-20.333,55.047,30.501],[-11.106,56.215,28.749],
[-10.977,54.432,18.996],[-12.828,58.345,19.702],[-13.044,58.164,29.040],
[-10.929,53.068,22.051],[-14.960,54.358,17.982],[-5.926,57.207,28.857],
[-19.418,58.205,16.553],[-19.755,54.958,26.349],[-15.080,57.332,31.137],
[-15.612,56.375,28.911],[-10.149,58.003,34.224],[-17.382,57.737,32.628],
[-13.615,53.759,27.364],[-15.009,59.110,33.191],[-11.070,55.750,31.450],
[-8.446,55.636,34.008],[-19.410,55.597,33.114],[-14.418,51.832,31.392],
[-14.234,53.044,33.846],[-4.859,57.593,15.976],[-5.663,54.703,14.872],
[-7.068,57.238,18.969],[-7.596,57.465,24.977],[-6.616,-59.171,27.676],
[-22.076,57.377,31.833],[-4.756,53.130,33.315],[-5.758,53.589,20.077],
[-21.314,54.592,24.018],[-22.150,58.513,25.862],[-4.038,54.279,30.393],
[-2.939,53.611,15.824],[-21.724,53.745,18.504],[-22.317,53.441,26.241],
[-6.480,52.927,30.661],[-12.135,55.478,25.043],[-5.357,52.510,25.480],
[-6.698,53.522,23.272],[-19.636,52.679,23.626],[-19.657,52.413,16.920],
[-15.377,53.852,21.778],[-9.424,54.557,23.370],[-7.433,52.211,18.394],
[-18.810,55.426,16.653],[-13.370,53.373,19.829],[-17.774,53.258,31.051],
[-17.865,53.270,33.786],[-9.153,52.880,33.092],[-10.944,-48.805,41.531],
[-12.576,-50.118,45.013],[-18.835,-52.543,43.839],[-10.436,-46.872,43.494],
[-18.658,-46.367,43.718],[-22.698,-51.270,39.842],[-21.521,-51.222,42.352],
[-10.391,-41.147,42.025],[-13.210,-47.627,43.257],[-4.549,-57.664,45.513],
[-9.370,-54.102,44.682],[-16.697,-56.411,41.868],[-10.527,-55.054,38.971],
[-20.852,-54.682,42.052],[-13.914,-50.347,41.119],[-4.691,-54.864,45.511],
[-18.342,-57.337,46.143],[-11.892,-45.925,40.347],[-16.025,-49.846,47.404],
[-5.810,-56.205,42.780],[-17.982,-55.446,44.361],[-17.715,-48.635,45.582],
[-20.422,-47.765,38.610],[-8.351,-46.219,45.000],[-7.311,-56.346,45.911],
[-13.462,58.344,43.396],[-10.178,-43.931,40.984],[-8.601,-45.894,38.967],
[-6.998,-47.842,47.468],[-11.865,-49.597,38.875],[-11.256,-53.675,34.085],
[-6.080,-50.443,46.817],[-7.364,-48.521,40.081],[-9.408,-53.048,47.815],
[-18.092,-49.393,40.835],[-15.707,-46.683,43.689],[-15.578,-49.978,44.307],
[-11.197,-44.821,45.731],[-14.615,-47.624,48.544],[-8.941,-52.417,35.940],
[-18.412,-50.255,48.306],[-14.198,-45.041,41.564],[-15.653,-46.245,50.913],
[-6.116,-58.331,37.477],[-16.531,-51.586,41.140],[-6.558,-43.094,35.159],
[-21.659,-48.108,46.781],[-21.315,-47.067,42.109],[-2.986,-45.006,50.483],
[-12.172,-45.903,48.126],[-11.605,-41.916,39.392],[-17.703,-41.920,40.089],
[-10.834,-43.266,34.845],[-9.292,-42.819,37.104],[-6.769,-53.578,50.036],
[-19.190,-46.940,47.526],[-16.398,-58.555,43.678],[-11.706,59.161,49.581],
[-8.699,-55.798,48.257],[-6.464,-51.438,34.284],[-11.516,-56.689,34.290],
[-15.986,-52.510,44.814],[-15.210,-54.508,39.989],[-10.886,-52.637,38.185],
[-10.677,-52.921,50.519],[-9.875,-52.686,41.321],[-18.684,-54.235,47.151],
[-15.133,-54.991,35.094],[-22.479,-56.893,40.624],[-20.155,-56.121,34.433],
[-11.492,-52.801,45.885],[-13.837,-56.446,42.110],[-17.487,-53.850,41.560],
[-15.879,-53.742,37.557],[-9.662,-56.761,44.358],[-17.507,-54.052,35.447],
[-6.348,-42.000,49.793],[-19.053,-49.193,51.186],[-13.009,-52.146,43.171],
[-15.605,-57.974,40.082],[-14.015,-54.142,47.095],[-13.848,-49.280,35.840],
[-6.260,-48.528,44.725],[-9.555,-55.183,35.667],[-11.961,-48.522,49.354],
[-8.337,-54.170,37.960],[-19.727,-49.719,43.997],[-11.277,-47.077,38.025],
[-23.736,-52.340,49.275],[-15.247,-52.296,49.998],[-9.359,-46.283,48.193],
[-3.259,-53.553,38.020],[-4.876,-47.445,51.295],[-4.812,-45.241,38.593],
[-8.966,-49.432,43.763],[-17.013,-46.189,41.292],[-19.583,-50.204,35.186],
[-14.940,-48.434,38.670],[-14.343,-45.706,45.908],[-15.990,-44.209,39.710],
[-12.079,-47.534,45.826],[-8.500,-47.498,34.650],[-7.422,-43.774,47.747],
[-5.898,-39.832,33.440],[-20.578,-55.741,48.066],[-8.909,-48.456,37.552],
[-21.598,-45.515,47.715],[-15.045,-46.408,32.719],[-15.106,-47.030,35.406],
[-17.800,-48.800,38.120],[-10.970,-57.390,46.805],[-8.877,-58.288,36.702],
[-12.852,-45.665,36.151],[-13.144,-54.310,38.321],[-11.553,-54.788,48.710],
[-12.443,-56.923,39.744],[-15.738,-45.151,37.070],[-14.928,-56.234,37.776],
[-4.798,-56.777,48.376],[-6.194,-49.721,42.265],[-18.062,-46.706,36.064],
[-20.197,-53.726,35.788],[-5.075,-42.939,45.930],[-19.124,-58.108,42.984],
[-19.698,-45.065,40.543],[-21.172,-43.931,49.994],[-15.242,-43.844,52.348],
[-10.740,-46.098,34.559],[-10.077,-46.418,51.113],[-8.292,-44.720,33.572],
[-21.265,-48.836,49.455],[-16.866,-49.388,34.792],[-16.054,-49.439,50.312],
[-12.764,-47.742,33.590],[-9.677,-50.252,48.565],[-20.262,-51.409,50.271],
[-16.167,-51.032,37.338],[-16.911,-52.429,47.739],[-14.345,-52.312,34.924],
[-11.200,-51.127,52.660],[-5.767,-48.886,34.489],[-14.930,-55.917,49.223],
[-15.922,-56.180,46.893],[-13.230,-51.498,47.613],[-8.558,-53.705,52.569],
[-8.290,-51.257,50.800],[-17.529,-54.126,50.952],[-14.033,-58.606,47.540],
[-12.882,-55.845,51.233],[-10.812,-57.700,51.342],[-8.191,-57.155,51.910],
[-18.048,-56.901,49.923],[-15.832,58.466,48.108],[-12.905,-58.987,52.784],
[-13.347,-40.740,36.185],[-20.129,-46.873,34.257],[-21.284,-41.056,40.612],
[-19.713,-42.486,42.024],[-3.359,-41.274,44.089],[-3.823,-43.188,48.817],
[-6.060,-46.460,53.719],[-6.112,-48.027,37.712],[-4.042,-49.493,40.678],
[-3.658,-46.755,43.927],[-5.385,-45.694,46.923],[-4.688,-50.511,37.139],
[-4.254,-53.078,35.571],[-4.588,-53.310,40.659],[-22.602,-49.217,44.257],
[-6.327,-49.460,49.492],[-4.544,-52.374,49.404],[-6.404,-54.929,36.124],
[-7.688,-55.046,40.682],[-8.052,-52.164,43.229],[-6.777,-53.158,46.717],
[-21.687,-53.641,49.557],[-3.141,-58.393,51.132],[-4.278,-57.633,39.638],
[-20.421,-55.204,45.231],[-23.744,-51.724,43.865],[-5.705,-56.005,51.217],
[-4.551,-56.919,35.713],[-22.006,-57.375,37.775],[-8.947,-41.755,39.675],
[-21.328,-57.571,46.176],[-18.415,-43.539,50.598],[-16.616,-45.020,48.392],
[-11.578,-41.719,51.115],[-9.161,-41.634,48.965],[-5.750,-58.900,49.709],
[-19.069,-41.683,48.557],[-18.595,-43.736,46.691],[-12.989,-42.722,45.663],
[-9.700,-43.029,46.837],[-8.064,-40.916,46.568],[-19.308,-40.888,45.755],
[-17.756,-42.990,43.878],[-15.325,-42.721,41.926],[-11.700,-58.264,42.211],
[-7.365,-58.993,44.679],[-4.342,-58.965,41.944],[-19.776,-57.699,40.375],
[-21.053,-59.154,35.846],[-13.810,58.782,40.683],[-8.595,-58.065,42.233],
[-8.180,-58.442,39.365],[-20.164,-42.834,37.288],[-18.319,-58.037,35.202],
[-13.705,-59.209,37.789],[-13.398,-43.037,34.822],[-4.502,-42.394,37.995],
[-19.567,-58.576,48.255],[-17.049,-58.923,51.580],[-18.707,-43.185,34.504],
[-6.250,-59.117,52.316],[-5.619,-44.472,50.463],[-10.098,-30.655,41.493],
[-12.471,-30.471,43.792],[-19.533,-31.356,44.068],[-8.503,-28.469,41.726],
[-18.220,-28.136,43.559],[-20.198,-34.110,39.434],[-24.675,-30.654,40.842],
[-10.986,-27.065,41.814],[-12.345,-27.721,44.076],[-3.893,-38.930,45.504],
[-8.873,-34.613,45.107],[-16.883,-34.816,42.105],[-7.955,-36.759,40.113],
[-21.518,-35.014,41.799],[-15.322,-32.317,41.318],[-4.044,-35.427,47.281],
[-17.550,-37.333,46.856],[-9.573,-29.367,39.132],[-17.094,-29.964,48.178],
[-7.331,-37.549,43.343],[-19.159,-34.384,43.707],[-18.321,-27.650,46.845],
[-21.593,-28.022,37.577],[-8.880,-27.531,44.941],[-5.781,-37.465,46.565],
[-14.669,-39.905,43.917],[-8.550,-25.525,43.216],[-8.360,-25.800,40.535],
[-5.918,-28.665,50.097],[-11.006,-31.926,39.110],[-11.703,-33.568,34.712],
[-4.119,-32.707,47.732],[-7.919,-31.526,39.331],[-10.606,-32.949,48.946],
[-17.037,-30.582,39.760],[-15.704,-27.492,43.495],[-17.031,-30.540,44.896],
[-11.546,-25.460,45.314],[-15.204,-27.142,48.188],[-10.423,-31.678,36.177],
[-18.235,-32.139,48.898],[-12.393,-26.959,39.054],[-16.883,-25.579,51.215],
[-5.092,-39.604,37.235],[-18.346,-32.168,37.750],[-4.561,-26.758,35.089],
[-19.915,-31.202,46.954],[-23.787,-25.689,42.610],[-2.132,-25.757,47.924],
[-12.657,-26.336,48.494],[-13.138,-23.685,38.456],[-19.053,-23.077,40.745],
[-8.673,-23.537,35.580],[-9.431,-23.065,38.938],[-6.924,-34.297,50.157],
[-19.769,-29.272,48.825],[-17.359,-39.062,44.556],[-11.585,-41.121,47.633],
[-8.813,-34.691,47.813],[-7.549,-31.049,34.661],[-14.672,-37.995,36.205],
[-16.607,-33.476,44.392],[-16.816,-34.527,39.548],[-11.159,-34.428,38.429],
[-12.769,-34.647,51.120],[-11.045,-35.862,41.489],[-18.439,-34.025,47.067],
[-17.170,-35.814,33.886],[-22.801,-38.047,40.807],[-20.744,-36.267,37.878],
[-11.155,-33.101,45.685],[-15.136,-37.193,40.961],[-20.179,-37.221,40.506],
[-16.569,-36.394,37.300],[-10.245,-37.367,44.325],[-16.735,-33.869,36.437],
[-5.850,-22.449,48.911],[-18.378,-29.550,51.433],[-14.064,-34.827,43.856],
[-18.063,-38.822,38.447],[-13.404,-36.267,46.420],[-14.647,-30.061,37.297],
[-7.358,-31.131,46.411],[-10.178,-38.301,35.478],[-12.633,-29.039,48.182],
[-9.761,-35.542,35.776],[-18.870,-30.001,41.701],[-11.843,-29.094,37.059],
[-22.170,-31.690,48.313],[-14.323,-32.655,51.743],[-10.010,-27.737,47.888],
[-3.945,-34.027,37.543],[-4.021,-29.850,47.975],[-5.009,-27.287,37.972],
[-9.477,-31.179,44.183],[-17.699,-26.990,38.730],[-19.764,-32.584,34.833],
[-14.887,-29.463,40.973],[-14.466,-26.251,45.855],[-16.930,-24.753,39.944],
[-10.563,-29.697,45.862],[-8.644,-28.278,35.560],[-7.085,-26.216,50.087],
[-4.114,-23.846,34.971],[-19.891,-36.227,48.911],[-7.167,-28.176,39.122],
[-20.086,-26.493,44.943],[-15.150,-27.830,33.820],[-16.711,-27.840,35.851],
[-19.464,-29.675,38.192],[-8.402,-38.393,46.237],[-10.423,-37.325,38.000],
[-14.291,-26.473,37.180],[-13.414,-36.094,39.004],[-12.779,-34.046,48.339],
[-10.981,-38.666,40.136],[-16.456,-24.534,37.220],[-16.937,-39.544,35.677],
[-6.655,-36.461,48.877],[-7.845,-32.861,42.614],[-19.254,-26.884,36.657],
[-19.133,-35.515,35.633],[-6.347,-26.162,44.897],[-19.852,-40.043,43.239],
[-19.931,-25.431,40.027],[-20.185,-25.357,51.956],[-14.727,-25.601,53.067],
[-11.975,-25.802,34.200],[-10.010,-27.519,50.676],[-7.469,-26.315,33.834],
[-21.490,-30.471,50.774],[-17.176,-30.437,35.436],[-15.709,-28.771,50.692],
[-11.115,-28.764,34.455],[-9.029,-30.811,48.750],[-19.605,-32.535,51.625],
[-15.296,-32.219,35.157],[-15.854,-34.499,46.717],[-13.715,-34.052,36.550],
[-10.465,-30.093,51.229],[-3.884,-29.395,33.809],[-15.081,-35.617,48.870],
[-15.034,-38.417,46.339],[-14.855,-32.337,49.090],[-8.706,-35.077,52.412],
[-8.895,-32.549,51.106],[-17.052,-34.148,50.520],[-16.060,-40.824,47.267],
[-13.977,-37.672,50.799],[-11.043,-35.999,49.343],[-8.464,-37.205,50.866],
[-18.079,-38.959,49.481],[-14.483,-40.578,50.469],[-12.821,-39.387,52.759],
[-13.136,-23.110,35.796],[-21.438,-27.469,34.316],[-21.330,-22.270,38.893],
[-22.811,-24.030,40.789],[-4.126,-24.316,44.321],[-4.547,-24.836,49.069],
[-6.236,-27.992,52.659],[-6.141,-29.885,36.890],[-4.506,-29.666,39.427],
[-5.227,-28.345,46.040],[-7.317,-26.938,47.514],[-4.031,-31.482,37.192],
[-4.071,-34.609,34.826],[-5.542,-32.715,40.747],[-22.313,-30.755,44.825],
[-6.053,-31.319,50.376],[-4.096,-33.529,50.531],[-7.503,-36.985,34.806],
[-7.230,-35.594,37.442],[-6.359,-34.977,44.036],[-6.428,-33.956,47.194],
[-21.884,-37.583,50.131],[-4.467,-42.324,52.050],[-6.478,-39.080,39.607],
[-19.096,-37.364,43.068],[-23.375,-35.706,45.408],[-4.475,-38.687,49.331],
[-5.120,-36.921,36.367],[-21.736,-39.099,38.420],[-7.147,-24.207,38.518],
[-20.858,-39.034,47.623],[-19.126,-26.866,49.962],[-16.689,-24.971,48.637],
[-10.685,-23.392,50.246],[-9.060,-22.014,48.578],[-5.222,-40.532,47.645],
[-19.025,-19.887,48.243],[-18.550,-22.524,46.749],[-12.864,-23.193,45.799],
[-9.836,-24.961,47.686],[-7.432,-23.743,46.291],[-19.990,-24.337,43.290],
[-17.196,-25.371,43.180],[-16.079,-22.348,41.977],[-12.864,-37.996,42.187],
[-4.996,-39.190,42.773],[-6.260,-41.993,40.578],[-16.614,-39.545,41.856],
[-20.249,-41.215,34.519],[-13.682,-40.756,41.558],[-8.476,-39.065,41.491],
[-7.665,-40.786,37.568],[-21.738,-25.542,36.335],[-19.335,-38.615,35.069],
[-13.761,-39.268,38.561],[-11.754,-25.397,36.753],[-3.712,-25.352,39.515],
[-20.383,-39.707,50.184],[-17.842,-40.864,51.420],[-16.998,-20.759,35.054],
[-9.317,-39.858,51.598],[-4.460,-26.311,51.344],[-10.911,-13.260,41.392],
[-12.655,-11.947,44.091],[-18.251,-14.155,43.837],[-6.657,-9.609,40.512],
[-18.666,-9.034,44.245],[-21.545,-14.616,39.997],[-22.029,-12.997,42.073],
[-10.270,-5.336,42.068],[-14.892,-10.357,43.333],[-4.701,-19.128,45.900],
[-8.686,-16.743,44.643],[-17.625,-18.114,40.962],[-9.656,-18.545,40.835],
[-20.900,-16.204,42.208],[-14.856,-15.155,41.145],[-3.118,-17.600,47.847],
[-18.111,-19.452,44.749],[-12.679,-9.909,40.745],[-16.361,-13.224,46.072],
[-7.426,-19.336,42.819],[-18.552,-16.988,43.318],[-17.217,-10.559,46.599],
[-20.307,-9.006,35.956],[-7.616,-9.357,43.687],[-7.193,-19.696,46.355],
[-13.774,-22.088,43.462],[-8.252,-7.031,41.668],[-8.935,-9.036,39.062],
[-7.135,-10.220,48.605],[-10.597,-11.159,38.895],[-11.985,-15.839,36.090],
[-4.657,-14.651,47.380],[-8.290,-11.997,40.993],[-10.163,-15.699,48.709],
[-16.307,-12.571,41.744],[-15.439,-7.254,43.564],[-17.332,-11.674,44.296],
[-11.786,-7.527,46.161],[-14.854,-9.787,49.570],[-9.935,-14.150,35.923],
[-17.645,-14.025,48.558],[-10.556,-6.903,39.411],[-16.825,-6.962,49.763],
[-4.843,-22.009,37.114],[-17.531,-13.031,38.892],[-4.919,-9.037,35.116],
[-20.375,-10.573,46.632],[-21.586,-8.081,44.691],[-3.103,-7.366,49.616],
[-13.668,-7.416,48.567],[-11.544,-4.518,39.958],[-17.815,-3.857,40.605],
[-7.676,-4.328,35.176],[-8.840,-4.559,39.519],[-6.216,-16.720,50.006],
[-18.285,-8.997,48.779],[-16.411,-21.504,44.818],[-11.746,-21.399,47.663],
[-8.151,-17.143,48.230],[-8.291,-13.200,33.974],[-12.043,-19.266,36.607],
[-16.210,-16.184,44.045],[-15.643,-16.586,39.104],[-10.305,-14.264,38.991],
[-10.802,-16.025,51.273],[-11.662,-17.613,42.455],[-19.582,-14.704,46.054],
[-15.615,-18.147,33.994],[-22.779,-19.175,41.288],[-21.622,-18.025,35.946],
[-11.452,-15.875,45.800],[-14.614,-18.467,41.300],[-21.057,-18.709,39.145],
[-16.715,-17.566,36.929],[-9.730,-20.054,45.629],[-16.261,-14.578,36.580],
[-7.386,-2.601,50.139],[-18.515,-11.582,49.066],[-12.713,-14.984,42.912],
[-17.931,-20.601,40.144],[-14.217,-16.761,46.018],[-12.012,-12.591,36.530],
[-6.456,-12.309,46.770],[-9.579,-20.121,35.734],[-10.886,-9.873,48.100],
[-8.938,-16.651,36.826],[-19.388,-13.020,41.512],[-11.921,-9.627,37.108],
[-20.662,-13.675,48.728],[-13.092,-13.395,50.528],[-10.635,-6.867,48.774],
[-3.299,-15.705,37.057],[-2.478,-10.168,48.099],[-6.435,-8.757,37.541],
[-9.982,-11.118,43.549],[-16.611,-9.040,41.301],[-18.600,-13.150,34.881],
[-15.429,-11.145,39.706],[-14.260,-9.329,46.100],[-15.830,-6.144,40.612],
[-9.099,-10.377,46.052],[-8.228,-10.111,34.824],[-7.771,-7.275,48.836],
[-5.204,-6.379,36.280],[-19.288,-17.752,46.499],[-7.087,-11.549,36.863],
[-20.072,-7.797,46.996],[-14.425,-8.724,33.875],[-15.119,-9.487,37.523],
[-19.067,-11.435,37.038],[-11.988,-18.560,47.299],[-9.691,-18.991,38.116],
[-13.331,-7.573,36.542],[-11.559,-17.237,39.410],[-12.870,-15.859,48.655],
[-11.858,-19.770,39.987],[-16.231,-6.879,36.309],[-16.705,-21.656,37.799],
[-6.373,-19.481,49.058],[-7.210,-14.897,43.170],[-17.701,-9.003,36.958],
[-20.659,-17.152,33.532],[-6.864,-5.359,45.653],[-18.783,-21.812,43.487],
[-17.889,-7.812,39.458],[-21.675,-4.448,50.573],[-15.737,-6.455,52.361],
[-10.705,-7.015,35.418],[-9.443,-9.717,50.516],[-7.934,-7.235,34.953],
[-20.068,-12.074,51.278],[-16.682,-11.018,35.042],[-13.421,-11.644,48.281],
[-11.016,-9.599,34.573],[-8.977,-12.093,48.543],[-19.338,-14.617,51.012],
[-14.595,-12.237,37.273],[-15.052,-14.581,48.102],[-14.347,-15.613,34.743],
[-10.607,-12.185,51.283],[-5.233,-13.435,32.078],[-14.556,-17.989,48.695],
[-15.397,-18.957,45.998],[-13.577,-13.984,45.641],[-7.478,-16.870,52.790],
[-8.093,-13.979,50.535],[-17.125,-16.246,50.925],[-15.161,-23.159,46.966],
[-12.954,-19.040,50.841],[-10.325,-19.331,49.903],[-7.954,-19.243,51.694],
[-18.699,-17.834,53.046],[-14.732,-23.191,49.978],[-12.739,-21.839,51.502],
[-14.830,-3.514,36.004],[-20.448,-9.197,33.063],[-21.067,-4.297,38.432],
[-20.417,-6.061,41.690],[-4.241,-4.342,45.691],[-3.957,-5.146,48.355],
[-6.686,-9.819,51.339],[-4.291,-11.375,36.763],[-5.985,-12.996,39.398],
[-4.841,-9.663,44.164],[-6.461,-7.891,46.003],[-5.247,-13.638,35.710],
[-4.792,-16.047,33.965],[-6.172,-15.851,40.143],[-21.201,-12.519,44.498],
[-5.778,-12.668,49.355],[-3.865,-15.106,50.237],[-5.717,-18.553,34.175],
[-6.910,-18.460,37.345],[-5.401,-17.369,43.793],[-6.828,-15.692,46.380],
[-22.331,-17.655,51.076],[-3.662,-21.778,53.169],[-5.520,-22.015,39.548],
[-20.600,-18.842,43.329],[-23.313,-14.609,44.374],[-3.117,-21.228,50.591],
[-3.372,-20.673,35.178],[-22.078,-20.346,37.005],[-6.361,-5.921,40.104],
[-21.411,-21.265,47.891],[-19.989,-9.530,51.656],[-16.951,-7.298,46.703],
[-12.923,-5.294,50.303],[-9.904,-4.497,49.430],[-4.664,-21.982,46.322],
[-20.057,-3.243,48.052],[-18.545,-5.058,46.999],[-12.695,-4.396,45.601],
[-9.336,-6.309,46.465],[-8.670,-3.539,46.911],[-20.368,-5.719,44.987],
[-17.979,-6.301,42.549],[-15.324,-4.667,43.391],[-10.910,-20.896,42.813],
[-5.477,-25.751,42.062],[-4.755,-22.841,42.261],[-20.540,-20.741,40.907],
[-20.695,-22.677,35.992],[-14.271,-21.482,39.893],[-8.155,-21.474,41.346],
[-7.407,-20.943,37.941],[-20.211,-6.819,37.569],[-18.447,-18.544,33.840],
[-14.418,-20.159,37.641],[-12.318,-4.990,36.359],[-4.292,-6.959,38.714],
[-17.884,-19.206,50.721],[-15.563,-21.060,51.600],[-18.834,-6.187,34.911],
[-7.976,-23.095,51.273],[-5.826,-5.981,50.636],[-9.219,3.361,40.793],
[-13.176,6.526,44.472],[-17.748,5.309,44.249],[-7.002,7.345,40.877],
[-18.624,10.056,44.883],[-20.275,2.323,39.387],[-22.712,6.025,41.216],
[-9.898,13.537,43.574],[-13.078,8.875,43.261],[-4.601,-1.326,45.197],
[-8.607,1.887,45.847],[-15.921,-0.261,42.465],[-8.651,0.074,41.045],
[-20.997,3.817,41.644],[-13.900,4.934,42.563],[-2.842,1.841,47.560],
[-18.059,0.484,43.933],[-12.001,8.441,39.879],[-15.573,6.466,46.668],
[-6.350,1.025,42.301],[-17.635,2.917,43.013],[-17.316,7.659,44.968],
[-21.963,10.561,39.682],[-8.364,8.968,44.515],[-7.655,-0.893,45.217],
[-13.862,-2.195,43.417],[-8.973,11.029,41.972],[-8.279,9.505,39.523],
[-6.756,8.401,47.889],[-10.252,4.904,38.662],[-11.865,5.679,34.287],
[-4.815,5.139,47.224],[-7.481,3.638,38.990],[-9.379,3.376,48.063],
[-16.385,4.874,41.282],[-15.867,10.392,44.296],[-15.312,4.189,45.031],
[-10.818,11.098,44.943],[-14.440,8.883,49.771],[-9.484,5.194,35.697],
[-18.229,6.727,47.699],[-11.050,12.479,40.086],[-15.810,11.670,48.720],
[-5.126,-3.834,35.297],[-18.393,5.200,38.597],[-5.363,8.790,36.357],
[-22.925,6.105,45.109],[-21.241,10.433,44.043],[-2.076,13.196,49.769],
[-13.091,10.328,48.027],[-13.484,13.812,38.271],[-18.330,13.776,41.298],
[-8.066,13.009,35.773],[-8.755,13.847,39.617],[-6.371,4.147,50.086],
[-18.386,10.323,49.658],[-16.510,-3.572,45.348],[-11.218,-3.047,47.640],
[-8.294,1.178,49.124],[-8.134,5.215,33.325],[-14.192,-0.919,37.365],
[-17.111,1.649,46.226],[-15.656,2.940,39.107],[-13.185,3.185,37.917],
[-10.995,3.871,50.155],[-10.846,2.156,42.243],[-18.077,2.415,48.693],
[-15.613,1.444,34.189],[-21.441,-1.100,41.293],[-19.181,-0.970,35.930],
[-10.739,4.159,44.576],[-13.535,0.854,41.846],[-18.137,1.478,40.840],
[-16.372,0.846,36.740],[-10.210,-0.954,46.108],[-16.879,3.692,35.786],
[-3.898,15.125,50.221],[-17.678,7.817,50.407],[-12.628,2.152,44.523],
[-18.416,-2.073,38.417],[-14.257,1.891,46.613],[-13.993,6.220,40.093],
[-6.044,5.934,44.667],[-9.836,-0.428,36.434],[-11.035,8.345,48.769],
[-9.502,2.200,35.321],[-19.469,8.157,42.790],[-12.330,7.721,36.167],
[-22.836,4.789,47.331],[-13.697,4.761,51.392],[-9.982,10.837,49.076],
[-3.080,2.128,36.572],[-4.752,10.310,48.557],[-8.564,11.098,37.384],
[-8.390,5.878,43.371],[-16.339,9.594,39.463],[-19.320,5.077,34.664],
[-15.824,8.027,42.445],[-13.836,13.537,47.780],[-15.159,12.333,40.281],
[-9.096,8.212,46.792],[-8.626,8.291,35.016],[-7.228,11.993,49.293],
[-2.894,12.846,34.517],[-20.328,1.322,48.511],[-7.727,6.748,37.130],
[-18.772,11.823,46.940],[-15.018,8.884,32.636],[-14.679,9.707,37.288],
[-18.206,7.935,37.446],[-12.569,-0.085,47.521],[-10.488,-2.997,37.489],
[-12.435,11.592,37.168],[-12.077,0.709,39.064],[-13.088,2.472,49.094],
[-12.610,-2.192,40.965],[-15.535,12.410,36.407],[-16.273,-1.541,35.061],
[-6.227,-0.872,48.305],[-6.778,3.734,42.278],[-17.596,10.460,36.914],
[-19.238,2.017,35.647],[-6.687,13.039,45.466],[-18.523,-1.772,42.466],
[-17.950,11.884,39.304],[-20.670,13.653,51.270],[-16.706,11.623,51.631],
[-11.842,11.249,34.203],[-8.895,10.718,51.565],[-5.899,11.902,34.814],
[-21.502,6.651,50.671],[-16.108,6.381,35.277],[-13.533,7.570,52.946],
[-12.262,8.516,33.404],[-8.229,5.922,48.166],[-20.000,4.009,50.353],
[-14.431,6.443,37.229],[-15.676,3.776,48.076],[-14.159,4.428,35.311],
[-11.362,6.668,50.775],[-5.369,6.844,34.281],[-16.886,-0.499,49.373],
[-17.228,-0.922,45.978],[-13.281,4.896,47.337],[-8.971,2.272,51.596],
[-8.926,5.354,51.256],[-16.466,3.409,50.742],[-14.198,-2.847,47.203],
[-12.722,2.086,51.645],[-11.138,0.887,49.942],[-9.105,-0.787,50.997],
[-18.090,0.805,51.362],[-14.945,-4.698,48.866],[-12.869,-2.623,50.580],
[-14.493,15.247,36.056],[-20.930,8.488,34.130],[-20.956,13.250,40.063],
[-20.366,12.153,42.314],[-1.947,14.178,45.634],[-4.234,12.823,47.828],
[-6.842,9.559,52.325],[-4.941,6.632,38.038],[-2.715,5.199,38.268],
[-4.892,8.400,43.023],[-6.405,10.167,46.112],[-4.007,5.106,35.887],
[-5.748,2.298,35.191],[-4.673,2.690,40.355],[-20.349,5.251,44.549],
[-6.763,7.144,50.521],[-4.117,3.612,49.187],[-6.334,-1.796,34.093],
[-6.795,0.203,37.609],[-5.048,1.465,45.743],[-5.938,1.862,48.098],
[-21.779,1.534,50.985],[-2.979,-2.578,52.659],[-5.310,-2.278,37.996],
[-21.082,0.758,43.240],[-22.415,3.511,44.019],[-4.372,-2.396,49.703],
[-4.560,-0.316,35.894],[-21.577,-1.789,37.401],[-6.477,15.311,40.601],
[-21.434,-1.079,47.489],[-20.471,11.062,52.143],[-15.508,9.303,47.004],
[-11.313,13.480,49.225],[-8.792,15.075,49.289],[-6.023,-3.395,47.491],
[-19.805,14.144,48.447],[-16.814,13.745,47.331],[-12.241,14.026,45.576],
[-9.259,12.642,46.901],[-9.429,15.150,46.162],[-18.128,12.666,44.323],
[-17.396,11.030,41.882],[-12.521,11.399,42.355],[-11.238,-0.630,42.652],
[-5.435,-5.238,43.157],[-3.487,-4.121,40.009],[-20.245,-3.378,41.276],
[-20.080,-4.073,35.921],[-15.291,-3.302,41.092],[-9.269,-3.251,44.247],
[-6.893,-2.801,40.224],[-20.292,11.703,37.406],[-20.633,-2.239,33.996],
[-13.390,-3.161,38.684],[-10.984,14.152,36.475],[-4.380,11.976,38.552],
[-18.776,-2.146,50.513],[-16.624,-3.032,52.056],[-18.016,13.251,34.922],
[-10.172,-3.241,52.296],[-5.328,11.928,51.267],[-9.685,24.808,40.835],
[-12.025,24.545,44.386],[-19.235,22.258,44.806],[-8.334,27.377,41.692],
[-17.194,27.172,42.722],[-21.667,21.801,39.996],[-19.106,23.689,41.113],
[-12.082,30.530,42.581],[-12.563,27.230,44.733],[-4.266,17.744,43.628],
[-8.737,20.045,43.759],[-17.435,19.308,41.195],[-8.348,17.652,40.062],
[-22.152,20.213,42.097],[-14.920,20.297,43.588],[-4.421,19.639,46.632],
[-18.552,17.759,45.010],[-10.446,27.565,39.673],[-16.070,24.860,47.412],
[-7.424,19.915,41.302],[-19.032,20.189,43.017],[-16.779,27.829,49.306],
[-19.520,28.316,37.848],[-7.745,27.234,45.618],[-5.900,17.459,45.813],
[-13.262,14.688,42.781],[-8.760,31.314,41.468],[-7.967,29.031,39.626],
[-5.387,26.538,48.149],[-11.138,23.373,38.780],[-10.560,22.400,35.085],
[-4.099,24.482,46.819],[-7.041,24.633,38.836],[-8.776,22.479,47.099],
[-16.931,25.389,40.147],[-15.111,27.982,44.498],[-15.925,24.175,44.550],
[-9.996,28.998,45.543],[-14.864,30.334,51.411],[-9.242,23.437,37.217],
[-18.223,23.939,48.966],[-11.264,30.682,39.445],[-16.715,30.997,49.481],
[-5.731,13.739,37.077],[-16.830,22.762,39.236],[-5.454,28.855,35.103],
[-20.602,26.618,46.272],[-21.624,29.660,44.370],[-1.751,30.326,48.803],
[-12.820,28.599,49.191],[-11.666,33.232,37.999],[-18.311,32.936,42.024],
[-9.918,31.156,36.254],[-7.834,31.623,39.000],[-5.403,21.956,50.489],
[-19.525,28.415,48.373],[-17.290,15.118,45.068],[-12.504,16.044,48.721],
[-9.236,20.008,48.601],[-8.764,24.589,34.665],[-11.522,19.031,36.546],
[-17.060,20.669,45.332],[-15.217,21.138,40.941],[-9.096,21.753,39.130],
[-13.005,21.603,51.075],[-9.989,20.845,41.528],[-18.521,20.766,48.493],
[-13.854,17.599,35.446],[-22.107,17.379,40.771],[-20.598,20.153,38.075],
[-10.381,22.051,45.090],[-14.273,18.929,39.780],[-19.014,21.203,39.934],
[-17.875,19.197,38.144],[-8.616,17.541,44.916],[-15.745,19.771,35.606],
[-7.190,32.804,50.888],[-19.555,25.689,50.703],[-12.275,20.910,43.227],
[-16.788,17.012,40.061],[-13.028,20.221,45.754],[-12.294,26.258,37.869],
[-6.557,24.098,47.543],[-8.897,16.442,35.310],[-11.085,26.739,48.081],
[-8.324,20.357,35.712],[-19.731,26.425,42.756],[-9.858,27.616,35.800],
[-21.096,23.875,47.611],[-14.904,23.264,50.754],[-9.961,29.098,48.641],
[-4.263,22.277,37.320],[-3.430,28.493,48.303],[-5.627,28.641,37.791],
[-8.308,24.674,43.453],[-14.985,29.827,39.951],[-18.240,24.915,36.955],
[-13.932,25.539,42.650],[-14.935,29.734,47.798],[-16.105,32.169,40.369],
[-10.102,26.098,45.345],[-7.905,27.456,33.917],[-7.432,31.062,48.639],
[-4.577,32.066,36.198],[-20.024,19.425,46.964],[-7.673,25.832,36.496],
[-19.377,30.062,45.781],[-14.424,29.073,33.819],[-15.052,27.009,35.516],
[-19.314,26.948,35.489],[-10.885,17.741,46.843],[-8.524,18.207,37.429],
[-12.315,28.890,36.680],[-13.010,21.167,38.143],[-13.649,20.950,48.602],
[-10.615,18.964,39.860],[-14.012,30.874,37.736],[-14.813,17.004,37.987],
[-7.062,18.064,48.179],[-7.279,22.025,44.567],[-17.064,29.225,36.280],
[-19.626,20.922,34.836],[-6.925,31.366,45.821],[-19.612,15.004,43.585],
[-18.212,29.954,39.854],[-21.184,32.320,50.460],[-15.953,32.602,52.521],
[-11.430,30.702,33.324],[-11.100,29.141,51.447],[-6.966,30.269,33.582],
[-22.149,25.694,50.196],[-17.369,25.821,33.863],[-14.935,25.787,49.797],
[-12.481,26.895,34.710],[-8.881,24.959,48.358],[-20.674,21.739,49.700],
[-14.399,24.328,35.960],[-16.059,22.235,48.360],[-16.534,22.472,34.376],
[-9.831,25.839,51.715],[-6.246,25.072,34.183],[-15.468,18.544,48.756],
[-15.170,18.355,45.420],[-13.750,22.920,46.792],[-7.915,21.751,52.954],
[-8.745,23.362,50.333],[-18.357,21.412,50.966],[-15.269,15.885,46.956],
[-14.089,18.586,51.158],[-11.629,19.316,50.043],[-9.267,18.223,50.909],
[-16.461,17.540,51.326],[-15.044,15.810,49.695],[-12.231,17.156,52.595],
[-14.344,33.598,36.231],[-21.681,27.005,34.482],[-20.507,33.740,40.724],
[-20.419,30.810,41.057],[-4.710,31.697,44.550],[-3.881,30.900,47.224],
[-5.316,28.363,50.798],[-4.908,26.376,36.319],[-4.013,24.097,39.411],
[-4.518,26.659,45.215],[-6.021,29.006,47.313],[-2.577,25.369,37.356],
[-4.883,21.165,34.849],[-6.376,22.096,39.487],[-21.568,24.865,43.918],
[-6.663,25.005,51.233],[-3.375,22.503,48.727],[-6.642,19.248,33.781],
[-6.610,19.703,38.595],[-5.532,20.103,43.185],[-6.375,21.410,48.176],
[-23.025,20.249,49.258],[-2.467,16.197,52.497],[-5.452,16.818,38.895],
[-20.531,17.530,43.095],[-21.942,22.113,44.450],[-2.757,17.294,49.383],
[-3.783,18.752,34.112],[-21.319,17.529,37.341],[-5.631,30.864,41.321],
[-20.007,16.468,47.014],[-19.486,30.385,50.124],[-16.871,29.217,46.209],
[-12.873,31.741,47.408],[-10.169,31.849,49.479],[-4.296,15.097,46.670],
[-18.976,33.674,45.987],[-16.135,31.820,46.491],[-11.916,33.462,45.080],
[-9.967,31.562,46.068],[-8.204,33.563,45.484],[-20.157,32.059,43.856],
[-16.956,30.084,42.741],[-14.641,31.521,42.869],[-12.123,17.255,42.762],
[-6.699,13.025,41.931],[-6.430,15.811,43.054],[-20.043,15.763,40.856],
[-19.484,14.608,38.287],[-13.589,15.733,40.316],[-9.256,15.508,41.809],
[-7.587,15.703,37.707],[-20.663,31.526,37.714],[-19.202,17.818,35.364],
[-10.823,15.967,39.482],[-11.142,32.998,34.816],[-4.042,31.375,39.037],
[-19.925,17.045,50.204],[-17.620,15.085,51.098],[-19.741,31.957,34.965],
[-8.669,16.285,52.695],[-4.210,30.835,50.094],[-9.558,42.210,40.568],
[-11.350,42.900,44.848],[-18.670,40.831,44.156],[-8.937,44.305,42.009],
[-17.969,45.984,44.628],[-21.661,40.352,39.287],[-19.898,42.269,41.348],
[-10.162,49.755,42.913],[-10.495,45.910,43.966],[-4.495,35.773,44.243],
[-9.278,38.549,46.067],[-16.981,37.085,41.785],[-8.331,36.910,40.263],
[-19.994,39.716,41.884],[-13.857,41.157,40.504],[-3.317,37.756,46.271],
[-18.259,36.399,46.590],[-11.003,45.770,39.387],[-13.996,44.048,45.263],
[-5.784,35.816,41.508],[-18.914,37.992,43.890],[-16.315,45.552,46.949],
[-20.625,46.528,37.120],[-5.584,46.243,42.301],[-7.759,35.879,46.641],
[-14.006,34.448,43.054],[-8.127,47.348,42.128],[-7.464,49.407,37.891],
[-5.112,46.170,49.133],[-12.409,43.071,38.726],[-10.456,40.663,35.717],
[-4.799,42.147,46.112],[-8.600,41.229,38.318],[-7.471,41.150,46.630],
[-16.002,42.781,40.463],[-15.707,46.663,42.586],[-14.959,41.828,43.822],
[-9.010,47.763,45.445],[-14.731,46.699,49.677],[-9.049,43.843,37.570],
[-18.008,40.949,47.160],[-11.961,48.978,40.800],[-17.175,48.739,50.065],
[-6.510,34.249,36.339],[-17.876,42.336,38.312],[-4.790,46.076,35.469],
[-18.345,43.550,48.035],[-20.560,46.477,45.628],[-3.233,49.227,50.451],
[-12.480,47.340,47.440],[-12.357,51.491,38.594],[-19.160,50.887,41.418],
[-8.720,50.794,35.444],[-9.533,51.237,39.205],[-6.216,39.856,49.742],
[-18.649,47.275,47.897],[-16.047,34.494,45.666],[-11.142,33.567,47.741],
[-7.879,38.267,48.590],[-8.408,42.055,34.514],[-11.318,38.391,33.886],
[-16.491,38.994,44.207],[-16.090,39.554,39.881],[-11.483,40.455,38.714],
[-11.060,41.106,50.125],[-9.519,39.443,40.817],[-19.488,38.990,46.336],
[-14.976,38.178,34.246],[-22.140,36.434,40.275],[-17.670,37.566,38.959],
[-11.849,40.709,46.371],[-14.234,37.153,42.598],[-19.043,40.283,39.519],
[-15.468,38.339,37.115],[-9.930,36.012,44.738],[-15.801,41.336,36.357],
[-5.508,53.198,50.715],[-19.131,45.786,49.972],[-13.460,39.875,42.988],
[-15.046,36.307,39.282],[-14.255,39.499,45.585],[-13.031,45.015,37.226],
[-6.655,44.459,46.789],[-9.349,38.367,37.294],[-11.660,44.554,48.337],
[-6.707,41.640,36.495],[-17.181,42.935,42.848],[-10.157,45.941,36.375],
[-20.859,42.544,49.011],[-15.927,41.979,50.397],[-9.726,47.907,48.119],
[-4.778,40.439,37.708],[-2.898,46.128,50.196],[-6.652,47.619,39.524],
[-7.198,42.694,44.129],[-15.856,47.080,39.877],[-18.267,41.685,34.615],
[-13.636,45.356,39.867],[-13.294,49.234,45.714],[-16.736,49.517,39.737],
[-8.948,44.382,45.666],[-9.311,44.477,33.675],[-7.399,50.378,48.963],
[-2.523,51.749,34.484],[-20.656,36.698,48.564],[-6.922,45.516,37.129],
[-19.333,49.199,45.702],[-14.377,46.662,33.068],[-15.751,46.493,35.792],
[-18.918,44.352,35.898],[-11.824,36.013,47.509],[-10.438,35.755,37.664],
[-12.796,47.734,36.597],[-12.494,38.807,36.637],[-13.137,38.435,47.721],
[-12.539,35.990,40.160],[-17.135,50.720,36.799],[-15.853,35.447,35.105],
[-7.021,35.850,49.683],[-7.946,39.650,43.678],[-17.917,47.855,37.256],
[-19.652,39.096,34.989],[-6.356,47.687,44.324],[-18.281,35.223,43.594],
[-19.096,46.850,40.279],[-19.192,50.070,48.899],[-14.528,50.304,51.584],
[-12.370,48.479,33.702],[-11.366,47.842,50.379],[-8.036,48.938,33.546],
[-19.237,44.137,52.201],[-17.658,46.152,33.597],[-11.654,42.884,52.204],
[-12.229,45.992,34.582],[-8.941,43.027,48.115],[-18.736,41.740,51.030],
[-14.725,43.840,35.724],[-15.411,40.718,47.706],[-13.155,43.711,32.992],
[-9.886,44.463,50.735],[-4.706,44.904,32.802],[-15.520,38.222,49.742],
[-15.237,36.839,46.882],[-13.268,42.576,48.542],[-8.800,39.401,51.135],
[-8.482,41.995,50.491],[-17.170,39.543,51.827],[-14.206,34.655,47.942],
[-13.413,37.960,51.874],[-10.564,37.963,48.813],[-9.738,36.651,51.177],
[-19.199,38.206,50.784],[-14.097,35.472,50.641],[-11.913,35.538,52.150],
[-14.209,50.934,36.207],[-20.792,45.696,34.427],[-21.221,51.276,39.772],
[-21.372,49.227,43.112],[-4.644,50.248,43.601],[-2.812,49.503,47.682],
[-5.606,46.538,51.990],[-4.772,43.753,37.696],[-4.413,42.377,39.965],
[-2.497,45.601,44.800],[-4.773,46.547,46.254],[-3.589,42.633,35.490],
[-4.908,39.200,35.063],[-6.882,41.125,41.591],[-20.121,44.128,43.642],
[-5.195,43.028,49.623],[-3.714,40.807,50.602],[-6.768,36.281,33.220],
[-6.766,37.962,37.147],[-6.866,37.107,44.261],[-6.046,38.921,46.726],
[-22.410,36.779,51.798],[-3.023,34.350,51.826],[-5.424,35.928,38.511],
[-19.696,36.502,41.447],[-21.518,39.566,44.349],[-4.130,35.415,49.383],
[-3.838,36.085,36.457],[-21.823,36.166,37.279],[-5.727,49.783,40.943],
[-21.631,34.623,44.813],[-20.628,48.132,50.252],[-15.299,48.150,47.543],
[-11.793,49.788,48.378],[-9.998,51.520,48.991],[-5.165,33.695,47.620],
[-18.116,51.647,46.552],[-16.287,49.334,45.072],[-11.939,51.182,44.268],
[-9.585,50.315,46.425],[-6.652,51.706,45.896],[-20.525,51.538,44.710],
[-17.653,48.792,42.253],[-14.260,49.196,42.858],[-11.166,35.578,42.401],
[-6.784,32.849,43.229],[-5.481,33.432,40.540],[-17.797,35.180,40.018],
[-19.215,33.915,37.247],[-13.329,32.816,40.708],[-8.846,33.905,42.340],
[-7.995,34.624,38.804],[-19.151,49.889,38.615],[-19.428,37.001,36.908],
[-13.467,36.381,36.298],[-11.434,50.544,36.152],[-4.286,49.926,35.939],
[-19.253,34.934,50.737],[-16.964,34.945,53.233],[-18.958,50.355,34.784],
[-9.874,33.472,51.999],[-5.707,50.735,51.324],[-3.736,56.299,45.337],
[-9.643,57.805,42.842],[-16.404,56.880,41.743],[-9.168,55.218,39.330],
[-21.346,56.925,41.612],[-3.512,59.010,46.828],[-17.791,55.642,45.339],
[-7.099,54.893,41.393],[-20.681,58.422,44.017],[-5.386,54.344,46.322],
[-14.603,51.701,43.645],[-4.855,52.479,36.466],[-7.521,56.784,50.769],
[-16.345,53.323,44.946],[-11.953,52.095,46.977],[-7.524,58.201,48.544],
[-11.881,54.284,34.074],[-16.630,57.843,44.221],[-16.230,58.184,39.415],
[-10.920,-59.416,39.934],[-18.545,58.889,46.115],[-14.130,55.999,34.843],
[-22.962,54.969,41.537],[-21.068,55.947,36.133],[-14.657,54.699,41.794],
[-18.845,56.920,40.537],[-15.815,57.077,36.961],[-8.719,54.978,44.888],
[-16.927,-58.254,37.645],[-12.330,55.757,43.770],[-16.180,55.349,39.541],
[-12.471,58.246,46.008],[-11.201,55.665,36.570],[-8.954,57.864,37.833],
[-3.392,58.685,37.830],[-19.285,54.681,47.477],[-11.650,55.244,46.363],
[-10.719,53.082,37.141],[-13.602,56.608,38.335],[-13.013,56.610,48.141],
[-11.759,56.174,40.482],[-16.007,53.605,36.106],[-5.059,56.643,47.887],
[-17.807,58.093,35.118],[-18.270,53.096,43.220],[-14.880,55.220,49.343],
[-15.330,55.448,46.538],[-9.675,58.830,52.758],[-17.217,57.006,51.292],
[-13.801,53.350,45.606],[-13.381,57.137,50.525],[-10.083,56.819,49.200],
[-8.979,54.127,52.748],[-19.738,55.747,50.217],[-15.506,52.901,50.288],
[-13.218,53.594,51.498],[-5.047,58.338,34.075],[-6.044,56.126,32.209],
[-6.598,57.590,38.560],[-7.220,57.043,43.585],[-9.133,59.004,46.579],
[-20.939,58.270,49.700],[-2.577,51.611,51.473],[-7.077,55.400,36.634],
[-20.732,54.810,43.638],[-23.040,-58.939,44.748],[-3.797,55.491,50.540],
[-4.518,55.786,34.389],[-20.940,54.456,38.264],[-21.596,54.051,46.267],
[-3.600,52.571,48.430],[-11.246,52.760,42.042],[-7.307,51.266,42.897],
[-6.232,53.094,39.597],[-20.919,52.807,42.035],[-20.146,51.895,37.035],
[-13.849,51.132,40.793],[-8.744,53.344,42.462],[-8.112,52.859,37.506],
[-18.080,55.475,35.466],[-13.544,54.035,37.184],[-17.692,53.133,48.823],
[-17.288,51.274,50.919],[-9.567,51.067,51.569],[-21.168,-53.033,57.568],
[-9.124,-56.054,58.137],[-10.000,-47.896,58.783],[-19.957,-47.707,57.067],
[-6.927,-47.155,57.589],[-10.810,-48.972,54.512],[-13.026,-53.193,51.930],
[-7.737,-49.724,58.013],[-9.596,-50.733,56.243],[-11.789,-44.852,57.716],
[-4.317,-58.860,54.330],[-18.606,-52.261,57.952],[-3.758,-44.864,53.309],
[-12.098,-39.887,57.616],[-19.415,-42.958,58.281],[-6.434,-42.257,55.251],
[-9.790,-40.979,56.358],[-8.376,-50.360,53.503],[-12.822,-55.883,53.952],
[-16.312,-53.737,58.545],[-11.785,-53.352,55.217],[-17.371,-56.289,52.508],
[-20.267,-56.196,57.153],[-20.079,-54.700,59.349],[-16.754,-54.202,55.981],
[-16.738,-53.209,53.554],[-16.781,-56.894,56.743],[-12.688,-49.843,56.607],
[-9.506,-55.860,53.875],[-9.595,-53.651,56.705],[-10.980,-46.298,55.124],
[-4.466,-53.563,54.549],[-5.728,-44.302,56.655],[-19.979,-49.875,53.697],
[-16.700,-48.312,57.901],[-15.299,-42.951,57.600],[-8.746,-47.421,53.398],
[-2.229,-42.101,53.942],[-7.587,-48.686,55.403],[-14.965,-46.976,53.575],
[-14.257,-47.673,56.477],[-21.518,-50.548,56.256],[-11.160,-55.885,56.021],
[-13.324,-44.571,55.531],[-13.467,-54.295,56.913],[-12.739,-57.183,58.752],
[-15.922,-45.054,56.147],[-15.556,-55.596,54.046],[-17.480,-45.376,53.108],
[-19.925,-54.902,53.149],[-21.135,-44.992,57.239],[-13.282,-45.061,50.487],
[-8.001,-44.327,53.772],[-18.005,-47.953,53.574],[-12.240,-47.030,52.776],
[-15.972,-49.787,54.430],[-14.894,-51.188,52.434],[-6.222,-49.332,52.301],
[-16.025,-40.573,53.984],[-20.444,-46.824,53.422],[-20.611,-41.393,-58.862],
[-4.396,-48.014,56.402],[-3.616,-50.377,58.888],[-4.500,-50.660,54.282],
[-5.552,-52.110,51.896],[-6.722,-51.886,59.298],[-6.519,-55.299,53.734],
[-6.132,-54.976,58.884],[-6.419,-57.373,57.602],[-3.270,-56.545,53.069],
[-22.196,-57.664,56.222],[-9.544,-42.618,58.860],[-18.985,-59.386,55.792],
[-15.063,-58.901,57.737],[-8.625,-58.712,58.157],[-20.868,-43.672,54.984],
[-20.360,-57.664,53.650],[-14.256,-57.099,55.973],[-11.775,-41.970,53.901],
[-5.672,-41.140,57.661],[-18.646,-42.427,53.723],[-20.597,-35.119,58.840],
[-9.384,-36.034,58.428],[-9.990,-29.020,58.889],[-20.588,-28.172,56.262],
[-6.730,-27.138,57.394],[-10.326,-31.218,57.103],[-11.383,-32.411,52.461],
[-7.975,-32.234,58.408],[-7.317,-33.029,53.498],[-9.546,-26.808,57.109],
[-5.337,-39.439,55.186],[-17.425,-32.069,55.799],[-3.650,-27.957,53.397],
[-13.160,-23.392,57.252],[-18.043,-23.837,56.827],[-8.804,-23.625,54.038],
[-9.394,-24.158,57.192],[-8.308,-30.368,52.612],[-12.200,-37.491,54.900],
[-16.538,-33.972,58.097],[-12.048,-33.469,57.911],[-16.075,-36.000,52.180],
[-21.850,-37.975,55.488],[-18.793,-37.813,58.576],[-16.991,-39.147,56.190],
[-16.840,-35.227,55.501],[-15.500,-37.617,-57.160],[-12.501,-31.457,54.732],
[-10.832,-37.492,52.115],[-10.430,-35.283,54.822],[-11.233,-28.724,55.885],
[-4.418,-35.372,55.801],[-4.119,-26.470,55.466],[-21.120,-32.698,55.397],
[-14.632,-28.364,59.558],[-15.985,-23.503,58.711],[-8.406,-28.440,54.868],
[-4.266,-24.609,53.311],[-7.741,-31.194,56.019],[-13.002,-30.536,50.693],
[-15.659,-27.632,55.055],[-18.276,-29.700,54.915],[-10.119,-38.233,56.737],
[-13.029,-25.914,56.053],[-13.324,-36.003,58.160],[-12.378,-38.057,-58.858],
[-15.538,-24.489,55.635],[-17.662,-37.534,54.188],[-18.178,-26.227,55.503],
[-21.240,-34.966,52.651],[-19.515,-26.249,58.101],[-12.840,-25.003,51.279],
[-9.181,-25.787,52.301],[-17.336,-27.662,52.698],[-10.967,-28.110,53.326],
[-15.315,-30.553,53.271],[-13.903,-35.090,54.872],[-4.975,-30.422,52.796],
[-14.631,-22.057,54.240],[-19.779,-27.658,53.836],[-20.539,-23.459,59.129],
[-5.753,-29.122,55.411],[-4.331,-33.155,58.034],[-4.271,-31.977,54.954],
[-4.503,-35.334,52.718],[-5.364,-31.531,-58.997],[-6.772,-39.483,52.473],
[-7.396,-36.339,56.490],[-6.986,-38.177,58.617],[-3.931,-38.592,52.464],
[-21.006,-39.521,57.904],[-8.476,-22.240,58.940],[-19.662,-39.326,55.204],
[-14.277,-39.750,-58.679],[-8.326,-40.695,54.192],[-20.677,-25.540,55.438],
[-18.450,-34.729,53.314],[-14.205,-38.085,56.676],[-11.673,-21.518,54.217],
[-2.866,-23.142,55.640],[-18.807,-22.958,51.673],[-20.764,-13.794,58.075],
[-10.498,-17.964,58.733],[-9.770,-8.302,56.935],[-22.180,-9.289,55.810],
[-6.096,-9.452,58.805],[-10.216,-12.524,57.607],[-9.600,-14.286,52.938],
[-6.625,-14.713,57.228],[-8.839,-14.125,55.740],[-11.454,-6.944,-59.632],
[-5.119,-21.721,55.600],[-16.535,-14.437,55.643],[-4.962,-5.775,53.177],
[-12.141,-3.661,57.525],[-18.698,-6.247,58.599],[-10.033,-7.299,54.250],
[-10.013,-5.459,56.894],[-7.000,-13.055,53.598],[-12.097,-18.873,54.920],
[-17.120,-16.730,58.374],[-11.482,-15.394,58.239],[-15.386,-18.090,51.596],
[-21.919,-17.150,56.835],[-20.079,-17.276,59.084],[-16.067,-17.140,55.709],
[-15.198,-15.729,53.163],[-16.949,-20.017,59.194],[-12.720,-11.414,57.202],
[-9.277,-20.547,53.672],[-9.175,-16.866,55.254],[-10.538,-10.478,55.335],
[-2.878,-16.920,54.306],[-6.268,-7.417,55.202],[-18.550,-11.214,53.339],
[-16.999,-10.881,56.631],[-16.328,-4.810,59.523],[-8.352,-10.141,53.519],
[-2.668,-4.872,54.127],[-7.709,-10.548,56.193],[-13.888,-9.218,53.097],
[-14.851,-9.497,55.654],[-19.492,-10.138,55.786],[-9.345,-19.545,56.397],
[-12.523,-8.440,56.937],[-14.014,-17.125,57.722],[-11.604,-20.413,57.451],
[-14.567,-6.668,55.077],[-16.694,-19.772,55.535],[-18.244,-8.149,53.355],
[-18.801,-15.117,53.658],[-21.294,-7.763,57.950],[-12.150,-6.490,52.691],
[-7.325,-7.375,52.709],[-15.998,-10.665,52.040],[-11.093,-9.915,52.897],
[-13.419,-12.070,54.472],[-12.433,-14.261,52.914],[-4.580,-12.138,52.215],
[-14.627,-4.319,53.598],[-21.794,-10.844,53.546],[-20.631,-4.076,58.521],
[-5.110,-10.254,55.967],[-4.277,-12.342,57.532],[-4.443,-14.733,53.670],
[-5.653,-16.783,55.331],[-5.060,-14.063,-59.374],[-5.837,-20.649,51.973],
[-7.541,-18.048,58.755],[-6.081,-20.946,58.903],[-4.340,-19.019,53.729],
[-21.278,-21.089,57.921],[-6.739,-4.447,-59.011],[-21.138,-19.764,55.855],
[-14.360,-21.283,58.782],[-8.067,-22.031,56.135],[-21.217,-6.719,54.721],
[-17.549,-22.168,54.381],[-14.285,-19.433,56.534],[-11.970,-4.795,54.909],
[-5.356,-5.033,55.883],[-19.282,-4.648,52.129],[-20.831,3.366,58.058],
[-9.000,1.168,58.604],[-10.058,8.810,58.051],[-22.116,7.834,55.725],
[-5.745,10.485,58.870],[-11.577,6.152,58.015],[-11.093,4.170,52.771],
[-6.496,6.047,57.119],[-9.991,6.626,53.707],[-11.671,11.748,57.837],
[-5.478,-0.679,54.805],[-17.053,5.843,56.456],[-4.222,10.370,52.890],
[-11.040,14.892,57.958],[-19.430,13.200,58.398],[-8.853,13.487,52.620],
[-8.960,12.727,56.691],[-6.994,7.121,54.277],[-10.942,0.548,53.196],
[-15.733,3.486,58.168],[-10.465,4.896,55.678],[-15.215,0.489,51.567],
[-20.733,0.724,56.506],[-18.787,1.502,58.247],[-16.563,1.594,55.423],
[-17.013,3.774,53.790],[-18.170,-0.642,56.295],[-14.949,7.592,57.090],
[-8.878,-1.277,53.912],[-8.949,1.728,55.590],[-11.328,9.469,55.664],
[-4.394,3.358,56.417],[-5.312,10.153,55.379],[-19.913,4.950,54.138],
[-13.845,9.716,-58.356],[-15.399,14.518,58.120],[-8.856,9.200,54.419],
[-4.454,14.014,52.802],[-7.190,8.712,56.745],[-14.041,10.266,52.361],
[-16.972,9.229,55.907],[-19.344,8.037,55.912],[-9.973,-0.952,56.666],
[-13.411,11.633,55.434],[-11.422,2.828,58.372],[-11.423,-1.452,59.240],
[-15.818,11.518,54.205],[-15.785,-1.066,53.831],[-18.487,10.412,53.916],
[-20.388,0.917,53.342],[-19.306,10.304,57.639],[-11.575,11.777,52.216],
[-7.738,11.679,54.366],[-16.475,8.605,52.669],[-11.888,9.422,51.084],
[-15.614,6.066,53.800],[-13.735,3.262,54.192],[-4.907,7.678,52.727],
[-13.748,15.091,54.638],[-19.895,7.961,53.036],[-21.513,13.573,-58.378],
[-4.505,8.355,57.571],[-4.619,5.351,58.845],[-4.894,4.851,54.232],
[-5.091,2.119,53.572],[-6.806,3.250,58.533],[-7.612,0.912,53.290],
[-6.729,0.949,57.132],[-6.293,-2.068,57.635],[-4.740,-0.213,52.402],
[-22.239,-1.634,57.105],[-7.111,12.988,58.778],[-21.900,-4.464,56.164],
[-13.903,-1.991,58.471],[-7.425,-3.491,54.812],[-18.630,12.401,55.630],
[-18.392,-0.787,53.800],[-13.482,-2.184,54.753],[-10.625,13.089,54.460],
[-3.692,11.759,57.152],[-19.397,14.951,53.490],[-20.957,22.496,55.760],
[-8.701,19.826,-59.590],[-9.459,27.280,57.949],[-21.212,26.697,55.751],
[-8.059,29.489,58.111],[-11.645,24.275,55.814],[-11.442,22.574,52.941],
[-7.979,23.944,58.923],[-9.236,23.732,54.195],[-10.642,29.087,-59.107],
[-6.543,15.323,54.305],[-16.688,23.596,55.258],[-3.924,28.955,52.808],
[-10.736,33.526,57.054],[-17.610,31.120,58.515],[-8.948,31.430,55.467],
[-10.403,30.953,57.898],[-6.702,24.740,54.232],[-11.698,18.178,55.358],
[-16.402,21.291,58.589],[-10.242,22.035,57.749],[-15.594,20.644,50.914],
[-21.352,19.436,56.338],[-19.302,20.486,58.013],[-16.482,20.650,55.967],
[-16.734,21.205,53.463],[-17.797,17.460,-59.264],[-14.523,25.947,56.348],
[-9.677,18.908,53.607],[-9.306,20.836,55.249],[-11.965,27.924,54.716],
[-4.783,20.797,55.980],[-5.437,29.813,54.934],[-21.050,24.166,53.210],
[-14.578,26.138,59.053],[-15.016,30.643,58.042],[-9.668,26.475,54.865],
[-4.559,31.865,53.262],[-7.858,25.459,56.676],[-15.081,28.607,53.542],
[-16.832,27.304,55.204],[-18.958,25.849,53.790],[-9.242,18.235,56.550],
[-13.469,29.829,55.638],[-14.509,22.389,57.006],[-11.526,17.752,58.205],
[-16.454,30.259,55.460],[-16.182,17.127,54.398],[-18.553,29.537,53.640],
[-20.645,20.375,53.920],[-20.333,31.356,57.844],[-12.610,31.357,50.447],
[-7.676,29.556,53.150],[-17.235,27.531,51.965],[-13.311,26.046,53.725],
[-15.868,25.028,53.178],[-13.754,23.435,54.278],[-4.655,25.698,52.647],
[-14.429,32.110,54.589],[-20.223,28.170,52.069],[-22.721,33.421,56.914],
[-4.840,27.011,56.879],[-3.304,24.756,57.590],[-3.964,24.746,54.936],
[-5.174,20.810,52.923],[-5.736,22.495,-59.032],[-6.144,18.597,50.810],
[-7.183,21.651,57.615],[-6.516,17.813,57.118],[-4.697,17.319,54.141],
[-20.134,16.671,58.138],[-7.957,32.739,58.420],[-20.355,14.911,56.146],
[-14.839,17.122,58.956],[-8.071,15.604,57.429],[-20.935,29.802,54.971],
[-18.978,18.081,52.812],[-14.029,18.818,56.940],[-11.977,33.112,54.099],
[-5.890,31.224,58.211],[-19.539,33.368,52.991],[-21.045,40.148,58.666],
[-9.042,35.732,58.053],[-9.020,46.533,57.777],[-21.721,45.072,56.949],
[-6.851,45.015,58.402],[-10.277,43.097,55.745],[-10.788,41.109,53.798],
[-8.976,43.451,57.848],[-8.058,42.306,54.153],[-10.314,48.266,-58.758],
[-4.985,34.124,55.186],[-16.981,42.407,56.942],[-4.416,48.616,54.976],
[-13.227,51.999,56.181],[-19.002,51.704,56.518],[-7.351,52.023,53.260],
[-8.541,50.162,55.842],[-5.231,44.223,53.571],[-10.922,38.480,52.878],
[-16.487,39.938,58.008],[-12.854,42.593,58.121],[-17.040,36.946,51.535],
[-19.778,37.423,55.913],[-18.869,38.805,58.053],[-16.718,38.132,56.175],
[-14.701,39.742,55.208],[-15.820,37.385,58.872],[-11.725,45.537,56.515],
[-8.367,37.166,53.470],[-10.728,39.094,55.497],[-10.413,45.384,54.159],
[-3.430,39.160,55.183],[-6.784,47.745,56.259],[-18.746,41.764,53.615],
[-15.422,45.415,58.327],[-16.392,50.366,58.137],[-7.972,46.220,53.012],
[-4.676,51.371,54.901],[-7.515,44.941,55.729],[-16.164,46.341,52.017],
[-15.042,47.339,54.749],[-18.894,44.086,56.334],[-10.265,35.951,54.889],
[-12.282,47.418,54.746],[-12.882,39.236,57.199],[-11.851,36.415,57.618],
[-15.874,50.403,54.894],[-16.015,34.965,55.794],[-17.367,45.779,54.481],
[-20.228,39.129,53.387],[-18.996,47.425,57.483],[-13.532,47.834,52.509],
[-7.661,49.215,52.922],[-14.497,43.954,51.874],[-12.383,45.201,50.851],
[-14.736,44.075,55.644],[-13.798,41.293,52.642],[-6.231,41.656,51.847],
[-16.283,52.936,52.909],[-21.792,45.566,52.386],[-20.766,51.172,58.352],
[-4.702,45.508,56.192],[-3.969,41.924,56.758],[-4.255,41.516,53.705],
[-5.091,39.116,52.838],[-8.600,39.397,-57.993],[-6.468,35.453,52.872],
[-5.425,37.634,58.431],[-6.482,35.247,57.744],[-3.367,36.789,53.534],
[-20.021,36.015,58.254],[-8.100,52.425,58.431],[-19.665,34.237,56.002],
[-13.532,33.144,57.201],[-7.750,34.625,55.297],[-19.759,48.552,55.122],
[-19.315,36.510,53.148],[-12.974,36.653,54.540],[-11.334,52.664,53.573],
[-6.140,50.221,57.896],[-18.678,50.916,53.769],[-10.762,55.737,58.836],
[-5.218,54.205,54.934],[-11.361,55.992,52.389],[-16.045,57.760,58.746],
[-14.940,55.878,52.351],[-19.634,56.814,54.862],[-19.374,57.114,58.228],
[-13.278,57.682,58.147],[-16.854,58.054,54.182],[-17.080,55.445,57.223],
[-7.245,55.932,54.688],[-9.954,57.817,57.155],[-3.404,58.132,56.740],
[-10.150,55.617,54.686],[-12.171,56.987,55.775],[-13.356,54.522,57.692],
[-16.719,54.932,54.629],[-19.433,58.527,52.335],[-4.256,58.137,51.575],
[-5.466,56.002,52.636],[-7.491,57.340,58.496],[-6.030,54.667,57.890],
[-3.568,53.593,52.810],[-21.755,56.135,56.426],[-21.734,54.183,54.141],
[-15.733,53.024,58.597],[-7.602,53.148,55.775],[-18.679,55.060,52.923],
[-13.383,54.197,54.443],[-16.330,-58.843,-14.423],[-17.792,59.322,-19.059],
[-8.554,-58.434,-14.994],[-18.413,-48.551,-29.506],[-11.717,-45.345,-31.354],
[-17.337,-57.468,-26.067],[-16.846,-49.998,-25.532],[-8.153,-57.769,-26.318],
[-7.398,-47.609,-27.815],[-11.793,-49.576,-30.548],[-13.575,-47.580,-32.730],
[-20.085,-48.726,-34.312],[-13.826,-53.061,-34.551],[-5.633,-53.672,-21.029],
[-20.192,-51.866,-36.718],[-17.293,-55.738,-36.077],[-10.487,-46.679,-24.355],
[-16.290,-50.582,-23.026],[-15.026,-52.373,-18.255],[-12.877,-53.842,-19.045],
[-19.139,-51.587,-18.361],[-15.401,-57.727,-16.962],[-23.304,-45.839,-34.515],
[-21.299,-45.815,-32.263],[-6.720,-44.172,-30.708],[-6.985,-56.723,-29.012],
[-4.734,-53.969,-15.418],[-17.410,-43.993,-29.522],[-13.066,-34.486,-37.687],
[-5.017,-36.917,-31.510],[-18.400,-29.653,-34.608],[-4.690,-28.687,-33.923],
[-17.992,-40.777,-38.855],[-17.979,-42.548,-41.819],[-11.680,-32.202,-33.203],
[-23.471,-33.461,-32.633],[-5.886,-31.018,-32.272],[-10.361,-27.352,-32.152],
[-16.006,-38.711,-25.439],[-6.749,-38.688,-38.547],[-7.081,-44.219,-42.894],
[-21.483,-27.326,-32.803],[-18.118,-30.955,-40.283],[-13.049,-42.924,-33.542],
[-4.915,-43.567,-34.623],[-8.783,-15.008,-47.825],[-15.677,-20.107,-43.658],
[-13.996,-13.788,-53.935],[-7.036,-19.840,-37.539],[-8.815,-21.490,-38.283],
[-6.877,-8.674,-44.511],[-10.383,-28.674,-48.659],[-10.916,-13.662,-43.951],
[-4.090,-27.153,-51.058],[-19.500,-20.457,-55.517],[-17.520,-25.745,-56.320],
[-11.960,-20.492,-38.181],[-7.814,-20.224,-34.032],[-5.696,-24.967,-51.695],
[-7.009,-11.549,-45.394],[-12.635,-17.774,-58.127],[-9.345,-3.000,-58.231],
[-4.991,-7.147,-51.441],[-17.995,-9.464,-54.120],[-15.226,-11.006,-50.163],
[-14.789,-4.100,-54.547],[-18.178,4.023,-51.577],[-5.771,-15.547,-56.037],
[-13.383,2.324,-54.404],[-12.763,-9.318,-47.585],[-11.272,-4.324,-50.837],
[-10.985,-7.064,-52.349],[-13.457,-6.495,-54.351],[-3.086,-10.061,58.878],
[-4.639,1.156,-51.399],[-12.160,-7.130,-49.692],[-16.593,-14.650,-57.919],
[-21.712,-10.762,-51.057],[-14.398,1.748,-49.561],[-14.152,-6.105,-43.833],
[-20.621,-10.729,-53.626],[-17.958,2.005,-56.357],[-13.047,6.486,-51.584],
[-18.192,5.883,-57.629],[-7.134,-11.360,-52.804],[-18.887,-12.516,-54.145],
[-18.108,8.406,59.149],[-9.570,12.633,-58.690],[-9.264,15.821,-56.327],
[-16.790,14.554,-55.980],[-6.073,16.537,-59.412],[-4.826,7.921,-52.892],
[-17.976,7.619,-52.930],[-12.625,-59.978,-0.699],[-17.485,-54.123,-0.259],
[-8.513,-57.427,-6.670],[-18.164,-58.320,0.695],[-13.536,-50.371,1.835],
[-11.040,-57.953,-1.793],[-12.335,-57.431,-11.391],[-21.792,-58.298,7.024],
[-16.953,-49.705,-3.999],[-23.414,-47.326,-14.840],[-19.236,-38.779,-17.174],
[-12.058,-51.549,-14.350],[-10.990,-47.021,-12.124],[-12.034,-41.709,-3.773],
[-18.658,-48.097,-11.185],[-17.127,-39.514,-15.679],[-17.952,-51.205,-13.807],
[-21.460,-50.706,-17.283],[-8.908,-39.702,-8.650],[-15.434,-44.853,-17.632],
[-18.378,-42.528,-6.301],[-12.775,-38.590,-4.110],[-5.211,-46.319,-20.368],
[-9.400,-45.028,-0.408],[-5.452,-35.630,-13.634],[-3.888,-44.342,-14.225],
[-20.646,-34.471,-7.512],[-17.705,-32.377,-11.183],[-5.214,-28.345,-16.136],
[-18.454,-31.513,-15.819],[-17.456,-22.180,-18.884],[-21.775,-26.258,-26.881],
[-18.051,-27.076,-22.568],[-10.325,-20.431,-22.197],[-15.163,-21.777,-17.259],
[-17.514,-29.065,-17.008],[-10.537,-24.926,-12.550],[-16.665,-32.470,-22.403],
[-7.857,-24.241,-18.800],[-20.064,-25.639,-20.292],[-10.643,-26.808,-27.475],
[-7.403,-26.692,-19.627],[-19.721,-27.174,-13.054],[-11.335,-27.422,-13.799],
[-13.213,-34.536,-19.784],[-12.917,-25.000,-32.077],[-4.213,-30.676,-26.806],
[-4.914,-24.142,-12.518],[-5.388,-30.864,-13.645],[-20.464,-37.699,-22.530],
[-19.215,-29.546,-8.688],[-22.609,-13.869,-32.119],[-13.153,-13.503,-35.013],
[-19.999,-13.295,-33.695],[-9.779,-14.601,-26.752],[-6.588,1.264,-27.397],
[-3.457,-17.790,-34.307],[-9.969,-12.238,-39.087],[-17.740,-13.593,-40.469],
[-13.988,-10.455,-20.867],[-21.870,-21.867,-30.207],[-8.433,7.816,-41.508],
[-22.220,5.175,-40.202],[-13.522,8.439,-40.408],[-15.852,8.044,-36.468],
[-17.409,1.936,-36.826],[-10.334,3.441,-42.447],[-15.032,4.840,-40.871],
[-20.160,6.837,-35.948],[-17.725,13.477,-36.965],[-5.650,6.106,-50.026],
[-10.901,10.004,-51.606],[-17.009,-3.287,-44.385],[-12.842,5.028,-38.629],
[-13.338,10.172,-35.933],[-10.586,-1.299,-41.987],[-14.134,8.743,-49.833],
[-20.700,9.334,-31.807],[-11.145,10.321,-33.584],[-4.215,7.470,-33.127],
[-7.452,-4.036,-39.914],[-20.123,14.431,-41.353],[-15.026,23.143,-47.964],
[-9.279,15.910,-48.462],[-20.350,17.711,-47.268],[-5.527,20.908,-41.672],
[-6.836,17.329,-42.302],[-10.591,26.121,-49.016],[-10.269,25.404,-57.422],
[-7.252,23.194,-42.101],[-17.992,14.980,-44.059],[-18.536,21.017,-43.958],
[-17.220,15.151,-53.135],[-10.618,21.558,-56.556],[-4.814,23.855,-56.761],
[-17.464,26.389,-54.985],[-21.143,24.229,-55.560],[-17.048,33.097,-45.220],
[-14.650,16.159,-54.693],[-11.898,19.804,-39.685],[-22.126,26.581,-53.594],
[-3.012,29.387,-52.858],[-4.732,21.685,-51.469],[-4.575,10.770,-51.844],
[-13.848,31.364,-51.390],[-19.479,27.298,-52.635],[-11.936,23.797,-59.564],
[-18.353,16.341,-36.842],[-15.754,33.421,-56.663],[-18.157,46.041,-53.031],
[-6.995,26.207,-57.939],[-20.329,43.265,-54.933],[-10.298,28.794,-53.883],
[-19.405,34.654,-47.224],[-10.578,-49.981,11.524],[-5.241,-49.147,3.906],
[-12.669,-42.965,6.013],[-20.749,-44.089,15.817],[-5.710,-46.279,3.891],
[-9.621,-58.041,8.887],[-16.216,-58.035,12.504],[-16.935,-52.420,3.653],
[-11.697,-50.913,4.090],[-12.249,-48.522,23.058],[-6.827,-49.874,18.211],
[-6.379,-57.355,12.772],[-4.886,-53.382,23.471],[-11.219,-39.637,15.066],
[-7.786,-38.948,11.937],[-21.917,-42.733,9.253],[-9.045,-58.489,16.014],
[-12.133,-46.251,2.828],[-17.571,-30.018,1.932],[-18.061,-33.761,1.506],
[-23.713,-28.965,2.770],[-12.352,-39.404,12.267],[-18.301,-39.385,0.951],
[-13.977,-34.837,-0.947],[-15.505,-30.457,12.452],[-15.371,-33.387,0.986],
[-21.351,-27.961,4.997],[-10.096,-41.529,3.104],[-16.706,-39.114,12.217],
[-5.533,-24.626,4.771],[-5.834,-30.547,4.477],[-6.043,-44.081,-0.136],
[-8.002,-39.169,2.887],[-16.688,-18.009,-5.515],[-19.098,-20.696,2.077],
[-17.951,-12.213,-2.670],[-9.982,-19.285,-8.442],[-18.966,-15.565,0.376],
[-5.409,-17.487,-14.428],[-13.491,-10.328,-2.387],[-14.173,-12.574,-13.768],
[-16.435,-23.173,-0.368],[-12.111,-20.819,-3.747],[-16.516,-23.252,-7.835],
[-12.845,-14.277,-0.485],[-11.580,-16.800,-11.505],[-20.746,-22.575,-9.920],
[-16.359,-7.317,0.690],[-15.933,-11.386,1.596],[-11.302,-16.650,4.949],
[-5.351,-21.555,-9.718],[-6.986,-18.866,-5.748],[-10.026,-8.631,-6.238],
[-8.169,-28.315,-0.455],[-12.113,-15.402,-14.430],[-6.529,-12.205,-11.921],
[-11.452,-21.192,8.280],[-15.668,-9.429,-8.990],[-12.555,2.875,-19.755],
[-7.004,-11.941,-14.834],[-18.627,-2.686,-17.078],[-15.791,-12.676,-19.377],
[-8.716,-6.017,-19.540],[-19.095,1.924,-15.642],[-12.606,-7.322,-14.209],
[-17.202,-10.866,-17.405],[-11.389,-1.158,-25.505],[-11.762,-3.073,-23.843],
[-17.744,-3.242,-7.684],[-12.246,-1.761,-6.510],[-6.065,6.349,-13.259],
[-4.803,2.861,-13.563],[-22.087,-7.813,-16.255],[-10.457,9.618,-15.436],
[-20.417,8.634,-13.117],[-20.493,7.312,-17.177],[-13.322,-7.392,-10.494],
[-20.514,-8.890,-11.643],[-12.362,13.384,-21.861],[-7.274,17.440,-25.606],
[-21.692,12.521,-21.453],[-17.481,18.987,-21.582],[-17.805,14.949,-24.689],
[-14.523,10.687,-14.680],[-9.935,6.054,-27.689],[-14.728,18.895,-16.610],
[-17.092,15.932,-18.967],[-5.223,10.555,-31.481],[-5.371,11.194,-12.045],
[-4.486,13.069,-14.133],[-22.589,9.144,-15.861],[-6.172,11.053,-15.671],
[-9.855,6.399,-21.171],[-23.913,28.046,-32.918],[-18.137,31.058,-35.516],
[-20.070,34.157,-31.062],[-14.208,22.461,-36.679],[-17.905,27.679,-34.005],
[-10.360,27.928,-32.301],[-24.076,25.433,-32.435],[-14.477,27.016,-30.939],
[-20.218,24.870,-33.199],[-18.623,23.007,-30.741],[-13.473,27.485,-28.100],
[-13.606,36.655,-29.371],[-6.127,32.792,-39.690],[-17.140,32.910,-37.628],
[-13.240,33.931,-40.834],[-7.813,28.869,-24.980],[-16.923,27.723,-41.926],
[-18.105,32.757,-23.783],[-4.570,40.032,-34.399],[-4.542,26.490,-38.798],
[-6.682,23.103,-37.176],[-20.904,38.931,-28.475],[-6.302,23.439,-30.368],
[-19.904,20.185,-36.096],[-18.078,42.363,-39.311],[-10.345,47.947,-37.552],
[-15.557,51.798,-39.574],[-5.081,47.695,-49.887],[-21.565,51.787,-43.481],
[-19.080,40.983,-37.029],[-12.136,42.643,-33.330],[-11.217,43.154,-44.060],
[-11.560,52.996,-39.360],[-4.304,49.537,-52.095],[-19.662,46.317,-48.844],
[-6.117,45.066,-35.101],[-7.636,45.365,-41.818],[-9.447,54.931,-44.829],
[-14.012,44.426,-51.002],[-14.059,50.687,-36.081],[-5.708,44.887,-50.837],
[-9.970,45.842,-30.409],[-21.248,45.247,-32.876],[-21.731,45.228,-50.784],
[-23.881,52.702,-49.492],[-4.912,49.000,-33.634],[-23.478,47.063,-32.714],
[-5.888,43.997,-32.067],[-12.672,54.848,-44.121],[-16.248,-39.645,29.049],
[-22.153,-39.594,24.227],[-10.717,-45.440,22.861],[-10.590,-40.384,37.087],
[-12.169,-44.293,27.594],[-12.550,-38.143,34.405],[-11.228,-40.906,23.522],
[-14.084,-43.407,18.118],[-13.725,-49.115,28.076],[-6.988,-44.108,37.490],
[-20.436,-47.193,24.962],[-4.635,-45.779,20.501],[-17.415,-43.062,36.992],
[-22.681,-35.062,22.658],[-7.547,-32.656,36.847],[-4.007,-46.191,23.747],
[-6.851,-47.212,27.955],[-8.229,-39.897,35.170],[-5.472,-43.531,42.676],
[-18.306,-20.934,19.253],[-11.701,-25.764,25.236],[-8.114,-29.135,27.563],
[-16.268,-22.717,18.622],[-6.488,-16.519,23.378],[-12.133,-19.556,13.538],
[-8.604,-29.340,13.727],[-16.833,-28.237,19.130],[-9.848,-29.391,25.621],
[-6.846,-13.633,23.200],[-13.455,-26.903,23.290],[-10.885,-30.796,17.486],
[-14.412,-32.098,20.892],[-17.506,-24.803,11.991],[-14.312,-25.876,29.683],
[-18.290,-24.194,30.638],[-20.507,-29.858,23.379],[-22.781,-27.337,23.696],
[-19.572,-15.927,19.149],[-12.607,-14.471,19.635],[-20.598,-23.235,10.998],
[-4.844,-10.532,16.423],[-9.799,-5.719,6.365],[-11.024,-12.124,9.066],
[-16.922,-4.182,4.641],[-20.651,-8.853,3.708],[-9.302,-7.802,-0.108],
[-6.972,-11.784,2.970],[-16.969,-9.845,19.523],[-14.939,-8.016,21.222],
[-21.947,-10.663,2.678],[-23.525,-2.539,7.348],[-6.083,0.038,18.867],
[-5.021,-12.862,4.767],[-23.104,-7.990,19.208],[-4.076,-11.268,24.396],
[-5.349,-8.941,21.440],[-19.947,0.133,15.205],[-17.572,-0.260,10.503],
[-19.151,-10.524,21.543],[-19.306,10.306,2.475],[-13.553,10.314,1.076],
[-8.558,3.442,8.861],[-5.955,10.419,6.875],[-22.849,1.387,2.802],
[-15.443,14.007,5.602],[-7.223,10.430,-6.267],[-12.966,9.322,-8.834],
[-15.987,4.466,-4.697],[-22.013,8.835,2.627],[-23.892,6.123,3.704],
[-11.952,24.307,-7.096],[-11.850,29.643,-12.349],[-21.342,28.071,-8.520],
[-20.074,29.407,-13.400],[-6.635,37.175,-5.680],[-8.107,26.323,-3.977],
[-10.260,15.961,-10.552],[-13.952,24.333,-14.302],[-14.153,30.767,-7.308],
[-19.994,30.360,-7.517],[-13.906,23.173,-17.013],[-19.553,24.621,-17.297],
[-20.382,26.681,-13.954],[-14.553,23.835,1.952],[-23.625,29.014,-13.394],
[-7.453,31.012,0.395],[-4.793,19.731,5.175],[-6.981,16.556,-0.982],
[-23.240,28.061,-16.055],[-11.727,39.100,-18.834],[-14.256,41.527,-16.471],
[-9.679,33.468,-13.815],[-4.596,41.364,-12.397],[-4.402,48.336,-13.948],
[-12.750,47.277,-13.698],[-19.524,46.437,-14.810],[-17.783,34.001,-18.292],
[-15.176,37.622,-11.581],[-10.039,33.826,-19.616],[-18.905,41.317,-16.484],
[-10.431,45.862,-14.033],[-13.304,33.998,-18.000],[-17.456,31.348,-18.364],
[-18.866,39.561,-25.590],[-10.722,41.914,-7.387],[-20.188,39.747,-6.610],
[-16.399,35.297,-6.259],[-21.852,35.074,-12.677],[-7.197,45.820,-23.214],
[-11.901,50.220,-15.026],[-17.452,49.288,-18.609],[-19.195,28.648,-21.028],
[-6.965,48.848,-13.591],[-8.238,58.629,-26.917],[-8.232,58.169,-32.832],
[-8.007,54.031,-25.847],[-17.225,54.169,-18.769],[-10.165,57.937,-17.338],
[-18.850,54.991,-16.772],[-19.338,58.300,-36.681],[-20.950,-33.442,45.879],
[-13.660,-27.110,41.612],[-17.575,-21.843,49.434],[-9.869,-26.856,38.207],
[-13.443,-28.878,54.051],[-19.878,-36.054,46.270],[-4.022,-36.790,42.987],
[-21.448,-29.548,40.102],[-12.687,-31.018,34.729],[-14.731,-32.829,56.491],
[-5.369,-31.879,43.396],[-4.790,-35.357,39.802],[-6.672,-36.523,53.713],
[-20.858,-23.768,45.819],[-6.083,-21.268,44.375],[-13.095,-38.813,48.284],
[-7.579,-39.543,49.450],[-10.664,-38.944,49.477],[-18.294,-10.739,39.745],
[-9.108,-10.821,32.343],[-21.205,-12.545,35.417],[-4.144,-6.962,41.585],
[-13.443,-6.748,39.115],[-12.460,-20.179,45.290],[-19.723,-9.026,41.808],
[-18.791,-23.606,37.937],[-3.926,-9.345,39.940],[-18.224,-15.359,40.135],
[-5.580,-12.615,44.149],[-6.447,-21.116,35.186],[-23.215,-17.118,37.963],
[-11.574,-22.521,40.598],[-22.487,-3.995,24.915],[-16.357,1.811,29.500],
[-16.955,5.442,32.860],[-11.927,5.187,22.368],[-18.295,7.637,21.189],
[-13.455,-8.569,24.191],[-18.086,-0.339,30.170],[-4.944,-5.709,23.877],
[-13.246,2.284,17.455],[-16.628,-4.318,38.133],[-15.909,-0.812,39.586],
[-4.488,1.682,21.241],[-7.564,2.426,32.836],[-4.619,-8.343,23.828],
[-11.727,11.956,29.663],[-19.854,8.362,25.783],[-19.761,-8.573,23.562],
[-5.821,5.016,18.427],[-9.692,-2.634,41.333],[-7.414,11.214,32.132],
[-11.046,15.690,17.159],[-5.110,11.740,22.182],[-17.948,11.533,18.916],
[-8.388,9.619,18.607],[-13.720,24.089,20.517],[-16.164,8.323,12.914],
[-12.714,21.910,21.855],[-15.536,19.723,25.960],[-10.657,17.456,10.076],
[-8.580,17.439,11.701],[-14.040,18.433,7.292],[-17.741,13.576,11.967],
[-8.266,7.440,13.806],[-22.873,9.003,16.080],[-13.959,25.892,16.183],
[-11.256,25.109,16.783],[-16.920,32.692,6.208],[-7.450,33.554,16.990],
[-12.240,29.529,6.773],[-18.109,28.108,2.770],[-20.397,26.102,8.923],
[-14.813,31.113,13.263],[-10.346,27.404,6.514],[-4.177,26.756,4.395],
[-9.679,35.603,10.337],[-19.933,29.731,15.345],[-16.873,34.150,1.449],
[-13.921,32.043,15.668],[-22.075,25.647,2.435],[-23.343,39.351,14.541],
[-18.178,30.797,-1.042],[-16.638,33.012,18.267],[-3.139,30.522,1.605],
[-6.674,25.675,3.127],[-3.310,28.959,21.004],[-5.060,22.372,6.543],
[-15.047,39.296,6.760],[-2.586,45.252,4.671],[-4.081,48.253,4.700],
[-16.695,53.408,0.318],[-19.681,44.686,3.759],[-8.071,53.134,-2.039],
[-9.772,45.718,-7.073],[-5.235,57.740,2.472],[-10.777,53.311,-11.220],
[-13.217,48.244,1.713],[-16.525,54.622,-6.361],[-13.302,49.600,5.826],
[-9.241,47.576,-0.457],[-10.774,47.763,-9.645],[-22.015,53.141,4.892],
[-15.094,42.893,-8.105],[-17.285,48.834,5.459],[-7.181,40.309,-1.413],
[-23.001,43.373,3.303],[-23.573,46.833,2.086],[-13.802,58.042,-3.264],
[-13.054,42.821,2.332],[-13.377,-15.848,55.320],[-7.160,-25.973,54.327],
[-13.643,-30.193,56.903],[-21.133,-29.981,58.383],[-16.438,-28.931,57.770],
[-3.531,-28.249,57.385],[-6.937,-19.625,55.225],[-19.388,-32.355,57.859],
[-20.829,-6.807,51.574],[-6.316,-6.490,57.972],[-18.693,-9.429,58.459],
[-17.883,-3.499,56.596],[-14.013,-14.293,57.675],[-17.262,-5.552,54.554],
[-19.577,-3.616,54.735],[-18.708,3.042,55.956],[-13.577,0.587,54.046],
[-14.912,0.147,44.833],[-17.866,-8.114,56.098],[-15.689,-12.490,49.743],
[-16.030,-7.564,58.162],[-11.091,-13.563,47.424],[-17.126,-2.917,47.888],
[-18.728,-12.860,56.089],[-13.008,-5.987,42.563],[-14.856,-0.263,56.743],
[-10.705,-7.969,43.466],[-15.434,-3.297,56.522],[-13.527,4.128,57.078],
[-21.979,-2.863,43.334],[-19.945,-16.445,49.218],[-7.427,3.794,53.558],
[-13.586,11.564,45.542],[-20.493,8.023,47.009],[-16.030,14.751,42.275],
[-18.806,9.091,40.181],[-13.231,16.743,45.427],[-14.916,17.805,42.478],
[-17.160,16.135,37.373],[-12.965,14.697,51.253],[-15.725,5.931,50.080],
[-4.095,8.904,39.442],[-4.284,0.547,38.512],[-19.987,1.406,45.445],
[-20.720,23.140,36.811],[-17.131,32.251,36.465],[-7.386,30.696,36.483],
[-16.819,34.882,37.588],[-23.403,30.415,34.300],[-12.294,16.434,31.198],
[-12.485,28.254,41.223],[-13.697,23.530,39.103],[-22.234,27.522,39.491],
[-19.830,23.532,28.309],[-9.056,34.615,35.167],[-4.821,28.356,23.282],
[-13.918,26.897,25.495],[-12.291,23.105,41.680],[-6.968,22.953,41.942],
[-13.154,28.720,23.060],[-14.739,27.539,38.243],[-8.791,32.257,33.557],
[-5.428,28.011,41.045],[-6.768,23.032,36.373],[-5.710,24.416,44.017],
[-23.276,24.147,41.461],[-18.338,32.696,30.736],[-6.191,17.953,36.017],
[-13.045,31.206,24.174],[-10.602,47.652,23.832],[-22.717,48.826,23.625],
[-18.231,47.712,17.993],[-16.241,34.813,29.999],[-11.294,38.044,22.583],
[-16.675,34.870,20.171],[-14.431,41.905,19.857],[-17.712,45.157,19.590],
[-19.698,40.909,18.821],[-15.799,53.007,28.999],[-13.509,40.622,34.167],
[-21.834,49.572,20.938],[-5.661,49.045,23.964],[-6.516,48.179,19.273],
[-17.004,50.905,24.068],[-12.583,51.126,21.569],[-17.836,39.647,36.990],
[-4.574,58.298,21.607],[-8.143,52.714,21.053],[-17.450,51.782,18.529],
[-13.187,54.472,12.890],[-15.384,58.049,5.922],[-19.987,51.516,20.231],
[-9.615,6.622,-58.878],[-20.833,7.913,58.502],[-19.164,22.718,59.126],
[-12.145,21.584,55.807],[-9.728,15.475,55.195],[-18.579,17.504,56.098],
[-11.016,22.138,48.594],[-18.204,23.735,52.048],[-16.944,11.555,59.147],
[-16.954,27.776,57.973],[-12.246,28.719,57.825],[-18.883,24.408,56.783],
[-20.249,39.729,48.927],[-20.617,31.579,47.457],[-13.329,29.946,45.274],
[-16.577,34.336,49.434],[-17.131,32.637,55.009],[-19.790,41.559,56.238],
[-7.052,44.556,50.904],[-13.132,41.840,55.350],[-23.108,41.309,41.997],
[-7.230,32.605,53.492],[-8.056,27.598,50.267],[-21.898,47.683,40.726],
[-11.843,59.236,35.819],[-19.757,47.291,43.129],[-6.065,48.474,47.351],
[-9.192,52.208,44.693],[-16.341,51.253,41.825],[-7.432,53.192,48.435],
[-22.155,43.946,41.899],[-17.311,45.243,38.649],[-15.850,52.053,39.027],
[-13.225,43.707,42.470],[-15.923,54.717,32.294],[-18.156,43.350,45.380],
[-15.043,51.310,47.403],[-9.522,54.178,49.732],[-6.688,58.424,41.216],
[-6.656,43.983,34.850],[-6.322,44.256,40.959],[-7.675,47.437,49.765],
[-18.854,56.340,37.945],[-4.904,44.433,44.421],[-20.491,43.483,38.276],
[-13.379,54.495,30.490],[-17.195,48.659,53.514],[-10.246,48.197,53.347],
[7.919,-49.174,-53.571],[6.728,-49.701,-49.028],[-0.009,-49.865,-48.003],
[10.767,-46.225,-51.494],[1.661,-44.615,-48.080],[-2.297,-50.442,-55.375],
[-3.238,-49.177,-52.612],[7.941,-41.804,-50.667],[6.082,-46.252,-50.226],
[13.827,-56.607,-47.767],[9.293,-52.379,-49.697],[2.332,-54.689,-51.121],
[10.671,-55.311,-54.425],[-2.253,-52.009,-50.804],[2.514,-51.709,-51.429],
[15.106,-54.609,-46.718],[3.087,-55.414,-48.642],[7.724,-46.479,-52.756],
[4.809,-47.887,-46.621],[11.990,-56.522,-50.695],[1.019,-52.912,-49.545],
[0.872,-46.775,-45.489],[-1.744,-46.450,-56.178],[12.125,-47.131,-49.116],
[11.708,-56.652,-45.757],[4.082,-58.608,-49.565],[9.637,-43.621,-51.863],
[10.231,-45.886,-54.594],[12.647,-47.531,-45.881],[4.233,-50.785,-53.329],
[6.251,-51.015,-58.839],[16.052,-52.165,-47.439],[9.803,-51.016,-54.617],
[9.198,-52.399,-44.869],[1.875,-49.071,-52.503],[3.223,-45.948,-50.198],
[2.342,-49.095,-48.542],[8.785,-45.492,-47.043],[4.552,-45.962,-43.764],
[9.267,-50.455,-57.372],[1.634,-51.159,-44.799],[6.440,-43.824,-52.168],
[1.117,-44.934,-42.851],[15.029,-57.493,-56.173],[1.622,-50.621,-54.574],
[14.206,-45.807,58.665],[-2.066,-48.284,-45.753],[-1.832,-46.449,-47.896],
[15.855,-44.711,-41.766],[7.239,-44.724,-44.938],[6.329,-39.809,-54.990],
[2.329,-41.542,-53.011],[10.315,-41.443,-57.773],[9.528,-43.230,-55.356],
[12.806,-51.430,-41.603],[-1.430,-45.695,-43.704],[5.603,-56.992,-48.101],
[5.928,-56.939,-44.582],[11.132,-54.368,-43.457],[12.765,-50.604,-58.002],
[6.015,-54.186,-57.718],[2.701,-52.553,-47.015],[4.470,-53.401,-54.235],
[6.794,-50.463,-55.840],[8.127,-53.476,-42.085],[9.890,-54.878,-51.744],
[0.007,-52.428,-46.677],[3.807,-55.107,-59.368],[-3.474,-56.484,-52.142],
[-3.910,-54.989,-57.677],[7.714,-51.769,-47.157],[5.458,-53.567,-51.601],
[-0.831,-55.058,-53.209],[1.286,-53.741,-53.923],[8.476,-55.374,-49.380],
[2.247,-52.768,-56.481],[12.464,-39.747,-41.967],[1.351,-47.772,-41.262],
[4.536,-51.647,-48.981],[1.772,-57.789,-52.078],[5.161,-53.242,-46.432],
[4.398,-47.310,-57.534],[11.605,-49.768,-48.002],[10.544,-52.465,-58.949],
[7.805,-47.627,-44.808],[11.998,-52.818,-56.381],[-0.043,-46.334,-49.879],
[6.957,-44.431,-55.039],[-2.601,-49.517,-43.509],[4.352,-51.486,-42.215],
[10.355,-46.799,-44.511],[15.598,-51.889,-58.375],[15.044,-47.462,-44.081],
[12.457,-43.879,-55.495],[10.000,-49.255,-51.793],[2.107,-44.231,-54.220],
[-0.404,-50.264,-58.919],[3.101,-46.737,-53.764],[4.796,-43.474,-47.017],
[4.726,-42.356,-54.651],[8.650,-48.425,-47.721],[9.411,-45.011,-59.108],
[9.923,-42.697,-45.328],[14.344,-42.835,-57.369],[-1.120,-55.325,-46.102],
[10.249,-47.886,-56.473],[-0.236,-44.155,-45.996],[1.718,-45.652,-59.449],
[2.106,-45.919,-56.717],[0.616,-47.574,-54.881],[9.439,-54.210,-47.060],
[9.846,-55.145,-56.992],[5.258,-44.177,-57.141],[7.200,-52.709,-53.968],
[6.628,-52.131,-43.955],[7.631,-57.754,-54.084],[2.663,-42.134,-56.517],
[2.397,-56.793,-57.291],[13.720,-55.226,-44.275],[11.797,-51.775,-50.407],
[0.158,-44.102,-57.018],[-2.993,-53.177,59.137],[12.025,-43.084,-48.424],
[-0.116,-57.601,-47.964],[-2.266,-46.112,-53.171],[-1.769,-42.321,-42.458],
[4.339,-42.884,-40.923],[3.635,-43.410,-59.464],[9.038,-45.582,-42.185],
[11.737,-43.605,-58.795],[-1.498,-47.355,-41.826],[1.360,-46.206,56.663],
[6.124,-49.469,-42.717],[6.735,-45.613,-59.668],[10.040,-49.521,-44.928],
[-0.338,-50.254,-42.253],[2.073,-48.774,-58.450],[3.981,-50.220,-45.166],
[1.975,-50.029,57.768],[9.207,-48.322,-42.233],[12.090,-46.348,-58.556],
[3.967,-53.480,-43.969],[2.701,-55.841,-45.088],[6.702,-49.806,-45.927],
[10.572,-52.840,-40.973],[9.040,-50.871,-42.797],[1.774,-52.245,-41.903],
[3.629,-58.542,-45.695],[4.990,-54.994,-41.694],[8.473,-55.152,-44.340],
[9.583,-56.451,-41.828],[0.825,-54.994,-40.755],[3.816,-59.106,-42.499],
[5.444,-57.395,-40.108],[6.270,-41.158,-59.054],[-1.401,-47.032,-58.931],
[-2.539,-41.963,-53.294],[-1.950,-43.157,-50.754],[14.688,-42.850,-49.444],
[13.721,-43.015,-46.236],[14.282,-46.847,-40.603],[13.702,-47.089,-55.713],
[13.106,-47.649,-52.719],[14.920,-45.606,-45.964],[11.625,-44.853,-46.268],
[15.216,-49.399,-56.613],[13.791,-53.623,58.944],[13.755,-51.254,-52.867],
[-2.746,-49.607,-48.172],[13.437,-49.576,-43.254],[15.474,-52.343,-42.171],
[12.579,-55.740,-58.208],[11.676,-52.962,-53.752],[11.695,-54.471,-48.922],
[12.234,-53.922,-46.123],[-0.644,-53.174,-43.987],[14.587,-59.488,-41.857],
[13.171,-56.452,-53.236],[-2.453,-56.365,-49.439],[-1.352,-53.852,-48.702],
[14.785,-56.177,-41.958],[14.578,-58.710,-58.883],[-3.516,-55.863,-55.107],
[11.334,-42.207,-53.234],[-1.642,-58.340,-45.876],[-1.784,-44.456,-41.037],
[3.097,-45.798,-46.238],[6.746,-42.390,-43.530],[9.360,-40.696,-42.868],
[13.231,-58.292,-44.016],[-0.698,-40.501,-45.709],[2.457,-41.631,-46.058],
[7.100,-40.392,-47.694],[9.149,-42.836,-48.269],[9.305,-40.495,-46.688],
[-2.200,-43.124,-48.024],[2.482,-41.939,-50.021],[4.833,-43.910,-49.761],
[6.640,-57.857,-50.791],[11.485,-59.022,-51.752],[14.510,-57.546,-51.171],
[-0.992,-57.317,-51.732],[-3.440,58.545,-55.204],[5.814,-56.195,-52.770],
[9.076,-57.742,-51.647],[11.631,-58.194,-54.939],[-2.271,-43.865,-55.828],
[-0.758,-56.351,-58.845],[5.692,-56.165,-55.386],[7.523,-43.181,-58.225],
[13.859,-42.089,-54.149],[1.100,-57.861,-43.950],[2.369,-57.469,-41.023],
[0.479,-41.774,-58.344],[9.758,-58.494,-43.866],[10.924,-42.332,-41.253],
[7.828,-30.688,-52.712],[5.956,-32.689,-49.512],[0.423,-32.033,-47.457],
[11.519,-28.508,-50.591],[0.318,-28.442,-49.411],[-3.848,-34.680,-52.675],
[-1.912,-31.011,-51.903],[9.096,-24.423,-49.744],[6.113,-28.600,-48.697],
[14.420,-36.835,-46.659],[9.539,-35.208,-47.390],[0.055,-36.715,-51.597],
[8.886,-35.912,-50.655],[-2.196,-33.268,-50.496],[4.168,-31.401,-50.872],
[14.180,-34.028,-45.116],[-0.091,-37.289,-47.538],[6.025,-28.630,-51.522],
[4.206,-29.197,-47.098],[11.319,-37.676,-50.990],[-0.936,-35.496,-49.511],
[-1.177,-30.336,-45.090],[-2.817,-28.037,-54.931],[10.119,-26.756,-48.659],
[11.405,-37.546,-47.299],[5.132,-40.141,-50.332],[9.885,-26.333,-51.261],
[10.340,-27.825,-53.292],[11.420,-28.774,-44.320],[6.723,-30.940,-56.284],
[8.452,-31.520,-59.285],[13.396,-31.603,-46.578],[10.937,-30.530,-52.997],
[8.487,-33.781,-44.446],[0.537,-32.475,-52.454],[2.870,-26.565,-48.755],
[2.704,-32.822,-49.109],[7.711,-26.831,-46.949],[3.050,-28.471,-44.556],
[8.359,-32.368,-54.651],[1.000,-31.712,-44.268],[7.174,-25.062,-52.110],
[1.889,-26.363,-43.278],[13.494,-39.480,-56.633],[-1.437,-30.937,-54.809],
[11.951,-24.859,-58.147],[-2.611,-29.906,-47.303],[-3.044,-27.181,-48.481],
[15.802,-26.503,-42.795],[5.319,-26.708,-44.923],[6.380,-23.188,-53.889],
[0.835,-23.878,-53.079],[8.983,-22.809,-58.699],[9.303,-24.545,-55.698],
[11.784,-33.641,-42.439],[-1.895,-28.366,-43.180],[2.131,-38.077,-49.344],
[6.340,-40.557,-45.270],[10.245,-35.651,-43.685],[11.452,-31.101,-58.134],
[6.711,-36.471,-56.556],[1.624,-34.961,-47.953],[1.876,-34.567,-53.675],
[6.913,-33.829,-56.646],[7.831,-35.240,-42.125],[8.192,-33.651,-51.748],
[-1.061,-34.778,-46.779],[4.278,-36.788,59.282],[-3.561,-37.277,-51.775],
[-2.772,-36.707,-56.542],[7.812,-33.443,-47.043],[4.266,-37.463,-51.007],
[-1.192,-34.592,-53.369],[3.632,-34.609,-55.597],[8.692,-37.612,-47.559],
[1.849,-35.739,-58.990],[12.218,-22.007,-41.144],[0.516,-29.636,-42.412],
[4.442,-34.874,-49.960],[1.701,-38.484,-52.481],[4.303,-34.241,-47.246],
[4.233,-30.084,-55.993],[13.208,-30.523,-49.035],[10.222,-37.293,-56.030],
[5.400,-29.991,-43.981],[9.097,-34.935,-55.293],[-0.161,-31.477,-50.064],
[7.673,-28.254,-56.284],[-3.811,-31.264,-44.353],[4.959,-32.336,-42.572],
[9.100,-25.926,-44.603],[14.513,-33.881,-55.599],[14.547,-29.341,-43.139],
[12.917,-26.093,-55.326],[9.683,-31.544,-48.365],[2.132,-25.993,-51.409],
[0.077,-32.045,-57.703],[2.923,-28.807,-50.879],[4.977,-25.690,-47.241],
[3.950,-24.187,-54.458],[9.606,-28.884,-46.842],[10.939,-27.399,-56.648],
[11.938,-24.364,-45.094],[15.776,-24.197,-58.575],[-2.358,-36.166,-44.089],
[9.522,-29.774,-55.197],[-3.068,-27.357,-45.730],[5.250,-27.052,-56.886],
[1.420,-27.990,-56.029],[-0.807,-29.278,-57.208],[5.674,-37.443,-45.821],
[7.357,-36.924,-53.827],[6.538,-25.043,-55.718],[5.786,-34.238,-53.737],
[5.268,-34.922,-44.317],[7.321,-39.509,-52.102],[2.488,-25.146,-58.412],
[1.304,-39.368,-55.109],[12.293,-38.156,-44.135],[11.246,-34.045,-49.288],
[0.317,-27.627,-58.967],[-0.549,-34.432,-58.541],[12.293,-24.337,-48.072],
[-1.574,-40.747,-48.842],[-0.211,-26.667,-52.628],[-2.400,-24.233,-41.591],
[3.122,-24.814,-41.345],[4.883,-24.577,59.337],[9.511,-27.674,-42.332],
[9.287,-25.624,-58.116],[-2.657,-30.075,-41.359],[-0.541,-30.481,59.100],
[4.502,-29.780,-41.530],[7.053,-28.898,-59.163],[9.868,-31.489,-45.545],
[-1.260,-32.159,-42.756],[2.573,-32.311,-56.663],[2.251,-34.042,-45.249],
[5.320,-31.726,-58.384],[7.188,-30.786,-40.789],[13.396,-28.453,-57.749],
[3.097,-36.472,-44.184],[2.633,-37.857,-46.589],[4.126,-31.767,-45.811],
[10.461,-34.640,-40.056],[9.840,-31.759,-42.873],[0.709,-33.861,-40.774],
[4.564,-39.320,-47.550],[5.405,-36.389,-41.284],[7.921,-37.436,-44.126],
[9.378,-37.230,-40.653],[0.343,-37.117,-42.279],[3.132,-40.956,-43.269],
[4.859,-39.605,-40.252],[3.452,-22.528,-57.638],[-2.057,-26.876,-58.047],
[-2.713,-22.019,-54.255],[-2.606,-24.658,-51.677],[14.488,-23.331,-49.094],
[14.607,-24.233,-45.809],[12.420,-28.600,-41.504],[12.826,-28.744,-54.098],
[13.455,-31.805,-53.959],[14.451,-29.206,-46.916],[11.627,-26.816,-46.044],
[14.447,-30.733,-56.476],[13.342,-34.785,-59.462],[11.361,-33.053,-52.294],
[-3.534,-29.940,-50.238],[12.525,-30.995,-43.655],[14.427,-32.830,-42.150],
[11.392,-36.439,-58.302],[11.316,-35.685,-53.508],[13.241,-36.692,-48.995],
[11.253,-33.925,-45.954],[-4.384,-36.014,-42.554],[15.921,-40.390,-42.637],
[13.286,-37.546,-54.108],[-2.695,-38.057,-48.869],[-4.606,-34.107,-49.508],
[14.891,-37.881,-41.744],[13.726,-37.587,-58.635],[-2.407,-39.484,-54.582],
[10.757,-25.065,-53.466],[-3.866,-38.961,-44.633],[-0.097,-27.062,-41.668],
[1.030,-27.440,-46.829],[6.914,-24.127,-44.025],[9.780,-21.792,-43.478],
[12.118,-40.226,-45.908],[-2.448,-22.977,-44.334],[-0.045,-22.679,-46.255],
[7.793,-22.154,-49.159],[9.956,-24.488,-47.023],[11.074,-22.048,-46.417],
[-2.078,-24.625,-47.698],[0.483,-25.016,-48.702],[4.969,-25.240,-50.555],
[6.712,-37.831,-49.948],[12.557,-39.575,-49.168],[13.415,-40.152,-51.908],
[-0.985,-40.066,-52.349],[-1.924,-41.236,-57.186],[4.214,-39.382,-53.199],
[9.430,-39.585,-49.755],[9.353,-39.518,-54.415],[-1.131,-24.822,-55.090],
[-0.375,-39.200,-58.389],[4.579,-37.509,-55.208],[6.098,-21.838,-58.695],
[13.533,-22.145,-54.381],[-0.615,-39.745,-42.897],[2.329,-40.468,-40.584],
[0.674,-21.749,-57.865],[10.092,-39.781,-40.103],[12.828,-25.343,-42.118],
[8.036,-13.312,-52.097],[7.571,-13.054,-49.344],[-1.046,-12.500,-47.645],
[10.689,-10.536,-51.349],[1.908,-11.566,-47.781],[-1.936,-15.214,-53.031],
[-3.102,-13.070,-52.125],[7.797,-6.170,-49.715],[8.417,-10.769,-48.217],
[14.366,-19.243,-47.566],[6.911,-17.548,-49.416],[2.082,-17.255,-51.090],
[9.675,-19.642,-52.926],[-1.930,-16.067,-50.340],[4.746,-12.801,-51.340],
[14.533,-16.140,-44.564],[1.276,-19.888,-46.684],[9.415,-9.706,-53.745],
[0.022,-9.251,-43.434],[11.041,-19.242,-50.816],[-0.191,-17.701,-49.077],
[0.094,-10.864,-45.839],[-1.846,-8.703,-55.068],[9.613,-8.148,-49.289],
[11.313,-19.699,-48.012],[3.954,-21.826,-49.014],[11.936,-6.586,-49.696],
[11.150,-7.553,-52.444],[11.956,-10.768,-45.215],[6.836,-11.378,-53.301],
[6.818,-15.691,-58.451],[15.765,-14.147,-46.581],[10.668,-15.145,-53.131],
[9.260,-14.724,-45.314],[-0.638,-12.359,-52.587],[4.428,-8.403,-50.328],
[3.273,-13.863,-47.771],[8.625,-8.039,-46.859],[4.346,-9.118,-44.075],
[9.096,-14.217,-57.071],[1.919,-13.030,-44.597],[7.974,-7.442,-53.418],
[1.544,-7.087,-44.342],[12.610,-20.386,-57.131],[2.541,-11.154,-55.313],
[15.510,-8.030,-55.446],[-2.919,-11.374,-45.554],[-3.126,-6.708,-48.623],
[15.843,-7.297,-44.055],[3.603,-10.064,-46.472],[5.061,-4.679,-54.107],
[0.471,-5.383,-52.011],[9.158,-6.799,-56.692],[8.395,-4.753,-55.028],
[11.328,-15.877,-43.609],[-2.724,-10.711,-42.870],[1.244,-22.125,-48.459],
[7.154,-22.034,-45.867],[9.725,-17.586,-46.205],[11.325,-13.204,-58.177],
[5.870,-19.585,-56.755],[2.458,-16.762,-48.352],[2.980,-14.963,-52.164],
[7.339,-14.184,-54.617],[7.398,-14.673,-43.140],[8.883,-15.741,-51.084],
[0.079,-16.637,-43.817],[3.335,-18.129,58.367],[-4.070,-17.404,-52.814],
[-1.192,-17.043,-54.861],[8.408,-15.471,-47.634],[5.198,-18.218,-51.936],
[1.029,-16.635,-53.441],[1.143,-18.716,-55.583],[8.612,-19.559,-48.420],
[1.407,-15.825,-57.555],[12.859,-3.810,-43.010],[0.302,-11.737,-43.093],
[5.657,-14.982,-48.856],[0.347,-19.087,-52.232],[2.998,-17.842,-45.922],
[6.711,-12.428,-56.495],[12.820,-12.771,-46.929],[8.483,-17.750,-57.960],
[6.659,-12.170,-44.849],[8.468,-16.537,-55.760],[1.579,-13.355,-50.589],
[7.575,-9.655,-55.753],[-1.830,-13.563,-43.010],[5.700,-13.315,-41.813],
[11.158,-5.746,-42.626],[16.361,-16.159,-55.865],[13.951,-10.289,-43.519],
[11.514,-8.102,-56.942],[10.291,-12.667,-49.479],[2.395,-7.823,-52.141],
[-1.436,-12.141,-58.622],[4.033,-10.686,-53.033],[4.043,-5.635,-45.289],
[2.432,-5.466,-53.832],[9.846,-12.225,-46.151],[9.606,-9.736,-58.647],
[10.509,-5.846,-45.578],[14.084,-6.725,-57.261],[-1.067,-18.120,-46.233],
[10.337,-11.084,-56.091],[-1.746,-8.496,-46.395],[4.892,-9.318,58.831],
[4.628,-9.386,-56.844],[-0.199,-10.878,-55.392],[6.050,-19.571,-46.795],
[9.977,-19.419,-56.419],[6.840,-7.682,-57.773],[5.234,-16.682,-54.168],
[6.321,-16.410,-45.316],[7.668,-20.989,-55.215],[2.592,-6.838,-56.683],
[2.472,-21.013,-55.163],[12.476,-19.653,-45.422],[11.890,-16.401,-50.440],
[0.559,-8.639,-56.547],[-0.562,-17.085,-59.073],[12.183,-4.671,-47.574],
[-1.244,-20.273,-48.516],[-1.276,-6.204,-54.148],[-2.734,-4.976,-42.478],
[2.022,-6.640,-41.115],[4.522,-6.545,-58.408],[8.633,-7.967,-39.338],
[9.239,-7.110,-59.483],[-1.226,-11.438,-40.514],[1.202,-11.475,-58.523],
[3.783,-11.437,-42.793],[6.990,-10.448,-58.254],[7.654,-9.326,-43.778],
[0.046,-14.102,-40.994],[2.846,-13.085,59.098],[3.319,-15.227,-44.804],
[4.550,-15.508,58.940],[10.133,-13.832,-41.018],[12.221,-10.407,-58.193],
[2.915,-18.878,-43.256],[3.605,-19.652,-47.877],[5.661,-10.152,-48.475],
[9.614,-16.856,-41.265],[10.402,-13.055,-43.531],[2.264,-15.702,-41.839],
[4.719,-22.967,-46.676],[5.050,-16.072,-41.952],[7.313,-18.235,-42.337],
[10.507,-18.888,-43.141],[1.026,-17.961,-41.475],[4.073,-22.565,-43.167],
[6.464,-20.795,-42.241],[4.785,-2.785,-57.175],[-1.954,-9.668,-57.710],
[-2.312,-1.681,-54.393],[-1.748,-6.643,-51.095],[16.816,-5.541,-48.968],
[14.348,-5.931,-46.082],[13.424,-9.168,-41.021],[13.943,-10.790,-54.698],
[13.777,-12.674,-52.959],[15.065,-7.600,-48.504],[13.416,-8.726,-46.116],
[13.642,-12.595,-56.942],[15.349,-15.801,-58.660],[12.572,-13.879,-50.773],
[-3.010,-13.699,-49.018],[13.111,-13.236,-43.853],[15.299,-13.710,-41.877],
[13.004,-18.285,-59.040],[11.803,-17.539,-53.492],[13.435,-19.785,-49.999],
[12.453,-17.094,-47.395],[-2.679,-17.899,-42.681],[16.763,-21.671,-40.032],
[13.228,-19.442,-54.329],[-3.112,-18.832,-50.609],[-1.436,-15.225,-47.352],
[13.315,-19.762,-42.623],[15.464,-19.465,-58.350],[-2.015,-19.895,-55.582],
[12.252,-4.814,-52.375],[-2.115,-20.383,-45.095],[-0.452,-6.792,-42.050],
[1.413,-8.679,-48.208],[6.455,-4.381,-42.806],[8.840,-4.566,-43.899],
[13.941,-22.054,-44.306],[-0.666,-4.005,-44.165],[2.668,-6.135,-47.644],
[6.455,-3.562,-48.156],[9.589,-5.207,-48.145],[10.214,-3.066,-46.083],
[0.375,-6.487,-49.191],[-1.180,-8.684,-49.307],[2.615,-5.621,-50.595],
[6.206,-20.073,-49.958],[12.361,-21.502,-51.817],[15.076,-22.320,-51.690],
[-0.223,-21.695,-51.522],[-1.878,-22.926,-57.287],[3.920,-21.463,-52.781],
[9.838,-22.249,-50.995],[10.033,-22.066,-54.922],[-0.203,-6.187,-57.359],
[-1.139,-19.899,-58.914],[3.148,-19.416,-57.695],[6.670,-4.764,-57.545],
[13.497,-5.215,-55.264],[-1.914,-20.609,-42.307],[0.895,-21.151,-42.597],
[1.796,-4.128,-56.796],[9.559,-21.197,-41.043],[13.742,-6.439,-41.730],
[7.978,5.995,-51.892],[7.233,5.981,-49.309],[0.178,5.930,-47.719],
[11.124,8.640,-51.788],[1.683,8.611,-49.036],[-1.706,4.727,-53.097],
[-3.360,5.479,-51.112],[8.777,13.695,-52.162],[6.982,8.453,-48.445],
[14.569,-0.110,-46.957],[9.489,2.859,-47.444],[0.033,0.657,-51.284],
[9.764,0.305,-53.036],[-2.671,2.583,-50.287],[4.175,4.794,-51.167],
[12.791,1.606,-45.310],[-0.146,-2.022,-46.089],[7.541,8.641,-53.486],
[3.093,6.998,-47.278],[13.193,-0.797,-50.399],[-0.352,2.445,-48.480],
[0.968,7.542,-45.445],[-1.675,8.624,-54.756],[11.508,8.639,-48.118],
[12.238,-1.762,-47.546],[3.715,-2.900,-49.552],[10.764,12.149,-50.427],
[8.898,10.832,-52.831],[12.958,8.930,-43.922],[4.934,4.924,-56.420],
[8.773,2.092,-58.950],[14.793,3.437,-45.725],[9.711,4.867,-54.198],
[9.111,4.570,-45.016],[2.958,7.251,-52.373],[4.826,10.464,-50.181],
[2.226,5.824,-49.613],[6.855,11.147,-47.281],[3.536,9.333,-43.693],
[8.698,4.281,-57.107],[-1.180,6.687,-43.795],[4.691,11.856,-52.717],
[0.905,10.357,-43.921],[11.789,-4.147,-57.308],[0.387,6.203,-54.633],
[14.557,11.253,58.946],[-2.669,7.043,-46.995],[-2.383,9.753,-48.053],
[16.534,11.017,-44.101],[5.625,11.370,-44.087],[6.046,13.993,-53.441],
[2.111,15.866,-51.989],[9.909,13.527,-57.747],[9.319,12.832,-54.749],
[11.757,3.087,-41.388],[-1.137,9.144,-45.707],[3.386,-0.432,-48.464],
[7.222,-2.413,-44.643],[8.948,1.828,-45.163],[11.659,4.995,-58.697],
[5.329,1.303,-59.353],[2.168,3.090,-47.685],[2.873,1.748,-52.748],
[6.845,2.155,-56.869],[6.083,2.078,-39.363],[7.988,3.417,-52.605],
[-0.246,3.051,-45.840],[3.375,0.562,-57.156],[-2.218,0.643,-52.802],
[-1.634,0.565,-56.695],[7.329,4.221,-47.187],[3.084,-0.181,-51.069],
[0.302,2.246,-53.604],[3.225,2.665,-55.314],[9.234,-0.185,-47.516],
[1.788,3.294,-58.155],[11.641,13.977,-42.631],[1.136,7.112,-42.672],
[6.846,3.063,-49.546],[1.242,-0.545,-53.270],[5.282,2.340,-47.031],
[5.582,7.634,-55.464],[12.441,6.118,-46.270],[8.974,-0.810,-58.405],
[6.159,7.386,-43.649],[10.355,1.587,-55.478],[0.181,6.087,-51.419],
[7.986,8.477,-57.109],[-2.649,4.602,-44.055],[4.540,4.355,-43.160],
[10.176,9.646,-45.403],[14.146,3.031,-55.398],[15.645,8.015,-44.342],
[14.465,10.467,-54.425],[9.909,6.524,-50.134],[1.999,10.722,-53.323],
[-2.592,7.071,-57.473],[4.760,9.126,-53.177],[5.458,13.434,-46.099],
[2.295,13.334,-52.511],[8.580,6.922,-46.982],[10.633,9.003,-57.473],
[9.640,11.938,-43.368],[14.165,12.898,-55.734],[-3.450,-0.280,-45.006],
[10.189,7.382,-55.487],[-0.342,12.056,-46.241],[4.257,10.473,59.525],
[3.749,9.797,-55.790],[0.959,8.850,-56.252],[6.979,0.438,-46.453],
[8.171,-0.535,-55.751],[5.910,11.352,-55.173],[6.195,1.693,-54.349],
[5.953,2.007,-44.282],[6.781,-1.614,-52.762],[2.616,11.927,-57.854],
[0.658,-1.220,-56.020],[12.248,-0.964,-44.557],[11.331,3.156,-49.718],
[-0.323,9.479,-59.759],[-3.399,2.569,-58.074],[11.985,13.240,-48.168],
[0.971,-3.635,-48.976],[-0.172,12.796,-53.657],[-3.000,13.663,-42.651],
[3.309,13.815,-40.921],[5.455,12.607,-58.050],[8.141,9.484,-43.259],
[12.686,12.717,-58.553],[-3.662,7.884,-42.861],[3.466,7.785,58.677],
[3.922,7.118,-42.021],[6.251,8.361,-59.566],[10.315,7.154,-44.676],
[-0.591,4.541,-41.383],[2.915,6.432,-55.152],[2.922,3.059,-44.842],
[4.223,5.019,-59.076],[8.050,6.506,-41.681],[13.853,7.436,-58.538],
[3.265,-0.619,-43.375],[3.448,0.441,-45.916],[5.043,4.949,-45.992],
[8.910,1.819,-40.572],[11.105,5.341,-42.485],[1.879,3.701,-42.224],
[4.181,-2.476,-45.803],[5.380,-0.111,-41.519],[7.852,0.532,-42.954],
[10.624,0.879,-43.372],[0.929,-0.049,-41.449],[3.941,-3.429,-43.160],
[5.893,-2.647,-40.493],[4.687,15.002,-57.780],[-1.469,10.209,-56.940],
[-2.561,14.677,-53.970],[-4.252,13.431,-50.861],[14.895,13.816,-49.378],
[14.164,12.988,-46.289],[12.726,9.713,-41.088],[16.693,7.199,-54.636],
[12.500,5.622,-54.180],[14.400,8.673,-47.406],[12.384,10.778,-46.447],
[13.711,5.317,-56.666],[14.301,1.176,-58.818],[11.329,3.543,-52.415],
[-1.638,7.384,-49.609],[13.726,6.176,-43.831],[13.780,3.092,-43.300],
[12.234,-0.276,-57.920],[12.464,1.258,-54.075],[12.421,0.982,-48.132],
[13.018,3.789,-47.613],[-1.748,0.834,-42.744],[15.837,-3.742,-41.167],
[12.882,-1.731,-55.458],[-1.148,-1.643,-48.656],[-2.190,4.679,-48.303],
[14.788,-2.279,-43.155],[15.128,-2.544,-56.975],[-3.805,0.577,-55.029],
[11.441,12.100,-53.086],[-2.877,-2.760,-45.900],[-1.542,9.399,-42.417],
[1.873,10.662,-46.561],[6.637,13.687,-42.467],[8.758,14.547,-44.517],
[13.571,-3.256,-45.653],[-0.726,16.528,-46.412],[2.308,12.775,-44.355],
[4.673,14.723,-50.640],[10.115,12.527,-45.893],[10.005,15.473,-47.431],
[-1.299,12.445,-51.158],[1.154,13.263,-48.276],[1.526,11.477,-50.509],
[6.451,-1.490,-50.045],[12.361,-3.491,-49.890],[15.608,-3.241,-50.261],
[-0.444,-2.768,-51.461],[-2.292,-3.992,-55.946],[4.357,-2.908,-52.116],
[9.292,-2.968,-49.666],[10.861,-3.673,-54.717],[-2.329,11.966,-55.084],
[-1.767,-2.131,-58.401],[5.155,-0.632,-55.593],[6.905,13.932,-56.206],
[13.815,15.482,-54.458],[1.184,-2.351,-42.682],[2.668,-4.127,-40.066],
[-0.043,12.780,-56.886],[9.747,-4.079,-40.972],[13.013,12.099,-43.954],
[8.192,24.737,-51.687],[6.981,24.815,-49.266],[-0.683,23.918,-48.944],
[10.047,27.081,-51.434],[0.163,27.678,-50.553],[-3.483,21.716,-53.654],
[-2.461,24.428,-51.922],[8.850,32.366,-50.074],[6.285,27.454,-50.019],
[14.188,17.511,-48.806],[11.361,20.241,-48.488],[1.889,19.346,-50.967],
[11.198,19.330,-52.903],[-2.166,22.213,-50.530],[4.054,22.314,-51.426],
[13.883,20.351,-46.591],[0.567,18.197,-48.607],[7.726,28.128,-52.725],
[2.578,25.549,-45.174],[10.529,18.274,-50.430],[-0.439,20.947,-48.629],
[0.861,27.598,-45.923],[-2.031,28.672,-57.646],[11.686,27.181,-49.141],
[11.408,18.131,-46.594],[8.581,12.684,-48.827],[12.143,31.427,-51.252],
[10.456,27.628,-54.200],[12.050,27.110,-44.752],[8.053,25.024,-56.532],
[7.856,23.013,-58.576],[13.965,22.838,-46.416],[10.908,24.763,-54.666],
[8.961,23.004,-47.399],[3.762,24.976,-52.786],[3.534,27.807,-50.327],
[2.270,25.428,-49.656],[9.228,30.288,-48.040],[4.859,29.907,-44.532],
[9.464,22.733,-55.142],[0.764,23.198,-46.149],[7.291,31.025,-51.886],
[2.304,29.963,-45.384],[13.095,15.085,-57.134],[1.058,24.668,-55.097],
[14.909,29.146,-58.722],[-2.239,26.236,-45.857],[-3.535,29.544,-48.033],
[15.827,29.685,-44.615],[7.862,30.048,-44.851],[6.637,34.087,-53.826],
[0.331,30.860,-54.499],[9.427,31.697,-57.706],[8.844,32.841,-55.232],
[12.878,21.752,-43.630],[-1.592,27.775,-43.305],[2.384,16.236,-48.285],
[7.326,16.090,-46.415],[11.330,19.404,-43.352],[10.381,23.968,-58.482],
[7.900,20.045,-57.830],[2.237,21.076,-48.648],[2.282,22.633,-53.602],
[7.432,22.850,-53.666],[5.715,21.460,-41.195],[9.493,21.489,-50.574],
[0.172,19.791,-46.304],[3.016,18.772,58.920],[-4.304,19.245,-52.405],
[-1.200,19.740,-53.883],[7.715,20.662,-46.372],[7.980,20.324,-55.064],
[-0.755,22.540,-54.005],[2.246,19.579,-55.918],[8.634,18.104,-47.762],
[3.118,21.572,-58.070],[10.243,34.556,-43.478],[-0.319,25.500,-43.172],
[6.645,22.127,-51.131],[1.029,18.261,-53.466],[4.839,20.770,-46.728],
[5.463,26.198,-56.366],[12.296,24.966,-47.781],[10.279,18.298,-57.638],
[6.160,27.581,-43.902],[12.102,19.797,-56.189],[0.364,24.880,-51.455],
[7.610,27.883,-55.841],[-2.268,23.280,-44.260],[3.774,22.685,-42.707],
[8.766,27.294,-44.359],[14.630,21.594,-56.579],[14.488,27.435,-43.617],
[14.732,29.398,-55.911],[9.699,24.913,-49.358],[3.404,28.602,-53.079],
[1.339,22.139,-56.083],[5.986,26.348,-53.676],[6.081,30.509,-47.067],
[1.918,32.391,-52.592],[8.042,25.328,-46.519],[9.610,28.753,-57.980],
[10.631,29.477,-44.984],[13.718,32.681,-58.742],[-1.120,18.797,-43.949],
[10.440,26.577,-56.945],[-0.326,29.382,-47.247],[5.170,28.800,-58.218],
[3.416,27.570,-55.655],[0.133,26.756,-56.519],[6.089,18.221,-44.860],
[7.220,16.549,-54.185],[6.790,30.619,-56.704],[4.612,20.015,-49.530],
[5.093,20.591,-43.951],[7.302,16.129,-51.191],[4.532,30.879,-54.837],
[3.875,17.744,-57.319],[14.990,18.071,-45.779],[11.768,23.347,-49.991],
[0.915,29.171,-56.683],[-0.859,21.477,-58.180],[12.023,31.698,-48.460],
[-0.378,15.668,-49.662],[-0.993,30.982,-52.248],[-1.000,32.060,-42.623],
[3.552,32.745,-40.102],[5.603,31.281,-58.996],[9.078,29.720,-42.373],
[12.027,29.658,-58.879],[-2.884,25.695,-42.212],[1.852,26.138,58.690],
[4.103,26.975,-42.293],[7.363,27.108,-58.449],[10.205,24.851,-44.786],
[-0.592,23.083,-41.743],[3.053,24.733,-58.082],[3.486,22.672,-45.418],
[4.363,23.529,58.403],[7.008,25.705,-41.964],[13.643,26.763,-59.136],
[2.371,20.377,-44.280],[1.894,17.627,-45.758],[6.374,23.251,-45.785],
[9.849,20.671,-40.747],[10.401,23.234,-42.512],[1.689,21.477,-41.750],
[4.549,16.672,-46.808],[4.440,18.944,-41.726],[7.599,20.451,-43.201],
[8.564,18.470,-41.923],[-0.566,20.057,-41.344],[4.564,14.982,-43.396],
[3.528,16.694,-40.839],[3.555,33.640,-57.541],[-0.394,30.363,59.124],
[-3.101,33.588,-54.420],[-3.161,30.963,-50.349],[13.657,33.197,-50.078],
[14.104,32.478,-46.107],[11.918,28.981,-41.875],[13.070,26.057,-55.954],
[13.262,23.710,-53.611],[14.209,27.313,-48.419],[12.414,29.269,-47.150],
[15.356,25.316,-57.364],[14.676,20.849,58.608],[11.227,22.066,-52.747],
[-1.963,26.203,-49.564],[13.114,25.018,-43.107],[14.729,22.904,-41.626],
[12.913,18.048,-59.203],[14.359,18.857,-55.203],[13.286,19.168,-50.754],
[10.770,20.731,-45.728],[-3.363,20.571,-43.766],[13.119,15.148,-40.855],
[12.073,17.655,-54.637],[-0.831,19.977,-51.232],[-4.348,22.875,-48.851],
[15.648,16.502,-42.498],[15.246,18.734,-58.099],[-2.295,17.164,-55.371],
[11.703,33.464,-54.115],[-3.126,16.451,-45.021],[-0.934,29.783,-41.056],
[2.868,30.072,-48.483],[6.601,32.377,-44.013],[10.842,31.778,-43.172],
[11.781,14.949,-45.241],[-1.765,32.712,-45.614],[1.273,32.299,-47.020],
[7.245,32.913,-47.825],[9.079,31.985,-45.773],[11.387,33.650,-45.989],
[-1.416,31.825,-48.401],[1.473,29.774,-51.773],[3.238,32.771,-49.863],
[6.680,18.253,-49.763],[12.505,15.403,-49.808],[14.439,16.270,-51.922],
[-2.313,17.157,-52.067],[-0.324,15.672,-56.884],[5.020,16.687,-52.731],
[11.145,16.380,-52.257],[11.192,15.143,-55.189],[-2.381,30.782,-55.588],
[-0.723,14.595,59.536],[4.383,18.758,-54.492],[6.697,33.949,-57.287],
[12.293,31.183,-55.729],[-1.302,16.565,-42.794],[1.549,16.088,-43.172],
[-1.159,31.707,-57.735],[8.842,14.878,-41.226],[13.107,30.278,-43.839],
[8.412,43.690,-51.932],[6.833,42.732,-49.244],[-0.174,42.493,-47.455],
[9.975,46.267,-51.337],[1.886,46.150,-49.475],[-2.883,40.545,-54.390],
[-2.553,41.536,-51.880],[7.926,51.291,-50.645],[7.233,46.555,-48.696],
[14.383,36.375,-47.708],[7.621,39.218,-47.749],[0.959,37.302,-50.205],
[10.677,36.930,-53.290],[-1.625,38.788,-51.628],[5.635,41.340,-52.226],
[14.071,40.703,-44.784],[-1.544,37.165,-47.361],[7.713,45.812,-53.313],
[2.140,43.614,-46.098],[13.452,35.956,-50.549],[0.648,39.944,-49.121],
[0.472,45.886,-46.850],[-1.889,45.817,-56.406],[12.203,48.335,-48.810],
[10.653,36.820,-45.578],[5.130,34.713,-49.809],[9.469,49.248,-52.712],
[12.494,47.569,-53.921],[12.949,45.951,-44.316],[6.682,43.565,-54.593],
[8.044,41.973,-58.667],[14.492,42.103,-47.164],[11.029,42.891,-53.382],
[9.227,41.765,-44.763],[1.620,43.858,-52.970],[4.390,48.709,-47.250],
[2.815,43.056,-50.354],[7.104,48.959,-47.870],[4.333,46.778,-43.878],
[9.932,42.687,-56.654],[0.098,41.523,-44.814],[6.925,49.423,-52.875],
[0.254,51.210,-44.027],[12.095,35.315,-57.349],[1.871,41.852,-55.617],
[12.491,48.091,-58.694],[-2.281,45.763,-46.525],[-3.012,48.403,-47.445],
[15.589,48.995,-44.891],[6.387,47.338,-42.034],[6.093,51.396,-54.611],
[0.143,51.590,-53.925],[8.476,51.315,-58.845],[8.414,50.073,-55.182],
[11.336,41.510,-42.810],[-1.492,46.537,-43.987],[1.900,34.824,-48.610],
[8.335,34.604,-45.322],[9.465,39.236,-43.691],[10.861,42.181,59.356],
[7.675,37.533,-58.982],[2.851,40.556,-48.040],[0.043,40.900,-51.715],
[7.442,40.092,-54.439],[6.236,41.633,-40.231],[8.073,39.307,-51.673],
[-0.374,38.606,-45.249],[3.254,35.709,-59.477],[-3.432,36.915,-53.088],
[-0.977,38.278,-56.427],[8.805,41.568,-47.645],[4.065,36.999,-51.138],
[-0.392,40.285,-54.707],[3.098,38.069,-55.708],[9.109,36.255,-47.820],
[2.735,41.423,-59.086],[11.632,51.164,-42.213],[-0.103,45.515,-41.565],
[5.245,39.783,-49.641],[1.357,35.677,-55.757],[5.013,39.225,-46.754],
[5.837,44.390,-57.082],[11.467,45.441,-47.358],[10.235,37.259,-58.469],
[8.509,46.053,-43.667],[10.554,40.582,-54.805],[-0.470,43.439,-50.181],
[7.060,46.905,-55.596],[-2.564,42.472,-45.257],[5.762,41.897,-42.937],
[9.102,49.116,-45.123],[14.538,40.420,-56.413],[15.061,45.182,-42.988],
[13.209,49.373,-55.903],[9.017,44.594,-49.105],[1.866,46.722,-52.424],
[-3.445,44.966,-58.358],[4.585,45.059,-53.166],[6.258,46.718,-46.001],
[2.091,51.271,-51.697],[8.583,45.261,-46.405],[9.401,47.086,-58.223],
[11.218,50.899,-45.370],[15.351,49.725,-58.215],[-3.251,37.626,-45.337],
[10.652,45.678,-55.479],[-0.460,48.881,-46.586],[3.937,46.360,-57.853],
[3.467,47.031,-55.082],[1.950,44.785,-56.851],[6.596,36.340,-46.710],
[8.588,37.642,-55.884],[5.476,48.648,-56.757],[5.838,37.946,-53.181],
[7.036,39.716,-44.859],[7.974,36.153,-52.444],[2.171,49.706,-56.151],
[3.987,35.365,-55.578],[12.209,38.801,-44.789],[10.352,41.651,-50.411],
[0.770,47.315,-55.058],[-0.808,39.643,-58.987],[11.029,50.841,-49.266],
[0.175,33.574,-50.714],[-2.831,47.243,-53.761],[-2.292,52.716,-40.602],
[2.538,48.788,-42.696],[5.607,48.683,-59.771],[9.135,48.594,-42.089],
[10.527,49.217,-56.672],[-2.820,44.846,-41.296],[1.683,45.742,-59.415],
[3.764,44.179,-43.244],[6.966,46.115,-58.799],[10.448,46.904,-45.399],
[-1.717,42.899,-42.422],[4.740,41.107,-55.079],[2.506,41.243,-43.600],
[5.658,41.075,-59.020],[7.526,44.458,-41.831],[14.528,45.866,-59.806],
[3.902,37.772,-43.418],[3.067,37.210,-46.042],[4.669,43.048,-46.881],
[8.227,38.833,-40.433],[10.320,44.110,-44.336],[2.193,38.805,-41.608],
[4.681,34.234,-46.867],[5.705,39.499,-41.492],[8.816,36.845,-42.530],
[10.097,35.752,-40.003],[-0.608,37.902,-42.596],[3.222,34.824,-43.007],
[6.027,36.670,-42.314],[4.139,51.151,-57.573],[-0.578,47.032,-58.476],
[-3.763,52.313,-51.788],[-2.343,48.524,-50.710],[15.511,51.745,-48.614],
[14.107,49.903,-47.271],[13.702,47.149,-41.471],[13.492,45.956,-57.121],
[13.870,42.884,-53.746],[13.860,45.692,-48.608],[13.298,47.518,-46.330],
[13.435,42.999,-57.169],[13.240,40.374,59.249],[12.518,39.574,-52.742],
[-3.565,43.245,-50.162],[13.080,43.352,-43.461],[14.882,41.909,-41.621],
[12.674,37.803,59.079],[11.660,38.407,-56.491],[12.671,38.816,-48.598],
[11.528,41.275,-47.040],[-2.329,39.777,-43.692],[12.393,33.567,-39.365],
[12.688,36.316,-55.070],[-1.537,36.120,-50.198],[-2.017,40.277,-47.729],
[14.602,34.704,-42.275],[15.069,36.311,59.446],[-1.200,36.174,-54.580],
[11.534,50.828,-52.916],[-3.430,34.790,-45.088],[-0.246,48.328,-42.452],
[2.123,48.120,-45.807],[7.347,50.579,-43.416],[8.942,52.582,-43.620],
[13.112,35.956,-45.352],[1.312,53.932,-46.582],[1.874,51.188,-46.385],
[5.865,52.023,-47.698],[8.830,51.037,-47.373],[10.126,53.243,-48.204],
[0.807,50.231,-48.654],[0.557,49.150,-51.291],[3.106,49.078,-50.706],
[7.283,37.216,-50.031],[10.859,34.213,-51.102],[14.248,33.694,-52.689],
[0.390,34.031,-53.819],[-1.957,35.937,-57.319],[4.008,34.127,-52.633],
[10.924,36.851,-50.648],[9.529,34.572,-57.485],[-1.854,49.771,-54.511],
[0.362,35.877,-59.069],[5.826,37.132,-56.386],[6.890,53.245,-57.300],
[13.857,51.903,-54.462],[-0.233,34.640,-43.806],[2.271,35.709,-40.473],
[-0.586,51.911,-58.072],[10.049,33.172,-40.798],[12.412,48.592,-44.032],
[13.684,55.467,-50.236],[8.771,58.127,-48.021],[2.970,56.156,-52.003],
[9.378,56.846,-52.889],[-2.723,58.929,-52.500],[13.359,58.034,-46.893],
[1.498,56.511,-47.000],[12.158,57.243,-51.955],[0.277,57.708,-49.078],
[12.171,55.254,-47.889],[4.992,51.075,-50.163],[13.928,52.516,-59.128],
[12.359,57.731,-42.323],[3.210,52.493,-48.512],[6.560,53.169,-44.516],
[9.715,57.706,-44.898],[5.816,55.887,-59.043],[2.724,58.626,-48.101],
[4.428,58.312,-52.403],[7.206,58.789,-52.573],[0.501,58.878,-46.363],
[3.880,54.791,58.110],[-1.750,56.228,-52.762],[-1.488,56.146,-56.151],
[5.415,55.778,-54.050],[0.494,-59.312,-56.101],[2.876,58.080,-55.577],
[9.885,55.914,-49.333],[5.132,-58.753,-58.096],[6.542,58.306,-49.898],
[1.458,54.308,-53.064],[6.614,-59.345,-46.942],[9.702,57.040,-58.758],
[9.186,-59.370,-55.923],[14.357,58.990,-56.481],[-1.840,55.960,-46.027],
[7.154,56.131,-46.029],[8.319,55.490,-56.868],[5.210,59.376,-55.192],
[6.075,58.965,-44.279],[7.920,54.355,-54.009],[2.932,55.392,-55.128],
[11.493,55.763,-44.419],[-1.004,58.870,-58.172],[-0.840,52.752,-48.214],
[2.606,57.641,-43.892],[4.130,56.199,-46.241],[9.487,58.651,-40.725],
[1.839,58.545,-41.163],[4.118,53.036,-45.398],[4.878,57.007,-42.340],
[7.532,57.171,-42.601],[8.919,55.127,-41.721],[-1.872,55.516,-40.447],
[2.900,51.507,-42.010],[4.810,53.192,-40.609],[12.026,-59.769,-58.680],
[13.539,55.756,-58.989],[12.347,58.211,-54.669],[13.194,59.546,-49.368],
[10.552,-58.659,-47.420],[-1.635,55.292,-43.229],[13.604,52.851,-42.452],
[12.455,55.119,-53.962],[-1.498,55.349,-48.967],[-2.076,-59.261,-48.883],
[14.664,56.620,-43.139],[16.258,53.708,59.560],[-3.671,54.893,-54.404],
[-3.937,52.367,-47.647],[13.227,52.730,-45.874],[5.495,55.174,-51.053],
[13.362,49.791,-51.172],[11.072,54.566,-51.380],[-0.480,52.334,-51.246],
[-2.658,52.316,-55.358],[4.493,52.866,-52.401],[9.402,52.535,-52.839],
[11.340,52.194,-55.181],[0.847,54.900,-57.039],[4.886,53.382,-55.502],
[-0.062,58.088,-43.719],[0.931,53.762,-42.313],[9.220,50.982,-41.216],
[8.765,-49.794,-33.728],[6.594,-50.732,-31.180],[-0.693,-50.591,-29.632],
[11.699,-46.454,-33.029],[0.338,-45.989,-30.082],[-1.618,-53.805,-34.765],
[-1.914,-50.169,-32.653],[7.901,-41.120,-31.503],[4.830,-47.015,-29.447],
[13.937,-57.091,-29.289],[8.948,-53.415,-29.138],[2.082,-55.195,-32.624],
[10.698,-54.696,-34.734],[-2.499,-52.819,-32.504],[2.314,-51.944,-32.890],
[15.090,-52.958,-26.647],[0.438,-56.585,-28.878],[6.283,-47.423,-35.583],
[3.323,-49.030,-28.642],[11.740,-56.710,-32.587],[-0.049,-53.111,-30.195],
[2.509,-47.333,-26.677],[-2.014,-46.847,-36.504],[10.164,-46.073,-30.613],
[12.833,-57.619,-26.836],[4.473,-58.259,-31.379],[8.761,-44.113,-32.393],
[9.880,-45.357,-34.655],[11.492,-47.439,-25.716],[6.405,-50.573,-35.776],
[8.558,-51.726,-39.858],[13.144,-50.369,-27.145],[11.377,-50.062,-35.286],
[8.906,-51.569,-26.559],[2.742,-48.972,-33.337],[4.030,-45.195,-31.548],
[3.502,-51.074,-30.311],[6.112,-43.995,-28.934],[2.756,-45.991,-24.401],
[10.419,-49.341,-38.117],[-0.532,-49.382,-26.790],[6.334,-44.142,-35.067],
[1.934,-43.386,-24.398],[12.236,-57.797,-37.563],[1.679,-50.079,-35.990],
[13.415,-43.273,-40.919],[-4.315,-49.013,-26.584],[-2.794,-45.957,-29.998],
[16.755,-44.000,-25.257],[4.589,-45.329,-26.596],[7.263,-41.111,-36.925],
[-0.799,-42.596,-33.570],[9.174,-41.980,-38.743],[8.593,-43.699,-36.896],
[11.603,-51.742,-23.973],[-0.239,-46.050,-25.563],[1.375,-59.147,-27.649],
[8.290,-59.482,-27.292],[9.064,-54.778,-26.180],[12.193,-48.998,-40.417],
[7.409,-55.558,-40.538],[2.621,-53.771,-29.359],[1.376,-52.719,-35.188],
[8.854,-52.086,-35.410],[6.704,-52.121,-23.250],[8.694,-52.285,-32.160],
[-0.358,-52.126,-26.353],[3.498,-53.214,-39.943],[-2.634,-56.242,-35.050],
[-1.517,-56.006,-37.702],[6.473,-51.584,-28.596],[4.783,-56.222,-33.380],
[-0.679,-54.949,-32.100],[3.379,-54.477,-37.499],[7.865,-55.728,-29.437],
[1.140,-52.974,-38.980],[11.595,-39.330,-24.215],[-0.397,-49.338,-24.242],
[4.907,-53.359,-30.797],[2.377,-57.276,-35.719],[6.303,-54.030,-28.315],
[4.938,-48.451,-38.978],[11.008,-50.047,-28.380],[9.679,-54.885,-38.913],
[6.620,-46.863,-25.718],[10.660,-53.279,-37.118],[0.976,-48.697,-30.249],
[7.919,-48.031,-37.920],[-2.799,-51.288,-25.337],[4.410,-50.925,-24.074],
[9.029,-46.333,-26.333],[13.646,-52.204,-37.452],[14.456,-46.776,-25.052],
[12.845,-46.261,-38.394],[10.582,-48.608,-30.911],[2.675,-45.134,-35.088],
[-1.130,-51.103,-39.431],[3.510,-48.186,-35.803],[3.634,-43.227,-28.477],
[1.722,-41.095,-33.832],[8.613,-48.688,-28.508],[9.997,-47.354,-39.848],
[10.687,-43.530,-25.779],[14.396,-42.119,-38.511],[-2.439,-55.795,-25.057],
[11.975,-47.321,-35.624],[-1.260,-45.138,-27.672],[4.374,-45.799,-40.384],
[3.227,-46.442,-37.924],[0.446,-47.925,-37.826],[6.540,-56.327,-27.004],
[9.444,-56.488,-36.379],[6.277,-45.473,-37.900],[7.192,-54.830,-37.654],
[5.635,-53.145,-25.761],[6.396,-55.108,-34.973],[3.892,-43.651,-38.259],
[4.084,-57.449,-37.884],[12.894,-54.752,-25.129],[10.894,-51.294,-30.760],
[0.784,-45.079,-37.151],[-2.012,-53.872,-39.410],[10.559,-42.572,-29.801],
[0.496,-57.761,-31.719],[-0.672,-45.189,-35.004],[-1.142,-43.704,-23.066],
[4.575,-42.363,-22.375],[7.143,-43.871,-40.241],[9.539,-45.949,-21.744],
[10.976,-44.280,-39.502],[-2.842,-48.366,-23.069],[2.043,-44.654,-40.320],
[4.386,-48.081,-24.132],[6.858,-47.639,-40.863],[9.572,-49.101,-26.152],
[-1.241,-51.883,-23.329],[3.877,-51.602,-35.168],[2.819,-52.415,-26.684],
[3.193,-50.371,-40.161],[6.658,-49.208,-23.816],[15.200,-48.692,-38.780],
[3.882,-55.640,-26.247],[3.336,-56.834,-29.046],[5.186,-50.221,-26.797],
[9.233,-53.378,-22.290],[9.027,-50.567,-24.029],[1.228,-53.777,-24.723],
[3.970,58.300,-25.774],[4.083,-54.132,-23.717],[6.800,-57.424,-24.339],
[9.694,-56.156,-22.845],[-0.125,-56.376,-23.489],[3.843,-58.482,-23.945],
[4.297,-56.508,-22.169],[4.839,-41.172,-38.029],[-2.827,-46.351,-39.378],
[-3.151,-41.388,-35.629],[-3.129,-41.892,-32.725],[13.834,-41.910,-29.848],
[13.172,-42.787,-27.154],[12.290,-46.926,-22.808],[14.586,-48.691,-35.930],
[13.591,-50.755,-34.151],[12.842,-47.856,-29.159],[13.795,-45.525,-27.212],
[16.272,-51.142,-38.051],[12.421,-53.983,-39.067],[11.486,-52.322,-33.469],
[-3.151,-50.415,-29.944],[13.105,-49.321,-24.127],[14.634,-51.537,-22.688],
[11.888,-56.652,-40.305],[13.427,-54.461,-34.441],[14.290,-53.820,-29.662],
[12.004,-53.136,-27.730],[-4.026,-55.326,-22.785],[16.659,-57.336,-22.190],
[13.413,-57.169,-34.620],[-1.971,-55.588,-29.725],[-2.907,-53.116,-30.049],
[14.142,-56.353,-22.828],[14.408,-55.910,-39.145],[-2.837,-58.663,-36.179],
[10.769,-42.344,-34.098],[-2.548,-58.809,-28.423],[-0.716,-46.605,-23.071],
[1.360,-44.538,-27.397],[7.384,-42.172,-25.241],[8.848,-39.553,-25.717],
[13.219,59.417,-25.056],[-0.388,-40.400,-26.496],[2.079,-41.904,-26.703],
[5.546,-41.282,-29.676],[8.244,-42.535,-27.772],[11.469,-39.285,-28.726],
[-1.568,-43.412,-29.977],[1.270,-44.016,-31.926],[3.419,-42.216,-31.358],
[6.834,-57.633,-32.322],[11.862,-58.640,-30.604],[14.542,-57.916,-32.054],
[-0.110,-59.071,-34.606],[-1.758,-58.512,-39.198],[4.654,-58.871,-35.022],
[9.198,-57.702,-31.222],[9.615,-57.438,-34.030],[-3.075,-44.157,-36.923],
[0.844,-57.088,-38.659],[6.693,-57.510,-36.695],[7.394,-41.333,-40.928],
[11.185,-42.664,-36.914],[-3.177,-58.889,-25.364],[2.233,-59.742,-22.190],
[0.035,-43.156,-39.429],[10.330,-59.688,-21.936],[13.096,-44.358,-23.973],
[8.734,-31.707,-33.371],[6.958,-31.865,-30.671],[0.340,-33.116,-29.002],
[12.489,-27.988,-31.224],[0.300,-28.300,-31.490],[-3.368,-35.677,-34.298],
[-1.645,-32.056,-34.099],[8.366,-23.290,-30.884],[6.417,-28.849,-30.964],
[13.868,-38.111,-28.653],[9.449,-34.861,-29.615],[2.050,-36.985,-31.978],
[9.278,-36.491,-34.772],[-1.313,-33.493,-31.477],[4.019,-31.733,-31.758],
[15.379,-35.802,-28.119],[0.205,-36.571,-27.828],[9.378,-28.441,-32.723],
[4.672,-30.557,-27.745],[11.143,-37.387,-32.897],[0.404,-35.679,-30.529],
[2.755,-28.192,-28.880],[-1.123,-27.596,-36.846],[10.924,-27.781,-28.863],
[12.082,-36.974,-26.813],[3.852,-39.333,-31.123],[11.444,-25.401,-32.176],
[10.484,-26.412,-34.868],[11.815,-29.631,-25.280],[6.551,-31.686,-36.286],
[7.811,-33.871,-39.184],[13.981,-32.643,-28.467],[12.913,-30.117,-35.882],
[9.225,-32.538,-25.899],[1.773,-30.178,-32.942],[4.570,-26.551,-30.892],
[1.327,-32.517,-31.495],[9.893,-25.750,-30.116],[2.377,-27.619,-26.203],
[9.125,-32.428,-36.796],[-1.005,-32.882,-26.418],[6.565,-26.103,-35.128],
[3.003,-25.310,-24.731],[14.314,-39.247,-38.521],[1.167,-31.925,-35.648],
[14.160,-25.978,-38.656],[-1.722,-28.979,-29.603],[-3.367,-25.274,-30.685],
[15.071,-25.789,-24.936],[5.702,-26.114,-25.680],[7.268,-23.160,-35.542],
[-0.559,-23.004,-33.249],[10.217,-24.194,-39.237],[7.718,-24.909,-37.671],
[12.838,-34.110,-23.083],[-0.210,-29.568,-27.404],[1.350,-39.777,-29.090],
[6.961,-39.707,-28.330],[11.524,-34.463,-25.405],[9.972,-30.331,-40.536],
[6.535,-36.608,-38.390],[2.690,-34.769,-29.219],[2.665,-33.549,-34.004],
[7.429,-33.905,-35.112],[7.341,-31.983,-23.688],[8.447,-34.899,-32.516],
[-1.635,-35.473,-25.766],[2.572,-36.770,-40.536],[-3.106,-38.010,-33.052],
[-3.151,-36.187,-37.262],[7.972,-34.857,-27.258],[4.088,-35.909,-33.414],
[-0.650,-34.732,-34.692],[1.887,-35.742,-35.203],[8.971,-37.668,-28.405],
[2.356,-35.892,-38.021],[13.569,-21.586,-22.496],[1.249,-29.813,-25.076],
[5.548,-34.824,-31.600],[-0.274,-37.489,-34.773],[6.039,-36.414,-28.027],
[4.469,-30.148,-37.247],[12.125,-30.303,-27.794],[7.309,-38.658,-40.107],
[8.011,-28.337,-25.596],[8.629,-35.591,-37.060],[-1.070,-30.855,-31.834],
[7.952,-29.125,-37.250],[-3.119,-32.832,-25.061],[4.669,-31.593,-24.488],
[8.092,-25.264,-25.880],[15.067,-34.284,-37.777],[14.056,-28.032,-26.105],
[13.517,-26.915,-35.797],[10.106,-30.477,-30.054],[2.564,-28.056,-35.721],
[-1.383,-34.105,-38.570],[6.040,-28.981,-34.545],[4.455,-25.588,-28.441],
[1.745,-23.021,-35.623],[9.337,-30.036,-27.581],[11.992,-30.113,-38.837],
[10.810,-27.079,-24.963],[14.536,-22.944,-38.482],[-1.344,-37.933,-26.383],
[11.130,-28.239,-36.596],[-1.012,-25.945,-29.384],[4.621,-27.131,-41.181],
[3.543,-27.736,-38.192],[-1.322,-30.581,-37.634],[6.177,-37.761,-25.676],
[7.956,-38.468,-35.789],[6.083,-26.998,-37.916],[6.233,-36.472,-35.321],
[5.852,-34.537,-25.580],[5.828,-38.791,-33.913],[2.318,-24.569,-37.989],
[2.821,-38.943,-38.439],[12.764,-36.729,-24.180],[11.329,-32.680,-30.511],
[0.979,-27.097,-38.681],[-0.870,-36.372,-40.132],[11.257,-23.374,-29.539],
[-1.752,-40.871,-30.746],[-1.109,-25.940,-34.252],[-1.447,-24.169,-23.738],
[1.612,-24.968,-21.883],[8.411,-26.193,-40.266],[7.431,-25.851,-23.233],
[11.073,-26.708,-39.873],[-2.579,-30.028,-23.161],[2.196,-29.030,-40.351],
[4.276,-28.579,-24.690],[6.996,-28.194,-41.982],[9.199,-29.947,-24.035],
[-0.499,-31.581,-23.947],[2.126,-31.962,-38.086],[2.002,-32.771,-27.120],
[3.209,-32.361,-40.602],[9.267,-31.271,-21.808],[14.164,-30.127,-40.190],
[2.516,-35.914,-25.443],[2.850,-37.369,-28.435],[7.094,-30.981,-26.350],
[10.235,-34.521,-22.755],[13.747,-32.703,-25.854],[0.502,-33.425,-21.581],
[4.394,-39.606,-27.109],[5.190,-36.310,-23.567],[8.333,-35.822,-24.246],
[9.800,-38.352,-22.102],[0.376,-34.700,-24.327],[4.162,-41.113,-25.056],
[5.343,-39.199,-22.491],[4.840,-22.539,-37.938],[-1.901,-28.084,-39.564],
[-2.150,-22.114,-35.794],[-3.306,-23.559,-33.303],[13.978,-23.589,-29.423],
[13.916,-23.640,-26.468],[12.766,-27.471,-23.140],[15.228,-28.701,-37.579],
[14.127,-31.603,-34.067],[14.438,-28.668,-29.338],[11.912,-25.915,-27.168],
[14.779,-31.832,-36.979],[13.109,-33.597,-39.570],[10.928,-33.084,-33.134],
[-3.468,-30.607,-30.759],[14.517,-30.359,-24.402],[15.116,-32.561,-22.750],
[11.816,-36.837,-41.447],[11.954,-35.637,-35.307],[11.976,-35.288,-30.495],
[11.514,-33.104,-27.852],[-4.168,-34.925,-23.908],[14.465,-40.053,-21.576],
[14.810,-36.166,-35.244],[0.068,-38.652,-31.162],[-3.347,-34.152,-29.711],
[14.918,-38.546,-23.827],[14.735,-35.890,-39.992],[-2.468,-38.852,-35.546],
[10.759,-23.172,-33.672],[-3.005,-39.849,-27.932],[-1.251,-27.170,-24.385],
[1.488,-25.577,-27.761],[6.805,-22.627,-23.335],[9.692,-22.600,-24.834],
[14.080,-39.708,-26.290],[-0.665,-21.748,-28.093],[2.920,-23.073,-28.927],
[6.761,-21.900,-28.710],[8.779,-23.699,-28.251],[11.802,-21.942,-26.879],
[-1.395,-23.466,-30.847],[1.742,-25.931,-31.239],[3.409,-23.937,-32.028],
[6.395,-37.320,-31.509],[12.988,-41.075,-33.008],[13.183,-38.672,-34.498],
[-1.047,-39.870,-33.178],[-2.259,-41.379,-38.176],[4.297,-41.154,-34.462],
[10.337,-40.241,-32.083],[10.586,-39.833,-37.517],[-0.103,-24.760,-36.821],
[-1.018,-38.157,-37.651],[5.476,-38.675,-37.122],[7.110,-23.801,-40.218],
[14.745,-24.491,-35.229],[-1.021,-38.990,-23.918],[2.639,-40.635,-22.785],
[0.284,-24.225,-40.770],[10.582,-41.091,-22.240],[13.605,-24.942,-22.803],
[8.912,-13.406,-33.512],[5.817,-10.742,-29.850],[-0.841,-12.684,-28.636],
[10.525,-9.680,-32.696],[1.751,-8.909,-30.173],[-2.142,-13.917,-35.878],
[-2.116,-11.592,-33.628],[9.025,-4.375,-32.545],[6.413,-8.205,-30.038],
[14.002,-19.028,-29.684],[9.891,-16.067,-29.314],[2.958,-17.411,-32.402],
[9.506,-18.926,-33.767],[-1.719,-15.801,-32.587],[3.527,-14.731,-32.505],
[15.197,-16.988,-28.216],[0.874,-19.070,-29.014],[6.460,-10.260,-32.748],
[4.378,-11.472,-27.635],[11.988,-18.170,-31.918],[0.256,-17.852,-31.306],
[0.919,-10.721,-27.442],[-2.131,-8.575,-36.640],[11.344,-9.676,-28.186],
[11.615,-18.637,-28.114],[5.070,-19.949,-29.850],[10.775,-6.907,-33.043],
[9.711,-9.144,-35.243],[12.125,-10.554,-25.707],[6.627,-14.374,-37.714],
[7.801,-14.892,-40.409],[14.117,-14.026,-27.716],[12.285,-11.420,-34.896],
[11.595,-14.252,-28.429],[3.026,-11.691,-31.926],[3.837,-7.852,-31.505],
[2.596,-12.677,-29.265],[8.375,-8.790,-27.294],[2.510,-10.974,-25.062],
[10.700,-14.372,-36.825],[2.715,-13.519,-25.911],[5.671,-7.988,-34.659],
[3.070,-7.687,-24.035],[13.037,-20.716,-38.790],[1.529,-13.828,-37.128],
[14.723,-8.417,-38.681],[-2.757,-11.268,-26.902],[-1.340,-10.136,-29.222],
[16.385,-7.626,-24.631],[4.641,-8.582,-26.217],[7.285,-3.700,-36.560],
[0.077,-4.464,-33.889],[9.831,-5.149,-38.434],[9.902,-5.478,-35.454],
[11.972,-14.872,-25.238],[-0.936,-9.419,-25.421],[2.805,-21.038,-30.883],
[8.819,-20.760,-27.222],[10.414,-17.283,-26.260],[12.007,-11.931,-41.675],
[6.293,-17.637,-39.454],[2.217,-16.818,-29.775],[2.750,-15.545,-34.992],
[8.253,-15.546,-35.729],[8.104,-16.049,-22.333],[10.123,-16.068,-33.991],
[-0.745,-16.673,-25.675],[1.873,-18.055,-38.902],[-1.890,-19.600,-32.632],
[-2.266,-17.986,-38.393],[7.667,-16.628,-27.017],[4.310,-19.439,-33.416],
[0.501,-16.979,-33.980],[3.936,-18.434,-36.804],[7.393,-18.020,-29.766],
[3.403,-15.995,-37.856],[11.085,-5.503,-23.960],[0.354,-10.650,-22.855],
[6.515,-14.884,-30.117],[-0.387,-18.152,-36.216],[5.140,-16.887,-28.707],
[5.147,-11.003,-36.391],[12.313,-11.711,-29.829],[9.851,-19.251,-39.377],
[6.708,-10.311,-25.303],[10.599,-16.959,-38.603],[0.187,-11.613,-31.653],
[7.779,-9.675,-37.308],[-2.658,-13.559,-25.237],[4.716,-15.041,-23.980],
[9.077,-6.415,-25.767],[15.112,-15.012,-37.430],[14.049,-9.484,-24.063],
[13.164,-7.430,-34.959],[9.409,-12.011,-31.277],[-0.292,-8.059,-32.300],
[-0.939,-13.149,-38.390],[3.371,-9.061,-33.829],[3.523,-6.267,-28.074],
[1.950,-4.188,-37.188],[7.764,-12.235,-28.703],[10.798,-9.678,-40.421],
[11.604,-8.102,-24.256],[15.442,-5.615,-39.292],[-2.333,-18.554,-24.107],
[10.610,-10.858,-37.005],[1.379,-5.812,-24.958],[4.864,-8.892,-41.103],
[3.057,-8.781,-38.972],[-0.387,-11.370,-36.209],[6.358,-19.342,-27.165],
[9.420,-18.090,-36.353],[4.802,-6.516,-37.763],[7.007,-19.373,-36.920],
[5.052,-16.920,-25.954],[6.532,-20.866,-34.450],[1.589,-6.753,-37.656],
[1.677,-20.212,-36.693],[14.722,-18.121,-25.738],[11.230,-15.806,-31.527],
[-0.365,-8.527,-38.822],[-0.757,-15.803,-38.124],[10.995,-5.323,-29.109],
[-0.284,-20.879,-30.587],[-1.524,-6.701,-34.608],[-1.409,-5.159,-23.010],
[4.017,-4.781,-21.459],[7.047,-7.145,-42.242],[9.302,-8.413,-22.974],
[11.586,-7.281,-38.901],[-2.163,-11.573,-23.623],[1.830,-10.080,-41.117],
[4.918,-12.171,-25.031],[6.984,-10.190,-40.378],[10.414,-12.344,-26.963],
[-0.775,-14.429,-22.784],[2.226,-11.379,-38.285],[2.526,-16.142,-27.159],
[4.402,-13.636,-39.419],[7.245,-13.186,-24.273],[14.010,-10.946,-38.464],
[2.437,-17.769,-24.279],[3.213,-18.810,-27.495],[5.782,-14.046,-26.676],
[9.889,-16.881,-20.586],[9.497,-13.090,-22.403],[2.317,-15.373,-22.790],
[4.886,-21.571,-26.559],[5.909,-17.555,-23.461],[8.126,-18.796,-25.265],
[9.498,-18.955,-22.615],[-0.038,-17.184,-23.162],[4.179,-22.796,-24.468],
[4.158,-20.214,-23.340],[3.942,-1.642,-38.677],[-1.748,-8.898,-41.065],
[-2.585,-4.352,-33.655],[-2.623,-6.513,-31.948],[14.901,-4.053,-30.103],
[14.784,-5.919,-26.974],[12.411,-10.004,-20.520],[15.699,-11.453,-36.176],
[14.381,-12.735,-34.129],[13.729,-9.235,-29.585],[12.546,-7.788,-26.833],
[13.566,-13.267,-39.603],[15.316,-16.189,-41.012],[12.505,-14.786,-34.645],
[-2.726,-11.870,-30.859],[13.755,-12.852,-25.247],[15.209,-14.110,-22.698],
[12.444,-17.393,-41.085],[12.314,-17.847,-35.397],[14.097,-16.234,-30.604],
[12.788,-16.419,-27.276],[-3.840,-16.103,-25.249],[16.688,-21.025,-24.245],
[13.948,-21.219,-36.324],[-2.572,-17.640,-30.782],[-2.403,-15.090,-29.976],
[15.116,-18.977,-23.181],[14.781,-18.997,-40.634],[-2.727,-19.630,-36.019],
[11.662,-4.392,-33.190],[-1.811,-18.898,-27.316],[0.217,-7.964,-22.577],
[1.792,-8.168,-26.328],[5.792,-7.410,-23.817],[9.812,-3.643,-25.164],
[13.690,-20.728,-25.265],[-0.732,-4.265,-25.735],[1.293,-5.052,-27.576],
[5.650,-4.093,-28.229],[8.239,-5.144,-28.866],[11.011,-3.113,-27.600],
[-1.267,-7.350,-28.722],[1.645,-6.093,-32.370],[4.469,-5.478,-30.458],
[7.446,-19.557,-31.959],[12.735,-21.223,-30.259],[13.897,-20.509,-33.096],
[0.150,-20.740,-34.678],[-1.690,-23.002,-39.159],[4.348,-23.164,-34.431],
[10.529,-20.865,-32.122],[11.209,-23.540,-36.543],[-1.065,-5.813,-36.972],
[0.521,-20.902,-39.510],[4.311,-20.735,-35.895],[6.075,-5.455,-40.325],
[14.370,-5.327,-36.215],[0.393,-20.285,-24.285],[2.466,-22.317,-22.240],
[0.014,-4.880,-39.428],[9.198,-21.766,-21.290],[13.683,-6.523,-24.059],
[10.031,5.677,-34.966],[6.396,6.574,-31.104],[-1.060,4.094,-29.764],
[11.087,8.189,-32.250],[-0.221,10.854,-30.717],[-3.764,3.529,-34.890],
[-2.123,5.556,-33.780],[7.924,14.126,-32.403],[6.409,9.435,-30.290],
[15.499,-0.119,-29.090],[9.957,2.470,-29.819],[1.826,1.315,-32.654],
[9.298,1.625,-35.967],[-1.146,3.335,-32.298],[4.311,3.933,-32.652],
[14.657,2.078,-27.526],[-0.296,0.238,-28.058],[9.296,8.780,-34.269],
[4.103,8.182,-30.606],[12.292,-0.160,-32.633],[1.393,1.314,-30.089],
[1.597,7.507,-28.719],[-2.729,10.311,-36.789],[11.528,9.697,-30.049],
[10.856,-0.122,-30.070],[5.182,-2.910,-31.020],[10.191,11.645,-31.507],
[11.643,11.153,-34.452],[12.775,8.595,-25.655],[6.723,8.427,-34.412],
[7.663,4.218,-40.244],[13.461,5.415,-29.288],[12.719,6.174,-35.393],
[9.834,4.068,-26.198],[1.901,8.366,-33.570],[2.637,10.305,-31.951],
[1.435,3.769,-31.688],[7.902,12.078,-28.962],[3.009,9.206,-26.637],
[9.527,4.643,-37.971],[1.427,5.086,-26.831],[6.223,13.393,-34.983],
[3.496,11.741,-24.373],[14.021,-3.051,-38.011],[0.249,6.023,-36.079],
[13.865,11.237,-38.889],[-2.685,7.969,-27.545],[-3.023,12.113,-30.441],
[15.426,11.773,-23.195],[5.540,10.197,-25.396],[9.252,15.093,-35.178],
[-0.778,14.814,-35.055],[10.750,12.344,-36.820],[8.717,11.780,-38.675],
[12.902,2.779,-23.014],[-0.794,9.935,-25.749],[1.722,-3.064,-30.177],
[7.160,-2.768,-26.433],[9.151,1.556,-26.124],[11.953,5.169,-38.870],
[7.060,-0.571,-39.308],[2.580,4.044,-29.094],[2.438,4.640,-35.119],
[7.615,3.836,-35.723],[8.343,3.429,-24.066],[9.218,2.766,-32.608],
[-1.359,3.363,-26.883],[3.709,2.047,-40.769],[-2.855,0.972,-33.398],
[-2.544,1.060,-37.277],[7.950,3.796,-28.294],[4.788,0.762,-33.991],
[-0.068,3.487,-34.780],[2.839,1.138,-37.806],[8.146,-1.324,-29.670],
[1.848,4.137,-38.873],[11.192,14.690,-23.396],[0.584,6.795,-24.482],
[6.041,3.130,-30.644],[0.892,0.715,-35.352],[5.357,0.500,-29.665],
[5.324,7.442,-36.480],[10.674,5.874,-29.389],[9.703,-0.563,-39.347],
[5.308,7.627,-26.758],[11.033,2.503,-38.466],[-0.105,6.525,-32.028],
[7.760,8.462,-37.633],[-4.060,6.268,-26.106],[3.386,5.355,-24.163],
[8.571,9.975,-26.468],[16.037,3.047,-38.262],[15.547,9.039,-24.507],
[13.948,9.828,-34.831],[10.644,5.268,-32.014],[3.770,10.082,-34.858],
[-1.456,5.934,-38.950],[4.812,6.995,-33.651],[3.675,11.916,-26.983],
[3.097,12.864,-34.136],[8.987,9.549,-29.641],[10.325,7.538,-38.274],
[10.504,12.257,-24.873],[12.875,14.143,-38.379],[-2.166,0.497,-25.711],
[12.013,8.923,-36.611],[-0.651,9.898,-28.253],[4.690,9.771,-40.983],
[4.016,9.595,-37.815],[0.312,8.109,-37.821],[7.444,0.126,-27.401],
[8.646,-1.081,-36.788],[6.310,13.123,-37.940],[6.793,1.085,-36.717],
[5.807,2.628,-26.936],[7.610,-2.223,-34.347],[2.816,12.032,-36.784],
[1.539,-1.657,-36.782],[12.254,0.073,-24.655],[11.839,2.962,-31.710],
[1.280,10.511,-38.815],[0.090,2.064,-39.785],[12.866,13.118,-28.843],
[-0.739,-4.305,-30.449],[0.309,12.313,-35.484],[-1.155,14.063,-23.269],
[3.408,12.281,-21.618],[6.663,11.471,-41.137],[10.146,9.523,-24.210],
[10.352,11.019,-41.048],[-2.092,7.830,-23.992],[0.802,9.512,-41.121],
[2.970,8.080,-23.161],[8.058,9.292,-40.184],[8.962,6.820,-27.359],
[1.213,4.590,-22.708],[3.737,6.339,-38.421],[3.132,2.833,-26.055],
[4.253,4.668,-40.375],[7.280,5.998,-23.557],[13.217,7.087,-40.651],
[3.622,0.246,-24.541],[3.029,1.192,-28.022],[5.241,5.536,-28.612],
[10.635,2.055,-23.821],[11.126,6.713,-25.858],[3.287,2.849,-22.866],
[3.920,-1.450,-28.520],[5.761,2.850,-24.008],[7.491,0.548,-24.437],
[9.656,-0.982,-24.946],[-1.062,0.990,-23.275],[3.940,-4.111,-25.423],
[7.062,-1.021,-22.229],[5.201,15.460,-39.109],[-1.711,9.971,-39.317],
[-2.978,15.346,-33.383],[-0.291,10.366,-33.322],[13.685,14.645,-31.036],
[15.553,13.816,-26.120],[12.311,8.985,-22.662],[15.181,9.075,-37.355],
[14.264,4.531,-32.367],[14.176,8.270,-30.141],[11.542,10.813,-27.235],
[14.787,5.716,-38.363],[13.547,3.031,-39.453],[12.080,2.945,-34.468],
[-2.567,6.106,-31.117],[12.381,4.995,-24.331],[15.024,5.060,-23.705],
[12.265,0.480,-40.864],[12.003,0.657,-36.819],[12.755,1.888,-29.749],
[12.073,3.577,-27.699],[-3.560,2.021,-23.281],[15.629,-2.222,-21.316],
[12.902,-1.680,-35.832],[-0.088,-1.150,-30.667],[-3.648,3.083,-30.178],
[13.065,-1.965,-22.622],[14.878,-0.012,-38.892],[-2.140,-0.932,-35.302],
[9.764,12.877,-33.822],[-3.628,-1.383,-26.862],[-0.456,11.265,-23.459],
[1.837,12.015,-29.156],[6.157,12.785,-25.188],[9.425,14.860,-25.264],
[12.925,-3.104,-25.839],[-1.177,15.533,-28.466],[0.752,14.122,-27.607],
[6.352,14.108,-29.760],[7.917,14.692,-27.559],[9.979,15.708,-29.185],
[-1.495,14.194,-30.748],[0.815,13.241,-31.822],[5.228,13.539,-32.154],
[6.328,-0.380,-32.035],[13.590,-1.234,-30.669],[14.226,-2.919,-33.257],
[0.525,-1.471,-33.700],[-2.695,-3.771,-36.471],[4.562,-5.173,-35.396],
[10.257,-2.185,-32.003],[10.873,-3.192,-36.781],[-1.983,13.231,-37.217],
[0.459,-1.728,-39.469],[4.155,-0.771,-36.047],[7.953,15.101,-38.480],
[13.416,12.481,-36.231],[-1.761,-2.025,-24.120],[2.715,-2.216,-21.636],
[1.082,12.952,-39.510],[9.393,-3.597,-22.343],[13.322,12.670,-24.928],
[9.362,24.220,-33.402],[6.158,24.701,-30.492],[0.639,24.203,-27.922],
[11.234,27.306,-33.481],[-0.496,28.520,-31.497],[-1.274,22.180,-35.656],
[-3.497,24.098,-32.439],[8.804,31.827,-31.431],[6.415,27.744,-30.096],
[13.747,19.041,-29.119],[9.615,20.549,-29.701],[2.526,20.003,-32.047],
[10.242,17.873,-34.295],[-1.910,21.695,-32.574],[3.762,22.963,-33.172],
[15.886,19.607,-26.752],[0.423,18.940,-28.796],[7.375,29.415,-35.769],
[3.106,25.505,-28.485],[11.667,17.697,-31.662],[-0.367,21.204,-30.136],
[0.724,26.948,-27.371],[-3.841,28.438,-36.259],[11.437,29.619,-30.223],
[12.144,17.463,-27.379],[3.726,15.576,-30.938],[10.744,30.006,-32.866],
[9.537,28.005,-35.493],[12.090,26.742,-26.099],[7.138,25.328,-34.495],
[8.072,22.478,-39.768],[13.447,22.793,-27.461],[11.089,22.725,-35.847],
[10.043,22.158,-25.800],[1.766,24.680,-32.541],[4.704,28.939,-32.012],
[2.375,23.065,-30.431],[8.536,29.937,-28.208],[3.454,28.279,-25.273],
[7.960,24.138,-37.439],[0.301,23.225,-25.331],[8.392,31.247,-34.153],
[1.781,30.530,-25.055],[13.962,16.547,-37.707],[0.481,24.321,-35.935],
[14.959,29.738,-40.187],[-2.826,26.081,-27.470],[-2.767,28.578,-30.200],
[15.638,30.132,-25.336],[5.997,28.339,-27.317],[8.170,33.441,-36.070],
[1.085,33.064,-34.090],[8.527,31.994,-38.273],[10.199,31.465,-36.053],
[12.764,21.576,-24.988],[-0.755,29.473,-25.766],[1.060,15.335,-30.115],
[5.538,16.333,-28.768],[9.231,19.711,-26.854],[9.053,24.832,-40.337],
[6.683,17.883,-40.041],[2.964,21.254,-28.662],[2.095,21.461,-34.968],
[6.513,23.001,-33.178],[7.928,23.698,-23.305],[7.060,21.914,-35.886],
[-1.749,22.798,-27.014],[1.253,18.681,-39.780],[-2.268,19.332,-33.431],
[-1.949,18.967,-39.487],[6.966,22.163,-28.751],[4.365,18.311,-33.049],
[0.061,19.339,-35.304],[2.174,18.894,-37.186],[7.906,18.498,-29.305],
[2.538,21.436,-39.155],[11.007,34.081,-23.875],[0.321,26.612,-24.272],
[4.813,21.319,-30.933],[1.856,17.108,-34.742],[4.900,18.764,-29.830],
[4.593,26.362,-36.308],[11.966,25.332,-28.689],[9.327,18.156,-38.977],
[6.864,25.657,-26.593],[9.631,21.421,-37.857],[-1.484,26.041,-31.910],
[7.265,26.676,-36.950],[-2.725,24.650,-25.440],[4.257,21.871,-23.562],
[7.761,29.310,-25.724],[14.554,21.622,-37.353],[14.485,27.199,-24.833],
[12.503,29.549,-37.614],[9.990,25.805,-30.750],[3.311,29.855,-34.502],
[-2.126,24.806,-36.845],[3.609,26.702,-33.896],[4.401,30.798,-27.513],
[3.582,32.534,-35.537],[10.460,28.043,-27.942],[8.464,28.753,-38.123],
[10.758,31.732,-25.197],[15.084,31.518,-36.610],[-0.716,19.863,-25.699],
[10.854,25.606,-35.670],[-2.701,29.751,-27.306],[2.522,28.577,-41.126],
[3.077,28.128,-38.124],[0.820,26.189,-38.072],[6.753,18.753,-26.658],
[8.295,18.167,-36.371],[5.451,29.118,-37.491],[4.670,20.383,-36.180],
[4.718,21.179,-26.335],[6.651,16.333,-35.189],[3.088,30.770,-37.686],
[1.996,16.112,-38.988],[13.729,19.116,-24.942],[10.895,20.556,-34.055],
[0.583,28.510,-39.231],[-1.000,22.456,-39.208],[11.593,31.532,-28.024],
[-3.451,16.945,-31.136],[-1.237,30.695,-35.389],[-3.938,33.310,-23.754],
[2.121,31.091,-20.930],[6.393,29.962,-41.279],[8.528,28.079,-23.264],
[10.159,29.530,-39.756],[-1.967,25.952,-23.069],[1.562,25.642,-40.494],
[4.897,25.877,-23.569],[6.393,27.154,-39.555],[9.402,24.603,-26.970],
[-1.505,22.518,-23.481],[2.628,24.089,-37.686],[1.981,21.448,-26.091],
[3.826,23.536,-40.187],[9.264,25.854,-24.375],[11.783,25.488,-40.563],
[3.363,19.265,-24.605],[3.956,18.680,-27.288],[4.900,23.947,-26.715],
[10.253,21.798,-21.952],[11.369,24.268,-24.812],[1.340,22.381,-22.809],
[2.543,16.786,-28.478],[6.737,21.065,-23.772],[8.019,20.409,-21.450],
[10.562,18.441,-22.385],[-2.139,18.500,-23.940],[4.191,14.463,-25.734],
[4.043,15.541,-23.426],[4.790,35.073,-39.951],[-2.269,27.543,-39.834],
[-3.534,31.818,-34.960],[-2.162,30.926,-32.804],[14.509,32.106,-29.814],
[14.908,32.497,-26.675],[12.938,28.136,-22.716],[12.671,26.848,-37.439],
[13.389,25.134,-34.128],[15.588,26.528,-29.244],[14.799,28.709,-27.569],
[15.121,23.865,-38.881],[13.649,19.958,-39.381],[13.845,21.599,-34.217],
[-2.533,24.144,-29.458],[13.978,23.919,-24.690],[14.461,23.064,-22.258],
[11.891,18.347,-40.922],[11.424,18.720,-36.911],[12.387,20.213,-31.382],
[11.727,20.512,-27.684],[-3.436,21.225,-21.259],[15.286,16.084,-21.646],
[14.135,18.404,-33.878],[-1.876,19.099,-30.545],[-3.899,21.917,-30.542],
[12.972,16.649,-24.075],[14.678,17.695,-40.232],[-2.588,17.473,-36.625],
[11.211,32.738,-33.820],[-1.865,17.564,-26.610],[-1.700,28.522,-23.278],
[1.675,30.430,-28.328],[4.525,30.849,-24.634],[7.791,33.180,-24.920],
[12.986,14.972,-26.306],[-0.871,34.751,-26.901],[2.211,33.214,-27.539],
[5.826,32.879,-28.761],[9.138,32.801,-28.976],[10.771,34.134,-27.115],
[-0.745,31.296,-29.694],[1.736,30.599,-31.929],[5.995,31.843,-32.170],
[6.689,17.861,-31.493],[10.848,14.795,-32.061],[12.564,15.843,-33.939],
[-0.428,16.178,-32.336],[-3.281,15.348,-38.108],[4.401,15.652,-33.734],
[8.985,16.603,-32.125],[11.200,15.911,-37.051],[-0.725,31.765,-38.192],
[-1.958,16.103,-40.245],[4.420,17.343,-36.691],[5.660,31.912,-38.324],
[12.580,32.327,-36.527],[0.250,15.896,-25.241],[1.381,15.303,-22.513],
[1.377,31.114,-39.846],[8.811,15.695,-22.606],[13.888,31.211,-23.488],
[9.225,43.010,-33.226],[7.006,44.011,-29.410],[-2.029,41.121,-30.470],
[9.837,47.357,-32.748],[2.527,46.819,-29.800],[-1.907,40.861,-36.219],
[-2.157,42.534,-34.028],[8.871,50.758,-31.202],[5.715,46.563,-29.243],
[15.064,36.562,-28.354],[11.075,39.027,-29.301],[2.295,38.886,-32.556],
[9.200,37.092,-32.758],[-0.470,40.644,-32.617],[2.163,41.450,-33.461],
[15.067,40.132,-25.542],[0.370,37.100,-27.579],[6.358,46.306,-34.344],
[3.714,44.525,-29.546],[11.664,37.760,-33.301],[0.258,39.486,-29.942],
[0.965,44.974,-28.442],[-3.029,45.587,-36.658],[12.240,47.469,-30.191],
[11.696,35.237,-29.293],[5.299,34.446,-31.048],[11.498,49.218,-32.132],
[9.743,47.827,-35.434],[11.609,48.138,-24.300],[6.611,43.439,-35.027],
[7.751,40.558,-38.315],[15.008,43.646,-28.518],[11.257,43.065,-35.395],
[10.461,41.469,-27.064],[1.878,45.603,-34.497],[4.722,48.412,-30.790],
[2.099,43.086,-31.231],[8.728,47.704,-27.777],[2.626,46.205,-25.146],
[10.242,42.814,-38.212],[1.253,41.620,-25.397],[7.512,49.428,-34.465],
[1.465,49.060,-24.498],[14.172,34.269,-37.259],[-0.740,44.525,-35.559],
[12.869,48.768,-38.460],[-1.482,43.775,-28.897],[-3.514,48.186,-29.744],
[14.104,49.111,-25.975],[5.550,46.896,-26.532],[6.635,52.754,-36.609],
[0.119,50.198,-33.723],[9.651,49.589,-38.901],[9.433,50.469,-36.202],
[12.576,40.612,-23.925],[-0.166,47.031,-25.790],[2.799,34.214,-30.028],
[7.889,34.703,-27.298],[10.357,38.646,-25.283],[12.227,40.503,-39.153],
[7.377,36.170,-40.044],[0.752,42.315,-27.973],[4.178,39.607,-34.557],
[7.642,40.812,-34.695],[5.338,39.804,-23.633],[6.780,38.524,-33.162],
[1.160,39.879,-27.455],[4.000,37.954,-39.625],[-3.530,37.680,-33.205],
[-2.481,39.266,-38.710],[8.689,40.933,-28.952],[4.313,37.038,-32.012],
[0.432,39.366,-34.775],[3.102,38.163,-36.462],[9.084,36.706,-29.182],
[2.550,40.362,-39.482],[11.185,50.806,-23.433],[0.476,44.309,-24.764],
[6.925,40.760,-30.930],[1.784,36.042,-34.130],[4.972,39.506,-28.848],
[4.733,44.369,-37.026],[12.551,44.586,-28.643],[9.908,38.634,-38.581],
[7.711,45.433,-26.329],[9.850,40.740,-36.252],[-0.235,43.925,-31.959],
[8.148,45.226,-36.598],[-2.987,43.907,-25.983],[3.846,42.186,-23.641],
[6.773,49.443,-27.205],[14.780,39.966,-37.135],[15.392,45.464,-26.884],
[15.387,47.663,-37.541],[10.673,43.492,-30.329],[2.283,48.778,-34.899],
[-1.611,43.793,-38.051],[4.018,43.974,-33.123],[4.175,50.098,-25.728],
[3.420,50.655,-33.345],[9.589,44.802,-28.176],[10.225,45.741,-39.344],
[10.670,49.942,-26.329],[14.144,51.853,-39.672],[-1.924,37.073,-25.528],
[11.282,45.752,-36.799],[-1.304,48.800,-27.376],[4.063,47.502,-40.271],
[2.913,47.880,-37.426],[0.624,46.114,-38.348],[6.413,37.473,-27.479],
[9.393,36.327,-37.151],[5.577,47.225,-36.952],[8.083,38.660,-36.477],
[7.123,40.421,-26.438],[7.858,36.332,-34.997],[4.411,49.634,-38.656],
[3.063,35.351,-36.858],[13.225,36.536,-24.935],[9.332,40.006,-32.228],
[-0.845,48.142,-39.265],[-0.261,39.401,-40.170],[11.140,50.444,-29.195],
[0.484,33.222,-31.125],[-1.336,48.128,-34.889],[-2.341,51.596,-23.505],
[2.658,51.092,-21.002],[6.812,49.362,-40.361],[8.868,47.803,-24.956],
[11.620,48.961,-40.733],[-2.707,45.341,-23.013],[3.262,45.204,-38.907],
[4.138,44.621,-22.659],[7.086,46.078,-39.375],[10.011,43.761,-25.576],
[-0.499,42.669,-22.752],[3.461,41.997,-37.324],[3.779,40.575,-25.886],
[3.846,42.363,-40.883],[6.472,43.803,-23.783],[13.682,44.251,-40.768],
[2.135,37.820,-25.668],[3.151,37.429,-28.183],[5.390,42.746,-27.782],
[10.999,39.145,-22.373],[10.921,42.588,-23.201],[1.036,40.212,-23.152],
[4.282,34.967,-27.312],[6.017,37.411,-22.588],[7.223,38.108,-25.008],
[9.498,36.230,-24.030],[-0.906,37.798,-22.515],[4.104,33.363,-25.023],
[4.140,36.123,-24.370],[4.024,54.640,-38.333],[-4.150,46.653,-38.900],
[-5.211,51.492,-33.084],[-2.375,50.035,-32.766],[13.200,51.879,-30.334],
[15.287,51.465,-27.322],[12.662,46.068,-22.867],[13.939,44.850,-35.979],
[14.044,42.024,-35.803],[14.813,46.509,-29.712],[13.113,47.333,-27.472],
[13.951,42.945,-38.149],[14.712,39.713,-39.958],[12.566,40.202,-34.327],
[-4.218,42.783,-29.908],[13.085,43.681,-24.356],[14.820,41.677,-22.687],
[11.564,38.053,-41.205],[11.268,38.140,-35.935],[13.605,39.827,-29.345],
[12.577,39.246,-26.659],[-3.608,38.606,-23.559],[15.288,34.197,-22.452],
[13.646,36.831,-36.143],[-0.864,35.766,-29.897],[-4.505,40.816,-31.578],
[12.150,36.524,-22.385],[15.700,36.879,-40.783],[-3.134,37.926,-36.008],
[11.663,52.102,-35.156],[-3.529,34.977,-26.298],[-0.930,47.778,-22.686],
[2.294,48.793,-27.366],[6.635,49.295,-24.261],[7.768,51.464,-25.096],
[13.503,34.587,-27.362],[-0.549,50.773,-25.456],[1.685,52.106,-29.058],
[5.649,51.329,-29.105],[8.588,51.052,-28.521],[8.947,53.638,-28.654],
[-0.509,50.114,-29.585],[2.128,49.604,-30.239],[6.385,50.066,-32.136],
[7.460,34.811,-32.914],[11.342,32.977,-30.930],[13.309,34.183,-34.692],
[-1.753,35.001,-34.020],[-1.336,34.179,-36.846],[3.800,34.024,-33.106],
[11.709,35.341,-32.187],[10.818,34.295,-36.387],[-2.730,48.665,-37.377],
[0.089,37.105,-38.887],[6.345,36.368,-37.253],[7.197,51.062,-38.415],
[13.941,50.084,-36.460],[-1.286,34.267,-23.668],[2.335,34.826,-22.563],
[0.010,52.201,-39.028],[9.838,33.312,-21.303],[14.085,49.508,-23.032],
[14.264,56.167,-30.132],[9.259,58.162,-30.415],[1.591,56.318,-32.745],
[10.608,57.734,-36.732],[-0.446,58.720,-31.663],[15.434,56.460,-27.266],
[-0.357,55.997,-28.100],[12.267,56.341,-33.952],[-0.415,56.220,-30.734],
[12.379,54.714,-27.703],[3.733,54.429,-30.333],[11.464,51.989,-39.714],
[13.438,-59.221,-22.457],[-0.371,53.477,-29.669],[6.466,52.740,-26.940],
[10.328,57.556,-24.384],[6.986,56.702,-39.496],[2.474,58.253,-30.777],
[2.405,-59.169,-32.998],[8.477,59.114,-33.976],[-0.692,-59.104,-25.838],
[3.722,56.652,-40.045],[-2.634,54.240,-33.467],[-0.106,55.369,-37.354],
[4.399,55.853,-33.378],[-0.681,56.681,-34.128],[3.056,57.879,-34.846],
[10.506,56.209,-29.070],[2.562,-59.555,-37.213],[6.606,58.413,-30.919],
[1.926,54.411,-34.718],[4.264,59.080,-28.864],[9.187,55.563,-38.375],
[9.279,-59.102,-37.657],[15.421,58.064,-37.546],[-1.583,57.073,-25.956],
[7.694,56.298,-25.618],[8.409,56.338,-35.518],[6.248,57.577,-34.401],
[6.209,57.702,-27.615],[6.864,54.270,-34.495],[2.814,56.752,-37.400],
[12.309,56.694,-26.098],[-0.250,57.955,-39.801],[-0.089,53.964,-32.503],
[1.406,58.209,-25.542],[3.139,55.858,-26.604],[7.854,58.872,-23.547],
[0.543,57.496,-23.109],[3.815,53.377,-26.000],[3.807,56.126,-23.727],
[7.607,55.971,-22.955],[10.393,55.213,-22.795],[-1.521,55.979,-23.306],
[4.698,51.825,-23.813],[5.481,54.132,-22.826],[14.574,56.949,-40.489],
[11.304,56.578,-39.513],[13.106,58.623,-39.071],[12.066,57.403,-31.352],
[13.695,58.826,-27.951],[-3.588,57.656,-24.419],[13.120,52.720,-22.853],
[13.829,54.611,-35.657],[-2.544,54.521,-30.636],[-2.753,57.754,-30.058],
[15.231,54.768,-24.577],[14.023,54.534,-39.535],[-2.849,55.563,-36.060],
[-3.636,53.359,-28.346],[13.284,53.008,-25.649],[7.109,55.614,-30.042],
[11.207,52.771,-32.126],[14.407,53.244,-32.341],[-0.758,53.229,-35.236],
[-2.163,53.154,-37.855],[4.619,53.086,-33.274],[8.244,54.115,-32.146],
[9.535,52.847,-37.999],[1.553,54.842,-39.848],[6.016,55.855,-37.034],
[-0.976,54.218,-25.482],[1.702,53.191,-23.326],[9.275,52.993,-23.582],
[8.519,-50.074,-15.410],[5.779,-50.076,-11.043],[0.401,-50.889,-9.433],
[11.400,-47.043,-14.265],[-0.114,-44.013,-11.124],[-3.403,-53.540,-17.822],
[-3.321,-49.767,-14.507],[9.149,-41.803,-13.656],[4.337,-47.815,-12.359],
[13.376,-56.558,-10.897],[9.963,-51.820,-10.713],[1.538,-54.666,-12.716],
[8.575,-57.668,-15.187],[-1.353,-51.664,-13.693],[3.875,-51.725,-14.296],
[16.187,-53.711,-8.045],[0.596,-54.963,-9.897],[7.578,-45.472,-15.681],
[3.832,-47.696,-9.658],[12.467,-55.259,-13.614],[0.736,-52.010,-11.752],
[0.881,-47.882,-9.276],[-1.806,-46.283,-15.747],[10.793,-46.175,-11.420],
[11.257,-57.307,-9.310],[2.633,57.464,-12.335],[10.440,-44.151,-13.738],
[10.065,-45.542,-16.550],[12.509,-45.411,-7.144],[5.392,-49.598,-16.409],
[8.033,-50.716,-21.387],[14.416,-50.103,-9.524],[11.448,-51.080,-17.327],
[9.822,-51.729,-7.644],[2.944,-48.171,-14.945],[2.328,-45.835,-12.191],
[2.466,-49.927,-12.007],[7.804,-44.489,-11.100],[2.606,-46.166,-7.805],
[8.471,-49.794,-18.675],[1.160,-50.585,-6.878],[7.226,-42.447,-15.928],
[2.366,-44.453,-4.704],[12.685,-58.296,-18.087],[0.570,-49.465,-16.689],
[14.652,-42.523,-22.469],[-3.843,-46.889,-8.370],[-2.856,-45.472,-11.301],
[14.948,-44.773,-5.921],[5.728,-45.694,-7.768],[7.099,-40.075,-18.099],
[-0.053,-41.679,-16.016],[9.836,-43.363,-20.941],[9.025,-43.005,-18.023],
[12.416,-51.909,-6.701],[-1.224,-46.189,-8.714],[2.256,-56.875,-10.684],
[6.043,-58.955,-7.341],[9.285,-54.376,-7.663],[10.652,-50.106,-21.857],
[7.005,-56.041,-21.960],[3.431,-52.912,-11.634],[2.160,-53.711,-16.956],
[6.904,-52.360,-17.914],[8.069,-52.520,-5.606],[8.146,-53.330,-13.210],
[0.293,-53.403,-7.699],[2.429,-54.499,-21.613],[-2.311,-55.056,-14.457],
[-1.646,-55.358,-19.341],[7.511,-51.838,-9.293],[3.299,-55.205,-14.499],
[-1.115,-53.258,-16.104],[0.492,-55.736,-17.571],[9.394,-55.865,-10.718],
[2.017,-52.264,-20.016],[11.148,-40.518,-5.087],[0.876,-47.876,-6.145],
[5.271,-53.979,-13.248],[0.572,-55.967,-15.041],[6.704,-54.367,-10.576],
[5.666,-49.450,-19.661],[11.660,-50.037,-9.702],[8.519,-54.713,-19.868],
[8.076,-47.175,-7.208],[9.643,-52.281,-19.079],[-0.665,-48.883,-13.973],
[6.795,-47.592,-17.783],[-3.290,-49.285,-6.993],[3.876,-52.535,-4.801],
[9.146,-44.825,-7.747],[14.169,-54.028,-19.268],[12.895,-47.804,-6.338],
[11.948,-43.763,-17.658],[9.646,-48.769,-12.954],[2.564,-45.058,-14.839],
[-1.466,-50.896,-20.850],[5.438,-46.826,-15.060],[4.542,-43.643,-9.518],
[3.418,-42.548,-17.129],[6.273,-46.787,-10.348],[9.333,-46.977,-19.160],
[11.574,-42.870,-7.728],[16.498,-42.137,-20.750],[-2.083,-54.658,-8.638],
[10.430,-48.487,-16.935],[-2.778,-43.045,-10.052],[3.320,-44.639,-21.755],
[3.789,-47.112,-19.706],[-0.961,-47.972,-18.443],[7.566,-56.292,-8.632],
[8.094,-55.374,-16.412],[5.308,-44.332,-18.926],[5.453,-54.741,-16.105],
[5.593,-53.285,-6.913],[6.119,-57.038,-17.157],[3.038,-42.151,-20.139],
[3.481,-56.707,-19.429],[12.468,-54.417,-7.673],[10.684,-53.762,-12.530],
[0.884,-45.034,-20.438],[-0.148,-53.778,-20.992],[12.452,-42.471,-10.404],
[-0.086,-58.404,-11.372],[0.057,-44.639,-15.970],[-1.005,-41.627,-4.984],
[1.714,-41.455,-4.031],[6.742,-43.970,-22.717],[9.264,-45.295,-4.043],
[12.727,-44.682,-20.689],[-1.558,-47.487,-5.080],[2.082,-47.699,-21.997],
[3.883,-47.679,-5.973],[5.929,-46.723,-21.904],[10.509,-47.687,-8.719],
[-1.336,-50.953,-5.301],[2.885,-49.702,-20.084],[2.829,-52.478,-8.114],
[4.148,-51.555,-21.481],[7.640,-49.702,-5.913],[13.192,-49.189,-21.338],
[2.645,-54.712,-6.068],[3.871,-54.815,-8.644],[4.741,-50.698,-8.564],
[9.788,-53.326,-3.263],[10.673,-49.444,-6.358],[0.821,-52.741,-4.933],
[3.158,-57.993,-8.128],[4.886,-55.797,-3.392],[7.076,-55.673,-5.881],
[10.208,-55.965,-4.575],[-0.525,-55.139,-5.698],[2.654,59.279,-5.819],
[4.778,-59.187,-3.654],[5.126,-40.586,-19.947],[-2.000,-45.567,-20.961],
[-2.926,-40.562,-17.172],[-2.717,-41.936,-14.788],[16.460,-41.896,-11.876],
[15.692,-41.849,-6.932],[11.856,-46.287,-3.842],[14.489,-48.473,-18.074],
[13.543,-50.321,-15.848],[13.839,-47.575,-11.104],[13.336,-44.921,-10.340],
[15.715,-49.850,-20.048],[12.966,-52.684,-21.223],[11.058,-53.220,-15.136],
[-1.872,-50.381,-11.182],[12.963,-50.202,-4.540],[15.752,-51.449,-4.844],
[11.801,-57.091,-21.339],[11.466,-53.977,-17.748],[14.269,-53.996,-11.834],
[12.794,-52.519,-10.031],[-3.973,-55.777,-6.613],[16.844,-59.114,-4.260],
[12.438,-57.729,-15.390],[-2.245,-55.011,-11.638],[-4.895,-50.288,-12.056],
[14.236,-56.967,-4.608],[14.499,-57.693,-20.538],[-3.623,-56.741,-18.028],
[11.415,-41.570,-16.147],[-1.803,-57.169,-9.596],[-0.464,-44.888,-5.508],
[1.961,-43.786,-9.128],[6.312,-43.347,-6.101],[8.877,-41.923,-6.368],
[13.902,-59.305,-8.661],[0.029,-39.157,-8.475],[0.864,-41.289,-9.826],
[6.282,-42.088,-10.903],[9.451,-42.324,-9.743],[10.272,-39.972,-7.996],
[-1.105,-41.283,-12.414],[3.518,-41.433,-11.121],[4.042,-43.095,-13.379],
[7.393,-56.054,-13.330],[13.828,-58.805,-13.075],[13.445,58.436,-15.466],
[-1.200,-58.710,-14.204],[-1.324,-58.126,-19.015],[4.773,-59.240,-14.540],
[9.144,-58.782,-12.522],[10.193,-58.385,-17.221],[-2.041,-43.582,-18.087],
[0.826,-56.458,-20.443],[5.980,-58.152,-19.776],[7.275,-41.075,-22.463],
[14.401,-42.898,-16.186],[-1.231,-57.532,-6.972],[1.461,-57.345,-5.820],
[0.252,-41.204,-22.116],[11.558,-58.387,-3.499],[13.113,-42.478,-5.167],
[7.892,-30.056,-12.752],[6.847,-32.517,-13.202],[-2.529,-33.197,-11.098],
[10.582,-28.675,-12.740],[0.810,-27.629,-11.885],[-2.445,-34.587,-17.839],
[-2.518,-30.859,-15.534],[9.653,-23.343,-13.965],[6.101,-28.285,-11.523],
[14.094,-37.615,-11.998],[8.860,-35.241,-10.474],[2.419,-35.778,-14.415],
[9.412,-37.805,-15.554],[-3.446,-34.619,-15.449],[3.982,-33.531,-15.120],
[14.613,-35.979,-8.897],[1.310,-37.014,-10.004],[9.018,-28.408,-14.942],
[3.714,-28.695,-10.097],[11.868,-37.855,-13.604],[0.370,-35.097,-12.051],
[1.002,-29.864,-10.051],[-1.422,-27.718,-18.132],[10.978,-26.778,-10.744],
[12.061,-38.461,-9.748],[4.525,-39.820,-13.142],[10.225,-25.897,-13.759],
[11.458,-27.741,-15.856],[11.077,-27.643,-7.274],[7.730,-30.550,-16.420],
[8.093,-33.786,-21.029],[14.281,-33.287,-8.151],[10.238,-32.654,-16.816],
[9.890,-33.619,-8.334],[5.043,-31.716,-16.322],[3.778,-28.282,-13.070],
[2.974,-32.399,-12.533],[8.081,-26.689,-10.921],[2.697,-27.118,-7.909],
[9.830,-31.725,-19.177],[0.228,-31.602,-6.745],[6.974,-25.803,-16.163],
[2.425,-24.807,-6.131],[12.248,-37.848,-19.093],[2.479,-31.317,-17.569],
[13.970,-25.834,-20.243],[-1.429,-29.332,-8.443],[-3.486,-26.353,-11.135],
[16.298,-23.638,-7.119],[5.412,-26.095,-8.633],[6.869,-23.147,-16.626],
[0.581,-24.406,-14.266],[8.441,-24.425,-20.952],[9.875,-23.641,-16.598],
[13.017,-34.419,-5.513],[-0.068,-26.813,-7.448],[3.691,-38.272,-10.727],
[7.308,-41.227,-8.436],[9.784,-35.438,-6.181],[11.984,-32.635,-20.852],
[5.960,-37.783,-19.373],[2.856,-35.019,-11.030],[0.716,-33.403,-16.999],
[7.043,-33.458,-17.487],[7.426,-33.447,-7.014],[9.202,-34.935,-15.506],
[0.326,-34.509,-9.337],[3.098,-34.764,-22.623],[-2.372,-37.759,-15.515],
[-0.480,-37.369,-18.883],[7.158,-33.268,-9.755],[2.671,-38.035,-13.179],
[-0.474,-34.922,-15.052],[3.617,-36.190,-18.539],[9.479,-38.112,-10.646],
[2.421,-33.628,-19.457],[11.963,-23.168,-5.400],[0.242,-29.203,-5.061],
[5.751,-35.334,-11.752],[1.325,-37.436,-16.286],[4.815,-36.400,-9.339],
[4.950,-29.441,-21.277],[13.557,-30.922,-9.381],[10.327,-39.504,-19.375],
[5.506,-29.013,-7.197],[10.152,-35.250,-19.928],[1.610,-30.341,-13.053],
[8.107,-29.028,-18.883],[-2.595,-32.269,-6.938],[4.075,-33.968,-5.274],
[7.335,-24.787,-7.029],[14.749,-33.645,-18.616],[14.403,-27.576,-6.623],
[13.159,-27.042,-17.906],[10.222,-30.963,-11.757],[3.238,-26.582,-15.969],
[-1.462,-31.319,-20.568],[4.239,-29.217,-16.732],[4.313,-23.596,-10.065],
[3.502,-24.334,-17.294],[7.743,-30.404,-9.873],[10.618,-28.791,-20.184],
[10.458,-24.884,-6.927],[15.050,-23.225,-20.862],[-3.003,-38.749,-7.793],
[11.330,-30.251,-16.847],[-0.735,-25.748,-10.239],[4.083,-27.193,-22.321],
[4.178,-27.625,-18.810],[0.395,-29.190,-19.321],[7.079,-37.900,-9.532],
[8.698,-37.684,-18.247],[6.199,-26.860,-20.625],[4.557,-37.030,-15.858],
[5.882,-35.688,-7.131],[6.651,-38.599,-15.264],[4.328,-24.272,-19.735],
[3.455,-38.333,-20.557],[12.193,-35.209,-7.755],[11.604,-32.813,-13.420],
[1.300,-26.504,-19.635],[-1.956,-34.366,-20.796],[11.668,-23.487,-11.838],
[0.510,-39.033,-11.574],[0.342,-25.732,-16.787],[-2.203,-24.475,-3.149],
[2.966,-25.659,-3.033],[5.308,-24.682,-22.061],[9.462,-27.503,-5.074],
[11.120,-25.727,-22.092],[-2.192,-30.580,-4.427],[2.147,-29.325,-21.164],
[2.920,-30.178,-6.756],[7.195,-28.943,-22.488],[10.450,-29.999,-7.047],
[-0.987,-32.999,-4.916],[3.851,-31.557,-20.340],[2.995,-33.393,-8.320],
[6.656,-31.485,-20.888],[8.906,-30.068,-4.905],[13.477,-29.809,-21.992],
[2.662,-35.828,-6.829],[2.625,-38.489,-7.210],[5.332,-31.709,-7.956],
[8.242,-34.084,-4.385],[9.836,-32.293,-6.008],[1.717,-33.100,-3.973],
[4.179,-40.702,-8.458],[5.708,-35.976,-4.031],[7.714,-37.099,-5.954],
[9.010,-36.934,-3.478],[-0.149,-36.001,-5.091],[3.603,-41.389,-5.913],
[5.740,-39.593,-6.427],[5.058,-21.768,-20.258],[-0.384,-28.525,-21.987],
[-2.140,-22.140,-17.085],[-2.737,-22.839,-13.967],[14.150,-24.651,-10.577],
[13.449,-23.727,-8.007],[11.984,-28.850,-4.854],[15.661,-28.824,-18.406],
[13.528,-31.973,-16.925],[14.748,-28.519,-9.875],[13.110,-26.530,-8.604],
[14.125,-30.798,-19.469],[14.952,-34.950,-21.591],[12.321,-34.229,-15.520],
[-1.920,-30.737,-12.467],[13.020,-31.193,-6.375],[15.508,-32.510,-3.985],
[13.743,-37.613,-21.383],[10.870,-36.090,-17.271],[11.714,-35.808,-10.344],
[11.680,-32.695,-10.282],[-3.712,-34.783,-6.158],[16.038,-39.822,-3.776],
[13.055,-37.325,-16.225],[-1.830,-37.013,-11.749],[-4.106,-35.414,-11.501],
[13.338,-38.884,-5.317],[15.201,-36.358,-19.419],[-2.963,-38.299,-18.626],
[12.299,-24.510,-15.499],[-2.156,-40.366,-10.053],[-0.073,-25.822,-4.885],
[2.211,-25.707,-10.230],[6.455,-21.881,-6.767],[8.988,-22.619,-7.134],
[13.584,-40.790,-8.562],[-0.128,-21.158,-8.829],[0.771,-23.387,-9.965],
[6.874,-23.045,-11.448],[9.393,-23.619,-9.952],[11.064,-21.871,-8.819],
[-1.039,-23.689,-12.056],[-0.974,-26.523,-14.484],[3.304,-24.895,-12.697],
[7.213,-37.440,-13.017],[13.351,-41.057,-12.684],[13.960,-39.665,-15.444],
[0.268,-39.278,-14.258],[0.444,-40.392,-19.632],[4.831,-40.738,-15.614],
[9.396,-39.208,-13.260],[12.609,-41.090,-18.497],[-2.274,-25.314,-16.919],
[-0.189,-38.539,-21.375],[3.564,-38.933,-17.666],[6.762,-24.778,-18.854],
[14.025,-24.061,-17.857],[-0.624,-38.909,-5.795],[3.535,-39.453,-4.034],
[-0.568,-23.137,-21.362],[8.728,-39.931,-4.236],[14.500,-25.241,-5.309],
[8.926,-14.008,-15.137],[8.247,-13.542,-11.756],[0.485,-13.896,-10.974],
[11.330,-10.077,-13.208],[1.674,-9.831,-13.363],[-0.011,-13.208,-14.244],
[-3.072,-13.134,-14.692],[7.931,-6.039,-12.432],[5.980,-9.740,-12.173],
[14.144,-19.600,-12.105],[9.025,-16.033,-11.014],[2.391,-17.315,-14.036],
[9.860,-18.503,-15.724],[-1.422,-15.283,-12.696],[2.799,-14.551,-14.802],
[14.931,-16.800,-8.891],[0.001,-18.384,-9.128],[6.854,-8.958,-16.079],
[2.864,-11.462,-10.203],[11.991,-18.911,-14.270],[1.520,-17.843,-11.335],
[0.087,-8.599,-9.686],[-1.703,-9.401,-18.576],[11.601,-8.246,-11.014],
[11.445,-19.840,-11.623],[4.280,-21.878,-12.801],[10.603,-6.902,-13.603],
[10.051,-8.453,-16.252],[13.358,-11.113,-7.510],[6.826,-11.740,-16.021],
[7.792,-14.972,-19.526],[16.324,-13.440,-9.040],[11.701,-12.944,-18.088],
[9.440,-14.111,-7.032],[2.400,-12.269,-13.606],[2.714,-8.635,-11.266],
[3.553,-13.749,-11.465],[8.442,-8.334,-10.907],[2.998,-10.522,-5.725],
[9.104,-12.491,-18.889],[-0.088,-13.196,-8.369],[8.233,-6.517,-15.082],
[0.808,-6.091,-6.952],[13.986,-21.204,-19.086],[1.100,-13.206,-18.097],
[14.469,-8.313,-19.721],[-1.989,-10.545,-7.616],[-2.797,-6.676,-12.439],
[15.927,-7.771,-5.702],[3.774,-6.895,-6.834],[6.097,-5.507,-16.660],
[0.658,-4.046,-15.850],[9.732,-5.485,-20.326],[10.088,-5.626,-17.212],
[13.367,-15.685,-5.792],[-0.119,-9.078,-6.520],[2.436,-21.370,-10.822],
[7.306,-21.633,-9.148],[11.006,-18.027,-7.198],[12.096,-13.898,-22.269],
[6.524,-17.569,-20.567],[2.319,-15.659,-9.782],[3.640,-16.712,-16.813],
[6.711,-14.543,-16.673],[7.901,-16.005,-5.846],[7.670,-16.072,-13.901],
[0.137,-15.903,-7.552],[2.175,-17.825,-21.641],[-1.783,-19.935,-15.480],
[-1.966,-16.483,-18.121],[7.539,-14.919,-9.250],[4.155,-19.458,-14.489],
[0.015,-16.592,-15.259],[3.482,-17.435,-19.296],[8.905,-19.162,-11.203],
[2.477,-14.958,-19.564],[11.017,-2.844,-5.181],[0.652,-11.706,-6.113],
[5.795,-14.875,-12.166],[2.030,-19.184,-16.112],[5.222,-16.214,-9.694],
[4.857,-10.664,-17.787],[11.604,-13.079,-8.492],[8.530,-19.226,-19.903],
[5.552,-10.041,-6.629],[10.088,-15.188,-17.745],[0.530,-11.011,-11.117],
[7.481,-9.929,-18.592],[-3.753,-13.269,-6.176],[6.040,-13.824,-6.372],
[8.960,-7.102,-6.935],[14.988,-16.332,-18.254],[15.220,-10.297,-4.885],
[12.169,-7.902,-18.280],[11.537,-12.254,-11.062],[3.552,-7.616,-16.597],
[-1.589,-12.340,-20.109],[4.327,-10.628,-14.826],[4.367,-5.798,-9.360],
[3.479,-5.368,-14.599],[9.410,-10.745,-9.227],[9.634,-10.021,-20.766],
[11.197,-8.418,-7.281],[14.453,-4.549,-20.537],[-0.229,-19.403,-6.421],
[10.346,-10.172,-18.178],[-2.684,-8.369,-9.122],[4.281,-8.151,-21.558],
[2.945,-8.878,-18.816],[-1.038,-12.394,-16.736],[6.764,-18.951,-8.713],
[9.643,-20.788,-17.889],[5.652,-7.931,-19.128],[6.250,-17.110,-17.937],
[5.353,-17.127,-7.154],[7.172,-19.691,-15.998],[3.244,-6.275,-19.603],
[2.914,-20.030,-19.880],[12.770,-19.101,-8.726],[11.446,-15.051,-12.326],
[0.656,-7.489,-18.927],[0.111,-15.815,-20.442],[11.626,-5.353,-11.127],
[0.063,-20.846,-11.848],[0.428,-7.279,-16.106],[-1.390,-4.493,-4.612],
[3.416,-5.067,-2.201],[7.368,-6.640,-21.061],[9.505,-8.652,-3.999],
[12.282,-7.214,-21.343],[-1.746,-11.950,-4.504],[2.823,-10.293,-22.178],
[5.328,-11.725,-4.385],[7.406,-10.014,-22.063],[10.686,-11.161,-6.860],
[-0.569,-14.720,-5.045],[4.884,-13.541,-18.759],[3.800,-15.036,-6.926],
[4.671,-13.216,-22.036],[8.406,-11.587,-4.085],[14.098,-11.651,-22.329],
[2.679,-17.438,-7.625],[4.299,-18.849,-11.178],[6.242,-11.864,-8.347],
[11.108,-16.464,-3.271],[10.247,-13.738,-4.145],[2.175,-15.313,-4.293],
[4.209,-20.601,-8.831],[6.596,-16.475,-3.688],[7.751,-18.755,-6.183],
[10.007,-19.318,-4.656],[0.799,-18.008,-4.396],[4.368,-22.769,-5.400],
[4.478,-20.169,-5.815],[2.518,-3.875,-18.391],[-2.958,-10.172,-21.512],
[-2.888,-4.098,-17.324],[-2.134,-6.050,-15.637],[14.205,-5.204,-10.548],
[15.062,-6.069,-7.983],[12.498,-9.173,-4.795],[13.575,-10.998,-18.285],
[13.912,-12.341,-15.990],[13.721,-9.743,-11.783],[13.236,-8.713,-8.740],
[14.398,-13.343,-20.177],[13.864,-15.929,-21.071],[12.150,-14.500,-15.648],
[-3.051,-12.257,-11.942],[13.262,-12.787,-5.326],[15.922,-14.779,-4.021],
[11.977,-17.848,-21.960],[11.865,-17.181,-16.984],[12.901,-17.455,-11.204],
[12.358,-16.175,-8.764],[-2.076,-18.662,-4.484],[16.282,-21.227,-2.498],
[12.370,-19.818,-17.502],[-0.353,-18.931,-13.606],[-3.923,-14.941,-11.473],
[16.152,-20.223,-5.379],[14.556,-18.539,-20.469],[-3.500,-18.418,-16.590],
[11.528,-3.814,-15.572],[-2.922,-19.651,-7.505],[-0.283,-6.954,-4.623],
[1.860,-6.679,-9.482],[6.261,-6.131,-6.034],[8.636,-4.268,-6.346],
[14.296,-22.149,-6.160],[0.238,-2.654,-7.428],[-0.393,-2.465,-10.035],
[6.016,-4.261,-10.747],[9.230,-4.935,-9.574],[10.798,-2.996,-8.271],
[-1.670,-4.342,-11.568],[1.223,-4.741,-12.773],[4.434,-6.271,-12.005],
[6.661,-19.994,-13.416],[13.509,-21.913,-13.447],[13.987,-21.758,-16.084],
[0.568,-21.825,-15.832],[-2.010,-21.096,-19.943],[3.378,-21.868,-15.392],
[9.535,-20.690,-14.102],[11.586,-22.694,-18.736],[-2.092,-6.669,-18.360],
[-0.403,-19.660,-21.569],[5.290,-19.870,-18.148],[6.488,-4.688,-19.270],
[13.833,-5.295,-17.506],[-0.163,-21.783,-4.608],[3.045,-23.140,-1.352],
[-0.136,-3.868,-20.242],[9.378,-22.734,-4.331],[13.284,-6.024,-5.215],
[7.194,5.998,-15.577],[8.034,5.377,-11.716],[-0.982,7.075,-10.612],
[9.749,7.486,-16.500],[1.675,8.340,-10.795],[-2.357,3.334,-16.211],
[-2.006,6.009,-15.170],[8.138,12.731,-12.490],[6.668,8.279,-11.798],
[14.710,0.414,-10.951],[10.104,1.571,-9.891],[2.235,1.770,-13.949],
[8.948,0.611,-14.908],[-1.381,4.323,-12.358],[4.783,4.049,-14.541],
[13.024,1.194,-8.104],[0.668,0.337,-9.829],[7.431,8.777,-16.008],
[3.942,7.005,-8.841],[11.577,0.502,-13.577],[0.857,4.174,-10.432],
[1.171,6.913,-8.526],[-2.165,9.619,-18.615],[10.818,9.439,-10.895],
[11.435,-1.690,-11.082],[4.448,-2.733,-12.740],[10.268,12.021,-14.027],
[10.012,10.214,-16.321],[12.046,9.236,-6.749],[6.596,5.568,-18.482],
[8.460,3.218,-21.119],[15.073,5.103,-9.014],[11.778,5.428,-16.775],
[9.526,4.609,-7.431],[2.498,6.012,-14.564],[4.001,9.134,-11.993],
[3.563,4.806,-11.957],[7.341,11.012,-10.583],[4.247,10.127,-6.359],
[9.220,5.141,-18.003],[0.204,4.598,-7.865],[6.961,12.186,-15.998],
[0.261,10.256,-6.464],[12.036,-2.971,-20.469],[0.522,5.393,-17.253],
[14.104,10.217,-20.768],[-3.587,8.574,-9.215],[-2.414,9.864,-11.536],
[15.689,11.930,-6.414],[5.871,12.133,-7.878],[7.520,14.809,-16.444],
[0.844,13.251,-17.172],[10.492,13.563,-20.893],[9.923,12.300,-17.591],
[12.001,3.547,-5.972],[-2.119,9.063,-7.004],[2.392,-1.644,-11.228],
[7.430,-3.167,-8.723],[9.174,1.946,-6.916],[10.174,5.992,-22.923],
[6.924,1.077,-19.863],[2.351,1.925,-11.141],[2.709,3.761,-16.389],
[7.389,3.256,-16.927],[7.509,3.581,-5.820],[9.196,3.291,-14.249],
[0.069,2.334,-6.596],[2.241,0.779,-20.876],[-3.110,-0.570,-14.357],
[-0.625,1.925,-19.587],[7.960,3.000,-9.352],[4.496,-0.123,-15.077],
[0.464,1.929,-16.007],[3.159,1.440,-17.918],[7.661,-1.199,-10.516],
[1.973,3.408,-20.345],[11.336,16.783,-4.724],[-0.090,6.905,-5.958],
[6.229,3.016,-12.561],[1.400,-0.867,-15.486],[4.507,2.876,-9.567],
[5.011,7.751,-18.443],[12.642,7.303,-10.227],[9.643,-0.966,-20.776],
[6.984,8.895,-6.249],[9.991,2.411,-18.774],[0.943,7.948,-13.287],
[8.074,9.862,-18.440],[-3.366,6.253,-7.232],[5.595,5.628,-5.274],
[9.486,10.462,-7.379],[15.053,2.775,-18.120],[14.776,8.898,-6.065],
[11.902,10.926,-19.354],[10.673,6.746,-12.538],[1.867,11.284,-15.846],
[-0.663,4.910,-19.804],[3.064,8.691,-15.204],[4.323,11.931,-10.257],
[3.234,13.831,-15.692],[7.410,8.407,-9.265],[9.467,8.661,-20.748],
[10.679,13.025,-6.297],[14.604,13.866,-20.161],[-1.492,-0.276,-8.615],
[11.229,6.651,-19.737],[-0.225,9.335,-9.102],[5.456,9.986,-22.522],
[3.237,9.967,-19.617],[0.174,7.954,-18.361],[5.979,-0.831,-8.226],
[9.615,-0.665,-18.082],[5.714,10.985,-18.935],[5.839,0.643,-17.589],
[5.965,2.144,-7.450],[7.598,-1.421,-15.871],[3.170,12.683,-18.389],
[3.082,-1.282,-19.214],[11.020,0.043,-6.626],[11.641,4.243,-13.336],
[0.974,10.080,-20.925],[-1.276,3.832,-22.104],[11.437,13.344,-10.113],
[-3.246,-2.260,-11.602],[-0.848,11.584,-15.969],[-0.908,13.128,-5.664],
[3.315,12.464,-3.892],[8.238,11.294,-23.219],[8.410,10.111,-3.815],
[12.450,12.077,-22.657],[-2.601,6.992,-4.573],[2.598,7.526,-20.603],
[3.205,6.730,-6.127],[7.662,8.491,-22.727],[9.581,7.751,-8.202],
[-1.072,4.395,-4.769],[3.773,5.275,-19.716],[2.716,3.453,-7.831],
[5.465,4.351,-21.468],[8.220,6.409,-4.354],[14.141,7.287,-21.530],
[3.197,0.589,-7.269],[3.857,0.306,-9.806],[6.120,5.016,-9.331],
[10.460,2.000,-2.205],[10.851,6.614,-5.387],[2.493,4.571,-4.634],
[4.246,-2.556,-9.340],[5.345,2.278,-4.608],[7.364,0.567,-5.733],
[9.175,0.168,-3.928],[0.215,0.419,-4.044],[4.497,-3.834,-6.364],
[4.538,-0.912,-5.263],[5.350,13.935,-19.772],[-0.423,6.857,-21.678],
[-1.800,14.500,-16.635],[-1.593,12.103,-13.237],[15.440,15.143,-10.498],
[13.494,13.208,-6.755],[11.944,9.077,-4.212],[12.522,8.865,-17.371],
[14.366,5.705,-16.534],[14.346,9.429,-11.217],[13.532,10.466,-8.594],
[13.705,5.429,-19.361],[13.740,2.657,-20.528],[12.410,2.585,-15.560],
[-2.891,6.706,-12.579],[12.288,6.291,-7.772],[14.717,3.095,-6.135],
[12.108,0.365,-21.140],[12.578,1.074,-17.752],[12.165,2.340,-11.329],
[11.723,3.971,-9.016],[-3.346,1.695,-4.489],[14.963,-4.573,-2.716],
[12.645,-1.856,-17.083],[-2.578,0.914,-12.225],[-3.244,2.956,-10.843],
[14.397,-2.418,-4.543],[14.674,0.175,-21.428],[-1.997,-0.517,-16.914],
[10.415,14.334,-15.827],[-3.995,-1.212,-7.569],[0.141,11.346,-3.608],
[2.030,10.697,-9.819],[6.705,13.349,-5.152],[8.714,15.005,-6.526],
[13.194,-3.758,-6.997],[-0.727,16.361,-8.331],[1.535,14.104,-9.385],
[5.214,14.346,-11.684],[8.469,13.304,-9.771],[10.233,15.619,-8.735],
[-1.067,12.591,-9.814],[1.616,12.524,-13.229],[5.019,11.748,-12.996],
[6.903,-0.100,-12.931],[13.471,-3.257,-12.066],[14.123,-2.565,-14.437],
[-1.198,-3.187,-14.189],[-1.991,-1.600,-19.256],[4.701,-3.288,-15.522],
[10.592,-2.893,-13.115],[8.788,-3.050,-19.536],[-1.437,12.315,-18.553],
[-0.157,-1.103,-21.489],[5.512,-2.085,-17.992],[7.787,12.509,-19.739],
[13.001,13.285,-17.437],[-2.674,-0.845,-4.787],[2.362,-2.694,-3.240],
[0.656,13.152,-21.278],[8.490,-2.476,-3.800],[13.842,15.469,-5.113],
[8.902,24.297,-14.801],[7.555,24.585,-11.654],[2.630,23.864,-10.090],
[12.470,27.995,-15.262],[1.438,27.166,-11.218],[-2.572,22.262,-16.281],
[-2.436,24.037,-13.917],[9.362,32.601,-12.354],[5.861,27.638,-10.785],
[14.334,19.614,-9.666],[8.250,20.624,-11.288],[1.533,20.366,-14.215],
[10.384,18.333,-15.265],[-1.198,21.678,-13.767],[3.631,21.126,-12.495],
[14.829,21.812,-7.714],[0.038,20.177,-8.068],[8.280,27.412,-14.463],
[3.603,26.322,-9.837],[13.360,18.905,-14.780],[-0.151,19.793,-12.087],
[1.193,26.756,-8.350],[-1.315,27.554,-17.699],[11.070,30.039,-11.388],
[11.447,16.341,-10.946],[3.973,16.621,-12.634],[9.419,29.877,-13.821],
[9.922,28.816,-16.483],[10.743,28.320,-8.692],[6.785,24.854,-16.944],
[6.989,23.380,-20.915],[14.161,24.282,-10.702],[11.502,24.150,-17.103],
[9.433,22.236,-7.790],[0.720,25.555,-13.983],[3.762,27.663,-12.509],
[3.009,24.933,-12.663],[8.356,28.722,-9.763],[4.231,28.287,-5.733],
[10.988,21.625,-18.424],[0.907,23.759,-7.701],[6.493,30.004,-15.978],
[2.023,29.665,-5.180],[12.868,16.093,-19.437],[1.847,24.851,-17.148],
[13.120,30.117,-20.743],[-3.816,26.671,-9.088],[-3.561,30.295,-11.725],
[15.771,28.370,-6.008],[5.948,28.560,-8.072],[6.817,32.749,-16.434],
[0.234,31.364,-14.199],[8.591,31.511,-23.110],[9.467,30.977,-18.425],
[12.428,22.568,-6.315],[-1.275,27.590,-7.381],[2.358,16.454,-10.558],
[6.561,14.806,-8.615],[10.084,20.042,-6.077],[9.633,25.120,-21.566],
[6.024,17.806,-23.375],[2.189,21.102,-10.170],[2.756,20.026,-17.006],
[9.715,21.325,-15.624],[5.436,21.103,-4.412],[9.831,19.411,-13.001],
[-1.783,22.524,-7.242],[4.344,19.342,-22.135],[-4.298,18.428,-14.791],
[-1.941,19.971,-19.223],[7.450,22.478,-9.558],[4.853,18.514,-14.464],
[0.302,21.674,-16.088],[3.530,19.341,-19.567],[9.555,18.630,-10.130],
[2.833,22.085,-20.386],[13.553,33.557,-5.440],[1.102,27.028,-5.710],
[5.612,21.362,-10.891],[0.465,18.336,-15.768],[4.108,18.840,-10.048],
[6.073,26.053,-19.142],[11.353,27.157,-11.158],[8.906,17.326,-20.453],
[8.129,27.031,-7.604],[9.206,19.854,-19.148],[0.798,23.073,-12.481],
[7.893,27.555,-17.982],[-2.610,24.573,-5.595],[4.849,23.554,-6.100],
[8.442,29.849,-7.191],[14.841,21.385,-19.857],[14.017,26.509,-6.256],
[11.925,28.838,-18.659],[9.583,23.135,-10.787],[3.303,29.481,-15.150],
[1.179,23.822,-19.555],[4.026,25.933,-15.228],[3.480,30.890,-9.840],
[2.682,32.439,-15.476],[9.530,25.508,-9.256],[10.510,27.776,-21.096],
[10.997,32.028,-7.345],[15.025,32.384,-20.182],[-3.391,19.099,-8.093],
[10.512,26.648,-18.281],[-1.225,30.215,-9.946],[3.495,29.138,-22.838],
[3.986,26.957,-20.382],[0.800,26.751,-19.246],[6.547,19.076,-9.035],
[9.474,17.119,-17.716],[5.577,29.376,-18.464],[7.465,19.740,-16.983],
[5.572,20.914,-7.525],[7.439,17.311,-15.696],[4.623,30.360,-20.789],
[4.041,16.593,-20.503],[11.866,20.078,-8.131],[11.729,22.170,-12.399],
[1.284,28.481,-21.517],[-0.518,21.445,-21.058],[11.290,32.718,-10.606],
[1.240,15.659,-12.841],[0.198,29.099,-15.907],[-2.721,31.701,-4.673],
[4.468,31.117,-3.931],[6.353,29.935,-22.712],[9.814,29.328,-4.350],
[11.092,30.347,-22.423],[-1.884,26.968,-4.783],[1.917,25.941,-22.097],
[3.345,25.546,-7.208],[7.378,27.359,-21.188],[10.584,24.517,-6.500],
[-1.260,22.994,-4.242],[4.236,23.603,-18.820],[3.081,21.670,-6.700],
[4.077,23.935,-21.798],[7.217,24.138,-3.842],[12.529,24.975,-21.585],
[2.606,19.749,-4.782],[2.612,19.171,-7.576],[5.025,23.563,-8.918],
[11.568,21.240,-4.115],[9.217,22.720,-5.096],[1.270,22.085,-4.789],
[2.761,16.412,-8.011],[4.440,17.781,-4.959],[7.532,19.496,-6.224],
[9.063,18.074,-4.370],[-1.600,20.434,-4.577],[4.256,14.253,-6.555],
[4.719,15.104,-3.832],[4.150,32.969,-21.630],[-1.913,28.016,-20.322],
[-3.496,33.994,-16.223],[-2.346,31.024,-14.254],[14.161,33.154,-10.557],
[13.597,30.543,-6.784],[14.818,27.530,-3.828],[13.144,26.037,-18.021],
[15.812,24.795,-17.453],[14.846,28.582,-11.128],[12.762,30.196,-9.213],
[14.799,24.382,-19.931],[12.641,20.648,-21.692],[12.978,22.629,-15.024],
[-3.232,25.072,-11.495],[12.670,24.465,-3.659],[14.619,23.614,-5.235],
[12.914,18.214,-21.078],[11.558,18.932,-17.935],[12.652,18.977,-11.973],
[11.949,23.220,-8.930],[-5.150,19.770,-3.519],[15.819,16.326,-3.134],
[12.685,16.696,-16.443],[-0.488,17.025,-11.229],[-4.546,23.005,-12.570],
[15.130,18.346,-5.815],[15.796,19.012,-21.401],[-1.608,17.037,-18.256],
[12.510,31.688,-14.776],[-4.163,15.451,-7.775],[-0.620,30.316,-3.718],
[1.249,29.931,-8.911],[6.866,32.570,-5.084],[10.686,34.058,-5.123],
[13.054,15.254,-8.671],[-0.666,33.428,-9.318],[1.944,33.204,-7.813],
[7.254,34.111,-9.759],[8.382,31.456,-9.510],[9.966,34.477,-8.501],
[-0.767,32.466,-11.890],[2.230,29.873,-12.056],[2.631,32.536,-12.260],
[7.056,16.082,-13.210],[13.166,14.675,-12.141],[14.279,15.925,-14.397],
[-1.156,16.056,-14.515],[-1.487,14.494,-20.157],[4.956,15.985,-15.676],
[9.983,15.512,-13.423],[10.843,14.711,-18.385],[-1.502,30.668,-17.622],
[-0.777,17.174,-21.459],[6.408,18.964,-19.650],[5.810,32.413,-18.991],
[13.395,31.103,-18.176],[-1.815,17.687,-5.818],[1.897,16.650,-3.755],
[0.315,33.009,-21.057],[8.996,15.657,-3.311],[13.948,31.188,-4.325],
[9.487,42.401,-14.793],[7.197,43.156,-11.643],[-1.447,45.099,-9.957],
[12.111,46.237,-14.494],[1.586,45.925,-10.964],[-2.644,41.093,-16.454],
[-2.057,42.965,-14.143],[7.357,50.363,-13.908],[7.057,46.666,-11.914],
[14.106,36.208,-10.042],[9.423,38.786,-9.093],[1.403,38.385,-13.516],
[8.298,37.053,-15.657],[-1.736,39.563,-13.831],[6.796,42.020,-14.092],
[15.183,38.720,-8.654],[0.610,36.843,-10.277],[8.775,45.245,-16.013],
[3.698,44.888,-9.920],[13.541,38.086,-12.982],[-0.093,39.713,-11.556],
[0.789,46.463,-7.956],[-0.234,46.088,-18.220],[12.016,47.701,-11.618],
[11.631,35.777,-10.093],[4.779,34.426,-10.880],[10.028,51.057,-14.973],
[10.282,48.208,-15.792],[11.640,44.903,-6.958],[6.824,43.452,-16.604],
[8.376,39.302,-22.306],[14.733,42.751,-10.355],[11.473,43.775,-16.004],
[9.005,40.817,-7.112],[2.448,40.817,-14.142],[1.933,47.970,-13.151],
[1.536,41.997,-11.857],[7.942,49.255,-11.544],[3.608,46.580,-6.374],
[8.259,41.845,-18.790],[0.936,42.948,-7.431],[7.993,48.487,-17.524],
[2.277,48.596,-5.303],[11.274,33.174,-18.156],[1.935,41.149,-16.752],
[12.834,48.466,-20.555],[-3.480,46.267,-8.348],[-1.731,47.255,-11.772],
[15.685,47.941,-5.596],[7.277,47.149,-6.463],[7.382,51.574,-16.970],
[0.619,50.194,-15.007],[8.923,51.863,-20.792],[9.756,50.435,-17.896],
[12.236,39.747,-7.167],[-1.417,45.596,-6.796],[1.880,34.569,-9.960],
[6.689,34.629,-7.119],[10.376,37.296,-6.539],[10.883,42.405,-19.774],
[5.580,37.859,-19.584],[2.267,39.672,-10.734],[4.923,39.121,-15.191],
[7.158,40.781,-16.645],[6.908,39.685,-4.341],[9.141,39.529,-14.114],
[0.555,41.571,-9.502],[2.332,37.913,-22.628],[-2.963,37.120,-14.543],
[-0.673,37.404,-19.018],[8.401,41.431,-9.543],[4.972,36.748,-13.456],
[-0.292,38.485,-16.357],[1.846,38.393,-18.697],[8.193,36.459,-10.558],
[1.474,40.860,-20.266],[11.206,52.872,-5.772],[0.538,44.255,-5.093],
[5.935,39.035,-11.957],[0.304,36.039,-15.486],[4.624,39.292,-9.680],
[4.432,44.419,-17.956],[10.957,46.251,-9.460],[9.180,36.551,-18.862],
[5.849,44.875,-5.788],[10.001,39.533,-19.723],[0.529,43.635,-13.790],
[7.923,45.609,-18.626],[-2.677,42.865,-7.161],[5.780,42.227,-5.684],
[9.527,48.649,-7.718],[15.481,40.370,-18.400],[13.643,46.966,-6.965],
[12.973,46.599,-17.245],[10.084,42.488,-11.980],[1.601,48.000,-16.107],
[-1.261,44.457,-20.967],[3.600,45.245,-15.548],[4.466,48.218,-10.655],
[2.945,50.062,-17.119],[7.613,45.019,-9.640],[11.071,46.314,-20.629],
[11.622,49.595,-5.816],[13.822,51.403,-20.659],[-1.364,39.433,-9.229],
[10.299,44.922,-18.415],[-0.726,49.202,-7.859],[5.120,47.085,-22.668],
[3.258,46.779,-20.743],[1.007,43.517,-19.200],[6.424,37.590,-8.316],
[6.790,35.806,-18.057],[5.798,47.404,-18.742],[4.625,36.834,-16.433],
[6.107,39.773,-7.085],[6.526,35.308,-15.300],[4.395,49.049,-20.179],
[2.432,36.302,-20.302],[12.833,37.256,-7.932],[11.591,40.233,-12.471],
[0.756,47.039,-20.514],[-1.166,40.348,-21.147],[11.080,51.021,-11.385],
[1.096,34.692,-12.563],[-1.355,47.978,-15.187],[-2.304,49.978,-5.359],
[3.982,50.481,-4.173],[7.374,48.683,-21.573],[9.480,48.637,-4.300],
[9.913,48.802,-20.970],[-2.168,45.799,-4.298],[1.553,44.968,-22.398],
[3.562,43.517,-4.739],[6.478,45.151,-20.996],[10.223,43.700,-9.017],
[-1.198,42.156,-4.890],[3.466,43.122,-20.500],[3.532,40.308,-7.476],
[6.435,42.406,-21.466],[8.192,42.808,-4.600],[12.791,44.175,-20.931],
[3.966,37.730,-6.723],[4.344,36.036,-8.795],[5.655,42.144,-8.817],
[9.891,39.787,-4.541],[10.955,41.923,-5.593],[1.567,41.010,-4.962],
[4.607,32.985,-8.389],[5.103,37.504,-3.914],[7.517,37.335,-5.742],
[10.279,36.384,-3.800],[0.640,37.167,-6.227],[3.803,32.959,-5.908],
[4.981,34.664,-4.002],[5.397,50.774,-18.476],[-2.990,46.760,-19.328],
[-2.340,52.579,-14.757],[-2.180,49.976,-12.905],[14.991,52.497,-11.022],
[15.292,49.645,-8.518],[13.282,47.942,-4.452],[14.156,44.428,-18.096],
[14.014,44.434,-14.408],[13.866,45.627,-11.579],[12.200,49.441,-8.702],
[13.942,42.158,-19.560],[12.713,40.105,-20.712],[11.719,40.500,-15.891],
[-3.388,43.999,-11.979],[13.549,43.231,-5.718],[15.434,41.939,-7.356],
[11.351,37.497,-20.129],[11.317,38.091,-17.148],[12.386,38.520,-10.475],
[12.257,41.874,-8.835],[-3.313,37.862,-4.557],[15.625,32.710,-2.338],
[11.611,35.396,-16.360],[-2.408,37.780,-11.648],[-2.735,40.438,-11.124],
[15.894,36.329,-6.074],[15.566,37.526,-20.779],[-4.017,37.278,-17.384],
[12.302,51.311,-16.636],[-1.818,35.860,-8.994],[-0.317,47.901,-4.426],
[2.125,48.705,-9.378],[6.520,49.863,-6.558],[8.780,51.696,-6.282],
[13.087,34.014,-8.355],[-1.617,52.333,-8.191],[1.035,52.559,-9.678],
[5.934,50.754,-10.458],[8.351,50.561,-9.298],[11.915,52.257,-8.226],
[-1.093,49.729,-10.366],[1.216,50.825,-12.107],[3.870,49.853,-13.237],
[7.328,34.550,-12.482],[12.538,33.588,-12.848],[14.270,35.362,-15.426],
[-1.455,33.786,-14.134],[-3.449,34.809,-18.831],[4.826,33.761,-14.004],
[10.269,35.174,-12.248],[9.045,34.309,-17.102],[-1.841,49.031,-17.603],
[-0.015,35.712,-21.203],[4.291,35.247,-18.563],[6.420,51.182,-21.638],
[14.151,49.878,-18.386],[-0.651,34.646,-6.535],[1.672,34.914,-3.430],
[-0.398,52.048,-21.605],[8.725,33.349,-3.333],[13.251,50.719,-3.950],
[15.354,55.247,-11.983],[10.218,57.964,-11.212],[1.414,55.167,-14.208],
[10.414,55.835,-16.603],[-0.147,57.219,-13.383],[16.551,58.285,-7.884],
[2.902,55.406,-10.033],[11.503,55.190,-12.658],[0.599,56.307,-11.104],
[10.312,55.626,-9.059],[3.884,52.255,-11.448],[12.913,54.083,-20.287],
[11.911,-58.486,-6.045],[1.585,53.485,-12.130],[6.753,53.268,-7.426],
[10.696,57.071,-6.723],[6.909,56.146,-20.130],[1.620,58.902,-9.998],
[3.686,57.050,-14.943],[7.718,56.717,-13.722],[-0.985,58.824,-8.840],
[2.801,56.408,-21.365],[-3.951,56.094,-14.067],[-1.421,56.367,-18.424],
[3.911,55.093,-12.798],[0.392,-58.517,-16.400],[3.101,-59.246,-17.886],
[8.543,55.661,-11.288],[4.749,57.939,-19.339],[6.636,59.132,-11.972],
[0.577,56.868,-16.503],[4.660,58.365,-10.287],[9.910,54.224,-19.861],
[9.600,57.542,-19.735],[15.157,57.418,-17.356],[-2.322,56.087,-9.152],
[5.867,55.599,-10.362],[8.102,54.494,-17.800],[5.487,58.182,-16.529],
[7.107,57.436,-8.009],[6.182,53.633,-15.548],[1.066,56.018,-19.088],
[13.120,56.664,-8.365],[-0.773,57.543,-20.729],[-0.957,52.717,-11.942],
[3.041,57.547,-7.774],[4.985,55.487,-7.934],[9.018,58.086,-3.714],
[1.160,57.213,-4.992],[3.995,53.020,-8.294],[3.847,56.561,-5.253],
[8.180,55.694,-6.121],[10.301,55.347,-4.232],[-0.180,54.968,-5.827],
[2.776,50.715,-7.412],[3.951,53.354,-5.319],[14.730,57.178,-20.621],
[12.677,56.174,-21.990],[11.740,57.806,-18.089],[13.410,57.323,-12.038],
[11.217,58.812,-8.592],[-2.146,57.265,-4.843],[15.890,51.525,-3.995],
[13.091,55.321,-17.687],[-2.035,55.158,-12.147],[-3.014,-59.781,-11.175],
[15.713,54.086,-5.501],[15.753,53.609,-21.393],[-3.204,54.435,-17.386],
[-4.519,52.520,-7.891],[13.226,55.005,-5.667],[6.597,54.062,-12.656],
[12.327,52.503,-13.062],[14.143,53.577,-15.068],[-0.731,54.371,-15.806],
[-0.074,53.194,-18.194],[3.030,54.093,-16.291],[8.971,53.546,-13.896],
[9.902,52.916,-16.749],[-0.609,54.590,-21.001],[5.883,53.488,-19.494],
[-0.675,52.091,-4.982],[0.936,52.743,-2.776],[9.144,51.476,-3.669],
[8.431,-49.993,3.723],[5.849,-51.694,5.712],[-0.372,-50.602,8.132],
[11.144,-47.792,5.205],[0.326,-45.994,6.327],[-0.851,-52.984,1.601],
[-2.483,-49.072,3.986],[7.872,-43.004,5.130],[6.347,-45.995,5.692],
[14.558,-56.530,7.675],[9.390,-54.263,8.810],[1.785,-55.070,5.369],
[10.174,-55.238,3.272],[-2.327,-52.180,4.279],[3.587,-50.729,2.904],
[15.309,-53.479,9.701],[-0.568,-55.638,7.110],[9.354,-47.989,1.415],
[2.501,-49.167,9.030],[12.713,-55.743,5.706],[0.051,-52.500,6.001],
[1.741,-46.771,10.223],[-1.808,-45.756,0.999],[11.510,-47.011,8.548],
[13.311,-56.084,10.150],[4.678,-58.655,6.657],[9.874,-45.729,6.309],
[9.355,-45.314,3.651],[10.969,-45.862,11.041],[6.420,-49.815,0.914],
[7.472,-50.472,-3.372],[14.426,-51.094,7.885],[11.064,-50.312,1.144],
[7.753,-52.502,10.332],[1.707,-48.435,3.014],[3.614,-45.347,5.702],
[3.155,-49.453,6.031],[7.721,-45.095,8.300],[3.690,-46.129,12.632],
[9.889,-51.587,-1.059],[0.357,-50.830,11.266],[7.224,-43.736,2.805],
[1.069,-43.638,13.000],[13.556,-56.368,-1.384],[1.073,-50.668,1.415],
[13.653,-43.783,-2.734],[-2.139,-50.064,10.291],[-3.359,-45.782,7.168],
[15.640,-44.976,12.412],[6.598,-43.754,10.406],[6.177,-40.727,0.768],
[0.571,-41.354,3.309],[8.491,-41.769,-2.000],[9.368,-43.330,1.176],
[12.270,-50.613,15.764],[-0.985,-47.976,11.364],[2.409,-56.995,7.460],
[6.054,58.526,10.383],[10.661,-56.022,10.496],[9.981,-49.608,-2.796],
[7.711,-56.047,-3.112],[3.057,-52.947,6.696],[3.861,-53.060,4.025],
[8.553,-51.844,1.516],[6.434,-51.396,12.935],[9.917,-52.551,4.068],
[0.562,-53.373,11.161],[2.426,-57.474,-3.295],[-2.716,-56.786,3.523],
[-1.508,-55.632,-0.295],[7.235,-51.367,7.995],[4.674,-56.590,3.779],
[-0.194,-55.156,3.500],[2.674,-52.533,-0.719],[8.899,-56.668,7.534],
[0.403,-53.481,-2.280],[10.897,-41.809,14.584],[0.826,-47.069,13.790],
[8.173,-53.058,5.965],[2.487,-57.388,1.418],[5.629,-54.238,5.582],
[4.940,-48.270,-1.238],[11.813,-49.604,8.297],[10.342,-55.977,-1.938],
[6.522,-47.615,11.437],[8.326,-54.523,-0.999],[0.120,-48.896,5.399],
[7.608,-47.796,-0.647],[-4.788,-50.791,11.191],[3.358,-50.471,11.542],
[9.027,-47.689,9.682],[15.850,-51.253,0.213],[14.402,-46.979,11.218],
[14.346,-46.317,2.047],[9.049,-48.940,6.382],[2.480,-45.857,3.106],
[-2.284,-49.119,-2.134],[4.462,-47.544,1.864],[4.181,-44.750,8.876],
[2.755,-44.294,0.582],[6.148,-48.590,6.465],[10.026,-47.277,-1.521],
[10.233,-43.475,12.151],[13.630,-40.828,-2.795],[-2.791,-54.874,9.069],
[12.094,-45.455,-0.786],[-0.711,-45.791,9.053],[3.974,-45.921,-2.890],
[2.847,-47.014,-0.159],[0.268,-47.849,-0.806],[6.575,-57.354,9.768],
[9.377,-56.576,0.527],[5.417,-45.591,-0.597],[6.557,-54.798,2.465],
[6.449,-54.743,9.834],[8.143,-57.207,2.964],[1.855,-42.570,-1.544],
[3.653,-56.885,-0.770],[12.851,-54.388,12.793],[11.095,-51.620,6.715],
[0.558,-45.315,-1.822],[-2.291,-54.595,-3.630],[10.882,-42.532,6.934],
[-0.507,-58.444,7.019],[-0.232,-45.337,3.320],[-2.547,-40.390,13.769],
[3.944,-43.269,13.720],[6.666,-44.181,-3.442],[8.171,-45.154,12.406],
[10.831,-43.553,-2.744],[-3.057,-48.196,14.684],[2.489,-48.219,-2.755],
[4.697,-48.717,13.543],[5.754,-48.129,-3.906],[11.153,-49.101,10.885],
[-0.679,-51.701,13.859],[3.663,-50.187,-0.174],[1.572,-53.768,8.703],
[4.475,-50.656,-2.863],[7.732,-49.054,13.872],[13.064,-48.473,-2.397],
[4.121,-53.862,11.902],[3.605,-56.047,9.750],[4.916,-51.004,9.381],
[8.756,-54.511,14.267],[9.479,-51.755,13.368],[2.181,-52.205,13.401],
[2.731,59.020,10.075],[5.921,-54.207,13.862],[8.113,-55.915,11.898],
[10.763,-56.293,14.849],[0.043,-54.419,13.720],[3.948,-57.386,12.368],
[5.589,-57.386,16.001],[5.026,-39.256,-1.661],[-2.088,-46.180,-1.824],
[-3.489,-40.733,2.058],[-2.378,-44.508,4.899],[14.997,-41.690,6.418],
[14.159,-42.054,11.055],[12.383,-46.776,13.132],[14.897,-48.231,-0.291],
[13.770,-50.096,1.399],[15.092,-46.688,8.337],[13.931,-44.610,9.938],
[14.182,-52.562,-2.553],[12.376,-52.830,-0.467],[12.385,-51.754,4.064],
[-2.021,-48.420,7.644],[13.642,-49.286,12.125],[14.791,-51.593,13.340],
[12.554,-54.485,-3.816],[12.585,-54.427,2.614],[13.114,-53.781,7.657],
[10.938,-52.707,10.581],[-2.883,-54.739,13.821],[16.822,58.860,12.035],
[12.918,-58.264,1.056],[-2.652,-54.436,5.731],[-2.532,-51.898,7.273],
[14.794,-58.196,12.676],[16.353,-56.314,-1.862],[-4.009,-55.301,0.563],
[11.952,-43.778,1.993],[-1.885,-57.816,9.778],[-2.003,-45.869,13.830],
[2.133,-43.953,10.572],[6.995,-42.628,12.801],[9.322,-40.971,11.857],
[11.980,58.870,11.793],[-0.490,-39.689,8.720],[1.888,-41.079,9.814],
[7.835,-41.019,7.572],[8.968,-42.841,9.986],[11.524,-42.854,9.718],
[-0.411,-43.150,7.685],[1.004,-43.523,5.021],[2.709,-41.669,5.591],
[6.386,-56.596,6.442],[11.755,-59.071,4.115],[15.286,-58.704,2.893],
[-0.011,-59.560,1.057],[-2.015,-58.288,-0.057],[5.987,-59.003,3.639],
[8.524,-58.408,5.224],[9.924,-59.148,1.564],[-0.717,-43.583,-0.306],
[0.319,-56.520,-1.982],[5.530,-57.220,1.226],[5.070,-43.076,-1.725],
[14.073,-42.165,1.088],[-0.773,-56.674,12.223],[2.962,-58.601,15.583],
[0.216,-40.338,-1.895],[10.556,-60.048,14.433],[12.871,-43.919,12.818],
[8.290,-31.387,2.616],[7.100,-30.988,6.795],[-0.121,-33.244,6.939],
[10.701,-28.293,4.753],[1.179,-28.130,6.110],[-2.940,-35.169,1.696],
[-1.907,-31.751,2.760],[6.548,-23.834,4.561],[6.295,-28.326,5.538],
[14.373,-37.409,8.566],[9.757,-35.757,6.256],[1.876,-36.330,4.391],
[10.461,-36.554,3.437],[-1.961,-33.675,4.701],[4.433,-32.428,5.377],
[14.192,-34.102,10.429],[0.987,-37.985,7.458],[7.446,-27.866,1.955],
[2.976,-28.518,8.311],[11.323,-38.041,5.684],[-0.411,-35.902,6.416],
[0.618,-30.291,7.779],[-1.676,-26.994,-1.221],[10.212,-28.407,7.293],
[10.621,-37.354,8.580],[5.285,-40.368,4.723],[10.240,-24.547,3.839],
[10.067,-27.823,2.299],[11.431,-28.436,11.078],[5.495,-31.112,1.816],
[6.680,-31.943,-3.781],[14.717,-31.010,9.052],[11.243,-32.421,1.198],
[10.755,-33.241,9.535],[1.240,-31.721,3.987],[3.769,-27.525,5.472],
[3.221,-30.378,6.468],[8.388,-26.684,8.361],[0.667,-29.268,10.543],
[9.056,-32.143,-1.515],[0.665,-32.173,10.248],[8.065,-26.035,5.179],
[1.932,-25.104,12.732],[12.736,-39.091,-0.860],[0.630,-30.961,1.165],
[14.423,-26.265,-1.776],[-1.943,-29.771,8.476],[-3.859,-26.556,7.511],
[14.218,-27.217,11.718],[4.323,-25.282,11.337],[8.217,-23.212,2.467],
[1.322,-22.437,2.929],[10.304,-23.777,-1.876],[9.714,-24.745,0.479],
[11.968,-34.251,11.841],[-1.531,-27.877,11.825],[2.865,-39.851,7.698],
[7.274,-40.337,10.102],[10.092,-35.489,10.653],[11.520,-32.192,-3.811],
[6.886,-37.014,-1.525],[0.832,-34.772,9.274],[2.988,-34.052,3.978],
[6.839,-34.147,0.521],[7.878,-34.233,13.812],[9.266,-33.341,4.686],
[-0.063,-37.025,10.976],[2.801,-36.213,-3.481],[-3.399,-38.143,3.799],
[-2.226,-36.569,-1.228],[7.975,-32.123,9.492],[3.925,-37.954,3.024],
[-0.255,-36.095,1.956],[3.567,-36.697,0.359],[8.177,-37.558,7.388],
[1.789,-34.646,-0.751],[11.663,-23.304,14.846],[-0.663,-30.718,12.647],
[7.096,-33.672,7.280],[1.152,-38.348,2.611],[5.090,-35.373,8.149],
[5.207,-27.943,0.595],[12.278,-30.210,7.807],[10.049,-37.667,-1.094],
[6.758,-28.156,11.208],[9.081,-34.943,-1.070],[-0.437,-29.733,4.940],
[6.907,-27.523,-1.805],[-1.855,-32.212,9.743],[5.782,-31.492,14.330],
[7.389,-25.509,11.324],[16.020,-34.074,-1.620],[16.113,-30.001,10.988],
[11.854,-27.978,0.051],[9.832,-31.217,7.684],[2.290,-26.925,3.195],
[-0.093,-31.195,-2.642],[3.938,-29.058,3.166],[4.707,-25.833,8.612],
[3.269,-23.716,1.482],[8.149,-29.358,9.317],[9.717,-29.694,-2.497],
[10.391,-24.960,12.643],[12.997,-23.747,-2.969],[-2.863,-37.035,11.076],
[10.236,-30.160,0.204],[-0.563,-27.557,8.914],[4.119,-27.441,-5.118],
[3.264,-27.922,-1.174],[0.391,-29.215,-0.929],[6.283,-37.486,9.933],
[8.507,-36.811,1.233],[7.198,-24.532,-0.749],[5.827,-36.163,2.042],
[5.164,-35.061,11.095],[6.802,-38.701,2.664],[4.517,-25.504,-0.575],
[2.543,-39.383,-0.787],[12.381,-37.170,10.543],[11.778,-32.932,6.660],
[1.064,-27.327,-2.575],[-0.340,-35.075,-2.483],[10.651,-25.141,7.738],
[-0.971,-41.049,6.270],[-0.059,-25.634,2.651],[-3.707,-24.575,14.248],
[5.453,-24.887,14.146],[6.910,-26.164,-4.332],[8.578,-27.072,13.268],
[11.369,-25.901,-4.246],[-2.359,-28.648,14.500],[2.297,-29.596,-3.357],
[3.238,-30.698,13.597],[6.207,-29.052,-3.921],[9.721,-30.534,11.061],
[-0.969,-32.694,14.450],[4.783,-31.804,-0.789],[2.613,-33.919,11.188],
[3.871,-31.818,-3.387],[7.470,-32.097,12.202],[13.426,-30.006,-2.870],
[2.786,-36.977,11.466],[3.134,-37.282,8.984],[5.057,-32.651,9.364],
[10.596,-34.645,14.046],[10.878,-32.038,12.906],[1.029,-34.909,13.391],
[5.001,-41.095,9.032],[5.442,-36.291,14.525],[7.598,-36.140,11.864],
[8.649,-37.147,14.330],[-1.629,-38.143,13.115],[4.478,-42.168,11.598],
[4.322,-38.739,14.194],[7.015,-20.977,-1.448],[-3.044,-30.113,-1.328],
[-1.537,-23.061,1.765],[-1.047,-24.402,4.760],[15.453,-23.281,7.985],
[13.687,-23.820,10.455],[12.234,-28.354,15.543],[14.327,-29.581,-0.001],
[14.107,-32.291,1.613],[12.603,-27.403,6.851],[12.506,-26.362,9.877],
[14.216,-32.261,-0.991],[13.126,-34.005,-2.698],[11.825,-34.226,3.245],
[-3.194,-30.582,5.393],[12.784,-30.959,11.068],[14.475,-33.007,13.222],
[12.177,-36.756,-4.071],[11.088,-35.353,0.819],[12.952,-35.187,5.578],
[12.764,-35.170,8.428],[-2.006,-35.183,13.193],[13.627,-41.600,13.785],
[13.358,-36.889,0.720],[-3.152,-36.767,6.222],[-4.228,-33.297,6.146],
[13.399,-39.330,12.215],[14.493,-37.852,-2.623],[-1.464,-38.276,1.307],
[13.864,-24.181,2.602],[-3.792,-39.682,8.560],[-1.019,-25.383,13.159],
[1.715,-25.929,8.852],[6.547,-23.254,12.488],[9.558,-22.636,11.658],
[12.690,-39.850,9.772],[-0.568,-21.574,10.084],[1.033,-23.427,8.162],
[5.734,-23.614,7.061],[8.346,-23.949,9.162],[10.532,-22.094,8.017],
[-1.276,-25.660,7.123],[1.444,-25.466,5.874],[4.140,-24.623,5.073],
[5.970,-36.979,5.748],[12.536,-40.093,6.954],[14.764,-41.097,3.962],
[-0.616,-39.015,4.412],[-2.104,-41.263,-0.517],[3.326,-43.215,3.226],
[9.825,-40.473,5.426],[9.169,-39.832,0.328],[-2.556,-26.539,1.942],
[-0.217,-37.961,-2.959],[3.328,-40.673,1.756],[6.755,-23.102,-3.790],
[12.397,-25.024,0.180],[-0.274,-40.591,11.712],[1.338,-40.159,14.080],
[0.445,-23.884,-2.621],[10.760,-39.278,13.533],[13.224,-24.838,13.087],
[6.883,-12.518,3.863],[6.962,-12.559,6.679],[-0.258,-13.674,8.454],
[12.072,-10.152,4.716],[1.110,-9.802,6.506],[-2.094,-13.457,1.775],
[-2.679,-11.388,3.534],[6.727,-5.563,6.107],[6.175,-9.995,6.132],
[14.003,-18.434,6.836],[9.320,-17.042,9.058],[0.339,-16.532,4.972],
[9.272,-16.877,2.301],[-2.358,-17.470,4.198],[3.502,-14.532,4.870],
[14.966,-17.098,9.900],[-0.149,-19.249,8.846],[6.805,-9.322,3.271],
[2.259,-11.929,10.111],[9.814,-19.192,4.309],[-1.065,-17.221,7.100],
[0.656,-9.706,9.108],[-2.430,-8.552,-0.315],[10.891,-8.158,7.054],
[12.268,-18.701,9.675],[4.016,-21.021,6.295],[9.282,-6.710,5.138],
[9.332,-9.461,3.965],[12.769,-9.506,11.155],[6.974,-11.190,-0.001],
[7.712,-13.929,-2.794],[14.057,-14.058,10.149],[9.350,-13.624,2.906],
[9.358,-14.431,10.448],[2.594,-11.577,4.223],[2.566,-7.512,7.358],
[2.652,-13.448,7.521],[7.111,-7.593,8.493],[2.789,-9.148,11.754],
[10.845,-14.503,-1.040],[0.844,-14.200,11.287],[4.581,-8.124,4.671],
[1.561,-7.970,13.894],[13.946,-21.386,-0.855],[0.567,-12.769,2.699],
[14.326,-7.343,-2.744],[-2.810,-12.389,8.860],[-2.882,-7.961,7.506],
[14.788,-6.586,9.967],[6.178,-8.337,12.025],[6.039,-3.364,0.816],
[-0.595,-5.052,3.793],[10.342,-4.788,-2.206],[10.918,-5.970,2.168],
[12.345,-15.798,13.673],[-1.657,-8.914,9.832],[2.063,-20.949,8.158],
[6.813,-22.004,10.060],[10.211,-16.451,12.264],[11.942,-12.277,-2.300],
[5.538,-18.409,-2.245],[2.954,-16.527,7.203],[3.247,-16.415,2.822],
[7.776,-14.905,1.367],[8.006,-15.459,13.253],[8.814,-16.441,5.837],
[-0.231,-16.366,9.873],[2.385,-18.099,-2.036],[-2.800,-20.407,2.780],
[-1.883,-17.908,1.239],[8.216,-14.618,8.084],[4.871,-19.729,3.836],
[0.417,-16.331,2.124],[2.707,-16.322,0.067],[9.051,-19.193,6.943],
[0.938,-14.782,-1.995],[13.733,-3.495,14.133],[-0.381,-11.956,12.501],
[6.020,-15.131,6.406],[2.656,-19.446,1.808],[5.554,-15.902,9.073],
[2.539,-9.379,-0.697],[11.227,-11.583,9.325],[9.880,-20.670,-1.572],
[7.808,-10.553,11.300],[10.372,-17.712,-0.569],[-0.143,-12.031,5.291],
[7.184,-7.750,-0.446],[-3.480,-12.789,11.665],[6.093,-13.610,13.864],
[8.654,-7.404,11.116],[15.211,-15.674,0.030],[15.702,-9.867,11.199],
[12.899,-7.883,-0.310],[10.426,-12.231,6.538],[1.924,-7.594,4.085],
[-1.214,-12.967,-1.317],[4.841,-10.899,2.828],[4.144,-6.111,8.978],
[3.279,-5.663,2.799],[9.046,-9.279,9.252],[9.776,-10.434,-1.519],
[10.825,-5.111,11.278],[15.187,-6.198,0.410],[-1.499,-18.464,11.446],
[10.023,-12.271,0.525],[-3.276,-6.085,9.246],[4.273,-8.587,-4.704],
[2.969,-7.598,-2.475],[0.139,-10.868,1.010],[6.048,-18.394,10.422],
[7.635,-17.734,-0.667],[8.090,-5.116,1.264],[5.633,-16.384,0.974],
[5.659,-15.899,12.256],[7.551,-19.919,2.977],[4.853,-6.626,-0.614],
[3.749,-20.496,-1.199],[10.459,-19.980,12.065],[12.455,-15.521,6.945],
[0.408,-7.265,-1.862],[-1.214,-16.551,-3.098],[11.912,-4.270,9.220],
[1.126,-20.761,5.348],[-0.072,-7.143,2.406],[-4.410,-7.067,14.189],
[0.837,-5.645,15.740],[6.152,-6.528,-3.132],[9.036,-9.541,13.480],
[11.248,-7.183,-2.720],[-3.615,-11.399,14.191],[1.416,-9.554,-3.718],
[3.858,-11.559,12.924],[6.929,-8.770,-4.472],[11.452,-10.980,13.317],
[-2.216,-13.901,13.687],[4.219,-11.845,-1.278],[2.999,-16.193,9.771],
[3.752,-14.054,-2.710],[6.958,-11.157,13.966],[13.747,-9.971,-2.277],
[3.134,-18.964,12.084],[3.680,-18.880,9.095],[4.301,-13.640,11.210],
[9.452,-16.714,15.386],[11.286,-13.815,12.262],[0.911,-14.609,14.219],
[3.980,-22.717,9.740],[6.848,-17.392,16.104],[8.107,-17.789,11.686],
[8.704,-19.222,14.671],[-0.109,-17.764,14.053],[3.441,-23.199,13.598],
[4.324,-20.624,14.039],[5.145,-2.584,-2.141],[-2.265,-8.880,-2.979],
[-1.477,-3.879,1.518],[-3.803,-7.502,4.270],[13.907,-5.130,7.576],
[13.322,-5.092,11.576],[12.213,-8.378,13.454],[14.342,-11.830,0.097],
[14.645,-13.088,2.344],[14.012,-9.754,6.564],[12.107,-7.287,9.893],
[14.266,-13.995,-1.916],[13.876,-16.599,-2.424],[11.748,-14.804,3.247],
[-2.855,-12.596,6.123],[13.554,-12.550,13.735],[15.477,-15.178,14.441],
[12.841,-18.203,-4.543],[12.028,-19.266,0.662],[11.263,-17.796,5.942],
[12.269,-15.991,10.805],[-2.920,-16.954,14.038],[15.489,-21.677,15.125],
[12.830,-21.562,1.698],[-2.006,-19.658,6.337],[-3.712,-15.391,6.600],
[13.274,-19.754,13.301],[15.760,-18.799,-2.901],[-3.212,-20.150,0.154],
[12.490,-4.979,4.577],[-2.983,-20.206,9.965],[-1.316,-6.921,14.032],
[1.704,-6.850,10.064],[6.049,-4.407,13.257],[8.213,-4.694,12.086],
[13.299,-21.005,10.752],[-1.193,-3.592,9.055],[0.698,-4.445,10.904],
[6.098,-4.081,8.391],[8.666,-5.187,9.225],[10.277,-2.221,10.293],
[-1.532,-5.570,7.135],[0.048,-7.366,6.234],[3.623,-5.486,5.924],
[6.539,-19.714,5.953],[12.192,-22.265,4.390],[14.686,-21.729,3.901],
[-1.056,-21.798,4.113],[-2.442,-22.808,-0.843],[5.550,-22.305,2.301],
[9.702,-21.618,5.224],[10.020,-20.846,1.770],[-1.159,-6.169,0.181],
[0.254,-19.987,-2.377],[5.593,-19.121,1.115],[7.848,-4.987,-1.429],
[13.379,-5.538,1.985],[-0.402,-20.520,13.760],[1.519,-21.934,15.363],
[0.669,-4.870,-2.938],[8.574,-22.119,14.274],[11.452,-5.318,13.855],
[7.919,5.477,4.242],[6.663,5.940,7.135],[0.761,4.553,6.969],
[9.566,8.555,5.399],[1.574,8.936,6.217],[-1.886,4.224,1.436],
[-3.098,4.305,4.267],[8.100,14.335,4.576],[5.042,9.136,6.418],
[16.159,0.202,7.240],[9.706,1.797,8.198],[0.307,0.110,3.456],
[10.044,0.202,3.078],[-1.648,2.223,3.995],[4.546,5.037,5.193],
[14.089,2.793,10.632],[0.753,-1.289,10.607],[7.264,8.766,2.607],
[3.542,9.150,8.662],[12.668,-0.055,5.839],[-0.619,1.688,6.720],
[1.864,7.770,11.299],[-1.735,8.858,0.266],[10.069,8.599,7.932],
[11.248,-0.675,7.956],[5.403,-3.051,5.776],[12.292,12.194,4.732],
[10.273,11.061,2.498],[13.748,9.188,10.548],[5.813,5.420,2.144],
[7.554,4.538,-0.380],[15.333,5.750,10.407],[11.357,6.137,2.477],
[9.111,4.626,9.804],[2.434,6.479,3.573],[3.450,10.962,5.329],
[2.791,6.507,7.034],[10.434,11.501,8.020],[4.171,9.053,11.356],
[11.019,6.412,-0.232],[0.885,5.231,11.471],[7.213,11.954,2.832],
[1.964,11.872,12.558],[12.935,-3.334,-1.755],[0.439,5.417,1.704],
[12.621,10.302,-1.257],[-2.117,8.034,9.153],[-2.705,10.632,8.338],
[15.691,10.149,12.625],[6.444,10.952,11.372],[7.073,14.838,1.516],
[-0.844,13.332,3.945],[9.761,13.342,-1.995],[9.554,13.159,0.771],
[12.154,3.784,14.896],[-0.976,10.393,11.636],[3.805,-0.676,6.630],
[7.085,-2.870,10.649],[10.285,2.040,12.572],[11.519,6.754,-2.706],
[6.298,0.443,-3.007],[1.910,2.049,7.402],[2.794,2.657,2.695],
[8.688,2.418,0.613],[7.508,3.795,14.058],[9.207,1.955,5.353],
[-1.081,3.389,9.123],[2.916,2.343,-3.107],[-2.683,-1.646,2.844],
[-1.773,1.521,0.882],[6.275,3.972,8.877],[3.296,-0.368,3.987],
[1.306,0.238,0.870],[2.676,3.207,0.087],[8.351,-0.839,9.046],
[1.410,4.431,-1.970],[11.382,15.728,13.455],[-0.873,8.023,12.991],
[5.835,2.198,6.560],[1.863,-2.325,1.920],[4.363,1.914,8.739],
[5.824,7.139,0.096],[12.056,6.857,8.273],[11.124,-0.872,-1.055],
[7.028,8.287,11.287],[11.748,3.833,-0.622],[0.145,6.589,5.023],
[8.337,7.113,-0.589],[-2.277,5.136,11.245],[3.579,4.098,11.880],
[9.068,10.365,10.332],[14.270,2.804,-0.942],[17.059,7.880,11.865],
[12.790,10.172,1.693],[9.246,5.775,6.792],[2.562,10.285,2.882],
[-1.244,6.222,-2.167],[4.305,8.067,2.043],[2.640,12.898,10.016],
[3.821,12.904,3.035],[6.742,8.244,8.605],[9.983,9.146,-1.447],
[11.448,12.917,13.013],[13.541,14.136,-2.435],[-1.376,1.047,10.633],
[9.794,8.165,1.568],[-0.436,12.304,9.544],[4.587,9.579,-3.581],
[3.445,9.252,-1.072],[0.859,7.458,-0.725],[5.811,-0.224,9.576],
[9.119,-0.691,0.527],[6.948,12.188,0.002],[5.652,2.314,1.268],
[6.207,2.285,11.406],[6.502,-2.063,3.076],[3.664,11.892,-1.112],
[4.194,-0.451,1.497],[11.949,-1.067,13.193],[11.096,4.057,6.896],
[0.401,11.069,-0.517],[-1.454,2.439,-2.124],[11.859,13.674,7.600],
[0.016,-3.103,6.462],[-0.064,10.248,2.188],[-2.813,13.468,14.856],
[3.805,13.566,15.025],[6.267,11.701,-2.627],[10.308,10.474,14.237],
[10.907,11.153,-3.088],[-2.770,6.851,14.891],[2.162,8.987,-4.537],
[4.566,6.719,13.141],[6.610,7.808,-2.370],[8.738,6.320,11.964],
[-0.861,5.125,13.470],[3.066,6.055,-0.348],[2.154,3.741,9.589],
[4.507,5.353,-2.673],[8.251,6.755,14.837],[14.589,8.675,-2.356],
[3.334,1.127,11.892],[2.831,-0.269,9.025],[4.650,6.186,9.016],
[9.704,2.032,15.363],[11.199,4.686,12.296],[1.671,4.017,14.304],
[3.517,-2.989,9.699],[5.102,1.280,14.397],[7.680,-0.313,11.780],
[9.269,-0.434,14.497],[0.279,1.674,13.005],[3.489,-4.703,12.504],
[5.074,-2.418,14.993],[3.104,14.896,-1.557],[-2.378,10.925,-2.348],
[-2.323,14.258,1.683],[-3.435,12.960,4.807],[14.464,13.285,8.227],
[14.797,12.883,10.892],[12.477,9.076,14.963],[13.309,7.734,0.856],
[14.562,5.046,3.043],[12.902,9.313,6.516],[12.445,11.664,10.154],
[14.070,5.748,-2.321],[13.098,2.235,-3.809],[12.242,3.581,3.686],
[-2.787,6.502,6.568],[14.327,7.551,12.460],[14.985,4.362,13.178],
[12.224,-0.402,-4.127],[12.428,1.229,1.418],[12.771,1.782,8.197],
[13.070,1.799,13.184],[-3.511,1.185,12.248],[16.303,-4.071,15.533],
[13.548,-3.144,0.812],[-1.779,-0.464,5.156],[-2.823,3.364,7.047],
[14.897,-1.779,12.529],[15.391,-1.076,-2.416],[-2.700,-0.689,-0.238],
[12.417,14.352,2.685],[-2.875,-1.253,9.814],[0.648,10.945,14.745],
[1.415,10.513,10.016],[7.164,13.513,12.313],[9.723,14.322,11.596],
[12.882,-2.123,10.714],[-0.000,14.556,11.418],[0.522,14.560,8.649],
[6.315,14.772,7.626],[8.531,13.451,7.368],[10.737,16.028,9.670],
[-1.141,13.785,6.627],[0.974,11.660,4.726],[2.054,13.972,5.721],
[6.585,-0.385,5.334],[12.535,-2.906,6.684],[13.168,-2.515,3.540],
[-0.018,-2.422,3.879],[-1.489,-3.031,-1.866],[3.927,-3.168,3.320],
[9.690,-3.280,5.041],[10.329,-3.148,-0.059],[-2.231,11.732,0.393],
[-0.189,-0.264,-1.465],[6.478,-0.263,-0.260],[6.895,14.326,-2.076],
[13.926,12.679,1.111],[-1.044,-0.835,12.623],[0.892,-2.125,14.060],
[0.632,14.970,-1.519],[10.033,-2.939,13.963],[13.488,11.596,14.003],
[9.788,25.778,4.384],[7.404,24.434,6.977],[-0.446,21.833,8.024],
[11.265,27.439,5.795],[0.515,27.661,6.868],[-2.231,22.962,2.244],
[-3.665,23.905,4.503],[9.262,32.443,4.297],[5.134,25.639,7.859],
[14.270,18.363,6.992],[9.385,20.739,8.321],[2.410,20.035,3.976],
[9.677,18.344,3.487],[-2.248,21.639,5.157],[4.834,23.869,5.715],
[15.705,19.766,9.780],[0.140,18.216,7.670],[7.527,27.097,1.689],
[1.204,25.825,8.929],[11.957,18.649,4.959],[0.664,20.942,5.823],
[2.187,28.202,9.482],[-1.016,27.603,0.324],[11.196,27.566,9.157],
[12.915,17.763,9.476],[5.302,14.909,5.290],[10.555,30.160,5.545],
[12.181,29.231,2.389],[11.703,26.786,11.830],[7.834,23.911,1.916],
[7.429,21.725,-1.217],[14.137,22.651,9.455],[10.008,25.558,1.707],
[9.592,23.218,9.485],[2.535,24.999,3.488],[3.170,27.360,4.995],
[2.788,23.469,7.977],[9.812,29.860,8.427],[1.753,27.422,12.232],
[9.986,23.244,-0.318],[1.467,23.916,11.395],[6.526,31.363,3.453],
[1.199,30.923,11.166],[13.819,15.284,0.134],[0.401,23.832,1.737],
[14.960,29.687,-2.427],[-4.116,25.578,10.184],[-4.262,29.188,5.983],
[14.718,29.306,12.408],[5.314,28.272,10.075],[6.818,33.017,1.479],
[-0.165,31.968,2.980],[9.079,30.929,-2.074],[10.955,30.787,0.508],
[12.614,21.809,13.802],[-0.483,28.878,11.618],[1.981,16.270,7.430],
[5.402,13.267,9.578],[10.517,19.969,11.806],[10.158,24.729,-2.636],
[6.508,18.755,-3.529],[1.602,20.461,8.551],[2.684,22.623,4.669],
[9.244,21.136,2.212],[6.370,22.959,13.176],[10.099,20.756,4.933],
[0.238,21.385,10.881],[2.341,18.705,-1.923],[-3.863,17.119,5.718],
[-1.641,19.826,0.559],[6.762,21.683,8.808],[4.682,18.744,4.783],
[-0.319,21.079,2.954],[1.295,19.511,0.850],[9.888,18.051,7.296],
[2.360,21.860,-1.233],[10.906,34.904,13.376],[0.221,25.546,13.926],
[6.898,21.390,5.077],[1.205,17.271,4.314],[4.920,20.106,7.820],
[6.999,24.400,-0.851],[12.549,25.573,7.474],[8.826,19.306,-2.022],
[7.399,27.252,11.478],[10.649,19.899,0.309],[0.698,25.108,5.336],
[5.852,27.721,-0.466],[-2.694,22.241,11.445],[3.431,22.145,12.929],
[8.574,30.360,10.801],[17.241,21.870,0.064],[14.364,26.455,12.338],
[14.469,29.125,0.169],[9.724,25.782,7.507],[3.884,29.318,0.937],
[-1.430,24.938,-1.992],[4.463,27.013,2.596],[4.547,30.042,8.398],
[1.941,30.470,3.375],[7.604,26.753,8.728],[10.169,27.508,-1.768],
[10.883,31.586,11.562],[12.749,33.633,0.039],[-1.117,19.123,12.114],
[11.673,26.956,0.294],[-1.559,29.559,8.764],[4.937,28.330,-3.053],
[3.406,24.510,-1.335],[0.750,25.786,-0.348],[6.210,18.535,10.248],
[8.049,19.221,0.839],[6.142,30.843,-0.492],[5.563,19.573,1.879],
[6.107,20.990,11.436],[6.847,17.431,2.827],[3.033,30.536,-1.702],
[2.602,16.748,0.511],[12.741,19.206,12.958],[10.288,23.105,6.767],
[0.969,28.975,-1.669],[-0.173,20.724,-1.941],[10.295,33.059,7.451],
[-1.335,16.295,7.201],[-1.156,28.976,2.602],[-2.937,32.568,13.227],
[2.073,30.807,15.226],[7.172,30.039,-4.035],[8.662,29.156,13.303],
[12.428,29.014,-3.918],[-3.213,25.503,14.023],[1.361,26.036,-2.970],
[4.454,27.078,12.472],[7.474,26.837,-4.650],[9.282,25.626,10.720],
[-1.159,22.725,13.835],[4.733,22.118,-0.178],[3.703,21.478,9.949],
[3.784,23.182,-3.592],[8.622,24.996,14.275],[13.010,26.648,-2.114],
[3.720,19.570,11.861],[3.470,18.418,9.243],[4.802,24.132,10.237],
[10.786,20.612,15.483],[9.978,23.754,12.380],[1.053,20.489,13.668],
[3.433,15.502,9.700],[5.533,19.137,14.671],[7.856,19.829,12.879],
[9.403,17.674,13.482],[0.221,17.765,14.472],[4.657,14.728,11.909],
[4.764,16.242,14.201],[3.904,32.880,-0.844],[-1.920,27.940,-2.276],
[-2.451,33.209,2.784],[-2.291,30.717,4.290],[13.640,32.515,6.821],
[14.684,32.404,11.085],[13.685,27.700,14.626],[14.120,26.812,1.546],
[13.807,24.384,2.647],[14.833,27.169,7.382],[12.852,30.159,10.056],
[15.881,24.244,-1.449],[13.710,21.675,-2.096],[12.720,21.615,3.651],
[-1.940,25.416,6.429],[12.496,23.189,11.297],[15.230,22.018,12.905],
[11.704,19.098,-2.351],[13.385,19.555,1.300],[11.974,20.428,7.127],
[12.717,20.445,10.476],[-3.916,19.420,13.780],[16.069,15.664,15.963],
[12.268,17.241,2.053],[-2.547,19.791,7.363],[-2.007,23.923,9.043],
[13.916,16.472,13.520],[14.024,18.136,-3.250],[-3.027,17.963,1.824],
[12.105,31.930,4.455],[-2.140,16.632,11.043],[0.015,28.582,14.042],
[1.791,30.939,8.115],[5.945,31.514,11.126],[8.203,33.342,12.253],
[13.145,14.888,11.465],[-1.404,34.208,10.124],[1.379,33.258,9.620],
[6.876,32.946,6.173],[7.452,32.713,8.953],[9.783,33.880,10.070],
[-1.562,31.431,6.972],[4.911,29.454,4.244],[2.822,29.737,5.998],
[7.380,16.879,5.635],[14.223,15.826,7.860],[14.384,16.223,3.725],
[-1.381,15.894,4.026],[-2.093,16.362,-0.776],[4.481,15.782,2.080],
[11.369,16.136,6.186],[10.452,15.682,1.112],[-0.361,30.472,0.252],
[-0.459,18.151,-2.895],[6.518,17.095,-1.186],[6.308,33.765,-1.774],
[13.289,32.430,2.258],[-1.609,15.794,13.958],[2.138,15.953,14.890],
[-0.669,33.283,-2.921],[9.387,15.552,15.351],[12.759,31.395,13.910],
[7.222,43.010,4.229],[7.429,42.939,6.722],[-0.582,41.747,7.072],
[10.865,45.086,3.734],[1.132,46.383,4.981],[-3.202,40.714,2.674],
[-2.299,42.790,3.974],[7.714,50.285,5.859],[6.878,45.697,8.150],
[13.098,36.644,8.670],[9.043,39.054,7.447],[0.510,36.904,3.622],
[10.227,37.299,1.353],[-1.294,39.386,4.141],[4.829,41.116,4.513],
[13.743,38.916,9.746],[0.782,36.709,8.346],[7.521,45.990,4.325],
[3.411,44.478,8.409],[11.089,37.262,4.730],[0.389,38.749,6.618],
[1.271,45.807,9.778],[-1.498,45.549,-0.049],[10.244,48.215,7.286],
[10.564,36.922,8.144],[4.287,33.771,5.358],[9.757,48.533,4.641],
[10.323,47.500,2.192],[12.523,45.205,10.357],[7.643,43.783,1.783],
[7.701,41.525,-1.943],[14.646,42.025,9.340],[11.140,43.224,1.323],
[9.468,41.032,10.479],[1.390,43.220,4.621],[3.454,47.803,6.349],
[3.217,42.039,6.718],[7.202,48.260,8.836],[2.873,47.078,11.552],
[9.258,42.736,-0.373],[0.475,42.911,10.520],[7.262,49.524,2.907],
[1.736,48.681,13.509],[12.709,34.223,-2.910],[1.811,42.058,1.386],
[14.251,49.734,-1.317],[-1.972,44.930,9.102],[-3.052,48.309,7.770],
[15.149,49.335,12.235],[5.470,46.960,12.636],[6.761,52.105,1.028],
[0.166,50.061,1.315],[9.116,49.929,-1.752],[9.218,50.911,1.290],
[11.832,39.768,11.984],[-1.749,45.790,11.686],[0.759,34.908,6.297],
[6.377,34.336,10.844],[9.683,38.535,13.138],[12.003,44.047,-3.352],
[7.484,35.701,-3.232],[1.346,40.390,8.537],[3.137,39.127,3.198],
[6.439,40.710,0.239],[7.062,40.427,13.591],[8.389,40.426,5.087],
[1.095,38.372,10.456],[2.208,38.002,-3.770],[-2.173,36.044,4.296],
[-3.721,38.206,1.823],[7.241,41.868,9.181],[4.633,36.762,6.030],
[-0.235,39.841,1.450],[2.800,38.664,-0.397],[6.976,36.230,7.733],
[4.740,39.698,-2.056],[11.266,51.932,14.506],[-0.153,45.196,14.101],
[5.984,39.239,6.397],[0.082,37.139,0.702],[3.318,38.003,8.907],
[5.374,43.605,-0.639],[11.570,43.310,8.203],[8.830,37.065,-1.380],
[6.712,45.013,11.146],[9.540,40.075,0.082],[0.524,44.243,7.195],
[6.692,46.192,1.579],[-2.756,42.378,10.637],[4.229,41.375,13.228],
[7.777,48.284,12.273],[15.797,40.163,0.134],[15.449,44.819,11.725],
[12.725,48.085,0.734],[10.219,44.549,6.229],[2.470,48.855,3.543],
[-1.324,43.328,-2.061],[3.086,45.293,3.193],[4.439,49.379,9.766],
[1.318,51.847,3.078],[8.952,46.478,10.374],[9.983,46.155,-2.665],
[9.411,49.622,10.690],[16.261,51.615,-1.314],[-1.611,38.561,10.667],
[9.722,45.274,0.422],[-0.563,47.844,8.801],[4.553,47.055,-3.621],
[4.306,46.169,-0.065],[1.534,44.814,-0.121],[5.862,37.145,10.183],
[8.417,34.907,0.230],[7.246,49.101,0.043],[6.191,38.539,1.795],
[5.201,39.743,11.181],[10.746,34.188,2.324],[2.996,49.175,0.563],
[3.581,36.131,-1.468],[12.611,36.906,12.377],[10.922,40.905,6.457],
[0.987,47.367,-1.127],[-2.399,40.450,-3.164],[13.677,49.345,7.545],
[-1.533,33.654,5.353],[0.431,47.520,2.163],[-1.933,51.251,13.727],
[3.552,51.262,14.469],[6.427,48.973,-3.880],[9.029,49.958,13.964],
[11.929,48.904,-2.322],[-2.500,43.861,13.979],[2.365,45.227,-3.054],
[3.720,44.316,12.992],[7.599,46.372,-3.952],[9.608,43.811,10.623],
[-1.382,41.295,13.839],[1.997,42.601,-1.600],[2.494,40.707,10.918],
[4.967,42.343,-2.802],[5.926,43.170,14.306],[14.794,44.852,-2.503],
[2.999,37.465,12.349],[3.013,35.266,7.754],[4.869,42.748,10.384],
[11.249,39.958,14.849],[10.928,42.628,12.768],[1.852,40.148,13.925],
[4.080,33.142,8.952],[5.232,37.998,14.022],[7.808,38.948,10.676],
[8.498,35.976,13.449],[-0.666,38.284,13.258],[3.850,33.896,11.571],
[4.424,34.012,14.705],[3.844,52.658,-2.343],[-1.697,47.630,-2.062],
[-5.027,52.595,4.178],[-1.466,50.162,4.347],[14.935,50.995,5.992],
[13.877,51.763,9.866],[11.330,46.452,13.999],[13.532,45.275,0.038],
[13.836,43.014,1.960],[12.734,46.365,6.580],[12.558,47.817,9.454],
[14.619,41.970,-1.811],[14.296,40.787,-4.240],[12.225,41.791,3.725],
[-2.664,43.532,6.958],[13.253,44.549,13.082],[13.538,41.705,13.409],
[12.373,38.179,-2.938],[12.245,38.375,0.011],[12.796,39.344,5.299],
[12.217,40.706,8.821],[-4.179,38.629,13.426],[15.083,35.066,15.668],
[12.973,36.032,1.668],[-2.835,37.618,6.671],[-3.751,40.617,6.616],
[15.858,35.064,12.583],[14.890,35.776,-3.636],[-3.158,35.485,0.655],
[12.056,50.867,1.676],[-2.701,34.629,7.692],[-2.930,47.535,13.712],
[2.399,50.062,11.431],[5.565,50.379,12.414],[9.430,53.480,12.211],
[12.483,33.574,9.834],[-0.604,51.103,10.347],[3.102,51.509,8.113],
[6.328,51.687,8.825],[8.939,51.343,8.353],[11.626,50.383,8.754],
[0.230,51.511,7.581],[0.759,48.860,5.942],[3.118,50.816,5.192],
[7.557,35.342,4.116],[12.254,34.944,6.303],[13.738,34.206,4.265],
[0.873,34.373,3.210],[-1.085,33.534,0.526],[3.333,35.848,3.731],
[9.423,35.642,6.166],[11.209,35.718,-1.152],[-2.003,48.386,0.781],
[-0.594,35.601,-1.238],[4.726,35.763,1.116],[6.474,51.016,-1.907],
[14.319,51.709,0.602],[-1.098,35.733,12.227],[1.755,34.885,13.143],
[0.048,50.336,-2.515],[9.177,33.861,15.102],[12.233,50.119,11.620],
[14.708,54.279,7.482],[9.777,58.497,6.963],[0.435,56.528,4.867],
[10.489,57.014,3.075],[-1.114,59.060,4.433],[14.451,58.881,10.196],
[0.720,55.806,9.287],[12.593,55.932,4.432],[-0.049,57.422,7.487],
[11.342,55.507,9.227],[5.330,52.599,5.485],[12.833,51.873,-1.497],
[13.028,55.724,14.231],[2.309,53.685,5.999],[7.064,53.058,11.217],
[9.578,57.532,10.914],[6.705,53.803,-2.956],[1.504,58.784,5.879],
[3.253,-58.771,3.857],[6.223,57.272,4.180],[-0.498,58.611,9.830],
[2.017,54.860,-3.768],[-3.582,55.677,2.824],[-1.842,56.969,0.177],
[3.918,55.137,4.257],[-0.233,57.038,2.335],[1.578,57.025,-0.020],
[9.360,55.610,5.602],[2.955,57.729,-2.541],[6.567,58.123,6.707],
[1.917,54.390,2.503],[3.616,56.889,8.611],[8.290,55.878,-1.930],
[9.000,57.295,0.389],[14.877,58.898,0.362],[-1.564,57.273,12.016],
[8.306,56.934,8.428],[8.613,54.767,0.700],[5.865,57.728,0.912],
[5.912,55.702,10.381],[5.903,55.271,2.039],[1.514,54.433,-0.365],
[13.707,55.969,11.344],[-0.273,58.253,-2.396],[-0.236,54.577,6.700],
[2.961,56.756,13.483],[2.837,55.577,11.098],[8.460,57.872,16.271],
[1.089,58.594,14.427],[1.978,53.213,10.065],[5.638,57.815,14.039],
[7.817,58.852,12.561],[10.208,54.773,15.046],[0.971,56.013,15.306],
[4.456,52.253,11.242],[4.263,54.219,13.862],[13.901,-59.272,-2.101],
[12.714,56.401,-2.763],[12.012,57.408,0.380],[11.954,57.028,6.767],
[11.936,58.681,8.960],[-2.916,57.070,14.301],[14.144,52.008,14.384],
[11.567,54.294,2.573],[-2.401,55.546,5.263],[-2.943,57.593,6.866],
[14.704,54.185,12.973],[14.876,54.616,-2.488],[-2.290,53.950,1.116],
[-2.894,54.799,11.050],[11.918,53.608,11.262],[5.085,55.079,6.745],
[12.240,50.054,5.038],[12.436,53.368,5.370],[-0.763,53.806,3.760],
[-1.121,52.764,-1.034],[4.130,52.497,2.800],[9.040,52.700,4.876],
[11.267,53.057,0.304],[-0.327,55.493,-2.617],[5.349,54.225,-0.579],
[0.040,54.809,12.325],[1.274,52.019,13.103],[8.138,52.810,14.809],
[8.650,-50.768,20.116],[5.188,-48.684,26.028],[-0.175,-50.919,27.015],
[12.174,-47.783,23.124],[0.309,-46.992,23.918],[-2.695,-52.846,20.530],
[-2.968,-51.266,22.697],[9.292,-43.856,24.554],[7.574,-46.846,26.223],
[15.043,-57.754,26.598],[10.039,-54.124,26.109],[1.144,-54.791,23.629],
[8.806,-54.390,21.919],[-1.236,-53.320,22.850],[3.874,-51.409,22.311],
[14.212,-53.777,27.695],[-0.230,-56.466,27.209],[8.892,-46.686,21.141],
[3.035,-49.402,27.560],[11.184,-55.130,23.617],[0.298,-53.159,25.582],
[2.927,-46.810,28.562],[-2.103,-46.746,19.355],[12.271,-45.930,25.172],
[12.970,-56.530,28.706],[4.470,-58.988,25.216],[9.584,-44.047,21.996],
[12.018,-46.383,18.833],[11.440,-46.038,30.180],[6.015,-50.801,20.739],
[8.247,-51.841,15.643],[14.129,-50.737,27.561],[10.663,-48.868,20.872],
[9.774,-52.011,29.221],[1.838,-48.797,22.217],[2.778,-47.416,25.500],
[3.015,-50.011,24.410],[9.479,-44.823,26.873],[4.228,-47.117,30.883],
[10.045,-51.032,17.568],[0.868,-49.171,28.956],[6.248,-42.837,21.248],
[1.988,-43.959,31.211],[12.354,-58.749,19.551],[-0.179,-49.975,20.343],
[13.606,-44.896,15.419],[-2.003,-48.509,28.638],[-3.411,-44.333,26.234],
[15.933,-44.026,30.009],[6.644,-45.347,31.032],[6.286,-41.096,18.972],
[0.133,-44.056,23.029],[9.905,-41.535,17.219],[8.603,-42.380,19.538],
[13.261,-52.510,31.824],[0.082,-46.247,29.754],[2.770,-56.707,24.797],
[4.858,-58.205,28.759],[11.494,-54.264,28.936],[10.736,-49.413,13.873],
[6.953,-54.398,17.077],[4.102,-55.944,26.930],[1.806,-53.490,21.179],
[6.819,-51.708,17.889],[6.825,-51.597,32.034],[7.815,-52.020,24.393],
[0.176,-53.248,28.427],[3.860,-55.158,15.815],[-3.640,-55.869,22.010],
[-3.359,-55.575,16.441],[7.727,-51.847,27.073],[4.677,-55.076,22.023],
[-0.898,-55.971,21.705],[2.270,-53.450,18.219],[9.145,-56.659,26.677],
[0.466,-52.855,16.000],[12.084,-39.860,32.459],[1.417,-50.295,31.958],
[4.986,-52.584,24.429],[1.606,-57.110,21.842],[6.061,-53.854,27.191],
[5.658,-48.650,18.892],[12.004,-49.287,27.094],[9.622,-54.628,16.702],
[7.959,-47.789,30.204],[9.981,-53.148,19.713],[-0.035,-50.666,23.682],
[7.834,-45.931,18.508],[-4.480,-51.999,30.362],[3.912,-52.060,31.433],
[8.825,-43.552,30.735],[14.615,-52.995,18.163],[13.849,-47.766,29.247],
[10.658,-44.029,19.472],[10.101,-47.732,24.950],[3.379,-43.916,22.636],
[0.194,-49.640,15.464],[4.669,-47.985,21.506],[4.264,-44.360,29.804],
[1.010,-42.263,20.033],[9.024,-49.664,27.415],[9.647,-47.081,15.500],
[11.288,-43.621,31.488],[13.989,-41.988,16.663],[-1.448,-55.163,29.517],
[9.450,-48.299,18.172],[-0.978,-45.383,26.916],[3.639,-46.112,17.307],
[1.579,-46.440,19.769],[-0.470,-48.549,17.792],[6.435,-55.888,28.879],
[8.542,-55.963,19.207],[5.145,-44.295,19.302],[5.013,-53.799,19.616],
[5.019,-54.645,31.580],[7.265,-56.997,21.423],[3.973,-43.025,16.447],
[3.042,-56.498,19.197],[14.965,-54.638,30.010],[10.815,-51.263,24.679],
[1.441,-45.196,15.914],[-2.284,-51.972,16.300],[11.750,-42.177,26.998],
[0.469,-58.237,24.123],[-2.545,-44.229,20.183],[-2.060,-43.866,33.206],
[2.074,-40.752,31.948],[6.647,-45.247,15.460],[9.694,-46.652,32.460],
[10.900,-44.858,14.503],[-2.942,-48.937,32.936],[2.883,-48.231,15.486],
[4.572,-49.956,33.349],[6.146,-47.962,16.332],[10.629,-49.023,30.342],
[-1.702,-50.909,29.969],[3.964,-50.780,18.987],[2.323,-51.241,29.373],
[5.116,-51.742,15.051],[8.434,-49.659,32.878],[12.773,-47.971,15.638],
[1.254,-53.863,30.648],[2.284,-56.567,29.267],[5.075,-50.987,28.395],
[9.617,-54.035,35.141],[10.263,-51.850,33.512],[1.071,-52.939,33.253],
[1.608,58.107,28.805],[7.335,-54.031,32.837],[10.481,-53.559,31.340],
[9.157,-56.026,33.080],[1.226,-56.102,32.282],[4.355,-59.280,32.229],
[4.833,-56.707,33.525],[5.301,-39.993,16.332],[-1.240,-46.242,16.532],
[-3.385,-40.821,21.102],[-3.057,-43.563,23.162],[14.929,-40.595,26.302],
[13.723,-42.426,30.020],[13.557,-46.290,32.839],[14.600,-47.160,19.274],
[13.234,-51.359,20.034],[14.434,-48.211,26.440],[13.163,-44.978,28.767],
[14.823,-50.581,16.970],[13.453,-53.297,15.251],[10.970,-51.439,21.707],
[-2.435,-49.704,25.119],[13.439,-49.572,31.546],[15.450,-51.443,34.207],
[14.276,-56.129,14.627],[12.911,-54.252,20.356],[13.314,-53.964,25.039],
[11.825,-52.135,27.231],[-1.930,-56.174,32.291],[15.940,-58.110,33.250],
[11.999,-56.561,20.780],[-1.998,-57.150,25.177],[-2.597,-52.997,25.938],
[14.132,-56.926,31.779],[15.598,-57.729,16.437],[-1.768,-56.252,18.916],
[10.938,-41.963,23.091],[-2.053,-57.868,28.651],[-1.708,-46.453,32.177],
[1.435,-44.343,28.429],[6.456,-42.430,31.259],[9.932,-40.337,30.433],
[13.257,-59.258,28.918],[0.559,-38.437,29.130],[0.534,-41.884,26.771],
[6.968,-41.003,26.303],[8.960,-42.633,28.235],[10.105,-40.005,27.479],
[-1.905,-42.075,25.200],[1.779,-44.335,25.778],[5.328,-43.006,24.796],
[6.759,-56.359,25.848],[12.036,-57.767,23.397],[14.542,-57.752,24.003],
[0.048,58.693,21.428],[-1.116,-59.397,19.069],[5.026,-57.822,22.583],
[9.330,-58.959,23.819],[9.791,-56.904,22.045],[-0.276,-43.949,17.679],
[0.140,-56.992,16.787],[5.788,-57.311,19.024],[6.859,-42.433,15.969],
[13.132,-42.019,19.695],[-0.974,-58.461,31.063],[-0.410,-58.307,33.600],
[0.798,-42.408,15.558],[11.032,-57.574,32.365],[13.157,-42.677,33.133],
[6.814,-31.943,22.639],[6.902,-31.995,25.262],[-1.288,-32.540,27.671],
[11.200,-28.287,24.450],[0.917,-28.727,25.277],[-2.467,-32.523,19.403],
[-2.055,-31.229,22.025],[8.783,-25.202,24.790],[6.240,-29.284,25.484],
[13.548,-38.319,27.177],[10.146,-35.231,26.317],[2.049,-36.506,22.875],
[9.948,-37.412,20.910],[-1.317,-33.892,22.874],[3.996,-33.098,22.839],
[15.561,-35.590,29.225],[0.074,-37.361,26.450],[6.621,-27.372,20.878],
[2.763,-28.977,27.134],[11.525,-37.995,23.503],[0.785,-34.842,24.666],
[-0.115,-28.734,27.904],[-2.188,-27.979,18.727],[12.864,-26.178,25.328],
[10.897,-37.518,28.313],[4.418,-40.167,25.322],[11.362,-26.357,22.100],
[9.846,-28.046,20.481],[12.254,-28.731,29.650],[7.635,-29.779,21.126],
[7.917,-32.576,15.905],[14.439,-32.676,27.239],[11.357,-30.417,22.309],
[9.021,-33.221,28.926],[1.982,-30.443,22.990],[4.215,-30.738,24.479],
[3.083,-32.983,25.331],[7.952,-26.392,28.049],[3.409,-26.441,29.624],
[9.013,-31.957,18.373],[1.969,-30.473,29.196],[6.897,-24.557,22.675],
[1.782,-24.764,31.187],[13.252,-38.097,17.512],[1.572,-32.053,19.911],
[12.961,-25.769,16.367],[-2.811,-29.373,27.989],[-1.897,-28.224,25.717],
[15.393,-26.218,31.802],[6.222,-26.569,30.210],[6.574,-22.790,19.229],
[0.431,-23.945,22.683],[8.696,-23.706,17.446],[8.007,-25.281,19.721],
[12.291,-33.382,32.643],[-1.495,-28.447,30.367],[1.680,-39.519,25.929],
[7.006,-40.667,28.984],[9.467,-35.859,30.300],[11.236,-31.420,16.438],
[4.270,-35.603,17.697],[2.069,-35.080,27.078],[3.046,-34.299,20.336],
[8.407,-32.997,20.792],[7.220,-33.637,31.406],[8.604,-34.227,23.539],
[-1.826,-33.290,30.850],[1.196,-36.839,16.816],[-3.592,-37.064,22.391],
[-1.250,-37.617,18.056],[6.506,-33.359,27.839],[4.295,-38.199,23.397],
[0.164,-34.785,20.602],[3.863,-36.811,20.356],[9.045,-37.795,26.469],
[2.220,-33.823,16.414],[11.901,-21.847,31.562],[0.447,-28.995,32.888],
[5.719,-34.649,24.496],[2.573,-38.970,19.119],[4.719,-35.435,27.193],
[4.224,-30.324,16.196],[12.077,-30.573,27.147],[8.550,-38.430,16.720],
[5.180,-29.736,28.615],[9.588,-35.186,16.698],[-0.500,-30.593,23.962],
[7.179,-28.822,18.136],[-5.222,-32.996,30.457],[3.500,-33.423,32.044],
[9.021,-27.043,30.721],[14.573,-34.850,18.149],[14.923,-28.672,30.656],
[13.031,-27.104,19.823],[9.374,-30.553,24.489],[2.079,-27.447,22.227],
[-1.533,-31.260,16.623],[4.065,-29.548,21.828],[3.635,-24.296,26.063],
[2.531,-24.028,21.284],[7.867,-29.907,27.693],[9.769,-27.834,16.889],
[9.791,-24.372,30.481],[13.966,-22.478,17.130],[-0.606,-35.092,28.631],
[10.782,-29.539,18.448],[-0.795,-25.448,26.213],[4.449,-27.893,14.656],
[3.733,-28.278,17.954],[-0.569,-30.141,19.399],[6.963,-36.361,27.533],
[7.675,-38.620,19.556],[5.733,-26.310,18.408],[6.663,-35.020,19.920],
[5.387,-34.863,29.786],[7.497,-37.199,22.039],[3.899,-25.918,16.301],
[2.425,-41.201,17.720],[12.069,-35.648,29.911],[11.344,-32.692,25.366],
[1.319,-26.379,17.449],[-0.475,-34.865,16.700],[11.251,-23.723,25.314],
[-0.010,-40.818,24.242],[-1.447,-25.918,21.280],[-1.852,-25.327,33.966],
[2.711,-22.117,33.951],[7.118,-26.594,15.938],[9.653,-27.167,33.735],
[10.099,-25.563,15.602],[-1.906,-30.404,31.979],[0.787,-28.267,13.579],
[3.909,-29.252,31.853],[7.368,-29.662,15.586],[9.879,-30.326,29.460],
[0.771,-31.429,32.624],[1.869,-30.906,17.271],[3.052,-33.584,29.053],
[5.306,-33.882,15.801],[8.921,-30.767,34.310],[13.565,-29.349,13.352],
[1.653,-35.672,29.759],[3.051,-37.401,28.001],[5.265,-31.963,30.336],
[9.775,-34.484,32.745],[9.930,-31.818,31.989],[0.980,-34.549,32.205],
[3.576,-40.139,28.069],[5.378,-35.511,34.024],[6.788,-36.253,31.689],
[9.144,-37.485,33.022],[-2.503,-36.897,33.540],[5.670,-39.508,31.063],
[3.772,-38.896,32.883],[4.506,-21.840,16.675],[-0.244,-28.780,16.090],
[-2.848,-21.957,19.790],[-2.420,-23.809,23.136],[13.873,-23.682,24.532],
[14.083,-23.993,28.273],[12.587,-27.925,32.806],[13.449,-29.856,18.276],
[13.769,-29.104,22.946],[13.700,-29.070,25.645],[11.653,-25.963,29.008],
[13.058,-32.707,19.405],[14.679,-32.904,16.293],[12.050,-33.120,22.673],
[-2.888,-32.265,24.875],[12.859,-30.562,31.657],[14.804,-32.797,32.398],
[11.459,-37.017,15.656],[10.655,-34.587,19.301],[12.812,-35.258,26.706],
[12.344,-32.515,29.075],[-3.871,-34.978,32.036],[15.080,-38.976,34.314],
[12.894,-39.219,19.910],[-0.576,-37.859,23.874],[-4.078,-34.768,25.950],
[13.403,-37.762,31.517],[14.072,-38.259,15.192],[0.120,-37.366,20.659],
[10.180,-23.439,22.606],[-2.358,-38.647,28.487],[0.165,-26.408,32.269],
[1.945,-26.037,27.304],[6.882,-23.810,31.378],[9.598,-21.547,30.188],
[12.818,-39.963,29.751],[-0.400,-23.107,28.472],[2.327,-23.491,28.755],
[6.730,-21.703,26.385],[8.327,-23.740,28.388],[11.167,-22.894,28.174],
[-2.268,-22.361,26.068],[1.718,-25.655,24.306],[4.057,-24.843,23.285],
[7.680,-36.569,24.590],[11.889,-40.106,25.334],[13.268,-40.746,22.262],
[1.699,-40.175,22.077],[-0.574,-40.934,17.518],[4.131,-41.118,22.607],
[9.351,-39.668,22.742],[10.376,-39.246,18.396],[-1.244,-25.183,18.335],
[-0.664,-38.918,15.442],[5.243,-38.616,18.571],[6.063,-24.003,16.776],
[13.346,-23.578,19.707],[-0.407,-39.803,31.388],[0.914,-38.525,33.653],
[-0.583,-24.383,15.664],[8.819,-40.398,33.774],[12.739,-24.954,31.422],
[7.408,-11.868,20.890],[5.231,-13.183,23.506],[-0.782,-12.942,27.413],
[9.770,-9.134,23.374],[0.645,-9.939,25.068],[-4.389,-15.507,22.148],
[-3.256,-13.216,22.611],[6.971,-5.101,24.449],[5.394,-10.154,25.868],
[13.798,-19.377,27.094],[9.511,-15.481,26.550],[2.636,-17.722,22.135],
[8.239,-18.622,22.466],[-2.020,-15.451,23.391],[2.455,-15.106,23.948],
[15.573,-17.234,27.783],[0.680,-18.579,25.936],[7.375,-9.218,21.671],
[3.016,-11.003,27.321],[11.091,-18.529,23.089],[0.516,-16.926,23.787],
[0.711,-10.484,28.145],[-1.286,-9.106,18.797],[10.355,-8.661,26.251],
[11.189,-19.219,28.387],[4.607,-21.026,24.751],[9.295,-5.753,22.640],
[12.237,-8.699,22.823],[12.180,-10.024,27.960],[7.371,-12.756,18.008],
[8.140,-14.325,15.863],[14.536,-14.445,27.634],[10.235,-13.141,20.149],
[8.231,-14.018,28.677],[1.674,-12.091,22.251],[2.843,-8.791,25.425],
[1.797,-12.554,25.452],[7.603,-8.070,25.850],[2.444,-9.836,30.027],
[10.120,-13.707,17.589],[0.548,-14.420,29.266],[6.936,-6.475,21.788],
[2.450,-6.778,30.292],[13.707,-20.188,19.261],[1.146,-12.964,18.305],
[14.038,-7.944,15.449],[-2.928,-10.392,30.221],[-3.715,-6.576,25.929],
[15.719,-8.229,31.747],[4.643,-8.615,30.459],[7.009,-4.213,20.463],
[-0.247,-7.440,21.178],[9.627,-4.167,16.805],[9.544,-7.141,18.328],
[12.787,-16.441,29.082],[-0.587,-9.298,30.355],[1.423,-21.243,26.817],
[7.074,-21.114,29.228],[7.762,-17.089,29.135],[10.791,-13.569,14.965],
[6.425,-18.898,18.309],[-0.092,-14.669,25.547],[1.841,-14.954,21.303],
[8.833,-15.954,18.966],[7.824,-14.543,31.289],[7.575,-15.808,21.278],
[-0.813,-16.479,27.551],[1.956,-18.704,15.485],[-3.547,-19.674,22.309],
[-2.943,-17.734,16.720],[6.657,-15.082,26.655],[3.662,-20.272,22.421],
[0.271,-17.060,20.912],[2.827,-17.190,19.384],[8.264,-18.908,27.273],
[2.296,-16.010,16.401],[12.193,-4.176,30.607],[0.710,-11.079,31.691],
[5.271,-16.026,24.314],[0.806,-19.623,22.775],[5.092,-16.866,27.994],
[3.789,-11.632,17.868],[10.994,-12.711,27.118],[9.027,-19.084,17.282],
[7.153,-9.796,29.959],[11.597,-16.624,18.398],[-1.070,-11.737,23.212],
[8.157,-9.582,16.448],[-2.102,-14.478,30.000],[5.546,-12.854,30.723],
[8.610,-7.233,30.296],[15.609,-15.079,18.076],[14.228,-10.040,29.970],
[12.073,-9.110,17.858],[9.954,-11.623,24.734],[2.159,-8.289,21.061],
[-0.966,-12.789,15.919],[3.891,-10.725,21.286],[4.147,-5.278,27.229],
[3.914,-5.469,20.246],[7.551,-11.494,27.685],[10.523,-10.450,16.042],
[11.132,-6.563,28.984],[12.642,-4.217,16.833],[-1.280,-19.072,31.061],
[9.855,-10.001,20.548],[-1.353,-7.419,27.367],[4.626,-8.186,14.473],
[2.734,-8.853,16.285],[-1.905,-13.354,18.299],[4.948,-19.311,31.137],
[10.167,-18.258,19.892],[6.595,-8.273,18.184],[5.181,-17.751,21.059],
[5.375,-15.734,30.789],[6.791,-20.411,20.611],[3.249,-6.478,17.588],
[2.885,-19.550,17.743],[11.967,-18.212,30.945],[11.323,-14.187,24.512],
[0.185,-10.151,16.412],[-0.462,-16.346,16.327],[10.725,-5.765,26.467],
[0.615,-21.966,24.492],[-2.641,-8.974,21.517],[-2.242,-5.484,32.556],
[4.011,-6.584,32.497],[6.788,-6.737,15.593],[8.324,-8.000,35.154],
[9.974,-7.389,15.196],[-1.785,-12.077,32.238],[1.912,-11.383,14.664],
[3.906,-11.560,32.935],[5.486,-10.426,16.312],[9.842,-11.310,29.317],
[-0.224,-14.385,32.048],[3.614,-14.265,18.417],[2.884,-14.572,30.794],
[4.545,-15.625,14.823],[8.276,-11.958,34.352],[14.680,-12.879,16.068],
[0.653,-16.941,30.808],[3.522,-19.063,27.169],[3.721,-14.496,26.530],
[9.536,-14.719,33.708],[10.785,-13.930,31.525],[1.588,-15.687,33.589],
[3.942,-21.968,27.274],[5.774,-18.437,33.728],[8.163,-17.331,31.944],
[9.806,-19.617,32.025],[0.459,-18.543,33.007],[2.781,-21.997,31.182],
[5.569,-21.790,32.858],[3.835,-3.644,17.409],[-2.320,-8.502,16.127],
[-5.032,-5.197,20.425],[-2.658,-4.751,22.179],[14.990,-4.694,25.960],
[14.485,-5.707,30.849],[12.448,-10.056,31.861],[14.252,-10.579,18.572],
[13.165,-12.400,20.439],[12.716,-9.544,25.393],[13.594,-6.925,27.266],
[12.760,-13.227,17.807],[13.888,-16.270,16.580],[11.889,-14.803,21.762],
[-2.353,-13.163,25.093],[12.552,-12.936,29.327],[15.193,-14.407,32.421],
[11.764,-18.389,15.617],[12.629,-17.674,20.789],[13.066,-16.642,26.359],
[10.535,-15.148,29.144],[-3.536,-17.844,32.044],[15.538,-21.406,33.644],
[11.355,-20.524,21.192],[-1.643,-18.503,23.425],[-4.332,-15.546,25.319],
[14.534,-19.427,30.383],[14.473,-19.062,15.895],[-2.972,-18.378,19.894],
[11.218,-6.711,20.974],[-2.466,-19.879,28.714],[0.098,-7.591,32.346],
[1.508,-7.189,27.588],[7.433,-4.766,31.282],[9.770,-4.194,32.515],
[13.833,-21.509,28.975],[-1.900,-4.365,29.745],[2.123,-3.564,29.339],
[7.571,-3.661,26.505],[8.322,-5.415,28.286],[11.853,-3.768,27.872],
[-1.613,-6.135,24.643],[1.157,-7.532,23.713],[4.656,-6.122,24.624],
[7.028,-20.883,23.543],[12.994,-21.250,24.768],[13.865,-21.853,22.334],
[-0.912,-21.589,22.263],[-1.290,-22.716,17.706],[4.956,-22.828,21.784],
[10.385,-20.938,23.806],[9.876,-22.485,19.554],[-2.513,-6.453,20.056],
[-0.202,-19.211,17.285],[4.245,-21.124,19.544],[7.348,-5.139,17.903],
[13.149,-5.199,19.484],[-2.079,-21.808,30.757],[0.451,-21.512,32.075],
[0.360,-5.740,18.384],[8.704,-21.747,33.129],[11.911,-6.552,32.016],
[8.393,7.048,20.280],[7.408,6.559,25.904],[1.571,4.535,26.780],
[10.686,8.659,23.082],[-0.260,8.556,22.855],[-2.441,3.357,20.745],
[-2.299,5.824,22.457],[7.677,13.861,23.852],[6.227,9.284,24.056],
[14.266,0.060,26.820],[10.181,1.442,26.523],[2.060,0.793,23.485],
[9.403,0.544,22.286],[-2.772,3.092,23.439],[4.356,6.051,22.936],
[13.999,1.701,30.493],[-0.576,-0.133,27.161],[6.979,9.944,21.112],
[2.047,7.073,27.611],[10.982,-1.171,23.710],[0.430,1.426,25.302],
[3.161,10.161,26.677],[-2.628,7.910,20.334],[11.095,9.098,25.703],
[11.055,-1.702,26.278],[3.918,-1.949,24.686],[9.787,11.208,23.794],
[9.858,9.957,20.610],[11.680,8.090,30.948],[5.256,5.759,20.332],
[8.027,4.394,16.504],[14.450,4.941,27.539],[10.701,6.286,21.368],
[10.252,4.834,28.294],[1.305,6.053,22.262],[3.456,8.371,24.314],
[2.335,5.835,24.654],[8.480,10.629,26.679],[2.972,8.828,29.169],
[10.295,4.903,18.971],[0.579,4.769,29.554],[6.945,12.743,20.986],
[3.243,9.787,31.613],[13.673,-2.054,19.054],[1.311,5.211,19.601],
[14.833,10.156,16.078],[-2.310,7.437,27.408],[-3.432,10.948,25.033],
[15.396,11.878,29.867],[5.569,11.642,31.051],[7.810,14.630,18.920],
[0.501,12.347,22.295],[10.581,13.445,16.905],[9.682,12.727,19.472],
[11.810,3.176,30.327],[-1.631,9.344,29.873],[1.512,-2.777,26.416],
[5.713,-2.939,28.274],[9.625,2.972,32.506],[10.786,5.556,16.345],
[6.683,1.783,17.616],[2.993,2.323,26.102],[2.135,3.498,23.433],
[6.660,3.471,20.311],[6.703,3.437,32.790],[8.384,2.987,23.323],
[-2.033,1.881,29.360],[3.086,-0.612,15.273],[-2.353,0.214,22.065],
[-2.518,1.151,17.518],[7.425,3.274,27.612],[3.027,-0.856,21.549],
[-0.313,2.130,22.450],[2.102,1.788,21.069],[8.439,-0.756,26.759],
[3.314,3.436,18.862],[11.956,14.246,30.738],[0.677,7.560,31.602],
[5.571,3.494,25.458],[0.258,-0.592,21.076],[4.857,2.049,27.985],
[6.080,8.010,18.940],[12.032,6.281,26.381],[7.857,-0.526,17.105],
[6.612,9.343,29.786],[9.426,2.428,18.788],[-2.491,5.181,25.289],
[8.370,9.087,18.600],[-2.152,5.189,29.177],[4.159,2.102,32.033],
[10.068,8.825,28.293],[13.818,2.249,17.772],[14.451,9.150,31.532],
[12.530,10.652,18.952],[8.078,7.150,23.330],[2.550,10.377,21.966],
[-1.930,4.029,16.069],[4.100,8.186,20.920],[5.004,12.649,27.291],
[2.820,12.926,20.333],[6.089,8.814,26.738],[10.052,8.267,15.896],
[11.064,14.536,27.686],[14.765,13.288,16.565],[-1.226,1.623,32.408],
[11.013,8.106,18.372],[-0.046,12.319,27.638],[4.073,11.008,15.303],
[2.628,8.817,17.687],[-0.617,7.159,17.656],[7.149,-0.420,29.343],
[8.986,-0.843,19.549],[5.147,10.244,17.769],[5.576,0.905,21.175],
[6.212,2.818,30.066],[7.238,-1.588,21.446],[2.830,12.106,17.699],
[1.411,-2.686,18.528],[12.383,-0.484,28.684],[10.763,4.308,24.536],
[0.163,9.971,18.680],[0.031,1.990,16.320],[11.929,12.144,26.387],
[1.922,-4.729,23.272],[-1.714,10.585,20.603],[-0.269,12.274,31.985],
[2.388,12.458,31.843],[7.828,11.315,13.821],[9.288,9.156,32.616],
[12.153,11.262,16.303],[-2.946,4.936,32.890],[2.069,6.508,15.702],
[4.804,6.115,32.583],[6.412,8.693,15.767],[7.468,6.862,29.713],
[-0.540,4.376,32.118],[3.712,6.182,17.762],[2.994,3.993,29.250],
[5.007,5.513,15.483],[8.612,5.464,32.470],[13.430,6.210,14.743],
[2.512,0.082,30.407],[3.046,-0.581,27.524],[4.320,5.768,26.696],
[8.321,1.746,34.804],[10.695,5.533,30.881],[1.538,2.856,32.801],
[5.270,-2.662,30.891],[4.429,-1.257,32.846],[6.274,0.488,31.716],
[8.957,-0.012,32.801],[0.964,-0.080,32.800],[5.155,-5.620,29.967],
[3.687,-4.081,34.213],[4.261,15.719,16.976],[-1.107,8.739,15.646],
[-4.225,14.477,22.263],[-2.106,12.365,22.814],[14.373,14.687,25.267],
[14.480,13.456,27.707],[11.970,9.524,33.337],[12.684,8.385,20.896],
[14.587,4.074,20.773],[13.644,9.920,26.688],[12.456,10.822,28.731],
[15.659,6.340,17.556],[15.936,3.114,16.375],[12.001,3.724,22.093],
[-4.548,7.683,24.462],[13.281,5.724,30.366],[14.335,4.513,32.460],
[12.234,0.948,15.617],[11.906,0.773,19.194],[13.008,2.171,25.290],
[12.385,2.730,27.788],[-4.581,-0.041,33.103],[14.053,-3.352,32.712],
[12.616,-1.045,21.228],[-0.816,-0.673,24.191],[-4.845,3.389,25.341],
[13.805,-0.501,32.401],[15.418,-0.484,17.465],[-2.206,-1.624,20.089],
[11.798,13.535,20.675],[-2.307,-1.151,29.466],[-0.060,10.525,34.126],
[2.810,11.567,29.404],[6.620,14.566,31.499],[9.285,14.631,30.153],
[14.884,-3.259,29.364],[-0.447,15.906,28.350],[1.771,14.097,28.947],
[6.591,14.533,26.204],[8.856,13.031,27.752],[9.212,15.543,26.060],
[-0.832,13.693,25.357],[1.011,10.609,24.584],[4.472,11.985,24.585],
[6.705,-0.636,24.162],[10.974,-3.852,23.467],[13.663,-2.706,22.884],
[-0.185,-3.593,21.685],[-2.185,-3.800,18.211],[4.441,-3.603,22.331],
[8.655,-2.834,23.923],[10.931,-2.437,18.334],[-2.343,11.484,18.144],
[-0.151,-2.835,16.607],[3.834,0.864,18.454],[7.088,12.401,17.111],
[14.425,12.660,19.943],[0.175,-2.109,30.998],[-0.063,-3.922,33.023],
[-0.196,13.513,14.199],[7.721,-2.875,33.832],[16.071,13.656,32.286],
[7.728,24.124,21.520],[7.370,24.927,25.640],[0.411,23.621,27.086],
[11.170,27.591,24.080],[0.681,28.134,25.136],[-2.077,22.023,20.516],
[-1.594,23.908,22.432],[7.786,32.314,23.392],[6.396,27.090,26.957],
[15.077,19.122,26.682],[9.865,20.691,25.509],[2.577,18.344,23.486],
[9.666,19.802,21.441],[-1.744,21.382,23.684],[4.634,23.702,22.774],
[16.087,20.591,28.852],[1.762,17.767,26.041],[7.863,27.573,22.581],
[2.343,25.327,28.450],[12.224,18.617,22.497],[0.271,20.410,24.864],
[0.385,27.150,27.612],[-1.943,26.417,20.256],[11.701,27.396,26.891],
[11.128,18.103,27.399],[3.093,14.791,22.911],[9.800,31.092,22.443],
[9.306,28.911,20.458],[11.284,27.794,29.405],[5.443,23.599,20.324],
[7.110,22.236,16.518],[14.678,23.298,28.215],[10.169,25.036,20.445],
[9.027,23.103,27.891],[1.712,24.960,22.869],[3.467,29.376,24.500],
[2.965,23.528,25.086],[7.685,29.880,28.433],[1.995,28.164,30.015],
[8.920,24.127,18.000],[1.155,22.056,29.647],[6.040,31.141,21.042],
[3.015,30.225,31.866],[13.965,15.124,18.686],[0.048,23.781,20.058],
[12.879,29.861,16.146],[-1.874,25.708,27.430],[-2.240,27.734,25.356],
[15.789,29.954,29.192],[6.212,28.404,30.314],[7.727,33.002,19.645],
[-0.226,32.764,23.135],[9.003,31.603,17.111],[10.406,31.426,19.439],
[12.421,21.099,31.706],[-1.293,27.355,29.930],[1.671,15.170,26.445],
[7.675,16.257,28.185],[10.545,20.577,28.500],[11.257,24.293,16.413],
[6.955,18.774,17.262],[3.472,22.237,27.391],[2.526,21.891,22.708],
[7.585,22.114,19.426],[8.537,20.498,30.543],[9.131,22.336,23.447],
[0.938,18.577,28.684],[2.784,18.209,16.088],[-3.567,17.701,22.599],
[-2.128,19.421,17.664],[8.006,20.603,27.302],[4.209,17.491,21.631],
[0.155,20.019,21.524],[2.719,19.834,18.657],[8.750,18.248,25.306],
[1.845,22.668,17.796],[11.461,34.470,33.158],[0.288,27.862,32.781],
[6.877,21.264,24.561],[1.201,17.617,21.283],[5.855,19.228,25.758],
[5.060,26.589,19.236],[13.263,24.680,25.903],[9.824,18.076,16.892],
[5.507,25.668,29.382],[9.130,20.430,18.101],[-0.861,26.215,23.608],
[7.336,28.230,17.820],[-3.083,24.810,29.777],[6.026,23.740,32.285],
[8.496,27.157,29.130],[14.359,21.302,18.226],[14.471,27.354,28.957],
[11.819,29.012,18.734],[9.371,25.139,23.851],[2.499,29.530,20.968],
[-1.755,24.195,17.809],[3.038,26.794,21.196],[4.906,29.623,27.902],
[3.254,31.754,22.257],[9.821,25.640,27.068],[11.114,27.252,15.707],
[10.620,30.424,30.293],[14.361,32.350,16.252],[-3.958,18.036,29.114],
[12.080,26.542,19.637],[-0.357,29.738,27.926],[3.486,28.489,15.040],
[2.596,27.953,17.603],[0.710,26.213,18.857],[6.318,18.701,28.618],
[8.765,17.189,19.154],[4.922,29.337,17.761],[5.325,20.045,19.355],
[5.194,21.097,29.976],[7.920,16.775,21.878],[2.268,30.512,18.505],
[3.015,16.814,19.351],[13.210,18.239,31.271],[11.904,22.407,24.909],
[0.097,28.983,18.114],[0.356,19.666,16.577],[11.652,32.188,26.603],
[-1.161,16.747,25.807],[-1.650,30.989,21.220],[-2.397,34.528,32.056],
[4.253,31.956,33.647],[6.047,28.883,13.863],[8.269,28.403,32.329],
[10.026,30.891,14.718],[-3.308,25.008,33.847],[1.423,25.254,16.166],
[3.821,25.235,32.275],[7.238,26.997,15.312],[10.163,25.207,29.762],
[-0.612,21.910,32.122],[4.754,24.337,17.764],[3.652,23.553,30.125],
[3.686,23.314,15.477],[8.386,25.594,32.658],[13.517,24.164,14.874],
[3.563,18.989,30.506],[3.773,19.229,27.942],[6.191,22.688,26.814],
[10.052,20.973,33.278],[11.866,23.447,27.890],[3.003,21.273,31.710],
[4.188,15.648,27.526],[6.758,21.023,32.430],[6.323,17.401,32.207],
[10.007,17.293,31.898],[1.126,18.941,33.124],[4.472,13.876,29.628],
[3.683,16.995,32.849],[5.210,32.975,18.467],[-1.466,26.987,15.968],
[-2.246,34.842,20.561],[-3.015,32.904,22.571],[15.552,33.379,24.946],
[12.924,31.973,29.782],[12.934,27.439,31.898],[14.719,26.563,19.248],
[13.759,24.663,21.212],[13.996,28.196,25.051],[12.764,29.938,27.928],
[13.685,24.210,18.286],[14.520,20.873,15.677],[11.802,22.107,21.223],
[-2.390,23.981,25.138],[12.705,24.480,30.630],[13.979,23.302,32.820],
[11.916,17.680,15.478],[13.546,19.489,20.267],[12.310,19.305,25.001],
[13.203,20.685,28.840],[-3.559,20.828,32.286],[16.124,16.271,33.410],
[13.560,16.868,20.550],[-1.005,18.357,23.302],[-4.139,21.010,25.233],
[13.419,14.994,33.262],[14.483,18.192,16.146],[-2.875,16.854,18.988],
[12.645,32.903,21.249],[-3.106,15.852,27.858],[-1.335,30.260,32.038],
[2.397,29.426,27.309],[6.212,31.611,30.839],[9.090,32.394,30.654],
[13.036,15.691,28.947],[-0.485,34.411,28.726],[0.829,32.359,26.580],
[6.247,32.519,26.729],[8.760,32.149,27.487],[10.894,33.917,28.284],
[-1.377,30.512,25.416],[0.182,27.786,21.384],[2.113,31.409,24.649],
[5.734,16.087,23.873],[11.889,13.183,23.626],[13.415,16.121,23.273],
[-1.564,16.308,21.478],[-1.966,13.917,20.499],[0.826,15.000,21.334],
[10.252,15.536,22.342],[11.385,15.531,18.458],[-1.855,31.227,18.580],
[-0.506,17.182,17.094],[5.323,15.565,19.637],[7.276,33.904,17.031],
[13.586,31.908,18.975],[1.897,15.912,30.859],[0.370,16.201,33.544],
[-0.327,32.396,14.542],[9.103,14.772,33.143],[13.726,29.818,31.049],
[9.604,43.827,22.154],[6.518,43.944,26.401],[-0.342,41.829,27.424],
[13.013,46.921,22.524],[0.796,45.718,25.154],[-1.576,41.152,20.361],
[-2.621,42.441,22.417],[6.759,50.208,25.356],[5.174,46.323,26.556],
[13.484,36.890,26.558],[10.233,39.148,27.123],[1.813,37.188,22.905],
[10.245,38.400,20.085],[-1.463,39.946,23.222],[4.490,40.712,23.456],
[15.326,38.687,27.988],[0.729,37.917,26.549],[7.827,46.778,21.805],
[2.760,42.709,28.167],[11.597,36.842,24.260],[1.207,40.021,24.554],
[2.761,45.375,27.532],[-2.301,45.544,19.258],[11.281,46.969,25.696],
[11.253,36.414,28.596],[4.068,34.259,25.468],[9.452,50.046,23.783],
[10.388,47.949,22.305],[10.978,46.081,29.956],[8.744,45.267,19.489],
[7.553,41.150,16.346],[15.066,41.163,27.115],[10.984,43.267,19.525],
[9.407,41.421,29.033],[2.177,42.878,23.045],[3.616,46.857,24.323],
[3.794,41.821,25.669],[7.494,48.310,27.543],[3.032,47.112,30.259],
[9.337,42.943,17.465],[0.892,41.533,30.115],[7.889,49.322,21.317],
[1.779,49.768,31.298],[13.255,34.386,17.525],[1.025,41.843,19.858],
[14.281,48.398,15.393],[-2.923,44.021,28.255],[-3.729,48.211,26.385],
[16.301,47.959,31.496],[5.467,50.814,28.653],[7.372,51.838,21.274],
[0.967,48.167,20.593],[9.731,50.882,17.738],[10.526,50.671,20.616],
[13.390,39.866,31.550],[-2.097,46.694,28.780],[1.617,35.042,26.793],
[8.776,35.095,29.576],[11.053,38.706,30.477],[10.932,43.426,15.336],
[6.396,35.847,15.276],[3.217,38.655,25.592],[3.032,39.745,20.959],
[6.533,40.285,18.704],[6.194,40.502,32.710],[9.800,38.343,22.862],
[-0.716,39.494,29.169],[2.051,37.699,15.264],[-2.654,36.958,22.315],
[-0.509,37.599,17.739],[8.269,41.785,26.365],[4.415,37.218,23.713],
[-0.143,38.216,21.496],[4.393,37.637,17.292],[8.487,36.772,26.611],
[2.099,40.043,17.647],[11.308,50.832,32.783],[0.382,43.475,32.054],
[7.137,40.493,23.656],[1.880,37.382,19.957],[5.455,38.911,26.748],
[5.183,45.010,18.750],[13.318,43.564,26.768],[8.426,36.586,16.995],
[6.616,45.416,29.016],[10.493,40.674,18.261],[-0.368,43.391,24.401],
[7.297,44.835,17.002],[-2.681,42.029,30.168],[4.092,41.287,30.259],
[6.719,47.733,30.259],[15.282,39.799,18.200],[13.911,46.568,31.488],
[12.188,49.943,18.628],[10.579,43.393,25.405],[4.161,46.829,20.513],
[-1.755,43.600,16.696],[4.470,44.384,21.724],[3.150,48.507,27.845],
[3.327,49.036,21.833],[8.949,45.656,26.221],[9.601,46.684,16.663],
[9.900,49.155,30.010],[14.772,50.711,16.884],[-1.547,37.143,28.368],
[12.015,45.934,19.568],[-1.866,49.219,27.779],[3.990,46.906,14.794],
[2.983,46.742,17.447],[0.081,44.136,18.715],[6.712,36.666,29.350],
[8.828,35.868,19.540],[6.396,49.432,18.031],[7.391,38.319,20.259],
[6.791,39.433,29.982],[7.843,34.942,21.767],[3.276,49.507,19.241],
[1.776,35.192,15.924],[13.950,36.889,29.587],[10.767,40.681,24.628],
[0.939,47.315,15.583],[0.020,39.623,15.808],[11.815,50.659,26.927],
[-0.885,34.315,25.983],[-0.516,46.228,21.496],[-0.386,50.934,32.709],
[3.597,50.086,33.553],[6.897,47.676,15.321],[8.429,48.313,33.082],
[11.331,48.660,16.154],[-1.995,44.281,31.782],[1.795,44.028,16.139],
[3.893,44.827,31.908],[7.525,45.339,14.219],[8.976,44.476,29.916],
[-1.508,41.811,33.094],[3.507,42.638,17.922],[1.823,40.202,27.238],
[4.054,40.503,15.733],[8.257,43.360,32.380],[12.608,45.304,16.163],
[4.516,37.913,30.265],[3.777,36.568,27.642],[6.256,41.366,28.195],
[11.092,39.343,33.334],[11.575,41.993,31.811],[1.112,39.977,32.570],
[3.953,32.728,28.283],[4.604,37.163,32.894],[8.500,39.166,32.489],
[9.779,36.490,32.037],[2.049,37.143,31.807],[4.699,34.622,31.639],
[5.816,34.454,34.029],[4.409,51.972,16.862],[-1.559,48.433,15.658],
[-3.553,52.241,21.026],[-1.705,48.768,22.536],[14.247,51.658,24.907],
[13.529,49.532,29.390],[11.893,46.461,33.239],[14.114,43.896,19.669],
[13.225,44.273,23.388],[14.088,46.419,25.699],[12.248,47.861,28.009],
[14.998,42.744,17.143],[13.431,40.638,16.378],[11.041,40.830,21.436],
[-3.294,42.511,26.138],[13.391,43.686,30.471],[15.308,41.807,31.738],
[10.930,37.211,15.573],[12.137,38.554,17.932],[12.867,39.201,25.440],
[12.812,40.009,28.307],[-1.771,37.127,31.335],[16.749,36.394,34.043],
[12.278,36.178,20.087],[-2.226,37.487,25.772],[-3.707,40.468,24.634],
[14.730,36.482,32.103],[13.555,37.496,15.833],[-3.037,36.001,18.143],
[12.969,52.797,21.883],[-3.378,35.055,26.848],[-0.309,47.594,32.613],
[0.266,46.941,29.991],[7.026,50.442,31.003],[9.279,52.311,30.389],
[12.848,34.569,30.319],[-1.369,51.133,29.865],[0.788,49.721,28.094],
[7.049,52.659,26.879],[8.745,50.675,28.009],[11.776,53.035,28.811],
[-0.617,49.953,25.057],[0.855,47.780,23.380],[3.274,49.814,25.279],
[6.482,34.762,24.408],[12.842,32.007,24.045],[13.761,35.132,22.789],
[-0.184,35.498,22.351],[-0.544,34.745,17.260],[5.219,33.540,22.516],
[9.142,34.945,24.317],[10.207,33.830,21.059],[-1.441,49.124,19.426],
[-0.808,36.414,15.001],[4.590,36.347,21.178],[5.996,50.418,15.506],
[14.502,50.817,19.923],[0.161,35.164,31.453],[0.317,34.065,33.933],
[-0.035,50.801,15.531],[8.347,33.324,33.781],[13.544,49.252,32.526],
[13.021,56.072,25.442],[9.013,59.298,27.195],[2.017,56.620,22.961],
[10.799,58.780,21.550],[-2.103,57.211,22.510],[15.070,57.185,28.171],
[-0.041,55.863,28.591],[8.919,57.513,24.968],[-0.101,57.114,24.444],
[12.731,55.340,28.041],[5.701,53.443,23.580],[13.021,52.709,18.159],
[11.426,58.300,32.900],[1.931,52.659,25.364],[6.814,53.065,29.710],
[8.561,58.891,31.276],[6.300,55.608,17.575],[3.520,57.128,25.006],
[3.135,57.899,20.860],[7.382,57.311,22.294],[-0.673,58.659,27.783],
[3.954,56.395,16.185],[-4.331,54.937,22.136],[-0.963,56.215,18.642],
[4.883,56.133,22.885],[-0.179,55.733,21.567],[5.333,56.094,20.279],
[9.853,56.165,28.997],[3.171,58.724,17.958],[6.444,58.913,26.809],
[1.293,53.444,22.008],[3.739,58.565,27.390],[10.480,55.564,18.277],
[11.070,58.152,18.410],[15.295,58.845,18.098],[-1.891,56.183,31.006],
[5.930,55.610,27.792],[8.586,53.520,19.616],[6.579,58.969,19.303],
[5.788,57.791,29.683],[7.978,54.503,22.390],[2.122,55.745,18.807],
[12.223,56.803,30.371],[-0.504,58.103,16.599],[-0.128,52.351,27.490],
[3.434,56.708,30.377],[2.657,55.986,27.608],[8.964,58.105,34.090],
[0.653,58.461,31.806],[4.649,53.386,28.242],[6.028,58.125,33.993],
[7.813,56.109,30.798],[10.261,55.769,32.732],[-1.467,55.014,33.661],
[4.292,51.937,31.257],[4.984,52.684,33.789],[13.070,58.771,15.911],
[13.514,55.744,16.743],[13.327,57.630,19.889],[12.863,58.711,25.371],
[11.512,-58.316,26.971],[-3.413,58.134,30.354],[14.205,53.274,32.337],
[13.955,54.773,20.232],[-1.757,54.510,23.618],[-2.805,58.034,25.375],
[14.150,54.982,30.432],[15.315,53.753,16.442],[-3.126,54.616,19.110],
[-2.688,54.469,26.221],[14.463,52.293,29.780],[6.987,55.390,25.092],
[14.527,54.041,23.918],[11.119,53.523,23.656],[-2.475,52.055,23.575],
[-0.211,52.932,19.673],[3.607,52.196,22.236],[8.499,52.768,24.462],
[11.209,53.453,19.989],[-0.727,54.364,16.724],[5.927,52.262,18.983],
[-0.627,53.807,31.130],[0.834,53.354,33.334],[7.987,52.576,32.744],
[7.880,-51.020,41.637],[6.563,-52.055,44.384],[-0.842,-50.794,45.853],
[9.184,-47.300,42.198],[0.857,-47.023,43.458],[-2.097,-52.108,40.109],
[-2.487,-50.963,42.686],[8.340,-41.648,43.205],[5.973,-46.476,42.526],
[13.260,-57.623,45.863],[9.393,-54.406,44.711],[3.643,-54.841,40.974],
[9.740,-56.388,40.022],[-0.222,-53.458,43.579],[5.252,-50.987,42.231],
[14.658,-53.681,46.724],[0.062,-56.245,46.408],[7.448,-46.311,40.333],
[2.466,-49.399,48.021],[11.865,-56.572,42.183],[1.853,-55.257,42.899],
[1.807,-46.966,47.702],[-1.380,-47.928,39.162],[10.565,-46.999,45.177],
[11.006,-56.449,45.229],[4.913,58.396,43.063],[10.805,-44.853,42.099],
[9.830,-45.376,39.648],[12.201,-48.323,47.270],[6.317,-49.889,39.676],
[8.029,-51.404,35.072],[13.876,-50.405,46.342],[10.097,-50.142,38.777],
[9.270,-51.972,48.236],[0.339,-49.213,41.277],[3.973,-46.227,44.209],
[1.208,-50.225,43.724],[8.485,-45.071,44.550],[4.731,-47.721,49.675],
[9.704,-52.270,36.989],[0.024,-51.164,48.529],[6.490,-43.506,41.413],
[2.065,-45.266,49.818],[11.817,-58.186,36.443],[0.857,-50.079,36.984],
[14.134,-44.229,35.396],[-3.005,-48.077,46.287],[-3.243,-44.171,44.491],
[16.093,-45.693,49.901],[6.434,-44.722,47.554],[5.996,-41.437,37.196],
[1.454,-41.275,40.810],[8.938,-42.401,36.145],[8.388,-42.562,38.916],
[11.156,-52.211,50.272],[-0.796,-46.642,48.859],[2.160,-57.851,45.809],
[6.777,-59.107,46.824],[10.930,-55.910,49.700],[10.034,-49.716,35.162],
[7.731,-56.294,35.460],[2.001,-53.773,45.756],[2.774,-52.469,42.891],
[7.257,-51.821,37.937],[7.203,-51.548,50.956],[9.420,-53.358,40.851],
[-0.641,-53.692,46.132],[4.707,-54.938,35.438],[-2.606,-54.402,41.971],
[-1.641,-55.194,36.518],[7.215,-53.751,46.267],[4.478,-57.223,42.849],
[0.630,-52.134,41.087],[3.413,-53.653,37.536],[8.818,-56.455,42.712],
[0.945,-52.858,36.272],[11.951,-41.875,51.610],[0.922,-47.427,50.918],
[6.587,-55.793,44.248],[2.551,-58.059,40.199],[4.561,-53.940,45.267],
[5.577,-49.101,37.208],[12.027,-51.219,44.318],[10.428,-55.040,37.824],
[6.642,-49.601,49.069],[11.863,-52.800,38.819],[-1.373,-48.740,43.979],
[7.505,-47.226,37.137],[-2.485,-52.203,47.799],[4.646,-51.619,50.138],
[9.284,-46.350,47.525],[14.499,-52.507,37.178],[13.480,-46.539,48.769],
[12.203,-45.522,37.874],[7.767,-47.938,44.380],[1.987,-45.695,40.996],
[-2.328,-50.998,35.216],[3.183,-48.485,38.650],[4.069,-43.340,49.140],
[1.800,-43.370,39.231],[8.057,-49.113,46.823],[10.126,-46.852,35.373],
[11.518,-44.642,49.354],[15.338,-41.747,35.522],[-2.550,-55.499,47.009],
[10.935,-47.835,37.696],[0.001,-44.172,46.792],[4.490,-45.019,32.619],
[3.449,-46.770,36.475],[-1.190,-48.372,36.174],[7.290,-56.546,46.859],
[9.319,-57.503,37.591],[5.465,-44.447,37.714],[7.462,-54.864,38.009],
[7.637,-53.999,49.014],[6.737,-56.222,40.334],[2.696,-44.077,36.744],
[2.960,-56.281,37.942],[14.373,-55.203,48.898],[10.271,-49.844,42.566],
[-0.017,-45.487,35.231],[-2.123,-53.728,34.005],[12.108,-42.668,44.730],
[0.300,-56.931,43.802],[-2.003,-45.215,39.534],[-1.523,-41.379,51.409],
[2.349,-42.925,51.201],[7.145,-45.368,35.103],[8.212,-46.889,49.956],
[10.789,-44.043,34.807],[-2.395,-47.998,50.654],[1.819,-46.249,32.333],
[2.993,-49.495,50.742],[7.138,-47.246,33.274],[9.676,-49.070,48.848],
[-1.644,-50.671,50.520],[3.060,-49.366,35.563],[2.899,-52.338,48.302],
[5.749,-52.484,34.457],[7.618,-49.174,52.912],[12.722,-48.634,34.078],
[4.018,-54.364,49.743],[2.833,-55.689,47.362],[5.265,-51.562,47.093],
[11.585,-54.328,51.929],[9.696,-50.487,51.521],[0.596,-52.774,52.818],
[4.199,-58.971,47.330],[6.116,-54.107,51.588],[8.072,-56.481,49.834],
[9.772,-56.164,52.547],[1.270,-55.515,50.300],[3.098,-59.253,50.090],
[5.897,-58.448,53.901],[4.449,-39.411,35.252],[-3.943,-47.155,35.947],
[-3.042,-40.848,39.681],[-1.584,-40.419,42.204],[15.246,-41.694,45.391],
[15.526,-43.394,47.828],[12.220,-48.289,55.479],[13.728,-48.118,37.348],
[12.404,-47.391,41.744],[13.394,-46.853,45.251],[11.406,-44.575,46.732],
[15.733,-49.411,35.863],[13.890,-53.520,34.464],[12.262,-51.242,40.987],
[-4.267,-50.455,44.592],[12.215,-49.809,49.572],[14.650,-51.092,50.106],
[12.179,-55.730,34.219],[12.604,-54.861,40.237],[11.676,-53.973,42.807],
[12.069,-52.091,47.712],[-1.532,-54.449,49.661],[15.338,-57.858,52.778],
[12.542,-57.639,39.569],[-2.217,-56.211,43.899],[-2.906,-52.833,44.848],
[13.338,-57.431,50.596],[13.345,-58.178,33.986],[-1.442,-57.394,39.257],
[11.953,-42.999,39.403],[-1.579,-58.212,46.169],[-0.535,-43.902,50.859],
[3.518,-44.788,47.009],[6.840,-42.176,50.001],[9.415,-42.559,49.072],
[12.346,-59.240,48.610],[-0.472,-41.627,48.202],[2.991,-41.083,47.248],
[6.245,-41.644,44.960],[9.430,-42.942,45.979],[12.028,-40.128,46.474],
[-0.916,-43.142,44.525],[-0.991,-43.912,41.896],[3.458,-43.548,41.553],
[7.000,-58.476,43.818],[11.225,-58.849,43.510],[14.241,-58.546,42.030],
[0.259,-59.045,41.077],[-1.901,-57.841,36.529],[5.242,-59.793,39.690],
[9.173,-58.845,41.241],[7.758,-59.529,38.769],[-0.617,-42.916,38.070],
[0.525,-56.943,36.780],[5.617,-57.391,38.142],[6.666,-42.938,33.716],
[16.061,-43.526,38.236],[-1.376,-57.876,49.036],[1.433,-58.291,51.965],
[-0.430,-41.819,35.101],[10.797,-58.358,51.031],[13.528,-43.933,51.028],
[8.621,-32.110,41.682],[6.226,-31.634,44.528],[0.106,-32.590,44.648],
[12.717,-27.570,43.612],[0.761,-28.143,43.679],[-1.785,-33.196,39.003],
[-3.430,-31.134,41.471],[7.577,-24.838,42.239],[6.169,-27.958,44.193],
[14.050,-38.401,45.593],[10.021,-35.562,45.281],[1.518,-36.316,41.903],
[9.494,-36.727,41.384],[-1.657,-35.028,41.406],[4.768,-31.977,40.937],
[13.953,-36.122,48.251],[0.113,-37.758,45.943],[8.286,-29.474,40.964],
[2.205,-31.265,46.298],[11.818,-37.589,43.850],[1.012,-35.336,44.759],
[1.795,-28.338,49.118],[-2.201,-29.301,37.004],[10.768,-26.331,45.902],
[11.804,-37.172,46.479],[2.803,-39.872,42.751],[9.548,-26.529,41.424],
[10.294,-26.401,38.609],[11.757,-30.763,47.557],[7.072,-31.833,39.086],
[7.315,-33.321,34.095],[13.780,-32.752,47.786],[10.971,-30.636,39.647],
[10.470,-32.542,48.881],[2.368,-30.256,41.100],[3.426,-27.200,43.717],
[3.197,-29.853,43.951],[7.467,-25.927,45.441],[4.172,-27.347,48.912],
[8.539,-31.809,36.860],[0.346,-32.179,48.271],[7.029,-25.554,39.265],
[1.831,-25.830,50.559],[13.416,-39.137,36.886],[1.429,-31.032,38.678],
[13.857,-25.828,34.574],[-1.862,-30.948,46.685],[-3.614,-27.758,43.855],
[15.365,-25.523,49.518],[6.544,-26.130,49.265],[6.441,-22.416,38.447],
[0.030,-23.757,40.941],[10.287,-22.932,35.684],[9.053,-23.637,38.092],
[11.629,-34.928,50.834],[-0.561,-28.843,47.910],[1.414,-39.949,45.399],
[8.782,-40.532,45.828],[9.767,-35.742,48.282],[11.808,-30.012,34.708],
[6.453,-37.731,36.798],[2.887,-33.375,44.410],[2.599,-33.935,41.138],
[7.982,-35.784,38.829],[7.085,-33.769,50.672],[6.747,-34.619,41.320],
[0.016,-35.426,48.140],[2.931,-36.243,33.012],[-4.099,-38.296,40.414],
[-2.064,-36.004,36.296],[7.019,-34.459,46.075],[4.069,-38.385,40.678],
[-1.516,-36.047,38.882],[2.376,-35.794,39.141],[9.401,-38.220,44.980],
[1.933,-35.674,36.352],[12.045,-21.406,49.422],[0.087,-30.108,50.340],
[5.434,-34.121,43.792],[1.566,-38.616,38.889],[4.775,-36.724,45.472],
[4.423,-29.557,36.674],[10.812,-32.902,45.834],[8.924,-37.546,35.838],
[6.968,-29.066,48.606],[9.406,-34.377,36.155],[-0.525,-30.462,42.372],
[7.004,-28.314,35.400],[-2.035,-31.936,49.614],[4.653,-32.323,50.707],
[9.445,-28.136,48.051],[14.817,-33.126,36.925],[14.657,-28.338,48.230],
[14.592,-26.818,38.993],[9.311,-30.952,44.363],[2.159,-26.329,40.642],
[-1.453,-30.782,34.565],[4.743,-29.540,39.732],[4.217,-26.043,46.039],
[3.060,-23.209,38.386],[8.781,-28.496,45.582],[10.113,-28.360,36.279],
[11.015,-24.614,48.978],[12.743,-24.604,36.565],[-1.861,-37.296,47.896],
[11.932,-27.640,40.609],[-0.052,-26.451,46.365],[4.006,-27.134,34.619],
[2.789,-27.350,37.220],[1.117,-29.500,36.433],[7.650,-37.057,46.947],
[8.156,-38.936,39.158],[5.633,-26.733,37.311],[4.937,-34.746,38.882],
[5.542,-34.265,48.328],[7.285,-38.697,42.028],[2.021,-24.478,35.803],
[2.167,-39.213,36.293],[11.704,-37.795,49.204],[10.223,-34.350,42.585],
[1.000,-26.918,35.030],[0.054,-35.867,34.252],[12.359,-23.942,46.003],
[-1.135,-39.912,44.998],[-0.629,-27.070,40.396],[-0.988,-24.567,50.161],
[3.016,-24.734,52.724],[5.865,-26.069,33.093],[8.533,-25.577,51.307],
[10.549,-24.680,33.743],[-2.956,-29.270,50.893],[3.052,-29.790,34.277],
[3.433,-30.563,48.557],[7.331,-28.360,32.548],[10.047,-30.179,49.932],
[-0.163,-32.553,51.628],[2.850,-32.384,36.654],[3.160,-33.167,48.378],
[5.415,-33.849,36.304],[7.029,-30.803,50.832],[12.360,-27.415,37.254],
[3.142,-35.414,49.427],[2.968,-37.939,47.125],[5.722,-31.247,47.235],
[8.627,-35.299,52.329],[9.907,-33.096,51.523],[0.501,-34.904,50.656],
[5.216,-39.936,47.686],[5.299,-36.067,51.125],[7.476,-36.604,49.639],
[10.290,-37.260,51.432],[0.507,-37.616,50.721],[4.463,-40.933,49.977],
[3.899,-39.124,52.100],[4.660,-22.904,35.652],[-1.412,-28.037,34.672],
[-2.360,-22.431,39.789],[-2.249,-25.142,41.910],[14.030,-22.586,43.300],
[14.746,-22.755,48.629],[12.368,-28.205,51.860],[14.077,-29.973,37.161],
[14.022,-31.730,39.609],[11.913,-29.188,45.548],[13.541,-26.653,46.321],
[13.678,-31.923,35.005],[12.811,-34.678,35.155],[11.606,-33.810,40.237],
[-2.426,-31.703,44.051],[14.470,-31.070,49.956],[14.138,-33.803,51.248],
[11.117,-36.979,34.650],[11.154,-36.625,37.738],[12.628,-35.382,42.401],
[12.032,-34.602,47.145],[-3.866,-36.248,50.096],[13.718,-39.358,51.267],
[12.776,-38.484,39.613],[-0.412,-37.546,43.364],[-3.699,-33.973,43.032],
[14.260,-39.087,48.412],[14.939,-36.575,35.989],[-1.280,-38.894,38.557],
[11.025,-22.811,39.628],[-2.487,-40.034,47.520],[-0.466,-27.211,50.031],
[2.114,-25.690,47.821],[6.565,-23.198,49.718],[9.482,-22.355,49.086],
[12.596,-41.413,49.167],[-1.754,-22.574,48.601],[-0.428,-21.068,45.145],
[7.044,-23.476,44.496],[8.548,-24.905,47.902],[9.729,-23.430,45.746],
[-1.840,-25.942,44.523],[0.747,-25.257,43.108],[3.408,-24.002,41.455],
[5.232,-36.955,42.415],[12.811,-39.849,42.750],[13.741,-41.198,39.904],
[-0.361,-38.095,40.860],[-2.231,-40.803,36.934],[4.352,-41.298,40.582],
[10.295,-39.971,42.247],[10.359,-40.540,39.651],[-2.121,-26.363,36.876],
[-0.893,-38.443,36.082],[4.211,-38.731,38.059],[7.211,-23.498,35.025],
[13.649,-24.138,39.939],[-1.938,-39.066,50.368],[0.576,-40.180,53.099],
[0.498,-23.686,33.792],[9.751,-39.851,51.955],[12.916,-25.358,50.482],
[8.811,-13.732,41.144],[7.460,-11.539,42.806],[0.597,-13.928,46.149],
[11.307,-9.912,41.451],[0.413,-9.782,43.842],[-3.603,-15.584,40.815],
[-3.570,-12.726,41.748],[8.245,-4.711,42.094],[5.567,-10.009,44.009],
[13.958,-19.263,45.436],[9.836,-16.377,45.290],[1.133,-18.616,40.488],
[9.464,-15.477,38.379],[-0.867,-18.420,42.481],[3.523,-14.300,41.382],
[14.560,-17.001,47.464],[1.151,-18.825,45.418],[8.202,-9.345,41.404],
[2.141,-11.892,47.775],[10.504,-17.495,40.614],[-0.434,-16.607,44.822],
[1.280,-9.633,46.599],[-1.941,-10.095,37.722],[10.329,-8.247,44.160],
[11.006,-19.037,45.433],[5.224,-22.731,42.804],[10.465,-7.335,41.082],
[10.143,-8.938,38.789],[12.096,-9.997,48.274],[7.476,-11.246,39.574],
[6.859,-14.131,33.785],[13.697,-14.517,46.782],[11.313,-12.426,38.702],
[10.394,-14.531,47.452],[1.615,-11.552,41.165],[3.195,-9.494,43.235],
[1.693,-12.028,44.373],[7.903,-7.815,45.628],[2.466,-9.600,50.195],
[8.380,-14.297,36.236],[0.653,-13.928,48.755],[8.164,-6.725,40.020],
[1.504,-6.975,49.582],[13.806,-22.338,37.845],[1.830,-13.940,38.634],
[13.583,-8.283,34.589],[-1.926,-9.374,45.418],[-3.178,-7.453,44.112],
[15.647,-7.694,49.397],[4.738,-7.908,49.410],[7.369,-4.936,38.150],
[1.049,-4.227,41.357],[9.116,-4.652,35.822],[10.297,-6.065,37.963],
[12.301,-16.291,51.855],[-0.305,-8.185,48.157],[2.461,-21.348,44.675],
[7.322,-21.823,47.363],[10.507,-17.422,47.949],[11.332,-12.268,33.840],
[7.585,-17.798,35.689],[2.091,-15.554,44.713],[2.502,-16.375,39.968],
[6.450,-14.367,38.188],[6.641,-15.344,50.732],[6.758,-17.724,41.794],
[0.596,-16.551,47.565],[2.959,-17.931,33.340],[-4.580,-20.182,41.160],
[-3.791,-18.130,36.358],[6.350,-14.380,45.632],[4.055,-18.901,41.812],
[-0.597,-16.139,40.198],[3.589,-17.588,37.536],[8.500,-18.992,44.396],
[2.373,-15.433,36.337],[12.476,-5.314,50.246],[0.019,-11.051,48.704],
[6.115,-14.068,42.746],[2.052,-20.821,39.394],[4.573,-16.658,45.786],
[6.814,-11.977,36.703],[12.942,-11.916,46.082],[9.617,-19.707,34.911],
[7.169,-10.785,46.376],[10.135,-16.612,35.841],[-0.750,-13.495,41.306],
[7.767,-9.566,37.536],[-2.909,-12.889,48.424],[4.528,-13.506,49.932],
[8.427,-7.892,48.794],[14.937,-14.910,37.391],[14.414,-10.587,49.381],
[11.589,-8.576,36.490],[10.120,-11.653,43.069],[2.572,-9.081,40.314],
[-0.513,-13.406,34.970],[4.409,-11.377,40.294],[5.159,-8.076,46.206],
[2.866,-6.032,39.895],[10.050,-10.723,45.770],[10.643,-8.934,33.739],
[11.144,-7.484,48.254],[12.478,-5.764,36.478],[-2.088,-19.074,45.858],
[10.218,-11.097,36.469],[-4.019,-8.003,47.228],[4.511,-8.675,33.954],
[4.046,-10.241,36.979],[0.557,-11.546,36.664],[7.250,-17.171,46.569],
[8.915,-18.638,38.004],[4.264,-7.139,36.383],[5.550,-15.930,40.220],
[6.228,-15.348,48.202],[6.455,-17.952,38.828],[1.625,-6.285,36.577],
[2.408,-19.908,36.590],[12.652,-18.397,49.580],[11.209,-15.290,43.097],
[1.314,-8.238,34.772],[-2.306,-17.174,34.411],[10.987,-4.927,45.176],
[-1.023,-22.768,43.289],[0.536,-7.261,39.324],[-3.183,-4.787,51.249],
[2.216,-5.513,52.428],[6.881,-7.552,32.239],[10.054,-8.841,50.475],
[10.613,-6.391,34.640],[-1.960,-11.646,50.557],[1.749,-10.812,34.473],
[5.184,-10.943,51.725],[6.379,-10.493,32.705],[9.904,-11.657,48.594],
[-1.243,-14.104,50.799],[4.289,-13.188,36.387],[3.337,-14.311,47.522],
[4.116,-14.522,33.343],[8.400,-11.705,50.990],[13.245,-12.731,31.947],
[3.836,-16.399,48.993],[3.182,-19.061,47.226],[4.076,-13.388,44.939],
[9.363,-16.445,50.929],[11.079,-14.610,50.223],[1.447,-16.167,50.350],
[4.682,-22.459,45.879],[5.453,-17.526,51.582],[7.225,-17.771,49.480],
[8.492,-19.145,51.172],[-0.042,-18.880,51.358],[4.714,-24.095,48.107],
[5.824,-20.433,50.131],[3.674,-3.566,36.936],[-1.394,-9.513,34.686],
[-1.086,-4.936,39.329],[-1.239,-7.475,42.026],[14.411,-6.282,44.833],
[13.690,-5.754,47.382],[11.248,-11.521,51.124],[15.276,-12.097,37.280],
[14.349,-11.604,39.799],[14.411,-9.541,43.189],[12.702,-8.891,45.514],
[13.759,-13.553,35.475],[13.640,-16.042,34.157],[12.536,-14.866,40.338],
[-3.205,-13.949,44.800],[12.961,-12.817,49.128],[13.892,-15.154,49.525],
[11.450,-17.929,33.741],[11.970,-17.451,38.165],[13.903,-18.008,42.998],
[12.498,-16.766,45.163],[-3.581,-18.508,51.200],[13.590,-21.963,51.688],
[11.864,-19.771,39.818],[-3.099,-19.978,43.527],[-2.841,-16.911,44.306],
[15.630,-19.603,50.873],[14.580,-18.833,34.230],[-2.348,-19.569,38.628],
[11.579,-4.772,40.069],[-3.798,-20.287,48.315],[-1.455,-9.264,51.311],
[3.083,-7.030,47.187],[6.158,-6.115,50.763],[9.041,-5.340,49.778],
[12.999,-21.407,46.854],[-0.187,-3.340,48.781],[1.146,-4.308,46.003],
[7.646,-3.863,44.917],[8.900,-5.542,46.904],[9.266,-2.759,46.623],
[-1.658,-5.114,45.560],[1.508,-7.302,42.612],[4.501,-6.321,43.730],
[6.964,-20.437,42.347],[12.895,-25.052,42.425],[14.639,-21.102,41.191],
[-0.217,-20.977,40.966],[-2.313,-23.110,36.535],[5.010,-20.813,40.355],
[9.772,-21.370,43.924],[10.170,-20.969,37.588],[-1.717,-7.046,37.705],
[0.202,-17.283,35.253],[4.951,-20.116,37.845],[5.911,-4.695,35.793],
[14.925,-5.373,37.605],[0.797,-21.409,49.324],[2.438,-20.975,51.476],
[-0.585,-6.243,34.724],[9.822,-21.542,51.681],[14.314,-7.159,51.926],
[9.870,4.334,39.808],[6.186,5.881,43.104],[0.152,4.614,42.025],
[11.566,8.548,42.868],[-0.116,9.864,43.634],[-1.804,2.566,39.047],
[-2.206,6.621,40.515],[7.787,13.994,40.907],[4.882,8.445,43.647],
[14.281,-0.570,44.990],[8.563,1.974,46.400],[1.279,-0.239,41.476],
[10.451,-0.553,39.462],[-1.673,2.032,42.069],[3.228,3.164,41.923],
[16.313,1.912,48.489],[1.994,-0.636,44.478],[7.485,8.470,41.947],
[2.995,8.292,47.394],[11.765,-0.243,42.221],[0.781,2.034,43.784],
[0.939,8.494,45.595],[-2.704,9.955,37.223],[9.597,7.047,44.148],
[11.825,0.267,45.038],[4.855,-3.087,43.849],[9.448,12.261,42.392],
[12.589,8.664,39.927],[11.076,8.549,48.313],[6.752,6.523,39.621],
[6.779,4.310,35.247],[14.467,3.230,46.815],[11.734,5.360,37.691],
[9.655,4.182,47.642],[2.674,5.333,40.784],[2.811,10.265,43.791],
[2.005,5.250,43.968],[8.730,10.898,44.751],[2.851,9.347,49.820],
[9.024,4.503,37.085],[0.376,6.755,47.794],[6.579,10.857,41.128],
[2.370,12.060,49.345],[12.780,-3.436,35.178],[-0.172,4.571,38.843],
[13.981,11.313,34.608],[-4.292,8.035,45.863],[-2.892,10.486,43.710],
[15.584,11.698,48.682],[6.312,9.306,47.093],[7.751,13.803,37.822],
[0.586,13.667,40.666],[8.820,12.693,35.151],[10.619,12.313,37.038],
[11.863,3.713,49.537],[0.153,11.068,48.286],[2.287,-3.522,43.680],
[7.448,-2.257,48.727],[9.412,1.843,49.144],[10.969,5.598,34.774],
[5.011,0.513,35.134],[2.288,1.842,46.129],[2.295,3.198,39.004],
[6.815,3.634,38.165],[7.588,2.539,50.986],[8.046,2.435,41.069],
[0.688,2.441,48.485],[0.310,1.194,35.038],[-4.791,-0.440,40.881],
[-1.207,-0.142,36.847],[7.287,4.559,45.400],[4.880,0.958,41.935],
[0.371,2.176,40.541],[3.097,0.903,37.613],[9.130,-0.444,44.066],
[1.324,3.911,35.645],[11.647,14.768,49.590],[0.095,8.414,50.234],
[5.544,3.192,43.683],[-0.773,-2.393,38.692],[4.686,2.706,46.399],
[4.625,7.353,38.178],[11.385,6.252,46.265],[9.730,0.058,36.488],
[7.476,7.990,48.886],[11.079,2.755,35.324],[-0.203,7.155,43.253],
[5.786,8.344,35.927],[-2.131,5.808,48.250],[5.292,3.979,50.054],
[8.962,11.323,47.569],[15.605,1.953,36.941],[13.641,9.063,49.877],
[12.610,10.220,37.176],[10.518,6.729,40.883],[2.890,9.930,38.635],
[-1.066,6.200,34.469],[3.619,8.064,40.701],[4.647,11.976,47.160],
[3.216,12.759,39.676],[9.097,8.557,46.465],[9.713,9.468,35.165],
[11.032,11.970,49.086],[15.219,13.681,34.970],[-0.871,-0.539,48.430],
[10.123,7.817,37.261],[-0.935,11.329,45.729],[4.090,9.780,34.192],
[2.241,7.802,35.967],[-1.547,7.574,36.543],[6.638,-0.543,46.867],
[6.573,0.703,37.497],[7.280,11.057,37.804],[5.199,1.908,39.534],
[6.603,2.209,48.286],[5.180,-1.164,40.405],[4.858,10.872,36.435],
[2.238,-1.884,34.498],[12.995,-0.827,50.290],[10.906,2.595,42.160],
[1.298,10.323,36.529],[-1.361,3.419,34.908],[11.200,13.141,45.269],
[-0.272,-2.812,43.839],[-0.426,11.089,38.687],[-2.516,12.340,52.677],
[2.936,13.205,51.545],[7.002,11.135,33.631],[9.016,9.852,50.285],
[11.169,12.614,33.583],[-4.068,6.525,49.955],[2.008,5.794,33.966],
[5.719,6.906,50.645],[7.080,7.728,33.087],[9.513,6.527,48.702],
[-1.098,3.933,49.988],[3.911,5.230,36.636],[2.910,4.134,48.432],
[4.185,3.786,34.362],[7.915,7.113,52.058],[13.334,7.053,33.835],
[3.847,0.840,48.892],[1.617,-1.098,47.581],[5.382,5.430,47.842],
[9.961,2.669,52.329],[9.579,5.098,50.920],[3.185,2.470,51.243],
[5.219,-3.167,47.523],[5.694,1.089,52.095],[7.574,-0.217,50.411],
[10.315,-0.799,50.985],[-0.016,2.102,51.507],[3.286,-2.639,49.731],
[5.454,-2.605,51.818],[4.244,14.862,36.243],[-2.585,9.162,34.354],
[-3.742,14.601,38.684],[-1.394,12.018,41.540],[14.668,13.881,45.112],
[13.188,12.284,47.211],[11.617,11.130,51.593],[13.538,7.613,37.212],
[13.489,6.246,41.223],[13.735,9.085,44.260],[11.502,10.240,46.169],
[13.521,4.239,35.432],[13.494,1.377,35.310],[12.722,3.290,40.261],
[-3.205,6.378,43.332],[12.004,6.434,50.375],[14.525,3.377,50.732],
[11.773,-0.978,34.339],[12.016,0.834,37.547],[12.187,3.155,45.136],
[13.696,1.541,49.023],[-3.651,0.528,49.683],[15.985,-1.793,55.366],
[12.566,-1.974,37.617],[-0.703,-0.258,43.459],[-4.274,3.262,43.480],
[14.466,-3.108,50.052],[14.821,-0.737,36.292],[-2.691,-2.235,36.419],
[13.600,12.567,39.417],[-1.913,-1.092,45.765],[-2.283,10.368,49.792],
[1.833,11.993,45.731],[6.201,14.035,48.441],[9.179,15.451,49.168],
[12.411,-3.248,47.464],[-0.696,15.549,47.561],[1.855,14.532,47.179],
[5.292,14.483,44.789],[8.490,13.737,46.440],[10.322,15.566,45.786],
[0.093,13.257,43.468],[1.907,10.259,41.183],[4.398,12.153,42.561],
[7.254,-2.212,42.793],[12.632,-2.922,42.290],[14.645,-3.720,40.182],
[-1.863,-2.507,41.588],[-3.511,-4.385,37.304],[3.923,-3.500,40.493],
[10.289,-2.689,43.470],[10.064,-3.154,38.135],[-1.459,13.327,36.693],
[-0.790,-1.614,34.616],[5.790,-2.182,37.886],[5.908,13.106,35.082],
[15.126,11.715,37.218],[0.468,-1.370,50.690],[1.961,-2.622,52.341],
[1.459,13.100,34.273],[8.464,-2.737,51.139],[13.615,13.079,50.927],
[8.429,25.749,41.586],[7.651,24.316,43.834],[-0.772,22.711,43.980],
[10.992,27.078,41.700],[1.299,27.207,42.753],[-2.424,21.593,39.373],
[-1.383,24.533,41.012],[8.043,32.310,42.313],[5.311,26.039,44.905],
[14.141,17.038,45.430],[9.111,20.145,42.941],[2.532,19.737,41.448],
[8.724,19.118,39.202],[-2.506,20.464,42.261],[3.503,22.090,42.682],
[14.322,19.585,47.183],[0.118,18.200,45.425],[6.877,28.138,38.463],
[2.308,24.928,47.705],[10.995,19.261,40.860],[-0.001,19.704,42.929],
[1.841,27.714,47.241],[-1.032,27.798,37.333],[10.474,28.921,44.818],
[11.605,17.825,45.925],[5.638,15.005,42.268],[9.860,29.791,41.769],
[9.862,28.943,38.859],[12.580,26.490,47.960],[7.310,25.371,39.239],
[8.862,23.494,35.194],[14.662,23.236,46.007],[14.458,27.247,38.925],
[9.839,22.493,47.515],[1.397,25.648,40.123],[3.696,28.007,43.714],
[2.609,23.711,44.893],[7.858,29.694,45.569],[1.712,28.658,49.782],
[10.932,21.884,36.425],[0.628,23.800,49.635],[5.769,30.969,38.943],
[2.046,31.140,50.822],[13.166,15.025,36.174],[2.307,23.808,38.240],
[14.493,29.995,35.452],[-1.555,26.454,45.961],[-3.064,28.897,44.348],
[15.345,30.668,48.689],[5.854,28.564,47.580],[6.311,34.640,38.005],
[1.039,32.788,40.750],[9.170,31.618,35.657],[8.841,33.006,38.841],
[12.614,21.554,49.461],[-0.673,27.835,48.753],[2.595,14.368,44.461],
[7.718,16.662,46.663],[10.717,19.921,48.544],[10.412,23.730,33.158],
[6.963,18.215,35.402],[2.125,20.462,44.675],[1.445,22.195,40.732],
[6.800,22.681,36.626],[6.070,22.319,50.947],[7.350,21.246,40.161],
[-0.214,21.854,47.844],[4.612,19.443,33.344],[-3.777,16.569,41.041],
[-1.176,19.508,37.495],[7.386,22.573,46.217],[4.803,18.955,40.039],
[0.704,20.386,39.014],[2.669,19.643,36.868],[9.079,17.713,44.491],
[2.393,21.588,35.235],[11.778,33.326,49.067],[-0.688,26.018,50.559],
[6.524,21.555,42.852],[2.054,16.711,40.168],[5.199,20.971,45.672],
[5.436,25.937,37.266],[12.023,24.179,46.117],[10.749,16.633,34.708],
[7.577,26.574,48.538],[9.563,19.619,35.754],[-1.019,25.960,43.130],
[8.677,27.603,36.508],[-1.475,24.407,48.055],[3.999,23.642,49.939],
[8.841,29.198,48.142],[14.346,22.436,37.345],[15.242,28.182,47.583],
[12.534,28.447,37.420],[10.501,25.167,43.384],[2.895,28.379,40.602],
[-0.963,23.619,34.059],[4.145,25.945,40.602],[3.663,29.481,46.388],
[4.004,31.100,40.913],[8.300,26.727,45.955],[10.133,27.286,34.173],
[11.554,30.885,47.836],[14.342,32.704,34.687],[-1.614,19.636,47.051],
[10.944,25.962,36.591],[0.134,30.908,44.809],[5.004,29.258,33.629],
[3.470,27.338,35.392],[-0.099,25.181,36.637],[7.888,20.126,47.259],
[9.548,17.174,37.092],[6.026,30.045,36.154],[5.541,20.337,36.923],
[5.475,20.636,48.439],[7.425,16.593,39.271],[3.187,31.322,36.935],
[2.244,16.984,37.447],[12.622,19.034,50.268],[11.395,22.177,44.107],
[1.147,28.578,35.324],[-0.335,20.282,34.969],[11.406,31.565,44.923],
[0.672,16.052,42.882],[0.402,28.854,39.618],[-1.788,31.205,51.465],
[4.893,31.509,52.383],[7.438,30.643,33.377],[8.779,28.065,51.245],
[12.232,30.460,33.536],[-3.454,26.162,50.235],[1.490,25.476,34.039],
[4.400,26.335,48.868],[6.367,26.885,34.672],[10.152,25.500,48.599],
[-1.254,21.965,50.370],[4.671,24.190,35.477],[2.689,21.629,48.063],
[4.996,21.793,34.337],[7.741,25.533,51.025],[12.789,25.664,33.863],
[3.115,19.018,49.176],[2.741,18.142,46.616],[5.077,23.540,47.217],
[11.153,21.102,52.232],[10.096,22.770,50.179],[1.559,21.633,51.207],
[4.407,15.781,46.928],[5.720,19.004,50.331],[8.213,17.949,49.011],
[8.744,19.011,52.029],[0.449,18.341,48.840],[3.912,14.814,49.591],
[6.259,16.566,52.946],[4.023,33.710,35.648],[-2.221,27.555,34.704],
[-2.661,33.178,37.908],[-1.944,30.309,41.377],[13.524,32.756,43.944],
[16.311,32.323,46.895],[13.617,28.905,50.601],[13.756,25.705,36.484],
[13.166,24.567,38.818],[14.302,26.960,44.772],[12.875,28.832,46.528],
[15.768,24.340,35.003],[13.781,21.016,34.584],[12.720,21.921,39.882],
[-3.080,24.092,43.218],[13.391,24.249,49.033],[15.260,22.393,49.701],
[11.736,19.039,33.584],[11.560,19.879,38.368],[12.618,19.338,43.465],
[12.292,21.332,46.468],[-3.731,19.880,49.177],[15.827,17.008,51.959],
[13.521,17.281,38.576],[-1.652,17.554,43.001],[-3.591,21.774,44.442],
[14.093,16.811,49.922],[14.230,18.020,34.115],[-3.762,18.504,37.412],
[12.407,30.848,39.972],[-3.146,17.385,45.852],[-0.670,28.776,51.531],
[0.632,31.456,47.898],[5.484,30.499,49.313],[8.020,31.755,49.928],
[13.638,15.359,47.804],[-0.449,33.708,45.158],[2.409,32.890,46.431],
[6.704,33.806,45.302],[8.945,32.259,45.124],[10.711,33.652,46.764],
[-2.129,32.028,43.529],[1.985,29.872,42.608],[4.043,31.185,43.620],
[8.334,17.259,41.603],[12.882,13.498,41.780],[12.124,16.144,40.554],
[-0.706,16.396,40.093],[-0.034,15.963,36.458],[4.594,16.171,40.076],
[10.316,15.293,42.449],[11.177,15.063,37.608],[-0.335,31.500,38.004],
[-1.791,17.778,35.567],[5.213,17.111,37.394],[6.564,32.523,35.750],
[13.516,32.147,37.375],[-0.545,16.063,50.124],[3.004,16.986,51.863],
[-1.266,32.319,35.447],[12.270,16.018,51.954],[13.283,31.693,50.560],
[8.943,42.235,41.612],[7.875,43.217,44.070],[2.233,42.739,44.719],
[10.779,46.731,42.115],[0.727,45.515,44.171],[-2.092,40.881,38.953],
[-2.208,41.872,41.455],[8.231,50.649,42.118],[6.414,46.369,41.806],
[15.501,36.850,45.130],[9.307,38.972,43.575],[1.383,36.825,40.828],
[9.989,37.302,40.140],[-2.838,37.729,41.279],[4.755,42.317,39.232],
[15.300,38.229,47.486],[0.389,36.708,45.269],[7.604,46.766,38.811],
[4.651,43.466,45.756],[12.727,36.206,42.812],[-0.609,39.547,41.701],
[2.136,44.035,47.335],[-3.359,46.488,37.775],[12.480,45.917,43.958],
[11.008,36.863,44.977],[4.926,33.746,43.010],[9.838,48.914,43.377],
[10.555,46.577,39.260],[13.163,45.301,48.849],[6.315,43.786,37.146],
[7.512,39.939,35.832],[14.077,41.981,47.284],[10.359,41.474,39.194],
[11.547,40.921,47.141],[3.642,44.660,40.925],[3.073,46.871,43.705],
[3.705,41.807,41.896],[8.272,46.780,45.867],[4.232,45.528,48.531],
[8.490,42.412,36.606],[0.071,42.574,48.544],[7.348,49.432,39.854],
[0.973,49.547,50.476],[13.546,35.082,35.283],[1.876,43.611,39.262],
[14.087,46.551,35.367],[-3.365,44.691,47.480],[-3.241,48.252,45.007],
[14.712,49.038,50.759],[6.006,47.664,47.872],[6.670,51.529,38.319],
[-0.059,50.714,42.049],[10.144,51.652,35.624],[9.464,50.955,38.132],
[12.467,39.790,49.990],[1.114,46.986,47.102],[2.361,34.603,44.318],
[6.803,33.073,47.870],[9.911,38.359,48.357],[9.891,41.664,34.078],
[7.565,36.702,34.010],[1.680,40.231,43.713],[2.750,40.280,39.334],
[7.466,40.925,39.234],[7.444,41.796,51.333],[9.146,39.660,41.030],
[0.324,41.131,46.197],[2.393,38.330,34.095],[-5.345,38.443,40.349],
[-2.184,39.124,36.962],[8.560,40.679,46.818],[4.668,37.618,41.471],
[0.119,38.731,39.056],[4.727,39.472,34.963],[7.425,36.925,44.073],
[2.322,40.935,34.914],[13.404,53.768,50.425],[0.616,45.392,50.040],
[6.024,40.424,41.618],[3.666,37.447,38.821],[2.333,39.139,46.399],
[5.292,46.228,36.475],[12.156,44.352,46.582],[9.240,37.763,35.794],
[6.915,44.515,48.262],[11.803,41.820,36.262],[1.102,42.667,41.886],
[7.868,47.372,35.751],[-2.648,42.172,47.960],[4.807,40.351,49.678],
[8.642,48.248,48.566],[14.181,39.983,36.794],[15.657,46.064,49.576],
[12.046,47.596,37.353],[9.948,41.796,44.768],[3.150,47.708,40.901],
[-0.909,42.875,35.506],[6.059,44.696,39.478],[5.215,49.391,45.544],
[1.353,50.177,39.563],[10.082,44.593,44.963],[9.404,45.320,34.425],
[10.234,50.351,48.301],[13.252,51.426,34.444],[-0.687,37.689,48.016],
[9.582,44.995,37.308],[-0.513,48.883,46.645],[4.640,47.367,32.612],
[2.504,46.731,34.777],[-0.297,45.226,36.654],[6.360,36.252,46.617],
[10.132,36.050,37.921],[5.619,49.443,36.972],[5.514,39.791,37.714],
[7.122,38.785,48.110],[7.585,36.915,39.262],[2.234,49.942,36.509],
[2.523,35.939,35.306],[12.086,35.716,47.784],[11.371,41.341,42.303],
[0.354,47.752,36.638],[-1.010,39.692,34.364],[12.593,48.622,44.438],
[0.447,34.766,42.402],[-0.443,48.133,40.723],[-0.978,50.823,49.312],
[3.512,50.069,52.071],[5.838,49.141,34.519],[9.926,47.023,50.702],
[11.355,48.850,34.880],[-2.076,43.929,51.502],[1.460,45.866,31.986],
[5.902,43.950,50.746],[6.694,45.646,33.659],[9.634,43.031,48.030],
[1.021,43.046,51.120],[4.380,42.827,35.588],[2.999,41.383,48.017],
[5.738,43.232,32.888],[8.821,44.224,51.284],[13.979,43.908,33.251],
[4.506,37.805,48.097],[3.170,37.290,44.574],[5.969,41.650,47.478],
[8.406,39.449,50.510],[10.287,41.806,50.189],[1.299,40.446,51.326],
[4.422,34.451,47.550],[5.642,38.025,50.571],[7.005,35.507,50.282],
[10.788,35.779,50.315],[0.044,38.304,50.492],[4.505,33.165,50.092],
[4.029,36.923,53.404],[2.459,52.434,35.087],[-2.448,45.676,34.264],
[-3.189,51.418,39.872],[-2.378,49.441,41.852],[14.788,49.998,43.827],
[14.731,49.592,46.475],[12.744,46.840,51.828],[14.429,43.017,37.328],
[14.454,43.048,41.093],[15.175,46.143,44.713],[13.603,47.235,46.533],
[15.643,42.251,35.067],[13.371,40.381,34.385],[12.632,40.514,40.120],
[-1.534,43.618,43.145],[12.971,42.607,49.654],[14.908,41.273,50.848],
[12.236,37.000,33.318],[11.256,38.750,37.699],[12.682,38.911,44.127],
[12.543,38.263,47.304],[-3.598,37.827,50.779],[14.620,33.468,52.171],
[13.637,35.690,38.253],[-1.305,36.537,43.272],[-2.289,41.164,44.746],
[15.475,36.538,49.820],[14.499,37.280,35.956],[-2.404,36.705,38.595],
[12.325,51.178,39.334],[-2.698,34.821,47.122],[-0.998,47.757,49.859],
[2.994,48.995,47.227],[7.296,50.650,47.466],[8.319,51.203,49.832],
[13.724,33.659,46.587],[0.309,52.121,47.188],[2.847,51.076,45.127],
[7.270,51.173,44.551],[9.137,49.828,45.726],[12.348,52.173,46.290],
[-2.135,51.031,45.521],[1.645,48.968,43.139],[5.232,49.369,42.399],
[6.678,35.392,41.329],[11.494,32.634,41.961],[14.031,33.865,41.267],
[-0.982,34.326,39.783],[-1.333,34.875,36.227],[4.212,34.687,40.537],
[9.861,34.830,42.144],[11.750,33.804,38.608],[-1.907,48.642,38.217],
[-0.671,37.419,35.608],[5.213,36.454,35.749],[7.178,52.694,35.287],
[13.772,49.824,37.219],[-0.030,35.011,47.411],[1.368,35.484,51.618],
[-1.496,49.410,35.391],[8.906,34.545,51.883],[11.876,50.084,50.375],
[13.195,55.463,44.830],[9.896,57.547,45.454],[2.544,56.569,41.282],
[8.864,56.576,38.593],[-2.248,58.022,43.422],[16.209,58.012,47.018],
[-0.833,55.667,46.378],[11.393,55.029,42.100],[0.241,57.519,42.530],
[10.968,55.059,46.565],[4.991,51.933,43.608],[13.362,52.888,37.166],
[12.347,57.396,50.201],[1.260,55.614,44.165],[5.995,52.788,48.848],
[10.057,57.410,48.137],[7.321,55.383,33.643],[0.568,58.181,45.314],
[2.832,58.169,39.017],[7.440,57.887,40.965],[-0.398,58.405,47.887],
[4.443,55.429,34.726],[-4.315,55.972,39.632],[-2.291,57.405,35.351],
[3.809,54.386,42.669],[-0.891,57.420,40.105],[4.455,58.644,36.035],
[8.881,54.709,44.710],[2.032,-59.243,36.088],[7.496,56.562,43.346],
[1.636,53.102,39.960],[4.074,57.524,45.945],[7.577,56.415,36.141],
[9.886,58.502,36.546],[14.512,56.833,37.733],[-1.115,54.146,48.580],
[6.536,55.373,47.382],[7.290,54.247,37.716],[5.327,56.442,37.661],
[7.397,57.738,48.465],[5.189,54.528,39.834],[1.772,55.013,37.918],
[13.430,57.116,47.290],[0.324,58.404,34.577],[-0.039,53.093,44.336],
[3.640,56.356,48.741],[2.758,55.064,46.425],[8.433,58.525,53.062],
[1.951,57.180,51.007],[5.634,53.177,45.982],[5.472,57.738,50.747],
[9.875,57.653,51.070],[10.617,55.055,51.255],[-0.694,54.594,51.968],
[3.570,53.633,50.161],[5.869,55.221,50.224],[13.413,57.448,35.302],
[13.054,55.895,33.019],[12.677,58.836,38.113],[12.866,57.911,44.146],
[9.794,-58.756,46.587],[-2.421,57.514,49.293],[15.120,51.591,51.135],
[12.835,55.037,39.262],[-2.281,54.898,43.166],[-4.919,58.570,44.343],
[14.964,55.969,50.795],[14.433,53.706,35.013],[-2.396,54.205,38.639],
[-2.837,53.597,45.643],[14.412,52.116,47.674],[6.819,53.754,43.128],
[10.733,51.630,43.799],[13.800,52.782,41.723],[-1.580,52.696,41.382],
[-2.352,51.573,37.173],[3.974,51.532,41.063],[9.198,53.305,42.136],
[9.943,54.174,39.272],[-0.362,54.362,36.421],[4.654,53.653,36.901],
[0.930,53.279,50.052],[1.760,51.974,52.445],[10.867,52.379,51.245],
[-2.469,-54.153,56.450],[10.223,-54.827,57.492],[8.704,-47.293,58.374],
[-1.715,-47.449,56.071],[10.263,-44.904,56.913],[7.892,-49.936,55.588],
[7.795,-52.322,53.417],[10.539,-49.691,59.110],[10.466,-50.232,56.390],
[6.993,-43.542,57.358],[11.962,-58.397,54.518],[-0.067,-52.916,55.475],
[14.030,-44.958,53.835],[6.661,-40.571,57.078],[0.553,-41.715,57.304],
[10.376,-42.657,53.765],[9.551,-41.154,56.016],[11.894,-50.800,53.282],
[6.485,-55.516,53.991],[1.319,-52.338,-59.700],[6.604,-51.838,57.267],
[1.524,-55.198,52.957],[0.115,-55.751,55.996],[-0.565,-54.240,58.147],
[3.619,-52.760,57.491],[3.234,-53.377,54.620],[1.322,-57.159,58.550],
[5.502,-48.805,56.776],[9.848,-54.207,54.838],[8.986,-52.341,57.837],
[7.487,-47.178,55.471],[15.834,-54.073,54.504],[13.539,-44.065,56.597],
[-1.642,-50.104,53.848],[3.440,-47.492,58.241],[3.685,-41.225,56.923],
[11.840,-46.562,52.836],[15.850,-41.021,53.351],[9.817,-48.340,54.293],
[3.946,-45.808,52.136],[3.691,-46.352,55.411],[-0.583,-49.877,57.144],
[8.760,-57.565,54.597],[4.641,-43.724,55.545],[5.353,-54.494,56.440],
[7.127,-55.993,58.284],[2.836,-41.738,53.487],[2.796,-55.857,55.340],
[1.185,-45.646,53.234],[-1.761,-54.265,53.146],[-0.544,-44.604,58.490],
[5.487,-43.704,51.812],[9.618,-45.118,51.469],[1.441,-49.318,53.173],
[4.889,-48.202,52.926],[3.229,-50.214,55.343],[4.357,-51.244,52.848],
[13.862,-48.279,51.592],[4.841,-40.519,54.585],[-1.033,-47.651,52.844],
[-3.529,-40.608,-59.826],[12.758,-48.264,58.239],[14.617,-49.529,56.769],
[14.594,-50.630,54.237],[13.595,-52.918,53.311],[12.571,-51.497,58.051],
[12.361,-55.624,54.615],[14.145,-54.210,56.490],[12.177,-56.861,58.101],
[15.172,-55.544,51.386],[-4.341,-56.278,55.769],[11.231,-41.614,57.959],
[-0.784,-58.849,55.748],[4.833,-59.367,58.048],[9.669,-57.711,57.455],
[-2.384,-44.460,56.148],[-0.636,-56.840,53.237],[5.377,-57.328,56.296],
[7.139,-42.348,54.740],[13.712,-42.304,58.746],[0.694,-43.215,53.920],
[-1.926,-35.105,57.804],[9.101,-36.226,58.328],[10.270,-29.316,58.775],
[-1.320,-28.614,54.850],[11.889,-26.876,58.921],[8.468,-31.778,56.731],
[7.642,-32.287,53.001],[11.020,-31.917,57.277],[10.861,-32.740,54.527],
[7.405,-24.935,58.632],[12.886,-40.063,54.141],[1.600,-31.676,55.723],
[14.189,-26.911,52.917],[6.688,-22.473,57.163],[-0.137,-25.115,59.021],
[9.791,-23.668,54.029],[9.428,-24.428,56.848],[10.714,-30.483,52.869],
[6.388,-38.513,53.871],[1.741,-33.919,57.649],[6.716,-34.273,57.581],
[2.526,-36.327,52.126],[-1.432,-36.924,52.959],[-1.969,-37.028,59.637],
[2.203,-36.566,56.589],[2.604,-34.871,54.294],[1.355,-38.176,58.814],
[6.545,-30.866,55.091],[9.180,-37.939,53.636],[9.215,-34.957,55.810],
[7.622,-28.635,56.524],[16.050,-33.619,55.826],[13.976,-25.242,55.631],
[-1.714,-31.387,54.250],[2.368,-29.650,-58.181],[2.674,-27.103,58.722],
[9.341,-27.928,54.254],[15.284,-24.338,52.847],[10.736,-29.568,55.788],
[5.017,-27.644,51.484],[3.644,-27.123,56.363],[1.250,-30.289,53.107],
[8.637,-38.226,56.461],[5.241,-25.007,56.668],[4.496,-35.264,56.381],
[7.389,-38.100,59.067],[2.165,-24.364,55.328],[2.503,-38.707,54.384],
[0.873,-27.573,53.360],[-2.116,-34.184,52.249],[-0.178,-27.585,56.999],
[5.532,-24.822,51.565],[11.220,-25.760,52.477],[3.288,-29.650,51.059],
[6.951,-28.612,53.121],[3.881,-31.952,53.466],[5.571,-34.371,53.247],
[14.031,-30.507,52.796],[2.677,-21.924,54.022],[-1.653,-26.251,52.350],
[-1.954,-23.039,58.336],[13.501,-28.782,55.277],[15.687,-33.018,-58.034],
[13.321,-32.022,55.298],[14.561,-34.702,53.639],[11.000,-33.620,-56.974],
[12.171,-37.766,53.351],[11.715,-36.484,57.402],[13.560,-37.942,55.929],
[15.611,-38.011,53.032],[-2.938,-37.724,57.066],[11.620,-22.562,-59.660],
[-2.593,-39.202,54.911],[4.563,-39.341,59.308],[10.870,-38.908,58.486],
[-2.031,-25.534,56.847],[-0.232,-37.526,55.565],[5.443,-37.749,56.800],
[7.327,-24.583,54.489],[14.046,-22.563,57.710],[0.272,-23.104,52.074],
[-3.945,-16.309,57.790],[9.879,-19.343,58.847],[7.711,-9.909,58.257],
[-3.465,-7.553,54.675],[10.956,-8.275,57.697],[7.848,-12.535,56.847],
[8.313,-14.063,52.423],[9.864,-12.184,58.496],[10.166,-13.482,55.333],
[6.098,-6.520,58.113],[13.910,-21.823,54.132],[0.867,-14.493,55.812],
[13.546,-6.690,54.439],[6.778,-4.164,56.944],[0.467,-8.500,-59.202],
[10.107,-5.806,54.469],[8.515,-7.602,56.655],[11.985,-11.157,54.869],
[6.233,-20.720,53.531],[1.742,-15.465,58.324],[7.449,-15.025,54.802],
[2.945,-18.144,51.193],[-2.366,-18.101,56.665],[0.321,-18.446,57.144],
[2.097,-16.985,55.901],[2.305,-15.724,52.741],[1.443,-20.162,58.933],
[4.737,-11.101,54.494],[8.936,-19.042,53.891],[10.319,-16.201,53.653],
[6.955,-10.288,55.798],[14.343,-14.882,56.410],[13.889,-8.100,57.068],
[-1.218,-12.395,53.203],[2.692,-10.481,57.937],[2.291,-5.406,59.335],
[10.190,-9.540,53.586],[14.781,-4.285,54.602],[9.654,-10.453,56.127],
[3.755,-7.935,52.750],[2.276,-9.148,54.684],[-0.562,-9.941,54.462],
[8.041,-19.954,56.642],[6.348,-7.122,54.019],[5.857,-17.338,57.462],
[6.870,-19.156,59.402],[2.059,-6.194,55.691],[1.580,-19.327,54.360],
[0.703,-7.858,52.387],[-0.336,-15.717,53.462],[-1.958,-7.517,58.700],
[8.435,-7.056,52.122],[11.395,-7.079,52.226],[1.347,-10.755,52.690],
[7.469,-9.820,52.884],[2.258,-12.515,54.464],[4.757,-14.555,52.966],
[13.881,-12.262,51.960],[4.593,-5.416,55.378],[-3.399,-10.310,53.582],
[-4.355,-4.280,58.485],[12.429,-11.112,58.082],[14.544,-13.205,58.646],
[14.369,-14.291,53.696],[13.944,-16.972,54.322],[12.333,-15.250,58.653],
[10.893,-18.773,51.938],[13.450,-17.602,57.384],[12.013,-19.970,57.480],
[13.465,-19.295,52.608],[-3.210,-20.602,58.049],[11.046,-4.943,59.187],
[-2.589,-20.366,55.367],[3.911,-22.293,58.628],[10.329,-21.869,56.421],
[-2.831,-5.792,56.674],[-0.776,-20.907,53.069],[4.956,-19.652,56.478],
[7.473,-4.407,52.811],[13.253,-4.366,56.863],[-0.407,-2.990,53.471],
[-2.050,3.774,58.818],[9.985,-0.184,57.806],[9.913,8.868,58.329],
[-2.759,8.964,54.616],[12.521,9.563,59.534],[7.763,6.998,56.271],
[7.719,4.501,52.883],[10.665,5.331,57.449],[10.622,4.641,54.671],
[7.261,10.303,57.385],[12.690,-2.515,54.309],[1.194,6.193,55.318],
[15.789,10.144,53.905],[6.523,15.238,58.183],[1.067,14.082,57.389],
[9.988,13.832,53.751],[8.248,11.985,54.864],[12.357,6.668,54.835],
[6.641,-0.363,55.037],[3.014,2.233,58.545],[6.942,4.606,55.362],
[1.686,0.448,52.609],[-3.096,0.920,55.427],[-0.713,1.745,-59.076],
[1.827,1.173,55.720],[1.852,3.658,53.455],[0.862,0.335,58.053],
[5.006,5.937,57.087],[8.593,0.403,53.294],[9.871,1.963,55.214],
[8.484,9.292,55.020],[15.662,2.716,55.049],[12.764,10.385,56.559],
[-2.635,6.340,53.868],[4.199,7.999,-57.628],[3.519,13.416,56.573],
[10.570,8.601,53.607],[13.766,11.844,53.621],[10.217,7.572,55.918],
[4.492,10.780,52.000],[3.734,9.738,56.611],[0.650,8.864,54.984],
[10.642,-1.127,55.402],[5.690,10.476,54.641],[6.542,3.537,58.169],
[7.034,-0.938,58.106],[4.102,12.591,53.926],[1.378,-1.407,54.956],
[1.688,10.231,53.009],[-2.015,1.913,53.104],[-1.936,12.116,-59.812],
[7.212,11.869,52.211],[11.144,11.424,54.463],[2.678,7.668,52.166],
[6.129,10.342,49.913],[3.719,6.124,54.227],[4.817,3.809,52.859],
[13.326,7.496,52.572],[3.004,15.190,54.142],[-1.049,8.233,52.731],
[-3.302,14.271,-58.635],[12.993,7.656,57.422],[15.328,5.286,56.501],
[14.599,4.993,53.734],[12.557,3.515,52.485],[11.628,2.224,-59.094],
[11.374,0.509,53.322],[12.043,0.875,56.603],[13.611,-1.642,57.067],
[13.761,-0.393,52.817],[-2.196,-0.401,58.001],[10.330,12.705,56.757],
[-2.551,-2.841,55.785],[4.710,-1.508,-59.438],[9.682,-3.576,56.353],
[-0.593,10.808,56.629],[-1.130,-0.569,52.455],[4.325,-1.720,54.373],
[6.614,13.908,53.985],[14.385,12.627,56.572],[0.196,13.153,52.452],
[-2.402,21.148,57.689],[9.735,18.417,58.646],[8.420,27.591,57.966],
[-2.163,26.749,56.046],[10.971,29.135,57.230],[6.244,24.347,55.420],
[8.088,23.430,52.730],[11.681,23.337,56.448],[10.123,23.190,54.187],
[8.075,30.443,-59.736],[13.247,14.590,54.912],[1.292,23.454,57.367],
[13.623,30.476,53.398],[7.217,33.234,55.931],[1.799,30.971,57.352],
[10.497,33.105,54.617],[9.321,32.133,57.397],[10.678,25.217,51.259],
[6.978,19.182,54.319],[2.754,21.332,57.451],[8.353,22.337,57.678],
[3.453,19.816,52.447],[-2.448,20.001,54.490],[0.343,20.462,58.579],
[2.318,19.544,55.749],[1.630,22.192,54.094],[1.000,16.612,58.596],
[4.468,26.245,56.447],[11.029,18.411,53.067],[9.498,20.682,54.678],
[7.126,27.339,55.508],[14.627,21.466,55.458],[11.817,27.916,54.081],
[-2.619,23.670,52.913],[4.142,26.743,59.470],[4.708,30.921,56.882],
[9.604,26.582,53.331],[14.397,32.096,55.682],[11.618,25.325,54.735],
[4.331,28.724,53.189],[3.005,28.216,55.568],[-1.439,23.615,56.228],
[9.394,17.821,55.889],[6.956,30.951,54.846],[5.105,21.263,56.272],
[6.916,17.923,59.410],[2.774,31.912,54.386],[1.718,17.648,54.283],
[0.192,28.261,54.849],[-0.243,20.254,53.168],[-1.578,30.702,56.141],
[6.587,29.706,51.804],[10.677,30.351,53.404],[3.345,26.928,51.487],
[6.638,26.313,53.294],[3.199,24.380,54.313],[4.828,22.434,53.367],
[13.316,24.361,51.839],[4.421,34.008,53.802],[-2.370,26.876,52.954],
[-3.747,32.757,59.244],[13.413,26.773,56.211],[15.608,25.219,57.146],
[13.999,24.227,55.222],[14.281,20.652,52.418],[12.297,22.563,-59.691],
[13.260,17.233,54.265],[12.228,20.111,56.753],[13.387,16.634,57.164],
[15.723,17.692,54.998],[-4.034,15.736,57.592],[11.031,32.933,-59.621],
[-0.148,15.564,54.715],[2.918,14.729,59.191],[9.728,15.241,57.253],
[-1.807,29.811,53.781],[-0.060,17.735,52.250],[5.828,17.009,55.962],
[6.841,33.322,52.918],[12.676,31.108,57.770],[-0.116,33.221,51.960],
[-1.831,40.154,57.532],[10.637,36.881,57.672],[8.039,45.986,57.163],
[-1.864,46.131,55.886],[11.225,48.167,57.915],[8.411,42.874,54.881],
[9.464,40.221,53.288],[11.438,45.147,58.503],[12.138,41.817,55.636],
[8.280,48.932,58.422],[13.335,34.740,54.191],[1.469,42.652,57.554],
[14.371,48.828,53.421],[5.689,52.069,55.886],[0.792,50.802,57.845],
[10.391,51.303,53.783],[9.914,50.376,56.635],[11.661,42.472,52.452],
[7.446,38.171,53.440],[1.138,39.893,58.047],[5.958,42.330,57.275],
[1.558,38.032,53.432],[-1.337,36.449,55.804],[-2.104,37.467,59.163],
[3.749,37.846,56.097],[2.793,42.327,53.132],[0.879,37.153,57.681],
[4.981,46.222,55.771],[10.211,37.397,53.401],[8.406,39.787,56.547],
[7.589,46.761,54.574],[14.670,41.832,56.345],[12.352,49.188,55.536],
[-0.808,43.207,53.891],[5.402,44.845,58.175],[2.189,48.580,58.677],
[9.977,46.304,53.279],[14.518,51.432,54.670],[10.924,44.274,55.319],
[4.191,47.470,52.511],[2.053,48.330,55.996],[0.640,45.955,56.595],
[8.542,37.031,55.683],[7.014,49.570,55.335],[5.592,39.818,55.612],
[5.176,36.984,58.370],[2.831,51.043,55.546],[2.104,35.546,55.852],
[1.511,46.188,53.537],[-1.702,39.920,52.717],[-1.135,48.873,58.239],
[6.375,48.707,51.471],[11.179,48.883,52.826],[4.132,44.712,52.539],
[7.035,45.991,52.086],[5.297,42.672,54.324],[5.680,40.129,52.573],
[13.753,44.345,51.664],[3.630,53.362,54.583],[-3.125,46.955,53.141],
[-3.549,50.904,57.861],[12.564,46.329,55.633],[14.018,43.444,58.599],
[14.820,42.325,53.576],[14.326,39.646,53.249],[10.385,39.465,58.943],
[12.676,37.167,52.374],[12.426,38.935,56.637],[12.703,34.333,57.575],
[16.041,37.461,52.598],[-2.838,34.377,56.798],[11.107,52.503,-58.903],
[-0.126,32.970,58.223],[4.954,34.036,57.439],[9.450,34.707,56.563],
[-1.095,48.709,55.350],[-0.654,36.591,52.791],[6.297,35.518,54.755],
[7.661,51.492,52.521],[15.010,53.120,57.207],[-0.061,51.301,55.046],
[10.094,54.457,58.088],[12.277,53.322,53.622],[7.858,56.154,52.139],
[3.095,57.190,-58.603],[3.726,55.370,52.558],[-1.001,55.523,54.571],
[-1.261,57.773,58.288],[3.494,57.472,56.702],[1.315,58.330,55.125],
[-1.037,54.186,58.611],[10.891,55.705,54.741],[9.686,57.726,56.089],
[14.633,56.836,56.676],[8.343,54.914,55.445],[6.920,57.305,55.193],
[7.198,54.536,57.888],[1.735,55.649,56.102],[-0.349,58.857,52.995],
[13.775,58.505,52.564],[13.111,55.933,53.311],[11.824,57.535,57.861],
[12.573,54.683,56.920],[14.809,53.861,52.734],[-2.436,55.667,57.094],
[-1.982,53.024,55.507],[4.283,52.102,58.725],[12.146,52.200,56.204],
[0.938,53.896,54.256],[5.372,54.976,55.751],[3.788,-57.398,-12.905],
[0.981,-59.286,-19.777],[10.888,58.157,-14.766],[-2.187,-47.986,-28.209],
[6.843,-46.077,-32.678],[1.354,-56.847,-25.986],[2.025,-49.889,-25.996],
[11.135,-56.718,-24.942],[10.946,-45.862,-28.050],[7.484,-48.039,-30.951],
[6.266,-48.622,-33.200],[-0.747,-48.262,-34.471],[3.697,-54.143,-34.319],
[11.189,-54.282,-20.347],[-0.593,-52.010,-36.970],[0.844,-54.980,-36.534],
[9.162,-47.951,-23.751],[1.286,-51.119,-23.262],[4.217,-52.902,-18.000],
[5.472,-53.693,-20.433],[-0.183,-52.559,-18.647],[3.200,-56.814,-16.581],
[-3.434,-46.984,-34.225],[-4.733,-45.396,-32.128],[12.513,-44.574,-29.911],
[11.915,-55.166,-29.441],[13.159,-55.857,-17.244],[1.111,-42.973,-29.567],
[5.667,-34.270,-37.579],[13.188,-35.351,-32.663],[0.141,-29.288,-35.029],
[15.290,-28.919,-34.818],[0.526,-40.417,-38.332],[1.422,-42.605,-41.844],
[6.215,-32.343,-33.271],[-4.706,-33.955,-32.529],[11.652,-30.111,-32.980],
[8.669,-25.888,-33.041],[3.471,-38.464,-24.563],[11.999,-37.541,-38.531],
[13.620,-45.118,-43.619],[-2.612,-28.209,-31.951],[0.003,-30.687,-39.940],
[5.897,-42.881,-32.036],[13.787,-42.962,-34.450],[10.929,-15.024,-48.229],
[3.315,-20.978,-45.301],[4.650,-13.878,-54.071],[11.121,-20.264,-36.978],
[9.147,-21.983,-37.930],[10.503,-8.338,-44.553],[8.390,-28.670,-49.832],
[8.170,-11.683,-42.154],[15.344,-27.876,-52.629],[-0.044,-21.920,-55.025],
[1.703,-24.054,-56.048],[7.285,-20.708,-39.638],[9.312,-21.091,-35.254],
[13.434,-24.506,-52.903],[11.423,-10.341,-47.847],[4.728,-16.893,-56.993],
[8.886,-3.682,-58.797],[13.509,-8.437,-50.748],[1.357,-9.235,-53.964],
[3.395,-10.774,-49.839],[3.163,-2.630,-54.493],[1.642,4.234,-52.337],
[11.984,-15.735,-56.599],[5.269,4.111,-53.767],[5.747,-8.384,-46.288],
[5.291,-5.202,-50.584],[7.665,-9.264,-51.230],[4.494,-7.556,-53.836],
[14.704,-10.204,59.058],[13.580,1.919,-50.595],[5.491,-6.336,-48.107],
[1.452,-13.296,-56.533],[-2.638,-11.486,-50.167],[4.300,2.655,-49.445],
[4.287,-6.303,-42.740],[-2.797,-9.134,-52.296],[-0.605,3.333,-56.034],
[5.628,6.661,-52.813],[-0.295,5.156,-57.914],[11.001,-12.169,-53.606],
[-4.589,-10.848,-54.171],[1.195,7.325,-58.904],[8.235,11.388,-57.081],
[8.487,15.706,-58.252],[2.047,12.657,-55.188],[11.893,15.018,-59.719],
[12.981,8.394,-53.789],[0.590,8.450,-52.658],[6.550,-58.162,-1.180],
[1.119,-54.525,0.811],[8.766,-58.644,-7.726],[1.933,-59.003,-1.081],
[5.124,-52.665,0.803],[9.661,-58.776,-1.424],[5.121,-57.301,-9.968],
[-3.214,-58.735,5.601],[1.162,-49.927,-4.307],[-4.547,-47.445,-14.630],
[-0.621,-39.437,-17.183],[6.413,-51.873,-14.860],[7.517,-46.855,-13.046],
[5.840,-41.163,-4.161],[-1.038,-47.438,-11.502],[2.318,-40.687,-15.072],
[1.195,-52.286,-14.996],[-2.051,-50.758,-17.131],[9.657,-37.444,-7.996],
[3.480,-45.324,-17.194],[1.081,-40.937,-6.692],[6.529,-38.518,-3.820],
[12.353,-46.733,-18.089],[8.505,-44.559,-1.347],[13.183,-35.279,-13.155],
[13.150,-44.843,-12.883],[-2.054,-36.187,-7.884],[0.212,-32.533,-13.754],
[14.841,-28.963,-15.911],[0.237,-31.118,-15.816],[1.609,-22.528,-19.388],
[-1.406,-27.175,-27.039],[1.341,-27.410,-23.594],[6.812,-19.841,-22.238],
[-0.125,-20.146,-17.687],[1.527,-28.590,-15.919],[6.481,-24.833,-13.705],
[2.363,-31.811,-22.804],[11.221,-25.378,-19.353],[-1.648,-25.696,-20.683],
[8.369,-27.148,-28.061],[8.881,-26.557,-17.727],[-3.585,-28.216,-13.570],
[5.775,-27.525,-14.768],[5.235,-34.963,-20.566],[6.358,-24.498,-32.978],
[14.762,-30.397,-27.156],[14.165,-25.004,-13.439],[13.073,-30.019,-13.571],
[-1.330,-36.032,-22.802],[0.634,-32.259,-11.276],[-3.940,-14.036,-31.967],
[5.412,-13.592,-35.148],[0.149,-13.159,-34.075],[8.736,-15.277,-24.797],
[12.094,0.746,-27.577],[13.623,-18.107,-37.677],[8.712,-12.626,-38.494],
[2.364,-12.889,-41.053],[5.253,-10.699,-20.800],[-3.562,-21.492,-30.813],
[10.257,8.106,-41.817],[-3.133,4.923,-40.641],[5.852,7.767,-39.952],
[2.655,7.580,-36.037],[-0.417,2.624,-37.228],[7.925,3.202,-42.717],
[0.737,6.740,-40.040],[-1.712,8.464,-34.970],[0.246,14.692,-37.651],
[13.295,4.954,-50.840],[8.574,9.395,-50.538],[1.879,-4.309,-44.904],
[6.356,4.790,-38.034],[6.421,10.713,-36.652],[8.373,-1.722,-41.427],
[4.746,7.458,-49.813],[-2.585,9.265,-31.760],[7.167,11.672,-31.788],
[13.620,7.466,-32.958],[11.960,-2.322,-39.865],[-0.181,14.096,-41.792],
[5.310,25.753,-46.764],[9.875,15.200,-50.137],[-2.712,19.762,-48.329],
[13.386,17.404,-43.517],[10.956,16.503,-42.752],[9.091,27.479,-48.024],
[8.873,24.986,58.194],[12.227,22.457,-40.390],[0.180,14.443,-44.837],
[-0.387,21.598,-44.274],[-0.037,15.573,-53.743],[5.407,23.147,-55.754],
[11.769,22.283,-57.001],[1.532,26.888,-52.963],[-1.763,24.943,-54.996],
[2.347,32.149,-43.612],[2.386,15.861,-54.878],[6.061,20.868,-38.500],
[-3.692,26.893,-53.902],[14.874,30.205,-51.144],[13.861,20.797,-52.874],
[13.890,12.848,-52.957],[4.618,31.591,-52.153],[-1.138,27.153,-53.214],
[5.973,24.808,-58.968],[-0.317,17.665,-37.899],[2.051,32.354,-55.842],
[-1.210,44.877,-53.346],[11.627,25.600,58.234],[-1.298,42.634,-56.694],
[8.934,30.154,-54.109],[-0.187,34.641,-46.900],[8.238,-50.025,10.419],
[13.324,-49.079,4.456],[5.189,-42.625,6.196],[-2.087,-42.302,15.552],
[13.730,-44.881,4.338],[8.872,-58.467,10.150],[1.203,-57.101,14.274],
[1.164,-52.888,3.385],[6.435,-51.890,3.172],[6.728,-47.909,23.782],
[12.640,-49.343,18.289],[12.279,-57.777,13.418],[13.465,-50.419,24.554],
[7.768,-40.364,14.468],[9.062,-38.373,10.992],[-2.658,-41.406,10.907],
[9.152,-58.391,17.561],[7.156,-47.369,3.275],[2.150,-28.858,1.133],
[-0.227,-33.347,0.600],[-5.339,-28.257,3.007],[6.669,-38.906,12.326],
[0.531,-40.194,0.834],[5.014,-35.040,-1.398],[3.641,-30.015,10.795],
[4.177,-33.741,1.353],[-3.288,-26.587,4.883],[9.126,-41.018,2.868],
[1.824,-37.688,14.208],[12.354,-25.287,5.412],[12.801,-31.072,3.795],
[11.815,-42.802,-0.476],[12.867,-39.722,2.310],[3.911,-17.628,-4.695],
[0.218,-20.339,0.817],[0.809,-12.317,-3.247],[9.507,-20.023,-7.759],
[-1.348,-15.843,-0.367],[13.287,-16.319,-14.027],[5.397,-9.617,-2.349],
[5.343,-12.398,-12.976],[1.084,-25.348,-0.171],[7.517,-20.318,-4.004],
[2.702,-22.508,-7.814],[5.703,-14.262,-0.598],[6.714,-17.546,-11.482],
[-2.539,-21.456,-10.618],[2.198,-6.959,0.374],[2.782,-11.524,1.013],
[7.069,-17.035,3.837],[13.378,-21.735,-10.314],[11.831,-20.630,-6.358],
[8.440,-9.900,-6.477],[9.055,-27.481,-0.132],[4.388,-13.024,-16.080],
[13.953,-13.825,-13.365],[7.417,-21.529,7.662],[3.206,-9.205,-8.277],
[5.120,2.996,-19.261],[11.134,-12.697,-13.986],[0.435,-2.002,-18.089],
[1.924,-12.124,-20.557],[9.038,-7.710,-18.997],[-0.868,0.886,-14.171],
[5.638,-7.429,-14.354],[2.430,-10.621,-16.891],[5.149,-1.769,-24.974],
[6.278,-3.967,-23.338],[2.337,-4.188,-8.414],[6.688,-2.290,-6.037],
[13.396,7.274,-13.646],[14.133,2.140,-13.135],[-3.997,-9.789,-14.445],
[8.625,9.528,-13.572],[-1.569,8.742,-14.105],[-2.242,5.792,-17.888],
[5.143,-8.330,-10.128],[-2.209,-9.464,-12.011],[6.445,14.283,-22.573],
[10.714,17.585,-24.932],[-1.398,11.416,-21.006],[1.367,19.403,-21.749],
[1.835,13.639,-24.678],[5.600,9.104,-14.055],[6.735,5.219,-26.192],
[4.451,17.800,-17.441],[1.861,16.656,-18.433],[13.150,11.040,-31.718],
[12.497,11.494,-11.672],[13.962,13.099,-14.739],[-3.584,10.768,-15.264],
[12.567,10.122,-14.305],[7.313,6.493,-20.896],[-4.641,28.404,-32.597],
[-0.764,28.239,-36.831],[-4.030,34.855,-30.259],[5.371,24.186,-38.090],
[0.955,26.615,-34.124],[8.037,27.392,-33.160],[-5.440,25.965,-32.992],
[3.534,26.727,-31.001],[-1.940,27.129,-34.483],[-0.854,23.879,-33.578],
[3.191,28.225,-28.112],[6.695,37.123,-30.614],[13.961,32.083,-41.564],
[1.744,33.209,-37.796],[6.621,33.247,-40.628],[10.611,28.853,-24.666],
[0.627,29.271,-43.459],[0.975,33.550,-24.922],[12.964,40.214,-31.815],
[13.792,27.319,-40.929],[12.227,24.019,-38.228],[-1.555,39.323,-27.842],
[12.166,22.876,-30.850],[0.269,20.937,-37.486],[1.062,41.777,-41.324],
[8.128,48.087,-37.597],[2.746,52.256,-38.407],[13.195,46.832,-51.080],
[-3.124,53.188,-44.514],[0.696,39.623,-37.442],[5.295,41.673,-33.077],
[7.100,43.361,-45.280],[7.090,53.826,-39.253],[14.291,52.699,-51.950],
[-0.284,47.670,-49.036],[11.531,45.439,-33.748],[11.093,45.895,-42.180],
[10.338,53.580,-45.643],[6.078,45.278,-50.841],[0.426,50.551,-36.529],
[13.221,43.379,-51.063],[8.435,46.032,-30.535],[-2.844,44.518,-31.529],
[-2.094,45.840,-50.848],[-5.451,52.988,-49.926],[13.864,49.397,-33.581],
[-2.775,47.367,-32.393],[13.280,45.229,-31.784],[3.798,54.381,-43.018],
[1.531,-41.677,29.580],[-3.817,-38.860,24.796],[6.906,-44.943,23.076],
[8.019,-39.955,36.521],[6.281,-43.981,27.741],[6.543,-38.334,33.953],
[7.903,-41.757,24.065],[3.755,-41.665,19.930],[5.597,-48.066,28.732],
[12.057,-42.844,36.764],[-2.876,-47.309,26.137],[15.191,-44.303,20.163],
[2.651,-41.190,38.104],[-4.107,-34.089,22.452],[11.134,-32.280,36.107],
[13.908,-44.749,22.816],[10.659,-47.036,27.803],[11.319,-40.773,35.129],
[12.993,-42.965,42.282],[0.430,-21.565,19.349],[5.846,-26.231,25.826],
[10.631,-28.181,27.580],[1.344,-23.750,17.809],[13.112,-16.943,23.370],
[6.095,-18.791,13.365],[10.702,-29.706,13.774],[0.578,-27.687,19.919],
[8.443,-27.769,25.759],[11.585,-12.192,22.641],[5.457,-27.323,23.301],
[6.083,-32.326,18.010],[4.949,-31.938,20.517],[0.086,-23.077,12.067],
[5.034,-23.810,29.048],[-1.157,-25.869,29.066],[-2.390,-27.898,22.891],
[-4.938,-27.307,23.866],[0.032,-15.391,18.814],[5.649,-16.088,18.136],
[-2.796,-23.490,11.584],[13.245,-10.618,15.806],[10.759,-5.352,6.940],
[6.999,-13.195,10.062],[1.738,-4.315,4.583],[-1.648,-9.084,5.134],
[9.847,-8.164,0.496],[11.912,-12.226,2.412],[1.201,-10.170,19.096],
[4.096,-8.916,19.173],[-2.779,-9.154,2.312],[-3.320,-2.941,5.617],
[13.179,-1.693,16.339],[13.273,-12.482,5.202],[-5.763,-7.560,19.047],
[14.445,-10.368,23.315],[13.530,-8.717,20.559],[-1.849,1.247,14.627],
[3.391,-1.904,12.281],[-2.128,-11.551,20.572],[-1.934,10.527,4.093],
[5.344,10.507,1.454],[11.744,4.135,9.652],[14.796,10.912,5.355],
[-4.210,1.727,2.445],[5.196,12.304,6.640],[13.245,11.593,-4.308],
[4.769,9.446,-9.219],[1.494,6.641,-3.395],[-3.785,8.434,2.305],
[-5.125,6.604,4.780],[7.221,24.441,-7.209],[7.107,30.248,-12.596],
[-3.226,29.135,-8.527],[-0.499,29.244,-12.317],[12.633,36.189,-5.360],
[10.569,26.575,-4.535],[8.084,15.650,-10.781],[5.193,23.472,-13.611],
[3.873,30.497,-7.127],[-0.539,31.403,-6.700],[4.850,23.065,-16.224],
[-0.859,24.464,-16.069],[-2.422,27.368,-12.909],[4.795,24.044,2.289],
[-4.863,28.244,-14.433],[12.155,31.446,-2.110],[14.215,20.122,5.073],
[11.880,16.024,-1.657],[-3.194,26.708,-16.063],[5.010,41.439,-18.878],
[4.549,41.945,-16.255],[9.248,34.159,-14.425],[12.876,42.044,-14.105],
[14.073,47.892,-13.241],[4.596,46.456,-13.475],[-1.667,45.627,-13.860],
[1.890,34.669,-17.647],[3.509,37.060,-11.089],[7.853,34.173,-19.784],
[-0.294,41.156,-15.450],[9.660,47.297,-13.415],[3.501,31.993,-17.875],
[0.724,31.632,-18.527],[-0.470,39.383,-25.480],[7.833,43.372,-7.414],
[0.436,39.414,-7.435],[2.869,35.432,-6.054],[-2.633,35.105,-11.608],
[10.041,45.353,-23.104],[5.662,48.714,-15.919],[0.209,49.775,-19.972],
[-0.852,30.663,-21.891],[11.835,49.498,-13.580],[10.898,59.111,-26.823],
[12.500,-59.625,-33.618],[10.448,54.527,-25.843],[3.318,54.022,-19.328],
[8.352,57.575,-17.226],[-1.854,58.324,-16.170],[-0.698,58.766,-37.072],
[-1.552,-33.618,46.511],[5.670,-27.326,41.557],[2.800,-23.361,50.121],
[8.481,-28.332,38.464],[4.369,-29.085,53.918],[-2.343,-36.737,45.219],
[15.337,-39.020,43.171],[-3.417,-27.681,40.999],[5.592,-30.675,34.594],
[4.046,-33.907,-59.178],[14.039,-32.220,44.007],[14.405,-34.306,40.277],
[11.862,-35.252,54.096],[-2.194,-23.177,45.908],[12.065,-20.257,42.450],
[5.287,-37.098,48.270],[10.227,-39.758,48.337],[7.082,-39.188,49.463],
[-0.401,-11.416,39.341],[9.415,-10.761,31.960],[-5.409,-11.446,34.067],
[15.390,-6.446,40.492],[4.880,-8.038,38.848],[6.198,-20.491,45.152],
[-1.566,-10.232,41.699],[0.215,-22.537,37.850],[13.992,-8.995,40.619],
[0.985,-15.320,42.222],[12.698,-12.140,43.454],[12.599,-20.477,34.545],
[-4.964,-18.292,39.338],[9.239,-20.551,40.329],[-3.026,-2.338,23.951],
[0.747,2.436,28.051],[3.052,6.382,30.470],[6.545,4.887,23.078],
[1.172,8.065,20.324],[4.773,-8.454,23.047],[0.396,1.486,30.465],
[13.169,-5.218,23.859],[3.703,2.972,15.921],[1.590,-4.197,38.628],
[2.501,-1.560,38.459],[12.580,1.245,22.497],[12.698,2.302,32.808],
[14.172,-7.479,24.408],[9.175,11.858,31.321],[-1.283,9.219,25.491],
[-2.064,-9.507,24.973],[13.292,5.106,17.516],[8.495,-2.599,40.451],
[12.040,11.641,31.083],[6.874,15.465,16.310],[14.483,13.108,22.491],
[0.018,12.541,19.475],[9.926,10.968,17.490],[2.875,24.546,19.670],
[2.887,8.814,13.846],[5.229,20.344,22.406],[3.217,20.005,25.502],
[7.508,15.324,9.912],[8.902,17.706,10.698],[4.946,16.981,7.071],
[2.319,14.579,12.740],[10.862,7.019,13.642],[-4.438,8.601,16.156],
[4.294,26.013,15.662],[7.849,26.293,19.551],[0.807,32.311,6.023],
[10.630,33.983,17.407],[6.685,29.605,6.373],[0.967,26.995,2.274],
[-1.626,26.375,10.632],[3.485,31.481,12.679],[8.454,27.741,5.794],
[13.783,27.838,4.839],[8.632,36.249,10.248],[-1.560,29.802,15.997],
[3.710,33.184,2.054],[5.101,31.229,16.020],[-1.520,26.042,3.228],
[-2.886,38.629,16.059],[1.306,32.473,-0.512],[2.800,32.926,17.222],
[15.132,28.868,2.879],[12.138,26.603,3.248],[14.359,29.296,20.326],
[13.628,23.099,6.224],[3.305,39.502,6.761],[14.945,45.174,4.656],
[15.118,48.423,5.101],[2.360,51.925,0.169],[0.467,44.717,2.496],
[8.760,52.975,-1.268],[9.530,45.840,-5.743],[13.878,57.403,2.514],
[7.336,52.912,-10.151],[4.845,48.036,2.227],[2.272,54.801,-7.273],
[5.642,48.823,5.174],[9.997,47.783,-0.497],[6.740,48.189,-9.124],
[-2.375,52.549,6.094],[3.904,43.927,-7.480],[1.646,49.354,8.437],
[11.381,40.521,-2.015],[-1.611,42.948,0.773],[-3.097,46.367,2.307],
[5.633,56.898,-2.695],[4.959,42.711,2.012],[4.745,-17.636,54.556],
[11.177,-26.279,55.324],[5.657,-31.569,57.735],[-2.622,-30.716,57.318],
[1.802,-29.012,55.579],[14.524,-27.130,58.803],[11.357,-20.903,53.953],
[-0.730,-32.805,56.949],[-4.235,-8.841,51.626],[13.565,-6.136,58.885],
[-0.121,-10.768,57.634],[2.586,-3.721,54.623],[5.350,-13.585,57.050],
[-0.440,-6.255,54.864],[-0.026,-2.156,57.528],[-0.217,3.092,55.472],
[3.889,0.861,53.855],[4.872,-0.276,44.526],[-1.042,-8.430,56.324],
[1.728,-12.551,50.766],[4.132,-8.025,56.619],[7.556,-13.170,48.955],
[2.787,-4.577,47.962],[-1.407,-15.348,57.177],[5.455,-5.845,40.654],
[4.312,-0.104,56.835],[7.571,-7.257,42.966],[3.958,-3.244,57.563],
[3.563,3.746,55.385],[-4.522,-2.701,42.593],[-1.088,-16.878,49.596],
[12.944,2.934,55.205],[6.175,11.878,44.850],[-1.078,6.094,45.776],
[2.833,14.712,41.796],[0.810,7.103,40.088],[5.254,17.562,45.286],
[3.058,17.556,42.866],[1.504,13.865,36.813],[6.746,14.632,50.790],
[1.988,6.240,49.919],[15.370,8.654,41.656],[14.128,1.267,39.299],
[-0.758,1.446,45.867],[-1.488,22.469,36.950],[1.415,33.644,36.808],
[11.024,31.131,37.589],[0.761,35.921,37.726],[-2.776,30.265,34.573],
[4.547,16.395,30.442],[5.681,28.310,41.724],[4.912,23.713,38.933],
[-3.068,28.584,39.482],[-1.869,22.888,28.230],[9.983,34.698,35.681],
[14.594,30.170,22.800],[3.746,26.615,24.709],[5.615,23.837,41.640],
[10.434,23.958,40.804],[4.829,28.533,22.146],[3.351,28.693,37.770],
[11.542,32.629,35.054],[13.303,28.625,40.922],[12.751,23.366,35.419],
[13.186,24.494,43.798],[-4.414,22.340,41.410],[-0.347,32.553,30.946],
[12.378,18.861,36.094],[5.486,31.311,24.188],[7.981,47.560,24.595],
[-3.624,47.020,23.403],[-0.296,46.734,17.711],[2.171,34.671,29.573],
[6.812,37.875,22.984],[1.500,34.417,19.100],[4.852,42.077,20.467],
[2.377,45.134,19.754],[-1.196,41.103,17.733],[2.338,52.139,28.977],
[2.322,43.596,34.095],[-3.459,47.545,20.652],[12.396,50.005,23.060],
[9.109,48.453,18.535],[1.085,50.771,22.537],[5.644,50.312,22.970],
[0.892,39.390,36.417],[15.671,58.367,21.298],[10.091,56.175,21.597],
[3.191,52.900,19.307],[7.835,55.275,13.267],[2.993,57.474,3.518],
[1.018,50.932,18.297],[8.883,6.537,-58.962],[-2.142,7.488,58.030],
[-0.964,24.902,59.236],[6.982,21.769,54.636],[9.040,16.444,53.383],
[-0.172,18.115,56.708],[7.742,23.449,48.998],[0.953,24.880,52.444],
[1.814,11.687,58.165],[0.866,28.478,57.444],[4.871,28.467,57.696],
[1.695,25.852,55.979],[-1.044,40.152,48.734],[-1.866,33.489,49.502],
[5.883,31.255,45.510],[1.996,36.623,48.431],[0.008,33.038,55.413],
[-0.534,39.825,55.038],[11.198,44.567,50.495],[3.593,40.309,57.228],
[-4.866,40.040,43.765],[10.802,32.797,51.692],[10.885,28.457,49.764],
[-3.688,47.136,40.290],[7.503,-59.051,35.946],[-0.629,48.063,44.079],
[10.742,47.639,46.465],[9.407,52.655,46.744],[1.997,52.621,42.686],
[11.033,53.572,48.844],[-3.428,44.863,41.460],[2.577,45.783,37.413],
[3.406,51.806,38.402],[5.020,44.630,43.309],[2.606,54.255,31.405],
[-0.620,44.488,46.525],[2.539,51.254,49.016],[8.184,53.683,51.050],
[11.622,57.990,41.704],[11.914,44.214,34.880],[11.439,44.207,40.761],
[11.754,47.596,48.916],[0.446,57.566,37.096],[12.480,42.871,44.287],
[-1.736,44.158,39.064],[5.416,54.676,31.613],[1.121,48.964,53.280],
[8.550,48.982,53.008],[28.062,-52.084,-52.561],[25.315,-48.985,-48.236],
[18.601,-51.006,-47.753],[28.171,-46.487,-50.296],[17.794,-46.076,-50.750],
[16.463,-50.153,-54.521],[16.287,-49.707,-51.590],[27.078,-42.919,-50.193],
[23.669,-46.549,-49.837],[33.104,-55.514,-46.046],[28.585,-54.425,-47.505],
[21.187,-54.130,-50.624],[28.761,-55.096,-52.596],[16.798,-52.649,-51.438],
[23.453,-51.077,-51.578],[34.195,-53.538,-44.717],[19.064,-56.084,-47.613],
[25.229,-45.937,-52.640],[21.669,-47.500,-48.130],[31.901,-55.055,-51.192],
[19.020,-53.759,-49.047],[19.425,-47.130,-46.316],[17.681,-45.682,-54.563],
[30.419,-45.461,-48.460],[29.813,-56.799,-46.668],[23.277,-59.378,-49.941],
[29.434,-44.505,-51.592],[27.822,-45.327,-53.723],[30.316,-47.839,-45.989],
[23.426,-47.219,-55.054],[25.715,-51.229,-58.041],[32.694,-52.302,-46.511],
[29.356,-49.680,-52.629],[27.918,-51.950,-45.200],[20.452,-50.569,-51.504],
[20.763,-45.050,-50.365],[21.356,-50.218,-48.952],[26.624,-44.344,-45.952],
[22.324,-47.075,-45.439],[28.769,-51.253,-56.551],[20.667,-51.816,-43.404],
[25.164,-45.401,-56.118],[20.190,-45.954,-43.398],[32.434,-57.671,-57.212],
[18.898,-51.391,-53.925],[33.109,-45.593,58.455],[15.089,-49.005,-46.612],
[15.986,-46.642,-48.628],[34.477,-45.173,-44.801],[24.441,-45.222,-44.146],
[25.102,-41.080,-54.962],[18.910,-41.258,-54.937],[28.253,-41.974,-57.532],
[27.703,-42.592,-54.623],[32.322,-53.329,-40.878],[17.029,-45.536,-44.167],
[21.514,-57.228,-47.760],[24.727,-58.381,-46.114],[29.153,-53.544,-43.016],
[30.525,-48.908,-58.002],[23.850,-55.983,-57.140],[22.686,-53.971,-48.024],
[21.446,-52.264,-53.229],[25.694,-51.634,-54.621],[26.673,-52.677,-43.006],
[29.295,-53.761,-50.137],[18.860,-53.447,-46.415],[21.353,-54.690,-58.141],
[15.486,-54.904,-51.603],[17.764,-55.197,-56.668],[23.666,-51.179,-47.358],
[24.344,-55.411,-51.376],[17.875,-55.223,-53.596],[21.385,-54.620,-54.250],
[28.887,-56.431,-49.777],[21.609,-52.820,-56.232],[31.094,-39.244,-41.874],
[18.205,-47.116,-42.108],[25.800,-53.196,-49.604],[20.313,-56.190,-52.156],
[25.164,-55.876,-48.152],[24.592,-49.469,-56.349],[31.722,-49.711,-47.244],
[27.258,-53.745,-57.890],[24.838,-48.765,-45.564],[29.419,-53.981,-55.088],
[19.141,-48.283,-48.910],[26.345,-47.540,-54.706],[16.505,-51.267,-44.693],
[23.327,-52.512,-42.654],[26.728,-47.038,-44.559],[33.741,-53.254,-55.650],
[31.492,-46.562,-42.680],[31.165,-45.739,-55.603],[28.927,-49.602,-48.147],
[20.504,-44.397,-53.920],[17.782,-49.730,-58.285],[22.685,-47.048,-52.432],
[23.364,-43.463,-46.902],[22.828,-42.631,-54.028],[28.263,-46.681,-47.477],
[26.602,-47.102,-57.850],[28.022,-42.702,-44.683],[32.331,-43.601,-58.297],
[17.416,-55.209,-44.964],[28.558,-48.666,-55.248],[16.676,-44.041,-47.259],
[22.032,-45.022,-59.530],[21.521,-46.210,-56.914],[19.555,-47.234,-55.368],
[26.655,-54.708,-45.584],[27.062,-55.229,-55.383],[24.145,-43.049,-56.347],
[24.518,-54.492,-54.033],[25.113,-51.269,-45.015],[26.827,-57.104,-53.011],
[21.275,-42.777,-56.247],[21.702,-57.871,-57.318],[30.246,-56.092,-43.829],
[30.337,-52.182,-47.975],[19.079,-44.025,-57.574],[19.308,-52.961,-57.830],
[30.054,-42.766,-49.594],[18.483,-58.565,-48.342],[18.299,-43.358,-52.472],
[17.947,-42.915,-43.005],[22.762,-43.397,-40.660],[25.153,-43.312,-58.854],
[27.605,-45.910,-41.739],[28.756,-44.009,-59.360],[15.574,-48.843,-41.775],
[19.428,-48.273,58.669],[23.035,-48.237,-42.871],[24.441,-46.901,-59.443],
[27.990,-49.240,-45.468],[18.137,-50.575,-42.482],[21.034,-49.915,-55.516],
[23.010,-52.986,-45.281],[22.759,-50.084,-58.799],[27.422,-49.560,-40.809],
[32.793,-47.489,-57.839],[21.425,-55.326,-42.370],[20.988,-54.736,-45.207],
[21.022,-51.258,-46.331],[28.456,-52.625,-40.424],[29.492,-50.112,-43.185],
[20.026,-51.986,-40.700],[22.117,-57.955,-45.234],[24.801,-54.369,-41.268],
[24.349,-55.079,-44.047],[28.876,-56.725,-41.356],[18.404,-56.260,-42.144],
[22.244,-59.326,-42.614],[23.894,-58.586,-39.854],[22.972,-40.509,-57.187],
[15.776,-46.621,-57.992],[16.112,-41.476,-52.773],[17.452,-41.924,-50.139],
[32.969,-42.261,-49.109],[32.387,-43.146,-45.007],[33.884,-45.820,-41.481],
[31.905,-48.331,-54.577],[32.229,-49.542,-51.523],[32.702,-47.590,-48.547],
[29.981,-45.202,-45.726],[32.528,-50.782,-56.222],[32.623,-54.871,-57.179],
[30.969,-52.173,-51.699],[16.421,-49.401,-48.924],[32.183,-50.686,-44.267],
[33.972,-52.092,-42.461],[30.651,-55.363,-58.621],[32.379,-53.790,-53.333],
[31.258,-55.001,-48.238],[30.684,-53.726,-45.297],[17.796,-53.172,-43.055],
[34.754,58.662,-41.117],[30.687,-57.372,-53.818],[16.218,-56.056,-49.332],
[14.093,-50.660,-49.193],[32.554,-57.360,-41.537],[35.171,-57.222,-58.602],
[15.225,-54.726,-54.430],[31.957,-43.337,-51.862],[16.348,-58.118,-47.016],
[18.086,-44.967,-40.332],[20.793,-44.223,-47.312],[25.386,-42.674,-43.529],
[28.470,-40.136,-42.992],[32.704,-57.857,-44.286],[16.401,-41.944,-45.079],
[20.839,-41.772,-46.033],[26.268,-40.196,-48.004],[28.221,-42.545,-47.393],
[29.596,-40.363,-45.745],[18.293,-41.980,-47.385],[20.288,-42.505,-50.234],
[23.365,-42.654,-49.800],[25.637,-58.145,-50.447],[30.740,-58.827,-51.117],
[33.137,-57.793,-52.053],[18.587,-58.370,-51.405],[16.954,-59.417,-56.897],
[23.886,-58.979,-52.720],[28.259,-59.183,-49.775],[29.491,-57.837,-56.401],
[16.284,-43.238,-55.740],[18.656,-57.567,-57.280],[22.696,-56.915,-54.473],
[25.191,-40.279,-59.079],[31.733,-42.925,-54.754],[18.517,-57.964,-44.565],
[19.374,-58.798,-41.090],[20.138,-40.969,-58.657],[26.784,-59.076,-42.986],
[32.070,-43.569,-42.130],[25.978,-30.500,-52.153],[25.678,-30.563,-49.202],
[18.763,-32.729,-47.792],[30.230,-28.944,-51.443],[18.703,-28.175,-48.707],
[14.138,-34.278,-52.593],[14.911,-30.908,-51.530],[26.504,-24.548,-49.456],
[23.783,-28.050,-47.955],[33.880,-38.999,-48.086],[27.932,-35.696,-49.307],
[20.593,-35.998,-51.419],[27.854,-37.141,-52.478],[16.900,-33.265,-50.064],
[22.264,-33.768,-50.479],[33.647,-35.540,-46.155],[19.931,-37.619,-48.006],
[24.924,-27.896,-53.798],[20.719,-30.395,-48.830],[29.995,-36.267,-51.338],
[18.283,-35.923,-49.429],[18.982,-29.002,-45.886],[15.705,-25.995,-56.192],
[28.581,-26.322,-48.657],[30.169,-37.988,-47.292],[23.694,-39.746,-49.828],
[28.003,-25.828,-51.547],[28.826,-27.429,-53.610],[29.920,-27.913,-43.674],
[25.135,-31.251,-56.309],[27.276,-34.464,-58.547],[32.801,-31.699,-45.594],
[28.799,-30.802,-53.127],[26.215,-32.017,-44.071],[20.867,-31.107,-52.318],
[21.095,-26.122,-49.597],[19.867,-33.137,-50.316],[26.214,-26.653,-46.729],
[21.332,-27.849,-45.381],[27.282,-32.884,-55.814],[19.161,-32.940,-43.360],
[25.188,-25.799,-51.649],[18.777,-26.776,-43.699],[30.723,-39.214,-57.850],
[17.706,-31.246,-53.978],[32.555,-25.389,-57.678],[17.103,-30.576,-46.797],
[15.127,-27.448,-49.163],[33.298,-26.474,-44.973],[24.228,-27.530,-45.251],
[25.582,-23.436,-53.446],[18.512,-24.577,-54.263],[27.946,-24.345,-56.864],
[27.822,-24.999,-54.291],[32.153,-32.304,-42.740],[16.542,-28.737,-44.865],
[20.658,-40.404,-48.705],[26.865,-39.425,-45.435],[30.251,-35.338,-45.339],
[28.662,-31.816,-57.991],[24.664,-36.005,-56.721],[20.752,-34.944,-48.243],
[20.366,-33.891,-52.983],[24.879,-33.212,-54.475],[26.733,-34.588,-42.410],
[29.542,-33.442,-51.968],[18.807,-34.758,-46.082],[22.418,-36.683,59.090],
[14.242,-37.728,-51.399],[16.233,-36.829,-56.851],[26.398,-33.620,-47.081],
[22.727,-37.496,-50.949],[16.755,-33.972,-53.486],[22.151,-34.910,-55.681],
[27.213,-36.925,-46.385],[19.665,-35.346,-57.098],[31.508,-22.150,-42.669],
[18.625,-30.248,-43.671],[24.497,-33.142,-49.470],[20.135,-38.802,-51.250],
[23.947,-35.007,-47.173],[21.909,-29.084,-55.373],[30.522,-30.456,-47.361],
[28.853,-38.200,-56.246],[25.195,-30.081,-46.051],[29.136,-34.668,-56.585],
[16.777,-29.760,-49.649],[25.968,-28.371,-56.313],[15.856,-31.744,-44.384],
[23.649,-33.109,-42.755],[27.037,-25.219,-44.333],[34.986,-32.789,-55.840],
[34.744,-28.738,-43.651],[31.276,-26.508,-55.477],[29.267,-31.444,-50.000],
[20.159,-26.537,-53.589],[17.971,-31.367,-57.645],[22.069,-28.302,-52.751],
[23.117,-24.599,-44.847],[21.652,-24.208,-54.359],[27.823,-30.772,-46.203],
[28.060,-27.081,-58.227],[30.347,-25.233,-42.664],[34.312,-23.551,-56.990],
[18.214,-36.830,-44.506],[27.762,-30.269,-55.497],[17.270,-26.958,-46.679],
[24.263,-26.774,-58.805],[21.355,-27.302,-57.580],[18.698,-27.694,-55.620],
[24.455,-37.667,-45.680],[26.385,-37.663,-55.280],[25.129,-25.515,-55.755],
[23.598,-35.306,-52.481],[23.800,-35.358,-44.099],[25.285,-37.441,-52.738],
[22.548,-24.875,-56.776],[20.400,-37.908,-54.949],[31.062,-37.843,-44.242],
[31.808,-32.439,-49.472],[19.104,-25.931,-57.867],[18.085,-33.804,-58.679],
[30.301,-24.383,-47.865],[17.891,-39.521,-48.687],[17.980,-27.087,-52.170],
[16.376,-23.799,-41.809],[20.856,-26.140,-41.507],[25.377,-23.490,-57.576],
[27.725,-27.359,-42.415],[30.408,-24.967,57.827],[16.453,-30.740,-41.939],
[18.740,-28.819,-58.169],[22.279,-30.349,-43.394],[24.367,-29.716,-58.393],
[28.288,-30.075,-43.254],[17.331,-33.475,-41.023],[21.164,-31.853,-55.057],
[21.280,-34.103,-45.591],[21.624,-32.429,-57.979],[25.315,-30.535,-41.967],
[30.504,-29.380,-57.932],[21.281,-35.999,-43.199],[21.047,-37.189,-45.510],
[23.131,-32.198,-45.560],[29.550,-35.116,-41.391],[29.190,-32.491,-42.413],
[20.416,-34.205,-41.226],[22.657,-39.610,-46.389],[23.647,-36.616,-41.152],
[25.880,-37.045,-43.388],[28.526,-37.832,-41.664],[19.104,-37.242,-41.799],
[22.966,-41.426,-43.839],[23.828,-39.158,-42.196],[22.434,-22.187,-56.102],
[16.042,-28.410,-57.880],[15.894,-23.739,-53.867],[15.333,-25.290,-51.127],
[32.467,-24.680,-49.600],[32.304,-23.945,-44.955],[30.788,-27.824,-41.162],
[30.530,-29.293,-55.063],[34.166,-30.117,-54.399],[32.569,-28.951,-46.684],
[29.962,-25.944,-45.617],[34.306,-30.478,-57.611],[31.723,-35.122,-57.984],
[30.425,-33.032,-54.834],[14.609,-32.657,-48.942],[32.363,-29.667,-43.475],
[34.520,-32.758,-40.353],[29.405,-36.820,-58.835],[30.120,-35.970,-54.122],
[31.126,-35.513,-48.203],[29.932,-33.082,-46.760],[14.901,-35.702,-43.158],
[35.621,-39.874,-39.593],[31.565,-39.202,-55.102],[16.088,-37.445,-49.410],
[15.364,-35.009,-48.355],[33.655,-37.979,-42.178],[32.269,-37.371,-59.514],
[16.942,-37.348,-54.001],[28.977,-21.705,-53.907],[15.239,-39.171,-45.491],
[17.920,-25.964,-40.873],[19.948,-25.496,-46.055],[25.108,-23.257,-43.404],
[28.417,-21.510,-42.925],[32.263,-40.457,-46.193],[16.745,-22.457,-44.605],
[18.084,-22.725,-47.244],[25.873,-22.064,-48.007],[27.585,-23.968,-46.925],
[27.749,-20.978,-46.262],[17.012,-25.215,-48.909],[18.667,-24.581,-51.130],
[23.094,-24.139,-50.772],[25.678,-37.473,-49.170],[31.534,-40.403,-50.204],
[32.439,-39.597,-52.541],[17.303,-39.116,-51.978],[16.723,-40.496,-56.351],
[22.749,-40.631,-52.320],[28.956,-39.708,-50.751],[29.265,-40.125,-53.682],
[17.607,-24.002,-56.489],[18.058,-39.382,59.116],[23.083,-38.728,-54.760],
[24.420,-20.985,-57.596],[31.677,-22.587,-54.497],[18.723,-40.004,-41.704],
[21.340,-39.744,-41.161],[18.723,-22.558,-58.543],[28.869,-40.698,-40.375],
[33.163,-26.331,-41.528],[26.579,-12.631,-51.295],[26.818,-12.656,-48.212],
[17.843,-12.491,-47.117],[28.908,-8.996,-52.284],[19.679,-10.162,-49.299],
[15.671,-14.813,-53.427],[17.762,-13.648,-51.312],[26.376,-5.618,-50.277],
[24.512,-11.417,-48.956],[33.269,-21.213,-48.105],[27.394,-16.482,-49.397],
[20.956,-17.598,-50.179],[27.953,-18.945,-51.147],[16.563,-16.407,-51.185],
[23.086,-16.482,-53.080],[33.960,-16.495,-45.488],[19.697,-19.016,-46.905],
[26.595,-10.339,-53.502],[22.582,-10.666,-43.773],[30.856,-18.596,-51.545],
[17.963,-17.042,-48.969],[20.298,-10.870,-46.090],[17.366,-9.495,-54.700],
[30.374,-7.472,-47.851],[29.224,-18.537,-47.463],[23.243,-23.008,-48.363],
[29.336,-6.240,-50.929],[30.845,-5.836,-54.009],[31.316,-11.517,-45.639],
[24.787,-12.050,-53.370],[27.008,-16.319,-58.155],[33.246,-13.405,-46.389],
[29.979,-12.987,-54.640],[27.993,-15.366,-44.683],[20.522,-10.514,-51.901],
[21.887,-7.740,-51.941],[22.525,-13.612,-49.799],[27.707,-7.141,-46.807],
[23.282,-8.085,-43.506],[28.214,-14.468,-56.796],[20.330,-13.959,-46.529],
[26.627,-7.468,-53.795],[19.871,-6.769,-43.281],[30.426,-19.898,-55.688],
[17.774,-13.460,-54.828],[32.992,-8.475,-57.528],[15.461,-10.819,-45.950],
[17.424,-8.406,-50.341],[32.885,-7.208,-43.203],[25.386,-7.425,-45.415],
[25.156,-3.875,-54.593],[19.617,-5.873,-52.069],[28.236,-5.421,-57.204],
[28.314,-5.218,-54.211],[31.224,-15.806,-43.358],[16.736,-9.998,-43.216],
[20.659,-20.766,-48.681],[25.586,-20.511,-43.952],[28.496,-18.965,-44.365],
[30.701,-12.762,-57.648],[26.279,-19.397,-58.623],[21.615,-16.754,-47.594],
[20.754,-14.686,-52.804],[27.116,-13.643,-54.082],[26.306,-14.153,-42.707],
[27.201,-16.047,-52.283],[19.392,-15.754,-44.218],[21.129,-17.771,59.269],
[14.310,-18.091,-52.216],[15.231,-18.529,-55.593],[24.452,-14.455,-46.943],
[23.218,-18.909,-50.350],[18.364,-16.627,-53.406],[21.042,-17.759,-54.281],
[26.731,-19.369,-48.592],[20.971,-16.006,-57.075],[30.848,-2.738,-42.190],
[19.512,-10.847,-43.374],[24.648,-15.066,-51.127],[19.311,-19.973,-53.627],
[24.318,-16.792,-48.982],[23.559,-11.435,-56.777],[29.803,-12.176,-48.061],
[28.810,-18.543,-57.465],[25.821,-9.081,-43.236],[27.310,-16.541,-54.850],
[17.195,-11.467,-49.778],[26.129,-11.151,-56.066],[16.890,-12.739,-44.488],
[22.623,-13.803,-42.301],[27.235,-6.646,-43.429],[33.451,-16.927,-56.354],
[31.714,-9.819,-43.055],[30.647,-7.823,-55.860],[29.311,-12.296,-50.696],
[19.235,-8.355,-52.922],[16.531,-12.577,-58.406],[22.096,-12.209,-53.577],
[23.756,-5.017,-45.442],[21.119,-5.661,-54.217],[26.120,-9.982,-46.216],
[28.903,-8.861,-57.426],[30.148,-5.395,-43.583],[33.609,-4.322,-56.731],
[17.271,-17.737,-46.405],[29.423,-10.187,-55.012],[18.676,-7.994,-46.283],
[21.782,-8.709,-59.063],[22.813,-8.213,-54.566],[18.318,-11.295,-56.837],
[25.093,-17.523,-46.313],[26.879,-20.591,-55.671],[24.884,-7.409,-57.079],
[25.566,-18.328,-53.632],[24.207,-15.657,-44.212],[22.997,-19.684,-53.138],
[21.018,-6.660,-56.945],[20.165,-19.816,-56.678],[31.326,-18.546,-44.831],
[30.016,-16.020,-50.654],[18.874,-8.548,-56.451],[18.405,-15.934,-58.591],
[30.344,-4.693,-47.463],[17.567,-20.599,-49.086],[17.931,-6.230,-54.167],
[16.415,-6.863,-41.448],[20.432,-5.179,-40.892],[22.241,-4.581,-58.317],
[27.858,-7.038,-40.641],[30.436,-6.440,-58.808],[16.320,-10.037,-40.392],
[21.104,-11.586,-58.449],[23.409,-9.968,-40.957],[25.479,-10.114,-58.534],
[28.428,-10.103,-43.381],[17.654,-12.263,-41.458],[20.961,-13.279,-55.677],
[21.854,-16.283,-44.950],[24.081,-14.407,-58.479],[26.004,-11.478,-41.943],
[33.149,-11.102,-58.632],[21.012,-18.630,-43.804],[22.685,-18.967,-46.966],
[22.783,-12.559,-46.215],[26.570,-16.445,-41.477],[29.033,-13.620,-43.093],
[19.818,-14.825,-41.549],[23.708,-21.930,-45.960],[24.158,-18.691,-41.997],
[26.172,-17.432,-43.810],[27.663,-18.867,-41.461],[19.049,-18.649,-42.175],
[21.613,-22.964,-43.037],[23.558,-21.370,-41.660],[22.686,-2.359,-56.197],
[16.598,-9.443,-57.877],[14.491,-3.117,-53.796],[16.816,-6.086,-51.576],
[34.738,-4.637,-47.483],[32.547,-6.170,-46.440],[30.674,-9.735,-40.331],
[32.081,-10.181,-55.118],[32.253,-12.993,-53.360],[31.428,-10.086,-48.002],
[31.514,-8.562,-45.659],[33.039,-12.899,-56.191],[33.650,-15.776,59.016],
[31.250,-15.679,-52.837],[15.057,-12.926,-49.213],[32.040,-12.780,-43.058],
[33.711,-14.748,-42.359],[30.942,-17.934,-58.864],[29.960,-17.619,-54.388],
[31.294,-17.587,-48.837],[31.416,-16.532,-46.461],[16.787,-17.241,-43.288],
[34.315,-21.066,-39.571],[32.474,-19.688,-53.979],[15.890,-18.732,-49.802],
[14.869,-15.644,-48.972],[30.688,-19.452,-41.977],[33.813,-18.658,-58.667],
[16.240,-20.177,-53.513],[31.808,-5.362,-51.481],[15.465,-19.813,-44.994],
[18.625,-8.316,-41.600],[21.268,-6.790,-45.470],[25.300,-3.902,-40.976],
[26.977,-4.140,-43.179],[32.370,-21.182,-45.450],[17.503,-1.240,-43.838],
[21.745,-4.511,-47.700],[25.522,-2.843,-49.091],[25.744,-4.686,-47.110],
[28.389,-3.155,-45.566],[17.909,-4.552,-46.598],[20.896,-7.777,-49.278],
[23.249,-5.650,-49.858],[24.878,-20.861,-51.594],[30.001,-21.802,-50.372],
[32.627,-20.954,-51.499],[17.711,-22.022,-51.872],[15.437,-21.755,-56.731],
[22.063,-21.966,-52.010],[27.325,-21.816,-50.668],[29.822,-21.940,-57.438],
[17.026,-6.500,-57.345],[18.784,-18.946,-58.656],[22.807,-19.320,-56.700],
[24.936,-4.324,-57.239],[32.571,-3.842,-54.194],[16.219,-20.747,-42.567],
[18.969,-21.897,-42.506],[18.747,-4.238,-58.476],[26.732,-21.514,-40.909],
[31.538,-6.086,-40.821],[26.853,7.715,-52.232],[26.142,6.130,-49.973],
[19.187,5.319,-47.115],[29.739,8.655,-52.456],[20.573,9.532,-50.287],
[17.168,3.654,-54.073],[17.234,5.937,-52.245],[27.214,12.842,-50.418],
[25.473,8.421,-48.929],[32.830,-2.231,-47.383],[27.751,1.605,-47.906],
[19.108,0.729,-50.881],[29.160,0.256,-53.376],[16.344,3.562,-51.036],
[21.531,3.641,-53.163],[33.737,1.618,-46.221],[18.950,-1.361,-46.199],
[24.687,8.184,-53.903],[22.177,6.309,-47.019],[29.860,-1.537,-51.445],
[17.845,2.072,-48.946],[18.895,8.659,-45.883],[17.394,9.680,-54.034],
[29.708,8.534,-47.994],[30.270,-0.912,-46.203],[22.852,-2.284,-48.975],
[29.730,11.324,-50.501],[27.804,10.714,-52.413],[30.760,8.600,-43.614],
[25.252,6.021,-56.215],[26.387,2.132,-58.854],[33.490,4.558,-45.649],
[29.049,5.304,-53.752],[28.534,3.778,-46.013],[20.695,6.024,-52.355],
[21.977,8.955,-47.821],[21.275,4.960,-49.802],[27.912,10.740,-48.204],
[21.511,9.726,-45.352],[29.622,4.534,-56.608],[19.686,5.373,-43.815],
[25.032,10.784,-52.444],[21.845,12.742,-44.296],[30.725,-0.462,-56.166],
[19.650,5.080,-54.493],[33.188,10.468,-58.057],[16.903,7.862,-47.451],
[16.260,9.950,-49.170],[34.341,11.632,-44.072],[24.855,10.181,-44.636],
[23.724,13.323,-53.613],[17.779,13.360,-52.296],[29.099,12.070,-57.309],
[26.604,11.229,-55.934],[32.218,1.701,-42.033],[19.287,12.691,-43.798],
[19.792,-1.891,-49.074],[25.985,-1.847,-44.384],[29.066,1.579,-44.821],
[30.662,7.401,-56.998],[24.112,0.340,-58.922],[21.825,2.734,-48.257],
[22.198,0.881,-53.586],[26.896,2.468,-53.298],[25.708,2.933,-41.156],
[28.582,1.787,-50.846],[18.857,2.488,-46.654],[19.786,0.925,-57.713],
[16.182,-1.682,-52.265],[16.580,0.582,-57.233],[25.954,3.905,-45.135],
[22.018,-1.274,-52.158],[17.200,0.948,-52.921],[21.951,1.989,-56.111],
[28.433,-1.238,-48.703],[20.691,3.662,-57.991],[30.455,14.036,-42.023],
[19.347,7.945,-42.905],[24.601,3.818,-49.742],[19.221,-1.226,-53.224],
[23.857,0.918,-47.269],[22.861,7.396,-55.978],[31.036,6.146,-47.206],
[27.461,0.379,-57.259],[25.443,7.022,-44.460],[28.915,1.996,-55.500],
[18.518,7.646,-50.301],[25.519,8.634,-56.928],[16.800,5.328,-44.556],
[22.463,4.390,-43.413],[27.952,9.713,-45.480],[31.627,2.913,-56.930],
[34.721,9.628,-45.554],[33.061,11.577,-55.032],[28.606,5.751,-50.350],
[20.213,11.185,-53.343],[16.107,6.560,-57.487],[22.016,8.428,-53.097],
[24.101,11.040,-47.240],[21.036,13.919,-53.485],[25.197,6.095,-46.855],
[28.335,9.344,-57.007],[29.144,11.571,-43.691],[33.356,13.096,-57.392],
[16.706,0.845,-45.397],[28.887,7.818,-54.914],[18.160,11.979,-46.305],
[21.185,11.390,-59.099],[22.669,9.060,-58.575],[19.086,7.780,-55.837],
[26.443,0.624,-45.815],[27.009,-0.938,-54.906],[23.952,10.461,-55.081],
[25.039,1.591,-55.195],[23.894,2.076,-44.837],[24.692,-0.806,-53.182],
[21.687,12.173,-56.431],[20.165,-2.160,-57.145],[31.195,-0.372,-43.693],
[29.920,3.611,-49.474],[19.404,10.554,-57.187],[18.056,2.689,-58.761],
[30.934,12.739,-48.225],[19.632,-4.566,-49.342],[18.176,10.636,-51.473],
[15.943,13.843,-43.121],[22.570,13.906,-41.752],[23.770,11.630,-57.991],
[28.103,9.806,-41.700],[31.008,12.689,-58.968],[16.971,6.689,-41.692],
[20.884,7.194,-57.738],[22.948,7.664,-43.085],[24.720,9.261,58.295],
[28.199,6.192,-44.987],[17.986,4.200,-41.381],[21.995,4.714,-55.750],
[21.111,3.502,-45.507],[23.286,4.945,-58.469],[26.403,5.917,-42.389],
[33.060,7.276,-59.099],[21.298,0.914,-44.102],[21.781,-0.639,-46.249],
[23.689,4.122,-46.590],[28.382,2.442,-41.198],[31.503,6.019,-43.353],
[20.189,2.192,-42.000],[24.148,-2.527,-46.558],[24.583,0.754,-40.246],
[24.896,0.613,-42.796],[29.281,-0.566,-41.295],[17.575,1.451,-41.995],
[23.586,-2.821,-43.820],[22.264,-1.862,-41.726],[22.142,14.262,-58.539],
[15.463,9.126,-56.788],[16.278,14.451,-54.750],[15.411,11.699,-51.064],
[33.359,14.075,-48.938],[31.726,12.962,-44.207],[31.466,9.513,-40.993],
[33.521,8.224,-56.273],[32.694,4.196,-53.953],[32.797,8.123,-48.655],
[31.117,10.681,-46.005],[32.820,5.472,-56.844],[33.850,1.954,-57.898],
[30.689,3.927,-52.015],[15.983,6.544,-49.746],[32.427,8.158,-45.660],
[33.325,4.277,-42.795],[29.995,1.038,-58.313],[31.670,1.699,-54.440],
[31.910,0.904,-48.598],[31.062,2.928,-46.892],[14.887,0.421,-42.251],
[33.307,-2.559,-40.474],[31.647,-1.110,-53.649],[15.807,-0.637,-49.411],
[15.858,4.073,-48.082],[33.489,-2.358,-43.161],[33.920,-1.604,-57.925],
[15.047,-0.243,-54.934],[29.133,13.416,-52.498],[16.158,-2.242,-46.332],
[17.600,9.322,-41.365],[21.375,11.792,-47.866],[25.270,13.661,-43.551],
[27.496,15.260,-44.603],[31.480,-3.120,-44.829],[18.004,17.085,-45.582],
[21.310,14.159,-46.776],[25.981,12.850,-48.171],[28.666,13.063,-46.512],
[29.373,15.649,-46.642],[17.564,13.810,-48.498],[20.078,14.739,-50.874],
[22.702,13.620,-51.024],[26.665,-0.151,-51.505],[32.119,-3.013,-50.216],
[34.463,-1.859,-50.589],[18.093,-3.568,-51.807],[16.044,-4.481,-58.598],
[23.323,-3.479,-51.331],[27.914,-3.236,-50.423],[29.862,-3.602,-55.681],
[17.774,12.222,-55.671],[17.640,-1.741,-56.137],[25.129,-1.107,-56.791],
[25.821,13.284,-57.418],[31.867,13.618,-53.160],[19.934,-1.505,-43.383],
[19.879,-2.028,-40.212],[19.416,13.369,-58.643],[28.857,-4.659,-41.085],
[31.920,12.570,-40.212],[24.225,24.620,-51.126],[26.258,24.730,-48.807],
[18.532,24.009,-47.877],[30.126,27.622,-51.388],[18.822,27.879,-50.168],
[15.443,22.291,-54.146],[16.637,24.087,-52.473],[27.448,32.004,-51.695],
[24.468,26.515,-48.167],[33.769,16.916,-48.748],[28.641,21.066,-48.705],
[19.192,22.601,-50.558],[27.875,20.069,-53.626],[16.422,21.305,-51.345],
[22.431,22.538,-50.775],[33.328,19.049,-46.295],[18.982,19.059,-47.334],
[27.012,27.007,-53.100],[21.782,26.123,-46.388],[31.560,18.135,-50.668],
[16.880,21.847,-48.488],[18.357,28.578,-45.188],[19.297,27.662,-58.456],
[30.151,27.900,-48.604],[30.011,18.207,-47.897],[23.642,15.187,-48.348],
[28.333,29.867,-53.304],[30.431,27.319,-54.045],[30.296,27.473,-44.257],
[24.666,25.179,-53.643],[28.468,21.071,-58.127],[32.605,22.406,-46.699],
[28.439,24.003,-54.251],[28.812,22.231,-46.092],[20.367,25.128,-52.383],
[21.758,27.871,-50.228],[20.813,24.920,-49.751],[28.033,29.785,-46.991],
[21.739,28.501,-43.418],[26.667,23.139,-56.479],[19.181,23.853,-45.148],
[25.691,29.687,-51.675],[18.288,31.459,-43.889],[32.621,15.583,-54.914],
[19.201,23.698,-54.808],[34.513,28.887,58.841],[15.922,24.655,-47.235],
[15.242,29.632,-47.487],[34.508,29.899,-45.598],[27.329,30.513,-44.173],
[25.341,33.684,-53.880],[18.192,31.392,-53.329],[30.256,31.048,-57.998],
[27.447,32.321,-55.294],[31.715,21.960,-44.086],[17.531,26.024,-45.411],
[20.655,16.715,-48.038],[25.673,15.182,-46.514],[29.136,18.903,-44.075],
[29.562,24.205,-57.690],[25.065,18.568,-57.627],[20.549,20.997,-48.616],
[21.110,22.618,-53.254],[25.869,22.195,-53.618],[24.389,22.199,-39.931],
[27.387,21.070,-51.157],[18.853,21.649,-46.542],[23.363,18.635,59.035],
[16.130,19.335,-52.965],[17.327,19.819,-55.636],[26.036,20.446,-46.203],
[22.747,19.054,-50.882],[18.421,21.593,-53.572],[20.027,18.535,-56.035],
[26.545,17.540,-47.294],[19.636,20.993,-57.323],[30.717,34.921,-42.393],
[20.831,25.478,-43.718],[25.659,22.062,-49.326],[20.179,19.196,-53.230],
[23.485,21.437,-47.287],[23.802,24.354,-56.107],[31.306,24.721,-50.214],
[27.994,18.125,-57.498],[25.724,28.024,-44.333],[30.458,21.799,-56.342],
[17.824,25.053,-50.251],[26.038,26.547,-56.290],[15.631,24.000,-44.167],
[22.415,23.496,-42.071],[27.537,28.398,-42.128],[33.113,19.633,-57.488],
[33.567,27.614,-44.634],[31.709,29.378,-56.729],[30.163,24.269,-47.247],
[20.967,28.987,-54.546],[17.575,23.571,-58.728],[23.528,27.624,-52.653],
[24.786,31.313,-47.276],[21.625,32.414,-52.695],[28.022,26.344,-46.743],
[29.847,26.351,-59.103],[31.352,32.396,-45.751],[32.726,33.029,-57.565],
[16.876,18.800,-43.963],[29.078,27.074,-56.300],[17.502,30.837,-46.536],
[23.918,28.361,-58.750],[23.204,27.091,-55.366],[17.989,25.556,-56.697],
[25.252,16.909,-44.680],[27.056,17.488,-54.940],[24.157,29.537,-55.852],
[23.123,20.830,-53.114],[24.517,21.473,-44.097],[25.282,17.499,-52.793],
[22.049,30.483,-57.545],[22.594,17.407,-57.559],[30.985,17.567,-45.577],
[29.405,22.643,-51.605],[19.580,29.894,-56.808],[17.431,20.725,-59.047],
[30.462,31.515,-48.388],[18.120,17.043,-49.059],[15.134,30.902,-53.714],
[16.021,32.802,-43.711],[21.564,32.777,-39.725],[24.809,31.119,-58.108],
[27.889,28.495,-39.190],[29.142,28.712,-58.442],[15.707,25.429,-41.344],
[19.870,25.385,59.295],[23.625,26.205,-42.872],[25.841,26.954,58.854],
[28.600,25.010,-44.393],[17.700,24.156,-42.627],[21.240,25.163,-57.235],
[21.532,22.468,-45.975],[24.384,21.290,58.911],[26.790,25.797,-42.563],
[31.683,29.051,59.420],[21.783,20.584,-43.975],[21.548,19.166,-46.634],
[24.274,24.240,-46.202],[28.619,21.535,-41.461],[25.269,23.440,-42.336],
[20.121,22.169,-42.303],[22.801,16.513,-46.274],[23.976,18.825,-43.201],
[26.615,18.822,-42.678],[28.428,18.647,-40.737],[19.292,19.374,-42.179],
[23.103,15.060,-44.198],[23.790,16.477,-41.933],[22.615,34.298,-56.489],
[16.923,27.179,59.477],[15.480,32.822,-55.808],[16.099,32.470,-51.310],
[34.142,33.544,-49.052],[34.531,32.535,-45.146],[31.743,27.671,-41.484],
[31.803,25.508,-56.064],[33.234,24.684,-53.974],[33.157,28.133,-47.441],
[31.068,29.611,-46.539],[32.465,24.333,-58.290],[31.531,19.377,58.791],
[30.670,21.089,-53.752],[15.007,25.268,-50.017],[31.725,24.747,-44.799],
[32.989,23.462,-41.775],[30.704,18.245,-57.794],[29.572,18.556,-55.219],
[32.046,20.667,-48.936],[30.986,20.237,-45.960],[15.183,20.243,-42.188],
[32.574,15.582,-41.797],[31.674,18.509,-53.495],[18.419,19.235,-51.082],
[13.986,21.965,-49.290],[34.657,16.954,-42.871],[31.882,15.609,-57.938],
[17.148,17.310,-55.052],[29.404,33.292,-53.315],[15.175,15.400,-45.586],
[17.679,29.729,-41.783],[20.809,30.297,-46.647],[24.968,31.982,-44.468],
[28.396,32.702,-42.888],[32.418,15.469,-44.520],[19.134,33.587,-46.153],
[21.123,33.093,-47.780],[26.325,33.270,-48.878],[27.433,32.324,-45.995],
[29.310,34.118,-47.185],[18.232,30.660,-49.196],[20.909,30.045,-51.480],
[19.992,32.314,-50.451],[25.104,17.300,-49.983],[31.350,15.819,-48.531],
[32.694,15.498,-51.393],[17.894,16.489,-52.163],[16.081,15.956,-57.905],
[23.236,15.772,-53.112],[29.914,16.056,-51.467],[30.053,13.430,-55.266],
[16.957,30.300,-57.433],[18.806,16.216,-56.968],[23.552,17.876,-55.048],
[25.241,33.754,-57.381],[31.467,30.873,-54.139],[18.306,16.574,-42.966],
[20.875,16.205,-41.776],[18.534,32.009,-58.742],[27.695,16.197,-41.974],
[31.869,29.832,-43.366],[26.862,43.248,-51.736],[25.221,42.798,-49.581],
[18.092,42.444,-47.034],[29.115,46.752,-50.954],[18.680,46.376,-49.613],
[15.137,40.220,-53.505],[15.738,42.355,-51.892],[25.271,50.412,-50.637],
[24.708,45.584,-48.002],[32.717,36.270,-48.312],[26.667,39.974,-49.333],
[19.550,38.043,-51.418],[28.260,37.328,-53.324],[16.802,39.657,-50.663],
[23.089,42.151,-51.509],[33.582,39.846,-46.930],[19.988,37.124,-48.441],
[28.176,45.367,-53.920],[20.895,43.034,-46.550],[31.201,37.252,-50.644],
[19.167,39.835,-49.110],[18.826,45.663,-46.774],[16.373,46.436,-55.957],
[29.475,48.687,-47.876],[29.138,36.921,-46.385],[23.555,33.259,-48.937],
[27.675,49.448,-51.473],[28.686,48.232,-53.756],[30.175,44.981,-44.643],
[25.069,43.698,-53.843],[26.337,39.904,-58.507],[35.816,41.679,-46.943],
[28.184,42.546,-54.044],[28.434,43.181,-46.860],[20.471,43.050,-51.426],
[21.146,47.264,-50.741],[22.371,44.352,-49.459],[25.480,48.942,-48.263],
[22.675,46.222,-44.540],[28.403,42.247,-56.751],[19.155,43.407,-43.791],
[25.863,48.770,-54.701],[18.839,50.379,-41.758],[31.731,35.923,-56.131],
[19.430,41.529,-55.704],[32.220,49.665,-57.049],[15.487,45.333,-46.046],
[16.841,49.088,-47.666],[34.531,48.673,-43.371],[24.990,49.339,-45.073],
[24.146,51.108,-55.003],[18.856,50.963,-53.755],[28.033,51.215,-57.322],
[28.115,51.281,-54.576],[31.104,39.968,-43.166],[18.249,47.143,-44.520],
[22.425,36.020,-48.723],[25.973,34.557,-44.262],[28.677,38.527,-43.705],
[30.460,42.525,-58.241],[24.956,36.450,-58.236],[21.651,40.435,-46.907],
[21.163,40.482,-53.084],[25.668,42.352,-56.563],[26.018,41.426,-42.280],
[27.632,40.156,-51.687],[19.111,39.981,-45.894],[22.009,35.967,-58.847],
[14.628,37.825,-52.083],[17.335,39.775,-56.269],[25.793,40.952,-46.974],
[22.229,37.611,-51.515],[17.833,39.155,-53.528],[21.791,39.191,-55.380],
[26.429,37.869,-47.851],[20.013,39.500,-57.640],[29.781,52.681,-42.861],
[19.613,45.810,-41.945],[21.726,39.886,-50.106],[19.016,36.097,-53.803],
[23.920,38.441,-48.773],[22.571,43.606,-59.067],[31.652,44.071,-46.842],
[27.572,35.479,-58.801],[27.052,45.794,-44.600],[28.956,39.305,-58.057],
[17.871,43.680,-50.921],[25.928,46.223,-55.915],[16.964,41.016,-44.254],
[24.034,41.649,-44.320],[27.480,48.115,-45.989],[33.374,39.790,-56.553],
[32.131,47.271,-44.658],[30.964,48.449,-54.920],[27.995,43.716,-49.468],
[22.422,47.446,-53.555],[16.210,43.609,-58.014],[22.146,44.698,-53.295],
[23.040,48.687,-47.477],[21.340,50.493,-53.219],[26.959,46.096,-49.147],
[27.151,47.230,-58.473],[29.106,50.101,-44.085],[34.794,51.544,-57.888],
[17.192,38.345,-44.325],[28.897,45.727,-56.516],[17.925,51.206,-45.157],
[22.550,46.435,-57.508],[21.115,44.678,-55.734],[19.223,44.429,-57.886],
[26.006,37.169,-45.220],[26.684,36.383,-55.295],[24.767,48.197,-57.353],
[24.673,38.749,-54.501],[26.616,39.704,-44.856],[26.308,35.622,-52.119],
[20.166,50.292,-57.209],[20.904,36.597,-55.777],[31.566,37.282,-45.241],
[31.218,40.057,-50.428],[19.768,47.175,-56.705],[17.496,38.959,-58.858],
[29.315,51.783,-47.845],[18.501,34.382,-50.599],[16.908,47.235,-52.557],
[16.115,51.815,-41.915],[21.826,50.056,-41.734],[23.459,48.087,57.427],
[26.942,48.962,-41.986],[29.613,49.042,-57.626],[16.666,44.133,-41.256],
[19.758,44.694,58.417],[22.234,43.494,-44.261],[25.376,44.696,-58.817],
[29.664,46.107,-46.891],[18.102,41.503,-41.748],[22.256,41.924,-56.242],
[21.222,41.186,-44.171],[23.933,38.915,-59.130],[25.461,44.600,-42.084],
[30.366,45.306,-58.736],[23.687,39.151,-45.642],[21.680,37.076,-46.034],
[23.969,43.318,-46.822],[29.543,40.538,-41.013],[28.873,43.230,-42.585],
[20.873,38.704,-42.680],[23.362,34.335,-45.954],[25.391,39.325,-41.009],
[26.370,37.217,-42.355],[28.666,36.381,-41.388],[17.986,38.400,-40.981],
[23.293,34.240,-43.215],[23.191,37.163,-43.179],[22.793,49.630,-58.047],
[17.421,45.951,-59.477],[16.481,52.298,-53.848],[16.299,49.512,-50.964],
[32.099,52.112,-48.723],[35.626,49.947,-45.699],[32.678,46.781,-41.426],
[30.885,44.860,-54.595],[32.311,43.706,-52.416],[31.138,45.554,-49.500],
[32.117,46.979,-47.497],[32.274,42.387,-56.097],[32.854,40.689,-59.318],
[30.804,38.974,-53.016],[16.311,43.865,-48.575],[32.768,44.133,-42.691],
[33.630,41.557,-42.352],[30.012,36.358,-58.235],[30.794,38.670,-55.888],
[31.037,39.010,-47.487],[30.854,41.536,-45.928],[15.402,39.346,-42.628],
[34.758,34.795,-39.847],[31.113,36.029,-53.443],[17.353,36.746,-49.571],
[14.471,40.581,-49.724],[35.616,36.466,-42.246],[33.399,36.310,-58.822],
[16.372,36.706,-54.024],[29.987,50.385,-52.591],[16.079,35.491,-44.898],
[16.792,47.701,-42.195],[20.816,47.629,-46.639],[24.839,51.259,-42.257],
[26.752,52.112,-44.108],[32.603,34.637,-44.865],[17.486,53.445,-47.729],
[21.642,51.153,-46.168],[24.397,52.213,-48.474],[26.852,50.631,-46.938],
[28.957,53.071,-45.332],[18.409,50.973,-49.158],[18.929,48.916,-51.511],
[22.714,49.261,-50.429],[25.206,36.040,-49.659],[31.114,34.071,-51.296],
[34.418,33.949,-52.206],[17.202,34.235,-53.052],[17.991,32.221,-55.947],
[23.439,34.608,-51.292],[28.179,36.388,-49.101],[29.231,33.322,-57.136],
[16.299,49.786,-55.159],[19.773,34.630,-57.623],[23.944,36.297,-55.260],
[25.034,52.383,-58.314],[31.460,51.502,-54.583],[19.378,34.756,-43.291],
[19.743,34.806,-40.437],[17.747,51.142,-58.287],[29.767,33.375,-40.174],
[31.696,49.231,-43.170],[33.552,54.833,-48.340],[27.864,58.476,-47.619],
[21.807,57.294,-52.005],[29.110,56.098,-53.825],[16.140,-59.634,-52.525],
[33.955,57.306,-47.043],[18.878,55.826,-46.221],[30.723,57.350,-51.657],
[19.513,56.182,-49.035],[29.449,56.045,-46.123],[22.892,52.064,-51.126],
[31.730,54.533,-58.726],[29.643,-59.572,-40.529],[20.053,53.736,-46.976],
[24.487,52.914,-45.743],[28.323,57.649,-44.351],[25.940,55.381,58.781],
[20.466,58.494,-47.829],[21.310,-58.536,-52.729],[27.724,57.731,-51.915],
[18.715,58.467,-45.371],[23.245,55.371,-58.151],[14.863,56.898,-52.755],
[16.275,56.530,-56.214],[24.821,57.213,-53.255],[18.505,-58.971,-54.421],
[21.874,58.403,-54.906],[26.760,54.249,-46.547],[22.833,58.512,-57.631],
[25.316,58.225,-48.941],[18.764,53.934,-54.911],[23.057,58.586,-47.294],
[29.417,56.929,-57.894],[28.006,57.932,-55.817],[35.386,57.113,-57.597],
[15.958,56.249,-45.789],[26.608,56.778,-46.118],[28.011,54.633,-56.442],
[25.291,58.336,-55.904],[23.640,58.214,-44.568],[26.952,54.540,-53.088],
[22.009,54.911,-55.006],[30.914,55.054,-44.095],[18.365,-58.025,58.364],
[19.839,52.909,-50.202],[21.227,57.120,-44.415],[22.193,56.102,-46.862],
[27.334,57.834,-41.298],[20.462,57.656,-41.777],[21.980,54.193,-48.745],
[23.713,56.984,-42.035],[25.817,56.990,-43.495],[27.347,53.876,-41.308],
[18.476,55.361,-41.493],[22.152,51.894,-43.377],[23.475,54.342,-41.221],
[30.643,-58.430,-58.770],[32.879,56.993,-59.136],[31.708,57.506,-54.196],
[32.217,57.829,-49.352],[29.591,-58.990,-45.109],[15.631,59.166,-44.635],
[32.112,53.820,-41.235],[31.657,53.963,-52.804],[16.824,56.671,-50.574],
[16.071,-59.568,-49.890],[33.808,53.842,-43.185],[34.359,54.226,-57.956],
[14.915,54.810,-54.959],[15.129,54.637,-47.788],[31.263,50.462,-46.223],
[23.716,55.924,-50.351],[30.266,52.755,-50.538],[31.273,55.345,-49.677],
[16.822,54.529,-52.490],[15.257,52.235,-56.710],[23.918,54.605,-52.888],
[28.818,55.222,-51.038],[30.975,54.626,-55.301],[18.371,54.168,-58.038],
[24.910,54.661,-55.718],[17.285,54.909,-43.825],[19.820,53.382,-43.105],
[30.241,52.478,-39.846],[26.904,-50.214,-34.000],[26.037,-50.607,-31.303],
[18.395,-51.069,-29.499],[28.840,-47.123,-33.740],[19.001,-47.366,-31.981],
[15.319,-52.586,-35.171],[15.787,-50.408,-33.009],[26.989,-42.608,-31.035],
[21.848,-47.595,-31.031],[33.066,-57.205,-30.584],[28.374,-53.779,-30.200],
[19.359,-56.616,-33.868],[29.803,-54.652,-35.385],[15.384,-53.297,-32.347],
[20.413,-51.644,-33.689],[33.461,-54.227,-25.794],[18.775,-55.651,-29.239],
[25.989,-47.605,-34.713],[20.196,-49.741,-27.758],[30.581,-55.827,-31.677],
[17.604,-53.657,-31.040],[20.165,-46.490,-28.637],[17.107,-47.214,-36.892],
[29.040,-46.281,-30.825],[29.445,-57.120,-27.330],[22.774,-58.435,-30.944],
[30.079,-43.562,-30.531],[27.456,-45.236,-34.731],[29.319,-46.171,-25.458],
[25.232,-49.817,-35.967],[26.064,-51.470,-39.148],[32.208,-50.996,-27.616],
[29.814,-50.239,-34.262],[27.432,-52.433,-26.762],[20.786,-48.927,-33.779],
[21.887,-44.856,-31.512],[20.737,-50.180,-31.286],[26.308,-45.417,-29.556],
[22.052,-46.593,-26.577],[28.793,-50.577,-37.223],[17.685,-51.035,-26.624],
[26.824,-42.798,-33.789],[20.170,-43.109,-24.051],[32.664,-56.041,-36.439],
[19.351,-50.214,-37.081],[33.067,-44.047,-39.478],[15.751,-48.198,-27.073],
[15.385,-44.474,-29.725],[35.102,-44.518,-26.057],[23.954,-43.981,-24.454],
[26.255,-41.019,-36.706],[19.605,-41.171,-33.950],[27.673,-42.549,-38.655],
[28.599,-43.192,-36.216],[31.577,-53.104,-23.945],[18.321,-46.190,-25.357],
[17.921,-58.133,-28.752],[25.340,-57.011,-27.248],[28.898,-53.465,-24.975],
[29.911,-49.986,-39.887],[24.049,-55.913,-38.836],[21.187,-53.151,-30.083],
[20.203,-53.691,-35.710],[27.475,-52.439,-35.495],[24.859,-51.771,-22.346],
[27.154,-52.564,-32.511],[19.300,-53.460,-27.627],[21.980,-53.697,-40.313],
[16.229,-55.951,-33.639],[17.036,-54.985,-36.131],[26.102,-52.169,-28.924],
[21.510,-55.983,-31.883],[17.915,-53.678,-33.716],[23.287,-53.145,-37.705],
[27.726,-56.811,-29.906],[19.705,-53.836,-38.706],[31.730,-39.969,-22.146],
[18.362,-48.745,-22.603],[23.717,-51.529,-30.316],[19.802,-58.469,-36.159],
[23.769,-53.776,-28.642],[24.080,-48.811,-38.030],[29.909,-49.038,-29.877],
[26.780,-55.796,-39.498],[25.752,-47.052,-25.547],[29.060,-54.101,-37.846],
[18.284,-50.181,-32.303],[27.314,-46.587,-37.033],[15.254,-50.549,-25.591],
[22.226,-51.824,-23.658],[27.054,-44.860,-26.918],[33.623,-51.828,-37.336],
[33.080,-46.838,-24.709],[31.126,-45.563,-38.276],[28.613,-49.091,-32.098],
[21.176,-46.334,-34.119],[17.149,-49.945,-40.232],[22.590,-47.636,-35.853],
[23.060,-42.950,-27.323],[20.847,-43.532,-35.052],[25.931,-49.701,-27.634],
[28.874,-47.277,-39.559],[29.159,-43.196,-25.720],[32.553,-41.548,-38.673],
[17.997,-55.380,-26.408],[30.130,-46.897,-36.062],[17.588,-44.464,-28.031],
[24.042,-46.372,-41.105],[22.024,-46.291,-38.181],[19.491,-47.417,-35.851],
[26.285,-54.920,-28.977],[26.765,-56.205,-36.867],[25.735,-43.991,-36.759],
[25.289,-54.072,-36.211],[24.934,-51.928,-25.900],[25.923,-56.592,-34.223],
[21.778,-42.890,-37.431],[21.679,-57.332,-38.999],[30.690,-54.820,-26.323],
[29.372,-51.369,-30.794],[18.788,-47.912,-38.759],[16.769,-54.504,-40.775],
[30.575,-41.974,-28.392],[19.057,-57.755,-31.074],[18.085,-43.895,-35.443],
[16.649,-41.559,-24.142],[22.146,-42.116,-22.516],[25.691,-44.435,-40.611],
[28.871,-44.293,-22.759],[29.264,-43.868,-40.406],[15.603,-48.717,-22.884],
[20.525,-46.793,-40.392],[22.316,-47.935,-24.281],[24.705,-48.903,-40.883],
[28.857,-49.748,-26.661],[18.922,-50.824,-24.128],[21.250,-49.100,-38.652],
[22.130,-52.140,-27.278],[22.643,-50.936,-40.273],[25.029,-48.947,-23.406],
[32.951,-48.679,-41.912],[21.885,-55.358,-25.424],[21.546,-56.540,-29.327],
[23.011,-49.566,-27.853],[27.325,-53.173,-22.480],[26.936,-50.243,-24.753],
[19.792,-52.545,-22.016],[21.618,-59.927,-26.870],[23.961,-53.963,-23.811],
[25.972,-56.029,-23.212],[28.762,-55.486,-23.228],[19.120,-56.879,-24.533],
[23.471,-59.365,-23.936],[23.975,-58.378,-20.978],[24.230,-40.346,-39.951],
[15.842,-45.698,-38.688],[16.194,-41.485,-36.068],[15.573,-41.513,-32.174],
[32.825,-42.297,-30.027],[32.833,-43.252,-26.409],[30.588,-47.279,-23.275],
[33.806,-49.093,-36.997],[32.408,-50.692,-35.091],[31.622,-47.582,-28.131],
[30.501,-45.007,-27.590],[32.034,-48.140,-38.886],[31.077,-52.237,-38.838],
[29.967,-52.613,-33.620],[15.237,-48.943,-29.676],[31.848,-49.847,-25.045],
[34.201,-51.445,-22.675],[30.740,-55.586,-39.722],[32.000,-54.143,-33.870],
[31.048,-53.281,-31.001],[29.936,-52.578,-27.875],[16.237,-55.642,-24.388],
[34.998,-58.077,-21.491],[31.847,-57.938,-34.617],[16.339,-56.488,-30.211],
[15.772,-51.503,-30.274],[33.927,-58.872,-23.972],[33.456,-57.856,-38.515],
[16.592,-57.418,-37.419],[29.156,-41.936,-32.907],[15.819,-58.503,-27.308],
[15.847,-45.816,-23.074],[20.352,-44.102,-27.054],[26.421,-40.695,-23.873],
[29.039,-41.049,-23.572],[32.009,-58.653,-26.158],[17.959,-41.332,-26.585],
[20.942,-41.015,-26.732],[25.143,-40.769,-29.084],[26.815,-42.695,-28.271],
[29.340,-40.475,-26.669],[16.516,-41.626,-29.149],[19.404,-44.385,-32.587],
[21.500,-41.818,-31.834],[24.010,-55.598,-30.399],[30.945,-58.022,-29.247],
[32.623,-59.271,-32.509],[18.391,-59.151,-33.438],[16.780,-57.647,-40.480],
[23.739,-58.204,-33.612],[28.093,-57.807,-32.596],[29.260,-57.596,-35.459],
[16.718,-43.187,-38.290],[19.146,-56.340,-39.116],[24.099,-56.705,-36.086],
[24.271,-42.337,-38.423],[30.210,-41.395,-35.309],[15.875,-58.538,-24.438],
[21.598,-57.213,-22.300],[19.530,-42.868,-40.213],[28.405,58.818,-22.852],
[31.365,-44.575,-24.034],[27.285,-28.389,-32.290],[25.971,-30.469,-30.831],
[19.806,-31.408,-29.608],[30.468,-27.704,-31.429],[20.156,-29.293,-31.442],
[16.235,-33.656,-35.342],[16.703,-30.669,-32.956],[27.031,-23.419,-31.563],
[24.943,-27.705,-30.654],[32.394,-37.806,-28.887],[28.101,-35.118,-29.990],
[18.816,-36.817,-34.543],[28.804,-37.728,-34.116],[16.047,-35.193,-32.654],
[23.671,-33.001,-34.806],[33.636,-35.595,-27.595],[18.627,-38.131,-30.080],
[26.302,-28.491,-34.875],[23.049,-30.512,-27.195],[29.551,-36.773,-31.634],
[18.245,-35.467,-31.123],[19.610,-28.452,-27.794],[17.019,-27.164,-36.472],
[29.769,-28.372,-28.850],[30.385,-36.976,-27.442],[23.268,-39.759,-30.731],
[29.332,-25.693,-32.897],[28.422,-26.700,-35.145],[30.485,-28.291,-23.593],
[25.487,-30.287,-36.809],[26.304,-33.300,-40.060],[32.420,-32.889,-29.167],
[29.354,-30.078,-33.817],[28.445,-32.629,-26.685],[21.585,-30.658,-33.919],
[22.216,-27.436,-30.361],[22.958,-32.525,-30.066],[29.682,-25.625,-29.950],
[21.868,-27.047,-26.178],[27.964,-31.799,-37.077],[19.347,-31.488,-26.960],
[25.386,-23.446,-33.644],[20.117,-25.126,-24.807],[31.844,-39.655,-36.496],
[18.354,-32.082,-34.576],[30.965,-29.116,-38.522],[17.269,-29.789,-27.869],
[14.817,-26.380,-30.912],[34.970,-25.384,-25.900],[24.360,-26.615,-25.188],
[25.588,-22.543,-36.413],[19.762,-20.440,-33.502],[28.392,-24.030,-40.306],
[27.687,-24.329,-36.430],[31.067,-34.556,-24.365],[17.958,-26.671,-26.254],
[19.760,-40.571,-29.296],[25.988,-40.235,-26.608],[28.666,-35.327,-25.713],
[28.904,-30.922,-40.195],[26.354,-36.639,-40.266],[20.651,-34.945,-29.442],
[20.877,-34.038,-33.016],[26.297,-34.267,-35.860],[26.642,-35.031,-22.586],
[27.606,-34.290,-32.965],[17.801,-34.483,-26.205],[21.220,-35.722,-39.308],
[15.424,-38.385,-32.928],[16.829,-36.219,-37.116],[26.617,-33.289,-28.510],
[22.143,-38.268,-32.572],[18.797,-34.312,-36.251],[21.629,-35.053,-36.681],
[26.786,-37.194,-28.844],[20.431,-33.303,-38.443],[31.362,-21.388,-21.931],
[17.886,-29.264,-25.406],[25.542,-34.291,-30.865],[21.284,-38.894,-35.280],
[23.285,-36.210,-29.544],[22.989,-29.340,-36.404],[30.147,-30.851,-27.800],
[27.823,-38.301,-38.720],[25.641,-31.223,-26.299],[28.907,-36.097,-37.916],
[18.232,-32.879,-31.960],[24.786,-27.290,-37.542],[16.452,-31.938,-25.494],
[23.825,-32.909,-23.830],[28.407,-27.639,-25.311],[33.487,-34.580,-38.082],
[33.288,-27.121,-24.857],[33.022,-27.350,-36.889],[29.586,-30.319,-30.795],
[21.713,-27.457,-34.429],[16.125,-31.537,-39.276],[23.653,-29.084,-33.264],
[25.116,-23.932,-26.956],[21.059,-24.592,-35.236],[27.132,-28.051,-28.169],
[28.427,-27.880,-39.371],[29.957,-24.858,-25.391],[32.239,-23.139,-38.898],
[17.695,-37.190,-26.081],[28.825,-29.150,-36.559],[17.871,-25.614,-28.745],
[23.430,-26.962,-40.251],[21.691,-27.455,-38.163],[19.482,-30.418,-37.129],
[24.753,-36.348,-26.500],[26.487,-38.181,-36.298],[24.496,-25.685,-35.526],
[22.976,-36.050,-34.082],[26.087,-33.977,-25.712],[25.048,-37.651,-34.011],
[21.731,-24.470,-38.863],[21.609,-38.745,-38.271],[32.581,-36.926,-25.281],
[29.290,-32.887,-31.041],[19.322,-28.268,-38.941],[18.143,-34.881,-38.897],
[30.656,-23.444,-28.202],[17.926,-40.087,-31.613],[18.454,-25.843,-34.642],
[17.184,-23.593,-24.606],[20.532,-24.138,-22.203],[25.625,-25.868,-41.727],
[27.547,-28.250,-21.357],[30.187,-25.418,-38.159],[16.652,-29.894,-22.677],
[21.380,-30.476,-39.471],[23.022,-29.553,-24.215],[25.619,-28.422,-40.380],
[28.044,-30.430,-25.033],[18.958,-31.832,-24.191],[21.627,-31.934,-36.644],
[21.199,-33.118,-27.500],[22.814,-32.594,-40.096],[26.554,-31.846,-23.233],
[31.697,-31.068,-40.387],[21.581,-34.724,-25.319],[21.142,-38.165,-28.476],
[24.087,-33.361,-27.476],[29.323,-35.273,-22.201],[30.306,-31.777,-24.261],
[19.827,-33.747,-22.405],[23.166,-39.645,-27.614],[24.011,-36.115,-23.332],
[26.396,-36.865,-24.608],[28.305,-37.971,-23.196],[18.827,-36.018,-23.620],
[23.520,-40.608,-25.168],[23.162,-39.176,-22.797],[23.030,-23.339,-36.624],
[16.405,-28.311,-40.055],[17.386,-22.922,-34.095],[15.404,-23.524,-31.998],
[32.357,-23.692,-30.351],[32.953,-23.495,-26.432],[32.435,-27.843,-21.862],
[34.003,-30.664,-38.676],[33.567,-31.551,-35.980],[32.261,-29.141,-29.997],
[31.486,-26.147,-27.652],[31.230,-31.824,-37.108],[30.436,-33.327,-39.198],
[30.805,-34.014,-33.479],[16.247,-31.155,-30.320],[32.282,-30.215,-24.887],
[33.693,-32.612,-23.851],[32.053,-36.478,-40.196],[30.512,-36.068,-35.611],
[31.099,-34.953,-30.443],[30.200,-34.277,-27.956],[15.230,-35.910,-25.223],
[34.793,-40.020,-21.263],[32.806,-37.625,-34.705],[16.079,-37.142,-30.566],
[16.681,-32.525,-28.080],[33.870,-39.028,-23.861],[33.444,-38.475,-39.188],
[15.859,-38.430,-36.222],[29.336,-22.885,-32.947],[16.389,-39.290,-27.749],
[17.004,-27.358,-23.719],[20.684,-25.381,-27.802],[26.266,-23.879,-24.374],
[28.674,-22.264,-24.699],[31.907,-39.958,-26.010],[18.244,-22.014,-27.948],
[20.347,-21.158,-26.411],[25.972,-22.002,-29.392],[27.790,-23.891,-28.633],
[29.452,-21.320,-27.092],[18.387,-23.670,-30.144],[19.293,-24.398,-32.679],
[21.949,-24.297,-31.966],[25.882,-39.426,-32.070],[32.604,-39.527,-30.926],
[31.542,-39.957,-33.420],[17.390,-39.592,-34.310],[16.976,-40.620,-38.856],
[24.419,-40.616,-34.168],[28.603,-39.178,-31.124],[29.021,-40.327,-37.476],
[17.212,-24.579,-37.203],[17.694,-37.696,-39.684],[23.564,-39.826,-36.570],
[25.045,-23.786,-40.081],[32.682,-24.619,-35.166],[17.968,-38.861,-23.723],
[20.530,-40.097,-23.914],[18.488,-23.915,-39.593],[27.206,-39.890,-21.511],
[33.124,-24.520,-23.597],[26.957,-11.882,-32.381],[25.199,-12.868,-30.296],
[18.845,-12.825,-28.910],[31.294,-9.417,-32.331],[17.968,-7.950,-30.280],
[17.740,-15.310,-35.681],[15.870,-11.834,-32.084],[28.100,-5.584,-31.808],
[24.817,-9.729,-28.790],[32.365,-18.683,-29.639],[28.221,-16.637,-29.895],
[21.751,-17.290,-32.791],[28.363,-19.010,-33.845],[16.318,-15.326,-32.339],
[23.399,-15.155,-31.165],[34.101,-16.568,-27.257],[19.536,-18.219,-29.191],
[25.208,-9.788,-32.376],[22.915,-10.770,-27.137],[29.511,-18.782,-31.260],
[18.973,-17.202,-32.013],[19.110,-9.461,-28.034],[17.671,-8.863,-35.469],
[30.050,-9.214,-29.673],[29.152,-18.508,-27.011],[23.134,-22.118,-29.929],
[27.817,-8.586,-32.129],[28.451,-9.203,-34.876],[29.078,-9.858,-25.399],
[24.897,-13.022,-37.426],[27.100,-14.381,-39.849],[31.371,-13.488,-27.112],
[29.727,-11.378,-33.643],[29.000,-14.834,-27.885],[19.317,-11.400,-33.695],
[22.794,-8.358,-31.238],[21.055,-11.458,-29.546],[27.462,-7.745,-29.246],
[22.009,-9.324,-24.657],[28.754,-13.995,-37.701],[19.554,-13.472,-26.182],
[26.270,-6.435,-34.432],[21.082,-4.979,-24.893],[31.740,-20.164,-37.328],
[19.832,-13.527,-35.502],[31.276,-8.553,-38.204],[17.262,-11.335,-26.620],
[15.465,-7.231,-29.522],[34.713,-7.059,-25.026],[23.758,-7.157,-26.124],
[25.347,-3.014,-36.044],[18.621,-4.052,-33.075],[27.782,-5.708,-38.175],
[28.278,-4.317,-35.590],[30.816,-14.619,-24.396],[18.609,-8.929,-25.556],
[20.399,-21.682,-31.057],[26.560,-21.742,-26.707],[29.742,-17.302,-24.673],
[30.154,-11.708,-38.213],[24.737,-17.643,-38.517],[21.059,-16.150,-30.079],
[21.131,-14.500,-32.651],[25.925,-14.195,-35.189],[26.403,-16.939,-22.151],
[26.906,-16.498,-34.324],[19.374,-15.934,-27.832],[22.078,-17.054,-40.447],
[15.302,-17.660,-34.019],[17.376,-17.560,-39.981],[26.270,-14.646,-28.129],
[22.439,-20.188,-32.212],[19.344,-18.143,-34.712],[21.576,-17.620,-37.394],
[27.228,-19.043,-29.173],[20.686,-15.092,-38.323],[29.472,-4.139,-21.134],
[19.168,-10.447,-23.176],[26.024,-15.370,-30.846],[20.372,-20.805,-36.313],
[23.159,-15.676,-28.375],[23.965,-10.378,-37.539],[30.658,-12.051,-29.401],
[27.194,-19.150,-38.829],[25.728,-11.084,-26.106],[28.230,-16.740,-37.556],
[18.152,-10.883,-31.024],[26.982,-8.027,-36.868],[16.575,-13.523,-25.226],
[22.914,-14.497,-23.918],[26.225,-8.221,-24.874],[32.095,-15.855,-36.503],
[32.939,-9.943,-24.365],[30.668,-7.642,-34.733],[27.903,-12.470,-29.890],
[21.484,-5.071,-33.461],[17.139,-12.364,-38.128],[21.777,-10.466,-32.465],
[21.960,-5.945,-28.875],[20.352,-4.417,-36.153],[27.270,-10.386,-28.290],
[28.345,-9.241,-38.979],[30.707,-5.676,-25.832],[32.174,-6.016,-38.162],
[17.919,-18.351,-25.381],[28.564,-11.780,-35.964],[17.771,-6.673,-27.676],
[22.435,-7.169,-40.884],[21.614,-8.561,-36.247],[19.578,-11.220,-37.023],
[24.708,-18.257,-27.929],[26.352,-19.791,-36.206],[24.523,-7.857,-38.039],
[23.506,-15.675,-37.072],[25.119,-15.467,-25.671],[24.654,-19.133,-34.239],
[22.275,-6.012,-37.953],[21.771,-20.712,-38.749],[31.797,-17.457,-26.156],
[29.961,-14.593,-30.754],[20.201,-7.294,-39.068],[18.007,-14.824,-39.466],
[29.949,-5.813,-28.864],[18.110,-20.508,-30.437],[18.569,-6.432,-34.074],
[15.673,-6.323,-22.274],[22.572,-5.070,-22.269],[24.991,-6.695,-40.448],
[27.966,-8.976,-22.990],[30.046,-7.810,-42.020],[16.323,-10.637,-23.568],
[20.844,-9.541,-41.021],[23.513,-11.479,-23.430],[25.996,-9.351,-40.204],
[28.149,-12.190,-26.539],[18.317,-13.860,-23.009],[22.323,-12.103,-36.555],
[21.964,-15.429,-26.118],[22.977,-14.325,-39.623],[25.838,-13.264,-23.899],
[32.984,-11.114,-39.979],[22.075,-17.720,-24.913],[21.871,-18.876,-27.556],
[23.738,-13.030,-26.034],[29.271,-16.017,-21.947],[28.721,-12.356,-23.383],
[21.277,-16.001,-22.102],[23.334,-21.488,-27.103],[23.038,-19.765,-23.327],
[26.459,-17.969,-24.984],[28.585,-18.838,-22.909],[18.898,-17.835,-22.869],
[23.348,-22.869,-24.858],[24.121,-21.894,-22.289],[22.508,-3.270,-38.392],
[18.331,-9.376,-38.602],[16.214,-5.096,-34.206],[15.340,-7.021,-32.237],
[33.388,-4.672,-29.884],[32.869,-5.720,-27.387],[30.955,-9.266,-22.481],
[34.618,-11.585,-35.339],[32.324,-12.905,-34.900],[32.632,-10.279,-29.227],
[31.345,-8.018,-26.819],[32.321,-13.459,-38.191],[32.041,-15.843,-39.609],
[30.505,-14.580,-33.713],[15.593,-11.818,-29.309],[30.977,-11.788,-25.008],
[33.704,-14.121,-22.785],[29.776,-17.003,-39.955],[29.709,-17.134,-35.033],
[31.730,-16.603,-31.243],[30.667,-16.732,-28.752],[16.537,-16.340,-23.142],
[34.828,-20.297,-21.383],[31.218,-19.494,-34.922],[16.546,-17.946,-30.476],
[16.533,-14.615,-29.626],[33.106,-19.105,-23.221],[32.094,-18.666,-39.558],
[16.027,-19.408,-37.914],[31.394,-4.875,-35.191],[16.442,-19.709,-27.290],
[18.451,-6.893,-22.644],[20.466,-6.982,-26.630],[24.693,-5.325,-23.970],
[27.626,-5.027,-23.977],[31.334,-21.079,-25.259],[18.207,-4.189,-24.172],
[19.382,-4.271,-27.072],[24.326,-3.628,-29.082],[27.101,-5.170,-28.518],
[29.976,-3.496,-27.399],[18.831,-5.395,-29.923],[20.432,-6.741,-31.516],
[23.374,-5.853,-31.267],[25.013,-19.598,-31.300],[30.680,-20.898,-29.493],
[32.339,-21.112,-33.230],[16.268,-20.085,-35.270],[18.231,-22.148,-37.722],
[23.513,-21.609,-34.250],[27.755,-20.859,-31.237],[30.346,-22.214,-35.794],
[17.155,-6.927,-37.368],[19.804,-18.848,-39.560],[23.746,-19.816,-36.824],
[25.083,-5.049,-38.243],[33.385,-7.897,-36.631],[19.193,-20.471,-23.879],
[21.250,-21.529,-21.703],[18.052,-5.050,-39.175],[28.726,-21.765,-21.977],
[32.203,-6.752,-23.567],[25.756,5.943,-34.575],[24.605,6.219,-29.902],
[17.098,4.414,-28.480],[29.806,9.216,-31.670],[19.161,8.918,-33.097],
[15.046,2.418,-34.760],[16.382,5.741,-33.977],[25.378,12.975,-32.921],
[25.198,8.882,-30.428],[31.406,-0.935,-27.735],[28.086,1.997,-30.427],
[20.663,0.964,-32.630],[30.164,-0.020,-34.819],[16.307,1.432,-32.199],
[23.072,4.889,-33.183],[33.444,1.128,-28.150],[18.618,1.156,-28.212],
[27.825,8.837,-33.268],[20.625,7.837,-29.009],[30.307,1.358,-31.847],
[18.629,2.313,-31.316],[18.358,8.407,-26.864],[18.878,9.961,-36.121],
[28.148,9.220,-27.950],[29.762,-1.213,-29.768],[24.463,-3.572,-31.870],
[28.648,11.634,-32.361],[28.278,10.315,-35.457],[30.858,8.406,-26.432],
[25.501,6.484,-37.094],[25.722,5.101,-39.493],[33.362,5.542,-28.679],
[30.350,6.107,-35.955],[29.375,3.824,-26.598],[20.658,6.759,-34.105],
[22.050,8.951,-31.266],[21.905,6.284,-30.887],[27.131,11.420,-29.947],
[22.303,10.540,-25.357],[28.744,3.753,-38.130],[19.525,5.768,-26.800],
[24.128,11.609,-34.837],[19.956,9.730,-23.819],[32.600,-2.990,-37.554],
[19.379,5.111,-35.789],[34.384,11.698,-39.739],[15.937,8.181,-27.942],
[15.400,10.744,-30.074],[34.294,11.863,-24.085],[24.971,10.622,-26.343],
[26.385,14.990,-37.494],[19.869,14.632,-34.243],[28.351,14.156,-39.707],
[27.536,13.057,-36.167],[30.442,2.944,-24.194],[17.659,10.796,-25.745],
[19.950,-1.780,-28.548],[25.805,-3.884,-26.551],[28.263,0.920,-26.160],
[28.396,5.238,-40.122],[25.719,1.072,-37.863],[19.845,4.666,-29.622],
[21.288,3.094,-34.534],[26.059,3.326,-35.604],[26.066,3.837,-24.757],
[28.431,2.027,-33.696],[16.980,2.213,-26.070],[21.911,0.666,-39.672],
[16.639,0.429,-34.576],[17.547,0.379,-37.403],[26.221,2.759,-27.925],
[23.272,0.010,-32.974],[18.664,2.383,-34.479],[20.479,1.019,-36.655],
[27.231,-0.584,-30.642],[21.327,3.429,-38.291],[30.483,14.741,-24.335],
[19.105,7.176,-24.569],[25.278,3.654,-31.374],[20.374,-1.476,-34.651],
[23.299,2.485,-28.484],[23.188,8.937,-37.928],[30.981,7.857,-29.398],
[28.262,0.379,-38.985],[23.679,8.559,-27.582],[28.020,1.643,-36.537],
[18.499,6.387,-32.342],[26.099,8.735,-35.572],[14.830,6.349,-26.150],
[21.548,4.769,-24.695],[26.797,8.433,-25.712],[34.599,3.473,-37.966],
[33.432,9.051,-25.556],[30.487,11.261,-36.773],[28.928,6.447,-33.203],
[21.889,9.834,-35.094],[17.606,5.597,-38.817],[23.364,7.639,-34.380],
[22.401,12.360,-27.619],[21.420,11.816,-33.372],[27.288,7.159,-29.793],
[29.096,8.507,-37.771],[28.914,12.871,-25.823],[32.555,13.235,-37.486],
[16.560,-0.535,-26.578],[30.674,8.565,-34.524],[17.604,12.010,-29.336],
[22.490,10.298,-42.308],[22.203,11.450,-39.966],[19.119,8.252,-38.399],
[26.198,-0.466,-27.079],[27.341,-1.017,-35.936],[25.238,11.901,-39.698],
[25.127,0.564,-35.319],[23.812,3.500,-25.975],[26.120,-1.792,-33.646],
[22.778,12.006,-37.482],[20.162,-1.687,-37.380],[31.401,-0.369,-23.899],
[29.291,4.278,-31.059],[19.843,11.144,-38.670],[18.816,2.454,-39.591],
[31.530,13.108,-29.380],[17.631,-3.238,-30.456],[17.466,11.410,-33.829],
[17.132,14.220,-22.855],[22.588,12.845,-23.484],[26.806,12.272,-41.823],
[27.484,10.227,-23.628],[29.578,11.839,-40.843],[16.856,6.976,-22.728],
[20.387,9.547,-40.758],[21.822,7.669,-23.744],[25.446,8.968,-41.931],
[28.233,6.241,-27.520],[18.006,4.301,-23.298],[21.743,6.882,-39.392],
[21.318,3.892,-27.263],[22.887,4.233,-40.632],[26.469,7.186,-23.064],
[32.366,6.699,-40.611],[19.358,0.132,-25.416],[21.689,0.171,-27.591],
[22.159,6.686,-26.540],[25.780,1.481,-21.888],[28.211,5.039,-22.966],
[21.415,1.899,-24.326],[22.472,-2.530,-27.348],[23.005,-0.232,-23.164],
[26.116,0.966,-24.480],[28.695,0.658,-23.314],[18.355,0.849,-22.846],
[22.559,-2.764,-24.382],[24.232,-2.642,-21.624],[23.610,14.345,-38.760],
[16.898,10.800,-39.204],[15.427,15.626,-33.885],[16.113,13.446,-31.842],
[32.759,15.127,-30.705],[34.081,13.705,-26.330],[30.489,9.792,-22.671],
[32.843,7.484,-36.271],[32.142,4.090,-35.773],[31.549,10.571,-30.239],
[30.188,10.970,-27.464],[32.742,5.681,-38.073],[31.173,4.080,-39.714],
[30.267,4.042,-33.746],[16.071,6.515,-31.096],[31.134,6.253,-24.529],
[35.009,4.842,-23.059],[30.680,1.237,-39.546],[31.167,1.439,-37.104],
[31.432,2.898,-29.786],[30.967,5.840,-27.441],[14.410,1.667,-24.915],
[33.798,-2.137,-21.549],[32.797,-0.618,-35.907],[17.981,-0.670,-30.575],
[15.196,2.654,-30.188],[31.962,-4.004,-22.762],[33.619,0.914,-38.823],
[15.933,-1.645,-36.111],[30.509,13.872,-35.694],[14.911,-2.648,-27.955],
[18.078,11.656,-23.139],[19.416,10.199,-28.603],[24.531,13.301,-25.729],
[26.443,15.111,-25.760],[29.905,-3.012,-24.534],[17.372,15.761,-27.072],
[18.797,13.271,-27.406],[23.465,13.477,-30.218],[26.711,12.641,-27.466],
[27.759,14.943,-28.384],[16.548,14.400,-29.143],[19.386,11.791,-31.782],
[21.508,14.088,-31.987],[24.492,-0.873,-30.561],[32.544,-0.099,-30.601],
[32.582,-2.952,-33.007],[17.650,-1.803,-33.832],[17.008,-3.975,-36.701],
[22.927,-2.648,-33.795],[29.839,-2.664,-32.230],[30.276,-2.469,-36.417],
[18.087,12.451,-36.841],[17.227,-1.065,-39.976],[23.849,-1.076,-37.115],
[25.499,14.625,-40.731],[32.899,12.663,-34.914],[15.779,-1.892,-23.927],
[21.461,-1.586,-21.465],[18.424,14.422,-38.672],[27.517,-2.612,-22.518],
[31.656,11.014,-24.933],[25.394,22.855,-32.761],[25.513,25.192,-30.034],
[18.927,22.667,-28.931],[30.180,27.733,-31.956],[16.859,27.863,-32.156],
[16.432,22.153,-35.401],[15.801,25.267,-31.887],[26.810,31.068,-31.883],
[24.400,27.509,-30.399],[32.228,17.284,-28.868],[28.753,20.929,-30.289],
[19.900,18.512,-32.357],[28.720,17.848,-35.511],[15.887,21.296,-32.365],
[22.815,23.748,-32.227],[34.037,21.023,-26.492],[19.153,19.615,-28.531],
[24.738,27.198,-33.190],[20.607,24.573,-27.441],[31.253,18.331,-33.509],
[18.126,21.153,-30.912],[18.710,26.392,-26.380],[16.617,28.620,-37.434],
[30.025,28.916,-29.538],[30.064,18.149,-30.436],[22.967,16.470,-31.369],
[30.104,30.318,-32.810],[28.870,28.948,-35.009],[30.146,27.364,-25.982],
[25.774,25.417,-35.105],[27.047,22.407,-39.568],[32.282,23.084,-27.676],
[29.114,25.463,-33.194],[29.959,22.302,-26.809],[20.834,24.569,-33.628],
[19.810,28.029,-31.231],[19.541,24.296,-31.037],[27.005,28.614,-28.164],
[21.282,28.906,-26.469],[27.422,24.447,-37.182],[17.680,23.731,-26.779],
[26.291,30.336,-34.659],[19.495,30.786,-25.869],[31.164,16.141,-39.473],
[19.725,23.649,-36.972],[33.041,30.269,-40.473],[14.921,25.364,-26.822],
[15.046,28.872,-30.538],[34.973,29.727,-24.680],[24.048,28.679,-27.805],
[26.257,32.916,-36.479],[22.838,34.818,-33.695],[27.409,31.542,-40.333],
[30.110,31.139,-36.001],[31.683,20.964,-25.115],[19.618,28.374,-24.475],
[20.842,17.252,-29.688],[24.873,15.176,-28.368],[28.904,20.132,-25.446],
[28.963,24.465,-39.666],[26.509,17.114,-39.564],[20.805,21.244,-30.375],
[20.904,21.102,-34.874],[25.705,22.326,-35.643],[24.447,21.894,-21.561],
[27.216,20.802,-32.930],[18.391,20.880,-26.136],[22.588,18.634,-40.690],
[16.841,18.135,-32.935],[17.141,20.233,-40.248],[27.339,22.364,-27.933],
[23.594,19.582,-33.520],[18.303,20.582,-33.842],[21.012,19.718,-37.062],
[27.440,18.482,-30.664],[21.290,20.986,-40.004],[30.655,34.135,-24.388],
[19.154,25.544,-23.992],[23.532,20.574,-31.054],[20.954,16.889,-35.302],
[23.109,19.957,-27.610],[23.736,26.282,-37.058],[31.004,25.903,-28.364],
[29.489,18.368,-38.242],[25.972,25.927,-25.872],[27.944,20.372,-36.522],
[18.710,25.900,-33.060],[26.629,27.580,-37.070],[15.674,22.237,-25.591],
[22.596,22.671,-24.610],[26.614,28.771,-25.321],[33.807,21.761,-37.314],
[32.495,27.633,-24.776],[31.144,30.733,-38.786],[27.873,23.773,-30.136],
[23.331,29.962,-33.823],[18.239,25.791,-40.481],[21.891,27.164,-33.871],
[22.415,31.138,-27.839],[22.601,32.313,-35.040],[28.437,25.247,-27.460],
[30.231,27.038,-38.467],[28.351,31.960,-26.371],[33.337,32.228,-38.716],
[18.231,18.038,-26.162],[28.934,26.411,-35.875],[16.758,30.999,-28.094],
[23.654,28.288,-40.960],[21.820,27.975,-37.078],[17.876,26.076,-37.848],
[25.807,18.169,-28.286],[26.574,18.242,-37.331],[24.324,29.085,-37.588],
[23.135,21.088,-36.413],[25.132,21.897,-25.425],[25.132,16.792,-35.500],
[22.112,30.760,-37.189],[21.781,16.560,-38.768],[30.477,17.735,-26.587],
[29.914,22.833,-32.284],[20.801,29.261,-39.775],[18.376,22.744,-40.528],
[32.347,31.527,-29.165],[17.570,17.392,-29.584],[17.366,29.925,-35.014],
[16.540,32.045,-23.176],[19.919,31.001,-21.819],[25.527,30.017,-41.614],
[26.714,28.058,-22.657],[30.006,30.915,-41.824],[15.348,25.504,-23.089],
[21.478,26.029,-41.163],[23.015,25.435,-23.971],[25.784,26.986,-39.826],
[28.839,25.427,-24.406],[17.624,22.995,-23.776],[20.502,26.255,-38.538],
[20.848,21.898,-27.057],[21.839,23.497,-39.391],[26.515,24.882,-23.047],
[31.383,25.205,-40.566],[21.809,19.177,-24.275],[22.816,17.283,-27.700],
[23.622,24.832,-27.809],[28.769,21.588,-21.923],[29.957,22.499,-24.175],
[20.154,21.548,-23.375],[21.616,15.061,-28.285],[24.423,19.639,-23.553],
[26.266,18.552,-25.430],[27.266,19.327,-23.046],[17.632,20.038,-23.627],
[23.078,15.784,-25.181],[23.215,16.760,-22.341],[23.198,35.371,-39.349],
[18.314,28.333,-39.467],[17.945,32.778,-35.453],[15.697,30.921,-32.872],
[32.322,31.944,-31.911],[31.694,32.094,-26.405],[31.527,28.797,-22.095],
[31.479,26.363,-35.579],[33.776,24.851,-35.048],[33.229,27.136,-29.794],
[31.990,29.365,-27.229],[34.236,23.828,-39.207],[31.663,23.175,-38.702],
[32.476,22.846,-33.354],[16.472,23.537,-30.001],[31.991,24.673,-25.057],
[33.820,23.588,-22.705],[30.808,19.855,-39.842],[30.900,19.943,-36.202],
[31.600,20.723,-31.789],[30.415,19.973,-28.377],[15.169,20.799,-23.287],
[36.475,16.835,-21.962],[31.342,17.152,-36.582],[16.071,19.378,-30.509],
[14.559,21.498,-29.761],[31.517,17.437,-23.719],[33.220,18.293,-40.549],
[16.060,18.527,-38.038],[30.436,32.982,-33.607],[14.919,17.113,-26.169],
[17.051,27.752,-23.912],[19.061,28.509,-28.506],[23.487,30.879,-25.296],
[26.140,32.499,-24.949],[31.638,15.015,-27.384],[17.465,32.535,-26.091],
[18.939,33.076,-28.754],[27.656,33.802,-29.894],[29.801,31.700,-29.756],
[28.910,35.146,-27.517],[21.172,31.362,-30.097],[20.704,29.962,-32.556],
[23.663,30.463,-30.586],[25.648,15.771,-32.468],[30.470,13.846,-32.170],
[32.054,15.632,-33.341],[18.671,15.993,-31.949],[16.656,14.955,-36.758],
[23.470,15.235,-33.936],[28.994,17.202,-32.705],[29.031,15.743,-37.665],
[17.618,30.936,-38.128],[19.743,18.638,-39.714],[23.337,18.262,-37.132],
[25.128,31.603,-38.818],[32.483,32.712,-35.888],[20.482,16.824,-24.862],
[19.840,15.228,-22.578],[19.945,31.641,-41.547],[27.663,14.880,-23.283],
[32.081,31.869,-23.660],[26.039,42.482,-33.733],[25.791,43.989,-31.105],
[16.369,41.375,-29.559],[30.457,46.459,-31.802],[18.786,47.503,-31.560],
[16.277,40.734,-34.696],[15.631,43.443,-33.968],[25.790,49.933,-32.625],
[24.808,45.868,-29.121],[33.612,36.603,-27.640],[29.830,39.484,-29.514],
[20.851,39.323,-32.136],[28.240,37.117,-34.098],[17.489,40.106,-32.250],
[22.625,41.651,-31.610],[33.251,40.899,-26.101],[19.523,38.049,-27.652],
[26.417,46.181,-33.601],[22.151,45.188,-28.205],[30.064,37.574,-31.660],
[20.282,39.753,-29.592],[19.226,45.565,-27.077],[16.266,46.077,-35.461],
[29.658,46.850,-29.328],[30.986,36.758,-29.315],[24.363,33.039,-29.120],
[28.530,49.524,-31.284],[29.845,46.856,-34.981],[31.228,45.520,-25.243],
[25.289,43.802,-36.190],[24.917,39.287,-38.260],[33.606,42.707,-28.112],
[30.459,41.992,-36.499],[28.066,41.835,-27.133],[18.397,43.421,-32.627],
[23.439,47.419,-32.534],[21.904,44.999,-30.885],[25.686,48.462,-29.710],
[21.174,47.582,-24.099],[28.095,42.707,-37.871],[19.095,42.719,-25.224],
[25.590,49.217,-35.213],[19.982,49.634,-25.058],[32.617,35.260,-38.210],
[18.843,44.405,-36.006],[31.378,50.140,-40.668],[17.631,47.095,-28.742],
[15.276,49.139,-29.152],[33.511,48.119,-25.243],[23.780,46.722,-26.227],
[25.704,52.146,-35.969],[18.817,50.663,-34.445],[29.570,49.730,-38.470],
[28.512,49.866,-35.983],[31.421,40.273,-24.333],[17.285,47.090,-25.912],
[21.749,35.053,-30.939],[26.308,34.203,-27.228],[28.817,38.432,-24.948],
[31.292,42.574,-41.227],[25.538,36.815,-39.787],[21.035,42.388,-29.589],
[21.587,40.221,-35.011],[26.294,41.037,-36.155],[25.021,40.686,-24.051],
[27.161,40.084,-33.418],[19.154,40.862,-27.381],[21.295,37.104,-40.489],
[16.156,36.064,-31.960],[16.912,39.031,-38.455],[25.361,41.295,-28.285],
[24.536,37.126,-30.900],[18.502,39.078,-34.561],[20.956,37.548,-36.310],
[27.943,37.413,-29.618],[22.407,39.778,-39.729],[29.668,51.562,-20.879],
[19.750,45.211,-24.301],[25.348,40.854,-31.165],[19.609,36.825,-33.923],
[23.258,39.236,-30.210],[23.570,44.817,-38.273],[30.975,43.968,-28.638],
[28.582,38.417,-38.408],[26.212,45.304,-26.180],[28.794,39.699,-35.776],
[19.093,45.012,-30.780],[27.072,45.719,-36.219],[16.697,42.413,-26.579],
[23.361,42.835,-24.257],[26.964,48.198,-26.040],[33.930,41.242,-37.616],
[35.829,46.619,-26.142],[32.476,48.654,-36.592],[29.264,43.858,-30.591],
[21.139,48.075,-34.042],[16.847,42.789,-38.002],[20.584,44.353,-33.827],
[22.483,50.098,-26.377],[21.454,51.012,-34.065],[28.817,46.093,-26.768],
[27.437,43.970,-40.184],[29.287,49.091,-27.439],[34.064,50.975,-39.711],
[15.743,37.248,-25.549],[29.515,45.383,-37.337],[17.909,49.733,-26.935],
[22.616,47.099,-40.054],[20.382,47.476,-37.818],[17.882,45.981,-38.139],
[25.232,37.909,-27.124],[26.275,36.147,-36.715],[24.748,48.320,-37.525],
[23.983,38.758,-35.785],[24.224,38.078,-24.703],[25.034,36.369,-34.264],
[21.195,50.785,-38.646],[20.423,34.374,-34.946],[31.222,37.080,-25.150],
[29.774,41.411,-31.562],[18.377,48.166,-39.804],[19.596,39.984,-39.234],
[31.536,50.290,-29.337],[18.680,37.634,-31.162],[17.295,48.920,-35.938],
[16.403,51.000,-22.029],[20.344,49.534,-22.221],[24.316,48.429,-42.112],
[27.680,46.734,-23.555],[30.085,47.120,-39.981],[15.215,45.249,-23.860],
[20.818,44.294,-37.957],[22.746,45.348,-23.300],[25.469,46.802,-40.062],
[28.492,44.104,-24.529],[17.448,42.777,-23.183],[21.853,41.648,-37.281],
[21.608,40.765,-25.847],[23.624,42.550,-40.645],[25.840,44.531,-23.446],
[30.668,44.487,-39.388],[21.248,38.132,-24.741],[22.480,38.773,-27.304],
[23.780,43.120,-26.891],[29.175,39.785,-22.214],[28.253,41.171,-24.128],
[19.678,40.138,-23.228],[22.153,34.301,-27.414],[22.477,39.975,-22.514],
[25.443,39.053,-21.846],[26.738,36.765,-24.461],[17.783,37.851,-23.280],
[24.007,34.307,-24.230],[23.149,33.901,-21.604],[23.336,52.226,-38.418],
[15.105,45.947,-39.408],[15.060,52.096,-34.768],[17.033,50.817,-31.876],
[33.812,51.973,-30.812],[33.751,51.903,-27.092],[32.721,46.193,-23.199],
[32.219,45.795,-36.027],[32.912,43.162,-35.958],[33.307,45.784,-28.226],
[31.467,47.590,-27.277],[33.034,44.747,-39.846],[32.238,39.876,-40.416],
[32.437,41.462,-32.791],[16.234,43.911,-31.005],[31.149,42.784,-25.307],
[33.015,41.627,-22.654],[31.168,37.031,-40.106],[31.200,38.574,-36.925],
[32.453,38.960,-28.769],[30.550,40.545,-27.257],[15.129,39.084,-22.913],
[34.245,31.618,-22.334],[32.882,36.081,-35.399],[16.168,38.315,-30.472],
[13.118,43.020,-33.290],[33.558,35.276,-23.120],[33.946,37.567,-40.421],
[17.100,35.495,-37.056],[29.771,51.788,-34.708],[18.449,30.546,-30.299],
[17.924,48.246,-22.369],[21.316,47.752,-27.064],[24.936,50.131,-24.764],
[29.689,50.440,-25.022],[31.420,34.971,-26.706],[17.647,51.694,-25.329],
[20.212,51.705,-27.455],[26.173,52.199,-29.733],[27.326,50.929,-27.628],
[30.167,51.768,-27.433],[18.032,49.869,-29.533],[20.882,49.557,-29.682],
[22.607,51.078,-31.594],[25.761,34.877,-31.865],[30.864,34.251,-30.199],
[32.017,34.878,-32.562],[16.785,35.025,-34.487],[17.090,33.339,-38.846],
[25.264,32.942,-33.617],[29.051,35.009,-32.090],[30.127,35.881,-36.502],
[16.058,50.224,-38.760],[18.946,36.919,-38.855],[23.416,35.760,-36.480],
[25.574,50.848,-39.256],[33.952,51.446,-36.705],[17.197,34.961,-24.582],
[19.766,35.114,-23.135],[18.596,50.438,-38.214],[28.771,33.787,-22.690],
[32.480,49.895,-23.437],[31.204,56.107,-29.527],[28.357,59.047,-29.007],
[21.501,55.610,-33.718],[29.376,56.199,-36.833],[15.694,58.628,-33.147],
[33.728,57.167,-27.927],[19.152,58.042,-28.655],[28.812,55.992,-31.088],
[19.213,57.698,-31.979],[30.892,54.231,-27.640],[23.836,53.463,-30.963],
[30.573,51.800,-37.486],[31.224,-59.236,-22.936],[21.281,54.260,-31.068],
[24.306,53.984,-27.051],[28.864,56.793,-27.122],[25.043,57.083,-39.630],
[21.356,58.180,-30.092],[21.114,-58.895,-33.382],[25.570,57.740,-33.676],
[17.895,58.929,-26.181],[22.439,56.571,-39.606],[15.397,55.724,-33.275],
[16.836,55.729,-39.125],[23.272,57.577,-32.459],[17.636,56.899,-34.253],
[20.066,57.418,-35.472],[26.474,57.243,-30.054],[20.452,58.969,-38.493],
[25.445,-59.227,-30.591],[19.855,53.642,-35.508],[24.086,57.589,-28.005],
[28.729,54.939,-39.017],[26.930,57.985,-37.864],[32.640,58.595,-37.622],
[17.946,56.115,-26.658],[26.372,55.824,-26.862],[27.089,55.192,-36.676],
[24.809,-59.339,-35.661],[24.946,-59.452,-26.351],[25.394,54.548,-34.442],
[21.255,55.308,-37.012],[31.250,56.999,-25.709],[17.899,58.486,-39.331],
[18.139,52.939,-29.802],[21.828,56.883,-24.125],[21.810,56.062,-28.280],
[25.625,58.289,-21.918],[20.215,-59.763,-22.825],[19.743,54.309,-27.950],
[24.719,56.637,-24.659],[27.535,55.716,-24.280],[29.118,55.996,-21.962],
[16.506,56.037,-22.673],[23.266,52.136,-24.156],[22.929,53.526,-21.475],
[32.351,59.274,-40.238],[29.626,56.165,-40.984],[30.768,57.670,-35.375],
[30.512,58.595,-30.689],[30.782,58.596,-27.754],[15.236,57.409,-24.709],
[31.750,53.381,-22.595],[31.015,54.379,-35.536],[16.910,56.075,-30.354],
[15.793,59.061,-29.835],[32.970,55.258,-24.850],[33.114,56.040,-39.622],
[16.211,55.465,-36.171],[15.211,53.518,-29.744],[31.848,52.706,-25.363],
[26.389,54.611,-31.825],[30.004,51.611,-32.031],[32.034,53.535,-32.320],
[17.569,53.081,-34.370],[16.474,52.734,-37.065],[22.904,53.211,-33.538],
[27.198,52.425,-32.980],[27.509,52.868,-37.855],[19.813,56.347,-39.451],
[23.779,55.941,-37.167],[19.320,54.531,-25.088],[20.412,52.222,-24.475],
[27.691,52.878,-22.093],[27.077,-50.276,-15.035],[26.114,-51.216,-12.165],
[19.360,-49.309,-11.110],[30.032,-47.250,-12.747],[20.497,-47.296,-12.367],
[13.697,-53.038,-16.316],[15.910,-49.542,-13.105],[27.825,-41.684,-13.157],
[24.831,-49.582,-10.394],[32.689,-57.093,-11.311],[27.713,-53.753,-10.988],
[19.785,-54.744,-13.232],[26.900,-55.810,-14.677],[17.261,-53.584,-12.525],
[22.706,-52.596,-13.403],[33.698,-54.144,-8.894],[19.123,-54.760,-10.566],
[25.507,-46.207,-16.409],[22.278,-49.083,-8.872],[30.844,-56.308,-13.441],
[18.986,-52.190,-11.021],[19.694,-47.493,-9.115],[16.945,-46.564,-18.167],
[28.985,-45.774,-10.688],[29.956,-57.543,-8.561],[23.047,-58.407,-11.646],
[29.115,-44.526,-13.886],[28.607,-46.743,-16.034],[31.499,-48.069,-7.569],
[24.137,-49.602,-16.319],[27.841,-49.546,-21.847],[33.490,-50.137,-9.759],
[29.844,-51.280,-17.659],[28.386,-51.769,-7.717],[20.737,-48.246,-15.653],
[22.802,-45.557,-12.314],[22.055,-50.840,-11.129],[26.770,-43.973,-8.996],
[21.807,-46.894,-7.204],[27.370,-49.660,-19.043],[19.649,-50.040,-8.236],
[25.609,-43.524,-14.512],[21.646,-44.589,-4.259],[32.039,-58.089,-19.849],
[19.892,-50.125,-18.097],[33.282,-44.467,-20.036],[16.165,-47.251,-9.380],
[16.002,-45.068,-11.658],[32.890,-45.865,-6.586],[24.783,-45.655,-7.655],
[25.438,-40.862,-17.925],[19.241,-42.587,-17.635],[28.564,-42.904,-19.955],
[27.410,-42.650,-16.440],[29.409,-52.994,-5.568],[17.324,-45.906,-6.903],
[20.944,-56.888,-10.867],[24.961,-58.977,-6.964],[27.692,-54.864,-7.352],
[29.374,-51.370,-20.422],[25.885,-56.361,-20.371],[21.740,-53.634,-10.942],
[20.744,-52.515,-16.049],[25.911,-51.815,-17.463],[26.640,-53.039,-5.433],
[28.046,-52.886,-13.600],[19.249,-53.611,-7.535],[21.052,-54.663,-22.778],
[15.225,-55.580,-14.907],[16.480,-55.437,-19.902],[26.368,-51.855,-9.545],
[23.609,-55.112,-13.786],[17.970,-54.015,-15.443],[21.242,-54.883,-18.334],
[27.803,-58.631,-11.062],[18.724,-53.868,-19.862],[30.243,-39.901,-5.242],
[19.007,-48.305,-6.122],[24.925,-53.555,-11.752],[19.830,-56.434,-16.012],
[22.744,-55.596,-9.530],[24.699,-48.562,-18.909],[30.788,-50.120,-9.172],
[28.544,-55.161,-20.561],[24.691,-48.044,-6.067],[26.431,-53.942,-19.213],
[18.584,-48.995,-13.937],[27.238,-47.128,-18.250],[14.530,-49.877,-6.963],
[22.999,-52.389,-4.970],[28.840,-45.945,-6.735],[32.892,-53.785,-18.693],
[32.914,-48.520,-5.532],[31.070,-45.921,-18.142],[28.678,-50.699,-12.253],
[21.069,-45.483,-14.614],[18.067,-50.773,-20.757],[22.887,-47.286,-14.562],
[22.422,-44.332,-8.493],[22.003,-42.839,-16.139],[26.234,-47.591,-9.165],
[29.004,-46.942,-20.908],[29.085,-42.634,-7.430],[32.132,-41.863,-20.261],
[17.912,-55.886,-8.008],[29.909,-48.673,-18.271],[17.067,-43.252,-9.112],
[21.845,-45.115,-23.358],[21.894,-45.727,-19.943],[18.936,-48.054,-20.098],
[24.597,-56.219,-7.843],[27.364,-55.856,-17.639],[25.478,-44.839,-19.246],
[23.369,-52.874,-16.250],[24.516,-53.503,-7.356],[25.262,-57.139,-16.698],
[22.528,-42.801,-19.764],[21.315,-57.402,-19.626],[32.134,-55.761,-7.503],
[31.165,-53.325,-13.228],[19.257,-45.469,-18.942],[16.841,-53.146,-21.727],
[31.140,-42.675,-10.393],[18.271,-57.510,-10.808],[18.526,-44.574,-16.198],
[17.800,-41.267,-5.342],[20.526,-42.209,-3.944],[25.962,-44.276,-22.462],
[27.766,-44.532,-4.372],[30.950,-44.910,-21.021],[15.083,-47.456,-5.020],
[20.768,-47.211,-22.113],[22.210,-47.538,-4.757],[24.216,-46.451,-22.384],
[29.244,-47.936,-9.240],[17.150,-50.666,-7.084],[21.926,-48.373,-19.493],
[21.667,-52.120,-8.621],[22.378,-52.349,-20.814],[26.139,-49.892,-4.723],
[32.060,-49.428,-21.957],[21.645,-54.611,-6.479],[20.404,-56.634,-8.094],
[24.345,-50.650,-7.896],[28.529,-53.904,-2.989],[28.706,-49.752,-5.504],
[19.371,-53.011,-4.224],[22.311,-58.390,-8.944],[23.317,-57.513,-5.198],
[25.798,-56.799,-5.410],[28.163,-56.355,-4.326],[18.454,-55.503,-3.575],
[21.558,-59.356,-5.428],[25.183,-58.891,-3.580],[22.558,-39.544,-20.095],
[16.449,-44.778,-20.263],[15.927,-40.807,-17.323],[16.323,-45.875,-15.564],
[34.025,-42.878,-9.605],[32.119,-42.725,-7.627],[30.899,-46.684,-3.858],
[33.022,-48.703,-17.609],[32.031,-50.468,-15.779],[32.704,-47.600,-11.078],
[30.673,-44.769,-8.507],[34.230,-49.197,-20.203],[33.106,-54.137,-21.982],
[29.964,-52.291,-15.264],[17.021,-50.149,-10.173],[30.984,-50.891,-6.311],
[33.795,-50.811,-4.092],[29.972,-57.307,-21.602],[29.658,-54.726,-17.903],
[30.366,-54.961,-10.720],[30.984,-52.890,-8.945],[15.809,-54.619,-4.533],
[35.067,58.703,-4.988],[30.672,-57.771,-16.881],[16.268,-55.925,-11.168],
[15.569,-51.754,-11.463],[31.819,-57.676,-5.198],[32.544,-56.670,-22.511],
[17.347,-57.172,-17.154],[30.909,-42.065,-14.914],[15.595,-56.969,-8.883],
[18.303,-44.071,-4.602],[19.780,-44.720,-9.026],[25.459,-43.894,-5.710],
[27.508,-41.721,-5.460],[31.966,59.452,-7.308],[18.888,-41.255,-7.894],
[21.375,-41.645,-8.577],[25.363,-41.623,-10.641],[28.423,-42.062,-10.062],
[30.570,-40.588,-8.875],[19.616,-42.665,-11.577],[22.208,-43.086,-10.828],
[23.074,-42.829,-13.267],[25.413,-56.403,-12.326],[31.336,-59.484,-12.480],
[33.331,58.667,-14.535],[16.615,-58.127,-13.667],[16.457,-59.231,-20.112],
[23.881,-59.416,-14.569],[28.370,-57.952,-13.869],[28.110,-58.394,-16.980],
[16.754,-43.334,-18.008],[19.229,-56.323,-20.847],[23.675,-58.556,-18.445],
[26.003,-42.232,-20.506],[32.310,-42.991,-17.189],[18.111,-58.186,-6.677],
[19.552,-57.759,-4.455],[19.447,-41.187,-21.788],[28.968,-59.629,-3.109],
[31.077,-44.139,-5.620],[25.536,-30.541,-14.333],[26.750,-32.143,-11.954],
[18.206,-33.069,-10.784],[31.437,-27.152,-12.416],[20.321,-28.657,-11.729],
[15.977,-33.492,-16.027],[15.391,-31.165,-14.745],[28.524,-21.836,-15.123],
[25.633,-28.758,-12.532],[34.227,-37.734,-10.603],[27.442,-34.757,-10.566],
[20.604,-36.124,-14.150],[29.091,-37.142,-15.295],[16.360,-34.366,-13.035],
[23.051,-33.472,-12.939],[33.420,-35.197,-9.434],[20.293,-37.877,-11.503],
[26.768,-28.513,-15.619],[21.584,-30.624,-7.466],[30.304,-38.206,-12.749],
[18.624,-35.579,-12.301],[19.360,-29.183,-8.414],[18.035,-26.806,-18.798],
[28.920,-25.715,-11.853],[31.302,-36.843,-10.081],[23.259,-40.109,-12.023],
[28.358,-27.197,-14.002],[29.172,-26.863,-16.539],[30.580,-28.258,-9.017],
[24.493,-31.663,-16.508],[27.236,-33.228,-20.920],[33.576,-30.649,-9.366],
[28.448,-31.096,-16.063],[28.690,-32.020,-10.064],[20.611,-31.285,-15.347],
[23.133,-28.441,-11.640],[21.878,-30.769,-12.894],[26.204,-25.043,-10.489],
[21.972,-27.675,-7.321],[28.296,-31.594,-18.873],[19.246,-33.254,-6.911],
[25.214,-25.069,-16.182],[18.930,-24.894,-5.306],[31.676,-38.121,-17.988],
[21.249,-32.492,-17.602],[32.399,-25.014,-20.405],[15.875,-29.656,-7.773],
[16.005,-26.907,-11.670],[34.819,-25.736,-7.255],[23.183,-25.148,-6.419],
[25.834,-22.255,-17.305],[17.866,-24.348,-14.807],[27.607,-24.555,-21.209],
[27.332,-24.354,-17.892],[31.746,-33.008,-7.046],[17.872,-27.000,-7.376],
[20.947,-40.009,-10.407],[25.507,-40.620,-7.021],[28.523,-34.917,-6.255],
[30.141,-33.204,-20.870],[25.016,-38.755,-20.615],[21.102,-35.197,-11.748],
[22.070,-34.509,-15.948],[25.726,-33.542,-18.098],[26.331,-33.252,-7.009],
[27.501,-34.026,-13.697],[19.013,-35.343,-9.295],[21.249,-36.073,-20.851],
[15.132,-36.614,-14.519],[18.515,-36.510,-19.034],[25.113,-32.841,-9.757],
[22.646,-38.086,-13.879],[18.792,-33.823,-17.074],[22.704,-35.511,-18.460],
[27.544,-38.060,-9.350],[21.685,-33.454,-20.185],[30.705,-23.747,-5.131],
[18.619,-30.638,-5.975],[25.314,-36.685,-13.668],[19.236,-36.942,-16.393],
[23.906,-35.158,-10.576],[25.236,-30.203,-19.004],[31.804,-30.384,-11.353],
[27.525,-37.624,-20.188],[25.457,-28.190,-7.346],[28.373,-35.413,-19.014],
[18.357,-30.552,-13.061],[26.797,-28.262,-18.218],[16.062,-32.029,-6.736],
[22.310,-32.621,-5.779],[26.996,-25.126,-6.699],[33.642,-34.466,-18.242],
[34.025,-27.575,-4.274],[31.804,-27.445,-18.316],[29.465,-30.104,-12.723],
[21.888,-26.685,-16.539],[16.443,-31.911,-20.287],[24.159,-28.670,-16.878],
[23.685,-24.228,-9.811],[20.601,-24.007,-15.906],[26.980,-28.572,-9.790],
[30.096,-29.117,-20.634],[29.593,-25.994,-6.922],[34.066,-23.037,-21.375],
[16.644,-36.958,-7.227],[30.044,-29.529,-17.843],[18.256,-27.387,-10.076],
[23.022,-26.500,-22.453],[22.449,-27.499,-19.147],[17.983,-29.571,-19.993],
[23.997,-37.665,-9.415],[27.894,-39.155,-17.461],[25.021,-26.313,-19.451],
[25.688,-35.634,-16.390],[23.971,-34.837,-7.168],[26.043,-40.026,-14.494],
[22.973,-24.980,-17.962],[20.454,-38.831,-18.266],[31.118,-35.474,-7.848],
[32.052,-32.797,-12.497],[20.234,-26.909,-21.053],[18.581,-33.873,-19.850],
[29.293,-22.879,-11.251],[18.350,-39.868,-11.270],[19.190,-26.423,-16.008],
[15.999,-23.500,-3.867],[20.952,-23.875,-3.838],[26.001,-26.396,-23.057],
[27.839,-25.725,-4.191],[29.912,-26.429,-21.631],[15.819,-29.677,-5.156],
[23.030,-29.162,-21.248],[23.155,-29.840,-5.285],[25.755,-29.115,-23.296],
[29.052,-28.913,-6.535],[18.157,-33.108,-4.081],[20.903,-30.645,-19.597],
[21.991,-33.561,-8.396],[24.261,-33.071,-20.801],[26.554,-29.932,-5.421],
[32.051,-31.124,-21.890],[21.219,-37.148,-6.361],[21.379,-36.834,-8.960],
[24.475,-31.304,-7.047],[28.154,-34.553,-3.584],[28.292,-31.970,-5.441],
[21.013,-34.714,-4.861],[22.739,-39.307,-7.664],[23.367,-36.333,-4.458],
[26.045,-36.053,-5.978],[28.096,-37.659,-5.378],[18.490,-35.513,-5.181],
[22.764,-41.887,-5.902],[22.889,-39.192,-4.954],[23.005,-23.094,-20.020],
[16.015,-27.991,-20.982],[16.632,-22.152,-16.576],[15.959,-22.963,-13.841],
[33.636,-24.005,-10.318],[30.785,-23.715,-8.152],[30.483,-27.505,-4.614],
[35.074,-29.304,-17.719],[32.012,-33.040,-16.688],[34.080,-28.032,-11.556],
[30.892,-25.898,-9.957],[33.631,-31.316,-18.482],[33.479,-33.981,-21.231],
[30.080,-33.641,-14.671],[16.050,-30.001,-11.790],[31.922,-30.362,-5.689],
[33.340,-32.638,-3.901],[30.625,-37.579,-21.289],[30.684,-35.493,-17.254],
[30.198,-34.723,-11.814],[30.895,-33.374,-9.648],[15.540,-35.335,-4.822],
[34.308,-39.245,-3.589],[32.135,-36.675,-15.308],[17.023,-37.675,-12.509],
[14.366,-33.909,-11.003],[32.453,-37.180,-5.660],[34.044,-36.928,-20.356],
[16.077,-36.217,-17.131],[29.530,-24.448,-15.466],[16.379,-38.364,-9.787],
[18.467,-26.296,-2.302],[20.218,-25.486,-8.044],[25.441,-23.204,-5.610],
[28.688,-22.311,-6.561],[32.258,-39.665,-7.056],[18.147,-20.883,-8.033],
[20.199,-23.709,-10.244],[26.179,-22.221,-10.972],[28.482,-24.750,-9.018],
[29.317,-21.485,-8.888],[16.903,-24.588,-11.231],[18.991,-25.627,-12.834],
[21.386,-25.953,-11.650],[25.671,-38.808,-11.553],[33.006,-41.414,-12.036],
[31.994,-39.798,-14.265],[16.315,-40.245,-14.547],[16.022,-39.053,-19.305],
[23.555,-40.785,-15.237],[27.873,-37.727,-13.177],[29.114,-40.213,-19.712],
[16.277,-24.929,-16.865],[19.821,-38.479,-20.804],[23.009,-38.742,-17.328],
[27.140,-22.288,-19.669],[34.280,-23.940,-18.809],[18.650,-38.572,-5.628],
[19.900,-39.573,-3.660],[17.823,-22.284,-21.940],[27.859,-39.970,-3.529],
[32.519,-26.249,-6.370],[26.992,-13.902,-14.759],[26.578,-11.532,-11.921],
[18.613,-13.537,-10.644],[29.654,-9.105,-13.013],[20.607,-9.625,-13.866],
[16.439,-14.428,-16.784],[16.332,-12.370,-14.997],[26.900,-6.206,-11.508],
[24.061,-10.097,-12.976],[33.082,-19.180,-11.642],[28.903,-16.428,-10.471],
[20.886,-16.803,-13.728],[29.480,-17.893,-15.296],[16.201,-16.059,-13.786],
[21.840,-14.282,-14.292],[34.138,-17.374,-9.485],[18.128,-18.690,-9.796],
[25.402,-9.143,-16.211],[21.261,-11.635,-10.211],[30.781,-19.111,-13.218],
[18.549,-16.356,-11.989],[19.261,-9.722,-10.608],[16.710,-8.749,-17.781],
[28.765,-8.928,-9.738],[30.657,-19.358,-10.020],[22.296,-23.539,-12.988],
[29.174,-6.309,-14.093],[29.343,-8.955,-16.003],[32.158,-9.886,-7.432],
[25.977,-11.938,-16.582],[26.978,-14.867,-20.672],[34.046,-12.817,-9.147],
[30.065,-13.157,-17.409],[27.763,-15.277,-7.679],[20.261,-12.136,-14.914],
[21.437,-8.135,-11.423],[22.135,-13.699,-11.433],[26.001,-7.354,-9.188],
[21.839,-10.428,-7.442],[27.819,-12.817,-19.099],[19.992,-12.687,-7.841],
[26.479,-6.619,-16.551],[19.677,-8.347,-6.643],[32.346,-20.489,-19.472],
[19.222,-13.023,-17.150],[32.144,-6.984,-20.712],[15.678,-10.443,-8.798],
[15.160,-6.691,-13.065],[35.278,-6.927,-6.959],[24.586,-8.365,-6.979],
[25.468,-3.971,-16.854],[19.431,-4.934,-15.624],[28.406,-5.270,-19.189],
[28.502,-4.963,-16.516],[31.347,-15.148,-6.724],[18.075,-7.930,-8.636],
[20.692,-20.965,-11.862],[26.314,-22.663,-8.244],[29.081,-16.950,-5.781],
[30.281,-13.486,-21.673],[24.650,-17.002,-19.585],[20.407,-15.978,-9.792],
[22.484,-15.644,-16.732],[25.996,-14.522,-17.272],[26.188,-15.878,-5.201],
[27.253,-16.092,-12.721],[19.145,-16.258,-7.040],[20.844,-18.673,-21.337],
[15.125,-18.144,-15.843],[17.324,-17.597,-18.961],[26.337,-15.886,-9.848],
[21.943,-19.566,-13.970],[18.781,-16.130,-16.086],[21.549,-17.291,-18.952],
[27.711,-18.991,-10.464],[20.575,-14.390,-19.038],[30.126,-3.056,-5.137],
[19.152,-11.127,-5.762],[24.815,-14.602,-11.681],[19.739,-18.371,-16.860],
[22.855,-16.769,-9.726],[23.484,-10.985,-18.003],[30.714,-10.662,-9.320],
[28.018,-19.364,-20.114],[25.095,-11.210,-7.147],[29.185,-15.910,-17.196],
[18.408,-11.876,-12.863],[25.783,-10.253,-19.127],[16.446,-15.194,-6.854],
[24.693,-13.528,-5.604],[27.104,-6.569,-6.884],[34.164,-15.420,-19.319],
[34.336,-9.504,-5.971],[30.197,-8.612,-19.804],[28.976,-12.630,-10.420],
[22.162,-8.150,-16.331],[16.803,-12.129,-20.220],[22.914,-11.274,-15.227],
[22.016,-5.575,-10.022],[22.079,-5.667,-15.301],[27.154,-10.897,-8.852],
[28.062,-10.078,-20.383],[29.969,-7.350,-6.256],[33.290,-4.524,-20.296],
[18.033,-18.573,-6.604],[29.268,-10.749,-18.037],[15.530,-7.884,-9.558],
[23.167,-8.070,-22.210],[23.459,-8.263,-19.442],[18.687,-10.365,-19.890],
[25.493,-19.520,-8.542],[28.331,-19.595,-17.375],[23.638,-5.992,-17.845],
[24.639,-17.511,-16.976],[24.919,-15.974,-7.397],[26.251,-19.773,-15.727],
[21.314,-6.203,-19.570],[21.232,-19.985,-19.005],[32.510,-18.933,-8.056],
[30.128,-14.518,-12.160],[19.830,-8.411,-18.717],[17.406,-15.092,-20.078],
[29.641,-6.248,-10.165],[17.870,-20.121,-12.233],[19.350,-7.822,-15.976],
[16.002,-4.403,-5.879],[22.551,-5.818,-3.441],[25.617,-7.224,-22.012],
[28.520,-8.476,-4.494],[29.529,-6.910,-24.028],[17.444,-11.795,-3.670],
[21.218,-12.150,-24.431],[22.213,-11.583,-4.863],[25.777,-10.395,-22.752],
[28.691,-12.475,-7.513],[18.328,-14.162,-4.685],[23.156,-13.533,-18.682],
[21.840,-15.104,-7.603],[23.848,-14.629,-21.452],[27.167,-10.686,-3.386],
[32.287,-11.677,-22.258],[22.052,-17.890,-7.346],[21.266,-18.802,-10.477],
[23.482,-13.087,-8.147],[28.255,-16.332,-3.336],[28.582,-13.759,-5.131],
[21.499,-15.790,-4.789],[23.103,-21.277,-9.447],[25.016,-17.866,-3.669],
[27.092,-18.408,-6.489],[28.816,-19.134,-3.615],[18.570,-16.930,-4.333],
[23.110,-22.254,-6.934],[23.157,-19.582,-4.216],[21.941,-3.611,-19.698],
[17.157,-7.952,-20.343],[15.608,-6.743,-15.598],[17.562,-5.262,-13.480],
[34.480,-4.308,-11.132],[33.648,-5.456,-8.576],[30.776,-9.936,-3.864],
[32.114,-10.114,-18.514],[32.606,-12.751,-16.264],[33.302,-10.032,-11.212],
[31.628,-7.535,-8.663],[32.739,-12.970,-19.738],[32.165,-15.747,-21.318],
[30.100,-14.320,-15.104],[15.184,-12.632,-11.363],[30.663,-12.169,-5.793],
[32.557,-13.673,-4.633],[30.659,-18.809,-21.268],[31.728,-17.871,-17.342],
[31.962,-16.474,-11.248],[31.306,-16.833,-8.711],[15.448,-17.588,-5.572],
[34.588,-20.614,-2.568],[31.072,-20.368,-16.471],[16.198,-17.874,-11.475],
[14.719,-15.501,-11.440],[34.266,-19.027,-4.814],[33.230,-18.059,-20.499],
[15.685,-19.420,-18.188],[32.110,-5.215,-15.045],[15.830,-19.735,-8.448],
[18.013,-7.080,-4.093],[21.065,-8.348,-8.925],[25.893,-4.201,-5.778],
[28.759,-4.549,-6.667],[32.479,-21.772,-6.906],[18.896,-3.499,-6.311],
[19.471,-4.376,-9.996],[24.451,-4.305,-10.718],[27.125,-4.154,-9.369],
[29.833,-2.752,-8.446],[18.193,-6.624,-11.258],[20.351,-6.778,-13.250],
[22.767,-5.662,-12.693],[25.282,-19.670,-12.332],[31.514,-21.500,-11.943],
[32.749,-21.051,-14.204],[18.675,-21.468,-14.434],[17.482,-21.745,-18.995],
[23.666,-22.834,-15.403],[28.538,-20.529,-12.626],[30.030,-22.119,-18.370],
[16.370,-5.827,-18.264],[18.254,-19.830,-20.976],[24.342,-20.074,-18.702],
[25.572,-5.440,-19.956],[31.974,-6.649,-17.710],[18.890,-21.892,-5.328],
[20.803,-20.775,-3.458],[18.157,-5.009,-20.799],[28.194,-21.759,-3.988],
[32.070,-5.798,-5.387],[27.070,5.215,-15.694],[26.256,4.770,-13.013],
[18.588,4.824,-11.446],[28.893,9.070,-13.331],[19.830,9.079,-11.450],
[16.937,4.264,-16.609],[16.351,6.241,-14.501],[27.619,13.910,-12.928],
[25.523,7.755,-12.165],[32.652,-0.305,-9.623],[28.399,1.450,-9.437],
[21.434,1.563,-14.165],[27.569,0.140,-15.514],[16.980,2.293,-13.035],
[23.605,5.747,-13.254],[33.526,1.643,-7.327],[20.079,1.928,-9.429],
[25.034,9.212,-17.094],[22.004,6.606,-9.950],[30.963,-0.468,-13.492],
[19.097,0.903,-12.430],[19.536,8.642,-8.325],[17.007,10.249,-17.611],
[28.472,10.665,-11.209],[29.934,-0.650,-10.059],[23.299,-2.178,-12.098],
[28.441,11.819,-14.215],[28.922,9.959,-15.990],[30.959,8.834,-8.608],
[23.877,6.455,-17.012],[25.584,4.321,-21.776],[32.185,4.442,-8.803],
[29.259,6.200,-16.818],[28.973,3.513,-7.926],[21.413,6.562,-15.572],
[21.887,10.655,-11.999],[20.769,6.113,-12.549],[26.626,11.490,-9.324],
[22.118,9.099,-8.434],[27.124,4.136,-19.688],[19.053,5.749,-8.290],
[25.779,11.770,-16.678],[20.479,11.036,-6.861],[31.453,-2.658,-19.981],
[20.655,6.830,-18.027],[32.302,10.035,-19.896],[16.313,7.861,-9.357],
[17.121,9.555,-11.379],[35.799,11.731,-6.898],[24.315,11.761,-7.695],
[23.944,13.824,-16.172],[19.471,14.465,-15.435],[29.396,13.668,-20.822],
[28.039,13.089,-18.107],[30.385,3.892,-4.552],[17.191,9.757,-7.257],
[20.485,-2.286,-11.729],[26.745,-2.248,-7.531],[28.700,0.538,-6.929],
[30.165,6.701,-22.164],[23.659,-0.033,-19.693],[21.518,3.898,-10.840],
[22.514,2.676,-16.448],[25.547,3.499,-17.176],[26.104,3.555,-6.401],
[26.280,2.448,-14.447],[18.723,2.502,-7.301],[21.933,1.224,-21.055],
[14.340,0.271,-15.080],[16.567,0.674,-19.535],[25.887,2.877,-9.717],
[21.964,-0.400,-16.145],[18.337,2.383,-15.395],[21.241,1.344,-18.262],
[26.203,0.103,-10.371],[19.477,2.391,-21.155],[30.790,17.507,-4.791],
[18.641,7.287,-5.286],[24.007,3.001,-12.612],[19.945,-2.307,-15.911],
[23.185,1.788,-10.029],[23.555,8.050,-19.254],[31.750,6.453,-10.813],
[28.044,-0.449,-20.798],[25.839,9.232,-6.965],[29.096,1.743,-19.333],
[18.207,7.249,-12.767],[27.618,9.246,-18.666],[15.033,6.037,-6.663],
[23.485,5.279,-6.003],[28.220,12.928,-7.778],[34.040,2.442,-18.689],
[33.240,8.434,-5.625],[31.328,12.228,-18.227],[27.898,6.385,-11.850],
[21.099,12.306,-16.515],[18.822,5.562,-19.963],[22.304,9.146,-16.581],
[24.453,12.459,-10.638],[21.972,15.917,-15.308],[27.087,8.170,-10.065],
[28.367,8.391,-21.011],[30.362,12.721,-6.188],[32.509,14.313,-20.345],
[16.921,-0.803,-9.502],[30.319,5.433,-19.685],[19.024,10.996,-9.627],
[21.535,8.536,-21.246],[21.452,10.060,-18.982],[18.642,8.240,-19.035],
[25.668,-0.003,-7.821],[26.403,-0.626,-18.722],[23.746,10.660,-20.101],
[24.171,0.722,-17.264],[23.756,2.386,-7.321],[25.062,-1.712,-15.161],
[21.980,12.725,-19.022],[20.337,-1.428,-18.876],[31.381,-0.139,-7.138],
[28.407,3.109,-12.501],[19.131,10.180,-20.853],[15.832,3.001,-22.445],
[31.972,14.307,-11.290],[17.297,-3.167,-11.399],[19.642,9.376,-16.168],
[17.400,14.115,-5.685],[22.788,13.003,-3.686],[24.413,10.980,-23.458],
[27.589,8.863,-4.886],[29.122,12.191,-22.982],[16.222,6.807,-4.407],
[19.937,5.889,-22.375],[21.485,7.338,-5.927],[25.477,9.395,-21.725],
[28.731,6.823,-8.486],[17.484,4.339,-5.791],[22.403,5.762,-20.278],
[21.463,4.161,-8.165],[23.137,3.459,-22.548],[26.118,6.597,-4.646],
[33.170,7.656,-22.189],[21.044,1.056,-6.460],[22.063,-0.739,-9.546],
[24.215,4.894,-8.995],[28.912,2.132,-2.578],[28.442,5.367,-6.278],
[19.323,3.171,-4.078],[23.165,-3.176,-8.304],[22.726,2.866,-4.366],
[24.576,0.955,-4.885],[27.772,0.228,-4.501],[18.961,0.603,-4.636],
[22.719,-4.136,-5.838],[22.650,-2.090,-3.726],[22.072,14.996,-20.724],
[16.515,9.426,-21.049],[16.231,13.770,-16.007],[16.565,12.096,-13.718],
[34.801,13.682,-10.805],[34.122,11.510,-8.972],[31.295,9.856,-4.782],
[32.164,7.570,-18.223],[31.957,5.781,-15.992],[33.116,8.778,-11.219],
[30.471,11.740,-9.132],[33.124,5.928,-20.130],[31.779,3.142,-20.193],
[30.130,4.379,-14.722],[15.717,6.609,-11.856],[31.061,6.325,-7.105],
[33.936,3.583,-5.228],[31.036,0.164,-20.635],[30.163,1.090,-16.751],
[30.738,1.962,-11.483],[31.175,1.941,-8.751],[15.110,0.274,-5.042],
[32.553,-3.409,-3.626],[29.771,-1.634,-16.096],[17.025,-0.531,-12.605],
[15.133,3.097,-10.700],[31.937,-0.953,-4.714],[33.564,0.763,-20.725],
[15.396,-1.790,-16.960],[29.700,14.045,-15.692],[14.872,-1.689,-7.342],
[19.264,11.731,-4.580],[21.970,11.774,-9.208],[25.044,13.959,-5.988],
[29.343,15.599,-5.843],[32.665,-3.002,-8.845],[18.141,15.389,-8.297],
[19.878,13.644,-9.693],[24.882,15.113,-11.678],[26.898,14.328,-9.898],
[30.094,14.859,-8.590],[17.080,13.082,-10.651],[19.232,11.084,-14.245],
[23.474,13.257,-13.385],[26.057,-1.571,-12.649],[30.900,-2.698,-12.163],
[32.656,-2.366,-14.312],[18.612,-2.416,-13.668],[16.872,-1.893,-19.139],
[21.983,-3.107,-14.396],[28.099,-3.084,-13.454],[30.603,-4.423,-17.997],
[15.482,12.309,-18.291],[18.493,-2.252,-21.278],[23.578,-2.243,-18.040],
[24.606,13.489,-19.674],[32.931,13.128,-16.365],[17.593,-1.266,-5.793],
[19.644,-2.292,-3.956],[19.913,12.954,-20.988],[27.262,-3.047,-3.641],
[31.963,13.260,-4.103],[27.087,24.244,-15.152],[24.920,24.949,-13.613],
[18.847,22.956,-10.598],[29.838,26.934,-15.005],[20.293,29.044,-10.196],
[16.026,22.614,-16.123],[15.443,24.534,-13.959],[28.492,32.970,-13.162],
[24.632,27.462,-12.498],[32.707,18.241,-11.723],[28.558,21.205,-10.487],
[19.409,19.755,-13.932],[28.323,18.527,-16.092],[17.455,21.913,-13.960],
[22.254,21.088,-13.737],[34.168,20.768,-9.837],[19.388,18.156,-8.642],
[25.794,28.244,-15.039],[21.487,25.775,-10.177],[30.955,18.609,-14.460],
[17.985,20.579,-11.587],[20.118,27.258,-7.209],[17.596,28.098,-17.979],
[28.924,27.760,-11.450],[30.283,17.385,-10.208],[22.462,16.369,-12.515],
[28.297,30.192,-13.227],[28.971,29.341,-17.150],[30.692,27.021,-7.359],
[24.767,24.895,-16.444],[26.700,22.912,-20.702],[33.672,23.311,-9.162],
[29.828,24.116,-17.102],[28.540,22.564,-7.941],[20.532,25.497,-14.646],
[21.180,27.231,-12.732],[21.200,23.863,-12.444],[26.517,29.237,-11.457],
[22.887,27.483,-6.725],[26.700,25.132,-18.336],[19.744,24.339,-8.311],
[25.182,31.845,-15.232],[19.030,29.581,-5.235],[31.133,17.051,-20.052],
[19.680,24.435,-19.903],[32.328,30.846,-20.234],[15.324,26.404,-8.633],
[17.191,27.678,-10.748],[33.894,29.392,-5.670],[24.528,29.463,-8.484],
[25.694,33.516,-17.265],[18.576,31.746,-16.159],[28.533,31.970,-20.549],
[27.953,31.601,-15.956],[31.207,22.343,-6.287],[17.599,27.529,-8.023],
[20.775,16.583,-10.148],[24.593,16.134,-7.847],[28.896,20.317,-6.058],
[28.902,24.317,-19.961],[25.731,17.325,-21.994],[21.342,21.530,-10.683],
[21.990,22.056,-17.564],[23.796,23.591,-18.755],[25.553,22.394,-4.447],
[27.516,21.206,-15.786],[19.284,20.913,-8.369],[22.312,19.849,-21.387],
[16.768,19.098,-15.541],[16.890,20.041,-19.022],[26.219,24.195,-9.251],
[23.825,18.668,-14.313],[19.655,21.960,-15.454],[19.170,19.260,-17.848],
[26.989,18.306,-10.129],[21.637,22.511,-20.500],[30.670,33.589,-6.109],
[18.240,25.310,-6.153],[24.982,22.064,-12.131],[19.485,17.384,-15.293],
[23.896,22.173,-9.696],[23.303,26.301,-18.840],[30.741,25.117,-10.895],
[28.884,16.330,-21.477],[25.700,27.541,-6.802],[28.770,18.589,-19.497],
[18.204,24.849,-13.554],[26.417,27.810,-18.257],[16.592,23.994,-7.731],
[24.289,24.024,-6.657],[27.142,30.063,-8.387],[34.135,21.041,-20.743],
[33.146,26.652,-6.240],[31.953,28.343,-19.324],[28.344,23.881,-11.457],
[19.992,28.987,-16.845],[17.390,25.527,-21.233],[22.846,26.809,-15.787],
[22.481,31.279,-10.255],[21.527,32.767,-16.440],[26.977,26.554,-10.391],
[29.500,28.477,-20.303],[29.581,30.324,-6.649],[33.230,32.729,-18.317],
[16.689,19.660,-7.961],[29.079,26.810,-18.213],[18.024,30.544,-9.696],
[22.784,28.473,-23.727],[22.184,27.772,-21.120],[20.462,27.062,-18.775],
[24.685,18.875,-8.004],[26.210,16.527,-18.790],[24.406,29.360,-19.863],
[25.696,21.990,-18.172],[25.012,21.610,-7.387],[25.374,18.332,-16.387],
[22.622,31.428,-20.082],[21.464,17.457,-19.115],[30.879,19.432,-8.282],
[30.321,22.946,-12.694],[19.369,29.580,-19.310],[17.535,21.574,-21.317],
[30.896,32.769,-11.111],[17.958,15.160,-12.311],[17.215,29.443,-15.522],
[16.648,30.702,-4.160],[22.909,31.521,-3.903],[26.012,31.063,-22.274],
[28.144,27.836,-7.570],[28.772,30.035,-22.518],[17.234,26.372,-3.949],
[19.942,26.850,-21.615],[21.705,25.196,-5.585],[26.132,26.364,-20.598],
[29.183,25.402,-8.598],[17.081,23.412,-4.349],[20.966,24.867,-17.513],
[21.675,22.444,-7.609],[23.486,24.656,-21.396],[26.269,24.921,-3.863],
[31.729,25.960,-22.385],[21.141,20.187,-6.002],[21.544,19.275,-9.684],
[24.135,24.848,-11.122],[28.273,21.954,-3.905],[29.004,23.924,-5.507],
[19.545,21.648,-4.224],[22.844,14.756,-9.845],[23.414,18.016,-5.554],
[25.496,19.675,-5.226],[27.798,18.854,-3.872],[18.218,19.173,-4.484],
[21.949,15.556,-5.881],[23.762,15.704,-3.815],[20.295,33.797,-21.000],
[14.856,27.560,-20.049],[14.226,33.666,-17.557],[16.046,32.138,-14.912],
[33.163,32.442,-12.933],[32.249,31.484,-8.460],[31.050,28.657,-5.111],
[33.746,26.287,-19.131],[32.482,25.673,-15.464],[32.459,27.137,-11.115],
[29.801,29.623,-9.232],[32.942,23.619,-20.102],[31.750,21.142,-22.016],
[31.990,23.078,-15.296],[16.678,25.014,-11.013],[31.541,25.124,-4.642],
[34.572,22.295,-6.432],[30.099,18.877,-21.798],[30.616,20.132,-16.664],
[31.001,20.331,-11.686],[30.888,22.174,-9.278],[15.440,20.848,-4.924],
[34.897,15.688,-4.931],[31.372,16.628,-17.272],[19.225,18.088,-11.722],
[15.154,22.159,-12.173],[33.245,17.592,-5.871],[33.341,18.449,-21.596],
[16.113,17.283,-19.576],[32.733,30.997,-15.185],[16.087,17.050,-8.496],
[18.398,28.817,-2.475],[20.234,30.475,-7.482],[25.054,31.336,-6.337],
[27.319,33.052,-6.568],[31.943,15.294,-5.819],[17.991,33.667,-9.856],
[20.832,33.425,-8.005],[25.708,34.170,-11.125],[27.550,32.231,-9.828],
[29.231,34.177,-8.909],[16.817,31.827,-11.707],[20.161,30.414,-14.484],
[22.089,30.959,-12.820],[25.671,17.465,-12.364],[29.314,13.423,-10.779],
[32.037,15.540,-14.759],[16.927,16.772,-14.291],[17.140,14.516,-19.466],
[25.882,15.414,-14.381],[28.730,16.507,-14.300],[30.071,14.773,-18.479],
[16.936,30.821,-18.169],[18.802,17.482,-19.650],[23.963,18.297,-19.985],
[25.720,31.845,-19.336],[31.234,30.947,-17.286],[17.495,17.120,-5.988],
[20.722,16.414,-2.860],[17.688,33.378,-21.075],[27.572,15.959,-3.650],
[31.132,31.452,-4.188],[26.277,42.645,-15.257],[25.471,43.823,-12.184],
[17.577,43.675,-8.990],[28.164,46.945,-13.183],[19.117,46.306,-11.682],
[16.533,40.739,-15.751],[16.748,44.417,-16.036],[25.599,51.408,-12.701],
[23.519,45.064,-10.741],[32.902,36.661,-11.070],[28.770,38.770,-9.664],
[20.824,38.389,-15.417],[29.508,36.819,-15.467],[16.930,39.514,-13.396],
[23.064,42.721,-12.936],[34.075,38.913,-8.891],[19.169,36.564,-10.109],
[26.969,45.842,-15.880],[21.404,46.486,-10.280],[31.600,37.531,-13.618],
[18.018,39.311,-11.039],[20.242,45.600,-6.996],[15.525,47.584,-18.338],
[30.531,47.222,-11.423],[30.137,36.025,-11.383],[23.580,35.182,-12.759],
[27.856,49.942,-14.299],[28.964,47.869,-16.231],[30.373,45.483,-6.911],
[25.513,43.932,-17.406],[26.192,41.793,-21.615],[32.275,42.560,-9.075],
[29.459,44.374,-15.652],[27.769,40.094,-7.377],[19.306,40.356,-14.162],
[23.010,46.614,-12.838],[21.347,42.195,-10.704],[27.909,48.307,-10.996],
[22.365,46.960,-7.198],[27.668,41.145,-17.959],[17.977,41.725,-6.792],
[25.859,48.156,-16.723],[20.258,48.413,-5.788],[29.805,35.379,-19.684],
[20.465,42.907,-17.933],[33.323,48.655,-21.030],[16.281,45.856,-9.879],
[16.158,47.622,-11.591],[32.178,48.131,-7.045],[25.368,47.133,-6.083],
[25.852,52.906,-17.728],[18.998,50.087,-15.531],[27.774,49.036,-20.154],
[27.899,51.390,-18.358],[30.948,40.951,-6.738],[17.356,45.926,-6.643],
[21.170,33.637,-10.804],[26.482,35.187,-8.398],[28.821,37.756,-6.928],
[29.024,42.981,-21.205],[24.513,36.818,-20.459],[20.927,39.511,-10.933],
[22.902,40.225,-14.666],[25.014,41.047,-17.482],[24.000,40.932,-2.352],
[27.709,40.485,-14.514],[19.059,41.026,-9.483],[22.706,36.882,-22.686],
[16.043,37.207,-15.505],[17.598,38.182,-18.831],[27.181,41.399,-9.578],
[23.234,37.141,-14.948],[18.321,38.406,-15.978],[20.257,39.846,-18.060],
[27.359,36.608,-10.744],[20.161,40.903,-20.570],[31.067,52.276,-5.386],
[20.455,42.795,-6.994],[24.390,40.682,-11.962],[20.001,35.688,-14.923],
[23.538,39.020,-9.620],[23.393,42.861,-18.809],[30.823,44.374,-10.495],
[27.329,37.085,-20.802],[24.851,44.439,-6.007],[29.611,39.658,-19.595],
[18.651,42.676,-13.551],[26.099,45.869,-18.896],[15.122,44.562,-7.613],
[23.585,42.098,-6.361],[27.792,48.496,-7.022],[33.938,39.987,-19.489],
[32.856,45.501,-6.166],[30.756,47.214,-19.991],[28.279,43.532,-12.187],
[21.300,47.683,-16.693],[16.681,42.764,-20.247],[21.931,44.880,-14.610],
[23.616,48.791,-9.367],[21.522,50.040,-14.698],[25.822,43.891,-8.623],
[28.229,46.391,-20.640],[29.727,50.408,-6.432],[33.602,51.374,-21.492],
[18.032,38.526,-8.471],[30.504,45.880,-17.674],[18.245,48.550,-10.185],
[23.168,49.123,-23.147],[22.436,46.564,-20.641],[18.335,44.350,-18.957],
[26.176,38.052,-8.724],[26.158,36.172,-18.516],[24.189,48.815,-19.066],
[26.221,39.004,-16.313],[24.918,40.061,-7.598],[25.522,35.842,-15.845],
[21.266,48.645,-19.490],[20.717,37.491,-19.486],[31.454,37.926,-8.413],
[30.149,39.757,-13.198],[19.004,47.066,-20.176],[17.818,39.623,-20.944],
[30.578,50.032,-10.903],[19.134,34.606,-12.431],[16.483,49.048,-15.879],
[17.750,50.976,-6.268],[21.661,49.946,-3.928],[25.834,49.327,-22.337],
[27.720,47.045,-4.422],[30.447,48.896,-22.000],[16.442,44.054,-4.833],
[19.183,44.656,-21.464],[22.313,44.795,-5.389],[24.954,46.546,-21.678],
[28.555,44.758,-8.951],[18.273,41.964,-4.371],[21.303,43.475,-20.532],
[21.938,40.912,-8.521],[23.913,43.862,-21.393],[24.960,43.346,-3.191],
[31.732,43.068,-21.025],[21.933,38.226,-7.504],[22.608,36.491,-9.779],
[24.319,42.209,-9.818],[27.359,39.140,-4.775],[29.077,42.794,-6.139],
[21.887,40.330,-4.773],[23.684,33.387,-9.335],[24.275,38.155,-4.164],
[25.924,36.750,-6.066],[28.607,36.314,-4.407],[18.172,38.256,-4.643],
[22.986,33.260,-5.802],[24.531,35.165,-4.219],[23.046,52.613,-18.876],
[15.144,46.128,-21.134],[15.302,51.426,-16.570],[16.408,51.230,-14.180],
[32.880,51.554,-11.331],[32.298,50.323,-8.453],[31.138,47.140,-4.789],
[33.202,44.513,-18.041],[32.044,44.184,-15.568],[32.701,45.539,-11.906],
[29.546,48.468,-8.980],[30.760,42.955,-17.932],[31.929,39.282,-21.227],
[30.244,40.558,-15.927],[16.752,44.127,-11.822],[31.649,43.325,-4.662],
[33.474,41.426,-4.937],[30.157,37.323,-21.984],[31.351,37.427,-18.016],
[31.578,39.089,-10.878],[29.892,41.241,-9.086],[14.999,38.400,-4.304],
[34.877,34.136,-3.821],[30.390,34.518,-16.743],[18.286,37.182,-13.296],
[14.790,40.497,-11.836],[33.445,35.774,-5.934],[33.287,36.974,-20.883],
[13.964,37.257,-18.597],[29.886,51.412,-16.222],[16.630,36.127,-8.568],
[18.415,48.213,-3.890],[20.532,48.794,-8.910],[25.301,50.724,-5.440],
[27.338,52.509,-6.113],[32.131,33.994,-8.758],[18.065,53.033,-8.325],
[18.946,51.635,-10.542],[24.679,50.901,-10.380],[27.602,51.247,-10.746],
[30.023,51.993,-9.156],[16.402,50.267,-11.173],[19.318,48.973,-12.954],
[23.394,49.846,-12.629],[25.722,37.865,-13.621],[30.884,34.465,-13.762],
[32.802,35.325,-15.465],[17.369,34.777,-14.499],[16.406,35.172,-19.521],
[23.396,34.211,-15.544],[27.722,36.027,-13.287],[28.285,33.420,-18.428],
[17.011,50.139,-18.985],[18.281,36.341,-21.128],[23.266,36.883,-17.606],
[26.104,51.510,-20.608],[32.463,49.703,-18.352],[17.760,34.175,-6.258],
[19.706,35.652,-4.579],[18.512,52.150,-20.354],[28.896,33.361,-4.058],
[31.792,49.847,-4.524],[32.611,54.450,-10.530],[28.694,57.507,-9.654],
[20.204,55.114,-14.297],[30.533,56.487,-16.454],[16.476,58.135,-13.194],
[33.974,57.092,-9.177],[19.902,55.054,-9.486],[30.768,56.984,-13.675],
[19.148,56.209,-11.836],[29.775,55.038,-9.362],[22.909,52.470,-11.540],
[32.220,52.756,-19.498],[30.924,58.599,-4.891],[20.503,51.637,-13.005],
[24.945,51.716,-7.804],[29.445,56.639,-6.585],[25.166,55.416,-21.164],
[20.898,57.710,-9.439],[21.741,57.189,-15.034],[24.491,56.628,-14.922],
[18.205,58.358,-10.186],[22.273,56.065,-21.564],[15.155,56.586,-14.863],
[17.128,57.306,-19.168],[22.147,53.443,-14.197],[18.619,-58.905,-15.373],
[21.479,58.198,-17.670],[27.107,56.593,-12.370],[21.867,58.314,-20.229],
[25.294,58.914,-12.043],[22.510,54.587,-16.590],[23.164,57.439,-11.164],
[28.178,54.876,-19.343],[28.313,58.429,-19.598],[33.222,57.312,-16.965],
[16.576,55.917,-9.493],[24.444,53.750,-9.617],[28.182,53.719,-16.597],
[24.120,57.566,-17.415],[26.128,56.067,-8.449],[25.118,53.944,-15.240],
[19.670,56.235,-19.289],[31.506,56.932,-8.657],[16.813,58.802,-22.652],
[18.298,53.156,-12.772],[24.057,58.095,-8.723],[22.523,55.465,-8.580],
[26.646,57.948,-2.471],[20.037,56.971,-5.090],[21.664,52.556,-8.984],
[23.248,56.999,-5.722],[26.532,55.789,-5.349],[29.288,54.478,-4.064],
[17.312,56.572,-4.842],[22.534,51.220,-6.682],[23.302,53.185,-3.458],
[32.963,57.601,-22.939],[31.796,55.714,-21.120],[31.079,58.037,-18.493],
[31.528,59.264,-9.906],[27.551,-58.831,-8.117],[14.568,57.569,-5.427],
[34.190,51.426,-4.297],[32.119,55.459,-18.426],[17.460,55.903,-14.304],
[15.517,-59.035,-10.961],[34.412,53.960,-6.186],[34.153,53.961,-20.822],
[15.776,53.481,-18.276],[15.234,53.475,-8.508],[31.837,53.931,-7.718],
[25.535,54.368,-12.609],[30.438,52.183,-12.692],[32.994,52.971,-13.783],
[18.594,53.137,-15.536],[18.861,51.696,-17.648],[23.500,52.021,-16.130],
[27.981,53.089,-13.539],[30.392,53.604,-17.804],[20.145,54.548,-22.372],
[23.861,55.304,-18.772],[18.132,53.488,-5.099],[20.132,52.373,-3.556],
[27.369,50.182,-3.631],[26.199,-49.738,3.793],[24.698,-49.345,6.764],
[17.668,-50.693,7.439],[30.202,-47.229,4.005],[17.265,-46.436,6.870],
[17.620,-53.078,1.316],[15.362,-51.320,4.680],[28.515,-42.101,6.035],
[24.172,-46.615,7.326],[32.703,-56.448,6.971],[28.633,-53.915,6.670],
[21.033,-56.154,3.893],[29.265,-54.769,3.108],[17.554,-53.692,4.471],
[23.129,-50.802,4.812],[34.063,-55.111,9.134],[18.785,-56.937,9.197],
[27.784,-47.808,2.585],[21.552,-49.897,10.050],[30.376,-56.225,5.181],
[18.700,-54.552,7.162],[19.084,-47.025,9.057],[16.656,-46.036,0.591],
[28.566,-46.458,7.514],[31.345,-55.492,9.711],[23.760,-58.031,7.067],
[28.477,-45.017,5.189],[29.498,-44.666,2.826],[30.808,-47.356,11.348],
[25.800,-49.803,0.950],[26.661,-51.393,-2.642],[33.413,-52.088,8.957],
[29.972,-49.502,2.229],[27.028,-52.722,10.922],[20.340,-49.452,4.012],
[20.460,-44.966,6.127],[21.951,-49.532,7.071],[26.888,-44.548,7.563],
[21.914,-46.354,11.663],[28.386,-50.714,0.128],[19.476,-51.140,11.278],
[25.381,-44.215,1.643],[20.250,-42.327,12.717],[31.791,-55.601,-1.532],
[18.903,-50.710,1.557],[33.051,-44.379,-3.397],[14.413,-49.010,9.508],
[15.126,-44.551,6.677],[35.284,-45.071,12.187],[25.556,-45.481,12.324],
[24.526,-41.650,1.367],[19.281,-40.273,3.045],[28.305,-41.974,-1.724],
[27.239,-42.395,0.741],[31.062,-51.330,13.064],[17.016,-48.372,11.279],
[20.170,-56.713,7.054],[25.412,58.972,8.729],[28.830,-56.731,11.229],
[29.411,-51.209,-2.471],[26.048,-55.964,-2.872],[20.183,-51.830,7.331],
[21.893,-53.924,2.922],[26.741,-53.654,0.970],[25.477,-52.491,13.727],
[27.645,-52.497,3.544],[18.540,-53.076,9.413],[23.439,-56.618,-2.707],
[15.199,-54.006,3.589],[14.922,-53.900,0.668],[25.881,-51.272,8.657],
[23.842,-54.634,4.695],[18.468,-56.044,2.806],[20.896,-52.503,0.710],
[28.371,-56.195,8.105],[19.673,-51.607,-1.646],[31.052,-41.749,14.775],
[18.073,-46.975,13.240],[26.014,-53.183,6.557],[17.763,-58.257,1.131],
[23.561,-53.100,8.771],[23.564,-47.878,-0.478],[30.088,-49.083,7.442],
[28.210,-56.172,-1.360],[24.693,-47.768,11.419],[29.786,-54.659,0.113],
[18.647,-47.601,4.746],[25.345,-47.091,1.320],[16.343,-51.114,10.802],
[22.644,-52.100,12.662],[26.711,-45.455,9.982],[33.669,-54.448,0.165],
[33.546,-46.282,10.409],[32.438,-45.293,1.044],[27.809,-50.817,7.030],
[20.483,-46.471,2.511],[16.991,-50.479,-2.517],[23.091,-48.432,2.452],
[24.011,-43.385,10.812],[21.234,-41.092,-0.220],[26.387,-47.802,8.713],
[28.871,-46.899,-1.711],[27.630,-42.159,11.111],[32.517,-42.082,-2.161],
[17.825,-56.393,11.647],[30.238,-46.917,0.611],[16.891,-45.761,9.962],
[23.263,-45.961,-2.540],[21.076,-46.509,-0.848],[18.424,-48.440,-0.672],
[25.915,-55.054,9.417],[27.574,-56.200,1.243],[23.154,-44.805,-0.140],
[24.396,-54.734,2.021],[24.469,-54.468,11.771],[25.975,-57.264,3.018],
[20.942,-43.793,-1.509],[22.077,-55.319,-0.117],[32.966,-54.004,11.449],
[30.534,-51.314,6.208],[19.199,-45.932,-2.537],[17.193,-53.283,-2.087],
[30.116,-44.264,7.369],[17.384,58.573,9.297],[18.342,-44.708,2.769],
[16.474,-42.379,12.697],[22.729,-42.849,13.888],[26.437,-44.136,-1.746],
[28.668,-45.409,11.729],[30.381,-43.848,-2.625],[15.098,-47.457,13.799],
[22.074,-48.684,-2.473],[22.553,-49.118,14.136],[26.175,-46.797,-2.422],
[27.810,-48.290,12.373],[17.618,-51.856,13.120],[20.828,-49.064,0.592],
[21.735,-53.540,10.800],[23.838,-50.684,-2.558],[25.388,-49.546,13.491],
[31.423,-49.472,-3.065],[22.479,-55.393,13.466],[21.770,-57.387,9.211],
[23.865,-50.930,10.329],[28.940,-52.832,14.468],[27.558,-50.995,13.057],
[20.012,-53.161,13.989],[23.700,-58.739,10.248],[24.803,-55.128,14.535],
[27.033,-54.505,13.044],[28.205,-56.619,15.675],[18.732,-56.491,16.743],
[22.282,-58.914,12.759],[23.869,-58.444,16.047],[23.018,-41.067,-2.585],
[16.647,-44.497,-2.244],[16.707,-40.849,1.716],[16.460,-43.991,4.464],
[32.850,-41.956,6.848],[32.267,-43.897,11.889],[31.969,-46.353,13.383],
[32.712,-47.899,0.104],[32.505,-50.288,1.359],[31.629,-47.201,8.322],
[30.723,-45.141,9.919],[33.445,-51.847,-0.602],[31.767,-53.197,-2.694],
[30.837,-52.304,3.143],[15.361,-48.542,6.448],[32.239,-49.011,13.154],
[33.516,-50.955,14.624],[30.869,-55.254,-4.739],[31.622,-54.497,2.047],
[31.298,-53.592,7.359],[30.391,-52.403,9.549],[15.467,-55.362,11.844],
[35.392,-59.601,14.254],[32.129,-57.188,2.258],[15.742,-56.224,5.149],
[16.501,-53.157,7.163],[33.387,-57.786,13.919],[34.624,-56.973,-1.230],
[15.402,-57.036,0.653],[29.211,-41.010,1.763],[16.212,-57.826,9.997],
[18.486,-44.477,12.060],[21.654,-45.756,8.425],[25.600,-42.547,13.314],
[29.850,-39.888,12.490],[32.004,-58.811,11.880],[18.197,-40.008,8.655],
[19.661,-42.307,8.858],[24.668,-40.477,7.163],[26.950,-41.807,8.657],
[30.465,-42.517,9.413],[17.401,-43.682,7.589],[19.235,-42.666,4.798],
[22.024,-41.894,5.794],[24.314,-55.359,7.209],[30.259,-58.905,5.023],
[33.358,59.271,3.386],[19.526,59.165,2.828],[17.695,-58.437,-1.768],
[23.607,-57.844,4.365],[27.730,-58.150,5.894],[28.024,-58.792,1.817],
[17.995,-43.126,0.275],[20.705,-54.537,-2.312],[22.755,-57.290,1.506],
[25.508,-41.628,-1.191],[32.195,-42.388,0.730],[17.440,-57.609,14.044],
[21.019,-58.090,14.906],[18.071,-41.826,-2.552],[28.566,-59.429,15.113],
[29.596,-43.326,12.846],[26.928,-31.327,3.054],[26.425,-32.086,5.615],
[18.964,-32.892,7.637],[29.669,-28.062,5.395],[18.051,-28.547,7.061],
[17.264,-32.963,1.700],[15.691,-31.414,3.543],[25.460,-22.429,4.846],
[25.115,-28.527,5.974],[33.319,-37.487,9.214],[28.509,-34.804,8.180],
[21.043,-35.726,4.632],[29.283,-36.138,3.387],[17.080,-33.495,5.772],
[23.629,-32.210,4.262],[32.895,-35.553,11.203],[19.218,-38.130,6.745],
[26.791,-28.341,2.269],[21.937,-28.617,8.227],[28.658,-37.433,5.446],
[18.788,-35.436,6.301],[19.288,-28.640,9.532],[16.420,-27.642,0.288],
[28.909,-28.130,7.900],[29.898,-37.412,8.183],[23.963,-40.430,4.457],
[28.180,-24.230,4.088],[29.031,-26.704,2.862],[30.513,-29.107,11.112],
[24.705,-30.033,1.586],[25.558,-32.214,-4.106],[32.368,-33.015,9.994],
[29.386,-31.266,1.483],[27.938,-33.212,10.281],[20.809,-31.029,3.186],
[21.779,-27.186,5.474],[21.830,-30.705,6.281],[27.635,-25.802,7.372],
[20.795,-27.161,11.728],[26.628,-32.069,-0.424],[18.981,-32.990,11.970],
[25.463,-24.797,3.179],[20.405,-25.378,13.635],[30.804,-39.785,-2.415],
[19.493,-31.827,0.499],[32.203,-26.299,-2.858],[14.533,-27.722,9.012],
[15.738,-26.253,6.607],[32.475,-26.542,12.558],[24.163,-26.729,11.057],
[24.914,-22.583,0.839],[18.604,-23.844,4.252],[27.885,-23.189,-1.480],
[27.844,-23.908,1.058],[31.044,-34.244,12.884],[17.682,-26.557,11.284],
[21.422,-39.885,7.342],[25.423,-40.708,11.129],[28.445,-35.500,11.733],
[30.081,-32.489,-3.319],[25.340,-35.471,-2.765],[21.203,-34.398,7.040],
[20.203,-33.537,2.951],[25.094,-33.059,1.534],[24.852,-33.606,11.683],
[26.490,-35.482,4.936],[17.513,-35.016,11.034],[20.387,-36.473,-2.682],
[15.550,-37.870,3.580],[15.894,-36.318,1.324],[25.522,-32.194,8.568],
[22.427,-37.852,3.045],[18.181,-35.865,2.572],[20.575,-35.422,0.959],
[26.548,-37.120,7.631],[20.680,-33.586,-2.113],[29.856,-21.827,13.471],
[18.456,-29.023,12.444],[24.386,-34.379,6.510],[20.322,-38.636,1.160],
[23.536,-34.052,9.059],[22.990,-28.981,-1.769],[31.484,-28.871,8.243],
[27.452,-37.298,-2.119],[25.506,-28.964,12.167],[26.988,-34.781,-0.565],
[17.937,-30.747,5.336],[25.542,-28.292,-0.372],[16.269,-32.755,10.699],
[23.512,-31.560,12.817],[27.063,-27.416,10.234],[33.856,-34.414,-1.269],
[34.288,-29.850,11.347],[30.197,-26.105,0.546],[28.388,-31.545,8.223],
[20.975,-25.987,2.491],[16.305,-30.822,-1.802],[22.478,-29.085,3.767],
[23.605,-24.522,9.790],[21.429,-23.129,3.342],[25.329,-29.056,8.930],
[27.748,-30.613,-2.890],[29.936,-24.230,11.746],[32.481,-23.342,-3.245],
[16.263,-37.476,10.851],[28.957,-29.873,-0.706],[17.382,-25.926,8.681],
[22.547,-27.404,-4.325],[22.065,-26.686,-1.165],[18.122,-29.403,-0.679],
[24.884,-37.171,9.773],[27.578,-37.472,0.645],[23.872,-26.170,1.076],
[24.888,-36.154,1.749],[23.590,-35.811,11.546],[25.476,-38.702,2.463],
[23.075,-24.304,-0.682],[21.657,-38.570,-1.801],[30.508,-36.997,10.785],
[30.184,-32.614,6.202],[19.315,-27.118,0.432],[17.858,-35.455,-2.657],
[30.110,-24.066,8.390],[17.457,-40.279,5.949],[18.199,-26.911,4.426],
[15.478,-24.524,14.764],[22.871,-24.727,14.022],[25.011,-26.647,-5.174],
[26.483,-26.391,13.719],[29.621,-25.857,-2.053],[16.308,-29.370,13.980],
[19.674,-30.619,-3.265],[22.684,-29.069,11.602],[25.383,-28.594,-3.373],
[27.997,-30.398,10.736],[17.641,-33.166,14.896],[23.911,-31.869,-1.175],
[21.310,-33.829,10.517],[23.262,-33.569,-3.271],[26.586,-31.269,13.119],
[30.600,-30.018,-3.236],[19.882,-36.051,11.904],[21.893,-37.500,8.577],
[23.035,-30.934,9.239],[28.912,-35.560,14.778],[28.631,-32.886,12.860],
[20.241,-34.392,14.880],[23.171,-40.267,9.399],[23.651,-35.906,14.624],
[25.935,-36.010,12.885],[28.786,-38.272,14.449],[17.842,-36.157,13.542],
[22.340,-41.252,11.724],[22.816,-39.220,14.855],[22.907,-21.452,-1.746],
[16.044,-28.481,-2.754],[16.548,-22.662,1.982],[15.054,-24.096,5.407],
[33.385,-24.150,8.895],[32.394,-23.470,11.421],[30.284,-28.478,13.784],
[30.334,-28.680,1.320],[32.527,-31.576,1.713],[32.553,-28.610,5.823],
[31.414,-26.493,9.559],[31.862,-31.976,-0.886],[30.597,-34.499,-1.135],
[30.415,-33.307,3.096],[15.572,-29.305,6.792],[32.046,-31.279,12.094],
[33.448,-32.050,14.433],[29.795,-37.335,-3.201],[28.911,-35.048,1.138],
[31.873,-35.375,4.075],[31.349,-34.861,8.403],[13.974,-35.934,12.425],
[33.751,-41.798,15.855],[31.440,-36.150,1.588],[15.974,-36.223,6.113],
[14.414,-33.117,5.384],[32.755,-39.120,13.059],[33.228,-36.529,-3.180],
[15.327,-38.830,0.311],[32.394,-25.073,2.317],[15.403,-39.844,8.213],
[17.675,-26.099,14.034],[19.765,-25.284,10.169],[25.067,-24.058,12.613],
[26.796,-22.101,12.259],[31.596,-39.492,10.381],[17.480,-20.599,11.127],
[19.068,-22.331,9.028],[24.570,-23.442,7.382],[26.990,-23.774,8.959],
[29.416,-21.415,7.819],[17.535,-23.906,6.863],[19.811,-25.759,6.149],
[23.012,-24.737,5.400],[24.866,-37.889,5.504],[30.772,-40.055,7.176],
[31.424,-40.266,3.559],[18.889,-37.898,4.170],[15.551,-41.783,-1.202],
[22.417,-40.664,2.296],[28.581,-39.908,4.246],[29.018,-39.384,-0.574],
[16.927,-25.146,0.662],[17.587,-38.060,-2.723],[23.026,-37.415,0.156],
[25.761,-22.988,-3.021],[32.173,-24.452,-0.507],[15.885,-39.728,12.924],
[19.194,-42.191,15.326],[18.908,-22.418,-2.488],[27.823,-40.665,13.878],
[31.393,-24.247,14.349],[26.718,-13.261,3.330],[26.330,-12.638,6.343],
[18.890,-13.433,8.752],[30.852,-9.981,4.331],[19.607,-8.915,5.383],
[16.607,-15.079,2.641],[16.759,-12.132,4.114],[25.659,-5.187,5.542],
[24.303,-7.928,7.049],[32.243,-19.198,8.232],[27.637,-16.152,6.976],
[20.732,-17.416,4.997],[27.645,-19.163,3.340],[15.219,-16.359,4.550],
[22.051,-14.126,4.223],[33.647,-16.724,9.926],[18.553,-18.760,9.391],
[25.623,-9.596,1.551],[21.767,-13.060,9.000],[30.579,-19.436,3.951],
[18.687,-16.504,7.618],[17.256,-10.949,9.153],[17.211,-8.160,0.020],
[27.423,-10.273,5.881],[30.228,-19.818,9.943],[23.238,-20.950,6.288],
[28.678,-8.322,3.960],[28.013,-10.819,1.814],[30.105,-9.917,11.143],
[23.944,-13.329,2.562],[26.925,-13.573,-2.977],[34.190,-14.364,8.565],
[28.734,-13.721,1.490],[27.554,-15.020,10.914],[20.491,-11.546,4.140],
[21.331,-6.982,6.565],[19.973,-13.662,6.024],[26.889,-7.807,8.108],
[22.088,-8.385,11.804],[29.634,-14.537,-1.639],[20.765,-13.492,11.744],
[24.478,-7.488,3.319],[18.877,-5.862,12.202],[32.339,-21.492,-0.948],
[18.914,-13.371,1.616],[32.843,-6.471,-2.093],[14.320,-11.207,8.876],
[15.532,-7.293,7.041],[34.237,-8.411,11.899],[24.636,-9.234,9.420],
[24.596,-2.298,1.187],[18.533,-5.316,3.095],[29.066,-5.343,-1.140],
[26.545,-4.306,0.838],[30.441,-14.831,14.168],[18.594,-9.301,10.997],
[20.585,-20.406,7.900],[25.010,-22.195,10.150],[29.310,-17.273,12.040],
[29.640,-12.119,-2.599],[25.425,-19.669,-1.532],[21.588,-16.328,7.219],
[19.967,-16.198,2.633],[25.855,-14.986,1.387],[25.827,-15.441,13.169],
[29.040,-15.895,3.775],[19.426,-16.742,11.228],[21.144,-18.110,-2.808],
[15.906,-19.133,4.812],[16.729,-17.922,2.073],[26.901,-14.119,8.515],
[22.892,-19.058,4.485],[18.097,-16.653,4.682],[20.956,-16.883,0.052],
[28.348,-18.575,8.364],[20.668,-14.929,-1.763],[32.627,-3.491,13.815],
[19.248,-10.891,14.703],[24.802,-15.022,5.915],[21.292,-19.956,2.342],
[23.747,-16.864,8.488],[23.649,-10.550,-0.035],[30.573,-9.673,6.980],
[26.087,-16.637,-1.604],[27.051,-10.238,10.932],[30.651,-17.215,-1.051],
[18.193,-11.379,6.541],[26.782,-11.431,-0.871],[16.538,-15.343,11.669],
[23.623,-13.201,12.612],[26.132,-7.407,11.061],[33.452,-15.551,0.021],
[33.021,-10.855,11.160],[31.564,-7.747,-0.205],[29.461,-12.545,6.887],
[20.976,-9.012,2.979],[17.131,-13.713,-1.670],[23.148,-9.538,4.617],
[22.902,-6.820,9.269],[21.877,-6.128,2.776],[28.567,-10.677,8.534],
[29.182,-9.510,-1.431],[29.360,-6.537,11.141],[33.641,-4.354,-0.215],
[17.464,-18.201,12.047],[30.991,-10.278,0.225],[17.239,-7.760,9.437],
[23.474,-8.737,-4.322],[21.640,-7.981,-1.670],[17.293,-11.171,-0.531],
[25.011,-19.545,10.268],[28.219,-19.184,0.358],[24.175,-7.766,-0.319],
[26.102,-17.474,1.078],[24.388,-16.964,11.041],[25.768,-20.290,1.500],
[22.368,-5.927,0.227],[21.642,-19.369,-0.757],[30.558,-19.490,12.580],
[30.283,-15.135,6.322],[19.337,-9.836,-1.972],[16.969,-16.733,-1.855],
[30.395,-6.983,6.342],[19.595,-21.671,5.834],[18.337,-8.179,2.804],
[15.754,-5.966,12.596],[21.286,-5.774,14.129],[25.382,-6.681,-3.955],
[27.840,-8.244,13.227],[30.126,-7.162,-2.758],[16.253,-12.433,13.498],
[20.557,-9.080,-4.235],[23.682,-10.547,11.925],[26.754,-8.091,-1.878],
[28.316,-12.559,10.591],[18.870,-13.220,13.607],[21.954,-12.688,-1.076],
[21.804,-15.794,10.002],[22.843,-13.757,-3.366],[26.897,-10.755,13.727],
[32.909,-9.330,-2.394],[21.711,-17.725,11.864],[22.230,-19.769,9.882],
[24.193,-14.462,10.173],[27.727,-15.923,15.055],[28.160,-13.368,13.222],
[20.635,-15.170,15.056],[22.275,-22.570,8.634],[25.136,-17.823,14.500],
[26.668,-18.348,11.944],[28.383,-18.911,14.417],[18.947,-17.746,14.489],
[22.556,-23.059,12.159],[22.799,-21.932,14.685],[21.869,-3.045,-0.637],
[16.391,-9.384,-2.361],[15.680,-4.438,2.963],[16.372,-6.746,4.225],
[32.116,-4.689,6.394],[32.627,-6.218,10.149],[30.965,-9.528,13.601],
[33.539,-11.540,0.276],[30.878,-12.261,2.030],[34.139,-9.459,7.027],
[32.122,-8.909,9.404],[34.035,-12.873,-2.155],[32.231,-15.108,-2.238],
[31.143,-14.987,2.565],[15.478,-12.576,6.871],[31.180,-12.520,12.517],
[33.922,-16.009,14.749],[30.931,-16.708,-3.970],[32.147,-17.754,1.716],
[30.392,-18.579,6.521],[31.190,-15.975,8.833],[14.561,-17.425,12.778],
[34.156,-20.917,16.383],[30.783,-20.076,1.155],[18.367,-19.327,6.730],
[16.034,-16.485,7.425],[33.344,-19.482,14.067],[32.822,-18.613,-2.300],
[15.914,-20.248,0.746],[29.350,-6.021,2.729],[15.383,-20.527,8.647],
[16.884,-9.679,13.471],[20.380,-8.002,9.766],[23.662,-4.440,13.053],
[28.038,-3.727,12.620],[31.947,-21.803,8.939],[18.037,-3.448,8.825],
[20.237,-3.765,10.261],[25.988,-3.325,8.294],[27.497,-5.319,9.721],
[28.803,-3.515,7.822],[17.595,-5.560,6.443],[19.959,-4.933,5.376],
[23.357,-5.243,6.806],[25.473,-19.159,5.603],[31.977,-21.493,6.172],
[32.340,-22.528,3.535],[16.799,-21.965,5.315],[17.065,-23.141,-0.869],
[23.441,-21.496,3.058],[28.275,-21.568,5.173],[29.090,-21.788,2.217],
[18.005,-5.806,-0.195],[19.031,-19.001,-1.541],[23.533,-18.846,1.218],
[26.375,-4.996,-1.820],[31.122,-5.481,0.881],[17.836,-20.398,13.916],
[20.540,-20.584,15.558],[19.727,-4.641,-2.244],[27.598,-21.579,15.002],
[30.548,-5.919,13.285],[25.958,5.934,3.276],[25.548,5.649,6.674],
[19.258,5.013,7.324],[28.055,8.411,6.689],[19.990,8.490,7.154],
[17.926,4.561,3.552],[15.532,4.097,5.670],[27.163,13.818,5.008],
[25.206,9.693,6.938],[32.889,0.093,8.267],[27.930,1.375,7.070],
[21.704,0.500,4.922],[29.038,0.912,2.714],[16.954,1.925,4.829],
[23.559,5.606,4.901],[33.675,1.363,10.877],[19.643,-1.009,8.270],
[26.013,9.114,2.651],[22.207,8.404,8.748],[30.775,0.543,4.719],
[18.554,1.869,7.316],[19.385,8.458,10.236],[17.062,8.855,0.615],
[29.727,8.345,8.744],[29.696,-1.069,8.511],[24.327,-2.686,6.085],
[28.934,11.667,6.609],[28.872,10.473,1.922],[30.176,7.834,11.411],
[23.771,5.810,1.632],[26.650,4.510,-3.238],[34.117,4.428,9.785],
[30.728,4.941,0.815],[29.071,3.779,11.788],[21.398,6.425,3.223],
[22.479,10.462,5.540],[22.713,5.339,7.516],[27.220,10.487,8.568],
[23.417,10.629,12.250],[29.916,6.965,-1.091],[19.898,5.746,11.039],
[27.134,11.936,3.154],[20.974,11.910,12.541],[31.513,-2.269,-0.933],
[21.186,5.257,0.554],[31.237,11.547,-2.227],[16.733,8.302,8.951],
[14.585,10.676,8.654],[34.927,11.502,12.499],[25.970,10.222,12.122],
[24.845,14.824,1.172],[17.520,13.223,3.284],[28.395,13.187,-2.038],
[28.597,13.172,1.292],[31.297,0.652,12.357],[18.464,10.887,11.662],
[21.585,-2.500,6.853],[24.503,-1.463,9.968],[28.305,1.307,11.157],
[29.512,6.148,-3.817],[26.227,1.429,-2.904],[19.951,1.589,9.671],
[22.048,3.332,2.570],[26.267,5.282,0.585],[25.844,3.720,13.710],
[27.227,2.429,4.396],[17.412,3.724,9.460],[21.070,0.487,-2.848],
[13.959,0.415,3.574],[16.544,1.286,-0.311],[27.695,4.372,7.455],
[24.210,-0.342,4.676],[19.511,2.197,3.277],[21.088,2.915,-1.266],
[27.295,-1.190,9.527],[19.859,5.312,-2.235],[31.767,14.950,13.263],
[19.600,6.937,14.031],[23.359,2.735,6.381],[19.791,-1.620,0.872],
[24.900,3.642,8.515],[24.521,9.493,-0.300],[31.524,7.115,6.432],
[28.346,-0.744,-2.147],[22.606,6.964,11.141],[27.873,3.204,0.439],
[18.706,7.227,4.112],[26.778,8.217,0.263],[17.893,5.282,12.535],
[22.749,3.230,13.640],[28.229,9.778,10.983],[33.964,2.906,-1.040],
[33.518,9.359,11.731],[30.535,9.787,-0.155],[28.302,6.241,4.852],
[21.868,10.866,2.819],[17.415,7.704,-2.208],[22.957,8.235,2.704],
[22.727,12.827,10.788],[21.289,13.516,1.332],[26.563,7.982,9.261],
[27.787,7.446,-2.331],[29.907,12.375,12.285],[32.895,14.003,-1.588],
[17.081,1.113,10.080],[29.122,7.616,2.067],[17.803,10.676,7.797],
[22.609,9.824,-3.803],[21.684,9.662,-0.947],[19.741,7.779,-0.588],
[25.806,1.436,9.843],[27.003,-1.348,1.147],[25.449,12.458,-0.329],
[24.294,2.469,0.871],[24.379,3.719,11.236],[25.836,0.945,2.290],
[21.115,12.283,-1.677],[22.148,0.495,0.020],[30.724,-1.918,12.251],
[30.541,4.742,5.412],[19.190,10.742,-0.154],[17.493,3.260,-1.853],
[32.100,12.069,7.671],[18.525,-2.669,6.099],[18.705,10.497,2.709],
[16.033,12.615,14.077],[21.839,12.587,16.277],[26.560,11.193,-2.886],
[27.407,11.756,13.629],[29.161,11.335,-3.708],[16.086,7.166,14.798],
[18.966,9.415,-3.391],[24.387,7.179,13.094],[25.254,8.661,-3.076],
[28.442,6.100,10.125],[17.470,3.448,14.192],[21.903,7.213,-2.395],
[21.527,4.049,9.656],[23.404,4.709,-1.191],[26.733,6.697,14.010],
[32.203,6.273,-3.316],[21.806,1.530,11.720],[22.016,-0.377,9.602],
[24.316,6.444,9.281],[28.603,2.777,14.433],[31.193,5.348,12.969],
[20.281,3.783,14.334],[23.278,-3.593,9.058],[24.665,0.786,14.476],
[26.031,-0.065,11.975],[28.399,-0.966,13.967],[18.782,1.205,13.351],
[22.487,-2.087,11.712],[21.958,-2.168,14.437],[22.820,14.772,-1.127],
[16.112,11.086,-3.812],[16.094,14.320,1.312],[15.311,13.800,5.100],
[33.894,14.256,8.175],[33.209,12.433,10.570],[30.862,10.053,15.177],
[32.458,7.744,-0.399],[33.602,5.304,0.493],[31.068,9.848,6.502],
[30.709,10.933,9.879],[30.860,4.163,-1.605],[32.375,2.879,-3.361],
[30.981,3.151,3.143],[14.260,6.190,7.239],[31.454,5.595,10.308],
[33.407,3.790,12.567],[31.192,0.448,-2.258],[30.347,2.143,0.226],
[32.021,2.405,6.092],[31.202,2.653,10.435],[15.763,1.064,12.425],
[34.962,-3.441,15.856],[30.416,-1.451,2.419],[18.605,-0.103,5.308],
[14.932,3.384,8.280],[33.952,-1.076,12.538],[33.470,-1.028,-2.129],
[16.867,-1.284,0.708],[31.453,12.763,3.340],[16.504,-1.728,10.185],
[18.712,12.132,14.386],[21.360,12.333,7.671],[25.311,14.912,12.669],
[27.929,14.299,12.465],[32.114,-2.999,10.105],[18.553,16.162,11.721],
[18.817,12.803,9.606],[25.715,15.074,7.074],[28.006,12.782,9.868],
[29.571,15.138,9.611],[17.008,14.771,8.605],[18.117,11.484,5.475],
[20.168,13.837,4.748],[25.808,-0.524,7.272],[30.113,-2.623,5.711],
[31.295,-4.136,3.583],[17.508,-2.150,3.407],[16.662,-3.751,-0.827],
[23.695,-3.894,3.464],[27.127,-2.842,5.131],[29.038,-2.779,-0.380],
[16.644,11.514,1.231],[18.057,-0.266,-2.251],[23.988,-1.260,-1.371],
[25.252,13.833,-2.510],[31.635,12.235,0.627],[17.410,-3.017,12.576],
[19.305,-1.404,13.710],[19.555,13.958,-3.040],[29.233,-3.031,15.539],
[32.361,11.751,13.221],[27.940,26.078,4.635],[26.235,24.501,7.343],
[18.453,24.087,7.872],[30.224,27.367,5.602],[18.626,28.535,6.234],
[16.709,23.260,2.377],[16.198,23.456,4.918],[27.353,32.094,5.362],
[23.338,25.377,6.916],[32.438,18.123,7.909],[28.228,20.363,8.173],
[21.358,17.931,5.742],[28.371,18.114,3.128],[17.724,21.184,4.886],
[23.173,23.497,4.770],[33.871,20.012,9.571],[18.381,17.811,8.462],
[25.763,28.050,2.710],[20.990,25.302,8.604],[31.173,18.348,4.972],
[19.173,21.345,7.402],[20.672,28.789,8.003],[17.284,26.785,1.444],
[30.344,27.824,8.730],[29.883,18.257,7.976],[22.861,15.798,6.731],
[28.102,29.659,5.189],[28.574,28.711,2.115],[30.177,28.632,11.626],
[26.280,25.481,2.638],[26.537,22.902,-1.749],[32.406,23.056,9.624],
[29.005,24.475,0.960],[28.320,23.040,8.625],[21.174,24.779,2.935],
[22.471,28.038,5.696],[21.214,23.275,6.712],[27.541,29.972,8.838],
[20.777,28.231,10.771],[30.881,22.671,-0.651],[19.724,23.868,11.686],
[25.429,30.358,4.374],[19.641,31.097,11.516],[31.247,14.875,0.943],
[19.370,23.190,1.208],[31.749,28.550,-1.801],[14.933,25.936,9.656],
[14.359,29.761,6.666],[33.881,29.995,12.549],[23.636,28.364,10.402],
[24.828,33.387,0.792],[19.826,32.623,3.219],[27.958,32.029,-1.947],
[27.627,32.946,2.628],[31.362,22.232,14.080],[18.024,28.618,11.197],
[20.347,15.711,7.475],[25.981,15.290,10.238],[29.238,21.203,12.692],
[30.548,25.262,-1.894],[25.380,17.527,-2.888],[21.109,19.990,8.901],
[22.227,21.691,1.407],[26.966,22.921,1.876],[26.689,21.673,12.943],
[28.268,20.965,3.794],[19.002,21.380,10.640],[20.963,19.209,-2.396],
[16.461,18.538,4.203],[15.229,19.801,-0.760],[25.680,21.698,8.353],
[23.417,19.099,3.877],[19.175,19.333,3.494],[19.790,20.339,1.244],
[27.919,18.691,5.899],[20.973,21.984,-1.645],[30.273,35.687,14.092],
[18.819,26.587,13.136],[25.783,21.537,5.519],[20.721,17.508,2.222],
[23.067,21.633,8.077],[24.708,26.377,0.425],[31.842,25.653,7.309],
[27.921,17.744,-1.549],[26.026,27.638,11.637],[28.538,20.238,-0.819],
[19.424,24.816,5.046],[27.162,27.444,-0.013],[15.691,23.931,11.276],
[24.047,22.065,14.014],[27.705,29.979,11.516],[33.496,21.496,-0.273],
[34.369,26.481,11.877],[31.110,29.626,1.176],[28.964,25.519,7.509],
[21.371,30.709,1.596],[18.531,23.596,-1.858],[23.334,26.613,3.595],
[23.947,31.414,8.787],[22.643,31.308,3.846],[27.830,27.218,9.665],
[29.112,27.948,-1.722],[30.259,31.324,12.320],[32.563,31.633,-1.756],
[17.734,18.861,11.867],[30.393,27.311,0.356],[16.782,29.223,8.084],
[21.752,29.184,-3.542],[22.040,26.782,-1.826],[19.544,25.818,-0.794],
[26.109,18.952,10.151],[26.747,19.530,1.383],[25.993,30.894,-0.113],
[24.956,21.875,3.173],[24.774,20.909,11.203],[26.988,15.327,2.739],
[22.844,31.382,-0.522],[22.564,17.234,0.210],[31.040,19.374,12.458],
[30.176,23.234,6.326],[20.649,29.606,-0.966],[17.127,20.942,-2.475],
[29.519,32.152,7.103],[17.788,16.398,6.198],[17.806,31.480,1.797],
[15.417,31.751,13.555],[20.008,31.262,15.416],[26.284,30.298,-3.600],
[27.478,28.516,13.876],[28.934,29.575,-3.895],[16.088,25.366,14.078],
[19.844,26.289,-3.495],[22.029,27.470,13.088],[24.513,27.246,-4.193],
[28.444,24.950,11.269],[17.919,22.657,14.169],[22.966,24.335,-0.415],
[22.103,23.603,10.390],[23.313,24.577,-3.297],[27.256,25.730,13.993],
[33.330,27.180,-3.679],[21.962,20.179,11.836],[23.442,18.998,9.667],
[24.743,23.994,9.689],[29.675,22.127,16.138],[29.164,23.872,13.931],
[20.559,22.028,13.853],[22.561,17.266,12.159],[24.838,19.574,14.744],
[26.858,19.009,12.710],[29.231,18.288,14.195],[19.851,19.518,13.131],
[22.834,13.982,13.409],[23.518,17.908,17.053],[22.779,33.625,-1.894],
[16.311,27.226,-1.282],[15.199,34.427,1.425],[16.149,32.957,3.442],
[32.771,31.880,6.520],[31.685,32.435,9.706],[30.862,27.426,14.424],
[33.058,26.417,0.145],[32.245,24.148,2.519],[34.094,27.211,7.317],
[33.141,28.698,9.707],[32.924,23.783,-1.861],[32.178,21.216,-3.009],
[30.603,20.039,2.756],[16.936,25.899,6.014],[31.294,24.489,12.099],
[33.844,22.488,13.329],[30.386,19.061,-2.743],[32.386,18.958,1.046],
[30.464,20.466,6.582],[30.587,21.322,10.197],[15.401,19.106,12.991],
[33.850,15.735,15.434],[30.651,17.202,2.138],[17.052,19.649,7.008],
[16.236,22.328,7.712],[32.095,17.429,14.204],[32.941,18.451,-2.873],
[15.158,17.456,1.531],[29.392,30.779,3.254],[16.289,16.854,9.978],
[16.692,28.311,13.829],[19.717,32.460,9.162],[24.587,32.844,12.669],
[29.008,34.696,11.942],[32.227,15.235,10.701],[17.466,35.038,10.111],
[19.806,34.610,11.275],[26.135,32.999,7.574],[26.876,32.353,10.231],
[29.059,34.008,9.244],[16.573,32.496,6.655],[18.863,29.820,3.540],
[22.031,30.582,6.351],[25.557,17.258,5.403],[32.307,14.898,6.158],
[32.475,15.377,3.549],[17.655,16.143,3.277],[16.585,14.395,-1.380],
[22.379,16.124,3.808],[30.020,15.435,4.943],[28.233,16.157,0.623],
[16.791,30.449,-0.425],[17.925,18.023,-0.995],[24.909,19.148,-0.655],
[25.656,33.274,-1.907],[32.281,31.690,2.619],[16.539,16.732,13.559],
[20.531,14.379,14.938],[19.882,33.132,-3.210],[27.610,16.306,14.841],
[31.990,31.299,14.288],[27.867,42.371,4.183],[25.496,42.857,5.666],
[18.550,41.696,6.965],[28.295,46.072,5.348],[19.418,47.028,6.618],
[16.378,41.742,2.331],[17.112,43.516,4.778],[26.583,52.146,5.424],
[24.430,46.024,7.487],[31.782,36.896,8.418],[27.237,39.067,7.266],
[21.497,37.623,3.995],[28.716,36.922,2.213],[17.132,40.664,4.731],
[23.100,42.961,4.444],[33.060,39.820,9.409],[18.629,37.803,9.879],
[26.098,46.796,3.172],[21.992,44.395,9.874],[30.757,37.042,4.896],
[17.906,38.848,7.110],[18.607,46.209,9.186],[17.789,46.983,1.506],
[29.950,47.845,7.184],[28.765,37.202,9.515],[23.409,33.549,6.856],
[28.624,49.877,5.200],[28.646,47.216,2.599],[29.818,45.991,11.657],
[26.045,44.172,0.714],[26.926,41.422,-2.655],[33.013,42.955,10.220],
[29.139,43.832,2.394],[26.560,42.163,10.397],[20.138,42.065,3.646],
[22.382,47.563,5.808],[21.183,43.662,7.638],[26.794,47.650,7.608],
[22.057,46.564,11.444],[28.035,43.704,-1.474],[19.218,43.530,10.224],
[26.659,49.202,1.506],[20.546,49.284,12.859],[32.601,34.609,-1.642],
[20.641,42.290,0.440],[32.085,49.286,-1.929],[16.117,45.032,9.209],
[16.763,46.533,6.929],[34.711,48.530,12.540],[25.041,48.404,10.875],
[25.733,51.878,1.745],[19.174,51.339,3.249],[29.022,50.907,-1.489],
[28.782,50.787,1.216],[31.369,40.336,12.297],[18.655,46.628,12.009],
[20.428,34.870,6.880],[24.415,33.802,10.285],[28.886,39.441,11.131],
[29.535,42.030,-3.473],[24.498,35.928,-1.811],[20.749,40.803,8.266],
[22.028,40.196,4.196],[26.087,41.822,-0.044],[26.483,40.370,12.509],
[25.899,40.092,4.801],[19.022,40.757,11.627],[22.018,36.555,-3.042],
[15.072,37.436,3.164],[15.456,37.499,-1.348],[25.507,41.145,7.861],
[23.545,35.673,3.433],[19.021,40.284,1.584],[21.415,39.003,1.634],
[26.943,36.320,7.658],[21.087,39.114,-2.271],[30.526,51.549,12.706],
[18.335,45.343,14.376],[23.453,40.585,6.317],[19.147,37.492,1.913],
[23.253,39.283,8.729],[22.935,43.409,-0.999],[30.849,44.579,7.895],
[26.864,37.065,-2.099],[25.873,45.636,10.516],[27.828,39.745,0.011],
[20.073,44.368,5.463],[26.315,46.805,-0.189],[16.403,42.253,11.334],
[24.272,40.286,14.024],[27.523,47.572,11.953],[33.590,40.379,-0.251],
[32.907,45.821,11.449],[31.177,47.429,1.786],[28.545,43.721,6.399],
[22.393,46.799,3.291],[17.146,43.791,-0.295],[21.652,44.408,2.051],
[24.376,49.235,8.310],[21.378,49.379,2.686],[28.524,46.638,9.277],
[29.376,46.098,-1.999],[29.786,49.339,11.438],[33.971,50.360,-0.503],
[15.632,37.446,10.877],[28.703,45.542,0.341],[19.383,48.752,10.381],
[22.632,47.586,-2.557],[23.297,46.320,0.054],[19.560,44.931,-0.288],
[24.704,36.860,9.490],[26.724,35.126,0.622],[25.655,49.329,-0.936],
[25.243,39.653,2.179],[23.811,38.762,11.517],[25.961,37.018,3.226],
[22.108,49.171,-0.325],[21.526,37.179,-0.345],[32.411,37.815,11.098],
[29.879,41.284,6.239],[19.521,48.246,-0.246],[17.499,40.015,-2.236],
[30.687,50.773,7.477],[16.290,35.604,4.956],[17.934,50.843,0.919],
[16.744,51.389,13.723],[21.723,49.174,15.603],[24.943,48.349,-3.901],
[28.775,48.801,13.972],[29.640,48.688,-3.144],[15.734,43.451,13.845],
[19.955,44.471,-4.022],[22.998,43.920,12.707],[25.937,45.730,-2.871],
[29.053,43.848,9.493],[18.187,42.755,13.618],[19.042,42.295,-1.803],
[21.656,40.647,11.081],[22.049,42.632,-3.248],[26.136,43.580,12.874],
[31.703,45.132,-2.623],[20.760,37.032,11.832],[21.501,37.289,9.067],
[24.195,41.459,11.044],[27.364,39.286,14.666],[28.887,42.344,12.792],
[20.854,39.943,13.640],[21.827,34.634,9.260],[23.384,37.296,13.699],
[26.496,37.714,12.008],[27.214,35.825,14.211],[17.927,38.176,13.096],
[21.442,33.167,12.632],[23.597,34.758,14.365],[23.457,51.678,-0.084],
[16.371,47.156,-2.615],[14.592,52.226,3.303],[16.804,50.066,3.289],
[33.618,51.178,7.138],[32.748,50.316,11.981],[30.890,46.264,14.380],
[33.003,45.214,0.354],[32.144,42.824,2.610],[33.304,45.592,8.132],
[31.714,48.671,9.031],[32.606,42.247,-1.831],[33.712,39.892,-2.803],
[30.379,40.950,3.500],[15.270,43.650,6.984],[31.248,43.470,12.591],
[33.031,40.856,14.514],[30.766,38.660,-2.851],[30.311,38.728,-0.078],
[31.315,39.092,7.037],[30.513,41.611,9.214],[14.702,38.845,13.834],
[36.118,35.447,15.556],[31.244,35.788,2.221],[14.496,36.870,6.384],
[15.523,40.780,6.815],[34.679,34.865,13.368],[32.812,36.942,-3.430],
[16.703,36.455,0.904],[30.479,52.572,2.804],[15.758,34.888,8.054],
[18.237,47.956,14.583],[21.473,50.605,9.503],[25.839,51.580,12.924],
[28.696,53.357,12.344],[31.856,35.287,11.004],[18.139,51.885,10.940],
[20.393,51.878,7.206],[24.901,52.578,7.567],[27.210,50.532,8.031],
[29.768,51.707,9.945],[17.494,50.328,7.318],[19.705,49.666,5.591],
[22.759,50.243,6.299],[25.959,34.676,4.883],[31.290,34.501,7.227],
[32.295,34.160,4.816],[18.926,35.081,3.703],[17.980,33.306,-0.598],
[24.549,33.036,3.484],[28.467,35.100,6.053],[29.373,35.367,-1.850],
[16.343,48.840,-0.095],[19.189,35.731,-0.767],[24.551,37.335,0.405],
[24.350,50.861,-2.775],[34.838,49.739,2.028],[17.990,33.009,12.847],
[20.855,35.597,14.747],[18.662,50.773,-1.894],[28.569,33.245,14.326],
[31.654,48.674,13.705],[33.593,54.673,7.804],[28.117,58.164,7.732],
[19.873,55.885,4.022],[29.250,57.701,2.645],[17.087,57.899,4.089],
[33.285,57.560,10.158],[18.686,55.926,9.304],[31.100,55.112,3.966],
[19.297,57.036,6.791],[31.233,55.773,8.753],[24.286,52.343,4.316],
[32.133,52.525,-1.484],[31.167,58.017,13.495],[22.130,53.326,6.262],
[25.872,52.903,10.593],[28.819,56.783,12.675],[26.794,55.187,-2.568],
[21.302,-59.367,6.407],[22.025,57.360,3.181],[26.488,58.931,4.021],
[19.645,-59.417,10.816],[21.965,55.012,-4.590],[15.137,55.037,3.361],
[18.216,58.025,-0.113],[23.037,54.594,3.529],[17.743,55.818,2.120],
[21.720,57.449,-0.603],[27.568,55.268,6.594],[20.958,-59.313,-2.371],
[24.357,58.202,6.136],[20.633,54.207,2.188],[22.818,57.745,8.982],
[29.074,56.200,-1.649],[28.166,57.970,-0.031],[33.141,-59.457,0.347],
[16.797,55.651,11.172],[25.824,56.095,8.577],[27.305,54.999,0.976],
[24.251,56.250,1.822],[24.328,56.731,10.987],[25.807,54.268,3.539],
[22.321,55.028,-1.938],[31.008,55.769,11.686],[18.705,57.852,-2.907],
[17.784,53.134,7.428],[22.611,55.643,12.686],[21.428,55.352,8.732],
[27.287,57.324,16.129],[20.441,58.195,14.094],[21.003,52.930,10.496],
[24.408,57.722,16.200],[26.181,57.791,13.503],[29.613,55.264,15.834],
[17.086,55.225,14.669],[22.864,51.000,11.982],[23.703,53.470,13.920],
[31.799,-59.608,-2.374],[31.473,56.187,-3.258],[31.262,57.502,-0.064],
[30.087,57.161,5.933],[30.648,59.235,9.407],[15.159,57.075,13.433],
[32.066,51.966,14.794],[31.906,54.779,1.042],[15.700,56.271,5.888],
[15.532,58.497,7.193],[33.259,54.368,12.537],[33.786,54.589,-2.746],
[14.840,55.705,0.265],[16.229,53.565,9.765],[32.379,52.662,10.195],
[24.575,54.807,6.250],[31.702,50.294,3.860],[31.780,52.747,5.324],
[18.900,53.410,4.668],[18.022,53.377,0.837],[21.813,51.682,3.876],
[29.359,53.391,5.862],[29.831,53.130,0.068],[17.837,54.014,-2.422],
[24.597,54.165,0.380],[19.287,54.720,11.758],[20.757,52.947,13.355],
[27.941,52.983,15.502],[27.669,-49.684,22.311],[23.798,-49.624,25.631],
[18.914,-49.679,26.039],[30.832,-47.695,23.284],[19.374,-46.652,24.126],
[16.077,-53.124,20.435],[15.575,-50.537,22.339],[26.370,-44.078,25.074],
[25.233,-46.814,26.235],[32.406,-57.417,27.183],[27.555,-53.635,26.493],
[20.956,-55.645,23.377],[26.519,-55.013,23.457],[17.160,-52.850,22.929],
[20.338,-52.443,23.332],[34.142,-52.800,28.573],[18.462,-56.173,26.608],
[26.966,-46.481,21.400],[21.174,-48.483,27.103],[29.652,-55.299,23.287],
[17.528,-53.910,25.434],[19.322,-46.489,27.182],[15.710,-44.585,17.710],
[30.680,-46.590,25.689],[29.926,-55.305,28.042],[21.791,-58.982,23.996],
[28.811,-45.442,23.715],[29.422,-45.888,21.173],[31.718,-48.739,30.085],
[25.745,-50.435,18.406],[26.584,-51.672,15.932],[33.187,-50.277,27.939],
[29.765,-48.896,20.763],[27.356,-51.907,29.157],[19.926,-48.748,22.342],
[22.624,-45.972,26.430],[21.295,-49.412,24.477],[28.464,-44.947,26.521],
[22.306,-46.362,30.727],[27.697,-50.809,19.998],[19.178,-49.133,29.136],
[24.998,-43.794,19.560],[18.014,-42.905,31.345],[31.743,-57.066,18.884],
[18.467,-49.825,20.355],[31.540,-46.178,16.717],[16.775,-47.510,28.273],
[13.716,-44.135,26.321],[33.139,-44.109,31.362],[22.918,-43.912,30.001],
[25.181,-41.215,18.363],[19.127,-42.764,24.427],[28.954,-42.812,16.960],
[27.447,-42.769,20.126],[31.013,-53.007,31.726],[18.445,-45.020,29.335],
[21.351,-58.396,27.140],[27.109,57.068,29.405],[28.772,-56.353,30.801],
[29.165,-50.116,15.426],[25.146,-55.269,17.586],[20.660,-53.656,25.647],
[21.236,-52.671,20.400],[26.506,-53.211,21.595],[25.081,-51.351,31.091],
[27.112,-51.654,23.936],[19.151,-52.112,27.375],[20.895,-55.181,15.813],
[14.160,-55.307,22.538],[14.273,-55.247,17.148],[25.858,-51.029,27.155],
[23.122,-55.161,21.602],[17.956,-55.714,21.343],[20.687,-54.668,18.460],
[26.705,-56.078,26.355],[18.611,-51.191,16.442],[30.184,-40.260,31.687],
[19.972,-49.157,33.754],[23.106,-52.711,23.936],[20.102,-57.308,20.542],
[23.143,-53.231,26.776],[24.118,-48.032,19.307],[30.707,-49.744,27.468],
[27.499,-54.188,16.355],[26.582,-46.798,29.194],[28.900,-54.129,20.161],
[17.661,-48.743,23.522],[24.346,-46.370,17.212],[15.593,-49.750,29.987],
[22.466,-50.813,32.507],[25.725,-44.494,30.428],[31.160,-53.262,18.785],
[33.612,-47.319,31.113],[30.267,-44.235,18.987],[28.605,-50.059,25.393],
[19.869,-44.537,22.357],[17.149,-49.165,15.236],[21.524,-46.725,21.157],
[24.819,-42.455,28.747],[20.260,-42.054,21.136],[27.350,-48.886,27.624],
[27.200,-47.737,16.421],[30.029,-42.286,29.661],[31.529,-41.972,18.009],
[17.098,-54.872,28.541],[28.479,-46.400,18.625],[16.322,-44.468,27.462],
[22.388,-45.874,14.868],[21.363,-46.521,17.536],[17.094,-48.435,18.042],
[25.410,-56.887,28.576],[27.196,-57.051,18.875],[24.038,-42.794,16.322],
[24.336,-53.536,19.593],[23.599,-55.004,29.321],[27.894,-56.564,21.645],
[21.681,-44.004,17.005],[21.763,-57.114,17.592],[32.265,-54.689,29.546],
[29.940,-52.408,25.135],[19.063,-45.498,15.617],[17.354,-54.757,14.053],
[28.706,-40.038,28.639],[19.110,-58.413,24.590],[17.589,-43.145,21.773],
[15.980,-43.578,33.649],[20.490,-43.196,32.874],[25.353,-45.244,14.974],
[28.384,-45.181,31.051],[28.649,-45.177,15.063],[15.706,-48.975,33.057],
[19.811,-48.485,16.427],[22.536,-48.128,33.173],[24.691,-48.782,16.121],
[28.321,-48.782,30.435],[18.360,-50.501,31.817],[22.934,-50.667,17.964],
[20.967,-51.006,29.388],[21.881,-51.742,15.518],[25.280,-49.167,33.386],
[31.647,-49.010,16.289],[20.386,-54.473,30.157],[20.888,-55.787,27.860],
[23.613,-50.820,28.816],[26.835,-53.313,33.459],[28.174,-51.620,31.760],
[20.362,-52.513,32.309],[21.856,58.711,29.641],[23.541,-53.625,32.632],
[26.349,-54.121,30.899],[26.422,-56.448,32.436],[19.150,-55.514,32.850],
[22.783,-59.938,32.537],[23.084,-58.671,34.829],[21.853,-41.229,16.933],
[16.597,-46.687,16.043],[15.609,-41.084,20.657],[15.100,-41.859,23.963],
[32.495,-43.088,25.535],[32.375,-43.818,28.150],[29.028,-47.910,33.236],
[31.497,-47.848,19.122],[32.163,-50.526,20.752],[32.786,-47.874,26.295],
[31.481,-45.852,30.099],[33.953,-50.249,17.557],[31.654,-52.510,15.861],
[30.338,-52.629,22.074],[16.212,-49.993,25.304],[30.949,-49.939,32.477],
[33.847,-52.309,32.939],[31.161,-56.638,16.117],[31.473,-55.063,20.894],
[31.534,-54.900,25.440],[29.490,-53.261,29.700],[16.776,-55.975,31.550],
[34.705,-58.493,33.889],[30.558,-57.391,21.553],[16.790,-56.265,24.054],
[15.472,-52.414,25.700],[31.977,-56.768,32.382],[33.824,-58.139,16.860],
[15.970,-57.826,19.576],[29.851,-42.458,24.418],[15.709,-57.262,29.329],
[17.212,-45.472,31.819],[20.895,-44.473,28.055],[25.236,-42.169,32.072],
[27.680,-42.239,30.997],[31.850,-58.211,30.208],[17.887,-39.587,27.729],
[18.877,-42.297,27.005],[25.421,-41.184,26.328],[27.559,-42.614,27.611],
[30.045,-39.882,26.385],[16.528,-43.971,24.570],[21.528,-44.796,24.544],
[23.045,-42.809,23.792],[23.939,-55.595,25.137],[30.490,-58.876,24.334],
[33.127,-58.715,23.587],[17.381,-58.465,22.552],[18.584,59.212,20.340],
[24.029,-58.961,22.653],[27.930,-58.261,24.975],[30.512,-59.961,20.371],
[17.310,-42.494,18.927],[17.776,-54.011,16.641],[24.590,-57.371,20.129],
[26.630,-41.888,15.934],[32.511,-43.126,20.295],[18.049,-58.405,30.133],
[18.518,-58.162,33.828],[16.586,-41.294,15.225],[29.058,-58.037,35.120],
[31.437,-46.490,33.282],[27.192,-31.906,21.574],[25.900,-30.686,24.973],
[18.830,-32.006,26.014],[30.482,-28.723,24.122],[20.005,-28.159,26.266],
[17.590,-33.162,20.538],[15.945,-31.648,22.441],[25.645,-23.094,22.429],
[23.384,-29.852,24.429],[32.907,-39.182,26.694],[26.949,-35.405,28.452],
[19.407,-37.544,23.416],[28.276,-37.844,21.831],[16.938,-34.119,23.272],
[20.015,-33.859,22.688],[33.769,-34.723,29.047],[18.716,-37.481,29.314],
[27.227,-28.657,21.145],[21.780,-30.974,27.243],[31.004,-36.025,22.969],
[17.953,-35.533,25.243],[21.143,-28.750,28.881],[16.135,-28.871,18.424],
[31.219,-25.792,26.016],[29.461,-37.266,27.879],[23.544,-40.006,24.619],
[29.394,-25.759,23.636],[30.462,-28.639,21.322],[30.509,-28.499,31.001],
[24.770,-30.206,21.133],[26.576,-33.372,15.375],[33.360,-31.547,27.776],
[29.757,-31.207,21.768],[29.395,-33.593,29.366],[19.499,-30.766,21.955],
[20.866,-29.166,24.121],[21.279,-33.078,25.229],[27.220,-26.694,27.250],
[22.633,-27.078,30.691],[28.780,-33.340,19.639],[19.219,-32.024,29.208],
[24.893,-25.403,21.391],[18.891,-26.802,30.513],[31.773,-39.368,19.130],
[20.108,-32.431,19.597],[31.402,-26.178,16.404],[15.430,-28.974,28.109],
[15.178,-26.037,26.557],[31.806,-25.738,30.970],[25.092,-25.982,30.595],
[24.751,-22.782,19.504],[18.557,-24.562,23.158],[27.878,-24.190,16.847],
[27.304,-24.492,19.749],[30.896,-34.994,32.215],[17.567,-29.739,29.702],
[20.991,-40.048,26.340],[26.367,-40.524,29.831],[28.456,-35.954,30.472],
[29.129,-32.540,16.570],[24.527,-39.011,16.865],[20.144,-35.397,27.037],
[20.174,-35.372,18.228],[25.459,-34.925,18.895],[25.673,-33.623,30.563],
[28.231,-35.000,21.919],[18.491,-34.893,29.931],[20.093,-38.066,14.115],
[14.613,-38.086,21.825],[16.282,-37.118,17.896],[25.783,-33.849,26.318],
[23.422,-37.308,23.535],[16.300,-35.474,20.603],[22.072,-37.062,19.751],
[27.162,-37.774,26.580],[22.428,-36.145,17.248],[30.964,-21.973,32.334],
[19.187,-29.878,31.809],[22.938,-34.475,23.751],[22.143,-39.041,21.473],
[22.586,-34.364,27.271],[22.634,-29.961,19.123],[30.423,-30.943,26.315],
[27.045,-37.882,16.806],[23.827,-29.432,30.009],[29.501,-35.592,17.765],
[18.195,-29.582,25.077],[25.360,-29.195,18.451],[15.071,-31.332,29.501],
[23.130,-33.545,32.145],[28.049,-26.211,29.762],[33.155,-34.939,17.836],
[31.550,-29.315,28.596],[30.997,-26.782,19.242],[28.522,-31.376,24.130],
[22.383,-27.435,21.054],[17.092,-31.034,16.452],[22.355,-30.306,21.873],
[24.521,-26.958,28.028],[20.168,-23.568,21.150],[26.710,-29.199,28.030],
[28.941,-28.697,16.557],[29.444,-24.346,31.426],[31.756,-23.709,18.452],
[16.657,-37.158,27.021],[29.010,-28.739,19.094],[17.692,-26.210,28.073],
[22.214,-27.073,14.937],[21.699,-28.182,17.576],[17.552,-30.515,19.991],
[23.631,-38.190,29.577],[27.994,-37.020,19.376],[24.438,-26.558,18.739],
[23.829,-35.544,21.380],[24.473,-35.351,28.749],[25.130,-38.749,22.015],
[22.083,-25.486,17.458],[20.296,-39.039,18.071],[31.247,-36.644,29.946],
[30.516,-33.390,24.737],[16.849,-26.431,16.454],[17.883,-35.054,16.689],
[29.910,-23.447,25.063],[18.131,-39.783,25.280],[17.564,-26.390,21.223],
[15.920,-24.190,34.112],[22.068,-24.475,33.478],[24.938,-26.897,16.062],
[28.258,-27.946,32.496],[28.565,-24.599,13.978],[16.453,-30.482,32.418],
[20.623,-29.147,14.429],[22.143,-30.375,31.770],[25.514,-30.304,16.050],
[28.669,-30.876,29.340],[18.272,-32.443,32.016],[21.094,-30.561,16.813],
[21.663,-33.466,29.752],[23.576,-32.872,16.020],[25.864,-30.135,32.542],
[29.057,-30.817,14.781],[21.363,-36.293,30.830],[23.388,-37.683,26.873],
[24.502,-31.594,27.515],[27.582,-35.087,33.446],[28.163,-32.539,31.763],
[20.201,-34.180,32.474],[22.496,-40.416,28.618],[24.233,-35.895,33.215],
[25.808,-36.937,31.038],[28.466,-37.803,32.529],[16.413,-35.897,33.585],
[23.013,-40.967,31.185],[22.299,-39.313,32.957],[23.058,-22.996,17.341],
[15.074,-28.465,15.978],[16.211,-22.812,19.684],[15.737,-23.943,22.418],
[33.196,-23.793,25.884],[31.948,-24.141,28.615],[30.854,-28.032,34.030],
[31.724,-29.615,18.645],[32.311,-32.036,20.944],[33.206,-29.783,24.642],
[30.468,-26.451,28.735],[32.197,-32.325,17.063],[31.373,-34.953,15.620],
[30.980,-34.307,20.663],[16.047,-31.120,25.864],[30.878,-32.028,31.358],
[33.125,-33.187,32.097],[31.221,-38.147,15.475],[30.954,-37.022,20.084],
[32.891,-35.761,25.020],[31.621,-33.675,27.262],[14.378,-35.339,31.670],
[32.145,-40.040,33.924],[29.551,-41.115,20.136],[16.292,-37.331,24.052],
[15.461,-34.640,25.878],[33.081,-38.302,31.726],[34.131,-39.006,16.812],
[17.235,-38.105,20.856],[29.324,-23.275,21.369],[15.962,-38.892,29.256],
[18.053,-25.784,32.935],[20.499,-25.582,27.675],[25.300,-23.078,32.249],
[27.399,-22.737,30.249],[32.235,-40.923,30.223],[18.013,-22.264,27.763],
[21.149,-21.053,28.476],[25.041,-24.268,26.810],[27.740,-24.011,27.452],
[29.854,-22.269,28.055],[17.759,-25.068,25.695],[20.191,-26.691,23.040],
[22.022,-24.009,24.329],[26.128,-37.011,24.086],[30.816,-38.571,24.200],
[31.668,-41.584,22.044],[18.553,-40.304,22.108],[15.731,-40.019,18.074],
[24.275,-41.262,21.049],[28.165,-39.779,23.981],[28.973,-39.955,17.482],
[15.950,-25.821,18.903],[17.464,-38.152,15.667],[22.906,-39.511,18.897],
[25.041,-23.392,15.478],[32.675,-24.604,21.165],[17.057,-40.265,31.210],
[19.488,-39.247,32.475],[19.593,-25.914,16.145],[27.601,-40.196,33.623],
[31.467,-24.920,33.982],[25.211,-12.205,22.346],[25.507,-12.332,24.854],
[18.724,-13.427,26.394],[29.915,-9.862,23.499],[19.482,-8.746,25.398],
[14.899,-15.692,20.949],[16.009,-12.973,22.798],[27.168,-5.303,24.594],
[25.044,-9.762,25.298],[33.131,-18.816,25.707],[27.040,-16.536,26.280],
[21.573,-16.973,23.765],[27.736,-18.665,21.826],[17.481,-15.210,23.322],
[21.340,-14.241,22.861],[31.384,-17.064,28.262],[19.599,-18.823,26.627],
[25.840,-8.159,20.992],[21.997,-10.427,29.191],[31.225,-18.878,23.691],
[18.765,-17.318,24.405],[19.581,-11.504,27.978],[17.430,-9.435,18.634],
[29.366,-9.396,26.302],[30.342,-19.808,27.036],[23.281,-21.891,24.601],
[28.231,-6.352,22.255],[28.497,-8.856,21.194],[29.413,-9.718,29.901],
[22.902,-13.790,20.076],[25.531,-13.730,15.565],[31.421,-13.112,28.298],
[28.768,-13.230,19.988],[29.056,-15.144,28.771],[18.978,-12.521,22.398],
[22.264,-10.187,23.431],[21.024,-12.523,24.764],[26.876,-8.203,27.586],
[23.767,-9.242,27.603],[27.652,-14.178,17.563],[19.573,-14.002,29.323],
[25.361,-5.488,21.869],[19.985,-9.040,33.025],[30.103,-21.384,17.567],
[19.318,-12.436,19.614],[32.219,-8.675,16.236],[14.898,-10.465,27.310],
[16.495,-8.514,25.294],[34.499,-8.541,30.731],[23.023,-8.214,30.230],
[24.799,-2.560,18.844],[18.727,-5.748,22.119],[27.698,-5.345,15.336],
[28.361,-5.906,18.613],[31.872,-14.365,30.931],[17.933,-8.550,29.182],
[20.466,-22.085,26.176],[25.905,-21.861,28.368],[28.930,-16.774,30.851],
[29.272,-12.853,15.814],[24.225,-17.858,17.008],[20.126,-15.661,26.657],
[19.664,-15.108,20.953],[26.138,-13.753,19.987],[25.478,-13.364,31.791],
[26.908,-16.143,21.365],[17.404,-15.838,26.214],[21.955,-17.587,14.635],
[14.394,-19.184,22.480],[16.316,-17.942,18.911],[27.656,-13.467,26.833],
[22.241,-20.294,21.972],[17.671,-17.091,21.106],[20.272,-17.937,19.740],
[27.378,-19.218,27.375],[21.023,-18.165,17.014],[30.389,-3.250,30.781],
[19.174,-11.731,31.842],[24.478,-15.930,23.964],[19.246,-19.890,21.322],
[23.039,-16.771,27.000],[22.268,-11.244,18.461],[30.832,-11.671,25.707],
[27.109,-19.704,16.713],[24.917,-10.181,30.033],[27.678,-16.921,17.847],
[18.008,-11.218,24.857],[24.852,-11.196,17.835],[15.518,-13.072,29.963],
[22.762,-14.559,31.854],[28.254,-6.574,29.767],[34.679,-15.190,17.734],
[32.892,-10.909,30.722],[29.936,-8.212,18.482],[28.182,-12.075,23.992],
[20.943,-8.228,21.543],[17.645,-12.180,16.392],[21.297,-11.492,21.054],
[24.323,-6.653,27.204],[22.118,-5.718,20.675],[27.125,-10.897,27.047],
[28.262,-8.394,16.194],[29.131,-5.374,32.065],[33.005,-5.648,15.488],
[18.234,-17.304,28.526],[29.927,-10.806,19.386],[18.373,-6.813,27.218],
[22.341,-8.403,14.457],[21.265,-7.414,16.684],[19.659,-10.670,17.350],
[24.062,-19.872,28.587],[27.935,-19.245,19.396],[25.231,-8.077,17.960],
[24.116,-18.198,20.853],[25.485,-15.207,29.596],[24.853,-20.602,21.034],
[20.404,-4.687,18.327],[20.909,-20.638,18.454],[30.279,-18.983,30.224],
[28.151,-14.767,24.622],[18.616,-8.599,16.163],[18.302,-16.460,16.927],
[29.827,-4.278,25.521],[16.357,-22.116,24.781],[15.990,-8.526,21.770],
[17.004,-4.725,30.863],[21.551,-5.666,33.287],[25.024,-6.321,15.231],
[27.553,-8.409,31.522],[30.393,-6.666,16.025],[15.822,-11.855,32.669],
[21.767,-10.958,15.621],[22.675,-11.454,31.933],[24.844,-9.241,15.130],
[28.091,-12.679,29.585],[18.544,-14.198,32.766],[20.989,-14.114,17.522],
[22.138,-15.067,29.132],[22.802,-13.512,15.203],[27.316,-12.006,34.170],
[33.826,-13.306,15.640],[20.757,-17.684,29.733],[22.207,-19.199,26.304],
[24.887,-14.412,27.072],[27.958,-14.929,34.401],[28.129,-14.100,31.888],
[20.676,-15.922,33.177],[23.126,-22.403,27.152],[23.507,-16.882,33.151],
[25.805,-17.711,31.020],[28.184,-19.217,32.182],[19.224,-18.117,32.300],
[22.946,-22.579,31.109],[23.933,-19.750,31.511],[22.692,-1.518,17.562],
[16.002,-9.841,16.009],[15.777,-3.287,21.087],[15.976,-5.683,22.204],
[33.881,-4.821,26.856],[33.187,-6.078,29.228],[30.846,-10.182,33.020],
[32.400,-10.724,18.673],[32.030,-12.444,20.754],[33.146,-9.073,26.460],
[31.329,-8.095,28.589],[31.891,-13.510,18.216],[31.856,-16.105,16.729],
[30.455,-14.811,21.363],[15.468,-12.500,25.331],[30.279,-12.116,30.868],
[34.728,-15.357,31.983],[29.804,-18.201,16.521],[29.800,-17.248,20.801],
[29.653,-17.426,25.991],[30.317,-14.639,26.381],[15.285,-16.757,30.624],
[34.821,-21.479,32.231],[29.927,-20.279,21.027],[16.428,-18.664,25.498],
[14.230,-14.933,24.601],[32.686,-19.971,31.757],[32.050,-19.546,16.708],
[16.642,-20.619,21.236],[31.156,-4.453,20.517],[16.728,-19.982,28.400],
[18.074,-6.959,32.327],[20.916,-8.177,27.961],[25.045,-6.309,30.222],
[27.548,-3.167,31.956],[32.486,-21.561,29.512],[20.277,-2.339,29.702],
[20.407,-4.317,27.306],[24.375,-3.337,26.363],[27.135,-5.489,27.243],
[28.846,-3.445,28.059],[17.540,-5.474,24.842],[20.434,-6.637,23.921],
[22.971,-5.305,24.554],[25.738,-19.379,23.968],[31.635,-21.287,25.003],
[32.660,-20.577,21.964],[18.622,-21.604,23.280],[17.194,-23.470,17.143],
[22.826,-22.945,21.779],[27.482,-21.817,25.766],[27.489,-21.874,18.670],
[16.368,-6.712,19.698],[18.567,-19.098,17.212],[24.356,-20.682,17.880],
[25.733,-4.170,17.138],[31.582,-5.855,18.274],[16.795,-21.421,31.200],
[19.306,-21.121,32.494],[18.899,-5.845,15.798],[27.690,-21.529,33.374],
[32.126,-7.592,31.314],[27.139,6.333,20.206],[25.207,5.975,25.162],
[18.116,2.994,26.396],[28.540,7.822,23.352],[18.944,9.523,24.693],
[16.476,2.033,19.232],[13.571,6.088,22.114],[26.282,15.390,23.010],
[25.222,8.858,24.801],[31.570,-0.775,26.016],[27.694,2.078,26.080],
[19.439,0.741,23.316],[28.717,1.114,21.237],[16.404,3.650,23.833],
[23.307,5.771,22.992],[34.185,0.820,27.993],[17.080,0.778,27.643],
[25.809,9.611,20.112],[20.203,7.169,28.363],[29.585,0.090,23.836],
[19.829,0.763,25.858],[19.468,9.666,27.284],[16.584,9.190,18.528],
[28.108,9.556,25.474],[29.012,-1.044,26.618],[22.133,-1.016,24.097],
[28.720,12.049,24.742],[29.001,10.287,20.485],[30.152,8.382,29.331],
[24.458,5.327,20.325],[26.867,4.425,16.267],[33.689,4.959,27.571],
[30.253,6.140,20.026],[28.866,3.805,28.900],[19.944,7.415,23.640],
[22.120,8.746,23.952],[21.871,6.029,25.310],[26.281,10.319,27.136],
[21.650,9.307,29.036],[28.716,5.826,17.342],[18.709,4.760,29.327],
[27.521,12.227,21.725],[21.251,10.533,32.685],[32.295,-3.469,17.567],
[20.119,4.491,18.976],[32.669,11.028,17.271],[14.799,7.513,27.583],
[15.075,11.831,25.371],[34.083,11.769,29.632],[23.425,11.629,30.649],
[27.356,15.430,18.932],[19.050,12.967,22.015],[29.888,14.973,16.853],
[28.741,12.927,19.184],[32.181,1.247,29.867],[18.376,10.096,29.884],
[20.980,-1.736,26.484],[26.584,-3.737,29.331],[27.914,1.807,30.918],
[31.244,7.105,16.299],[25.433,0.370,17.382],[20.381,3.517,24.734],
[22.109,3.905,21.572],[25.679,3.372,18.595],[25.560,4.182,33.016],
[28.082,2.952,23.521],[17.196,2.469,29.906],[21.841,0.957,14.575],
[15.044,0.608,23.636],[18.202,0.053,17.718],[25.860,5.770,29.744],
[24.511,-0.163,22.662],[18.624,3.212,22.072],[21.772,1.207,20.691],
[26.581,-0.161,26.628],[22.071,2.745,18.682],[30.607,16.226,33.730],
[17.921,7.442,30.135],[23.814,3.438,25.595],[19.017,0.173,20.794],
[24.614,0.891,27.980],[23.428,8.171,18.544],[30.740,6.968,26.173],
[28.339,0.302,16.578],[24.727,9.668,29.199],[28.435,2.965,17.784],
[17.798,5.814,23.553],[26.099,7.277,17.851],[15.713,6.206,29.672],
[23.181,4.826,31.275],[28.493,10.479,30.035],[33.440,2.185,18.282],
[32.653,8.491,30.954],[31.491,9.808,19.422],[27.546,5.376,26.913],
[21.737,10.761,21.767],[17.896,5.971,16.010],[24.382,8.321,22.234],
[24.224,12.140,27.632],[21.272,14.132,21.184],[25.183,7.640,27.410],
[28.569,8.409,16.172],[29.316,13.107,29.307],[33.941,14.201,17.590],
[15.869,-0.319,29.996],[29.630,8.108,18.471],[17.875,12.317,28.306],
[23.187,10.571,14.894],[21.921,9.778,17.271],[18.157,6.520,18.631],
[25.156,-1.097,29.719],[28.210,-0.961,19.116],[24.723,10.499,17.934],
[24.176,2.156,21.113],[24.713,3.055,29.564],[25.656,0.149,20.364],
[22.281,12.557,18.805],[19.845,-1.921,18.789],[29.414,1.290,28.533],
[30.676,4.412,24.551],[19.225,9.699,18.439],[18.886,2.590,16.367],
[30.382,13.250,26.854],[16.277,-2.798,23.744],[16.737,11.239,21.073],
[18.751,13.173,31.427],[21.044,13.006,33.923],[25.667,12.309,15.765],
[27.371,9.523,32.350],[29.329,12.308,15.622],[16.263,6.631,32.927],
[21.390,7.016,16.715],[23.673,8.076,31.396],[26.180,9.688,15.261],
[27.358,7.993,29.427],[17.791,4.094,33.402],[23.041,5.287,18.035],
[22.289,2.902,28.790],[24.128,6.367,15.731],[27.852,6.574,32.661],
[31.920,8.217,13.621],[20.901,0.764,31.434],[21.989,0.195,28.362],
[23.782,5.023,27.860],[27.449,1.670,33.598],[28.890,4.402,31.637],
[20.392,3.838,32.404],[23.081,-2.592,28.651],[23.259,0.509,32.591],
[25.738,-0.861,32.287],[29.830,-0.725,32.542],[18.533,0.546,33.145],
[22.057,-4.510,30.672],[24.069,-2.286,33.556],[22.683,15.620,16.173],
[17.488,9.560,14.956],[16.804,15.035,20.566],[16.764,14.027,23.817],
[32.781,13.746,25.033],[32.669,13.843,29.076],[30.855,8.870,34.570],
[32.587,7.410,19.107],[32.733,4.713,20.400],[33.103,9.014,25.239],
[29.990,10.738,27.849],[33.782,5.762,17.357],[33.332,3.454,15.244],
[30.395,3.243,21.567],[15.740,6.825,24.610],[31.487,6.109,28.762],
[34.546,3.229,31.854],[31.041,1.454,15.523],[30.702,1.103,18.734],
[32.025,1.790,23.251],[31.203,3.551,27.555],[16.055,2.347,32.483],
[31.744,-3.708,32.961],[32.361,-1.418,19.380],[18.883,-2.089,24.840],
[15.547,2.818,26.083],[32.373,-1.157,31.293],[34.565,-1.777,17.817],
[16.848,-1.010,19.914],[31.567,14.378,21.626],[17.146,-2.151,28.430],
[18.189,10.606,32.866],[20.535,11.823,29.833],[24.595,13.927,30.656],
[27.346,13.870,31.153],[31.123,-2.327,28.315],[18.230,16.124,27.863],
[20.479,13.863,27.730],[25.657,14.729,25.643],[27.385,12.911,27.224],
[30.058,16.316,27.927],[17.565,14.070,26.465],[20.931,12.521,24.027],
[23.246,10.728,25.428],[25.929,-2.676,22.287],[31.224,-3.638,23.149],
[33.844,-3.055,22.078],[18.387,-2.552,21.452],[15.815,-4.275,18.305],
[22.775,-3.284,22.842],[27.086,-1.601,24.365],[29.655,-3.135,18.680],
[16.811,11.794,17.792],[19.577,-3.366,15.436],[23.063,-1.320,20.476],
[25.102,13.606,18.517],[32.260,12.476,19.503],[16.416,-2.035,31.991],
[20.992,-2.525,33.067],[18.122,13.820,16.648],[25.540,-4.650,33.490],
[31.217,12.437,32.668],[27.025,24.686,22.913],[25.409,24.207,25.214],
[18.534,23.833,24.659],[28.652,27.504,22.833],[18.237,27.556,26.212],
[16.522,20.570,19.527],[16.111,23.917,22.186],[26.845,33.202,23.742],
[24.185,27.122,25.438],[32.126,18.015,27.489],[28.479,21.731,26.770],
[20.955,19.621,22.703],[29.677,19.288,21.725],[16.696,21.688,23.604],
[22.807,22.479,23.082],[33.333,20.518,27.408],[18.050,18.967,26.462],
[26.926,26.198,20.531],[21.358,25.822,27.244],[31.138,18.338,23.886],
[19.722,20.394,24.841],[20.200,28.320,28.219],[16.948,28.089,20.007],
[29.439,26.985,27.634],[29.127,18.961,27.099],[23.618,13.005,23.244],
[28.363,30.532,23.500],[28.197,28.597,20.262],[30.630,27.149,29.942],
[26.349,24.316,18.705],[26.731,22.138,15.997],[32.251,23.600,28.343],
[29.556,23.846,21.263],[28.150,22.765,29.158],[21.075,24.848,22.212],
[21.021,29.457,24.531],[21.265,23.219,25.106],[27.829,29.640,27.323],
[21.539,28.749,30.400],[28.717,23.147,18.700],[20.334,23.874,29.277],
[23.693,31.607,19.611],[20.503,31.626,30.626],[32.211,16.579,17.962],
[19.616,24.455,19.710],[32.046,29.278,16.396],[15.797,25.700,26.868],
[14.937,30.700,25.756],[34.356,29.494,31.107],[23.758,28.506,28.597],
[25.396,33.852,19.860],[18.322,32.649,24.770],[27.627,31.210,16.031],
[28.518,30.345,18.307],[30.981,20.808,32.022],[17.516,27.757,28.897],
[19.831,15.543,24.444],[25.948,15.095,28.822],[28.406,20.140,30.066],
[29.933,24.633,16.513],[25.566,19.717,17.547],[19.225,22.570,26.873],
[20.745,21.882,20.989],[25.250,21.922,19.097],[26.731,20.790,31.926],
[27.721,21.600,22.884],[18.828,20.843,29.104],[21.411,17.877,14.982],
[15.298,17.807,22.245],[18.331,18.865,17.559],[25.733,22.369,27.552],
[22.652,17.624,20.286],[18.437,19.932,21.579],[20.618,19.615,19.572],
[27.007,17.100,26.260],[20.912,22.091,17.797],[30.251,32.855,32.415],
[18.436,25.521,30.285],[25.569,21.325,24.921],[20.165,17.453,20.912],
[23.795,20.015,26.559],[24.441,26.183,18.859],[31.483,25.569,26.235],
[28.760,17.621,16.832],[24.412,25.988,29.683],[28.008,20.380,17.496],
[17.479,26.358,23.260],[25.853,27.717,16.136],[16.385,22.735,30.504],
[24.443,23.552,32.342],[27.357,27.999,29.845],[34.326,20.958,18.574],
[33.288,28.798,28.704],[30.960,29.800,19.178],[28.452,25.649,24.895],
[21.565,29.281,20.524],[16.486,25.031,17.409],[21.478,26.684,19.655],
[23.688,30.941,29.237],[21.590,31.851,21.107],[26.462,26.085,27.572],
[28.740,27.677,16.482],[28.625,30.999,29.769],[32.858,32.912,16.452],
[17.749,18.471,29.017],[30.450,27.048,18.365],[18.256,30.388,27.539],
[23.115,28.814,15.218],[21.018,28.440,16.845],[19.129,26.749,18.297],
[25.635,18.362,28.299],[27.851,18.501,19.326],[25.771,29.304,18.898],
[25.827,19.555,20.855],[24.059,21.040,29.057],[27.801,17.130,21.621],
[21.134,30.713,18.285],[20.869,17.039,17.655],[30.861,17.997,30.700],
[30.019,22.972,24.768],[18.651,29.544,17.682],[18.218,21.620,16.824],
[30.423,30.666,25.291],[16.913,16.603,25.254],[19.011,29.245,21.165],
[16.096,34.402,30.909],[22.041,31.710,33.448],[24.973,30.493,14.489],
[26.881,29.554,32.542],[29.641,29.847,14.916],[15.350,25.684,31.242],
[20.748,25.752,16.273],[21.452,26.160,31.182],[24.625,26.799,13.807],
[28.361,25.455,30.511],[18.677,22.198,32.088],[22.735,24.335,18.060],
[21.612,21.250,29.559],[22.697,24.444,14.182],[26.032,23.632,30.109],
[32.324,25.146,14.778],[20.774,18.750,30.116],[21.960,18.106,26.737],
[22.978,23.809,28.554],[28.828,21.793,34.015],[29.435,23.304,31.569],
[21.329,20.841,32.458],[23.225,15.122,27.588],[24.470,19.922,33.370],
[24.860,19.020,30.676],[26.991,17.619,33.910],[19.050,18.143,32.561],
[21.562,14.172,30.622],[25.019,15.822,33.799],[24.066,34.073,17.120],
[16.481,27.933,16.804],[15.417,33.197,21.130],[17.045,31.335,22.391],
[34.285,33.333,24.669],[32.317,31.303,28.160],[30.849,28.399,32.317],
[32.977,27.901,19.003],[32.098,26.520,22.489],[32.715,27.817,25.525],
[30.421,29.525,27.844],[33.473,23.428,16.998],[34.047,20.969,15.610],
[30.474,21.686,22.533],[14.852,25.930,23.934],[31.457,25.294,31.577],
[33.279,22.933,32.249],[31.220,19.636,15.888],[32.180,20.118,20.632],
[30.949,20.491,25.739],[31.131,20.429,29.073],[15.939,20.904,32.948],
[33.888,16.447,34.039],[32.461,17.255,20.607],[17.100,19.091,23.987],
[16.536,21.743,26.368],[33.619,17.201,30.435],[33.638,18.219,16.423],
[16.156,17.241,19.112],[29.412,31.109,21.074],[15.365,16.976,29.014],
[17.963,29.955,32.218],[20.934,30.629,27.259],[25.268,32.716,30.295],
[28.714,33.812,30.559],[29.630,15.652,30.789],[17.654,35.027,28.806],
[19.577,33.176,27.750],[25.057,34.206,26.968],[27.357,33.127,27.898],
[29.718,34.449,27.896],[17.735,30.037,24.748],[19.743,31.358,22.871],
[22.208,31.707,23.729],[23.315,15.588,24.589],[30.327,14.386,24.220],
[33.304,16.140,23.712],[17.856,16.916,22.547],[18.555,15.927,18.989],
[23.573,15.362,21.708],[28.759,16.911,24.289],[29.846,15.993,19.342],
[16.283,30.712,19.201],[18.923,16.614,15.619],[24.997,16.422,19.421],
[26.163,32.240,17.817],[32.045,31.644,21.850],[15.916,18.323,31.783],
[20.447,15.743,33.359],[17.387,32.510,15.568],[28.176,14.740,33.466],
[31.296,30.856,30.599],[26.262,42.901,22.682],[24.875,43.858,25.117],
[19.698,42.045,26.826],[29.512,46.446,23.983],[19.394,46.295,23.838],
[15.669,41.105,20.465],[15.799,42.844,22.857],[26.744,50.950,25.004],
[23.889,46.486,25.775],[31.866,36.287,27.558],[28.955,39.900,26.449],
[20.469,37.939,22.884],[29.564,37.257,21.352],[16.550,39.840,22.526],
[23.627,41.042,23.159],[34.168,38.709,29.188],[17.893,37.593,26.870],
[26.075,46.304,20.040],[21.382,45.379,28.980],[30.913,37.034,23.584],
[17.458,40.089,25.548],[19.594,46.582,31.033],[16.287,46.586,19.495],
[27.940,47.095,26.105],[29.404,37.521,27.911],[23.421,33.878,24.827],
[28.691,49.296,23.466],[28.236,45.382,21.591],[31.090,45.627,30.434],
[25.099,43.435,20.092],[27.258,42.687,15.225],[33.392,42.377,27.820],
[27.884,42.879,20.249],[28.633,40.628,28.904],[20.931,43.260,22.467],
[21.937,47.781,24.210],[20.770,44.207,25.010],[26.058,49.508,27.200],
[22.522,46.757,30.981],[26.602,41.506,18.205],[19.124,42.038,29.492],
[25.602,49.004,20.759],[20.620,49.542,32.161],[33.630,35.671,17.818],
[19.975,42.662,19.947],[32.844,47.611,17.815],[13.483,45.620,28.684],
[16.101,48.601,27.055],[34.388,48.704,30.833],[24.773,47.412,30.035],
[24.419,51.406,19.303],[19.762,50.064,20.149],[27.455,48.305,16.742],
[28.310,52.040,20.132],[31.579,40.384,30.637],[16.527,46.362,28.536],
[20.538,34.754,25.419],[26.452,35.365,29.531],[29.374,38.704,31.137],
[29.952,43.130,15.707],[24.115,37.495,16.339],[22.140,39.367,25.216],
[20.937,40.686,22.037],[27.195,40.132,20.708],[25.256,41.083,32.950],
[26.963,40.088,24.189],[19.773,39.527,29.097],[21.396,38.673,15.968],
[14.228,37.737,22.426],[17.215,38.625,19.677],[26.020,40.131,27.466],
[23.493,37.979,23.430],[20.000,38.996,20.197],[22.633,38.521,18.559],
[27.547,36.125,25.780],[20.946,40.599,18.057],[30.547,51.779,33.503],
[18.741,44.078,31.123],[24.439,41.179,25.631],[18.946,36.409,19.716],
[21.901,37.657,27.388],[22.985,44.603,16.866],[31.492,42.878,26.014],
[28.335,37.173,16.367],[27.675,46.211,31.123],[29.518,40.735,18.788],
[17.312,44.744,24.006],[25.557,46.373,17.106],[16.575,41.784,29.154],
[22.763,41.586,30.581],[27.441,49.341,30.002],[33.725,40.753,18.343],
[33.751,45.547,30.816],[31.737,49.609,20.213],[27.814,44.036,25.198],
[22.606,47.580,20.479],[17.131,44.285,16.594],[23.106,45.060,22.362],
[23.179,48.845,28.001],[22.250,50.328,21.873],[27.000,47.316,28.538],
[28.455,45.392,16.178],[29.497,48.137,30.772],[32.552,50.229,16.666],
[17.222,37.582,29.526],[29.121,44.310,18.539],[17.888,48.520,29.456],
[23.294,48.430,13.416],[22.858,47.760,17.733],[18.735,45.191,18.850],
[24.903,37.576,29.149],[27.078,36.595,19.106],[25.063,49.386,17.589],
[24.693,39.026,20.611],[25.419,40.436,30.052],[25.426,35.354,22.151],
[22.149,49.940,19.205],[21.916,35.837,17.922],[32.299,37.569,30.705],
[29.515,41.744,24.661],[20.874,46.534,15.173],[18.708,39.882,16.133],
[31.108,49.386,26.138],[17.839,34.897,26.209],[18.879,47.929,21.903],
[16.373,51.245,33.006],[23.140,50.604,33.625],[25.812,48.341,14.615],
[26.781,47.424,33.570],[30.472,48.224,16.706],[16.589,45.226,32.377],
[20.618,43.437,14.789],[22.804,44.108,32.062],[27.246,45.786,13.927],
[28.525,43.923,30.015],[17.773,41.442,33.132],[20.826,43.344,17.318],
[22.115,40.924,27.768],[22.407,41.505,15.719],[26.161,44.274,32.778],
[32.494,44.130,15.099],[22.440,38.468,30.380],[23.196,35.537,28.790],
[24.391,42.549,28.209],[28.258,39.176,33.900],[27.372,41.478,31.503],
[20.229,40.008,33.971],[22.408,32.963,27.673],[23.479,36.392,31.772],
[26.743,37.765,31.940],[28.954,36.726,33.073],[18.366,38.320,32.712],
[22.736,33.157,31.026],[23.391,34.373,33.665],[23.169,51.303,16.618],
[16.461,47.919,17.186],[15.256,51.442,22.506],[16.989,49.755,23.715],
[32.953,51.804,26.454],[32.043,49.540,29.503],[30.191,46.027,32.737],
[30.517,46.592,18.857],[32.921,44.127,21.065],[31.544,45.833,25.457],
[30.967,47.452,27.983],[31.522,42.964,18.349],[31.496,40.259,16.711],
[29.640,40.912,21.804],[15.753,43.300,25.707],[31.806,43.305,30.750],
[32.878,41.447,33.054],[31.041,37.107,16.351],[29.748,38.145,18.772],
[31.641,38.874,26.303],[31.079,41.068,28.015],[16.012,39.110,31.879],
[34.398,34.479,33.992],[32.001,36.570,19.931],[16.058,36.159,24.745],
[13.573,41.578,24.424],[33.850,35.437,31.116],[33.122,36.514,14.629],
[15.541,35.663,20.007],[30.174,52.928,22.034],[13.853,34.042,27.084],
[18.577,47.820,33.732],[20.685,48.190,28.909],[25.641,50.823,31.601],
[29.064,53.573,31.433],[31.879,33.721,29.468],[18.529,53.025,28.396],
[19.789,50.729,28.828],[23.944,51.330,26.693],[27.314,51.470,28.566],
[30.280,51.929,28.493],[18.169,50.895,26.171],[19.393,48.866,24.666],
[21.490,50.132,26.357],[24.286,36.618,25.468],[29.448,33.164,22.688],
[31.643,34.272,21.433],[18.149,35.659,22.357],[17.007,34.670,17.700],
[22.832,34.280,20.690],[28.348,35.409,23.153],[28.483,34.032,19.525],
[17.306,49.261,19.597],[18.180,37.179,16.965],[24.320,36.634,19.595],
[26.128,51.029,15.740],[33.622,51.074,18.975],[18.250,35.806,31.521],
[19.587,35.988,33.915],[19.795,50.876,16.786],[26.855,34.146,33.023],
[31.935,49.560,32.331],[32.547,54.451,26.498],[28.295,58.242,27.152],
[19.996,56.665,22.769],[27.295,58.700,19.852],[16.467,58.908,24.880],
[33.037,56.646,28.000],[19.611,55.260,26.940],[28.478,57.833,24.113],
[18.668,57.193,25.104],[30.861,54.649,28.719],[24.300,53.558,24.683],
[32.714,53.323,17.512],[29.789,56.804,35.193],[20.197,53.327,24.434],
[26.172,53.532,29.713],[29.081,56.413,31.348],[25.181,55.083,16.319],
[20.310,58.002,27.548],[22.140,-59.281,20.007],[26.057,57.853,22.317],
[17.389,57.742,29.640],[22.319,56.372,15.590],[14.791,56.784,23.440],
[15.832,56.124,18.673],[22.168,54.861,23.627],[17.528,55.494,21.127],
[23.367,56.603,20.487],[27.065,55.155,27.464],[19.927,-59.916,18.004],
[26.016,57.781,25.638],[21.495,54.752,20.610],[23.758,-59.887,26.657],
[27.682,55.833,18.398],[29.404,-59.210,17.981],[33.495,58.649,18.802],
[17.238,55.071,30.357],[24.709,56.460,27.638],[26.061,53.673,18.851],
[24.661,58.799,19.476],[24.637,58.337,29.692],[26.907,52.876,22.871],
[20.379,57.144,20.124],[31.175,57.255,30.136],[18.228,58.240,15.470],
[17.529,53.711,26.092],[21.996,56.430,31.920],[19.744,55.760,29.461],
[27.980,59.165,33.259],[19.877,58.563,32.393],[23.667,54.013,28.189],
[25.049,58.323,33.819],[26.476,56.990,31.908],[28.157,55.330,33.710],
[17.592,55.454,33.058],[22.925,51.595,30.947],[22.652,54.009,33.777],
[31.328,59.171,16.246],[32.640,54.852,15.278],[31.648,56.530,18.861],
[30.671,57.060,26.438],[31.412,59.176,27.996],[15.161,58.902,30.897],
[32.656,52.933,32.051],[30.173,56.397,21.357],[15.744,55.731,26.006],
[18.256,-58.966,26.950],[33.297,54.531,30.062],[35.167,54.464,17.811],
[16.167,53.497,19.079],[15.893,54.056,28.254],[29.749,50.748,31.095],
[26.715,55.659,24.159],[31.556,52.824,24.431],[29.586,54.803,24.189],
[17.413,52.745,23.540],[17.458,52.165,16.786],[23.600,53.004,21.677],
[28.349,53.016,26.159],[30.244,53.827,19.284],[18.175,54.943,17.217],
[23.283,54.900,18.058],[18.774,52.784,32.516],[19.792,54.267,34.374],
[27.544,51.795,33.397],[26.128,-49.667,41.123],[24.282,-48.690,44.800],
[19.075,-51.109,46.267],[28.896,-47.580,42.039],[19.510,-45.920,43.897],
[14.997,-52.254,40.417],[16.232,-49.745,40.226],[26.804,-42.438,43.186],
[24.414,-46.310,43.710],[32.655,-57.669,43.541],[28.388,-56.670,45.807],
[21.161,-55.674,40.762],[27.719,-54.543,41.087],[16.872,-53.462,42.161],
[23.298,-54.816,43.579],[32.965,-55.140,46.120],[18.778,-56.516,45.593],
[25.431,-45.433,39.374],[21.602,-49.913,47.312],[29.847,-56.257,42.316],
[17.916,-53.122,44.636],[20.513,-46.995,47.186],[14.985,-46.093,39.199],
[29.348,-45.507,44.205],[30.883,-56.694,46.729],[22.508,57.653,43.253],
[29.201,-43.431,42.260],[28.601,-45.735,39.659],[30.726,-46.960,49.342],
[24.964,-49.159,36.645],[25.360,-51.344,35.189],[31.424,-51.291,46.534],
[28.754,-49.351,40.062],[29.000,-51.271,47.751],[19.660,-48.612,41.055],
[21.796,-45.361,42.509],[21.108,-51.969,44.328],[26.672,-45.495,45.225],
[22.714,-46.337,48.842],[29.148,-51.390,36.792],[19.347,-51.049,49.147],
[25.926,-42.843,40.431],[20.233,-44.062,49.792],[31.558,-57.725,36.943],
[20.012,-50.646,39.064],[31.699,-44.064,35.106],[15.237,-47.951,46.893],
[16.185,-46.221,44.365],[36.582,-44.329,49.498],[25.278,-46.739,48.272],
[23.599,-41.228,38.583],[19.800,-41.184,40.179],[27.176,-42.529,35.178],
[27.822,-42.491,38.225],[30.574,-53.958,50.160],[18.751,-46.561,49.260],
[20.988,-57.825,45.183],[26.417,-57.634,47.463],[28.409,-53.590,48.836],
[28.804,-51.056,34.302],[25.570,-56.726,35.184],[19.827,-53.580,46.424],
[21.693,-53.090,39.548],[25.105,-52.165,38.458],[25.658,-51.359,50.795],
[27.728,-53.583,43.807],[17.041,-55.076,46.973],[23.054,-55.728,34.964],
[15.191,-55.107,40.697],[16.074,-54.417,36.425],[26.044,-52.930,45.826],
[22.619,-57.767,43.013],[20.497,-52.477,41.897],[21.288,-52.266,36.794],
[25.976,-55.708,44.528],[19.497,-53.500,34.978],[30.670,-42.988,51.908],
[18.473,-48.942,50.909],[24.370,-51.701,43.334],[18.902,-58.357,40.292],
[23.916,-54.626,46.311],[23.057,-47.611,38.080],[30.097,-49.180,44.889],
[28.294,-54.128,35.634],[24.370,-49.389,47.736],[28.464,-54.255,38.239],
[17.442,-49.799,43.022],[26.270,-46.823,37.361],[16.410,-51.024,47.335],
[23.271,-50.973,49.522],[27.676,-45.395,47.524],[34.032,-52.544,36.797],
[33.107,-45.924,48.272],[32.522,-45.480,38.986],[27.034,-48.699,43.747],
[21.624,-45.851,39.449],[17.992,-50.875,34.813],[23.045,-49.552,40.316],
[23.195,-42.691,46.813],[21.272,-43.331,40.633],[26.772,-49.926,46.064],
[26.996,-47.447,34.879],[29.883,-44.369,49.065],[32.651,-43.031,37.233],
[17.643,-56.123,49.637],[28.918,-46.663,36.924],[17.780,-47.487,46.154],
[22.117,-45.214,34.043],[21.565,-45.841,36.652],[18.281,-47.652,36.612],
[26.183,-54.805,47.778],[27.000,-56.444,37.565],[25.282,-44.839,35.882],
[23.678,-54.426,37.370],[25.486,-52.243,48.268],[25.995,-56.732,40.254],
[21.588,-43.232,37.882],[21.735,-57.031,37.081],[33.173,-55.718,50.580],
[30.016,-51.179,41.974],[19.459,-44.038,36.716],[17.092,-52.968,32.604],
[30.481,-42.187,45.130],[18.784,-58.114,43.523],[18.141,-44.087,40.795],
[16.991,-42.247,50.518],[21.259,-44.067,52.385],[26.878,-44.222,32.907],
[27.266,-44.868,50.436],[28.985,-44.846,34.500],[15.825,-48.647,49.564],
[19.532,-47.115,31.772],[21.862,-48.712,50.098],[24.637,-46.097,33.366],
[27.874,-48.231,48.511],[20.280,-51.136,51.970],[20.009,-49.895,36.428],
[21.902,-53.057,48.329],[22.703,-50.439,35.292],[26.245,-48.796,51.417],
[31.378,-48.858,34.993],[21.997,-55.135,49.755],[21.471,-55.910,47.056],
[23.571,-51.868,46.276],[29.172,-53.564,52.391],[29.103,-50.462,50.207],
[19.502,-54.175,49.051],[24.790,-59.149,49.069],[22.719,-55.224,52.410],
[24.358,-56.295,49.891],[27.535,-56.212,49.816],[17.610,-58.304,50.859],
[21.825,58.305,48.907],[24.198,-58.906,52.225],[22.645,-40.780,36.025],
[16.707,-46.059,34.491],[16.744,-41.660,40.939],[15.578,-43.989,42.578],
[35.405,-40.872,43.815],[32.983,-43.203,47.297],[31.257,-47.720,53.700],
[31.286,-47.828,37.776],[32.702,-48.763,39.916],[33.428,-46.064,45.502],
[30.146,-46.473,46.590],[32.356,-50.516,37.454],[30.892,-53.025,35.570],
[30.293,-52.502,39.421],[14.921,-48.924,43.638],[31.799,-50.226,48.991],
[32.904,-52.553,50.879],[29.593,-55.755,33.504],[30.814,-55.402,39.012],
[30.481,-53.676,43.233],[30.328,-54.000,46.096],[16.303,-53.272,49.180],
[34.565,-58.370,50.593],[31.139,-57.850,40.357],[15.957,-56.745,43.255],
[14.643,-52.648,44.127],[30.258,-56.606,51.290],[32.988,-56.703,34.844],
[14.767,-58.344,38.369],[30.782,-43.215,39.589],[16.152,-57.717,46.137],
[18.191,-45.299,51.671],[20.348,-44.275,46.811],[24.972,-42.564,50.436],
[29.679,-41.230,49.183],[31.237,-58.670,48.270],[18.330,-42.106,48.248],
[20.065,-40.263,47.168],[24.973,-41.184,44.654],[27.047,-42.742,46.276],
[30.390,-39.695,45.896],[17.892,-44.436,45.726],[19.499,-43.148,43.266],
[22.570,-42.340,43.696],[25.165,-58.063,43.817],[29.913,-58.411,43.973],
[32.076,59.135,41.829],[16.905,59.538,41.887],[16.805,-58.166,36.205],
[23.840,-59.448,40.425],[27.316,-59.018,42.574],[28.178,-58.940,38.798],
[17.902,-41.601,38.304],[19.127,-57.504,37.700],[24.331,-57.985,37.959],
[24.904,-41.275,34.822],[34.484,-43.106,39.558],[19.715,-59.760,51.073],
[20.927,-57.739,52.395],[18.323,-42.111,34.798],[28.875,-58.614,50.179],
[32.213,-44.671,50.491],[25.276,-30.391,41.815],[24.656,-31.471,44.212],
[18.857,-30.929,45.390],[29.682,-27.583,43.553],[18.506,-27.934,44.348],
[17.335,-32.721,39.524],[15.918,-32.123,42.172],[25.612,-23.976,42.538],
[24.487,-28.361,43.374],[33.385,-38.694,45.067],[27.316,-35.634,44.199],
[21.360,-36.442,41.818],[28.319,-36.429,39.657],[17.122,-34.425,41.650],
[23.529,-32.440,41.847],[33.145,-35.070,48.209],[19.093,-37.046,46.674],
[26.643,-28.433,40.137],[21.238,-30.110,48.165],[30.079,-37.309,42.396],
[18.577,-34.902,44.494],[18.964,-27.924,47.532],[17.236,-27.531,38.971],
[29.843,-26.758,46.633],[30.766,-36.674,46.145],[22.742,-38.993,43.249],
[29.516,-24.391,41.653],[29.567,-27.365,40.872],[29.948,-28.704,49.280],
[25.212,-30.588,39.049],[25.870,-32.867,34.463],[32.634,-32.363,46.094],
[29.751,-32.009,39.109],[28.019,-33.786,48.022],[20.810,-30.461,41.073],
[21.700,-27.845,43.251],[21.912,-30.578,43.544],[24.145,-23.544,45.492],
[21.528,-27.192,47.432],[27.321,-32.249,37.651],[18.695,-31.920,47.734],
[25.731,-25.720,39.053],[20.449,-24.600,50.448],[32.311,-39.410,36.670],
[20.072,-31.692,38.482],[32.674,-26.709,35.552],[15.409,-30.610,46.975],
[15.212,-25.797,44.358],[33.886,-26.249,49.673],[24.005,-26.476,49.125],
[23.049,-22.385,38.027],[19.182,-22.030,39.739],[28.280,-23.494,35.421],
[28.282,-24.480,39.190],[30.303,-34.934,51.026],[17.313,-29.700,48.688],
[20.897,-38.617,45.104],[26.448,-40.246,47.530],[29.954,-35.798,48.439],
[29.676,-30.736,34.183],[24.708,-37.981,36.251],[20.908,-35.081,45.621],
[21.063,-33.888,40.261],[25.184,-33.219,39.668],[25.452,-33.075,51.240],
[26.534,-34.574,41.557],[18.420,-34.641,47.473],[21.226,-36.837,33.927],
[15.429,-36.626,40.743],[16.451,-35.153,38.480],[25.978,-32.888,46.087],
[23.257,-38.333,40.425],[18.682,-36.468,40.145],[21.070,-36.552,38.829],
[28.002,-38.223,44.956],[20.811,-34.167,36.835],[31.201,-22.355,50.906],
[19.169,-29.174,51.829],[22.615,-32.823,45.097],[20.502,-38.853,41.264],
[23.419,-36.517,46.278],[23.211,-29.217,36.226],[30.345,-30.934,47.135],
[27.755,-37.420,35.194],[24.919,-29.221,48.463],[27.331,-35.516,37.021],
[18.230,-30.038,42.074],[26.246,-29.290,36.961],[16.244,-33.442,49.070],
[23.536,-31.555,49.732],[27.103,-27.739,48.431],[33.092,-34.286,36.984],
[33.622,-27.573,46.660],[30.526,-27.507,36.926],[28.840,-30.287,42.391],
[20.919,-26.850,39.778],[16.542,-32.459,34.654],[23.348,-28.176,39.230],
[23.703,-26.644,45.844],[21.780,-24.185,39.272],[26.305,-28.813,45.354],
[28.426,-28.288,35.245],[29.223,-24.256,49.937],[32.514,-23.098,35.423],
[16.721,-37.162,48.085],[29.099,-29.711,37.693],[17.040,-25.960,46.531],
[23.248,-27.085,33.562],[21.615,-27.254,36.828],[17.643,-30.194,37.758],
[26.023,-35.903,46.860],[27.082,-38.464,37.828],[23.963,-26.017,36.032],
[24.037,-36.118,39.431],[23.668,-33.819,48.084],[26.099,-37.930,40.304],
[21.363,-24.495,36.246],[20.793,-38.426,36.080],[32.065,-37.114,49.573],
[29.882,-32.579,43.432],[19.344,-27.403,34.931],[18.062,-34.355,36.265],
[31.094,-25.182,44.811],[18.084,-38.891,44.639],[18.590,-25.350,40.242],
[17.920,-24.635,51.661],[21.858,-24.116,52.718],[25.421,-25.659,33.403],
[27.160,-26.010,51.099],[28.834,-25.767,34.416],[16.130,-29.317,51.369],
[21.109,-29.131,34.041],[21.776,-28.428,50.245],[25.504,-28.484,34.347],
[27.726,-30.473,48.619],[18.520,-31.864,51.434],[23.393,-32.244,36.966],
[21.161,-32.874,48.060],[23.492,-33.958,34.888],[26.002,-30.498,51.521],
[32.020,-29.860,32.412],[22.661,-36.126,49.298],[21.414,-37.775,47.486],
[23.110,-30.480,46.283],[27.491,-35.150,51.406],[28.416,-32.786,50.628],
[20.841,-33.511,50.975],[23.210,-39.692,46.889],[23.676,-35.420,51.916],
[25.724,-35.084,49.258],[28.469,-37.469,50.560],[18.672,-35.369,52.162],
[21.631,-40.237,50.841],[23.386,-37.921,52.426],[22.713,-21.479,35.656],
[16.928,-29.306,34.962],[16.251,-23.455,40.207],[17.353,-25.480,42.459],
[33.774,-24.546,44.728],[33.750,-23.833,47.893],[32.514,-28.694,49.200],
[32.255,-30.696,38.016],[31.527,-30.226,41.082],[31.656,-28.906,45.021],
[30.692,-24.318,47.633],[33.035,-32.153,35.313],[32.090,-34.392,34.497],
[30.677,-34.170,41.142],[15.867,-30.337,44.290],[30.731,-31.598,50.591],
[33.069,-33.464,50.715],[29.776,-35.515,34.844],[30.192,-36.130,37.702],
[29.944,-35.208,43.848],[30.396,-33.993,46.445],[14.751,-36.655,50.888],
[32.678,-41.133,51.139],[31.706,-38.055,38.963],[17.260,-36.752,42.791],
[15.256,-34.556,43.452],[33.647,-37.568,47.736],[31.674,-37.398,34.873],
[15.376,-38.419,38.673],[30.331,-22.988,38.463],[16.301,-39.792,46.864],
[17.887,-26.620,49.642],[19.875,-25.307,48.057],[25.083,-21.832,50.195],
[27.478,-22.535,48.676],[32.202,-39.899,48.647],[18.166,-19.896,49.081],
[19.262,-23.816,45.913],[26.103,-21.593,45.047],[26.486,-24.839,46.699],
[28.795,-21.747,45.932],[16.942,-23.205,44.626],[20.244,-25.646,42.818],
[22.964,-25.248,43.485],[24.888,-37.139,42.696],[31.887,-39.769,43.177],
[31.875,-40.549,40.247],[17.792,-39.241,41.589],[17.383,-40.059,36.334],
[24.636,-40.701,41.818],[29.003,-39.968,43.097],[28.844,-39.759,39.425],
[16.236,-25.331,37.768],[18.250,-38.062,34.792],[22.702,-38.499,37.946],
[25.496,-24.121,36.884],[32.195,-25.101,37.664],[17.055,-39.070,50.224],
[19.464,-40.196,53.232],[18.370,-23.303,34.874],[28.672,-39.473,52.367],
[31.380,-26.084,50.183],[29.511,-12.117,40.772],[26.416,-12.668,43.686],
[19.281,-14.039,45.295],[30.279,-9.284,42.239],[19.564,-8.755,43.955],
[16.551,-13.873,39.351],[15.376,-13.312,42.274],[28.311,-5.663,42.661],
[23.927,-10.148,44.513],[32.934,-21.762,44.461],[28.795,-15.493,44.385],
[20.448,-17.919,42.234],[28.079,-19.674,40.263],[16.416,-16.619,41.613],
[22.710,-15.145,42.497],[33.328,-17.846,48.107],[19.860,-18.575,45.101],
[27.525,-9.953,41.158],[22.174,-10.614,47.060],[30.077,-20.175,41.969],
[17.890,-16.459,44.212],[19.703,-11.563,46.154],[16.992,-9.870,37.369],
[28.991,-8.458,44.763],[31.049,-19.346,46.741],[22.618,-21.101,43.227],
[28.551,-7.322,40.489],[30.240,-8.185,38.553],[30.762,-9.944,47.577],
[26.516,-11.911,39.038],[25.314,-14.404,34.799],[32.974,-14.975,46.700],
[30.506,-12.390,37.614],[27.739,-15.040,48.679],[21.849,-12.228,41.760],
[22.334,-8.712,42.633],[21.732,-12.668,44.488],[25.751,-8.399,45.516],
[22.847,-8.701,49.916],[27.573,-13.592,36.770],[18.409,-13.835,48.084],
[25.766,-8.036,40.550],[18.061,-6.054,49.209],[32.162,-20.750,37.107],
[19.102,-14.413,38.937],[33.214,-8.899,34.365],[16.573,-10.486,46.311],
[17.109,-10.134,43.597],[34.068,-7.672,49.063],[23.670,-8.198,47.402],
[25.968,-4.768,40.902],[19.468,-3.557,40.674],[28.993,-5.846,34.910],
[29.044,-5.822,37.586],[30.493,-14.771,48.032],[18.745,-9.725,48.061],
[20.983,-21.178,45.710],[28.341,-20.084,48.054],[29.348,-17.254,48.411],
[30.690,-13.513,34.834],[25.292,-20.189,33.843],[21.246,-16.364,44.890],
[20.784,-16.005,40.183],[24.154,-13.322,37.937],[27.004,-16.120,51.395],
[28.054,-16.988,41.677],[19.520,-16.690,47.008],[21.816,-19.280,33.550],
[14.415,-18.440,40.098],[18.140,-16.459,36.304],[26.367,-15.331,46.077],
[23.425,-16.743,40.308],[18.636,-17.522,39.881],[22.081,-17.428,35.971],
[26.573,-18.739,45.816],[20.475,-15.069,36.446],[31.022,-3.331,51.577],
[20.508,-10.090,50.127],[25.409,-15.534,42.613],[20.410,-19.537,39.896],
[23.989,-17.085,46.039],[25.275,-11.571,36.174],[31.476,-12.629,46.344],
[29.432,-18.651,36.560],[25.066,-10.600,47.986],[27.163,-16.729,36.413],
[18.193,-12.571,42.041],[26.815,-9.032,37.198],[15.323,-12.667,47.597],
[22.681,-14.425,50.548],[26.423,-7.860,48.582],[33.764,-15.704,36.801],
[32.057,-9.924,49.953],[31.489,-7.108,36.267],[28.979,-11.484,43.349],
[20.524,-7.835,40.517],[16.827,-13.848,34.762],[24.061,-11.280,40.288],
[23.932,-6.021,45.785],[22.044,-6.058,38.807],[27.410,-10.563,45.832],
[29.283,-8.941,35.892],[30.104,-7.379,48.321],[33.046,-5.499,34.737],
[17.194,-18.115,47.186],[28.735,-10.408,38.694],[18.110,-7.183,46.470],
[22.836,-8.777,33.034],[23.082,-10.201,37.369],[17.914,-12.511,37.106],
[24.414,-19.335,48.594],[26.599,-19.419,36.828],[24.234,-7.844,36.829],
[26.595,-16.950,39.255],[24.703,-15.726,48.545],[25.279,-21.113,40.908],
[21.878,-8.259,35.666],[20.944,-19.731,37.185],[31.428,-17.573,49.964],
[30.506,-16.316,42.355],[19.385,-7.019,36.866],[16.690,-16.438,33.213],
[30.365,-5.896,45.735],[18.373,-21.323,43.948],[18.873,-9.034,39.056],
[16.029,-5.066,51.466],[21.149,-5.312,53.024],[25.529,-7.510,33.843],
[27.388,-8.663,51.341],[31.081,-7.269,33.540],[16.005,-13.106,50.512],
[21.678,-11.134,35.085],[22.512,-11.758,51.124],[25.228,-10.437,32.848],
[29.458,-12.238,48.394],[18.555,-12.257,50.818],[21.309,-12.886,37.845],
[21.888,-15.380,47.553],[21.039,-13.146,33.559],[25.417,-11.923,51.089],
[32.437,-12.394,33.051],[21.788,-17.548,49.073],[21.982,-20.638,48.472],
[23.209,-12.973,47.888],[30.065,-16.805,51.968],[28.227,-13.380,50.620],
[20.100,-15.898,50.664],[24.202,-21.973,47.759],[22.979,-16.882,51.958],
[26.501,-17.723,48.677],[27.986,-18.517,51.523],[18.220,-18.479,51.371],
[22.351,-22.594,50.268],[23.100,-19.470,50.964],[22.163,-3.758,35.809],
[16.660,-9.515,34.267],[17.366,-4.021,38.940],[16.779,-4.702,41.792],
[32.319,-4.944,44.192],[33.127,-5.823,47.163],[30.014,-8.096,51.834],
[34.063,-11.452,37.429],[32.928,-14.049,39.093],[33.573,-10.148,44.174],
[31.758,-8.030,45.755],[34.811,-13.008,34.881],[31.757,-16.479,35.329],
[30.025,-14.694,39.537],[15.596,-12.126,44.473],[32.065,-13.938,50.580],
[34.588,-15.139,51.487],[31.143,-17.490,32.515],[29.568,-16.155,37.403],
[32.617,-17.535,43.740],[31.151,-16.607,45.869],[15.504,-15.810,51.608],
[33.925,-23.026,52.499],[30.524,-19.868,38.977],[16.584,-18.811,43.920],
[15.136,-15.713,43.884],[33.870,-21.252,50.346],[32.418,-19.303,34.564],
[15.914,-20.431,36.454],[31.142,-4.748,38.896],[14.651,-20.068,48.429],
[17.952,-8.817,50.671],[20.876,-8.006,48.162],[25.902,-4.231,49.050],
[29.138,-5.041,49.802],[32.169,-21.566,48.145],[18.117,-2.743,48.788],
[18.652,-3.545,46.284],[26.159,-4.770,44.329],[27.909,-5.632,47.055],
[30.312,-2.845,46.362],[17.301,-6.369,44.002],[19.238,-5.879,42.189],
[22.329,-4.988,43.788],[25.385,-18.986,42.755],[32.015,-24.007,40.644],
[32.739,-21.327,41.573],[17.276,-20.765,41.152],[16.565,-22.939,37.078],
[22.400,-21.713,40.629],[28.277,-21.759,43.280],[27.566,-22.025,37.329],
[17.000,-7.256,37.988],[18.188,-19.295,35.059],[23.735,-19.920,38.645],
[24.593,-4.915,36.314],[32.761,-7.099,38.692],[17.560,-21.920,50.722],
[21.058,-21.098,52.179],[18.842,-5.652,34.571],[28.576,-21.857,51.284],
[31.995,-6.204,50.588],[27.175,4.247,40.152],[25.519,4.809,44.921],
[18.164,3.543,44.863],[29.235,8.409,43.226],[19.196,9.038,43.010],
[16.246,3.430,38.831],[16.266,5.516,41.311],[27.259,13.264,43.149],
[24.226,6.076,43.077],[32.261,-0.517,46.163],[28.031,3.837,45.349],
[20.613,0.172,42.236],[29.082,1.592,40.776],[15.956,2.734,42.035],
[21.847,3.472,40.391],[32.762,2.266,47.025],[19.948,-1.515,45.050],
[25.051,8.927,41.569],[21.069,6.194,46.576],[30.198,-0.925,42.897],
[18.581,1.583,43.095],[18.933,7.990,46.329],[15.900,8.828,37.060],
[29.307,10.490,45.145],[29.272,-0.123,45.736],[24.119,-3.338,42.777],
[29.642,11.192,42.057],[28.354,10.632,39.759],[29.387,8.855,48.043],
[25.672,6.818,39.389],[25.185,3.265,35.895],[32.458,5.572,46.023],
[29.044,6.547,39.017],[27.868,4.004,48.013],[20.091,5.870,40.636],
[21.891,10.497,44.239],[21.636,5.090,43.996],[26.882,11.253,46.142],
[20.462,8.377,49.421],[29.074,3.502,38.063],[18.648,5.270,47.833],
[25.979,11.817,40.465],[20.776,12.814,49.108],[30.938,-3.595,36.028],
[17.899,5.575,39.057],[31.927,11.922,35.878],[14.118,6.190,45.816],
[15.986,10.804,44.343],[35.292,13.566,48.282],[24.608,9.811,47.501],
[26.353,14.028,38.653],[19.391,14.022,41.083],[27.472,13.233,35.882],
[28.881,13.101,38.295],[30.765,4.422,52.334],[18.487,11.210,48.889],
[21.285,-2.611,42.785],[26.214,-3.136,46.706],[29.008,2.134,50.679],
[29.919,3.791,34.705],[23.808,-0.461,35.804],[20.446,2.439,45.176],
[19.256,3.066,38.255],[24.765,2.803,38.919],[26.550,3.102,50.263],
[26.139,2.299,43.802],[18.946,1.527,47.930],[22.346,1.672,34.952],
[14.639,-0.814,41.580],[17.622,-0.256,36.624],[25.931,2.498,46.819],
[23.420,0.898,41.774],[18.170,3.049,40.770],[21.976,2.482,37.780],
[27.292,-1.822,44.627],[20.141,3.882,35.351],[31.430,13.455,51.782],
[17.735,7.615,49.617],[23.492,3.133,43.180],[19.274,-0.196,39.380],
[23.574,0.987,45.560],[23.626,7.213,37.177],[30.183,6.409,44.738],
[26.889,-1.397,34.809],[25.856,6.505,49.372],[27.891,1.839,36.173],
[19.279,5.966,43.366],[25.737,8.563,36.398],[15.091,5.449,48.498],
[22.228,4.002,50.386],[27.122,9.489,49.503],[34.160,3.776,35.758],
[32.711,8.337,50.459],[31.064,10.153,38.660],[27.604,4.957,42.776],
[21.968,9.745,40.309],[18.237,6.066,35.894],[21.371,7.503,38.870],
[22.798,11.733,47.518],[21.416,12.267,40.135],[27.841,7.703,45.741],
[28.251,8.498,35.080],[29.659,11.816,49.731],[33.816,12.767,33.801],
[14.324,-0.819,47.842],[29.044,9.014,37.477],[18.554,10.677,45.153],
[24.144,10.291,33.102],[22.406,9.541,36.460],[19.530,8.405,37.217],
[27.157,-0.173,47.424],[26.750,-0.867,37.797],[24.668,10.022,38.599],
[25.080,0.462,39.503],[24.207,3.307,48.772],[24.042,-1.648,40.871],
[21.958,12.502,36.640],[20.664,-1.269,36.784],[29.651,-1.998,48.994],
[29.613,2.635,43.347],[19.810,11.001,35.486],[18.002,2.254,35.292],
[30.340,13.440,45.400],[18.499,-3.421,43.683],[18.574,11.780,39.440],
[16.453,13.588,50.192],[21.199,11.884,52.480],[26.304,10.676,34.884],
[26.888,10.001,52.662],[28.757,11.436,33.639],[14.948,6.063,51.048],
[20.920,6.849,35.146],[23.050,6.578,49.522],[25.704,7.997,33.714],
[28.393,6.568,48.622],[17.029,4.027,49.677],[22.338,5.023,37.899],
[21.142,3.695,47.808],[22.887,4.702,34.639],[27.053,7.365,51.556],
[30.614,6.522,31.866],[22.995,-0.738,49.093],[21.293,-0.009,47.088],
[23.423,4.772,46.628],[26.151,1.565,52.400],[28.324,4.753,51.775],
[19.335,3.104,50.383],[23.208,-4.014,47.811],[23.866,1.631,50.928],
[25.706,-0.284,50.150],[28.757,-0.880,51.346],[19.218,0.878,52.182],
[22.240,-3.794,50.207],[23.255,-2.886,52.536],[23.020,14.310,34.517],
[16.582,8.910,34.362],[16.663,14.968,38.508],[16.408,12.008,41.606],
[35.130,14.861,45.070],[32.712,12.877,48.541],[30.284,9.773,51.622],
[29.974,6.387,36.427],[33.191,5.346,38.297],[33.669,9.071,44.756],
[31.543,10.836,47.538],[32.586,5.384,34.136],[32.055,2.169,34.740],
[30.761,4.018,41.276],[15.375,6.757,43.477],[30.399,6.393,50.417],
[33.263,3.911,50.849],[29.767,-0.999,35.384],[31.684,2.497,37.820],
[30.738,3.747,45.567],[30.935,1.212,48.718],[15.524,0.610,50.989],
[35.057,-2.065,52.229],[32.505,-1.441,37.262],[16.676,-0.311,43.539],
[14.775,4.090,44.104],[32.718,-1.763,49.927],[33.720,0.144,35.417],
[16.009,-0.612,38.892],[30.979,12.740,40.167],[16.215,-2.487,46.008],
[15.887,8.433,48.067],[20.453,9.901,47.403],[24.084,12.403,49.636],
[28.404,16.044,49.447],[33.434,-3.072,47.302],[17.415,13.650,47.444],
[20.461,14.752,46.675],[26.197,15.395,45.019],[27.925,13.506,47.032],
[29.911,15.601,46.899],[17.839,13.312,44.354],[20.036,11.849,42.475],
[22.921,12.819,42.814],[26.453,-0.292,42.349],[30.819,-3.611,41.258],
[33.236,-4.546,40.880],[16.860,-2.103,41.457],[16.703,-3.010,36.066],
[22.462,-4.127,40.692],[29.457,-3.753,43.996],[28.967,-2.751,38.302],
[17.647,12.754,37.011],[18.634,-2.773,34.331],[23.939,-2.692,38.489],
[24.884,13.088,36.290],[33.236,12.510,38.146],[16.609,-1.761,50.897],
[19.704,-2.343,51.193],[18.198,13.026,34.316],[26.265,-2.508,52.278],
[32.393,10.870,50.265],[26.389,26.360,39.927],[26.040,24.728,44.051],
[19.131,22.530,44.299],[29.361,27.457,41.812],[19.570,27.192,42.757],
[15.467,21.483,39.977],[17.370,23.664,39.757],[27.008,32.435,42.507],
[24.609,26.938,44.980],[33.083,17.088,45.364],[28.599,20.443,44.822],
[19.724,19.400,41.163],[28.597,18.584,39.928],[17.286,21.416,41.809],
[23.110,21.230,40.970],[32.530,19.835,45.877],[18.507,16.965,47.589],
[23.510,30.356,39.219],[21.923,24.545,47.076],[30.123,19.635,41.669],
[16.202,20.055,44.097],[19.733,26.191,47.327],[17.342,27.104,38.713],
[29.155,26.916,45.233],[30.192,18.331,46.190],[24.470,15.833,42.745],
[29.898,29.983,41.677],[28.075,29.670,39.311],[31.663,26.616,48.108],
[26.543,23.843,38.844],[26.220,22.745,33.971],[32.865,24.558,46.176],
[30.514,26.406,38.839],[27.588,22.824,48.169],[20.193,24.024,41.428],
[21.579,28.691,41.595],[21.356,24.707,44.147],[28.034,29.110,43.932],
[22.217,28.084,49.943],[26.839,22.872,36.472],[19.735,22.491,47.656],
[25.857,31.495,40.245],[20.641,30.616,49.192],[31.963,14.695,37.350],
[20.915,23.369,38.238],[31.270,29.096,35.222],[17.218,26.395,45.991],
[14.970,29.521,43.688],[35.322,31.168,49.192],[24.913,28.688,48.773],
[24.686,33.839,38.514],[19.087,33.024,40.226],[27.878,32.340,35.056],
[27.400,32.937,37.966],[31.193,22.228,50.096],[17.750,28.755,48.196],
[19.848,17.689,44.886],[26.326,16.854,47.387],[28.922,20.808,49.142],
[28.100,24.666,33.586],[25.319,19.280,35.896],[20.853,20.231,45.032],
[20.963,20.910,39.419],[23.696,22.423,35.358],[23.489,21.886,50.615],
[26.035,20.435,40.711],[18.168,20.719,46.335],[21.538,18.515,33.721],
[14.883,18.250,40.716],[16.035,20.463,36.322],[25.712,21.774,45.872],
[21.912,17.159,39.709],[18.212,21.158,39.246],[20.451,18.442,36.473],
[28.039,17.500,45.053],[20.879,21.029,35.572],[29.872,31.552,51.154],
[19.077,26.594,50.802],[24.762,21.780,43.006],[19.200,18.116,38.861],
[23.454,20.046,44.832],[24.972,25.955,37.217],[29.171,23.987,45.944],
[27.800,15.854,35.842],[26.373,25.352,49.360],[27.958,20.448,36.250],
[17.791,24.939,42.710],[26.764,28.473,36.572],[16.925,23.237,47.623],
[22.203,24.148,50.118],[27.252,29.763,47.391],[32.303,21.790,37.327],
[33.887,28.878,47.853],[31.027,29.322,38.277],[28.618,25.010,43.278],
[19.737,27.983,39.705],[19.561,23.386,34.262],[22.580,26.234,40.366],
[23.031,30.861,47.692],[20.362,30.745,39.757],[26.718,27.173,46.570],
[28.026,27.375,34.502],[30.510,31.945,48.630],[33.017,32.062,35.534],
[16.966,18.962,48.194],[27.836,25.483,36.892],[19.062,29.072,45.204],
[23.954,28.728,31.380],[21.903,28.117,35.072],[18.307,26.717,35.956],
[27.191,19.613,47.043],[27.227,17.946,37.574],[23.869,29.230,36.527],
[24.024,19.365,38.759],[24.680,20.841,48.390],[24.621,16.657,40.154],
[21.651,31.275,36.933],[22.331,16.697,35.425],[31.228,19.430,49.221],
[28.734,22.108,42.870],[19.758,29.391,34.041],[18.466,20.355,34.253],
[30.011,31.079,45.469],[18.627,15.693,43.077],[17.223,30.962,39.949],
[16.043,30.636,51.430],[20.638,31.306,52.066],[25.128,31.718,33.212],
[26.993,29.110,50.850],[29.065,29.870,33.931],[15.127,26.325,49.582],
[22.100,25.386,34.211],[23.911,26.064,48.577],[25.185,27.187,33.402],
[28.876,25.409,48.219],[16.963,24.196,50.379],[21.613,25.686,37.212],
[22.228,21.473,47.069],[21.817,23.460,31.648],[26.579,25.013,52.076],
[30.659,26.249,34.063],[23.130,18.590,49.458],[21.726,18.721,46.999],
[25.314,24.313,46.880],[28.374,20.847,51.966],[28.507,24.106,50.593],
[20.173,21.151,50.249],[23.379,16.596,47.410],[24.603,18.908,51.711],
[26.821,19.135,50.025],[29.183,17.348,51.784],[19.768,17.972,50.439],
[21.702,15.326,50.001],[23.557,14.878,52.331],[22.987,33.861,36.398],
[17.580,27.041,32.824],[15.870,33.244,39.319],[14.932,31.033,41.545],
[32.324,31.970,43.739],[32.791,31.304,46.359],[31.375,27.689,50.604],
[32.607,27.326,36.825],[32.279,24.443,39.346],[32.489,27.618,45.543],
[31.169,29.337,47.856],[33.223,24.277,36.671],[32.143,21.788,34.499],
[31.848,21.716,40.135],[15.185,25.877,41.255],[31.592,24.797,50.303],
[33.674,22.544,51.192],[29.522,19.003,33.651],[30.270,20.210,38.340],
[31.103,22.116,44.941],[31.110,22.356,47.524],[15.377,19.252,50.449],
[32.468,15.743,52.196],[31.708,17.030,39.592],[17.661,18.274,42.370],
[16.225,22.972,43.839],[33.231,17.678,50.332],[32.352,18.602,33.237],
[16.306,17.600,38.023],[32.223,31.634,40.911],[16.932,16.000,45.423],
[18.423,29.425,50.933],[21.399,29.349,46.241],[24.869,31.578,49.653],
[28.061,32.823,48.897],[32.983,15.250,47.351],[18.625,34.027,46.322],
[21.195,32.169,45.871],[25.730,33.379,44.790],[27.503,32.171,46.334],
[30.120,33.807,45.574],[16.911,32.434,44.088],[18.733,31.210,42.437],
[22.776,30.800,42.783],[27.387,17.871,42.391],[30.822,13.853,42.658],
[31.361,16.497,42.368],[17.401,16.191,40.766],[17.383,15.231,35.638],
[23.356,14.389,40.080],[28.664,15.647,41.276],[29.300,15.945,38.053],
[16.324,31.658,36.797],[17.916,18.029,35.711],[24.110,16.531,37.522],
[25.360,32.020,36.671],[32.948,31.787,38.266],[18.803,14.995,49.483],
[20.896,14.797,52.811],[18.203,31.870,34.494],[28.676,13.803,51.375],
[32.320,30.055,50.272],[28.443,42.182,41.462],[26.445,42.798,43.821],
[18.410,41.956,44.491],[28.621,44.962,41.833],[19.650,45.667,43.237],
[15.220,40.663,39.477],[16.869,42.836,39.984],[26.984,49.351,43.000],
[24.697,45.380,44.738],[32.114,35.987,44.522],[28.186,38.712,45.654],
[19.781,38.903,40.837],[27.688,37.405,40.447],[16.549,39.199,41.296],
[23.281,42.290,41.084],[33.797,38.956,48.148],[18.799,36.646,44.663],
[26.123,46.309,40.109],[22.121,43.194,47.168],[29.705,38.099,42.427],
[18.268,39.213,45.027],[19.838,44.358,45.485],[17.530,45.890,38.398],
[32.059,48.007,42.026],[30.174,36.436,46.206],[22.730,33.571,43.142],
[29.072,47.637,42.983],[28.165,47.928,38.648],[30.674,45.159,47.830],
[25.587,42.847,37.253],[27.164,41.907,34.853],[31.450,42.646,47.910],
[27.557,44.248,38.750],[28.080,41.472,46.538],[21.523,44.690,40.919],
[22.266,46.774,43.247],[22.555,43.710,44.143],[26.261,48.178,45.517],
[22.853,46.377,46.780],[28.794,43.137,36.455],[19.368,42.102,47.095],
[26.213,48.996,40.336],[19.730,49.086,49.604],[30.756,34.334,34.645],
[20.761,41.516,40.106],[32.604,47.550,34.696],[15.128,44.792,47.065],
[17.537,48.301,44.728],[33.371,48.307,49.147],[24.430,47.414,49.121],
[25.567,51.509,39.270],[18.894,51.909,39.769],[28.124,49.471,35.737],
[28.095,50.963,38.193],[31.120,40.066,49.969],[18.066,45.930,47.655],
[20.691,34.833,44.895],[25.933,34.246,47.463],[28.512,38.522,48.699],
[30.274,41.058,33.210],[25.805,38.629,35.124],[21.566,38.355,43.253],
[20.958,39.065,38.443],[26.172,41.159,39.891],[25.905,40.811,50.823],
[26.132,39.194,41.723],[16.882,41.050,47.235],[22.082,38.134,33.288],
[14.470,37.718,41.327],[16.549,39.344,37.693],[25.840,40.204,44.984],
[22.233,37.465,40.785],[18.388,40.728,38.809],[21.540,41.030,36.451],
[27.714,36.417,44.522],[20.057,42.829,37.694],[31.883,53.506,50.057],
[18.964,44.232,50.260],[23.599,40.151,42.581],[19.074,36.680,38.835],
[22.856,40.097,45.219],[23.494,45.886,37.369],[31.390,45.030,45.149],
[28.662,38.029,36.604],[26.561,45.137,48.436],[30.271,40.823,35.993],
[19.161,43.062,42.133],[26.798,46.242,36.865],[17.918,41.916,49.643],
[22.333,40.136,50.170],[27.358,48.036,48.025],[32.964,40.583,36.436],
[33.428,45.452,49.264],[31.761,47.588,37.330],[28.828,43.855,44.535],
[20.690,47.259,40.868],[19.455,42.748,34.678],[23.168,46.390,39.902],
[24.296,48.951,46.934],[20.896,49.726,40.130],[27.317,45.762,46.005],
[28.103,45.450,34.626],[31.538,50.538,48.934],[33.437,50.151,35.287],
[17.679,37.209,47.888],[29.815,45.624,37.660],[18.621,49.402,47.002],
[23.331,47.144,33.452],[21.027,47.007,34.901],[19.120,45.284,35.912],
[25.370,36.666,45.867],[26.417,36.470,37.059],[24.916,48.086,38.052],
[24.083,40.550,37.578],[24.313,38.906,47.862],[24.726,36.800,39.306],
[20.851,49.550,35.408],[21.514,36.507,37.277],[30.605,36.140,49.561],
[28.548,40.377,43.462],[18.957,48.418,36.795],[16.115,39.228,34.731],
[31.532,48.151,44.497],[18.551,34.469,43.035],[18.491,48.573,39.782],
[16.651,50.581,48.907],[21.509,49.402,52.849],[25.380,49.517,34.339],
[27.154,48.060,50.675],[29.395,48.373,33.590],[16.490,43.744,51.324],
[20.249,45.012,33.394],[22.982,45.029,50.122],[24.722,45.519,34.917],
[28.876,42.839,48.744],[18.818,39.340,51.509],[23.016,43.420,36.495],
[21.686,40.632,47.519],[22.870,41.198,33.840],[25.317,43.511,50.626],
[32.463,44.284,33.124],[21.643,38.063,48.252],[22.390,37.267,45.574],
[24.419,41.518,47.059],[26.990,38.467,50.929],[28.413,41.479,51.152],
[20.494,41.470,51.628],[23.173,33.979,46.052],[24.180,38.628,51.540],
[25.082,36.524,49.064],[27.778,36.093,50.475],[19.251,36.684,50.364],
[23.428,33.750,48.623],[22.684,34.008,51.693],[21.150,52.146,35.057],
[16.674,46.068,35.196],[15.530,52.246,39.445],[16.506,49.997,41.298],
[33.028,50.395,43.868],[33.185,50.145,46.580],[31.503,45.877,51.289],
[35.854,44.937,39.289],[32.836,42.906,39.045],[33.784,45.589,44.345],
[32.795,46.853,46.603],[32.852,42.955,35.416],[31.466,38.790,34.382],
[31.227,40.765,39.562],[15.228,42.990,43.861],[31.770,42.984,50.474],
[33.770,40.869,50.383],[31.585,36.340,33.017],[31.449,38.387,37.305],
[30.864,38.753,44.712],[30.886,40.086,47.195],[16.632,39.162,50.128],
[34.378,35.189,51.800],[30.206,37.394,40.054],[16.872,36.783,42.794],
[14.858,39.776,45.548],[34.196,36.254,49.384],[33.338,37.052,35.581],
[15.960,36.806,39.121],[30.782,47.394,39.970],[16.219,35.069,47.238],
[17.214,48.170,50.063],[21.198,48.746,46.999],[25.870,50.709,49.043],
[28.523,50.113,49.421],[31.986,34.069,47.909],[16.955,53.079,47.676],
[20.499,51.161,46.060],[25.670,51.053,44.542],[28.539,49.534,45.922],
[30.770,51.061,45.159],[16.891,50.784,45.627],[19.510,48.697,42.896],
[23.296,49.685,43.907],[24.871,35.760,42.418],[29.602,32.695,42.097],
[31.444,34.644,42.188],[17.180,34.881,40.908],[17.426,33.953,35.781],
[22.754,32.469,40.517],[28.015,35.135,42.009],[31.281,34.459,39.575],
[16.624,49.257,38.062],[17.434,36.480,36.772],[23.989,36.310,36.093],
[26.044,51.628,36.135],[32.026,50.088,38.164],[18.443,34.418,49.130],
[20.764,33.187,49.961],[18.747,51.808,36.069],[27.004,33.879,51.979],
[32.232,50.152,51.714],[32.421,55.774,44.435],[26.910,58.029,46.202],
[20.343,56.425,42.112],[30.062,56.942,37.962],[14.453,56.936,41.934],
[33.981,56.286,48.627],[18.341,55.885,46.346],[31.103,54.647,40.738],
[18.212,57.953,43.730],[31.130,54.525,46.413],[23.554,52.807,43.278],
[29.901,53.018,35.822],[31.093,58.960,51.414],[20.580,53.910,43.881],
[25.483,52.574,47.095],[28.882,57.812,48.494],[25.417,54.413,34.580],
[20.403,57.765,45.708],[20.956,58.925,41.322],[27.446,57.468,40.678],
[19.192,-59.070,48.225],[22.538,56.954,35.012],[15.813,54.663,40.764],
[16.223,56.266,35.728],[23.154,55.978,41.246],[16.541,57.739,39.400],
[20.726,-59.651,38.565],[27.714,55.535,44.497],[20.663,-59.531,35.666],
[25.349,58.007,43.796],[20.611,55.635,39.017],[23.025,57.470,46.004],
[26.857,55.140,36.558],[28.242,58.536,36.564],[33.436,59.488,38.008],
[17.505,55.408,48.766],[25.574,55.343,46.308],[25.641,55.309,39.116],
[23.121,57.026,38.566],[25.654,56.941,48.898],[24.476,53.617,40.745],
[21.818,54.358,36.521],[32.131,57.792,46.948],[16.898,58.359,34.097],
[18.380,52.933,45.373],[21.262,55.578,47.912],[23.164,54.390,45.558],
[27.303,58.868,52.131],[21.904,58.645,52.555],[22.063,52.690,47.261],
[24.495,57.073,51.904],[27.991,55.752,49.959],[29.924,54.569,51.922],
[17.757,55.829,52.753],[21.933,53.859,49.781],[24.515,54.412,50.742],
[32.432,58.858,35.048],[31.463,54.487,34.155],[32.925,56.911,39.344],
[29.802,56.838,45.651],[29.279,-59.443,46.438],[15.730,58.324,49.705],
[35.470,51.409,51.520],[31.538,53.768,38.181],[16.089,56.531,44.113],
[15.713,-59.626,44.363],[34.501,53.662,49.533],[33.962,54.892,34.853],
[15.990,54.520,37.664],[14.718,53.300,44.820],[32.896,52.495,47.554],
[25.415,55.070,42.938],[29.633,50.134,41.948],[31.732,51.512,40.704],
[17.580,52.580,42.080],[16.140,51.798,36.856],[22.138,51.988,40.901],
[28.112,53.033,42.751],[28.849,53.863,38.231],[19.017,54.806,36.828],
[24.549,53.405,37.690],[16.855,53.587,50.555],[20.081,51.694,52.060],
[29.594,52.071,50.770],[16.874,-53.000,57.684],[28.717,-55.813,58.615],
[26.962,-48.148,57.478],[17.123,-47.146,56.253],[28.776,-46.329,58.212],
[26.137,-49.389,55.091],[26.226,-53.709,52.273],[29.559,-50.093,58.156],
[28.119,-51.274,56.528],[25.942,-44.508,57.597],[29.630,-57.626,55.323],
[19.017,-51.380,56.033],[32.462,-45.226,53.466],[25.840,-41.571,57.427],
[16.653,-42.809,57.749],[29.083,-41.235,54.591],[28.977,-43.367,56.682],
[29.793,-50.352,52.971],[25.790,-56.596,52.846],[20.121,-50.953,58.494],
[25.772,-52.233,55.714],[19.534,-55.165,52.282],[16.127,-56.828,55.139],
[19.007,-54.738,58.808],[20.799,-53.631,55.994],[19.198,-53.014,53.929],
[20.931,-56.589,59.000],[24.116,-48.847,57.159],[25.786,-55.169,55.279],
[28.972,-53.838,56.476],[26.470,-44.491,54.833],[31.395,-51.940,56.065],
[31.639,-44.173,56.296],[16.919,-50.281,52.963],[21.884,-47.357,57.959],
[21.247,-42.966,57.926],[28.846,-47.075,52.835],[33.020,-42.434,54.804],
[29.194,-48.300,56.079],[22.979,-46.597,51.514],[23.175,-46.408,55.735],
[19.521,-48.414,55.692],[27.124,-57.393,57.117],[23.624,-43.028,55.633],
[24.402,-54.119,57.760],[26.022,-56.549,-58.993],[20.007,-42.675,55.562],
[20.255,-57.350,55.150],[20.256,-45.921,54.692],[16.624,-52.873,52.152],
[18.008,-45.553,59.084],[23.785,-44.324,53.285],[29.824,-44.540,53.811],
[21.327,-48.687,52.975],[25.827,-46.313,52.877],[22.090,-50.479,56.282],
[23.019,-51.612,52.921],[31.667,-49.133,51.459],[22.585,-40.373,54.296],
[16.747,-46.874,53.312],[15.297,-40.351,-59.194],[31.904,-49.309,56.658],
[33.491,-50.463,58.525],[32.816,-50.345,53.897],[32.139,-53.006,53.394],
[29.651,-52.581,59.487],[30.310,-55.090,54.384],[31.318,-54.226,57.826],
[34.401,-56.027,57.223],[33.527,-55.419,54.214],[14.943,-56.080,58.610],
[30.382,-41.823,-59.316],[15.397,59.071,55.103],[24.119,-59.120,-59.829],
[29.308,-58.694,57.957],[16.598,-43.659,55.301],[18.266,-58.128,53.415],
[24.323,-57.664,56.659],[25.792,-41.738,54.316],[32.711,-42.377,57.979],
[18.230,-42.793,53.237],[16.127,-34.393,58.511],[27.107,-38.398,-58.058],
[27.668,-29.072,58.224],[17.766,-29.248,57.077],[30.368,-27.917,58.280],
[27.380,-31.688,55.668],[27.075,-32.565,53.158],[27.886,-33.592,57.737],
[29.563,-33.002,53.295],[26.791,-25.133,58.983],[32.018,-39.938,53.774],
[20.546,-31.045,56.913],[32.838,-26.962,52.601],[25.098,-22.629,57.527],
[19.199,-24.694,57.989],[29.511,-23.160,53.274],[27.353,-24.009,56.521],
[28.794,-29.674,51.677],[25.402,-36.435,53.640],[19.970,-35.096,58.723],
[24.666,-32.514,57.200],[21.187,-36.708,51.521],[17.567,-36.877,56.079],
[17.417,-36.629,-59.296],[21.325,-35.659,56.634],[20.560,-34.490,54.170],
[19.753,-37.811,-58.308],[24.749,-30.858,54.802],[27.534,-37.476,55.466],
[26.846,-34.421,55.188],[26.353,-28.031,55.941],[33.838,-33.599,55.982],
[31.061,-26.806,55.972],[16.310,-31.184,54.165],[21.259,-29.984,59.712],
[21.886,-25.162,58.926],[28.518,-28.557,54.319],[33.695,-24.560,54.902],
[30.205,-30.587,56.564],[22.388,-26.799,52.433],[21.731,-26.279,54.910],
[19.949,-29.280,54.412],[27.237,-36.467,58.003],[24.126,-26.028,56.706],
[24.330,-35.493,56.765],[24.753,-37.406,-59.102],[22.429,-23.897,56.350],
[23.051,-36.982,54.886],[18.893,-26.894,55.830],[16.429,-33.398,52.530],
[17.096,-26.761,58.909],[24.480,-24.457,51.413],[29.682,-25.699,52.291],
[21.378,-30.879,51.132],[25.910,-28.212,53.027],[21.748,-31.237,54.313],
[23.576,-33.170,53.509],[32.756,-30.184,52.602],[22.448,-21.756,54.364],
[17.715,-27.544,53.436],[17.002,-22.307,58.222],[32.509,-28.918,54.968],
[33.779,-32.891,-58.882],[32.020,-32.130,54.584],[32.231,-35.044,52.817],
[31.013,-32.539,-59.117],[30.948,-37.946,51.997],[29.977,-35.768,57.376],
[31.761,-39.744,56.955],[33.724,-36.617,54.558],[15.518,-37.959,57.819],
[28.162,-21.080,58.905],[18.360,-39.897,56.409],[22.334,-39.606,58.333],
[29.421,-38.823,58.039],[16.380,-24.750,57.175],[20.270,-37.531,54.115],
[24.463,-38.387,56.791],[26.448,-23.226,54.039],[32.293,-23.036,56.675],
[19.251,-22.870,53.746],[16.794,-15.270,57.756],[28.294,-17.661,58.649],
[27.765,-8.914,59.049],[14.448,-9.695,55.044],[30.513,-10.820,59.089],
[26.083,-12.972,57.067],[26.334,-13.682,53.220],[28.188,-13.648,-59.303],
[28.362,-13.493,55.209],[25.698,-6.102,-59.833],[32.049,-19.926,55.837],
[19.335,-13.059,56.428],[33.473,-7.875,53.688],[26.837,-4.246,56.791],
[18.961,-7.133,-58.785],[28.791,-5.038,53.480],[28.435,-6.068,57.712],
[29.529,-12.903,52.784],[25.579,-19.390,54.952],[20.220,-15.347,58.557],
[27.424,-15.523,57.344],[20.390,-17.696,52.683],[16.193,-17.403,55.869],
[19.362,-17.319,56.749],[21.480,-16.043,56.390],[20.163,-15.417,53.854],
[21.079,-20.652,-59.118],[23.556,-11.708,56.402],[28.280,-20.051,53.909],
[28.652,-16.395,54.750],[26.532,-10.644,55.445],[32.896,-14.158,55.906],
[30.312,-8.182,57.051],[17.705,-13.255,53.056],[20.956,-11.258,57.428],
[20.953,-6.176,58.687],[29.148,-10.448,54.152],[34.081,-4.105,54.511],
[29.071,-10.914,56.952],[23.770,-9.586,52.411],[20.655,-9.756,55.205],
[17.292,-10.936,54.823],[28.217,-19.758,56.614],[25.474,-7.832,55.476],
[24.789,-17.145,58.611],[25.582,-19.992,57.627],[22.585,-7.940,54.424],
[20.986,-18.672,54.950],[20.275,-7.813,52.627],[17.339,-15.849,53.792],
[16.150,-8.071,58.496],[24.791,-7.154,51.604],[29.798,-7.365,54.493],
[19.346,-10.517,52.966],[24.276,-11.736,53.875],[22.998,-14.289,54.521],
[24.931,-15.866,53.048],[31.296,-10.933,52.604],[22.694,-4.330,55.103],
[15.301,-9.931,52.498],[13.780,-3.258,59.303],[32.100,-11.057,56.600],
[33.889,-12.310,57.822],[32.607,-13.808,53.269],[32.539,-16.987,53.220],
[31.720,-13.624,58.813],[30.550,-19.682,51.521],[30.002,-15.751,57.753],
[31.899,-20.261,58.790],[32.648,-20.216,52.931],[15.452,-20.024,57.860],
[30.300,-3.993,59.047],[16.210,-21.423,55.684],[23.176,-22.311,59.377],
[29.981,-22.047,55.868],[16.195,-6.508,56.232],[18.414,-20.345,53.507],
[22.865,-18.792,57.266],[24.414,-5.310,53.524],[32.187,-4.264,56.655],
[17.779,-3.195,52.921],[15.543,2.785,58.361],[28.656,-1.679,-59.883],
[27.987,9.895,57.610],[14.572,8.021,55.087],[30.666,9.507,-59.270],
[26.380,6.726,57.176],[26.167,4.613,53.069],[29.549,5.288,57.368],
[27.598,5.116,55.331],[26.073,11.620,57.773],[32.192,-1.927,55.396],
[18.079,4.182,57.448],[34.382,9.609,52.380],[25.386,14.478,57.818],
[18.844,11.859,57.989],[29.765,14.306,54.074],[27.913,13.123,55.975],
[30.607,5.330,54.906],[26.304,-0.455,54.208],[20.980,3.289,57.917],
[27.244,3.335,57.709],[21.668,-0.490,52.067],[15.437,0.691,56.808],
[15.914,-0.972,59.185],[22.501,1.226,56.168],[21.449,1.988,53.283],
[19.095,0.203,58.317],[23.308,8.255,56.005],[28.943,0.580,53.510],
[26.843,2.151,55.010],[26.885,8.713,55.394],[34.514,2.870,55.292],
[30.771,9.711,56.689],[18.324,4.834,52.035],[24.005,6.706,58.108],
[21.625,13.357,57.997],[29.336,8.382,53.670],[34.490,14.209,54.160],
[29.217,7.592,56.090],[22.443,9.376,51.764],[20.139,7.993,56.277],
[17.402,7.853,55.203],[28.080,-0.834,56.389],[25.880,11.046,54.763],
[25.232,1.174,56.968],[25.759,-2.267,58.417],[22.240,10.526,54.923],
[22.307,-1.472,55.081],[19.228,8.921,52.672],[16.593,0.380,53.354],
[17.243,9.759,57.998],[25.890,12.559,51.776],[29.341,11.263,54.605],
[21.352,6.860,53.539],[24.811,8.659,50.900],[21.964,5.503,55.749],
[23.381,4.262,53.527],[32.040,7.072,53.089],[22.472,15.010,55.706],
[16.727,7.679,52.382],[15.920,13.431,-59.027],[32.019,7.229,56.601],
[32.560,5.048,58.187],[33.539,5.396,55.397],[32.705,3.191,53.612],
[30.705,3.212,58.948],[31.633,-0.883,52.803],[30.190,0.309,57.862],
[32.299,-1.433,58.731],[34.232,-0.074,54.055],[17.751,-2.703,58.063],
[28.606,12.938,58.861],[17.445,-4.123,55.633],[23.513,-2.392,-59.051],
[29.477,-3.756,56.005],[17.975,11.233,55.504],[18.812,-0.837,54.454],
[25.244,-3.014,54.881],[24.840,13.616,54.517],[31.440,12.162,57.243],
[18.036,13.979,52.215],[16.408,19.672,56.699],[28.987,18.949,-59.884],
[27.965,28.350,58.109],[14.525,29.328,56.344],[30.671,30.277,57.418],
[25.893,24.708,57.206],[26.541,22.730,53.661],[30.677,22.748,55.149],
[28.132,23.322,56.256],[26.918,30.369,59.469],[31.983,14.759,56.231],
[20.910,23.122,57.144],[33.813,29.625,52.551],[25.970,32.628,57.438],
[20.164,33.871,56.980],[27.787,30.160,53.526],[27.628,30.652,56.272],
[28.703,24.387,53.757],[25.459,18.055,54.404],[21.610,21.114,59.201],
[27.066,21.620,57.947],[22.057,20.701,52.272],[17.786,19.714,51.373],
[18.080,18.180,-58.925],[20.772,17.904,56.449],[20.742,20.391,54.674],
[20.640,18.584,59.620],[23.508,24.558,56.296],[29.585,19.005,54.364],
[27.666,20.858,55.400],[24.954,27.040,56.028],[34.934,22.686,53.782],
[31.931,28.262,55.183],[16.713,26.420,54.422],[22.933,26.062,58.831],
[21.723,31.777,58.528],[28.984,27.149,53.029],[32.517,32.118,55.000],
[28.683,26.405,56.283],[23.245,28.977,52.384],[22.214,27.477,54.789],
[19.266,25.976,54.555],[27.430,18.397,56.656],[23.814,29.634,55.384],
[22.884,20.235,56.767],[25.903,17.570,58.928],[21.191,30.185,54.995],
[22.237,17.001,53.839],[19.658,28.250,53.039],[17.698,20.946,53.837],
[17.882,29.783,58.288],[25.240,30.857,52.243],[30.789,29.622,53.205],
[21.789,26.506,52.012],[25.616,27.495,52.580],[21.396,24.457,54.341],
[23.829,24.229,52.850],[31.219,24.674,53.453],[22.819,33.096,54.420],
[16.022,27.941,52.063],[16.155,33.883,58.415],[31.163,27.163,57.564],
[35.479,24.895,56.885],[32.866,24.911,56.210],[31.337,20.581,52.204],
[31.777,21.756,-59.081],[32.660,18.237,53.349],[30.192,20.585,56.897],
[32.506,17.538,57.008],[34.124,16.683,55.079],[16.203,16.619,57.370],
[31.243,33.349,58.358],[17.304,15.375,54.903],[21.799,16.133,58.543],
[28.881,15.406,56.702],[16.556,30.554,54.481],[18.714,17.317,52.769],
[23.521,17.458,56.683],[26.340,32.231,54.341],[33.013,31.345,57.467],
[17.548,32.663,52.370],[16.487,41.113,58.661],[28.995,38.265,57.884],
[26.091,46.399,58.256],[15.199,46.583,55.492],[31.451,47.365,58.721],
[26.062,43.732,55.750],[27.795,43.034,53.919],[28.879,40.825,56.449],
[29.470,41.031,53.652],[26.071,49.058,58.852],[31.934,34.601,55.301],
[20.574,42.440,56.649],[31.836,47.974,53.222],[26.374,51.011,56.631],
[19.076,51.791,58.107],[30.406,50.799,55.173],[28.137,48.583,56.908],
[30.187,43.666,52.753],[25.397,38.154,53.846],[19.629,41.179,59.024],
[26.709,42.219,58.583],[19.967,38.144,53.485],[17.452,38.496,54.699],
[16.614,37.914,57.570],[22.700,37.596,54.004],[21.814,40.313,53.813],
[19.967,36.798,57.560],[22.420,44.597,55.804],[28.036,36.818,53.546],
[27.195,38.890,55.988],[26.810,46.196,55.660],[33.592,41.416,55.986],
[30.938,48.276,56.315],[18.147,42.622,53.897],[22.501,45.756,58.445],
[21.150,49.922,58.079],[29.732,46.296,54.603],[33.344,52.554,53.396],
[29.979,45.203,57.238],[22.946,47.342,51.918],[22.320,47.193,54.975],
[18.815,45.285,53.455],[28.115,36.212,56.568],[25.441,48.632,55.324],
[24.276,38.870,56.615],[25.700,36.675,58.055],[21.217,49.700,55.529],
[20.981,35.701,54.261],[20.031,46.413,49.336],[16.923,40.378,52.906],
[18.997,48.299,58.628],[24.748,49.241,51.254],[29.664,49.211,52.440],
[23.504,44.771,52.742],[26.252,45.890,52.177],[21.323,43.031,53.588],
[24.324,41.952,52.946],[32.434,41.921,53.415],[22.445,52.322,53.419],
[15.779,46.774,52.178],[16.250,50.877,58.241],[33.756,45.078,57.166],
[32.665,43.352,59.240],[31.760,44.543,55.380],[33.360,39.088,53.991],
[29.992,40.551,58.941],[29.693,38.772,51.785],[31.619,38.279,-59.685],
[30.885,35.831,57.645],[34.951,35.600,54.498],[14.654,36.041,56.693],
[30.142,52.445,58.710],[17.059,33.202,55.784],[23.606,34.012,58.561],
[28.745,33.487,56.720],[16.602,49.390,55.963],[17.576,35.426,53.971],
[25.516,35.619,55.274],[26.355,52.830,52.525],[32.431,51.166,57.258],
[19.444,52.087,55.112],[29.745,55.410,57.827],[29.672,53.463,54.866],
[26.618,55.573,52.139],[20.779,-59.433,58.079],[20.236,54.557,51.700],
[18.035,57.271,55.109],[16.437,56.832,58.979],[22.173,57.533,56.781],
[20.275,58.653,54.897],[19.484,56.939,57.695],[29.164,57.461,53.542],
[27.772,59.291,55.941],[33.382,57.435,55.139],[27.514,55.886,56.042],
[23.137,55.258,57.629],[25.330,53.621,56.464],[21.778,54.609,55.086],
[16.549,58.658,52.433],[33.195,-58.811,53.058],[31.479,55.941,54.169],
[29.842,57.826,57.246],[32.272,55.683,56.995],[33.751,55.331,52.325],
[16.602,55.145,56.553],[16.804,52.340,55.418],[21.800,52.268,56.784],
[28.099,53.167,57.158],[17.645,52.740,52.944],[24.107,54.288,53.665],
[19.461,-57.785,-13.273],[19.221,-59.109,-19.907],[29.341,58.761,-15.216],
[18.353,-48.073,-29.449],[27.039,-44.994,-32.035],[20.174,-57.701,-27.116],
[21.152,-50.063,-25.376],[30.269,-57.284,-24.629],[28.024,-47.599,-27.858],
[24.817,-48.265,-31.177],[23.797,-46.978,-33.243],[17.117,-49.598,-35.302],
[22.420,-54.199,-34.222],[30.635,-53.577,-20.427],[17.893,-52.277,-36.102],
[21.416,-56.046,-36.177],[27.556,-46.705,-23.095],[20.838,-50.048,-21.696],
[22.914,-51.096,-18.521],[23.735,-53.783,-19.155],[18.046,-52.105,-17.354],
[22.672,-57.048,-15.802],[15.239,-45.928,-34.659],[14.222,-44.571,-31.963],
[31.664,-45.890,-30.263],[30.689,-54.925,-29.026],[32.188,-55.324,-16.731],
[19.649,-43.089,-30.096],[24.414,-33.498,-37.981],[32.405,-36.143,-32.617],
[19.187,-29.343,-34.363],[32.799,-28.839,-34.693],[19.515,-40.444,-39.119],
[20.765,-42.726,-42.890],[25.606,-31.107,-33.557],[15.095,-34.312,-30.279],
[32.031,-30.863,-33.029],[27.276,-26.030,-30.886],[21.389,-38.042,-25.165],
[30.581,-38.426,-38.684],[29.539,-44.419,-43.233],[16.724,-27.533,-32.914],
[19.335,-31.150,-41.140],[24.130,-43.354,-31.828],[30.509,-44.427,-33.965],
[28.706,-14.937,-47.399],[21.036,-21.487,-45.796],[22.943,-15.132,-55.666],
[29.013,-19.375,-36.698],[28.777,-21.990,-38.744],[29.359,-7.700,-44.531],
[26.613,-27.874,-49.466],[25.154,-12.483,-44.668],[32.231,-27.432,-50.872],
[17.570,-20.243,-56.034],[19.456,-22.163,-55.326],[25.090,-20.537,-39.156],
[27.272,-21.477,-34.187],[30.682,-24.744,-51.820],[27.791,-10.203,-48.545],
[24.589,-17.620,-56.289],[27.533,-2.906,-57.737],[31.188,-8.523,-50.487],
[20.007,-10.545,-54.680],[23.225,-11.019,-51.341],[20.684,-3.179,-54.528],
[19.099,3.626,-51.267],[30.518,-15.823,-56.529],[24.105,3.464,-52.596],
[23.026,-7.457,-47.260],[25.792,-3.521,-52.157],[25.841,-10.059,-50.833],
[23.368,-5.657,-55.447],[32.505,-8.980,58.554],[32.979,1.102,-51.677],
[23.967,-8.104,-49.703],[18.957,-13.207,-58.938],[15.115,-10.397,-50.689],
[23.248,1.324,-50.228],[22.353,-5.177,-43.002],[14.314,-7.173,-52.844],
[19.432,1.183,-54.878],[23.475,5.671,-53.734],[18.563,5.283,-57.199],
[31.107,-10.537,-52.479],[16.906,-11.528,-52.876],[18.694,7.625,59.286],
[26.495,10.661,-58.662],[28.380,14.819,-57.865],[20.317,14.384,-55.828],
[30.055,16.452,58.882],[31.499,8.680,-54.726],[19.329,8.228,-52.921],
[25.125,-56.651,0.126],[18.994,-55.582,-0.318],[28.911,-58.257,-5.990],
[20.523,-58.295,0.188],[23.187,-50.563,0.822],[29.653,-58.358,-0.769],
[25.215,-58.229,-9.422],[17.234,-58.167,7.029],[20.602,-50.361,-4.397],
[14.541,-47.060,-13.552],[18.009,-39.071,-17.354],[25.462,-52.499,-14.579],
[25.113,-46.947,-13.011],[25.061,-41.093,-4.004],[17.431,-47.513,-11.663],
[20.668,-40.348,-15.401],[19.278,-51.914,-14.148],[16.224,-49.708,-16.458],
[30.073,-38.153,-7.974],[23.205,-44.607,-17.850],[20.727,-40.332,-6.148],
[25.271,-38.236,-3.946],[31.791,-47.473,-20.402],[27.500,-45.940,0.523],
[32.261,-36.307,-12.370],[32.196,-45.635,-13.799],[17.011,-34.041,-8.131],
[19.914,-33.035,-13.186],[31.668,-28.665,-15.743],[17.495,-30.182,-15.884],
[20.476,-24.142,-19.240],[14.701,-25.770,-27.794],[20.623,-27.428,-23.597],
[25.974,-19.605,-21.998],[20.837,-21.036,-15.992],[19.835,-28.887,-16.794],
[25.557,-24.676,-13.537],[21.148,-31.314,-22.557],[31.258,-24.693,-17.899],
[18.321,-24.835,-20.844],[24.546,-26.649,-27.832],[28.691,-26.265,-19.247],
[16.560,-27.190,-14.394],[24.613,-27.070,-14.662],[25.630,-35.711,-20.323],
[24.686,-25.020,-31.281],[32.651,-30.529,-27.564],[31.856,-23.978,-12.887],
[31.122,-31.268,-14.433],[17.361,-37.080,-21.422],[17.608,-30.792,-9.564],
[12.686,-13.672,-31.836],[23.384,-13.973,-34.985],[17.603,-13.370,-33.740],
[27.820,-14.618,-25.567],[30.535,1.723,-27.416],[32.559,-17.022,-33.866],
[27.601,-11.619,-39.130],[19.756,-12.597,-40.025],[23.714,-10.461,-20.859],
[15.614,-20.944,-30.955],[29.099,6.797,-42.012],[15.173,4.322,-41.186],
[24.320,7.123,-40.386],[21.596,6.429,-36.531],[18.626,3.038,-37.191],
[29.457,4.397,-43.466],[20.737,5.324,-41.491],[17.143,8.003,-35.203],
[20.742,14.010,-37.031],[32.326,4.758,-49.400],[27.875,9.038,-50.213],
[20.898,-3.747,-44.996],[23.708,4.477,-37.486],[25.709,11.091,-37.016],
[26.743,-1.435,-40.577],[23.476,6.202,-51.137],[15.955,8.870,-32.208],
[24.421,9.716,-32.747],[31.488,5.978,-31.412],[30.771,-1.565,-39.190],
[17.991,14.621,-41.327],[23.325,23.746,-48.647],[27.779,15.729,-49.520],
[16.187,19.537,-49.243],[31.457,18.208,-42.713],[29.852,15.737,-43.850],
[26.470,28.762,-48.972],[28.177,24.322,58.709],[30.347,23.877,-42.588],
[20.209,15.441,-44.355],[17.321,21.324,-43.572],[19.691,16.388,-54.104],
[23.120,21.553,-56.045],[33.026,22.227,-56.357],[20.251,26.362,-54.931],
[17.162,22.453,-56.127],[21.003,32.065,-44.109],[22.953,14.610,-55.863],
[25.624,19.751,-40.084],[15.343,26.929,-53.932],[33.097,29.497,-51.844],
[31.936,21.196,-51.377],[31.600,11.012,-52.605],[23.476,30.812,-50.563],
[17.704,26.591,-52.474],[24.818,23.467,-58.619],[19.071,17.658,-37.441],
[20.695,32.397,-55.950],[19.715,45.792,-53.139],[30.524,24.486,57.740],
[16.698,42.871,-55.388],[27.086,29.069,-55.439],[16.039,34.561,-48.608],
[26.513,-49.687,10.587],[31.833,-49.125,5.110],[23.469,-43.428,3.757],
[16.341,-44.177,14.990],[34.090,-45.429,3.311],[26.335,-58.234,11.196],
[20.644,-56.256,11.318],[19.827,-52.419,4.095],[25.019,-52.449,3.628],
[25.485,-48.071,23.735],[30.506,-50.262,18.579],[30.021,-57.752,13.749],
[32.898,-51.560,23.354],[25.897,-38.659,14.434],[27.238,-38.702,11.495],
[16.290,-42.298,9.588],[26.591,-58.888,17.092],[23.761,-46.211,3.727],
[17.407,-29.979,2.119],[18.025,-35.434,-0.061],[14.717,-28.638,2.415],
[22.901,-38.371,11.966],[18.377,-39.804,-0.509],[23.239,-34.496,-0.591],
[20.951,-30.994,12.075],[22.766,-34.027,2.077],[14.203,-26.807,4.249],
[26.211,-41.596,3.465],[20.384,-38.802,11.384],[30.647,-25.066,4.619],
[30.527,-30.160,3.887],[29.596,-43.555,0.396],[29.828,-38.417,2.147],
[21.279,-20.285,-6.417],[19.037,-21.061,1.698],[19.976,-12.409,-3.166],
[29.188,-18.793,-7.861],[18.255,-15.463,0.171],[31.354,-16.514,-13.940],
[23.850,-10.726,-2.666],[24.478,-13.078,-14.247],[20.086,-24.317,-0.537],
[25.266,-20.340,-5.772],[20.729,-22.215,-8.051],[24.589,-13.800,-1.009],
[23.790,-17.358,-12.108],[16.601,-22.556,-9.592],[20.246,-7.704,0.741],
[21.327,-12.013,1.384],[24.947,-17.130,3.617],[33.102,-21.025,-9.565],
[30.762,-19.130,-5.759],[26.810,-9.473,-6.130],[27.368,-26.293,-0.548],
[24.137,-15.963,-14.586],[31.661,-12.455,-11.309],[26.198,-21.444,7.875],
[24.482,-9.776,-9.853],[23.565,3.043,-19.335],[30.214,-11.562,-15.089],
[17.575,-3.283,-16.996],[21.140,-11.447,-19.898],[27.386,-7.875,-19.095],
[17.273,-0.118,-15.443],[26.299,-8.802,-13.661],[21.001,-10.574,-17.014],
[23.727,0.413,-25.824],[25.248,-1.849,-24.488],[20.566,-5.101,-7.506],
[25.134,-1.554,-5.350],[30.685,6.987,-13.696],[31.846,1.854,-14.641],
[13.320,-8.775,-14.350],[26.412,10.070,-14.605],[16.932,9.066,-15.128],
[16.350,6.660,-19.176],[23.918,-7.156,-10.696],[16.754,-10.364,-11.355],
[24.908,14.351,-22.991],[29.021,17.530,-24.018],[17.158,11.979,-20.353],
[19.593,19.425,-21.272],[20.299,13.429,-25.066],[23.643,10.758,-14.545],
[25.430,6.295,-26.337],[21.527,19.464,-16.403],[20.341,14.749,-18.530],
[31.319,11.270,-32.714],[30.996,10.285,-11.850],[30.958,13.011,-13.549],
[14.664,10.572,-15.973],[33.333,11.433,-12.365],[26.000,6.718,-19.902],
[14.723,29.183,-34.799],[19.623,29.621,-37.084],[16.299,34.411,-29.789],
[24.554,23.816,-37.732],[19.850,26.583,-35.969],[27.678,28.106,-32.840],
[12.774,26.725,-30.989],[22.393,26.249,-31.334],[17.074,27.162,-34.757],
[17.947,23.210,-33.368],[21.777,27.052,-28.303],[22.383,36.720,-28.836],
[32.652,33.140,-41.227],[17.170,35.002,-41.769],[24.256,32.831,-41.123],
[30.578,30.108,-24.593],[19.478,27.673,-42.279],[20.511,33.862,-25.248],
[33.351,38.674,-31.252],[33.007,26.599,-37.919],[30.337,23.122,-36.355],
[16.213,39.256,-27.912],[30.356,22.933,-29.513],[17.582,18.844,-35.615],
[19.125,43.803,-40.115],[27.286,48.375,-38.850],[20.483,52.921,-40.270],
[32.384,48.434,-51.288],[15.543,52.456,-44.448],[19.303,40.108,-36.692],
[23.960,39.983,-33.458],[26.681,42.969,-44.534],[25.122,54.073,-39.225],
[33.215,49.680,-54.053],[16.184,46.586,-49.014],[30.461,44.335,-34.200],
[28.762,45.910,-42.426],[27.375,54.904,-43.799],[24.988,44.996,-51.261],
[20.502,49.530,-36.382],[32.046,43.060,-49.875],[27.136,46.636,-31.154],
[16.264,45.781,-32.777],[15.476,45.080,-51.199],[13.104,53.090,-49.410],
[32.345,49.957,-34.114],[16.212,48.434,-32.692],[32.619,44.264,-31.833],
[22.001,54.372,-44.645],[19.960,-41.259,29.295],[14.085,-38.558,24.478],
[23.770,-45.873,22.810],[29.870,-40.749,35.950],[25.492,-44.837,27.861],
[25.742,-38.500,33.453],[27.068,-42.355,22.831],[22.742,-42.689,19.144],
[23.446,-47.619,28.396],[29.555,-43.472,36.623],[16.164,-46.683,24.957],
[30.722,-44.058,22.208],[20.637,-40.537,37.707],[13.599,-35.453,23.806],
[30.112,-33.043,36.147],[32.490,-45.532,23.608],[29.674,-47.222,28.478],
[29.692,-42.067,33.724],[32.023,-42.541,41.979],[18.480,-21.448,18.778],
[26.883,-25.121,24.824],[29.692,-28.273,27.037],[19.827,-23.984,18.291],
[30.720,-16.341,23.597],[22.923,-19.618,13.104],[27.678,-28.604,14.264],
[20.011,-29.324,19.270],[27.591,-28.657,25.198],[30.591,-13.471,23.678],
[24.680,-27.945,23.057],[26.686,-31.929,18.017],[22.969,-32.494,19.940],
[17.297,-23.801,10.540],[23.228,-24.469,28.887],[17.768,-24.268,30.362],
[15.930,-28.407,24.603],[14.964,-26.235,23.456],[17.999,-14.111,18.296],
[23.107,-16.570,19.187],[15.835,-22.001,12.502],[30.207,-10.527,16.713],
[28.280,-5.528,5.087],[24.834,-12.124,8.968],[20.140,-3.047,3.301],
[15.635,-9.363,4.310],[27.665,-7.748,1.347],[29.019,-12.265,4.207],
[19.040,-7.289,19.396],[22.802,-8.758,19.222],[16.215,-10.336,1.849],
[14.776,-1.782,5.582],[31.165,-1.022,16.575],[31.868,-12.600,5.182],
[15.456,-7.953,17.642],[32.327,-11.342,23.483],[31.735,-9.916,21.397],
[16.143,0.646,14.889],[18.987,-0.716,10.992],[16.983,-11.007,20.851],
[16.738,9.057,5.238],[24.266,11.782,2.030],[30.444,3.635,7.897],
[31.828,8.581,4.094],[14.754,2.415,2.228],[23.107,13.067,5.598],
[29.382,10.262,-6.554],[24.905,9.450,-9.516],[20.871,5.235,-4.755],
[14.960,8.443,2.820],[11.760,7.108,5.510],[24.447,26.728,-9.212],
[24.298,32.176,-11.962],[15.451,29.528,-8.509],[18.549,29.491,-12.594],
[30.823,36.145,-6.290],[29.187,26.584,-4.442],[28.263,16.124,-11.366],
[24.909,22.267,-14.876],[22.706,31.262,-7.627],[17.644,31.556,-7.374],
[22.117,23.389,-15.162],[17.771,26.090,-16.076],[16.685,27.651,-13.461],
[24.148,24.173,2.289],[13.362,29.384,-13.215],[29.783,31.995,0.675],
[33.531,20.878,4.137],[31.240,16.827,-0.986],[14.624,29.848,-15.831],
[24.286,39.337,-19.365],[23.155,42.721,-16.031],[27.495,34.155,-15.250],
[33.315,41.037,-14.181],[32.656,48.309,-13.262],[23.165,48.069,-15.033],
[16.100,46.406,-14.016],[20.131,35.249,-17.810],[22.846,38.094,-12.483],
[27.034,34.244,-20.579],[18.080,42.329,-16.575],[25.376,46.996,-13.859],
[22.314,33.936,-18.950],[20.290,32.086,-18.809],[17.847,39.246,-25.476],
[26.564,42.218,-5.687],[19.956,39.750,-6.540],[22.240,35.799,-5.789],
[14.976,34.857,-12.736],[30.146,45.813,-22.589],[25.042,50.195,-15.180],
[18.781,47.942,-17.670],[17.526,28.583,-21.245],[30.573,49.673,-14.119],
[27.707,-59.485,-26.209],[29.635,59.066,-33.396],[28.886,53.865,-25.446],
[20.430,53.638,-18.413],[26.961,57.497,-17.283],[19.261,57.057,-16.590],
[16.973,59.234,-35.686],[16.711,-33.236,45.860],[24.716,-26.407,41.488],
[19.353,-22.554,48.678],[27.949,-26.213,37.137],[23.344,-29.699,52.694],
[16.523,-36.530,45.443],[32.767,-36.523,43.088],[15.503,-28.804,41.017],
[23.975,-31.028,34.109],[22.802,-33.975,58.551],[33.421,-33.726,41.590],
[33.576,-35.307,39.469],[30.323,-35.929,54.746],[17.208,-21.009,46.238],
[29.195,-18.376,44.948],[25.187,-38.974,50.125],[29.865,-38.496,48.285],
[27.210,-37.759,48.309],[21.117,-10.827,39.418],[27.962,-10.946,31.810],
[13.583,-10.839,34.251],[32.960,-7.822,41.499],[23.111,-8.800,39.805],
[23.786,-19.735,45.451],[17.202,-9.890,40.878],[19.950,-22.247,36.616],
[32.335,-10.047,39.769],[19.924,-14.848,42.682],[32.743,-12.755,43.519],
[30.202,-20.824,34.919],[15.007,-17.694,37.428],[25.819,-23.046,39.332],
[13.473,-2.543,25.787],[19.773,1.856,28.326],[20.787,6.660,30.994],
[25.231,3.952,23.201],[19.550,7.783,20.776],[23.660,-7.892,22.898],
[18.525,-0.068,29.998],[32.497,-5.799,24.064],[24.369,3.379,15.932],
[19.773,-3.551,37.557],[22.131,0.188,38.909],[30.501,-1.266,21.646],
[31.812,3.612,31.613],[32.448,-8.329,23.562],[26.465,11.741,29.591],
[15.913,9.286,25.514],[18.681,-9.680,22.697],[31.376,4.186,17.448],
[27.205,-2.208,40.613],[30.512,9.858,31.762],[25.388,15.139,15.790],
[32.989,11.641,22.328],[19.173,12.491,19.227],[27.894,10.526,17.607],
[23.516,23.590,20.607],[21.957,8.338,13.745],[25.336,22.699,22.641],
[22.708,18.068,24.221],[27.603,16.645,8.333],[28.739,17.190,11.678],
[23.442,19.126,6.982],[20.816,14.435,11.443],[29.449,6.628,14.469],
[13.585,7.785,18.090],[23.144,26.423,16.350],[25.619,24.882,16.135],
[19.467,31.591,6.680],[29.081,34.182,16.601],[24.479,28.284,7.489],
[20.705,27.474,3.677],[18.403,25.897,10.074],[22.635,30.421,12.587],
[27.380,27.659,6.804],[32.883,26.467,5.015],[26.852,35.098,10.296],
[15.852,30.142,15.688],[22.308,33.923,1.927],[23.381,31.180,16.642],
[15.821,25.915,3.411],[16.292,40.835,15.639],[20.511,32.922,-0.193],
[21.280,33.376,17.831],[33.292,28.644,1.995],[30.233,25.817,3.058],
[33.707,28.344,21.386],[32.561,22.054,6.822],[20.604,38.182,6.735],
[33.580,45.316,4.436],[33.418,48.300,4.752],[20.832,51.600,0.756],
[19.043,44.664,2.681],[27.029,53.180,-1.172],[27.419,45.334,-6.710],
[32.627,57.139,2.221],[27.319,54.428,-10.436],[24.249,48.645,2.497],
[21.052,53.559,-6.600],[25.246,49.232,5.135],[28.984,48.168,-0.079],
[26.129,47.503,-9.146],[16.187,53.268,5.234],[22.868,44.313,-8.233],
[21.637,47.848,8.626],[30.410,41.502,-0.331],[16.301,44.591,2.011],
[14.912,46.972,2.777],[23.956,57.592,-2.296],[23.672,42.224,1.472],
[24.162,-16.432,55.596],[28.419,-26.320,55.594],[23.963,-29.780,57.305],
[15.130,-30.309,58.202],[19.983,-27.881,58.149],[33.384,-27.321,57.260],
[30.306,-18.413,54.657],[18.038,-32.045,57.482],[17.254,-7.068,52.692],
[32.711,-6.255,58.557],[18.819,-9.692,58.050],[20.227,-3.096,54.373],
[22.251,-13.215,58.608],[19.893,-5.773,55.515],[20.015,-3.701,57.182],
[19.028,3.155,55.000],[23.729,0.228,53.536],[24.760,-1.541,44.767],
[18.306,-8.161,55.474],[20.832,-12.886,49.084],[22.921,-8.926,57.081],
[26.207,-13.137,47.752],[20.289,-4.342,48.284],[16.742,-13.459,55.653],
[23.604,-6.434,41.560],[21.229,-1.521,58.291],[26.804,-8.223,43.089],
[23.092,-5.352,57.557],[24.853,4.152,56.453],[15.320,-3.325,43.393],
[16.793,-15.936,48.924],[30.001,2.373,56.076],[24.477,12.118,44.930],
[16.910,6.009,45.971],[21.346,15.563,43.405],[18.027,8.778,39.673],
[23.779,17.495,44.733],[22.272,18.870,42.178],[20.220,14.692,36.798],
[26.157,15.284,51.697],[20.262,5.742,50.443],[30.834,7.034,41.009],
[30.945,-0.390,40.002],[16.467,0.429,46.391],[18.352,22.172,36.626],
[19.121,30.179,37.171],[29.013,30.399,36.709],[19.728,34.529,36.863],
[15.234,31.954,31.703],[23.032,16.834,30.130],[24.282,28.095,41.292],
[23.506,23.825,38.872],[14.731,30.166,38.820],[17.393,24.006,28.108],
[28.304,34.953,35.746],[33.125,29.654,23.550],[22.698,25.254,24.119],
[25.345,23.981,41.242],[29.044,23.178,40.523],[23.927,28.617,22.468],
[22.263,28.098,38.269],[30.671,33.164,37.172],[33.258,28.513,42.961],
[30.274,23.840,35.932],[31.723,24.176,43.428],[14.054,23.304,41.738],
[17.864,32.266,30.361],[31.285,18.323,36.654],[24.721,31.472,23.892],
[26.050,48.412,24.350],[14.305,48.955,24.806],[19.494,48.144,18.199],
[20.460,35.517,29.360],[26.550,37.882,22.524],[20.024,34.135,20.811],
[22.862,41.152,20.269],[21.299,45.618,19.146],[18.016,41.913,18.062],
[21.308,53.175,28.861],[20.311,41.514,31.876],[15.212,48.505,21.815],
[31.127,50.382,22.802],[29.537,48.061,21.016],[19.582,50.929,22.978],
[24.792,50.488,22.877],[18.934,38.740,36.464],[32.771,57.827,21.297],
[27.617,55.081,21.542],[21.354,52.333,19.298],[26.644,55.042,12.949],
[21.993,57.001,6.499],[17.738,52.218,20.928],[27.933,5.253,-59.287],
[16.203,6.752,58.806],[17.954,22.636,57.382],[24.570,20.901,53.919],
[27.224,15.673,54.519],[18.264,18.285,55.425],[26.249,21.831,50.550],
[19.090,24.084,52.433],[20.256,12.873,55.853],[18.886,27.103,57.385],
[20.331,29.287,57.584],[21.152,25.819,56.819],[18.898,39.606,48.273],
[15.386,33.450,49.271],[25.344,29.673,43.277],[21.213,35.404,47.719],
[19.545,33.482,53.880],[17.493,40.815,56.039],[28.521,44.386,50.832],
[23.689,41.899,56.655],[14.128,40.403,42.401],[31.401,33.933,52.103],
[28.731,28.436,48.941],[14.962,47.320,39.581],[25.626,58.055,37.797],
[17.255,47.133,42.495],[29.536,47.109,46.132],[28.105,51.975,45.489],
[20.857,51.044,43.076],[30.197,52.341,47.757],[14.461,46.005,41.960],
[20.209,45.978,38.560],[21.391,51.691,38.057],[24.299,44.586,42.051],
[21.319,53.886,31.604],[17.365,44.641,44.587],[20.414,51.573,49.277],
[27.256,52.998,49.374],[29.564,57.847,42.810],[30.507,44.158,34.843],
[30.475,44.184,39.740],[29.894,47.445,49.148],[18.599,57.848,36.964],
[30.874,43.289,42.829],[17.746,41.975,36.506],[24.883,53.525,32.057],
[18.742,48.547,52.431],[27.502,47.835,53.473],[46.530,-51.972,-52.718],
[44.237,-49.363,-48.390],[37.581,-50.866,-45.803],[49.147,-46.372,-50.893],
[38.090,-46.073,-49.948],[35.411,-51.787,-53.763],[35.044,-50.424,-51.570],
[45.198,-42.501,-50.148],[45.424,-46.981,-48.509],[51.517,-58.581,-49.287],
[47.037,-53.817,-48.395],[40.052,-55.522,-51.608],[46.367,-54.763,-53.369],
[35.407,-53.016,-50.442],[41.603,-52.834,-50.329],[53.831,-53.824,-44.001],
[39.095,-57.298,-46.564],[43.189,-45.535,-52.613],[42.455,-47.580,-47.297],
[50.384,-56.230,-50.034],[37.695,-52.646,-48.938],[39.001,-48.249,-44.771],
[35.750,-46.450,-53.885],[48.281,-47.103,-48.308],[49.274,-57.328,-45.751],
[41.869,-60.006,-50.550],[46.970,-44.688,-51.277],[48.006,-45.362,-53.803],
[49.942,-48.439,-44.483],[43.690,-48.634,-53.150],[44.023,-52.741,-58.986],
[53.627,-50.715,-45.791],[47.876,-50.030,-54.391],[48.048,-52.176,-44.794],
[39.028,-49.965,-52.434],[41.696,-47.297,-49.689],[41.506,-50.226,-48.730],
[44.903,-44.308,-46.761],[39.515,-45.401,-44.447],[45.247,-50.907,-57.272],
[41.431,-53.190,-44.565],[44.963,-43.699,-55.230],[38.380,-43.500,-43.477],
[50.323,-59.074,-56.732],[38.307,-50.785,-55.805],[50.910,-45.803,-59.457],
[36.029,-47.860,-45.793],[34.290,-45.530,-48.719],[52.955,-43.935,-44.439],
[43.200,-45.411,-45.094],[43.910,-41.452,-52.887],[38.374,-42.114,-53.295],
[47.846,-42.405,-57.554],[47.744,-42.171,-53.011],[50.257,-50.879,-42.300],
[36.857,-46.148,-43.999],[40.436,-57.041,-49.028],[45.665,-59.543,-44.854],
[48.390,-55.120,-43.769],[47.269,-49.196,-57.594],[42.782,-54.325,-57.122],
[39.834,-52.027,-47.330],[39.375,-52.817,-52.715],[44.885,-51.221,-54.552],
[43.579,-52.233,-41.111],[44.476,-50.608,-50.832],[36.454,-51.949,-43.504],
[41.449,-56.055,58.937],[34.486,-55.698,-52.767],[35.655,-55.280,-57.173],
[43.225,-51.428,-46.108],[43.202,-55.265,-52.847],[36.533,-54.148,-52.660],
[39.772,-55.276,-54.533],[46.672,-56.945,-47.867],[38.855,-53.394,-56.393],
[49.257,-41.666,-41.850],[36.800,-48.604,-43.044],[44.391,-52.169,-48.572],
[38.031,-57.377,-52.549],[42.261,-53.796,-47.271],[42.619,-50.272,-55.515],
[48.487,-49.729,-46.990],[46.850,-54.238,-59.118],[42.803,-47.828,-43.680],
[46.545,-53.241,-55.816],[37.213,-48.105,-51.530],[45.079,-48.225,-55.479],
[33.498,-48.071,-44.364],[44.249,-49.900,-42.221],[46.104,-45.398,-44.084],
[51.969,-52.963,-56.638],[52.134,-48.677,-42.466],[51.947,-45.222,-56.064],
[48.491,-49.190,-50.876],[40.652,-45.447,-54.140],[35.392,-50.241,-57.653],
[41.135,-48.298,-53.335],[41.506,-43.866,-46.755],[40.525,-42.724,-54.860],
[44.249,-48.826,-45.589],[47.258,-46.462,-57.746],[47.961,-43.708,-42.975],
[51.587,-40.753,-54.934],[39.908,-54.725,-46.185],[48.269,-47.517,-55.282],
[37.469,-44.718,-47.411],[42.507,-46.571,-58.345],[38.379,-46.448,-55.491],
[35.959,-49.105,-55.407],[44.756,-56.664,-46.055],[45.377,-56.021,-55.948],
[42.083,-44.366,-56.549],[42.411,-52.825,-53.990],[45.552,-55.188,-43.728],
[46.990,-57.844,-52.680],[39.402,-43.478,-56.989],[40.048,-56.371,-57.828],
[51.523,-54.542,-45.737],[48.557,-52.093,-50.361],[36.123,-44.007,-58.358],
[33.692,-52.318,-58.172],[49.461,-42.518,-47.663],[37.317,-58.800,-48.710],
[34.632,-44.075,-53.790],[35.464,-42.540,-42.069],[40.586,-43.030,-41.408],
[42.759,-44.096,59.468],[46.087,-45.454,-41.224],[49.227,-44.840,-57.598],
[35.310,-48.089,-40.754],[37.759,-48.478,-57.913],[41.006,-47.427,-41.852],
[43.251,-49.010,-57.765],[46.401,-48.297,-43.840],[33.490,-50.579,-39.969],
[40.245,-49.026,-56.708],[39.146,-52.293,-43.136],[41.881,-51.538,-57.605],
[45.990,-48.611,-40.697],[51.593,-48.625,-59.270],[40.258,-56.452,-41.782],
[42.229,-56.634,-44.886],[38.988,-48.980,-47.689],[47.663,-53.219,-41.740],
[47.157,-50.548,-42.476],[38.206,-52.078,-39.625],[42.645,-58.891,-46.260],
[43.111,-55.430,-41.833],[44.065,-57.642,-43.131],[46.756,-57.684,-40.811],
[37.964,-54.710,-41.696],[42.238,-59.650,-42.800],[40.319,-58.971,-40.456],
[41.762,-41.754,-57.446],[35.924,-46.419,-56.854],[35.298,-41.413,-53.997],
[34.107,-41.773,-51.676],[50.561,-43.506,-50.350],[50.148,-42.972,-44.436],
[50.603,-46.237,-40.647],[50.529,-48.902,-55.013],[51.020,-48.861,-51.997],
[51.252,-47.925,-48.249],[50.655,-45.103,-47.169],[51.596,-50.274,-57.154],
[48.903,-52.122,-57.301],[50.229,-51.659,-52.386],[35.132,-49.139,-49.310],
[50.959,-51.451,-45.359],[52.431,-52.923,-41.808],[47.810,-56.823,-58.415],
[50.750,-54.926,-54.990],[51.388,-53.630,-50.223],[50.948,-52.677,-47.856],
[35.505,-56.244,-40.488],[53.407,58.861,-41.180],[50.646,-57.800,-54.154],
[36.250,-55.286,-47.912],[34.669,-51.699,-48.353],[49.414,-58.049,-42.615],
[53.337,-59.341,-57.986],[34.010,-56.711,-55.308],[50.083,-39.447,-52.895],
[31.071,57.778,-44.232],[36.751,-45.122,-41.231],[40.036,-46.191,-47.208],
[43.067,-42.435,-42.595],[45.985,-42.151,-44.391],[52.022,-58.187,-44.599],
[36.532,-40.014,-46.347],[39.181,-41.905,-45.780],[43.279,-41.978,-48.201],
[46.435,-42.316,-47.696],[47.389,-40.330,-46.248],[35.092,-42.991,-46.833],
[38.024,-43.478,-50.868],[41.683,-43.709,-50.140],[43.783,-56.794,-49.280],
[49.085,-59.226,-51.558],[51.668,-59.001,-52.059],[36.037,-57.964,-50.893],
[35.088,-59.090,-56.437],[43.262,-58.613,-52.762],[46.783,-58.494,-50.123],
[48.396,-59.428,-54.668],[34.509,-42.413,-56.612],[38.377,-58.391,-57.198],
[41.241,-57.847,-55.259],[45.403,-41.317,-58.434],[50.437,-43.291,-54.093],
[35.908,-57.399,-43.423],[37.058,59.485,-41.729],[37.714,-41.258,-56.208],
[47.810,58.708,-40.800],[50.398,-45.436,-43.499],[45.773,-31.425,-52.175],
[44.121,-32.185,-49.704],[36.957,-31.919,-48.397],[49.046,-28.359,-50.430],
[39.312,-27.764,-50.110],[33.042,-33.914,-53.399],[34.815,-32.004,-52.723],
[45.636,-23.536,-50.670],[43.633,-28.155,-50.421],[50.619,-37.729,-48.129],
[46.801,-34.864,-47.400],[38.687,-38.183,-50.254],[47.034,-36.810,-52.389],
[35.152,-34.518,-51.297],[40.656,-32.011,-50.534],[51.884,-35.282,-45.212],
[38.157,-37.001,-47.850],[44.325,-28.058,-53.437],[40.339,-28.423,-47.204],
[47.880,-38.026,-48.694],[38.419,-34.033,-49.577],[37.189,-29.114,-44.825],
[33.474,-28.237,-56.228],[48.383,-27.345,-47.643],[47.668,-38.523,-44.099],
[41.339,-40.466,-49.699],[46.875,-25.838,-51.672],[47.697,-27.147,-54.279],
[49.975,-28.985,-46.294],[42.105,-31.906,-53.787],[44.747,-34.996,-58.823],
[52.318,-31.843,-47.648],[48.754,-30.587,-51.992],[47.109,-33.666,-45.039],
[39.581,-30.455,-52.283],[41.656,-25.982,-49.646],[40.498,-32.464,-47.573],
[44.508,-25.826,-48.007],[39.498,-27.558,-44.727],[45.724,-31.090,-56.347],
[37.256,-32.205,-45.598],[43.488,-25.340,-52.585],[39.171,-24.018,-43.611],
[50.653,-41.365,-57.378],[37.652,-31.960,-53.461],[49.966,-25.167,-57.546],
[35.282,-29.524,-46.960],[34.735,-27.193,-48.618],[52.883,-26.820,-42.965],
[43.818,-27.295,-45.275],[43.826,-20.956,-55.152],[36.600,-23.123,-53.435],
[47.885,-23.334,-56.728],[45.755,-24.278,-53.658],[49.639,-32.974,-44.170],
[36.846,-26.648,-43.656],[38.886,-40.246,-48.179],[42.657,-41.618,-45.207],
[49.058,-35.609,-45.076],[48.640,-33.011,-56.296],[43.807,-36.954,-56.698],
[40.554,-35.219,-48.616],[40.134,-34.539,-53.614],[44.811,-32.659,-54.041],
[44.051,-34.032,-40.536],[44.772,-34.810,-51.580],[37.410,-34.676,-46.814],
[41.469,-35.192,-58.554],[32.724,-37.376,-50.870],[36.289,-37.073,-57.085],
[45.066,-32.318,-46.753],[41.601,-37.338,-51.288],[37.523,-35.082,-52.767],
[42.310,-37.713,-53.991],[45.746,-36.900,-49.746],[38.898,-34.113,-57.744],
[48.326,-23.165,-41.610],[38.973,-31.340,-43.752],[42.269,-34.551,-50.508],
[39.336,-37.160,-53.275],[43.096,-35.750,-47.316],[41.482,-29.797,-56.383],
[49.742,-31.027,-47.729],[46.013,-37.941,-58.318],[44.789,-29.145,-43.684],
[45.953,-35.560,-56.382],[37.038,-30.405,-50.738],[44.088,-28.807,-56.134],
[34.882,-33.498,-44.049],[41.985,-33.305,-42.645],[46.466,-25.458,-43.881],
[53.545,-34.622,-55.431],[51.670,-27.575,-45.172],[50.152,-26.922,-54.767],
[47.135,-30.897,-49.332],[38.276,-25.479,-53.245],[36.634,-31.827,-58.037],
[41.726,-27.616,-52.303],[42.000,-25.674,-46.517],[41.069,-24.583,-53.469],
[46.128,-29.683,-46.991],[45.169,-28.300,-58.866],[49.135,-26.366,-43.358],
[50.663,-22.459,-56.910],[35.336,-36.476,-44.091],[47.098,-27.713,-56.905],
[35.777,-25.297,-46.625],[42.366,-27.167,-57.669],[39.477,-27.786,-57.011],
[37.457,-29.740,-56.382],[44.306,-39.211,-46.264],[45.315,-39.735,-55.857],
[43.137,-24.962,-55.441],[42.654,-34.305,-55.008],[42.705,-34.778,-44.630],
[44.129,-38.885,-52.424],[39.039,-25.272,-55.848],[39.019,-37.498,-55.952],
[49.896,-38.433,-45.553],[48.336,-33.125,-49.796],[37.135,-26.758,-58.199],
[36.052,-34.437,-59.145],[48.828,-24.000,-47.595],[35.643,-40.854,-49.644],
[36.754,-27.635,-54.035],[35.241,-24.542,-42.475],[39.443,-27.124,-40.382],
[40.520,-22.653,-58.240],[46.268,-27.062,-41.711],[47.604,-25.004,-59.508],
[35.059,-30.209,-41.502],[39.365,-30.783,-58.099],[41.873,-29.362,-44.631],
[42.252,-29.916,-58.974],[47.750,-28.758,-44.783],[36.600,-32.786,-42.040],
[39.458,-32.123,-55.484],[39.901,-34.315,-45.607],[41.063,-32.630,-59.016],
[43.917,-32.044,-43.557],[49.459,-28.913,-58.534],[40.976,-36.494,-43.602],
[41.673,-38.501,-46.643],[42.677,-31.881,-45.794],[48.565,-34.588,-42.146],
[46.367,-31.069,-42.902],[39.027,-33.344,-41.222],[39.430,-38.830,-44.524],
[42.517,-36.654,-41.239],[45.700,-35.188,-42.579],[46.061,-37.880,-42.151],
[37.963,-36.455,-42.741],[40.774,-40.473,-42.961],[42.694,-40.063,-41.118],
[41.586,-22.413,-55.565],[32.807,-28.331,-58.777],[34.115,-22.469,-53.508],
[33.871,-25.494,-51.804],[51.905,-24.282,-47.783],[49.702,-23.169,-44.455],
[50.276,-27.489,-41.125],[48.831,-29.563,-55.243],[51.601,-29.119,-55.051],
[51.166,-28.204,-48.523],[50.272,-25.625,-45.832],[50.980,-31.579,-55.735],
[51.788,-34.984,59.178],[48.518,-32.848,-53.602],[34.274,-30.785,-49.780],
[49.112,-30.133,-42.779],[51.651,-32.226,-42.651],[49.368,-36.066,59.398],
[48.735,-36.179,-54.783],[50.429,-35.604,-49.740],[49.433,-34.071,-47.633],
[32.946,-35.439,-42.747],[52.135,-39.292,-40.651],[51.253,-37.359,-54.835],
[35.345,-36.524,-49.161],[33.763,-34.446,-48.791],[50.387,-38.891,-42.666],
[50.771,-38.839,-58.951],[35.083,-37.380,-53.012],[48.428,-24.207,-53.701],
[34.336,-39.316,-45.073],[36.698,-26.481,-40.881],[39.106,-25.654,-46.628],
[42.791,-25.779,-43.098],[47.058,-22.199,-43.962],[51.469,-41.064,-46.115],
[36.752,-23.866,-44.625],[39.963,-23.043,-46.598],[44.233,-22.924,-48.055],
[46.272,-24.197,-46.184],[47.691,-21.789,-46.969],[36.450,-23.222,-48.790],
[37.484,-24.818,-50.853],[41.143,-23.300,-50.302],[43.544,-38.449,-49.524],
[48.440,-40.488,-49.337],[50.949,-39.230,-50.374],[37.271,-39.203,-54.162],
[34.303,-39.353,-56.130],[41.191,-41.071,-52.395],[46.310,-39.786,-51.025],
[48.432,-40.446,-55.022],[35.580,-25.287,-55.373],[37.011,-38.910,-59.208],
[41.991,-39.281,-56.270],[44.679,-23.408,-57.561],[52.440,-22.944,-54.907],
[36.634,-38.786,-43.686],[37.849,-39.666,-41.025],[37.345,-22.132,-57.949],
[46.643,-40.554,-42.295],[52.085,-24.959,-41.294],[45.102,-12.121,-51.227],
[44.233,-13.011,-48.477],[36.010,-12.891,-46.821],[48.107,-9.711,-50.528],
[38.474,-9.862,-49.941],[34.390,-14.467,-53.283],[36.036,-13.049,-51.746],
[44.617,-5.529,-50.189],[42.597,-10.636,-48.651],[51.520,-19.570,-49.055],
[45.007,-16.870,-48.070],[38.502,-17.069,-50.739],[47.670,-19.026,-53.220],
[35.411,-16.256,-50.647],[39.409,-14.125,-50.579],[51.944,-16.037,-45.370],
[38.697,-18.304,-47.182],[45.507,-10.053,-53.055],[40.431,-11.851,-45.995],
[48.081,-18.470,-50.639],[36.818,-15.513,-48.110],[37.419,-10.569,-47.326],
[36.060,-9.870,-55.315],[46.803,-7.802,-48.873],[48.577,-18.835,-46.610],
[42.076,-21.007,-48.738],[46.533,-6.316,-51.690],[48.031,-9.652,-54.232],
[49.457,-11.270,-45.113],[43.865,-12.402,-53.638],[46.700,-14.249,58.634],
[51.516,-13.432,-46.165],[47.694,-14.412,-54.046],[47.541,-14.668,-44.200],
[40.317,-13.718,-53.327],[40.739,-8.293,-51.150],[38.571,-13.523,-46.965],
[45.139,-10.275,-49.262],[40.462,-7.663,-44.363],[45.074,-13.959,-56.438],
[37.702,-13.268,-44.011],[45.079,-7.115,-53.961],[38.546,-6.217,-43.087],
[48.883,-20.142,-55.537],[39.316,-15.028,-55.537],[51.771,-7.555,-56.847],
[34.273,-10.044,-47.033],[35.144,-8.050,-48.709],[52.288,-6.520,-43.459],
[43.029,-8.482,-43.901],[43.009,-4.640,-53.557],[39.126,-5.044,-50.621],
[46.702,-6.277,-56.734],[46.262,-3.053,-55.039],[49.510,-15.960,-42.369],
[34.829,-9.655,-43.415],[39.216,-21.205,-48.437],[44.542,-22.975,-44.312],
[46.818,-17.460,-43.722],[47.797,-13.833,-57.170],[43.717,-20.449,-58.504],
[39.710,-16.049,-48.306],[40.947,-16.046,-51.716],[45.355,-15.703,-53.798],
[42.836,-16.615,-40.500],[47.322,-16.070,-50.964],[37.770,-16.056,-45.346],
[40.721,-18.493,-59.093],[33.687,-17.742,-52.441],[36.409,-18.107,-56.809],
[46.305,-14.780,-46.762],[40.824,-18.730,-52.106],[36.618,-16.611,-53.163],
[39.551,-17.557,-56.889],[45.323,-20.293,-46.570],[39.895,-15.258,-58.192],
[48.538,-4.551,-42.319],[38.188,-9.924,-44.222],[42.611,-16.225,-49.442],
[38.738,-21.015,-53.272],[42.798,-16.583,-46.339],[43.543,-11.614,-56.337],
[48.505,-11.914,-47.547],[45.211,-17.904,-57.787],[42.167,-11.278,-44.036],
[47.081,-16.710,-55.673],[35.522,-11.171,-49.365],[45.767,-10.460,-55.689],
[35.318,-14.633,-44.427],[41.542,-14.560,-41.913],[45.741,-7.992,-43.385],
[52.401,-16.036,-55.446],[50.467,-9.444,-43.359],[49.280,-7.295,-54.870],
[47.164,-12.758,-49.507],[40.122,-8.572,-54.436],[35.402,-12.273,-58.686],
[41.741,-10.464,-53.416],[41.704,-5.457,-45.795],[40.339,-5.527,-53.082],
[44.136,-10.577,-46.014],[49.742,-9.476,-56.906],[47.764,-4.475,-45.679],
[51.477,-4.779,-57.462],[35.957,-18.269,-46.971],[48.843,-12.067,-55.648],
[34.623,-7.688,-45.501],[42.725,-9.049,-58.968],[41.854,-9.373,-56.436],
[37.087,-11.972,-56.612],[42.876,-19.493,-44.933],[45.881,-19.139,-55.176],
[44.432,-8.050,-56.676],[41.279,-17.322,-54.309],[42.924,-14.966,-44.192],
[44.855,-19.660,-52.447],[39.313,-6.343,-56.154],[39.585,-20.243,-57.442],
[50.029,-17.986,-44.300],[49.351,-14.602,-49.394],[38.757,-9.518,-57.015],
[37.092,-14.628,-57.761],[48.683,-5.543,-48.340],[36.615,-20.398,-48.413],
[37.251,-7.245,-53.642],[34.628,-4.834,-41.792],[39.591,-7.545,-39.844],
[42.457,-6.618,-57.747],[47.016,-7.920,-40.956],[49.649,-6.962,-58.393],
[34.498,-12.198,-42.122],[39.987,-8.943,59.401],[40.355,-11.237,-42.299],
[45.027,-10.406,59.057],[47.209,-10.953,-43.409],[36.526,-13.987,-41.351],
[40.821,-11.995,-55.875],[40.296,-15.117,-45.448],[42.465,-13.721,-58.130],
[43.863,-12.385,-40.063],[50.986,-11.426,-58.603],[40.715,-17.734,-42.231],
[40.267,-17.985,-44.826],[42.498,-13.998,-46.798],[45.833,-16.567,-40.561],
[46.526,-14.178,-41.777],[39.025,-15.186,-40.988],[40.631,-20.038,-46.399],
[42.023,-19.462,-39.453],[44.192,-18.077,-42.419],[46.999,-19.444,-41.537],
[37.318,-17.571,-40.985],[41.274,-22.414,-43.513],[41.701,-21.495,-41.015],
[41.401,-4.148,-56.858],[35.498,-8.861,-57.929],[35.624,-3.249,-53.393],
[35.713,-5.078,-51.509],[51.517,-4.969,-49.052],[50.056,-6.036,-45.035],
[50.119,-10.112,-39.871],[51.149,-11.166,-54.694],[52.529,-13.420,-53.544],
[50.747,-9.816,-49.614],[49.627,-8.150,-47.264],[50.910,-13.344,-56.565],
[52.178,-16.814,59.462],[50.077,-14.129,-51.885],[32.980,-11.913,-48.848],
[49.205,-13.000,-42.556],[51.632,-13.413,-43.393],[50.186,-18.330,-58.170],
[49.344,-16.359,-53.320],[50.176,-17.368,-48.414],[49.388,-15.777,-46.484],
[35.872,-17.162,-43.250],[53.487,-21.911,-41.318],[51.218,-20.757,-53.867],
[33.924,-18.144,-48.877],[34.256,-14.274,-48.907],[51.119,-20.696,-42.734],
[51.634,-20.610,-58.680],[35.747,-19.207,-54.132],[48.985,-5.349,-51.272],
[35.371,-20.188,-45.122],[36.107,-7.415,-42.391],[38.935,-8.499,-46.371],
[43.913,-5.709,-44.131],[45.658,-3.790,-43.451],[49.757,-20.620,-45.230],
[36.459,-5.625,-45.430],[39.805,-4.590,-48.031],[44.036,-3.102,-48.258],
[45.729,-5.533,-46.898],[48.358,-2.009,-45.355],[37.476,-6.756,-48.079],
[37.183,-7.419,-50.731],[42.196,-5.570,-49.410],[44.826,-20.099,-49.484],
[48.540,-23.566,-50.792],[50.521,-20.587,-51.342],[35.340,-20.737,-51.294],
[34.202,-20.608,-55.901],[41.327,-21.432,-52.936],[46.932,-21.143,-51.057],
[46.351,-21.602,-53.762],[34.833,-7.125,-55.949],[36.155,-19.684,-59.496],
[42.763,-18.729,-56.111],[44.064,-5.003,-56.313],[51.462,-5.418,-54.737],
[34.354,-22.096,-43.359],[37.483,-20.578,-42.203],[37.036,-5.266,-56.803],
[45.844,-21.876,-41.577],[49.414,-7.241,-41.662],[46.141,6.598,-51.761],
[44.233,7.862,-50.136],[37.347,5.373,-48.548],[49.803,8.232,-50.340],
[37.879,9.351,-50.114],[35.397,3.639,-53.399],[35.834,6.124,-51.963],
[46.553,13.980,-50.745],[43.440,9.501,-48.192],[51.708,-0.934,-47.958],
[46.873,2.524,-48.561],[38.896,-0.276,-51.045],[48.011,0.331,-52.771],
[36.096,1.884,-51.252],[41.159,4.008,-52.419],[51.186,1.038,-45.935],
[37.717,0.341,-46.169],[45.023,10.351,-52.745],[40.463,8.062,-46.717],
[50.317,-0.227,-50.552],[37.496,2.630,-48.948],[37.358,8.497,-47.672],
[34.535,8.099,-53.403],[47.948,9.103,-47.640],[48.325,-0.359,-47.560],
[41.542,-2.212,-48.835],[47.967,11.656,-50.037],[48.816,11.515,-53.460],
[49.572,7.700,-44.466],[44.564,7.111,-54.297],[46.069,4.266,-57.869],
[51.323,3.834,-46.541],[47.951,5.778,-53.448],[46.247,3.492,-45.297],
[40.297,6.462,-53.379],[40.578,10.112,-48.680],[39.789,3.981,-49.786],
[44.815,11.288,-46.800],[40.852,9.384,-43.961],[48.793,5.010,-55.879],
[38.360,4.951,-44.858],[42.472,10.908,-51.451],[38.980,11.931,-42.787],
[50.786,-2.442,-56.179],[37.871,6.105,-54.372],[51.817,9.433,-58.390],
[34.712,7.183,-46.970],[33.531,11.209,-47.565],[53.133,11.887,-42.084],
[43.767,10.423,-43.346],[44.721,14.962,-54.494],[37.050,12.927,-52.422],
[47.262,12.478,-56.802],[46.419,12.716,-54.088],[47.847,2.767,-42.917],
[37.347,9.916,-44.942],[38.935,-2.037,-47.022],[45.387,-1.598,-45.755],
[47.431,0.677,-44.743],[48.136,6.308,-57.984],[44.628,-0.888,-57.661],
[39.223,2.887,-46.729],[40.805,1.449,-51.478],[44.747,4.303,-55.068],
[43.728,3.475,-41.027],[46.276,2.121,-51.439],[36.364,2.577,-45.544],
[39.692,0.284,-58.192],[35.876,-0.058,-53.519],[34.789,1.969,-55.392],
[44.447,5.337,-47.187],[41.426,-1.155,-51.806],[38.147,1.327,-53.365],
[40.693,1.455,-55.786],[46.112,-0.099,-49.164],[39.542,3.143,-57.504],
[49.317,14.975,-41.814],[37.278,7.508,-43.229],[43.051,3.717,-48.774],
[39.579,-0.931,-53.619],[41.661,1.904,-46.690],[41.481,7.930,-55.707],
[48.625,5.454,-46.539],[46.267,1.868,-56.400],[44.286,8.072,-44.457],
[47.290,3.004,-53.922],[38.455,6.433,-50.735],[44.074,8.895,-57.789],
[35.663,5.880,-44.491],[42.560,5.269,-42.961],[45.334,12.838,-44.623],
[51.955,2.421,-55.987],[51.716,9.118,-43.776],[52.289,10.585,-53.626],
[46.583,6.768,-48.885],[40.858,10.076,-53.585],[35.080,6.552,-57.522],
[42.896,8.438,-52.757],[41.326,11.836,-46.717],[39.618,13.626,-53.179],
[46.503,7.074,-46.229],[47.380,8.751,-56.774],[47.786,12.249,-43.789],
[53.762,14.700,-58.188],[35.671,-0.057,-44.185],[47.112,8.467,-53.899],
[35.979,11.761,-46.541],[42.205,10.232,59.022],[40.070,9.531,-57.284],
[38.098,8.771,-55.483],[43.484,0.297,-45.173],[45.985,-0.452,-54.590],
[42.783,10.280,-55.704],[43.360,1.518,-51.551],[43.216,3.564,-44.877],
[45.364,-0.396,-51.940],[39.788,12.614,-57.658],[38.544,-0.663,-56.008],
[49.550,-0.350,-43.652],[48.509,4.466,-49.520],[38.251,10.029,-59.537],
[36.490,3.663,-58.795],[48.136,14.063,-48.554],[37.089,-4.346,-49.094],
[35.131,9.722,-51.135],[34.710,13.995,-42.670],[41.214,13.230,-41.161],
[42.413,11.838,-57.771],[46.278,10.727,-41.847],[49.897,11.967,-57.695],
[34.874,7.134,-41.668],[39.613,6.083,-59.163],[39.871,6.483,-42.993],
[43.890,8.304,58.332],[46.883,8.175,-43.474],[36.304,4.676,-41.724],
[40.880,4.995,-55.712],[40.477,3.522,-44.318],[42.077,4.230,59.388],
[44.646,6.819,-41.905],[50.806,6.905,-58.880],[39.882,0.877,-42.496],
[40.423,-1.546,-44.282],[41.287,5.196,-46.510],[46.086,1.693,-40.912],
[47.851,5.463,-43.697],[38.509,3.313,-42.351],[42.398,-2.568,-45.944],
[42.242,1.906,-42.920],[44.033,0.212,-42.118],[46.897,-0.946,-42.135],
[37.301,0.710,-41.322],[40.766,-3.617,-42.520],[39.430,-1.107,-40.370],
[42.534,14.280,-57.170],[36.219,9.276,-57.378],[34.797,15.093,-53.182],
[34.761,12.807,-50.990],[51.899,14.073,-49.210],[50.857,11.921,-44.831],
[50.810,10.177,-40.672],[51.477,7.324,-55.908],[50.825,6.704,-53.192],
[51.187,8.981,-47.248],[47.940,11.900,-46.471],[51.719,5.032,-57.079],
[51.361,2.353,-58.990],[49.372,4.006,-51.869],[34.929,6.407,-49.549],
[49.726,5.203,-41.928],[50.408,3.063,-43.905],[49.896,-0.225,-58.092],
[50.054,1.876,-54.140],[50.103,1.522,-48.262],[48.781,2.721,-46.269],
[34.342,0.104,-41.931],[50.296,-2.586,-41.234],[50.004,-1.113,-53.717],
[37.141,-0.931,-48.888],[34.710,3.317,-48.025],[52.579,-2.593,-42.647],
[52.322,-1.661,-58.110],[33.968,-1.586,-55.104],[49.498,13.585,-51.401],
[35.268,-3.176,-45.095],[36.425,9.841,-42.125],[38.697,12.138,-47.252],
[43.839,13.184,-42.142],[46.822,14.086,-42.241],[50.332,-3.426,-43.991],
[36.856,16.917,-45.243],[38.981,15.564,-46.619],[42.778,14.235,-48.060],
[45.554,14.019,-47.111],[47.149,15.334,-45.016],[36.384,13.295,-48.536],
[39.249,12.821,-50.486],[42.060,13.471,-50.939],[43.565,-0.944,-49.917],
[49.254,-2.623,-48.722],[51.423,-2.740,-50.584],[37.416,-2.541,-51.037],
[36.605,-2.613,-56.187],[41.518,-3.643,-50.995],[46.758,-3.182,-49.556],
[49.123,-4.792,-55.947],[35.793,11.060,-53.798],[36.896,-0.524,-57.969],
[43.050,-0.575,-55.143],[45.123,13.827,-57.795],[52.272,12.520,-55.733],
[36.482,-2.244,-42.163],[38.795,-4.046,-40.606],[36.210,13.472,-58.256],
[45.188,-4.754,-40.883],[49.661,12.385,-41.890],[43.715,24.332,-51.575],
[45.487,23.281,-49.785],[38.091,23.039,-47.484],[49.446,27.215,-51.050],
[38.876,27.361,-49.811],[33.775,21.858,-53.567],[34.070,24.774,-51.191],
[46.201,33.009,-50.404],[42.841,27.391,-48.267],[51.833,18.278,-48.231],
[47.657,20.791,-48.304],[40.911,20.483,-51.708],[46.468,19.678,-53.105],
[34.901,21.982,-51.105],[40.766,23.246,-51.312],[52.666,20.632,-45.816],
[38.479,18.879,-48.096],[45.923,27.178,-52.111],[39.779,24.870,-45.593],
[50.201,18.234,-50.846],[36.506,21.233,-48.950],[37.422,28.534,-46.724],
[36.998,28.432,-55.009],[49.040,28.125,-48.415],[48.173,18.354,-46.821],
[42.590,16.066,-50.554],[46.821,31.279,-52.537],[47.995,28.779,-54.060],
[49.679,27.459,-45.327],[44.668,24.560,-54.558],[44.388,22.610,-58.980],
[52.605,23.201,-46.678],[47.417,24.614,-54.046],[47.173,22.346,-44.236],
[40.120,25.607,-52.522],[41.461,27.429,-50.466],[38.774,24.769,-49.686],
[44.713,29.842,-47.668],[41.186,29.294,-43.725],[46.070,24.815,-57.070],
[37.437,23.532,-44.454],[44.561,29.606,-52.843],[38.643,30.175,-44.196],
[50.650,14.585,-55.499],[37.903,23.858,-54.702],[52.615,29.813,-58.409],
[35.331,26.262,-46.480],[33.778,30.715,-48.358],[53.261,30.507,-43.560],
[44.482,29.734,-44.166],[44.191,31.608,-54.450],[37.497,32.955,-53.165],
[45.914,32.107,-57.029],[47.526,31.518,-55.039],[50.412,21.766,-45.113],
[36.433,27.857,-43.619],[41.063,17.796,-48.590],[44.314,15.604,-44.450],
[47.551,19.448,-43.887],[48.923,23.671,-57.576],[44.228,17.364,-57.354],
[40.763,21.847,-48.860],[40.549,22.374,-54.849],[44.897,22.091,-53.711],
[43.734,23.081,-42.317],[46.940,22.203,-51.947],[36.876,20.919,-45.728],
[41.102,19.287,-59.181],[35.108,17.617,-51.725],[35.663,20.905,-55.801],
[46.013,22.701,-46.938],[40.752,17.738,-51.741],[37.609,21.398,-53.034],
[39.843,19.456,-55.807],[45.625,18.805,-47.760],[39.168,22.295,-58.960],
[49.753,32.863,-41.715],[37.982,25.519,-42.914],[42.964,20.699,-50.009],
[36.908,18.591,-54.700],[43.279,20.492,-47.216],[42.224,24.427,-55.939],
[50.254,25.575,-47.882],[47.859,19.932,-57.952],[42.945,28.198,-45.780],
[47.228,22.000,-56.021],[36.245,26.628,-50.323],[44.174,27.019,-57.170],
[34.566,23.826,-44.707],[41.835,23.646,-44.027],[46.365,27.594,-44.891],
[53.115,21.413,-55.717],[52.058,28.094,-43.719],[50.468,29.640,-55.979],
[47.859,24.816,-48.680],[41.686,30.033,-53.572],[35.451,24.825,-58.930],
[42.489,27.105,-53.151],[43.451,32.076,-46.173],[40.356,32.464,-52.386],
[45.138,25.423,-47.709],[47.273,27.639,-58.178],[47.388,31.192,-43.949],
[51.683,33.098,-56.146],[35.302,19.210,-44.300],[48.184,26.047,-56.156],
[37.197,29.762,-49.305],[43.131,30.127,-59.395],[40.897,28.064,-55.855],
[36.393,26.122,-56.392],[44.146,18.611,-45.388],[46.328,18.332,-55.602],
[43.451,28.991,-55.470],[42.492,20.545,-54.105],[44.265,22.576,-45.036],
[43.775,17.021,-52.860],[40.593,31.233,-55.630],[40.241,17.192,-57.563],
[50.355,19.089,-45.018],[50.419,23.441,-51.282],[38.744,27.919,-57.372],
[35.898,21.471,-58.476],[48.823,31.936,-47.368],[37.308,15.964,-49.111],
[33.891,29.740,-54.637],[36.805,32.330,-43.104],[41.393,31.826,-41.151],
[43.604,32.653,-59.299],[46.132,28.444,-42.253],[48.702,33.291,-56.779],
[34.790,26.186,-41.118],[38.190,26.363,-59.570],[41.904,26.657,-42.976],
[44.457,27.857,58.957],[47.219,24.838,-45.032],[35.671,23.247,-42.172],
[38.983,24.490,-57.202],[39.716,21.920,-45.219],[41.608,23.272,-58.222],
[45.307,25.527,-39.274],[51.715,26.944,-58.406],[40.525,21.064,-42.643],
[40.714,19.368,-45.223],[42.864,25.098,-46.057],[47.326,22.132,-41.442],
[48.939,23.862,-42.740],[38.503,22.918,-41.923],[41.940,16.862,-45.298],
[42.539,19.385,-40.891],[44.058,20.060,-43.118],[46.815,18.471,-40.691],
[35.996,18.953,-41.702],[39.910,13.761,-45.011],[42.549,16.324,-41.048],
[42.025,34.007,-57.646],[36.101,28.568,-57.677],[35.186,33.206,-54.717],
[35.417,31.385,-51.285],[51.619,32.108,-49.120],[51.087,32.875,-46.469],
[51.578,28.140,-41.244],[50.726,26.606,-55.744],[51.835,23.563,-53.787],
[51.696,27.447,-49.087],[50.902,29.623,-46.454],[51.439,23.885,-56.584],
[53.100,20.001,-59.726],[49.151,22.050,-53.597],[33.827,25.805,-48.736],
[50.886,25.780,-43.330],[51.486,23.177,-41.948],[50.262,19.102,-59.740],
[49.728,19.282,-55.540],[50.269,20.572,-48.649],[48.998,22.875,-47.143],
[33.140,20.335,-42.325],[51.381,15.852,-40.468],[51.449,17.072,-54.520],
[36.975,18.346,-50.158],[34.037,23.142,-48.989],[51.643,17.217,-42.966],
[52.176,17.247,-57.756],[34.269,18.745,-54.863],[49.704,32.842,-53.387],
[34.394,16.688,-45.676],[36.723,29.997,-41.658],[40.107,29.600,-48.636],
[43.937,33.277,-43.485],[47.001,33.771,-43.411],[49.656,16.112,-44.730],
[38.274,33.276,-45.010],[39.148,31.691,-47.185],[44.697,33.797,-47.973],
[46.042,31.994,-46.328],[47.751,34.594,-45.818],[36.642,32.301,-48.562],
[39.737,29.517,-51.506],[41.557,32.127,-48.428],[44.795,18.046,-50.396],
[49.895,15.929,-49.167],[51.676,15.773,-51.253],[37.641,15.860,-51.882],
[34.934,15.269,-56.481],[42.242,14.730,-53.421],[47.934,16.504,-51.229],
[48.173,15.255,-53.735],[36.099,30.913,-56.300],[37.960,15.430,-56.719],
[41.807,17.748,-54.698],[43.822,33.933,-55.413],[52.608,31.329,-53.930],
[37.396,14.678,-43.091],[38.608,16.260,-41.177],[36.723,33.155,-57.678],
[45.310,15.667,-40.904],[50.457,30.314,-44.069],[46.238,42.768,-51.758],
[43.379,43.253,-49.803],[37.921,43.715,-48.523],[47.977,46.118,-51.187],
[38.680,46.222,-49.868],[34.295,40.321,-54.168],[34.812,42.449,-52.510],
[44.595,50.830,-50.431],[43.627,45.642,-48.842],[51.674,37.015,-46.803],
[46.880,39.358,-48.324],[38.408,39.103,-51.368],[47.150,38.301,-53.874],
[35.644,39.421,-51.376],[41.628,42.788,-51.510],[51.373,39.265,-45.149],
[37.806,36.692,-46.346],[45.591,45.584,-53.765],[39.634,43.791,-46.749],
[50.041,37.070,-50.359],[37.251,40.006,-48.850],[39.029,45.581,-44.220],
[36.157,46.864,-54.717],[48.765,48.693,-48.374],[49.006,37.069,-47.837],
[43.128,33.837,-50.475],[46.852,49.101,-50.586],[47.623,48.145,-53.354],
[49.713,45.245,-44.362],[44.443,43.264,-53.773],[44.894,40.188,-58.095],
[52.220,41.922,-47.768],[48.648,44.527,-53.319],[46.580,40.814,-45.181],
[39.286,42.534,-52.597],[41.254,46.509,-49.780],[39.743,41.802,-49.489],
[44.049,48.354,-49.398],[41.852,47.042,-44.545],[45.094,43.176,-57.227],
[37.419,42.971,-45.073],[44.439,49.586,-53.113],[38.183,48.880,-43.470],
[50.468,35.341,-57.715],[38.176,42.330,-55.377],[50.364,48.943,-57.985],
[34.518,44.971,-46.346],[34.143,48.395,-48.715],[53.573,47.757,-44.190],
[43.951,49.422,-45.640],[43.569,52.649,-54.954],[37.634,51.259,-51.894],
[47.089,50.764,-57.850],[47.291,50.651,-54.785],[49.100,40.438,-43.120],
[36.071,45.532,-44.100],[39.457,34.114,-48.328],[45.313,34.806,-45.127],
[47.005,38.238,-43.745],[46.891,44.664,-58.862],[45.471,37.545,-57.916],
[40.352,39.964,-47.539],[40.577,40.370,-51.907],[44.298,41.291,-55.218],
[45.215,41.110,-42.679],[46.648,39.935,-51.006],[36.650,39.384,-45.986],
[40.683,38.610,-58.146],[34.061,37.139,-51.079],[36.189,39.348,-55.926],
[43.812,39.851,-46.880],[41.750,36.527,-50.702],[37.473,40.466,-53.712],
[40.755,38.488,-54.950],[46.366,36.891,-46.770],[39.433,41.274,-58.026],
[49.206,52.762,-42.253],[37.627,43.953,-42.359],[42.932,40.485,-49.480],
[37.966,37.156,-55.347],[42.844,37.792,-48.781],[43.080,44.399,-55.881],
[47.545,42.959,-46.809],[47.914,38.902,-57.982],[44.059,45.721,-44.670],
[46.956,40.962,-55.846],[37.308,44.156,-51.260],[46.026,47.377,-55.783],
[34.025,42.261,-44.933],[42.215,42.434,-42.936],[45.822,47.851,-43.729],
[50.963,40.159,-55.250],[52.438,46.310,-42.407],[49.787,47.869,-55.635],
[46.380,44.044,-49.142],[40.393,47.623,-52.263],[35.988,43.590,-57.957],
[41.427,44.386,-53.719],[41.262,49.371,-45.851],[40.969,50.192,-53.221],
[45.250,47.128,-46.858],[48.227,46.639,-57.844],[48.455,49.992,-45.821],
[49.792,51.891,-58.527],[34.514,38.762,-44.614],[47.403,44.646,-55.800],
[36.684,48.015,-47.499],[39.697,48.578,-59.161],[40.304,47.956,-55.237],
[37.453,44.923,-56.033],[44.179,37.266,-45.041],[46.231,36.107,-55.115],
[43.239,48.136,-55.324],[43.703,40.354,-52.501],[42.587,39.565,-44.304],
[44.486,35.735,-52.664],[38.739,50.141,-56.857],[39.612,35.602,-57.303],
[49.217,37.299,-45.141],[49.561,40.146,-50.481],[38.915,46.872,-57.310],
[35.752,41.002,-58.369],[49.875,51.345,-47.996],[36.477,34.287,-50.640],
[35.173,48.453,-52.787],[35.500,51.413,-42.441],[40.379,50.098,-40.749],
[43.270,50.135,-59.042],[47.085,49.842,-41.290],[47.923,48.960,59.275],
[35.262,44.976,-41.321],[37.220,45.268,59.048],[40.522,44.488,-42.080],
[42.733,47.002,-58.749],[46.913,45.517,-44.937],[36.773,41.483,-42.696],
[41.150,41.610,-55.813],[40.250,41.431,-45.029],[42.455,42.410,-58.517],
[47.034,42.666,-41.429],[51.750,42.718,-59.729],[40.281,37.810,-44.404],
[40.330,36.906,-47.870],[42.714,43.054,-46.679],[48.108,39.220,-40.659],
[48.075,42.997,-43.904],[39.311,41.342,-41.579],[42.034,34.728,-47.744],
[42.893,39.820,-40.770],[44.805,37.867,-41.759],[47.427,36.622,-41.773],
[37.860,38.488,-41.911],[40.341,34.104,-42.867],[41.542,35.973,-41.340],
[40.595,51.746,-55.514],[36.180,46.619,-57.856],[32.328,51.033,-51.254],
[34.788,51.180,-52.306],[53.364,50.838,-47.159],[51.186,49.534,-46.143],
[49.712,46.133,-41.649],[51.717,44.508,-56.632],[51.236,43.737,-53.277],
[50.081,46.494,-49.221],[50.326,46.786,-46.593],[50.509,41.757,-57.558],
[52.435,39.668,59.446],[49.030,42.663,-51.406],[34.814,43.419,-50.086],
[49.718,42.965,-41.075],[52.461,41.541,-40.563],[51.534,38.411,-57.267],
[50.266,37.601,-55.056],[50.783,39.198,-48.059],[49.430,41.124,-45.963],
[35.396,39.268,-42.220],[51.329,35.473,-42.291],[51.725,35.455,-53.942],
[37.489,36.846,-49.890],[33.616,41.043,-49.409],[50.140,37.601,-42.782],
[53.430,34.343,-58.942],[33.548,37.391,-54.353],[50.937,51.301,-52.834],
[35.484,35.314,-45.024],[36.220,47.852,-41.448],[39.596,47.073,-46.534],
[43.362,52.002,-42.678],[46.954,51.729,-43.718],[50.424,34.028,-44.177],
[36.105,52.682,-45.282],[38.659,50.907,-46.065],[44.994,52.492,-48.271],
[46.360,49.967,-47.562],[47.207,53.492,-48.971],[36.655,50.564,-48.858],
[38.984,49.250,-49.472],[41.541,50.077,-50.219],[45.611,36.679,-49.295],
[49.047,32.711,-50.735],[52.060,34.864,-50.356],[37.801,35.631,-52.987],
[34.291,35.015,-56.753],[40.726,35.364,-52.959],[47.168,34.858,-52.026],
[48.921,35.675,-53.917],[32.921,46.560,-53.823],[36.669,37.617,-58.133],
[42.788,36.731,-57.123],[44.619,50.836,-56.913],[50.408,50.996,-55.563],
[37.854,35.636,-43.556],[38.556,34.847,-40.877],[37.850,51.405,-59.084],
[46.662,34.270,-40.828],[49.524,50.255,-43.090],[51.715,56.566,-47.524],
[46.595,57.547,-47.551],[38.526,57.021,-51.088],[47.893,56.887,-53.715],
[35.224,58.425,-50.572],[54.712,57.845,-45.123],[37.686,54.867,-46.628],
[48.281,55.502,-50.829],[37.185,57.136,-48.509],[49.531,54.490,-47.031],
[42.619,52.907,-49.521],[50.193,53.873,-56.573],[50.579,58.629,-42.148],
[38.404,53.771,-49.201],[45.049,52.551,-45.414],[48.361,57.026,-42.858],
[44.037,53.645,-58.962],[39.821,58.916,-48.876],[40.532,59.262,-53.520],
[46.858,57.888,-51.515],[36.229,57.734,-45.528],[41.273,55.636,58.055],
[34.195,56.403,-52.438],[36.716,56.267,-55.304],[42.096,57.210,-52.809],
[37.305,58.928,-53.205],[38.936,58.281,-55.713],[45.490,55.077,-46.212],
[42.319,-59.433,-58.424],[44.400,58.354,-50.017],[39.296,52.510,-53.176],
[40.086,58.312,-46.121],[46.604,55.827,-58.295],[45.988,58.166,-57.083],
[52.777,56.737,-58.553],[34.308,54.871,-45.745],[43.399,55.286,-47.905],
[46.207,55.302,-55.091],[43.234,58.309,-56.026],[43.930,57.529,-46.254],
[43.926,54.911,-53.395],[38.801,54.307,-55.167],[51.199,57.051,-44.735],
[36.348,59.215,-59.292],[35.703,53.080,-48.576],[40.696,55.969,-44.774],
[41.011,55.481,-49.185],[44.582,-59.880,-41.223],[38.527,56.533,-43.426],
[42.554,52.692,-46.374],[41.495,56.445,-42.006],[44.993,56.504,-43.768],
[47.405,54.659,-41.369],[36.030,55.640,-40.766],[40.736,52.073,-43.438],
[42.092,53.757,-41.137],[49.615,-58.726,59.434],[50.250,57.522,-58.597],
[48.744,56.847,-56.357],[49.469,58.122,-49.313],[49.703,58.983,-46.738],
[32.655,56.825,-42.329],[51.066,53.813,-40.818],[48.940,54.430,-54.266],
[35.517,55.072,-50.308],[34.309,-59.273,-48.275],[52.706,53.986,-42.997],
[53.523,53.557,-57.844],[34.073,55.546,-55.202],[33.889,52.328,-46.824],
[49.138,52.394,-44.927],[45.265,55.089,-50.783],[48.573,51.869,-50.784],
[50.390,53.994,-51.777],[36.365,53.752,-52.765],[33.889,52.467,-55.572],
[42.306,52.482,-52.179],[45.797,52.759,-52.044],[47.178,53.455,-56.748],
[37.713,54.297,-58.042],[41.871,53.499,-57.285],[35.852,55.642,-43.436],
[37.925,51.967,-43.403],[46.178,52.234,-41.044],[44.660,-49.023,-32.992],
[43.677,-51.036,-31.342],[37.017,-50.642,-30.001],[48.098,-47.811,-33.656],
[36.608,-46.145,-30.832],[34.593,-52.243,-34.967],[36.133,-50.651,-33.227],
[45.610,-42.110,-30.576],[43.826,-45.939,-30.440],[50.099,-56.368,-29.669],
[43.223,-54.645,-30.988],[38.019,-55.059,-30.584],[47.929,-55.377,-34.074],
[33.578,-52.445,-31.342],[39.033,-51.729,-33.094],[50.530,-53.987,-28.152],
[36.720,-55.488,-28.364],[44.993,-46.538,-34.931],[38.773,-49.251,-28.501],
[48.921,-57.026,-32.110],[36.303,-52.825,-31.267],[40.789,-47.013,-28.146],
[35.142,-46.621,-35.977],[49.630,-47.038,-31.730],[48.739,-56.387,-27.002],
[42.089,58.505,-31.885],[47.121,-44.754,-31.456],[47.470,-45.423,-34.170],
[50.121,-47.867,-25.212],[43.848,-49.285,-36.202],[45.160,-50.938,-38.821],
[51.993,-51.240,-28.681],[48.356,-49.677,-35.487],[45.991,-53.155,-26.935],
[42.188,-50.092,-34.122],[40.721,-45.705,-30.889],[40.447,-50.670,-30.052],
[47.138,-44.260,-28.661],[41.074,-45.066,-25.992],[47.109,-49.163,-37.631],
[36.709,-48.890,-26.122],[42.595,-44.894,-35.559],[38.977,-43.141,-24.227],
[49.437,-57.214,-37.437],[37.723,-50.450,-35.857],[50.783,-43.267,-40.271],
[34.120,-48.325,-26.563],[34.120,-46.183,-29.203],[53.751,-44.331,-25.147],
[44.278,-44.398,-25.453],[41.427,-42.572,-36.145],[36.427,-43.508,-34.508],
[45.778,-43.124,-38.342],[45.912,-42.961,-35.541],[49.792,-52.244,-23.839],
[37.522,-45.928,-25.129],[39.460,-57.381,-29.201],[45.204,-57.895,-26.946],
[47.237,-55.313,-24.997],[47.928,-50.583,-39.766],[44.462,-56.426,-39.675],
[40.800,-53.356,-31.075],[39.052,-52.773,-36.730],[46.270,-51.296,-34.420],
[43.653,-51.461,-23.663],[46.526,-53.018,-32.075],[36.071,-52.691,-26.761],
[41.390,-53.240,-39.881],[33.545,-56.102,-33.163],[35.804,-56.304,-37.649],
[44.618,-51.424,-28.790],[41.541,-55.556,-32.790],[37.193,-53.603,-34.308],
[41.956,-53.304,-36.860],[44.818,-56.032,-29.023],[38.038,-54.623,-38.615],
[48.177,-39.896,-23.969],[38.066,-48.472,-23.803],[42.485,-52.766,-33.363],
[39.046,-56.872,-34.286],[41.763,-54.131,-28.798],[42.170,-46.958,-37.581],
[48.587,-49.362,-29.770],[47.481,-55.196,-39.307],[44.229,-47.459,-25.548],
[46.825,-51.927,-36.936],[38.601,-48.941,-31.616],[45.771,-46.698,-38.221],
[34.179,-51.368,-25.439],[41.178,-52.193,-24.483],[46.342,-46.320,-26.450],
[50.827,-54.186,-36.898],[52.143,-46.388,-24.660],[50.430,-46.000,-37.870],
[46.960,-49.796,-32.122],[40.213,-44.167,-33.836],[35.797,-50.935,-38.657],
[40.312,-46.756,-34.355],[42.213,-42.719,-26.791],[38.540,-41.874,-34.821],
[43.753,-47.421,-28.131],[48.158,-47.267,-39.259],[46.915,-43.443,-25.684],
[52.247,-41.253,-38.721],[36.105,-55.369,-25.801],[48.267,-46.817,-36.370],
[36.255,-44.980,-28.389],[40.834,-45.514,-40.005],[40.180,-45.141,-37.060],
[37.445,-48.791,-38.414],[43.815,-55.649,-25.537],[45.691,-56.161,-36.281],
[43.023,-43.654,-37.996],[44.537,-53.430,-35.478],[43.074,-52.520,-26.728],
[44.924,-55.177,-33.308],[40.309,-42.581,-38.918],[40.198,-56.199,-37.710],
[50.313,-55.182,-25.087],[48.155,-51.886,-30.308],[37.373,-45.484,-37.990],
[35.736,-53.407,-39.834],[49.248,-43.003,-30.330],[37.365,-57.943,-31.048],
[35.403,-45.636,-33.102],[35.859,-42.585,-22.699],[42.158,-43.398,-23.638],
[43.542,-44.408,-40.681],[47.893,-43.985,-22.880],[48.144,-43.892,-40.044],
[35.033,-48.893,-23.699],[38.515,-47.243,-40.503],[41.028,-47.854,-24.779],
[43.667,-47.258,-40.000],[46.537,-48.903,-26.992],[36.509,-51.457,-24.023],
[41.010,-49.046,-36.303],[39.992,-52.574,-27.230],[40.317,-48.943,-39.389],
[43.672,-48.788,-23.525],[50.049,-48.839,-40.308],[40.160,-55.037,-24.782],
[41.324,-56.012,-27.053],[41.815,-49.917,-26.428],[44.465,-53.377,-21.787],
[47.089,-51.118,-24.945],[38.807,-52.622,-23.608],[40.044,-57.902,-25.802],
[42.864,-55.159,-22.966],[45.392,-53.422,-24.231],[46.616,-57.231,-23.013],
[37.732,-55.366,-22.759],[41.938,-59.406,-24.027],[42.306,-57.959,-21.796],
[41.919,-40.037,-38.500],[34.587,-46.513,-38.699],[33.176,-41.951,-35.422],
[34.015,-41.842,-32.648],[51.121,-40.841,-30.838],[51.634,-42.408,-26.063],
[49.494,-45.994,-22.313],[51.328,-48.552,-36.462],[50.665,-50.758,-34.834],
[51.031,-47.207,-28.666],[49.692,-44.644,-27.869],[50.259,-50.731,-37.999],
[51.082,-53.173,-39.435],[49.749,-49.926,-32.555],[33.430,-48.873,-29.937],
[51.794,-49.562,-26.278],[51.927,-50.496,-23.275],[49.988,-54.848,-41.343],
[48.187,-54.790,-36.698],[48.973,-54.337,-31.068],[49.081,-51.600,-27.411],
[35.099,-55.943,-23.296],[53.616,-57.822,-22.148],[50.292,-56.173,-35.141],
[33.907,-54.720,-29.890],[34.913,-50.481,-28.118],[51.310,-56.427,-22.907],
[52.421,-55.724,-40.615],[35.764,-56.810,-34.583],[47.526,-42.690,-33.133],
[34.696,-57.811,-26.934],[35.022,-45.233,-23.359],[38.570,-44.264,-26.966],
[43.980,-41.633,-24.845],[45.226,-39.344,-24.468],[49.403,58.856,-25.298],
[36.283,-41.796,-26.034],[39.242,-41.387,-28.194],[43.649,-38.861,-29.969],
[45.347,-42.609,-27.687],[48.839,-41.710,-27.881],[35.572,-42.420,-30.241],
[38.373,-46.309,-32.651],[39.023,-42.427,-31.813],[43.114,-57.400,-31.441],
[49.065,-59.500,-30.421],[51.206,58.168,-34.183],[37.206,58.160,-33.809],
[35.799,-58.599,-39.092],[44.218,-57.668,-34.179],[46.419,-57.859,-32.351],
[47.068,-58.522,-35.340],[34.803,-43.783,-37.468],[38.198,-57.487,-39.361],
[42.948,-57.600,-36.534],[45.049,-41.321,-40.537],[51.412,-43.489,-36.995],
[36.678,-58.401,-23.691],[37.584,-58.351,-21.244],[38.021,-43.435,-39.783],
[47.981,59.247,-21.514],[49.598,-44.516,-25.047],[46.998,-31.084,-33.477],
[45.199,-31.591,-30.949],[37.623,-32.970,-29.721],[50.295,-26.138,-31.749],
[38.788,-28.974,-31.872],[33.197,-34.310,-35.141],[34.712,-30.403,-33.253],
[45.971,-23.799,-31.220],[43.802,-29.292,-30.308],[51.397,-38.120,-29.026],
[46.322,-35.236,-28.555],[37.748,-36.443,-32.962],[47.245,-37.574,-35.722],
[35.209,-35.218,-33.247],[41.597,-33.240,-32.750],[51.931,-35.424,-27.425],
[36.287,-37.013,-28.253],[43.747,-27.595,-34.135],[41.012,-28.405,-28.450],
[48.128,-37.590,-33.045],[36.873,-35.462,-30.379],[38.390,-29.670,-28.806],
[35.676,-28.117,-37.041],[48.618,-28.492,-28.426],[48.828,-37.148,-28.870],
[41.708,-42.194,-30.361],[47.797,-25.248,-32.343],[47.513,-28.054,-32.635],
[48.236,-29.576,-23.836],[44.871,-30.914,-36.173],[44.997,-31.613,-40.002],
[51.392,-32.995,-28.787],[48.659,-31.443,-35.677],[46.567,-32.869,-27.067],
[38.897,-31.739,-31.982],[40.922,-26.826,-30.872],[40.678,-31.828,-29.592],
[46.963,-26.199,-28.526],[39.209,-27.629,-25.493],[46.828,-32.447,-37.493],
[38.233,-31.835,-27.075],[44.464,-24.824,-33.890],[37.946,-24.610,-25.403],
[48.803,-42.259,-37.808],[39.159,-31.847,-36.145],[51.259,-27.889,-38.689],
[35.661,-30.614,-28.076],[34.418,-25.945,-29.920],[51.821,-26.263,-24.690],
[44.650,-26.688,-25.328],[43.679,-22.780,-36.150],[36.929,-23.836,-32.852],
[46.763,-25.370,-38.381],[47.402,-26.636,-35.555],[48.238,-34.840,-22.338],
[36.822,-27.234,-28.077],[39.337,-38.588,-29.035],[44.777,-40.010,-27.508],
[47.196,-37.204,-25.808],[47.639,-31.039,-40.371],[44.899,-36.632,-39.973],
[39.501,-34.731,-30.718],[42.383,-34.923,-35.015],[45.555,-34.434,-35.246],
[43.599,-34.085,-23.300],[46.292,-33.675,-32.678],[36.943,-34.546,-27.097],
[40.134,-38.036,-41.356],[34.437,-38.363,-32.921],[35.265,-36.779,-37.889],
[44.123,-32.545,-28.439],[40.820,-38.529,-32.421],[36.265,-36.359,-35.313],
[39.744,-35.563,-35.232],[46.002,-37.878,-28.745],[40.179,-35.624,-38.375],
[49.776,-20.752,-23.381],[37.741,-29.833,-25.442],[44.224,-34.388,-30.997],
[38.373,-39.098,-33.280],[43.341,-35.139,-28.443],[42.156,-30.867,-36.909],
[49.244,-31.379,-28.437],[47.010,-38.037,-39.114],[43.218,-30.488,-26.020],
[45.907,-35.326,-37.861],[36.336,-29.805,-31.104],[44.754,-28.091,-37.422],
[35.633,-31.922,-25.582],[41.337,-32.802,-24.409],[47.865,-27.605,-25.732],
[53.972,-33.981,-37.788],[52.456,-28.833,-24.993],[50.681,-26.402,-36.051],
[48.078,-30.039,-30.770],[40.127,-27.734,-34.948],[35.658,-32.531,-37.830],
[41.924,-30.034,-34.360],[40.882,-24.337,-29.597],[38.960,-23.813,-34.613],
[46.245,-29.821,-27.944],[46.955,-28.535,-39.622],[47.815,-24.736,-25.738],
[50.885,-22.829,-40.093],[35.625,-36.873,-25.461],[47.009,-29.721,-37.210],
[36.085,-24.616,-28.121],[42.028,-27.164,-40.623],[39.954,-28.236,-37.780],
[37.520,-30.049,-38.421],[43.215,-37.846,-27.499],[45.576,-39.824,-36.586],
[43.029,-25.322,-36.919],[43.220,-35.904,-37.414],[43.201,-34.410,-25.873],
[43.709,-40.203,-34.109],[39.889,-23.615,-37.193],[38.610,-37.950,-37.955],
[49.968,-37.047,-25.973],[48.575,-32.792,-30.966],[37.764,-26.253,-37.833],
[36.112,-34.808,-39.439],[49.276,-24.664,-29.108],[36.242,-39.703,-30.859],
[35.942,-26.398,-34.715],[35.858,-24.657,-23.312],[38.883,-24.228,-21.299],
[44.264,-26.234,-39.112],[45.889,-27.588,-23.032],[49.319,-25.514,-39.388],
[34.747,-30.445,-22.364],[41.484,-30.548,-41.278],[40.974,-29.689,-24.653],
[43.696,-29.188,-40.294],[45.296,-29.265,-25.150],[37.836,-32.476,-24.274],
[39.977,-30.899,-38.818],[39.591,-34.336,-26.088],[41.299,-32.889,-40.110],
[45.059,-31.650,-23.977],[51.144,-30.075,-40.034],[41.117,-36.068,-23.274],
[40.161,-37.119,-26.777],[41.666,-31.966,-27.208],[46.446,-34.312,-24.126],
[48.664,-33.027,-24.922],[39.506,-34.028,-22.747],[41.505,-39.945,-27.299],
[43.546,-37.407,-22.589],[44.519,-36.608,-25.576],[46.487,-38.107,-22.311],
[37.292,-35.030,-24.007],[41.161,-41.331,-24.816],[41.153,-39.177,-22.521],
[41.735,-23.745,-39.116],[34.585,-27.891,-39.313],[33.142,-22.339,-36.185],
[34.614,-22.879,-33.428],[51.506,-23.458,-31.006],[50.462,-24.359,-26.468],
[51.819,-28.122,-22.563],[51.807,-30.225,-37.219],[51.558,-31.172,-34.872],
[51.641,-28.789,-30.337],[50.792,-26.859,-28.645],[51.152,-32.429,-38.952],
[50.747,-34.951,-38.667],[48.815,-33.417,-33.790],[34.345,-31.385,-30.398],
[50.703,-30.756,-24.626],[51.265,-33.425,-24.141],[49.407,-36.484,-40.566],
[48.960,-36.694,-37.697],[49.085,-35.465,-31.283],[48.979,-34.105,-28.326],
[33.962,-35.443,-23.460],[52.808,-40.276,-22.701],[50.531,-38.983,-36.205],
[34.151,-36.988,-30.569],[34.584,-34.249,-30.364],[51.253,-38.221,-23.801],
[51.716,-38.007,-38.183],[34.585,-39.045,-36.707],[48.153,-22.368,-33.035],
[35.139,-39.206,-27.156],[36.291,-27.598,-24.670],[39.293,-26.099,-28.335],
[43.283,-24.477,-24.907],[47.246,-22.119,-25.644],[49.983,-39.880,-26.371],
[36.940,-21.042,-28.128],[38.351,-23.120,-27.657],[43.705,-23.407,-29.513],
[46.306,-23.014,-28.363],[47.865,-20.560,-27.826],[35.691,-23.486,-30.431],
[38.344,-26.117,-33.055],[39.872,-23.380,-32.048],[42.082,-36.530,-30.980],
[48.319,-40.428,-31.529],[50.418,-39.379,-33.600],[35.418,-40.393,-34.605],
[36.084,-41.526,-36.924],[40.995,-40.342,-34.540],[45.140,-39.276,-32.299],
[48.216,-39.706,-37.240],[35.164,-25.085,-37.116],[37.005,-37.389,-40.061],
[41.258,-37.557,-36.502],[43.719,-23.813,-40.923],[49.269,-24.277,-34.653],
[36.522,-39.645,-23.633],[38.532,-38.752,-21.633],[36.731,-23.832,-39.221],
[45.544,-40.563,-21.620],[50.748,-23.674,-22.804],[46.578,-12.505,-35.213],
[44.069,-12.320,-30.324],[37.003,-13.247,-28.422],[48.682,-9.821,-32.346],
[39.199,-9.070,-29.577],[34.765,-15.167,-34.858],[34.545,-12.905,-32.898],
[45.282,-3.353,-32.696],[43.473,-9.686,-29.435],[50.647,-19.349,-28.945],
[46.615,-16.389,-29.325],[39.028,-18.381,-32.641],[46.350,-17.541,-34.210],
[35.216,-15.460,-31.911],[39.865,-14.579,-31.684],[52.456,-17.157,-24.723],
[38.199,-18.487,-29.090],[43.312,-9.388,-34.441],[39.403,-12.767,-26.937],
[48.635,-19.525,-31.380],[37.664,-16.615,-31.212],[37.176,-9.461,-27.615],
[37.271,-9.177,-33.736],[48.093,-9.817,-29.380],[49.628,-17.250,-27.349],
[41.564,-21.815,-30.640],[48.746,-6.162,-31.507],[46.845,-8.378,-34.020],
[48.905,-10.636,-26.110],[44.578,-13.062,-37.418],[44.122,-16.111,-38.174],
[51.739,-12.434,-28.087],[48.886,-14.546,-35.028],[47.845,-15.125,-27.276],
[39.134,-10.714,-32.437],[40.903,-8.675,-31.853],[39.380,-12.139,-29.720],
[45.726,-8.423,-27.860],[40.692,-9.824,-24.593],[46.314,-15.144,-36.621],
[38.522,-14.444,-24.367],[42.856,-6.640,-34.546],[39.383,-6.828,-23.882],
[50.941,-20.716,-37.412],[38.458,-12.579,-36.808],[52.159,-7.465,-40.787],
[35.669,-11.484,-26.758],[34.594,-7.420,-29.399],[53.324,-9.445,-23.199],
[42.656,-8.464,-25.961],[45.106,-3.497,-36.344],[37.827,-5.015,-33.219],
[46.927,-5.487,-38.587],[46.143,-6.176,-35.928],[50.740,-15.172,-25.752],
[36.479,-9.849,-24.738],[38.559,-21.684,-30.406],[43.711,-21.606,-25.889],
[47.336,-18.392,-26.535],[48.352,-12.105,-40.372],[42.973,-18.370,-37.091],
[38.453,-15.365,-28.795],[37.407,-15.439,-36.001],[44.066,-14.964,-34.924],
[45.369,-16.007,-21.749],[45.943,-15.367,-31.917],[36.620,-16.051,-26.780],
[39.994,-17.542,-39.439],[34.033,-18.904,-32.689],[35.612,-17.302,-38.048],
[44.504,-14.794,-29.371],[40.977,-19.983,-33.337],[36.624,-17.046,-33.814],
[40.331,-18.342,-36.926],[46.068,-18.977,-30.165],[39.754,-15.265,-37.896],
[48.989,-3.461,-23.738],[37.294,-12.134,-21.834],[42.995,-16.553,-31.091],
[37.497,-18.813,-35.648],[42.182,-15.093,-28.366],[42.999,-11.010,-36.639],
[49.193,-12.233,-29.208],[44.739,-19.613,-39.935],[45.289,-10.022,-25.666],
[46.918,-17.214,-38.237],[36.570,-10.342,-31.097],[45.367,-10.143,-36.733],
[33.482,-12.606,-25.318],[41.234,-14.584,-24.111],[48.252,-8.174,-25.390],
[52.166,-15.793,-37.282],[51.212,-9.060,-25.199],[52.491,-8.295,-38.148],
[46.737,-13.426,-29.763],[40.499,-7.978,-35.020],[34.877,-13.074,-37.816],
[40.909,-11.118,-34.625],[40.609,-6.574,-26.281],[38.022,-7.778,-36.631],
[44.559,-11.710,-27.704],[46.549,-10.213,-39.492],[48.272,-5.563,-26.808],
[51.945,-4.832,-39.699],[35.118,-18.901,-24.724],[47.884,-9.705,-36.539],
[36.287,-7.157,-27.105],[41.088,-8.679,-41.813],[41.462,-9.464,-39.127],
[35.939,-10.444,-37.449],[43.789,-18.238,-27.968],[45.167,-19.730,-37.077],
[43.085,-8.051,-37.598],[42.499,-17.344,-34.669],[42.055,-17.211,-25.553],
[43.509,-20.086,-34.907],[40.809,-5.941,-38.131],[39.178,-22.176,-39.201],
[49.457,-18.147,-24.601],[49.363,-13.976,-30.862],[37.159,-8.788,-38.893],
[37.188,-14.419,-38.653],[49.303,-4.586,-29.026],[36.601,-20.115,-30.593],
[33.436,-9.184,-34.066],[34.310,-5.909,-22.305],[41.859,-6.044,-22.848],
[43.460,-7.219,-40.560],[46.786,-9.051,-22.797],[49.194,-7.475,-38.958],
[35.069,-11.414,-23.169],[38.975,-10.200,-40.260],[42.328,-12.044,-23.712],
[43.243,-10.097,-41.633],[46.157,-12.615,-25.860],[35.725,-14.080,-25.170],
[41.285,-13.200,-36.862],[40.161,-15.625,-26.742],[40.022,-12.877,-39.490],
[47.011,-11.959,-22.980],[51.559,-12.300,-40.866],[40.768,-19.529,-25.316],
[40.989,-18.520,-28.698],[43.235,-13.850,-25.986],[48.039,-15.873,-22.626],
[48.477,-13.255,-24.849],[39.829,-17.055,-24.326],[41.331,-21.050,-27.649],
[42.900,-19.436,-23.813],[44.916,-17.724,-24.821],[46.800,-19.629,-23.726],
[35.655,-17.931,-22.455],[40.075,-22.735,-24.203],[42.282,-20.433,-21.550],
[40.657,-3.019,-39.040],[34.346,-7.423,-39.342],[34.962,-4.475,-33.739],
[35.496,-6.844,-31.957],[52.042,-5.070,-29.965],[51.917,-5.653,-26.269],
[49.151,-9.258,-20.909],[50.429,-12.387,-36.031],[50.473,-12.353,-33.125],
[50.859,-9.439,-30.591],[50.323,-7.598,-27.662],[51.034,-13.429,-38.502],
[50.305,-16.406,-39.697],[48.453,-15.942,-32.740],[34.401,-11.840,-30.137],
[51.742,-11.981,-25.405],[51.009,-13.829,-23.510],[49.866,-19.567,-39.859],
[49.015,-16.673,-36.818],[49.658,-17.190,-30.347],[50.981,-14.942,-28.611],
[33.612,-16.387,-24.480],[53.512,-21.827,-22.984],[49.894,-19.735,-34.960],
[35.224,-17.850,-30.254],[33.407,-14.676,-29.191],[52.614,-19.451,-23.283],
[52.319,-18.216,-39.065],[34.213,-17.903,-35.820],[49.229,-5.359,-34.368],
[33.701,-19.764,-26.927],[36.708,-7.298,-23.351],[42.539,-7.069,-28.501],
[44.665,-6.747,-23.717],[45.621,-4.118,-25.232],[49.944,-21.189,-25.903],
[37.079,-4.467,-24.719],[39.789,-5.645,-28.537],[44.367,-3.645,-29.638],
[45.875,-5.479,-28.026],[45.716,-1.687,-27.058],[36.795,-5.466,-29.502],
[38.827,-6.965,-31.450],[41.998,-5.839,-31.234],[43.786,-18.936,-32.568],
[49.531,-21.736,-29.796],[50.703,-21.493,-32.876],[35.863,-20.135,-34.222],
[36.371,-21.491,-37.412],[42.221,-22.519,-33.634],[45.932,-20.737,-32.167],
[48.143,-21.810,-35.646],[36.082,-4.733,-36.148],[36.970,-19.732,-39.413],
[41.410,-20.766,-36.460],[43.413,-4.197,-38.656],[52.443,-6.056,-36.507],
[35.840,-21.260,-23.769],[39.877,-21.717,-21.626],[37.258,-5.836,-39.093],
[47.242,-22.285,-22.709],[50.238,-6.242,-24.134],[46.167,5.533,-34.132],
[43.716,6.263,-30.646],[36.674,4.935,-29.551],[47.282,8.961,-31.856],
[37.385,10.517,-30.606],[34.387,2.987,-34.877],[32.883,6.599,-33.620],
[45.492,12.118,-30.294],[43.769,9.709,-30.460],[52.158,-0.336,-28.596],
[47.583,2.242,-30.239],[39.344,0.692,-30.980],[46.556,-0.544,-32.785],
[36.900,4.412,-32.403],[41.582,4.769,-33.574],[52.178,2.426,-27.374],
[38.323,1.322,-27.310],[46.416,8.197,-34.777],[40.453,7.422,-27.891],
[49.121,-0.157,-30.876],[36.160,1.852,-29.454],[37.858,7.225,-28.178],
[34.235,9.970,-37.044],[47.776,8.221,-29.239],[48.573,-1.735,-28.407],
[41.983,-2.946,-31.157],[46.185,13.556,-32.653],[47.288,11.360,-34.042],
[49.308,8.514,-26.068],[43.526,6.137,-35.214],[46.142,4.843,-40.139],
[51.747,4.800,-28.815],[48.615,5.579,-35.516],[46.720,5.260,-26.732],
[39.024,6.650,-33.214],[41.229,9.556,-31.889],[40.927,5.840,-30.537],
[46.062,9.909,-27.708],[39.877,9.497,-25.636],[47.171,4.098,-37.773],
[39.032,6.895,-25.444],[44.632,12.458,-34.558],[39.686,12.002,-24.559],
[51.012,-3.920,-36.101],[37.454,5.336,-36.278],[50.768,13.447,-39.669],
[35.555,8.691,-27.056],[33.147,11.367,-27.464],[53.116,12.169,-24.317],
[42.693,11.284,-25.430],[42.343,13.991,-35.663],[36.138,13.605,-33.514],
[44.883,13.320,-39.549],[45.647,14.916,-35.035],[50.068,3.276,-23.528],
[36.344,11.477,-25.824],[39.496,-3.379,-29.716],[43.690,-3.786,-26.822],
[47.205,1.255,-25.788],[47.991,6.647,-38.603],[43.770,0.596,-39.427],
[39.101,3.957,-29.498],[39.341,3.978,-34.755],[44.704,3.280,-35.758],
[41.517,2.611,-23.958],[46.763,2.744,-33.105],[35.387,2.702,-26.947],
[41.101,1.083,-40.065],[34.348,-0.930,-33.123],[35.074,0.689,-36.639],
[44.876,3.739,-28.455],[41.592,-0.176,-33.467],[37.581,2.194,-33.975],
[40.226,1.285,-36.946],[45.313,0.146,-29.352],[39.464,3.494,-39.521],
[48.671,13.910,-22.054],[36.416,6.949,-24.699],[44.189,3.807,-30.948],
[38.464,-0.537,-33.698],[42.595,2.320,-28.500],[42.016,7.774,-37.131],
[50.182,7.175,-28.359],[46.317,-0.963,-39.544],[42.335,9.213,-28.292],
[46.129,1.435,-37.890],[36.608,7.551,-32.417],[44.121,9.554,-37.615],
[33.449,6.021,-25.791],[39.777,4.899,-22.890],[44.547,8.862,-25.736],
[51.366,3.414,-37.258],[51.419,9.966,-25.232],[49.215,10.463,-36.731],
[47.528,5.524,-29.176],[40.031,9.975,-34.282],[35.847,6.173,-38.637],
[41.467,7.574,-34.211],[40.921,12.800,-28.038],[39.327,14.261,-34.447],
[45.025,7.786,-28.807],[46.488,8.832,-39.024],[47.456,12.421,-25.297],
[52.684,14.218,-37.927],[36.242,0.127,-25.369],[48.909,8.066,-35.604],
[35.677,10.809,-28.499],[40.649,10.584,-40.972],[41.774,10.221,-38.778],
[37.615,8.042,-38.286],[43.108,-1.434,-28.476],[44.750,-0.720,-36.876],
[43.654,12.258,-37.337],[42.953,1.050,-35.624],[42.529,4.799,-25.188],
[43.945,-1.668,-34.059],[40.023,12.165,-36.671],[39.822,-1.833,-36.571],
[51.130,1.302,-25.361],[48.875,4.543,-32.062],[38.825,10.289,-38.886],
[36.020,2.053,-39.673],[47.983,13.412,-30.115],[36.750,-2.505,-28.513],
[37.054,10.089,-36.501],[36.087,14.640,-23.454],[39.998,13.122,-22.231],
[43.612,10.997,-40.420],[46.489,9.695,-23.727],[47.515,12.551,-39.689],
[35.381,9.521,-23.466],[39.018,8.394,-41.234],[41.635,7.683,-24.427],
[43.396,7.861,-39.647],[46.749,7.094,-24.613],[37.566,3.632,-21.917],
[40.515,5.132,-37.537],[40.646,2.845,-26.685],[41.678,5.191,-40.244],
[44.110,7.009,-23.751],[50.692,7.323,-40.527],[40.420,0.195,-25.106],
[40.855,0.133,-27.989],[42.295,5.383,-27.986],[47.015,3.186,-23.781],
[48.593,5.202,-24.642],[37.746,2.662,-25.056],[40.869,-2.660,-26.945],
[42.876,1.174,-21.768],[44.579,1.945,-24.365],[46.682,-0.466,-23.792],
[37.243,0.874,-22.031],[41.511,-4.471,-25.129],[41.260,-2.616,-22.243],
[40.806,15.339,-39.086],[35.547,9.331,-39.554],[34.841,15.701,-34.820],
[34.262,11.347,-33.006],[52.326,14.097,-28.797],[51.935,13.521,-26.183],
[49.326,10.474,-23.413],[51.754,8.177,-35.981],[51.240,5.628,-35.158],
[51.040,9.217,-30.804],[49.149,11.221,-29.799],[51.023,6.024,-37.992],
[50.671,2.361,-39.588],[49.328,3.068,-34.348],[34.506,6.467,-31.136],
[51.085,5.966,-25.139],[52.358,5.061,-23.094],[49.078,1.008,-41.179],
[48.534,1.031,-36.862],[50.199,3.086,-29.917],[49.881,3.815,-26.613],
[33.595,1.408,-25.475],[51.653,-2.904,-22.643],[49.865,-1.589,-36.187],
[36.343,-0.049,-31.561],[34.249,3.249,-30.302],[50.214,-0.848,-23.884],
[51.488,0.012,-40.914],[34.932,-1.993,-35.829],[49.217,13.734,-34.054],
[33.765,-2.798,-27.261],[37.211,11.554,-23.321],[38.724,11.314,-27.352],
[43.431,13.473,-23.583],[45.842,14.779,-25.764],[50.030,-3.466,-26.421],
[36.988,16.366,-27.245],[37.299,13.895,-28.503],[43.538,14.319,-30.281],
[46.304,12.711,-27.652],[47.130,15.297,-28.499],[35.644,13.835,-30.754],
[38.431,13.734,-31.683],[41.643,13.193,-32.039],[43.121,-0.074,-31.211],
[50.497,-2.689,-30.459],[50.987,-2.763,-33.455],[36.977,-2.742,-31.285],
[35.613,-3.474,-38.354],[41.784,-3.796,-34.025],[47.920,-3.534,-32.640],
[47.706,-3.456,-36.223],[36.928,12.087,-38.314],[36.474,-1.012,-38.907],
[42.508,-2.215,-36.448],[43.794,15.492,-38.500],[50.125,13.146,-37.151],
[34.325,-2.995,-24.103],[39.255,-2.839,-24.133],[38.175,13.561,-40.385],
[46.439,-3.489,-22.549],[49.830,11.632,-26.715],[44.591,26.338,-32.707],
[44.219,24.361,-30.062],[37.454,23.262,-28.643],[49.745,27.018,-31.602],
[37.483,27.873,-29.401],[35.031,22.425,-35.010],[34.864,24.824,-31.859],
[45.652,30.455,-30.665],[43.404,27.347,-29.012],[52.135,18.720,-29.574],
[46.680,20.973,-30.125],[39.052,19.401,-31.120],[46.754,18.773,-35.188],
[34.475,21.369,-32.110],[41.484,21.806,-31.153],[52.575,20.334,-27.479],
[37.256,18.343,-29.103],[44.317,28.453,-34.689],[41.223,25.152,-28.666],
[47.676,18.582,-32.351],[36.726,21.378,-30.538],[38.893,26.021,-27.616],
[33.504,29.269,-37.610],[48.479,28.544,-29.179],[49.154,17.843,-27.966],
[41.172,16.299,-31.551],[49.021,29.695,-32.837],[47.811,27.584,-34.056],
[49.724,26.914,-24.943],[44.553,24.728,-35.488],[43.953,23.066,-39.402],
[51.672,23.480,-28.068],[46.625,24.777,-33.599],[47.403,21.942,-26.305],
[38.615,25.115,-34.009],[41.410,28.932,-31.678],[39.353,23.721,-30.859],
[46.268,27.175,-30.409],[40.345,28.391,-25.451],[46.827,24.232,-37.594],
[39.334,23.667,-26.458],[43.004,30.571,-33.678],[39.118,30.994,-25.926],
[50.782,16.459,-37.340],[40.143,23.732,-35.864],[51.276,30.334,-40.049],
[33.401,26.634,-26.981],[33.323,29.604,-31.245],[53.443,29.894,-25.351],
[44.205,29.714,-27.351],[44.246,33.711,-36.283],[37.828,33.008,-33.695],
[46.790,31.478,-40.370],[46.425,31.952,-35.950],[51.578,22.205,-25.838],
[37.339,27.874,-26.393],[38.993,16.211,-29.291],[43.679,16.097,-26.310],
[48.110,19.576,-25.147],[47.659,25.596,-40.618],[43.358,18.770,-38.453],
[39.384,21.303,-29.084],[39.835,22.744,-33.336],[43.171,22.269,-35.600],
[43.639,22.152,-22.548],[45.845,20.638,-32.949],[36.993,20.816,-27.524],
[38.549,18.938,-40.659],[34.284,19.270,-33.310],[35.270,19.577,-38.701],
[44.987,21.903,-28.132],[41.293,19.215,-32.841],[36.986,20.400,-33.219],
[40.091,20.990,-35.456],[46.298,17.945,-28.628],[39.619,21.508,-39.658],
[49.823,32.565,-23.327],[38.991,26.016,-24.553],[44.140,21.611,-31.131],
[38.566,17.162,-34.203],[41.949,22.621,-27.991],[42.575,26.556,-36.717],
[49.988,25.162,-29.524],[47.537,18.465,-37.860],[45.476,26.421,-25.221],
[46.672,21.520,-35.714],[36.939,26.519,-31.531],[45.135,27.874,-37.119],
[34.797,23.773,-26.649],[41.457,21.917,-24.524],[44.481,29.015,-24.779],
[51.397,22.823,-37.659],[52.579,26.584,-24.502],[50.285,28.881,-36.500],
[47.379,24.906,-30.872],[40.824,28.817,-35.343],[36.883,23.735,-38.345],
[41.775,26.245,-33.894],[41.418,30.871,-27.850],[40.044,32.313,-34.818],
[46.003,26.069,-28.035],[45.020,28.221,-39.956],[48.473,31.502,-26.459],
[51.741,32.514,-38.207],[36.005,18.455,-26.189],[47.573,26.791,-37.912],
[34.973,29.858,-27.343],[41.688,28.709,-40.641],[39.986,27.877,-38.418],
[37.654,26.450,-37.929],[43.435,17.632,-28.735],[44.840,17.144,-36.179],
[42.971,30.311,-36.790],[41.854,19.254,-35.993],[43.378,20.944,-26.297],
[43.847,18.463,-34.196],[40.334,30.795,-37.684],[39.970,17.881,-38.655],
[50.386,18.330,-25.375],[49.098,23.961,-32.689],[38.830,29.736,-39.947],
[36.360,21.543,-40.137],[47.742,33.200,-29.563],[34.928,16.856,-30.193],
[36.800,29.996,-35.097],[35.417,32.393,-25.456],[41.659,30.432,-24.508],
[43.662,30.141,-41.760],[46.465,29.071,-21.866],[48.270,29.222,-40.080],
[35.320,26.031,-24.860],[38.828,25.035,-40.372],[41.280,25.673,-22.853],
[42.405,26.252,-39.601],[48.018,24.663,-26.325],[36.123,22.502,-24.437],
[40.526,25.353,-37.947],[39.609,21.178,-26.379],[41.499,23.969,-40.858],
[45.684,23.666,-23.750],[50.307,25.100,-40.568],[40.411,18.683,-24.998],
[41.005,18.760,-27.894],[43.079,24.620,-26.401],[47.598,21.446,-23.213],
[47.993,25.257,-23.627],[38.922,21.396,-23.666],[41.432,15.521,-28.283],
[43.180,18.490,-24.836],[45.538,19.717,-24.061],[47.746,17.680,-22.865],
[36.502,19.420,-23.643],[40.924,14.488,-25.116],[42.259,16.028,-22.023],
[42.245,33.622,-38.933],[35.602,28.088,-38.962],[34.880,32.316,-34.633],
[35.248,31.637,-32.152],[51.951,32.113,-29.985],[51.345,31.810,-26.013],
[50.642,28.016,-22.243],[53.079,25.030,-36.347],[49.607,22.979,-35.561],
[51.350,27.686,-29.546],[50.374,29.787,-27.367],[53.445,24.192,-38.820],
[51.938,21.168,-39.593],[50.926,22.257,-33.249],[33.054,24.308,-30.044],
[50.472,24.342,-25.170],[52.224,23.067,-23.125],[49.331,19.567,-39.511],
[48.784,20.933,-37.294],[49.321,20.574,-32.182],[49.820,20.666,-28.387],
[33.949,20.522,-23.662],[52.824,15.984,-21.918],[49.166,17.452,-34.811],
[34.631,19.541,-29.253],[34.456,22.206,-29.073],[52.545,17.179,-24.521],
[52.120,18.467,-40.532],[34.406,17.750,-36.538],[48.258,32.359,-32.664],
[33.823,16.447,-26.867],[33.877,27.924,-22.497],[39.462,29.993,-29.841],
[44.217,31.780,-24.294],[47.423,33.846,-24.481],[50.317,15.554,-26.513],
[37.501,35.113,-27.367],[38.188,32.495,-28.104],[44.212,33.332,-29.632],
[46.147,31.569,-27.909],[47.404,34.528,-27.052],[36.443,30.656,-29.549],
[38.765,30.499,-32.549],[42.387,31.797,-31.281],[43.721,18.446,-31.520],
[50.318,14.217,-31.234],[51.127,16.280,-33.481],[36.665,15.920,-32.008],
[35.403,15.413,-37.971],[41.638,15.994,-34.145],[48.246,15.837,-32.572],
[48.296,15.165,-36.360],[35.693,31.065,-38.385],[35.979,16.957,-40.037],
[42.034,16.703,-36.929],[45.502,33.102,-38.507],[49.141,31.051,-35.519],
[37.567,16.857,-24.459],[39.257,16.227,-22.016],[37.505,32.380,-39.921],
[46.240,15.377,-22.841],[51.009,30.111,-23.789],[45.158,41.930,-33.128],
[43.435,43.129,-30.674],[36.781,42.047,-29.018],[47.098,46.762,-32.252],
[37.325,46.985,-31.308],[33.716,40.131,-34.677],[35.077,43.248,-33.091],
[46.249,50.102,-30.842],[42.806,46.063,-29.258],[51.566,37.102,-27.159],
[46.209,38.953,-29.370],[39.462,37.312,-31.828],[47.650,38.105,-36.051],
[36.311,40.708,-32.953],[39.017,40.917,-32.993],[51.459,40.829,-25.330],
[37.981,37.505,-29.061],[44.450,45.484,-33.693],[40.477,43.512,-27.863],
[47.746,37.403,-32.829],[37.135,39.673,-30.492],[38.461,46.646,-28.888],
[36.455,46.630,-36.404],[47.660,45.211,-30.176],[49.063,35.523,-28.940],
[41.303,34.526,-31.619],[46.689,49.435,-33.678],[45.991,46.967,-35.123],
[49.693,44.141,-25.153],[43.403,42.999,-35.665],[45.694,40.402,-40.045],
[52.144,40.667,-27.821],[47.069,43.080,-35.031],[47.096,40.606,-26.727],
[38.979,44.151,-32.910],[40.616,46.621,-31.041],[39.311,41.369,-30.044],
[45.111,47.629,-28.880],[39.694,45.919,-25.047],[46.036,42.667,-37.803],
[37.965,41.895,-25.858],[43.001,49.135,-33.660],[38.963,48.921,-23.992],
[51.458,35.401,-38.567],[39.372,43.089,-35.450],[50.840,48.140,-38.592],
[36.320,45.012,-28.126],[35.045,47.783,-29.638],[53.781,47.792,-24.911],
[43.342,48.393,-26.856],[43.233,51.702,-36.075],[37.393,50.300,-33.736],
[48.044,50.083,-38.591],[45.912,50.450,-36.172],[48.815,40.555,-24.693],
[36.164,48.854,-24.828],[40.576,36.376,-29.540],[44.824,33.822,-27.006],
[46.893,38.399,-25.093],[48.639,41.664,-38.955],[44.187,35.519,-39.819],
[40.698,38.967,-29.973],[40.917,39.786,-34.952],[43.667,41.253,-38.028],
[43.276,39.952,-23.757],[46.598,39.797,-32.417],[37.507,39.799,-27.612],
[40.418,38.932,-40.616],[34.141,36.902,-33.113],[36.021,39.138,-37.462],
[44.991,41.073,-28.608],[41.672,35.889,-34.061],[38.329,38.499,-34.450],
[40.596,37.084,-36.540],[46.053,36.073,-29.233],[40.365,41.027,-37.535],
[47.664,51.994,-23.949],[37.339,44.531,-25.089],[43.378,39.113,-30.823],
[37.678,37.002,-36.594],[42.321,41.323,-28.802],[42.034,45.958,-37.721],
[49.916,42.856,-29.696],[47.226,37.545,-38.812],[43.203,45.326,-26.525],
[47.525,40.763,-36.376],[38.677,43.885,-29.906],[44.419,44.887,-37.403],
[33.519,43.863,-25.008],[40.934,41.140,-24.516],[45.253,47.563,-25.105],
[49.845,39.333,-38.013],[52.079,46.379,-27.406],[50.423,48.254,-35.949],
[46.127,42.781,-30.412],[38.000,46.975,-34.091],[35.128,43.253,-39.442],
[41.467,45.190,-33.567],[40.174,50.537,-26.758],[40.204,49.717,-33.902],
[45.633,44.756,-28.206],[47.180,47.379,-40.077],[47.691,48.446,-27.815],
[51.503,52.036,-38.683],[36.523,37.300,-26.426],[47.320,45.209,-37.284],
[36.550,48.959,-27.683],[40.849,47.120,-40.982],[39.539,46.699,-38.059],
[37.780,44.440,-37.516],[45.012,37.285,-26.852],[47.182,35.292,-35.955],
[42.948,48.066,-36.172],[43.367,39.112,-36.010],[43.586,39.933,-26.569],
[45.286,37.388,-34.713],[41.033,49.107,-38.182],[39.127,37.260,-38.955],
[49.078,36.657,-26.175],[48.394,41.807,-31.996],[37.294,48.074,-38.720],
[38.356,39.932,-39.272],[49.239,49.573,-29.745],[37.518,34.142,-30.146],
[34.963,46.596,-34.312],[36.342,51.243,-23.554],[38.881,51.301,-22.453],
[43.223,48.341,-40.102],[47.873,47.626,-23.414],[49.590,48.685,-40.848],
[35.540,44.030,-22.875],[37.947,45.848,-40.383],[42.159,44.497,-24.074],
[45.275,44.861,-40.261],[47.131,44.344,-25.783],[37.303,41.738,-23.340],
[40.283,43.813,-38.160],[40.170,39.612,-27.520],[41.905,42.552,-40.072],
[44.924,43.796,-23.272],[49.222,44.882,-39.307],[40.723,38.028,-24.943],
[42.112,37.567,-27.534],[43.038,42.783,-26.780],[48.597,38.350,-20.704],
[47.279,40.137,-22.472],[39.431,40.533,-22.397],[40.582,34.524,-27.248],
[42.731,37.856,-22.364],[43.376,36.924,-24.850],[45.665,36.353,-23.630],
[35.910,36.968,-23.822],[40.306,34.561,-24.108],[42.209,34.403,-22.263],
[42.019,51.668,-38.673],[34.839,46.328,-38.571],[33.854,52.409,-33.671],
[34.919,49.817,-32.823],[53.363,51.852,-31.252],[50.197,49.765,-26.740],
[51.430,46.422,-24.459],[51.156,45.720,-35.801],[52.747,43.282,-34.196],
[50.113,46.038,-29.428],[49.073,46.528,-26.261],[51.171,43.645,-37.497],
[50.792,39.409,-40.589],[49.861,41.050,-34.334],[34.815,43.508,-30.505],
[48.694,42.713,-22.542],[52.091,41.374,-22.853],[49.610,36.595,-40.042],
[50.313,38.276,-35.556],[49.662,38.316,-29.593],[49.826,39.251,-27.209],
[34.048,39.108,-23.761],[51.773,34.746,-23.392],[51.070,34.607,-35.395],
[36.604,37.313,-31.806],[33.885,41.091,-30.327],[50.419,37.927,-24.015],
[52.513,37.387,-40.793],[35.106,36.643,-37.423],[48.936,51.223,-34.627],
[33.675,33.620,-27.443],[36.600,48.410,-22.105],[39.044,48.197,-26.633],
[42.772,49.764,-24.372],[45.217,52.233,-24.998],[50.209,34.165,-25.583],
[37.148,51.932,-28.255],[39.211,51.670,-31.032],[43.627,50.219,-29.073],
[45.640,49.922,-26.749],[47.974,51.304,-26.790],[36.458,49.888,-30.358],
[39.194,49.198,-29.690],[39.553,52.857,-33.705],[44.741,36.033,-31.994],
[49.054,35.220,-32.127],[51.240,34.177,-32.759],[37.554,35.675,-33.556],
[36.844,34.261,-36.147],[43.052,33.526,-33.452],[45.828,34.085,-33.916],
[49.006,33.822,-37.065],[35.053,49.131,-37.367],[36.576,37.153,-39.755],
[42.342,37.297,-38.517],[44.071,49.930,-38.121],[50.763,53.087,-35.937],
[36.209,34.406,-23.923],[38.398,36.155,-24.643],[36.968,50.378,-40.138],
[47.946,33.146,-21.563],[49.524,49.564,-23.863],[50.464,55.734,-28.851],
[46.384,58.405,-30.166],[39.721,56.472,-30.995],[47.986,55.658,-35.802],
[34.678,57.394,-32.439],[52.156,57.264,-27.408],[37.779,56.429,-28.542],
[48.317,56.365,-32.913],[36.869,58.376,-30.730],[49.322,53.252,-28.475],
[42.021,52.458,-31.578],[49.314,54.177,-38.296],[50.434,59.101,-22.540],
[40.525,54.144,-30.045],[45.323,53.219,-27.953],[47.901,57.005,-23.762],
[44.184,56.632,-39.880],[39.641,-59.778,-30.278],[39.779,59.550,-33.617],
[46.022,58.007,-33.430],[38.589,58.832,-27.608],[41.780,55.754,-39.333],
[33.178,55.517,-33.894],[36.017,57.533,-38.879],[41.561,54.936,-32.792],
[35.881,55.751,-34.286],[40.772,57.295,-34.958],[46.370,55.012,-30.369],
[39.031,58.878,-38.008],[43.632,59.383,-29.705],[38.847,55.551,-34.626],
[41.566,57.746,-28.925],[46.049,57.524,-37.301],[48.329,58.950,-38.174],
[52.010,57.816,-37.401],[35.988,54.924,-25.245],[42.852,55.470,-27.943],
[46.205,54.397,-37.401],[43.972,57.801,-35.250],[42.906,59.346,-26.751],
[43.767,54.204,-34.909],[40.539,54.147,-37.368],[49.323,56.067,-26.089],
[38.282,57.584,-40.300],[36.121,53.261,-30.449],[40.271,57.561,-25.928],
[40.124,55.497,-27.704],[44.103,56.766,-21.596],[38.708,58.910,-24.090],
[41.160,53.963,-25.668],[43.261,56.928,-25.241],[45.381,58.067,-23.763],
[47.224,54.915,-21.906],[34.964,55.702,-22.585],[40.842,51.634,-24.079],
[41.447,54.126,-22.843],[51.270,59.284,-39.489],[49.283,56.531,-40.127],
[48.799,58.300,-35.469],[47.573,55.912,-28.207],[48.845,58.501,-28.064],
[35.894,57.638,-24.454],[50.240,52.351,-22.371],[49.298,53.849,-33.604],
[35.584,55.763,-30.259],[34.192,58.740,-30.010],[51.818,53.757,-24.380],
[51.578,55.612,-38.716],[33.324,55.813,-37.004],[33.558,54.630,-28.880],
[49.432,53.189,-25.311],[43.860,56.264,-32.509],[48.186,51.363,-31.942],
[51.776,54.090,-31.662],[36.985,52.723,-32.827],[36.998,51.367,-36.246],
[40.194,51.696,-36.450],[45.634,52.655,-31.834],[46.903,52.666,-35.151],
[38.790,55.929,-38.334],[42.835,55.842,-36.838],[38.449,53.724,-24.979],
[39.378,54.849,-21.152],[45.954,50.487,-22.746],[45.466,-49.863,-15.269],
[45.455,-50.379,-12.528],[35.941,-51.376,-11.138],[48.416,-46.119,-14.323],
[37.688,-47.165,-13.899],[33.446,-52.989,-16.051],[34.851,-50.459,-14.989],
[47.558,-43.238,-15.082],[43.714,-47.047,-10.700],[50.399,-57.182,-10.872],
[46.397,-54.216,-10.874],[39.150,-54.351,-14.175],[45.361,-54.727,-15.972],
[36.489,-52.194,-13.486],[41.603,-49.081,-12.629],[53.006,-55.186,-7.949],
[36.867,-56.163,-9.869],[43.361,-45.675,-15.927],[41.852,-48.613,-9.352],
[47.915,-56.668,-15.667],[37.821,-53.584,-10.909],[36.441,-47.170,-8.252],
[33.712,-45.664,-17.235],[48.855,-47.214,-10.467],[48.219,-56.275,-8.198],
[40.902,58.847,-12.502],[47.356,-44.358,-12.414],[46.662,-44.949,-17.088],
[49.385,-48.212,-6.719],[42.729,-51.050,-17.012],[43.802,-49.895,-21.227],
[51.567,-49.314,-8.155],[48.239,-50.786,-15.532],[45.840,-52.761,-8.347],
[39.182,-49.537,-14.893],[40.525,-45.902,-12.061],[39.731,-50.660,-10.512],
[47.203,-44.484,-9.712],[39.685,-47.143,-7.878],[45.860,-51.705,-20.478],
[38.154,-51.039,-6.823],[43.338,-42.149,-15.515],[38.428,-45.388,-4.773],
[49.158,-58.369,-19.053],[37.155,-50.355,-16.503],[50.601,-43.074,-22.488],
[33.911,-47.088,-8.644],[34.413,-44.170,-12.216],[51.759,-45.407,-7.331],
[42.642,-45.276,-5.144],[44.439,-40.797,-18.463],[36.344,-42.753,-14.418],
[48.662,-42.098,-20.701],[45.854,-42.510,-17.355],[49.724,-52.789,-5.095],
[37.695,-45.021,-7.143],[39.297,-57.620,-10.440],[44.033,-58.932,-7.977],
[47.380,-53.795,-6.356],[48.186,-50.093,-22.528],[44.399,-56.127,-20.717],
[40.579,-53.507,-10.504],[39.396,-52.200,-15.870],[46.128,-51.153,-17.504],
[44.673,-52.421,-4.993],[46.636,-52.803,-14.681],[37.019,-53.745,-8.275],
[40.060,-56.424,-21.928],[34.354,-56.063,-14.170],[34.003,-56.118,-19.920],
[44.712,-51.044,-10.048],[41.816,-55.261,-14.653],[36.918,-53.825,-16.031],
[39.523,-54.426,-17.606],[46.826,-57.049,-10.409],[38.712,-53.970,-20.244],
[48.331,-41.618,-3.192],[37.527,-48.187,-5.944],[43.291,-52.750,-12.093],
[38.077,-56.359,-15.583],[43.402,-54.194,-9.810],[43.265,-46.539,-18.617],
[49.507,-49.829,-11.999],[46.831,-56.789,-19.753],[44.205,-47.779,-7.196],
[46.283,-54.086,-19.203],[37.685,-49.737,-12.616],[46.004,-47.417,-16.331],
[35.367,-49.669,-7.596],[41.995,-51.437,-5.594],[45.768,-44.922,-7.228],
[51.653,-54.588,-18.561],[52.455,-47.794,-5.167],[50.424,-45.359,-17.734],
[47.347,-48.169,-12.944],[39.518,-45.219,-14.908],[37.148,-50.944,-19.841],
[41.641,-47.971,-16.789],[40.996,-44.386,-8.140],[40.013,-42.459,-17.043],
[46.420,-47.142,-9.209],[46.510,-49.038,-20.589],[48.072,-44.046,-5.758],
[52.875,-42.607,-21.233],[36.850,-56.227,-7.087],[47.528,-47.241,-18.929],
[35.559,-45.295,-9.881],[42.318,-45.943,-23.481],[40.346,-46.897,-18.985],
[36.847,-48.243,-19.214],[44.121,-56.024,-8.124],[45.078,-56.212,-17.927],
[43.192,-44.853,-21.133],[43.001,-53.739,-17.686],[43.341,-53.621,-7.071],
[44.144,-57.212,-15.704],[41.593,-43.068,-19.111],[39.578,-57.166,-19.328],
[50.405,-54.598,-6.856],[47.959,-51.894,-12.066],[36.408,-45.462,-19.844],
[35.615,-54.045,-21.497],[49.204,-42.520,-10.249],[37.923,-59.150,-12.240],
[36.717,-45.439,-15.638],[36.721,-42.396,-6.228],[38.200,-42.591,-3.514],
[45.481,-46.083,-23.087],[47.218,-46.692,-5.543],[48.849,-44.486,-19.737],
[35.822,-47.809,-3.426],[39.182,-47.841,-21.670],[40.447,-47.292,-5.027],
[41.743,-47.376,-21.194],[46.647,-49.969,-7.762],[35.709,-52.411,-5.645],
[39.318,-49.863,-17.745],[39.884,-52.523,-8.001],[40.959,-52.891,-21.468],
[44.879,-49.591,-5.065],[50.517,-48.569,-22.178],[39.303,-54.627,-6.498],
[39.237,-55.444,-8.829],[42.478,-51.624,-8.642],[46.935,-53.456,-3.627],
[47.607,-51.092,-5.430],[38.022,-52.974,-4.320],[41.192,-57.978,-8.460],
[41.982,-54.475,-4.593],[45.220,-55.002,-5.578],[46.665,-56.484,-3.710],
[38.041,-56.035,-4.480],[43.254,58.663,-4.997],[41.348,-57.844,-5.346],
[39.712,-41.100,-19.599],[34.078,-46.735,-21.413],[35.064,-41.232,-16.725],
[33.814,-43.284,-14.868],[51.628,-42.249,-11.529],[50.057,-43.199,-7.908],
[50.009,-47.795,-3.668],[49.423,-49.130,-17.947],[51.155,-50.230,-15.933],
[51.447,-48.112,-10.850],[48.922,-45.691,-7.711],[51.593,-49.550,-19.752],
[50.714,-52.522,-21.100],[50.119,-52.453,-14.405],[34.312,-49.642,-12.470],
[50.861,-50.158,-5.472],[52.595,-52.401,-5.841],[48.827,-56.648,-21.521],
[48.875,-54.564,-17.673],[49.299,-54.494,-10.256],[51.058,-52.123,-8.816],
[32.548,-53.398,-5.721],[54.195,-59.667,-3.711],[50.354,-57.714,-16.464],
[35.596,-58.210,-10.992],[33.068,-52.601,-11.208],[51.416,-56.982,-5.734],
[51.797,-58.212,-20.367],[34.316,-57.306,-16.559],[48.343,-41.040,-16.247],
[34.426,-57.674,-8.667],[35.058,-45.552,-4.742],[39.179,-42.306,-7.861],
[44.356,-42.867,-6.291],[46.772,-41.406,-5.902],[51.286,-59.308,-7.558],
[37.589,-40.265,-7.641],[38.644,-40.284,-10.327],[44.264,-40.710,-10.770],
[44.983,-42.785,-8.947],[46.249,-40.030,-8.348],[37.922,-43.888,-9.643],
[37.683,-44.564,-13.036],[40.231,-43.291,-10.872],[44.222,-56.173,-12.226],
[48.432,-59.093,-11.556],[50.360,-59.462,-14.382],[35.681,-58.684,-13.954],
[35.499,58.919,-17.340],[41.700,-58.792,-14.952],[47.499,-59.159,-14.075],
[46.555,-59.100,-18.507],[34.880,-43.465,-18.367],[37.110,-55.952,-19.528],
[43.605,-58.501,-19.304],[43.289,-42.081,-20.886],[52.117,-42.128,-18.123],
[37.677,-58.892,-5.849],[38.217,-59.329,-2.920],[37.501,-41.210,-21.314],
[49.127,-57.657,-4.282],[50.539,-45.066,-4.699],[45.992,-32.906,-12.858],
[44.419,-31.043,-11.450],[36.625,-31.272,-10.666],[49.183,-28.072,-14.388],
[39.201,-30.746,-11.107],[35.347,-33.752,-16.349],[34.105,-30.176,-14.957],
[45.939,-25.384,-12.301],[43.554,-29.364,-13.364],[51.108,-37.849,-11.683],
[46.885,-35.864,-10.375],[39.545,-35.990,-13.556],[46.663,-37.245,-15.148],
[34.807,-36.130,-12.807],[41.132,-32.785,-15.333],[50.655,-35.520,-8.692],
[37.121,-37.742,-9.480],[44.724,-27.742,-15.219],[41.286,-30.216,-8.864],
[48.803,-37.628,-12.913],[38.370,-33.890,-10.723],[38.759,-27.622,-9.807],
[35.863,-26.842,-18.662],[48.747,-28.415,-11.367],[49.636,-37.947,-9.474],
[43.067,-40.657,-13.283],[47.444,-26.409,-14.078],[48.167,-26.198,-16.651],
[51.169,-29.063,-7.760],[44.775,-30.361,-16.535],[45.366,-32.857,-21.449],
[51.954,-31.686,-6.108],[48.114,-32.860,-17.597],[47.501,-34.064,-7.865],
[38.196,-30.715,-15.702],[39.576,-27.598,-12.482],[39.934,-30.331,-13.787],
[45.317,-26.353,-9.668],[39.964,-27.390,-7.128],[46.089,-31.403,-19.328],
[38.004,-32.281,-7.465],[44.458,-25.071,-15.820],[39.020,-24.743,-6.576],
[50.558,-40.148,-17.356],[37.514,-31.808,-17.976],[52.788,-26.690,-20.291],
[34.084,-28.469,-6.908],[34.490,-24.759,-12.966],[52.402,-26.208,-5.623],
[43.763,-25.915,-6.506],[44.380,-22.410,-16.305],[38.307,-23.128,-14.883],
[46.344,-23.372,-20.561],[47.300,-23.307,-16.822],[50.033,-33.773,-6.455],
[35.896,-28.093,-9.016],[39.690,-39.690,-12.664],[43.133,-40.438,-7.197],
[48.730,-36.325,-5.587],[48.391,-31.942,-20.746],[42.046,-37.530,-20.256],
[41.164,-32.512,-12.417],[39.896,-33.442,-17.615],[45.295,-33.050,-17.400],
[45.285,-32.997,-6.881],[44.931,-34.681,-14.613],[38.116,-35.008,-7.673],
[39.549,-36.041,-21.030],[34.533,-38.764,-13.789],[35.779,-36.180,-18.227],
[45.723,-32.880,-9.823],[41.575,-38.293,-14.251],[36.868,-35.517,-14.837],
[41.230,-36.891,-17.007],[46.060,-38.494,-11.533],[38.908,-33.343,-20.140],
[49.538,-22.457,-4.563],[37.953,-30.494,-5.408],[42.427,-34.534,-13.344],
[38.902,-35.581,-16.408],[42.959,-33.538,-10.329],[42.875,-29.873,-18.661],
[48.020,-31.312,-8.875],[45.093,-37.409,-19.897],[45.036,-28.162,-7.074],
[46.618,-34.544,-19.750],[37.191,-29.329,-13.051],[45.513,-28.753,-19.231],
[35.667,-30.984,-7.802],[42.583,-33.173,-4.895],[47.213,-26.534,-7.275],
[50.590,-34.140,-18.606],[53.120,-28.810,-5.338],[50.651,-26.911,-17.651],
[48.392,-31.305,-13.460],[40.171,-25.251,-14.587],[35.582,-32.719,-19.980],
[40.072,-28.421,-15.965],[42.101,-23.985,-7.418],[39.793,-23.563,-18.082],
[46.709,-29.030,-9.792],[48.076,-27.816,-18.986],[48.826,-23.649,-7.348],
[51.284,-23.562,-20.034],[35.167,-38.945,-7.851],[48.725,-30.457,-18.494],
[36.141,-25.467,-9.552],[42.051,-27.120,-23.598],[40.789,-27.976,-18.666],
[36.743,-30.291,-20.440],[44.510,-37.887,-7.620],[43.414,-38.462,-17.827],
[43.472,-27.033,-17.963],[42.943,-34.760,-16.741],[43.234,-35.229,-7.913],
[44.005,-38.603,-14.994],[41.523,-25.352,-19.333],[39.832,-38.333,-19.265],
[50.005,-38.606,-6.727],[48.219,-32.822,-11.362],[38.514,-26.159,-19.334],
[36.932,-35.192,-21.129],[48.859,-24.581,-10.677],[37.305,-41.167,-12.576],
[38.515,-26.092,-16.388],[34.316,-23.873,-5.493],[38.774,-24.496,-3.783],
[43.220,-25.262,-21.663],[46.208,-27.040,-3.767],[47.964,-26.797,-21.615],
[34.964,-30.135,-4.926],[38.627,-29.100,-22.911],[41.377,-29.342,-5.052],
[43.175,-27.982,-20.834],[47.526,-29.168,-7.235],[35.655,-33.005,-5.329],
[40.767,-31.023,-19.724],[40.796,-33.821,-8.308],[42.425,-32.288,-21.903],
[44.399,-30.667,-5.839],[50.131,-29.702,-20.693],[40.417,-35.160,-5.570],
[40.786,-36.556,-8.233],[43.025,-32.198,-8.011],[47.188,-34.379,-4.183],
[47.528,-31.792,-5.465],[38.238,-33.232,-4.363],[41.528,-38.684,-10.140],
[43.385,-35.863,-3.753],[45.584,-35.577,-6.030],[46.322,-36.816,-3.709],
[35.825,-35.817,-4.799],[41.058,-41.723,-5.720],[41.499,-39.513,-4.029],
[42.145,-22.700,-20.102],[36.095,-27.947,-21.739],[33.711,-22.334,-16.634],
[35.281,-22.464,-14.145],[52.257,-23.832,-10.588],[51.437,-24.070,-7.671],
[50.746,-27.527,-3.954],[51.957,-29.839,-17.663],[51.686,-31.143,-14.493],
[51.170,-28.140,-10.438],[49.641,-26.206,-8.331],[52.375,-31.092,-19.849],
[50.955,-35.401,-21.988],[48.593,-34.025,-14.936],[34.657,-30.980,-12.529],
[49.638,-30.202,-5.773],[50.409,-32.376,-3.918],[49.263,-37.484,-22.017],
[48.253,-37.837,-19.206],[49.786,-35.210,-11.457],[50.508,-32.991,-9.162],
[33.742,-34.652,-6.150],[51.238,-40.909,-3.842],[48.715,-37.901,-16.622],
[36.693,-38.611,-12.070],[34.282,-33.571,-11.519],[52.850,-38.669,-4.841],
[50.889,-39.461,-20.173],[33.978,-38.425,-16.609],[48.956,-24.126,-14.961],
[36.013,-40.680,-9.751],[36.846,-25.817,-5.620],[42.039,-25.988,-9.159],
[44.136,-23.222,-5.003],[46.717,-22.084,-7.410],[50.912,-40.822,-7.942],
[37.432,-20.999,-7.528],[38.895,-23.067,-9.935],[42.788,-21.633,-10.467],
[45.386,-23.439,-9.670],[48.267,-21.749,-9.673],[36.883,-24.064,-11.988],
[36.419,-26.667,-12.338],[41.632,-24.820,-12.237],[44.006,-36.978,-12.820],
[49.227,-41.270,-13.216],[50.463,-38.906,-14.560],[37.733,-37.889,-14.275],
[34.713,-39.315,-18.851],[40.895,-41.013,-15.051],[45.970,-40.944,-13.592],
[46.948,-40.258,-19.276],[35.099,-24.877,-16.120],[37.118,-37.707,-19.723],
[41.145,-39.755,-17.433],[46.017,-24.693,-22.960],[50.863,-23.539,-16.971],
[37.006,-38.294,-5.752],[38.981,-39.614,-2.881],[37.244,-22.185,-20.753],
[46.342,-39.517,-4.078],[49.579,-25.376,-5.350],[45.918,-10.629,-14.127],
[44.212,-12.054,-12.058],[36.645,-11.966,-10.602],[49.790,-8.412,-12.871],
[38.421,-9.650,-12.935],[33.349,-15.408,-16.061],[35.160,-12.749,-15.457],
[46.496,-6.024,-15.019],[42.878,-9.139,-11.642],[51.829,-17.279,-10.041],
[47.259,-15.705,-10.370],[39.005,-15.213,-14.663],[47.702,-16.671,-14.851],
[35.633,-14.929,-13.474],[40.616,-13.417,-16.582],[52.739,-16.537,-7.420],
[37.743,-17.491,-10.117],[44.539,-9.875,-16.397],[40.067,-10.850,-10.487],
[49.532,-19.508,-13.885],[38.091,-16.164,-12.360],[37.900,-9.190,-10.265],
[34.142,-8.720,-17.603],[47.307,-8.425,-8.896],[48.492,-17.694,-8.636],
[41.894,-21.582,-12.981],[46.873,-8.254,-12.463],[47.802,-7.822,-17.556],
[49.276,-10.847,-6.945],[45.332,-12.189,-18.037],[43.593,-14.311,-19.115],
[52.259,-14.356,-8.821],[47.512,-13.211,-16.415],[46.015,-16.762,-8.409],
[38.779,-12.412,-14.677],[40.416,-7.992,-11.036],[39.447,-13.349,-11.811],
[44.967,-7.063,-10.464],[41.403,-10.847,-7.953],[46.906,-14.196,-20.002],
[38.090,-13.711,-7.642],[44.037,-6.359,-16.210],[39.004,-6.595,-5.668],
[48.733,-22.325,-18.875],[37.807,-13.764,-17.530],[52.751,-6.652,-20.684],
[35.770,-10.217,-8.414],[33.292,-7.322,-11.686],[53.346,-7.024,-6.264],
[41.776,-8.584,-6.747],[44.492,-4.371,-18.213],[36.885,-5.424,-14.817],
[48.497,-4.449,-21.097],[46.994,-5.035,-18.582],[49.807,-15.586,-4.528],
[37.537,-8.505,-7.026],[38.651,-20.220,-10.041],[44.557,-20.411,-7.199],
[48.322,-17.505,-6.051],[49.124,-13.282,-21.545],[43.436,-18.104,-20.614],
[39.058,-15.142,-9.591],[40.636,-16.117,-16.730],[45.146,-15.182,-15.494],
[44.496,-15.900,-3.064],[44.952,-16.138,-12.992],[36.246,-16.271,-8.386],
[40.245,-18.788,-21.839],[34.243,-17.605,-13.185],[35.782,-17.928,-19.646],
[45.280,-13.955,-10.396],[42.079,-18.979,-13.043],[36.420,-15.430,-15.860],
[41.127,-17.389,-19.065],[47.055,-19.110,-12.128],[39.610,-16.043,-21.697],
[49.654,-3.323,-5.766],[37.870,-11.041,-6.206],[42.703,-14.412,-12.218],
[37.998,-17.483,-16.069],[41.727,-15.869,-10.048],[41.223,-10.359,-18.567],
[47.489,-11.094,-9.352],[46.312,-18.261,-20.390],[44.299,-11.488,-7.845],
[46.417,-16.177,-18.051],[36.451,-11.393,-13.394],[44.234,-10.224,-19.422],
[34.479,-12.453,-6.251],[42.315,-13.710,-5.679],[45.760,-6.977,-6.585],
[52.639,-15.240,-19.064],[50.869,-8.740,-5.632],[50.795,-8.394,-18.682],
[46.640,-12.910,-12.609],[41.311,-7.755,-16.572],[35.174,-12.382,-20.421],
[41.666,-10.425,-15.219],[40.761,-5.637,-9.593],[39.827,-3.533,-14.786],
[45.385,-10.015,-10.712],[46.564,-10.934,-20.419],[48.549,-5.745,-5.999],
[50.530,-4.340,-19.485],[36.123,-18.411,-6.664],[48.630,-11.703,-18.942],
[34.163,-8.274,-9.424],[41.351,-8.534,-21.715],[39.971,-7.856,-18.620],
[36.838,-10.211,-18.326],[44.189,-18.589,-8.958],[45.200,-19.337,-16.593],
[42.479,-7.527,-19.013],[43.709,-16.821,-17.303],[43.118,-15.706,-7.515],
[43.686,-17.727,-14.717],[40.194,-5.722,-20.311],[40.628,-20.437,-19.372],
[50.714,-18.818,-7.011],[48.578,-14.887,-12.758],[37.783,-7.002,-19.849],
[36.656,-14.823,-22.130],[48.009,-5.725,-9.582],[37.115,-21.367,-12.380],
[38.087,-6.162,-17.152],[34.386,-4.685,-5.714],[39.562,-4.276,-4.428],
[45.200,-6.544,-20.361],[47.207,-8.602,-4.921],[48.448,-6.939,-22.003],
[35.778,-11.144,-3.780],[39.853,-11.383,-22.379],[41.870,-11.114,-5.327],
[44.591,-10.760,-23.077],[46.876,-12.204,-7.070],[35.748,-14.831,-5.758],
[39.569,-14.497,-19.377],[40.375,-15.160,-7.306],[41.264,-13.607,-21.495],
[46.384,-11.372,-4.339],[49.574,-10.628,-23.008],[41.142,-17.593,-6.651],
[40.251,-18.054,-9.167],[42.668,-13.238,-9.229],[46.980,-16.183,-3.655],
[47.494,-14.837,-5.790],[39.714,-15.116,-4.690],[41.958,-20.509,-8.140],
[42.399,-17.133,-4.032],[45.065,-16.773,-5.898],[46.530,-19.177,-4.389],
[36.771,-16.774,-4.484],[41.578,-22.090,-5.399],[41.412,-19.819,-3.952],
[42.529,-3.366,-19.432],[35.652,-8.422,-20.195],[33.102,-3.651,-17.608],
[34.583,-5.725,-13.441],[49.310,-5.594,-12.002],[52.749,-4.978,-8.234],
[50.460,-8.894,-2.561],[51.524,-10.947,-17.965],[51.568,-12.476,-14.178],
[49.435,-10.380,-11.411],[49.660,-7.522,-8.143],[50.376,-13.723,-18.548],
[50.617,-15.539,-20.890],[49.901,-14.371,-15.622],[33.969,-12.000,-12.820],
[50.066,-13.685,-6.527],[52.710,-14.888,-5.449],[49.573,-18.291,-21.587],
[50.169,-17.938,-16.992],[49.255,-17.789,-11.349],[49.924,-15.337,-9.659],
[33.844,-16.735,-6.249],[53.978,-21.590,-2.351],[49.006,-20.738,-16.671],
[36.880,-18.589,-13.473],[33.892,-14.812,-11.267],[53.137,-19.905,-4.913],
[51.488,-20.142,-20.618],[35.623,-19.743,-16.018],[48.983,-4.857,-16.834],
[35.500,-19.467,-9.173],[36.391,-7.366,-4.607],[38.495,-6.955,-9.116],
[43.862,-6.279,-4.631],[45.261,-4.258,-6.117],[50.957,-21.463,-6.517],
[36.154,-3.720,-7.220],[37.067,-3.950,-9.758],[43.214,-4.378,-11.386],
[45.864,-4.745,-8.669],[48.064,-2.311,-9.155],[36.920,-6.186,-11.413],
[38.741,-6.909,-13.605],[40.712,-5.522,-12.423],[44.468,-19.902,-11.601],
[48.262,-21.838,-13.766],[51.343,-21.792,-14.147],[38.107,-20.508,-14.939],
[35.297,-21.489,-18.766],[41.558,-20.777,-16.040],[46.309,-21.827,-11.679],
[46.713,-20.590,-18.871],[34.540,-6.131,-18.297],[37.930,-19.743,-19.807],
[42.700,-19.607,-18.171],[43.714,-4.075,-21.584],[51.379,-6.138,-17.532],
[37.506,-22.019,-4.952],[40.168,-22.066,-2.940],[36.552,-4.660,-20.676],
[46.814,-22.168,-3.860],[51.095,-6.000,-4.756],[46.723,6.173,-14.829],
[43.843,5.226,-12.989],[37.565,6.207,-11.562],[47.721,9.123,-13.502],
[37.863,9.249,-11.908],[33.136,3.543,-16.373],[34.190,3.756,-13.710],
[46.067,14.349,-13.189],[43.672,10.627,-11.333],[51.930,-0.291,-9.779],
[46.555,2.202,-10.230],[38.822,-0.176,-13.134],[45.673,0.866,-16.757],
[36.243,2.333,-14.467],[39.865,3.259,-13.802],[51.990,2.290,-9.201],
[37.603,0.540,-10.352],[43.789,8.771,-16.427],[39.547,8.019,-9.981],
[48.961,-0.025,-14.492],[37.245,2.790,-12.184],[37.708,9.236,-7.752],
[34.921,9.928,-17.927],[48.552,8.967,-10.812],[48.521,-0.276,-10.811],
[41.344,-2.595,-12.252],[47.016,11.508,-15.333],[46.461,8.433,-16.512],
[49.387,9.193,-8.338],[44.068,6.241,-18.289],[45.300,4.190,-21.453],
[50.960,4.788,-9.248],[48.370,6.412,-17.345],[46.661,2.953,-7.537],
[39.129,6.155,-14.618],[40.847,10.049,-11.799],[41.286,4.682,-11.631],
[46.254,10.618,-12.064],[40.482,11.463,-7.158],[46.752,4.839,-19.173],
[38.334,4.642,-7.512],[44.398,11.524,-16.433],[38.716,11.790,-4.787],
[52.101,-1.968,-19.552],[37.270,6.086,-16.736],[51.473,11.789,-21.556],
[34.415,5.869,-9.308],[35.318,9.727,-12.552],[53.267,11.456,-6.000],
[43.646,11.018,-7.631],[45.024,15.386,-16.942],[39.324,13.081,-16.333],
[46.286,12.543,-20.504],[47.297,13.358,-17.552],[49.545,2.943,-6.429],
[36.093,6.869,-7.467],[39.093,-2.307,-10.787],[45.007,-2.178,-7.888],
[47.161,0.039,-8.133],[47.535,5.873,-22.047],[42.125,-0.781,-20.384],
[39.994,1.421,-11.266],[39.483,3.565,-16.401],[45.150,3.634,-17.324],
[44.852,3.441,-5.508],[45.392,3.791,-14.776],[37.245,2.250,-7.547],
[39.453,1.318,-20.788],[34.337,-1.122,-16.006],[36.365,1.063,-19.417],
[46.419,5.090,-9.359],[41.891,-0.674,-14.902],[35.965,3.698,-16.735],
[39.032,0.715,-18.117],[45.279,0.074,-12.075],[39.073,3.665,-19.192],
[48.630,14.913,-3.513],[37.943,7.337,-5.235],[44.564,2.583,-12.160],
[39.491,-1.162,-16.006],[42.581,1.423,-10.583],[41.893,8.237,-18.486],
[50.532,6.830,-11.125],[45.580,0.596,-21.636],[44.925,8.755,-6.960],
[46.155,2.105,-19.350],[36.562,7.034,-14.073],[45.727,10.195,-18.626],
[33.629,5.994,-6.361],[43.121,6.021,-4.379],[45.986,12.107,-7.367],
[52.054,2.978,-17.984],[50.909,9.642,-6.044],[50.983,11.273,-17.991],
[46.532,5.639,-12.193],[40.729,10.256,-16.967],[37.864,5.943,-20.403],
[41.006,7.460,-16.071],[41.738,11.460,-9.592],[41.905,12.900,-16.990],
[45.907,9.526,-9.682],[48.226,8.555,-21.767],[48.355,13.820,-6.096],
[52.817,14.947,-19.230],[35.531,0.088,-8.747],[48.036,8.946,-18.396],
[34.648,8.973,-8.738],[41.684,11.000,-21.259],[38.960,8.235,-17.825],
[36.513,8.064,-18.914],[43.138,-0.506,-8.241],[45.039,-0.504,-19.116],
[43.177,11.169,-18.922],[43.780,1.218,-14.558],[42.837,2.061,-7.175],
[44.164,-1.238,-16.119],[39.876,11.926,-19.024],[39.970,-2.206,-19.453],
[50.194,0.109,-6.896],[49.183,2.842,-11.588],[38.536,10.590,-20.935],
[35.050,1.982,-23.273],[48.459,11.851,-10.996],[33.686,-1.336,-11.856],
[37.537,9.730,-15.333],[34.463,12.167,-3.320],[41.418,13.434,-3.705],
[44.527,10.547,-21.371],[45.787,11.175,-3.917],[48.745,11.411,-20.438],
[35.429,7.351,-4.469],[39.815,8.081,-21.069],[40.781,7.661,-5.157],
[42.859,8.307,-21.109],[46.982,7.235,-7.722],[36.383,4.992,-5.491],
[40.787,6.002,-20.101],[42.052,5.776,-7.079],[42.580,3.713,-21.164],
[45.446,5.923,-5.969],[50.277,6.879,-22.516],[40.407,0.934,-6.704],
[40.587,-1.691,-8.447],[42.888,4.277,-9.195],[46.522,2.535,-2.715],
[47.437,4.180,-5.224],[39.810,2.891,-2.825],[42.639,-3.482,-8.539],
[42.039,2.490,-4.356],[45.010,0.778,-6.442],[46.896,0.247,-4.834],
[37.374,2.466,-4.940],[41.541,-3.547,-6.182],[40.787,-1.328,-4.076],
[42.173,15.572,-19.154],[35.135,9.380,-20.715],[35.187,14.499,-15.682],
[35.456,11.915,-14.055],[53.798,13.305,-12.466],[51.434,13.343,-9.691],
[49.216,8.669,-3.523],[50.357,8.252,-20.007],[50.912,6.791,-16.311],
[52.298,9.223,-10.521],[49.287,12.208,-8.194],[51.913,5.527,-20.499],
[50.794,2.527,-20.310],[49.529,4.371,-16.002],[34.078,5.985,-11.760],
[50.073,6.369,-7.048],[51.720,4.865,-4.896],[48.406,1.715,-21.133],
[48.673,1.216,-18.011],[53.683,0.679,-11.555],[49.558,1.502,-9.001],
[34.320,0.399,-4.602],[50.625,-2.669,-2.980],[50.784,-2.885,-17.388],
[35.099,0.877,-11.754],[34.241,3.303,-10.972],[51.963,-1.587,-5.184],
[51.887,0.016,-21.544],[34.020,-1.181,-18.725],[48.484,16.328,-15.943],
[34.436,-1.708,-6.566],[36.429,9.858,-4.462],[39.131,11.841,-9.523],
[43.080,13.216,-5.863],[45.768,14.690,-6.324],[50.308,-3.703,-9.295],
[36.314,16.586,-9.400],[37.664,14.497,-7.818],[42.054,15.806,-10.310],
[46.615,13.558,-9.820],[47.706,15.896,-8.329],[37.115,12.295,-11.280],
[38.991,12.312,-13.249],[41.978,12.473,-12.532],[43.882,-1.837,-13.070],
[47.600,-3.745,-13.477],[50.723,-2.521,-14.148],[35.678,-3.046,-13.404],
[35.674,-3.403,-17.745],[41.467,-3.681,-16.901],[45.304,-3.558,-14.897],
[48.207,-2.148,-18.599],[37.191,13.132,-18.131],[37.457,-2.084,-20.330],
[41.700,-0.330,-17.604],[45.647,15.078,-19.605],[49.489,13.641,-19.126],
[37.414,-0.183,-6.201],[37.839,-1.684,-3.916],[37.084,13.179,-21.149],
[47.078,-2.510,-4.883],[50.347,12.179,-5.857],[46.336,22.957,-14.828],
[45.139,24.161,-12.285],[37.502,23.910,-10.506],[48.729,26.420,-14.839],
[38.439,28.262,-11.031],[34.105,20.565,-15.691],[34.694,24.142,-15.393],
[45.869,32.050,-12.888],[44.756,26.442,-13.801],[52.086,18.850,-8.763],
[47.301,21.306,-9.881],[39.958,19.187,-13.062],[46.797,18.633,-16.213],
[35.771,21.536,-13.792],[40.081,21.905,-13.715],[53.075,21.206,-7.791],
[38.125,18.409,-11.241],[43.777,28.243,-15.567],[39.926,26.234,-10.192],
[49.171,18.686,-13.634],[37.052,20.878,-11.608],[37.873,26.117,-8.387],
[36.899,28.178,-17.179],[48.298,27.918,-11.380],[49.037,18.697,-8.892],
[41.959,15.243,-13.168],[47.442,30.115,-14.269],[47.785,28.731,-16.939],
[49.346,26.695,-8.857],[43.921,25.741,-16.562],[44.212,24.828,-21.478],
[52.289,22.943,-9.813],[47.372,24.786,-16.686],[47.605,21.730,-7.027],
[39.422,26.168,-14.874],[41.286,26.736,-13.012],[40.423,23.944,-12.058],
[45.826,28.981,-10.576],[40.375,29.023,-5.727],[45.640,24.354,-18.795],
[38.136,23.700,-7.016],[44.134,31.166,-15.609],[37.414,29.476,-6.589],
[50.446,15.995,-19.571],[38.924,25.005,-17.464],[52.380,29.841,-20.437],
[32.849,26.109,-8.840],[35.669,27.550,-10.363],[53.526,28.431,-4.848],
[42.257,29.615,-7.836],[44.337,32.781,-17.783],[38.113,31.066,-16.600],
[46.091,31.286,-20.180],[46.772,32.038,-17.656],[49.458,21.606,-4.896],
[35.522,27.926,-7.600],[39.506,15.941,-12.060],[43.453,15.491,-7.559],
[46.830,18.875,-7.046],[48.098,24.100,-20.997],[43.848,18.107,-22.264],
[40.360,21.354,-11.031],[40.177,21.002,-16.175],[44.522,22.931,-16.857],
[43.785,23.060,-4.877],[47.119,20.464,-14.191],[38.287,20.577,-9.213],
[41.053,18.783,-22.319],[33.788,17.921,-14.169],[35.014,19.363,-18.321],
[44.248,23.082,-9.798],[42.571,19.558,-14.898],[36.750,21.398,-16.305],
[40.743,22.285,-18.776],[45.809,18.468,-10.932],[39.055,21.862,-21.046],
[50.469,34.141,-5.165],[35.871,27.073,-4.774],[43.088,22.296,-12.108],
[38.102,18.902,-16.466],[42.744,19.628,-11.331],[42.696,25.113,-19.000],
[49.994,24.956,-11.060],[46.333,19.623,-20.629],[44.144,27.543,-7.050],
[46.693,22.111,-19.694],[36.854,24.938,-13.889],[45.246,27.837,-19.398],
[34.869,24.843,-7.148],[40.789,23.069,-3.811],[44.790,30.173,-8.045],
[52.821,23.004,-18.507],[52.803,26.022,-6.417],[50.222,28.316,-19.611],
[47.395,24.281,-10.477],[39.968,28.882,-15.102],[36.117,24.037,-21.033],
[41.260,27.129,-17.471],[41.724,30.432,-10.917],[40.414,31.638,-15.191],
[45.575,26.343,-10.214],[46.992,26.516,-21.220],[47.413,30.722,-6.989],
[53.118,32.431,-20.943],[34.144,18.707,-8.095],[48.486,26.584,-18.402],
[36.975,30.224,-9.233],[39.889,28.334,-22.486],[41.411,27.287,-20.456],
[37.654,25.910,-19.567],[43.670,18.144,-8.914],[44.970,17.957,-18.497],
[42.909,29.271,-19.352],[43.211,20.720,-17.775],[43.361,22.114,-7.470],
[44.542,17.694,-14.968],[40.361,29.718,-19.032],[40.111,17.715,-20.048],
[50.327,19.657,-6.628],[48.340,22.716,-12.597],[38.057,28.746,-20.293],
[36.762,21.156,-20.046],[48.376,32.171,-10.611],[36.462,15.066,-12.680],
[35.436,29.072,-15.224],[34.157,31.992,-5.038],[40.022,31.673,-2.488],
[43.690,29.659,-21.916],[45.947,28.444,-5.354],[48.928,30.542,-21.628],
[34.540,24.548,-4.474],[38.491,25.366,-21.980],[40.085,26.020,-4.817],
[44.181,27.041,-22.843],[46.715,25.687,-7.574],[37.661,24.639,-4.080],
[40.261,24.811,-20.044],[40.134,22.212,-7.956],[41.143,22.834,-21.881],
[44.995,25.854,-4.511],[50.393,25.412,-22.197],[40.849,20.272,-6.222],
[40.702,18.870,-8.769],[41.944,24.577,-9.467],[46.641,21.962,-4.126],
[47.319,24.350,-5.336],[37.832,20.916,-6.495],[40.796,14.919,-8.032],
[43.086,17.912,-5.893],[45.128,20.607,-5.634],[46.515,18.547,-4.437],
[37.767,18.504,-4.621],[41.479,15.495,-5.557],[42.179,16.930,-2.670],
[40.610,32.802,-20.800],[35.582,27.628,-20.605],[34.803,32.510,-16.004],
[36.124,31.515,-13.852],[51.097,32.362,-11.378],[50.925,32.255,-8.741],
[50.671,29.309,-3.512],[52.146,27.305,-18.043],[50.666,24.894,-17.088],
[50.825,27.849,-10.791],[49.251,29.619,-9.000],[50.750,23.247,-19.973],
[50.400,20.669,-20.856],[49.804,23.272,-14.923],[34.447,24.811,-11.085],
[50.043,25.051,-6.337],[52.123,24.016,-4.605],[49.470,18.328,-20.501],
[50.907,21.006,-16.237],[50.175,20.510,-11.031],[49.521,22.843,-9.243],
[34.928,19.391,-5.027],[51.316,15.568,-3.605],[49.921,17.976,-17.666],
[35.416,17.813,-11.939],[33.105,22.717,-11.727],[52.240,18.250,-4.670],
[52.254,18.557,-20.932],[34.127,16.842,-19.398],[49.202,31.729,-16.166],
[33.755,15.697,-8.418],[35.936,30.311,-3.869],[39.391,28.719,-8.461],
[44.228,31.836,-5.081],[46.292,33.041,-7.025],[50.255,15.660,-6.730],
[36.611,33.633,-9.191],[40.140,33.103,-7.769],[42.759,33.399,-10.892],
[45.569,31.620,-10.228],[46.993,34.856,-8.851],[35.206,31.484,-10.937],
[38.158,31.012,-11.654],[40.622,32.574,-12.200],[43.639,17.249,-12.314],
[49.351,15.948,-12.829],[50.867,15.185,-15.098],[36.646,16.725,-14.969],
[34.574,14.285,-18.622],[41.056,16.408,-15.620],[46.973,17.092,-13.216],
[47.605,16.868,-18.788],[34.557,30.359,-17.595],[38.068,15.182,-19.836],
[42.674,18.470,-19.589],[43.644,32.527,-20.552],[48.757,30.485,-18.660],
[36.759,17.936,-7.056],[38.876,16.154,-5.523],[36.516,33.101,-21.545],
[46.017,15.443,-3.579],[49.981,30.411,-6.383],[43.917,42.571,-14.409],
[45.512,43.346,-12.080],[36.695,42.855,-10.739],[47.232,45.109,-13.244],
[37.873,46.514,-11.531],[34.923,41.997,-18.027],[34.614,43.224,-15.492],
[45.567,50.362,-12.428],[43.775,45.813,-11.684],[51.634,36.797,-9.167],
[47.937,38.855,-10.561],[39.187,37.036,-13.533],[45.582,37.097,-15.949],
[35.433,38.994,-14.101],[41.155,41.420,-13.310],[52.379,39.062,-7.062],
[38.073,37.080,-11.069],[45.994,43.771,-16.076],[40.679,45.694,-11.044],
[48.099,37.638,-13.189],[36.443,39.550,-11.704],[38.120,46.590,-8.457],
[36.372,46.001,-15.614],[48.732,46.992,-10.434],[48.423,35.766,-11.147],
[42.012,35.006,-13.111],[46.993,49.185,-14.442],[48.439,46.578,-15.262],
[47.835,45.867,-5.457],[44.736,41.588,-19.258],[44.819,40.958,-21.750],
[51.177,41.207,-9.218],[49.079,43.900,-14.947],[45.467,40.814,-6.987],
[38.792,43.364,-14.952],[39.656,47.441,-13.373],[40.014,42.835,-11.377],
[45.994,48.112,-10.761],[40.294,46.132,-6.612],[47.646,40.380,-17.125],
[36.564,42.204,-7.028],[44.124,48.458,-15.561],[38.656,48.173,-6.144],
[49.926,33.472,-18.138],[39.710,43.124,-17.567],[52.174,48.426,-20.434],
[33.198,45.157,-8.825],[34.428,48.324,-11.304],[52.018,47.812,-6.294],
[42.386,47.800,-7.088],[41.717,52.594,-18.175],[37.152,49.809,-15.194],
[46.445,48.981,-19.683],[44.505,51.091,-17.373],[49.900,39.275,-4.387],
[35.488,46.454,-6.634],[39.296,33.455,-10.143],[44.292,34.050,-8.687],
[47.572,39.396,-5.918],[46.835,42.692,-20.566],[44.567,36.331,-20.662],
[39.169,39.714,-12.214],[40.112,40.687,-16.470],[44.791,40.478,-16.679],
[43.878,41.645,-4.978],[45.637,39.383,-14.265],[36.792,39.921,-8.838],
[38.777,37.744,-22.650],[33.040,38.564,-15.629],[35.813,37.364,-18.529],
[45.491,41.265,-9.843],[41.171,36.786,-15.405],[37.880,38.136,-15.675],
[38.678,37.774,-19.004],[45.586,36.254,-10.692],[40.649,39.113,-20.456],
[48.089,52.084,-5.084],[38.132,44.452,-6.609],[44.688,39.148,-11.413],
[38.082,35.350,-16.995],[41.466,40.299,-10.739],[42.448,45.677,-18.336],
[49.216,42.915,-9.049],[47.547,35.911,-20.567],[42.418,44.682,-7.860],
[46.405,39.647,-19.920],[36.739,43.946,-13.444],[45.048,46.204,-17.577],
[33.908,43.136,-7.021],[41.130,41.156,-5.067],[46.567,48.322,-8.163],
[52.500,41.378,-18.823],[49.914,45.329,-7.917],[50.536,48.269,-16.758],
[49.015,44.292,-11.329],[38.576,47.721,-16.042],[36.119,41.651,-20.733],
[41.582,44.382,-15.082],[41.924,49.331,-9.177],[39.349,50.413,-16.507],
[44.345,43.634,-9.630],[46.407,45.860,-20.632],[48.355,50.281,-6.961],
[50.971,51.005,-18.935],[36.262,37.190,-8.769],[47.536,45.323,-18.412],
[36.634,48.537,-9.922],[41.896,47.501,-23.112],[39.987,44.639,-20.241],
[36.101,44.372,-19.382],[43.741,37.291,-8.237],[44.709,35.659,-18.088],
[43.584,48.375,-18.690],[42.837,38.852,-16.002],[42.574,39.539,-6.843],
[43.981,34.898,-15.194],[41.432,48.943,-20.212],[41.340,36.379,-20.576],
[51.210,36.757,-6.548],[47.807,41.609,-11.355],[37.280,46.838,-20.141],
[36.326,39.222,-21.751],[48.597,49.711,-9.900],[38.704,34.624,-12.476],
[34.387,47.998,-17.080],[34.660,48.482,-4.929],[39.476,50.783,-4.231],
[44.452,48.334,-21.595],[45.826,47.396,-3.508],[49.332,48.249,-19.995],
[34.097,45.155,-3.809],[38.443,45.053,-22.318],[40.911,43.918,-5.085],
[43.776,45.694,-21.516],[46.978,44.159,-8.422],[36.506,41.725,-2.865],
[40.738,42.051,-19.694],[39.833,41.926,-7.671],[42.164,40.870,-21.512],
[45.049,43.197,-2.184],[51.188,44.316,-21.902],[40.220,38.372,-6.430],
[39.455,39.068,-9.196],[43.028,41.609,-8.490],[45.811,39.900,-4.068],
[47.970,42.841,-5.843],[38.376,41.476,-4.919],[41.711,35.214,-8.784],
[41.896,38.817,-4.128],[45.124,37.570,-5.277],[43.490,35.280,-6.535],
[35.799,36.513,-4.708],[41.748,31.930,-6.248],[39.950,34.457,-2.713],
[41.073,52.164,-20.809],[34.399,46.295,-20.707],[34.850,51.612,-16.005],
[34.344,50.463,-13.807],[51.976,51.007,-10.478],[51.823,49.866,-8.219],
[50.978,45.798,-4.775],[50.123,45.203,-17.548],[51.705,43.361,-16.067],
[51.894,45.339,-10.888],[49.678,48.020,-7.820],[49.950,41.852,-20.260],
[51.595,39.146,-20.079],[48.918,38.356,-15.746],[34.309,43.194,-12.308],
[50.613,43.131,-5.310],[52.248,40.901,-4.273],[49.657,35.623,-22.227],
[47.775,37.434,-18.271],[49.963,38.589,-8.683],[49.779,40.497,-6.847],
[32.788,38.833,-6.091],[50.347,34.206,-2.284],[50.547,35.864,-16.670],
[35.300,36.900,-12.437],[33.751,40.983,-11.036],[53.214,35.095,-4.904],
[51.919,35.035,-20.114],[35.125,34.988,-16.742],[51.349,53.068,-14.740],
[34.058,35.593,-8.916],[36.932,47.569,-3.805],[39.411,48.815,-8.666],
[42.992,51.518,-4.552],[46.738,49.698,-4.692],[49.558,35.328,-8.311],
[36.032,50.991,-7.436],[38.879,51.490,-9.463],[42.837,51.603,-10.554],
[45.774,50.836,-9.586],[48.127,52.531,-8.457],[36.189,51.224,-10.314],
[37.324,50.007,-12.570],[39.704,50.991,-13.229],[42.437,38.118,-13.015],
[48.919,32.794,-13.148],[51.088,33.740,-14.681],[36.557,36.026,-14.725],
[35.165,34.668,-19.829],[40.586,34.162,-15.918],[47.103,35.137,-14.148],
[47.389,34.639,-17.517],[35.519,48.963,-19.537],[37.096,36.424,-20.974],
[42.426,34.419,-19.040],[43.633,50.849,-21.506],[48.702,50.529,-17.664],
[37.073,34.588,-6.752],[38.499,33.491,-4.578],[37.030,51.359,-20.234],
[46.109,33.565,-3.800],[49.718,48.288,-4.668],[52.590,55.439,-10.309],
[47.562,57.580,-10.083],[39.387,54.490,-13.537],[46.079,55.977,-14.681],
[36.107,57.439,-14.707],[52.366,57.335,-8.464],[38.168,55.609,-8.998],
[48.841,56.773,-14.527],[38.062,56.471,-11.451],[48.565,54.932,-7.594],
[42.771,51.526,-13.474],[50.821,53.694,-19.670],[50.083,58.259,-4.467],
[40.383,53.487,-11.317],[45.362,51.776,-7.015],[47.477,57.111,-5.372],
[43.809,53.519,-21.008],[38.851,58.846,-9.064],[38.770,57.363,-13.906],
[46.012,57.706,-12.332],[36.376,57.942,-9.951],[40.532,56.953,-23.228],
[32.929,55.568,-15.020],[34.557,56.104,-19.313],[41.566,56.361,-14.395],
[37.929,-59.384,-15.185],[40.131,58.283,-16.474],[46.812,54.952,-9.827],
[39.602,58.714,-21.386],[44.099,57.128,-10.251],[40.048,55.059,-16.328],
[41.462,58.714,-9.734],[46.117,55.244,-19.504],[47.273,57.453,-19.318],
[53.947,59.522,-16.870],[35.815,55.101,-8.005],[43.209,54.897,-9.434],
[45.879,54.159,-16.687],[43.777,57.950,-17.406],[43.330,57.470,-7.518],
[43.192,53.910,-15.892],[40.506,56.474,-18.632],[49.812,55.478,-5.423],
[36.320,58.094,-21.402],[37.140,54.024,-11.598],[40.822,58.291,-6.594],
[40.855,56.074,-9.107],[45.157,58.594,-3.150],[38.492,57.319,-5.590],
[41.914,52.801,-8.332],[41.747,55.664,-6.257],[44.906,56.188,-5.318],
[47.552,54.643,-3.838],[36.449,55.029,-4.189],[40.862,51.091,-6.513],
[42.422,54.248,-3.887],[51.577,57.072,-19.598],[50.813,56.852,-23.928],
[49.634,57.129,-17.683],[50.255,56.795,-9.890],[49.029,-59.006,-8.903],
[33.070,56.384,-5.437],[52.839,52.306,-3.690],[49.826,54.583,-16.963],
[35.014,55.766,-12.900],[34.387,58.278,-11.858],[52.423,54.936,-4.406],
[50.080,55.478,-21.549],[33.839,53.850,-17.147],[34.074,52.489,-8.890],
[50.776,51.942,-5.746],[43.788,55.363,-13.251],[49.911,52.280,-12.289],
[51.292,55.824,-13.231],[37.488,52.521,-15.945],[34.653,51.727,-18.798],
[40.994,52.361,-15.335],[46.891,52.619,-13.221],[47.912,53.268,-18.619],
[36.863,54.155,-20.376],[43.159,55.031,-18.587],[37.354,52.943,-6.107],
[38.610,53.044,-3.356],[45.553,52.604,-3.869],[44.267,-50.060,3.984],
[42.518,-49.684,6.355],[37.233,-52.348,9.097],[48.310,-48.600,5.687],
[37.474,-46.338,7.171],[35.644,-52.988,1.835],[34.348,-49.825,3.752],
[46.590,-43.007,6.539],[42.462,-46.854,6.868],[51.974,-55.341,9.031],
[46.842,-53.848,8.762],[39.846,-55.566,4.343],[47.347,-55.754,3.754],
[36.177,-51.722,5.246],[41.422,-50.994,4.118],[51.909,-52.511,11.234],
[37.854,-55.828,7.849],[46.772,-48.674,3.324],[40.576,-46.921,9.291],
[50.210,-55.638,5.676],[37.883,-54.280,5.308],[37.159,-47.439,9.715],
[35.938,-45.557,0.953],[48.328,-46.181,7.736],[48.831,-57.083,7.645],
[41.870,-59.115,6.736],[47.022,-45.554,5.164],[48.518,-44.543,2.061],
[49.591,-48.218,11.732],[45.057,-49.866,0.185],[44.462,-51.644,-2.145],
[52.062,-51.450,8.864],[49.254,-50.671,0.782],[45.583,-51.865,12.521],
[38.933,-48.960,3.498],[41.063,-45.514,4.603],[39.491,-50.058,6.818],
[45.331,-45.189,8.586],[40.181,-46.038,12.215],[46.877,-51.422,-1.034],
[39.529,-50.199,9.784],[43.797,-42.782,4.625],[39.201,-43.292,13.053],
[48.357,-59.525,0.175],[37.201,-50.663,0.968],[51.400,-43.785,-1.614],
[33.557,-49.352,9.712],[32.613,-44.664,6.058],[52.262,-45.040,12.710],
[43.035,-45.578,11.908],[42.997,-42.724,2.067],[37.635,-41.548,3.861],
[45.935,-42.807,-2.501],[46.484,-43.095,1.354],[49.498,-51.914,12.163],
[34.816,-48.408,12.027],[39.901,-57.504,6.123],[45.269,-58.128,9.758],
[47.916,-54.331,12.843],[48.707,-49.622,-1.974],[43.708,-55.518,-2.822],
[39.418,-53.768,7.896],[40.554,-53.624,2.573],[45.988,-51.988,1.543],
[43.810,-50.673,15.908],[45.497,-53.480,3.864],[38.002,-53.913,11.402],
[40.398,-55.947,-3.183],[33.757,-53.931,3.777],[36.704,-55.357,0.219],
[44.663,-52.167,9.324],[43.072,-54.308,3.329],[36.733,-55.121,3.007],
[41.329,-53.864,-1.252],[46.370,-56.321,7.649],[38.738,-51.638,-1.516],
[47.954,-42.221,13.787],[37.112,-47.727,14.218],[43.090,-54.610,6.082],
[38.548,-57.750,3.490],[42.388,-53.281,8.113],[42.568,-48.442,0.171],
[49.249,-48.505,8.940],[46.536,-54.346,-1.093],[45.903,-48.036,12.041],
[48.533,-53.110,0.659],[37.061,-49.297,5.457],[45.208,-47.229,0.913],
[36.450,-50.245,10.860],[42.573,-51.496,12.810],[45.942,-44.952,11.493],
[51.428,-53.025,0.483],[52.491,-47.754,12.945],[50.169,-46.343,1.115],
[47.337,-49.924,7.628],[38.136,-46.005,2.530],[35.721,-50.834,-2.254],
[41.582,-48.071,3.976],[39.616,-44.479,8.160],[38.841,-42.876,1.778],
[45.724,-47.905,9.223],[47.126,-47.165,-1.977],[48.015,-43.326,11.113],
[52.123,-41.174,-0.440],[35.867,-55.839,10.889],[47.982,-47.897,0.927],
[36.188,-44.791,9.371],[41.997,-44.848,-0.988],[41.229,-44.851,1.754],
[37.531,-49.170,-1.220],[43.314,-56.229,8.150],[46.294,-56.788,0.642],
[44.064,-44.706,0.691],[43.837,-54.840,0.485],[43.190,-54.559,10.734],
[44.411,-56.554,3.895],[39.344,-43.989,-0.992],[41.200,-57.681,-1.125],
[51.153,-55.239,11.726],[48.252,-51.416,5.387],[38.444,-46.693,-1.485],
[35.516,-55.135,-4.240],[48.872,-42.122,8.310],[36.973,-58.209,6.515],
[34.775,-43.119,1.877],[34.981,-42.368,13.413],[40.671,-43.280,15.341],
[44.311,-45.485,-2.046],[47.787,-46.156,12.932],[47.881,-44.622,-3.173],
[33.977,-47.637,14.622],[41.233,-48.048,-2.385],[42.369,-48.305,12.524],
[44.641,-47.227,-3.962],[47.541,-50.322,10.559],[38.062,-50.196,13.024],
[39.844,-50.322,0.765],[40.556,-52.818,10.268],[40.963,-50.688,-3.305],
[44.440,-48.884,13.936],[51.523,-50.693,-3.038],[40.618,-55.402,12.236],
[40.164,-56.418,9.383],[42.438,-50.603,8.890],[48.392,-53.769,15.584],
[47.622,-50.360,14.950],[38.666,-52.627,13.798],[42.232,59.243,10.650],
[42.338,-53.883,13.646],[45.476,-54.426,13.588],[47.339,-56.274,15.082],
[37.459,-55.422,14.632],[42.172,-58.452,13.197],[42.352,-58.034,16.376],
[41.563,-40.782,-1.146],[34.245,-45.877,-1.337],[34.231,-40.666,3.181],
[32.694,-42.817,3.992],[52.482,-42.766,7.787],[50.656,-42.469,10.280],
[48.943,-46.323,15.966],[52.509,-47.951,2.505],[51.953,-50.732,1.890],
[51.748,-48.079,8.328],[50.123,-45.477,10.402],[52.456,-49.239,-0.331],
[49.649,-52.745,-2.340],[50.081,-52.387,2.923],[35.468,-49.493,8.015],
[50.889,-49.835,13.574],[52.764,-51.483,14.810],[50.073,-55.044,-3.391],
[49.626,-55.249,2.580],[49.062,-52.092,8.106],[49.126,-54.615,10.161],
[34.029,-54.364,14.081],[53.410,-58.850,15.440],[50.816,-57.501,2.497],
[35.527,-55.914,5.363],[34.316,-53.759,6.794],[50.805,-56.197,14.376],
[51.485,-56.825,-1.748],[34.961,-57.544,1.449],[47.428,-43.196,3.825],
[34.268,-58.137,9.006],[36.839,-45.063,14.806],[38.567,-45.131,10.387],
[42.649,-42.663,13.316],[45.039,-42.278,12.051],[51.150,-58.678,11.550],
[36.506,-41.327,9.213],[39.024,-41.840,10.665],[44.314,-41.120,7.417],
[46.018,-42.711,9.297],[48.277,-40.554,10.657],[36.684,-42.851,6.844],
[37.873,-44.534,5.053],[40.194,-43.173,5.868],[44.452,-59.473,6.579],
[49.413,-58.616,4.842],[52.263,-59.726,3.049],[37.775,-59.998,1.952],
[35.791,-59.487,-0.556],[41.808,-59.022,3.995],[46.833,-58.497,5.214],
[48.124,-58.146,2.537],[35.333,-41.910,-0.578],[37.150,-56.990,-1.926],
[42.540,-57.799,1.421],[44.122,-42.142,-0.508],[51.152,-43.434,1.244],
[35.630,-59.323,11.181],[39.236,-58.855,13.855],[35.249,-42.877,-3.294],
[48.472,-59.334,16.870],[50.308,-42.995,13.231],[46.373,-32.440,5.106],
[44.212,-31.942,6.548],[36.938,-32.925,7.469],[49.747,-27.490,3.808],
[38.401,-28.547,5.477],[35.862,-33.922,1.964],[34.532,-30.920,3.606],
[45.724,-23.443,5.909],[43.013,-29.759,7.514],[51.746,-38.339,8.555],
[47.446,-36.218,7.538],[38.086,-35.380,5.404],[47.189,-38.036,3.262],
[35.684,-33.984,4.717],[38.681,-33.689,3.282],[51.896,-35.591,10.002],
[37.750,-38.036,7.924],[43.263,-27.888,3.756],[39.801,-29.899,7.783],
[48.861,-37.493,5.193],[36.174,-35.311,7.360],[38.069,-28.495,9.500],
[36.926,-27.207,0.182],[47.509,-27.966,6.256],[48.724,-38.957,8.516],
[41.736,-41.128,6.677],[47.195,-25.914,4.311],[45.845,-29.262,4.025],
[49.760,-28.339,12.488],[43.986,-29.734,1.720],[44.797,-32.443,-2.107],
[52.180,-32.684,9.581],[47.507,-31.297,2.267],[46.389,-33.505,10.555],
[37.064,-31.226,4.213],[41.304,-27.070,5.394],[38.994,-32.078,5.778],
[46.294,-26.205,8.036],[39.235,-27.055,12.774],[45.777,-32.430,0.496],
[37.263,-31.929,11.994],[44.083,-25.477,2.876],[38.508,-24.397,13.401],
[49.684,-40.433,0.551],[39.831,-31.300,0.788],[51.524,-26.866,-0.644],
[35.469,-31.473,9.161],[33.805,-26.654,8.234],[52.325,-26.531,11.213],
[43.057,-27.364,12.091],[44.007,-22.506,3.193],[38.844,-24.360,3.774],
[46.918,-23.192,-1.076],[45.707,-23.748,1.470],[50.103,-33.878,11.397],
[36.897,-26.795,10.995],[39.053,-40.477,7.962],[44.217,-40.383,10.306],
[46.853,-36.409,11.858],[47.501,-31.859,-2.634],[43.816,-35.997,-1.232],
[39.332,-35.576,7.774],[38.591,-34.404,0.437],[44.223,-32.086,3.123],
[43.431,-33.474,14.039],[46.626,-35.582,4.810],[37.036,-34.611,12.295],
[38.772,-36.266,-3.817],[33.233,-38.114,3.944],[34.077,-36.160,1.493],
[42.694,-33.910,7.570],[40.917,-38.354,3.373],[36.995,-36.023,3.072],
[40.055,-36.643,1.154],[45.016,-38.458,8.223],[39.820,-33.109,-1.761],
[48.288,-21.930,12.774],[37.371,-29.210,12.734],[43.961,-34.177,4.994],
[38.363,-39.582,1.053],[42.342,-36.358,8.421],[42.053,-27.915,-1.896],
[48.831,-31.162,9.703],[45.816,-37.933,-1.431],[43.351,-30.237,11.258],
[46.906,-35.507,-0.662],[35.570,-29.822,6.143],[45.566,-28.251,0.228],
[34.694,-32.689,12.074],[40.591,-32.877,12.877],[45.467,-26.172,10.603],
[52.027,-34.148,-0.464],[51.064,-31.172,11.580],[49.311,-26.647,0.959],
[47.691,-30.545,7.289],[40.415,-26.140,2.531],[34.946,-31.669,-1.819],
[40.067,-28.924,2.407],[41.395,-26.068,9.045],[39.249,-22.356,1.895],
[45.353,-29.788,9.187],[48.356,-28.069,-1.481],[47.812,-24.824,12.104],
[51.349,-24.997,-2.431],[34.806,-37.583,11.519],[48.004,-30.835,-0.149],
[36.565,-26.339,8.421],[40.833,-26.246,-4.837],[39.760,-27.877,-0.260],
[38.613,-30.672,-1.518],[42.682,-37.960,10.450],[45.526,-37.094,1.139],
[43.845,-26.069,-0.485],[42.924,-36.248,1.782],[42.774,-34.783,11.071],
[44.379,-38.556,2.975],[41.396,-24.428,-1.589],[41.145,-38.035,-1.111],
[49.177,-36.556,10.341],[48.657,-32.869,6.353],[36.825,-26.333,-2.572],
[36.724,-34.782,-2.102],[48.179,-24.248,7.872],[35.244,-40.601,6.270],
[37.509,-26.976,3.268],[34.051,-25.679,15.106],[41.168,-24.646,14.873],
[43.817,-25.904,-3.513],[45.596,-24.050,13.598],[48.821,-25.609,-2.852],
[32.872,-29.289,14.701],[38.368,-28.560,-3.450],[40.661,-29.246,14.684],
[43.963,-28.646,-3.801],[45.813,-31.161,12.082],[36.153,-33.091,14.536],
[42.233,-31.498,-0.588],[39.226,-34.090,10.201],[40.514,-31.730,-3.951],
[43.836,-30.519,13.989],[50.744,-29.896,-2.418],[39.937,-36.167,12.142],
[40.389,-37.041,5.745],[41.740,-31.979,9.774],[45.439,-35.493,14.220],
[48.115,-34.432,13.561],[38.616,-33.847,14.626],[41.333,-40.540,10.052],
[41.599,-36.398,14.350],[44.116,-37.281,12.675],[46.794,-37.844,14.358],
[36.089,-36.267,14.470],[40.678,-41.011,12.895],[41.628,-38.995,14.851],
[41.250,-22.048,-0.283],[34.790,-27.941,-1.298],[34.854,-22.660,2.186],
[34.169,-24.895,5.905],[52.249,-23.830,6.262],[49.892,-24.067,10.073],
[48.564,-28.383,15.067],[50.674,-29.252,0.731],[50.233,-31.840,1.780],
[50.293,-28.839,6.476],[48.807,-26.630,10.075],[49.759,-33.174,-1.333],
[50.699,-35.192,-3.816],[48.260,-34.070,2.909],[32.848,-31.193,7.021],
[48.526,-30.860,13.172],[50.685,-33.218,13.801],[48.949,-37.453,-2.901],
[50.197,-36.375,0.349],[50.040,-34.867,5.233],[49.044,-34.217,8.780],
[33.797,-35.028,14.024],[50.069,-39.434,15.359],[52.102,-38.908,2.237],
[35.508,-37.723,6.159],[33.065,-34.033,6.468],[51.279,-37.936,11.582],
[50.804,-38.601,-0.980],[35.564,-38.537,1.978],[48.549,-23.745,1.694],
[34.080,-40.170,8.865],[35.251,-27.270,13.003],[39.247,-24.971,10.739],
[41.585,-25.369,12.019],[45.428,-20.923,12.227],[50.021,-40.225,12.841],
[36.393,-20.586,10.423],[37.013,-23.299,9.227],[42.659,-23.801,7.386],
[45.822,-23.705,9.080],[47.938,-21.851,9.522],[35.115,-22.159,7.112],
[37.415,-25.551,5.947],[40.300,-24.890,6.646],[43.642,-38.436,5.522],
[49.900,-42.406,5.873],[50.832,-39.199,4.722],[37.546,-38.644,4.275],
[34.235,-38.573,-0.603],[41.412,-41.093,3.656],[46.844,-39.912,5.612],
[47.359,-40.263,-1.006],[34.739,-24.990,0.419],[36.622,-37.879,-3.112],
[43.075,-39.388,0.610],[43.874,-23.189,-2.281],[51.335,-24.333,0.877],
[37.322,-40.816,12.292],[38.457,-40.710,15.045],[37.211,-23.725,-1.679],
[45.990,-41.410,15.449],[49.939,-25.579,13.666],[45.172,-11.997,3.677],
[45.224,-12.485,6.736],[36.705,-12.926,8.707],[48.529,-9.040,6.103],
[38.053,-9.626,5.967],[35.292,-15.326,2.295],[33.833,-13.552,3.494],
[44.622,-5.736,4.960],[44.283,-9.568,7.506],[51.064,-19.692,9.449],
[46.990,-16.288,8.739],[39.364,-16.876,4.624],[48.340,-20.491,3.061],
[34.644,-16.956,4.563],[41.293,-14.978,3.664],[52.183,-17.109,9.648],
[37.132,-18.959,8.418],[44.640,-9.290,2.275],[41.095,-12.438,9.946],
[48.600,-18.214,4.811],[37.012,-17.554,6.254],[38.744,-11.095,10.187],
[34.060,-8.802,-0.102],[47.134,-8.340,8.297],[48.347,-19.020,9.151],
[41.468,-20.849,5.976],[47.255,-7.439,4.321],[47.419,-10.232,2.773],
[49.830,-10.248,11.269],[43.054,-12.075,1.558],[46.187,-13.629,-3.183],
[52.312,-14.413,8.829],[48.210,-12.706,2.330],[46.882,-14.619,10.715],
[37.831,-11.082,3.160],[39.919,-6.900,6.694],[39.103,-11.691,7.690],
[44.585,-7.058,9.417],[39.354,-8.618,9.855],[47.281,-14.317,-0.838],
[37.527,-13.020,12.541],[42.673,-7.251,2.857],[38.869,-6.620,11.591],
[50.006,-22.227,-0.603],[35.736,-12.239,1.467],[51.854,-6.832,-1.351],
[34.115,-11.626,8.905],[33.074,-7.098,7.334],[53.890,-7.540,12.153],
[41.850,-8.026,11.379],[43.128,-3.069,1.093],[37.080,-4.160,3.640],
[46.800,-5.200,-2.231],[46.632,-5.131,2.403],[49.614,-14.546,13.065],
[36.114,-10.333,10.212],[38.948,-21.914,8.464],[44.168,-21.699,9.823],
[47.648,-18.652,12.114],[48.961,-12.708,-2.743],[43.076,-20.209,-1.240],
[39.101,-15.260,7.165],[38.474,-14.977,2.259],[46.306,-14.265,1.579],
[43.570,-14.820,13.192],[46.864,-16.075,3.738],[37.121,-15.795,8.901],
[40.040,-17.109,-2.789],[35.195,-20.246,3.988],[35.490,-17.544,1.063],
[44.868,-14.693,8.383],[41.880,-18.195,3.266],[37.090,-17.138,3.452],
[40.386,-16.342,-0.313],[45.490,-18.677,8.040],[39.458,-13.748,-1.929],
[48.272,-4.726,14.086],[37.332,-9.982,12.899],[43.557,-14.149,5.526],
[38.257,-18.078,0.590],[42.379,-16.029,7.369],[41.561,-10.072,0.511],
[49.901,-12.674,9.566],[45.113,-18.309,-0.140],[43.341,-10.581,11.079],
[47.369,-17.025,-1.099],[37.798,-12.578,5.577],[44.824,-8.280,-0.995],
[33.410,-14.858,12.099],[40.484,-13.867,12.021],[45.442,-7.177,12.199],
[51.627,-15.053,-0.263],[52.630,-9.695,11.998],[49.238,-8.919,1.374],
[46.374,-12.103,9.139],[39.913,-7.364,4.073],[36.649,-13.292,-1.768],
[40.833,-9.722,2.973],[42.009,-7.966,7.775],[40.252,-3.928,1.101],
[45.891,-9.399,10.374],[47.502,-8.934,-1.736],[48.371,-5.746,11.432],
[53.596,-4.488,-2.451],[36.466,-17.662,11.221],[47.227,-11.123,-0.099],
[36.592,-7.988,8.674],[41.335,-8.064,-4.247],[39.615,-11.205,-0.703],
[37.478,-10.180,0.590],[43.770,-19.014,10.002],[46.971,-17.359,1.447],
[42.128,-5.901,0.315],[43.190,-16.531,-0.304],[43.503,-16.348,10.357],
[44.189,-19.511,2.850],[38.985,-6.116,-0.605],[40.201,-19.623,-1.130],
[50.866,-19.441,13.455],[48.558,-15.254,5.763],[37.696,-8.706,-2.345],
[35.155,-15.076,-3.190],[48.664,-6.013,8.384],[38.946,-22.021,5.642],
[34.893,-8.048,3.297],[34.313,-6.054,12.927],[40.102,-6.424,14.399],
[42.296,-7.647,-1.859],[46.267,-9.293,13.728],[48.887,-6.524,-3.498],
[34.539,-10.714,13.548],[40.206,-10.581,-3.187],[40.688,-10.826,12.523],
[44.992,-9.044,-3.425],[47.317,-11.720,11.999],[35.682,-14.347,13.539],
[41.841,-12.880,-0.957],[39.239,-16.262,10.317],[41.770,-15.098,-2.303],
[46.420,-12.210,14.751],[50.883,-11.495,-4.190],[39.098,-18.160,13.951],
[40.868,-18.266,8.560],[43.652,-13.318,10.510],[47.737,-16.507,14.745],
[46.697,-14.411,13.242],[38.911,-15.113,14.145],[41.278,-20.636,9.691],
[42.370,-17.474,12.987],[44.974,-16.991,12.664],[46.076,-18.989,14.166],
[36.328,-17.706,14.119],[40.438,-22.574,12.920],[41.745,-21.538,15.396],
[42.579,-3.312,-1.562],[35.521,-6.836,-2.174],[34.421,-3.784,3.024],
[33.237,-6.292,4.436],[51.333,-6.254,8.426],[51.090,-5.603,11.008],
[51.019,-9.846,14.211],[51.249,-10.747,0.277],[51.745,-13.094,2.152],
[51.247,-10.281,8.901],[49.756,-8.156,9.820],[49.687,-13.171,-0.106],
[50.004,-16.450,-1.529],[49.729,-14.850,3.305],[34.569,-12.598,6.272],
[50.532,-12.146,12.723],[52.133,-14.193,13.691],[48.920,-18.153,-3.170],
[49.711,-17.641,2.428],[50.098,-17.043,7.944],[49.444,-17.237,10.770],
[33.567,-17.623,12.409],[52.673,-21.496,13.792],[50.772,-20.163,1.320],
[34.923,-19.499,6.801],[33.086,-16.914,7.060],[52.793,-19.193,11.431],
[50.748,-20.124,-2.566],[33.964,-19.727,0.362],[49.229,-4.946,3.632],
[33.597,-19.849,10.486],[37.010,-7.111,13.829],[38.874,-6.063,9.101],
[43.468,-5.537,12.225],[46.331,-2.991,13.118],[50.905,-21.814,11.223],
[36.311,-4.145,9.223],[38.184,-3.827,11.369],[44.270,-3.032,8.747],
[45.434,-4.638,10.537],[48.102,-3.333,9.984],[35.325,-4.653,6.029],
[37.823,-5.315,6.859],[41.979,-5.897,5.200],[44.286,-18.465,5.570],
[50.662,-21.302,6.816],[50.933,-21.944,3.510],[36.375,-22.847,4.173],
[35.333,-22.414,-0.448],[41.407,-22.615,3.800],[46.044,-20.879,4.814],
[46.612,-20.683,0.292],[35.146,-6.395,1.196],[37.207,-20.863,-2.485],
[41.952,-18.623,0.770],[44.438,-4.993,-0.622],[51.029,-5.451,1.232],
[36.316,-20.757,14.382],[38.985,-20.477,15.389],[36.312,-4.158,-3.795],
[45.073,-21.652,14.803],[49.933,-7.083,13.383],[44.771,6.126,3.834],
[43.585,6.349,6.646],[37.277,4.371,7.260],[45.565,9.290,5.776],
[36.999,9.557,7.017],[35.739,4.013,1.729],[34.925,5.949,4.035],
[46.130,13.739,5.849],[42.691,9.549,6.806],[51.563,-0.329,9.121],
[47.460,1.372,8.333],[39.496,1.273,5.342],[46.635,-0.010,3.020],
[37.257,2.648,4.107],[42.124,4.312,5.093],[52.669,1.117,11.023],
[36.025,-0.666,7.762],[43.131,5.202,-0.120],[40.441,7.946,8.971],
[47.221,-0.380,6.010],[37.071,1.691,6.672],[38.123,7.774,10.748],
[35.847,7.089,1.685],[46.575,7.844,8.449],[48.893,-0.973,8.752],
[41.576,-3.986,7.674],[47.613,11.573,5.436],[47.997,9.758,3.458],
[49.465,7.881,10.650],[42.893,3.988,2.489],[44.000,3.771,-2.615],
[50.597,5.844,9.702],[47.704,5.376,3.562],[45.877,3.993,11.390],
[40.395,6.610,3.967],[41.128,11.694,5.740],[41.121,5.872,7.511],
[47.281,11.726,8.392],[40.150,9.459,11.439],[45.915,6.438,-0.937],
[38.996,4.792,12.054],[44.947,12.781,3.311],[39.542,11.592,13.527],
[51.045,-1.843,-0.083],[38.458,5.329,2.135],[51.068,10.690,-3.556],
[37.014,6.785,8.350],[34.071,9.148,6.664],[52.310,10.911,9.804],
[44.143,11.864,11.680],[43.327,14.902,1.276],[37.303,13.856,4.318],
[47.747,13.979,-1.151],[46.104,13.565,1.124],[49.724,4.239,14.432],
[36.529,9.542,13.207],[38.576,-1.676,8.206],[43.525,-2.922,11.600],
[48.077,2.460,11.828],[48.623,4.479,-2.012],[44.689,1.266,1.372],
[39.371,1.264,8.534],[39.774,2.994,2.112],[45.527,4.426,1.847],
[44.268,4.047,13.974],[46.660,2.448,4.114],[37.550,3.301,10.110],
[41.653,0.850,-2.310],[33.949,-0.969,2.846],[35.345,0.745,-0.935],
[44.993,2.823,8.744],[40.421,-1.028,4.828],[37.532,1.383,1.168],
[40.152,1.553,-0.304],[45.807,-0.621,8.586],[38.186,4.807,-0.389],
[49.939,14.748,12.681],[37.340,6.784,13.085],[42.765,1.856,6.276],
[39.768,-0.877,1.419],[42.335,1.756,9.283],[43.870,7.789,0.325],
[48.415,5.502,7.742],[45.903,0.112,-1.057],[42.866,8.304,11.094],
[47.442,2.006,-0.283],[37.694,6.017,4.552],[45.906,7.758,-3.622],
[34.961,6.135,11.419],[42.292,4.857,12.631],[45.484,9.410,11.986],
[50.540,3.465,0.443],[50.908,8.969,12.529],[49.345,11.162,1.001],
[46.299,6.220,6.495],[40.555,9.603,4.332],[35.138,6.615,-1.769],
[40.786,8.232,1.949],[40.580,12.168,10.876],[39.417,12.985,2.997],
[45.151,10.490,16.611],[47.505,9.779,-1.135],[48.349,10.175,12.028],
[53.570,13.718,-2.628],[36.691,0.687,11.117],[47.384,8.096,0.796],
[35.813,10.907,9.276],[41.537,9.421,-3.122],[40.214,8.941,-0.714],
[37.545,7.810,-0.487],[43.426,-0.520,10.014],[45.408,-1.498,1.191],
[43.208,12.605,-0.495],[42.073,0.186,2.963],[43.362,2.282,11.840],
[44.571,-2.732,3.657],[40.142,11.966,-1.839],[40.419,-1.221,-1.098],
[49.828,-0.662,11.335],[48.523,2.789,5.892],[37.925,11.141,0.015],
[36.564,3.433,-2.573],[50.081,12.954,8.105],[37.707,-2.257,5.165],
[38.181,10.344,3.301],[34.338,13.207,14.510],[40.202,12.279,16.788],
[43.409,10.289,-4.964],[46.610,11.108,14.521],[48.456,11.754,-3.075],
[33.159,5.919,14.717],[38.198,9.236,-2.702],[42.849,8.046,13.715],
[44.078,9.444,-2.239],[47.188,6.360,10.702],[36.425,3.761,13.157],
[40.772,6.037,0.551],[40.151,3.610,9.524],[41.356,5.128,-2.184],
[45.063,6.636,14.099],[48.966,6.158,-4.357],[40.800,0.963,12.183],
[40.807,-1.328,9.829],[44.472,6.558,9.649],[48.053,2.142,15.319],
[47.082,4.268,13.804],[39.957,3.191,13.956],[41.146,-4.925,10.227],
[43.521,0.347,13.705],[45.988,1.650,13.621],[46.731,-0.507,15.011],
[37.635,0.576,13.902],[41.404,-4.076,13.496],[43.572,-3.448,14.762],
[41.088,14.983,-0.876],[34.498,10.210,-1.434],[34.415,13.264,2.643],
[34.349,13.629,5.400],[53.403,14.129,9.433],[50.587,12.481,11.047],
[49.199,9.533,14.722],[50.545,7.729,0.605],[50.278,5.457,2.624],
[50.015,7.840,8.010],[49.465,10.381,9.613],[51.091,5.762,-1.607],
[51.270,2.923,-2.436],[48.827,2.588,2.280],[33.942,6.110,7.503],
[50.737,5.892,12.464],[52.752,4.040,13.593],[49.433,1.022,-3.027],
[49.243,0.173,0.724],[50.702,0.864,6.450],[51.183,3.299,11.145],
[34.935,1.606,13.655],[51.811,-2.239,16.230],[50.134,-1.203,2.679],
[35.085,-1.190,5.146],[34.586,2.628,6.798],[50.910,-0.914,13.623],
[52.227,-0.594,-2.385],[35.201,-2.260,0.680],[48.464,12.682,3.083],
[34.712,-1.466,9.998],[36.965,12.225,13.978],[38.629,11.123,9.084],
[44.135,14.727,13.530],[46.884,13.552,12.644],[50.084,-3.222,12.031],
[37.269,16.027,10.854],[38.172,13.660,10.026],[44.147,15.929,8.492],
[45.362,13.397,8.866],[47.573,14.788,7.843],[36.588,14.626,7.993],
[38.486,12.028,6.012],[40.413,13.270,8.137],[43.575,-0.617,5.478],
[49.235,-3.611,6.659],[50.913,-3.178,4.349],[37.095,-1.269,2.147],
[36.397,-4.729,-0.565],[42.027,-3.158,4.017],[46.058,-3.279,5.636],
[48.477,-3.812,0.672],[35.302,11.773,0.779],[36.785,-1.943,-1.532],
[42.999,-0.582,-0.412],[43.973,13.750,-3.083],[52.017,13.109,0.564],
[35.412,-1.130,14.737],[37.977,-2.605,14.224],[38.224,13.848,-3.105],
[46.640,-2.907,16.167],[51.017,11.529,13.840],[46.208,25.356,5.179],
[45.590,23.212,6.909],[37.863,23.111,8.266],[48.402,26.754,6.210],
[37.258,26.235,7.300],[35.735,22.348,1.185],[35.142,23.420,4.713],
[44.395,32.527,5.635],[42.429,27.473,7.924],[51.895,16.676,8.486],
[46.839,20.845,8.341],[38.789,19.001,5.097],[46.766,19.590,5.103],
[36.218,20.288,4.444],[40.810,22.919,4.681],[51.735,20.566,9.679],
[36.788,17.793,8.514],[42.719,28.187,3.898],[39.293,25.111,8.918],
[48.758,16.771,5.820],[37.554,20.721,6.849],[37.844,27.679,9.681],
[35.751,28.553,0.683],[47.950,26.460,8.775],[48.866,16.959,8.992],
[41.532,15.866,7.795],[47.632,28.992,4.813],[46.659,29.200,2.163],
[49.869,27.359,10.523],[45.161,25.160,2.217],[45.700,23.765,-2.582],
[50.680,23.177,9.830],[48.290,25.276,3.237],[46.071,23.649,10.442],
[38.477,22.796,1.689],[39.858,28.349,5.307],[41.225,25.622,5.674],
[46.783,29.969,7.752],[40.608,27.701,10.751],[46.951,23.997,0.429],
[38.418,23.540,11.340],[43.398,31.221,3.299],[39.029,30.848,13.417],
[50.502,15.316,-0.355],[36.849,24.687,2.614],[49.819,31.148,-1.837],
[32.480,25.852,9.940],[35.301,28.888,5.555],[53.808,29.541,12.143],
[41.839,29.124,12.666],[43.282,33.734,1.014],[37.094,32.805,4.152],
[47.130,31.553,-0.856],[45.699,32.037,1.724],[49.778,21.217,14.107],
[36.904,29.213,11.454],[38.957,16.313,7.875],[43.385,14.563,10.704],
[47.587,19.026,10.826],[49.423,24.383,-3.620],[42.857,19.371,-1.587],
[39.363,20.810,9.061],[40.453,20.936,2.750],[45.102,22.332,1.548],
[45.124,22.360,13.197],[44.864,21.842,4.198],[37.303,21.202,10.838],
[40.185,19.283,-3.446],[33.720,18.071,3.918],[34.928,20.174,-2.137],
[43.841,22.952,8.992],[41.348,18.805,4.750],[37.880,20.231,2.053],
[40.976,19.187,0.477],[46.011,17.792,6.999],[39.515,22.353,-1.327],
[48.263,33.429,13.624],[37.785,26.158,12.279],[43.846,20.673,6.434],
[38.331,15.282,1.931],[41.761,21.257,8.122],[42.811,25.745,0.147],
[50.491,25.654,8.371],[46.502,18.915,-1.037],[43.288,27.746,10.888],
[47.769,20.787,0.839],[38.859,24.320,5.842],[44.277,27.574,1.044],
[34.990,22.717,10.967],[42.471,22.250,14.156],[45.772,29.164,10.638],
[49.946,22.673,-0.390],[52.340,27.073,12.545],[51.299,29.684,0.162],
[48.426,23.862,6.722],[38.777,29.027,1.702],[35.994,23.191,-1.546],
[39.601,25.959,3.569],[42.395,31.316,9.992],[40.791,31.327,1.999],
[44.858,25.374,8.160],[46.417,27.198,-2.109],[47.826,28.081,12.249],
[52.013,32.958,-0.415],[36.152,18.718,12.011],[48.377,27.113,0.615],
[37.048,29.931,8.768],[41.705,27.341,-3.063],[39.934,28.581,-1.011],
[37.483,26.897,-0.505],[44.069,18.954,9.477],[46.568,16.127,-0.271],
[43.802,30.668,0.829],[43.021,19.964,2.567],[43.048,21.377,11.299],
[43.501,17.042,3.148],[40.054,32.965,-0.202],[40.941,16.377,1.720],
[50.007,20.024,11.740],[48.013,21.276,3.617],[37.589,29.661,-1.695],
[37.299,20.906,-1.790],[48.482,32.013,7.309],[36.743,16.302,5.950],
[35.951,30.270,2.756],[35.397,32.019,12.983],[39.370,31.829,15.969],
[42.481,30.109,-1.526],[46.470,29.571,15.640],[47.331,30.087,-3.331],
[35.012,26.147,14.539],[38.790,28.050,-3.596],[40.433,26.144,13.602],
[44.248,27.995,-3.129],[45.602,26.320,11.603],[36.530,23.713,13.538],
[39.582,25.526,0.912],[40.661,21.423,12.205],[43.208,23.371,-1.238],
[42.878,25.010,14.478],[51.645,26.683,-3.790],[40.545,18.701,12.512],
[41.303,18.561,9.932],[43.237,24.497,11.296],[47.276,20.985,13.636],
[47.314,24.045,13.070],[38.583,21.538,14.178],[40.911,15.581,10.936],
[42.439,19.839,15.749],[45.523,19.328,12.357],[47.313,18.356,14.406],
[37.879,19.070,13.883],[41.486,15.814,14.140],[43.494,17.826,14.178],
[42.301,33.183,-2.067],[34.850,27.731,-1.707],[34.532,33.635,2.706],
[34.774,31.365,4.880],[51.354,33.175,7.437],[51.661,30.840,9.187],
[48.802,27.734,15.254],[52.287,26.818,-0.043],[51.802,24.411,2.972],
[51.407,28.488,7.558],[49.068,29.998,10.291],[51.128,24.761,-1.451],
[51.974,21.879,-2.172],[49.843,22.640,2.360],[34.460,24.250,8.149],
[49.909,24.968,11.935],[51.366,23.323,14.861],[49.811,19.589,-2.732],
[50.492,20.124,0.088],[48.977,20.970,6.735],[48.367,21.963,10.665],
[33.988,19.348,13.481],[49.445,13.721,15.171],[50.362,17.360,2.131],
[34.695,19.101,6.593],[35.010,21.672,7.577],[50.145,17.495,12.981],
[51.871,17.980,-1.722],[34.675,17.028,1.270],[48.412,31.539,2.546],
[34.522,16.837,10.759],[36.738,29.705,14.457],[39.480,30.318,10.485],
[43.701,31.137,13.062],[46.423,31.243,13.143],[50.556,14.911,10.078],
[35.986,35.912,10.613],[36.959,32.343,10.436],[43.649,34.729,7.646],
[45.483,32.896,9.468],[48.311,32.844,10.845],[35.815,31.905,7.530],
[38.292,30.549,3.972],[41.379,30.722,5.242],[43.474,16.537,5.734],
[50.294,12.211,5.306],[50.784,14.957,5.044],[35.680,15.866,3.608],
[34.671,16.248,-1.464],[41.027,14.972,5.035],[46.112,16.153,4.809],
[47.935,15.136,1.964],[34.940,30.535,-0.996],[37.638,16.288,-1.642],
[44.259,17.957,0.238],[44.815,32.792,-1.203],[49.881,33.904,2.456],
[35.218,15.935,13.101],[38.574,16.593,13.583],[36.984,32.630,-2.573],
[46.189,16.142,15.010],[49.199,30.992,13.954],[45.941,42.905,2.878],
[44.523,42.890,6.658],[37.317,41.773,6.613],[46.715,45.806,3.239],
[37.749,46.928,5.684],[34.992,39.991,1.907],[34.971,42.481,4.575],
[44.961,53.010,5.602],[42.842,47.177,7.304],[50.857,35.736,7.941],
[46.768,38.868,8.803],[38.816,38.030,4.830],[46.144,36.357,3.146],
[34.308,38.968,4.604],[42.324,44.032,4.654],[53.346,39.026,9.515],
[37.307,37.574,8.757],[44.039,46.053,3.190],[40.709,44.606,9.142],
[50.447,38.156,4.890],[37.230,39.280,6.561],[38.256,46.051,10.123],
[37.461,47.513,1.248],[48.116,48.361,6.776],[49.162,37.117,9.613],
[41.603,33.335,6.696],[47.258,50.022,4.731],[48.541,47.561,2.291],
[49.191,45.130,11.209],[44.221,43.884,0.741],[43.929,40.795,-2.052],
[51.993,43.013,6.925],[48.588,43.849,3.316],[45.831,40.408,11.027],
[39.119,42.293,4.024],[40.470,46.829,5.822],[40.355,43.037,6.520],
[47.237,48.273,9.259],[40.322,45.827,11.739],[46.748,42.462,-0.321],
[37.844,42.837,9.869],[44.595,50.365,4.284],[38.650,50.457,13.093],
[49.670,35.768,0.348],[39.125,42.035,0.378],[50.951,49.948,-2.369],
[35.545,44.691,9.712],[34.296,48.344,8.036],[51.251,48.377,11.653],
[42.442,47.521,11.959],[42.502,50.847,1.575],[37.851,51.051,2.571],
[46.671,50.376,-1.132],[45.030,49.823,1.836],[48.963,40.679,12.103],
[35.573,46.033,11.872],[38.620,34.979,8.903],[45.398,33.899,12.229],
[47.212,38.616,12.643],[48.163,43.270,-2.855],[47.274,34.492,-1.146],
[39.088,39.866,9.090],[39.911,40.222,2.351],[44.645,41.338,1.093],
[44.714,40.045,13.869],[46.644,38.772,5.647],[35.874,40.500,9.059],
[39.327,36.910,-3.794],[33.705,36.185,3.750],[34.532,37.804,-1.256],
[44.873,40.920,8.417],[40.510,35.831,4.961],[36.739,40.186,4.077],
[39.224,38.922,-0.038],[44.794,37.083,7.728],[39.020,39.921,-2.544],
[50.152,52.788,13.153],[37.839,45.902,13.010],[42.050,39.627,4.740],
[37.344,36.659,0.071],[42.008,39.247,8.686],[42.472,45.835,0.327],
[48.684,45.093,8.051],[45.524,35.935,-2.259],[44.349,45.393,12.295],
[46.803,38.193,-1.570],[37.217,44.243,5.497],[45.622,46.219,0.767],
[35.773,41.775,11.835],[42.098,42.450,13.505],[45.618,47.970,11.507],
[51.343,39.919,0.405],[52.113,43.956,11.364],[49.911,47.224,-0.211],
[46.794,43.099,5.258],[41.318,47.153,2.735],[35.255,43.710,-1.403],
[39.620,44.893,4.084],[41.587,49.679,9.755],[40.536,50.109,3.247],
[46.107,45.555,10.141],[46.739,45.382,-1.568],[46.166,50.205,13.161],
[51.284,52.325,-1.288],[35.911,38.731,11.097],[48.228,44.836,0.689],
[36.584,48.079,9.440],[41.060,47.531,-4.190],[40.549,47.142,-1.377],
[37.895,44.454,0.420],[43.612,36.177,9.961],[44.898,35.928,0.204],
[43.883,47.768,-1.288],[44.600,38.792,2.026],[43.103,40.765,11.064],
[43.494,36.430,2.646],[40.058,49.916,-1.525],[39.834,36.181,-0.435],
[50.528,37.751,11.610],[48.375,41.544,6.949],[38.002,48.252,-1.248],
[36.383,39.081,-2.756],[48.954,50.097,9.005],[34.817,33.952,5.731],
[38.203,48.394,3.421],[35.548,51.165,14.433],[39.903,51.425,15.615],
[43.925,49.128,-5.024],[47.147,46.203,12.784],[48.420,48.683,-2.409],
[34.958,44.160,14.212],[38.624,45.518,-3.718],[42.060,45.127,13.472],
[43.746,44.997,-4.101],[47.410,42.495,9.302],[37.699,42.787,13.636],
[39.811,43.936,-1.601],[38.073,40.454,11.264],[41.081,41.609,-2.323],
[44.703,42.722,12.415],[50.546,44.685,-2.643],[42.075,37.582,11.881],
[40.181,37.256,9.567],[42.487,42.652,8.403],[47.149,40.179,15.668],
[47.254,42.915,12.041],[40.733,40.199,13.351],[41.348,34.824,9.315],
[42.398,37.771,14.494],[45.329,36.754,12.316],[46.394,37.143,14.731],
[37.471,39.156,13.621],[41.266,33.467,12.359],[42.247,34.938,14.417],
[41.324,52.062,-2.168],[34.920,47.308,-1.311],[33.386,52.368,2.811],
[35.217,50.625,4.621],[51.064,50.810,7.353],[51.890,51.022,12.142],
[49.473,46.206,14.180],[50.828,45.451,1.578],[50.837,42.157,2.925],
[51.469,45.999,7.126],[50.011,47.121,9.252],[52.034,42.647,0.041],
[51.344,38.901,-2.197],[49.597,39.850,2.908],[34.990,43.723,6.798],
[49.721,43.506,13.164],[51.472,41.281,13.755],[49.178,36.971,-2.790],
[48.879,38.946,0.486],[48.893,38.199,7.119],[49.636,40.059,9.187],
[33.795,39.061,12.826],[52.144,34.270,15.180],[51.288,37.302,2.493],
[35.495,36.885,5.891],[32.934,41.323,7.031],[51.548,36.338,13.750],
[51.914,36.310,-1.902],[34.070,36.195,0.842],[49.866,51.679,2.021],
[35.983,34.632,8.218],[36.478,47.708,14.300],[38.939,49.729,10.230],
[44.262,50.397,11.349],[47.698,52.378,13.553],[50.231,34.719,11.751],
[36.838,52.204,9.308],[38.692,51.347,7.547],[41.989,52.347,7.960],
[46.207,51.041,7.561],[47.732,52.370,9.804],[36.017,50.360,7.246],
[38.322,49.657,5.747],[42.333,49.054,4.775],[43.530,36.341,5.235],
[47.477,33.216,4.596],[50.973,35.087,4.675],[37.856,35.529,3.183],
[35.029,33.122,-0.216],[41.817,33.519,3.937],[46.672,35.608,5.726],
[47.123,34.370,1.633],[36.885,50.626,0.114],[36.008,35.201,-1.784],
[42.411,37.612,0.245],[43.842,51.074,-1.563],[51.015,49.930,0.325],
[37.764,36.531,12.967],[38.733,33.358,13.611],[36.633,51.031,-2.910],
[45.048,34.553,15.268],[50.479,49.630,14.142],[51.444,54.842,8.057],
[46.804,59.360,7.860],[38.447,55.778,5.268],[47.076,57.353,1.837],
[36.633,58.940,4.350],[50.658,57.722,9.666],[36.369,55.765,8.604],
[49.122,56.511,4.264],[36.727,57.827,6.867],[49.602,55.348,10.489],
[42.686,52.299,4.389],[50.675,54.127,0.872],[49.189,58.081,14.560],
[40.204,53.707,6.210],[45.414,53.418,10.653],[47.501,57.368,11.746],
[45.206,55.095,-2.236],[39.297,58.711,6.995],[39.890,58.008,4.274],
[45.486,58.026,4.260],[37.881,59.165,9.786],[39.868,55.111,-4.303],
[33.573,55.737,4.500],[38.573,56.209,-1.007],[42.509,57.430,4.658],
[36.231,56.708,2.953],[40.295,57.645,0.532],[46.985,55.539,9.443],
[40.488,57.791,-3.245],[45.322,57.485,9.990],[38.471,54.802,2.644],
[42.938,57.209,8.035],[48.137,54.515,-0.963],[45.454,57.531,-0.467],
[51.549,59.046,0.192],[35.970,55.453,13.068],[43.849,54.242,8.163],
[44.965,55.022,0.847],[44.835,59.291,1.587],[42.869,56.336,10.881],
[43.876,54.762,3.302],[39.968,54.362,0.371],[49.823,55.681,13.255],
[36.178,57.064,-2.279],[37.063,53.559,6.778],[39.882,56.841,11.884],
[39.859,55.766,9.121],[45.029,57.115,15.085],[39.393,56.733,16.300],
[39.977,52.346,9.638],[42.418,57.242,13.753],[44.553,59.400,12.532],
[47.176,55.570,15.481],[36.796,54.332,15.541],[42.360,52.141,11.177],
[42.803,53.904,14.670],[51.046,-59.141,-2.518],[49.209,56.307,-2.601],
[49.945,56.753,-0.083],[49.339,58.438,7.120],[48.542,-59.088,9.955],
[33.649,57.162,13.790],[51.279,53.753,15.311],[48.555,53.572,3.674],
[35.731,55.586,5.776],[33.350,58.043,7.364],[52.248,54.337,12.751],
[51.992,54.937,-1.544],[34.042,56.461,-0.352],[35.156,54.066,10.360],
[50.469,52.732,10.453],[45.119,55.768,5.919],[49.799,50.951,4.659],
[50.259,54.287,5.605],[35.839,53.961,3.184],[35.174,52.842,0.806],
[40.988,53.783,2.954],[48.074,52.761,6.313],[47.101,52.365,0.992],
[36.497,54.464,-0.956],[42.679,53.842,-0.771],[37.686,53.252,12.231],
[39.403,54.281,15.266],[45.123,52.333,15.121],[44.256,-48.501,21.600],
[43.875,-49.256,24.412],[36.828,-51.171,25.855],[48.559,-47.410,23.845],
[38.024,-47.567,26.050],[35.511,-52.206,20.539],[34.842,-49.786,21.884],
[47.072,-42.885,24.752],[41.808,-47.952,25.736],[49.447,-57.424,26.647],
[46.423,-53.473,26.681],[39.485,-54.303,23.764],[47.711,-54.448,21.118],
[35.739,-52.490,23.611],[42.004,-53.573,22.767],[53.545,-54.519,27.702],
[37.411,-56.364,25.439],[45.919,-44.962,20.537],[39.726,-49.448,28.022],
[47.962,-55.784,23.661],[35.754,-54.217,26.103],[39.387,-46.393,28.554],
[34.106,-46.215,19.138],[47.354,-45.455,26.170],[48.496,-55.353,28.115],
[42.213,59.190,26.166],[47.190,-45.043,22.755],[47.620,-45.430,18.501],
[49.598,-47.705,29.647],[44.022,-50.956,19.088],[45.933,-52.516,16.466],
[51.002,-50.343,27.093],[48.599,-49.405,21.286],[47.426,-50.467,27.843],
[39.840,-49.023,19.610],[39.916,-46.791,24.181],[39.759,-51.274,25.365],
[44.928,-44.201,25.737],[38.384,-46.108,31.161],[46.735,-49.803,18.671],
[37.768,-50.283,29.983],[43.772,-43.476,21.566],[38.032,-44.179,33.073],
[50.607,-57.899,18.828],[37.499,-49.948,20.476],[50.144,-43.298,17.042],
[34.786,-47.543,28.408],[34.543,-45.579,26.305],[53.306,-43.656,31.290],
[40.758,-44.417,31.700],[43.920,-41.336,19.325],[37.754,-43.528,20.021],
[45.116,-43.426,18.001],[46.456,-42.009,20.177],[49.544,-51.539,32.028],
[35.657,-45.584,30.644],[37.578,-58.918,27.028],[44.098,-58.232,28.776],
[47.019,-54.655,30.149],[49.082,-49.012,16.502],[43.830,-54.371,15.792],
[39.302,-53.931,26.441],[40.228,-51.556,20.944],[45.190,-52.855,20.800],
[45.106,-52.439,32.224],[45.265,-53.911,23.428],[36.980,-52.416,28.207],
[40.075,-54.606,15.104],[33.819,-54.065,22.180],[33.169,-54.890,16.584],
[44.356,-51.647,27.679],[41.710,-54.959,25.246],[35.876,-55.171,20.893],
[38.288,-54.797,18.477],[46.460,-56.372,26.593],[38.282,-52.383,16.833],
[48.823,-40.783,31.633],[36.255,-48.195,30.908],[43.076,-51.930,24.966],
[38.559,-56.505,22.294],[42.355,-53.585,27.588],[42.516,-48.445,18.676],
[48.916,-48.534,27.135],[44.108,-54.563,18.668],[44.144,-47.013,30.022],
[47.214,-52.628,18.999],[36.679,-49.438,23.833],[45.142,-47.454,19.038],
[33.714,-50.501,30.735],[40.429,-50.358,31.402],[44.851,-43.439,29.896],
[51.350,-52.725,19.165],[51.374,-46.075,30.776],[50.384,-45.821,18.110],
[45.716,-51.189,24.527],[39.697,-44.565,21.913],[36.625,-49.709,16.166],
[41.323,-47.400,21.769],[41.017,-43.704,26.906],[38.806,-41.948,22.205],
[45.876,-48.807,26.415],[46.575,-47.822,16.514],[48.260,-43.798,30.821],
[52.690,-41.739,17.392],[35.870,-55.469,29.807],[48.813,-47.738,19.387],
[35.263,-43.251,28.289],[40.941,-46.445,15.242],[40.167,-45.987,18.953],
[37.219,-47.123,19.525],[43.533,-55.078,29.307],[45.352,-56.974,19.676],
[43.075,-45.160,18.520],[41.209,-54.514,19.972],[41.169,-55.507,31.378],
[43.969,-55.414,21.641],[40.243,-43.216,17.821],[40.776,-56.083,17.590],
[49.964,-53.546,30.422],[48.822,-52.893,23.989],[38.426,-46.584,16.996],
[34.861,-53.032,16.655],[50.219,-43.286,26.793],[37.594,-59.486,23.311],
[36.098,-45.400,21.263],[34.547,-43.618,34.193],[39.342,-41.746,33.395],
[43.333,-44.354,15.778],[45.764,-45.683,32.157],[46.390,-44.845,15.939],
[33.783,-49.138,32.975],[40.105,-48.699,13.748],[41.960,-48.037,31.054],
[43.708,-47.273,15.927],[47.077,-48.988,30.704],[36.157,-50.068,32.790],
[41.059,-49.321,16.318],[40.289,-52.829,29.519],[41.187,-52.266,15.556],
[45.241,-49.110,34.137],[51.758,-48.499,15.699],[38.682,-54.648,30.859],
[40.707,-57.111,28.369],[42.696,-50.845,30.232],[49.676,-54.542,35.841],
[47.617,-51.699,30.233],[38.050,-52.059,33.163],[40.551,-59.557,29.385],
[42.885,-54.021,32.922],[45.516,-56.983,31.113],[46.788,-56.125,33.199],
[40.169,-53.540,33.702],[41.145,-58.321,31.887],[41.625,-59.592,36.693],
[41.478,-39.965,17.652],[35.735,-46.471,16.914],[34.357,-40.992,20.996],
[33.703,-43.084,23.156],[52.154,-43.405,24.185],[49.729,-41.951,29.188],
[50.579,-46.346,33.439],[52.542,-47.745,18.233],[51.553,-49.577,21.243],
[50.018,-46.056,25.663],[51.171,-45.089,28.369],[53.325,-50.806,17.968],
[51.087,-53.496,16.533],[49.975,-51.825,21.676],[34.375,-50.354,25.256],
[49.861,-48.849,32.113],[52.205,-50.940,32.602],[49.629,-56.030,16.973],
[51.535,-54.279,21.536],[51.348,-53.121,24.930],[51.885,-52.647,28.756],
[34.176,-54.957,32.019],[53.531,-57.582,32.944],[50.537,-56.793,21.484],
[34.154,-56.335,25.334],[32.987,-52.589,26.108],[50.874,-56.469,32.082],
[53.106,-55.264,18.085],[35.227,-56.503,18.461],[48.651,-41.101,21.229],
[34.808,-57.510,28.250],[34.872,-46.552,33.139],[37.331,-44.731,28.831],
[42.735,-41.789,33.383],[45.659,-42.063,32.775],[50.209,-59.056,30.499],
[36.152,-40.418,29.714],[38.263,-42.636,27.178],[43.662,-40.113,26.126],
[47.202,-42.554,28.014],[48.922,-40.895,26.633],[36.069,-40.891,27.135],
[37.994,-44.760,23.930],[41.208,-43.499,23.985],[43.639,-57.089,23.777],
[49.579,-58.641,23.323],[52.353,59.542,21.624],[34.183,-57.075,21.389],
[36.363,-59.524,18.527],[41.378,-59.446,22.587],[46.577,-58.342,23.892],
[48.009,-57.435,20.329],[33.628,-43.699,17.847],[37.142,-57.335,16.978],
[41.357,-56.310,22.298],[43.277,-41.528,16.355],[51.495,-42.799,19.501],
[35.059,-58.757,30.613],[38.947,-58.610,34.479],[36.232,-41.250,16.190],
[48.718,-58.235,32.927],[50.366,-43.914,32.303],[44.850,-30.354,23.374],
[45.075,-31.897,25.674],[38.134,-30.805,26.144],[49.763,-28.607,24.659],
[38.199,-28.457,24.770],[34.688,-33.521,20.531],[34.450,-30.098,21.441],
[45.828,-23.231,23.373],[42.744,-30.249,25.286],[51.673,-37.463,27.452],
[46.696,-35.643,26.274],[37.856,-37.122,23.476],[47.020,-38.424,21.888],
[35.661,-33.221,23.751],[43.584,-32.754,23.363],[51.267,-34.695,29.336],
[37.858,-37.325,26.255],[43.705,-28.707,21.258],[40.901,-30.828,27.596],
[48.633,-37.214,23.471],[37.799,-34.643,24.714],[37.390,-29.249,28.337],
[35.219,-26.820,19.913],[47.480,-26.206,24.990],[49.095,-37.546,28.059],
[41.535,-40.695,24.514],[47.999,-24.967,22.197],[46.623,-27.271,20.493],
[49.257,-28.977,32.093],[44.788,-31.199,20.773],[45.279,-33.518,16.143],
[50.875,-32.067,28.032],[48.521,-29.729,21.757],[46.276,-33.183,30.112],
[39.607,-30.785,23.299],[42.553,-26.879,25.249],[40.169,-32.714,25.391],
[46.771,-26.202,27.893],[40.422,-28.638,29.030],[46.812,-32.405,17.918],
[38.758,-31.033,30.037],[44.336,-25.861,21.010],[40.860,-25.340,32.661],
[49.447,-40.943,18.143],[38.163,-31.866,18.461],[51.835,-26.577,16.590],
[34.052,-28.921,27.963],[33.902,-26.484,26.902],[53.128,-25.989,30.567],
[43.235,-28.735,30.067],[43.769,-22.377,20.968],[38.322,-23.591,23.884],
[46.135,-22.648,17.173],[46.120,-24.066,20.354],[49.512,-36.302,31.403],
[34.980,-28.695,31.184],[38.895,-39.917,26.724],[43.486,-40.498,30.167],
[46.303,-35.709,30.537],[47.906,-32.355,15.420],[42.471,-36.268,17.945],
[39.203,-34.815,27.692],[40.575,-34.753,19.841],[43.895,-33.128,19.129],
[44.307,-34.456,32.068],[46.498,-35.277,22.532],[37.234,-33.373,28.963],
[38.665,-37.024,16.772],[32.933,-39.239,22.567],[35.303,-36.122,19.001],
[45.555,-33.635,27.494],[40.447,-37.215,23.574],[37.606,-31.530,21.410],
[40.136,-37.371,19.288],[46.651,-38.268,26.752],[40.097,-34.361,16.867],
[50.339,-23.485,31.699],[37.695,-29.971,32.989],[42.706,-34.527,25.313],
[37.550,-37.536,20.663],[41.648,-34.376,27.525],[42.923,-30.208,18.644],
[49.427,-30.938,26.195],[46.524,-36.913,17.301],[41.949,-31.321,30.263],
[46.415,-34.217,19.700],[36.550,-30.294,23.531],[44.456,-27.866,18.396],
[33.526,-31.098,30.584],[41.889,-33.110,32.644],[45.514,-25.730,30.341],
[52.225,-32.054,18.251],[51.869,-29.166,30.769],[49.412,-25.945,19.734],
[47.338,-31.919,22.910],[39.650,-27.322,21.283],[35.577,-31.661,16.762],
[40.773,-29.778,20.713],[42.493,-24.785,27.060],[38.709,-24.326,20.911],
[44.805,-28.565,27.220],[47.154,-26.805,17.535],[47.991,-24.450,29.883],
[51.978,-23.681,15.597],[36.589,-36.107,28.356],[46.358,-29.528,19.209],
[36.549,-26.699,26.982],[40.700,-27.016,16.238],[40.782,-27.997,18.572],
[35.966,-29.296,18.233],[43.519,-38.357,28.447],[45.104,-38.782,19.437],
[43.443,-24.524,18.883],[42.295,-35.948,21.531],[44.334,-35.669,28.961],
[43.961,-38.305,22.038],[40.190,-25.192,18.328],[38.306,-39.862,18.043],
[47.418,-37.938,30.419],[48.648,-33.798,24.608],[37.337,-26.897,17.525],
[36.327,-34.636,17.051],[48.493,-24.370,27.035],[37.673,-41.910,24.845],
[37.168,-26.982,22.271],[34.414,-25.439,32.693],[38.974,-23.679,33.342],
[44.720,-25.829,16.498],[46.326,-28.863,32.553],[47.842,-25.169,15.491],
[34.970,-30.832,33.136],[37.419,-27.381,14.851],[42.196,-30.098,32.744],
[42.784,-27.008,14.670],[46.844,-29.143,29.954],[36.665,-32.948,32.234],
[39.884,-31.104,16.597],[40.508,-33.659,30.059],[42.738,-32.163,16.815],
[45.414,-31.919,32.722],[50.451,-31.200,15.539],[40.567,-37.695,30.760],
[40.639,-37.545,27.178],[43.461,-33.279,29.372],[46.431,-35.563,34.358],
[47.854,-33.512,32.470],[39.283,-34.607,32.281],[41.347,-40.731,28.303],
[42.729,-36.671,32.879],[45.153,-38.068,31.792],[48.271,-38.457,33.150],
[36.923,-35.598,31.209],[41.221,-40.629,31.505],[40.463,-38.536,33.495],
[41.334,-22.531,18.690],[35.048,-28.075,15.760],[34.466,-22.502,20.986],
[33.995,-22.688,23.573],[50.850,-24.119,25.553],[50.910,-24.848,28.971],
[51.190,-27.294,33.448],[50.617,-28.629,19.394],[50.650,-31.553,21.122],
[52.263,-29.858,27.991],[49.334,-27.252,28.938],[49.415,-31.707,18.510],
[51.235,-35.012,18.039],[48.797,-34.673,20.982],[35.495,-31.345,25.964],
[50.003,-30.990,30.474],[52.612,-32.661,33.009],[49.125,-36.583,15.141],
[47.249,-36.839,19.824],[49.717,-36.027,25.937],[48.836,-33.702,27.428],
[34.271,-35.730,32.273],[52.381,-39.555,34.704],[49.892,-38.135,19.773],
[35.806,-38.442,24.287],[34.108,-33.927,26.446],[50.918,-38.733,32.236],
[52.248,-37.026,16.483],[35.132,-37.967,21.328],[48.846,-23.378,19.667],
[33.875,-38.529,28.987],[36.732,-27.404,33.051],[38.868,-26.282,28.644],
[43.684,-24.009,29.495],[47.943,-21.633,31.325],[51.171,-39.488,29.473],
[36.733,-22.151,27.783],[38.866,-23.513,28.629],[44.244,-21.033,24.769],
[45.373,-24.410,26.389],[47.271,-21.456,26.493],[36.004,-24.834,24.786],
[39.816,-26.503,23.793],[40.988,-24.065,24.826],[42.502,-37.370,25.341],
[49.701,-42.348,24.164],[50.380,-39.709,22.488],[36.850,-40.047,22.189],
[34.239,-41.141,18.452],[40.693,-39.505,21.953],[47.603,-39.610,24.571],
[46.856,-40.179,17.789],[33.917,-25.560,17.990],[36.797,-38.530,15.465],
[42.338,-38.623,19.807],[43.395,-23.218,16.523],[51.004,-24.315,18.572],
[35.936,-38.394,31.337],[37.320,-39.471,33.426],[36.489,-24.143,15.137],
[48.107,-42.165,34.055],[49.845,-26.162,31.340],[45.228,-12.520,22.503],
[44.779,-12.528,25.382],[37.576,-14.044,26.294],[48.453,-8.702,23.109],
[36.946,-8.880,25.051],[32.836,-16.731,20.276],[34.993,-12.794,22.225],
[44.551,-6.603,25.559],[43.884,-9.926,25.600],[51.711,-19.179,26.044],
[46.536,-16.310,26.059],[38.893,-16.659,23.398],[47.077,-18.762,21.893],
[33.201,-14.785,22.254],[42.449,-13.022,23.347],[52.703,-18.493,28.945],
[38.585,-18.953,25.098],[43.815,-9.966,21.001],[40.350,-11.910,27.742],
[50.302,-19.311,23.337],[37.752,-14.197,23.387],[36.944,-9.688,28.221],
[34.902,-9.083,18.407],[48.087,-8.970,25.619],[48.810,-19.946,28.900],
[40.776,-20.626,24.663],[47.331,-5.195,21.389],[46.305,-7.738,21.210],
[49.469,-10.296,29.688],[43.567,-12.663,18.644],[45.721,-14.745,16.070],
[50.228,-12.663,28.194],[48.266,-13.581,20.431],[46.502,-15.108,28.971],
[39.251,-12.636,21.789],[40.178,-9.208,24.416],[40.430,-14.151,25.322],
[45.795,-8.321,28.022],[39.665,-9.340,29.542],[46.150,-12.118,17.975],
[37.176,-12.931,28.891],[44.501,-5.730,22.100],[38.972,-5.891,31.030],
[50.365,-22.135,17.377],[38.680,-12.687,19.059],[51.450,-7.421,15.553],
[34.433,-11.527,28.451],[34.643,-7.384,25.033],[51.216,-8.597,32.586],
[41.938,-6.952,31.356],[43.323,-4.166,19.227],[37.795,-5.990,21.094],
[48.153,-5.365,16.787],[45.681,-5.934,19.262],[50.189,-15.075,29.363],
[37.555,-8.193,30.627],[39.095,-21.235,26.771],[44.538,-21.809,28.143],
[48.110,-17.563,30.172],[48.205,-12.834,16.648],[44.042,-17.978,17.557],
[40.423,-15.306,28.017],[40.405,-15.688,21.383],[45.632,-14.394,19.763],
[44.110,-15.135,32.939],[47.398,-15.932,22.136],[36.820,-15.901,28.003],
[41.435,-18.743,15.895],[33.721,-17.466,23.123],[34.065,-17.993,17.923],
[44.424,-14.360,27.231],[41.155,-19.493,22.316],[37.084,-16.286,20.961],
[38.858,-18.202,17.670],[45.901,-19.018,24.819],[39.664,-15.354,16.622],
[48.723,-3.884,30.747],[37.741,-11.127,31.277],[43.899,-15.866,25.056],
[38.219,-19.450,21.807],[41.909,-17.402,26.247],[41.608,-11.572,19.801],
[49.140,-12.253,25.698],[46.075,-19.804,16.623],[44.283,-10.182,29.584],
[46.546,-16.749,17.677],[37.166,-11.785,24.266],[44.609,-9.796,18.174],
[34.821,-14.586,29.291],[40.445,-14.089,32.878],[45.041,-7.808,30.668],
[52.737,-15.311,18.155],[48.817,-10.480,32.307],[48.693,-7.326,20.278],
[46.870,-11.321,24.568],[39.076,-8.106,22.219],[36.753,-13.749,16.776],
[41.371,-9.784,21.966],[41.712,-7.522,26.764],[40.471,-5.235,20.460],
[46.560,-11.134,27.522],[46.808,-9.626,16.415],[47.943,-7.556,30.734],
[52.619,-5.022,16.780],[34.373,-17.137,29.056],[47.981,-10.440,19.079],
[35.686,-7.178,28.522],[41.771,-8.691,16.205],[39.424,-9.405,17.467],
[37.249,-11.171,17.471],[43.529,-18.891,29.827],[45.712,-19.572,19.311],
[43.192,-7.412,19.015],[43.567,-16.823,20.151],[43.634,-16.052,29.823],
[43.247,-19.686,20.691],[41.000,-5.730,17.757],[39.130,-20.964,18.671],
[50.786,-17.715,30.826],[48.291,-14.198,24.112],[37.028,-8.957,15.773],
[36.844,-16.500,16.529],[48.280,-5.469,27.603],[35.076,-21.477,25.731],
[36.543,-9.058,21.282],[36.046,-5.635,32.231],[40.864,-5.787,33.843],
[43.745,-7.021,14.504],[45.909,-5.881,32.255],[48.673,-7.703,15.543],
[35.426,-12.840,31.702],[39.584,-9.090,14.482],[41.293,-11.161,31.200],
[44.152,-9.855,15.218],[46.849,-12.397,29.794],[36.827,-14.612,33.680],
[41.037,-11.444,16.898],[40.679,-13.807,30.108],[43.736,-12.769,15.616],
[45.518,-12.581,33.514],[52.597,-13.070,16.297],[38.984,-18.372,30.765],
[39.492,-18.090,28.073],[43.385,-11.817,27.557],[47.905,-16.873,33.916],
[47.900,-15.071,31.321],[38.601,-15.989,34.845],[41.878,-20.981,28.225],
[41.786,-17.359,31.834],[45.854,-16.981,31.128],[45.703,-19.742,31.895],
[37.202,-17.333,32.566],[39.969,-20.674,32.492],[42.447,-19.881,33.378],
[42.074,-2.933,17.123],[34.092,-10.611,16.233],[33.640,-4.713,19.753],
[35.085,-6.396,22.698],[52.971,-4.007,26.804],[51.094,-5.995,29.779],
[49.514,-10.129,35.280],[50.391,-11.541,18.942],[51.102,-12.318,21.250],
[51.139,-10.119,26.881],[48.716,-8.190,28.232],[50.366,-13.847,17.531],
[50.723,-16.517,15.437],[51.367,-15.914,20.496],[34.404,-12.540,25.343],
[50.723,-12.911,30.981],[54.063,-14.942,32.276],[48.697,-19.657,15.994],
[49.621,-16.327,17.863],[49.461,-17.683,27.376],[48.833,-15.180,26.697],
[34.473,-18.012,31.573],[53.523,-21.859,32.447],[50.194,-18.512,19.377],
[36.280,-17.552,24.196],[34.391,-15.279,26.069],[52.405,-19.580,31.561],
[51.617,-20.267,15.925],[32.597,-19.464,19.560],[49.893,-6.318,22.632],
[34.638,-20.076,28.665],[38.311,-8.150,33.379],[38.903,-7.631,27.325],
[44.492,-4.178,30.603],[46.460,-2.897,31.701],[51.546,-21.785,28.168],
[36.443,-1.902,28.754],[39.427,-4.741,27.592],[43.078,-2.883,26.996],
[45.472,-5.351,28.357],[46.933,-2.977,28.032],[36.984,-6.132,26.117],
[38.968,-6.775,24.733],[41.692,-6.249,24.075],[44.645,-17.816,22.619],
[49.878,-21.098,25.706],[50.409,-21.796,22.492],[36.552,-21.381,22.742],
[34.464,-22.845,18.136],[41.548,-22.612,22.480],[47.426,-20.960,23.532],
[48.022,-20.906,18.954],[35.352,-6.453,19.047],[36.909,-21.822,16.652],
[41.399,-19.261,18.941],[44.690,-4.994,17.079],[50.867,-6.229,18.720],
[36.977,-22.674,30.646],[37.545,-21.465,33.540],[37.390,-5.555,16.067],
[45.682,-22.220,33.936],[49.802,-6.128,32.639],[46.149,6.600,20.705],
[44.433,6.471,25.271],[37.986,5.039,26.484],[48.641,9.213,23.962],
[38.814,8.886,24.099],[34.616,2.551,20.524],[35.147,4.946,22.071],
[45.373,13.862,23.983],[44.941,9.136,26.079],[51.973,0.810,28.057],
[45.925,2.195,26.682],[39.787,0.669,23.371],[47.442,-0.307,22.633],
[35.742,2.733,23.502],[42.361,5.512,23.548],[51.445,2.669,30.132],
[36.478,0.049,27.025],[45.967,8.130,23.176],[39.657,6.878,27.946],
[49.869,-0.769,24.282],[36.841,2.654,26.346],[37.613,8.828,27.073],
[34.038,9.611,20.363],[48.262,8.518,26.450],[48.367,-0.991,26.696],
[40.905,-3.698,25.040],[46.597,11.367,23.832],[48.353,10.196,21.624],
[49.272,8.014,30.112],[44.345,3.756,18.750],[45.645,3.999,16.234],
[51.511,5.080,28.860],[48.586,5.641,20.403],[45.468,4.179,30.681],
[39.524,7.227,22.005],[41.468,10.312,24.517],[40.107,6.314,24.807],
[47.489,10.979,26.517],[39.204,9.190,29.348],[46.636,5.723,18.156],
[38.256,3.556,30.843],[44.597,12.416,21.486],[38.674,12.086,30.766],
[50.938,-3.684,18.260],[39.754,4.462,19.857],[50.097,11.619,16.744],
[34.818,7.775,27.641],[34.985,12.442,26.261],[52.652,12.671,31.477],
[41.790,10.332,29.670],[45.232,15.270,21.042],[39.210,13.569,21.815],
[48.701,15.297,17.381],[46.780,13.433,19.832],[48.836,4.355,29.899],
[36.209,10.100,30.324],[38.614,-2.240,27.190],[44.866,-1.648,29.746],
[48.517,1.297,31.325],[49.450,5.600,17.011],[44.177,1.037,17.218],
[39.756,2.124,26.012],[41.346,3.105,21.879],[46.497,3.870,21.154],
[44.118,4.245,32.917],[46.263,1.909,23.130],[35.964,2.589,28.910],
[40.650,1.081,16.125],[34.234,-0.156,21.792],[37.468,2.811,19.148],
[43.309,3.882,27.837],[42.491,-1.074,22.282],[38.608,2.753,21.846],
[40.174,-1.063,20.542],[45.746,-0.560,27.385],[40.638,1.560,18.973],
[47.649,15.959,32.411],[37.011,7.436,31.268],[43.204,3.291,24.781],
[37.841,-0.379,21.598],[42.024,1.541,29.774],[43.818,8.276,19.447],
[49.551,6.255,27.293],[46.680,-0.400,18.103],[43.927,8.054,29.567],
[47.816,2.183,18.048],[37.222,5.687,23.733],[45.106,7.286,16.813],
[35.760,5.612,29.439],[41.774,5.730,32.211],[45.834,9.594,30.644],
[54.868,3.958,18.110],[51.029,9.114,31.673],[49.826,10.831,19.364],
[47.182,5.370,27.783],[39.479,10.897,22.341],[35.903,4.879,16.025],
[41.314,8.692,20.371],[42.663,13.440,28.880],[41.594,14.794,20.480],
[46.561,8.487,28.474],[47.342,8.967,16.975],[47.886,13.206,30.503],
[52.740,13.174,16.061],[35.283,0.347,30.646],[48.334,8.228,19.367],
[36.969,11.326,28.132],[41.627,10.501,15.210],[39.847,9.202,17.717],
[37.564,7.554,18.080],[44.952,1.748,29.079],[45.748,-2.736,19.977],
[42.960,10.445,18.075],[43.853,1.820,20.796],[41.818,2.796,32.228],
[45.076,-0.877,21.956],[41.058,11.705,20.093],[39.666,-1.973,17.867],
[49.358,-0.777,29.615],[48.025,3.046,25.145],[38.147,9.894,15.348],
[37.065,2.471,16.302],[49.167,15.199,27.065],[35.947,-3.095,26.210],
[36.987,10.157,20.825],[36.000,12.630,31.811],[39.661,12.558,34.578],
[44.022,10.259,14.208],[46.841,10.408,35.554],[47.714,12.471,17.329],
[34.553,7.356,32.698],[39.172,5.345,17.094],[41.613,8.229,31.470],
[42.257,8.298,16.587],[46.381,6.584,31.643],[36.360,5.051,32.378],
[41.912,5.315,18.505],[41.121,4.837,29.506],[42.631,5.667,15.613],
[44.913,8.569,34.388],[52.190,6.687,14.818],[39.110,1.246,32.081],
[39.982,-0.528,28.751],[42.183,6.606,27.106],[46.893,0.898,34.518],
[46.812,4.663,33.835],[39.308,3.709,33.261],[41.969,-2.551,29.633],
[43.006,0.162,31.679],[45.312,1.545,31.707],[47.323,-1.644,33.908],
[36.488,0.434,34.242],[41.183,-4.456,31.456],[42.873,-3.440,33.004],
[41.226,14.915,16.881],[36.407,12.048,16.500],[33.940,14.933,20.380],
[35.887,13.494,22.188],[50.859,13.606,25.941],[53.317,14.467,29.210],
[47.823,8.546,33.248],[50.991,7.517,19.785],[51.509,5.194,20.811],
[50.875,9.357,26.544],[49.604,12.053,28.391],[52.385,6.799,17.597],
[50.450,2.405,16.603],[48.921,3.946,22.806],[34.454,6.218,25.006],
[50.830,5.921,31.473],[52.843,3.272,33.003],[49.284,-0.345,16.270],
[49.295,2.345,20.511],[50.785,2.556,25.571],[49.164,2.923,27.698],
[33.487,0.669,32.718],[50.375,-3.615,33.608],[51.224,-2.626,20.749],
[37.072,0.430,24.410],[33.542,2.737,25.916],[51.917,-1.902,31.858],
[51.643,-0.006,17.959],[35.914,-1.622,20.111],[49.204,13.499,20.964],
[33.786,-1.867,27.454],[35.488,10.316,33.065],[39.896,11.582,27.694],
[42.969,12.498,31.370],[45.469,13.857,31.893],[50.740,-3.160,28.765],
[36.453,15.034,27.742],[39.012,14.762,28.958],[43.941,14.731,26.290],
[46.191,13.435,28.192],[46.874,16.984,28.792],[36.861,13.486,24.695],
[39.326,11.992,24.783],[42.523,12.586,24.289],[44.971,-0.462,24.807],
[48.507,-3.690,25.542],[51.866,-2.818,23.441],[36.780,-4.144,23.054],
[36.675,-4.013,18.506],[42.343,-4.005,21.985],[46.488,-2.976,24.051],
[48.924,-3.522,20.235],[35.080,11.845,18.770],[37.373,-0.539,17.563],
[42.598,-1.080,19.226],[43.616,12.959,17.432],[51.791,12.721,20.911],
[34.639,-4.037,29.811],[38.878,-4.182,32.928],[36.338,14.859,16.828],
[45.368,-3.481,34.194],[50.064,13.831,32.664],[45.937,23.738,22.797],
[43.846,24.915,25.274],[37.439,24.534,27.073],[47.780,27.605,23.842],
[37.937,27.290,25.069],[35.769,22.369,20.653],[33.095,22.409,22.275],
[46.173,32.046,23.149],[42.443,27.066,25.801],[52.314,17.590,26.841],
[47.633,19.667,28.501],[39.644,19.577,22.956],[46.543,20.219,21.379],
[35.309,20.917,23.161],[40.048,23.061,23.239],[52.166,21.808,29.022],
[37.510,18.351,28.216],[45.065,27.565,21.896],[39.464,25.836,28.949],
[48.297,17.794,22.819],[37.782,22.054,24.904],[37.484,27.892,27.945],
[36.071,28.954,19.292],[47.828,27.908,26.799],[49.937,17.969,27.673],
[41.499,15.988,25.107],[47.063,30.521,25.303],[47.109,28.835,19.703],
[48.608,27.701,29.540],[45.212,24.897,20.544],[44.973,22.472,16.022],
[51.352,24.880,27.815],[48.323,24.665,21.676],[47.313,22.404,29.117],
[37.930,24.715,23.040],[40.395,28.511,25.011],[39.885,23.848,26.245],
[44.935,29.171,26.741],[40.458,28.441,29.472],[47.613,22.702,17.894],
[38.112,23.788,30.205],[43.287,31.824,22.115],[39.236,31.486,31.124],
[50.966,14.670,18.631],[38.473,23.634,20.137],[51.577,29.051,16.211],
[34.861,26.650,27.140],[33.611,30.378,26.005],[51.740,30.675,30.370],
[43.132,29.250,29.598],[43.611,33.206,19.687],[36.906,32.907,22.705],
[47.943,31.245,19.329],[45.833,32.219,20.560],[50.288,21.841,31.090],
[36.333,27.110,30.355],[39.362,16.139,26.438],[44.311,15.751,28.726],
[47.554,20.879,31.353],[49.238,24.882,15.899],[44.583,17.993,16.754],
[38.691,21.356,27.268],[41.376,21.268,21.771],[44.388,22.250,19.621],
[45.105,22.304,31.829],[44.432,20.782,23.235],[38.037,20.719,29.807],
[39.625,17.893,16.607],[33.603,18.773,22.607],[36.132,19.154,18.148],
[44.271,23.014,27.323],[42.080,18.466,23.487],[37.121,20.017,21.286],
[39.601,20.600,19.605],[46.673,18.107,25.971],[39.862,22.411,17.723],
[48.594,32.796,32.745],[35.325,25.824,32.324],[42.238,22.455,24.899],
[38.662,17.505,21.400],[41.306,21.049,27.786],[42.614,26.117,19.711],
[48.829,25.090,26.768],[47.882,18.034,18.046],[45.234,24.129,29.820],
[46.871,20.577,16.511],[36.295,26.571,23.092],[44.909,27.336,19.191],
[33.982,25.651,29.661],[41.417,24.569,31.227],[45.995,29.137,29.640],
[51.572,22.152,18.866],[51.774,27.939,30.147],[49.890,29.320,19.990],
[46.419,24.186,25.667],[39.416,28.993,21.053],[33.720,26.095,17.001],
[39.545,26.291,21.855],[43.079,32.353,29.669],[39.156,32.303,21.194],
[46.044,26.454,28.563],[47.218,27.171,17.380],[48.527,31.327,30.179],
[52.267,31.735,16.537],[34.917,19.289,29.412],[48.643,26.159,19.285],
[35.911,29.961,27.622],[41.266,28.475,15.085],[40.032,27.175,19.136],
[37.287,26.778,18.189],[43.646,19.651,27.943],[45.744,17.603,19.485],
[42.367,29.041,19.679],[43.311,19.478,21.092],[42.707,20.787,30.089],
[43.164,15.394,22.770],[40.186,30.165,18.626],[40.285,16.714,18.873],
[49.234,18.891,30.564],[47.050,21.309,24.442],[38.899,28.937,16.158],
[37.993,20.273,16.641],[49.360,31.081,27.095],[36.739,16.978,25.679],
[36.033,29.596,22.359],[34.924,32.415,32.486],[40.127,32.309,33.691],
[43.661,29.351,16.631],[45.854,28.592,32.471],[49.051,30.183,16.943],
[33.162,25.011,33.810],[39.591,26.301,16.409],[40.224,26.804,32.049],
[44.546,27.446,14.509],[47.269,26.239,31.415],[36.217,23.205,32.528],
[41.787,24.739,17.398],[40.421,21.949,30.205],[40.324,23.794,15.441],
[45.216,25.530,32.804],[51.665,26.397,15.381],[40.574,19.202,30.923],
[40.396,18.386,28.411],[41.827,24.689,28.222],[47.324,21.497,34.113],
[48.073,23.587,31.605],[38.641,21.800,32.510],[41.770,15.742,27.865],
[42.635,19.427,33.209],[44.929,19.441,30.763],[46.085,18.901,33.163],
[38.161,19.213,32.843],[40.334,14.298,31.571],[39.926,16.855,32.559],
[40.935,32.990,17.823],[35.516,28.322,16.649],[33.582,32.894,19.916],
[34.631,32.006,22.331],[50.836,33.508,25.714],[52.053,31.528,27.907],
[49.181,28.109,34.160],[51.200,27.218,18.343],[51.045,24.314,20.703],
[51.069,26.750,25.115],[50.617,28.810,27.850],[51.715,24.533,17.831],
[51.815,20.629,16.597],[49.374,20.900,22.003],[34.062,25.646,24.217],
[49.602,25.156,29.974],[52.946,22.473,31.779],[49.759,18.909,15.360],
[49.884,19.699,18.176],[49.850,21.049,24.823],[49.611,21.385,27.653],
[35.771,20.435,31.633],[53.604,14.262,35.052],[50.918,18.259,20.843],
[37.192,19.156,24.134],[33.387,22.942,25.215],[53.076,16.404,31.543],
[54.054,15.496,15.097],[35.931,16.922,20.163],[48.378,33.105,21.506],
[34.673,17.141,27.621],[36.757,28.987,32.613],[38.740,31.175,28.186],
[42.256,30.864,32.155],[46.097,32.628,30.311],[49.611,15.604,29.990],
[36.109,33.870,29.655],[37.071,33.405,26.894],[45.113,33.714,27.347],
[46.681,31.295,27.888],[47.983,33.485,28.262],[36.283,31.375,24.894],
[38.915,30.899,23.504],[42.503,30.246,25.537],[43.971,17.635,25.677],
[48.557,15.010,24.186],[51.379,15.908,24.452],[36.132,16.263,22.696],
[36.726,13.910,19.583],[40.674,16.488,22.602],[45.798,16.446,23.739],
[47.964,15.929,19.873],[34.126,30.815,18.472],[36.808,17.589,16.052],
[43.177,17.567,18.999],[44.849,30.451,18.808],[50.585,31.915,19.241],
[35.887,15.375,30.780],[37.126,16.785,32.935],[36.689,32.566,15.164],
[48.126,14.698,34.866],[49.896,28.786,31.669],[45.414,43.437,21.951],
[43.616,42.944,24.625],[35.636,41.928,25.961],[48.842,46.725,22.384],
[39.209,45.416,24.647],[34.448,41.622,20.777],[33.852,43.033,24.093],
[45.257,50.425,24.139],[43.550,46.119,27.389],[51.437,36.462,26.314],
[46.491,39.488,26.900],[38.348,37.937,23.150],[47.297,38.008,22.403],
[36.613,41.621,23.028],[41.550,41.687,23.083],[51.548,37.887,28.714],
[36.696,37.459,27.772],[42.740,45.915,20.724],[38.494,43.417,27.479],
[49.750,37.142,23.538],[38.559,38.770,25.754],[40.176,45.990,27.350],
[35.332,45.111,19.717],[48.765,47.663,25.511],[48.098,37.499,28.758],
[40.563,33.989,25.296],[47.465,49.355,22.882],[46.897,47.769,20.518],
[50.225,45.072,30.464],[43.974,43.182,19.365],[44.025,43.089,16.601],
[52.409,40.201,27.887],[47.755,44.419,21.361],[45.682,38.640,29.107],
[38.061,44.015,22.221],[41.329,47.185,24.679],[40.373,42.820,25.215],
[45.533,48.084,27.699],[38.842,46.345,31.285],[46.527,43.010,17.838],
[37.968,42.785,30.005],[44.087,49.715,21.214],[39.011,49.466,31.554],
[50.465,34.450,17.267],[38.548,42.805,19.669],[51.892,48.686,16.602],
[31.189,44.767,27.913],[33.662,48.132,26.425],[54.429,49.423,31.388],
[43.426,47.377,30.635],[42.883,52.384,19.374],[38.987,50.367,19.378],
[45.769,48.650,16.407],[46.447,51.099,20.612],[52.044,40.550,30.522],
[36.524,45.899,29.542],[39.272,35.925,27.049],[44.050,35.147,29.412],
[48.546,38.174,31.352],[47.916,42.961,15.630],[43.195,37.024,17.174],
[38.807,40.706,28.372],[39.047,40.190,20.842],[43.733,40.465,18.085],
[43.998,41.207,33.896],[45.687,38.926,24.523],[36.203,39.947,28.154],
[40.099,38.765,16.028],[33.524,36.515,22.906],[35.662,38.455,18.366],
[43.236,41.772,26.983],[41.230,37.820,22.806],[36.111,39.327,21.531],
[40.610,38.207,19.660],[45.932,36.318,26.779],[39.081,41.323,17.638],
[48.357,51.224,34.412],[35.459,45.471,32.923],[42.907,39.909,24.947],
[37.535,37.460,20.058],[41.103,39.659,27.100],[42.719,45.335,18.015],
[49.654,44.839,26.971],[45.918,36.972,17.425],[46.140,46.344,29.360],
[48.070,40.583,18.669],[36.879,44.656,25.634],[44.824,47.173,18.403],
[35.352,41.771,29.754],[41.687,40.990,32.165],[45.959,48.741,30.303],
[50.358,40.190,19.798],[52.748,46.773,30.888],[49.801,46.574,17.913],
[47.054,43.782,24.454],[39.146,47.593,22.270],[34.894,43.950,17.015],
[41.502,44.552,23.208],[42.010,49.058,28.713],[40.574,51.778,22.609],
[45.305,43.591,27.176],[47.144,45.816,15.824],[48.531,49.179,29.827],
[49.632,51.584,16.657],[36.111,36.774,30.679],[47.043,45.459,18.708],
[36.508,48.198,28.010],[41.673,47.727,14.822],[39.871,45.972,18.068],
[37.300,44.623,18.005],[43.766,38.103,27.006],[44.928,35.799,19.693],
[42.377,48.212,18.714],[44.396,38.320,19.848],[43.797,40.178,30.707],
[43.235,35.973,21.826],[40.144,48.688,17.213],[40.750,35.869,16.408],
[50.204,36.205,30.346],[47.770,40.672,24.848],[38.088,46.928,16.183],
[37.010,40.182,16.264],[49.996,51.337,26.088],[36.746,34.682,24.688],
[34.955,48.585,23.007],[36.278,50.830,32.410],[41.836,50.725,33.456],
[44.394,48.541,14.161],[45.546,47.207,33.148],[49.050,48.463,15.998],
[34.580,43.161,31.911],[40.556,44.515,15.566],[41.494,45.888,31.704],
[44.305,46.200,15.347],[47.617,44.268,30.888],[36.823,41.763,32.512],
[41.120,43.244,18.196],[40.868,42.023,29.344],[40.793,41.798,15.572],
[46.188,44.438,33.614],[50.513,43.432,15.955],[40.327,38.001,31.104],
[42.540,37.289,29.965],[43.266,42.930,29.666],[48.138,40.182,33.285],
[46.324,41.953,31.817],[39.157,40.407,31.837],[41.488,34.508,28.627],
[42.646,38.721,33.298],[45.483,37.367,31.549],[46.120,37.435,34.265],
[36.625,38.606,32.536],[41.281,33.395,31.460],[41.761,34.382,34.002],
[42.218,51.093,16.889],[35.238,47.985,16.770],[33.504,52.530,21.609],
[34.634,51.317,23.905],[53.085,50.874,24.274],[51.704,49.815,27.709],
[50.734,47.047,32.993],[51.797,44.887,18.843],[50.456,42.653,21.224],
[51.683,46.136,25.555],[49.376,47.588,28.019],[50.039,42.422,18.364],
[49.504,38.269,18.030],[47.764,40.545,21.865],[34.281,45.375,26.164],
[49.827,42.723,28.337],[50.922,42.381,32.203],[49.037,37.565,14.938],
[51.033,37.820,20.858],[48.517,38.277,25.685],[49.531,39.887,27.789],
[34.268,39.188,31.993],[53.990,35.331,32.267],[49.262,35.797,21.111],
[35.959,37.339,24.884],[33.128,40.614,25.256],[49.479,35.324,33.080],
[51.506,37.185,16.459],[34.734,36.411,20.430],[49.506,51.675,21.748],
[34.333,34.676,27.684],[36.952,48.100,32.690],[39.370,48.175,28.783],
[43.510,50.165,30.932],[46.060,51.479,30.861],[50.146,33.525,30.625],
[37.696,53.167,28.057],[37.502,50.580,27.313],[44.305,53.144,26.595],
[45.057,50.554,28.529],[48.637,52.380,29.097],[35.745,49.526,25.749],
[38.306,48.577,24.969],[40.361,50.633,25.414],[43.206,34.545,24.589],
[49.676,31.503,24.114],[50.505,34.389,22.936],[37.178,35.806,22.227],
[36.228,34.889,19.198],[41.284,33.817,21.101],[48.069,34.924,24.183],
[48.179,34.115,18.918],[35.438,48.778,20.025],[37.967,37.107,16.296],
[42.030,35.598,18.885],[44.748,50.321,18.377],[51.138,50.436,19.927],
[39.234,35.501,31.933],[37.239,35.827,33.632],[37.214,49.827,16.337],
[46.786,35.112,32.766],[50.399,49.893,32.626],[51.523,55.072,25.946],
[46.428,58.017,26.132],[36.433,56.555,22.724],[47.364,58.704,22.081],
[33.401,57.445,23.797],[53.518,56.933,28.938],[35.294,54.553,27.615],
[48.379,56.419,24.242],[37.387,57.759,25.578],[49.483,54.915,27.844],
[42.388,52.603,24.620],[48.715,52.090,19.240],[49.344,57.938,32.602],
[38.894,53.363,25.132],[45.272,53.512,29.166],[46.926,56.803,29.047],
[44.263,54.651,17.867],[39.542,57.721,27.443],[39.715,57.308,22.940],
[44.509,58.465,23.212],[37.225,57.547,29.767],[41.859,55.208,17.144],
[32.991,55.146,22.589],[35.196,56.237,20.172],[40.833,55.063,24.387],
[38.146,57.292,20.302],[41.325,58.436,17.761],[46.143,55.147,26.604],
[37.545,-58.627,20.572],[44.022,57.568,27.311],[39.627,54.773,21.306],
[41.793,56.297,27.006],[45.376,57.090,17.879],[46.666,-59.674,19.398],
[51.587,58.114,19.100],[35.601,55.955,31.404],[43.051,55.044,29.016],
[45.105,54.542,20.332],[43.536,58.850,19.957],[43.664,57.887,29.770],
[45.959,56.068,22.472],[38.847,53.967,18.129],[49.918,56.710,30.036],
[37.834,-59.651,16.044],[36.247,52.645,25.885],[40.441,56.797,30.097],
[39.800,54.880,28.479],[43.797,-59.074,33.199],[38.226,-58.722,31.877],
[42.718,51.839,28.348],[40.623,57.786,32.669],[45.151,57.127,32.929],
[47.469,55.735,33.361],[37.985,54.841,31.975],[41.192,51.830,30.659],
[42.322,55.127,33.683],[51.312,58.610,16.563],[49.534,56.149,16.622],
[48.912,57.627,18.993],[48.969,58.794,25.573],[48.967,58.847,28.312],
[33.200,57.026,32.149],[51.118,52.187,31.276],[49.767,55.120,20.000],
[34.974,55.716,25.155],[35.422,-59.218,24.816],[51.665,56.027,31.852],
[52.142,55.763,17.440],[33.215,54.603,20.004],[34.199,52.020,28.975],
[47.339,54.079,30.918],[44.055,54.905,24.352],[48.858,53.699,24.048],
[51.523,55.079,22.348],[35.932,53.751,22.705],[35.549,51.887,16.961],
[42.709,53.827,21.815],[46.318,53.477,24.371],[45.977,52.635,18.019],
[37.111,56.470,18.036],[42.362,56.089,19.879],[37.769,52.403,30.863],
[40.228,53.089,33.700],[45.549,53.300,33.312],[46.859,-50.713,41.351],
[43.263,-50.097,44.198],[35.984,-51.396,47.406],[46.001,-47.074,41.586],
[37.224,-46.708,45.260],[32.325,-52.321,40.920],[35.166,-50.382,42.227],
[44.931,-43.160,42.443],[44.209,-46.503,44.102],[50.983,-57.652,44.465],
[47.851,-53.983,44.557],[38.839,-56.223,41.192],[46.406,-55.483,40.927],
[34.929,-52.852,41.338],[42.204,-50.267,41.600],[51.664,-54.202,46.727],
[37.332,-55.643,45.648],[44.102,-46.571,39.304],[40.802,-49.056,45.883],
[50.706,-55.492,41.906],[37.156,-53.782,42.749],[38.194,-48.152,47.344],
[36.121,-45.476,38.191],[47.890,-46.792,45.018],[48.723,-56.455,46.407],
[41.281,59.236,43.331],[47.575,-44.254,41.797],[47.011,-45.961,39.272],
[49.331,-47.027,48.525],[43.733,-50.481,36.946],[45.820,-51.608,35.243],
[52.554,-51.156,45.382],[48.182,-49.899,39.080],[47.352,-51.686,47.259],
[39.197,-48.853,41.564],[41.007,-46.475,44.116],[39.200,-50.773,43.153],
[45.649,-44.380,45.012],[40.716,-47.743,48.319],[47.823,-49.914,35.945],
[37.821,-51.555,49.284],[44.685,-44.092,39.930],[39.108,-45.511,51.501],
[48.517,-56.713,37.375],[37.343,-50.560,39.556],[49.826,-44.898,36.029],
[34.139,-49.308,47.001],[32.419,-43.955,44.473],[54.046,-44.055,50.397],
[42.657,-45.747,47.593],[43.679,-40.999,38.871],[38.293,-41.155,40.633],
[45.905,-43.345,35.242],[45.562,-42.282,37.714],[49.609,-53.822,50.485],
[35.320,-47.134,49.076],[39.367,-56.272,43.802],[43.857,-58.132,47.199],
[47.321,-54.415,48.746],[47.750,-50.144,33.299],[42.587,-55.671,36.249],
[38.468,-51.050,45.975],[39.511,-52.044,40.822],[45.305,-52.020,39.559],
[45.001,-52.939,50.412],[45.504,-53.089,42.318],[37.928,-53.565,47.244],
[40.596,-56.075,34.361],[34.617,-56.414,41.748],[33.987,-55.277,36.960],
[45.373,-52.295,45.020],[41.919,-55.343,42.147],[36.492,-55.030,40.224],
[41.458,-51.166,38.793],[45.942,-57.076,45.089],[38.336,-52.606,36.159],
[48.133,-41.231,51.853],[38.255,-48.025,50.104],[41.619,-52.404,43.198],
[38.217,-58.205,39.100],[41.171,-53.214,45.793],[43.217,-47.943,37.032],
[49.937,-49.532,45.975],[46.616,-54.862,35.791],[43.594,-46.975,49.770],
[47.284,-53.228,37.774],[36.733,-47.896,42.461],[44.341,-45.490,35.393],
[35.085,-51.134,50.040],[40.588,-50.489,50.410],[45.327,-45.311,48.736],
[50.936,-52.487,37.264],[51.793,-46.900,49.729],[51.790,-44.788,39.881],
[46.371,-49.474,44.099],[38.135,-46.563,40.219],[34.551,-50.339,35.383],
[41.701,-47.888,40.550],[42.572,-44.373,45.105],[41.618,-44.656,40.735],
[44.993,-47.192,46.841],[47.549,-47.631,33.165],[47.511,-44.651,47.240],
[52.709,-41.933,36.155],[37.334,-55.637,48.776],[47.452,-47.504,37.094],
[36.515,-44.781,46.918],[41.722,-45.967,35.076],[39.175,-46.540,36.012],
[37.634,-47.994,37.633],[44.363,-54.804,46.550],[45.383,-56.195,37.799],
[42.375,-44.756,37.663],[43.455,-53.955,38.565],[42.245,-53.666,50.181],
[43.460,-56.516,39.745],[39.880,-44.440,38.261],[37.952,-55.829,37.464],
[49.172,-56.461,49.062],[48.681,-51.647,42.959],[36.594,-45.863,35.365],
[35.474,-54.038,34.984],[48.739,-42.811,45.684],[36.514,-58.216,43.701],
[37.504,-43.785,39.763],[34.716,-43.825,51.302],[40.467,-41.221,52.342],
[43.565,-44.190,32.782],[46.691,-45.326,51.449],[47.882,-45.009,33.767],
[35.032,-48.614,51.576],[40.478,-48.430,34.250],[39.678,-51.868,53.053],
[43.383,-46.770,32.796],[47.574,-49.023,47.187],[37.034,-50.793,51.698],
[40.919,-49.710,36.689],[40.582,-52.142,48.354],[40.814,-51.253,34.558],
[44.727,-48.815,51.068],[50.220,-49.266,34.787],[39.747,-54.609,49.582],
[40.306,-55.455,47.101],[42.984,-50.640,47.633],[47.913,-52.969,52.290],
[46.946,-50.885,50.646],[37.932,-53.294,51.859],[40.366,-57.620,49.173],
[43.844,-56.494,51.723],[44.118,-55.587,48.986],[46.594,-56.219,50.780],
[38.115,-56.897,51.024],[40.414,58.495,50.010],[42.803,-59.138,52.443],
[41.544,-40.117,35.284],[33.919,-46.481,35.955],[34.948,-40.293,40.826],
[35.101,-43.171,42.124],[51.783,-41.217,45.450],[51.842,-42.981,47.489],
[48.641,-46.868,52.969],[50.899,-47.982,37.329],[51.243,-49.702,39.320],
[50.891,-47.975,42.993],[50.051,-45.513,46.251],[52.603,-50.388,36.748],
[50.393,-51.969,34.557],[51.031,-52.672,41.054],[33.932,-48.589,44.456],
[50.278,-50.680,50.354],[52.807,-51.805,50.426],[49.025,-54.001,33.100],
[49.831,-54.412,39.267],[50.183,-52.613,45.064],[49.715,-52.725,48.012],
[33.445,-53.286,48.280],[52.453,-59.399,50.887],[50.182,-57.252,39.671],
[35.041,-54.758,44.335],[32.999,-51.905,43.705],[49.757,-58.260,51.752],
[50.389,-57.084,34.995],[33.298,-56.455,39.255],[47.347,-43.109,39.326],
[35.218,-56.949,47.372],[36.775,-44.652,52.802],[40.036,-44.665,48.338],
[44.843,-41.864,50.068],[46.977,-40.970,48.905],[50.591,-58.985,48.598],
[36.178,-41.669,48.968],[38.796,-41.658,46.796],[42.604,-41.068,46.312],
[45.739,-41.926,46.552],[47.565,-39.735,46.617],[35.890,-42.175,46.303],
[37.400,-45.157,42.753],[41.213,-42.629,43.621],[43.632,-57.594,42.023],
[48.194,-56.429,43.241],[51.913,-59.197,40.984],[34.543,59.087,40.397],
[35.463,-57.571,36.299],[41.996,-58.897,40.193],[46.424,-58.492,42.512],
[47.246,-59.657,39.662],[35.635,-43.219,36.668],[37.405,-55.892,34.857],
[41.092,-57.147,37.880],[43.043,-42.240,36.267],[51.161,-42.282,41.107],
[36.090,-59.862,48.322],[37.935,-59.737,51.679],[36.961,-42.165,34.631],
[47.461,-58.801,50.537],[51.012,-43.319,50.456],[43.648,-32.334,41.469],
[43.879,-30.627,43.535],[37.242,-32.393,46.243],[47.830,-29.690,41.841],
[36.511,-27.403,43.898],[34.401,-32.762,38.773],[36.481,-30.275,41.021],
[46.416,-23.972,42.480],[42.020,-28.582,44.416],[52.362,-38.348,45.615],
[46.548,-34.554,45.433],[40.499,-35.602,41.054],[46.486,-36.983,40.966],
[35.296,-35.416,42.730],[40.628,-32.908,41.806],[51.084,-36.074,46.404],
[38.109,-37.457,46.323],[45.110,-29.350,41.706],[39.012,-29.383,45.450],
[48.630,-37.956,42.488],[37.266,-34.757,44.552],[39.313,-28.298,49.182],
[33.762,-28.356,37.832],[47.534,-27.184,44.542],[48.863,-37.449,45.285],
[42.706,-40.280,42.728],[48.442,-25.545,41.380],[48.469,-27.670,39.814],
[49.477,-27.909,48.096],[45.205,-31.377,38.882],[45.242,-33.372,36.258],
[50.326,-30.185,47.059],[47.937,-30.952,39.667],[45.946,-32.158,48.364],
[39.011,-28.975,41.493],[41.954,-25.154,42.737],[41.297,-31.249,43.903],
[45.881,-25.566,46.571],[41.633,-27.010,48.393],[46.614,-30.933,36.326],
[37.916,-30.646,48.501],[43.981,-24.442,39.158],[39.121,-26.089,50.781],
[51.372,-40.939,38.856],[38.510,-30.985,39.242],[50.050,-24.796,34.472],
[34.862,-30.564,47.477],[33.113,-27.060,43.566],[54.465,-25.696,49.945],
[42.841,-25.017,50.559],[44.474,-22.098,36.972],[37.443,-24.041,39.053],
[47.620,-23.978,32.985],[47.072,-23.822,37.823],[49.298,-33.969,50.725],
[35.739,-28.179,48.231],[40.445,-39.591,45.050],[44.872,-39.543,47.467],
[46.160,-35.545,48.037],[48.200,-30.766,34.042],[44.336,-36.914,35.914],
[39.816,-34.635,43.904],[39.013,-33.887,39.647],[43.985,-34.815,38.404],
[44.070,-32.675,50.844],[45.572,-34.333,42.067],[39.092,-35.181,46.722],
[38.622,-36.663,33.996],[33.685,-37.701,40.516],[34.978,-36.128,35.817],
[44.327,-32.866,45.710],[42.279,-37.894,41.055],[37.211,-33.941,41.574],
[40.412,-36.596,37.821],[46.667,-37.070,44.010],[40.586,-35.072,35.110],
[49.866,-22.546,50.140],[37.170,-29.420,51.540],[42.484,-35.441,43.729],
[39.380,-38.427,39.691],[42.169,-34.406,46.406],[42.257,-29.294,36.457],
[47.262,-30.048,46.696],[46.814,-37.996,35.378],[42.889,-29.467,48.062],
[48.167,-35.756,36.549],[38.504,-31.394,42.671],[44.938,-28.978,37.474],
[35.276,-33.098,48.665],[41.389,-32.497,50.425],[45.258,-27.762,48.890],
[52.218,-33.853,36.983],[52.106,-27.829,50.217],[49.381,-26.813,36.126],
[46.710,-30.487,44.113],[39.338,-26.763,39.756],[36.124,-32.141,35.644],
[41.353,-29.048,39.390],[43.215,-26.066,46.541],[40.736,-23.749,38.640],
[45.206,-28.419,45.837],[47.040,-28.273,35.054],[48.730,-25.070,49.340],
[52.303,-22.949,34.792],[36.563,-37.911,49.142],[48.509,-29.073,37.327],
[36.093,-26.119,46.752],[40.683,-28.079,33.742],[39.932,-28.322,37.170],
[36.332,-29.353,37.874],[43.847,-37.103,46.913],[46.172,-37.324,38.211],
[43.679,-26.938,35.724],[43.444,-35.539,41.157],[42.900,-34.629,49.372],
[44.516,-38.689,39.736],[40.003,-25.650,35.294],[39.305,-38.680,36.204],
[48.866,-36.227,48.380],[47.540,-33.130,43.304],[37.509,-27.305,36.199],
[37.102,-34.692,35.094],[50.103,-23.441,45.355],[37.486,-39.293,44.232],
[36.187,-26.840,39.443],[36.012,-24.169,50.469],[40.589,-24.075,52.305],
[43.903,-26.861,32.375],[45.575,-25.348,50.781],[48.178,-26.701,33.297],
[34.809,-30.351,50.503],[39.476,-30.844,35.032],[41.297,-29.480,50.959],
[44.391,-29.268,34.702],[47.954,-29.952,49.406],[36.548,-32.577,50.933],
[39.783,-32.876,37.143],[40.059,-32.616,47.407],[43.166,-34.387,34.812],
[43.980,-30.057,51.135],[51.220,-30.284,33.196],[40.025,-34.595,49.224],
[39.230,-38.532,49.514],[42.905,-32.333,48.086],[46.804,-34.719,52.643],
[46.981,-31.803,51.087],[39.201,-33.613,51.580],[41.366,-42.318,48.342],
[41.593,-38.208,51.920],[44.250,-36.217,51.276],[46.578,-38.249,51.525],
[37.339,-35.556,49.309],[42.242,-39.810,48.629],[42.139,-39.771,54.004],
[42.964,-22.708,32.086],[34.094,-28.468,33.949],[34.565,-23.056,39.043],
[34.587,-23.180,42.320],[52.187,-23.170,43.780],[51.494,-24.723,49.266],
[49.393,-27.451,50.796],[51.347,-27.231,38.054],[50.917,-29.709,41.192],
[49.753,-27.776,43.088],[48.681,-25.921,46.700],[51.103,-32.108,35.348],
[51.721,-34.589,34.567],[48.332,-34.848,41.465],[35.109,-31.271,44.194],
[50.500,-30.227,49.980],[52.285,-33.379,50.983],[50.144,-37.462,35.061],
[49.733,-36.448,38.386],[49.477,-34.727,44.155],[48.834,-33.711,46.854],
[34.834,-36.494,50.667],[51.180,-41.649,52.579],[50.277,-38.594,40.333],
[35.102,-38.097,43.067],[32.767,-34.781,45.038],[51.416,-39.284,50.385],
[52.539,-37.027,34.169],[35.237,-37.025,38.309],[48.667,-22.895,40.006],
[35.129,-39.607,47.565],[36.375,-26.957,50.492],[38.628,-25.837,47.761],
[43.976,-22.768,49.435],[46.819,-21.480,49.482],[49.123,-42.533,48.712],
[37.346,-21.371,47.209],[39.824,-22.894,46.375],[43.489,-23.611,44.269],
[47.321,-23.830,45.148],[48.718,-22.562,47.403],[36.636,-24.766,44.104],
[39.160,-24.749,41.349],[40.844,-25.804,45.089],[44.074,-37.984,43.875],
[48.364,-40.870,41.479],[51.297,-39.357,43.072],[36.941,-38.451,40.907],
[34.133,-40.767,37.947],[40.957,-41.658,40.996],[46.209,-39.758,42.647],
[48.152,-40.441,38.752],[35.212,-25.197,37.127],[36.190,-39.088,36.625],
[42.511,-38.598,37.958],[44.851,-24.566,35.722],[50.634,-24.592,38.517],
[35.427,-39.929,50.573],[37.980,-39.729,51.999],[36.609,-25.023,34.498],
[45.709,-40.394,53.211],[49.577,-24.763,52.173],[45.851,-12.520,38.777],
[44.034,-12.330,44.081],[38.120,-14.118,44.950],[48.166,-10.523,42.021],
[37.931,-10.749,43.364],[32.931,-16.204,40.430],[34.580,-12.154,41.022],
[46.129,-6.357,43.747],[43.194,-10.009,45.079],[50.770,-19.160,43.798],
[47.718,-14.888,45.485],[38.717,-17.922,42.511],[47.077,-19.118,41.576],
[35.226,-16.555,41.964],[41.290,-14.126,41.225],[52.428,-16.691,46.707],
[38.740,-19.021,45.675],[44.615,-10.246,40.474],[38.769,-11.145,46.321],
[49.884,-18.070,41.200],[36.906,-17.155,44.899],[37.444,-8.948,45.055],
[34.700,-8.619,37.839],[48.514,-7.879,46.198],[48.240,-17.908,47.099],
[40.664,-21.516,42.797],[47.875,-7.195,41.820],[46.901,-8.958,40.048],
[51.097,-17.244,49.056],[43.317,-12.483,39.956],[45.897,-14.186,35.762],
[51.923,-13.745,46.688],[48.624,-13.028,40.568],[47.415,-14.261,48.157],
[39.670,-11.804,41.026],[41.258,-8.414,43.012],[40.826,-12.932,44.747],
[45.102,-7.578,46.713],[41.274,-8.637,49.585],[47.784,-12.376,36.114],
[37.504,-13.579,47.673],[45.496,-5.728,40.125],[39.565,-5.877,50.088],
[50.344,-21.853,37.663],[36.819,-13.350,38.676],[51.244,-7.802,35.654],
[35.979,-10.691,47.613],[34.558,-7.508,43.936],[53.625,-6.996,48.901],
[43.986,-7.888,49.260],[44.902,-3.281,38.114],[38.150,-5.190,41.253],
[47.655,-5.171,35.349],[46.970,-6.410,37.752],[48.837,-15.978,50.075],
[37.922,-7.948,49.658],[38.508,-22.344,44.361],[46.568,-20.737,46.862],
[44.724,-17.322,44.980],[48.326,-13.948,33.941],[43.328,-17.707,35.084],
[41.039,-16.050,43.685],[40.653,-16.614,39.259],[43.074,-15.064,39.538],
[43.803,-15.798,51.707],[45.744,-15.591,40.832],[38.234,-16.283,47.160],
[39.789,-18.174,33.524],[32.598,-18.728,41.367],[34.372,-18.501,36.150],
[45.059,-14.808,45.956],[42.442,-18.102,41.043],[37.426,-17.125,40.218],
[38.755,-17.626,37.846],[46.409,-19.257,44.203],[39.019,-14.721,37.572],
[48.892,-4.311,50.692],[38.299,-11.042,49.242],[43.318,-15.280,42.492],
[37.969,-20.079,39.864],[41.866,-16.837,46.307],[43.385,-12.214,36.916],
[50.275,-11.716,45.604],[45.120,-19.659,35.165],[44.594,-11.605,47.394],
[45.938,-16.910,35.777],[36.838,-13.050,41.941],[45.481,-10.248,37.194],
[35.534,-15.235,47.799],[41.574,-14.402,51.350],[46.159,-9.275,48.464],
[51.999,-15.627,37.205],[52.146,-12.079,48.672],[49.925,-8.274,38.196],
[47.868,-12.208,44.190],[39.715,-8.953,40.577],[36.812,-11.410,33.702],
[41.243,-10.892,38.875],[41.955,-7.818,45.853],[40.343,-6.613,41.299],
[47.124,-11.434,46.557],[46.884,-9.836,34.709],[48.567,-5.909,48.224],
[51.746,-4.098,36.676],[36.016,-18.321,47.561],[48.341,-11.304,38.511],
[35.389,-7.936,46.759],[41.274,-8.924,33.762],[40.586,-8.758,36.349],
[36.825,-10.520,37.838],[44.196,-19.347,47.541],[46.538,-18.986,37.498],
[42.873,-7.128,35.852],[44.383,-16.966,38.438],[43.562,-16.679,48.242],
[44.052,-21.459,40.717],[40.640,-5.877,36.597],[40.873,-18.665,36.142],
[49.664,-19.415,50.135],[46.914,-14.432,43.002],[38.304,-7.751,36.900],
[35.149,-17.074,34.188],[50.389,-2.264,47.911],[37.466,-22.677,41.474],
[37.271,-7.734,39.859],[36.541,-5.283,50.419],[39.363,-5.442,52.943],
[43.783,-7.827,33.156],[46.939,-8.788,51.202],[45.227,-6.346,34.825],
[35.706,-11.900,50.839],[40.023,-11.183,35.235],[43.095,-10.742,49.755],
[43.394,-10.605,32.923],[47.814,-11.224,49.562],[37.002,-14.429,50.134],
[41.883,-13.551,35.107],[40.242,-13.682,48.507],[41.483,-16.280,34.184],
[45.377,-11.922,51.581],[51.399,-12.672,33.582],[40.707,-17.115,48.911],
[41.347,-19.400,45.730],[42.701,-13.753,47.016],[46.620,-16.535,51.563],
[47.111,-13.858,51.042],[39.341,-15.866,50.986],[42.189,-21.561,47.397],
[42.594,-18.345,52.140],[45.225,-21.073,52.081],[47.079,-19.327,51.290],
[37.970,-19.546,52.120],[41.067,-23.211,49.435],[42.730,-20.491,50.484],
[40.329,-3.186,36.518],[35.953,-9.417,35.388],[33.570,-4.349,38.127],
[35.473,-6.146,41.521],[54.496,-4.923,45.928],[50.936,-7.245,47.937],
[49.682,-9.483,50.819],[51.012,-10.756,37.608],[50.794,-13.244,38.771],
[50.815,-8.252,44.569],[46.182,-9.839,44.461],[50.452,-12.953,36.146],
[50.754,-15.450,34.825],[50.069,-14.971,41.515],[34.603,-12.562,46.133],
[49.816,-13.038,50.789],[52.538,-14.272,50.826],[49.320,-19.157,32.715],
[48.398,-16.422,39.734],[49.484,-16.649,44.323],[49.849,-15.459,47.232],
[33.826,-18.684,50.905],[53.472,-20.632,52.936],[49.762,-19.384,38.304],
[35.008,-18.915,43.748],[35.363,-14.605,44.173],[52.170,-19.110,50.530],
[51.993,-19.807,34.196],[34.888,-19.935,38.388],[50.070,-5.225,42.535],
[33.666,-19.836,45.965],[35.818,-8.836,51.446],[39.215,-7.074,46.230],
[43.792,-5.031,49.729],[46.566,-4.280,49.305],[51.051,-21.063,47.625],
[37.755,-3.498,48.408],[38.563,-4.435,45.626],[44.177,-4.515,45.091],
[46.556,-5.440,46.402],[47.819,-3.017,47.230],[36.774,-6.362,44.823],
[37.637,-7.765,42.615],[42.535,-5.895,43.249],[44.313,-19.190,42.762],
[50.206,-23.704,42.152],[49.861,-20.665,41.200],[35.172,-20.035,41.158],
[35.423,-22.457,35.184],[40.898,-21.309,39.734],[47.534,-21.698,42.683],
[47.140,-21.667,36.372],[35.898,-5.892,38.184],[36.679,-19.899,35.704],
[43.515,-19.679,37.462],[42.998,-4.035,35.770],[49.614,-5.522,38.224],
[38.085,-20.887,49.778],[40.193,-21.645,51.372],[37.941,-5.103,35.263],
[48.621,-21.853,52.507],[50.547,-6.066,52.075],[46.072,7.766,41.151],
[42.747,5.042,44.655],[37.230,4.561,44.684],[47.091,10.249,42.403],
[36.297,8.832,43.929],[34.279,3.049,39.421],[33.329,5.598,41.258],
[44.593,13.870,42.822],[43.455,8.586,44.205],[50.619,-0.967,43.728],
[46.955,2.208,45.871],[40.625,1.482,42.277],[45.935,0.105,41.011],
[34.349,3.234,41.914],[41.239,4.378,42.468],[51.144,0.345,48.491],
[38.388,-0.925,45.180],[44.578,10.080,40.815],[41.404,7.275,47.191],
[48.657,0.419,42.564],[36.976,2.038,43.215],[36.410,9.066,46.863],
[34.895,8.897,37.795],[48.097,9.418,45.148],[47.245,-0.486,48.105],
[43.195,-3.078,43.027],[47.068,13.198,41.702],[48.989,11.451,40.138],
[49.892,9.117,48.318],[43.665,6.139,40.784],[44.379,3.084,35.923],
[51.599,4.590,47.233],[49.242,6.090,41.585],[46.226,4.744,47.973],
[38.409,6.096,40.362],[40.795,9.705,45.015],[39.056,6.635,43.194],
[45.254,10.201,45.835],[38.993,7.940,48.053],[46.400,5.097,36.782],
[37.173,5.848,48.317],[43.221,13.598,40.257],[39.402,10.216,49.732],
[48.833,-2.362,36.324],[37.269,4.129,38.012],[51.266,10.129,35.673],
[32.475,7.541,47.699],[34.362,11.331,43.473],[53.475,12.069,48.828],
[42.029,10.111,49.160],[42.768,15.269,37.364],[37.820,13.456,40.610],
[46.096,13.329,35.820],[47.195,12.221,37.977],[49.833,4.054,49.473],
[36.944,11.067,48.775],[40.645,-2.577,44.475],[45.135,-3.096,47.242],
[48.049,1.872,48.911],[48.068,3.111,35.879],[42.335,-1.280,36.404],
[39.334,2.595,44.447],[40.252,4.288,39.831],[44.665,4.039,38.489],
[45.276,2.283,51.395],[46.049,1.453,43.431],[37.553,1.819,48.798],
[41.070,1.872,34.826],[33.061,-0.499,42.345],[35.937,1.094,37.787],
[44.561,3.215,45.883],[40.059,-1.144,42.123],[36.604,2.062,40.516],
[38.979,1.497,38.046],[45.812,-0.924,45.798],[41.453,3.819,37.112],
[48.850,14.432,49.169],[37.376,7.311,51.028],[42.087,2.371,44.380],
[36.688,-1.088,39.762],[42.641,1.440,47.073],[40.226,7.821,38.753],
[50.063,6.299,45.167],[44.880,-0.286,36.002],[44.190,8.761,48.240],
[47.745,3.877,38.715],[36.234,6.502,41.845],[44.082,9.438,38.219],
[33.935,5.738,48.904],[41.175,5.294,49.729],[46.441,9.585,49.514],
[53.258,3.219,35.655],[52.705,9.785,50.563],[49.283,10.316,37.576],
[47.614,6.146,44.106],[40.310,10.606,39.244],[35.903,6.093,34.802],
[41.847,8.818,41.106],[41.671,12.665,47.440],[40.549,12.989,40.462],
[46.168,7.881,46.720],[47.287,7.591,35.788],[48.283,11.375,49.179],
[51.380,12.758,35.401],[33.966,0.686,49.154],[47.263,9.495,39.071],
[36.216,12.011,45.915],[40.746,10.666,32.125],[40.463,9.436,35.407],
[37.695,6.780,36.792],[44.533,-0.650,48.181],[46.642,0.484,38.513],
[44.150,11.228,36.446],[42.860,0.307,38.704],[42.841,3.062,49.929],
[43.708,-1.551,40.697],[41.883,12.490,38.085],[39.378,-1.726,34.466],
[48.884,-0.105,50.341],[49.155,3.438,43.641],[37.938,11.137,36.572],
[36.706,3.400,34.512],[49.215,13.479,44.617],[37.040,-3.002,43.769],
[37.950,10.727,40.186],[35.238,12.102,50.821],[39.047,12.418,51.855],
[44.723,11.864,33.451],[46.187,10.870,52.426],[48.672,12.165,35.169],
[34.368,6.679,51.543],[39.585,7.519,33.610],[42.923,7.721,50.728],
[43.591,9.828,32.194],[47.994,7.050,49.031],[35.699,3.870,49.642],
[40.619,6.309,36.067],[39.743,4.684,47.147],[43.672,5.995,34.842],
[45.472,6.920,50.895],[50.266,7.068,34.125],[41.402,0.695,49.268],
[39.590,0.430,47.099],[42.629,4.477,47.374],[48.094,2.503,51.735],
[47.419,5.098,50.784],[39.215,4.014,51.004],[42.501,-3.947,47.589],
[42.753,3.775,52.551],[44.380,-0.362,51.173],[47.100,-0.088,52.283],
[38.391,1.182,51.253],[41.683,-3.903,51.712],[43.767,-3.031,52.895],
[42.675,15.659,33.325],[33.738,9.239,35.147],[33.097,14.581,41.053],
[34.964,12.758,40.473],[51.474,14.657,45.988],[50.892,13.218,48.211],
[49.871,9.426,51.373],[49.259,7.576,37.672],[50.828,5.765,39.594],
[50.625,8.963,44.711],[49.898,11.347,46.405],[49.954,5.174,36.997],
[50.593,1.473,35.834],[47.525,2.871,41.378],[34.300,5.465,43.816],
[50.399,6.793,50.273],[52.370,4.714,50.689],[49.975,-0.991,33.275],
[49.543,1.494,38.927],[49.111,0.933,45.382],[48.869,3.905,46.781],
[33.326,1.165,51.818],[53.057,-2.475,52.753],[49.432,-1.529,38.774],
[34.902,0.362,44.324],[33.145,2.998,44.457],[50.978,-1.263,51.210],
[51.475,-1.565,36.036],[33.858,-1.713,39.799],[49.419,14.483,40.184],
[36.050,-1.747,47.208],[34.254,10.423,47.672],[38.644,11.141,46.354],
[43.913,14.067,49.917],[46.701,16.143,49.638],[51.858,-4.764,49.303],
[37.738,14.296,49.081],[39.877,15.271,46.773],[44.093,16.284,44.891],
[44.721,13.745,47.184],[46.928,15.546,46.938],[36.935,13.302,43.455],
[39.297,11.333,42.390],[42.192,11.761,42.311],[45.230,-1.151,43.257],
[49.819,-3.504,44.624],[52.125,-3.439,41.537],[35.573,-3.400,41.359],
[35.944,-3.759,36.487],[40.791,-3.929,41.767],[47.086,-3.400,43.733],
[47.711,-3.787,38.975],[36.423,13.326,37.667],[36.521,-2.382,34.176],
[42.181,-3.101,38.790],[44.263,15.452,35.164],[50.150,12.858,38.024],
[37.672,-2.545,50.914],[40.434,-1.634,51.597],[36.518,13.713,34.483],
[46.116,-4.341,53.262],[50.181,12.352,51.538],[45.206,26.332,41.923],
[43.954,25.367,44.375],[37.656,23.577,45.192],[48.365,27.664,41.119],
[37.946,27.785,41.898],[34.438,22.266,38.783],[33.755,25.959,41.310],
[45.239,33.893,43.536],[43.148,28.018,45.230],[50.902,17.398,45.668],
[46.434,21.012,44.594],[39.889,20.180,42.590],[46.127,19.033,37.714],
[35.106,22.863,41.743],[41.292,22.301,41.802],[51.084,19.551,47.669],
[36.409,18.412,45.242],[44.477,28.360,38.410],[39.923,25.881,45.788],
[48.331,18.426,42.430],[36.946,20.923,44.119],[39.807,28.271,47.451],
[34.502,28.968,37.107],[48.509,27.931,44.511],[48.623,18.708,46.674],
[42.119,15.070,42.365],[46.863,30.832,40.937],[46.815,28.859,39.314],
[49.397,26.359,47.992],[43.541,23.081,37.210],[44.255,22.978,34.563],
[51.099,24.115,46.517],[47.451,25.390,40.500],[46.724,22.964,46.449],
[38.465,24.143,41.212],[40.611,28.676,41.643],[40.193,24.555,43.585],
[46.639,29.853,45.265],[40.344,29.244,49.980],[47.016,23.982,36.366],
[37.605,25.571,47.620],[44.091,32.048,41.114],[39.287,32.093,50.206],
[49.452,15.376,37.343],[37.995,24.249,38.464],[51.857,29.585,37.504],
[35.775,27.264,46.317],[34.397,29.796,44.935],[54.230,30.414,49.857],
[42.958,29.262,48.439],[44.142,33.443,38.783],[37.280,31.847,41.860],
[47.151,32.645,35.645],[44.952,30.980,38.790],[49.190,22.167,50.643],
[37.469,29.417,48.863],[37.367,15.923,46.379],[44.263,16.179,48.100],
[47.531,20.516,48.981],[48.113,25.109,33.734],[44.360,18.542,35.481],
[39.306,21.033,45.077],[39.512,21.719,39.444],[45.995,23.475,38.736],
[43.878,22.701,50.985],[45.790,23.545,41.537],[36.493,21.542,48.939],
[41.159,18.097,35.021],[34.249,17.636,40.359],[36.147,19.377,36.903],
[44.114,22.397,45.950],[41.771,20.039,39.669],[37.120,20.601,41.109],
[40.382,19.741,37.178],[46.402,17.687,45.580],[39.190,20.840,35.014],
[49.624,31.878,49.235],[37.682,26.875,50.326],[43.406,23.049,43.252],
[38.302,17.691,40.310],[41.957,21.300,44.337],[42.653,26.640,37.673],
[48.080,25.174,45.820],[47.118,17.513,35.715],[44.197,26.746,47.878],
[47.782,21.338,37.312],[35.631,26.902,43.111],[45.828,28.477,35.998],
[35.106,23.875,47.578],[41.664,24.591,50.641],[46.512,30.244,48.330],
[52.210,19.911,36.601],[52.172,27.343,48.893],[48.817,28.867,36.902],
[47.648,25.820,43.081],[38.900,28.736,39.175],[35.473,25.886,35.189],
[40.472,26.506,39.842],[41.276,31.442,48.128],[39.701,31.254,40.647],
[46.017,26.972,45.622],[48.699,26.119,36.060],[48.462,31.336,46.568],
[53.633,33.041,34.579],[35.207,20.224,46.757],[47.018,26.674,38.024],
[36.964,30.049,45.711],[41.526,28.687,34.069],[39.866,27.810,36.377],
[37.647,26.115,36.610],[43.958,19.272,47.367],[45.297,17.800,39.995],
[43.087,30.014,36.794],[43.645,20.437,37.625],[42.387,21.580,48.234],
[42.426,17.407,40.880],[40.575,31.313,37.694],[39.134,16.553,36.252],
[49.103,18.241,49.420],[49.093,22.862,45.232],[36.526,28.665,35.424],
[36.187,20.507,34.408],[49.333,30.809,44.143],[37.096,16.143,43.134],
[36.786,30.085,39.865],[33.547,32.433,51.247],[40.968,32.100,52.193],
[42.873,31.484,34.585],[44.996,28.352,49.930],[48.687,30.709,34.462],
[33.766,26.274,50.270],[38.095,25.548,32.913],[41.345,26.830,49.102],
[43.923,26.955,34.605],[46.730,25.506,48.377],[36.262,24.100,50.575],
[41.088,24.696,36.123],[40.584,22.854,46.549],[40.415,23.749,33.603],
[46.403,25.966,51.080],[50.893,25.208,34.180],[40.699,19.763,49.376],
[39.120,18.166,45.957],[42.641,24.613,47.037],[47.100,21.201,52.147],
[48.763,25.175,50.462],[38.085,21.981,51.159],[41.463,16.983,49.038],
[42.425,19.676,51.834],[44.977,20.592,49.811],[46.620,18.255,52.073],
[36.226,19.234,51.150],[39.883,14.894,51.102],[43.417,16.465,51.812],
[42.696,32.822,36.766],[33.680,27.844,33.369],[36.414,32.827,39.384],
[34.868,30.279,41.889],[51.143,32.668,44.008],[51.657,29.992,49.004],
[50.059,27.423,51.216],[50.998,26.802,37.018],[49.950,25.415,40.679],
[52.347,27.318,43.550],[50.433,29.348,46.447],[51.328,23.718,36.678],
[50.395,21.803,34.380],[48.237,21.846,40.468],[34.295,24.554,43.865],
[51.017,24.465,49.225],[53.000,23.127,51.108],[49.040,18.809,33.509],
[50.198,20.534,38.518],[49.625,20.731,43.294],[49.636,21.928,47.827],
[34.084,20.373,49.739],[53.187,16.060,53.236],[49.690,17.477,39.773],
[35.413,18.468,42.623],[34.205,21.779,44.376],[51.070,15.760,51.479],
[51.449,16.843,33.851],[33.956,17.531,37.576],[49.314,32.787,39.441],
[34.727,17.790,47.810],[36.216,30.157,51.621],[39.107,30.821,47.003],
[44.463,31.824,50.127],[47.114,32.932,50.064],[50.982,16.147,48.752],
[36.452,34.892,47.047],[39.758,33.761,47.937],[43.325,33.666,45.751],
[45.640,32.115,45.899],[47.809,33.914,45.385],[35.293,32.987,43.461],
[39.186,30.361,43.678],[41.846,31.084,43.112],[44.350,19.647,42.881],
[49.754,15.797,43.223],[51.474,15.864,41.195],[35.990,15.642,40.396],
[34.412,14.603,36.373],[40.438,15.697,39.708],[46.047,16.846,43.073],
[47.392,16.664,38.589],[35.283,31.403,37.124],[36.771,16.468,35.241],
[42.673,17.843,37.652],[44.612,33.779,33.993],[51.185,30.919,39.754],
[36.145,16.424,50.029],[37.659,16.152,52.493],[37.370,32.809,33.645],
[48.617,15.317,52.123],[50.533,30.043,51.351],[45.455,44.300,40.562],
[44.806,42.404,42.545],[38.535,42.396,45.849],[46.028,47.233,42.615],
[37.485,44.313,44.244],[34.978,40.812,39.155],[35.416,41.992,41.532],
[46.827,51.680,41.544],[43.947,46.844,44.402],[51.935,37.530,43.020],
[46.905,39.668,45.324],[39.654,38.350,41.427],[47.199,37.551,40.238],
[34.196,39.236,41.859],[40.448,41.231,41.497],[52.444,38.954,47.786],
[37.483,36.875,45.157],[44.784,47.145,39.797],[40.058,43.531,47.874],
[47.382,38.353,43.047],[37.735,39.088,43.873],[39.926,46.359,45.698],
[36.810,47.217,37.785],[48.718,46.907,42.973],[49.038,36.191,45.856],
[41.407,34.554,43.048],[47.854,49.173,41.739],[47.373,47.638,39.598],
[49.770,45.076,48.069],[44.639,42.773,37.940],[45.093,39.436,36.058],
[51.841,42.789,47.554],[46.507,40.720,38.082],[45.357,42.027,48.044],
[40.217,44.217,40.995],[41.268,46.190,43.234],[39.907,43.410,43.482],
[48.223,48.447,46.648],[40.368,46.292,48.421],[47.869,43.761,36.299],
[37.756,42.072,48.335],[44.961,49.715,40.458],[38.530,48.544,49.755],
[51.988,32.595,36.905],[39.195,43.183,38.730],[51.943,49.165,34.882],
[33.593,43.512,46.331],[34.408,48.116,44.827],[54.027,48.771,49.379],
[42.869,46.491,47.296],[45.707,52.847,38.113],[37.101,51.326,41.951],
[46.861,48.741,35.076],[47.173,50.657,38.035],[50.157,39.676,49.530],
[37.464,46.783,47.895],[39.327,34.720,45.693],[46.230,34.376,47.671],
[47.292,38.156,49.708],[48.923,42.699,34.052],[43.730,36.273,33.574],
[40.146,40.191,44.313],[39.340,40.464,39.299],[44.054,39.985,38.757],
[45.863,39.471,51.723],[45.786,40.006,40.770],[37.825,39.699,46.539],
[40.205,38.997,34.197],[33.814,36.478,41.502],[34.286,38.526,38.052],
[45.037,41.530,45.326],[42.021,39.201,40.491],[36.954,39.853,40.733],
[39.942,38.872,37.148],[46.145,36.566,45.208],[39.240,43.968,35.878],
[50.269,51.647,50.797],[36.300,45.195,50.089],[42.386,41.307,43.688],
[40.710,36.473,39.561],[42.994,39.180,44.919],[43.769,45.318,37.039],
[49.486,43.420,45.980],[47.476,38.142,36.577],[44.789,45.414,49.066],
[48.397,41.212,36.265],[37.577,43.611,41.156],[46.667,46.719,37.244],
[34.896,42.489,48.637],[40.518,41.767,50.063],[45.762,47.124,46.964],
[50.964,40.343,37.553],[52.132,46.535,49.738],[51.597,47.747,37.179],
[46.611,44.697,43.588],[40.695,47.736,40.815],[35.050,41.667,34.870],
[42.259,45.534,39.994],[42.525,49.720,45.457],[40.381,50.386,40.477],
[45.625,44.637,46.048],[48.356,45.948,33.560],[49.038,50.371,49.007],
[52.264,51.822,35.684],[36.479,37.795,48.294],[49.459,45.418,37.790],
[35.511,48.889,47.349],[40.882,48.253,34.460],[41.435,45.798,35.824],
[35.619,44.734,36.710],[43.725,37.616,47.032],[45.328,35.956,38.574],
[43.156,48.686,37.782],[42.444,37.803,38.038],[42.731,41.107,48.636],
[43.335,35.815,40.569],[39.853,49.732,36.950],[40.057,35.791,35.668],
[49.257,38.663,47.078],[47.409,41.596,42.827],[38.967,46.905,36.112],
[37.906,40.613,35.078],[49.759,49.057,44.516],[38.252,34.518,43.254],
[36.542,47.828,40.507],[36.528,50.297,49.248],[39.730,49.541,52.452],
[44.251,49.347,33.465],[45.789,47.016,50.883],[48.295,48.115,32.301],
[34.256,43.691,51.413],[38.409,45.437,33.854],[42.361,44.875,49.798],
[43.182,45.972,33.706],[47.015,44.384,48.263],[36.662,42.173,50.775],
[40.510,41.706,36.934],[40.645,40.837,46.806],[41.705,43.349,34.112],
[44.056,44.952,51.915],[52.126,44.616,32.609],[40.574,38.790,49.026],
[40.801,37.361,45.223],[42.270,42.832,46.356],[46.632,40.098,54.171],
[47.200,42.288,49.856],[39.145,39.547,52.894],[41.566,35.620,47.451],
[43.034,39.898,51.200],[44.769,37.560,49.848],[47.801,36.966,52.096],
[37.543,37.882,51.531],[41.621,34.101,49.907],[42.162,34.872,52.358],
[41.475,52.160,35.863],[35.301,47.654,35.455],[35.547,52.465,39.862],
[34.778,49.790,41.731],[51.611,51.857,45.073],[51.867,50.425,48.180],
[50.129,46.194,52.307],[52.020,45.333,38.431],[51.208,42.730,39.604],
[52.945,46.454,44.677],[51.361,48.182,46.829],[50.783,42.865,36.374],
[50.852,39.181,35.143],[48.979,41.260,40.100],[35.133,43.157,44.126],
[49.997,42.443,49.297],[52.625,41.074,49.708],[48.825,37.636,34.187],
[49.940,38.508,40.074],[50.075,40.161,42.880],[48.673,41.348,46.986],
[33.347,38.323,51.303],[51.263,34.304,52.965],[50.542,35.565,39.104],
[36.375,37.226,42.644],[33.643,40.375,44.561],[50.998,35.719,48.825],
[51.805,36.281,34.249],[33.727,35.237,38.896],[49.343,50.316,39.604],
[34.117,36.131,46.581],[35.778,47.848,50.722],[41.148,49.767,48.094],
[44.262,49.785,48.755],[46.934,51.335,51.159],[50.519,33.910,46.582],
[38.025,51.626,47.065],[39.889,49.926,45.604],[45.425,51.039,44.795],
[46.651,50.484,47.627],[48.302,51.026,45.501],[35.727,50.768,44.158],
[38.576,48.829,43.732],[42.374,48.822,42.934],[44.116,37.890,42.623],
[49.119,32.957,41.994],[51.057,34.915,41.591],[36.203,34.934,41.428],
[35.858,34.081,36.151],[41.165,33.857,40.067],[47.994,35.192,42.949],
[48.825,34.320,37.093],[35.355,49.560,37.740],[35.858,38.357,35.006],
[43.007,35.837,36.323],[44.956,50.561,35.961],[52.038,50.392,38.498],
[37.193,33.283,49.203],[38.862,34.570,52.155],[37.456,51.601,34.974],
[48.777,34.425,51.530],[51.812,48.682,51.868],[51.947,54.565,45.145],
[47.320,58.404,44.154],[38.939,57.346,42.746],[48.111,56.356,39.850],
[34.192,57.383,42.676],[52.790,56.985,47.336],[37.162,55.623,45.895],
[48.601,56.468,42.763],[36.869,58.210,44.764],[49.407,55.216,46.869],
[43.402,51.994,42.959],[49.005,53.547,35.944],[50.348,57.641,49.918],
[39.691,53.552,45.649],[44.169,52.303,47.689],[47.712,56.523,48.833],
[43.709,53.477,35.578],[39.423,55.824,47.357],[40.492,57.823,40.043],
[44.990,58.563,41.499],[38.668,-59.350,47.045],[39.372,55.328,34.724],
[33.956,54.802,41.023],[35.241,57.249,36.615],[41.187,56.187,43.552],
[36.796,57.914,41.295],[40.676,56.555,36.736],[46.560,55.509,44.562],
[38.810,58.888,37.063],[43.698,-59.044,44.472],[39.055,55.527,40.378],
[41.115,58.452,47.298],[45.276,56.752,35.708],[47.030,-58.879,36.468],
[51.253,-59.419,37.342],[36.588,54.886,48.354],[44.297,54.527,45.753],
[45.955,55.623,38.128],[43.848,57.159,39.316],[43.565,57.560,48.071],
[43.825,53.925,39.838],[39.449,53.668,37.255],[49.865,55.028,49.427],
[37.668,57.413,34.384],[36.729,53.188,44.314],[41.582,55.632,49.751],
[41.963,55.447,46.486],[45.328,58.856,52.053],[38.220,56.906,49.584],
[42.230,52.718,45.564],[43.091,57.630,50.857],[45.881,58.434,49.048],
[48.543,53.829,51.872],[36.396,54.604,51.931],[42.271,52.231,49.900],
[43.499,53.926,52.526],[51.226,58.901,34.486],[49.288,56.727,35.337],
[48.961,57.933,37.881],[50.332,58.740,43.665],[47.729,-59.296,46.826],
[35.735,57.559,50.280],[51.577,53.511,52.397],[50.478,55.281,38.243],
[34.904,55.096,43.629],[34.264,58.842,45.617],[52.303,54.306,49.233],
[52.164,54.053,33.773],[34.214,54.408,38.418],[34.654,54.108,46.243],
[50.645,52.866,47.623],[44.875,54.388,42.802],[49.766,51.242,42.592],
[51.112,54.069,42.513],[34.222,52.571,42.532],[34.864,52.209,36.661],
[41.547,52.655,40.895],[48.716,53.411,44.030],[48.715,53.580,39.873],
[36.703,54.278,36.394],[43.160,55.675,36.709],[37.567,53.195,49.979],
[39.211,52.129,52.213],[48.776,51.217,53.654],[34.901,-52.706,57.860],
[48.211,-56.479,57.762],[46.228,-47.471,57.723],[32.979,-46.956,55.776],
[48.078,-45.016,58.224],[44.552,-50.021,58.684],[44.802,-52.953,54.239],
[48.718,-48.079,58.612],[46.610,-51.091,57.002],[45.530,-45.185,59.530],
[49.604,-59.150,55.591],[38.716,-51.670,56.848],[50.951,-45.787,53.815],
[42.121,-40.776,56.770],[38.145,-43.745,58.348],[46.910,-43.400,53.139],
[45.023,-43.376,56.377],[47.270,-50.545,54.069],[43.231,-55.113,54.101],
[39.805,-52.173,-59.419],[42.809,-51.598,57.192],[39.626,-55.276,52.609],
[35.523,-54.130,55.593],[37.594,-55.553,56.760],[41.123,-53.570,57.550],
[38.529,-53.546,55.036],[38.126,-57.208,58.681],[42.583,-48.861,55.751],
[45.903,-56.176,53.519],[47.173,-53.463,55.489],[44.143,-46.416,56.391],
[52.036,-52.407,55.041],[47.832,-43.788,55.690],[36.051,-49.745,54.006],
[40.219,-47.528,-59.318],[40.781,-42.544,58.142],[46.621,-45.901,54.357],
[51.248,-41.238,56.624],[48.100,-48.107,55.629],[40.935,-46.909,53.709],
[40.957,-46.884,56.720],[36.969,-48.639,56.596],[45.108,-57.107,56.629],
[42.505,-45.188,54.910],[44.088,-54.661,57.500],[43.916,-57.765,58.943],
[40.062,-44.274,56.228],[39.946,-56.135,55.241],[37.462,-44.775,55.648],
[34.931,-54.161,52.239],[36.543,-46.300,59.014],[44.314,-44.053,51.331],
[49.047,-44.143,51.909],[38.403,-48.918,52.889],[43.429,-47.412,53.008],
[41.097,-51.070,55.265],[42.276,-50.859,52.661],[50.417,-48.692,52.305],
[41.518,-42.496,54.657],[35.177,-46.568,54.053],[34.227,-40.003,57.971],
[51.627,-47.305,57.212],[50.681,-50.987,57.512],[50.874,-49.344,55.217],
[51.128,-52.113,52.580],[49.278,-50.343,-59.214],[48.493,-56.114,52.810],
[49.797,-54.417,56.143],[50.803,-57.342,57.274],[51.365,-55.788,52.973],
[35.184,-58.391,57.500],[48.258,-41.729,58.232],[35.285,-57.696,54.662],
[41.331,-58.792,58.093],[46.588,-58.741,58.462],[35.173,-43.245,57.042],
[36.925,-55.859,53.414],[42.666,-56.563,56.306],[44.174,-42.423,54.124],
[51.079,-42.975,58.990],[35.538,-42.155,53.221],[32.988,-34.800,58.519],
[46.258,-37.164,58.113],[45.715,-30.940,59.094],[35.744,-28.925,55.417],
[49.263,-28.065,57.062],[44.016,-30.274,55.256],[45.221,-32.589,53.573],
[48.035,-32.022,59.275],[46.276,-31.445,56.172],[44.414,-24.647,59.061],
[49.985,-39.530,54.769],[38.289,-31.581,58.358],[51.794,-25.898,53.258],
[44.744,-22.193,56.995],[36.884,-24.919,58.155],[47.020,-23.039,55.685],
[45.178,-24.765,55.902],[47.933,-32.183,53.553],[44.240,-38.174,53.154],
[39.165,-35.164,58.366],[45.639,-33.805,57.848],[40.187,-35.873,51.412],
[36.024,-35.386,56.357],[36.780,-38.047,56.897],[40.954,-35.727,55.616],
[39.223,-33.121,54.711],[39.273,-38.016,58.154],[42.201,-29.083,57.199],
[47.341,-38.468,54.323],[46.988,-34.069,55.383],[45.162,-28.399,56.998],
[52.150,-33.435,55.783],[51.032,-25.771,56.023],[34.800,-31.680,53.631],
[39.728,-28.480,59.225],[39.299,-23.376,58.477],[46.998,-27.963,54.788],
[52.086,-22.884,53.432],[49.208,-31.192,55.999],[41.507,-26.712,52.093],
[40.606,-26.695,54.595],[38.026,-30.641,54.745],[45.611,-36.368,54.518],
[43.278,-26.724,55.858],[42.913,-35.238,57.850],[42.938,-37.537,-59.237],
[40.320,-24.461,55.962],[40.517,-38.308,55.511],[38.131,-27.315,54.924],
[35.267,-34.380,52.494],[35.504,-28.210,59.035],[42.880,-25.078,53.725],
[47.014,-25.248,52.937],[39.385,-29.090,53.049],[44.649,-27.379,53.425],
[41.193,-31.105,55.615],[42.528,-33.650,52.775],[50.107,-29.354,52.946],
[41.079,-22.035,54.300],[35.846,-27.134,53.104],[34.967,-23.328,58.755],
[51.030,-30.027,57.517],[50.165,-33.461,57.725],[50.777,-32.017,53.160],
[52.130,-35.105,53.541],[50.633,-32.239,-58.535],[49.445,-37.403,52.331],
[48.022,-36.317,56.284],[49.911,-38.305,57.371],[52.755,-38.099,54.405],
[33.894,-37.136,57.292],[49.806,-22.738,-59.448],[34.731,-39.282,54.920],
[41.084,-39.684,-59.441],[45.948,-40.944,56.588],[35.531,-26.011,56.056],
[37.514,-38.108,54.053],[43.090,-37.769,56.319],[44.433,-22.808,54.059],
[51.641,-24.647,58.536],[38.633,-22.931,54.151],[35.331,-14.484,57.013],
[47.990,-18.524,58.616],[46.954,-9.172,57.180],[34.433,-9.127,56.288],
[49.545,-9.940,57.195],[42.717,-12.788,56.725],[45.476,-14.229,53.051],
[47.943,-11.431,59.665],[46.752,-13.648,56.031],[44.611,-6.883,59.085],
[49.987,-23.010,56.618],[38.190,-12.626,54.786],[52.338,-7.902,55.163],
[44.182,-3.128,55.519],[37.670,-5.700,57.785],[47.729,-4.840,55.976],
[46.212,-4.379,59.163],[48.403,-13.908,53.590],[43.915,-19.151,55.491],
[39.126,-15.122,57.416],[44.420,-15.145,57.917],[39.883,-18.119,53.048],
[34.874,-18.422,53.468],[36.804,-17.090,58.550],[39.222,-17.228,55.567],
[39.596,-14.947,54.139],[39.132,-20.735,58.652],[42.934,-10.460,55.196],
[46.510,-16.840,55.448],[48.423,-15.786,57.057],[45.112,-11.090,56.405],
[52.369,-16.835,56.356],[47.830,-7.447,55.276],[35.745,-13.593,53.547],
[41.810,-11.035,58.456],[41.174,-6.287,58.807],[47.748,-9.949,53.818],
[52.573,-5.335,54.060],[48.806,-12.746,57.250],[41.173,-8.859,52.595],
[39.586,-10.172,55.704],[35.638,-11.195,55.174],[46.583,-19.634,56.534],
[44.999,-7.788,55.325],[42.607,-17.058,57.658],[44.865,-18.634,58.574],
[41.408,-5.778,55.805],[39.392,-19.800,55.223],[38.053,-8.205,54.200],
[37.075,-15.523,54.777],[35.880,-6.917,59.269],[45.118,-6.930,52.047],
[47.918,-6.814,52.869],[38.641,-10.262,52.312],[45.010,-9.830,53.413],
[40.753,-12.350,53.967],[42.563,-14.589,53.989],[54.710,-6.641,55.761],
[41.282,-3.494,54.367],[33.883,-10.676,53.072],[33.142,-3.853,59.360],
[51.300,-11.871,56.228],[51.215,-13.532,58.325],[52.355,-13.224,54.081],
[49.903,-16.043,54.560],[49.807,-15.509,-59.076],[49.542,-18.274,52.681],
[49.167,-18.334,56.010],[50.307,-20.452,57.544],[51.601,-19.256,54.860],
[35.250,-19.625,56.737],[48.743,-4.115,-58.432],[35.055,-22.287,55.907],
[42.115,-22.601,58.207],[46.997,-22.054,58.557],[35.018,-6.203,56.013],
[36.438,-22.005,52.301],[41.320,-19.167,57.044],[43.695,-5.733,54.014],
[50.277,-6.177,55.677],[36.530,-5.638,53.274],[34.563,3.068,57.852],
[46.009,1.711,-59.156],[47.412,10.710,58.852],[32.972,8.992,54.991],
[49.782,9.260,58.736],[43.688,5.567,57.200],[45.642,4.644,53.119],
[48.202,5.302,58.017],[46.788,3.951,55.780],[46.010,13.349,58.451],
[50.814,-3.411,54.896],[39.443,5.155,56.793],[50.604,10.932,54.053],
[44.163,14.671,57.265],[35.800,14.008,57.936],[46.487,14.331,55.158],
[47.748,12.453,56.688],[48.259,5.515,53.725],[44.210,-0.383,55.488],
[39.296,2.864,58.613],[45.377,3.381,57.788],[41.207,0.713,52.001],
[35.566,0.737,56.481],[35.411,0.893,59.302],[39.374,1.946,56.055],
[37.889,3.372,53.855],[38.101,0.343,57.965],[42.501,7.807,55.927],
[47.372,0.440,55.096],[45.156,1.888,54.944],[45.767,7.160,56.832],
[53.909,2.686,55.287],[50.503,11.519,57.528],[36.470,5.045,52.271],
[42.169,6.896,-59.138],[40.941,13.397,58.166],[47.656,7.759,51.957],
[51.570,13.818,53.661],[48.042,8.333,56.208],[40.748,10.183,52.024],
[41.124,10.073,55.147],[36.607,6.568,54.983],[46.362,-1.696,56.612],
[43.725,10.518,56.336],[44.033,0.868,57.819],[42.489,-0.902,-59.873],
[41.668,12.491,54.197],[40.754,-0.886,54.749],[38.728,10.872,54.371],
[35.579,1.957,53.034],[35.294,10.427,57.791],[43.373,11.214,52.410],
[48.375,12.878,54.043],[38.730,8.447,53.056],[44.429,8.854,53.848],
[41.372,5.495,54.887],[40.406,2.925,53.607],[50.931,6.330,53.573],
[41.670,15.288,54.905],[36.473,9.251,54.365],[33.030,13.954,58.720],
[51.006,7.098,56.261],[51.744,5.369,58.183],[50.550,4.145,55.831],
[49.846,1.567,53.384],[48.630,2.720,59.294],[50.704,-0.746,54.106],
[47.887,0.351,57.701],[50.948,-0.535,57.646],[54.238,0.137,53.547],
[34.600,-2.078,57.386],[49.020,13.821,59.139],[36.962,-3.062,56.527],
[43.600,-3.139,-58.571],[49.075,-5.329,58.191],[34.731,11.942,55.408],
[37.393,0.494,54.404],[41.295,0.466,57.106],[44.153,12.864,54.998],
[52.082,12.662,-58.757],[36.742,12.782,53.265],[33.904,21.041,56.870],
[47.566,18.204,59.048],[46.816,26.720,57.047],[34.672,27.320,56.290],
[47.876,29.094,57.622],[43.895,24.268,55.537],[45.113,22.861,53.487],
[48.126,24.252,58.362],[47.325,24.457,55.498],[45.946,30.001,59.646],
[51.102,15.609,55.591],[37.796,23.564,54.988],[51.016,30.861,53.744],
[43.967,33.290,57.131],[38.595,31.679,58.670],[46.947,32.740,54.898],
[46.319,31.273,57.367],[48.670,22.193,54.320],[43.547,18.816,54.855],
[41.081,21.748,57.683],[46.494,22.433,57.279],[39.371,19.676,51.902],
[36.480,20.196,56.765],[38.806,19.946,58.294],[38.780,18.914,55.764],
[40.200,21.884,54.873],[37.902,15.782,-59.619],[41.186,25.971,56.643],
[48.822,17.686,54.023],[46.324,20.000,54.414],[44.063,26.929,56.236],
[55.018,22.149,55.546],[49.064,30.942,55.831],[33.791,25.273,53.656],
[41.786,27.360,59.013],[42.039,31.381,57.417],[48.218,26.012,53.021],
[54.048,33.119,53.228],[48.769,27.289,55.385],[41.545,29.085,53.417],
[39.817,27.542,54.715],[36.411,25.858,54.426],[46.080,17.619,56.049],
[45.140,30.284,54.693],[43.330,20.539,58.236],[44.785,17.648,58.883],
[40.270,31.161,55.382],[40.107,17.542,53.297],[37.603,29.331,54.588],
[37.282,20.835,53.992],[37.139,29.086,58.307],[43.755,30.172,52.304],
[48.995,29.307,53.356],[40.809,27.109,51.802],[43.537,26.533,52.100],
[40.281,24.694,54.578],[42.524,23.555,53.172],[50.755,24.909,52.466],
[41.954,33.819,55.214],[35.411,26.815,52.180],[34.107,33.727,58.910],
[51.802,27.862,56.195],[51.037,24.513,58.276],[50.130,24.519,55.424],
[51.201,22.170,52.694],[48.197,21.702,-59.694],[49.821,19.715,52.396],
[47.952,19.773,56.709],[50.502,16.803,58.151],[51.613,18.529,54.160],
[34.714,16.275,59.477],[46.644,32.843,-59.333],[36.588,15.744,54.878],
[40.405,16.231,58.722],[47.576,15.762,58.119],[34.502,29.677,55.457],
[36.288,18.191,53.728],[41.030,17.456,56.285],[44.468,33.486,53.582],
[50.708,29.748,58.556],[36.524,32.847,52.608],[35.622,40.690,57.871],
[46.264,35.853,58.940],[44.983,45.106,58.001],[35.077,46.963,56.073],
[49.379,45.521,58.551],[43.442,43.999,55.434],[46.169,42.292,52.585],
[48.929,42.794,58.852],[46.010,42.709,55.747],[45.153,48.961,58.187],
[49.706,33.662,55.150],[39.253,44.104,57.476],[49.930,48.558,53.854],
[43.930,51.234,56.909],[37.638,50.949,56.915],[46.444,53.124,53.672],
[46.729,50.417,56.118],[49.005,42.163,52.114],[44.423,38.022,54.440],
[38.502,41.182,58.282],[45.387,41.803,58.616],[42.059,37.658,52.315],
[35.010,38.272,56.024],[36.584,38.193,58.147],[39.204,38.360,56.330],
[39.968,42.046,53.118],[38.564,36.306,58.190],[42.836,47.274,56.102],
[46.886,38.180,55.988],[48.294,40.682,56.634],[46.078,47.250,56.264],
[51.924,38.558,54.239],[49.222,47.428,56.226],[37.334,41.973,53.738],
[42.101,44.959,57.631],[40.228,50.524,57.513],[47.821,46.167,53.217],
[52.222,50.260,54.076],[48.154,45.053,55.727],[40.476,47.221,51.100],
[40.538,46.559,54.988],[37.399,45.135,54.802],[46.940,35.385,55.021],
[44.202,49.279,54.584],[42.454,37.537,57.750],[43.661,35.301,58.911],
[41.537,49.958,55.047],[39.626,35.119,54.713],[38.319,44.006,52.058],
[37.015,39.974,55.610],[35.791,47.954,58.583],[42.970,48.671,52.083],
[48.444,48.198,51.659],[41.244,44.885,53.006],[44.764,46.795,53.570],
[40.905,43.132,55.346],[42.205,42.739,52.035],[51.317,43.880,51.301],
[41.156,53.181,54.293],[33.972,44.969,53.843],[35.363,50.497,58.510],
[50.710,43.297,56.715],[51.292,40.677,56.761],[50.582,42.007,54.158],
[49.739,38.969,52.517],[49.925,39.701,59.143],[49.276,36.961,54.386],
[48.229,37.951,58.115],[50.633,35.716,56.444],[53.327,35.832,53.255],
[33.991,36.215,57.458],[47.239,52.447,59.024],[36.044,32.396,57.884],
[40.932,35.426,-59.570],[47.955,34.153,57.393],[34.953,49.733,55.848],
[37.574,36.652,54.843],[44.532,35.916,56.335],[43.886,51.761,54.253],
[50.070,50.585,58.179],[37.160,51.629,53.717],[47.016,54.828,58.089],
[51.187,51.970,55.987],[45.364,55.968,52.389],[39.042,58.988,58.854],
[41.126,55.787,52.372],[36.614,56.087,54.261],[36.966,56.623,-59.726],
[40.457,58.113,56.554],[38.356,58.642,54.475],[39.087,53.513,58.060],
[47.784,55.826,54.053],[46.205,58.732,54.710],[51.986,57.615,56.031],
[45.042,56.500,55.518],[43.531,58.762,56.913],[43.680,54.307,57.226],
[39.405,56.231,54.864],[35.602,58.467,53.267],[52.097,58.063,53.185],
[50.210,55.632,53.248],[48.718,57.947,57.831],[49.571,55.285,55.828],
[53.118,55.039,53.854],[35.023,55.841,56.561],[35.300,53.328,55.165],
[41.059,52.189,59.438],[48.207,52.619,56.502],[38.088,53.830,55.189],
[42.614,55.526,54.795],[39.573,-56.825,-13.003],[37.858,-59.402,-18.870],
[47.278,58.338,-16.357],[36.119,-48.148,-28.723],[44.171,-43.976,-32.121],
[37.515,-57.716,-26.785],[39.235,-50.328,-25.697],[49.409,-57.705,-24.197],
[48.065,-46.993,-28.448],[45.137,-48.639,-30.036],[42.029,-47.584,-32.424],
[37.411,-47.614,-34.959],[40.578,-53.976,-34.651],[47.995,-53.136,-21.295],
[35.898,-53.578,-37.090],[37.874,-55.573,-36.102],[47.591,-47.310,-24.175],
[40.598,-50.343,-22.478],[41.228,-50.894,-19.406],[43.224,-52.334,-19.801],
[35.724,-53.514,-18.553],[41.788,-56.372,-18.187],[32.941,-45.245,-35.479],
[33.518,-44.542,-31.615],[52.446,-44.175,-28.324],[47.550,-53.992,-28.961],
[51.192,-55.102,-15.317],[37.989,-43.422,-29.485],[41.091,-33.300,-37.296],
[49.905,-35.942,-34.844],[37.376,-28.934,-35.266],[51.563,-28.745,-34.179],
[38.460,-40.699,-38.434],[37.744,-41.087,-43.473],[44.139,-31.896,-33.521],
[33.341,-33.093,-32.359],[49.481,-30.499,-33.038],[44.771,-26.373,-31.311],
[39.444,-39.055,-24.923],[49.620,-39.771,-39.657],[48.177,-46.813,-41.970],
[35.086,-27.418,-32.062],[37.966,-29.647,-41.072],[42.401,-42.748,-33.562],
[49.359,-44.195,-35.362],[47.697,-16.539,-48.260],[39.088,-21.346,-44.807],
[42.883,-15.105,-53.638],[48.199,-19.546,-37.358],[46.687,-21.479,-38.672],
[48.377,-8.046,-44.670],[46.101,-27.864,-49.508],[44.517,-12.703,-43.740],
[50.890,-25.624,-50.194],[38.279,-18.207,-54.522],[38.904,-22.660,-55.524],
[43.848,-21.979,-39.082],[45.289,-22.139,-34.399],[50.837,-23.464,-52.354],
[47.185,-9.956,-46.261],[42.935,-16.206,-56.726],[47.172,-0.998,-57.060],
[51.644,-7.276,-51.181],[39.037,-11.372,-53.600],[41.016,-11.785,-50.917],
[38.254,-3.927,-54.291],[38.137,3.914,-51.891],[49.690,-15.886,-55.929],
[42.570,2.473,-54.361],[41.556,-7.818,-47.274],[44.917,-2.986,-51.916],
[44.652,-8.372,-51.251],[42.044,-6.461,-55.098],[51.841,-9.110,58.400],
[52.682,1.229,-50.998],[44.173,-7.321,-48.076],[38.188,-12.201,-59.066],
[36.165,-9.913,-52.350],[39.466,0.819,-48.277],[41.231,-5.890,-41.444],
[32.459,-7.739,-52.939],[37.477,3.471,-55.149],[43.389,5.883,-51.966],
[37.610,5.800,-57.134],[49.523,-11.077,-52.165],[36.952,-13.950,-54.458],
[37.137,7.304,58.413],[44.827,11.266,-56.907],[47.562,15.700,-56.538],
[37.760,12.119,-55.854],[49.538,16.278,-58.388],[51.966,8.724,-51.767],
[37.406,8.854,-52.967],[44.077,-58.553,-1.251],[38.423,-53.399,0.680],
[45.818,-57.395,-6.488],[38.997,-58.959,-0.362],[42.764,-52.362,0.993],
[46.784,-58.767,-1.852],[41.802,-56.227,-10.860],[34.196,-58.296,5.355],
[38.145,-50.056,-4.094],[34.044,-47.320,-14.814],[37.662,-38.111,-17.179],
[43.529,-51.866,-14.646],[44.174,-47.512,-13.999],[43.502,-41.831,-3.664],
[35.912,-48.491,-10.527],[38.023,-40.623,-15.417],[39.515,-52.076,-13.070],
[34.390,-51.081,-17.612],[47.268,-37.093,-7.854],[37.528,-42.893,-18.347],
[39.587,-39.356,-6.088],[43.835,-38.583,-3.346],[50.178,-47.120,-19.975],
[47.219,-45.181,-0.122],[50.791,-33.405,-13.422],[51.533,-45.981,-14.668],
[35.616,-33.463,-8.928],[38.760,-33.368,-13.818],[51.577,-28.634,-15.434],
[36.438,-28.084,-15.624],[37.824,-21.735,-17.966],[34.152,-27.565,-27.818],
[38.485,-26.160,-23.240],[44.960,-20.848,-20.991],[39.361,-19.697,-17.126],
[38.011,-28.829,-17.783],[44.245,-23.635,-12.910],[39.613,-31.453,-22.415],
[48.233,-25.119,-19.482],[36.250,-24.901,-20.525],[43.742,-28.090,-27.775],
[45.543,-25.914,-19.979],[34.243,-27.036,-14.470],[41.996,-26.998,-13.950],
[43.214,-33.315,-19.268],[41.322,-25.463,-33.578],[51.028,-29.850,-27.409],
[51.244,-25.743,-15.188],[50.690,-30.406,-12.118],[35.761,-37.601,-22.230],
[38.278,-29.920,-8.572],[32.647,-14.056,-31.785],[43.822,-12.692,-33.136],
[39.062,-14.242,-34.251],[46.221,-15.194,-24.444],[49.500,0.858,-27.519],
[50.246,-16.856,-34.506],[47.859,-12.426,-37.735],[36.617,-11.872,-39.673],
[42.611,-11.097,-21.179],[33.412,-21.238,-30.589],[47.730,7.073,-41.118],
[34.236,4.749,-40.242],[40.308,7.577,-39.114],[39.531,7.716,-36.129],
[37.343,2.794,-37.055],[45.461,4.507,-42.748],[38.384,5.861,-40.486],
[35.032,8.490,-34.648],[38.346,14.416,-37.365],[50.504,6.379,-48.325],
[46.774,9.222,-50.223],[39.375,-4.492,-44.985],[41.893,3.605,-35.799],
[43.132,10.231,-34.996],[43.813,-2.168,-40.390],[41.732,6.097,-49.242],
[34.924,9.278,-31.325],[45.293,10.473,-32.578],[51.103,6.751,-31.762],
[49.311,-1.008,-39.363],[34.356,14.189,-40.031],[41.010,24.767,-48.079],
[46.690,16.639,-48.848],[35.705,18.710,-47.661],[49.579,18.780,-42.112],
[47.908,16.888,-42.497],[46.371,27.526,-48.892],[45.075,25.006,59.205],
[49.255,22.287,-39.664],[39.792,16.220,-43.738],[37.577,20.764,-42.964],
[39.000,16.760,-54.484],[42.977,20.824,-56.810],[50.958,21.439,-58.574],
[38.844,26.401,-55.142],[35.200,23.677,-55.696],[41.427,32.267,-44.428],
[40.721,14.975,-55.471],[44.891,20.812,-40.668],[34.463,27.228,-53.188],
[51.029,29.560,-50.983],[51.474,20.372,-53.234],[51.467,11.051,-50.380],
[42.632,30.920,-50.800],[37.556,26.797,-52.902],[42.658,24.207,58.138],
[37.420,18.385,-36.755],[39.570,32.895,-57.313],[38.582,44.891,-53.664],
[49.637,26.765,58.319],[35.226,42.466,-55.519],[45.807,29.432,-56.601],
[36.953,34.386,-47.065],[44.112,-49.905,10.917],[50.957,-49.999,4.208],
[43.045,-43.554,7.014],[33.513,-44.595,15.146],[51.273,-46.773,4.649],
[46.023,-57.219,12.868],[38.640,-57.471,11.313],[38.534,-52.085,3.848],
[44.860,-52.561,6.284],[45.432,-46.856,24.169],[50.087,-50.245,18.510],
[49.007,-57.797,12.521],[52.584,-51.878,22.936],[44.362,-39.381,15.085],
[46.189,-39.358,12.191],[33.759,-43.299,9.120],[46.265,-57.288,17.367],
[44.747,-47.366,4.996],[36.406,-29.741,1.706],[36.236,-36.485,-0.146],
[32.696,-28.720,2.547],[42.946,-39.832,12.626],[36.826,-39.326,-1.010],
[41.391,-34.808,-0.646],[40.060,-29.451,11.932],[41.160,-33.598,1.610],
[34.475,-26.744,3.330],[45.349,-40.956,2.813],[39.658,-39.014,11.642],
[49.845,-24.990,5.557],[48.905,-30.499,4.610],[48.716,-42.957,-0.328],
[48.047,-40.909,2.628],[39.699,-19.971,-6.173],[36.859,-19.978,1.811],
[38.213,-12.628,-4.104],[46.568,-19.763,-9.480],[37.154,-15.559,-0.153],
[50.390,-16.670,-14.087],[43.145,-10.461,-1.995],[41.518,-11.487,-12.717],
[39.008,-25.340,-0.394],[44.380,-20.945,-3.638],[39.890,-22.321,-7.564],
[44.591,-14.380,-0.712],[43.404,-17.371,-11.255],[35.954,-22.711,-9.264],
[37.797,-5.915,1.769],[40.441,-12.611,1.882],[44.042,-16.381,3.686],
[50.971,-20.666,-10.301],[48.656,-20.212,-6.031],[44.565,-9.365,-6.192],
[46.774,-25.792,-0.608],[42.961,-13.152,-14.775],[49.050,-13.013,-10.992],
[44.219,-21.427,7.040],[42.844,-8.137,-9.159],[42.999,2.187,-18.429],
[49.066,-12.016,-13.789],[38.248,-3.517,-17.261],[38.897,-10.103,-20.146],
[45.092,-7.888,-18.234],[36.989,0.052,-16.023],[41.513,-5.603,-14.941],
[39.077,-10.726,-17.127],[43.275,0.380,-26.208],[43.224,-2.778,-24.058],
[39.236,-4.368,-7.842],[43.351,-1.649,-5.295],[50.363,8.352,-13.722],
[50.700,2.209,-14.091],[32.280,-8.180,-15.108],[45.043,10.051,-14.337],
[33.710,8.185,-14.032],[34.672,6.195,-17.795],[44.465,-5.757,-13.414],
[35.654,-8.864,-12.380],[43.328,13.420,-20.811],[48.065,16.954,-25.580],
[34.398,12.700,-21.447],[38.143,18.976,-21.181],[38.151,14.360,-25.701],
[41.951,10.226,-14.333],[44.381,6.275,-26.709],[39.555,19.848,-18.521],
[39.283,16.726,-17.815],[51.231,11.591,-31.867],[51.493,11.139,-12.126],
[51.224,13.760,-12.918],[32.514,10.202,-15.843],[50.158,11.575,-14.819],
[46.156,7.380,-19.498],[33.449,29.540,-34.419],[38.297,28.652,-36.751],
[33.767,35.594,-30.429],[41.300,22.440,-37.950],[38.326,27.851,-33.893],
[46.075,29.906,-33.316],[32.362,26.154,-32.573],[40.140,26.556,-31.282],
[36.452,26.404,-35.393],[36.618,23.075,-32.842],[40.707,28.053,-28.561],
[43.525,36.523,-29.738],[49.167,32.044,-39.292],[39.353,33.427,-37.721],
[44.100,32.603,-40.782],[48.183,28.957,-24.269],[39.245,27.921,-42.208],
[37.897,32.358,-23.848],[50.204,39.194,-32.072],[50.134,27.680,-38.971],
[50.029,25.174,-37.787],[34.946,39.593,-27.990],[49.493,22.710,-30.176],
[37.381,21.690,-36.171],[38.291,43.119,-39.745],[46.203,47.852,-37.589],
[37.998,53.108,-39.825],[50.448,47.049,-52.971],[33.327,51.117,-44.305],
[36.493,41.694,-37.860],[42.207,41.527,-33.245],[45.258,43.127,-44.619],
[43.994,53.666,-39.156],[52.216,49.191,-53.849],[36.309,45.787,-48.802],
[48.662,46.269,-34.620],[47.210,45.479,-42.400],[48.432,55.021,-44.579],
[43.676,45.017,-51.797],[38.730,49.350,-36.394],[49.592,42.935,-48.720],
[44.542,46.344,-31.145],[34.864,45.984,-31.773],[34.383,46.565,-51.047],
[31.344,53.802,-46.580],[50.246,49.388,-33.337],[32.859,48.437,-31.453],
[50.970,44.095,-32.389],[42.442,54.120,-43.821],[39.427,-42.802,29.859],
[34.263,-40.827,24.738],[42.745,-45.555,22.964],[45.841,-40.480,35.499],
[44.830,-46.247,27.706],[43.903,-39.256,33.680],[44.549,-40.977,22.483],
[40.757,-41.755,20.141],[42.107,-46.167,27.629],[48.254,-42.813,36.624],
[34.162,-47.938,23.671],[49.291,-44.539,21.056],[39.029,-41.888,37.662],
[32.830,-33.766,23.110],[48.585,-33.148,36.113],[50.685,-45.819,22.930],
[46.830,-45.170,28.871],[50.139,-40.828,36.236],[50.757,-45.294,43.067],
[36.947,-21.877,19.914],[44.896,-25.862,23.777],[47.091,-31.295,28.047],
[37.649,-24.159,18.135],[51.274,-15.544,23.680],[41.743,-20.033,12.755],
[46.072,-27.988,13.943],[37.354,-28.738,20.428],[46.455,-28.600,24.682],
[48.586,-11.816,22.452],[42.210,-26.372,22.235],[45.001,-30.508,16.356],
[42.149,-32.119,20.946],[36.083,-24.498,12.046],[41.703,-25.909,29.907],
[35.774,-24.843,29.176],[35.401,-27.590,24.327],[32.783,-26.306,23.955],
[38.928,-15.357,19.024],[43.304,-15.301,17.546],[34.621,-21.988,12.100],
[50.778,-9.486,16.886],[45.979,-5.283,7.360],[41.899,-11.799,7.575],
[39.563,-4.482,4.907],[35.217,-10.456,4.691],[46.676,-7.401,0.626],
[49.885,-10.860,3.999],[39.233,-10.350,20.206],[40.340,-7.864,19.543],
[33.048,-10.782,2.945],[32.635,-1.955,6.658],[49.313,-1.234,18.891],
[48.501,-11.652,7.070],[32.924,-7.501,20.255],[52.275,-11.790,23.513],
[50.871,-9.458,22.075],[33.925,0.729,16.058],[40.024,-1.663,12.341],
[34.403,-10.731,20.488],[36.215,8.520,4.646],[43.032,9.942,2.397],
[49.137,3.451,9.277],[50.365,8.660,5.258],[31.977,0.845,1.856],
[43.365,13.320,5.528],[48.126,10.304,-5.659],[41.970,7.925,-9.030],
[39.405,5.036,-4.650],[33.317,7.961,2.063],[32.064,5.724,3.331],
[44.192,24.504,-7.288],[44.321,29.656,-13.198],[33.093,29.048,-8.800],
[36.430,29.048,-12.758],[48.000,35.993,-6.006],[48.936,27.040,-3.608],
[44.725,15.235,-10.022],[43.678,23.646,-14.542],[39.117,31.217,-5.643],
[34.853,30.937,-7.704],[41.514,24.038,-16.792],[35.951,25.703,-17.339],
[34.946,26.895,-13.524],[41.833,23.736,2.094],[33.244,29.476,-12.987],
[49.362,33.106,-0.121],[49.940,18.906,4.765],[48.220,17.413,-2.211],
[33.607,27.877,-17.066],[43.697,38.724,-19.426],[43.024,43.134,-17.303],
[46.267,33.056,-15.349],[50.332,39.636,-12.057],[50.401,46.752,-12.527],
[41.215,47.400,-15.587],[34.560,45.864,-13.750],[38.885,34.742,-19.552],
[41.041,37.506,-10.710],[46.134,33.935,-19.814],[37.134,40.997,-16.261],
[45.230,46.605,-14.144],[41.572,32.141,-17.864],[38.168,31.883,-20.129],
[36.167,39.888,-25.158],[45.214,43.974,-6.102],[36.167,39.850,-5.473],
[40.254,35.498,-6.255],[34.916,34.456,-11.503],[49.206,45.494,-22.875],
[41.970,50.065,-16.436],[38.847,49.509,-19.685],[35.701,30.619,-19.989],
[49.929,49.337,-14.191],[45.620,58.272,-26.464],[48.257,-59.550,-33.036],
[46.820,54.266,-25.763],[39.100,52.108,-18.680],[44.670,-59.751,-14.711],
[37.448,55.246,-15.541],[34.379,57.837,-35.257],[35.130,-33.729,45.737],
[42.559,-28.248,41.866],[38.335,-23.388,49.169],[46.678,-26.435,37.463],
[42.278,-28.666,53.851],[34.849,-36.432,45.736],[51.161,-35.515,41.934],
[34.346,-28.738,42.177],[42.357,-31.639,34.923],[40.695,-33.406,57.283],
[50.744,-32.176,44.136],[50.924,-34.322,39.298],[49.738,-35.039,54.592],
[36.050,-21.472,43.908],[48.664,-20.779,44.824],[41.953,-37.218,49.139],
[48.940,-38.817,49.981],[44.097,-38.909,50.129],[37.246,-10.725,40.331],
[45.893,-9.810,32.404],[32.254,-11.193,35.477],[51.858,-7.722,40.068],
[42.218,-8.364,38.931],[43.757,-20.947,44.628],[35.220,-9.263,41.493],
[37.684,-22.084,36.861],[50.721,-10.415,42.845],[38.623,-14.593,40.412],
[50.260,-14.128,44.628],[49.365,-22.190,34.786],[32.661,-17.931,38.335],
[46.575,-21.633,39.201],[33.827,-1.601,24.468],[39.091,1.876,28.902],
[39.575,6.661,30.956],[44.345,4.133,22.565],[37.145,6.116,20.550],
[43.751,-8.233,23.271],[37.662,-0.810,30.796],[51.683,-5.347,24.692],
[42.177,3.095,15.907],[40.262,-4.725,39.209],[40.044,-1.113,39.289],
[49.684,-0.411,21.465],[49.369,3.326,32.856],[50.805,-8.226,24.795],
[44.091,10.976,28.821],[35.813,9.327,24.756],[34.827,-9.898,23.254],
[51.507,3.437,18.817],[46.305,-3.276,41.205],[48.633,10.471,31.553],
[43.795,15.433,16.335],[51.395,11.760,23.728],[38.368,11.533,19.192],
[46.677,10.682,19.369],[41.186,23.904,20.753],[40.428,6.951,13.523],
[43.071,23.368,22.660],[40.533,19.822,25.491],[46.181,17.015,9.665],
[47.157,16.561,12.096],[42.258,18.625,7.238],[39.302,14.240,12.825],
[48.463,6.819,14.125],[33.843,8.226,15.703],[42.409,27.154,16.992],
[44.528,24.965,17.898],[38.927,32.852,7.125],[47.752,34.953,16.224],
[44.053,29.902,5.932],[37.198,27.242,3.086],[36.292,25.327,10.008],
[41.574,31.471,14.426],[45.884,27.553,6.837],[50.993,26.221,5.760],
[47.449,35.477,11.131],[34.277,30.435,15.488],[40.863,35.187,2.037],
[44.754,31.853,15.878],[34.088,26.078,2.718],[33.627,38.490,16.577],
[37.412,31.886,0.467],[38.092,33.193,17.767],[50.942,29.879,3.201],
[50.396,27.335,3.268],[52.069,28.085,20.910],[50.757,23.162,5.173],
[41.577,37.687,6.522],[51.005,44.462,4.735],[50.145,47.688,4.735],
[39.521,51.772,0.384],[36.848,45.274,2.727],[48.489,52.086,-1.979],
[45.026,46.737,-6.094],[49.911,58.517,2.456],[45.003,53.188,-11.061],
[41.016,48.713,0.670],[39.597,54.048,-7.516],[45.062,48.248,5.897],
[47.312,47.754,-0.250],[45.248,46.063,-8.782],[34.470,53.410,5.516],
[38.762,44.128,-9.313],[40.769,49.229,7.062],[48.795,40.555,-1.840],
[34.679,43.417,2.212],[33.979,47.033,2.187],[42.966,57.058,-2.695],
[42.069,42.351,1.893],[41.828,-17.258,54.793],[48.331,-25.388,55.172],
[43.014,-31.759,57.692],[33.899,-30.496,58.720],[39.032,-26.307,57.717],
[51.818,-27.402,58.167],[48.670,-20.924,55.033],[36.022,-32.054,56.463],
[34.225,-5.396,52.035],[52.086,-6.080,59.157],[36.583,-9.364,57.930],
[38.215,-2.891,53.735],[40.704,-13.345,58.521],[39.374,-4.301,56.237],
[36.337,-2.905,-59.020],[37.085,3.043,56.810],[42.589,1.543,54.189],
[43.073,-0.652,45.254],[39.324,-7.852,57.095],[40.578,-11.815,50.696],
[42.570,-7.873,57.035],[44.740,-14.063,49.510],[39.912,-2.020,49.103],
[37.070,-12.254,57.268],[42.801,-5.594,40.142],[39.321,-1.911,58.947],
[44.226,-8.271,43.016],[41.537,-3.409,57.567],[42.038,3.284,56.383],
[33.604,-2.508,44.332],[36.383,-17.146,50.233],[50.015,1.578,56.155],
[43.342,12.196,45.470],[35.284,6.079,46.481],[39.692,13.965,43.733],
[36.178,8.515,40.091],[41.974,17.302,46.336],[40.301,17.179,42.811],
[39.975,14.137,37.218],[44.533,14.327,52.884],[40.147,7.556,50.398],
[51.266,8.244,41.209],[50.832,-0.312,40.928],[35.687,2.241,46.886],
[37.158,23.253,35.692],[38.320,32.577,36.395],[48.047,31.338,37.599],
[38.038,35.331,37.838],[32.289,30.666,33.133],[42.924,17.676,30.192],
[42.778,27.190,41.642],[41.600,23.998,39.000],[33.537,29.279,39.810],
[35.471,22.695,28.111],[46.194,35.327,35.804],[52.527,30.145,23.022],
[40.892,25.666,24.199],[43.513,22.472,40.467],[48.918,23.431,42.588],
[42.318,29.136,22.544],[41.624,28.860,38.962],[50.283,32.963,34.790],
[51.054,29.236,41.620],[48.814,23.890,38.494],[50.345,25.533,43.571],
[32.672,20.550,42.464],[36.327,31.094,30.279],[49.126,18.579,36.937],
[44.938,29.842,23.715],[44.710,47.219,23.369],[32.128,47.918,24.014],
[37.656,47.534,18.767],[38.523,36.819,29.673],[43.964,39.553,22.592],
[39.341,35.495,19.394],[41.600,41.239,20.009],[39.843,44.942,20.531],
[36.460,41.735,18.474],[40.112,51.601,28.145],[39.522,43.341,32.199],
[33.319,47.288,20.984],[49.992,49.593,23.858],[48.644,49.128,18.884],
[37.857,51.159,22.844],[42.424,50.041,23.344],[37.334,39.520,37.622],
[49.869,57.724,21.968],[47.613,53.713,20.700],[37.892,52.648,20.321],
[45.909,55.086,13.021],[40.710,56.275,6.729],[35.776,51.075,21.209],
[46.531,7.948,-59.485],[34.687,6.786,57.633],[36.860,23.010,58.411],
[44.171,21.695,55.895],[46.062,16.383,53.673],[36.651,17.170,57.233],
[44.885,23.158,48.478],[38.598,25.373,52.548],[37.905,12.367,57.669],
[40.199,29.531,-59.108],[41.907,28.582,56.131],[39.551,24.111,57.848],
[38.421,40.121,50.215],[34.486,33.657,48.997],[44.756,30.763,43.460],
[39.061,36.008,49.402],[37.248,33.137,55.351],[36.194,43.163,57.011],
[48.218,44.544,50.825],[41.193,40.792,57.444],[32.552,41.247,42.197],
[48.330,31.646,52.664],[47.756,28.330,49.973],[33.612,47.910,39.574],
[44.377,-59.026,35.749],[36.526,46.874,43.654],[48.469,45.772,45.725],
[46.717,53.329,46.570],[39.994,51.854,43.154],[48.329,52.782,48.864],
[33.665,44.703,41.753],[39.126,46.142,38.881],[38.852,52.812,40.108],
[42.588,43.911,42.628],[40.531,54.557,31.502],[36.230,44.336,47.060],
[39.550,51.071,49.813],[45.752,53.602,49.550],[49.561,-58.115,42.092],
[50.580,45.019,34.950],[49.311,44.648,40.432],[49.948,47.889,49.062],
[36.398,56.126,39.098],[49.373,44.485,43.504],[37.012,42.189,37.285],
[43.440,53.420,31.640],[35.417,49.109,53.115],[45.949,49.157,52.451],
[55.030,-49.884,-48.120],[57.970,-46.479,-49.058],[52.848,-50.892,-54.776],
[54.216,-49.641,-52.213],[55.042,-53.493,-51.069],[53.104,-51.996,-51.979],
[56.655,-56.550,-46.623],[56.554,-52.237,-48.422],[56.996,-47.038,-44.359],
[55.994,-46.875,-54.112],[57.229,-49.234,-53.782],[57.840,-49.434,-50.522],
[55.904,-49.810,-45.054],[59.674,-43.533,-43.232],[57.330,-51.430,-55.023],
[53.790,-48.021,-46.659],[52.957,-46.195,-50.919],[55.316,-42.390,-51.627],
[53.845,-46.644,-44.498],[58.319,-56.480,-49.142],[58.091,-52.872,-46.060],
[58.056,-51.516,-52.193],[55.619,-53.615,-46.367],[-58.446,-56.638,-58.583],
[54.099,-56.557,-50.861],[54.534,-54.252,-56.723],[55.092,-52.835,-54.437],
[-59.651,-53.789,-55.607],[57.929,-52.657,-57.573],[55.586,-47.453,-41.796],
[56.785,-55.951,-51.960],[55.197,-47.015,-49.401],[54.182,-50.806,-42.535],
[57.632,-46.118,-52.180],[54.128,-49.471,-58.638],[57.493,-41.814,-53.567],
[54.454,-56.507,-45.162],[54.413,-45.016,-46.627],[54.585,-47.862,-56.440],
[57.458,-57.053,-56.526],[57.754,-44.862,-56.331],[53.806,-52.549,-58.689],
[56.035,-58.040,-49.692],[53.369,-44.534,-53.570],[51.824,-41.267,-43.174],
[59.051,-41.731,-40.467],[52.767,-48.001,-39.815],[58.274,-46.980,-58.275],
[56.899,-49.981,-42.268],[58.827,-50.211,-57.020],[58.610,-50.620,-44.409],
[58.478,-55.710,-44.250],[-59.211,-55.267,-46.618],[56.481,-53.010,-41.647],
[55.884,-55.757,-43.194],[53.263,-46.164,-58.553],[53.707,-40.980,-53.321],
[51.658,-41.654,-51.723],[52.797,-49.400,-49.787],[51.982,-55.874,-43.193],
[54.226,-56.715,-48.027],[53.720,-52.188,-48.937],[52.441,-55.608,-52.926],
[53.850,-59.025,-46.479],[55.301,-44.289,-42.930],[57.569,-44.278,-44.812],
[55.179,-41.540,-44.350],[56.833,-41.112,-46.386],[53.880,-40.931,-48.199],
[55.566,-43.810,-49.047],[56.453,-59.280,-51.982],[54.731,-59.018,-55.106],
[54.697,-43.553,-56.502],[55.724,-57.937,-59.638],[55.398,-58.519,-42.330],
[58.616,-58.070,-40.328],[55.460,-41.424,-58.697],[56.219,-32.576,-47.500],
[57.241,-28.979,-48.227],[51.517,-33.725,-53.802],[52.220,-30.072,-52.134],
[57.316,-36.734,-50.307],[53.733,-34.009,-50.376],[56.393,-38.378,-46.787],
[56.021,-35.924,-47.986],[57.495,-28.708,-45.236],[54.015,-26.799,-54.989],
[58.236,-30.440,-52.365],[58.918,-33.629,-49.718],[57.215,-31.837,-44.459],
[58.965,-26.415,-44.063],[58.252,-32.388,-54.366],[52.503,-29.930,-45.204],
[53.272,-26.348,-48.389],[56.764,-22.771,-52.536],[54.705,-28.558,-45.002],
[58.329,-39.827,-48.275],[58.780,-33.057,-46.810],[57.372,-33.182,-51.967],
[54.586,-34.469,-45.480],[59.220,-36.287,-58.147],[52.842,-36.598,-52.304],
[53.904,-36.529,-59.190],[56.443,-35.549,-52.718],[58.792,-35.337,-54.405],
[56.723,-33.559,-57.072],[55.997,-29.398,-42.450],[58.083,-37.863,-52.801],
[55.223,-30.000,-51.863],[54.575,-31.651,-45.065],[-59.285,-27.850,-52.722],
[53.936,-31.198,-56.655],[57.659,-24.768,-56.455],[53.499,-36.890,-43.817],
[53.459,-25.645,-45.588],[55.374,-29.525,-54.761],[58.034,-38.807,-56.665],
[54.576,-25.614,-57.781],[54.105,-33.714,-58.201],[57.191,-39.950,-51.133],
[52.626,-25.532,-53.246],[52.444,-23.234,-43.627],[-59.336,-25.301,-40.856],
[52.438,-29.634,-42.191],[56.269,-29.361,-58.136],[54.295,-32.828,-42.600],
[58.865,-31.357,-56.807],[58.186,-32.432,-42.011],[58.483,-37.369,-42.799],
[58.748,-36.243,-45.494],[56.732,-35.020,-42.442],[55.590,-37.533,-41.592],
[53.348,-27.853,-57.184],[54.543,-21.527,-53.907],[53.545,-22.899,-51.542],
[53.152,-29.915,-49.656],[52.523,-36.306,-41.306],[53.742,-38.473,-49.854],
[53.071,-36.102,-48.131],[53.552,-38.455,-55.008],[52.540,-38.548,-46.540],
[55.086,-25.080,-41.939],[56.308,-24.655,-45.059],[54.255,-22.965,-45.945],
[58.759,-22.530,-48.111],[55.364,-24.757,-47.773],[57.506,-23.887,-50.171],
[55.746,-39.165,-53.350],[55.117,-40.586,-55.715],[55.583,-25.151,-53.764],
[55.406,-38.005,-57.325],[54.367,-40.054,-42.398],[56.933,-40.106,-41.274],
[55.845,-22.684,-57.302],[54.196,-12.141,-46.410],[57.535,-9.659,-49.507],
[54.296,-16.657,-52.934],[53.085,-14.230,-51.125],[57.499,-16.867,-50.480],
[54.307,-16.499,-50.052],[56.873,-19.351,-46.862],[56.400,-16.997,-48.060],
[56.822,-10.546,-45.839],[53.385,-9.734,-56.035],[57.861,-12.218,-50.461],
[56.927,-14.337,-48.582],[56.520,-13.963,-43.922],[56.265,-6.804,-43.638],
[57.179,-13.218,-55.271],[52.353,-11.007,-44.860],[52.604,-8.045,-48.488],
[56.356,-4.897,-51.717],[54.410,-8.594,-43.770],[58.786,-19.559,-49.014],
[58.444,-13.718,-46.307],[58.769,-14.764,-52.085],[54.963,-15.191,-46.547],
[58.972,-17.167,-58.295],[52.148,-18.091,-51.597],[53.741,-18.088,-57.201],
[56.101,-14.686,-51.544],[56.966,-15.737,-53.955],[-59.608,-16.012,-55.474],
[56.641,-10.662,-43.007],[58.221,-18.766,-52.303],[55.542,-12.069,-49.160],
[53.982,-14.311,-44.312],[57.891,-8.569,-52.052],[52.994,-13.303,-58.690],
[58.860,-5.889,-52.515],[54.236,-17.823,-44.974],[52.842,-8.490,-45.840],
[57.122,-10.004,-57.007],[58.351,-19.455,-56.358],[56.909,-6.632,-56.607],
[54.347,-15.383,-57.788],[55.193,-21.535,-48.428],[55.255,-7.970,-53.397],
[54.314,-4.835,-41.885],[58.715,-5.796,-39.116],[53.664,-11.141,-41.849],
[57.839,-11.338,58.549],[54.760,-14.095,-41.604],[-58.537,-9.043,-55.953],
[58.523,-16.520,-43.422],[56.407,-19.068,-43.665],[59.208,-17.733,-47.235],
[56.711,-16.154,-41.099],[55.425,-19.084,-40.689],[53.971,-8.369,-58.228],
[52.345,-4.484,-52.293],[53.715,-5.535,-50.263],[52.597,-12.342,-48.958],
[52.461,-16.336,-42.642],[54.335,-19.011,-49.544],[52.548,-15.024,-48.105],
[53.022,-18.597,-54.596],[53.751,-20.091,-46.462],[57.465,-8.483,-41.298],
[58.404,-7.010,-47.900],[54.956,-3.273,-46.091],[57.848,-4.454,-46.972],
[53.801,-5.827,-47.360],[56.153,-5.872,-48.643],[56.819,-20.851,-50.567],
[55.449,-19.986,-56.246],[54.083,-6.422,-56.051],[54.770,-20.505,-58.792],
[53.943,-20.360,-43.604],[58.102,-22.675,-41.254],[55.723,-3.383,-56.612],
[56.084,6.359,-46.047],[57.106,8.226,-48.024],[52.108,2.974,-53.056],
[53.295,5.800,-51.708],[57.925,0.374,-50.188],[55.456,1.314,-51.034],
[57.218,-0.028,-47.728],[54.169,1.903,-48.660],[56.296,9.391,-45.815],
[53.650,8.078,-53.884],[-57.041,8.308,-53.631],[58.495,4.706,-46.054],
[54.733,2.133,-43.828],[56.267,11.912,-43.461],[56.340,6.504,-54.635],
[53.721,8.106,-45.502],[54.662,9.288,-49.764],[55.754,12.871,-53.047],
[54.139,10.250,-43.929],[58.313,-3.775,-49.837],[57.599,2.651,-47.958],
[58.795,2.717,-52.743],[54.749,3.023,-46.359],[58.924,0.867,-59.558],
[52.534,-0.805,-52.948],[54.010,1.096,-57.030],[55.680,1.008,-53.885],
[58.537,0.804,-55.113],[58.381,2.652,-57.303],[54.289,7.611,-40.193],
[57.653,-0.488,-52.858],[55.728,4.350,-48.603],[53.593,5.194,-45.395],
[58.247,10.457,-54.016],[53.450,6.291,-58.765],[58.462,13.511,-52.907],
[55.150,-1.063,-44.445],[54.890,11.700,-46.301],[56.292,7.350,-57.543],
[59.679,-0.609,-57.107],[56.405,10.464,-58.067],[53.869,3.508,-58.340],
[55.563,-3.179,-49.032],[54.957,10.236,-53.058],[51.919,14.317,-43.053],
[56.940,13.059,-41.264],[52.676,6.919,-42.546],[58.251,8.903,-59.027],
[55.130,5.682,-42.435],[58.451,5.862,-56.459],[58.358,2.412,-44.113],
[56.617,0.751,-42.395],[-58.034,0.721,-45.581],[56.948,3.312,-41.730],
[54.320,-0.501,-41.845],[54.349,9.134,-56.528],[53.072,13.857,-52.459],
[54.337,12.031,-50.603],[53.783,6.809,-47.824],[52.081,1.066,-43.326],
[54.506,-0.631,-49.485],[52.627,4.503,-48.997],[52.072,-0.168,-55.572],
[52.345,-2.230,-45.517],[55.763,10.134,-41.172],[57.922,11.320,-45.839],
[53.857,14.680,-46.214],[58.147,14.640,-45.786],[55.974,13.809,-48.145],
[56.844,11.163,-48.731],[54.726,-2.839,-51.513],[53.575,-4.108,-54.837],
[55.526,11.764,-55.702],[56.791,-0.552,-57.178],[55.895,-2.507,-41.286],
[58.197,-4.067,-41.697],[56.759,13.435,-57.943],[55.440,23.433,-48.137],
[55.487,28.841,-50.189],[54.474,22.482,-53.792],[53.726,24.572,-51.389],
[57.304,19.425,-51.021],[53.759,21.705,-51.468],[56.784,18.535,-47.433],
[55.451,20.874,-49.462],[59.182,27.510,-45.837],[53.084,28.100,-55.555],
[57.291,23.905,-50.455],[-59.557,23.518,-48.207],[56.294,23.421,-45.225],
[58.288,30.258,-43.697],[56.600,22.467,-55.818],[52.436,26.490,-45.937],
[53.311,29.951,-48.549],[56.769,31.601,-51.155],[54.550,27.775,-45.146],
[56.939,16.159,-50.054],[57.991,21.245,-46.716],[58.482,22.602,-52.744],
[55.172,20.707,-46.490],[58.828,19.085,-59.008],[52.917,18.258,-51.508],
[55.232,19.968,-56.210],[55.573,19.801,-53.162],[58.123,20.343,-55.647],
[-58.984,21.772,-57.688],[56.294,25.535,-43.313],[57.577,17.950,-53.190],
[55.668,25.951,-49.615],[53.377,23.174,-44.134],[58.179,29.231,-52.194],
[53.311,24.681,-58.450],[57.516,31.772,-53.880],[53.689,18.927,-44.058],
[55.509,30.084,-46.247],[56.588,26.241,-57.371],[57.426,18.193,-57.060],
[55.398,29.390,-57.314],[55.110,22.279,-58.037],[54.398,15.186,-49.437],
[55.120,29.329,-52.877],[54.289,32.538,-41.849],[58.315,32.407,-40.986],
[53.039,25.536,-41.773],[59.173,26.849,59.692],[54.964,23.518,-41.953],
[58.328,24.162,-57.111],[-58.692,23.373,-43.033],[58.386,21.542,-43.988],
[59.315,17.852,-47.309],[57.294,20.662,-41.492],[56.178,18.234,-43.047],
[54.844,27.482,59.667],[53.049,33.647,-52.802],[53.736,31.827,-51.081],
[52.871,24.936,-48.894],[53.299,21.096,-42.212],[54.967,18.249,-49.931],
[52.852,21.761,-49.003],[53.883,17.979,-55.018],[52.252,16.961,-45.725],
[54.892,30.204,-40.771],[58.734,30.197,-46.364],[55.468,34.338,-44.840],
[57.259,32.418,-46.213],[55.878,31.170,-48.737],[58.088,29.981,-49.768],
[55.194,15.627,-52.311],[53.246,15.203,-55.599],[53.667,31.148,-56.319],
[54.748,18.033,-58.292],[54.156,16.095,-42.712],[57.931,16.418,-41.500],
[54.834,33.138,-55.046],[56.536,43.135,-47.756],[57.302,47.475,-49.879],
[54.161,41.122,-54.479],[53.479,42.108,-52.085],[57.236,38.328,-51.378],
[54.399,38.852,-51.656],[55.891,38.297,-47.024],[56.774,40.191,-49.041],
[56.725,45.686,-46.131],[54.214,47.098,-55.551],[58.305,44.543,-51.970],
[58.523,43.114,-49.693],[56.220,43.007,-45.208],[57.602,49.287,-42.336],
[57.971,41.639,-54.819],[52.300,44.554,-45.119],[53.101,48.249,-47.816],
[57.583,51.380,-53.060],[56.575,45.107,-43.489],[58.306,34.977,-47.501],
[58.295,39.578,-46.157],[58.620,41.832,-52.340],[54.898,41.142,-46.540],
[-59.470,38.013,-58.978],[52.231,37.202,-51.954],[52.934,38.721,-54.326],
[55.944,41.054,-51.977],[58.688,39.098,-53.796],[56.499,38.199,-55.359],
[54.226,44.587,-41.168],[58.448,36.295,-53.675],[55.390,44.190,-50.345],
[52.662,41.735,-44.819],[57.047,47.958,-53.886],[54.099,44.403,58.709],
[59.570,49.356,-54.217],[54.123,39.133,-44.783],[55.578,48.536,-45.942],
[56.275,44.043,-58.066],[59.090,38.771,-56.302],[57.437,47.241,-57.705],
[55.166,40.106,-58.425],[56.216,33.689,-49.358],[54.642,47.947,-53.008],
[52.599,49.050,-41.894],[58.277,50.442,-39.417],[51.752,44.351,-40.414],
[58.839,44.033,59.352],[54.453,42.704,-42.848],[59.072,42.544,-57.129],
[58.530,41.571,-43.452],[56.921,37.396,-44.583],[-58.920,37.061,-46.749],
[56.613,39.850,-41.951],[57.639,37.592,-41.143],[53.713,47.917,-58.200],
[53.235,53.408,-52.868],[53.892,50.090,-51.403],[52.047,44.308,-49.542],
[52.463,39.638,-42.850],[54.587,35.829,-48.092],[52.662,40.255,-49.865],
[54.298,35.849,-53.484],[52.915,34.549,-45.645],[55.679,47.088,-41.232],
[58.384,50.156,-45.052],[55.931,52.536,-45.222],[56.873,52.630,-48.579],
[55.762,50.075,-48.156],[56.666,49.919,-50.663],[55.075,35.418,-50.971],
[53.535,35.400,-56.417],[55.201,50.924,-54.270],[55.963,35.081,-57.627],
[54.083,35.185,-42.742],[56.351,34.435,-41.371],[56.380,49.899,-58.586],
[58.556,56.487,-50.851],[53.884,59.481,-50.126],[56.425,54.815,-46.733],
[55.944,56.939,-50.811],[58.108,53.426,-50.823],[57.649,57.163,-48.393],
[59.273,-58.824,-53.312],[57.073,-59.095,-45.063],[58.431,56.142,58.203],
[52.522,56.589,-53.084],[54.934,57.139,-56.469],[55.057,55.638,-53.946],
[57.434,57.978,-55.341],[59.480,59.100,-57.163],[57.469,53.992,-54.036],
[54.471,55.857,-43.465],[58.402,56.555,-57.980],[55.420,57.247,-59.091],
[54.557,52.319,-50.025],[57.810,56.571,-44.971],[-58.633,56.256,-47.193],
[58.247,59.232,-42.390],[53.977,56.238,-40.955],[51.378,56.234,-41.901],
[54.270,54.839,-50.028],[55.608,58.775,-48.149],[52.910,55.351,-55.894],
[51.601,53.081,-45.609],[55.825,53.475,-51.970],[52.662,50.208,-57.249],
[56.023,54.411,-57.201],[53.676,52.566,-40.239],[56.393,51.646,-40.751],
[55.640,-51.342,-28.666],[56.305,-46.636,-31.300],[52.142,-54.040,-34.700],
[53.167,-51.592,-33.656],[57.129,-54.524,-32.194],[53.151,-53.785,-31.959],
[55.882,-55.876,-30.444],[55.264,-53.236,-30.438],[57.912,-47.117,-27.030],
[53.939,-47.254,-34.559],[55.649,-50.668,-32.763],[58.332,-51.161,-30.736],
[56.479,-51.253,-26.011],[58.932,-44.571,-23.808],[57.863,-50.307,-36.801],
[53.505,-47.538,-26.910],[54.703,-44.300,-30.206],[56.662,-42.999,-35.480],
[55.708,-47.041,-25.580],[58.447,-57.648,-30.633],[58.354,-52.574,-28.480],
[58.101,-51.978,-33.323],[53.023,-54.013,-26.449],[-58.982,-55.062,-39.939],
[52.432,-56.678,-33.230],[53.243,-55.340,-37.630],[55.399,-53.430,-34.064],
[58.987,-53.545,-37.457],[58.389,-52.448,-39.897],[57.331,-46.920,-23.626],
[57.485,-57.203,-33.526],[55.962,-49.173,-30.432],[53.831,-51.489,-25.458],
[-58.609,-45.412,-33.476],[52.979,-50.922,-39.481],[-59.627,-43.395,-34.997],
[54.917,-56.315,-27.342],[55.093,-44.679,-27.674],[54.289,-47.548,-37.420],
[57.800,-55.888,-38.201],[57.362,-44.321,-40.044],[55.259,-54.771,-39.135],
[55.682,-58.500,-31.737],[54.152,-44.697,-35.401],[54.851,-42.265,-23.176],
[58.326,-42.799,-21.020],[54.412,-49.348,-23.982],[55.170,-49.582,-39.669],
[56.771,-51.446,-23.330],[58.417,-49.701,-39.458],[59.199,-51.613,-25.837],
[57.922,-54.346,-25.710],[57.700,-56.320,-28.266],[57.028,-54.413,-22.849],
[55.002,-53.926,-24.401],[53.204,-45.257,-39.937],[53.377,-41.630,-35.954],
[52.629,-41.473,-33.092],[53.014,-48.664,-29.844],[52.163,-53.344,-24.008],
[52.884,-56.573,-30.486],[53.361,-51.172,-30.972],[54.196,-56.843,-35.684],
[51.963,-56.632,-27.636],[55.932,-44.627,-23.002],[56.999,-42.850,-26.622],
[54.698,-39.899,-27.469],[58.505,-40.267,-28.289],[53.912,-41.880,-29.015],
[55.093,-44.727,-32.834],[53.987,-58.997,-34.077],[52.849,-57.701,-38.381],
[54.532,-43.340,-38.099],[56.082,-56.959,-40.135],[54.803,-58.600,-24.661],
[57.185,-58.578,-22.835],[54.840,-41.471,-40.077],[56.940,-33.496,-29.016],
[57.135,-28.615,-32.335],[52.105,-33.956,-34.885],[54.490,-32.298,-33.218],
[57.449,-36.338,-33.112],[54.326,-35.060,-32.616],[55.539,-37.282,-28.606],
[55.520,-37.495,-31.821],[57.497,-30.774,-28.132],[55.926,-26.712,-37.410],
[58.013,-31.905,-32.606],[57.790,-34.234,-31.455],[56.326,-32.684,-26.254],
[58.309,-24.758,-23.100],[56.341,-31.390,-35.261],[54.611,-30.610,-28.426],
[53.227,-25.989,-29.364],[54.052,-23.768,-34.444],[55.862,-29.883,-26.184],
[57.787,-38.920,-30.608],[58.561,-36.177,-29.378],[-59.347,-34.585,-33.596],
[55.228,-35.118,-26.845],[58.835,-38.450,-40.253],[52.831,-37.412,-33.965],
[52.334,-35.985,-36.456],[55.636,-33.895,-35.126],[57.444,-35.900,-36.012],
[59.561,-34.975,-39.141],[54.768,-29.218,-23.613],[57.151,-38.789,-35.217],
[55.139,-30.463,-31.371],[53.729,-32.704,-26.467],[-58.873,-27.071,-32.448],
[54.501,-31.426,-38.217],[56.775,-23.758,-35.195],[55.938,-37.732,-25.878],
[55.694,-26.457,-27.932],[56.667,-29.939,-37.976],[58.799,-37.231,-38.000],
[56.614,-28.194,-40.023],[55.166,-35.039,-40.409],[54.602,-39.307,-30.110],
[53.769,-26.561,-35.421],[53.114,-23.466,-24.950],[-58.891,-25.467,-21.194],
[53.225,-31.327,-22.768],[58.903,-29.397,-39.286],[55.702,-32.388,-23.634],
[58.166,-32.181,-39.073],[59.030,-33.520,-27.025],[57.917,-35.612,-26.130],
[59.282,-37.894,-26.920],[56.187,-34.840,-22.508],[57.209,-37.687,-23.540],
[53.696,-27.667,-39.593],[51.891,-22.370,-35.550],[54.429,-25.188,-32.215],
[53.114,-31.771,-30.511],[52.517,-35.868,-24.970],[52.422,-37.424,-31.349],
[54.516,-34.634,-29.714],[53.983,-38.408,-36.593],[52.439,-38.522,-26.460],
[54.163,-25.736,-25.884],[57.389,-25.630,-25.795],[56.063,-22.559,-25.518],
[56.794,-23.709,-28.017],[54.065,-23.747,-28.470],[56.473,-24.889,-30.181],
[55.058,-40.028,-34.174],[55.380,-40.859,-37.428],[52.364,-25.302,-38.279],
[55.704,-38.732,-39.210],[54.906,-39.861,-24.719],[57.398,-41.269,-23.203],
[55.109,-24.338,-38.437],[55.577,-13.975,-28.070],[55.111,-9.409,-30.425],
[53.424,-15.142,-34.989],[55.300,-14.494,-31.378],[58.306,-17.794,-31.240],
[52.518,-16.664,-32.752],[56.641,-19.127,-29.156],[55.746,-17.434,-31.120],
[57.998,-10.643,-25.100],[55.263,-8.760,-36.187],[58.123,-11.820,-32.686],
[58.599,-13.196,-28.981],[56.530,-12.892,-25.333],[57.650,-6.052,-24.958],
[57.692,-14.217,-34.855],[53.067,-10.051,-27.489],[53.064,-7.581,-29.280],
[56.541,-6.104,-34.831],[56.271,-8.435,-24.478],[57.803,-21.330,-30.701],
[57.624,-16.281,-28.638],[58.209,-15.114,-31.741],[54.862,-16.695,-28.519],
[-59.467,-17.929,-39.690],[52.467,-19.654,-33.584],[54.619,-17.005,-37.530],
[55.785,-16.018,-33.932],[57.111,-17.113,-36.168],[57.876,-15.758,-38.684],
[56.166,-10.336,-22.789],[57.464,-18.959,-34.023],[56.611,-11.542,-30.235],
[53.947,-13.932,-25.928],[58.365,-9.208,-34.475],[54.259,-12.250,-39.451],
[58.595,-6.939,-36.399],[55.156,-18.753,-26.909],[55.395,-6.961,-27.781],
[55.721,-12.250,-36.806],[56.324,-20.044,-36.913],[54.973,-8.838,-39.085],
[53.934,-16.294,-40.355],[55.459,-21.548,-29.325],[54.759,-8.375,-33.522],
[53.587,-4.753,-23.211],[59.582,-6.721,-21.887],[53.533,-12.352,-23.641],
[56.512,-10.833,-40.123],[55.557,-13.610,-21.947],[-59.007,-14.210,-36.807],
[57.626,-15.373,-26.091],[56.813,-17.571,-25.096],[59.015,-19.955,-27.021],
[57.490,-15.885,-23.058],[55.370,-17.831,-22.930],[52.927,-11.209,-37.322],
[51.522,-6.196,-32.326],[55.046,-4.662,-32.995],[53.993,-11.920,-29.697],
[52.980,-16.989,-21.691],[52.521,-18.683,-30.656],[52.835,-15.780,-30.302],
[53.009,-19.207,-36.311],[52.884,-21.031,-28.851],[56.081,-6.408,-22.227],
[59.327,-6.392,-27.487],[54.862,-4.241,-27.396],[57.484,-3.067,-28.883],
[55.633,-5.038,-30.037],[56.818,-7.391,-31.757],[55.397,-21.024,-34.353],
[53.379,-21.011,-38.537],[54.328,-6.107,-39.003],[56.018,-21.795,-39.001],
[56.298,-20.020,-24.330],[56.310,-22.875,-22.792],[56.444,-5.864,-40.766],
[56.345,5.224,-29.977],[57.434,8.895,-31.267],[52.236,3.093,-34.419],
[53.272,5.515,-33.267],[58.878,1.398,-31.371],[53.877,2.088,-32.157],
[56.685,-0.645,-29.840],[55.658,2.118,-29.930],[58.762,7.489,-25.950],
[54.622,9.933,-36.700],[57.806,6.481,-34.369],[58.717,4.271,-30.555],
[56.316,4.089,-25.534],[57.862,12.086,-24.591],[58.361,5.554,-36.971],
[52.613,7.429,-26.792],[51.732,10.521,-28.196],[55.861,13.918,-32.423],
[55.389,7.341,-27.245],[57.940,-2.998,-31.770],[57.369,2.951,-27.749],
[58.906,1.825,-34.147],[54.699,3.899,-27.687],[58.564,-0.761,-40.582],
[53.390,0.676,-35.017],[53.132,0.830,-38.979],[55.575,2.035,-34.398],
[59.634,0.407,-36.400],[57.472,3.185,-39.003],[56.329,7.769,-24.566],
[57.095,-2.763,-34.148],[55.840,6.612,-32.317],[54.046,5.591,-25.307],
[57.648,9.300,-34.784],[54.509,4.640,-39.462],[-59.396,13.003,-35.210],
[54.487,0.268,-25.353],[53.910,12.023,-27.704],[55.764,7.648,-36.903],
[-59.154,-1.528,-38.326],[53.281,11.558,-39.323],[55.550,1.488,-39.919],
[54.903,-1.427,-27.724],[52.741,10.691,-34.459],[54.786,14.670,-23.409],
[57.888,13.453,-21.343],[53.450,7.478,-21.850],[57.845,7.442,-39.222],
[55.217,4.579,-23.121],[-57.920,5.798,-37.054],[-59.343,3.401,-25.949],
[59.254,0.771,-26.602],[57.688,-1.912,-26.431],[57.292,2.411,-23.597],
[54.271,0.857,-22.660],[52.179,8.719,-38.847],[52.052,13.709,-33.754],
[53.862,12.120,-32.055],[53.524,5.667,-30.423],[53.505,2.600,-24.807],
[52.905,-0.657,-32.065],[52.759,2.613,-29.926],[52.575,-1.525,-36.558],
[52.639,-1.845,-26.331],[55.577,12.125,-23.046],[58.776,11.127,-28.160],
[54.095,16.033,-28.860],[58.124,13.822,-28.756],[55.552,11.458,-29.896],
[56.811,11.844,-33.946],[54.588,-2.788,-35.114],[54.585,-2.371,-38.370],
[54.408,12.476,-36.862],[57.783,-3.328,-39.236],[55.674,-3.092,-25.116],
[56.210,-1.265,-22.585],[54.618,13.854,-39.670],[55.795,24.642,-28.016],
[55.329,27.794,-30.480],[53.253,22.459,-34.819],[52.243,23.430,-30.981],
[57.661,18.318,-32.261],[53.238,20.087,-33.132],[57.176,18.547,-26.762],
[57.799,20.692,-30.659],[57.142,27.736,-28.130],[53.253,28.173,-36.495],
[58.746,25.090,-33.471],[-58.800,24.551,-29.886],[54.513,24.423,-25.522],
[58.660,29.470,-24.811],[57.665,22.374,-35.278],[51.395,26.193,-27.152],
[53.205,29.443,-30.356],[55.554,31.882,-34.044],[55.068,27.856,-25.765],
[57.139,16.584,-28.569],[58.603,20.476,-28.232],[58.510,22.516,-32.796],
[55.842,20.954,-27.580],[-57.784,16.932,-41.173],[51.482,18.781,-34.542],
[54.516,19.250,-40.069],[56.071,20.831,-32.729],[57.345,19.298,-34.679],
[58.171,21.536,-38.912],[54.996,25.583,-23.151],[57.832,16.770,-35.162],
[57.185,25.371,-30.189],[54.429,21.870,-25.494],[58.334,27.700,-34.109],
[55.834,23.810,-36.938],[57.662,30.161,-34.903],[54.284,19.115,-25.668],
[55.223,29.795,-27.955],[57.069,25.772,-38.463],[58.566,18.639,-37.434],
[56.940,28.241,-39.659],[55.597,22.032,-39.841],[54.587,16.270,-31.514],
[52.505,29.554,-33.774],[52.865,32.140,-23.846],[59.127,31.844,-20.416],
[53.004,25.334,-21.499],[58.157,25.623,-41.118],[55.406,22.882,-22.334],
[-57.814,24.993,-38.391],[-59.105,22.579,-25.356],[-59.030,19.482,-25.708],
[-59.539,17.968,-28.515],[57.170,21.421,-24.409],[54.958,18.451,-23.200],
[54.907,26.760,-40.109],[52.282,32.173,-35.322],[52.325,31.822,-32.521],
[53.553,25.838,-29.056],[52.815,20.544,-23.611],[55.076,18.942,-31.084],
[54.312,21.615,-29.860],[53.247,19.736,-36.877],[52.233,17.280,-27.375],
[55.960,29.501,-23.728],[58.710,30.022,-27.470],[54.868,35.461,-26.938],
[56.933,32.389,-27.917],[54.922,31.474,-31.354],[57.016,30.292,-29.905],
[54.870,17.659,-34.347],[53.586,15.170,-35.352],[54.031,30.672,-37.106],
[56.942,18.507,-39.857],[55.419,16.405,-25.544],[57.118,16.320,-22.209],
[55.637,32.396,-39.048],[54.402,42.227,-27.892],[55.945,47.095,-28.992],
[52.413,40.753,-36.446],[53.961,41.462,-32.429],[56.591,35.884,-32.027],
[52.229,39.819,-33.880],[55.715,37.304,-28.733],[55.230,39.006,-31.905],
[57.053,44.446,-27.979],[55.137,46.494,-36.467],[56.975,43.795,-32.150],
[58.447,41.406,-31.383],[55.591,43.328,-25.056],[57.630,50.580,-24.883],
[57.222,41.839,-34.382],[51.406,43.533,-27.222],[52.037,48.049,-29.763],
[56.034,50.716,-32.006],[56.909,45.873,-25.682],[58.906,33.518,-31.584],
[57.315,39.699,-29.686],[58.824,38.737,-34.998],[58.256,39.748,-27.168],
[-58.423,36.644,-41.083],[52.317,37.237,-33.260],[54.256,37.443,-38.442],
[54.377,39.304,-35.327],[-58.236,36.530,-35.965],[59.163,38.900,-39.326],
[56.302,44.362,-22.325],[56.358,37.225,-35.405],[55.639,42.447,-30.195],
[54.080,40.970,-25.688],[59.085,45.464,-32.517],[56.013,43.349,-37.793],
[-59.126,50.003,-35.393],[54.089,37.909,-25.659],[54.668,48.661,-27.204],
[57.501,44.709,-36.047],[57.680,37.002,-38.041],[56.785,46.795,-38.794],
[54.374,39.576,-40.248],[55.472,34.629,-29.559],[54.940,48.674,-34.783],
[52.360,50.002,-22.889],[-59.183,51.201,-22.227],[53.469,44.786,-23.090],
[56.830,44.153,-40.368],[55.198,41.821,-22.771],[59.116,43.853,-38.309],
[59.163,40.659,-24.282],[56.719,37.117,-25.598],[59.050,37.414,-28.296],
[57.416,40.218,-21.920],[55.140,36.657,-22.482],[53.043,45.899,-38.714],
[52.770,51.208,-33.817],[53.989,48.415,-32.152],[53.134,44.466,-29.581],
[53.237,39.195,-23.541],[53.604,35.843,-31.268],[51.808,40.612,-30.520],
[52.977,36.530,-35.961],[52.766,33.682,-27.145],[55.351,48.530,-22.281],
[57.319,49.455,-28.248],[53.794,51.613,-28.215],[57.580,52.197,-28.486],
[54.397,49.381,-29.798],[57.215,48.102,-31.771],[53.727,34.657,-34.141],
[54.495,33.568,-36.656],[52.610,50.689,-36.565],[55.986,34.979,-38.299],
[55.054,34.200,-24.498],[56.868,34.699,-22.167],[55.687,50.747,-38.105],
[56.491,59.102,-34.365],[53.507,58.749,-31.573],[56.045,53.665,-26.822],
[55.850,57.634,-31.966],[58.988,53.392,-31.432],[56.276,-59.958,-29.192],
[58.994,-59.656,-35.314],[54.986,58.089,-26.663],[58.578,54.052,-41.143],
[53.407,56.455,-33.536],[54.513,56.845,-38.465],[55.552,56.179,-35.803],
[59.226,57.497,-37.267],[57.489,58.314,-40.046],[57.832,54.544,-34.058],
[53.834,55.324,-24.865],[59.381,55.230,-38.413],[55.149,59.346,-39.067],
[54.875,53.520,-29.676],[56.672,55.481,-25.167],[58.247,55.781,-27.469],
[56.769,58.160,-24.551],[54.890,55.953,-22.410],[52.625,58.673,-24.907],
[53.770,55.799,-30.590],[51.529,-59.955,-29.660],[53.035,55.241,-36.062],
[51.483,53.311,-27.038],[54.781,53.042,-33.595],[54.554,52.916,-36.760],
[56.032,54.658,-39.171],[54.622,50.745,-25.463],[56.895,52.552,-22.720],
[54.803,-51.023,-10.837],[58.152,-46.502,-10.733],[52.546,-52.297,-17.079],
[54.874,-49.080,-14.223],[57.066,-55.793,-13.321],[52.990,-51.736,-14.361],
[56.857,-56.224,-10.061],[55.291,-54.116,-11.723],[58.990,-46.024,-7.202],
[55.704,-47.024,-18.584],[58.331,-49.222,-14.702],[58.363,-50.843,-11.090],
[56.753,-50.317,-8.886],[56.907,-41.922,-4.708],[55.270,-51.980,-16.923],
[53.921,-47.164,-9.098],[53.178,-44.484,-11.178],[55.265,-42.158,-15.449],
[55.445,-45.629,-7.204],[58.484,-57.756,-11.756],[57.199,-53.288,-9.653],
[57.770,-53.068,-16.119],[54.266,-53.087,-9.362],[58.799,-56.035,-21.336],
[51.493,-57.178,-13.827],[54.332,-56.410,-19.101],[54.947,-53.742,-14.811],
[58.586,-55.760,-17.534],[57.524,-54.373,-19.595],[57.036,-47.433,-5.368],
[55.934,-55.859,-16.915],[56.129,-48.662,-11.159],[53.810,-50.714,-7.697],
[58.116,-45.376,-14.267],[53.696,-50.493,-21.052],[58.132,-43.307,-16.259],
[56.119,-55.449,-7.381],[53.405,-43.573,-8.622],[56.265,-49.849,-18.362],
[-59.361,-57.524,-19.415],[56.920,-46.389,-21.088],[53.494,-53.183,-21.359],
[55.922,-59.294,-11.394],[52.810,-44.535,-16.423],[52.248,-42.611,-6.114],
[59.317,-41.950,-2.253],[53.769,-48.210,-2.697],[57.493,-48.984,-21.932],
[55.394,-51.322,-5.390],[-58.999,-49.718,-19.954],[58.466,-50.734,-4.899],
[59.081,-54.653,-7.011],[58.908,-57.361,-7.829],[56.952,-53.897,-4.179],
[54.935,-55.591,-5.133],[53.842,-45.902,-20.498],[52.815,-40.428,-15.145],
[51.625,-42.970,-14.309],[52.000,-51.237,-11.592],[52.972,-54.273,-3.621],
[53.642,-55.759,-13.029],[51.747,-53.682,-12.843],[52.814,-56.891,-17.131],
[53.044,-57.983,-9.551],[53.877,-44.570,-4.868],[56.990,-44.134,-9.849],
[53.874,-40.834,-7.811],[56.633,-41.619,-8.605],[54.412,-42.256,-11.946],
[55.896,-44.646,-12.237],[55.229,-57.887,-14.461],[53.635,58.851,-19.690],
[55.176,-44.392,-17.943],[56.070,-56.059,-21.059],[53.846,-59.398,-6.550],
[57.026,-58.117,-4.077],[55.766,-42.861,-20.419],[57.716,-30.965,-9.945],
[55.626,-27.682,-12.447],[52.340,-32.591,-16.613],[53.810,-32.636,-13.309],
[-59.379,-36.750,-11.376],[54.321,-35.496,-14.870],[56.475,-37.434,-11.002],
[55.321,-34.851,-12.426],[57.233,-27.817,-9.583],[53.993,-27.230,-17.315],
[57.752,-30.300,-15.204],[58.233,-31.101,-12.411],[56.942,-31.906,-6.589],
[58.266,-25.766,-6.519],[57.119,-33.763,-17.190],[53.748,-30.065,-8.644],
[53.235,-26.757,-11.610],[54.892,-24.571,-16.100],[55.157,-27.447,-7.104],
[58.854,-40.602,-12.015],[57.382,-34.065,-10.561],[58.088,-35.597,-15.063],
[54.735,-34.530,-9.586],[-58.756,-36.395,-20.314],[51.347,-36.506,-16.104],
[51.960,-36.385,-19.644],[54.036,-34.735,-17.688],[59.209,-36.101,-17.898],
[57.423,-34.919,-19.915],[56.272,-29.459,-5.297],[56.128,-37.313,-13.799],
[55.597,-30.539,-13.807],[54.371,-32.468,-7.112],[59.125,-26.915,-16.175],
[52.771,-33.763,-20.188],[57.801,-24.236,-16.221],[56.721,-37.133,-8.334],
[52.858,-26.432,-8.470],[55.890,-31.294,-18.813],[58.064,-37.891,-20.895],
[57.416,-28.122,-20.359],[53.534,-34.656,-22.767],[55.607,-40.094,-11.998],
[53.709,-26.287,-14.311],[53.401,-24.470,-4.041],[59.012,-24.813,-3.956],
[53.694,-31.246,-3.624],[57.024,-30.661,-21.479],[56.046,-32.101,-4.187],
[58.986,-31.479,-19.989],[59.030,-32.624,-8.079],[58.872,-35.397,-6.913],
[-59.323,-37.250,-8.718],[-59.357,-32.672,-5.225],[56.490,-36.764,-5.601],
[54.328,-29.097,-20.992],[53.818,-22.560,-17.387],[53.941,-23.693,-13.362],
[55.339,-30.244,-11.149],[53.241,-36.136,-4.026],[53.793,-37.290,-11.135],
[52.420,-34.134,-10.829],[53.337,-37.791,-17.859],[52.871,-38.434,-8.489],
[55.605,-26.532,-4.542],[57.873,-24.832,-9.701],[55.354,-22.779,-8.453],
[58.189,-21.474,-10.594],[55.431,-24.709,-10.936],[56.735,-26.457,-14.943],
[55.168,-40.123,-17.486],[53.892,-40.638,-19.581],[52.845,-24.940,-18.139],
[55.101,-38.136,-20.663],[55.579,-39.189,-6.111],[57.987,-39.250,-4.648],
[53.685,-23.109,-20.616],[54.456,-12.682,-9.409],[56.865,-8.637,-9.620],
[52.735,-13.876,-16.799],[54.153,-13.400,-14.562],[57.586,-17.423,-13.510],
[53.130,-16.460,-13.706],[56.648,-18.847,-11.117],[54.515,-14.624,-12.073],
[56.657,-10.072,-6.899],[55.824,-7.723,-15.948],[56.345,-12.107,-13.913],
[58.207,-12.488,-10.648],[56.984,-12.942,-7.922],[58.890,-6.266,-6.693],
[57.011,-14.034,-16.473],[52.897,-9.692,-7.464],[52.330,-7.046,-11.683],
[55.900,-5.093,-15.543],[53.931,-9.242,-5.051],[58.223,-24.019,-12.348],
[55.602,-15.125,-9.482],[59.163,-16.098,-16.120],[55.031,-15.376,-6.828],
[58.683,-18.961,-22.955],[52.706,-18.915,-14.318],[53.204,-17.678,-17.780],
[55.833,-15.618,-14.809],[56.825,-15.950,-18.570],[58.783,-17.036,-20.628],
[57.304,-9.592,-4.054],[56.078,-18.037,-17.111],[56.048,-10.761,-11.123],
[52.048,-12.054,-7.309],[58.469,-8.536,-15.563],[54.820,-13.463,-18.903],
[58.740,-6.109,-18.486],[55.745,-18.322,-7.449],[52.680,-7.673,-9.167],
[56.236,-11.200,-19.559],[59.140,-19.495,-19.555],[56.094,-8.705,-20.539],
[55.084,-15.918,-20.592],[55.911,-21.818,-12.125],[53.219,-9.061,-13.775],
[52.908,-4.143,-5.086],[58.132,-5.020,-4.155],[53.667,-12.292,-4.820],
[58.572,-12.037,-20.783],[55.425,-14.255,-4.340],[-58.876,-15.069,-19.125],
[58.203,-14.523,-5.089],[57.706,-15.895,-7.613],[57.757,-19.915,-8.420],
[57.134,-17.146,-3.518],[55.942,-19.657,-4.796],[53.590,-9.622,-20.068],
[53.454,-4.692,-14.623],[50.809,-5.104,-14.745],[53.693,-11.598,-12.531],
[51.927,-17.400,-4.438],[53.489,-18.830,-11.646],[51.953,-13.687,-11.370],
[52.511,-20.513,-16.431],[53.038,-19.387,-8.097],[56.013,-6.518,-5.301],
[57.121,-5.985,-9.086],[54.895,-3.554,-9.113],[57.615,-3.378,-9.901],
[53.432,-4.764,-11.330],[58.336,-6.112,-14.657],[56.075,-20.740,-14.755],
[53.634,-20.574,-19.307],[54.566,-6.492,-18.448],[55.818,-18.805,-19.970],
[54.255,-21.850,-6.064],[56.414,-22.870,-4.174],[54.154,-3.997,-20.555],
[55.589,5.668,-10.734],[57.447,9.372,-12.853],[52.374,4.220,-15.481],
[54.007,6.252,-14.594],[58.018,2.031,-13.760],[54.252,2.608,-13.859],
[56.014,-0.987,-10.146],[55.500,2.818,-11.603],[57.290,9.095,-8.023],
[53.342,8.858,-18.157],[58.398,6.744,-15.146],[57.904,6.425,-11.888],
[56.843,5.390,-7.070],[57.808,11.653,-3.756],[56.643,4.555,-18.258],
[52.026,7.354,-8.471],[54.802,10.534,-12.554],[56.784,11.207,-16.984],
[54.646,9.172,-7.572],[58.121,-1.548,-11.952],[56.910,2.990,-9.229],
[57.950,3.439,-16.251],[54.446,2.084,-8.726],[59.149,-1.863,-22.004],
[51.880,0.378,-15.539],[53.806,0.900,-18.942],[55.054,3.030,-16.546],
[-58.219,2.407,-18.453],[58.124,2.329,-19.173],[56.063,7.693,-5.717],
[57.566,-0.738,-14.992],[56.349,4.381,-14.112],[53.984,5.311,-7.823],
[-58.767,10.556,-14.452],[55.150,5.532,-20.604],[59.336,13.197,-17.157],
[53.873,0.624,-5.791],[53.825,11.763,-8.595],[55.419,8.123,-19.820],
[-59.335,-0.753,-16.979],[57.188,9.495,-22.557],[55.115,2.803,-20.914],
[55.164,-3.179,-12.295],[54.999,9.790,-15.135],[52.638,14.108,-5.472],
[-58.702,12.251,-4.591],[53.646,6.783,-4.024],[57.816,7.161,-21.311],
[54.961,4.685,-4.914],[59.192,4.617,-20.742],[58.110,3.106,-5.786],
[57.688,0.762,-7.806],[59.318,-2.116,-8.043],[55.890,2.312,-4.434],
[56.511,-0.312,-5.530],[53.627,10.090,-21.757],[53.839,13.584,-15.632],
[52.919,11.535,-14.346],[52.876,7.444,-12.444],[52.025,2.385,-4.900],
[53.045,-1.734,-11.981],[52.851,4.277,-11.365],[52.965,-1.950,-16.111],
[52.276,-2.302,-7.856],[54.726,11.100,-3.546],[56.955,12.666,-7.951],
[54.604,15.034,-8.504],[57.524,15.009,-9.558],[56.046,12.766,-11.118],
[56.791,12.344,-13.654],[55.473,-2.785,-16.429],[53.481,-4.124,-17.646],
[54.040,12.549,-18.528],[57.269,-1.395,-19.853],[54.552,-2.165,-6.588],
[57.259,-2.302,-2.914],[53.544,13.476,-21.310],[56.458,24.902,-8.966],
[58.027,27.694,-11.567],[55.464,22.822,-15.115],[53.579,24.699,-16.119],
[57.357,18.980,-13.892],[53.789,20.756,-14.115],[56.183,18.245,-10.274],
[56.040,20.953,-11.842],[56.257,27.536,-9.500],[55.391,28.493,-18.128],
[58.614,26.607,-14.121],[59.015,24.453,-11.729],[55.023,22.588,-8.982],
[57.792,29.558,-6.364],[58.182,24.868,-16.812],[53.349,26.367,-9.234],
[53.300,27.742,-11.620],[56.076,32.290,-15.761],[55.169,28.017,-7.176],
[58.724,17.208,-11.457],[57.536,21.315,-9.598],[58.677,21.417,-14.805],
[55.524,20.057,-8.113],[-59.264,16.649,-23.172],[52.279,18.050,-14.378],
[54.644,18.230,-17.984],[56.022,20.671,-17.239],[-58.783,20.709,-18.223],
[58.994,20.315,-21.247],[56.692,27.539,-4.874],[58.205,19.413,-16.782],
[55.276,25.824,-11.530],[53.959,23.612,-6.516],[57.192,29.029,-14.449],
[55.387,23.145,-18.325],[58.756,30.774,-15.891],[54.070,18.119,-7.014],
[55.315,30.241,-9.375],[54.654,26.025,-19.393],[58.843,18.006,-19.086],
[56.912,27.565,-20.476],[54.500,20.895,-20.185],[55.372,15.578,-12.096],
[53.630,29.822,-15.485],[52.221,31.664,-3.954],[58.425,29.649,-2.788],
[54.694,25.813,-4.227],[57.144,27.211,-23.435],[57.123,23.444,-5.433],
[-58.698,23.147,-17.447],[59.043,24.066,-7.705],[59.074,21.272,-7.217],
[58.642,18.389,-8.748],[55.973,20.871,-5.394],[56.742,18.316,-5.718],
[53.453,28.162,-22.441],[52.486,32.167,-16.484],[53.788,32.332,-13.882],
[52.668,25.016,-11.615],[52.377,20.999,-5.252],[54.914,18.349,-12.547],
[53.155,20.643,-11.489],[52.165,20.293,-18.519],[52.124,16.011,-8.603],
[55.210,31.017,-5.098],[58.220,29.443,-8.934],[53.902,36.712,-10.539],
[55.909,33.770,-8.979],[53.489,31.483,-11.147],[57.492,33.242,-12.077],
[56.334,15.484,-14.621],[54.791,15.364,-17.378],[54.614,31.078,-18.553],
[55.436,16.387,-20.080],[54.240,16.397,-4.883],[56.595,14.927,-4.516],
[55.890,32.534,-20.463],[56.547,42.798,-9.427],[53.756,46.412,-9.358],
[52.259,40.065,-16.095],[52.672,41.811,-13.488],[57.491,36.087,-13.541],
[53.654,39.292,-13.716],[56.815,36.675,-8.719],[55.448,41.066,-11.683],
[56.869,46.757,-9.561],[52.651,46.473,-17.646],[58.894,43.428,-15.111],
[57.435,42.992,-12.258],[56.705,41.754,-6.854],[58.728,47.878,-3.998],
[57.124,42.821,-17.810],[52.303,43.974,-8.435],[55.237,48.584,-11.053],
[57.751,51.371,-16.420],[53.226,45.005,-6.081],[59.061,35.966,-11.271],
[56.619,40.015,-9.542],[58.205,40.890,-15.915],[53.918,41.299,-9.342],
[57.971,37.333,-22.728],[52.167,36.784,-14.900],[53.897,39.009,-18.438],
[56.784,39.676,-13.966],[57.371,38.477,-16.982],[58.881,39.111,-19.949],
[55.036,44.807,-3.916],[56.389,35.945,-16.340],[55.608,44.757,-13.632],
[53.408,42.189,-6.795],[57.408,47.568,-15.475],[53.911,43.428,-20.615],
[58.940,48.729,-17.108],[54.728,37.328,-6.854],[54.441,48.244,-7.873],
[55.496,45.468,-17.865],[57.325,35.279,-18.932],[57.088,46.903,-20.374],
[54.109,40.263,-20.957],[56.400,35.316,-11.230],[54.374,46.436,-15.543],
[52.896,49.263,-4.058],[56.586,49.313,-2.231],[52.699,43.554,-3.694],
[58.600,45.869,-22.464],[55.401,42.194,-4.679],[56.635,42.494,-20.401],
[58.791,42.102,-5.073],[57.523,38.925,-6.960],[-59.291,36.412,-7.986],
[56.353,39.047,-4.389],[54.850,36.717,-3.263],[53.626,46.250,-20.297],
[52.202,49.966,-15.685],[52.347,50.441,-13.068],[54.246,43.519,-11.455],
[53.018,38.062,-4.602],[54.856,36.437,-13.609],[52.670,39.284,-10.776],
[53.165,36.387,-17.881],[53.080,34.505,-7.762],[54.984,47.596,-4.211],
[58.634,49.384,-9.303],[54.863,51.100,-8.883],[57.590,52.150,-9.200],
[56.219,51.876,-11.720],[57.181,49.766,-13.600],[53.953,34.505,-15.642],
[54.309,33.903,-18.188],[53.560,49.648,-18.030],[54.689,35.335,-20.205],
[55.742,34.228,-5.918],[58.012,34.808,-4.196],[53.546,51.089,-20.413],
[58.386,56.877,-13.713],[52.775,58.154,-13.535],[55.300,56.391,-9.740],
[55.359,57.525,-12.973],[58.457,53.779,-13.345],[57.191,58.582,-8.923],
[-59.060,58.218,-15.648],[55.879,-57.979,-7.879],[57.481,55.745,-22.565],
[52.700,55.815,-16.423],[54.010,55.600,-20.003],[56.744,58.465,-15.770],
[58.889,58.473,-18.118],[57.448,57.995,-20.697],[57.613,54.663,-15.785],
[54.428,54.301,-7.655],[56.731,55.506,-19.236],[54.969,58.553,-22.123],
[54.152,53.401,-11.051],[58.233,56.631,-6.487],[58.975,54.714,-8.412],
[56.339,57.431,-4.511],[55.128,54.200,-4.604],[52.519,57.379,-5.706],
[53.873,55.489,-12.950],[52.333,58.064,-11.021],[52.191,52.993,-17.475],
[52.659,52.384,-8.271],[55.459,53.593,-14.320],[54.664,51.864,-17.077],
[53.213,53.515,-21.823],[55.543,51.916,-6.392],[57.827,53.174,-3.282],
[55.034,-49.999,8.793],[55.999,-46.711,6.305],[53.546,-52.965,2.220],
[54.092,-50.432,3.785],[59.514,-52.274,5.332],[54.558,-54.010,4.578],
[55.072,-56.181,8.886],[55.705,-54.303,6.999],[58.846,-46.645,10.515],
[55.143,-46.284,-0.169],[56.617,-48.678,3.707],[58.357,-50.346,7.252],
[56.387,-50.418,11.256],[57.589,-44.753,12.198],[58.193,-51.799,2.387],
[52.871,-46.903,10.356],[52.065,-45.442,8.234],[56.456,-42.598,2.107],
[54.743,-45.177,11.391],[58.013,-56.036,7.304],[56.591,-53.801,9.373],
[-59.063,-55.119,3.275],[54.228,-54.449,10.935],[58.724,-56.058,-3.018],
[52.392,-55.660,3.963],[53.274,-54.926,-0.811],[57.199,-54.492,3.672],
[-59.301,-53.435,0.738],[58.398,-51.789,-1.505],[56.493,-47.399,13.004],
[57.624,-57.357,4.551],[56.161,-49.270,6.353],[53.045,-49.778,11.181],
[56.927,-45.236,3.324],[54.284,-51.303,-2.249],[-58.425,-42.521,1.943],
[55.836,-56.781,11.262],[55.237,-46.568,9.197],[56.254,-49.752,0.334],
[57.363,-55.723,-0.731],[56.503,-46.283,-2.857],[55.419,-53.687,-1.790],
[56.274,59.339,5.668],[54.074,-45.369,2.420],[52.498,-43.056,14.681],
[56.688,-43.184,16.085],[54.404,-49.708,13.422],[58.765,-46.617,-1.464],
[54.676,-52.401,12.655],[58.281,-49.158,-1.850],[58.489,-52.047,9.801],
[58.921,-54.570,10.865],[58.303,-57.419,9.794],[57.523,-52.616,14.180],
[56.522,-55.112,13.979],[51.729,-46.589,-1.294],[50.798,-41.231,2.893],
[52.983,-42.655,4.070],[53.439,-48.531,6.127],[52.542,-54.117,13.825],
[54.075,-56.405,6.399],[52.255,-53.645,7.066],[53.664,-57.213,1.669],
[53.138,-57.090,10.866],[55.114,-43.388,13.156],[57.414,-43.632,9.503],
[55.645,-40.499,10.815],[58.037,-40.491,9.395],[53.634,-40.158,8.074],
[56.457,-43.745,6.358],[55.164,-58.953,3.429],[54.203,-59.652,0.487],
[54.174,-43.426,0.428],[55.607,-57.870,-1.509],[54.367,-58.901,12.852],
[56.344,-59.411,15.209],[56.580,-42.557,-0.814],[54.877,-31.446,8.935],
[57.211,-28.034,6.975],[52.730,-33.744,2.677],[53.685,-31.435,4.265],
[57.242,-36.088,4.723],[53.568,-34.435,5.281],[55.766,-37.641,8.043],
[55.306,-34.674,7.750],[57.894,-28.536,10.423],[54.751,-28.203,1.111],
[57.118,-31.642,4.430],[58.089,-33.045,7.185],[56.359,-31.250,11.353],
[57.299,-25.547,12.500],[57.899,-33.531,2.104],[52.730,-29.151,10.019],
[52.716,-27.840,6.990],[56.499,-24.329,2.901],[54.700,-28.456,11.732],
[58.191,-39.336,5.142],[57.147,-33.801,9.931],[59.507,-33.826,5.063],
[54.074,-34.177,10.889],[58.578,-35.826,-3.716],[54.113,-37.134,3.618],
[53.312,-36.644,-1.286],[55.842,-34.156,3.493],[59.440,-35.745,1.206],
[56.648,-35.306,-1.235],[55.538,-30.171,13.792],[56.890,-38.523,2.959],
[55.412,-30.341,6.278],[53.766,-31.291,11.814],[-59.196,-26.963,3.225],
[53.305,-31.633,-1.040],[58.805,-24.857,1.573],[55.778,-37.796,12.115],
[55.725,-26.627,8.671],[55.053,-30.982,0.780],[58.646,-37.201,-0.909],
[55.003,-26.006,-1.494],[54.614,-34.095,-2.457],[55.763,-39.696,6.271],
[55.431,-27.115,4.245],[52.033,-23.909,12.849],[57.305,-24.378,14.864],
[52.886,-30.086,13.851],[56.719,-29.902,-2.450],[55.101,-33.327,13.593],
[57.610,-31.685,-0.744],[-58.662,-32.953,14.809],[58.320,-36.241,12.302],
[58.581,-36.579,9.165],[57.512,-34.028,14.081],[55.131,-36.338,15.428],
[53.666,-28.794,-2.678],[52.875,-23.881,3.181],[54.756,-24.670,4.937],
[52.348,-30.500,6.493],[52.932,-34.576,15.268],[53.683,-37.714,6.328],
[52.178,-35.622,7.299],[54.641,-38.259,0.860],[53.617,-38.703,10.644],
[54.014,-27.385,14.256],[58.959,-25.413,10.296],[54.705,-21.572,9.918],
[57.226,-22.050,8.132],[55.208,-23.691,7.535],[57.412,-25.439,5.454],
[54.338,-40.312,3.016],[55.136,-39.764,-1.311],[54.316,-25.135,1.253],
[55.710,-37.584,-3.022],[54.255,-40.125,13.299],[56.133,-38.810,14.987],
[56.348,-23.609,-1.705],[55.664,-12.890,7.647],[57.880,-10.151,6.106],
[52.717,-15.540,2.335],[53.021,-14.242,4.827],[56.464,-18.240,4.381],
[53.895,-16.685,4.715],[56.973,-18.391,8.099],[56.805,-16.043,6.460],
[57.337,-9.694,10.543],[53.280,-10.668,2.255],[56.900,-12.987,4.878],
[58.914,-14.000,6.394],[56.408,-14.138,10.868],[58.309,-6.668,12.476],
[56.373,-13.516,2.074],[52.515,-12.578,10.754],[53.345,-9.635,7.420],
[56.018,-3.775,2.852],[54.793,-9.223,10.460],[58.175,-20.026,6.021],
[58.470,-14.883,9.069],[58.410,-15.517,4.262],[55.958,-16.403,9.552],
[-58.675,-18.127,-2.469],[51.410,-19.382,4.017],[53.240,-17.777,-0.289],
[55.999,-15.774,3.220],[58.056,-17.313,2.106],[58.327,-16.588,-1.099],
[55.697,-10.846,13.077],[57.710,-20.190,3.022],[54.793,-11.216,5.627],
[53.986,-15.109,11.545],[57.418,-8.019,3.247],[54.057,-13.153,-1.835],
[58.541,-5.913,1.906],[54.641,-18.850,9.456],[54.600,-6.220,9.800],
[55.936,-10.764,1.482],[57.947,-19.116,0.289],[55.932,-7.097,-0.145],
[54.399,-16.111,-2.262],[55.065,-20.794,7.110],[54.158,-8.278,1.447],
[52.419,-5.812,13.651],[57.693,-6.611,16.465],[53.117,-11.710,13.771],
[56.227,-9.339,-1.597],[57.346,-13.046,13.160],[58.426,-13.975,-0.518],
[59.198,-15.647,11.590],[57.106,-17.868,11.807],[-59.417,-17.526,9.280],
[57.547,-15.810,13.799],[56.105,-18.300,14.323],[54.165,-8.217,-2.468],
[53.234,-3.796,2.961],[52.008,-6.007,3.647],[52.545,-12.412,6.972],
[52.312,-16.920,13.111],[53.664,-19.687,4.998],[54.247,-15.880,7.334],
[53.491,-19.859,1.855],[52.520,-22.151,8.459],[56.052,-6.924,14.216],
[57.977,-6.983,9.580],[55.205,-3.703,10.652],[57.538,-4.335,9.260],
[54.199,-6.132,7.168],[56.792,-7.174,6.782],[55.079,-21.953,3.275],
[53.214,-22.348,0.485],[53.849,-5.630,1.436],[55.911,-19.309,-1.721],
[55.132,-20.231,12.569],[56.503,-21.506,14.637],[56.585,-6.734,-2.579],
[56.278,5.450,8.240],[55.100,9.695,6.373],[53.707,5.547,1.986],
[52.569,5.154,4.531],[57.862,0.441,5.644],[53.418,1.700,4.610],
[56.684,0.584,8.285],[55.598,2.686,6.525],[58.034,8.903,10.255],
[54.212,8.457,1.243],[-59.451,6.659,3.639],[-59.541,5.269,5.970],
[56.644,4.883,10.841],[58.246,11.884,13.002],[56.482,4.865,2.910],
[52.679,6.710,8.506],[52.243,10.045,7.183],[55.242,13.228,2.605],
[55.346,7.568,11.081],[56.668,-2.037,8.272],[58.041,2.705,7.743],
[58.324,3.313,4.455],[55.283,3.240,9.072],[-59.120,-1.206,-4.023],
[53.355,-1.590,4.172],[51.925,0.996,0.338],[54.059,2.591,1.398],
[57.307,3.896,0.471],[58.392,1.934,-2.175],[57.533,8.987,13.734],
[56.043,1.625,3.871],[56.800,6.333,5.671],[53.757,4.790,10.809],
[59.018,9.296,3.501],[52.927,4.297,-0.392],[58.857,10.670,1.097],
[54.400,-0.718,9.384],[54.151,9.171,8.986],[54.995,6.482,-0.364],
[56.845,0.459,-0.312],[54.558,8.809,-1.538],[53.629,1.837,-2.038],
[55.634,-3.588,6.119],[55.781,10.365,2.880],[52.683,14.124,13.093],
[-59.014,11.923,15.196],[54.635,6.071,13.079],[57.835,7.324,-3.539],
[57.518,6.373,13.008],[-58.878,6.110,-0.217],[57.314,3.024,12.593],
[57.517,-0.095,13.128],[59.199,-0.575,8.775],[57.185,2.625,15.120],
[54.797,-0.001,12.668],[51.955,8.303,-2.096],[52.469,15.221,2.028],
[52.378,12.877,3.487],[53.760,4.671,7.188],[51.337,1.910,13.997],
[53.340,-0.420,6.673],[52.212,2.722,8.629],[54.084,-0.557,-0.276],
[52.547,-2.955,9.352],[55.794,10.801,12.151],[59.005,12.250,9.595],
[56.038,13.854,9.728],[56.608,11.964,7.876],[54.911,13.782,7.043],
[57.604,11.116,5.062],[56.065,-1.512,4.431],[53.301,-3.412,-0.072],
[54.334,11.401,0.210],[55.491,-0.349,-2.981],[57.088,-3.173,12.910],
[56.328,-1.261,15.125],[56.362,13.807,-1.597],[55.008,24.280,8.730],
[56.959,27.890,6.047],[52.095,20.897,2.801],[54.145,24.561,4.366],
[56.931,19.320,4.257],[54.244,20.609,4.302],[55.459,16.988,8.261],
[55.783,19.964,6.697],[57.020,28.751,10.509],[55.279,28.992,-0.151],
[57.757,25.122,3.775],[58.036,23.390,7.223],[55.002,20.875,12.275],
[58.204,29.407,12.926],[54.839,23.753,1.030],[52.843,26.382,9.418],
[53.669,29.671,6.892],[55.056,31.842,3.942],[54.871,25.892,11.591],
[57.984,18.028,6.369],[57.631,21.167,9.833],[-59.799,22.130,3.723],
[54.972,21.117,9.480],[58.948,18.492,-3.736],[52.738,16.937,3.903],
[54.343,19.334,-0.789],[56.079,22.589,4.958],[58.709,19.991,1.152],
[-59.675,20.894,-2.307],[56.306,27.356,13.245],[57.224,18.059,1.914],
[56.141,25.226,6.079],[53.750,23.478,11.083],[59.503,28.179,4.505],
[54.150,23.563,-2.672],[58.707,29.890,1.394],[53.872,18.392,11.034],
[54.177,28.979,9.551],[55.677,26.364,0.016],[57.762,17.450,-0.766],
[56.078,27.956,-2.343],[54.269,20.877,-3.328],[56.829,15.518,6.384],
[54.775,28.860,2.657],[54.123,32.444,13.588],[57.997,32.064,13.484],
[53.816,25.427,14.358],[57.305,25.272,-3.161],[54.545,22.900,14.374],
[58.648,25.049,-0.324],[58.750,24.782,13.537],[58.978,18.419,11.439],
[59.263,19.308,8.769],[56.581,20.789,14.533],[55.021,18.061,13.451],
[52.995,29.138,-2.292],[52.754,33.559,3.229],[51.244,31.094,5.741],
[53.088,25.363,6.988],[52.035,19.413,14.224],[53.259,17.611,6.297],
[53.215,21.507,7.283],[53.422,17.754,1.119],[53.467,19.038,8.465],
[55.691,29.927,13.938],[58.230,31.046,9.952],[55.216,33.719,11.034],
[-58.981,32.981,8.779],[54.122,32.226,6.926],[57.303,30.484,4.017],
[55.722,15.774,2.276],[53.845,16.163,-0.964],[53.021,31.214,1.985],
[55.799,18.250,-3.197],[54.754,15.613,12.158],[56.648,15.102,14.141],
[56.392,32.128,-2.310],[56.492,43.401,9.425],[55.586,46.676,5.283],
[53.286,40.659,2.495],[53.266,42.876,4.089],[58.159,37.198,5.729],
[53.770,39.987,5.325],[56.485,37.323,8.642],[56.234,39.313,6.380],
[56.817,46.190,9.743],[53.629,45.282,-0.306],[57.947,43.457,4.615],
[57.535,42.165,7.065],[58.288,43.105,11.359],[58.224,48.505,13.101],
[58.334,42.779,1.893],[53.433,44.208,9.168],[52.783,48.318,7.751],
[55.656,50.222,2.451],[53.692,46.530,11.438],[57.813,34.009,7.766],
[57.150,40.302,9.053],[58.036,40.802,4.107],[56.116,39.941,11.701],
[58.470,37.553,-4.219],[53.041,36.916,4.735],[54.157,38.445,-0.206],
[55.274,38.309,2.272],[58.304,39.858,0.461],[58.253,39.597,-2.353],
[56.694,45.464,12.621],[57.815,37.524,2.821],[55.563,43.926,5.751],
[54.244,41.802,10.969],[57.931,46.870,3.551],[54.318,41.008,0.147],
[57.869,49.975,0.472],[54.177,37.042,11.783],[55.267,48.263,9.010],
[55.959,42.936,0.607],[58.445,36.978,-0.303],[56.488,46.413,-1.404],
[55.324,40.331,-2.168],[55.035,34.888,7.657],[55.044,47.487,2.834],
[53.157,49.221,13.542],[57.769,50.682,16.316],[53.347,44.228,14.084],
[56.981,43.526,-2.624],[55.754,42.714,13.424],[-59.153,42.127,-2.498],
[59.040,41.331,13.539],[58.253,38.192,10.569],[57.602,35.836,11.019],
[58.416,38.577,13.319],[54.180,37.280,14.679],[52.630,47.412,-1.919],
[53.213,52.000,2.604],[53.730,48.917,4.893],[54.622,42.070,7.449],
[52.134,39.004,14.429],[54.066,37.378,7.428],[51.864,39.653,7.218],
[52.362,35.582,0.633],[52.652,34.259,10.337],[55.420,48.333,14.558],
[57.150,50.542,11.539],[54.561,50.721,10.162],[56.486,53.326,10.513],
[54.204,50.803,7.203],[56.978,49.282,4.872],[56.791,34.381,1.905],
[54.587,33.552,0.773],[53.608,48.713,0.681],[56.290,35.349,-0.907],
[54.737,34.839,14.312],[59.599,34.540,14.240],[55.755,52.040,-1.730],
[55.603,55.294,4.532],[53.679,57.605,4.505],[53.456,55.875,9.431],
[56.444,56.083,8.734],[-58.980,52.469,7.151],[57.652,-59.214,7.855],
[58.771,58.596,4.397],[55.123,-58.977,10.063],[59.228,55.581,-3.180],
[52.656,55.262,4.634],[53.856,57.030,-0.664],[56.788,57.629,2.789],
[59.519,57.955,0.611],[58.714,-59.944,-1.792],[56.731,54.787,1.755],
[54.707,55.334,11.955],[57.826,55.962,-0.678],[56.196,57.938,-1.964],
[57.365,53.421,6.636],[57.790,57.133,12.917],[-59.695,55.314,9.089],
[-58.834,58.082,14.497],[53.993,57.044,14.294],[51.914,58.207,13.367],
[54.579,55.432,7.015],[52.225,57.812,6.925],[53.011,54.920,1.928],
[53.432,53.199,8.257],[56.936,52.050,4.405],[53.529,50.480,-1.274],
[54.697,54.428,-0.290],[54.613,52.012,13.251],[57.018,53.859,15.772],
[54.735,-51.862,27.012],[56.848,-45.931,25.580],[53.648,-52.278,20.473],
[54.540,-50.200,22.265],[58.528,-55.380,23.779],[54.053,-54.550,23.705],
[55.583,-56.236,26.815],[56.184,-53.582,25.474],[56.356,-46.326,28.906],
[54.523,-45.903,19.288],[57.214,-50.467,21.762],[58.146,-51.063,25.217],
[56.173,-50.688,29.230],[58.139,-44.009,31.948],[56.992,-52.784,20.079],
[51.555,-47.699,27.552],[53.902,-46.920,26.241],[55.282,-42.980,22.344],
[54.238,-47.406,30.668],[58.622,-58.732,24.793],[57.543,-52.707,28.156],
[58.872,-52.882,22.260],[54.342,-53.203,30.309],[59.406,-54.752,15.841],
[53.371,-57.031,22.806],[54.864,-54.529,20.192],[56.298,-53.547,22.754],
[58.923,-54.431,18.920],[56.774,-52.905,17.162],[56.892,-46.647,32.427],
[57.314,-57.150,21.795],[55.675,-49.898,24.686],[53.730,-49.340,28.223],
[58.299,-46.602,21.628],[55.559,-50.657,16.053],[57.869,-43.921,22.485],
[54.967,-55.883,29.729],[53.837,-44.637,28.829],[56.018,-50.235,19.123],
[57.143,-56.557,18.442],[56.096,-45.905,17.159],[54.508,-53.834,15.814],
[55.912,-57.638,24.428],[53.777,-46.055,21.876],[51.202,-41.948,33.749],
[-59.203,-42.190,33.462],[55.043,-48.820,33.015],[58.517,-47.737,15.411],
[55.031,-51.358,32.257],[58.551,-50.815,15.964],[58.997,-50.627,30.244],
[-59.377,-53.935,29.476],[58.790,-54.083,26.210],[56.754,-53.839,32.945],
[54.352,-55.276,32.438],[53.924,-46.647,15.492],[53.598,-41.329,20.449],
[53.089,-40.756,22.881],[53.025,-49.485,25.434],[51.669,-53.982,32.604],
[53.095,-56.823,25.511],[50.564,-50.013,24.201],[53.079,-57.076,20.062],
[52.705,-57.055,28.836],[55.611,-44.453,33.169],[58.734,-43.436,28.764],
[53.576,-41.738,29.262],[57.877,-41.506,26.960],[53.908,-43.271,26.551],
[56.302,-42.860,25.294],[55.728,-58.069,19.893],[54.287,-59.464,18.324],
[55.058,-43.272,19.000],[55.649,-56.433,16.168],[57.552,-57.112,31.018],
[57.343,-58.433,34.102],[55.036,-41.112,15.971],[55.942,-32.467,26.864],
[56.823,-27.830,25.825],[55.203,-32.233,18.834],[52.735,-31.601,22.889],
[55.806,-35.646,24.334],[54.385,-34.028,22.602],[56.437,-38.883,27.458],
[54.311,-33.569,25.497],[56.853,-29.488,28.442],[53.612,-27.895,18.184],
[57.649,-31.072,23.285],[58.667,-32.019,25.722],[56.744,-32.361,29.690],
[57.379,-24.054,31.423],[55.918,-31.231,21.151],[54.546,-30.923,29.087],
[53.689,-27.942,26.414],[54.082,-24.256,21.542],[53.887,-28.156,29.323],
[58.507,-39.151,25.295],[58.447,-34.432,28.575],[58.351,-34.031,21.882],
[55.572,-35.159,27.792],[58.824,-38.376,15.971],[51.429,-37.181,22.707],
[53.582,-35.417,20.386],[56.087,-36.343,21.514],[57.892,-36.136,19.322],
[58.761,-34.908,16.300],[58.433,-29.944,31.653],[56.615,-38.528,22.881],
[55.574,-30.189,25.203],[53.079,-32.739,30.285],[58.950,-26.741,21.772],
[54.450,-31.887,16.129],[58.566,-23.977,20.431],[53.994,-37.090,28.558],
[54.734,-25.816,27.690],[55.757,-29.203,19.615],[56.234,-38.474,19.540],
[56.970,-27.021,16.205],[56.289,-34.052,17.063],[55.002,-40.124,24.712],
[55.338,-27.140,21.520],[52.929,-24.918,32.943],[-58.841,-23.661,35.438],
[54.013,-30.352,32.270],[55.397,-29.261,16.259],[55.298,-32.626,32.195],
[57.360,-31.316,17.635],[56.663,-34.900,31.263],[57.151,-36.897,29.258],
[59.430,-37.024,27.636],[58.438,-32.668,33.135],[55.159,-37.799,30.951],
[52.039,-29.643,16.791],[53.107,-22.570,19.569],[51.879,-23.809,23.070],
[52.319,-31.033,25.552],[53.502,-35.611,32.141],[53.654,-37.351,23.958],
[51.527,-34.003,25.412],[53.120,-38.612,21.249],[53.212,-39.618,26.968],
[55.464,-25.854,33.320],[58.133,-26.129,28.382],[55.939,-23.356,28.262],
[57.205,-22.580,25.028],[53.839,-24.765,24.841],[56.795,-25.674,24.303],
[56.278,-41.106,20.268],[53.780,-39.397,18.106],[53.640,-25.147,18.564],
[52.715,-37.978,14.071],[54.783,-39.831,32.385],[57.128,-39.322,34.173],
[54.891,-23.580,15.416],[53.780,-13.888,27.370],[55.594,-8.533,25.549],
[54.688,-15.436,19.975],[52.961,-14.114,22.319],[56.568,-18.110,23.720],
[55.174,-15.566,23.049],[56.153,-18.711,26.416],[55.883,-15.980,25.584],
[57.628,-11.020,25.236],[54.456,-10.783,18.115],[58.665,-11.291,22.390],
[54.545,-13.229,24.431],[55.180,-13.797,29.939],[55.053,-6.795,30.452],
[56.591,-12.408,20.547],[53.843,-9.255,27.424],[52.791,-7.157,26.301],
[55.415,-5.138,21.442],[53.688,-9.643,31.085],[58.596,-19.638,26.646],
[58.429,-14.931,28.581],[58.242,-15.100,23.673],[55.347,-16.207,28.479],
[58.781,-19.897,15.160],[52.664,-17.353,22.284],[52.939,-17.752,16.707],
[57.144,-16.337,21.391],[58.187,-17.609,19.181],[57.549,-16.363,16.336],
[56.337,-11.219,32.263],[58.867,-20.165,24.007],[55.649,-11.424,22.772],
[52.928,-15.219,29.451],[57.885,-8.952,23.354],[55.142,-13.436,18.049],
[57.734,-5.149,19.249],[55.224,-18.822,29.508],[54.774,-6.275,27.713],
[57.330,-9.472,18.429],[57.986,-22.571,17.514],[55.981,-8.975,16.015],
[54.982,-15.959,17.140],[54.992,-20.813,24.952],[54.953,-8.322,20.789],
[58.401,-9.223,33.440],[57.341,-5.213,32.538],[53.438,-12.369,31.715],
[57.555,-11.409,15.380],[54.502,-12.799,34.116],[59.075,-11.962,19.578],
[57.157,-14.804,31.485],[57.154,-18.779,31.413],[58.167,-17.562,28.951],
[57.220,-16.863,33.259],[54.430,-17.890,32.270],[53.643,-8.969,14.632],
[52.696,-4.669,20.273],[53.034,-6.416,22.275],[51.947,-12.590,26.022],
[51.349,-15.442,31.937],[53.528,-17.131,24.848],[51.516,-15.204,26.251],
[52.501,-20.061,18.851],[54.301,-20.760,27.541],[55.737,-7.275,33.178],
[57.598,-6.194,29.888],[54.342,-4.114,29.362],[57.835,-3.695,27.588],
[54.064,-4.392,24.155],[56.798,-6.029,25.303],[53.332,-21.334,22.828],
[54.917,-22.239,17.575],[54.816,-5.600,18.570],[54.159,-19.964,16.472],
[55.552,-21.187,30.814],[56.968,-20.708,33.056],[55.658,-3.879,16.165],
[55.084,7.526,26.729],[58.312,9.402,25.485],[53.165,2.553,20.944],
[53.161,4.156,23.207],[57.258,0.456,22.225],[55.039,5.319,24.945],
[56.577,-0.782,28.015],[57.901,1.636,25.517],[56.539,8.768,28.503],
[54.328,8.918,18.951],[57.536,6.966,23.466],[-58.446,5.648,26.461],
[56.743,5.441,29.940],[58.522,10.601,31.605],[57.983,4.388,19.746],
[53.584,8.504,29.405],[52.674,11.361,26.107],[56.131,13.328,22.372],
[55.951,9.850,30.893],[57.876,-1.345,24.045],[56.830,4.684,26.965],
[57.777,4.366,22.346],[54.616,3.214,28.629],[58.395,-0.152,16.284],
[52.499,0.223,24.855],[54.413,1.184,18.667],[55.908,2.781,23.862],
[58.759,1.856,20.063],[58.337,2.562,17.425],[56.077,7.573,32.338],
[56.702,-2.085,20.923],[55.142,7.393,22.787],[54.133,5.476,30.299],
[58.230,9.693,21.535],[54.507,5.686,15.741],[58.869,13.165,20.719],
[53.571,-1.134,29.761],[54.493,11.869,28.073],[55.549,6.607,18.076],
[-59.761,-2.239,18.093],[57.077,9.540,16.346],[53.518,2.447,15.951],
[55.636,-2.509,25.269],[53.427,9.455,21.529],[55.226,14.917,32.162],
[-57.986,11.387,34.544],[53.239,7.488,31.949],[57.377,6.150,15.892],
[55.585,4.015,33.790],[-59.030,4.796,16.622],[58.027,2.399,29.538],
[-49.928,0.466,34.970],[-58.353,10.057,29.679],[57.558,2.037,32.841],
[54.249,0.449,32.507],[52.718,10.762,16.564],[52.422,15.558,21.739],
[53.967,11.962,22.769],[52.037,6.625,23.840],[51.236,1.276,32.447],
[55.167,0.209,24.691],[53.772,3.030,25.784],[52.389,-0.100,20.449],
[54.112,-1.424,27.116],[54.574,11.592,33.432],[57.833,11.233,28.732],
[56.158,14.040,29.701],[56.574,14.497,26.413],[54.580,13.428,24.763],
[57.393,12.128,24.967],[54.130,-1.818,22.159],[53.924,-2.782,17.899],
[54.364,11.621,19.592],[56.534,-1.197,18.125],[54.813,-3.408,32.321],
[53.551,-2.754,34.545],[55.712,13.221,15.574],[56.471,23.523,26.849],
[55.054,28.155,25.042],[53.340,22.625,20.930],[53.173,25.198,22.098],
[58.524,19.246,24.404],[54.023,20.817,23.941],[56.163,18.818,27.109],
[56.517,21.539,24.964],[58.246,28.544,27.105],[54.439,28.802,19.552],
[57.910,24.824,21.783],[58.829,23.578,25.534],[56.924,23.943,29.662],
[58.354,31.893,31.640],[58.112,22.993,19.871],[53.162,27.032,27.335],
[51.761,29.860,25.601],[54.764,33.741,20.302],[54.999,26.553,29.899],
[57.969,16.703,26.614],[59.334,21.253,27.089],[59.291,21.800,22.977],
[54.870,21.988,29.401],[58.327,17.594,14.670],[51.249,18.665,23.807],
[53.317,19.610,18.715],[56.438,20.456,21.797],[58.465,20.052,19.769],
[59.084,19.749,16.626],[57.404,26.339,31.208],[57.459,17.851,21.319],
[55.525,26.265,22.908],[52.866,24.706,30.027],[58.385,27.410,22.764],
[54.693,24.373,17.830],[58.885,31.195,20.804],[53.776,19.865,27.868],
[54.445,30.088,27.495],[56.668,26.689,18.515],[-58.741,16.752,17.346],
[56.132,29.010,17.236],[54.290,20.419,15.795],[54.999,16.236,24.832],
[56.222,29.888,22.223],[51.703,32.474,32.574],[58.509,29.757,33.986],
[52.377,26.292,32.175],[56.325,24.812,14.135],[54.756,24.549,32.037],
[58.385,24.892,16.610],[59.087,22.106,31.248],[57.261,21.011,29.361],
[58.745,18.917,28.123],[56.620,21.435,34.120],[54.708,19.392,33.000],
[53.844,27.259,16.770],[52.376,34.923,20.986],[52.792,32.860,22.703],
[53.421,25.669,24.969],[51.951,19.627,33.325],[55.165,18.449,23.190],
[52.002,21.719,25.905],[52.306,17.059,18.281],[52.171,19.008,30.590],
[54.762,27.867,32.239],[57.024,30.779,28.917],[54.717,34.805,27.948],
[56.372,32.814,26.985],[54.191,31.635,25.434],[55.657,32.459,23.212],
[55.088,15.714,22.007],[54.395,14.934,19.337],[53.546,31.562,18.996],
[55.841,18.078,15.997],[55.472,17.279,30.508],[58.593,17.662,31.078],
[55.924,31.701,16.526],[56.279,43.175,28.236],[55.774,46.733,24.976],
[53.232,41.078,20.373],[54.987,41.878,22.847],[56.540,37.677,23.156],
[52.634,40.298,23.255],[56.541,37.570,26.717],[55.368,39.498,24.812],
[56.690,46.857,29.400],[53.270,44.942,22.678],[57.149,43.742,23.127],
[56.648,41.683,25.580],[54.640,41.700,29.829],[57.185,50.186,30.553],
[55.492,42.635,20.374],[52.629,45.550,27.952],[53.926,48.758,26.405],
[55.651,50.482,21.813],[54.901,44.783,29.637],[58.960,36.176,26.582],
[58.321,39.651,27.049],[57.580,40.964,22.785],[55.250,40.098,27.796],
[57.257,36.502,15.158],[53.862,37.817,21.232],[53.468,38.623,17.101],
[55.833,39.453,20.185],[58.662,39.146,19.979],[57.412,41.348,15.610],
[56.710,42.965,31.604],[56.685,36.316,20.284],[55.295,44.594,26.274],
[52.470,43.552,29.883],[57.963,46.599,21.588],[53.230,42.475,16.530],
[58.498,49.666,21.409],[54.941,37.471,30.586],[55.709,49.232,28.617],
[54.564,44.113,18.483],[57.960,36.265,17.960],[58.070,46.138,16.999],
[54.357,40.470,14.417],[55.743,35.160,25.533],[54.928,47.060,21.912],
[53.179,51.494,33.135],[59.627,48.932,33.490],[55.191,45.602,32.164],
[57.753,44.005,14.655],[53.935,42.423,32.346],[56.837,42.921,17.925],
[58.771,41.693,30.488],[57.246,39.494,29.584],[59.241,37.756,29.702],
[56.082,40.556,32.243],[57.218,38.047,32.316],[53.164,46.087,15.870],
[52.130,52.491,22.382],[53.270,49.502,21.668],[52.835,43.791,25.236],
[52.435,39.586,33.206],[53.209,37.351,24.125],[50.862,40.796,25.552],
[52.511,36.445,18.926],[53.763,37.554,26.742],[56.593,48.371,33.086],
[59.026,49.283,28.463],[55.109,52.598,28.433],[57.650,51.320,27.587],
[55.370,50.943,26.169],[58.210,48.135,23.657],[55.196,35.230,22.519],
[54.451,33.804,17.627],[53.533,49.222,18.948],[55.140,36.567,17.328],
[53.240,33.963,30.045],[56.632,35.470,33.441],[55.110,50.661,16.645],
[56.940,54.040,22.816],[52.512,58.678,24.281],[55.600,56.295,27.444],
[55.461,57.730,24.456],[58.462,54.026,25.272],[56.515,58.699,26.578],
[59.375,58.921,22.070],[55.806,-58.417,29.228],[57.852,56.570,15.903],
[53.378,57.001,21.942],[53.982,56.768,19.231],[55.560,58.976,22.091],
[58.967,57.486,19.996],[58.507,-59.301,17.884],[57.064,55.017,20.303],
[54.893,57.545,31.026],[58.151,55.058,17.936],[55.670,57.752,16.952],
[55.620,54.448,25.534],[58.289,58.841,29.116],[58.522,56.358,27.805],
[57.035,59.240,31.753],[54.169,55.940,33.387],[52.595,58.665,31.580],
[53.541,55.939,24.622],[53.716,-59.106,27.176],[52.240,53.724,19.087],
[52.363,52.970,28.304],[54.453,54.579,21.478],[52.300,51.554,17.160],
[55.045,54.994,17.289],[54.312,54.090,31.260],[57.583,51.787,32.713],
[54.958,-49.788,46.063],[55.897,-47.598,44.209],[53.019,-52.113,39.301],
[52.723,-50.820,41.705],[57.600,-54.831,43.266],[53.927,-53.773,41.565],
[56.790,-56.317,46.133],[55.223,-53.171,43.875],[56.903,-46.749,46.600],
[53.897,-46.145,37.101],[58.124,-47.994,42.282],[58.045,-51.028,43.186],
[56.575,-50.672,48.083],[56.746,-43.480,49.164],[55.712,-52.448,38.827],
[52.127,-48.112,46.471],[53.241,-46.527,44.370],[54.208,-42.693,40.704],
[54.516,-47.512,48.596],[58.619,-57.414,43.742],[56.877,-52.708,45.856],
[57.958,-51.002,40.150],[54.709,-52.859,47.736],[58.574,-56.062,33.263],
[53.072,-55.492,39.707],[52.544,-55.204,36.037],[57.851,-54.163,40.566],
[57.306,-55.114,38.206],[58.692,-53.467,34.820],[55.519,-46.820,50.913],
[55.610,-56.418,39.662],[55.864,-49.802,41.805],[53.995,-50.363,48.600],
[59.703,-45.646,40.866],[52.735,-48.114,35.045],[58.600,-43.305,39.760],
[54.058,-55.610,47.234],[54.763,-45.078,47.148],[55.110,-48.499,36.404],
[57.645,-57.397,36.310],[57.108,-46.466,35.502],[53.700,-52.872,34.475],
[56.350,-57.450,42.179],[54.215,-45.989,40.138],[53.186,-41.560,50.436],
[-58.822,-44.124,54.559],[53.197,-48.723,51.126],[57.468,-48.844,34.344],
[53.813,-50.941,52.860],[58.247,-50.002,37.490],[59.119,-51.789,48.898],
[58.993,-55.865,48.796],[59.070,-57.860,47.099],[57.315,-53.669,50.009],
[55.827,-56.019,49.427],[52.832,-44.621,33.921],[53.175,-40.223,40.833],
[52.713,-43.213,43.117],[53.755,-49.232,43.735],[52.559,-54.635,50.612],
[53.498,-56.377,44.328],[52.192,-54.073,43.589],[52.837,-57.483,37.761],
[53.106,59.549,46.584],[56.378,-43.842,53.913],[58.425,-44.390,46.671],
[54.434,-40.978,47.923],[56.495,-42.210,46.612],[55.093,-44.302,44.122],
[57.303,-43.704,42.367],[54.776,-59.314,40.344],[53.939,-58.295,35.451],
[54.311,-43.340,37.566],[55.680,-56.126,34.594],[54.758,-58.752,49.762],
[57.105,58.716,52.533],[55.252,-41.972,34.257],[56.056,-31.363,45.510],
[55.557,-27.347,43.536],[53.784,-33.309,38.927],[54.211,-30.800,41.070],
[57.910,-36.448,43.639],[52.912,-33.233,41.752],[56.166,-36.782,45.643],
[55.545,-34.528,44.031],[57.408,-28.549,49.025],[53.982,-27.592,37.413],
[57.096,-30.973,40.458],[57.033,-32.501,42.918],[55.044,-31.493,48.466],
[57.881,-24.407,49.296],[56.900,-32.858,37.089],[52.672,-28.866,47.337],
[52.844,-26.771,45.110],[55.608,-24.193,39.660],[54.838,-28.504,48.794],
[57.650,-38.865,45.547],[58.053,-34.409,45.955],[58.552,-33.585,40.758],
[55.632,-35.785,48.196],[-59.131,-39.768,34.539],[53.410,-36.248,40.156],
[53.557,-36.060,37.365],[55.961,-34.378,40.917],[59.109,-34.845,38.448],
[58.975,-35.425,33.776],[54.175,-28.375,52.680],[58.089,-38.470,41.474],
[57.689,-29.647,43.820],[52.333,-31.525,48.700],[58.374,-27.971,39.538],
[54.591,-32.444,35.142],[59.541,-24.359,39.108],[53.025,-37.258,48.137],
[54.710,-26.194,47.197],[55.681,-29.473,36.212],[56.818,-36.997,38.337],
[55.941,-28.504,33.677],[55.875,-35.485,35.670],[54.180,-39.598,43.843],
[54.600,-26.986,40.221],[52.806,-21.741,50.530],[58.190,-22.833,52.496],
[52.595,-30.451,51.816],[58.145,-30.134,34.646],[54.936,-31.834,51.072],
[-58.419,-30.765,36.411],[57.872,-31.636,47.846],[58.393,-36.856,52.358],
[58.184,-35.781,48.347],[58.159,-33.845,52.220],[53.868,-38.305,50.779],
[53.107,-29.107,34.880],[52.493,-21.932,40.948],[53.054,-25.039,41.224],
[53.572,-29.871,44.558],[51.329,-36.642,50.261],[53.882,-36.677,43.678],
[52.894,-34.083,44.340],[52.501,-38.564,38.298],[50.295,-39.265,47.376],
[56.054,-24.874,52.215],[57.992,-26.580,46.999],[54.169,-21.010,48.134],
[56.073,-23.040,47.030],[54.541,-24.817,44.089],[56.683,-24.867,42.329],
[55.791,-40.250,41.630],[55.123,-40.182,38.686],[56.322,-25.253,35.940],
[55.076,-38.927,36.145],[53.458,-40.193,52.824],[56.253,-41.203,51.279],
[55.091,-23.337,34.474],[56.689,-12.738,45.221],[56.882,-8.507,44.046],
[53.717,-14.814,39.312],[52.482,-12.412,41.191],[56.535,-18.204,41.829],
[53.658,-17.262,41.706],[55.609,-19.629,46.076],[55.076,-16.288,43.821],
[56.667,-10.149,46.934],[53.666,-8.873,37.959],[57.920,-12.609,40.748],
[-59.808,-13.317,44.334],[57.172,-13.655,47.833],[57.164,-8.170,49.189],
[57.315,-14.734,38.189],[53.879,-11.596,46.600],[53.130,-9.590,44.871],
[55.893,-4.982,40.263],[54.173,-10.683,49.418],[57.956,-19.676,43.933],
[57.501,-17.054,44.942],[58.658,-16.757,40.440],[55.271,-16.520,46.354],
[-59.427,-21.544,33.758],[52.196,-19.291,41.345],[54.629,-18.289,35.092],
[56.119,-15.481,40.532],[59.360,-16.555,37.761],[57.632,-15.039,35.271],
[56.951,-10.199,51.332],[57.132,-20.310,40.275],[55.450,-13.169,42.062],
[54.363,-13.949,47.681],[58.316,-9.026,39.307],[54.934,-14.206,36.660],
[59.044,-5.190,39.120],[53.946,-17.808,48.624],[55.852,-7.113,46.986],
[55.253,-11.461,36.890],[59.051,-18.547,36.143],[57.289,-9.585,36.118],
[53.522,-15.868,34.965],[56.993,-22.768,44.371],[55.793,-7.700,40.070],
[53.340,-6.300,51.457],[56.526,-6.133,50.956],[52.590,-11.442,51.388],
[57.823,-12.076,34.386],[55.180,-13.583,50.436],[-58.815,-14.203,36.176],
[58.441,-15.332,50.177],[57.822,-17.146,47.801],[-59.099,-17.968,46.056],
[56.514,-15.653,52.120],[57.639,-17.986,52.539],[54.347,-8.993,34.868],
[52.949,-5.045,39.041],[53.417,-6.220,41.705],[53.712,-12.387,43.780],
[51.711,-16.706,52.000],[53.682,-20.768,44.004],[52.725,-15.254,44.482],
[50.692,-17.857,35.983],[53.030,-19.137,46.451],[54.546,-8.580,51.228],
[58.256,-7.876,46.573],[54.745,-4.634,49.698],[57.275,-4.134,45.688],
[56.699,-4.060,42.818],[57.425,-6.761,42.214],[55.301,-21.681,41.888],
[53.867,-19.891,37.408],[54.126,-6.606,36.307],[56.379,-20.557,35.548],
[54.826,-19.163,50.888],[57.445,-20.504,51.183],[56.485,-5.356,35.095],
[56.911,4.712,46.788],[55.571,9.486,43.306],[51.934,3.324,38.239],
[52.462,4.383,41.089],[56.851,-1.988,41.511],[55.863,0.955,42.130],
[55.278,0.976,46.942],[-57.581,4.849,43.681],[57.180,9.421,47.745],
[54.305,9.658,38.273],[57.109,6.718,42.149],[51.406,1.867,42.397],
[56.488,6.365,49.198],[59.265,10.731,49.026],[56.175,5.249,39.582],
[52.399,7.939,46.602],[53.197,10.533,44.566],[55.663,13.246,42.102],
[54.092,9.346,48.127],[59.237,-4.582,43.935],[57.616,2.192,45.819],
[58.317,4.059,40.844],[54.399,4.202,47.846],[57.322,-0.010,34.675],
[53.510,0.162,41.240],[53.063,0.273,34.975],[54.339,2.623,40.023],
[-58.701,2.125,38.533],[58.568,3.374,36.340],[55.469,8.399,50.219],
[56.684,-1.997,36.179],[54.808,5.449,42.038],[53.586,6.497,48.885],
[58.529,8.002,40.045],[53.548,6.628,35.641],[59.612,13.062,38.327],
[53.619,0.471,49.523],[55.440,11.652,47.146],[55.627,7.286,37.646],
[-58.999,-0.017,36.635],[56.736,9.287,36.027],[56.077,6.568,34.821],
[55.078,-1.744,43.996],[54.421,11.019,40.508],[52.901,13.371,51.310],
[57.029,12.802,53.859],[53.024,7.000,51.781],[58.115,8.683,33.766],
[55.120,5.022,51.241],[58.669,7.123,35.908],[57.906,3.521,48.946],
[55.824,2.114,49.782],[59.642,-0.190,47.365],[58.101,1.565,52.000],
[56.019,-0.021,51.296],[54.083,10.026,35.546],[52.516,14.574,38.800],
[52.589,12.041,42.319],[52.644,6.036,43.987],[51.621,2.065,50.519],
[53.682,0.729,44.271],[51.798,2.997,44.886],[52.190,-2.185,38.882],
[53.140,-2.164,48.192],[55.489,11.692,51.732],[58.573,12.039,46.098],
[54.834,14.804,47.361],[57.917,14.779,45.711],[54.210,13.610,44.489],
[56.473,12.081,44.493],[55.837,-2.372,38.983],[54.647,-4.063,36.780],
[53.278,12.139,37.802],[56.922,-2.416,33.429],[54.234,-2.042,50.607],
[56.808,-2.831,50.939],[56.217,13.878,34.612],[56.205,24.116,44.747],
[54.714,28.305,43.432],[53.056,22.111,38.741],[52.856,24.306,41.593],
[57.169,18.853,42.705],[52.621,20.439,42.964],[55.555,18.914,46.458],
[55.471,20.825,42.722],[57.727,27.741,47.075],[56.564,27.313,38.474],
[58.010,23.156,40.629],[57.745,25.681,42.858],[56.483,24.924,48.134],
[-59.427,28.743,49.882],[56.931,24.138,37.932],[51.572,26.757,46.156],
[52.356,30.125,44.103],[56.298,31.778,38.872],[55.346,28.461,48.109],
[58.735,16.520,43.394],[56.890,22.079,46.391],[57.751,20.012,40.362],
[54.762,21.039,47.689],[58.732,18.629,33.873],[52.807,18.466,40.919],
[54.735,18.486,35.597],[55.736,21.976,39.052],[57.845,20.165,36.906],
[59.357,22.067,34.718],[54.892,27.262,51.961],[56.871,17.262,38.094],
[54.999,25.523,42.577],[53.097,23.073,48.151],[58.155,29.381,39.942],
[54.288,23.696,34.590],[59.632,31.701,38.373],[53.188,19.044,45.415],
[55.490,30.511,46.339],[55.899,25.588,35.967],[58.912,17.160,36.225],
[55.911,28.252,35.760],[55.200,21.383,36.271],[55.481,16.441,45.274],
[54.211,28.546,39.440],[51.442,32.694,50.922],[58.682,31.918,52.056],
[53.103,25.717,50.662],[58.073,26.688,34.358],[55.626,22.800,51.626],
[-59.131,23.863,36.946],[56.658,20.943,49.804],[58.068,18.149,48.000],
[58.158,19.303,45.494],[56.840,19.193,52.184],[54.364,18.275,51.607],
[53.465,28.602,34.810],[52.834,33.573,39.312],[53.395,31.663,41.211],
[53.522,24.756,44.659],[51.573,20.403,50.381],[53.452,17.125,43.411],
[51.512,22.925,43.524],[52.251,16.352,36.471],[53.711,17.273,47.693],
[56.781,30.248,50.775],[57.894,29.871,45.397],[54.873,34.150,46.756],
[57.577,32.441,44.131],[54.847,31.587,43.804],[56.826,31.915,41.410],
[54.917,14.946,39.761],[54.768,16.034,37.057],[54.109,30.920,37.701],
[56.021,17.132,33.661],[53.810,15.616,50.195],[57.461,14.343,51.695],
[56.417,32.845,33.125],[56.258,42.042,45.473],[56.679,46.271,42.488],
[53.960,40.768,38.589],[52.910,42.061,41.739],[57.258,38.358,41.725],
[53.632,38.914,41.282],[56.506,37.139,44.466],[55.967,40.703,41.995],
[56.614,45.904,46.407],[56.325,47.359,37.891],[57.975,43.889,41.993],
[-59.713,42.382,44.987],[58.080,42.439,48.355],[58.871,48.634,49.037],
[56.272,42.423,38.910],[52.487,45.365,47.255],[52.537,49.460,44.650],
[56.581,50.454,40.130],[54.899,46.030,48.564],[58.229,34.999,44.891],
[57.797,39.631,44.097],[58.371,41.576,40.319],[56.072,39.469,46.407],
[-58.281,34.883,32.444],[54.375,36.926,38.715],[53.582,38.526,36.494],
[56.090,39.487,39.464],[-59.770,38.235,36.718],[57.628,40.626,34.575],
[56.637,44.405,49.672],[57.446,37.182,39.204],[55.117,43.292,42.813],
[54.705,42.607,47.854],[58.019,48.192,40.753],[53.227,42.127,34.954],
[59.113,48.381,37.795],[55.898,36.822,48.327],[54.528,48.383,46.626],
[55.765,45.122,36.106],[57.268,37.230,35.296],[58.071,46.781,34.773],
[54.985,38.911,34.330],[55.233,34.283,44.091],[55.006,47.400,40.521],
[53.347,51.336,51.811],[-58.991,52.614,53.129],[53.917,44.815,51.053],
[57.344,44.097,34.057],[55.248,42.293,50.473],[58.380,42.236,36.745],
[-58.939,40.571,48.425],[57.738,38.868,49.158],[58.646,37.572,46.249],
[57.111,40.832,51.490],[55.364,37.393,51.182],[53.343,45.829,34.604],
[51.476,52.143,40.458],[53.575,50.098,40.760],[52.558,44.233,43.318],
[52.428,37.858,50.794],[53.695,35.543,41.527],[52.440,39.647,45.173],
[52.071,36.203,37.117],[52.598,35.241,45.566],[55.271,47.170,51.007],
[57.077,49.106,46.925],[55.958,51.644,47.243],[58.400,51.749,45.668],
[55.533,50.804,43.092],[57.610,49.664,44.232],[57.810,34.360,40.277],
[54.841,33.781,37.002],[54.322,49.063,37.409],[54.371,36.034,35.215],
[53.952,34.085,50.607],[56.887,33.541,50.629],[56.015,51.360,35.061],
[57.913,54.155,41.388],[54.351,57.738,42.576],[54.336,55.358,45.399],
[56.524,56.858,45.083],[57.675,54.277,44.474],[58.433,58.506,43.738],
[59.078,-59.063,40.901],[56.482,-59.765,47.307],[58.872,57.592,34.151],
[53.625,55.081,40.675],[53.540,57.307,36.206],[57.314,56.804,41.016],
[58.440,58.523,38.558],[56.280,58.782,36.506],[56.704,53.840,37.077],
[55.569,55.950,48.025],[57.987,56.283,36.685],[54.523,58.658,33.716],
[55.227,52.860,44.732],[57.821,55.434,49.514],[59.046,56.261,45.962],
[56.555,57.929,49.840],[54.721,54.689,51.758],[53.988,56.571,50.008],
[55.573,55.284,42.947],[52.412,57.236,44.334],[53.357,55.023,38.026],
[54.132,53.601,47.463],[54.597,52.566,41.104],[54.819,51.822,37.787],
[54.373,53.637,35.782],[55.619,52.048,50.084],[57.504,53.855,52.039],
[53.455,-52.486,57.285],[54.716,-45.764,56.411],[56.169,-51.415,59.534],
[57.513,-43.570,-59.259],[56.654,-53.324,57.476],[58.209,-56.029,52.451],
[54.552,-54.258,55.181],[54.387,-55.320,58.267],[58.438,-55.788,55.627],
[58.671,-53.506,53.428],[58.241,-56.866,58.438],[54.826,-51.343,55.249],
[-58.518,-43.171,-59.385],[57.790,-48.487,55.362],[58.022,-58.287,54.675],
[56.173,-47.343,53.392],[56.116,-52.989,52.442],[55.032,-43.794,58.557],
[57.978,-49.141,52.275],[59.280,-50.844,54.068],[53.162,-48.040,53.645],
[52.971,-40.796,-59.375],[52.789,-57.036,55.418],[54.191,-59.097,53.700],
[53.961,-45.369,53.471],[54.085,-55.840,53.220],[52.994,-42.864,54.550],
[54.330,-34.194,57.969],[53.941,-28.990,56.624],[56.695,-32.877,57.223],
[56.017,-24.554,59.009],[57.044,-34.678,-59.402],[58.557,-39.529,54.442],
[53.943,-36.435,56.702],[55.876,-37.571,58.136],[58.962,-34.078,57.228],
[57.686,-34.647,54.877],[57.375,-39.175,59.403],[53.483,-32.381,53.710],
[58.898,-26.479,-58.376],[56.198,-31.101,54.958],[-59.444,-36.727,55.118],
[56.445,-27.814,54.181],[54.768,-34.532,52.266],[54.858,-26.898,58.235],
[57.818,-29.892,52.437],[58.941,-32.258,54.318],[51.991,-28.235,54.489],
[53.493,-22.977,-58.916],[52.628,-39.001,57.118],[56.387,-40.991,55.834],
[53.392,-23.800,56.511],[55.294,-38.760,54.932],[55.481,-22.854,53.859],
[54.003,-15.288,58.051],[54.295,-10.004,56.274],[56.673,-13.751,57.280],
[55.080,-5.460,58.209],[57.236,-15.442,-59.289],[58.416,-20.362,53.799],
[54.814,-18.220,56.085],[55.331,-18.045,59.275],[58.928,-15.822,57.088],
[59.238,-16.118,53.718],[57.551,-19.653,-59.388],[54.987,-13.658,53.474],
[58.728,-6.346,56.929],[56.958,-11.148,55.961],[57.779,-18.191,55.387],
[57.035,-9.306,53.978],[53.926,-16.576,54.003],[54.489,-8.238,58.248],
[57.252,-12.646,52.497],[58.976,-13.545,54.471],[54.053,-9.947,53.518],
[51.320,-3.477,58.930],[53.209,-20.090,57.968],[55.364,-21.361,56.001],
[52.603,-4.763,56.771],[55.912,-19.992,53.368],[54.845,-4.568,52.590],
[54.250,3.661,57.828],[54.560,9.448,56.027],[57.594,5.405,56.848],
[56.181,11.924,58.806],[57.537,3.119,58.683],[-58.837,-0.208,52.160],
[53.600,0.272,56.215],[54.921,1.064,-59.487],[-58.186,2.645,58.961],
[57.600,2.966,55.525],[57.204,-1.004,58.794],[53.230,4.423,53.239],
[59.071,11.887,-59.553],[55.263,7.099,53.638],[57.044,0.412,54.559],
[56.880,9.508,52.337],[55.431,2.631,52.617],[53.442,10.375,58.577],
[57.597,6.776,51.666],[58.039,6.254,54.346],[52.328,8.639,53.736],
[52.828,15.373,58.146],[53.717,-1.118,58.421],[53.252,-2.187,55.280],
[54.421,12.421,56.210],[55.883,-2.182,53.515],[54.352,13.415,53.771],
[53.475,22.242,58.142],[54.776,27.138,56.706],[57.719,23.655,56.848],
[57.127,31.505,57.067],[58.272,21.853,-59.745],[-59.099,18.356,52.195],
[55.182,19.480,54.639],[56.912,19.244,57.938],[58.767,21.312,56.327],
[58.391,20.916,53.488],[56.995,16.551,58.936],[53.010,24.336,54.058],
[-58.811,31.266,57.246],[56.209,26.773,54.308],[57.509,18.074,54.849],
[55.027,29.062,54.081],[53.715,21.327,53.382],[53.547,29.620,57.612],
[57.636,27.030,52.058],[-59.158,25.230,55.305],[52.423,27.773,52.894],
[51.057,32.703,58.214],[54.284,17.714,57.728],[55.157,14.916,57.060],
[53.290,31.485,55.137],[55.074,16.775,54.829],[56.312,31.377,53.522],
[54.189,41.497,58.150],[55.722,46.016,56.364],[57.513,42.881,56.981],
[57.205,49.810,56.597],[56.713,40.317,57.674],[58.250,35.908,51.653],
[53.253,36.813,57.165],[54.659,38.314,58.468],[59.033,39.147,57.631],
[59.395,39.384,53.677],[56.813,37.580,-58.471],[52.874,42.327,52.562],
[-59.169,49.526,58.663],[57.265,44.138,54.472],[58.401,37.231,55.180],
[55.152,47.015,53.724],[53.825,39.984,52.161],[55.353,46.807,-59.964],
[58.542,45.149,52.352],[-59.070,42.940,54.860],[52.868,45.406,53.341],
[53.963,51.289,59.554],[53.017,33.746,56.460],[55.618,34.861,56.691],
[52.074,48.126,55.603],[56.038,36.682,53.880],[57.257,51.262,52.999],
[57.465,58.799,59.098],[-57.203,56.732,52.608],[54.906,56.689,54.656],
[54.468,55.201,58.301],[59.155,58.711,56.586],[-58.883,58.471,53.960],
[57.030,53.640,57.281],[57.313,55.900,55.516],[54.576,57.624,52.140],
[52.296,54.409,56.713],[53.641,52.438,54.643],[55.765,53.908,54.272],
[59.272,-58.048,-14.307],[56.332,-59.679,-18.501],[55.817,-48.760,-27.788],
[57.597,-58.303,-25.747],[57.993,-49.213,-24.952],[58.481,-48.082,-29.635],
[57.187,-48.256,-33.434],[54.321,-52.896,-36.458],[56.518,-55.314,-35.851],
[59.084,-50.912,-22.225],[54.981,-53.655,-19.141],[51.248,-46.368,-34.360],
[52.236,-44.587,-31.375],[56.331,-41.488,-29.756],[54.989,-28.977,-33.901],
[58.199,-40.082,-37.326],[57.138,-42.559,-42.168],[52.005,-34.193,-31.654],
[53.658,-27.907,-31.843],[55.998,-31.424,-40.800],[57.541,-21.240,-44.947],
[58.076,-26.035,-48.808],[55.706,-18.604,-54.087],[57.510,-20.803,-54.250],
[57.480,-25.197,-39.261],[57.556,-10.978,-53.506],[59.109,-6.686,-55.299],
[58.445,-1.957,-55.130],[55.942,4.613,-51.366],[58.242,-2.476,-58.186],
[56.074,-13.057,-59.023],[54.147,-9.874,-50.112],[53.122,-9.565,-52.723],
[56.932,3.741,-55.046],[56.264,4.400,-58.107],[54.621,-12.070,-51.856],
[54.766,8.539,-59.231],[58.920,11.072,-56.960],[57.580,13.367,-55.243],
[56.100,7.360,-52.014],[56.121,-54.379,1.061],[56.884,-57.557,1.644],
[-59.301,-58.352,-4.959],[51.790,-58.459,5.869],[56.242,-49.299,-3.592],
[52.089,-48.580,-13.658],[58.745,-43.868,-11.815],[56.058,-37.278,-17.511],
[54.392,-46.991,-12.241],[57.748,-39.146,-17.186],[58.227,-41.301,-18.688],
[-59.106,-40.217,-20.734],[55.529,-51.749,-13.069],[53.625,-49.635,-17.543],
[57.423,-46.673,-16.623],[58.935,-40.982,-6.812],[59.460,-45.405,-3.995],
[59.437,-39.157,-24.425],[53.902,-35.254,-6.609],[56.960,-33.215,-14.351],
[54.454,-30.032,-16.193],[59.141,-24.514,-18.807],[56.114,-22.321,-19.631],
[53.298,-28.135,-27.518],[56.918,-27.598,-23.756],[58.165,-21.248,-17.852],
[56.527,-28.854,-17.354],[58.529,-32.370,-24.326],[56.020,-25.644,-20.276],
[53.228,-29.558,-12.845],[55.787,-22.794,-32.387],[54.188,-37.641,-23.291],
[56.324,-29.770,-8.095],[51.882,-13.503,-31.050],[54.701,-12.909,-34.271],
[56.451,-5.230,-37.671],[57.349,-10.020,-37.287],[58.722,-8.360,-38.964],
[57.033,-13.212,-38.978],[58.951,-13.719,-22.928],[53.179,-21.397,-31.870],
[52.314,4.451,-41.251],[59.190,10.255,-37.693],[56.351,0.807,-37.323],
[57.728,5.334,-43.273],[54.534,8.378,-34.226],[57.149,12.507,-37.172],
[57.673,-2.741,-44.545],[56.620,8.690,-43.150],[-58.222,5.578,-43.437],
[58.673,7.696,-45.629],[54.391,8.753,-31.326],[-59.300,9.760,-51.600],
[54.242,16.539,-39.400],[56.821,-2.037,-36.818],[54.329,17.481,-47.128],
[56.704,14.589,-43.409],[55.662,21.103,-43.716],[-59.455,14.812,-50.383],
[58.149,15.806,-54.557],[57.352,26.501,-54.509],[54.119,24.379,-55.843],
[56.358,32.170,-43.620],[52.608,27.331,-52.624],[56.987,26.906,-51.784],
[56.520,16.745,-37.701],[58.419,36.015,-56.320],[57.236,44.312,-54.865],
[54.129,44.027,-55.006],[57.952,35.420,-50.655],[-59.110,33.450,-50.156],
[57.772,29.423,-55.477],[53.983,32.569,-47.085],[59.208,-50.974,12.451],
[58.711,-43.500,18.910],[52.554,-44.733,17.092],[59.005,-49.380,9.725],
[58.393,-47.997,5.480],[57.988,-58.637,12.156],[55.857,-51.862,5.301],
[53.351,-42.211,11.500],[-56.314,-56.322,16.386],[56.822,-29.208,2.862],
[54.754,-34.029,0.427],[52.798,-29.784,2.286],[57.804,-44.733,0.026],
[57.281,-39.609,0.238],[-59.332,-33.319,0.117],[52.722,-26.962,3.347],
[57.880,-39.797,12.580],[-58.407,-30.550,-6.769],[58.418,-42.336,7.333],
[-56.564,-20.977,-5.420],[55.862,-20.484,1.084],[56.592,-12.111,-3.005],
[55.946,-15.702,-0.195],[57.641,-22.373,0.690],[58.243,-22.743,-7.716],
[53.724,-21.266,-10.686],[58.249,-14.592,-13.479],[56.347,-4.553,0.213],
[58.324,-3.517,-15.403],[56.807,-4.131,-18.350],[58.485,-9.699,-18.850],
[54.450,0.006,-14.472],[56.777,-10.547,-16.372],[58.014,-4.392,-21.024],
[57.178,-3.532,-6.996],[50.915,-8.318,-15.532],[55.322,7.571,-16.720],
[54.019,5.214,-18.032],[53.461,-9.813,-10.388],[59.171,-9.916,-13.564],
[55.523,11.420,-20.302],[56.346,18.870,-21.056],[56.135,13.765,-26.170],
[58.186,16.050,-16.920],[-58.570,19.935,-32.094],[52.345,8.922,-15.493],
[51.599,27.025,-34.484],[58.423,33.097,-34.569],[57.637,27.930,-36.944],
[51.819,35.321,-29.409],[55.685,27.165,-35.000],[52.503,26.036,-31.876],
[55.131,25.121,-32.502],[56.227,23.069,-31.206],[58.028,33.241,-38.099],
[57.523,27.969,-42.295],[56.293,31.564,-25.414],[57.427,25.522,-25.822],
[-57.330,33.985,-40.423],[53.871,39.392,-29.093],[55.938,21.085,-36.829],
[58.136,41.650,-40.561],[57.783,52.591,-37.528],[52.719,51.090,-44.119],
[57.009,40.682,-37.921],[54.186,45.557,-47.696],[57.860,48.129,-36.641],
[55.003,45.630,-31.223],[54.177,47.576,-50.250],[52.350,53.818,-48.215],
[51.835,47.091,-32.448],[58.542,53.828,-44.423],[56.943,-41.300,29.818],
[51.571,-40.657,25.067],[57.754,-41.127,22.909],[56.824,-48.405,26.800],
[56.083,-47.725,22.610],[57.758,-40.378,36.662],[57.479,-43.085,35.365],
[56.306,-45.276,38.039],[51.211,-34.056,22.047],[57.074,-48.923,31.051],
[55.770,-20.359,19.569],[58.501,-22.002,22.096],[56.766,-25.301,18.829],
[58.216,-29.081,20.622],[59.154,-25.260,17.463],[57.493,-22.089,10.772],
[54.846,-25.694,11.903],[55.762,-26.176,30.432],[53.367,-28.778,23.684],
[51.773,-26.629,24.048],[57.277,-14.892,18.811],[54.455,-22.922,12.255],
[59.305,-6.448,7.126],[58.842,-3.368,5.554],[54.922,-6.860,4.331],
[59.440,-7.732,19.478],[-58.070,-4.000,0.233],[51.554,-7.649,6.027],
[52.891,-3.216,6.670],[52.184,-9.017,19.086],[53.404,-0.437,15.075],
[57.271,-1.388,10.654],[53.642,-11.566,21.074],[52.980,7.739,5.588],
[56.602,10.696,-1.582],[58.863,14.233,5.111],[50.878,1.813,3.641],
[58.544,9.175,-5.237],[57.489,4.200,-3.067],[51.726,8.400,3.016],
[49.573,5.927,5.449],[59.154,13.338,-0.018],[52.086,29.685,-9.437],
[55.520,29.227,-12.113],[54.401,32.241,-7.322],[57.753,27.649,-16.803],
[55.919,25.064,-14.076],[53.215,27.253,-14.614],[-59.073,29.326,-17.990],
[50.633,29.816,-12.877],[51.161,28.836,-15.763],[57.954,46.963,-12.806],
[52.883,46.008,-13.337],[58.390,32.931,-17.977],[55.061,39.659,-16.283],
[57.539,30.358,-18.892],[55.972,38.786,-23.693],[54.890,39.966,-6.624],
[-59.361,42.656,-27.202],[53.295,34.251,-11.201],[56.040,49.754,-18.923],
[54.908,29.998,-21.218],[56.628,52.512,-19.023],[55.287,56.217,-15.533],
[54.746,58.538,-36.342],[53.279,-32.179,46.322],[55.846,-22.341,49.903],
[58.432,-41.352,41.990],[53.902,-35.199,46.359],[58.917,-26.501,37.178],
[52.793,-28.954,39.413],[58.325,-26.221,33.746],[53.486,-22.843,46.180],
[58.184,-40.847,49.399],[56.011,-10.639,40.683],[52.816,-11.274,35.708],
[54.430,-8.984,41.979],[55.703,-22.473,37.178],[59.605,-14.418,33.501],
[57.176,-12.138,29.218],[57.552,-15.421,42.805],[-58.632,-10.213,25.132],
[52.782,-17.579,39.095],[50.743,-2.497,26.016],[51.098,1.931,22.885],
[-59.730,6.815,30.932],[55.644,5.760,20.525],[59.648,-2.362,33.808],
[58.275,-4.056,36.835],[55.020,10.553,24.871],[53.479,-8.887,23.562],
[56.718,13.328,18.703],[58.302,15.625,19.003],[-58.664,8.741,12.463],
[57.924,14.388,11.817],[58.160,14.522,23.915],[59.103,14.149,27.877],
[50.302,8.656,17.433],[58.711,22.560,15.266],[58.495,7.001,18.361],
[56.412,30.784,7.717],[57.244,27.779,2.889],[56.072,26.719,8.937],
[57.351,34.409,4.970],[54.013,29.794,16.022],[58.678,33.891,-1.318],
[53.784,26.411,2.367],[50.405,40.097,16.196],[56.172,31.252,1.077],
[57.627,33.750,16.165],[58.624,27.469,16.036],[59.284,40.054,6.270],
[59.481,25.816,9.402],[-59.127,48.403,-0.049],[57.574,52.560,0.630],
[56.427,45.149,2.047],[58.692,50.935,-2.092],[58.336,53.727,-5.984],
[53.347,52.831,5.545],[57.173,46.695,-6.209],[58.073,49.896,7.353],
[53.322,45.153,2.474],[51.822,47.690,2.395],[52.528,-29.746,-59.237],
[56.519,-29.840,58.095],[56.651,-22.089,58.531],[57.824,-23.011,55.874],
[58.234,-25.462,56.987],[54.017,-31.617,57.754],[51.545,-8.604,52.335],
[55.471,-10.491,59.569],[58.340,-1.821,56.403],[57.514,-6.618,54.367],
[55.540,-3.123,59.529],[56.104,1.198,56.886],[56.895,-8.317,56.487],
[58.928,-11.756,48.888],[57.555,-2.402,48.424],[-59.376,-6.771,48.792],
[53.740,-12.493,57.344],[-59.258,-1.756,41.399],[52.403,-2.664,45.128],
[55.749,-15.812,49.021],[57.952,7.106,45.129],[55.030,7.850,45.457],
[58.615,13.800,43.221],[55.151,7.866,40.381],[57.721,14.791,37.597],
[-58.841,6.592,49.871],[59.510,3.554,52.781],[59.264,12.827,30.440],
[52.774,2.071,47.125],[58.236,30.018,36.731],[54.090,23.831,37.209],
[55.429,31.235,35.285],[58.154,32.584,35.324],[56.776,35.329,37.574],
[51.257,29.945,34.020],[52.158,26.999,39.558],[53.780,24.081,27.468],
[-58.805,27.777,30.494],[57.279,25.798,25.447],[57.094,30.528,25.016],
[58.857,22.560,43.351],[51.021,21.779,40.770],[55.042,31.114,31.091],
[56.467,33.497,29.795],[51.936,47.901,23.376],[55.877,47.847,19.395],
[57.193,35.926,29.366],[57.632,33.400,18.821],[57.330,45.082,19.569],
[55.199,40.805,17.440],[58.532,54.115,30.006],[-59.559,42.204,33.233],
[59.362,45.834,12.501],[51.468,47.591,20.502],[57.316,51.512,23.839],
[56.504,39.499,36.723],[57.572,52.357,19.336],[54.045,51.815,20.022],
[58.968,49.888,18.847],[-59.617,8.042,56.176],[53.172,7.621,57.593],
[56.985,26.052,58.212],[57.857,15.722,56.508],[58.149,24.089,52.544],
[57.781,10.437,56.464],[56.948,29.145,58.906],[58.466,28.779,56.124],
[54.939,39.594,48.844],[52.578,33.293,48.374],[57.602,35.015,47.709],
[57.956,33.931,53.515],[55.002,42.602,55.480],[51.459,42.000,44.454],
[51.743,47.966,39.891],[55.306,47.770,44.115],[58.903,51.588,42.174],
[51.608,47.164,42.447],[57.202,44.903,38.678],[58.252,51.106,37.573],
[56.988,56.384,32.362],[54.063,44.085,45.694],[58.335,52.110,49.232],
[-57.971,46.747,48.456],[55.422,56.829,38.372],[55.543,42.167,36.136],
[57.445,48.459,52.084]]

